package ZipTie::Adapters::CheckPoint::CPSMS;

use strict;
use warnings;
use v5.10;
use Tie::File;
use File::Path qw(make_path);

use Cwd;

use ZipTie::Adapters::CheckPoint::CPSMS::CPTool;
use ZipTie::Adapters::CheckPoint::CPSMS::FileUtilities qw(getFileToString getFilePath);
use ZipTie::Adapters::CheckPoint::CPSMS::Parsers qw(parse_system parse_chassis parse_CPSMS_interfaces parse_CPSMS_object_groups parse_filters parse_static_routes parse_snmp);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);

use ZipTie::ConnectionPath;
use ZipTie::Credentials;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

sub backup
{
	my $packageName = shift;
	my $backupDoc   = shift;    # how to backup this device
	my $responses   = {};       # will contain device responses to be handed to the Parsers module

	# translate the backup operation XML document into ZipTie::ConnectionPath
	my ($connectionPath) = ZipTie::Typer::translate_document( $backupDoc, 'connectionPath' );

	# at this point we have the CPSMS IP address and the discovered end point's IP address
	# get CPSMS IP address 
	my $managementServerIp = $connectionPath->get_management_server_ip_address();
	# get discovered end point IP address
	my $ip_addr = $connectionPath->get_ip_address();

	# set up the XmlPrint object for printing the ZiptieElementDocument (ZED)
	my $filehandle = get_model_filehandle( 'Check Point CPSMS', $connectionPath->get_ip_address() );
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

	# get package's path 
	( my $pkgDir = __PACKAGE__ ) =~ s!::!/!g;
	my ($pkgPath) = $INC{"$pkgDir.pm"} =~ m!^(.+)/!;
	# set CPT directory relative to CPSMS
	my $cpt_cpsms_dir = "../../../../cptool/cpsms";
	# set CPT directory relative to CPTool
	my $cpt_dir = "../".$cpt_cpsms_dir;
	# set CPT XML output directory 
	my $out_dir = "/usr/share/ziptie-server/osgi-config/scratch"; 
	my $xml_dir = $out_dir. "/logs";
	# set XSL directory relative to CPSMS
	my $xsl_dir = "../../../../../xsl";
	# create a new instance for CheckPoint Tool (CPT)
	my $cpt = ZipTie::Adapters::CheckPoint::CPSMS::CPTool->new(connectionPath => $connectionPath, out_dir => $out_dir, 
																managementServerIp => $managementServerIp,
																ip_addr => $ip_addr);
	
	my $recording_dir = $ENV{RECORDING_DIR};
	my $cpsms_dir = $out_dir."/logs/".$managementServerIp."/".$ip_addr;
	
	# Check to see if this is a simulation
	my $isSimulation = 0;
	my $simulationFilePath = "";
	my $lineInFile = "";
	my $line = "";
	my $xmlTag = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
	my $tag = "false";
	my $tagValue;	
	
	open FILE, "<$pkgPath/$cpt_cpsms_dir/cpmi.conf" or $LOGGER->fatal_error_code("File is not available: $pkgPath/$cpt_cpsms_dir/cpmi.conf");
	foreach $lineInFile (<FILE>)
	{       
		if ($lineInFile =~ m/^adapters.cpsms.simulation.filesource=(\/.*)/)
        {
			$isSimulation = 1;
			$simulationFilePath = $1;
        }
	}
	close FILE;	
	if ($isSimulation)
	{
		# Not calling the CPT
		$LOGGER->debug("Running the simulation. Using $simulationFilePath as the recording file.");
		open R_FILE, "<$simulationFilePath" or $LOGGER->fatal_error_code("File is not available: $simulationFilePath");
		foreach $line (<R_FILE>)
		{
			if($line =~ m/<(.*?)>/ && ($1 eq "services.xml" || $1 eq "network_objects.xml" || $1 eq "statuses.xml" || $1 eq "fw_policies.xml" || $1 eq "properties.xml" || $1 eq "/services.xml" || $1 eq "/network_objects.xml" || $1 eq "/statuses.xml" || $1 eq "/fw_policies.xml" || $1 eq "/properties.xml"))
			{
				$tagValue = $1;
				if ($tagValue eq "services.xml" || $tagValue eq "network_objects.xml" || $tagValue eq "statuses.xml" || $tagValue eq "fw_policies.xml" || $tagValue eq "properties.xml")
				{
					$tag = "true";
					unless (-d $cpsms_dir){
						make_path $cpsms_dir;
					}
					open FILE, ">$cpsms_dir/$tagValue" or $LOGGER->fatal_error_code("File is not available: $cpsms_dir/$tagValue");
					print FILE $xmlTag;
					next;
				}
				else
				{
					$tag = "false";
					close FILE;
					$LOGGER->debug("Simulation file $cpsms_dir$tagValue generated.");
				}
			}
			if($tag eq "true")
			{
				print FILE $line;
			}
		}
		close R_FILE;		
	}
	else
	{
		# execute Check Point Tool
		_callCPT($cpt, "fw_policies",$cpt_dir);
		_callCPT($cpt, "network_objects",$cpt_dir);
		_callCPT($cpt, "properties",$cpt_dir);
		_callCPT($cpt, "services",$cpt_dir);
		_callCPT($cpt, "statuses",$cpt_dir);
		
	}

	# Check to see if recording is enabled and create a recording
	if ( $ENV{ENABLE_RECORDING} )
    {
    	$LOGGER->debug("Running Recording...");
    	# Calculate the time of the recording
		my ( $sec, $min, $hour, $day, $month, $year ) = localtime();
		my $current_time = sprintf(
			'%s-%02s-%02s_%02s-%02s-%02s',
			( $year + 1900 ),
			( $month + 1 ),
			$day,
			$hour,
			$min,
			$sec
		);
    	my $recordingFileName="CPSMS_backup_".$ip_addr."_".$current_time.".xml";    	
    	
		open FILE, ">$recording_dir/$recordingFileName" or $LOGGER->fatal_error_code("File is not available: $recording_dir/$recordingFileName");
		print FILE $xmlTag;
		print FILE "<tables>\n";
		close FILE;

		opendir (DIR, $cpsms_dir) or $LOGGER->fatal_error_code("Directory not available: $cpsms_dir");
		my @xmlFiles = grep {/.*?\.xml/} readdir DIR;
		close DIR;

		open FILE, ">>$recording_dir/$recordingFileName" or $LOGGER->fatal_error_code("File is not available: $recording_dir/$recordingFileName");
		for my $xmlFile(@xmlFiles)
		{
			print FILE "<$xmlFile>\n";
			my $xmlFilePath = $cpsms_dir."/".$xmlFile;
			
			tie my @lines, 'Tie::File', $xmlFilePath or $LOGGER->fatal_error_code("File is not available: $xmlFilePath");
			for (my $count = 1; $count < scalar (@lines); $count++) 
			{
				print FILE $lines[$count];
			}

			untie @lines;
			print FILE "\n";
			print FILE "</$xmlFile>\n";
			
			
		}
		print FILE "</tables>";
		close FILE;
		$LOGGER->debug("Recording has been generated at $recording_dir/$recordingFileName");
    }
	
	# check if recording was enabled
	my $env_rec = (defined ($ENV{ENABLE_RECORDING}) )? 1 : 0; 
	# set Check Point SMS client's context
	$responses->{cpsms_ctx}={
		connectionPath => $connectionPath,
		cpt_dir => $cpt_dir,
		xml_dir => $xml_dir,
		xsl_dir => $xsl_dir,
		env_rec => $env_rec,
	};
	
	$LOGGER->debug("ZipTie::Adapters::CheckPoint::CPSMS; parsing system...");
	parse_system($responses, $printer);
	$LOGGER->debug("ZipTie::Adapters::CheckPoint::CPSMS; parsing chassis...");
	parse_chassis($responses, $printer);
	$LOGGER->debug("ZipTie::Adapters::CheckPoint::CPSMS; parsing object groups...");
	parse_CPSMS_object_groups($responses, $printer);
	# clean the unused responses
	delete $responses->{'services'};
	delete $responses->{'serviceGroup'};
	$LOGGER->debug("ZipTie::Adapters::CheckPoint::CPSMS; parsing filters...");
	parse_filters($responses, $printer);
	$LOGGER->debug("ZipTie::Adapters::CheckPoint::CPSMS; parsing interfaces...");
	parse_CPSMS_interfaces( $responses, $printer );
	$LOGGER->debug("ZipTie::Adapters::CheckPoint::CPSMS; parsing snmp...");
	parse_snmp( $responses, $printer );
	$LOGGER->debug("ZipTie::Adapters::CheckPoint::CPSMS; parsing static complete...");
	parse_static_routes( $responses, $printer );

	$printer->close_model();                # close out the ZiptieElementDocument
	close_model_filehandle($filehandle);    # Make sure to close the model file handle
	$LOGGER->debug("ZipTie::Adapters::CheckPoint::CPSMS; backup() complete!");
}

sub _callCPT
{
	# grab our arguments
	my ($cpt, $table_name, $appdir) = @_;
	
	# clean possible previous CheckPoint tools traces
	my $temp_dir = $cpt->{out_dir};
	$cpt->{out_dir} = $cpt->{out_dir}."/logs"."/".$cpt->{managementServerIp}."/".$cpt->{ip_addr};
	my $out_dir = $cpt->{out_dir};
	$cpt->clean($cpt->{out_dir});
	# run CPT
	$cpt->{out_dir} = $temp_dir;
	$cpt->run($table_name, $appdir);
	# clean temporary created files by cpsms_client
	$cpt->deletefiles($out_dir,"CKP");
	$cpt->deletefiles($out_dir,"ICA");
	$cpt->deletefiles($out_dir,"ops_prng");
	$cpt->deletefiles($out_dir,"session");
	
}
1;

__END__

=head1 NAME

ZipTie::Adapters::CheckPoint::CPSMS - Adapter for performing various operations against Check Point Security Management Server(s).

=head1 SYNOPSIS

    use ZipTie::Adapters::CheckPoint::CPSMS;
	ZipTie::Adapters::CheckPoint::CPSMS::backup( $backup_document );

=head1 DESCRIPTION

This module represents an adapter that can be used to perform various operations against Check Point Security Management Server(s).

Generally, this module is run through the ZipTie C<invoke.pl> script.

=head1 PUBLIC SUB-ROUTINES

=over 12

=item C<backup($backup_document)>

Performs the backup of the device that is described in the specified XML document. This XML document contains
a "connectionPath" element that contains the IP/hostname, protocol, credential, and file server information that
may be needed to connect to, authenticate with, and back up the device in question. This XML will be parsed into
a C<ZipTie::ConnectionPath> object to help easily access this vital information.

The act of backing up a device means that information about a device will be retrieved, as well as the configuration
data/files for that device.  All of this collected information will be used to populate the ZipTie Element Document,
which is an open and extensible format for describing a network device.

The return value for this method will be a string containing the successfully populated ZipTie Element Document.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_callCPT($cpt, $table_name, $appdir)>

Sub-routine handles Check Point Tool (CPT) client application execution. 
Passed parameters: 
- Check Point Tool object, 
- Check Point OPSEC table's name (i.e network_objects, statuses,...) for which the content is desired to be retrieved, 
- CPT's application directory.   

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

Portions created by IBM are Copyright (C) 2013, IBM Corp.

=head1 AUTHOR

Contributor(s): Toni Barabas (toni.barabas@ca.ibm.com) Gomed Saini (gomeds@ca.ibm.com)
Date: May 16, 2013

=cut
