package ZipTie::Adapters::CheckPoint::CPSMS::Parsers;

use strict;
use warnings;

use MIME::Base64 'encode_base64';
use XML::Simple;
use XML::LibXML;
use v5.10;
use ZipTie::Adapters::CheckPoint::CPSMS::CPSMSConverter;
use ZipTie::Adapters::CheckPoint::CPSMS::FileUtilities qw(getFileToString getFilePath);
use ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser;
use ZipTie::Adapters::CheckPoint::OPSEC::Parsers qw(parse_rules_file generate_filter_list get_any);
use ZipTie::Adapters::CheckPoint::Utilities::FileUtility qw(getFileandParse);
use ZipTie::Adapters::Utils qw(create_network_element get_port_expression mask_to_bits get_any_address_address);

use Exporter 'import';
our @EXPORT_OK =
  qw(parse_chassis parse_CPSMS_interfaces parse_CPSMS_object_groups parse_filters parse_system parse_static_routes parse_snmp);

my $LOGGER = ZipTie::Logger::get_logger();
use constant FAUX_SINGLE_INTERFACE_NAME => "eth0";

sub parse_system
{
	my ( $in, $out ) = @_;
	
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	# extract system from network_objects.xml
	$in->{xsl} = "system";
	$in->{device_system} = _cp_to_sed_xml($in, {"ipaddr"=>"ipaddr"});
	
	# get and store CP fw_policies.xml
	$in->{xml} = "fw_policies";  
	$in->{fw_policies} = _cp_to_doc($in);
		
	# get device's ip address 
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# create a new parser for network objects 
	my $parserno = XML::LibXML->new();	
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");
	
	# create a new parser for firewall policies
	my $parserfw = XML::LibXML->new();
	# create a new XPath context and register namespace for firewall policies
	my $XPCFW = _xpcregistration($parserfw,$in->{fw_policies},"fw_policies");
	
	# get gateway's host name, interfaces
	my $gateway = _getGateway($XPCNO, $ip_addr, 0);
	
	# create a new DOM
	my $dom = $parserno->parse_string($in->{device_system});
	# and set document's root
	my $doc = $dom->documentElement();
	
	# check NAT existence
	_checkNAT($in,$XPCNO, $XPCFW, $gateway);
	
	 # print nodes and elements	
	_printelement($doc,'core:systemName',$out);
	$out->open_element('core:osInfo');
		_printelement($doc,'core:make',$out);
		_printelement($doc,'core:name',$out);
		_printelement($doc,'core:version',$out);
		_printelement($doc,'core:osType',$out);
		if ($in->{isNAT}) {
			# get an addressTranslationParser object
			my $addressTranslationParser = ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser->new ($in);
			my $packetProcessingOrdering = $addressTranslationParser->getPacketProcessingOrdering();
			$out->print_string($packetProcessingOrdering);
		} 
	$out->close_element('core:osInfo');
	_printelement($doc,'core:biosVersion',$out);
	_printelement($doc,'core:deviceType',$out);
}

sub parse_chassis
{
	my ( $in, $out ) = @_;
	
	my $chassis = {};
	# get and store CP statuses.xml
	$in->{xml} = "statuses";  
	$in->{statuses} = _cp_to_doc($in);
	
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	
	# get device's ip address 
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# create a new parser 
	my $parserno = XML::LibXML->new();
	
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");

	# get gateway's host name, interfaces
	my $gateway = _getGateway($XPCNO, $ip_addr, 1);
	_getClusteringInfo($XPCNO, $gateway);
	
	# create a new parser 
	my $parserst = XML::LibXML->new();
	
	# create a new XPath context and register namespace for statuses
	my $XPCST = _xpcregistration($parserst,$in->{statuses},"statuses");

	# add total memory info to gateway
	_getMemoryStat($XPCST,$gateway);
	my $model = _getModel($XPCNO,$XPCST, $gateway);
	
	my $memory = {
		kind => "RAM",
	};
	
	foreach my $totalmem (@{$gateway->{MemoryStat}}){
		if (exists $totalmem->{totalVirtMem}){
			$memory->{size} = $totalmem->{totalVirtMem};
			push (@{$chassis->{memory}}, $memory);	
		}
	} # end foreach totalmem
	
	# set vendor and model
	$chassis->{'core:asset'}->{'core:assetType'} = 'Chassis';
	$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:make'} = 'Check Point';
	if (defined $model){
		$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $model;
	} else {
		$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = "Unknown";
	}
	
	$out->print_element('chassis', $chassis) if (defined $chassis) ;
		
}

sub parse_CPSMS_object_groups 
{	
	my ( $in, $out ) = @_;
	
	# get device's ip address
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# extract networkGroup from network_objects.xml
	$in->{xml} = "network_objects";
	$in->{xsl} = "networkGroup";
	$in->{networkGroup} = _cp_to_sed_xml($in);
	# get and store CP network_objects.xml
	$in->{network_objects} = _cp_to_doc($in);
	
	# create a new parser for network object
	my $parserno = XML::LibXML->new();
	
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");
		
	# extract pseudo protocolGroup and serviceGroup from services.xml
	$in->{xml} = "services";
	$in->{xsl} = "pseudoProtocolServiceGroup";
	$in->{xmlstring} = _cp_to_sed_xml($in);

	# get and store CP services.xml
	$in->{services} = _cp_to_doc($in);
	
	# create a new parser for services
	my $parsersrv = XML::LibXML->new();
	
	# create a new XPath context and register namespace for services
	my $XPCSRV = _xpcregistration($parsersrv,$in->{services},"services");
	
	# extract protocolGroup and serviceGroup from $in->{xmlstring} xmlstring
	# by default $in->{xml} = "";
	$in->{xsl} = "protocolServiceGroup";
	$in->{protocolServiceGroup} = _cp_to_sed_xmlstring($in);
												
	# create a new parser 
	my $parser = XML::LibXML->new();
	
	# networkGroup
	# create a new XPath context and register namespace
	my $XPC = _xpcregistration($parser, $in->{networkGroup},"ngns");
	
	$out->open_element('objectGroups');
	parse_network_groups ($out, $XPC, $ip_addr, $in, $XPCNO, $LOGGER);
		 	
	# create a new XPath context and register namespace
	$XPC = _xpcregistration($parser, $in->{protocolServiceGroup},"psgns");
	
	# protocolGroup
	parse_protocol_groups ($out, $XPC);
	
	# serviceGroup	
	parse_service_groups ($out, $XPC, $XPCSRV, $in);
		
	$out->close_element('objectGroups');
}

sub parse_filters {
	
	my ($in, $out) = @_;
	
	$out->open_element("filterLists");

	# check NAT existence 
	if ($in->{isNAT}){
		# print out packetFiltering
		# this happens before translating and routing regardless of 
		# 'translate destination on client side' (nat_dst_client_side) switch value from properties.xml
		$out->open_element('filterListsSection');	
			_packetFiltering($in,$out);
		$out->close_element('filterListsSection');
		# print out address-translation
		_addressTranslation ($in, $out);
	} else {
		_packetFiltering($in,$out);
	}
	
	$out->close_element("filterLists");
	
}	

sub _packetFiltering {
	
	my ( $in, $out ) = @_;
	
	# get and store CP fw_policies.xml
	$in->{xml} = "fw_policies";  
	$in->{fw_policies} = _cp_to_doc($in);
	
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	
	# get and store CP properties.xml
	$in->{xml} = "properties";
	$in->{properties} = _cp_to_doc($in);
	
	# get and store CP services.xml
	$in->{xml} = "services";
	$in->{services} = _cp_to_doc($in);
	
	# get and store CP statuses.xml
	$in->{xml} = "statuses";
	$in->{statuses} = _cp_to_doc($in);
	
	# get device's ip address
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $managementServerIp = $connectionPath->get_management_server_ip_address();
	my $ip_addr = $connectionPath->get_ip_address();
	
	# get CPT directory
	my $cpt_dir = $in->{cpsms_ctx}->{cpt_dir};
	
	# set CPT xml output directory
	my $xml_dir = $in->{cpsms_ctx}->{xml_dir};
	$xml_dir = $xml_dir."/".$managementServerIp . "/". $ip_addr;
	
	# get recording's flag
	my $env_rec = $in->{cpsms_ctx}->{env_rec};
	
	# create parsers, XPath contexts, and register nampespaces
	
	# create a new parser for firewall policies
	my $parserfw = XML::LibXML->new();
	# create a new XPath context and register namespace for firewall policies
	my $XPCFW = _xpcregistration($parserfw,$in->{fw_policies},"fw_policies"); 

	# create a new parser for network object
	my $parserno = XML::LibXML->new();
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");	

	# create a new parser for properties
	my $parserpr = XML::LibXML->new();
	# create a new XPath context and register namespace for properties
	my $XPCPR = _xpcregistration($parserpr,$in->{properties},"properties"); 
	
	# create a new parser for services
	my $parsersrv = XML::LibXML->new();
	# create a new XPath context and register namespace for properties
	my $XPCSRV = _xpcregistration($parsersrv,$in->{services},"services");
	
	# create a new parser for statuses
	my $parserStatuses = XML::LibXML->new();
	# create a new XPath context and register namespace for properties
	my $XPCSTS = _xpcregistration($parserStatuses,$in->{statuses},"statuses");
	
	# get gateway's host name, interfaces
	my $gateway = _getGateway($XPCNO, $ip_addr, 1);
	
	# get clustering info if exists and store for NAT
	my $clusterInfo = _getClusteringInfo($XPCNO, $gateway);
	$in->{clusterInfo} = $clusterInfo;
	
	# get properties enablement by parsing properties and store the enabled properties elements plus the positional elements
	$in->{enabled_services} = _getPropertiesEnablement($XPCPR);
	
	# initialize default $filterList
	my $filterList = {
        mode => 'stateful',
    };
	
	# generate the appliesTo element that contains the interfaces where the rules are applied
	# +++wayne I don't think this is right.  Aren't filters enabled by applied interface?
	# +++wayne but here I go piling on even if it is a bug.  There's a possibility that we
	# didn't get interface detail for the local system.  In this case, there is a fake
	# name we've concocted for a single syntheized interface
	
	if (defined $gateway->{intfs})	{
		foreach my $intf (@{$gateway->{intfs}}) {
			push(@{$filterList->{appliesTo}}, {source => $intf->{officialname}});
			
		}
	}
	else	{
		push(@{$filterList->{appliesTo}}, {source => FAUX_SINGLE_INTERFACE_NAME});
	}
	my @appliesTo = @{$filterList->{appliesTo}};
	
	# disabled the implied_rules.C and recimpliedFeSets files generation
	# rem out the next line if one wants to re-create these files 
	# note: this procedure requires expert credentials plus SSH protocol enabled 
	# on Protocols page through CSM UI
	$env_rec = 0; 
	
	# generate impliedFeSets in case $env_rec is enabled 
	if ($env_rec) {
		# set the full path name of the new recorded impliedFeSets
		$in->{impliedFeSets} = getFilePath("recimpliedFeSets",$xml_dir);
		# set the full path name for the local implied_rules.C
		$in->{impliedFile} = getFilePath("implied_rules.C",$xml_dir);
		getFileandParse($in,$connectionPath);	
	}
	
	# FOR EACH implied rule, we....
	#    * generate the filter list
	#    * _prune_ the list for entries which have either (any src/any dest) plus no
	#          qualifying protocol, since we're right now believing that those aren't
	#          actually being acted on by the firewall
	#    * unless this is the "final" implied rule list (which we have to generate even
	#           if it's empty owing to its important defaujltAction), if it is emnpty,
	#           bypass generating it since it's just confusing noise.
	
	# generate first implied rules filterList 
	my $properties = getFileToString("impliedFeSets", $cpt_dir);
	my $impliedFeSets = _processProperties($XPCPR, $properties, $in->{enabled_services});
	
	my $serviceGroups = $in->{portServiceGroups};
	
	_prune_global_rules($impliedFeSets->{'first'}, $serviceGroups);
	
	# if NAT configuration is enabled then set processingPhase
	if ($in->{isNAT}) {
		$in->{NAT}->{settings}->{processingPhase} = "packet-filtering";
	}
	if (defined $impliedFeSets->{'first'} && scalar(@{$impliedFeSets->{'first'}})) {		
		generate_filter_list( { out => $out, filterEntries => $impliedFeSets->{'first'}, 
				filterName => "Implied-first", filterDesc => "Global rules - 'first' entries", 
				applicableInterfaces => \@appliesTo, defaultAction => "none", in => $in});
	}
	
	# generate OPSEC user rules filterList
	my $processOrder = 0;
	$out->open_element('filterList');
	$out->print_element('defaultAction', 'none');
	$out->print_element('description', 'OPSEC user rules');
	$out->print_element('appliesTo',\@appliesTo);
	
	my $installedPolicy;
	# get installed policy from statuses
	if ($clusterInfo->{isGatewayCluster}){
		foreach my $member (@{$clusterInfo->{cluster_members}}){
			$installedPolicy = _getInstalledPolicy($XPCSTS, $member);
			last if $installedPolicy ne "";
		}
	} else {
		$installedPolicy = _getInstalledPolicy($XPCSTS, $gateway->{hostname});
	}
	# store for NAT
	$in->{installedPolicy} = $installedPolicy;
	# find object nodes from rules and create filterEntry
	RULE: foreach my $libxmlobject ($XPCFW->findnodes("/fw_policies/object[collection='referenced_object: $installedPolicy']/rule/rule/object")){
		my $filterEntry = {};
		# check if rule is disabled
		my $isDisabled = $libxmlobject->findvalue('./disabled');
		if ($isDisabled eq "True"){
			next RULE;
		}
		
		# find object_uid -> ruleUID essential for rule counting
		my $rule_uid = $libxmlobject->findvalue('./object_uid');
		if ($rule_uid ne "" && $rule_uid ne "empty_object_uid") {
			$filterEntry->{ztSignature} = $rule_uid;
		}
		
		my $installObjValue = $libxmlobject->findvalue('./install/object/empty_field_name/empty_field_name');
		my @targetArray = split ("referenced_object: ",$installObjValue);

		# check if is cluster member and compare target with cluster gateways
		my $isGatewayClusterTarget;
		if ($clusterInfo->{isClusterMember} && scalar @{$clusterInfo->{gateway_clusters}}){
			foreach my $gatewayCluster (@{$clusterInfo->{gateway_clusters}}){
				if ($gatewayCluster ~~ @targetArray){
					$isGatewayClusterTarget = 1;
					last;
				}
			}
		}
		if ( ('Any' ~~ @targetArray) ||  ('Gateways' ~~ @targetArray) || ($gateway->{hostname} ~~  @targetArray) ||$isGatewayClusterTarget ) {
		#if ( (@targetArray eq 'Any') ||  ($target eq 'Gateways') || ( $target eq $gateway->{hostname}) ||$isGatewayClusterTarget ) {
			# find rule's destination value
			my $dstValue = $libxmlobject->findvalue('./dst/object/empty_field_name/empty_field_name');
			my @dstArray = split ("referenced_object: ",$dstValue);
			
			foreach my $dst (@dstArray){
				if ($dst eq "Any") {
	    			push (@{$filterEntry->{destinationIpAddr}}, get_any());
				} elsif ($dst ne ""){
					if (defined $in->{networkGroups}->{$dst}){
						push (@{$filterEntry->{destinationIpAddr}}, _set_objectGroupReference($dst));
					} else{
						$LOGGER->debug("_packetFiltering; Object group reference $dst NOT FOUND. Dropping current rule.");
						next RULE;
					}
				};
			}; #end foreach dst
			
			# find rule's services 
			foreach my $libxmlempty_field_name ($libxmlobject->findnodes('./services/object/empty_field_name')) {
				my $empty_field_name = $libxmlempty_field_name->findvalue('./empty_field_name');
				my @services = 	split ("referenced_object: ",$empty_field_name);
				foreach my $service (@services){
					if ($service ne "" && $service !~ /Any/) {
						# check for __src serviceGroups
						_checkServiceGroup($in,$service,$filterEntry);	
					}	
				} # end foreach services
			} # end foreach empty_field_name
			
			# find rule's track method
			my $track = _substring_after($libxmlobject->findvalue('./track/track'),"referenced_object: ");
			if( $track eq "Log"){
				$filterEntry->{log} = "true";
			}else {
				$filterEntry->{log} = "false";
			}

			my $name = $libxmlobject->findvalue('./name');
			if ($name ne "") {
				$filterEntry->{name} = $name;
			}
			
			# find type (permission) values
			my $type = $libxmlobject->findvalue('./action/action/object/type');
			if ($type eq "accept"){
				$type = "permit";
			}
			if ($type eq "reject"){
				$type = "deny";
			}
			# other possible action type drop, encrypt, decrypt, key install
			if ($type ne ""){
				$filterEntry->{permitAction} = $type;
			}

			# find source value
			my $srcValue = $libxmlobject->findvalue('./src/object/empty_field_name/empty_field_name');
			my @srcArray = split ("referenced_object: ",$srcValue);
			
			foreach my $src (@srcArray){
				if ($src eq "Any") {
	    			push (@{$filterEntry->{sourceIpAddr}}, get_any());
				} elsif ($src ne ""){
					if (defined $in->{networkGroups}->{$src}){
						push (@{$filterEntry->{sourceIpAddr}}, _set_objectGroupReference($src));
					} else{
						$LOGGER->debug("_packetFiltering; Object group reference $src NOT FOUND. Dropping current rule.");
						next RULE;
					}
				};	
			}; #end foreach src 
			
		_printFilterEntry($filterEntry, ++$processOrder, $out);
		} # end if target		
	} #foreach libxmlobject
	
	# close the one-time elements for OPSEC user rules	
	$out->print_element('mode', 'stateful');
	$out->print_element('name', 'OPSEC');
	if ($in->{isNAT}) {
		# get an addressTranslationParser object	
		my $addressTranslationParser = ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser->new ($in);
		my $packetProcessingPhase = $addressTranslationParser->getPacketProcessingPhase();
		$out->print_string($packetProcessingPhase);
	} 
	$out->close_element('filterList');
	
	# generate the unordered implied rules
	_prune_global_rules($impliedFeSets->{'unordered'}, $serviceGroups);
	
	if (defined $impliedFeSets->{'unordered'} && scalar(@{$impliedFeSets->{'unordered'}})) {		
		generate_filter_list({out => $out,filterEntries =>  $impliedFeSets->{'unordered'}, 
				filterName => "Implied", 
				filterDesc => "Global unordered rules", 
				applicableInterfaces => \@appliesTo, defaultAction => "none", in => $in});
	}

	# generate the before last implied rules
	_prune_global_rules($impliedFeSets->{'before last'}, $serviceGroups);
	
	if (defined $impliedFeSets->{'before last'} && scalar(@{$impliedFeSets->{'before last'}})) {		
		generate_filter_list({out => $out,filterEntries =>  $impliedFeSets->{'before last'}, 
				filterName => "Implied-before-last", 
				filterDesc => "Global rules to apply before default last rules", 
				applicableInterfaces => \@appliesTo, defaultAction => "none", in => $in});
	}

	# generate the last implied rules
	if (defined $impliedFeSets->{'last'} && scalar(@{$impliedFeSets->{'last'}})) {	
		generate_filter_list({out => $out, filterEntries =>  $impliedFeSets->{'last'}, 
				filterName => "Implied-last", filterDesc =>  "Global default last rules", 
				applicableInterfaces =>\@appliesTo, 
				defaultAction => "drop", forceGeneration => 1, in => $in});
	}
	
}

sub _getClusteringInfo{
	my ($XPCNO, $gateway) = @_;
	
	my $hostname = $gateway->{hostname};
	my @noObjectArray = $XPCNO->findnodes("/network_objects/object[object_name='$hostname']");
	if (scalar @noObjectArray eq 0){
		return;
	}
	my $noObject = $noObjectArray[0];
	my $type = $noObject->findvalue('./type');
	if ($type eq "gateway_cluster"){
		$gateway->{isGatewayCluster} = 1;
		# get cluster members cluster_members
		my @clusterMembers = $noObject->findnodes('./cluster_members');
		foreach my $member (@clusterMembers){
			my $memberValue = $member->findvalue('.');
			$memberValue = _substring_after($memberValue,"referenced_object: ");
			push @{$gateway->{cluster_members}}, $memberValue;
		}
	} elsif ($type eq "cluster_member"){
		$gateway->{isClusterMember} = 1;
		my @gatewayClusterArray = $XPCNO->findnodes("/network_objects/object[type='gateway_cluster']");
		foreach my $gatewayCluster (@gatewayClusterArray){
			my $gatewayClusterName = $gatewayCluster->findvalue('./object_name');
			# get cluster members cluster_members
			my @clusterMembers = $gatewayCluster->findnodes('./cluster_members');
			foreach my $member (@clusterMembers){
				my $memberValue = $member->findvalue('.');
				$memberValue = _substring_after($memberValue,"referenced_object: ");
				if ($memberValue eq $hostname){
					push @{$gateway->{gateway_clusters}}, $gatewayClusterName;
				}
			}
		}
	}
	return $gateway;
}

sub _getInstalledPolicy {
	my ($XPCSTS, $hostname) = @_;
	
	my $installedPolicy = "";
	foreach my $libxmlobject ($XPCSTS->findnodes('/statuses/object')){
		my $appObj = $libxmlobject->findvalue('./application_obj');
		$appObj= _substring_after($appObj,"referenced_object: ");
		# find the matching hostname cpshared_application
		if ($appObj =~ /firewall_application_$hostname/){
			$installedPolicy = $libxmlobject->findvalue('./fwStatus/object/fwPolicyStat/object/fwPolicyName');
			return $installedPolicy;
		}
	}
}

sub _printFilterEntry {
	my ($filterEntry, $processOrder, $out) = @_;
	$out->open_element ('filterEntry');
	foreach my $val (@{$filterEntry->{destinationIpAddr}}){
			$out->print_element ('destinationIpAddr',$val);
	}
	if ( defined(@{$filterEntry->{destinationService}}) and (scalar @{$filterEntry->{destinationService}}) ){
			foreach my $service (@{$filterEntry->{destinationService}}){
				$out->open_element('destinationService');
				$out->print_element('objectGroupReference', $service);
				$out->close_element('destinationService');	
			}		
	}
	$out->print_element('log', $filterEntry->{log});
	$out->print_element('name',$filterEntry->{name}) if defined $filterEntry->{name};
	$out->print_element('primaryAction',$filterEntry->{permitAction});
    $out->print_element ('processOrder', $processOrder);
    foreach my $val (@{$filterEntry->{sourceIpAddr}}){
			$out->print_element ('sourceIpAddr',$val);
	}
	if ( defined (@{$filterEntry->{sourceService}})  and (scalar @{$filterEntry->{sourceService}})){
			foreach my $service (@{$filterEntry->{sourceService}}){
					$out->open_element('sourceService');
					$out->print_element('objectGroupReference', $service);
					$out->close_element('sourceService');	
			}
	} 
	$out->print_element('ztSignature', $filterEntry->{ztSignature}) if defined $filterEntry->{ztSignature};	
	$out->close_element ('filterEntry');
}

sub _addressTranslation {
	
	my ($in, $out) = @_;
	
	# set manual NAT as undefined by default
	$in->{manualNAT}->{available} = 0;
			
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	
	# get and store CP properties.xml
	$in->{xml} = "properties";
	$in->{properties} = _cp_to_doc($in); 
	
	# get and store CP fw_policies.xml
	$in->{xml} = "fw_policies";
	$in->{fw_policies} = _cp_to_doc($in); 

	# create a new parser for firewall policies
	my $parserfw = XML::LibXML->new();
	# create a new XPath context and register namespace for firewall policies
	my $XPCFW = _xpcregistration($parserfw,$in->{fw_policies},"fw_policies");
	
	# create a new parser for network object
	my $parserno = XML::LibXML->new();
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");	

	# create a new parser for properties
	my $parserpr = XML::LibXML->new();
	# create a new XPath context and register namespace for properties
	my $XPCPR = _xpcregistration($parserpr,$in->{properties},"properties");
		
	$in->{isCPSMS} = 1;
	$in->{XPCFW} = $XPCFW;
	$in->{XPCNO} = $XPCNO;
	$in->{XPCPR} = $XPCPR;
	
	# get an addressTranslationParser object
	my $addressTranslationParser = ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser->new ($in);
	# store global NAT settings under $in->{NAT}->{settings}
	$addressTranslationParser->getNATProperties();
	
	#set processingPhase
	$in->{NAT}->{settings}->{processingPhase} = "address-translation";
	my $packetProcessingPhase = $addressTranslationParser->getPacketProcessingPhase();
	
	# check where the automatic NAT is applied, at the server side or at the client side
	if ( defined($in->{NAT}->{settings}->{nat_dst_client_side}) and  $in->{NAT}->{settings}->{nat_dst_client_side} eq 'True')  {
		$out->open_element('filterListsSection');
			# if translation is executed at the client side 
			my $routingNATFilterList = $addressTranslationParser->routingNAT('automatic');
			$out->print_string($routingNATFilterList);	
			# add the phase discriminator
			$in->{NAT}->{settings}->{processingPhase} = 'routing';
			$in->{NAT}->{settings}->{processingPhasePosition} = 'before';
			my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
			$out->print_string($phaseDiscriminatorPhase);
		$out->close_element('filterListsSection');		
	} else {
		$out->open_element('filterListsSection');
			# else translation is executed at the server side
			my $routingNATFilterList = $addressTranslationParser->routingNAT('automatic');
			$out->print_string($routingNATFilterList);
			# add the phase discriminator
			$in->{NAT}->{settings}->{processingPhase} = 'routing';
			$in->{NAT}->{settings}->{processingPhasePosition} = 'after';
			my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
			$out->print_string($phaseDiscriminatorPhase);
		$out->close_element('filterListsSection');
	}
	
	#check if manual NAT is defined
	$addressTranslationParser->checkManualNAT();
	if (defined ($in->{manualNAT}->{available}) and $in->{manualNAT}->{available} == 1){
		#check where the manual NAT is applied, at the server side or at the client side
		if ( defined($in->{NAT}->{settings}->{nat_dst_client_side_manual}) and $in->{NAT}->{settings}->{nat_dst_client_side_manual} eq 'True'){
			$out->open_element('filterListsSection');
				# if translation is executed at the client side
				my $routingNATFilterList = $addressTranslationParser->routingNAT('manual');
				$out->print_string($routingNATFilterList);	
				# add the phase discriminator
				$in->{NAT}->{settings}->{processingPhase} = 'routing';
				$in->{NAT}->{settings}->{processingPhasePosition} = 'before';
				my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
				$out->print_string($phaseDiscriminatorPhase);		
			$out->close_element('filterListsSection');	
		} else {
			$out->open_element('filterListsSection');
				# else translation is executed at the server side
				my $routingNATFilterList = $addressTranslationParser->routingNAT('manual');
				$out->print_string($routingNATFilterList);
				# add the phase discriminator
				$in->{NAT}->{settings}->{processingPhase} = 'routing';
				$in->{NAT}->{settings}->{processingPhasePosition} = 'after';
				my $phaseDiscriminatorPhase = $addressTranslationParser->getPhaseDiscriminator();
				$out->print_string($phaseDiscriminatorPhase);	
			$out->close_element('filterListsSection');
		}	
	} # check manual NAT existence
		
	
}

sub parse_CPSMS_interfaces {
	
	my ( $in, $out ) = @_;
	
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# create a new parser 
	my $parserno = XML::LibXML->new();
	
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");

	# get gateway's host name, interfaces
	my $gateway = _getGateway($XPCNO, $ip_addr, 1);
	
	$out->open_element ('interfaces');
		if (defined $gateway->{intfs})	{
			foreach my $intf (@{$gateway->{intfs}}) {
				$out->open_element('interface');
					$out->print_element ('adminStatus','up');
					$out->print_element('ifIndex',$intf->{ifIndex});
						$out->open_element ('interfaceIp');
							my $ref = create_network_element($intf->{ipaddr}, $intf->{netmask});
							$out->open_element('ipConfiguration');
								$out->print_element('ipAddress', $ref->{address});
								$out->print_element('mask', $ref->{mask});
							$out->close_element('ipConfiguration');
						$out->close_element('interfaceIp');
						my $physical;
						if ($intf->{officialname} =~ /eth/){
							$out->print_element('interfaceType','ethernet');
							$physical = 'true';
						} elsif ($intf->{officialname} =~ /lo/){
							$out->print_element('interfaceType','softwareLoopback');
							$physical = 'false';
						} else {
							$out->print_element('interfaceType','virtual');
							$physical = 'false';
						}
						$out->print_element('name',$intf->{officialname});
						$out->print_element('physical', $physical);
				$out->close_element('interface');
			}
		}
		else  { # gateway interfaces not defined
			# We have to create a single interface here since we lack the detail we need to place this.
			# We have the ip addr to assign to it, but everything else must be cobbled together.
			# I hope to be able to remove this code someday when we understand better what we can back
			# up, and what discovering nodes like this really means.
			$LOGGER->debug("parse_CPSMS_interfaces; synthesizing single interface for " .
				$ip_addr . " because interface detail not present in network_objects");

			$out->open_element('interface');
			$out->print_element ('adminStatus','up');
			$out->open_element ('interfaceIp');
			my $ref = create_network_element($ip_addr, "255.255.255.0"); # I know, I know....
			$out->open_element('ipConfiguration');
			$out->print_element('ipAddress', $ref->{address});
			$out->print_element('mask', $ref->{mask});
			$out->close_element('ipConfiguration');
			$out->close_element('interfaceIp');
			$out->print_element('interfaceType','ethernet');
			$out->print_element('name',FAUX_SINGLE_INTERFACE_NAME);						
			$out->print_element('physical', "true");
			$out->close_element('interface');
		}
	$out->close_element('interfaces');
	
}

sub parse_static_routes 
{
	
	my ( $in, $out ) = @_;
	
	my $staticRoutes;
	# get and store CP statuses.xml
	$in->{xml} = "statuses";  
	$in->{statuses} = _cp_to_doc($in);
	
	# get and store CP network_objects.xml
	$in->{xml} = "network_objects";
	$in->{network_objects} = _cp_to_doc($in);
	
	# get device's ip address 
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# create a new parser 
	my $parserno = XML::LibXML->new();
	
	# create a new XPath context and register namespace for network objects
	my $XPCNO = _xpcregistration($parserno,$in->{network_objects},"network_objects");

	# get gateway's host name, interfaces
	my $gateway = _getGateway($XPCNO, $ip_addr, 1);
	
	# create a new parser 
	my $parserst = XML::LibXML->new();
	
	# create a new XPath context and register namespace for statuses
	my $XPCST = _xpcregistration($parserst,$in->{statuses},"statuses");
	
	# add static routes info to gateway
	_getStaticRoutes ($XPCST, $gateway);	
	
	if (defined $gateway->{osRouting})
	{
		foreach my $osRouting (@{$gateway->{osRouting}}){
			if ($osRouting->{osRoutingDest} ne '127.0.0.0'
				&& $osRouting->{osRoutingDest} ne '127.0.0.1' 
				&& $osRouting->{osRoutingGateWay} ne '0.0.0.0') {
					my $route = {
						destinationAddress => $osRouting->{osRoutingDest},
						destinationMask    => mask_to_bits($osRouting->{osRoutingMask}),
					};
					$route->{gatewayAddress} = $osRouting->{osRoutingGateWay};
					$route->{interface}      = $osRouting->{osRoutingIntrf};
					if ( ( $route->{destinationAddress} eq '0.0.0.0' ) && ( $route->{destinationMask} eq '0' ) ){
						$route->{defaultGateway} = 'true';
					}
					else{
						$route->{defaultGateway} = 'false';
					}
					push( @{ $staticRoutes->{staticRoute} }, $route );	
			} # end for if osRouting
		} # end foreach osRouting
	}
	else {
		$LOGGER->debug("parse_static_routes; no static routes defined");
	}
	
	$out->print_element( 'staticRoutes', $staticRoutes ) if ( defined $staticRoutes );
		
}

sub parse_snmp {
	
	my ($in, $out) = @_;
	
	$in->{xml} = "network_objects";
	$in->{snmp} = _cp_to_doc($in);
	my $connectionPath=$in->{cpsms_ctx}->{connectionPath};
	my $ip_addr = $connectionPath->get_ip_address();
	
	# create a new parser 
	my $parser = XML::LibXML->new();
	
	# create a new XPath context and register namespace
	my $XPC = _xpcregistration($parser,$in->{snmp},"snmp");
	
	# find SNMP settings for the given ip_addr 
	foreach my $libxmlobject ($XPC->findnodes('//object')){
		my $ipaddr = ($libxmlobject->findvalue('./ipaddr'));
		if ($ipaddr eq $ip_addr ) {
			# find object node from SNMP
			foreach my $libxmlSNMP ($libxmlobject->findnodes('./SNMP')) {
				foreach my $libxmlObj ($libxmlSNMP->findnodes('./object')){
					#find values for a non-empty sysName values
					my $sysName = $libxmlObj->findvalue('./sysName');
					my %community;
					if(defined ($sysName) && ($sysName ne "")){
						$out->open_element('snmp');
							#search for read_community
							my $read_community = $libxmlObj->findvalue('./read_community');
							if (defined ($read_community) && ($read_community ne "")){
								my $accessType = "RO";
								my $communityString = $read_community;
								$community{$communityString}=$accessType;
							}
							
							#search for readwrite_community
							my $write_community = $libxmlObj->findvalue('./write_community');
							if (defined ($write_community) && ($write_community ne "")){
								my $accessType = "RW";
								my $communityString = $write_community;
								$community{$communityString}=$accessType;
							}
							
							#print community element 
							foreach my $key (keys %community){
								$out->open_element('community');
									$out->print_element('accessType',$community{$key});
									$out->print_element('communityString',$key);
								$out->close_element('community');
							}
							
							#sysContact
							my $sysContact = $libxmlObj->findvalue ('./sysContact');
							if(defined ($sysContact) && ($sysContact ne "")){
								$out->print_element('sysContact', $sysContact);
							}	
							
							#sysDescription
							my $sysDescr = $libxmlObj->findvalue ('./sysDescr');
							if(defined ($sysDescr) && ($sysDescr ne "")){
								$out->print_element('sysDescr', $sysDescr);
							}
							
							#sysLocation
							my $sysLocation = $libxmlObj->findvalue ('./sysLocation');
							if(defined ($sysLocation) && ($sysLocation ne "")){
								$out->print_element('sysLocation', $sysLocation);
							}
							
							$out->print_element('sysName', $sysName);
						$out->close_element('snmp');		
					} #end if sysName
				} #end foreach libxmlObj	
			} #end foreach libxmlSNMP
								
		} #end if ipaddr
	} #end foreach libxmlobject	
			
}

sub _xpcregistration {
	
	my ($parser, $in, $ns) = @_;
	
	# create a new DOM
	my $dom = $parser->parse_string($in);
	# set document's root
	my $doc = $dom->documentElement();
	# create a new XPath context and register namespace
	my $XPC     = XML::LibXML::XPathContext->new($doc);
	$XPC->registerNs( $ns => 'http://www.ziptie.org/model/common/1.0' );
	
	return $XPC;
	
}

sub _printnodes{
	
	my ($doc, $in, $out) = @_;
	
	my $ns = undef;
	my $item = undef;
	
	if (($in !~ /^core/) && ($in =~ /\:/)){
		($ns, $item) =split (':',$in);
	} else {
		$item = $in;
	}
	
	# find core:x nodes
	foreach my $libxmlvar ($doc->findnodes('//'.$in)) {
		if (defined $libxmlvar and $libxmlvar ne ''){
			my $xmlinvar = $libxmlvar->toString;
			my $ref = XMLin($xmlinvar, ForceArray =>1);
			$out->print_element($item, $ref);	
		} else {
			$out->print_element($item, 'unknown');
		}		
	}
}

sub _printelement{
	
	my ($doc, $in, $out) = @_;
	my $ns = undef;
	my $item = undef;
	
	if (($in !~ /^core/) && ($in =~ /\:/)){
		($ns, $item) =split (':',$in);
	} else {
		$item = $in;
	}
	
	# find core:x element's value
	foreach my $libxmlvar ($doc->findvalue('//'.$in)){
		if (defined $libxmlvar and $libxmlvar ne ''){
			my $ref = $libxmlvar;
			$out->print_element($item, $ref);	
		} else {
			$out->print_element($item, 'unknown');
		}
		
	} 
}

sub _cp_to_sed_xml
{
	my ($in,$xp_params) = @_;
	
	# create a new CPSMSConverter session
	my $cptxml = ZipTie::Adapters::CheckPoint::CPSMS::CPSMSConverter->new(connectionPath => $in->{cpsms_ctx}->{connectionPath},
																			cpt_dir => $in->{cpsms_ctx}->{cpt_dir},
																			xml_dir => $in->{cpsms_ctx}->{xml_dir},
																			xsl_dir => $in->{cpsms_ctx}->{xsl_dir},
																			xml => $in->{xml},
																		    xsl => $in->{xsl},
																		    xp_params => $xp_params);
	my $result = $cptxml->cp2sedxml();
	return $result;																		    															   		
}

sub _cp_to_sed_xmlstring
{
	my ($in,$xp_params) = @_;
	
	my $xml = exists $in->{xml} ? $in->{xml} : "";
	# create a new CPSMSConverter session
	my $cptxml = ZipTie::Adapters::CheckPoint::CPSMS::CPSMSConverter->new(connectionPath => $in->{cpsms_ctx}->{connectionPath},
																			cpt_dir => $in->{cpsms_ctx}->{cpt_dir},
																			xml_dir => $in->{cpsms_ctx}->{xml_dir},
																			xsl_dir => $in->{cpsms_ctx}->{xsl_dir},
																			xml => $xml,
																			xmlstring => $in->{xmlstring},
																		    xsl => $in->{xsl},
																		    xp_params => $xp_params);
	my $result = $cptxml->cp2sedxmlstring();
	return $result;																		    															   		
}

sub _cp_to_doc
{
	my ($in,$xp_params) = @_;
	
	# create a new CPSMSConverter session
	my $cptxml = ZipTie::Adapters::CheckPoint::CPSMS::CPSMSConverter->new(connectionPath => $in->{cpsms_ctx}->{connectionPath},
																			cpt_dir => $in->{cpsms_ctx}->{cpt_dir},
																			xml_dir => $in->{cpsms_ctx}->{xml_dir},
																			xsl_dir => $in->{cpsms_ctx}->{xsl_dir},
																			xml => $in->{xml},
																		    xsl => $in->{xsl},
																		    xp_params => $xp_params);
	my $result = $cptxml->cpxml2doc();
	return $result;																		    															   		
}

# returns substring that is found after a passed string
sub _substring_after
{
	my ($in,$after) = @_ ;
	my $result;
	if ($in =~ /$after(.*)/){
		$result = $1;
	}
	return $result;
}

# set objectGroupReference
sub _set_objectGroupReference{
	
	my ($in) = @_;
	my @obj;
	push (@obj ,{objectGroupReference => $in});
	return @obj;
	
}

# get Gateway's information: hostname, interfaces {ipaddr, netmask, interface's name}
sub _getGateway {
	
	my ($XPCNO, $ip_addr, $forceLocalAddrResolve) = @_;
	
	# if '$forceLocalAddrResolve' is nonzero, then we will be using force majeur to 
	# resolve this address as the local gateway.
	
	my $gateway;
	my $provisionalHostname;
	# find and store host interface settings for a given ipaddr 
	foreach my $libxmlobject ($XPCNO->findnodes('/network_objects/object')){
		my $hostname = $libxmlobject->findvalue('./object_name');
		if (! exists ($gateway->{hostname})  ){
			$gateway = {};
			foreach my $libxmlinterface ($libxmlobject->findnodes('./interfaces')){
				my $interface = {};
				foreach my $libxmlintfobj ($libxmlinterface->findnodes('./object')) {
					# store interface's ip address 
					$interface->{ipaddr} = $libxmlintfobj->findvalue('./ipaddr');
					# store interface's ifindex
					$interface->{ifIndex} = $libxmlintfobj->findvalue('./ifindex');
					# store interface's mask
					$interface->{netmask} = $libxmlintfobj->findvalue('./netmask');
					# store interface's name
					$interface->{officialname} = $libxmlintfobj->findvalue('./officialname');
					# check if this interface is external or internal
					foreach my $libxmltypeintf ($libxmlintfobj->findnodes('./security/object/netaccess/object')) {
						my $typeofIntf = $libxmltypeintf->findvalue('./leads_to_internet');
						if ($typeofIntf eq 'True'){
							$interface->{typeofIntf} = 'external';	
						} else {
							$interface->{typeofIntf} = 'internal';
						}
					}
				} # end foreach libxmlintfobj
				
				# store gateway's interfaces and host name
				if (exists $interface->{ipaddr}) {
					push (@{$gateway->{intfs}}, $interface);
					if ($interface->{ipaddr} eq $ip_addr){
						# found gateway required to be backed up
						$gateway->{hostname} = $hostname;
						# the next two variables are added for later crosschecking (i.e. parse_rules) 
						$gateway->{my_ip} = $interface->{ipaddr};
						$gateway->{my_ip_and_local_cluster}->{my_ip} = $interface->{ipaddr};
					}
				} #end if store gateway's interfaces and host name
			} # end foreach libxmlinterface
			# We may have a circumstance where the actual gateway object may not have
			# interface-level detail.  In this case, the only IP addr in the node 
			# we care about will come about from the {ipaddr} node.  This will really not
			# ultimately result in great SED detail, but we want to be able to tiptoe
			# through the backup to produce something topologically placeable at one interface
			# This and parse_interfaces will require special handling to that end.
			#
			# I hope this code can be removed someday when we get a better idea of what kinds of
			# gateways we can actually support.
			#
			if (not defined $gateway->{hostname}  and not defined $provisionalHostname) 
			{
				if ($forceLocalAddrResolve)
				{
					my $locIpAddr = $libxmlobject->findvalue('./ipaddr');
					if (defined($locIpAddr) and $locIpAddr eq $ip_addr)
					{
						# we may find the definitive entry later, but let's store this for now.
						# if multihomed interface data is not initialized on the gateway, it might
						# be as good as we can get.
						$provisionalHostname = $hostname;
					}
				}
			}
			 	
		} #end if not exists		
	} #end foreach libxmlobject
	
	# now, do we have to resort to the provisional gateway definition?  Can we?
	if ($forceLocalAddrResolve)
	{
		if (not defined $gateway->{hostname} and defined $provisionalHostname)
		{
			$gateway->{hostname} = $provisionalHostname;
			$gateway->{my_ip} = $ip_addr;
			$gateway->{my_ip_and_local_cluster}->{my_ip} = $ip_addr;
		}
	}
	
	return $gateway;	
}

# get CP osRouting information from statuses.xml and collect them under gateway->{osRouting}
sub _getStaticRoutes {
	
	my ($XPCST, $gateway) = @_;
	
	my $hostname = $gateway->{hostname};
	# find and parse application_obj and osRoutingInfo for a given hostname 
	foreach my $libxmlobject ($XPCST->findnodes('/statuses/object')){
		my $appObj = $libxmlobject->findvalue('./application_obj');
		$appObj= _substring_after($appObj,"referenced_object: ");
		# find the matching hostname cpshared_application
		if ($appObj =~ /cpshared_application_$hostname/){
			foreach my $libxmlOsRoute ($libxmlobject->findnodes('./osStatus/object/osRoutingInfo/object/osRoutingEntry/object')) {
				# create an empty bucket for routing
				my $osRouting ={};	
				$osRouting->{osRoutingDest} = $libxmlOsRoute->findvalue('./osRoutingDest');
				$osRouting->{osRoutingGateWay} = $libxmlOsRoute->findvalue('./osRoutingGateWay');
				$osRouting->{osRoutingIntrf} = $libxmlOsRoute->findvalue('./osRoutingIntrf');
				$osRouting->{osRoutingMask} = $libxmlOsRoute->findvalue('./osRoutingMask');
				# push routing information under osRouting bucket 
				push (@{$gateway->{osRouting}}, $osRouting);
			} # end foreach libxmlOsRoute  		
		} # end of if find matching hostname 
	} #end foreach libxmlobject	
}

# get CP MemoryStat-uses and collect them under gateway->{MemoryStat}
sub _getMemoryStat {
	my ($XPCST,$gateway) = @_;
	
	my $hostname = $gateway->{hostname};
	# find and parse application_obj and osRoutingInfo for a given hostname 
	foreach my $libxmlobject ($XPCST->findnodes('/statuses/object')){
		my $appObj = $libxmlobject->findvalue('./application_obj');
		$appObj= _substring_after($appObj,"referenced_object: ");
		# find the matching hostname cpshared_application
		if ($appObj =~ /cpshared_application_$hostname/){
			foreach my $libxmlMemoryStat ($libxmlobject->findnodes('./osStatus/object/osPerfStat/object/MemoryStat/object')) {
				# create an empty bucket for MemoryStat
				my $MemoryStat ={};	
				$MemoryStat->{totalVirtMem} = $libxmlMemoryStat->findvalue('./total_virtual_mem');
				push (@{$gateway->{MemoryStat}}, $MemoryStat);
			} # end foreach	
		} # end of if find matching hostname
	} # end of foreach libxmlobject	
}

sub _getModel {
	my ($XPCNO,$XPCST, $gateway) = @_;
	
	my $version;
	my $hostname = $gateway->{hostname};
	# need to use gateway cluster hostname if cluster member
	if ($gateway->{isClusterMember}){
		if ($gateway->{gateway_clusters} > 0){
			$hostname = $gateway->{gateway_clusters}[0];
		}
	}
	
	foreach my $libxmlobject ($XPCNO->findnodes('/network_objects/object')){
		my $network_object_name = $libxmlobject->findvalue('./object_name');
		$version = $libxmlobject->findvalue('./svn_version_name');
		if ($network_object_name eq $hostname){
			$version = $libxmlobject->findvalue('./svn_version_name');
			if ($version ne ""){
				return $version;
			}
		}
	}
	# if version still not found use statuses
	# find and parse application_obj and osRoutingInfo for a given hostname 
	foreach my $libxmlobject ($XPCST->findnodes('/statuses/object')){
		my $appObj = $libxmlobject->findvalue('./application_obj');
		$appObj= _substring_after($appObj,"referenced_object: ");
		# find the matching hostname cpshared_application
		if ($appObj =~ /cpshared_application_$hostname/){
			$version = $libxmlobject->findvalue('./osStatus/object/CPsharedInfo/object/CPsharedVersion');
			last;
		}	
	}
	return $version; 	
}

# List of exceptions names for positional properties
my %AlternateServicePositionalPropertyMappings = (
		# R75.40 Gaia
		"r_access_enable_p" => "raccessenable", 
		#"allow_gw_local_services_p" => "allow_gw_local_services"
);

# Function checks properties enablement.
# It returns a hash variable containing the properties relevant for implied_rules
# A valid property has to be enabled (i.e property = True) and 
# has an associated positional property (i.e property_p = '<first> <before last> <last>')  
sub _getPropertiesEnablement {

	my ($XPCPR, $property) = @_;
	my $properties = {};
	
		foreach my $libxmlobject($XPCPR->findnodes('/properties/object')){
			my $objname = $libxmlobject->findvalue('./object_name');
			if ($objname =~ /firewall_properties/){
				if (defined $property){
				# check if $property is available. The found property is a good candidate to build unordered rules 
					my $enProperty = undef;  	
					$enProperty = $libxmlobject->findvalue("./$property");	
					if (defined ($enProperty) and ($enProperty eq 'True')){
						$properties->{$property}->{enabled} = 1;
					}			
				}else {
					# find all properties that have a positional property
					foreach my $posProperty ($libxmlobject->findnodes("./*[contains(local-name(), '_p')]")) {
						my $propertyName = undef;
						my $valProperty = undef;
						# node's name
						my $nodeName = $posProperty->nodeName;
						if ($nodeName =~ /(^\w+)_enable_p$/) {
							# check for special property names
							my $property = $AlternateServicePositionalPropertyMappings{$nodeName};
							if (defined ($property) and ($property ne "") ){
								$propertyName = $property;
							}else {
								$propertyName = $1;
							}
						} elsif ($nodeName =~ /(^\w+)enable_p$/){
							$propertyName = $1;
						} elsif ($nodeName =~ /(^\w+)_p$/){
							$propertyName = $1; 
						# could not find a matching property corresponding to the positional property		
						} else {
							$propertyName = $AlternateServicePositionalPropertyMappings{$nodeName};
						} 
						
						if (defined ($propertyName)){
							# alternative possible property name
							my $altPropertyName = $propertyName."enable";
							# search for properties associated with the positional properties 
							my $enProperty = undef;
							$valProperty = $libxmlobject->findvalue("./$altPropertyName");
							if ((defined $valProperty) and ($valProperty eq 'True')){
								$enProperty = $altPropertyName; 
							}
							$valProperty = $libxmlobject->findvalue("./$propertyName");
							if ((defined $valProperty) and ($valProperty eq 'True')) {
								$enProperty = $propertyName;	
							}		
							# check if property is enabled
							if (defined ($enProperty)){
								$properties->{$enProperty}->{enabled} = 1;
								# positional's node value
								$properties->{$enProperty}->{position} = $posProperty->findvalue(".");
							}
						} #end of if defined propertyName		
						
					} #end foreach posProperty
				} # end else	
			} # end 
		} # end foreach $XPCPR
		
	return $properties;	
}

# process properties and builds $impliedFeSets that is going to be feed 
# to generate_filter_list 	
 sub _processProperties{
 	my ($XPCPR, $prop,$prop_with_pos) = @_;
 	
 	my $impliedFeSets = {};
 	# parse property names from impliedFeSets file and compare them
 	# with the properties that are enabled and have a defined positional property
 	for my $propName (keys %{$prop}) {
 		my $struct =  ${$prop}{$propName};
       	    	
    	for my $posProp (keys %{$prop_with_pos}) {
    		if  ($propName eq $posProp ){ 
	    		# push-in the ordered rules
	    		my $position = ${$prop_with_pos}{$posProp}{position};
	    		foreach my $valofStruct (@{$struct}){
	    			push (@{$impliedFeSets->{$position}}, $valofStruct);	
	    		}
    		} 	 	
    	} #end for $posProp
    	
    	my $property = _getPropertiesEnablement($XPCPR,$propName);
    	if (($property->{$propName}->{enabled}) and ( not defined(${$prop_with_pos}{$propName}{position})) ){
    		foreach my $valofStruct (@{$struct}){
    			push (@{$impliedFeSets->{'unordered'}}, $valofStruct);
    		}
    	} #end if $property->{$propName}->{enabled}
        	
    }#end for $propName
        
 	return $impliedFeSets;	
 	
 }

sub _prune_global_rules {
	my $fPruneList = shift;
	my $serviceGroups = shift;
	
	if (not defined $fPruneList) {
		return;
	}
	
	for (my $pruneIdx = scalar(@{$fPruneList} - 1); $pruneIdx >= 0; $pruneIdx--) {
		my $fPruneEnt = $fPruneList->[$pruneIdx];
		my $anyAddrEnt = undef;
		
		# (re) assign process order
		$fPruneEnt->{processOrder} = $pruneIdx + 1;
		
		# Check for a source, dest of any, if it doesn't match, we can move to the next 
		# filter entry	
		foreach my $xAddrList ($fPruneEnt->{sourceIpAddr}, $fPruneEnt->{destinationIpAddr}) {
			for (@{$xAddrList}) {  # check for not qualified by object group ref and address is any addr
				if ( (not defined $_->{objectGroupReference}) and
					($_->{network}->{address} eq get_any_address_address()))	{
					$anyAddrEnt = $_;
					last;
				}
			}
			
			if (defined $anyAddrEnt) {
				last;
			}									
		}
		# did we match in src or dest?
		if (not defined $anyAddrEnt) {
			# yes.  There's no matching wilcard-any addr, so let's leave this alone and move on to the next entry
			next; 
		}
		
		# if protocol for src or dest isn't specified, then delete
		if ( (not defined $fPruneEnt->{protocol}) and
			(not defined $fPruneEnt->{destinationService}) and
			(not defined $fPruneEnt->{sourceService})) {
				# delete at this list entry.  Be careful to delete from the array, not the
				# entry itself, which would just assign an undef to this slot.
				splice (@{$fPruneList}, $pruneIdx, 1);				
		}		
		
		foreach my $destService ($fPruneEnt->{destinationService}){
			foreach my $destServiceEntry (@{$destService}){
				if (defined $destServiceEntry->{objectGroupReference}){
					my $destServiceRef = $destServiceEntry->{objectGroupReference};
					unless (defined $serviceGroups->{$destServiceRef}){
						splice (@{$fPruneList}, $pruneIdx, 1);
						$LOGGER->debug("_prune_global_rules; Implied rule with missing service being DROPPED"); 
						last;
					}
				}
			}
		}
	}
}
 
# Function parses services.xml and search for src_port definitions
# if it is available it will push-in a new serviceGroup__src structure   
sub _createSrcPortServiceGroup{
 	
 	my ($id, $XPCSRV,$in, $servObjList) = @_;
 	
 	foreach my $libxmlSRVObj ($XPCSRV->findnodes('/services/object')){
		my $libxmlObjName = $libxmlSRVObj->findvalue('./object_name');
		my $portExpression;
		if ($id eq $libxmlObjName){
			my $srcPort = $libxmlSRVObj->findvalue('./src_port');
			my $protocol = $libxmlSRVObj->findvalue('./type');
			if ((defined $srcPort) and ($srcPort ne "")){
					my $servObj = { id => $id . '__src' };
					if ( $srcPort =~ /^\d+$/ ){
						$portExpression = {
							port     => $srcPort,
							operator => 'eq',
							protocol => lc ($protocol),
						};
						$servObj->{portExpression} = $portExpression;
					}elsif ( $srcPort =~ /^(\d+)\s?-\s?(\d+)$/ )
					{
						$portExpression = {
							portStart => $1,
							portEnd   => $2,
							protocol  => lc($protocol),
						};
						$servObj->{portRange} = $portExpression;
					}		
					push (@{$servObjList}, $servObj);
					# make sure we know we had a source mapping here.		
					$in->{portServiceGroups}->{$id . '__src'} = $portExpression;							
			} #end if defined srcPort 
		} #end if $id			
 	} #end foreach libxmlSRVObj			 
 } #end sub _createSrcPortServiceGroup
 
 sub _checkServiceGroup {
 	
 	my ($in, $service, $filterEntry) = @_;
 	
 	if (defined $in->{portServiceGroups}->{$service.'__src'}) {
 		push (@{$filterEntry->{sourceService}}, $service.'__src');
 	}
 	if (defined $in->{portServiceGroups}->{$service}) {
 		push (@{$filterEntry->{destinationService}}, $service);
 	}
 }
 
# parse network_objects to check if NAT is defined.
# looking after firewalling_obj elements and store the NAT elements of the network objects
sub _checkNAT {
	
	my ($in, $XPCNO, $XPCFW, $gateway) = @_;
	$in->{isNAT} = 0;
	
	# check if automatic NAT is defined 
 	#loop through network_objects to find a node that has a different object name than CP_default_Office_Mode_addresses_pool
	foreach my $libxmlobject ($XPCNO->findnodes('/network_objects/object')){
		my $object_name = $libxmlobject->findvalue('./object_name');
		if ($object_name ne 'CP_default_Office_Mode_addresses_pool'){		
			foreach my $libxmlNAT($libxmlobject->findnodes('./NAT')) {
				foreach my $libxmlNATObj ($libxmlNAT->findnodes('./object')){
					# store NAT settings for future references 
					# NAT rules are installed on the Gateway specified in the_firewalling_obj
					$in->{NAT}->{$object_name}->{the_firewalling_obj} = $libxmlNATObj->findvalue('./the_firewalling_obj');
					# store NAT translation method (i.e static or hide (dynamic) )
					$in->{NAT}->{$object_name}->{netobj_adtr_method} = $libxmlNATObj->findvalue('./netobj_adtr_method');
					# valid address name
					$in->{NAT}->{$object_name}->{valid_addr_name} = $libxmlNATObj->findvalue('./valid_addr_name');
					# valid NAT IP address
					$in->{NAT}->{$object_name}->{valid_ipaddr} = $libxmlNATObj->findvalue('./valid_ipaddr');
					
					# found a network_objects that has a reference to the analyzed gateway
					# NAT rules can be installed on a specific gateway or on all available managed gateways
					if( ($in->{isNAT} == 0 and defined($in->{NAT}->{$object_name}->{the_firewalling_obj}) ) 
					and (	($in->{NAT}->{$object_name}->{the_firewalling_obj} =~ /All/) or 
							($in->{NAT}->{$object_name}->{the_firewalling_obj} =~ /Gateways/) or  
							(defined ($gateway->{hostname}) and  ($in->{NAT}->{$object_name}->{the_firewalling_obj} =~ /$gateway->{hostname}/)))){
						$in->{NAT}->{installed_on_gateway} = $gateway;						
						$in->{isNAT} = 1;
					}
				}#end foreach $libxmlNATObj
			} #end foreach $libxmlNAT
						
		}  
	} #end foreach libxmlobject XPCNO
	
	# check if manual NAT is defined
	foreach my $libxmlobject ($XPCFW->findnodes('/fw_policies/object/rule_adtr/rule_adtr/object')){
		foreach my $libxmlinstall ($libxmlobject->findnodes('./install')) {
			my $install = $libxmlinstall->findvalue('./install');
			
			if ( ($in->{isNAT} == 0 and defined($install)) 
				and (	($install =~ /Any/ or $install =~ /Gateways/) or 
				(defined $gateway->{hostname} and $install =~ /$gateway->{hostname}/)  ) ){
					$in->{NAT}->{installed_on_gateway} = $gateway;						
					$in->{isNAT} = 1;			
			} 
			
		} # end foreach install	
	} #end foreach libxmlobject XPCNO
}


sub parse_network_groups {
	my $out     = shift;
	my $XPC     = shift;
	my $ip_addr = shift;
	my $in      = shift;
	my $XPCNO   = shift;
	my $LOGGER  = shift;

	# find and print networkGroup node
	foreach my $libxmlNG (
		$XPC->findnodes(
			'//ngns:networkGroup')) {
		$out->open_element('networkGroup');
			
			my $id = $XPC->findvalue('.//ngns:id',$libxmlNG);
			$out->print_element('id', $id);
			
			# find host elements			
			my $host = $XPC->findvalue('.//ngns:host', $libxmlNG);
			if (defined ($host) && ($host ne "")){
				$out->print_element('host',$host);
				# additionally store the networkGroup elements for later crosschecking (i.e. parse_rules)
				$in->{networkGroups}->{$id}={ host => $host};
				if ($host eq $ip_addr){
					my $gateway = _getGateway($XPCNO, $host, 1);
					if ( (exists $gateway->{my_ip}) and ($gateway->{my_ip} eq $ip_addr) ){
					# build networkGroups->{my_ip} element. It is used for later crosschecking (i.e. parse_rules) 
						$in->{networkGroups}->{"my_ip"} = $gateway->{my_ip};
					}
					if ( (exists $gateway->{my_ip_and_local_cluster}->{my_ip}) and ($gateway->{my_ip_and_local_cluster}->{my_ip} eq $ip_addr) ){
					# build networkGroups->{my_ip} element. It is used for later crosschecking (i.e. parse_rules) 
						$in->{networkGroups}->{"my_ip_and_local_cluster"} = $gateway->{my_ip_and_local_cluster}->{my_ip};
					}	
				}
					
			} # end of if defined host

			# find subnet elements
			foreach my $libxmlsub ($XPC->findnodes('.//ngns:subnet',$libxmlNG)) {
				my $xmlinvar = $libxmlsub->toString;
				my $ref = XMLin($xmlinvar, ForceArray =>1);
				$out->print_element('subnet', $ref);	
				# additionally store the networkGroup elements for later crosschecking (i.e. parse_rules)
				$in->{networkGroups}->{$id}={ subnet => $ref};
			}
			
			# find objectGroupReference-s
			foreach my $libxmlsub ($XPC->findnodes('.//ngns:objectGroupReference',$libxmlNG)) {
				my $xmlinvar = $libxmlsub->toString;
				if ($xmlinvar eq "<objectGroupReference/>"){
					$LOGGER->debug("parse_CPSMS_object_groups; Empty object group reference being DROPPED"); 
					next;
				}
				my $ref = XMLin($xmlinvar, ForceArray =>1);
				$out->print_element('objectGroupReference', $ref);
				
				# additionally store the networkGroup elements for later crosschecking (i.e. parse_rules)
				$in->{networkGroups}->{$id}={ objectGroupReference => $ref};
			}
			
			# find IP ranges (or IP pools)
			foreach my $libxmlsub ($XPC->findnodes('.//ngns:range',$libxmlNG)) {
				my $xmlinvar = $libxmlsub->toString;
				my $ref = XMLin($xmlinvar, ForceArray =>1);
				$out->print_element('range', $ref);
				# additionally store the networkGroup elements for later crosschecking (i.e. parse_rules)
				$in->{networkGroups}->{$id}={ range => $ref};
			}
			
		$out->close_element('networkGroup');	
	} #end $libxmlNG
	
	# add extra networkGroup nodes i.e my_ip, my_ip_and_local_cluster
	if (exists $in->{networkGroups}->{"my_ip"}) {
		$out->open_element('networkGroup');
			$out->print_element('id','my_ip');
			$out->print_element('host', $in->{networkGroups}->{"my_ip"});
		$out->close_element('networkGroup');	
	}
	
	if (exists $in->{networkGroups}->{"my_ip_and_local_cluster"}) {
		$out->open_element('networkGroup');
			$out->print_element('id','my_ip_and_local_cluster');
			$out->print_element('host', $in->{networkGroups}->{"my_ip_and_local_cluster"});
		$out->close_element('networkGroup');
		
	}
}

sub parse_protocol_groups {
	my $out = shift;
	my $XPC = shift;

	foreach my $libxmlPG (
		$XPC->findnodes(
			'//psgns:protocolGroup')) {
		my $id = $XPC->findvalue('.//psgns:id', $libxmlPG);
	
		foreach my $protocol ($XPC->findnodes('.//psgns:protocol',$libxmlPG)){
			if (defined ($protocol) && ($protocol) ne ""){
				my $xmlinvar = $protocol->toString;
				next if ($xmlinvar =~ /<protocol\/>/);
				my $ref = XMLin($xmlinvar, ForceArray =>1);
				$out->open_element('protocolGroup');
					$out->print_element('id', $id);
					$out->print_element('protocol', $ref);
				$out->close_element('protocolGroup');
			}	
		} # end foreach protocol
	
		my $objPGR = 0;
		foreach my $objectGroupReference($XPC->findnodes('.//psgns:objectGroupReference',$libxmlPG)){
			if (defined ($objectGroupReference) && ($objectGroupReference ne "")){
				my $xmlinvar = $objectGroupReference->toString;
				my $ref = XMLin($xmlinvar, ForceArray =>1);
				if ($objPGR == 0){
					$out->open_element('protocolGroup');
					$out->print_element('id', $id);
				}	
				$out->print_element('objectGroupReference', $ref);
				$objPGR = 1;	
			} #end if		
		} # end foreach protocolGroup objectGroupReference
		if ($objPGR == 1){
			$out->close_element('protocolGroup');
		}				
	} # end foreach protocolGroup
}

sub parse_service_groups {
	my $out    = shift;
	my $XPC    = shift;
	my $XPCSRV = shift;
	my $in     = shift;

	foreach my $libxmlSG (
		$XPC->findnodes(
			'//psgns:serviceGroup')) {
		my $id = $XPC->findvalue('.//psgns:id', $libxmlSG);
		my $servObjList = [];
		
		# check and define <serviceGroup>_src if there is a src_port associated with this $id
		# the newly created serviceGroup_src will be stored within $servObjList		
		_createSrcPortServiceGroup($id,$XPCSRV,$in,$servObjList);
		if (defined $in->{portServiceGroups}->{$id . '__src'}) {
			# create a new serviceGroup in case there is defined src_port for this $id
			foreach my $exp(@{$servObjList}){
				$out->open_element('serviceGroup');
					$out->print_element('id',$exp->{id});	
				for my $keyname (keys %{$exp}) {									
	    			if  ($keyname eq "portRange" ){ 
						$out->print_element('portRange',$exp->{portRange});
	    			}
	    			if ($keyname eq "portExpression"){
	    				$out->print_element('portExpression',$exp->{portExpression});
	    			}
				} #end for $servObjList
				$out->close_element('serviceGroup');				
			} #end foreach $exp
		} # end if in->{portServiceGroups}->{id__src}
												
		# flag to indicate portExpression availability
		my $flagPE = 0;
		foreach my $libxmlPE($XPC->findnodes('.//psgns:portExpression',$libxmlSG )){
			if (defined ($libxmlPE) && ($libxmlPE) ne ""){
				$flagPE = 1;
				my $port = $XPC->findvalue('.//psgns:port',$libxmlPE);
				next if(defined ($port) && ($port eq "NaN"));
				$out->open_element('serviceGroup');
					$out->print_element('id',$id);
					$out->open_element('portExpression');
						$out->print_element('port', $port);	
						my $operator = $XPC->findvalue('.//psgns:operator',$libxmlPE);
						$out->print_element('operator',$operator);
						my $protocol = $XPC->findvalue('.//psgns:protocol',$libxmlPE);
						$out->print_element('protocol',$protocol);
					$out->close_element('portExpression');
				$out->close_element('serviceGroup');
				# additionally store the serviceGroup element for later crosschecking (i.e. parse_rules)
				$in->{portServiceGroups}->{$id} = {
					portExpresion =>{
							operator => $operator,
							port => $port,
							protocol => $protocol,
					},	
				}				
			} 
			#there is no libxmlPE defined
			else {
				$flagPE = 0;
			}
		} # end foreach portExpression
		
		# flag to indicate portRange availability
		my $flagPR = 0;
		foreach my $libxmlPR($XPC->findnodes('.//psgns:portRange',$libxmlSG )){
			if (defined ($libxmlPR) && ($libxmlPR) ne ""){
				$flagPR = 1;
				$out->open_element('serviceGroup');
					$out->print_element('id',$id);
					$out->open_element('portRange');
						my $portStart = $XPC->findvalue('.//psgns:portStart',$libxmlPR);
						$out->print_element('portStart', $portStart);	
						my $portEnd = $XPC->findvalue('.//psgns:portEnd',$libxmlPR);
						$out->print_element('portEnd',$portEnd);
						my $protocol = $XPC->findvalue('.//psgns:protocol',$libxmlPR);
						$out->print_element('protocol',$protocol);
					$out->close_element('portRange');
				$out->close_element('serviceGroup');
				# additionally store the serviceGroup element for later crosschecking (i.e. parse_rules)
				$in->{portServiceGroups}->{$id} = {
					portRange =>{
							portStart => $portStart,
							portEnd => $portEnd,
							protocol => $protocol,
					},	
				}
				
			} #end if defined libxmlPR
			# there is no libxmlPR defined
			else{
				$flagPR = 0;
			}
		} # end foreach portRange
		
		# parse serviceGroup containing objectGroupReferences
		my $serviceGroupReference = {};
		# flag to indicate objectGroupReference availability
		my $flagGR = 0;
		foreach my $objectGroupReference($XPC->findnodes('.//psgns:objectGroupReference',$libxmlSG)){
			if (defined ($objectGroupReference) && ($objectGroupReference ne "")){
				#found a referenced objectGroupReference
				$flagGR = 1;
				
				my $xmlinvar = $objectGroupReference->toString;
				my $ref = XMLin($xmlinvar, ForceArray =>1);
						
				#check to see if exists a src_port with ref = id
				if (exists $in->{portServiceGroups}->{$ref . '__src'}) {
					push (@{$serviceGroupReference->{$id.'__src'}}, $ref.'__src');
				} 
				# simply push-in the found objectGroup reference  
				# these references will be referred later as destination group services	
				push (@{$serviceGroupReference->{$id}}, $ref); 					
								
			}
			# there is no objectGroupReference defined
			else{
				$flagGR = 0;
			}		
		} # end foreach serviceGroup objectGroupReference
		 	
		#print objectGroupReference 
		if ( (exists $serviceGroupReference->{$id}) or (exists $serviceGroupReference->{$id.'__src'}) ){
			if (exists $serviceGroupReference->{$id}) {
				$out->open_element('serviceGroup');
				$out->print_element('id', $id);
					
				foreach my $groupReference (@{$serviceGroupReference->{$id}}){
					if ((defined $groupReference) and ($groupReference ne "")) {
						$out->print_element('objectGroupReference', $groupReference);
					}		
				} # end foreach groupReference	
				$out->close_element('serviceGroup');
				
				# additionally store the serviceGroup element for later crosschecking (i.e. parse_rules)
				$in->{portServiceGroups}->{$id} = {
					objectGroupReference =>  $serviceGroupReference->{$id}
				}
			} #end exists $serviceGroupReference->{$id}
			
			if (exists $serviceGroupReference->{$id.'__src'}) {
				$out->open_element('serviceGroup');
				$out->print_element('id', $id.'__src');
				
				foreach my $srcGroupReference (@{$serviceGroupReference->{$id.'__src'}}){
					if ((defined $srcGroupReference) and ($srcGroupReference ne "")) {
						$out->print_element('objectGroupReference', $srcGroupReference);
					}		
				} # end foreach srcGroupReference
				$out->close_element('serviceGroup');
				
				# additionally store the serviceGroup element for later crosschecking (i.e. parse_rules)
				$in->{portServiceGroups}->{$id .'__src'} = {
					objectGroupReference =>  $serviceGroupReference->{$id.'__src'}
				}
			} #end exists $serviceGroupReference->{$id.'__src'}		
		} # end if exists $serviceGroupReference
			
		# check for serviceGroups that contains only id element
		if (($flagGR == 0) and ($flagPR == 0) and ($flagPE == 0)){
			foreach my $libxmlSRVObj ($XPCSRV->findnodes('/services/object')){
				 my $libxmlObjName = $libxmlSRVObj->findvalue('./object_name');
	
					if ($id eq $libxmlObjName){
						my $type = $libxmlSRVObj->findvalue('./type');
						my $protocol = $libxmlSRVObj->findvalue('./protocol');
						# translate Check Point protocol's name to SED's name format
						if (defined($type) and ($type =~ /icmp/i)){
							$protocol = "icmp"; 
						} 
						if (defined ($type) and ($type =~ /dcerpc/i)){
							$protocol = "ms-rpc";
						}
						if (defined ($type) and ($type =~/other/i) and defined($protocol) and ($protocol == 55)){
							$protocol = "mobile-ip";
						}
						# get protocol definition
						if (defined ($protocol)){
							my $tempExp = get_port_expression($protocol);
							$out->open_element('serviceGroup');
								$out->print_element('id',$id);
								foreach my $exp(@{$tempExp}){
									for my $keyname (keys %{$exp}) {
    									if  ($keyname eq "portRange" ){ 
											$out->print_element('portRange',$exp->{portRange});
											# additionally store the serviceGroup element for later crosschecking (i.e. parse_rules)
											$in->{portServiceGroups}->{$id} = {
												portRange => ${$exp}{$keyname}	
											}
    									}
    									if ($keyname eq "portExpression"){
    										$out->print_element('portExpression',$exp->{portExpression});
    										# additionally store the serviceGroup element for later crosschecking (i.e. parse_rules)
											$in->{portServiceGroups}->{$id} = {
												portExpression => ${$exp}{$keyname}
											}
    									}		
									}		
								} #end foreach $tempExp							
							$out->close_element('serviceGroup');
						} # end if defined protocol
							
					} # end if $id eq $libxmlObjName
			} # end $libxmlSRVObj
		} # end if flag conditions				
	} # end foreach serviceGroup
}
