#!/usr/bin/perl
use strict;
use warnings;
use XML::Simple;
use ZipTie::Logger;
use ZipTie::Client;
use ZipTie::Adapters::CheckPoint::CPSMS;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

# Set Ziptie Env
setEnv();
my $ziptieClient = ZipTie::Client->new();

my ($cpsms_ip_addr) = @ARGV;
my $nwo_location = "/tmp/nwo/".$cpsms_ip_addr;
my $cpt_dir = "../../../../../cptool/cpsms";

$LOGGER->debug("Initiating Discovery from $cpsms_ip_addr server.");

# This is to get the credential based on the ip addr
my $cpt = getCredential($ziptieClient, $cpsms_ip_addr, $nwo_location);

my $cptPkgPath = $cpt->{pkgPath};
my $isSimulation;
open FILE, "<$cptPkgPath/$cpt_dir/cpmi.conf" or $LOGGER->debug("File is not available: $cptPkgPath/$cpt_dir/cpmi.conf");
        foreach my $lineInFile (<FILE>)
        {
                if ($lineInFile =~ m/^adapters.cpsms.simulation.filesource=(\/.*)/)
        		{
                        $isSimulation = 1;
        		}
        }
if ($isSimulation){
        $LOGGER->debug("Attempting to run simulated Discovery...");
} else { 
		# Get the network objects by calling cpmi client
        ZipTie::Adapters::CheckPoint::CPSMS::_callCPT($cpt, "network_objects", $cpt_dir);
}

# Add\Update devices that are discovered
discoverDevices($ziptieClient, $cpsms_ip_addr, $nwo_location);

sub getCredential
{
	my ($ziptieClient, $cpsms_ip_addr, $nwo_location) = @_;
	my $credP = $ziptieClient->credentials();
	my @credentials = $credP->getCredentialSetsByIpAddress(ipAddress => $cpsms_ip_addr);
	my @protocol = $credP->getProtocolConfigByIpAddress(ipAddress => $cpsms_ip_addr);
	my $credential;
	my $isCorrectCred = 0;
	my %cred;
	my $adaptorProtocol = "CPSMS";
	my $adaptorPort;

	foreach (@credentials)
	{
		$credential = $_;
		if ($credential->{name} ne "default set")
		{
			$isCorrectCred = 1;
			last;
		}
	}

	foreach (@{$protocol[0]->{protocols}})
	{
		if ($_->{name} eq $adaptorProtocol)
		{
			$adaptorPort = $_->{port};
			last;
		}
	}

	if ($isCorrectCred && defined $adaptorPort) 
	{
		foreach (@{$credential->{credentials}})
		{
			$cred{$_->{name}}=$_->{value};
		}
	}
	else
	{
		$LOGGER->fatal_error_code("Credential/Port undefined for $cpsms_ip_addr. Please make sure correct credential set is entered in a defined Network Groups with the IP address of the CPSMS and is not falling back to default set in the Credentials pane. Furthermore only one credential set and CPSMS IP address can be added for a defined Network Groups.");
	}

	# Prevent any uninitialized value in concatenation errors when building the connection document.
	foreach my $credType ('snmpAuthPassword', 'snmpUsername', 'roCommunityString', 'username',
		'opsecSslca', 'opsecEntitySicName', 'snmpPrivPassword', 'enableUsername', 'password',
		'opsecApplicationObjectSicName', 'enablePassword')
	{
		if (!defined($cred{$credType}))
		{
			$cred{$credType} = '';
		}
	}

	my $connection_xml = <<__CONNECTIONXML__;
		<connectionPath host=\"$cpsms_ip_addr\" managementServer=\"$cpsms_ip_addr\">
        	<credentials>
        		<credential name=\"snmpAuthPassword\" value=\"$cred{snmpAuthPassword}\"/>
            	<credential name=\"snmpUsername\" value=\"$cred{snmpUsername}\"/>
            	<credential name=\"roCommunityString\" value=\"$cred{roCommunityString}\"/>
            	<credential name=\"username\" value=\"$cred{username}\"/>
            	<credential name=\"opsecSslca\" value=\"$cred{opsecSslca}\"/>
            	<credential name=\"opsecEntitySicName\" value=\"$cred{opsecEntitySicName}\"/>
            	<credential name=\"snmpPrivPassword\" value=\"$cred{snmpPrivPassword}\"/>
            	<credential name=\"enableUsername\" value=\"$cred{enableUsername}\"/>
            	<credential name=\"password\" value=\"$cred{password}\"/>
            	<credential name=\"opsecApplicationObjectSicName\" value=\"$cred{opsecApplicationObjectSicName}\"/>
            	<credential name=\"enablePassword\" value=\"$cred{enablePassword}\"/>
        	</credentials>
        	<protocols>
            	<protocol name=\"$adaptorProtocol\" port=\"$adaptorPort\">
					<properties/>
            	</protocol>
        	</protocols>
        	<fileServers/>
    	</connectionPath>
__CONNECTIONXML__

	my $connectionPath = ZipTie::ConnectionPath::from_xml($connection_xml);
	my $cpt = ZipTie::Adapters::CheckPoint::CPSMS::CPTool->new(connectionPath => $connectionPath, out_dir => $nwo_location, managementServerIp => $cpsms_ip_addr, ip_addr => $cpsms_ip_addr);

	return $cpt;
}

sub discoverDevices
{
	my ($ziptieClient, $cpsms_ip_addr, $nwo_location) = @_;
	$nwo_location = $nwo_location."/logs/".$cpsms_ip_addr."/".$cpsms_ip_addr."/network_objects.xml";
	my $adapter = 'ZipTie::Adapters::CheckPoint::CPSMS';
	my $deviceService = $ziptieClient->devices();
	my $deviceSearchService = $ziptieClient->devicesearch();
	my $xmlin = new XML::Simple;
	my $data = $xmlin->XMLin($nwo_location);
	my $devices = get_devices($deviceSearchService);

	foreach my $e (@{$data->{object}})
	{
		if (defined $e->{ipaddr} && defined $e->{type} && ( ($e->{type} eq "gateway") or ($e->{type} eq "gateway_cluster") or ($e->{type} eq "cluster_member") ) )
		{
			# Skip gateway node objects -> these are not Check Point firewalls
			next if ($e->{type} eq "gateway" && (defined $e->{firewall} && $e->{firewall} eq "not-installed")); 
			my $device = $devices->{$e->{ipaddr}};
			if (defined $device)
			{
				if ($device->{managementServerIp} ne $cpsms_ip_addr)
				{
					$device->{managementServerIp} = $cpsms_ip_addr;
					$deviceService->updateDevice(ipAddress => $e->{ipaddr}, managedNetwork => 'Default', device => $device);
					$LOGGER->debug("Updated $e->{ipaddr} node.");
					print("$e->{ipaddr}, Node updated.\n");
				}
				else
				{
					$LOGGER->debug("Skipped $e->{ipaddr} node.");
					print("$e->{ipaddr}, Node skipped as no properties have changed.\n");
				}
			}
			else
			{
				$deviceService->createManagedDevice(ipAddress => $e->{ipaddr}, adapterId => $adapter, managedNetwork => 'Default', managementServerIp => $cpsms_ip_addr);
				$LOGGER->debug("Added $e->{ipaddr} node.");
				print("$e->{ipaddr}, Node added.\n");
			}
		}
	}
}

sub get_devices
{
	my ($deviceSearchService) = @_;
	my %deviceHash;
	my $pageSize = 500;
	my $offset = 0;
	my $total = undef;

	while (!defined($total) || ($offset < $total))
	{
		my $page = $deviceSearchService->search(
				scheme => 'ipAddress', query => '',
				pageData => {pageSize => $pageSize, offset => $offset}
			);
		$offset += $pageSize;
		
		if ( (ref $page->{devices} eq 'HASH')  and (defined($page->{devices}->{ipAddress})) ){
			$deviceHash{$page->{devices}->{ipAddress}} = $page->{devices};	
		} elsif ( ref $page->{devices} eq 'ARRAY') {
			foreach my $device (@{$page->{devices}}) {
				$deviceHash{$device->{ipAddress}} = $device;
			}	
		} 
		
		# Total is only returned for the first search
		if (!defined($total))
		{
			$total = $page->{total};
		}
	}
	return \%deviceHash;
}

sub setEnv
{
	my $pwd = `pwd`;
	$pwd =~ s/\n//;

	chdir('/usr/share/ziptie-server');
	# Need to add psftp to the path
	my $lsCommand = 'ls -d /usr/share/ziptie-server/core/org.ziptie.adapters_*' .
        '/bin/linux';
	my $psftpPath = `$lsCommand`;
	$psftpPath =~ s/\n//;
	$ENV{'PATH'} .= ":$psftpPath";

	# ZipTie user/pass
	my $tomcatProperties = '/opt/qradar/conf/tomcat.frameworks.properties';

	# We need to perform this check so the script works with QRM 7.1.0.
	if (-f '/opt/qradar/conf/tomcat-rm.frameworks.properties')
	{
        $tomcatProperties = '/opt/qradar/conf/tomcat-rm.frameworks.properties';
	}

	# Obtain CSM username
	my $csmUser = `grep ziptie.auth.user $tomcatProperties`;
	if (!($csmUser =~ /^ziptie\.auth\.user=(.+)$/))
	{
        print "ERROR: Failed to obtain CSM username.\n";
        exit(1);
	}
	$csmUser = $1;

	# Obtain CSM password
	my $csmPass = `grep ziptie.auth.pass $tomcatProperties`;
	if (!($csmPass =~ /^ziptie\.auth\.pass=(.+)$/))
	{
        print "ERROR: Failed to obtain CSM password.\n";
        exit(1);
	}
	$csmPass = $1;

	# Need to set ZIPTIE_AUTHENTICATION string
	$ENV{'ZIPTIE_AUTHENTICATION'} = "https://$csmUser:$csmPass\@127.0.0.1:8082/" .
        'server/';

	# Add ZipTie core to PERL5LIB env variable
	$lsCommand = 'ls -d /usr/share/ziptie-server/core/org.ziptie.adapters_*' .
        '/scripts';
	my $corePath = `$lsCommand`;
	$corePath =~ s/\n//;
	$ENV{'PERL5LIB'} = $corePath;

	# Add ZipTie common to PERL5LIB env variable
	$lsCommand = 'ls -d /usr/share/ziptie-server/core/org.ziptie.adapters.' .
        'common_*/scripts 2>&1';
	my $commonPath = `$lsCommand`;
	$commonPath =~ s/\n//;
	unless ($commonPath =~ /No such file or directory/)
	{
        $ENV{'PERL5LIB'} = ":$commonPath";
	}

	# Add generic adapter to PERL5LIB env variable
	$lsCommand = 'ls -d /usr/share/ziptie-server/adapters/' .
        'ziptie.adapters.generic.xml_*/scripts';
	my $genericPath = `$lsCommand`;
	$genericPath =~ s/\n//;
	$ENV{'PERL5LIB'} .= ":$genericPath";

	# Add the ZipTie::Client to the PERL5LIB env variable
	$ENV{'PERL5LIB'} .= ':/usr/share/ziptie-server/scripts';

}
