#!/usr/bin/perl

use strict;
use warnings;

use ZipTie::Client;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Adapters::CheckPoint::CPSMS::CPTool;

# Redirect warnings to the Logger so they don't pollute Tool output
my $LOGGER = ZipTie::Logger::get_logger();
local $SIG{__WARN__} = sub {
	my $warning = shift;
	chomp $warning;
	$LOGGER->debug($warning);
};


my ($managementServerIp, $applicationObjectSicName, $password) = @ARGV;

# check the ipAddress
if ($managementServerIp !~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/){
	print "ERROR,Not a valid IP address\n"; 
}

# check the OPSEC Application Object SIC name's content
if (($applicationObjectSicName eq "") || ($applicationObjectSicName !~ /CN\=/)){
	print "ERROR,Not a valid OPSEC Application Object SIC name\n";
}

# check the one-time password's content
if ($password eq "") {
	print "ERROR,One time password not available\n"; 
}

# set opsec_pull_cert application directory
my $opsec_pull_cert_dir = "/opt/qradar/bin"; 

# set default pulled certificate folder
my $out_dir = "/usr/share/ziptie-server/osgi-config/scratch";

# create a connectionPath object
my $connection_xml = <<__CONNECTIONXML__;
<connectionPath host="undefined_yet" managementServer="$managementServerIp">
		<protocols>
			<protocol name="CPSMS" port="not_defined" validatedOnDevice="false">
				<properties/>
			</protocol>
		</protocols>
		<credentials>
			<credential name="opsec_sic_name" value="$applicationObjectSicName"/>
			<credential name="cert_password" value="$password"/>
		</credentials>
	</connectionPath>
__CONNECTIONXML__

my $connectionPath = ZipTie::ConnectionPath::from_xml($connection_xml);

#pull certificate
# create a new instance of CheckPoint Tool (CPT) for opsec_pull_cert
my $cpt = ZipTie::Adapters::CheckPoint::CPSMS::CPTool->new(connectionPath => $connectionPath, out_dir => $out_dir, managementServerIp => $managementServerIp);
_call_opsec_pull_cert($cpt, $opsec_pull_cert_dir);


sub _call_opsec_pull_cert
{
	# receive opsec_pull_cert arguments through a previously built Check Point Tool object
	my ($cpt, $appdir) = @_;
	
	# clean possible previous CheckPoint 
	my $temp = $cpt->{out_dir};
	$cpt->{out_dir} = $cpt->{out_dir}."/certificates"."/".$cpt->{managementServerIp};
	$cpt->clean($cpt->{out_dir});
	# restore back the original out_dir
	$cpt->{out_dir} = $temp;
	# run opsec pull certificate
	$cpt->opsec_pull_cert($appdir);
	
}

1;

__END__

=head1 NAME

pullcertificate.pl - Script tool used to pull OPSEC application object's certificate from a Check Point certificate authority IP

=head1 SYNOPSIS

  usage: pullcertificate.pl 192.168.0.1 18190 CN=cp_mgmt_vm101-CPSMS1-GW3,O=vm1000-CPSMS..bs7ocx CN=cpsms101,O=vm100-CPSMS..bs7ocx password

=head1 DESCRIPTION

This script tool is used to pull OPSEC application object's certificate from a Check Point certificate authority IP

Generally you would run this script tool through Get Certificate button from Credentials tab page of the Ziptie CSM.


=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_call_opsec_pull_cert($cpt, $opsec_pull_cert_dir)>

The passed arguements received from the Get Certificate dialog, that is launched from Credentials CSM tab page, is bundled into
a Check Point Tool object C<$cpt>. The obtained C<$cpt> object is passed to the C<opsec_pull_cert> method along with the opsec_pull_cert application's directory. 
The C<opsec_pull_cert> method will retrieve the OPSEC application object's certificate.

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

Contributor(s): Toni Barabas(toni.barabas@ca.ibm.com)
Date: April 12, 2013

=cut
