#!/usr/bin/perl

package ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser;

use 5.010_000;
use strict;
use warnings;

use ZipTie::Adapters::CheckPoint::OPSEC::Parsers qw(object_to_hash);
use ZipTie::Adapters::Utils qw(get_crep get_port_number mask_to_bits);
use ZipTie::Logger;
use ZipTie::Model::AddressTranslation;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

sub new {
	
	my ( $proto, $in ) = @_;
    my $package = ref($proto) || $proto;
	my $self = {};
	
	$self = {
		in => $in,
		logger => ZipTie::Logger::get_logger(),
		zoneMapperHash => _setZoneMapperHash($in),
	};

	 bless($self, $package);
}

# This function returns packetProcessingOrdering element in case NAT is defined 
# for the analyzed gateway
sub getPacketProcessingOrdering {
	
	my ($self) = @_;

	if (!$self->{in}->{isNAT}){
		return undef;
	}

	my $packetProcessingOrdering = '';

	open (my $packetProcessingOrderingFh, '>', \$packetProcessingOrdering);
	
	my $packetProcessingOrderingPrinter =  ZipTie::Model::XmlPrint->new(\*$packetProcessingOrderingFh, 'core');
	
	$packetProcessingOrderingPrinter->open_element('core:packetProcessingOrdering');
	$packetProcessingOrderingPrinter->open_element('core:packetProcessingElement');
	$packetProcessingOrderingPrinter->print_element('core:packetProcessingPhase','packet-filtering');
	$packetProcessingOrderingPrinter->print_element('core:processingPoint','source');
	$packetProcessingOrderingPrinter->close_element('core:packetProcessingElement');
	$packetProcessingOrderingPrinter->open_element('core:packetProcessingElement');
	$packetProcessingOrderingPrinter->print_element('core:packetProcessingPhase','address-translation');
	$packetProcessingOrderingPrinter->close_element('core:packetProcessingElement');
	$packetProcessingOrderingPrinter->close_element('core:packetProcessingOrdering');

	close($packetProcessingOrderingFh);

	return ($packetProcessingOrdering);
}

# This function returns a processingPhase element
sub getPacketProcessingPhase {
	
	my ($self) = @_;

	if (!$self->{in}->{isNAT}){
		return undef;
	}
	
	my $packetProcessingPhase = '';
	
	open (my $packetProcessingPhaseFh, '>', \$packetProcessingPhase);
	
	if (defined $self->{in}->{NAT}->{settings}->{processingPhase}){
		my $packetProcessingPhasePrinter = ZipTie::Model::XmlPrint->new(\*$packetProcessingPhaseFh, 'core');
		$packetProcessingPhasePrinter->open_element('processingPhase');
		$packetProcessingPhasePrinter->print_element('core:packetProcessingPhase', $self->{in}->{NAT}->{settings}->{processingPhase});
		$packetProcessingPhasePrinter->close_element('processingPhase');	
	}
	
	close($packetProcessingPhaseFh);

	return ($packetProcessingPhase);
	
}

# This function returns a phaseDiscriminator element
sub getPhaseDiscriminator {
	
	my ($self) = @_;

	if (!$self->{in}->{isNAT}){
		return undef;
	}
	
	my $phaseDiscriminator = '';
	
	open (my $phaseDiscriminatorFh, '>', \$phaseDiscriminator);
	
	if (defined $self->{in}->{NAT}->{settings}->{processingPhasePosition}){
		my $phaseDiscriminatorPrinter = ZipTie::Model::XmlPrint->new(\*$phaseDiscriminatorFh, 'core');
		$phaseDiscriminatorPrinter->open_element('phaseDiscriminator');
		
		my $packetProcessingPhase =  $self->getPacketProcessingPhase();
		$phaseDiscriminatorPrinter->print_string($packetProcessingPhase);
		$phaseDiscriminatorPrinter->print_element('processingPhasePosition', $self->{in}->{NAT}->{settings}->{processingPhasePosition});
		$phaseDiscriminatorPrinter->close_element('phaseDiscriminator');	
	}
	
	close($phaseDiscriminatorFh);

	return ($phaseDiscriminator);
	
}

# This function returns global NAT settings from properties.xml or possible future objects.C file
# and store them under in->{NAT}->{settings}  
sub getNATProperties {
	
	my ($self) = @_;

	
	if (!$self->{in}->{isNAT}){
		return undef;
	}
	
	$self->{in}->{NAT}->{settings} = {};
	# check if it is a CPSMS environment
	if ((defined $self->{in}->{isCPSMS}) and ($self->{in}->{isCPSMS}==1)){
		my $XPCPR = $self->{in}->{XPCPR};
		foreach my $libxmlobject ($XPCPR->findnodes('/properties/object')) {
			my $object_name = $libxmlobject->findvalue('./object_name');
			if (defined $object_name and $object_name =~ /firewall_properties/){
				# manually translate destination on client side
				$self->{in}->{NAT}->{settings}->{nat_dst_client_side_manual} = $libxmlobject->findvalue('./nat_dst_client_side_manual');
				# automatically translate destination on client side					
				$self->{in}->{NAT}->{settings}->{nat_dst_client_side} = $libxmlobject->findvalue('./nat_dst_client_side');
			}
		}		
	} else {
		# SecurePlatform environment
		if (defined $self->{in}->{firewall_properties}){
			# manually translate destination on client side
			$self->{in}->{NAT}->{settings}->{nat_dst_client_side_manual} = $self->{in}->{firewall_properties}->{nat_dst_client_side_manual};
			# automatically translate destination on client side					
			$self->{in}->{NAT}->{settings}->{nat_dst_client_side} = $self->{in}->{firewall_properties}->{nat_dst_client_side};	
		} 
	} 
}

# This function generates routing NAT
sub routingNAT {
	
	my ($self, $typeNAT) = @_;
	
	if (!$self->{in}->{isNAT}){
		return undef;
	}
	my $RoutingNATFilterList = '';

	# open a printer to hold pre-routing NAT FilterList 
	open (my $RoutingNATFilterListFh, '>', \$RoutingNATFilterList);
	my $RoutingNATFilterListPrinter = ZipTie::Model::XmlPrint->new(\*$RoutingNATFilterListFh, 'core'); 
	
	
	if ($typeNAT eq 'automatic') {
		$self->_generateAutomaticNATXML($RoutingNATFilterListPrinter);	
	}
	if ($typeNAT eq 'manual'){
			$self->_generateManualNATXML($RoutingNATFilterListPrinter);
	}
		
	close($RoutingNATFilterListFh);
	return $RoutingNATFilterList;

}

# This function returns gateway's name
sub _getGatewayName{
	my ($self) = @_;
	my $name;
	
	my $gateway =  defined($self->{in}->{NAT}->{installed_on_gateway}) ? $self->{in}->{NAT}->{installed_on_gateway}: undef;
	# for CPSMS case
	if (exists $gateway->{hostname}) {
		$name = defined $gateway->{hostname} ? $gateway->{hostname}:undef; 
	}
	# for SecurePlatform's case
	if (exists $gateway->{name}) {
		$name = defined $gateway->{name} ? $gateway->{name}:undef;
	}	
	
	return $name;
	
}
# This function generates the XML content for automatic NAT
sub _generateAutomaticNATXML {
	
	my ($self,$routingNATFilterListPrinter) = @_;
		
	# hold the structure of NAT's object
	my $structNAT = {};
	# default protocol is 'ip' although CP enables for automatic NAT 'any' protocol 	
	my $protocol = 'ip';
	# NAT rule
	my $rule = '';
	# direction
	my $direction = '';
	# hostname contains the gateways name on which the NAT rules are installed
	my $gatewayName;
	# manual NAT flag is equal to "0" for dynamic NAT
	my $isManualNAT = 0; 

	# create a new AddressTranslation object
	my $addressTranslationObject = ZipTie::Model::AddressTranslation->new('any','any');
	
	# set the interfaceZoneMapper based on the on the gateway's interfaces characteristic
	$addressTranslationObject->setInterfaceZoneMapper(_getInterfaceZoneMapper($self));
						
	#loop through the NAT rules objects
	for my $NATObj (keys %{$self->{in}->{NAT}}){
			
		# confirm that NAT's rule objects are available within in->{networkGroup}
		if (defined ($self->{in}->{networkGroups}->{$NATObj})){
				
			# the address translated to
			my $mapIP = '';
			# real Networks Object Address 
			my $realAddr = '';
			
			# if it is available store the NATObj's specific structure
			$structNAT =  ${$self->{in}->{NAT}}{$NATObj};
		
			# check if the found NAT rule object it is applicable to the analyzed gateway
			my $gateway = $self->_getGatewayName();
			
			# parse the rule only if the referenced NATing gateway is going to be applied to All or it is hiding behind the analyzed firewall/gateway
			if (defined $gateway and ($structNAT->{the_firewalling_obj} =~ /$gateway/ or $structNAT->{the_firewalling_obj} =~ /All/) ) {
				
			} else { 
				# otherwise just skip to the next NAT rule
				next;
			}
			
			# test if the mapped address is a host and store it in mapIP
			# this will be re-updated again for dynamic or hide NAT rule
			if ($structNAT->{valid_ipaddr} =~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/){				 
				$mapIP = $structNAT->{valid_ipaddr};						
			} 
			
			# the real Address is going to be name of the NAT object 
			$realAddr = $NATObj;
			
			if ($structNAT->{netobj_adtr_method} =~ /static/) {
				foreach $direction ("source", "destination") {	
					# call createNATRule for a single, 'one' type of source or destination translation
					$self->_createNATRule('one',$addressTranslationObject, $isManualNAT, undef, $structNAT, $direction, $protocol, $mapIP, $realAddr);							
				}		
			}# end if structNAT contains static NAT rule
						
			if ($structNAT->{netobj_adtr_method} =~ /adtr_hide/) {
				$direction = 'source';
				# call createNATRule for a single, 'one' type of source or destination translation
				$self->_createNATRule('one',$addressTranslationObject, $isManualNAT, undef, $structNAT, $direction, $protocol, $mapIP, $realAddr);
			}# end if structNAT contains hide (dynamic) NAT rule	
			
		} # check the NAT-ed object is available between in->{networkGroups}
			
	} # end for NATObj
	foreach $direction ("source"){
		# generate translation list
		$rule = _generateTranslationList($addressTranslationObject, $direction,"automatic_NAT");
		$routingNATFilterListPrinter->print_string($rule);	
	}
	 		
} # end _generateAutomaticNATXML

# This function generates XML for manual NAT
sub _generateManualNATXML{
	
	my ($self, $routingNATFilterListPrinter) = @_;
	
	my $isManualNAT = 1;
	my $rule = '';
	# direction
	my $direction = '';
	
	# create a new AddressTranslation object
	my $addressTranslationObject = ZipTie::Model::AddressTranslation->new('any','any');
	
	# set the interfaceZoneMapper based on the on the gateway's interfaces characteristic
	$addressTranslationObject->setInterfaceZoneMapper(_getInterfaceZoneMapper($self));
	
	$self->_getNATFWPolicies();
	foreach my $rule (@{$self->{in}->{manualNAT}->{rules}}) {	
		$self->_createNATRule (undef, $addressTranslationObject, $isManualNAT, $rule, undef, undef, undef, undef, undef);
	}
	
	foreach $direction ("source"){
		# generate translation list
		$rule = _generateTranslationList($addressTranslationObject, $direction,"manual_NAT");
		$routingNATFilterListPrinter->print_string($rule);	
	}
	
}

# This function retrieves NAT rules for a specific Check Point environment i.e CPSMS or SPLAT 
sub _getNATFWPolicies{
	
	my ($self) = @_;
	
	# store manual NAT configuration
	$self->{in}->{manualNAT}->{rules} = [];
	
	if ((defined $self->{in}->{isCPSMS}) and ($self->{in}->{isCPSMS}==1)){
		# CPSMS environment
		_getCPSMSNATRules($self);	
	}else{
		# SecurePlatform environment
		_getSPLATNATRules($self);
	}
	
}

# This function checks the existence of manual NAT
sub checkManualNAT{
	
	my ($self) = @_;
	
	if ((defined $self->{in}->{isCPSMS}) and ($self->{in}->{isCPSMS}==1)){
		# CPSMS environment
		_getCPSMSNATRules($self, 'checkManualNAT');	
	}else{
		# SecurePlatform environment
		_getSPLATNATRules($self,'checkManualNAT');
	}
}


#This function parses rules.C file and stores manual NAT rules 
sub _getSPLATNATRules {
	
	my ($self,$check) = @_;
	$self->{in}->{manualNAT}->{available} = 0;
	
	
	my $checkManualNAT = defined ($check) ? $check :undef;
	open( RULES, $self->{in}->{rulesFile} ) or die("Unable to open file '$self->{rulesFile}': $!");
	my $rulesWhitespace;          # leading whitepsace on the rules definition
	my $currentRuleWhiteSpace;    # leading whitespace on each rule definition
	my $currentRule;
    my $type;
    

	while ( my $line = <RULES> ){
		if( $line =~ /^\s*#.*$/ ){
			next;
		}
		if (defined($rulesWhitespace)){
			if ( $line =~ /^$rulesWhitespace\)/ ){
				$rulesWhitespace = undef;
			    $type = undef;
			}
			elsif ( ( $line =~ /(^\s+):\s*\(/ ) && ( !$currentRuleWhiteSpace ) ){
				$currentRuleWhiteSpace = $1;
				$currentRule .= $line;
			}
			elsif ( $currentRuleWhiteSpace && ( $line =~ /^($currentRuleWhiteSpace|\t)\)/ ) ){
			    $currentRule .= $line;
			    if ($type eq 'rules-adtr'){
			    	if (not defined($checkManualNAT)){
			    		$self->_process_rule($currentRule);	
			    	}
			    }
			    $currentRule           = undef;
				$currentRuleWhiteSpace = undef;			
			}
			elsif ($line =~ /^\(/){
			    next;
			}
			elsif ($line =~ /^#/){
			    next;
			}
			else{
				$currentRule .= $line;
			}    
		}
		elsif ($line =~ /(^\s+):rules-adtr\s+/) {
			# analyze NAT rules only if they are available
			if ($line =~ /(^\s+):rules-adtr\s+\(\)/){
				next; 
			}
			$self->{in}->{manualNAT}->{available} = 1;
			$rulesWhitespace = $1;
			$type = 'rules-adtr';
		}
	} # end while	
	
	close(RULES);
}

# This function parses a given NAT rule passed by _getSPLATNATRules 
# and stores found attributes in a rule variable
sub _process_rule{
	my ($self,$currentRule) = @_;
	
	my $NATRule = ZipTie::Adapters::CheckPoint::OPSEC::Parsers::object_to_hash($currentRule);
	my $name         = ( keys(%{$NATRule}) )[0];
    my $o            = $NATRule->{$name};
    my $rule 		= {};
    # flag indicating the rule it's installed on gateway
    my $onGateway;

    return unless (ref $o);
    #get gateway's name
	my $gateway = $self->_getGatewayName();
    # find where the NAT rules are installed
	my $install = ( keys(%{$o->{install}}) )[0];
	$rule->{install} = $install;
	# check if the found NAT rule object it is applicable (installed) to the analyzed gateway or for 'Any' or 'Gateways' gateways	
	if ( defined $gateway and ( ($rule->{install} eq 'children' and $gateway ~~ @{$o->{install}->{children}}) 
								or ($rule->{install} =~ /$gateway/ or $rule->{install} =~ /Any/ or $rule->{install} =~ /Gateways/) 
								) ) {
		if ($rule->{install} eq 'children' and $gateway ~~ @{$o->{install}->{children}}){
			$rule->{install} = $gateway;
		};							
		$onGateway =1 ;
	}	
	

	if ($onGateway){
				
		#dst_adtr node
		$self->_setManualNATRuleArg($rule,'dst_adtr','realDstAddr', $o);
		
		# dst_adtr_translated node
		$self->_setManualNATRuleArg($rule,'dst_adtr_translated','mapDstAddr', $o);
				
		# services_adtr node
		$self->_setManualNATRuleArg($rule,'services_adtr','realServices', $o);
				
		# services_adtr_translated node
		$self->_setManualNATRuleArg($rule,'services_adtr_translated','mapServices', $o);
		
		# src_adtr node
		$self->_setManualNATRuleArg($rule,'src_adtr','realSrcAddr', $o);
		
		# src_adtr_translated node
		$self->_setManualNATRuleArg($rule,'src_adtr_translated','mapSrcAddr', $o);
		
		if (not defined ($rule->{protocol}) ){
			# store the default protocol value
			$rule->{protocol} = 'ip';
		}
	}	 
           
    push (@{$self->{in}->{manualNAT}->{rules}}, $rule);
	
}

sub _setManualNATRuleArg {
	
	my ($self, $rule, $CPRuleArg, $ruleArg, $o) = @_;
	
	while (my ($key,$value) = each %{$o->{$CPRuleArg}}) {
		if (defined($key) and $key ne "" and $key !~ /(AdminInfo|Any|children|adtr_method)/) {
			$rule->{$ruleArg} = $key;
			
		}

		if ( $key =~ /children/){
			$rule->{$ruleArg} = ${$o->{$CPRuleArg}->{$key}}[0];
			if (($CPRuleArg =~ /services/) and (not defined($rule->{protocol})) ) {
				#i.e. services_adtr or services_adtr_translated				
				$rule->{protocol} = $self->_getServicesProtocol(${$o->{$CPRuleArg}->{$key}}[0]);
			}		
		}
		if( ($CPRuleArg =~ /services_adtr_translated/) and ($key =~ /adtr_method/) ){
			# store nat methods i.e static or dynamic (hide)
			if ( $value =~ /static/ ){
				$rule->{isStaticNATMapping} = 1;
			}		
			if ( $value =~ /hide/ ) {
				$rule->{isStaticNATMapping} = 0;
			}				
		} # services_adtr_translated
		 
	} #while
	
}
# This function parses fw_policies.xml and store information about manual NAT as they are available under fw_policies/object/rule_adtr
sub _getCPSMSNATRules {
	
	my ($self, $check) = @_;
	my $XPCFW = defined($self->{in}->{XPCFW}) ? $self->{in}->{XPCFW}:undef;
	$self->{in}->{manualNAT}->{available} = 0;
	
	my $checkManualNAT = defined ($check) ? $check :undef;
	
	my $installedPolicy = $self->{in}->{installedPolicy};
	
 	if (defined $checkManualNAT) {
		foreach my $libxmlobject ($XPCFW->findnodes("/fw_policies/object[collection='referenced_object: $installedPolicy']/rule_adtr/rule_adtr/object")) {
			# found manual NAT node configuration
			$self->{in}->{manualNAT}->{available} = 1;
		}	
	}
	
	# it is called to parse the manual NAT rules
	if (not defined ($checkManualNAT) ){

		# find object nodes from adress translation rules
		foreach my $libxmlobject ($XPCFW->findnodes("/fw_policies/object[collection='referenced_object: $installedPolicy']/rule_adtr/rule_adtr/object")){
			
			my $rule = {};
			# check if rule is disabled
			my $isDisabled = $libxmlobject->findvalue('./disabled');
			if ($isDisabled eq "True"){
				next;
			}
			
			# find where the NAT rules are installed  
			foreach my $libxmlinstall ($libxmlobject->findnodes('./install')) {
				my $install = $libxmlinstall->findvalue('./install');
				my @installs = split ("referenced_object: ",$install);
				@{$rule->{install}} = @installs; 
			} # foreach install
			
			# check if the found NAT rule object it is applicable (installed) to the analyzed gateway
			my $gateway = $self->_getGatewayName();
			# check if is cluster member and compare target with cluster gateways
			my $clusterInfo = $self->{in}->{clusterInfo};
			my $isGatewayClusterTarget;
			if ($clusterInfo->{isClusterMember} && scalar @{$clusterInfo->{gateway_clusters}}){
				foreach my $gatewayCluster (@{$clusterInfo->{gateway_clusters}}){
					if ($gatewayCluster ~~ @{$rule->{install}}){
						$isGatewayClusterTarget = 1;
						last;
					}
				}
			}
			# parse the rule only if the referenced NATing gateway is going to be applied to on the analyzed gateway or 'Any' or 'Gateways'
			if (defined $gateway and ($gateway ~~ @{$rule->{install}} or "Any" ~~ @{$rule->{install}} or "Gateways" ~~ @{$rule->{install}} or $isGatewayClusterTarget ) ){
				# do nothing
			} else {
				# otherwise just skip to the next NAT rule	
				next;
			}
			
			# find dst_adtr nodes 
			foreach my $libxmldst_adtr ($libxmlobject->findnodes('./dst_adtr')) {
				my $dst_adtr = $libxmldst_adtr->findvalue('./dst_adtr');
				if ( ($dst_adtr ne "") and ($dst_adtr !~ /Any/) and ($dst_adtr =~ /referenced_object: (\S+)/) ){
					$rule->{realDstAddr} = $1; 
				}			
			} # foreach dst_adtr
			
			# find dst_adtr_translated nodes
			foreach my $libxmldst_adtr_translated($libxmlobject->findnodes('./dst_adtr_translated/object')){
				my $empty_field_name = $libxmldst_adtr_translated->findvalue('./empty_field_name');
				if (($empty_field_name ne "") and ($empty_field_name !~ /Any/) and ($empty_field_name =~ /referenced_object: (\S+)/) ){
					$rule->{mapDstAddr} = $1;		
				}	
			} # foreach dst_adtr_translated
			
			# find services_adtr nodes 
			foreach my $libxmlservices_adtr ($libxmlobject->findnodes('./services_adtr')) {
				my $services_adtr = $libxmlservices_adtr->findvalue('./services_adtr');
				if ( ($services_adtr ne "") and  ($services_adtr !~ /Any/) and ($services_adtr =~ /referenced_object: (\S+)/) ){
					$rule->{realServices} = $1;
					# store the associated protocol
					if ( (not defined($rule->{protocol})) or ( defined($rule->{protocol}) and ($rule->{protocol} eq 'ip') ) ){
						$rule->{protocol} = $self->_getServicesProtocol($1);
					}
				}			
			} # foreach services_adtr
			
			# find services_adtr_translated nodes
			foreach my $libxmlservices_adtr_translated($libxmlobject->findnodes('./services_adtr_translated/object')){
				# type of the translated services
				my $empty_field_name = $libxmlservices_adtr_translated->findvalue('./empty_field_name');
				if (($empty_field_name ne "") and ($empty_field_name !~ /Any/) and ($empty_field_name =~ /referenced_object: (\S+)/) ) {
					# store the mapped (translated to) services
					$rule->{mapServices} = $1;
					# store the associated protocol
					if ( (not defined($rule->{protocol})) or ( defined($rule->{protocol}) and ($rule->{protocol} eq 'ip') ) ){
						$rule->{protocol} = $self->_getServicesProtocol($1);
					}
					 
				} 
				# type of manual NAT i.e static or hide (dynamic)
				my $adtr_method = $libxmlservices_adtr_translated->findvalue('./adtr_method');
				if ( defined($adtr_method) and ($adtr_method ne "") ) {	
					# store nat methods i.e static or dynamic (hide)
					if ( $adtr_method =~ /static/ ){
						$rule->{isStaticNATMapping} = 1;
					}		
					if ( $adtr_method =~ /hide/ ) {
						$rule->{isStaticNATMapping} = 0;
					}			
				}		
			} # foreach services_adtr_translated
					
			# find src_adtr nodes 
			foreach my $libxmlsrc_adtr ($libxmlobject->findnodes('./src_adtr')) {
				my $src_adtr = $libxmlsrc_adtr->findvalue('./src_adtr');
				if ( ($src_adtr ne "" ) and ($src_adtr !~ /Any/) and ($src_adtr =~ /referenced_object: (\S+)/) ){
					$rule->{realSrcAddr} = $1;
				}		
			} # foreach src_adtr
			
			# find src_adtr_translated nodes
			foreach my $libxmlsrc_adtr_translated($libxmlobject->findnodes('./src_adtr_translated/object')){
				my $empty_field_name = $libxmlsrc_adtr_translated->findvalue('./empty_field_name');
				if ( ($empty_field_name ne "") and ($empty_field_name !~ /Any/) and ($empty_field_name =~ /referenced_object: (\S+)/) ){
					$rule->{mapSrcAddr} = $1;			
				} # end foreach mapped src_adtr_translated
			} # foreach src_adtr_translated	
			
			# if it's still not available then set a default value for protocol 
			if (not defined ($rule->{protocol}) ){
				# store the default protocol value
				$rule->{protocol} = 'ip';
			}
			push (@{$self->{in}->{manualNAT}->{rules}}, $rule);	
		}# for libxmlobject
	} #end if checkManualNAT
	
	
} # end _getCPSMSNATRules

# This function lookup a given service and returns the associated protocol. 
# By default it returns an 'ip' protocol if it couldn't find an associated service 
sub _getServicesProtocol{
	
	my ($self,$service) = @_;
	
	my $protocol = undef;
	
	
	if ((defined $self->{in}->{isCPSMS}) and ($self->{in}->{isCPSMS}==1)){
		# CPSMS environment
		if (scalar keys %{$self->{in}->{portServiceGroups}->{$service}} != 0){
			foreach my $key (keys %{$self->{in}->{portServiceGroups}->{$service}}){
				if ( (ref($self->{in}->{portServiceGroups}->{$service}->{$key}) eq 'HASH') and (exists $self->{in}->{portServiceGroups}->{$service}->{$key}->{'protocol'}) ){
					$protocol = defined ($self->{in}->{portServiceGroups}->{$service}->{$key}->{'protocol'}) ?
						$self->{in}->{portServiceGroups}->{$service}->{$key}->{'protocol'}:'ip';
				}
			}
		}else {
			$protocol = defined($self->{in}->{portServiceGroups}->{$service}->{'portExpression'}->{'protocol'}) ? 
				$self->{in}->{portServiceGroups}->{$service}->{'portExpression'}->{'protocol'} : 'ip'; 
			$protocol = defined ($self->{in}->{portServiceGroups}->{$service}->{'portRange'}->{'protocol'}) ?
				$self->{in}->{portServiceGroups}->{$service}->{'portRange'}->{'protocol'} : 'ip';
		}
				
	}else {
		# SecurePlatform environment
		$protocol = defined($self->{in}->{portServiceGroups}->{$service}->{'protocol'}) ? 
			$self->{in}->{portServiceGroups}->{$service}->{'protocol'} : 'ip';
	}
	
	
	return $protocol;
}


# This function assembles the required NAT translation arguments and calls _createAddressTranslationRule 
# to add the rule and the address rule criterion to an address translation object
sub _createNATRule {
	
	my ($self, $setOfRules, $addressTranslationObject, $isManualNAT, $rule, $structNAT, $direction, $protocol, $mapIP, $realAddr) = @_;
	
	my $ruleAddrTranslKey = undef;
	# by default isStaticNATMapping is undefined
	my $isStaticNATMapping = undef;

	if (!$isManualNAT){
		# automatic NAT
		
		# if static NAT
		if ($structNAT->{netobj_adtr_method} =~ /static/){
			$isStaticNATMapping = 1;	
		} # end if static
			
		# if hide (dynamic) NAT
		if ($structNAT->{netobj_adtr_method} =~ /adtr_hide/){
			$isStaticNATMapping = 0;
			# check if it is hidding behind a gateway or a given IP address
			if ( ($structNAT->{valid_ipaddr} =~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/) and ($structNAT->{valid_ipaddr} =~ /0.0.0.0/) ){
				# if it is hiding behind Gateway's ip address
				# then the gateway's ip address is the address to be mapped to
				my $gateway = $self->_getGatewayName();
				$mapIP = $gateway;
									
			} else {
				#it is hiding behind a specific address
				$mapIP = $structNAT->{valid_ipaddr};
			}
		} # end if hide NAT
		
		#create address translation keys
		$ruleAddrTranslKey = _createAddressTranslationKeys($direction, $isStaticNATMapping, $mapIP, $realAddr);
		
	} #end automatic NAT
	
			
	# test if the mapped address is ...net
		
	# test if the mapped addess is ... objectGroupReference		
	
	my $addRuleArguments = {};
	my $criterionArguments = {};
	
	if ($isManualNAT){
		#manual NAT

		# set to static or hide (dynamic)
		if ( (not defined ($addRuleArguments->{isStaticNATMapping}) ) and (defined $rule->{isStaticNATMapping}) ){
			$addRuleArguments->{isStaticNATMapping} = $rule->{isStaticNATMapping};		
		}
		
		# set protocol
		if ( (not defined($addRuleArguments->{protocol}) ) and (defined $rule->{protocol}) ) {
			$addRuleArguments->{protocol} = $rule->{protocol};
		}	
		 	
		if ( defined($rule->{mapSrcAddr}) ) {
			# handle it as objectGroupReference	
			$addRuleArguments->{sourceAddrNameMapped} = $rule->{mapSrcAddr};
		} #mapSrcAddr			
			
		if ( defined($rule->{mapDstAddr}) ){
			# handle it as objectGroupReference
			$addRuleArguments->{destAddrNameMapped} = $rule->{mapDstAddr};
		} # mapDstAddr 
		
		if ( defined($rule->{realSrcAddr})) {
			# handle it as objectGroupReference
			$criterionArguments->{sourceIpObjectGroup} = $rule->{realSrcAddr};	
		}# realSrcAddr					
		
		if ( defined($rule->{realDstAddr})){
			# handle it as objectGroupReference
			$criterionArguments->{destIpObjectGroup} = $rule->{realDstAddr};					
		} # realDstAddr
			
		if ( defined($rule->{realServices}) ){
			$criterionArguments->{sourceServiceObjectName} = $rule->{realServices}; 
		}
				
		if ( defined($rule->{mapServices}) ) {
			$criterionArguments->{destServiceObjectName} = $rule->{mapServices};
		}
	} else {
		#automatic NAT
							
		# set to static or hide (dynamic)
		if ( (not defined ($addRuleArguments->{isStaticNATMapping}) ) and (defined $isStaticNATMapping) ){
			$addRuleArguments->{isStaticNATMapping} = $isStaticNATMapping;		
		}
	
		if ( (not defined($addRuleArguments->{protocol}) ) and (defined $protocol) ) {
			$addRuleArguments->{protocol} = $protocol;
		}
		
		# find mapIP's type
		if ( defined($mapIP) and $mapIP =~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/) {
			# overwrite the default value of $ruleAddrTranslKey->{ruleAddr} given by createAddressTranslationKeys
			if ($ruleAddrTranslKey->{ruleAddr} =~ /source/){
				$addRuleArguments->{sourceAddrMapped} = $mapIP;	
			}
			if ($ruleAddrTranslKey->{ruleAddr} =~ /dest/) {
				$addRuleArguments->{destAddrMapped} = $mapIP;
			}		
		} else {
			# keep the default value of $ruleAddrTranslKey->{ruleAddr} given by createAddressTranslationKeys
			$addRuleArguments->{$ruleAddrTranslKey->{ruleAddr}} = $mapIP;
		}
		
		# must be an objectGroup or host
		$criterionArguments->{$ruleAddrTranslKey->{criterionAddr}} = $realAddr;
			
		 
	} # end else automatic NAT			

	_createAddressTranslationRule ($addressTranslationObject, $addRuleArguments, $criterionArguments);
}

# This function maps the name arguements required by createAddressTranslationRule function
sub _createAddressTranslationKeys {
	
	my ($direction, $isStaticNATMapping, $mapIP, $realAddr) = @_;
	
	my $addrDirection = '';
	my $ruleAddr = '';
	my $criterionAddr = '';		
	$mapIP = (defined($mapIP)? $mapIP:undef);
	$realAddr = (defined ($realAddr)? $realAddr:undef);
	
	if ($direction eq 'source'){
		$addrDirection = 'source';		 
	} else {
		$addrDirection = 'dest';
	}
	
	if ($isStaticNATMapping) {
		# for static NAT
		$ruleAddr = $addrDirection.'AddrMapped';
		$criterionAddr = _setCriterionAddr($realAddr, $addrDirection);	
	} else{
		# for hide (dynamic) NAT
		$ruleAddr = $addrDirection.'AddrNameMapped';
		$criterionAddr = _setCriterionAddr($realAddr, $addrDirection);
		
	}
	
	return { addrDirection => $addrDirection, ruleAddr => $ruleAddr, criterionAddr => $criterionAddr }; 
}

sub _setCriterionAddr{
	my ($realAddr, $addrDirection) = @_;
	my $criterionAddr;
	
	my $IpPattern = qr/^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/;
	
	if (defined($realAddr) and ($realAddr =~ $IpPattern) ) {
			$criterionAddr = $addrDirection.'Ip';	
	}elsif (defined($realAddr) and ($realAddr !~ $IpPattern) ){
		$criterionAddr = $addrDirection.'IpObjectGroup';
	}
	
	return $criterionAddr;
	
}

# This function creates a new address translation rule in $addressTranslation.  
# $addRuleArguments and $criterionArguments are both references to a hash of arguments to be passed into addRule and
# addRuleAddressCriterion methods.  
#
# For the passed in argument sets for addRuleAddressCriterion the rule argument
# does not need to be set.  This is set automatically after the call to addRule is completed.
sub _createAddressTranslationRule
{
	my ($addressTranslationObject, $addRuleArguments, $criterionArguments) = @_;

	my $rule = $addressTranslationObject->addRule($addRuleArguments);

	$criterionArguments->{rule} = $rule;
	$addressTranslationObject->addRuleAddressCriterion($criterionArguments);

}

# This function generate translation list for an address translation object
sub _generateTranslationList
{
    my ( $addressTranslationObject, $direction, $type ) = @_;
	
	my $nonDefaultListName = defined $type ? $type : undef;
	my $translationList = '';
	
	open (my $translationListFh, '>', \$translationList);
	
	my $translationListPrinter = ZipTie::Model::XmlPrint->new(\*$translationListFh, 'core'); 
	$addressTranslationObject->generateTranslationList($translationListPrinter, $direction, $nonDefaultListName);

	close($translationListFh);

	return ($translationList);

}

# addressTranslation's setInterfaceZoneMapper takes in a function pointer to a function that returns a hash
# containing the zone mappings.  This function returns that function pointer.
sub _getInterfaceZoneMapper
{
	my ($self) = @_;
	
	return sub { return $self->{zoneMapperHash}; };
}

# This function returns a hash (zoneMapperHash) with keys containing the mapped interfaces 
sub _setZoneMapperHash
{
	my ($in) = @_;
	
	my %zoneMapperHash;
	my @interfaces;
	
	if ((defined $in->{isCPSMS}) and ($in->{isCPSMS}==1)) {
		foreach my $intf (@{$in->{NAT}->{installed_on_gateway}->{intfs}}){
	
		#push (@{$zoneMapperHash{$intf->{typeofIntf}}}, $intf->{officialname});
		push (@interfaces, $intf->{officialname});
		}	
	} else {
		while (my($key,$value) = each %{$in->{NAT}->{installed_on_gateway}->{interfaces}}){
			push (@interfaces, $key);
		}
	}
	
	$zoneMapperHash{'any'} = \@interfaces;
	return \%zoneMapperHash;
}

=head1 NAME

	ZipTie::Adapters::CheckPoint::OPSEC::AddressTranslationParser
	Parse Check Point CPSMS and SecurePlatform NAT configuration

=head1 DESCRIPTION

	This module implements the NAT parsers for Check Point Security Management Server and Check Point SecurePlatform adapters.  
	It uses ZipTie::Model::AddressTranslation for the low level XML generation.

=head1 AUTHOR(S)

	Toni Barabas
	toni.barabas@ca.ibm.com

=head1 COPYRIGHT

	Copyright (c) 2013, IBM Corporation. All Rights Reserved.
	This module is free software. It may be used, redistributed
	and/or modified under the same terms as Perl itself.

=cut

1;
