package ZipTie::Adapters::Cisco::CatOS;

use warnings;
use strict;

use ZipTie::CLIProtocolFactory;
use ZipTie::CLIProtocol;
use ZipTie::Adapters::Cisco::CatOS::AutoLogin qw(issue_command $invalid_commands_RE);
use ZipTie::Adapters::Cisco::CatOS::RestoreConfig;
use ZipTie::Adapters::Cisco::CatOS::GetConfig qw(get_config);
use ZipTie::Adapters::Cisco::CatOS::Parsers
  qw(parse_routing create_config parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_interfaces parse_static_routes parse_vlans parse_stp parse_vtp parse_services);
use ZipTie::Adapters::Cisco::CatOS::Disconnect
	qw(disconnect);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::Model::XmlPrint;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Adapters::GenericAdapter;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

# Specifies that this adapter is a subclass of ZipTie::Adapters::BaseAdapter
use ZipTie::Adapters::BaseAdapter;
our @ISA = qw(ZipTie::Adapters::BaseAdapter);


sub backup
{
	my $package_name = shift;
	
	# Retrieve the operation XML document that contains all of the IP, protocol, credential, and file server information
	# that is needed to successfully backup a device.
	my $backup_doc = shift;

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ( $connection_path ) = ZipTie::Typer::translate_document( $backup_doc, 'connectionPath' );

	# Connect to the device and capture the ZipTie::CLIProtocol that is created as a result of the connection.
	# Also be sure to capture the device prompt that is returned as a result of a successful connection.
	my ( $cli_protocol, $enable_prompt_regex ) = _connect( $connection_path );

	# Grab an output filehandle for the model.  This usually points to STDOUT
	my $filehandle = get_model_filehandle( 'CatOS', $connection_path->get_ip_address() );

	# initialize the model printer
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'cisco', "http://www.ziptie.org/model/cisco/1.0 cisco.xsd" );
	$printer->open_model();

	# Begin executing commands on the device.  The results of each command will
	# be stored in a single hashtable ($parsers) and fed into each parsing method
	my $responses = {};
	
    my $logging = issue_command('show logging', $cli_protocol, $enable_prompt_regex);
    if ($logging =~ /Logging\s+Session:\s+Enable/i)
    {
    	$cli_protocol->send_and_wait_for( "set logging session disable", $enable_prompt_regex );
    }
	
	$responses->{config} = get_config( $cli_protocol, $connection_path );


	$responses->{version} = issue_command('show version', $cli_protocol, $enable_prompt_regex);
	$responses->{bootinfo} = issue_command('whichboot', $cli_protocol, $enable_prompt_regex);
	$responses->{module} = issue_command('show module', $cli_protocol, $enable_prompt_regex);
	$responses->{mod_ver} = issue_command('show mod ver', $cli_protocol, $enable_prompt_regex);
	$responses->{'system'} = issue_command('show system', $cli_protocol, $enable_prompt_regex);
	$responses->{show_fs} = issue_command('show flash devices', $cli_protocol, $enable_prompt_regex);
		
	foreach my $fileSys ( _get_file_systems( $responses->{show_fs} ) )
	{
		if ($fileSys =~ /disk\d+|slot\d+|bootdisk|bootflash/i)
		{
			$responses->{file_systems}->{$fileSys} = issue_command("show flash " . $fileSys . ":", $cli_protocol, $enable_prompt_regex);
		}
	}

	parse_system( $responses, $printer );
	parse_chassis( $responses, $printer );
	delete $responses->{version};
	delete $responses->{module};
	delete $responses->{'system'};
	delete $responses->{show_fs};

	create_config( $responses, $printer );

	$responses->{ifalias} = issue_command('show snmp ifalias', $cli_protocol, $enable_prompt_regex);
	if ( $responses->{ifalias} =~ /\bUsage:/mi )
	{
		$responses->{ifalias} = issue_command('show port ifIndex', $cli_protocol, $enable_prompt_regex);
	}
	
	$responses->{interfaces} = issue_command('show interface', $cli_protocol, $enable_prompt_regex);
	
	$responses->{ports} = issue_command('show port', $cli_protocol, $enable_prompt_regex);
	
	$responses->{stp} = issue_command('show spantree', $cli_protocol, $enable_prompt_regex);

	parse_interfaces( $responses, $printer );
	delete $responses->{interfaces};
	delete $responses->{ports};
	delete $responses->{ifalias};

	parse_local_accounts( $responses, $printer );

	parse_services( $responses, $printer );

	parse_snmp( $responses, $printer );
	delete $responses->{config};

	parse_stp( $responses, $printer );
	delete $responses->{stp};

	$responses->{static_routes} = issue_command('show ip route', $cli_protocol, $enable_prompt_regex);
	parse_static_routes( $responses, $printer );
	delete $responses->{static_routes};

	$responses->{vlans} = issue_command('show vlan', $cli_protocol, $enable_prompt_regex);
	parse_vlans( $responses, $printer );
	delete $responses->{vlans};

	$responses->{vtp_info} = issue_command('show vtp domain', $cli_protocol, $enable_prompt_regex);
	parse_vtp( $responses, $printer );
	delete $responses->{vtp_info};
	
	# Disconnect from the specified device
	disconnect($cli_protocol);

	# close out the ZiptieElementDocument
	$printer->close_model();

	# Make sure to close the model file handle
	close_model_filehandle($filehandle);
}

sub commands
{
	my $package_name = shift;
	my $command_doc = shift;
	my ( $connection_path, $commands ) = ZipTie::Typer::translate_document( $command_doc, 'connectionPath' );
	my ( $cli_protocol, $enable_prompt_regex ) = _connect( $connection_path );
    	my $termLen = $cli_protocol->send_and_wait_for( "set length 0", $enable_prompt_regex );
    	if ($termLen =~ /$invalid_commands_RE/i)
    	{
    		# set the --more-- prompt if the term length 0 didn't go through
        	$cli_protocol->set_more_prompt( '<--- More --->\s*$', '20');
    	}
	my $result = ZipTie::Adapters::GenericAdapter::execute_cli_commands('CatOS', $cli_protocol, $commands, '(#|\$|>)\s*$|'.$enable_prompt_regex);
	disconnect($cli_protocol);
	return $result;
}

sub restore
{
	my $package_name = shift;
	my $command_doc  = shift;
	my ( $connection_path, $restoreFile ) = ZipTie::Typer::translate_document( $command_doc, 'connectionPath' );
	
	# Check to see if either TFTP or SCP are supported
	my $tftp_protocol = $connection_path->get_protocol_by_name("TFTP") if ( defined($connection_path) );
	my $scp_protocol  = $connection_path->get_protocol_by_name("SCP")  if ( defined($connection_path) );
	
	if ( $restoreFile->get_path() =~ /config/i )
	{
		if ( defined($tftp_protocol) )
		{
			my ( $cli_protocol, $enable_prompt_regex ) = _connect($connection_path);
			ZipTie::Adapters::Cisco::CatOS::RestoreConfig::restore_via_tftp( $connection_path, $cli_protocol, $restoreFile, $enable_prompt_regex );
			disconnect($cli_protocol);
		}
		elsif ( defined($scp_protocol) )
		{
			ZipTie::Adapters::GenericAdapter::scp_restore( $connection_path, $restoreFile );
		}
		else
		{
			$LOGGER->fatal("Unable to restore CatOS config.  Protocols SCP and TFTP are not available.");
		}
	}
	else
	{
		$LOGGER->fatal( "Unable to promote this type of configuration '" . $restoreFile->get_path() . "'." );
	}
}

sub _connect
{
	# Grab our arguments
	my $connection_path = shift;

	# Create a new CLI protocol object
#	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	# Make a connection to and successfully authenticate with the CatOS device
#	my $enable_prompt_regex = ZipTie::Adapters::Cisco::CatOS::AutoLogin::execute( $cli_protocol, $connection_path );
	my ($enable_prompt_regex, $cli_protocol) = ZipTie::Adapters::Cisco::CatOS::AutoLogin::execute( $connection_path );

	# Store the found prompt as "enablePrompt" on the specified CLI protocol.
#	$cli_protocol->set_prompt_by_name( "enablePrompt", $enable_prompt_regex );

	# Return the created ZipTie::CLIProtocol object and the enable prompt encountered after successfully connecting to a device.
	return ( $cli_protocol, $enable_prompt_regex );
}

sub _get_file_systems
{

	# Given the output of "show flash devices", this method returns
	# an array of the flash file systems
	my @showFileSystems = shift;
	my @fs;
	foreach (@showFileSystems)
	{
		if ($_ =~ /(\S+,.+$)/m)		# Expect a comma delimited list of file systems ...
		{
			@fs = split(/,\s/, $1);
		}
                elsif ($_ =~ /^(\w+)\s*$/m) 	# ... or a single file system.
                {
                        push (@fs, $1);
                }
	}
	return @fs;
}

1;

__END__

=head1 NAME

ZipTie::Adapters::Cisco::CatOS - Main adapter module for backing up Cisco CatOS network devices

=head1 VERSION

This document describes the current version of ZipTie::Adapters::Cisco::CatOS (version number not specified)

=head1 SYNOPSIS

    use ZipTie::Adapters::Cisco::CatOS;
    backup ($backup_document);
    commands ($command_document); - DEPRICATED
    restore ($command_document);  - DEPRICATED

=head1 DESCRIPTION

This module is an adapter that allows the ZipTie framework to connect to Cisco CatOS devices.
It retrieves configuration information from a specified device and transforms it into a Standard Element Document.


=head1 SUBROUTINES/METHODS

=over 12


=item C<backup($backup_document)>

This function takes a backup doccument, uses the information that it contains to connect to the specified device, and reads configuration
information from that device. It then transforms that information into Standard Element Document format and prints it into an xml document.
This document provides information about the device's configuration in a standard format to the caller.

=item C<commands($command_document)>

B<Depricated function, not maintained. This code should not be used.>

=item C<restore($command_document)>

B<Depricated function, not maintained. This code should not be used.>

=back

=head1 DIAGNOSTICS

N/A

=head1 CONFIGURATION AND ENVIRONMENT

N/A

=head1 DEPENDENCIES

N/A

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

 No known defects at this time.

 This adapter does not back up ACLs.
 This adapter does not back up routing protocol information.

=head1 LICENSE AND COPYRIGHT

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

Portions created by IBM are Copyright (C) 2013, IBM Corp.

=head1 AUTHOR

Contributor(s): David Tickner dtickner@ca.ibm.com, Prior contributors unknown as no documentation was attached to the Original Code

Date: November 2013

=cut

