package ZipTie::Adapters::Cisco::CatOS::Telemetry;

use strict;
use warnings;

use ZipTie::Logger;
use ZipTie::Typer;
use ZipTie::CLIProtocolFactory;
use ZipTie::Adapters::Cisco::CatOS::AutoLogin qw(issue_command);
use ZipTie::Adapters::Cisco::CatOS::Disconnect qw(disconnect);
use ZipTie::Adapters::Cisco::CatOS::Parsers qw(parse_arp parse_cdp parse_mac_table parse_telemetry_interfaces);
use ZipTie::Adapters::Utils qw(choose_admin_ip get_model_filehandle close_model_filehandle);
use ZipTie::Model::XmlPrint;

my $LOGGER = ZipTie::Logger::get_logger();

sub invoke
{
	my $pkg               = shift;
	my $connectionPathDoc = shift;

	# setup work
	my ($connectionPath, $discoveryParams) = ZipTie::Typer::translate_document( $connectionPathDoc, 'connectionPath' );

#	my $cliProtocol = ZipTie::CLIProtocolFactory::create($connectionPath);
#
#	# Make a connection to device and get the data necessary to fill out the discovery event document 
#	my $promptRegex = ZipTie::Adapters::Cisco::CatOS::AutoLogin::execute( $cliProtocol, $connectionPath );
#	$cliProtocol->send_and_wait_for('set length 0', $promptRegex);
	my ($promptRegex, $cliProtocol) = ZipTie::Adapters::Cisco::CatOS::AutoLogin::execute( $connectionPath );


####### Set up the XML Printer ###

	my $filehandle = get_model_filehandle( "CatOS", $connectionPath->get_ip_address() );
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'telemetry');
	$printer->attributes(1);
	$printer->open_discovery_event();
	
	
	
####### Issue Commands and Parse Responses ###

	my $responses = {};

	$printer->open_element('neighbors');

	# process the ARP table
#	$responses->{arp} = $cliProtocol->send_and_wait_for('show arp', $promptRegex);
	$responses->{arp} = issue_command('show arp', $cliProtocol, $promptRegex);
	parse_arp($responses, $printer);
	delete $responses->{arp};

	# process the CDP neighbors
#	$responses->{cdp} = $cliProtocol->send_and_wait_for('show cdp neighbors', $promptRegex);
	$responses->{cdp} = issue_command('show cdp neighbors', $cliProtocol, $promptRegex);
#	$responses->{cdp} .= $cliProtocol->send_and_wait_for('show cdp', $promptRegex);
	$responses->{cdp} .= issue_command('show cdp', $cliProtocol, $promptRegex);
	
	parse_cdp($responses, $printer);
	delete $responses->{cdp};

	# process the MAC address table 
#	$responses->{mac} = $cliProtocol->send_and_wait_for('show cam dynamic', $promptRegex);
	$responses->{mac} = issue_command('show cam dynamic', $cliProtocol, $promptRegex);
	parse_mac_table($responses, $printer);
	delete $responses->{mac};

	$printer->close_element('neighbors');

	# process the interfaces and their status
#	$responses->{port_status} = $cliProtocol->send_and_wait_for('show port status', $promptRegex);
	$responses->{port_status} = issue_command('show port status', $cliProtocol, $promptRegex);
	
#	$responses->{interface} = $cliProtocol->send_and_wait_for('show interface', $promptRegex);
	$responses->{interface} = issue_command('show interface', $cliProtocol, $promptRegex);


# The inside of the while loop was commented out, so I got rid of the whole block.
# No sense in running a while loop that does nothing.
	
#	my ( $status_blob ) = $responses->{port_status} =~ /^[-\s]+^(.+)/imsg;
#	while ( $status_blob =~ /^\s*(\S+)(?:\s+\S+)?\s+(\S+)\s+(\d+)\s+\S+\s+\S+\s+\S+\s+(?:No Connector|\S+)\s*$/mig )
#	{
#		#$responses->{"port_counters_$1"} = $cliProtocol->send_and_wait_for("show counters $1", $promptRegex,90);
#	}
	my $interfacesHash = parse_telemetry_interfaces($responses, $printer);
	delete $responses->{interface};
	delete $responses->{port_status};
	while ( ( my $key, my $value ) = each%{$responses} )
	{
		delete $responses->{$key};
	}

	my $adminIp    = $connectionPath->get_ip_address();
	if ( $discoveryParams->{calculateAdminIp} eq 'true' )
	{
		$adminIp = choose_admin_ip($connectionPath->get_ip_address(), $interfacesHash);
	}
	$printer->print_element( 'adminIp', $adminIp );
	$interfacesHash = 0;

	# tear down work
	$printer->close_element('DiscoveryEvent');
	
	
	
	
####### Disconnect from the device ###
	
	close_model_filehandle($filehandle);
	disconnect($cliProtocol);
	
	return;
}

1;

__END__

=head1 NAME

ZipTie::Adapters::Cisco::CatOS::Telemetry - Main adapter module for backing up Cisco CatOS network devices

=head1 VERSION

This document describes the current version of ZipTie::Adapters::Cisco::CatOS::Telemetry (version number not specified)

=head1 SYNOPSIS

    use ZipTie::Adapters::Cisco::CatOS::Telemetry;
    invoke ($connectionPathDoc);

=head1 DESCRIPTION

This module is the telemetry portion an adapter that allows the ZipTie framework to connect to Cisco CatOS devices.
It retrieves information about neighbors of those devices and transforms it into a Discovery Event document.


=head1 SUBROUTINES/METHODS

=over 12


=item C<invoke($connectionPathDoc)>

This function takes a connection path document, uses the information that it contains to connect to the specified device, and reads configuration
information from that device. It then transforms that information into Discovery Event document format and prints it into an xml document.
This document provides information about the device's neighbors in a standard format to the caller.


=back

=head1 DIAGNOSTICS

N/A

=head1 CONFIGURATION AND ENVIRONMENT

N/A

=head1 DEPENDENCIES

N/A

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

No known defects at this time.

=head1 LICENSE AND COPYRIGHT

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

Portions created by IBM are Copyright (C) 2013, IBM Corp.

=head1 AUTHOR

Contributor(s): David Tickner dtickner@ca.ibm.com, Prior contributors unknown as no documentation was attached to the Original Code

Date: November 2013

=cut

