package ZipTie::Adapters::Cisco::IOS;

use strict;
use warnings;
use 5.010_000;

use ZipTie::Adapters::BaseAdapter;
use ZipTie::Adapters::Cisco::IOS::AutoLogin;
use ZipTie::Adapters::Cisco::IOS::GetStartupConfig qw(get_startup_config);
use ZipTie::Adapters::Cisco::IOS::Parsers qw(parse_vtp parse_static_routes parse_vlans parse_routing parse_access_ports parse_interfaces create_config parse_local_accounts);
use ZipTie::Adapters::Cisco::IOS::Parsers qw(parse_chassis parse_filters parse_snmp parse_system parse_stp parse_services parse_mpls parse_qos parse_object_groups parse_dynamic_routes);
use ZipTie::Adapters::Cisco::IOS::Disconnect qw(disconnect);
use ZipTie::Adapters::GenericAdapter;
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::CLIProtocolFactory;
use ZipTie::CLIProtocol;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

use Exporter 'import';
our @EXPORT_OK = qw(clean_responses);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

# Specifies that this adapter is a subclass of ZipTie::Adapters::BaseAdapter
our @ISA = qw(ZipTie::Adapters::BaseAdapter);

sub backup {
	my $package_name = shift;

	# Retrieve the operation XML document that contains all of the IP, protocol, credential, and file server information
	# that is needed to successfully backup a device.
	my $backup_doc = shift;
	my $responses  = {};

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ($connection_path) = ZipTie::Typer::translate_document($backup_doc, 'connectionPath');

	# Connect to the device and capture the ZipTie::CLIProtocol that is created as a result of the connection.
	# Also be sure to capture the device prompt that is returned as a result of a successful connection.
	my ($cli_protocol, $enable_prompt_regex) = _connect($connection_path);

	# Attempt to disable pagination
	my $termLen = $cli_protocol->send_and_wait_for('terminal length 0', $enable_prompt_regex);

	# Grab an output filehandle for the model
	my $filehandle = get_model_filehandle('IOS', $connection_path->get_ip_address());

	# initialize the model printer
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, 'cisco', 'http://www.ziptie.org/model/cisco/1.0 cisco.xsd');
	$printer->open_model();

	$responses = get_command_responses($cli_protocol, $enable_prompt_regex, $connection_path);
	clean_responses ($responses);
	
	parse_system($responses, $printer);
	parse_chassis($responses, $printer);
	create_config($responses, $printer);
	parse_access_ports($responses, $printer);
	parse_dynamic_routes($responses, $printer);
	parse_object_groups($responses, $printer);
	parse_filters($responses, $printer);
	my ($subnets, $filterAssignments) = parse_interfaces($responses, $printer);
	parse_local_accounts($responses, $printer);
	parse_mpls($responses, $printer);
	parse_qos($responses, $printer);
	parse_routing($responses, $printer);
	parse_services($responses, $printer);
	parse_snmp($responses, $printer);
	parse_stp($responses, $printer);
	parse_static_routes($responses, $printer, $subnets);
	parse_vlans($responses, $printer);
	parse_vtp($responses, $printer);

	$printer->close_model;

	# Close the model output file handle
	close_model_filehandle($filehandle);
}

sub _connect {

	# Grab our arguments
	my $connection_path = shift;

	# Create a new CLI protocol object
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	# Make a connection to and successfully authenticate with the IOS device
	my $enable_prompt_regex = ZipTie::Adapters::Cisco::IOS::AutoLogin::execute($cli_protocol, $connection_path);

	# Store the found prompt as "enablePrompt" on the specified CLI protocol.
	$cli_protocol->set_prompt_by_name('enablePrompt', $enable_prompt_regex);

	# Return the created ZipTie::CLIProtocol object and the enable prompt encountered after successfully connecting to a device.
	return ($cli_protocol, $enable_prompt_regex);
}

# Given the output of 'show file systems', this method returns an array of the file systems
sub _get_file_systems {
	my $showFileSystems = shift;
	my @fs;

	while ($showFileSystems =~ /^\*?\s+\d+\s+\d+\s+\b(\S+)\b\s+[A-Za-z]+\s+(\S+):/mg) {
		my ($type, $name) = ($1, $2);
		if ($type !~ /opaque|nvram/i) {
			push(@fs, $name);
		}
	}

	return @fs;
}

sub get_command_responses {
	my ($cli_protocol, $enable_prompt_regex, $connection_path) = @_;
	
	my $responses = {};

	$responses->{version} = $cli_protocol->send_and_wait_for('show version', $enable_prompt_regex);
	$responses->{startup_config} = get_startup_config($cli_protocol, $connection_path);
	$responses->{snmp} = $cli_protocol->send_and_wait_for('show snmp', $enable_prompt_regex);
	$responses->{install_run} = $cli_protocol->send_and_wait_for('show install running', $enable_prompt_regex);
	$responses->{show_fs} = $cli_protocol->send_and_wait_for('show file systems', $enable_prompt_regex);
	foreach my $fileSys ( _get_file_systems( $responses->{show_fs} ) ) {
		$responses->{file_systems}->{$fileSys} = $cli_protocol->send_and_wait_for("show $fileSys", $enable_prompt_regex);
	}
	# On switches and 7600 routers execute...
	if ($responses->{version} =~ /\bWS-C\d{4}|Cisco\s*76\d{2}/i) {
		
		my $shMod = $cli_protocol->send_and_wait_for( 'show module', $enable_prompt_regex );

		# Only set {show_module} if the command was successfull
		if ( $shMod !~ /Invalid|Incomplete/mi )	{
			$responses->{show_module} = $shMod;
		}
		
		$responses->{show_power}     = $cli_protocol->send_and_wait_for( 'show power',     $enable_prompt_regex );
		$responses->{show_inventory} = $cli_protocol->send_and_wait_for( 'show inventory', $enable_prompt_regex );
		$responses->{show_mod_ver}   = $cli_protocol->send_and_wait_for( 'show mod ver',   $enable_prompt_regex );
	} else {

		# Try show diagbus first
		my $diagbus = $cli_protocol->send_and_wait_for('show diagbus', $enable_prompt_regex);
		if ($diagbus =~ /Invalid/mi) {

			# Use 'show diag' when diagbus fails
			my $diag = $cli_protocol->send_and_wait_for('show diag', $enable_prompt_regex);
			if ($diag !~ /Invalid|Incomplete/mi) {
				$responses->{show_diag} = $diag;
			}
		} else {
			$responses->{show_diag} = $diagbus;
		}
	}
	$responses->{interfaces} = $cli_protocol->send_and_wait_for('show interfaces', $enable_prompt_regex);
	$responses->{ospf_ints} = $cli_protocol->send_and_wait_for('show ip ospf interface', $enable_prompt_regex);
	# Build up a hash of EIGRP state for parse_dynamic_routes to work with
	$responses->{eigrp}->{interfaces} = $cli_protocol->send_and_wait_for('show ip eigrp interface', $enable_prompt_regex);
	$responses->{eigrp}->{neighbors} = $cli_protocol->send_and_wait_for('show ip eigrp neighbors', $enable_prompt_regex);
	$responses->{eigrp}->{routes} = $cli_protocol->send_and_wait_for('show ip route eigrp', $enable_prompt_regex);
	$responses->{eigrp}->{topology} = $cli_protocol->send_and_wait_for('show ip eigrp topology', $enable_prompt_regex);
	$responses->{protocols} = $cli_protocol->send_and_wait_for('show ip protocols', $enable_prompt_regex);
	$responses->{ospf}->{routes} = $cli_protocol->send_and_wait_for('show ip route ospf', $enable_prompt_regex);
	# Only supported in IOS >= 12.4
	my $nwObjResponse = $cli_protocol->send_and_wait_for('show object-group', $enable_prompt_regex);
	$responses->{object_groups} = $nwObjResponse;
	$responses->{access_lists} = $cli_protocol->send_and_wait_for('show access-lists', $enable_prompt_regex);
	$responses->{standby} = $cli_protocol->send_and_wait_for('show standby', $enable_prompt_regex);
	$responses->{vrrp} = $cli_protocol->send_and_wait_for('show vrrp', $enable_prompt_regex);
	$responses->{glbp} = $cli_protocol->send_and_wait_for('show glbp', $enable_prompt_regex);
	$responses->{protocols} = $cli_protocol->send_and_wait_for('show ip protocols', $enable_prompt_regex);
	$responses->{stp} = $cli_protocol->send_and_wait_for('show spanning-tree', $enable_prompt_regex);
	$responses->{show_ip_route} = $cli_protocol->send_and_wait_for('show ip route', $enable_prompt_regex);
	$responses->{vlans} = $cli_protocol->send_and_wait_for('show vlan', $enable_prompt_regex);
	$responses->{vtp_status} = $cli_protocol->send_and_wait_for('show vtp status', $enable_prompt_regex);
	
	# Disconnect from the specified device
	disconnect($cli_protocol);
	
	return $responses;
}


sub clean_responses {
	my $responses = shift;

	foreach my $command ( keys %$responses ) {
		$responses->{$command} =~ s/[\b]\s?//mig;
		$responses->{$command} =~ s/^\s--More--(\s+)?//mg;
	}
}

1;

__END__

=head1 NAME

ZipTie::Adapters::Cisco::IOS - Adapter for performing various operations against Cisco IOS devices.

=head1 SYNOPSIS

    use ZipTie::Adapters::Cisco::IOS;
	ZipTie::Adapters::Cisco::IOS::backup( $backup_document );

=head1 DESCRIPTION

This module represents an adapter that can be used to perform various operations against against Cisco IOS devices.

Generally you would run this module through the ZipTie C<invoke.pl> script.

=head1 PUBLIC SUB-ROUTINES

=over 12

=item C<backup($backup_document)>

Performs the backup of the device that is described in the specified XML document.  This XML document contains
a "connectionPath" element that contains the IP/hostname, protocol, credential, and file server information that
may be needed to connect to, authenticate with, and back up the device in question.  This XML will be parsed into
a C<ZipTie::ConnectionPath> object to help easily access this vital information.

The act of backing up a device means that information about a device will be retreived, as well as the configuration
data/files for that device.  All of this collected information will be used to populate the ZipTie Element Document,
which is an open and extensible format for describing a network device.

The return value for this method will be a string containing the successfully populated ZipTie Element Document.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_connect($connection_path)>

Creates the initial CLI connection to the device.  A list is returned containing two elements: the first is a
C<ZipTie::CLIProtocol> object that is a CLI client that can communitcate with the device, and the second is a
regular expression that can be used to match the primary prompt of the device.  This is useful when sending
commands to and receiving responses from the device through the C<ZipTie::CLIProtocol> object and being able to
know when a command has generated all the output and returned back to the primary prompt.

=item C<_get_file_systems($show_file_systems_response)>

Parses the response of the "show file systems" and returns an array representing the file system of the Cisco Security
Appliance device.

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

Contributor(s): rkruse, Dylan White (dylamite@ziptie.org)
Date: August 10, 2007

=cut
