package ZipTie::Adapters::Cisco::IOS::AddressTranslationParser;

# Great, brief documentation on IOS NAT
# http://www.cisco.com/en/US/technologies/tk648/tk361/tk438/technologies_white_paper09186a0080091cb9.html

use 5.010_000;
use strict;
use warnings;

use Data::Dumper;
$Data::Dumper::Terse = 1; # Do not print $VAR1 before each dumped variable
$Data::Dumper::Sortkeys = 1; # Sort hash keys in dumped hashes

use ZipTie::Adapters::Utils qw(get_crep int_port invert_mask mask_to_bits);
use ZipTie::Addressing::Subnet;
use ZipTie::Logger;
use ZipTie::Model::AddressTranslation;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

# Set logging to verbose for extra NAT parsing logs
#$LOGGER->set_logging_level(-1);

use Exporter qw(import);
our @EXPORT_OK = qw(parseAddressTranslation);

sub parseAddressTranslation {
	my ($args) = @_;

	# Handle arguments
	my $in               = $args->{in};
	my $interfaces       = $args->{interfaces};
	my $interface_mapper = $args->{interface_mapper};
	my $mapper_context   = $args->{mapper_context};
	my $only_direction   = $args->{only_direction} // 'both'; # Not currently passed as an argument, will always be 'both'
	my $startup_config   = $in->{startup_config};

	# Allow for "any" address here. "Careful Ned, careful now". '|any' can cause gotchas with regexes.
	my $ip_regex_any = get_crep('cipm') . '|any';

	# Our array of address translation objects to return
	my @address_translations;
	my $address_translation;

	# There are two "zones" in IOS, inside and outside.
	my @sides = qw(inside outside);

	foreach my $side (0..1) {
		foreach my $direction ('source', 'dest') {
			NAT_RULE: while( $in->{startup_config} =~ /(^ip nat\s+$sides[$side]\s+$direction\S*\s+([a-z]+)\s+(.+?))\n/mig ) {
				my ( $line, $definition, $blob ) = ( $1, $2, $3 );

				if( not defined $address_translation ) {
					$address_translation = _create_address_translation($sides[$side], $sides[$side-1], $interfaces, $direction, $only_direction);
				}

				if( $definition ~~ 'static') {
					my $rule = _parse_static_nat( $line, $blob, $direction, $interface_mapper, $mapper_context );

					if( defined $rule ) {
						# Add the created NAT rule to the address translation obect
						$rule->{rule} = $address_translation->addRule( $rule );
						$address_translation->addRuleAddressCriterion( $rule );
					}
				} elsif( $definition ~~ 'list' ) {
					my $rule = _parse_dynamic_nat( $line, $blob, $direction, $interface_mapper, $mapper_context, $startup_config );

					if( defined $rule ) {
						# Now we are in a position to allocate the rule and apply the address criteria to it.
						# This may end up resulting in a different numbering and ordering of generated rule entries,
						# but the point of the abstraction maintained into AddressTranslation is to shield the
						# per-config-regime caller from having to sweat that.
						my $rule_handle = $address_translation->addRule( {protocol => 'ip'} );

						# Set dynamic NAT to not allow reverse connection opens. This is explicitly controlled through this method
						$address_translation->setEstablishedReverseConnectionTrafficOnly($rule_handle, 'true');

						# We have separate sets of elements to add - address translation and criterion rule data.
						foreach my $atRule ( @{ $rule->{translations} } ) {
							$atRule->{rule} = $rule_handle;
							$atRule->{usePATOverload} = $rule->{pat_enabled} // 0;
							$address_translation->addRuleAddressTranslationMapping( $atRule );
						}

						foreach my $critHash ( @{ $rule->{criterion} } ) {
							$critHash->{rule} = $rule_handle;
							$address_translation->addRuleAddressCriterion( $critHash );
						}
					}

				# TODO We do not handle route-maps currently
				} elsif( $definition ~~ "route-map" ) {
					$LOGGER->warning($NAT_RULE_DROPPED, "Dropping rule, route-map rules are not supported by QRM: $line\n");
				}
			}

			if (defined $address_translation) {
				push @address_translations, $address_translation;
				undef $address_translation;  # it's in the list now, reset
			}
		}
	}

	return \@address_translations;
}

# Creates a ZipTie::Model::AddressTranslation object
sub _create_address_translation {
	my ($side_a, $side_b, $interfaces, $direction, $only_direction) = @_;

	my $address_translation = ZipTie::Model::AddressTranslation->new($side_a, $side_b);

	# Set interface zone mapper for address translator
	$address_translation->setInterfaceZoneMapper( sub { return $interfaces } );

	# TODO For TCP and UDP the PAT range should start at 1
	# Below it starts at 0, which should only be for ICMP

	# Add policy for IOS "overload" PAT ranges
	$address_translation->addPortTranslationRange( (0,511) );
        $address_translation->addPortTranslationRange( (512,1023) );
        $address_translation->addPortTranslationRange( (1024,65535) );

	# Throttle any translation that may come about from direction
	given($only_direction) {
		when ('source') { $address_translation->setSourceTranslateOnly(); }
		when ('destination') { $address_translation->setDstTranslationOnly(); }
	}

	$address_translation->setTranslationName("$side_a\_$direction\_NAT");

	return $address_translation;
}

sub _parse_static_nat {
	my ($line, $blob, $direction, $interface_mapper, $mapper_context) = @_;
	my $ip_regex_any = get_crep('cipm') . '|any';

	my $rule;

	# Build a verbose log statement for the parsing of this NAT rule
	my $logging = "Parsing NAT line: '$line'\nNAT desc: static;";

	# esp - special case for navigation across IPSec VPN tunnels, added in 12.2(13)T
	if( $blob && $blob =~ /^(udp|tcp|esp)?\s*(.+)$/i ) {
		my $protocol = $1 // 'ip';
		$blob = $2;

		if( $protocol ) {
			$logging .= "$protocol;";
		}

		# Our collection of values to be passed to the address translation object
		$rule = {
			protocol           => $protocol,
			isStaticNATMapping => 1
		};

		# The unique case of static network-based NAT.  This basically says, map network a to network b
		# statically, using the governing mask of the numeric prefix length as generated.
		if( $blob =~ /^network\s+($ip_regex_any)\s+($ip_regex_any)\s*(\/\s*(\d+))?\s*(.*)$/ )
		{
			$rule->{ $direction . 'Ip' } = $1;
			$rule->{ $direction . 'AddrMapped' } = $2;

			# Note the prefix is used identically on the mapped and criterion addresses
			# (static network-based NAT), so its being assigned to two keys intentionally
			if( defined $3 ) {
				$rule->{ $direction . 'AddrMaskMapped' } = $4;
				$rule->{ $direction . 'Prefix' } = $4 || '0.0.0.0';
			}

			$blob = $5;
			$logging .= 'network;';
		}
		elsif( $blob =~ /^interface\s+(\S+)\s+(.+)$/ )
		{
			$rule->{ $direction . 'Ip' } = $interface_mapper->($1, $mapper_context);
			$blob = $2;
			$logging .= 'interface;';
		}
		elsif( $blob =~ /^($ip_regex_any)\s+(.+)$/ )
		{
			$rule->{$direction . 'Ip'} = $1;
			$blob = $2;
			$logging .= 'address;';

			if( $blob =~ /^(\S+)\s+(($ip_regex_any).+)$/ ) {

				$rule->{ $direction . 'ServicePort' } = int_port($1);
				$blob = $2;

				$logging .= 'port;';
			}
		} else {
			$logging .= "\nrule = " . Dumper($rule);
			$LOGGER->verbose($logging);
			$LOGGER->warning($NAT_RULE_DROPPED, "Unable to detect first address at '$blob'\n");
			return;
		}

		if( $blob ) {
			# Start with the possibility of a static mapping constrained by route map
			if( $blob =~ /^($ip_regex_any)\s+route-map\s+(\S+)\s*(.*)$/ ) {
				$rule->{ $direction . 'AddrMapped' } = $1;

				# TODO Do something with route map, analyze it against $in->{routeMapFilters}
				my $staticNATRouteFilter = $2;

				$blob = $3;
				$logging .= 'route-map;';
				$logging .= "\nrule = " . Dumper($rule);
				$LOGGER->verbose($logging);

				$LOGGER->warning($NAT_RULE_DROPPED, "Dropping rule, route-map rules are not supported by QRM: $line\n");
				return;

			} elsif( $blob =~ /^interface\s+(\S+)\s*(.*)$/ ) {

				$rule->{ $direction . 'AddrMapped' } = $interface_mapper->($1, $mapper_context);
				$blob = $2;
				$logging .= 'second interface;';

			} elsif( $blob =~ /^($ip_regex_any)\s*(.*)$/ ) {

				$rule->{ $direction . 'AddrMapped' } = $1;
				$blob = $2;
				$logging .= 'second address;';

				if( $blob && $blob =~ /^(\S+)\s*(.*)$/ ) {
					my $possible_port = $1;
					my $tmp_blob = $2; # do not shorten blob until definitly port below

					# Attempt to reslove the string where the port number / expression should be
					if( $possible_port !~ /^\d+$/ ) {

						my $resolved_possible_port = int_port($possible_port);

						if( $possible_port ne $resolved_possible_port ) {
							$rule->{ $direction . 'PortMapped' } = $resolved_possible_port;
						}
					} else {
						$rule->{ $direction . 'PortMapped' } = $possible_port;
					}
					if( defined $rule->{ $direction . 'PortMapped' } ) {
						$logging .= 'second port; ';
						$blob = $tmp_blob;
					}
				}
			}
		}

		# Log our parsed NAT rule verbosely
		$logging .= "\nrule = " . Dumper($rule);
		$LOGGER->verbose($logging);

		if( not defined $rule->{ $direction . 'AddrMapped' } ) {
			$LOGGER->warning($NAT_RULE_DROPPED, "Unable to detect second address at '$blob'\n");
			return;
		}

		if( $blob ) {

			# Do not use "extendable" rules
			if( $blob =~ /extendable/ ) {
				# Used to set as exemptions, but as far as I can see this just
				# reorders the rule to the front of the NAT list.
				# $address_translation->setExemptionRule( $rule->{rule} );

				$LOGGER->warning($NAT_RULE_DROPPED, "Dropping rule, extendable rules are not supported by QRM: $line\n");
				return;
			}
		}
	}

	return $rule;
}

sub _parse_dynamic_nat {
	my ($line, $blob, $direction, $interface_mapper, $mapper_context, $startup_config) = @_;

	my $rule;

	# Build a verbose log statement for the parsing of this NAT rule
	my $logging = "Parsing NAT line: '$line'\nNAT desc: dynamic;list;";

	# Parse the access list
	if( $blob =~ /^(\S+)\s+(.+)$/ ) {
		my $list_name = $1;
		$blob = $2;

		$rule->{criterion} = _parse_nat_access_list( $startup_config, $list_name, $direction );

		if (not defined $rule->{criterion}) {
			$LOGGER->warning($NAT_RULE_DROPPED, "Unable to find or parse NAT access list '$list_name'");
			return;
		}
	}

	# Parse the interface or pool to translate to
	if( $blob =~ /^interface\s+(\S+)\s*(.*)$/ ) {
		my $interface_name = $1;
		$blob = $2; # might have matched nothing

		$logging .= 'interface;';

		# Call caller's passed interface name -> addr function
		my $address = $interface_mapper->($interface_name, $mapper_context);

		if( defined $address ) {
			push @{$rule->{translations}},  { $direction . 'AddrMapped' =>  $address };
		} else {
			$LOGGER->warning($NAT_RULE_DROPPED, "Unable to find address for interface '$interface_name'");
			return;
		}

	} elsif( $blob =~ /^pool\s+(\S+)\s*(.*)$/ ) {
		my $pool_name = $1;
		$blob = $2; # might have matched nothing

		$logging .= 'pool;';

		$rule->{translations} = _parse_nat_pool($startup_config, $pool_name, $direction);
	}

	if( $blob && $blob =~ /overload/) {
		$rule->{pat_enabled} = 1;
		$logging .= 'pat;';
	}

	# Log our parsed NAT rule verbosely
	$Data::Dumper::Terse = 0;
	$Data::Dumper::Varname = 'criteria';
	$logging .= "\n" . Dumper( @{ $rule->{criterion} } );
	$Data::Dumper::Varname = 'translation';
	$logging .= Dumper( @{ $rule->{translations} } );
	$Data::Dumper::Terse = 1;
	$LOGGER->verbose($logging);

	return $rule;
}

# Parsing for the pool by name, entry of its data into a list suitable for calling
# AddressTranslation::addRuleTranslationMapping with
sub _parse_nat_pool {
	my ($rc, $pool_name, $direction) = @_;
	my $ip_regex = get_crep('cipm');

	my $entry;
	my @entry_list;
	my $search_for_addresses = 0;
	my $netmask;

	foreach my $line ( split /\n/, $rc ) {

		# There are at least two addresses in each standard pool entry
		if( $line =~ /^(ip\s+nat\s+pool\s+$pool_name\s+($ip_regex)\s+($ip_regex)\s*(.*))$/ ) {
			my $pool = $1;
			$entry->{$direction . 'AddrMapped'}    = $2;
			$entry->{$direction . 'AddrEndMapped'} = $3;

			my $blob = $4;

			$LOGGER->verbose("Parsing pool $pool_name: '$pool'");

			# See if there are any trailing arguments to the nat pool entry
			if( defined $blob and $blob =~ /^(\S+)\s+($ip_regex|\d+)\s*(.*)$/ ) {

				# These map to the same param
				my $mask_type = $1;
				if( $mask_type eq 'netmask' or $mask_type eq 'prefix-length' or $mask_type eq 'prefix' ) {
					$entry->{$direction . 'AddrMaskMapped'} = $2;
				}

				$blob = $3;
				if( defined $blob and $blob =~ /type\s+([\w\-]+)/ ) {
					my $option = $1;

					# Could also be 'no-alias' or 'rotary' but we do not care about their effect
					if( $option eq 'match-host' ) {
						$entry->{useAppliedAddressRules} = 'true';
					}
				}
			}
			push @entry_list, $entry;
			undef $entry;

		# The next two elsifs support 'flexible pool configuration', a pool defined over multiple lines with an IP range on each
		} elsif( $line =~ /^(ip\s+nat\s+pool\s+$pool_name\s+(\S+)\s+($ip_regex|\d+)\s*(.*))$/ && !$search_for_addresses ) {
			my $pool = $1;
			my $mask_type = $2;

			undef $netmask;
			if( $mask_type eq 'netmask' or $mask_type eq 'prefix-length' or $mask_type eq 'prefix' ) {
				$netmask = $3;
			}

			my $blob = $4;
			$search_for_addresses = 1;

			$LOGGER->verbose("Parsing multi-line pool $pool_name: '$pool'");

		# Found the beggining of a pool defined over multiple lines, interpret them as such
		} elsif( $search_for_addresses ) {
			if( $line =~ /^(\s*address\s+($ip_regex)\s+($ip_regex)\s*)$/ ) {
				my $pool_line = $1;
				$entry->{$direction . 'AddrMapped'}    = $2;
		                $entry->{$direction . 'AddrEndMapped'} = $3;

				if( defined $netmask ) {
					$entry->{$direction . 'AddrMaskMapped'} = $netmask;
				}

				$LOGGER->verbose("'$pool_line'");

				# Add this range to the list
				push @entry_list, $entry;
				undef $entry;
				
			# If the line is not an address line or a comment line, reset the logic
			} elsif( $line !~ /^\s*!/ ) {
				$LOGGER->verbose("End of multi-line pool");
				$search_for_addresses = 0;
			}
		}
	}

	return \@entry_list;
}

sub _parse_nat_access_list {
	my ($rc, $list_name, $direction) = @_;
	my $ip_regex = get_crep('cipm');
	my $inverse_direction = _resolve_inverse_direction( $direction );

	my @criterion_list;

	# This could be a single line entry, or a multiline extended entry.  So, let's start by iterating on all access lists
	# so identified, and it is safe to snarf that up to create a blob for the line(s) which follow.  Pick up EVERYTHING
	# which follows up to the next new line which *isn't* followed by white space.  Make that last check ungreedy.
	while( $rc =~ /^((ip\s+)?access-list\s+(extended|dynamic|standard)?\s*$list_name\s+(.+?)(?=\n\S))/smig) {
		my $access_list = $1;
	 	my $type = $3 // _resolve_access_list_type($list_name);
		my $blob = $4;

		$LOGGER->verbose("Parsing access-list $list_name: '$access_list'");

		while( $blob =~ /(permit|deny)\s+(tcp|udp|ip|esp|icmp)?\s*([\w\s\.]+?)$/mig ) {
			my ($action, $inner_blob) = ($1, $3);
			my $criterion = {ipProtocol => $2 // 'ip'};

			if( $action =~ /deny/ ) {
				$criterion->{negateMatch} = 1;
			}

			# Parse the extra extended part to begin with
			if( $type and $type ~~ 'extended' ) {
				if( $inner_blob =~ /^any\s+([\w\s\.]+)/ ) {
					$criterion->{ $direction . 'Ip' } = 'any';
					$inner_blob = $1;
				
					# we've moved changed the blob so the non-extended parsing code can deal with it

				} elsif( $inner_blob =~ /^host\s+($ip_regex)\s*(.*)/ ) {

					$criterion->{ $direction . 'Ip' } = $1;
					$inner_blob = $2;  # pick up what we can here...

					# trim off "any"
					if( $inner_blob =~ /^any\s*(.*)/ ) {
						$inner_blob = $1;

					} elsif( $inner_blob =~ /^host\s+($ip_regex)\s*(.*)/ ) {

						$criterion->{ $inverse_direction . 'Ip' } = $1;
						$inner_blob = $2;
					}
				}
			}

			$criterion = _parse_nat_access_rule( $criterion, $inner_blob, $direction );

			if( defined $criterion ) {
				push @criterion_list, $criterion;
			}
		}
	}

	return \@criterion_list;
}

sub _parse_nat_access_rule {
	my ( $criterion, $blob, $direction ) = @_;
	my $ip_regex = get_crep( 'cipm' );
	my $inverse_direction = _resolve_inverse_direction( $direction );

	foreach my $current_direction ( $direction, $inverse_direction ) {
		if( not defined $criterion->{ $current_direction . 'Ip' } && not defined $criterion->{ $current_direction . 'IpObjectGroup' } ) {

			# Parse IP address (the one required element here)
			if( $blob =~ /^($ip_regex)(\s+($ip_regex))?\s*([\w\s\.]*)?/ ) {

				$criterion->{ $current_direction . 'Ip' } = $1;

				if( defined $2 ) {
					# TODO Handle a mask given in hex format, this can be set with "ip netmask-format hexadecimal"

					# Cisco inverse order handling
					my $prefix = $3;
					$prefix = mask_to_bits(invert_mask($prefix));

					if( $prefix != 32 ) {
						$criterion->{ $current_direction . 'Prefix' } = $prefix || '0.0.0.0';
					}
				}

				$blob = $4;

			} elsif( $blob =~ /^(\w+)\s*([\w\s\.]*)?/ ) {
				$blob = $2;
				if( $1 ~~ 'any' ) {
					$criterion->{ $current_direction . 'Ip' } = 'any'; # AddressTranslation can in fact deal with this
				} elsif( $1 ~~ 'host' ) {
					if ($blob =~ /^\s*($ip_regex)\s*(.*)/) {
						$criterion->{ $current_direction . 'Ip' } = $1;
						$blob = $2;
					}
				} else {
					$criterion->{ $current_direction . 'IpObjectGroup' } = $1;
				}
			}
		}

		# Now go for service/port
		if( defined $blob ) {
			if( $blob =~ /^(eq|lt|gt|ne)\s+(\S+)\s*(.*)$/) {
				my $operator = $1;
				my $port = int_port($2);
				$blob = $3;

				if( $port !~ /^\d+$/ ) {
					my $resolved_port = int_port($port);

					if( $port != $resolved_port ) {
						$port = $resolved_port;
					} else {
						$LOGGER->warning($INVALID_PORT, "Unable to parse or map presumed NAT filter entry source port/service: '$port'");
						undef $port;
					}
				}

				if( defined $port ) {
					$criterion->{ $current_direction . 'ServicePort' } = $port;

					if( $operator ne 'eq' ) {
						$criterion->{ $current_direction . 'ServicePortOperator' } = $operator;
					}
				}

			} elsif( $blob =~ /^range\s+(\d+)\s+(\d+)\s*(.*)/ ) {
				my ($port_start, $port_end) = ($1, $2);
				$blob = $3;

				$criterion->{ $current_direction . 'ServicePort' } = $port_start;
				$criterion->{ $current_direction . 'ServicePortEnd' } = $port_end;
			}
		}

		if( not defined $criterion->{ $direction . 'Ip' } && not defined $criterion->{ $direction . 'IpObjectGroup' }) {
			$LOGGER->warning($INVALID_CONFIGURATION, "Unable to pick out access-list NAT $direction at '$blob'");
			return;
		}
	}

	return $criterion;
}

# If passed the current direction, either 'source' or 'dest', returns the other
sub _resolve_inverse_direction {
	my $direction = shift;

	if( $direction eq 'source' ) {
		return 'dest';
	}

	return 'source';
}

# Figures out the "implied" type of a numerically identified access list
sub _resolve_access_list_type {
	my $id = shift;
	my $type;

	if( (1 <= $id && $id <= 99) or (1300 <= $id && $id <= 1399) ) {
		$type = 'simple';
	} elsif( (100 <= $id && $id <= 199) or (2600 <= $id && $id <= 2699)) {
		$type = 'extended';
	}
	
	return $type;
}

=head1 NAME

    ZipTie::Adapters::Cisco::IOS::AddressTranslationParser - Parser for IOS NAT and PAT.

=head1 VERSION

    This document refers to version 1.1 of ZipTie::Adapters::Cisco::IOS::AddressTranslationParser, released 2015 Q4.

=head1 SYNOPSIS

  use ZipTie::Adapters::Cisco::IOS::AddressTranslationParser;
  my $addressTranslation = Ziptie::Model::AddressTranslation->new();
  parseAddressTranslation(
    in                 => $in,
    interfaces         => parse_interfaces($in),
    addressTranslation => $addressTranslation);
  
  $addressTranslation->generateTranslationList($out);

=head1 DESCRIPTION

    This module scans the elements of Cisco IOS NAT, and handles "ip nat"
    configuration statements, nat pools, and application to interfaces.
    It sets those up by allocating and acting as the compositional
    interface point to Ziptie::Model::AddressTranslation.

=head2 Overview

   Mostly provides the interface to do NAT parsing, plus callbacks and
   interface points to the Ziptie::Model::AddressTranslation it gets
   at parse time.

=head2 Constructor and initialization

    None; this is an interface package

=head2 Class and object methods

    None; this is an interface package

=over

=item "parseAddressTranslation"
    routine which will parse a running config for address translation
    attributes. It can be limited to looking for source or destination
    address and port translations.

    Parameters:  (passed as named interfaces in hash, "$args")

=over 6

=item * in => running configuration of device (required)

=item * interfaces => output from parse_interfaces() in IOS parser (required)

=item * interface_mapper => reference to function which will be passed through/called back as intefaceZoneMapper for AddressTranslation object

=item * interface_mapper_context =>

=item * only_direction => "source" or "destination" will cause the parsing and generation to be focused entirely on source NAT or destination NAT. (optional).

=back

    Returns: list of AddressTranslation object instances reflecting NAT configuration detected
    in $args->{in}->{running-config}.  There is roughly one for all "ip nat source..." rules and
    one for all "ip nat [inside|outside] [source | destination]..." rule set permutation, leading
    to a max of four AddressTranslation instances in the list.

=back

=head1 ENVIRONMENT

=head1 DIAGNOSTICS

=over

=item "error message that may appear"
    Explanation of error message

=back

=head1 BUGS
    Description of known bugs (and any work-arounds).
    Usually also includes an invitation to send the author bug reports.

=head1 FILES
     Ziptie::Model::AddressTranslation

=head1 SEE ALSO

    http://wiki.q1labs.inc/display/QRM/Address+Translation+Support+in+ZipTie+and+CSM

=head1 AUTHOR(S)
    Wayne Tackabury
    wtackabury@us.ibm.com

    Jak Devaney
    john.devaney@uk.ibm.com

=head1 COPYRIGHT
    Copyright (c) 2012, IBM Corporation. All Rights Reserved.
    This module is free software. It may be used, redistributed
    and/or modified under the same terms as Perl itself.
=cut

1;
