package ZipTie::Adapters::Cisco::IOS::Telemetry;

use strict;
use warnings;
use 5.010_000;

use ZipTie::Adapters::Cisco::IOS::Parsers qw(parse_arp parse_cdp parse_telemetry_interfaces parse_mac_table parse_routing_neighbors);
use ZipTie::Adapters::Cisco::IOS::AutoLogin;
use ZipTie::Adapters::Cisco::IOS::Disconnect qw(disconnect);
use ZipTie::Adapters::Utils qw(choose_admin_ip get_model_filehandle close_model_filehandle);
use ZipTie::CLIProtocolFactory;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

use Exporter qw(import);
our @EXPORT_OK = qw(open_telemetry_neighbors close_telemetry_neighbors);

my $LOGGER = ZipTie::Logger::get_logger();

sub invoke {
	my ($pkg, $connectionPathDoc) = @_;
	
	# Setup work
	my ($connectionPath, $discoveryParams) = ZipTie::Typer::translate_document($connectionPathDoc, 'connectionPath');
	my $filehandle = get_model_filehandle('IOS', $connectionPath->get_ip_address());
	my $cliProtocol = ZipTie::CLIProtocolFactory::create($connectionPath);
	my $responses = {};
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, 'telemetry');
	$printer->attributes(1);
	$printer->open_discovery_event();

	# Make a connection to device and get the data necessary to fill out the discovery event document 
	my $promptRegex = ZipTie::Adapters::Cisco::IOS::AutoLogin::execute($cliProtocol, $connectionPath);
	$cliProtocol->send_and_wait_for('terminal len 0', $promptRegex);
	open_telemetry_neighbors($responses, $printer);
	
	# Process the ARP table
	$responses->{arp} = $cliProtocol->send_and_wait_for('show ip arp', $promptRegex);
	parse_arp($responses, $printer);
	delete $responses->{arp};

	# Process the CDP neighbors
	$responses->{cdp} = $cliProtocol->send_and_wait_for('show cdp neighbors detail', $promptRegex);
	parse_cdp($responses, $printer);
	delete $responses->{cdp};
	
	# Process the MAC address table 
	$responses->{mac} = $cliProtocol->send_and_wait_for('show mac address-table dynamic', $promptRegex);
	parse_mac_table($responses, $printer);
	delete $responses->{mac};
	
	# Process the routing neighbors
	$responses->{ospf}->{neighbors} = $cliProtocol->send_and_wait_for('show ip ospf neighbor', $promptRegex);
	$responses->{eigrp}->{neighbors} = $cliProtocol->send_and_wait_for('show ip eigrp neighbors', $promptRegex);
	$responses->{bgp}->{neighbors} = $cliProtocol->send_and_wait_for('show ip bgp neighbors', $promptRegex);
	parse_routing_neighbors($responses, $printer);
	delete $responses->{ospf};
	delete $responses->{eigrp};
	delete $responses->{bgp};
	
	close_telemetry_neighbors($responses, $printer);
	
	# Process the interfaces and their status
	$responses->{interfaces} = $cliProtocol->send_and_wait_for('show interfaces', $promptRegex);
	$responses->{standby} = $cliProtocol->send_and_wait_for('show standby', $promptRegex);
	my $interfacesHash = parse_telemetry_interfaces($responses, $printer);
	delete $responses->{interfaces};
	
	my $adminIp = $connectionPath->get_ip_address();
	if ($discoveryParams->{calculateAdminIp} eq 'true') {
		$adminIp = choose_admin_ip($connectionPath->get_ip_address(), $interfacesHash);
	}
	$printer->print_element('adminIp', $adminIp);
	
	# Tear down work
	$printer->close_element('DiscoveryEvent');
	close_model_filehandle($filehandle);
	disconnect($cliProtocol);
}

# Routines to allow external call to open and close the telmetry-level "neighbors" elements
sub open_telemetry_neighbors {
	my ($in, $out) = @_;
	$out->attributes(1);
	$out->open_element('neighbors');	
}

sub close_telemetry_neighbors {
	my ($in, $out) = @_;
	$out->close_element('neighbors');
}

1;
