# ZipTie::Nexus::AutoLogin : Implement login and access interface for Nexus
#
# Author:	Wayne Tackabury (wtackabury@us,ibm.com)
package ZipTie::Adapters::Cisco::Nexus::AutoLogin;

use strict;

use v5.10;
use ZipTie::CLIProtocol;
use ZipTie::Response;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Netconf qw(get_netconf_message_termination_seq);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

###########################################
#
# PUBLIC METHODS
#
###########################################

# Connects to and authenticates with an NxOS-based device that is being connected to by a specified CLIProtocol implementation.
# Attempt to handle all of the edge cases that are encountered and successfully put the device into
# exec mode.  If all is successful, the device prompt is parsed out and is returned to be used as the primary
# regular expression to wait for when sending commands.
#
# Input:	$cli_protocol -	The CLIProtocol object that is being used to communicate with the device.
#			$connection_path -	Reference to a ZipTie::ConnectionPath object that contains the IP address
#								and any credentials that would need to be used for the current connection.
#
# Returns:	The regular expression that will match the device's exec mode prompt that was found.
sub execute
{
	my $cli_protocol    = shift;
	my $connection_path = shift;

	# Pull off the ZipTie::Credentials instance from the ZipTie::ConnectionPath object
	my $credentials = $connection_path->get_credentials();

	# Store all the information needed to connect to the device
	my $ip_address = $connection_path->get_ip_address();
	my $port       = $cli_protocol->get_port();
	my $username   = $credentials->{username};
	my $password   = $credentials->{password};

	my $protocol_name = $cli_protocol->get_protocol_name();
	$LOGGER->debug("Protocol in use: $protocol_name");

	# Attempt to connect to the device
	$cli_protocol->connect( $ip_address, $port, $username, $password );
	$LOGGER->debug("Verifying the initial connection ...");

	# If all goes well during the login process, the last method to be called will be "_calculate_prompt_regex"
	# and its return value will be returned all the way to here
	my $prompt_regex = _initial_connection( $cli_protocol, $credentials );

	# Print that the login process has been completed
	$LOGGER->debug("Login has successfully completed!\n");
	
	$LOGGER->debug("Attempting to disable paging...");
	# Get rid of the more prompt	
	my $termLen = $cli_protocol->send_and_wait_for( 'terminal length 0', '(' . $prompt_regex . ')');
    	
    #If disable paging failed then set the more prompt
    if ($termLen =~ /Permission denied for the role/i) {
    	$LOGGER->debug("Unable to disable paging, setting the more prompt");
    	# set the --more-- prompt if the term length 0 didn't go through
       	$cli_protocol->set_more_prompt( '\e\[\d*m*--More--\e\[\d*m*\s*\n*$', '20');
    }

	# Return the prompt regular expression that was found
	return $prompt_regex;
}

# Note that this interface, based on our testing, is ONLY available via SSH/CLI, not netconf/xml.
sub connect_vdc
{
	my ($vdcInvokeString, $cli_protocol) = @_;
	my $responsePrompt = "";
	my $vdcPromptRegex = '\S+#\s*';
	

	if ( $cli_protocol->get_protocol_name() =~ /Netconf/i)
	{
		$LOGGER->fatal("Attempt to change to VDC via \' $vdcInvokeString \' over unsupported Netconf/XML transport");		
	}
	
	#  We issue the string, waiting for no password prompt, but calculating a prompt regex
	
	my @responses = (
		ZipTie::Response->new( $vdcPromptRegex, \&_calculate_vdc_regex )		
	);
	
	$cli_protocol->send($vdcInvokeString, 0);
	
	my $response = $cli_protocol->wait_for_responses( \@responses );
	
	if ($response)
	{
		if ($cli_protocol->last_match() =~ /($vdcPromptRegex)/)
		{
			$responsePrompt = $1;
		}
	}
	
	return $responsePrompt;
}
###########################################
#
# PRIVATE SUB-ROUTINE
#
###########################################

sub _initial_connection
{
	my $cli_protocol = shift;
	my $credentials  = shift;

	# Specify the responses to check for
	my @responses = ();
	# +++ wayne spend some time trying various failure modes for the device
	push( @responses, ZipTie::Response->new( 'sername:\s*',            \&_send_username ) );
	push( @responses, ZipTie::Response->new( 'assword:\s*',            \&_send_password ) );
	push( @responses, ZipTie::Response->new( '(^|\n|\r)[^#^\n^\r]+#\s*$|[^#^\n^\r]+#\s*\S+#\s*$', \&_calculate_prompt_regex ) );
	#  responses.  The first is the "hello" message.
	push (@responses, ZipTie::Response->new( '\/hello>\W+?' . ZipTie::Netconf::get_netconf_message_termination_seq(), \&_issue_netconf_hello_resp ));

	# Since we just want to see what the initial prompt of the device is, there is no reason
	# to send anything, so let's just wait for a match
	my $response = $cli_protocol->wait_for_responses( \@responses );

	# Based on the response of the device, determine the next interaction that should be executed.
	my $next_interaction = undef;
	if ($response)
	{
		$next_interaction = $response->get_next_interaction();
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}

	# Call the next interaction if there is one to call
	if ($next_interaction)
	{

		# Return the  prompt found
		return &$next_interaction( $cli_protocol, $credentials );
	}
}

sub _send_username
{
	my $cli_protocol = shift;
	my $credentials  = shift;

	# Specify the responses to check for
	my @responses = ();
	push( @responses, ZipTie::Response->new( 'invalid', undef, $INVALID_CREDENTIALS ) );
	push( @responses, ZipTie::Response->new( 'assword:\s*|PASSCODE:\s*', \&_send_password ) );

	# Send username credential
	my $username = $credentials->{username};
	if ( !$username )
	{
		$LOGGER->fatal_error_code( $INVALID_CREDENTIALS, $cli_protocol->get_ip_address(), "This device requires a username." );
	}

	$LOGGER->debug("Sending username credential ...");
	$cli_protocol->send($username, 0);
	my $response = $cli_protocol->wait_for_responses( \@responses );

	# Based on the response of the device, determine the next interaction that should be executed.
	my $next_interaction = undef;
	if ($response)
	{
		$next_interaction = $response->get_next_interaction();
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}

	# Call the next interaction if there is one to call
	if ($next_interaction)
	{

		# Return the prompt found
		return &$next_interaction( $cli_protocol, $credentials );
	}
}

sub _send_password
{
	my $cli_protocol = shift;
	my $credentials  = shift;

	# Specify the responses to check for
	my @responses = ();
	push( @responses, ZipTie::Response->new( 'sername:\s*',              undef, $INVALID_CREDENTIALS ) );
	push( @responses, ZipTie::Response->new( 'assword:\s*',              undef, $INVALID_CREDENTIALS ) );
	push( @responses, ZipTie::Response->new( 'PASSCODE:\s*',             undef, $INVALID_CREDENTIALS ) );
	push( @responses, ZipTie::Response->new( 'invalid',               undef, $INVALID_CREDENTIALS ) );
	push( @responses, ZipTie::Response->new( 'Bad passwords',         undef, $INVALID_CREDENTIALS ) );
	push( @responses, ZipTie::Response->new( 'Authentication failed', undef, $INVALID_CREDENTIALS ) );
		push( @responses, ZipTie::Response->new( '(^|\n|\r)[^#^\n^\r]+#\s*$|[^#^\n^\r]+#\s*\S+#\s*$', \&_calculate_prompt_regex ) );

	# Send password credential
	my $password = $credentials->{password};
	$LOGGER->debug("Sending password credential ...");
	$cli_protocol->send($password, 1);  # pass "secure" flag
	my $response = $cli_protocol->wait_for_responses( \@responses );

	# Based on the response of the device, determine the next interaction that should be executed.
	my $next_interaction = undef;
	if ($response)
	{
		$next_interaction = $response->get_next_interaction();
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}

	# Call the next interaction if there is one to call
	if ($next_interaction)
	{

		# Return the prompt found
		return &$next_interaction( $cli_protocol, $credentials );
	}
}

# a subset of _calculate_prompt regex
sub _calculate_vdc_regex
{
	my $cli_protocol = shift;

	# Grab the last match that the CLIProtocol implementation captured
	my $last_match = $cli_protocol->last_match();
	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[LAST MATCH CAPTURED]");
	$LOGGER->debug($last_match);

	# Use a regular expression to strip the prompt from the match
	$last_match =~ /([^\s][^\r\n]*).$/;

	# Grab the extracted device prompt
	my $device_prompt = $1;

	# Clean up the device prompt for regular expression use
	$device_prompt =~ s/^\s+//;
	$device_prompt =~ s/\s+$//;
	$device_prompt = quotemeta($device_prompt);

	# Append the device prompt for the spaces that will likely follow it
	my $device_prompt_regex = $device_prompt . "\\s*\$";

	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[REGULAR EXPRESSION TO MATCH DEVICE PROMPT]");
	$LOGGER->debug($device_prompt_regex);
	$LOGGER->debug("---------------------------------------------------------");

	return $device_prompt_regex;
}

sub _calculate_prompt_regex
{
	my $cli_protocol = shift;

	# Grab the last match that the CLIProtocol implementation captured
	my $last_match = $cli_protocol->last_match();
	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[LAST MATCH CAPTURED]");
	$LOGGER->debug($last_match);

	# Use a regular expression to strip the prompt from the match
	$last_match =~ /([^\s][^\r\n]*).$/;

	# Grab the extracted device prompt
	my $device_prompt = $1;

	# Register the device prompt with the ZipTie::Recording singleton instance module
	use ZipTie::Recording;
	my $recording = ZipTie::Recording::get_recording();
	$recording->set_device_prompt($device_prompt);

	# Clean up the device prompt for regular expression use
	$device_prompt =~ s/^\s+//;
	$device_prompt =~ s/\s+$//;
	$device_prompt = quotemeta($device_prompt);

	# Append the device prompt for the spaces that will likely follow it
	my $device_prompt_regex = $device_prompt . "\\s*\$";

	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[REGULAR EXPRESSION TO MATCH DEVICE PROMPT]");
	$LOGGER->debug($device_prompt_regex);
	$LOGGER->debug("---------------------------------------------------------");
	
	# more prompt
	$cli_protocol->set_more_prompt( '--More--\s*$', '20' );

	return $device_prompt_regex;
}

sub _issue_netconf_hello_resp
{
	my $cli_protocol = shift;
	my $lastMatch = $cli_protocol->last_match();

	# We want to be sure we're only being called on a netconf instance
	if ( $cli_protocol->get_protocol_name() !~ /Netconf/i)
	{
		$LOGGER->fatal("Netconf response invoked on non-netconf SSH session from \'" . $lastMatch . "\'");		
	}
	
	# The response we issue will have no response from the device itself.  If we matched to get here
	# however, we know what the general command response will be
	my $device_prompt_regex = '^' . ZipTie::Netconf::get_netconf_message_termination_seq();  # no more, no less 

	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[NETCONF_HELLO_REQUEST_MATCHED]");
	$LOGGER->debug($lastMatch);

	# grab session ID
	my ($ncSessionID) = ($lastMatch =~ /<session\-id>(\d+)<\/session\-id>/);
	
	if ($ncSessionID)
	{
		$cli_protocol->set_netconf_session_id($ncSessionID) if ($ncSessionID);
	}
	else
	{
		$LOGGER->debug("Cisco::AutoLogin::_issue_netconf_hello_resp : unable to identify session ID  in hello from device");			
	}
	
	# Netconf module takes care of this for us
	$cli_protocol->send_netconf_hello_response();
	
	# more housekeeping on the netconf session for our adaptation
	$cli_protocol->set_netconf_local_namespace("xmlns:nxos", "http://www.cisco.com/nxos:1.0");
	$cli_protocol->set_netconf_message_headers("nxos:exec-command", "nxos:cmd");
	
	# with no further response called for from the device, it is ready to receive commands.
	return $device_prompt_regex;
}

1;

__END__

=head1 NAME

ZipTie::Adapters::Cisco::Nexus::AutoLogin - Automates the connection and authentication with a Cisco Nexus device.

=head1 SYNOPSIS
    use ZipTie::Adapters::Cisco::Nexus::AutoLogin;
	my $cli_protocol = ZipTie::CLIProtocolFactory::create( $connection_path );
	my $prompt_regex = ZipTie::Adapters::Cisco::Nexus::AutoLogin::execute( $cli_protocol, $connection_path );

=head1 DESCRIPTION

The autologin module for an adapter provides an abstract sequence of commands and responses
that allow the user to login to the device.  Usually this is via a CLI protocol such as Telnet
or SSH.  Once fully connected and authenticated with the device, the autologin module must be able
to calculate and return the a regular expression that will be leveraged by users of the module
to match the primary prompt for the device.  This is crucial in being able to send commands and wait
for their responses where leveraging the command line interface (CLI) clients provided by the ZipTie
Perl framework, such as C<ZipTie::Telnet> and C<ZipTie::SSH>.  Particular enhancements have been made
to support C<ZipTie::Netconf>, which is netconf-over-SSH.

This module should be able to handle all the possible scenarios that might be encountered when
connecting to and authenticating with a device.

The autologin module should be abstract enough that scripts outside of the adapters can use it
as well.  For example, a script to make mass changes to a device should be able to leverage
the autologin sequence just like an adapter does during backup.

=head1 PUBLIC SUB-ROUTINES

=over 12

=item C<execute($cli_protocol, $connection_path)>

The main entry point for connecting to and authenticating with a network device.  This sub-routine
will connect to the device using the specified C<ZipTie::CLIProtocol> object as the command line interface (CLI)
client and using the hostname or IP address specified within the C<ZipTie::ConnectionPath> object.

This method must return a regular expression that can be used to match the primary prompt of the device that
has been connected to.  This is required so that users of the autologin module can have a way to match the
prompt after sending a command to a device and waiting for the response.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_inital_connection($cli_protocol, $connection_path)>

Once an inital connection has been established with the network device, the first text response over the CLI is
examined.  Depending on the response, various sub-routines may be called.

=item C<_send_username($cli_protocol, $connection_path)>

This sub-routine is called if a response from the device indicates that the "username" credential should be sent to the
device.

=item C<_send_password($cli_protocol, $connection_path)>

This sub-routine is called if a response from the device indicates that the "password" credential should be sent to the
device.

=item C<_calculate_prompt_regex($cli_protocol, $connection_path)>

This sub-routine is called once the device has been connected to and authentication has completed successfully.  The
last match from the device will be examined the primary prompt will be parsed from it and converted into a regular
expression that can be used to match the prompt of the device.  This regular expression will be returned so that it can
be used by it's caller.

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

Contributor(s): rkruse, Dylan White (dylamite@ziptie.org), Wayne Tackabury (wtackabury@us.ibm.com)
Date: November 4, 2012

=cut
