package ZipTie::Adapters::Cisco::Nexus::Parsers;

use 5.010_000;
use strict;
use warnings;

use ZipTie::Adapters::Cisco::Nexus::Utils qw(expand_interface_name get_interface_connected_to_network is_conforming_xml);
use ZipTie::Adapters::Utils qw(trim get_crep parseCIDR create_network_element get_any_address_encoding get_port_number);
use ZipTie::Addressing::Subnet;
use ZipTie::Addressing::RouteFilter;
use ZipTie::Ranges;

use Data::Dumper;
$Data::Dumper::Terse = 1; # Do not print $VAR1 before each dumped variable
$Data::Dumper::Sortkeys = 1; # Sort hash keys in dumped hashes

use MIME::Base64;
use XML::Simple;
use XML::Twig;

use Exporter 'import';
our @EXPORT_OK = qw(parse_system parse_chassis create_config parse_access_ports parse_interfaces parse_object_groups parse_access_lists parse_static_routes parse_dynamic_routes open_vdc_section close_vdc_section get_default_vrf_name parse_vlans);

use ZipTie::Logger;
my $LOGGER = ZipTie::Logger::get_logger();

use constant SUPPORTED_ROUTING_PROTOCOLS => qw(eigrp rip ospf);

# +++wayne modullarize these
# If this is XML over SSH, we issue commands with the " | xml" CLI designator
use constant NEXUS_XML_COMMANDS_XML_SSH => 1;
# if this is a real Netconf connection, then we issue Netconf requests to the port requested
use constant NEXUS_XML_COMMANDS_NETCONF => 2;
# The real kicker is that at least one device (the Nexus 4000) in fact does not support Netconf or
# XML command formatting *at all*.  We actually have to back off to a pure CLI text response mode of
# parsing, and the command as issued has no XML form or request designator (although we'll try the latter
# and glean if we got an error to see if we're in this predicament
use constant NEXUS_XML_NO_XML_TEXT_ONLY => 3;
use constant DEFAULT_VRF_NAME => "default";

my $ip_regex = get_crep('cipm');
my $cidr_regex = get_crep('cidr');

# Parse the ziptie core sections describing top level system attributes in the ZED.
#
# PARAMS: 
# $in - input context, featuring a {running_config} section
# $out - output XML writer context.
sub parse_system{
	my ($in, $out) = @_;
	my $sysData = _parse_system($in);
	my $lastReboot = undef;

	# The following vairables are required by the schema, however if they are missing it's
	# not important enough to fail the backup.  Report debug errors.
	if (!defined($sysData->{version}))
	{
		$LOGGER->debug('Failed to parse required element core:version from \'show version\'.');
	}
	if (defined($sysData->{utDays}) && defined($sysData->{utHrs}) && defined($sysData->{utMins}) && defined($sysData->{utSecs}))
	{
		# Calculate lastreboot (based of of Cisco IOS)
		$lastReboot = time();
		$lastReboot -= $sysData->{utDays} * 24 * 60 * 60 if (defined($sysData->{utDays}));
		$lastReboot -= $sysData->{utHrs} * 60 * 60 if (defined($sysData->{utHrs}));
		$lastReboot -= $sysData->{utMins} * 60 if (defined($sysData->{utMins}));
		$lastReboot -= $sysData->{utSecs} if (defined($sysData->{utSecs}));
	}else{
		$LOGGER->debug('Failed to parse and calculate required element core:lastReboot from \'show system\'.');
	}

	$out->print_element('core:systemName', $sysData->{systemName});

	$out->open_element('core:osInfo');
	$out->print_element('core:fileName', $sysData->{fileName}) if (defined($sysData->{fileName}));
	$out->print_element('core:make', 'Cisco');
	$out->print_element('core:version', $sysData->{version}) if (defined($sysData->{version}));
	$out->print_element('core:osType', 'NX-OS');

	# Here is where packetProcessingOrdering would go.  Because Cisco Nexus currently doesn't support NAT
	# the packetProcessingOrdering is not currently required.

	$out->close_element('core:osInfo');

	$out->print_element('core:biosVersion', $sysData->{biosVersion}) if (defined($sysData->{biosVersion}));
	$out->print_element('core:deviceType', 'Switch');
	$out->print_element('core:contact', $sysData->{contact}) if (defined($sysData->{contact}));
	$out->print_element('core:lastReboot', $lastReboot) if (defined($lastReboot));

	return;
}

sub _parse_system{
	my ($in) = @_;
	my $sysData = {};
	my $parseSystemReturn;

	given ($in->{xmlForm})
	{
		when (NEXUS_XML_COMMANDS_XML_SSH) {$parseSystemReturn =  _parse_system_xml($in, $sysData);};
		when (NEXUS_XML_COMMANDS_NETCONF ) {$parseSystemReturn =  _parse_system_xml($in, $sysData);};
		when (NEXUS_XML_NO_XML_TEXT_ONLY) {$parseSystemReturn = _parse_system_cli_text($in, $sysData)};
	}

	# Avoid partial assignment
	if (!$parseSystemReturn){
		$sysData = undef;
	}
	
	return $sysData;
}

sub _parse_system_xml{
	my ($in, $sysData) = @_;

	# Check for version data.  If not present, then this is a VDC and
	# get what we need to where we can.
	if (not $in->{version}){
		$LOGGER->debug('Cisco::Nexus::Parsers::parse_version() : no version data found in device response.');
	}else{
		my $twig = XML::Twig->new(twig_roots => {
					host_name => _createSimpleTwigSetter(\$sysData->{systemName}),
					isan_file_name => _createSimpleTwigSetter(\$sysData->{fileName}),
					sys_ver_str => _createSimpleTwigSetter(\$sysData->{version}),
					bios_ver_str => _createSimpleTwigSetter(\$sysData->{biosVersion}),
					kern_uptm_days => _createSimpleTwigSetter(\$sysData->{utDays}),
					kern_uptm_hrs => _createSimpleTwigSetter(\$sysData->{utHrs}),
					kern_uptm_mins => _createSimpleTwigSetter(\$sysData->{utMins}),
					kern_uptm_secs => _createSimpleTwigSetter(\$sysData->{utSecs})
				} );
		$twig->parse($in->{version});
	
		# We will overwrite what is in the version host_name with the definition of hostname in the
		# running-config, if present.  This is necessary for subcontexts
		if ($in->{running_config} =~ /^(host|switch)name (\S+)/m)
		{
			my $hostname = $2;
			$sysData->{systemName} = $hostname;
		}
	
		# This looks convoluted, but it has to handle some five different cases
		if ($in->{snmp} && $in->{snmp} =~ /^sys contact: +(\w)/m)
		{
				($sysData->{contact}) = $in->{snmp} =~ /^sys contact: +(.*)$/m;
				chomp($sysData->{contact});		
		}
	}
	
	return scalar (keys ( %{$sysData}));
}

sub _parse_system_cli_text
{
	my ($in, $sysData) = @_;

	return scalar (keys ( %{$sysData}));
}

# parse_chassis
#
# Parse the ziptie core sections describing asset data and cards (and their asset data)
#
# PARAMS: 
# $in - input context, featuring a {running_config} section
# $out - output XML writer context.  
#
sub parse_chassis
{
	my ($in, $out) = @_;

	# Check for a lack of version data
	if (not $in->{version})
	{
		$LOGGER->debug('Cisco::Nexus::Parsers::parse_chassis() : no version data found in device response.');
		return;		
	}

	my $modules = _parse_chassis($in);
	
	if (not defined $modules)
	{
		# _parse_chassis returned error and logged it
		return;
	}
	
	# Use the first two tokens in the description (chassis_id) as the chassis modelNumber.
	my $modelNumber = undef;
	if ($modules->{_chassis}->{description} =~ /^\s*(\S+(?:\s+\S+)?)/){
		$modelNumber = $1;
	}else{
		$modelNumber = 'Unknown';
	}

	$out->open_element('chassis');

	$out->open_element('core:asset');
	$out->print_element('core:assetType', 'Chassis');

	$out->open_element('core:factoryinfo');
	$out->print_element('core:make', 'Cisco');
	$out->print_element('core:modelNumber', $modelNumber);

	# Serial numbers can be the same on multiple virtual devices, QRM needs unique serial numbers so appending hostname
	my $serial_number = $modules->{_chassis}->{serialNumber};
	if($in->{hostname}){
		$serial_number .= '_' . $in->{hostname};
	}
	$out->print_element('core:serialNumber', $serial_number);

	$out->close_element('core:factoryinfo');

	$out->close_element('core:asset');

	$out->print_element('core:description', $modules->{_chassis}->{description}) if (defined($modules->{_chassis}->{description}));

	foreach my $card (values(%{$modules->{_cards}}))
	{
		if (!defined($card->{modelNumber}))
		{
			# modelNumber must be defined (required by schema).  If it's not defined this isn't
			# serious enough to stop the adapter.
			$LOGGER->debug('Required element core:modelNumber missing from card with id: ' . $card->{id});
			next;
		}

		$out->open_element('card');

		$out->open_element('core:asset');
		$out->print_element('core:assetType', 'Card');
		$out->open_element('core:factoryinfo');
		$out->print_element('core:hardwareVersion', $card->{hardwareVersion}) if (defined($card->{hardwareVersion}));
		$out->print_element('core:make', 'Cisco');  # +++waynet need to parse this somehow for third party cards
		$out->print_element('core:modelNumber', $card->{modelNumber});
		$out->print_element('core:serialNumber', $card->{serialNumber}) if (defined($card->{serialNumber}));
		$out->close_element('core:factoryinfo');
		$out->close_element('core:asset');

		$out->print_element('core:description', $card->{description}) if (defined($card->{description}));
		$out->print_element('portCount', $card->{portCount}) if (defined($card->{portCount}));
		$out->print_element('softwareVersion', $card->{softwareVersion}) if (defined($card->{softwareVersion}));
		$out->print_element('status', $card->{status}) if (defined($card->{status}));

		$out->close_element('card');
	}

	if (defined($modules->{_chassis}->{cpuType}))
	{
		trim($modules->{_chassis}->{cpuType});

		$out->open_element('cpu');
		$out->print_element('cpuType', $modules->{_chassis}->{cpuType});
		$out->close_element('cpu');
	}

	if (defined($modules->{_chassis}->{memInKb}))
	{
		$out->open_element('memory');
		$out->print_element('kind', 'RAM');
		$out->print_element('size', ($modules->{_chassis}->{memInKb} * 1024));  # Memory is sed-encoded in bytes
		$out->close_element('memory');
	}

	$out->close_element('chassis');

	return;
}

# This is the interior routine for parsing chassis data.  It will build up a hash in the return ref, keyed on "global"
# attributes and cards
sub _parse_chassis
{
	my $in = shift;
	my $modules = {};
	my $parseChassisReturn = 0;

	given ($in->{xmlForm})
	{
		when (NEXUS_XML_COMMANDS_XML_SSH) {$parseChassisReturn =  _parse_chassis_xml($in, $modules);};
		when (NEXUS_XML_COMMANDS_NETCONF ) {$parseChassisReturn =  _parse_chassis_xml($in, $modules);};
		when (NEXUS_XML_NO_XML_TEXT_ONLY) {$parseChassisReturn = _parse_chassis_cli_text($in, $modules)};
	} 

	# mark error for return to caller if nothing was set
	$modules = undef unless ($parseChassisReturn);
	
	return $modules;
}

sub _parse_chassis_xml
{
	my ($in, $modules) = @_;
	
	my $twig = XML::Twig->new(twig_roots => {
				proc_board_id => _createSimpleTwigSetter(\$modules->{_chassis}->{serialNumber}),
				chassis_id => _createSimpleTwigSetter(\$modules->{_chassis}->{description}),
				cpu_name => _createSimpleTwigSetter(\$modules->{_chassis}->{cpuType}),
				memory => _createSimpleTwigSetter(\$modules->{_chassis}->{memInKb})
			} );
	$twig->parse($in->{version});

	
	# Check for a lack of chassis data
	if (not $modules->{_chassis})
	{
		$LOGGER->debug('Cisco::Nexus::Parsers::parse_chassis() : no chassis data found in device response.');
		return 0;		
	}
	# Serial Number is very important to topology.
	elsif (!defined($modules->{_chassis}->{serialNumber}))
	{
		$LOGGER->debug('Failed to parse required element core:serialNumber for chassis from \'show system\'.');
		return 0;
	}

	$twig = XML::Twig->new(twig_roots => {
				ROW_modinfo => sub {
						my ($twig, $element) = @_;
	
						my $id = $element->first_child('modinf')->text();
						$modules->{_cards}->{$id} = {};
						$modules->{_cards}->{$id}->{id} = $id;

						$modules->{_cards}->{$id}->{portCount} = $element->first_child('ports')->text();
						$modules->{_cards}->{$id}->{description} = $element->first_child('modtype')->text();
						$modules->{_cards}->{$id}->{modelNumber} = $element->first_child('model')->text();
						$modules->{_cards}->{$id}->{status} = $element->first_child('status')->text();
	
						return 1;
					},
				ROW_modwwninfo => sub {
						my ($twig, $element) = @_;

						my $id = $element->first_child('modwwn')->text();

						$modules->{_cards}->{$id}->{softwareVersion} = $element->first_child('sw')->text();
						$modules->{_cards}->{$id}->{hardwareVersion} = $element->first_child('hw')->text();

						return 1;
					},
				ROW_modmacinfo => sub {
						my ($twig, $element) = @_;

						my $id = $element->first_child('modmac')->text();

						$modules->{_cards}->{$id}->{serialNumber} = $element->first_child('serialnum')->text();

						return 1;
					},
				ROW_xbarinfo => sub {
						my ($twig, $element) = @_;

						my $id = $element->first_child('xbarinf')->text();
						$id .= '_fabric';
						$modules->{_cards}->{$id} = {};
						$modules->{_cards}->{$id}->{id} = $id;

						$modules->{_cards}->{$id}->{portCount} = $element->first_child('xbarports')->text();
						$modules->{_cards}->{$id}->{description} = $element->first_child('xbartype')->text();
						$modules->{_cards}->{$id}->{modelNumber} = $element->first_child('xbarmodel')->text();
						$modules->{_cards}->{$id}->{status} = $element->first_child('xbarstatus')->text();

						return 1;
					},
				ROW_xbarwwninfo => sub {
						my ($twig, $element) = @_;

						my $id = $element->first_child('xbarwwn')->text();
						$id .= '_fabric';

						$modules->{_cards}->{$id}->{softwareVersion} = $element->first_child('xbarsw')->text();
						$modules->{_cards}->{$id}->{hardwareVersion} = $element->first_child('xbarhw')->text();

						return 1;
					},
				ROW_xbarmacinfo => sub {
						my ($twig, $element) = @_;

						my $id = $element->first_child('xbarmac')->text();
						$id .= '_fabric';

						$modules->{_cards}->{$id}->{serialNumber} = $element->first_child('xbarserialnum')->text();

						return 1;
					}
			} );
	$twig->parse($in->{module});
	
	return scalar (keys ( %{$modules}));
}

sub _parse_chassis_cli_text
{
	my ($in, $modules) = @_;
	
	# We need to parse through sections: cards/ports, hardware/software, serial #, then the same three for "xbar"  
	# Each has a pattern of a header, a dashed line, and the data directly beneath for each module, 1-n.
	# it closes with a blank line for each section (xbar sw closes it)
	
	if ($in->{version} =~ /^Hardware(.+?)\n\S/gism)
	{
		my $hardwareVersionBlob = $1;
		# first line is description
		if ($hardwareVersionBlob =~ /^\s+(.+?)$/ig)
		{
			$modules->{_chassis}->{description} = $1;
		}
		
		if ($hardwareVersionBlob =~ /\s+Processor Board ID (\S+)/ig)
		{
			$modules->{_chassis}->{serialNumber} = $1;
		}
		else
		{
			# This is an error, since it will create an unusable SED
			$LOGGER->debug('Failed to parse required element core:serialNumber for chassis from \'show version\'.');
			return 0;
		}
		
		# first mention of CPU is cpu type
		if ($hardwareVersionBlob =~ /^\s+(.+? CPU)\s+with (\d+)/ig)
		{
			$modules->{_chassis}->{cpuType} = $1;
			$modules->{_chassis}->{memInKB} = $2;
		}
	}
	else
	{
		# This is an error, since it will create an unusable SED
		$LOGGER->debug('Failed to parse required element core:serialNumber (hardware info) for chassis from \'show version\'.');
		return 0;		
	}
	
	# parse board-level serial $, description, CPU name, and memory out of $in->{module}
	
	if ($in->{module} =~ /^Mod\s+Ports\s+Module\-Type\s+Model\s+Status(.+?)Mod/mis)
	{
		my $cardPortsBlob = $1;
		
		# We can parse this line by line. The difference between the description and model is tough since they can run together.
		# I'm having to just use the fact that the module is always uppercase, digits, and dashes, e.g., N7K-M108X2-12
		while ($cardPortsBlob =~ /^(\d+)\s+(\d+)\s+(.+?)\s+([A-Z\d\-])\s+(\S+)/gi)
		{
			my $id = $1;
			$modules->{_cards}->{$id}->{portCount} = $2;
			$modules->{_cards}->{$id}->{description} = $3;
			$modules->{_cards}->{$id}->{modelNumber} = $4;
			$modules->{_cards}->{$id}->{status} = $5;			
		}
		
	}

	# parse hw, sw out of $in->{version}
	
	if ($in->{module} =~ /^Mod\s+Sw\s+Hw.(.+?)Mod/mis)
	{
		my $hwSwBlob = $1;
		
		while ($hwSwBlob =~ /^(\d+)\s+([\d\.\(\)]+)\s+([\d\.]+)/gi)
		{
			my $id = $1;
			$modules->{_cards}->{$id}->{softwareVersion} = $2;
			$modules->{_cards}->{$id}->{hardwareVersion} = $3;
		}
		
	}

	# parse serial # out of the MAC addr/serial # section in $in->{version}
	
	if ($in->{module} =~ /^Mod\s+Mac\-Address\(es\)\s+Serial\-Num(.*?)Mod/mis)
	{
		my $serialBlob = $1;
		
		# move right to the serial number here
		while ($serialBlob =~ /^(\d+)\s+([\da-z\- ])+([A-Z\d+]+)/gi)		
		{
			my $id = $1;
			$modules->{_cards}->{$id}->{serialNum} = $3;
		}
		
	}

	# Parse crossbar ports, modujle, type, status
	if ($in->{module} =~ /^Xbar\s+Ports\s+Module\-Type\s+Model\s+Status(.+?)Xbar/mis)
	{
		my $xbarPortsBlob =  $1;
		
		# We can parse this line by line. The difference between the description and model is tough since they can run together.
		# I'm having to just use the fact that the module is always uppercase, digits, and dashes, e.g., N7K-M108X2-12
		while ($xbarPortsBlob =~ /^(\d+)\s+(\d+)\s+(.+?)\s+([A-Z\d\-])\s+(\S+)/gi)
		{
			my $id = $1 . '_fabric';
			$modules->{_cards}->{$id}->{portCount} = $2;
			$modules->{_cards}->{$id}->{description} = $3;
			$modules->{_cards}->{$id}->{modelNumber} = $4;
			$modules->{_cards}->{$id}->{status} = $5;			
		}
		
	}

	# parse hw, sw out of $in->{version}
	
	if ($in->{module} =~ /^Xbar\s+Sw\s+Hw.(.+?)Mod/mis)
	{
		my $xbarHwSwBlob = $1;
		
		while ($xbarHwSwBlob =~ /^(\d 	+)\s+([\d\.\(\)]+)\s+([\d\.]+)/gi)
		{
			my $id = $1 . '_fabric';
			$modules->{_cards}->{$id}->{softwareVersion} = $2;
			$modules->{_cards}->{$id}->{hardwareVersion} = $3;
		}
		
	}
	# parse xbar serial # out of the MAC addr/serial # section in $in->{version}.  This closes that section, so 
	# just scan to the end of it
	
	if ($in->{module} =~ /^Xbar\s+Mac\-Address\(es\)\s+Serial\-Num(.*)/mis)
	{
		my $xbarSerialBlob = $1;
		
		# move right to the serial number here
		while ($xbarSerialBlob =~ /^(\d+)\s+([\da-z\- ])+([A-Z\d+]+)/gi)		
		{
			my $id = $1 . '_fabric';
			$modules->{_cards}->{$id}->{serialNum} = $3;
		}
		
	}

	return scalar (keys ( %{$modules}));
}

# parse_interfaces
#
# Go through the interface section(s) parsing of the running config.
#
# PARAMS: 
# $in - input context, featuring a {running_config} section
# $out - output XML writer context.  CAN BE undef, in which case parsing and subnet formation will occur
#    without XML output.
#
# RETURNS:
#   a list:
#   	$subnets - a hash of Ziptie::Addressing::Subnets instances, indexed by interface name
#       $filterAssignments - a hash indexed by interface name, indicating assigned ingress and egress filters.
#
sub parse_interfaces {
	my ($in, $out) = @_;
	my $locInterfaces = {};

	# If we have no interface data to parse then return
	if( not defined $in->{interface} ) {
		$LOGGER->debug('parse_interfaces: No interfaces data found in device response.');
		return;
	}

	_parse_vlan_information($in);

	my $interface_networks;

	# Create a simple hash of the standalone snmp index interface element
	my $ifIndices;
	if ($in->{interface_ifindex}) {
		my $twig = XML::Twig->new(
			twig_roots => {
				ROW_interface => sub {
						my ($twig, $element) = @_;
	
						my $ifName = $element->first_child('interface')->text();
						if( defined $element->first_child('ifindex-dec') ) {
							$ifIndices->{$ifName} = $element->first_child('ifindex-dec')->text();
						}
						return 1;
					}
		} );
		if (is_conforming_xml($in->{interface_ifindex})) {
			$twig->parse($in->{interface_ifindex});
		}
	} else {
		$LOGGER->debug('parse_interfaces: No interface index data found in device response.');
	}
	
	my $twig = XML::Twig->new(twig_roots => {
			ROW_interface => sub {
					my ($twig, $element) = @_;
					my $ifName = $element->first_child('interface')->text();

					if (_parse_interface_brief($in, $ifName, $element, $locInterfaces))
					{
						# Assign if index
						if (defined $ifIndices->{$ifName}) {
							$locInterfaces->{$ifName}->{ifIndex} = $ifIndices->{$ifName};
						}

						# Now, dive into the running-config section for this adapter to pull up layer 3 properties.
						# We also need to collect subnets and filter assignments here.
						if ($in->{running_config} =~ /^interface\s+($ifName)(.+?)\n(?=\w)/smi)
						{
							my $ifCfgBlob = $2;
							my $ipOptions = undef;

							if ($ifCfgBlob =~ /^\s+description\s+([\S ]*)/mi)
							{
								$locInterfaces->{$ifName}->{description} = $1;
							}
							
							while ($ifCfgBlob =~ /^\s+ip\s+(\S.*)$/mig)
							{
								my $ipLine = $1;
								given ($ipLine) 
								{
									when (/^directed\-broadcast/i)	{$ipOptions->{directedBroadcast} = "true";}
									when (/^proxy\-arp/i) {$ipOptions->{localProxyArp} = "true";}
									when (/^route-cache/i) {$ipOptions->{routeCache} = "true";}
									when (/^dhcp\s+relay\s+address/)
									{
										if ($ipLine =~ /^dhcp\s+relay\s+address\s+($ip_regex)/)
										{
											push (@{$ipOptions->{udpForwarder}}, $1);
										}
									}
									when (/^address\s+$cidr_regex/)
									{
										# assign ipConfiguration											
										if ($ipLine =~ /^address\s+($ip_regex)\/(\d+)/)
										{
											my $sn = new ZipTie::Addressing::Subnet($1, $2);
											push (@{$ipOptions->{ipConfiguration}}, 
												{ ipAddress => $sn->getHostIp(), mask => $sn->getMask()});
											push (@{$interface_networks->{$ifName}}, $sn);
										}
									}
									when (/^access\-group\s+(\S+)\s+(\w+)/)
									{
										# access list interface application 
										given ($2)
										{
											when ("in") {$locInterfaces->{$ifName}->{'ingressFilter'} = $1;}
											when ("out") {$locInterfaces->{$ifName}->{'egressFilter'} = $1;}
											default {$LOGGER->debug("parse_interfaces: unable to parse access list direction \'$2\'");}; 												
										}
									}										
								}
								$locInterfaces->{$ifName}->{interfaceIp} = $ipOptions;									
							}
						}
					}												
					return 1;
				}
	} );

	# Need to clean interface due to bug in 'show interfaces brief' XML response
	my ($cleanInterfaces) = _clean_interfaces($in);
	$twig->parse($cleanInterfaces);

	# Print interfaces if the xml printer was passed in
	if (defined $out) {
		$out->open_element('interfaces');

		for my $nextIf (keys %{$locInterfaces} ) {
			_printInterface( $locInterfaces->{$nextIf}, $out);
		}

		$out->close_element('interfaces');
	}

	$in->{parsed}->{interface_networks} = $interface_networks;

	return;
}

sub create_config
{
	my ($in, $out) = @_;

	my $repository = {};
	$repository->{'core:config'} = [];

	# The "root folder" of where we put things is dependent on whether we have contexts we'll need
	# to rake over or not.
	$repository->{'core:name'} =  "/";

	$LOGGER->debug("Cisco::Nexus::Parsers::create_config: setting storage for admin context'...");
	# Back up global context....
	my $dcAdminContextFolder;
	$dcAdminContextFolder->{'core:name'} = "system space";
	
	push (@{$dcAdminContextFolder->{'core:config'}}, _createSingleContextConfigFromCommand($in, 'running-config', 'running_config',
			'text/plain', 'active', 'false'));

	if ($in->{startup_config})
	{
		push (@{$dcAdminContextFolder->{'core:config'}}, _createSingleContextConfigFromCommand($in, 'startup-config', 'startup_config',
				'text/plain', 'boot', 'true'));
	}
	push ( @{ $repository->{'core:folder'} }, $dcAdminContextFolder);		 
	
	if ($in->{contexts})
	{
		for my $vdcName (keys %{$in->{contexts}})
		{
			$LOGGER->debug("Cisco::Nexus::Parsers::create_config: setting storage for context \'" . $vdcName . "\'...");
			my $dcFolder;
			$dcFolder->{'core:name'} = $vdcName;
			
			push ( @{ $dcFolder->{'core:config'} },  
				_createSingleContextConfigFromCommand($in->{contexts}->{$vdcName},'running-config', 'running_config', 
					'text/plain',	'active', 'false')); 
			push ( @{ $repository->{'core:folder'} }, $dcFolder);		 
		}			
	}

	$out->print_element('core:configRepository', $repository);

	return;
}

# parse_access_ports
#
# Parse the ziptie core sections describing diagnostic and chassis-level ports to access the 
# system at console level.
#
# PARAMS: 
# $in - input context, featuring a {running_config} section
# $out - output XML writer context.  
sub parse_access_ports
{
	my ($in, $out) = @_;

	my $accessPortsOpen = 0;
	while ($in->{running_config} =~ /^line (\S+)/mg)
	{
		my $type = $1;

		if (!$accessPortsOpen)
		{
			$out->open_element('accessPorts');
			$accessPortsOpen = 1;
		}

		$out->open_element('accessPort');
		$out->print_element('type', $type);
		$out->close_element('accessPort');
	}
	if ($accessPortsOpen)
	{
		$out->close_element('accessPorts');
	}

	return;
}

# parse_object_groups
#
# Parses the output of the show object groups commands and generates the objectGroups section of the SED.
#
# PARAMS: 
# $in - input context, featuring a {object_group} section
# $out - output XML writer context.
sub parse_object_groups
{
	my ($in, $out) = @_;

	# Currently the Nexus only supports network groups and service groups
	my $objectGroups = {};
	$objectGroups->{networkGroups} = [];
	$objectGroups->{serviceGroups} = [];

	# If no object groups, no worries.  Totally optional element.  But don't pass it to Twig::parse, or it
	# will NPE.  
	if (not $in->{object_group})
	{
		$LOGGER->debug('Cisco::Nexus::Parsers::parse_object_groups() : no object group data found in device response (which is perfectly acceptable).');
		return;
	}	

	# This hash maps the group type integers found in the Nexus XML to the correct parser to use to parse each object group.
	# The first argument of each parser will be the XML::Twig element to parse, and the second argument to each parser will
	# be the $objectGroups hash, as seen in the XML::Twig handler bellow for the ROW_ogroup element.
	my %GROUP_PARSERS = (

			# ipv4 object group parser
			100 => \&_networkGroupParser,

			# port object group parser
			102 => \&_serviceGroupParser
		);

	my $twig = XML::Twig->new(twig_roots => {
				ROW_ogroup => sub {
						my ($twig, $element) = @_;

						my $groupTypeElement = $element->first_child('group_type');
						if (!defined($groupTypeElement))
						{
							$LOGGER->debug("parse_object_groups: Failed to parse element group_type from object group");
							return 1;
						}

						my $groupType = $groupTypeElement->text();
						if ($groupType == "101")
						{
							$LOGGER->debug("parse_object_groups: Dropping IPv6 object group");
						}
						elsif (defined($GROUP_PARSERS{$groupType}))
						{
							$GROUP_PARSERS{$groupType}->($element, $objectGroups);
						}
						else
						{
							$LOGGER->debug("parse_object_groups: Unknown object group type: $groupType");
						}

						return 1;
					} });
	$twig->parse($in->{object_group});

	# This is where we actually output the XML, if there is anything to output
	if (scalar(@{$objectGroups->{networkGroups}}) || scalar(@{$objectGroups->{serviceGroups}}))
	{
		$out->open_element('objectGroups');

		foreach my $networkGroup (@{$objectGroups->{networkGroups}})
		{
			$out->open_element('networkGroup');
			$out->print_element('id', $networkGroup->{id});

			if (defined($networkGroup->{host}))
			{
				foreach my $host (@{$networkGroup->{host}})
				{
					$out->print_element('host', $host);
				}
			}

			if (defined($networkGroup->{subnet}))
			{
				foreach my $subnet (@{$networkGroup->{subnet}})
				{
					$out->print_element('subnet', $subnet);
				}
			}

			$out->close_element('networkGroup');
		}

		foreach my $serviceGroup (@{$objectGroups->{serviceGroups}})
		{
			$out->open_element('serviceGroup');
			$out->print_element('id', $serviceGroup->{id});

			if (defined($serviceGroup->{portExpression}))
			{
				foreach my $portExpression (@{$serviceGroup->{portExpression}})
				{
					$out->print_element('portExpression', $portExpression);
				}
			}

			if (defined($serviceGroup->{portRange}))
			{
				foreach my $portRange (@{$serviceGroup->{portRange}})
				{
					$out->print_element('portRange', $portRange);
				}
			}

			$out->close_element('serviceGroup');
		}

		$out->close_element('objectGroups');
	}

	return;
}


# parse_access_lists
#
# Parses the access lists declared within the Nexus configuration and "show ip access-lists expanded"
# command
#
# PARAMS: 
# $in - input context, featuring a {access_lists} section
# $out - output XML writer context.  ALSO assigns the generated_access_lists $in 
#        hash bucket
sub parse_access_lists
{
	my ($in, $out) = @_;
	my $fListsDeclared = 0;
	my $fEntryName;
	my $fEntryDesc;	
	my $fEntryDescNext;
	my $alNID;
	my @alSpecTerms; 
	my $alSignatureAssembly;

	# Back away unless actual access lists were declared as part of the 
	# configuration
	return unless ($in->{access_lists});

	# Look over $in->{access_lists}, Enumerate on
	#     ROW_ip_ipv6_mac entries
	#     +++waynet MAC acl including ethertype (*not* supported on layer 3 interfaces)
	#     +++waynet VLAN access map
	
	my $twig = XML::Twig->new(twig_roots => {
				ROW_ip_ipv6_mac => sub {
					my ($twig, $element) = @_;
					@alSpecTerms = ();
						
					if (not $fListsDeclared)
					{ 
						$out->open_element("filterLists");
						$fListsDeclared = 1;
					} 
					# pick up the acl name here...
					$fEntryName = $element->first_child_text('acl_name');
					$fEntryDesc = "";

					$out->open_element('filterList configPath="running-config"');
									
					# Now, specs are sequenced by priority.  It sure looks like they come in order,
					# but I'm not one to say necessarily.  To that end, put them in an array which will 
					# index by the eval order number

					# if this as applied to an i/f doesn't match traffic, default is deny...
					$out->print_element("defaultAction", "deny");

					# Filter signature.  I don't recall why, but there's a reason that we need to put
					# this after default signature.
					$alSignatureAssembly = $out->attach_signature_assembly('filterList', "ztSignature");

					# Process the TABLE_seqno/ROW_seqno subpaths
					for ($element->first_child('TABLE_seqno'))
					{
						if(defined($_)){
							foreach my $feElement ($_->children('ROW_seqno'))
							{
								my $alSpecTerm = {};
								($alNID, $fEntryDescNext) = _parse_alSpecTerm($in, $fEntryName, $feElement, $alSpecTerm);

								# Description is to be found in "remark" element.  We latch the first such thing found to be the
								# overall filter description.  This often comes unterminated by the netconf *ssh CLI,*
								if ($fEntryDescNext and not $fEntryDesc)
								{
									$fEntryDesc = $feElement->first_child_text("remark");
									chomp($fEntryDesc);
								}

								$alSpecTerms[$alNID] = $alSpecTerm;
							}
						}
					}
					# was there a description to elevate?
					if ($fEntryDesc)
					{
						$out->print_element("description", $fEntryDesc);											
					}
					
					# Now iterate our list

					my $alIndex = 1;
					for $alNID (0 .. $#alSpecTerms)
					{
						if (defined $alSpecTerms[$alNID])
						{
							# assign the "name" element as being the index here
							$alSpecTerms[$alNID]->{'name'} = $alNID;
							$alSpecTerms[$alNID]->{'processOrder'} = $alIndex++;
							$out->print_element("filterEntry", $alSpecTerms[$alNID] );
						}
					}
					
					$out->print_element("mode", "stateful");
					$out->print_element("name", $fEntryName);
					
					$out->close_element('filterList');
					return 1;
				}
		}
	);
	$twig->parse($in->{access_lists});
		
	$out->close_element("filterLists") if ($fListsDeclared); 
	return;
}

sub parse_static_routes {
	my ($in, $out) = @_;
	$LOGGER->debug('Parsing static routes');

	my $staticRoutes = _parse_static_route_data($in);
	my $static_route_count = 0;

	if( defined $staticRoutes ) {
		$static_route_count = scalar @{ $staticRoutes };

		$out->open_element('staticRoutes');
		$out->print_element('staticRoute', $staticRoutes);
		$out->close_element('staticRoutes');		
	}

	$LOGGER->debug("Found $static_route_count static routes");
	return;
}

sub parse_vlans
{
	my ($in, $out) = @_;

	parse_interfaces($in, undef);

	my @vlans = values(%{$in->{parsed}->{vlans}});
	if( scalar @vlans ) {  
		$out->open_element('vlans');
		$out->print_element('vlan', \@vlans);
		$out->close_element('vlans');
	}

	return;
}

sub _parse_vlan_information {
	my ($in) = @_;

	if( $in->{parsed}->{vlans} && $in->{parsed}->{interface_vlans} ) {
		return;
	}

	my %vlans = ();
	my %interface_vlans = ();

	my $twig = XML::Twig->new(twig_roots => {
		ROW_vlanbrief => sub {
				my ($twig, $element) = @_;
				my $vlan_element = $element->first_child('vlanshowbr-vlanid-utf') // $element->first_child('vlanshowbr-vlanid');

				if( not defined $vlan_element ) {
					return;
				}
				
				my $vlanId = $vlan_element->text();
				if( not defined $vlans{$vlanId} ) {
					$vlans{$vlanId} = {
						number => $vlanId,
						interfaceMember => [] # This array is populated in parse_interfaces
					};
				}

				$vlans{$vlanId}->{enabled} = 'false';
				if( $element->first_child('vlanshowbr-shutstate')->text() eq 'noshutdown' ) {
					$vlans{$vlanId}->{enabled} = 'true';
				}

				my $vlanNameElement = $element->first_child('vlanshowbr-vlanname');
				if( defined $vlanNameElement ) {
					$vlans{$vlanId}->{name} = $vlanNameElement->text();
				}

				for my $vlanPortListElement ( $element->children('vlanshowplist-ifidx') ) {
					for my $portRange ( split /,/, $vlanPortListElement->text() ) {
						my $ports = _expand_port_range($portRange);

						for my $port (@{$ports}) {
							if (not defined $interface_vlans{$port} ) {
								$interface_vlans{$port} = ZipTie::Ranges->new();
							}
							$interface_vlans{$port}->add_range($vlanId);
						}
					}
				}
	}});
	
	if( is_conforming_xml($in->{vlan}) ) {
		$twig->parse($in->{vlan});
	}

	$in->{parsed}->{vlans} = \%vlans;
	$in->{parsed}->{interface_vlans} = \%interface_vlans;

	return;
}

sub _expand_port_range
{
	my ($portRange) = @_;

	my $ports = [];

	if ($portRange =~ /^(\S+?)(\d+)-(\d+)$/)
	{
		my $prefix = $1;
		my $startPort = $2;
		my $endPort = $3;

		for (my $i = $startPort;$i <= $endPort;$i++)
		{
			push(@{$ports}, ($prefix . $i));
		}
	}
	else
	{
		push(@{$ports}, $portRange);
	}

	return $ports;
}

# Create the common:dynamic_routes section of the ZED by looking at the routing state data
#
# PARAMS: 
# $in - input context, featuring a {<routing protocols>->{routes}} section. 
# $out - output XML writer context.  
sub parse_dynamic_routes {
	my ($in, $out) = @_;
	my @dynamic_routes = ();
	my @vrfs = (DEFAULT_VRF_NAME);

	$LOGGER->debug('Parsing dynamic routes');

	# We do not have the capability to model anyting but the default VRF
	#while ($in->{running_config} =~ /^vrf\s+context\s+(\S+)/mig) { push @vrfs, $1; }

	for my $vrf ( @vrfs ) { 
		my $vrf_routes = _parse_dynamic_route_data($in, $out, $vrf);
		if( defined $vrf_routes ) {
			push @dynamic_routes, @{ $vrf_routes };
		}
	}

	my $dynamic_route_count = scalar @dynamic_routes;
	if( $dynamic_route_count ) {
		$out->open_element('dynamicRoutes');
		$out->print_element('dynamicRoute', \@dynamic_routes);
		$out->close_element('dynamicRoutes');
	}

	$LOGGER->debug("Found $dynamic_route_count dynamic routes");
	return;
}

sub _parse_dynamic_route_data {
	my ($in, $out, $vrfName) = @_;
	my @dynamic_routes;

	$LOGGER->debug("Parsing VRF '$vrfName' routes");
	foreach my $protocol ( SUPPORTED_ROUTING_PROTOCOLS ) {
		$LOGGER->debug("Parsing '$protocol' routes");

		my $twig = XML::Twig->new( twig_roots => { ROW_vrf =>
			sub {
				my ($twig, $element) = @_;

				# Now, let's look for the routes which should be presumably most if not all about routes learned from neighbor systems.
				if ($element->first_child_text('vrf-name-out') !~ /^$vrfName/i) {
					return;
				}

				# Now, we have a lot of navigation to do here.  First, we want to be sure we're on the IPv4 routing table
				for ($element->first_child('TABLE_addrf')) {
					for my $afElement  ($_->children('ROW_addrf')) {
						if ($afElement->first_child_text('addrf') ne 'ipv4') {
							next;
						}
						# If we have routes...
						if (not	$afElement->first_child('TABLE_prefix')) {
			                        	next;
						}

						# Within this, there is a single TABLE_prefix, with ROW_prefix rows.
						for ($afElement->first_child('TABLE_prefix')) {
							for my $prefixElt ($_->children('ROW_prefix')) {

								my $dRouteElt = { advertisingRoutingProcess => $protocol };								
								($dRouteElt->{routeNetwork}->{address}, $dRouteElt->{routeNetwork}->{mask}) = ($prefixElt->first_child_text('ipprefix') =~ /^([\d\.]+)\/(\d+)/);

								# Dive down to find next hop and where we heard about it from
								my $tblNH = $prefixElt->first_child('TABLE_path');
								if ($tblNH) {
									foreach my $pathElt ($tblNH->children('ROW_path')) {
										my $nh;

										my $nextElt = $pathElt->first_child_text('metric');
										if ($nextElt) {
											$nh->{adjacencyMetric} = $nextElt;
										}

										$nextElt = $pathElt->first_child_text('ipnexthop');
										if ($nextElt) {
											$nh->{adjacencyNode} = $nextElt;
										}

										$nextElt = $pathElt->first_child_text('ifname');
										if ($nextElt) {
											$nh->{adjacencyInterface} = expand_interface_name($nextElt);
										} else {
											my $interface = get_interface_connected_to_network($in, $nh->{adjacencyNode});
											if( $interface ) {
												$nh->{adjacencyInterface} = $interface;
											} else {
												$LOGGER->debug("parse_dynamic_routes: Unable to find interface for next hop " . $nh->{adjacencyNode});
											}
										}

										if ($nh) {
											push (@{$dRouteElt->{routeNextHop}}, $nh);
										} else {
											$LOGGER->warning($ROUTE_DROPPED, 'parse_dynamic_routes: Unable to parse next hop for network ' . $prefixElt->first_child_text('ipprefix'));
										}
									}
								} else { # TABLE_path not defined in XML input
									$LOGGER->warning($ROUTE_DROPPED, 'parse_dynamic_routes: Not able to find any next hop information for route ' . $dRouteElt->{routeNetwork}->{address} . '/' . $dRouteElt->{routeNetwork}->{mask} );
									next;
								}

								if($dRouteElt) {
									push @dynamic_routes, $dRouteElt;
								}
							}
						}
					}
				}
			}
		});

		if( $in->{$protocol}->{routes} ) {
			$LOGGER->debug("Parsing routes from 'show ip route $protocol'");
			$twig->parse($in->{$protocol}->{routes});
		}
	}

	return \@dynamic_routes;
}

sub open_vdc_section
{
	my ($in, $out, $subelementSet)  = @_;
	
	if (defined $subelementSet) 
	{
		if (not defined $in->{subelementContexts})
		{
			# Track our generation of vdc's / subelements this way
			map { $in->{subelementContexts}->{"$_"} = 1; } @{$subelementSet};
			$out->open_element("subelements");
		}
	}
	else
	{
		# this is for testing
		$out->open_element("subelements");
	}		

	$out->open_element("subelement");

	return;
}

sub close_vdc_section
{
	my ($in, $out, $thisSubelement) = @_;
	
	# Close out the subelement.  that's the easy part.  We also need to identify the 
	# subelement type as  a mandatory element
	$out->print_element("subelementType", "virtual");
	$out->close_element("subelement");
	
	if (defined $thisSubelement)
	{
		# Now, remove the subelement name...
		if (exists $in->{subelementContexts}->{"$thisSubelement"})
		{
			delete $in->{subelementContexts}->{"$thisSubelement"};
			
			# Now, check if there's a nonempty hash key
			for (keys %{$in->{subelementContexts}})
			{
				return if ($_ =~ /\w/);   # if this is a nonempty key, then we've still got contexts to generate
			}
			
			# If we get here, then we can close out the entire 'subelements' section
			$out->close_element("subelements");
		}		
	}
	else
	{
		# this is for testing
		$out->close_element("subelements");
		
	}

	return;
}

sub get_default_vrf_name { return DEFAULT_VRF_NAME; }

##### Functions private to this module #################

sub _createSingleContextConfigFromCommand
{
	my ($in,  $name, $command, $mediaType, $context, $promotable) = @_;

	my $config = {};
	# The listing format encodes the time of query into it, which is a bad thing for
	# persisting our Nexus configs, since it means all of them will display differences.
	# So, make a copy and eliminate that line for storage
	my $configText = $in->{$command};
	$configText =~ s/^\s*!Time[\w\s\d:]+\n//m;

	$config->{'core:name'} = $name;
	$config->{'core:textBlob'} = encode_base64($in->{$command});
	$config->{'core:mediaType'} = $mediaType;
	$config->{'core:context'} = $context;
	$config->{'core:promotable'} = $promotable;

	return $config;
}

sub _parse_static_route_data {
	my $in = shift;
	my @staticRouteData;

	while ($in->{running_config} =~ /^(ip route ([\d\.\/]+)\s+(.*))/mig) {
		my ($line, $cidr, $blob) = ($1, $2, $3);
		my $staticRoute;

		$LOGGER->verbose("Parsing static route line: '$line'");

		if ($cidr =~ /$cidr_regex/) {
			my ($ip, $mask) = split(/\//, $cidr);

			$staticRoute = {
				destinationAddress => $ip,
				destinationMask => $mask,
				defaultGateway => 'false'
			};

			if( $ip eq '0.0.0.0' && $mask == 0 ) {
				$staticRoute->{defaultGateway} = 'true';
			}
		} else {
			$LOGGER->warning($ROUTE_DROPPED, "parse_static_routes: Unable to interpret '$line'");
			next;
		}

		# Following element is either an interface or a next hop. This is matching an IP address followed
		# *not* directly by a nonspace charactger (which we're presuming might be a CIDR address, to be 
		# matched in the next "elsif" clause)
		if ($blob =~ /^($ip_regex)(\s+|$)(.*?)$/) {
			$staticRoute->{gatewayAddress} = $1;
			$blob = $3;

			my $interface = get_interface_connected_to_network($in, $staticRoute->{gatewayAddress});
			if( $interface ) {
				$staticRoute->{interface} = $interface;
			} else {
				$LOGGER->warning($ROUTE_DROPPED, "parse_static_routes: Unable to find interface connected to '$staticRoute->{gatewayAddress}'");
	                        next;
			}
		}
		# This is either an interface or a next hop network.  There are three variations really.  One is the
		# identification of a next hop network.  The ZED doesn't really support that, so we have to match
		# it up with an interface matching the network, if we can.
		elsif( $blob =~ /^($ip_regex)\/(\d+)\s*(.*?)$/ ) {
			my $network = new ZipTie::Addressing::Subnet($1, $2);
			$blob = $3;

			my $interface = get_interface_connected_to_network($in, $network);
                        if( $interface ) {
                                $staticRoute->{interface} = $interface;
                        }

			# See bug #42869; If a gateway host disguised as a net convert to host
			if ($network->getMask() ~~ 32) {
				$staticRoute->{gatewayAddress} = $network->getHostIp();
			}

			if (not defined $staticRoute->{interface} and not defined $staticRoute->{gatewayAddress}) {
				$LOGGER->warning($ROUTE_DROPPED, "parse_static_routes: Unable to resolve network " . $network->to_string() . " to an interface or gateway address");
				next;
			}
		# Else, we're going with an interface here
		} else {
			if ($blob =~ /^(\S+)(\s+\d+\/\d+)?\s*(.*)/) {
				my $nhIF = $1; $nhIF .= $2 // "";
				$blob = $3;

				# See if this is just a discard route
				if ($nhIF =~ /^null/i) {
					$staticRoute->{discard} = 'true';
				}

				# Either way, we always set interface - required element
				$staticRoute->{interface} = $nhIF;
				
				# See if a next hop was specified here as well as an IP addr
				if ($blob =~ /^($ip_regex)\s*(.*)/) {
					$staticRoute->{gatewayAddress} = $1;
					$blob = $2 // '';
				}
			}
		}

		# See if we can infer a route pref or metric it will presumably just be a
		if( $blob ) {
			$LOGGER->verbose("Remaining blob: '$blob'");
			if ($blob =~ /\S+/) {
				if ($blob =~ /^(\d+)/) {
					$staticRoute->{routePreference} = $1;
				}
			}
		}
		
		if (defined $staticRoute) {
			$LOGGER->verbose('Creating static route: ' . Dumper($staticRoute));
			push @staticRouteData, $staticRoute;

			# This route gets stored in a special place
			if ($staticRoute->{defaultGateway} eq 'true') {
				$in->{defaultGatewayRoute} = $staticRoute;
			}
		} else {
			$LOGGER->warning($ROUTE_DROPPED, "parse_static_routes: Unable to parse a static route from '$line'");
		}
	}

	return \@staticRouteData;
}

########  Twig subhandlers and utilities ######################
#
#  This takes the element context of the ROW_Interface 
#  element for a single element of "show interfaces brief".  The
#  interface name, as passed, is already parsed.
#
#  All element context is filled in in $zInterfaces->{$ifName}
#
#  RETURNS: 0 or nonzero - if nonzero, then the interface has been
#   acknowledged as pertinent to the ZED (something was filled in zInterfaces)
#   and parsing/processing on this interface should continue.  This routine will
#   ONLY set up applicable interface-first level and interfaceEthernet properties.

sub _parse_interface_brief
{
	my ($in, $ifName, $tElement, $zInterfaces) = @_;

	if ($in->{interfaces}->{$ifName})
	{
		# Add to Interfaces hash, marking admin status from interface brief.
		$zInterfaces->{"$ifName"}->{adminStatus} = $tElement->first_child('state')->text();
		# Now, for speed.  The default reporting is in megabits it appears, unless appended by "G"
		
		my $ifSpeed = 0;
		
		if ($tElement->first_child_text('speed') =~ /^(\d+)(\w?)/)
		{
			if ($2)
			{
				given ($2)
				{
					when ('G') { $ifSpeed = 1073741824; } # 1 Gb in bits
					when ('M') { $ifSpeed = 1048576; }
					default { 
						$LOGGER->debug("parse_interfaces: Unexpected speed modifier $2");
						$ifSpeed = 1048576; # for some basis		
					}
				}
			}
			else
			{
				# base 1 MB/sec basis
				$ifSpeed = 1048576; 
			}
			
		}
		else
		{
			# This is one place we pick this up
			if ($tElement->first_child_text('speed') =~ /auto/i)
			{
				$zInterfaces->{"$ifName"}->{interfaceEthernet}->{autoSpeed} = "true";
			}
		}
		
		if ($ifSpeed)
		{
			# Now, just scale appropriately
			$ifSpeed *= $1;
			$zInterfaces->{"$ifName"}->{speed} = $ifSpeed;
		}
		elsif (not defined  $zInterfaces->{"$ifName"}->{interfaceEthernet}->{autoSpeed})
		{
			# There's not a great way with the data we have to determine, "virtual", so this is a blunt one...
			if ($ifName !~ /^(loopback|port\-channel|Vlan)/)
			{
				$LOGGER->debug("parse_interfaces : Unable to determine speed on interface $ifName");						
			}
		}
		
		if ($tElement->first_child("desc"))
		{
			$zInterfaces->{"$ifName"}->{description} = $tElement->first_child_text("desc");
			# for some reason, this often remains with a trailing eol in the xml output
			chomp($zInterfaces->{"$ifName"}->{description});
		}
		
		my $portmode = $tElement->first_child_text("portmode");
		if ($portmode =~ /^route/i)
		{
			push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "routing");			
		}
		elsif ($portmode eq 'trunk' && defined($in->{parsed}->{interface_vlans}->{$ifName}))
		{
			foreach my $vlanRange (@{$in->{parsed}->{interface_vlans}->{$ifName}->get_ranges()})
			{
				if ($vlanRange->{lowInt} == $vlanRange->{highInt})
				{
					push(@{$zInterfaces->{$ifName}->{interfaceVlanTrunks}}, { startVlan => $vlanRange->{lowInt} });
				}
				else
				{
					push(@{$zInterfaces->{$ifName}->{interfaceVlanTrunks}}, { startVlan => $vlanRange->{lowInt}, endVlan => $vlanRange->{highInt} });
				}
			}
		}

		# Record vlan membership information, used in parse_vlans
		my $vlan = undef;
		my $vlanElement = $tElement->first_child("vlan");
		if (defined($vlanElement))
		{
			$vlan = $vlanElement->text();
		}
		elsif ($ifName =~ /^Vlan(\d+)$/)
		{
			$vlan = $1;
		}

		if (defined($vlan) && ($vlan =~ /^\d+$/) && defined($in->{parsed}->{vlans}->{$vlan}))
		{
			push(@{$in->{parsed}->{vlans}->{$vlan}->{interfaceMember}}, $ifName);
		}
		
		# Pull what we can out of the full interface entry for this device for Ethernet address and other properties
		my $fTwig = XML::Twig->new(twig_roots => {
				ROW_interface => sub {
						my ($twig, $element) = @_;

						$zInterfaces->{"$ifName"}->{name} = $element->first_child_text("interface");

						# We need to make our searches for interfaces for virtual ethernet ports be prepended with "veth", since the 
						# tags will be different.
						my $ethernetPropertiesTagPrefix = ($ifName =~ /^Vether/i) ? "veth" : "eth";

						# MTU can come from two different places
						my $mtu = $element->first_child_text($ethernetPropertiesTagPrefix . "_mtu") // $element->first_child_text("svi_mtu");

						if ($mtu)
						{
							# See bug 83444, this is fix for a bug in the Nexus
							# https://tools.cisco.com/quickview/bug/CSCtt41070
							if ($mtu =~ /^[^\d]+(\d+)[^\d]+$/)
							{
								$mtu = $1;
							}

							$zInterfaces->{"$ifName"}->{mtu} = $mtu;
						}
						
						if ($element->first_child_text($ethernetPropertiesTagPrefix . "_hw_addr") =~ /^([\w\d\.]+)/ )
						{
							# Get rid of the delimiting dots
							$zInterfaces->{"$ifName"}->{interfaceEthernet}->{macAddress} = $1;
							$zInterfaces->{"$ifName"}->{interfaceEthernet}->{macAddress} =~ s/\.//g;
						}
						
						if ( $element->first_child($ethernetPropertiesTagPrefix . "_duplex"))
						{
							if ($element->first_child_text($ethernetPropertiesTagPrefix . "_duplex") =~ /^(full|half)/)
							{
								# if interface is down, this can also just contain the value "auto", which
								# isn't a legal assigment to operational duplex, you see....
								$zInterfaces->{"$ifName"}->{interfaceEthernet}->{operationalDuplex} = $1;
							}

							# Nexus default
							$zInterfaces->{"$ifName"}->{interfaceEthernet}->{autoDuplex} = "true";
						}

						# Going to base the loopback determination here as to whether octet-level stats are returned in
						# an element named "loopXXX"
						if ($element->first_child("loop_in_bytes"))
						{
							$zInterfaces->{"$ifName"}->{interfaceEthernet}->{loopBack} = "network";
							$zInterfaces->{"$ifName"}->{interfaceType} = "softwareLoopback";
							$zInterfaces->{"$ifName"}->{physical} = "false";
							push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "switching");
						}
						elsif ($ifName =~ /^mgmt/)  # Only way I can figure this...
						{
							# management interface
							# Disabling admin behavior here.  If this is the only L3 interface (which it will be, on a Nexus 1000v,
							# ) then topology would drop the device.
							# push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "administrative");
							
							# Adding this on to provide ability to form profile for bridged forwarding of L2 traffic
							push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "bridging");							
							$zInterfaces->{"$ifName"}->{physical} = "true";
							$zInterfaces->{"$ifName"}->{interfaceType} = "ethernet";
						}
						else
						{
							# make some bold determinations here
							$zInterfaces->{"$ifName"}->{interfaceType} = "ethernet";
							$zInterfaces->{"$ifName"}->{physical} = "true";
							push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "switching");
						}

						# we may override some of this by looking at port mode.
						if ($portmode eq "access")
						{
							push (@{$zInterfaces->{"$ifName"}->{interfaceBehavior}}, "bridging");														
						}

						given ($tElement->first_child_text("type"))
						{
							when ('virt')
							{
								$zInterfaces->{"$ifName"}->{physical} = "false";
								$zInterfaces->{"$ifName"}->{interfaceType} = "virtual";								
							}
							
							when ('eth')
							{
								$zInterfaces->{"$ifName"}->{physical} = "true";
								$zInterfaces->{"$ifName"}->{interfaceType} = "ethernet";																
							}
						}
												
						if ($element->first_child("svi_arp_type"))
						{
							$zInterfaces->{"$ifName"}->{interfaceEthernet}->{encapsulation} =
								$element->first_child_text("svi_arp_type"); 
						}

						# This is a second place to assess auto_speed
						given ($element->first_child_text($ethernetPropertiesTagPrefix . "_autoneg"))
						{
							when ('on') { 	$zInterfaces->{"$ifName"}->{interfaceEthernet}->{autoSpeed} = "true"; } 
							when ('off'){ 	$zInterfaces->{"$ifName"}->{interfaceEthernet}->{autoSpeed} = "false"; }

						}

						# Now, let's deal with trunking vs. access (VLAN) and EtherChannel trunking
						# relationships.
						# +++wayne not sure this is meaningful, or if it is (which it probably is), what this
						# means to core and how to uniquely associate port-channel 
#						if ($zInterfaces->{"$ifName"}->{interfaceType} ~~ "ethernet")
#						{
#							if ($element->first_child_text($ethernetPropertiesTagPrefix . "_hw_desc") =~ /^Port\-channel/i)
#							{
#								push ( @{$zInterfaces->{"$ifName"}->{interfaceGrouping}}, 
#			   						{ifGroupName => $ifName,
#			    				ifGroupingCharacteristic => "primaryInterface"});
#							}
#						}

				}
		} );
		if ($in->{interfaces}->{"$ifName"})
		{
			$fTwig->parse($in->{interfaces}->{"$ifName"}); # full interface entry for this i/f		
		}
		else
		{
			$LOGGER->debug('parse_interfaces: no interface specific data found in device response for interface $iName.'); 
		}

		return 1;
	}
	else
	{
		# This isn't an explicitly configured interface,.  This is normal for things like Vlan 1 (the "all vlans" meta-vlan)
		return 0;
	}
}

# This subroutine parses both ipv6 and ipv4 network groups from the Nexus
sub _networkGroupParser
{
	my ($element, $objectGroups) = @_;

	my $networkGroup = {};

	my $nameElement = $element->first_child('group_name');
	if (!defined($nameElement))
	{
		$LOGGER->debug("parse_object_groups: Failed to parse name from network object group.");
		return;
	}
	$networkGroup->{id} = $nameElement->text();

	my $groupMembersTableElement = $element->first_child('TABLE_seqno');
	if (!defined($groupMembersTableElement))
	{
		$LOGGER->debug("parse_object_groups: Can't find members for network object group: " . $networkGroup->{id});
		return;
	}

	# Parses an ipv4 or an ipv6 host
	my $parseHost = sub {
			my ($host) = @_;

			if (!defined($networkGroup->{host}))
			{
				$networkGroup->{host} = [];
			}
			push(@{$networkGroup->{host}}, $host);

			return 1;
		};

	# Parses an ipv4 or an ipv6 subnet
	my $parseSubnet = sub {
			my ($subnet) = @_;

			my $address = undef;
			my $mask = undef;
			if ($subnet =~ /^([^\/]*)\/(\d+)/)
			{
				$address = $1;
				$mask = $2;
			}
			else
			{
				return 0;
			}

			if (!defined($networkGroup->{subnet}))
			{
				$networkGroup->{subnet} = [];
			}
			push(@{$networkGroup->{subnet}}, { address => $address, mask => $mask });
		};

	foreach my $groupMember ($groupMembersTableElement->children('ROW_seqno'))
	{
		# Parse out any ipv4 hosts if they exist
		my $hostElement = $groupMember->first_child('hostaddr');
		if (defined($hostElement))
		{
			my $success = $parseHost->($hostElement->text());

			if (!$success)
			{
				$LOGGER->debug("parse_object_groups: Failed to parse ipv4 host: " . $hostElement->text());
			}
		}

		# Parse out any ipv4 subnets if they exist
		my $netElement = $groupMember->first_child('net_ip');
		if (defined($netElement))
		{
			my $success = $parseSubnet->($netElement->text());

			if (!$success)
			{
				$LOGGER->debug("parse_object_groups: Failed to parse ipv4 subnet: " . $netElement->text());
			}
		}
	}
	if (!defined($networkGroup->{host}) && !defined($networkGroup->{subnet}))
	{
		$LOGGER->debug("parse_object_groups: Can't find members for network object group: " . $networkGroup->{id});
		return;
	}

	push(@{$objectGroups->{networkGroups}}, $networkGroup);
}

# This subroutine parses port object groups from the Nexus (stored as service groups in the SED).
sub _serviceGroupParser 
{
	my ($element, $objectGroups) = @_;

	my $serviceGroup = {};

	my $nameElement = $element->first_child('group_name');
	if (!defined($nameElement))
	{
		$LOGGER->debug("parse_object_groups: Failed to parse name from service object group.");
		return;
	}
	$serviceGroup->{id} = $nameElement->text();

	my $groupMembersTableElement = $element->first_child('TABLE_seqno');
	if (!defined($groupMembersTableElement))
	{
		$LOGGER->debug("parse_object_groups: Can't find members for service object group: " . $serviceGroup->{id});
		return;
	}

	foreach my $groupMember ($groupMembersTableElement->children('ROW_seqno'))
	{
		# Parse any port operators
		my $portOpElement = $groupMember->first_child('_port_op');
		if (defined($portOpElement))
		{
			# When a port operator is used we expect a single port value stored in port0_num
			my $portElement = $groupMember->first_child('port0_num');
			if (defined($portElement))
			{
				my $operator = $portOpElement->text();
				my $port = $portElement->text();

				# Nexus operators match operators avaiable in the SED schema, except for the not equals operator.
				if ($operator eq 'neq')
				{
					$operator = 'ne';
				}

				if (!defined($serviceGroup->{portExpression}))
				{
					$serviceGroup->{portExpression} = [];
				}
				push(@{$serviceGroup->{portExpression}}, { port => $port, operator => $operator });
			}
			else
			{
				$LOGGER->debug("parse_object_groups: Faild to parse port0_num for service object group: " . $serviceGroup->{id});
			}
		}

		# Parse any port ranges
		my $portRangeElement = $groupMember->first_child('_port_range');
		if (defined($portRangeElement))
		{
			# When a port range is used we expect two port values storted in port1_num and port2_num
			my $portStartElement = $groupMember->first_child('port1_num');
			my $portEndElement = $groupMember->first_child('port2_num');
			if (defined($portStartElement) && defined($portEndElement))
			{
				my $portStart = $portStartElement->text();
				my $portEnd = $portEndElement->text();

				if (!defined($serviceGroup->{portRange}))
				{
					$serviceGroup->{portRange} = [];
				}
				push(@{$serviceGroup->{portRange}}, { portEnd => $portEnd, portStart => $portStart });
			}
			else
			{
				$LOGGER->debug("parse_object_groups: Faild to parse port1_num and port2_num for service object group: " . $serviceGroup->{id});
			}
		}
	}
	if (!defined($serviceGroup->{portExpression}) && !defined($serviceGroup->{portRange}))
	{
		$LOGGER->debug("parse_object_groups: Can't find members for service object group: " . $serviceGroup->{id});
		return;
	}

	push(@{$objectGroups->{serviceGroups}}, $serviceGroup);
}


# This takes one argument, a reference to a scalar.
# It returns reference to a closure that can be used
# as a twig_handler to simply set the scalar referenced
# to the text context of the element the handler was
# triggered to handle.
#
# Note: If there is more than one instance of each element
#       the referenced scalar will end up being set to the
#       value of the last element.  createSimpleTwigSetter
#       probably will not be of any use if you expect
#       more than one instance of the element.
sub _createSimpleTwigSetter
{
	my ($varRef) = @_;
	return sub {
		my ($twig, $element) = @_;
		$$varRef = $element->text();
		return 1;
	};
}

sub _printInterface
{
	my ($zInterface, $out) = @_;
	my @orderedInterfaceZedFields = (
			"adminStatus", "description", "egressFilter",
			"failover", "ifIndex", "ingressFilter", 
			"interfaceEthernet", "interfaceGrouping",
			"interfaceIp", "interfaceBehavior", "interfaceType",
			"interfaceVlanTrunks", "mtu", "name", "physical",
			"speed"
	);	
	
	# start with an interface entry here...
	$out->open_element("cisco:interface");
	
	for my $nextIfField (@orderedInterfaceZedFields)
	{
		if (defined $zInterface->{$nextIfField})
		{
			$out->print_element($nextIfField, $zInterface->{$nextIfField});
		}	
	}
	
	$out->close_element("cisco:interface");	
}

#
#  This will scan the "RowSeqno"  term scoped here for the term data
#  to comprise a filter spec.
# 
#  $in - entire response data from context
#  $twigElement - element corresponding to ROW_Seqno in an access list
#  $filterEntry - term to "fill in" with ZED filterEntry conformant element names/values
#
#  RETURNS 'sequence number' of term (required element in Nexus access list spec)
sub _parse_alSpecTerm
{
	my ($in, $alName, $twigElement, $filterEntry) = @_;
	my $seqno = $twigElement->first_child_text('seqno');
	
	defined $seqno ||$LOGGER->debug("Cisco::Nexus::parse_access_lists; cannot find term sequence number in $alName");

	$filterEntry->{name} = $seqno;
	
	my $fEltText;
	my $nlPortHi;
	my $nlPortLo;
	my $nlPortOp;
	my $feDescription;
	my $primaryProtocol;
	
		# Assign mandatory elements with default values
	$filterEntry->{log} = "false";
	$filterEntry->{primaryAction} = "next"; # if nothing but metadata is in this filter, then simply pass to the next one

	foreach ($twigElement->children())
	{
		# Throw out structural metadata stuff
		next if (not ($_->is_elt && $_->string_value));
		$fEltText = $_->string_value;
		my $fEltTag = $_->name;		
		
		given($fEltTag)
		{
			when (/^dest_(\S+)/) { 
			    
			    # we have to special case the port numbers, so that we can locally
			    # assign them for fixup once the filter spec has been put together
			    if ($1 =~ /^port(\d)_num/)
			    {
					given ($1)
					{
					    when (1) { $nlPortLo->{destination} = $fEltText; }
					    when (2) { $nlPortHi->{destination} = $fEltText; }
					}
			    }
			    elsif ($1 =~ /^port(\d)_str/)
			    {
					given ($1)
					{
						# +++wayne this will return a single port number.  Do we need to worry about
						# multiple noted port numbers logically associated with one port (e.g., netbios)?
					    when (1) { $nlPortLo->{destination} = get_port_number($fEltText, $primaryProtocol ); }
					    when (2) { $nlPortHi->{destination} = get_port_number($fEltText, $primaryProtocol); }
					}			    	
			    }
			    elsif ($1 =~ /^port_op/)
			    {
					$nlPortOp->{destination} = $fEltText;
			    }
			    else
			    {
				# let the common dest/src argument handler deal with it
					_parse_access_list_entry_direction_options($fEltTag, $fEltText, 'dest', 'destination',
							       $filterEntry );

			    }
				
			}
			 
			# ""established" ... shows up as ""<established>enable</established>"
			when (/^established/) { $filterEntry->{established} = ($fEltText =~ /^enable/i) ? "true" : "false";} 
			# "log" is similarly encoded with an "enable" 
			when (/^log/) { $filterEntry->{log} = ($fEltText =~ /^enable/i) ? "true" : "false";} 
			when ("precedence_str") { $filterEntry->{precedence} = $fEltText; }  # nonenumerated anyways
			# dscp <-> tos : specify numerically
			when ("dscp") { $filterEntry->{tos} = $fEltText; } 
			
			when ("proto_str") 
			{
				 push(@{$filterEntry->{protocol}},  $fEltText);   # All cases I've seen match enumeration
				 $primaryProtocol = $fEltText;
			}
			when ("permitdeny") { $filterEntry->{primaryAction} = $fEltText; }
			when ("remark") { $feDescription = $fEltText unless ($feDescription); }
			# ICMP message filters go into "protocol filter"
			when ("icmp_str") { $filterEntry->{protocolFilter} = $fEltText; }
			# "[permit|deny] ip ..." seems to turn into a field *called* ip.  *Don't* assign $primaryProtocol here.
			when ("ip")  { push(@{$filterEntry->{protocol}}, "ip") if ($fEltText =~ /^ip$/i); }  # "ip" seems to be the only value, too  			
			# All source fields
			when (/^src_(\S+)/) { 
			    
			    # we have to special case the port numbers, so that we can locally
			    # assign them for fixup once the filter spec has been put together
			    if ($1 =~ /^port(\d)_num/)
			    {
					given ($1)
					{
					    when (1) { $nlPortLo->{source} = $fEltText; }
					    when (2) { $nlPortHi->{source} = $fEltText; }
					}
			    }
			    elsif ($1 =~ /^port(\d)_str/)
			    {
					given ($1)
					{
						# +++wayne this will return a single port number.  Do we need to worry about
						# multiple noted port numbers logically associated with one port (e.g., netbios)?
					    when (1) { $nlPortLo->{source} = get_port_number($fEltText, $primaryProtocol ); }
					    when (2) { $nlPortHi->{source} = get_port_number($fEltText, $primaryProtocol); }
					}			    	
			    }
			    elsif ($1 =~ /^port_op/)
			    {
					$nlPortOp->{source} = $fEltText;
			    }
			    else
			    {
				# let the common dest/src argument handler deal with it
					_parse_access_list_entry_direction_options($fEltTag, $fEltText, 'src', 'source',
							       $filterEntry );

			    }
				
			}
		}
	}
	
	# Now, we have to see how to assign the port op plus arguments....
	foreach my $gDir ("source", "destination")
	{
		my $gServ;
		my $gOp = $nlPortOp->{$gDir};
		
		if (not defined $gOp)
		{
			; # nothing
		}
	    # There has to be an operator for value, and the one we special case is 'range'...
	    elsif ($gOp eq 'range')
	    {
			if (defined $nlPortLo->{$gDir} and defined $nlPortHi->{$gDir})
			{
			    $gServ->{portRange}->{portStart} =	$nlPortLo->{$gDir};
			    $gServ->{portRange}->{portEnd} =	$nlPortHi->{$gDir};
			}
			else
			{
			   $LOGGER->debug("Cisco::Nexus::parse_access_lists; could not parse port arguments " .
					  "for range  in spec $seqno of access list $alName");
			}
	    }
	    else  # if defined otherwise, 'tis unary operator
	    {
			if (defined $nlPortLo->{$gDir})
			{
				# Mapping for range
				$gOp =~ s/neq/ne/;
			    $gServ->{portExpression}->{operator} = $gOp;
			    $gServ->{portExpression}->{port} =	$nlPortLo->{$gDir};
			}
			else
			{
		    	$LOGGER->debug("Cisco::Nexus::parse_access_lists; could not parse port argument " .
					  "for operator \' $gOp \'  in spec $seqno of " .
				  	" access list $alName");
			}
		}
		
		if ($gServ)
		{
			push (@{$filterEntry->{$gDir . "Service"}}, $gServ);
		}
	}

	return ($seqno, $feDescription);
}

# This parses "source" or "dest" arguments for a filter spec.  NOTE that actual port numbers and port ops are 
# parsed by the parent, since they have to be "fixed up", depending on port operator.
sub _parse_access_list_entry_direction_options
{
	my ($elementTag, $elementSData, $tagDirection, $generatedDirection, $filterEntry) = @_;
	my $gAddr;
	
	# Reality check that we're parsing the intended directional element
	if ($elementTag !~ /^$tagDirection/i )
	{
		$LOGGER->debug("Cisco::Nexus::parse_access_lists; _parse_access_list_entry_direction (direction $tagDirection) called with element $elementTag");
		return;	
	}	
	
	given ($elementTag)
	{
		when (/^${tagDirection}_(\S+)/) 
		{
			given ($1)
			{
				when ("ip_prefix") {
					if ($elementSData =~ /^$cidr_regex/)
					{
						my $cAddr = parseCIDR($elementSData);
							
						# see if this is a /32
						if ( $cAddr->{prefix} == 32)
						{
							$gAddr->{host} = $cAddr->{host};
						}		
						else
						{
							# a real network
							$gAddr->{network} = create_network_element($cAddr->{host}, $cAddr->{network});
						}
					}
					else
					{
						$LOGGER->debug("Cisco::Nexus::parse_access_lists; cannot parse CIDR dest mask: $elementSData");
					}
					
				} # end of "ip_prefix"
				when ("any") {$gAddr->{network} = get_any_address_encoding(0); }
				when ("addrgrp") {$gAddr->{objectGroupReference} = $elementSData; }
				when ("portgrp") {push (@{$filterEntry->{$generatedDirection . "Service"}}, 
								{ objectGroupReference => $elementSData}); }
			} # end of "given $1 "	
		} # end of "when $tagDirection"...
	} # end of "given element tag"

	if ($gAddr)
	{
		push(@{$filterEntry->{$generatedDirection . "IpAddr"}}, $gAddr);						
	}
}

sub _clean_interfaces {
	my $in = shift;

	my $interfacesValid = $in->{interface};
	$interfacesValid =~ s/(<\/.*)nftlv\(output\):.*/$1>/mg;
	$interfacesValid =~ s/nftlv\(output\):.*|^\s+>$//mg;
	$interfacesValid =~ s/^\s*\n+//mg;
	return ($interfacesValid);
}

1;

__END__

=pod

=head1 NAME

    Parsers.pm

=head1 DESCRIPTION

    Parsing logic for Cisco Nexus configuration.

=head2 Methods

=over

=item parse_system() 

=item parse_access_ports()

=item parse_chassis()

=item parse_interfaces()

=item parse_object_groups()

=item parse_access_lists()

=item parse_static_routes()

=item parse_dynamic_routes()

=item open_vdc_section()

=item close_vdc_section()

=item create_config()

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

Portions created by IBM are Copyright (C) 2012, IBM Corp.

=head1 AUTHOR

Contributor(s): Wayne Tackabury (wtackabury@us.ibm.com) David Payne (david.payne@ca.ibm.com)
Date: November 15, 2012

=cut
