package ZipTie::Adapters::Cisco::Nexus::Telemetry;

use strict;
use warnings;

use ZipTie::Logger;
use ZipTie::CLIProtocol;
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle choose_admin_ip);
use ZipTie::Model::XmlPrint;

use ZipTie::Adapters::Cisco::Nexus::Disconnect qw(disconnect);
use ZipTie::Adapters::Cisco::Nexus::TelemetryParsers qw(parse_telemetry_interfaces parse_arp parse_mac_table parse_cdp parse_routing_neighbors);
use ZipTie::Adapters::Cisco::Nexus::ResponseHandler qw(populate_virtual_device_context delete_instance_data get_virtual_device_contexts netconf_elements_extract);

my $LOGGER = ZipTie::Logger::get_logger();

# If this is XML over SSH, we issue commands with the " | xml" CLI designator
use constant NEXUS_XML_COMMANDS_XML_SSH => 1;
# if this is a real Netconf connection, then we issue Netconf requests to the port requested
use constant NEXUS_XML_COMMANDS_NETCONF => 2;
# The real kicker is that at least one device (the Nexus 4000) in fact does not support Netconf or
# XML command formatting *at all*.  We actually have to back off to a pure CLI text response mode of
# parsing, and the command as issued has no XML form or request designator (although we'll try the latter
# and glean if we got an error to see if we're in this predicament
use constant NEXUS_XML_NO_XML_TEXT_ONLY => 3;

# The following show commands are piped to xml on the Nexus device.
my $XML_COMMANDS = [ 'interface brief', 'ip arp', 'mac address-table', 'ip route' ];

# The following show commands are not piped to xml on the Nexus device.
my $TEXT_COMMANDS = [ 'cdp entry all' ];

sub invoke
{
	my $pkg = shift;
	my $backup_doc = shift;

	my $responses = {};

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ($connection_path, $discovery_params) = ZipTie::Typer::translate_document($backup_doc, 'connectionPath');

	# Create a new CLI protocol object
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	# Make a connection to and successfully authenticate with the IOS device
	my $exec_prompt_regex = ZipTie::Adapters::Cisco::Nexus::AutoLogin::execute($cli_protocol, $connection_path);

	# Store the found prompt as "enablePrompt" on the specified CLI protocol.
	$cli_protocol->set_prompt_by_name('execPrompt', $exec_prompt_regex);

	if ($cli_protocol->get_protocol_name() =~ /Netconf/i)
	{
		# Since we're here, add the known element namespace(s) for later strip
		$cli_protocol->add_netconf_element_namespace("mod");
	}
	else
	{
		# Get rid of the more prompt--only applicable for terminal protocols
		my $termLen = $cli_protocol->send_and_wait_for('terminal length 0', $exec_prompt_regex);		
	}	

	# Grab an output filehandle for the model
	my $filehandle = get_model_filehandle('Nexus', $connection_path->get_ip_address());

	# initialize the model printer
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, 'telemetry');
	$printer->attributes(1);
	$printer->open_discovery_event();

	# Need to extract a host name.  It may be domain qualified.  We will try to extract it with 
	# XML to be able to determine what the deal is with whether we can assume documented XML behavior
	# or if that feature is snipped off (Nexus 4000)
	my $hostName = netconf_elements_extract( "show hostname | xml", $exec_prompt_regex, $cli_protocol );
	
	if ($hostName)
	{
		# fine, there's a conformant SSH -> XML formatter
		$responses->{xmlForm} = NEXUS_XML_COMMANDS_XML_SSH;
		# get the hostname out of the XML
		if ($hostName =~ /\<hostname\>(\S+?)\<\/hostname/im)
		{
			$hostName = $1;  # just the host, at the most
		}
	}
	else
	{	
		$responses->{xmlForm} = NEXUS_XML_NO_XML_TEXT_ONLY;
		# by the way, we *do* need a host name
		$hostName = $cli_protocol->send_and_wait_for('show hostname', $exec_prompt_regex );
		$hostName =~ s/.*?\n([\w_\-\.]+)[\s\r\n]+.*/$1/sm;
	}
	
	my $deviceContextList = get_virtual_device_contexts($responses, $hostName, $cli_protocol, $exec_prompt_regex);

	# Keep track if the subdevices element is open or not
	my $subdevicesOpen = 0;

	foreach my $vdcName (undef, @{$deviceContextList})
	{
		my $parseResponses;

		# We are collecting data for a VDC
		if (defined($vdcName))
		{
			# Issue the commands for the context
			populate_virtual_device_context({responses => $responses, cliProtocolInst => $cli_protocol, promptRegex => $exec_prompt_regex,
				xmlShowCommands => $XML_COMMANDS, textShowCommands => $TEXT_COMMANDS, vdcName => $vdcName});

			if (!defined($responses->{contexts}->{$vdcName}))
			{
				$LOGGER->debug("Cisco::Nexus::Telemetry::invoke : error encountered in populating virtual device context '$vdcName'");
				next;
			}

			$parseResponses = $responses->{contexts}->{$vdcName};

			if (!$subdevicesOpen)
			{
				$printer->open_element('subdevices');
				$subdevicesOpen = 1;
			}

			$printer->open_element('subdevice');
			$printer->print_element('subdeviceName', $vdcName);
		}
		# We are collecting data for the admin device
		else
		{
			# Issue the commands
			populate_virtual_device_context({responses => $responses, cliProtocolInst => $cli_protocol, promptRegex => $exec_prompt_regex,
				xmlShowCommands => $XML_COMMANDS, textShowCommands => $TEXT_COMMANDS});

			$parseResponses = $responses;
		}

		$printer->open_element('neighbors');

		parse_cdp($parseResponses, $printer);

		parse_arp($parseResponses, $printer);

		parse_routing_neighbors($parseResponses, $printer);

		parse_mac_table($parseResponses, $printer);

		$printer->close_element('neighbors');

		my $interface_hash = parse_telemetry_interfaces($parseResponses, $printer);

		if (defined($vdcName))
		{
			$printer->close_element('subdevice');
		}
		else
		{
			# Only print admin IP address if this isn't a VDC
			my $admin_ip = $connection_path->get_ip_address();
			if ($discovery_params->{calculateAdminIp} eq 'true')
			{
				$admin_ip = choose_admin_ip($admin_ip, $interface_hash);
			}
			$printer->print_element('adminIp', $admin_ip);
		}
	}

	if ($subdevicesOpen)
	{
		$printer->close_element('subdevices');
	}

	# tear down work
	$printer->close_element('DiscoveryEvent');
	close_model_filehandle($filehandle);
	disconnect($cli_protocol);

	# remove instance data
	delete_instance_data({responses => $responses, textShowCommands => $TEXT_COMMANDS, xmlShowCommands => $XML_COMMANDS});
}

1;

__END__

=head1 NAME

ZipTie::Adapters::Cisco::Nexus::Telemetry - Impliments the Telemetry operation for the Cisco Nexus adapter.

=head1 SYNOPSIS

	use ZipTie::Adapters::Cisco::Nexus::Telemetry;
	ZipTie::Adapters::Cisco::Nexus::Telemetry::invoke( $backup_document );

=head1 DESCRIPTION

This module represents an the Telemetry operation for the Cisco Nexus adapter.  Collects Neighbor Data from the Cisco Nexus device.

Generally you would run this module through the ZipTie C<invoke.pl> script.

=head1 PUBLIC SUB-ROUTINES

=over 12

=item C<invoke($backup_document)>

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

Portions created by IBM are Copyright (C) 2013, IBM Corp.

=head1 AUTHOR

Contributor(s): David Payne (david.payne@ca.ibm.com)
Date: February 22, 2013

=cut
