package ZipTie::Adapters::Cisco::Nexus::TelemetryParsers;

use strict;
use warnings;

use XML::Twig;

use ZipTie::Model::XmlPrint;
use ZipTie::Logger;
use ZipTie::Adapters::Utils qw(get_interface_type strip_mac trim get_crep);
use ZipTie::Adapters::Cisco::Nexus::Utils qw(expand_interface_name);


use Exporter 'import';
our @EXPORT_OK = qw(parse_telemetry_interfaces parse_arp parse_mac_table parse_cdp parse_routing_neighbors);

my $LOGGER = ZipTie::Logger::get_logger();

sub parse_telemetry_interfaces
{
	my ($in, $out) = @_;

	my @interfaces;

	my $parseTelemetryInterface = sub {
			my ($twig, $element) = @_;

			my $name = $element->first_child('interface')->text();

			# Any of the following tags will be accepted as state.
			my $stateElement;
			foreach my $stateTag ('state', 'svi_admin_state')
			{
				$stateElement = $element->first_child($stateTag);

				if (defined($stateElement))
				{
					last;
				}
			}

			my $state;
			if (!defined($stateElement))
			{
				$LOGGER->debug("Interface $name, unable to determine state.  Interface not parsed.");
				return 1;
			}
			else
			{
				$state = $stateElement->text();
			}

			# Convert the state string insto something accepted by the schema.
			if ($state eq 'up')
			{
				$state = 'Up';
			}
			elsif ($state eq 'down')
			{
				$state = 'Down';
			}
			elsif ($state eq 'administratively down')
			{
				$state = 'Down';
			}
			else
			{
				$LOGGER->debug("Interface $name, unknown state: $state.  Marking state as 'Unknown'.");
				$state = 'Unknown';
			}

			my $ip = $element->first_child('eth_ip_addr');
			my $mask = undef;
			if (defined($ip))
			{
				$ip = $ip->text();
				my $maskElement = $element->first_child('eth_ip_mask');
				if (defined($maskElement))
				{
					$mask = $maskElement->text();
				}
			}

			my $type = get_interface_type($name);

			my $interface = {};
			$interface->{name} = $name;
			$interface->{operStatus} = $state;
			$interface->{type} = $type;
			$interface->{inputBytes} = 0; # In layer2 inputBytes will be made optional, for now set to 0.
			if (defined($ip))
			{
				my $ipEntry = {};
				$ipEntry->{ipAddress} = $ip;
				if (defined($mask))
				{
					$ipEntry->{mask} = $mask;
				}
				$interface->{ipEntry} = [$ipEntry];
			}
			push(@interfaces, $interface);

			return 1;
		};

	my $twig = XML::Twig->new(twig_roots => { ROW_interface => $parseTelemetryInterface });
	$twig->parse($in->{interface});

	if (scalar(@interfaces))
	{
		$out->open_element('interfaces');

		foreach my $interface (@interfaces)
		{
			$out->print_element('interface', $interface);
		}

		$out->close_element('interfaces');
	}

	# Return the hash expceted by choose_admin_ip
	my $interfaceHash = {};
	$interfaceHash->{interface} = \@interfaces;
	return $interfaceHash;
}

sub parse_arp
{
	my ($in, $out) = @_;

	my @arpEntries;

	my $parseArpEntry = sub {
			my ($twig, $element) = @_;

			my $interface = $element->first_child('intf-out');
			if (!defined($interface))
			{
				$LOGGER->debug('Failed to get interface name for arpEntry.  arpEntry not parsed.');
				return 1;
			}
			$interface = $interface->text();

			my $ipAddress = $element->first_child('ip-addr-out');
			if (!defined($ipAddress))
			{
				$LOGGER->debug('Failed to get ipAddress for arpEntry.  arpEntry not parsed.');
				return 1;
			}
			$ipAddress = $ipAddress->text();

			my $macAddress = $element->first_child('mac');
			if (!defined($macAddress))
			{
				$LOGGER->debug('Failed to get macAddress for arpEntry.  arpEntry not parsed.');
				return 1;
			}
			$macAddress = strip_mac($macAddress->text());

			my $arpEntry = {};
			$arpEntry->{interface} = $interface;
			$arpEntry->{ipAddress} = $ipAddress;
			$arpEntry->{macAddress} = $macAddress;
			push(@arpEntries, $arpEntry);

			return 1;
		};

	my $twig = XML::Twig->new(twig_roots => { ROW_adj => $parseArpEntry });
	$twig->parse($in->{arp});

	if (scalar(@arpEntries))
	{
		$out->open_element('arpTable');

		foreach my $arpEntry (@arpEntries)
		{
			$out->print_element('arpEntry', $arpEntry);
		}

		$out->close_element('arpTable');
	}
}

sub parse_mac_table
{
	my ($in, $out) = @_;

	my @macEntries;

	my $parseMacEntry = sub {
			my ($twig, $element) = @_;

			my $macAddress = $element->first_child('disp_mac_addr');
			if (!defined($macAddress))
			{
				$LOGGER->debug('Failed to get macAddress name for macEntry.  macEntry not parsed.');
				return 1;
			}
			$macAddress = strip_mac($macAddress->text());

			my $interface = $element->first_child('disp_port');
			if (!defined($interface))
			{
				$LOGGER->debug('Failed to get interface name for macEntry.  macEntry not parsed.');
				return 1;
			}
			$interface = trim($interface->text());

			$interface = expand_interface_name($interface);

			# vlan is optional
			my $vlan = $element->first_child('disp_vlan');
			if (defined($vlan))
			{
				$vlan = $vlan->text();
			}

			my $macEntry = {};
			$macEntry->{macAddress} = $macAddress;
			$macEntry->{interface} = $interface;
			if (defined($vlan))
			{
				$macEntry->{vlan} = $vlan;
			}
			push(@macEntries, $macEntry);

			return 1;
		};

	my $twig = XML::Twig->new(twig_roots => {  ROW_mac_address => $parseMacEntry });
	$twig->parse($in->{mac});

	if (scalar(@macEntries))
	{
		$out->open_element('macTable');

		foreach my $macEntry (@macEntries)
		{
			$out->print_element('macEntry', $macEntry);
		}

		$out->close_element('macTable');
	}
}

sub parse_cdp
{
	my ($in, $out) = @_;

	my @cdpEntries;

	my $ipRe = get_crep('cipm');

	while ($in->{cdp} =~ /([^-]\S.*?)(?=-{6,}|$)/sg)
	{
		my $cdpText = $1;

		# All fields are optional, but we want at least one peice of information on
		# the remote system before we add the cdp entry.
		my $remoteData = 0;

		my $systemName = undef;
		if ($cdpText =~ /^System Name:(.+?)\s*$/m)
		{
			$systemName = $1;
			$remoteData = 1;
		}

		my $systemDescription = undef;
		if ($cdpText =~ /^Version:(.+?)Advertisement Version/ms)
		{
			$systemDescription = trim($1);
			$remoteData = 1;
		}

		my $localInterface = undef;
		my $remoteInterface = undef;
		if ($cdpText =~ /^Interface: (\S+), Port ID \(outgoing port\): (\S+)/m)
		{
			$localInterface = $1;
			$remoteInterface = $2;
			$remoteData = 1;
		}

		# Schema only allows one IP address, so only take the last one.
		# We take the last one because if it exists it should be the management IP address.
		# TODO: Parse IPv6 addresses
		my $ipAddress = undef;
		if ($cdpText =~ /.*IPv4 Address: ($ipRe)/s)
		{
			$ipAddress = $1;
			$remoteData = 1;
		}

		my $platform = undef;
		if ($cdpText =~ /^Platform: (.+?)\s*$/m)
		{
			$platform = $1;
			$remoteData = 1;
		}

		if ($remoteData)
		{
			my $cdpEntry = {};
			$cdpEntry->{protocol} = 'CDP';
			$cdpEntry->{sysName} = $systemName if (defined($systemName));
			$cdpEntry->{sysDescr} = $systemDescription if (defined($systemDescription));
			$cdpEntry->{localInterface} = $localInterface if (defined($localInterface));
			$cdpEntry->{remoteInterface} = $remoteInterface if (defined($remoteInterface));
			$cdpEntry->{ipAddress} = $ipAddress if (defined($ipAddress));
			$cdpEntry->{platform} = $platform if (defined($platform));
			push(@cdpEntries, $cdpEntry);
		}
	}

	if (scalar(@cdpEntries))
	{
		$out->open_element('discoveryProtocolNeighbors');

		foreach my $cdpEntry (@cdpEntries)
		{
			$out->print_element('discoveryProtocolNeighbor', $cdpEntry);
		}

		$out->close_element('discoveryProtocolNeighbors');
	}
}

sub parse_routing_neighbors
{
	my ($in, $out) = @_;

	my @routes;

	my $parseRoute = sub {
			my ($twig, $element) = @_;

			my $protocol = $element->first_child('clientname');
			if (!defined($protocol))
			{
				$LOGGER->debug("Failed to parse protocol for route.  Route not parsed.");
				return 1;
			}
			$protocol = $protocol->text();

			if ($protocol =~ /eigrp/i)
			{
				$protocol = 'EIGRP';
			}
			elsif ($protocol =~ /static/i)
			{
				# Static explicitly not supported for 7.1 MR1.  Add support later.
				return 1;
			}
			elsif ($protocol =~ /local/i)
			{
				# We don't care about local routes.
				return 1;
			}
			elsif ($protocol =~ /hsrp/i)
			{
				# Possibilly add support to the schema for HSRP routes, for now parse as UNKNOWN.
				$protocol = 'UNKNOWN';
			}
			elsif ($protocol =~ /direct/i)
			{
				# TODO: Find out what direct routes are, and see if we should set protocol to something else.
				$protocol = 'UNKNOWN';
			}
			# The following is untested.
			elsif ($protocol =~ /ospf/i)
			{
				$protocol = 'OSPF';
			}
			# The following is untested.
			elsif ($protocol =~ /igrp/i)
			{
				$protocol = 'IGRP';
			}
			# The following is untested.
			elsif ($protocol =~ /bgp/i)
			{
				$protocol = 'BGP';
			}
			# The following is untested.
			elsif ($protocol =~ /isis/i)
			{
				$protocol = 'ISIS';
			}
			# The following is untested.
			elsif ($protocol =~ /rip2/i)
			{
				$protocol = 'RIP2';
			}
			# The following is untested.
			elsif ($protocol =~ /rip/i)
			{
				$protocol = 'RIP';
			}
			else
			{
				$LOGGER->debug("Unknown protocol: $protocol for route.  Route not parsed.");
				return 1;
			}

			my $nextHop = $element->first_child('ipnexthop');
			if (!defined($nextHop))
			{
				$LOGGER->debug("Failed to parse next hop for route.  Route not parsed.");
				return 1;
			}
			$nextHop = $nextHop->text();

			my $interface = undef;
			my $ifShortName = $element->first_child('ifname');
			if (defined($ifShortName))
			{
				$interface = expand_interface_name($ifShortName->text());
			}

			my $route = {};
			$route->{interface} = $interface if defined($interface);
			$route->{ipAddress} = $nextHop;
			$route->{protocol} = $protocol;
			$route->{routerId} = $nextHop; # TODO: RouterID has been made optional in layer2, this line should be removed in layer2.
			push(@routes, $route);

			return 1;
		};

	my $twig = XML::Twig->new(twig_roots => { ROW_path => $parseRoute });
	$twig->parse($in->{route});

	if (scalar(@routes))
	{
		$out->open_element('routingNeighbors');

		foreach my $route (@routes)
		{
			$out->print_element('routingNeighbor', $route);
		}

		$out->close_element('routingNeighbors');
	}
}

1;

__END__

=pod

=head1 NAME

    TelemetryParsers.pm

=head1 DESCRIPTION

    Parsing logic for Cisco Nexus telemetry configuration.

=head2 Methods

=over

=item parse_telemetry_interfaces() 

=item parse_arp() 

=item parse_mac_table()

=item parse_cdp()

=item parse_routing_neighbors()

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

Portions created by IBM are Copyright (C) 2013, IBM Corp.

=head1 AUTHOR

Contributor(s): David Payne (david.payne@ca.ibm.com)
Date: February 22, 2013

=cut
