package ZipTie::Adapters::Cisco::SecurityAppliance::AclToRoute;

use strict;
use warnings;

use ZipTie::Adapters::Cisco::SecurityAppliance::Common qw(break_network_object_into_hosts_subnets_ranges);

use Net::CIDR::Set;

sub new
{
	my ($class) = @_;

	my $self = {
			_permitCidrSet => undef,
			_denyCidrSet => undef
		};

	bless($self, $class);

	return $self;
}

sub addPermitAddresses
{
	my ($self, $addresses) = @_;

	if (!defined($self->{_permitCidrSet}))
	{
		$self->{_permitCidrSet} = Net::CIDR::Set->new();
	}

	# Generate a new set with the addresses
	my $newAddresses = Net::CIDR::Set->new(@{$addresses});

	# This will remove any addresses in _denyCidrSet from the newAddresses set
	if (defined($self->{_denyCidrSet}))
	{
		$newAddresses = $newAddresses->diff($self->{_denyCidrSet});
	}

	# Add the remaining addresses to the _permitCidrSet
	$self->{_permitCidrSet}->merge($newAddresses);
}

sub addDenyAddresses
{
	my ($self, $addresses) = @_;

	if (!defined($self->{_denyCidrSet}))
	{
		$self->{_denyCidrSet} = Net::CIDR::Set->new();
	}

	$self->{_denyCidrSet}->add(@{$addresses});
}

sub getCidrList
{
	my ($self) = @_;

	if (defined($self->{_permitCidrSet}))
	{
		my @cidrs = $self->{_permitCidrSet}->as_cidr_array();

		# Convert IP to CIDR
		foreach my $cidr (@cidrs)
		{
			if ($cidr !~ /\//)
			{
				# Convert to IPv4 CIDR
				if ($cidr !~ /:/)
				{
					$cidr = "$cidr/32";
				}
				# Convert to IPv6 CIDR
				else
				{
					$cidr = "$cidr/128";
				}
			}
		}

		return \@cidrs;
	}
	else
	{
		return [];
	}
}

sub convertAclToRoute
{
	my ($self, $aclName, $filterLists, $networkGroups) = @_;

	foreach my $filterEntry (@{$filterLists->{$aclName}})
	{
		my $filterEntryDestinations = [];

		foreach my $destinationIpAddr (@{$filterEntry->{destinationIpAddr}})
		{
			my $hosts = [];
			my $subnets = [];
			my $ranges = [];

			if (defined($destinationIpAddr->{host}))
			{
				push(@{$hosts}, $destinationIpAddr->{host});
			}
			if (defined($destinationIpAddr->{network}))
			{
				push(@{$subnets}, $destinationIpAddr->{network});
			}
			if (defined($destinationIpAddr->{range}))
			{
				push(@{$ranges}, $destinationIpAddr->{range});
			}
			if (defined($destinationIpAddr->{objectGroupReference}))
			{
				my ($refHosts, $refSubnets, $refRanges) = break_network_object_into_hosts_subnets_ranges($networkGroups, $destinationIpAddr->{objectGroupReference});
				push(@{$hosts}, @{$refHosts});
				push(@{$subnets}, @{$refSubnets});
				push(@{$ranges}, @{$refRanges});
			}

			push(@{$filterEntryDestinations}, @{$hosts});
			foreach my $subnet (@{$subnets})
			{
				push(@{$filterEntryDestinations}, $subnet->{address} . '/' . $subnet->{mask});
			}
			foreach my $range (@{$ranges})
			{
				push(@{$filterEntryDestinations}, $range->{startAddress} . '-' . $range->{endAddress});
			}
		}

		if (!scalar(@{$filterEntryDestinations}))
		{
			# $filterEntryDestination empty in ACL, adding 0.0.0.0/0
			push(@{$filterEntryDestinations}, '0.0.0.0/0');
		}

		my $action = $filterEntry->{primaryAction};

		if ($filterEntry->{primaryAction} eq 'permit')
		{
			addPermitAddresses($self, $filterEntryDestinations);
		}
		else
		{
			addDenyAddresses($self, $filterEntryDestinations);
		}
	}
}

sub getRoutes
{
	my ($self, $interface) = @_;

	my $cidrList = getCidrList($self);

	my $routes = [];

	foreach my $cidr (@{$cidrList})
	{
		if ($cidr =~ /([^\/]+)\/(\d+)/)
		{
			my $ip = $1;
			my $mask = $2;

			my $route = {
					defaultGateway => 'false',
					destinationAddress => $ip,
					destinationMask => $mask,
					interface => $interface,
					routeMetric => '0'
				};

			push(@{$routes}, $route);
		}
	}

	return $routes;
}

=head1 NAME

	ZipTie::Adapters::Cisco::SecurityAppliance::AclToRoute
	Convert ACLs into routes for ASA P2P tunnels.

=head1 DESCRIPTION

	This module generates routes based on the destination portion of access lists.
	It access-lists and network groups are passed in as a perl data structure
	representing the SED schema.

=head1 AUTHOR(S)

	David Payne
	david.payne@ca.ibm.com

=head1 COPYRIGHT

	Copyright (c) 2013, IBM Corporation. All Rights Reserved.
	This module is free software. It may be used, redistributed
	and/or modified under the same terms as Perl itself.

=cut

1;
