#!/usr/bin/perl

package ZipTie::Adapters::Cisco::SecurityAppliance::AddressTranslationParser;

use strict;
use warnings;

use ZipTie::Logger;
use ZipTie::Model::AddressTranslation;
use ZipTie::Adapters::Utils qw(get_crep get_port_number mask_to_bits);
use ZipTie::Model::XmlPrint;
use ZipTie::Adapters::Cisco::SecurityAppliance::NameMapper;
use ZipTie::Adapters::Cisco::SecurityAppliance::Common qw(break_network_object_into_hosts_subnets_ranges);

use Net::IP;
use Math::BigInt;
use Data::Dumper;

sub new
{
	my ($class, $in, $security_levels) = @_;

	my $isUnderNatControl = $in->{running_config} =~ /^nat-control/mig;

	# Store the interface IPs in an array (they were in a hash to dedup).
	my $interfaceIps = {};
	foreach my $key (keys(%{$in->{_interface_ips}}))
	{
		my @keys = keys(%{$in->{_interface_ips}->{$key}});
		$interfaceIps->{$key} = \@keys;
	}

	my $self = {
		_in => $in,
		_logger => ZipTie::Logger::get_logger(),
		_zoneMapperHash => _parseZoneMapperHash($in->{running_config}),
		_is_under_nat_control => $isUnderNatControl,
		_network_groups => $in->{_network_groups},
		_interface_ips => $interfaceIps,
		_security_levels => $security_levels
	};

	bless($self, $class);
	return $self;
}

# Generates the NAT XML.  The generated XML will not be printed to the documnet, but can be retreived via the
# getPreRoutingFilterLists and getPostRoutingFilterLists methods.
sub generateXml
{
	# Pre-8.3 NAT order of operations.
	# http://www.cisco.com/en/US/docs/security/asa/asa82/configuration/guide/nat_overview.html#wp1079279

	# 8.3 and after NAT order of operations.
	# http://www.cisco.com/en/US/docs/security/asa/asa83/upgrading/migrating.html#wp60328
	# http://www.cisco.com/en/US/docs/security/asa/asa83/configuration/guide/nat_overview.html#wp1118157

	my ($self) = @_;

	# Store the pre-routing and post-routing in scalars, for future retreaval.
	open(my $preRouteFh, '>', \$self->{_preRouting});
	open(my $postRouteFh, '>', \$self->{_postRouting});
	my $preRoutePrinter = ZipTie::Model::XmlPrint->new(\*$preRouteFh, 'core');
	my $postRoutePrinter = ZipTie::Model::XmlPrint->new(\*$postRouteFh, 'core');

	# Print NAT exemption rules.
	if (defined($self->{_addressTranslationObjects}->{exemption}))
	{
		foreach my $interface (keys(%{$self->{_addressTranslationObjects}->{exemption}}))
		{
			# Add the NAT exepmption rules to the start of both the pre-routing filterListsSection and the
			# post-routing filterLists section.
			my $addressTranslation = $self->{_addressTranslationObjects}->{exemption}->{$interface};

			$addressTranslation->setTranslationName("NAT_Exemption_$interface" . '_pre-routing');
			$addressTranslation->generateTranslationList($preRoutePrinter, 'source');

			$addressTranslation->setTranslationName("NAT_Exemption_$interface" . '_post-routing');
			$addressTranslation->generateTranslationList($postRoutePrinter, 'source');
		}
	}

	# Print the sections one at a time.
	_printNATSection($preRoutePrinter, $postRoutePrinter, $self->{_addressTranslationObjects}->{1});
	_printNATSection($preRoutePrinter, $postRoutePrinter, $self->{_addressTranslationObjects}->{2});
	_printNATSection($preRoutePrinter, $postRoutePrinter, $self->{_addressTranslationObjects}->{3});

	close($preRouteFh);
	close($postRouteFh);
}

# Return the pre-routing filter lists.  If it doesn't exist return an empty string.
sub getPreRoutingFilterLists
{
	my ($self) = @_;

	if (defined($self->{_preRouting}))
	{
		return $self->{_preRouting};
	}

	return '';
}

# Return the post-routing filter lists.  If it doesn't exist return an empty string.
sub getPostRoutingFilterLists
{
	my ($self) = @_;

	if (defined($self->{_postRouting}))
	{
		return $self->{_postRouting};
	}

	return '';
}

sub parse82Nat
{
	my ($self) = @_;

	# Order of operations for 82 ASA NAT:
	# http://www.cisco.com/en/US/docs/security/asa/asa82/configuration/guide/nat_overview.html#wp1079279

	_parse82StaticCommands($self);

	_parse82GlobalCommands($self);

	_parse82NatCommands($self);
}

# This method generates a post routing string of filterListReference elements and the corrisponding string
# of filterList elements for nat-control.  Returns them as an array ($filterListReferences, $filterLists).
# If nat-control isn't enabled then it returns undef.
sub returnNatControlFilterListAndReference
{
	my ($self) = @_;

	if (!$self->{_is_under_nat_control})
	{
		return;
	}

	my $filterListReferences = '';
	my $filterLists = '';
	open(my $filterListReferencesFh, '>', \$filterListReferences);
	open(my $filterListsFh, '>', \$filterLists);
	my $filterListReferencesPrinter = ZipTie::Model::XmlPrint->new(\*$filterListReferencesFh, 'core');
	my $filterListsPrinter = ZipTie::Model::XmlPrint->new(\*$filterListsFh, 'core');

	foreach my $sourceInterface (keys(%{$self->{_security_levels}}))
	{
		my $sourceSecurityLevel = $self->{_security_levels}->{$sourceInterface};

		foreach my $destInterface (keys(%{$self->{_security_levels}}))
		{
			my $destSecurityLevel = $self->{_security_levels}->{$destInterface};

			if ($sourceInterface eq $destInterface)
			{
				next;
			}

			# Usually NAT control only applies to traffic traveling from higher security levels to lower
			# security levels but can also apply in any direction if dynamic NAT is applied.
			if ($sourceSecurityLevel > $destSecurityLevel ||
					defined($self->{_nat_control_dynamic_nat_interfaces}->{$sourceInterface}->{all}) ||
					(defined($self->{_nat_control_dynamic_nat_interfaces}->{$sourceInterface}->{same}) &&
							($sourceSecurityLevel == $destSecurityLevel)))
			{
				my $key = $sourceInterface . '_' . $destInterface;
				my $name = "NAT_control_$sourceInterface" . "_to_$destInterface";
				my $processOrder = 1;

				$filterListReferencesPrinter->open_element('filterListReference');
				$filterListReferencesPrinter->open_element('appliesTo');
				$filterListReferencesPrinter->print_element('source', $sourceInterface);
				$filterListReferencesPrinter->print_element('destination', $destInterface);
				$filterListReferencesPrinter->close_element('appliesTo');
				$filterListReferencesPrinter->print_element('direction', 'egress');
				$filterListReferencesPrinter->print_element('name', $name);
				$filterListReferencesPrinter->close_element('filterListReference');

				$filterListsPrinter->open_element('filterList');

				if (defined($self->{natControlMappedAddrs}->{$key}))
				{
					foreach my $filterEntryTokens (@{$self->{natControlMappedAddrs}->{$key}})
					{
						$filterEntryTokens->{processOrder} = $processOrder++;
						$filterEntryTokens->{primaryAction} = 'none';
						$filterEntryTokens->{log} = 'false';

						$filterListsPrinter->print_element('filterEntry', $filterEntryTokens);
					}					
				}
				else
				{
					$filterListsPrinter->open_element('filterEntry');
					$filterListsPrinter->print_element('log', 'false');
					$filterListsPrinter->print_element('primaryAction', 'deny');
					$filterListsPrinter->print_element('processOrder', '1');
					$filterListsPrinter->close_element('filterEntry');
				}

				$filterListsPrinter->print_element('mode', 'stateful');
				$filterListsPrinter->print_element('name', $name);
				$filterListsPrinter->open_element('processingPhase');
				$filterListsPrinter->print_element('core:packetProcessingPhase', 'packet-filtering');
				$filterListsPrinter->close_element('processingPhase');
				$filterListsPrinter->close_element('filterList');
			}
		}
	}

	close($filterListReferencesFh);
	close($filterListsFh);

	return ($filterListReferences, $filterLists);
}

sub parseTwiceNat
{
	my ($self) = @_;
	my @nat_array = $self->{_in}->{running_config} =~ m/(^nat\s.*)/mig;
	
	# Twice NAT exists in the global portion of the config, with no whitespace between the
	# beginning of the line and the rest of the line
	foreach my $line (@nat_array) 
	{
		# Filter out 8.2 style NAT
		if ($line =~ /^nat\s+\([^)]+\)\s+\d+/)
		{
			next;
		}

		if ($line =~ /^nat(?<ifc>\s+\((?<realIfc>[^,]+),(?<mappedIfc>[^)]+)\))?
				(?<afterAuto>\s+after-(object|auto))?
				\s+source\s+(?<staticDynamic>static|dynamic)\s+(?<srcRealObj>\S+)\s+
				((?<patPool>(?<srcMappedObj>\S+)?\s*pat-pool\s+(?<patSrcMappedObj>\S+))|(?<srcMappedObj>(?!interface)\S+))?
				(?<roundRobin>\s*round-robin\S+)?
				(?<extended>\s*extended\S+)?
				(?<flat>\s*flat\s*(?<includeReserve>include-reserve)?)?
				(?<interface>\s*interface)?
				(?<destination>\s*destination\s+static\s+(?<destMappedObj>\S+)\s+(?<destRealObj>\S+))?
				(?<service>\s*service\s+(?<realSrcMappedDestService>\S+)\s+(?<mappedSrcRealDestService>\S+))?
				(?<dns>\s*dns)?
				(?<unidirectional>\s*unidirectional)?
				(?<no_proxy_arp>\s*no-proxy-arp)?
				(?<route_lookup>\s*route-lookup)?
				(?<inactive>\s*inactive)?
				(\s*description\s(?<description>.+))?/x)
		{

			# Skip NAT rule if rule is inactive
			if (defined($+{inactive}))
			{
				next;
			}

			# By default real/mapped ifc is any
			my $realIfc = 'any';
			my $mappedIfc = 'any';
			if (defined($+{realIfc}))
			{
				$realIfc = $+{realIfc};
			}
			if (defined($+{mappedIfc}))
			{
				$mappedIfc = $+{mappedIfc};
			}

			# Section of the NAT table
			my $natSection = 1;
			if (defined($+{afterAuto}))
			{
				$natSection = 3;
			}

			# Process service first, so we can obtain protocol if it exists
			my $realSrcMappedDestService = undef;
			my $mappedSrcRealDestService = undef;
			my $protocol = undef;
			if (defined($+{service}))
			{
				# We expect both of the following to be defined, if not we can't parse the NAT rule.
				if (defined($self->{_in}->{splitServiceGroups}->{$+{realSrcMappedDestService}}) &&
					defined($self->{_in}->{splitServiceGroups}->{$+{mappedSrcRealDestService}}))
				{
					$realSrcMappedDestService = $self->{_in}->{splitServiceGroups}->{$+{realSrcMappedDestService}};
					$mappedSrcRealDestService = $self->{_in}->{splitServiceGroups}->{$+{mappedSrcRealDestService}};
					$protocol = _protocol_from_split_service_group($realSrcMappedDestService, $self->{_in}->{object_groups});
				}
				else
				{
					$self->{_logger}->debug("Failed to parse twice NAT line: $line");
					next;
				}
			}

			# We will gather all the correct arguments first
			my $sourceAddRuleArguments = {};
			my $sourceCriterionArguments = {};
			if (defined($protocol))
			{
				$sourceAddRuleArguments->{protocol} = $protocol;
			}
			else
			{
				$sourceAddRuleArguments->{protocol} = 'ip';
			}

			if ($+{srcRealObj} eq 'any')
			{
				$sourceCriterionArguments->{sourceIp} = 'any';
			}
			else
			{
				$sourceCriterionArguments->{sourceIpObjectGroup} = $+{srcRealObj};
			}

			# Set the real destination ip if it exists.
			if (defined($+{destMappedObj}))
			{
				if ($+{destMappedObj} eq 'any')
				{
					$sourceCriterionArguments->{destIp} = 'any';
				}
				elsif ($+{destMappedObj} eq 'interface')
				{
					if (defined($self->{_interface_ips}->{$realIfc}->[0]))
					{
						$sourceCriterionArguments->{destIp} = $self->{_interface_ips}->{$realIfc}->[0];
					}
					else
					{
						$self->{_logger}->debug("Failed to create destination mapping in twice NAT line: \"$line\", " .
								"IP for interface $realIfc doesn't exist.");
						next;
					}
				}
				else
				{
					$sourceCriterionArguments->{destIpObjectGroup} = $+{destMappedObj};
				}
			}

			# Set the real destination port if it exists.
			if (defined($mappedSrcRealDestService->{destination}))
			{
				$sourceCriterionArguments->{destServiceObjectName} = $mappedSrcRealDestService->{destination};
			}

			my @addressMappings;

			if ($+{staticDynamic} eq 'static')
			{
				# Define as static mapping.  The static mapping flag has control over the established
				# flag for the return rule, if it's static then established will not be set to true.
				# For unidirectional static NAT we want the established flag to be set to true on the return
				# rules.
				if (!defined($+{unidirectional}))
				{
					$sourceAddRuleArguments->{isStaticNATMapping} = 1;
				}

				if (defined($+{srcMappedObj}) && $+{srcMappedObj} eq 'any')
				{
					my $addressMapping = {};
					$addressMapping->{sourceAddrMapped} = 'any';
					$addressMapping->{useAppliedAddressRules} = 'true';
					push(@addressMappings, $addressMapping);
				}
				elsif (defined($+{interface}))
				{
					my $interfaceMappings = _createInterfaceMappings($self, $mappedIfc, 1, 0);

					if (!defined($interfaceMappings))
					{
						$self->{_logger}->debug("Failed to parse twice NAT line: $line, " .
								"no IP defined for interface $mappedIfc");
						next;
					}

					push(@addressMappings, @{$interfaceMappings});
				}
				else
				{
					if (!defined($+{srcMappedObj})){
						$self->{_logger}->debug("Failed to parse twice NAT line: $line. No srcMappedObj defined.");
						next;
					}
					my $addressMapping = {};
					$addressMapping->{sourceAddrNameMapped} = $+{srcMappedObj};
					$addressMapping->{useAppliedAddressRules} = 'true';
					push(@addressMappings, $addressMapping);
				}

				# Define source ports if applicable
				if (defined($realSrcMappedDestService->{source}) &&
						defined($mappedSrcRealDestService->{source}))
				{
					$sourceCriterionArguments->{sourceServiceObjectName} = $realSrcMappedDestService->{source};

					foreach my $addressMapping (@addressMappings)
					{
						$addressMapping->{sourcePortNameMapped} = $mappedSrcRealDestService->{source};
					}
				}
			}
			else
			{
				if (defined($+{patPool}))
				{
					if (defined($+{srcMappedObj}))
					{
						my $srcMappedObjAddressMappings =
								_createDynamicSrcMappedObjAddressMappings($self, $+{srcMappedObj});
						push(@addressMappings, @{$srcMappedObjAddressMappings});
					}
					my $patPoolAddressMapping = {};
					$patPoolAddressMapping->{sourceAddrNameMapped} = $+{patSrcMappedObj};
					$patPoolAddressMapping->{usePATOverload} = 1;
					push(@addressMappings, $patPoolAddressMapping);
				}
				elsif (defined($+{interface}))
				{
					my $interfaceMappings = _createInterfaceMappings($self, $mappedIfc, 0, 1);

					if (!defined($interfaceMappings))
					{
						$self->{_logger}->debug("Failed to parse twice NAT line: $line, " .
								"no IP defined for interface $mappedIfc");
						next;
					}
					push(@addressMappings, @{$interfaceMappings});
				}
				else
				{
					if (!defined($+{srcMappedObj})){
						$self->{_logger}->debug("Failed to parse twice NAT line: $line. No srcMappedObj defined.");
						next;
					}
					my $srcMappedObjAddressMappings =
							_createDynamicSrcMappedObjAddressMappings($self, $+{srcMappedObj});
					push(@addressMappings, @{$srcMappedObjAddressMappings});
				}
			}

			# Create the rule with the parsed arguments.
			my $sourceAddressTranslation = _getAddressTranslationObject($self, $realIfc, $mappedIfc, $natSection);
			_createFlatIncludeReserveTranslationRules($sourceAddressTranslation, $sourceAddRuleArguments,
					$sourceCriterionArguments, \@addressMappings, $+{includeReserve}, $+{flat});

			# If a destination transaltion exists create it
			my $destRule = undef;
			my $destAddressTranslation = undef;
			if (defined($+{destination}))
			{
				# We will gather all the correct arguments first
				my $destAddRuleArguments = {};
				my $destCriterionArguments = {};
				my @addressMappings;

				# Destination NAT is always static.
				$destAddRuleArguments->{isStaticNATMapping} = 1;

				if (defined($protocol))
				{
					$destAddRuleArguments->{protocol} = $protocol;
				}
				else
				{
					$destAddRuleArguments->{protocol} = 'ip';
				}

				# Set the real source ip if it exists.
				if ($+{srcRealObj} eq 'any')
				{
					$destCriterionArguments->{sourceIp} = 'any';
				}
				else
				{
					$destCriterionArguments->{sourceIpObjectGroup} = $+{srcRealObj};
				}

				# Set the real destination port if it exists.
				if (defined($realSrcMappedDestService->{source}))
				{
					$destCriterionArguments->{sourceServiceObjectName} = $realSrcMappedDestService->{source};
				}

				# More details on the interface handling for the dynamic addresses
				# http://www.cisco.com/en/US/docs/security/asa/asa83/configuration/guide/nat_rules.html#wp1264263
				if ($+{destMappedObj} eq 'interface')
				{
					if (defined($self->{_interface_ips}->{$realIfc}->[0]))
					{
						$destCriterionArguments->{destIp} = $self->{_interface_ips}->{$realIfc}->[0];
					}
					else
					{
						$self->{_logger}->debug("Failed to create destination mapping in twice NAT line: \"$line\", " .
								"IP for interface $realIfc doesn't exist.");
						next;
					}
				}
				else
				{
					$destCriterionArguments->{destIpObjectGroup} = $+{destMappedObj};
				}

				if ($+{destRealObj} eq 'any')
				{
					my $addressMapping = {};
					$addressMapping->{destAddrMapped} = 'any';
					$addressMapping->{useAppliedAddressRules} = 'true';
					push(@addressMappings, $addressMapping);
				}
				else
				{
					my $addressMapping = {};
					$addressMapping->{destAddrNameMapped} = $+{destRealObj};
					$addressMapping->{useAppliedAddressRules} = 'true';
					push(@addressMappings, $addressMapping);
				}

				if (defined($mappedSrcRealDestService->{destination}) &&
						defined($realSrcMappedDestService->{destination}))
				{
					$destCriterionArguments->{destServiceObjectName} = $realSrcMappedDestService->{destination};

					foreach my $addressMapping (@addressMappings)
					{
						$addressMapping->{destPortNameMapped} = $mappedSrcRealDestService->{destination};
					}
				}

				my $destAddressTranslation = _getAddressTranslationObject($self, $realIfc, $mappedIfc, $natSection, 1);
				_createAddressTranslationRule($destAddressTranslation, $destAddRuleArguments,
						$destCriterionArguments, \@addressMappings);
			}
		}
		else
		{
			$self->{_logger}->debug("Failed to parse twice NAT line: $line");
		}
	}
}

sub parseObjectNat
{
	my ($self) = @_;

	my $ipRe = get_crep('cipm');

	my @staticNatRules;
	my @dynamicNatRules;

	my $inObject = 0;
	my $currentObject;
	foreach my $line (split(/\n/, $self->{_in}->{running_config}))
	{
		unless ($line =~ /^\s/)
		{
			$inObject = 0;
		}

		if ($inObject)
		{
			# Parse static NAT
			if ($line =~ /^\s+nat(?<ifc>\s+\((?<realIfc>[^,]+),(?<mappedIfc>[^)]+)\))?\s+static\s+
					(?<mappedObj>\S+)
					(?<dns>\s*dns)?
					(?<service>\s*service\s+(?<tcpUdp>(tcp|udp))\s+(?<realPort>\S+)\s+(?<mappedPort>\S+))?
					(?<no_proxy_arp>\s+no-proxy-arp)?
					(?<route_lookup>\s+route-lookup)?/x)
			{
				my %tokens = %+;
				$tokens{mappedObj} = $self->{_in}->{name_mapper}->get_ip($tokens{mappedObj});
				push(@staticNatRules, { tokens => \%tokens, objectName => $currentObject,
						objects => $self->{_network_groups}, line => $line } );
			}
			# Parse dynamic NAT
			elsif ($line =~ /^\s+nat(?<ifc>\s+\((?<realIfc>[^,]+),(?<mappedIfc>[^)]+)\))?\s+dynamic\s+
					((?<patPool>(?<srcMappedObj>\S+)?\s*pat-pool\s+(?<patSrcMappedObj>\S+))|
					(?<inlineHost>(?<inlineHostIp>\S+)(?<inlineHostInterface>\s*interface)?))
					(?<roundRobin>\s*round-robin\S+)?
					(?<extended>\s*extended\S+)?
					(?<flat>\s*flat\s*(?<includeReserve>include-reserve)?)?
					(?<dns>\s*dns)?/x)
			{
				my %tokens = %+;

				if (defined($tokens{inlineHostIp}))
				{
					# Handle named IP addresses
					$tokens{inlineHostIp} = $self->{_in}->{name_mapper}->get_ip($tokens{inlineHostIp});

					unless($tokens{inlineHostIp} =~ /^$ipRe$/)
					{
						# inlineHost isn't really an inline host because it's not an IP address,
						# it's actually a srcMappedObj.
						$tokens{srcMappedObj} = $tokens{inlineHost};
						$tokens{inlineHost} = undef;
						$tokens{inlineHostIp} = undef;
					}
				}

				push(@dynamicNatRules, { tokens => \%tokens, objectName => $currentObject,
						objects => $self->{_network_groups}, line => $line } );
			}
			elsif ($line =~ /^\s+nat/)
			{
				$self->{_logger}->debug("For object $currentObject failed to parse NAT line: $line");
				next;
			}
		}
		else
		{
			if ($line =~ /^object\s+network\s+(\S+)/)
			{
				$inObject = 1;
				$currentObject = $1;
			}
		}
	}

	# Sort the NAT lines based on the order of operations.
	my @staticNatSorted = sort _compareObjectNatObjectGroup @staticNatRules;
	my @dynamicNatSorted = sort _compareObjectNatObjectGroup @dynamicNatRules;

	foreach my $staticNat (@staticNatSorted)
	{
		# By default real/mapped ifc is any
		my $realIfc = 'any';
		my $mappedIfc = 'any';
		if (defined($staticNat->{tokens}->{realIfc}))
		{
			$realIfc = $staticNat->{tokens}->{realIfc};
		}
		if (defined($staticNat->{tokens}->{mappedIfc}))
		{
			$mappedIfc = $staticNat->{tokens}->{mappedIfc};
		}

		# Determine the protocol
		my $protocol = 'ip';
		my $mappedPort = undef;
		my $realPort = undef;
		if (defined($staticNat->{tokens}->{service}))
		{
			$protocol = $staticNat->{tokens}->{tcpUdp};
			$mappedPort = _int_port($staticNat->{tokens}->{mappedPort});
			$realPort = _int_port($staticNat->{tokens}->{realPort});
		}

		# We will gather all the correct arguments first
		my $addRuleArguments = {};
		my $criterionArguments = {};
		my @addressMappings;

		$addRuleArguments->{isStaticNATMapping} = 1;

		$addRuleArguments->{protocol} = $protocol;

		$criterionArguments->{sourceIpObjectGroup} = $staticNat->{objectName};

		if ($staticNat->{tokens}->{mappedObj} eq 'interface')
		{
			my $interfaceMappings = _createInterfaceMappings($self, $mappedIfc, 1, 0);

			if (!defined($interfaceMappings))
			{
				# Failed to parse NAT rule.
				$self->{_logger}->debug('For object ' . $staticNat->{objectName} . ' failed to parse NAT line: ' .
						$staticNat->{line} . ", no IP defined for interface $mappedIfc");
				next;
			}

			push(@addressMappings, @{$interfaceMappings});
		}
		elsif ($staticNat->{tokens}->{mappedObj} =~ /^$ipRe$/)
		{
			# mappedObj is an inline IP address

			my $addressMapping = {};

			# We need to determine if the real object is a host/range/subnet
			# http://www.cisco.com/en/US/docs/security/asa/asa84/command/reference/no.html#wp1832363
			my $realObject = $self->{_network_groups}->{$staticNat->{objectName}};
			if (defined($realObject->{host}))
			{
				# If it's a host, simply add the host.
				$addressMapping->{useAppliedAddressRules} = 'true';
				$addressMapping->{sourceAddrMapped} = $staticNat->{tokens}->{mappedObj};
			}
			elsif (defined($realObject->{range}))
			{
				# If the real object is a range, then create a new range starting with the mapped object, with
				# the same amount of addresses as the real object.
				my $startAddress = $staticNat->{tokens}->{mappedObj};

				# Determine how many addresses are in the object group, and then determine the end address by adding
				# that number to the start address.
				my $numberAddressesInRange = _countAddressesInNetworkGroup($staticNat->{objectName},
						$self->{_network_groups});
				my $endAddress = Net::IP->new("$startAddress + $numberAddressesInRange")->last_ip();

				$addressMapping->{useAppliedAddressRules} = 'true';
				$addressMapping->{sourceAddrMapped} = $startAddress;
				$addressMapping->{sourceAddrEndMapped} = $endAddress;
			}
			elsif (defined($realObject->{subnet}))
			{
				# If the real object is a subnet, then create a new subnet using the mappedObject setting the mask
				# the same as the real object.  Since this is a network object, not an object group, there should
				# only be one subnet, so use the index of 0.
				$addressMapping->{useAppliedAddressRules} = 'true';
				$addressMapping->{sourceAddrMapped} = $staticNat->{tokens}->{mappedObj};
				$addressMapping->{sourceAddrMaskMapped} = $realObject->{subnet}->[0]->{mask};
			}
			else
			{
				# Failed to parse NAT rule.
				$self->{_logger}->debug('For object ' . $staticNat->{objectName} . ' failed to parse NAT line: ' .
						$staticNat->{line});
				next;
			}

			push(@addressMappings, $addressMapping);
		}
		else
		{
			my $addressMapping = {};

			# This is an object group
			$addressMapping->{useAppliedAddressRules} = 'true';
			$addressMapping->{sourceAddrNameMapped} = $staticNat->{tokens}->{mappedObj};

			push(@addressMappings, $addressMapping);
		}

		if (defined($realPort))
		{
			$criterionArguments->{sourceServicePort} = $realPort;

			foreach my $addressMapping (@addressMappings)
			{
				$addressMapping->{sourcePortMapped} = $mappedPort;
			}
		}

		# Create the translation object
		my $addressTranslation = _getAddressTranslationObject($self, $realIfc, $mappedIfc, 2);
		_createAddressTranslationRule($addressTranslation, $addRuleArguments, $criterionArguments,
				\@addressMappings);
	}

	foreach my $dynamicNat (@dynamicNatSorted)
	{
		# By default real/mapped ifc is any
		my $realIfc = 'any';
		my $mappedIfc = 'any';
		if (defined($dynamicNat->{tokens}->{realIfc}))
		{
			$realIfc = $dynamicNat->{tokens}->{realIfc};
		}
		if (defined($dynamicNat->{tokens}->{mappedIfc}))
		{
			$mappedIfc = $dynamicNat->{tokens}->{mappedIfc};
		}

		# We will gather all the correct arguments first
		my $addRuleArguments = {};
		my $criterionArguments = {};
		my @addressMappings;

		$addRuleArguments->{protocol} = 'ip';

		# Set the real object
		$criterionArguments->{sourceIpObjectGroup} = $dynamicNat->{objectName};

		if (defined($dynamicNat->{tokens}->{inlineHost}))
		{
			# Inline hosts use dynamic PAT
			my $addressMapping = {};
			$addressMapping->{sourceAddrMapped} = $dynamicNat->{tokens}->{inlineHostIp};
			$addressMapping->{usePATOverload} = 1;
			push(@addressMappings, $addressMapping);

			# Handle interface keyword
			# http://www.cisco.com/en/US/docs/security/asa/asa84/command/reference/no.html#wp1832373
			if (defined($dynamicNat->{tokens}->{inlineHostInterface}))
			{
				my $interfaceMappings = _createInterfaceMappings($self, $mappedIfc, 0, 1);

				if (!defined($interfaceMappings))
				{
					$self->{_logger}->debug('For object ' . $dynamicNat->{objectName} .
							' failed to parse NAT line: ' . $dynamicNat->{line} .
							", no IP defined for interface $mappedIfc");
					next;
				}

				push(@addressMappings, @{$interfaceMappings});
			}
		}
		else
		{
			if (defined($dynamicNat->{tokens}->{patPool}))
			{
				if (defined($dynamicNat->{tokens}->{srcMappedObj}))
				{
					my $srcMappedObjAddressMappings =
							_createDynamicSrcMappedObjAddressMappings($self, $dynamicNat->{tokens}->{srcMappedObj});
					push(@addressMappings, @{$srcMappedObjAddressMappings});
				}

				my $patPoolAddressMapping = {};
				$patPoolAddressMapping->{sourceAddrNameMapped} = $dynamicNat->{tokens}->{patSrcMappedObj};
				$patPoolAddressMapping->{usePATOverload} = 1;
				push(@addressMappings, $patPoolAddressMapping);
			}
			else
			{
				if ($dynamicNat->{tokens}->{srcMappedObj} eq 'interface')
				{
					# Handle interface keyword
					# http://www.cisco.com/en/US/docs/security/asa/asa84/command/reference/no.html#wp1832373
					my $interfaceMappings = _createInterfaceMappings($self, $mappedIfc, 0, 1);

					if (!defined($interfaceMappings))
					{
						$self->{_logger}->debug('For object ' . $dynamicNat->{objectName} .
								' failed to parse NAT line: ' . $dynamicNat->{line} .
								", no interface defined for interface $mappedIfc");
						next;
					}

					push(@addressMappings, @{$interfaceMappings});
				}
				else
				{
					my $srcMappedObjAddressMappings =
							_createDynamicSrcMappedObjAddressMappings($self, $dynamicNat->{tokens}->{srcMappedObj});
					push(@addressMappings, @{$srcMappedObjAddressMappings});
				}
			}
		}

		# Create the translation object
		my $addressTranslation = _getAddressTranslationObject($self, $realIfc, $mappedIfc, 2);
		_createFlatIncludeReserveTranslationRules($addressTranslation, $addRuleArguments,
				$criterionArguments, \@addressMappings, $dynamicNat->{tokens}->{includeReserve},
				$dynamicNat->{tokens}->{flat});
	}
}

# This subroutine is used to create the translation rules for dynamic object and twice NAT.  It doesn't do anything
# special unless either the $includeReserve or $flat arguments are defined.  If $includeReserve is defined then
# it manually generates a PAT rule on the port range 1 - 65535, if only flat is defined then it manually generates
# two PAT rule on the port ranges 1 - 1023 and 1024 - 65535.
#
# flat and include-reserve keywords for object nat:
# http://www.cisco.com/en/US/docs/security/asa/asa84/configuration/guide/nat_objects.html#wp1542090
#
# flat and include-reserve keywords for twice nat:
# http://www.cisco.com/en/US/docs/security/asa/asa84/configuration/guide/nat_rules.html#wp1509792
sub _createFlatIncludeReserveTranslationRules
{
	my ($addressTranslation, $addRuleArguments, $criterionArguments, $addressMappings, $includeReserve, $flat) = @_;

	# If flat isn't defined then allow the addressTranslation object handle any PAT rules.
	if (!defined($flat))
	{
		_createAddressTranslationRule($addressTranslation, $addRuleArguments,
				$criterionArguments, $addressMappings);
	}
	# If include reserve is defined then we only need to generate one large PAT rule.
	elsif (defined($includeReserve))
	{
		# We only require one rule for flat include-reserve
		$criterionArguments->{sourceServicePort} = '1';
		$criterionArguments->{sourceServicePortEnd} = '65535';

		foreach my $addressMapping (@{$addressMappings})
		{
			if (defined($addressMapping->{usePATOverload}))
			{
				$addressMapping->{usePATOverload} = undef;
				$addressMapping->{sourcePortMapped} = '1';
				$addressMapping->{sourcePortEndMapped} = '65535';
			}
		}

		_createAddressTranslationRule($addressTranslation, $addRuleArguments,
				$criterionArguments, $addressMappings);
	}
	# Only the flat keyword is defined, generate two rules.
	else
	{
		# Create an array of all the PAT mappings.
		# We need to do this because usePATOverload will be set to undef after creating the first rule,
		# and we wouldn't be able to test for it on the second rule.
		my @patMappings;
		foreach my $addressMapping (@{$addressMappings})
		{
			if (defined($addressMapping->{usePATOverload}))
			{
				push(@patMappings, $addressMapping);
			}
		}

		# We will require two rules
		$criterionArguments->{sourceServicePort} = '1';
		$criterionArguments->{sourceServicePortEnd} = '1023';

		foreach my $addressMapping (@patMappings)
		{
			$addressMapping->{usePATOverload} = undef;
			$addressMapping->{sourcePortMapped} = '1';
			$addressMapping->{sourcePortEndMapped} = '1023';
		}

		_createAddressTranslationRule($addressTranslation, $addRuleArguments,
				$criterionArguments, $addressMappings);

		$criterionArguments->{sourceServicePort} = '1024';
		$criterionArguments->{sourceServicePortEnd} = '65535';

		foreach my $addressMapping (@patMappings)
		{
			$addressMapping->{usePATOverload} = undef;
			$addressMapping->{sourcePortMapped} = '1024';
			$addressMapping->{sourcePortEndMapped} = '65535';
		}

		_createAddressTranslationRule($addressTranslation, $addRuleArguments,
				$criterionArguments, $addressMappings);
	}
}

# http://www.cisco.com/en/US/docs/security/asa/asa82/command/reference/s8.html#wp1512466
sub _parse82StaticCommands
{
	my ($self) = @_;

	my $ipRe = get_crep('cipm');

	while ($self->{_in}->{running_config} =~ /^(static\s+.*$)/mg)
	{
		my $line = $1;

		my %tokens;

		# Set to 1 if the static NAT rule has failed to parse.
		my $failedToParse = 0;

		# Parse the static command:
		# http://www.cisco.com/en/US/docs/security/asa/asa82/command/reference/s8.html#wp1512466
		if ($line =~ /^static\s+\((?<realIfc>[^,]+),(?<mappedIfc>[^)]+)\)\s+
				(?<tcpUdp>tcp|udp)?\s*(?<mappedIp>\S+)\s+(?<realIp>\S+)\s*(?<remainder>.+)?$/x)
		{
			%tokens = %+;

			$tokens{mappedIp} = $self->{_in}->{name_mapper}->get_ip($tokens{mappedIp});

			# Static PAT
			if (defined($tokens{tcpUdp}))
			{
				$tokens{mappedPort} = $tokens{realIp};
				$tokens{realIp} = undef;

				if (defined($tokens{remainder}) && ($tokens{remainder} =~ /^(\S+)\s+(.+)$/))
				{
					if ($1 eq 'access-list')
					{
						unless ($2 =~ /^(\S+)/)
						{
							$failedToParse = 1;
						}

						$tokens{accessList} = $1;
					}
					else
					{
						$tokens{realIp} = $self->{_in}->{name_mapper}->get_ip($1);

						unless ($2 =~ /^(\S+)\s*(?:netmask\s+(\S+))?/)
						{
							$failedToParse = 1;
						}

						$tokens{realPort} = $1;

						if (defined($2))
						{
							$tokens{mask} = $2;
						}
					}
				}
				else
				{
					$failedToParse = 1;
				}

			}
			# Static NAT (no PAT)
			else
			{
				if ($tokens{realIp} eq 'access-list')
				{
					$tokens{realIp} = undef;

					unless (defined($tokens{remainder}) && ($tokens{remainder} =~ /^(\S+)/))
					{
						$failedToParse = 1;
					}

					$tokens{accessList} = $1;
				}
				else
				{
					$tokens{realIp} = $self->{_in}->{name_mapper}->get_ip($tokens{realIp});

					if (defined($tokens{remainder}) && ($tokens{remainder} =~ /^netmask\s+($ipRe)/))
					{
						$tokens{mask} = $1;
					}
				}
			}
		}
		else
		{
			$failedToParse = 1;
		}

		if ($failedToParse)
		{
			$self->{_logger}->debug("Failed to parse ASA 82 static line: $line");
			next;
		}

		if (defined($tokens{accessList}))
		{
			my $protocol = undef;
			if (defined($tokens{tcpUdp}))
			{
				$protocol = $tokens{tcpUdp};
			}

			my $rules = _parsePolicyNat($self, $tokens{accessList}, $protocol, 0);

			if (!defined($rules))
			{
				$self->{_logger}->debug("Failed to parse ASA 82 static line: $line, no ACL " . $tokens{accessList} .
						' found.');
				next;
			}

			my @addressMappings;

			if ($tokens{mappedIp} eq 'interface')
			{
				my $interfaceMappings = _createInterfaceMappings($self, $tokens{mappedIfc}, 1, 0);

				if (!defined($interfaceMappings))
				{
					$self->{_logger}->debug("Failed to parse ASA 82 static line: $line, " .
							'no IP defined for interface ' . $tokens{mappedIfc});
					next;
				}

				if (defined($tokens{mappedPort}))
				{
					foreach my $addressMapping (@{$interfaceMappings})
					{
						$addressMapping->{sourcePortMapped} = _int_port($tokens{mappedPort});
					}
				}

				push(@addressMappings, @{$interfaceMappings});
			}
			else
			{
				my $addressMapping = {};
				$addressMapping->{useAppliedAddressRules} = 'true';
				$addressMapping->{sourceAddrMapped} = $tokens{mappedIp};

				if (defined($tokens{mappedPort}))
				{
					$addressMapping->{sourcePortMapped} = _int_port($tokens{mappedPort});
				}

				push(@addressMappings, $addressMapping)
			}

			my $addressTranslation = _getAddressTranslationObject($self, $tokens{realIfc}, $tokens{mappedIfc}, 1);
			
			foreach my $rule (@{$rules})
			{
				my ($addRuleArguments, $criterionArguments) = @{$rule};

				if (!defined($addRuleArguments))
				{
					$self->{_logger}->debug($criterionArguments . " for ASA 82 static line: $line");
					next;
				}

				$addRuleArguments->{isStaticNATMapping} = 1;

				_createAddressTranslationRule($addressTranslation, $addRuleArguments, $criterionArguments,
						\@addressMappings);

				_recordMappedAddressesForNatControl($self, $tokens{realIfc}, $tokens{mappedIfc}, \@addressMappings,
						$addRuleArguments->{protocol});

				# Because this is static we must record a mapping for the oppisite direction as well.
				_recordMappedAddressesForNatControl($self, $tokens{mappedIfc}, $tokens{realIfc},
						_reverseCriterionArgumentsForReverseNatControlMapping($self, $criterionArguments),
						$addRuleArguments->{protocol});
			}
		}
		else
		{
			# We will gather all the correct arguments first
			my $addRuleArguments = {};
			my $criterionArguments = {};
			my @addressMappings;

			if (defined($tokens{tcpUdp}))
			{
				$addRuleArguments->{protocol} = $tokens{tcpUdp};
			}
			else
			{
				$addRuleArguments->{protocol} = 'ip';
			}

			$addRuleArguments->{isStaticNATMapping} = 1;

			# Set the mask of the real and mapped addresses
			my $mask = undef;
			# http://www.cisco.com/en/US/docs/security/asa/asa82/command/reference/s8.html#wp1512538
			if (defined($tokens{mask}))
			{
				$mask = mask_to_bits($tokens{mask});
			}
			else
			{
				$mask = _getMaskFromAddressClass($tokens{realIp}, 1);
			}
			if ($mask == 32)
			{
				$mask = undef;
			}

			if ($tokens{mappedIp} eq 'interface')
			{
				my $interfaceMappings = _createInterfaceMappings($self, $tokens{mappedIfc}, 1, 0);

				if (!defined($interfaceMappings))
				{
					$self->{_logger}->debug("Failed to parse ASA 82 static line: $line, " .
							'no IP defined for interface ' . $tokens{mappedIfc});
					next;
				}

				if (defined($mask))
				{
					foreach my $addressMapping (@{$interfaceMappings})
					{
						$addressMapping->{sourceAddrMaskMapped} = $mask;
						$addressMapping->{useAppliedAddressRules} = 'true';
					}
				}

				if (defined($tokens{mappedPort}))
				{
					foreach my $addressMapping (@{$interfaceMappings})
					{
						$addressMapping->{sourcePortMapped} = _int_port($tokens{mappedPort});
					}
				}

				push(@addressMappings, @{$interfaceMappings});
			}
			else
			{
				my $addressMapping = {};
				$addressMapping->{sourceAddrMapped} = $tokens{mappedIp};

				if (defined($mask))
				{
					$addressMapping->{sourceAddrMaskMapped} = $mask;
					$addressMapping->{useAppliedAddressRules} = 'true';
				}

				if (defined($tokens{mappedPort}))
				{
					$addressMapping->{sourcePortMapped} = _int_port($tokens{mappedPort});
				}

				push(@addressMappings, $addressMapping);
			}

			$criterionArguments->{sourceIp} = $tokens{realIp};

			if (defined($mask))
			{
				$criterionArguments->{sourcePrefix} = $mask;
			}

			if (defined($tokens{realPort}))
			{
				$criterionArguments->{sourceServicePort} = _int_port($tokens{realPort});
			}

			my $addressTranslation = _getAddressTranslationObject($self, $tokens{realIfc}, $tokens{mappedIfc}, 1);
			_createAddressTranslationRule($addressTranslation, $addRuleArguments, $criterionArguments,
					\@addressMappings);

			_recordMappedAddressesForNatControl($self, $tokens{realIfc}, $tokens{mappedIfc}, \@addressMappings,
					$addRuleArguments->{protocol});

			# Because this is static we must record a mapping for the oppisite direction as well.
			_recordMappedAddressesForNatControl($self, $tokens{mappedIfc}, $tokens{realIfc},
					_reverseCriterionArgumentsForReverseNatControlMapping($self, $criterionArguments),
					$addRuleArguments->{protocol});
		}
	}
}

sub _parse82GlobalCommands
{
	my ($self) = @_;

	my $ipRe = get_crep('cipm');

	while ($self->{_in}->{running_config} =~ /^(global\s+.*$)/mg)
	{
		my $line = $1;

		my @addressMappings;

		if ($line =~ /^global\s+\((?<mappedIfc>[^)]+)\)\s+
				(?<id>\d+)\s+
				(?<ip>\S+)
				(\s*netmask\s+(?<mask>$ipRe))?/x)
		{
			my %globalTokens = %+;

			if ($globalTokens{ip} eq 'interface')
			{
				# Create interface mappings using PAT.
				my $interfaceMappings = _createInterfaceMappings($self, $globalTokens{mappedIfc}, 0, 1);

				if (!defined($interfaceMappings))
				{
					$self->{_logger}->debug("Failed to parse ASA 82 global line: $line, " .
								'no IP defined for interface ' . $globalTokens{mappedIfc});
					next;
				}

				push(@addressMappings, @{$interfaceMappings});
			}
			else
			{
				my $isRange = 0;

				# The following is possible with named IPs:
				#   name 172.16.55.75 dash-name
				#   name 172.16.55.77 dash-name-two
				#
				#   global (inside) 8 dash-name-dash-name-two
				#
				# To support this we will loop through every '-' character, checking the prematch and
				# postmatch for valid IP addresses.  We will take the first valid pair of IP addresses
				# as a range.
				while ($globalTokens{ip} =~ /-/gp)
				{
					my $startIp = $self->{_in}->{name_mapper}->get_ip(${^PREMATCH});
					my $endIp = $self->{_in}->{name_mapper}->get_ip(${^POSTMATCH});

					if ($startIp !~ /^$ipRe$/ || $endIp !~ /^$ipRe$/)
					{
						next;
					}

					$isRange = 1;

					my $originalStart = $startIp;
					$startIp = Net::IP->new($startIp)->intip();
					$endIp = Net::IP->new($endIp)->intip();

					# This is the only sane explination I could find on how to handle the netmak keyword:
					#
					# http://onlinestudylist.com/archives/ccie_security/2012-February/029050.html
					#
					# Quote:
					#     The 'netmask' option at the end of the global statement is only applicable
					#     for a global NAT pool, and not global PAT address.
					#     When the netmask is specified, the ASA determines what the network and
					#     broadcast IPs are, for that network, and if they are included in the global
					#     pool, it will not use them.
					if (defined($globalTokens{mask}))
					{
						my $mask = mask_to_bits($globalTokens{mask});

						my $network = _buildIpFromCidr($originalStart . '/' . $mask);
						my $networkIp = $network->intip();
						my $broadcastIp = $network->last_int();

						# If the first IP is the same as the network IP, increment by 1.
						if (!$startIp->bcmp($networkIp))
						{
							$startIp->binc();
						}

						# If the last IP is the same as the broadcast IP, decrement by 1.
						if (!$endIp->bcmp($broadcastIp))
						{
							$endIp->bdec();
						}
					}
 
 					# No PAT for ranges.
					my $addressMapping = {};
					$addressMapping->{sourceAddrMapped} = _bigintToIpString($startIp);
					$addressMapping->{sourceAddrEndMapped} = _bigintToIpString($endIp);
					push(@addressMappings, $addressMapping);

					last;
				}

				if (!$isRange)
				{
					my $ip = $self->{_in}->{name_mapper}->get_ip($globalTokens{ip});

					# A single IP uses PAT.
					my $addressMapping = {};
					$addressMapping->{sourceAddrMapped} = $ip;
					$addressMapping->{usePATOverload} = 1;
					push(@addressMappings, $addressMapping);
				}
			}

			if (!defined($self->{_global_address_pools}))
			{
				$self->{_global_address_pools} = {};
			}

			if (!defined($self->{_global_address_pools}->{$globalTokens{id}}))
			{
				$self->{_global_address_pools}->{$globalTokens{id}} = {};
			}

			if (!defined($self->{_global_address_pools}->{$globalTokens{id}}->{$globalTokens{mappedIfc}}))
			{
				$self->{_global_address_pools}->{$globalTokens{id}}->{$globalTokens{mappedIfc}} = [];
			}

			push(@{$self->{_global_address_pools}->{$globalTokens{id}}->{$globalTokens{mappedIfc}}},
					@addressMappings);
		}
		else
		{
			$self->{_logger}->debug("Failed to parse ASA 82 global line: $line");
		}
	}
}

sub _parse82NatCommands
{
	my ($self) = @_;

	my $ipRe = get_crep('cipm');

	my @regularNatTokens;

	while ($self->{_in}->{running_config} =~ /^(nat\s+\([^)]+\)\s+\d+.*$)/mg)
	{
		my $line = $1;

		# Parse regular NAT
		if ($line =~ /^nat\s+\((?<realIfc>[^)]+)\)\s+
				(?<id>\d+)\s+
				(?!access-list)(?<realIp>\S+)\s+
				(?<mask>$ipRe)?
				(?<dns>\s*dns)?
				(?<outside>\s*outside)?/x) # Note, there are actually more tokens that can be parsed, but
				                           # they will not have any affect on the resulting XML.
		{
			my %tokens = %+;
			$tokens{line} = $line;

			$tokens{realIp} = $self->{_in}->{name_mapper}->get_ip($tokens{realIp});

			# Determine the real IP based on the following criteria:
			# http://www.cisco.com/en/US/docs/security/asa/asa82/command/reference/no.html#wp1769622
			if ($+{realIp} eq '0.0.0.0' || $+{realIp} eq '0')
			{
				$tokens{realIp} = 'any'; 
			}
			else
			{
				# Determine the mask based on the following criteria:
				# http://www.cisco.com/en/US/docs/security/asa/asa82/command/reference/no.html#wp1769584
				if (defined($+{mask}))
				{
					$tokens{mask} = mask_to_bits($+{mask});
				}
				else
				{
					$tokens{mask} = _getMaskFromAddressClass($+{realIp});
				}
			}

			# Identify NAT
			if ($tokens{id} == 0)
			{
				my $addressTranslation = _getAddressTranslationObject($self, $+{realIfc}, 'any', 'exemption');

				my $addRuleArguments = {};
				my $criterionArguments = {};

				$addRuleArguments->{protocol} = 'ip';

				$criterionArguments->{sourceIp} = $tokens{realIp};

				if (defined($tokens{mask}))
				{
					$criterionArguments->{sourcePrefix} = $tokens{mask};
				}

				my $translationRule = $addressTranslation->addRule($addRuleArguments);

				$criterionArguments->{rule} = $translationRule;
				$addressTranslation->addRuleAddressCriterion($criterionArguments);

				_recordMappedAddressesForNatControlFromCriterion($self, $+{realIfc}, undef, $criterionArguments,
						$addRuleArguments->{protocol});
			}
			else
			{
				# Dynamic NAT isn't parsed based on the order they are in the config, so they have to be stored and
				# sorted.
				# http://www.cisco.com/en/US/docs/security/asa/asa82/configuration/guide/nat_overview.html#wp1079279
				push(@regularNatTokens, \%tokens);
			}
		}
		# Parse policy NAT
		elsif ($line =~ /^nat\s+\((?<realIfc>[^)]+)\)\s+
				(?<id>\d+)\s+
				access-list\s+
				(?<accessList>\S+)
				(?<dns>\s*dns)?
				(?<outside>\s*outside)?/x) # Note, there are actually more tokens that can be parsed, but
				                           # they will not have any affect on the resulting XML.
		{
			my $ignorePorts = 0;
			if ($+{id} == 0)
			{
				$ignorePorts = 1;
			}

			my $rules = _parsePolicyNat($self, $+{accessList}, undef, $ignorePorts);

			if (!defined($rules))
			{
				$self->{_logger}->debug("Failed to parse ASA 82 nat line: $line, no ACL " . $+{accessList} .
						' found.');
				next;
			}

			if ($+{id} == 0)
			{
				my $addressTranslation = _getAddressTranslationObject($self, $+{realIfc}, 'any', 'exemption');

				foreach my $rule (@{$rules})
				{
					my ($addRuleArguments, $criterionArguments) = @{$rule};

					if (!defined($addRuleArguments))
					{
						$self->{_logger}->debug($criterionArguments . " for ASA 82 nat line: $line");
						next;
					}

					my $translationRule = $addressTranslation->addRule($addRuleArguments);

					$criterionArguments->{rule} = $translationRule;
					$addressTranslation->addRuleAddressCriterion($criterionArguments);

					_recordMappedAddressesForNatControlFromCriterion($self, $+{realIfc}, undef, $criterionArguments,
							$addRuleArguments->{protocol});
				}
			}
			elsif (defined($self->{_global_address_pools}->{$+{id}}))
			{
				foreach my $rule (@{$rules})
				{
					my ($addRuleArguments, $criterionArguments) = @{$rule};

					if (!defined($addRuleArguments))
					{
						$self->{_logger}->debug($criterionArguments . " for ASA 82 nat line: $line");
						next;
					}

					foreach my $interface (keys(%{$self->{_global_address_pools}->{$+{id}}}))
					{
						# Skip current mapping if designated to do so by the outside keyword.
						if (_outsideKeywordSkipCheck($self, $+{outside}, $+{realIfc}, $interface))
						{
							next;
						}

						my $addressTranslation = _getAddressTranslationObject($self, $+{realIfc}, $interface, 2);

						my $addressMappings = $self->{_global_address_pools}->{$+{id}}->{$interface};

						_createAddressTranslationRule($addressTranslation, $addRuleArguments, $criterionArguments,
								$addressMappings);

						_recordMappedAddressesForNatControl($self, $+{realIfc}, $interface, $addressMappings,
								$addRuleArguments->{protocol}, $+{outside}, 1);
					}
				}
			}
			else
			{
				$self->{_logger}->debug("No global pool defined for for ASA 82 nat line: $line");
			}
		}
		else
		{
			$self->{_logger}->debug("Failed to parse ASA 82 nat line: $line");
		}
	}

	my @sortedRegularNatTokens = sort _compare82RegularNatTokens @regularNatTokens;

	foreach my $natTokens (@sortedRegularNatTokens)
	{
		# We will gather all the correct arguments first
		my $addRuleArguments = {};
		my $criterionArguments = {};

		$addRuleArguments->{protocol} = 'ip';

		$criterionArguments->{sourceIp} = $natTokens->{realIp};

		if (defined($natTokens->{mask}))
		{
			$criterionArguments->{sourcePrefix} = $natTokens->{mask};
		}

		if (defined($self->{_global_address_pools}->{$natTokens->{id}}))
		{
			foreach my $interface (keys(%{$self->{_global_address_pools}->{$natTokens->{id}}}))
			{
				# Skip current mapping if designated to do so by the outside keyword.
				if (_outsideKeywordSkipCheck($self, $natTokens->{outside}, $natTokens->{realIfc}, $interface))
				{
					next;
				}

				my $addressTranslation = _getAddressTranslationObject($self, $natTokens->{realIfc}, $interface, 3);

				my $addressMappings = $self->{_global_address_pools}->{$natTokens->{id}}->{$interface};

				_createAddressTranslationRule($addressTranslation, $addRuleArguments, $criterionArguments,
						$addressMappings);

				_recordMappedAddressesForNatControl($self, $natTokens->{realIfc}, $interface, $addressMappings,
						$addRuleArguments->{protocol}, $natTokens->{outside}, 1);
			}
		}
		else
		{
			$self->{_logger}->debug('No global pool defined for for ASA 82 nat line: ' . $natTokens->{line});
		}
	}
}

sub _parsePolicyNat
{
	my ($self, $accessList, $protocol, $ignorePorts) = @_;

	my $rules = [];

	if (defined($self->{_in}->{_acl_terms}->{$accessList}))
	{
		foreach my $term (@{$self->{_in}->{_acl_terms}->{$accessList}})
		{
			# The action should always be permit according to the following document:
			# http://www.cisco.com/en/US/docs/security/asa/asa82/command/reference/s8.html#wp1512478
			# If it's not permit ignore this term.
			if ($term->{primaryAction} ne 'permit')
			{
				next;
			}

			# We will gather all the correct arguments first
			my $addRuleArguments = {};
			my $criterionArguments = {};

			if (defined($protocol))
			{
				$addRuleArguments->{protocol} = $protocol;
			}
			elsif (defined($term->{protocol}))
			{
				$addRuleArguments->{protocol} = $term->{protocol};
			}
			# Note, if protocol is defined using a protocol group, we will use that protocol.  However, if
			# the protocol group has more than one protocol defined, we wil default to using ip, since the
			# AddressTranslation.pm doesn't support more than one protocol.
			elsif (defined($term->{protocolObjectGroupReference}) &&
					defined($self->{_in}->{protocolGroups}->{$term->{protocolObjectGroupReference}}->{protocol}) &&
					defined($self->{_in}->{protocolGroups}->{$term->{protocolObjectGroupReference}}->{protocol}) &&
					scalar(@{$self->{_in}->{protocolGroups}->{$term->{protocolObjectGroupReference}}->
					{protocol}}) == 1)
			{
				$addRuleArguments->{protocol} =
						$self->{_in}->{protocolGroups}->{$term->{protocolObjectGroupReference}}->{protocol}->[0];
			}
			else
			{
				$addRuleArguments->{protocol} = 'ip';
			}

			if (defined $term->{sourceIpAddr} && scalar(@{$term->{sourceIpAddr}}) > 1)
			{
				push(@{$rules}, [undef, 'More than 1 sourceIpAddr for processOrder ' . $term->{processOrder} .
						" in accessList $accessList"]);
				next;
			}

			if (defined $term->{sourceIpAddr} && scalar(@{$term->{sourceIpAddr}}) == 0)
			{
				$criterionArguments->{sourceIp} = '0.0.0.0';
				$criterionArguments->{sourcePrefix} = '0';
			}
			elsif (defined($term->{sourceIpAddr}->[0]->{host}))
			{
				$criterionArguments->{sourceIp} = $term->{sourceIpAddr}->[0]->{host};
			}
			elsif (defined($term->{sourceIpAddr}->[0]->{network}))
			{
				$criterionArguments->{sourceIp} = $term->{sourceIpAddr}->[0]->{network}->{address};
				$criterionArguments->{sourcePrefix} = $term->{sourceIpAddr}->[0]->{network}->{mask};
			}
			elsif (defined($term->{sourceIpAddr}->[0]->{range}))
			{
				$criterionArguments->{sourceIp} = $term->{sourceIpAddr}->[0]->{range}->{startAddress};
				$criterionArguments->{sourceIpEnd} = $term->{sourceIpAddr}->[0]->{range}->{endAddress};
			}
			elsif (defined($term->{sourceIpAddr}->[0]->{objectGroupReference}))
			{
				$criterionArguments->{sourceIpObjectGroup} = $term->{sourceIpAddr}->[0]->{objectGroupReference};
			}
			else
			{
				push(@{$rules}, [undef, 'Failed to parse sourceIpAddr for processOrder ' . $term->{processOrder} .
						" in accessList $accessList"]);
				next;
			}

			if (!$ignorePorts && defined($term->{sourceService}) && scalar(@{$term->{sourceService}}) == 1)
			{
			 	if (defined($term->{sourceService}->[0]->{portExpression}->{operatior}) &&
			 			$term->{sourceService}->[0]->{portExpression}->{operatior} eq 'eq')
			 	{
			 		$criterionArguments->{sourceServicePort} = $term->{sourceService}->[0]->{portExpression}->{port};
			 	}
			 	elsif (defined($term->{sourceService}->[0]->{objectGroupReference}))
			 	{
			 		$criterionArguments->{sourceServiceObjectName} =
			 				$term->{sourceService}->[0]->{objectGroupReference};
			 	}
			}

			if (defined $term->{destinationIpAddr} && scalar(@{$term->{destinationIpAddr}}) > 1)
			{
				push(@{$rules}, [undef, 'More than 1 destinationIpAddr for processOrder ' . $term->{processOrder} .
						" in accessList $accessList"]);
				next;
			}

			if (defined $term->{destinationIpAddr} && scalar(@{$term->{destinationIpAddr}}) == 0)
			{
				# There is no destinationIpAddr, don't set it.
			}
			elsif (defined($term->{destinationIpAddr}->[0]->{host}))
			{
				$criterionArguments->{destIp} = $term->{destinationIpAddr}->[0]->{host};
			}
			elsif (defined($term->{destinationIpAddr}->[0]->{network}))
			{
				$criterionArguments->{destIp} = $term->{destinationIpAddr}->[0]->{network}->{address};
				$criterionArguments->{destPrefix} = $term->{destinationIpAddr}->[0]->{network}->{mask};
			}
			elsif (defined($term->{destinationIpAddr}->[0]->{range}))
			{
				$criterionArguments->{destIp} = $term->{destinationIpAddr}->[0]->{range}->{startAddress};
				$criterionArguments->{destIpEnd} = $term->{destinationIpAddr}->[0]->{range}->{endAddress};
			}
			elsif (defined($term->{destinationIpAddr}->[0]->{objectGroupReference}))
			{
				$criterionArguments->{destIpObjectGroup} = $term->{destinationIpAddr}->[0]->{objectGroupReference};
			}
			else
			{
				push(@{$rules}, [undef, 'Failed to parse destinationIpAddr for processOrder ' . $term->{processOrder} .
						" in accessList $accessList"]);
				next;
			}

			if (!$ignorePorts && defined($term->{destinationService}) && scalar(@{$term->{destinationService}}) == 1)
			{
			 	if (defined($term->{destinationService}->[0]->{portExpression}->{operatior}) &&
			 			$term->{destinationService}->[0]->{portExpression}->{operatior} eq 'eq')
			 	{
			 		$criterionArguments->{destServicePort} =
			 				$term->{destinationService}->[0]->{portExpression}->{port};
			 	}
			 	elsif (defined($term->{destinationService}->[0]->{objectGroupReference}))
			 	{
			 		$criterionArguments->{destServiceObjectName} =
			 				$term->{destinationService}->[0]->{objectGroupReference};
			 	}
			}

			push(@{$rules}, [$addRuleArguments, $criterionArguments]);
		}
	}
	else
	{
		return;
	}

	return $rules;
}

# Compares two 82 regular dynamic NAT tokens based on the mask.  When used with built in sort it sorts from smallest\
# (narrowest) mask to the largest mask.
# http://www.cisco.com/en/US/docs/security/asa/asa82/configuration/guide/nat_overview.html#wp1079279
# This subroutine is called by sort.  $a and $b are provided as the objects to compare as global vairables.
sub _compare82RegularNatTokens
{
	my $aMask;
	if ($a->{realIp} eq 'any')
	{
		$aMask = 0;
	}
	else
	{
		$aMask = $a->{mask};
	}

	my $bMask;
	if ($b->{realIp} eq 'any')
	{
		$bMask = 0;
	}
	else
	{
		$bMask = $b->{mask};
	}

	return $aMask <=> $bMask;
}

# Compares two NAT object group and compares them based on the criteria in the following document:
# http://www.cisco.com/en/US/docs/security/asa/asa83/configuration/guide/nat_overview.html#wp1118157
# This subroutine is called by sort.  $a and $b are provided as the objects to compare as global vairables.
sub _compareObjectNatObjectGroup
{
	# Step 1, smallest amount of addresses goes first.
	my $amountAddressesInA = _countAddressesInNetworkGroup($a->{objectName}, $a->{objects});
	my $amountAddressesInB = _countAddressesInNetworkGroup($b->{objectName}, $b->{objects});
	if ($amountAddressesInA > $amountAddressesInB)
	{
		return 1;
	}
	elsif ($amountAddressesInA < $amountAddressesInB)
	{
		return -1;
	}

	# Step 2, compare the smallest IP's in each object group.  If they are different, than this is the result.
	my $smallestInA = _getSmallestAddressInNetworkGroup($a->{objectName}, $a->{objects});
	my $smallestInB = _getSmallestAddressInNetworkGroup($b->{objectName}, $b->{objects});
	my $ipCmp = $smallestInA->bcmp($smallestInB);
	if ($ipCmp)
	{
		return $ipCmp;
	}

	# Step 3, if they are the same, then compare the names.
	return $a->{objectName} cmp $b->{objectName};
}

# Takes in the outside keyword, realIfc and mappedIfc.  Returns true if the current mapping should be skipped
# based on the following criteria:
#
# Configuration Guide:
# http://www.cisco.com/en/US/docs/security/asa/asa82/configuration/guide/nat_dynamic.html#wp1077564
# Command Reference:
# http://www.cisco.com/en/US/docs/security/asa/asa82/command/reference/no.html#wp1769614
#
# Summary:
#   If outside keyword is provided then only perform NAT between lower security levels to higher security levels.
#   If outside keyword isn't provided then only perfrom NAT between higher security levels and lower security levels.
sub _outsideKeywordSkipCheck
{
	my ($self, $outsideKeyword, $realIfc, $mappedIfc) = @_;

	my $realIfcSecLevel = $self->{_security_levels}->{$realIfc};
	my $mappedIfcSecLevel = $self->{_security_levels}->{$mappedIfc};

	if (defined($outsideKeyword))
	{
		# outside keyword is defined.
		if ($realIfcSecLevel <= $mappedIfcSecLevel)
		{
			# Going from outside to inside, don't skip the mapping.
			return 0;
		}
		else
		{
			# Going from inside to outside, skip the mapping.
			return 1;
		}
	}
	else
	{
		# outside keywoard not defined.
		if ($realIfcSecLevel >= $mappedIfcSecLevel)
		{
			# Going from inside to outside, don't skip the mapping.
			return 0;
		}
		else
		{
			# Going from outside to inside, skip the mapping.
			return 1;
		}
	}
}

# On the ASA, earlier sections are evaluated first, so we print one seciton at at time, in order.
sub _printNATSection
{
	my ($preRoutePrinter, $postRoutePrinter, $section) = @_;

	if (!defined($section))
	{
		return;
	}

	# Print "source" filterLists to the post-route seciton, and "destination" filterLists to the pre-route section. 
	foreach my $key (keys(%{$section}))
	{
		# If the rule is defined from the standpoint of destination NAT then the forward direction must
		# come before routing and the destination after routing.  This ensures that destination NAT occurs
		# before routing.
		if ($key =~ /_destinationNat$/)
		{
			$section->{$key}->generateTranslationList($preRoutePrinter, 'source');
			$section->{$key}->generateTranslationList($postRoutePrinter, 'destination');
		}
		# This is how they rules will usually be printed out, when defined from the source NAT perspective.
		else
		{
			$section->{$key}->generateTranslationList($postRoutePrinter, 'source');
			$section->{$key}->generateTranslationList($preRoutePrinter, 'destination');
		}
	}
}

# AddressTranslation's setInterfaceZoneMapper takes in a function pointer to a function that returns a hash
# containing the zone mappings.  This function returns that function pointer.
sub _getInterfaceZoneMapper
{
	my ($self) = @_;
	
	return sub { return $self->{_zoneMapperHash}; };
}

sub _parseZoneMapperHash
{
	my ($running_config) = @_;
	
	my %zoneMapperHash;
	my @interfaces;
	
	while ($running_config =~ /^interface[^!]*nameif\s(\S+)/gm)
	{
		$zoneMapperHash{$1} = [ $1 ];
		push(@interfaces, $1);
	}
	
	$zoneMapperHash{'any'} = \@interfaces;
	
	return \%zoneMapperHash;
}

# We only want one translation object for each real -> mapped ifc.  If the translation object already exists
# it will return it, otherwise it will create a new translation object and return that.
# If $backwards is true, then destination NAT is performed in the forward direction (usually it's performed in the
# reverse direction.
sub _getAddressTranslationObject
{
	my ($self, $realIfc, $mappedIfc, $natSection, $backwards) = @_;

	my $key = $realIfc;
	if (defined($mappedIfc))
	{
		$key .= "_to_$mappedIfc";
	}
	else
	{
		$mappedIfc = 'any';
	}
	if ($backwards)
	{
		$key .= "_destinationNat";
	}

	if (!defined($self->{_addressTranslationObjects}))
	{
		$self->{_addressTranslationObjects} = {};
	}

	if (!defined($self->{_addressTranslationObjects}->{$natSection}))
	{
		$self->{_addressTranslationObjects}->{$natSection} = {};
	}

	if (!defined($self->{_addressTranslationObjects}->{$natSection}->{$key}))
	{
		my $addressTranslationObject = ZipTie::Model::AddressTranslation->new($realIfc, $mappedIfc);
		$addressTranslationObject->setInterfaceZoneMapper(_getInterfaceZoneMapper($self));
		$addressTranslationObject->setTranslationName("NAT_section_$natSection" . "_$key");

		# PAT range is always the same on the ASA
		$addressTranslationObject->addPortTranslationRange(0, 511);
		$addressTranslationObject->addPortTranslationRange(512, 1023);
		$addressTranslationObject->addPortTranslationRange(1024, 65535);

		$self->{_addressTranslationObjects}->{$natSection}->{$key} = $addressTranslationObject;
	}

	return $self->{_addressTranslationObjects}->{$natSection}->{$key};
}

# Count addresses in a network group
sub _countAddressesInNetworkGroup
{
	my ($id, $objects, $seen) = @_;

	my $count = 0;

	# Seen will be used to keep track of the object groups we have already counted.
	if (!defined($seen))
	{
		$seen = {};
	}

	# If we have already counted then return now, otherwise we may recurse forever.
	if (defined($seen->{$id}))
	{
		return 0;
	}

	# Mark this object group as seen.
	$seen->{$id} = 1;

	# If this isn't a network group, return 0.
	if (!defined($objects->{$id}))
	{
		return 0;
	}

	# If this object group has already been counted, return the value.
	if (defined($objects->{$id}->{count}))
	{
		return $objects->{$id}->{count};
	}

	if (defined($objects->{$id}->{subnet}))
	{
		foreach my $subnet (@{$objects->{$id}->{subnet}})
		{
			$count += _countAddressesInSubnet($subnet);
		}
	}

	# There is only one address in a host, so add the number of hosts to the count.
	if (defined($objects->{$id}->{host}))
	{
		$count += scalar(@{$objects->{$id}->{host}});
	}

	if (defined($objects->{$id}->{range}))
	{
		foreach my $range (@{$objects->{$id}->{range}})
		{
			$count += _countAddressesInRange($range);
		}
	}

	# This is where we get recursive.
	if (defined($objects->{$id}->{objectGroupReference}))
	{
		foreach my $objectGroupReference (@{$objects->{$id}->{objectGroupReference}})
		{
			$count += _countAddressesInNetworkGroup($objectGroupReference, $objects, $seen);
		}
	}

	# Save the count for future reference.
	$objects->{$id}->{count} = $count;

	return $count;
}

# When first called takens in two arguments, $objectName and $objects.  $objectName must be a string, the id of a
# network group.  $objects is a reference to $self->{_network_groups}.  Returns the smallest IP address in the
# network group.  This function is recursive, $seen is only used internally by this function for recursive calls.
sub _getSmallestAddressInNetworkGroup
{
	my ($objectName, $objects, $seen) = @_;

	# The smallest IP found so far, initilized to infinity (very large).
	my $smallestIp = Math::BigInt->binf();

	# Seen will be used to keep track of the object groups we have already checked.
	if (!defined($seen))
	{
		$seen = {};
	}

	# If we have already checked this object group then return now, otherwise we may recurse forever.
	if (defined($seen->{$objectName}))
	{
		return Math::BigInt->binf();
	}

	# Mark this object group as seen.
	$seen->{$objectName} = 1;

	# If this isn't a network group, return inf.
	if (!defined($objects->{$objectName}))
	{
		return Math::BigInt->binf();
	}

	# If this object group has already checked, return the value.
	if (defined($objects->{$objectName}->{smallIp}))
	{
		return $objects->{$objectName}->{smallIp};
	}

	# Go through each of the subnets, hosts and ranges, if we find a smaller IP than the current smallest,
	# replace it.
	if (defined($objects->{$objectName}->{subnet}))
	{
		foreach my $subnet (@{$objects->{$objectName}->{subnet}})
		{
			if (defined $subnet->{mask}) {
				my $ip = _buildIpFromCidr($subnet->{address} . '/' . $subnet->{mask})->intip();
				if ($smallestIp->bcmp($ip) > 0)
				{
					$smallestIp = $ip;
				}
			} else {
				ZipTie::Logger::get_logger()->debug(qq{AddressTranslationParser::_getSmallestAddressInNetworkGroup: subnetmask not defined } . Dumper($subnet));
			}
		}
	}

	if (defined($objects->{$objectName}->{host}))
	{
		foreach my $host (@{$objects->{$objectName}->{host}})
		{
			my $ip = Net::IP->new($host)->intip();
			if ($smallestIp->bcmp($ip) > 0)
			{
				$smallestIp = $ip;
			}
		}
	}

	if (defined($objects->{$objectName}->{range}))
	{
		foreach my $range (@{$objects->{$objectName}->{range}})
		{
			my $ip = Net::IP->new($range->{startAddress})->intip();
			if ($smallestIp->bcmp($ip) > 0)
			{
				$smallestIp = $ip;
			}
		}
	}

	# This is where we get recursive.
	if (defined($objects->{$objectName}->{objectGroupReference}))
	{
		foreach my $objectGroupReference (@{$objects->{$objectName}->{objectGroupReference}})
		{
			my $ip = _getSmallestAddressInNetworkGroup($objectGroupReference, $objects, $seen);
			if ($smallestIp->bcmp($ip) > 0)
			{
				$smallestIp = $ip;
			}
		}
	}

	# Save the smallest for future reference.
	$objects->{$objectName}->{smallIp} = $smallestIp;

	return $smallestIp;
}

sub _countAddressesInSubnet
{
	my ($subnet) = @_;
	my $returnVal = 0;
	if (defined $subnet->{mask}) {
		$returnVal =  (2 ** (32 - $subnet->{mask})); 
	} else {
		ZipTie::Logger::get_logger()->debug(qq{AddressTranslationParser::_countAddressesInSubnet: subnetmask not defined, returning 0}. Dumper($subnet));
	}
	return $returnVal;
}

sub _countAddressesInRange
{
	my ($range) = @_;

	my $startInt = Net::IP->new($range->{startAddress})->intip();
	my $endInt = Net::IP->new($range->{endAddress})->intip();

	return ($endInt->bsub($startInt)->numify() + 1);
}

# This subroutine takes in the ID of an objectGroupReference that is used for dynamic NAT.  Because dynamic NAT
# performs PAT on hosts, and non PAT on ranges the objectGroupReference will have to be broken apart.  If possible
# it will preserve the objcetGroupReference (if only hosts or ranges are used), but if it has to it will create
# PAT mappings with the hosts, and not-PAT mappings with the ranges, loosing the name of the objectGroupReference.
# It returns an array of addressMappings.
#
# http://www.cisco.com/en/US/docs/security/asa/asa84/configuration/guide/nat_rules.html#wp1509102
# http://www.cisco.com/en/US/docs/security/asa/asa84/configuration/guide/nat_objects.html#wp1541768
#
# NOTE: The 8.3 documentation doesn't make any mention of this.  I checked the 8.4 release notes here:
# http://www.cisco.com/en/US/docs/security/asa/asa84/release/notes/asarn84.html
# The release notes doesn't make any mention of PAT fallback functionality changing, so I made the assumption
# the PAT fallback functionality works the same in ASA 8.3.  I have no way to test to make sure this assumption is
# correct.
sub _createDynamicSrcMappedObjAddressMappings
{
	my ($self, $id) = @_;

	# $subnets is ignored
	my ($hosts, $subnets, $ranges) = break_network_object_into_hosts_subnets_ranges($self->{_network_groups}, $id);

	my $addressMappings = [];

	# We only have hosts, perform PAT with the object group reference.
	if ( (defined @{$hosts} && scalar(@{$hosts})) && (!defined @{$ranges} || !scalar(@{$ranges})))
	{
		my $patPoolAddressMapping = {};
		$patPoolAddressMapping->{sourceAddrNameMapped} = $id;
		$patPoolAddressMapping->{usePATOverload} = 1;
		push(@{$addressMappings}, $patPoolAddressMapping);
	}
	# We only have ranges, perform dynamic NAT without PAT with the object group reference.
	elsif ( (!defined @{$hosts} || !scalar(@{$hosts})) && (defined @{$ranges} && scalar(@{$ranges})))
	{
		my $patPoolAddressMapping = {};
		$patPoolAddressMapping->{sourceAddrNameMapped} = $id;
		push(@{$addressMappings}, $patPoolAddressMapping);
	}
	# This shuoldn't happen, so print out a debug message and don't return any new translations.
	elsif ( (!defined @{$hosts} || !scalar(@{$hosts})) && (!defined @{$ranges} || !scalar(@{$ranges})))
	{
		$self->{_logger}->debug("Trying to create dynamic NAT mapping with object group $id " .
				"when no ranges or hosts exist in the object group.");
		return [];
	}
	# We hav both hosts and ranges.  According to the Cisco documentation dynamic NAT will be performed
	# with the ranges, and PAT fallback will occur with the hosts.  We can no longer maintain the objcet
	# group reference.
	else
	{
		# Add hosts with PAT.
		foreach my $host (@{$hosts})
		{
			my $addressMapping = {};
			$addressMapping->{sourceAddrMapped} = $host;
			$addressMapping->{usePATOverload} = 1;
			push(@{$addressMappings}, $addressMapping);
		}

		# Add ranges without PAT.
		foreach my $range (@{$ranges})
		{
			my $addressMapping = {};
			$addressMapping->{sourceAddrMapped} = $range->{startAddress};
			$addressMapping->{sourceAddrEndMapped} = $range->{endAddress};
			push(@{$addressMappings}, $addressMapping);
		}
	}

	return $addressMappings;
}

# Creates mappings for an interface, one mapping per IP address.  Returns an array of mappings.
# If the interface isn't defined in $self->{_interface_ips} undef is returned as an error.
# If useAppliedAddressRules is set to true, then the useAppliedAddressRules is set to true in each mapping.
# If usePATOverload is set to true, then usePATOverload is set to 1 in each mapping.
sub _createInterfaceMappings
{
	my ($self, $interface, $useAppliedAddressRules, $usePATOverload) = @_;

	if (!defined($self->{_interface_ips}->{$interface}))
	{
		return;
	}

	my @addressMappings;

	foreach my $ip (@{$self->{_interface_ips}->{$interface}})
	{
		my $addressMapping = {};

		$addressMapping->{sourceAddrMapped} = $ip;

		if ($useAppliedAddressRules)
		{
			$addressMapping->{useAppliedAddressRules} = 'true';
		}

		if ($usePATOverload)
		{
			$addressMapping->{usePATOverload} = 1;
		}

		push(@addressMappings, $addressMapping);
	}

	return \@addressMappings;
}

# This function takes in a realIfc, a mappedIfc and a reference to a list of address mappings.
# If nat-control is enabled it records the mapped ip addresses for nat-control.
#
# $isOutsideDynamicNat is optional.  It should only be set when performing outside dynamic NAT.  It's
# used to keep track of the interfaces outside dynamic NAT is performed on.  Normally nat-control only
# applies to traffic going from higher security levels to lower security levels, but the following two
# exceptions exist:
#
# http://www.cisco.com/en/US/docs/security/asa/asa82/configuration/guide/nat_control.html#wp1081400
# http://www.cisco.com/en/US/docs/security/asa/asa82/configuration/guide/nat_control.html#wp1081410
#
# The list of interfaces that dynamic NAT is performed on allows us to model the above exceptions.
sub _recordMappedAddressesForNatControl
{
	my ($self, $realIfc, $mappedIfc, $addressMappings, $protocol, $isOutsideDynamicNat, $isDynamicNat) = @_;

	# Don't do anything if we are not under nat control
	if (!$self->{_is_under_nat_control})
	{
		return;
	}

	if ($realIfc eq $mappedIfc)
	{
		return;
	}

	# Keep track of interfaces dynamic NAT is performed on.
	if ($isOutsideDynamicNat)
	{
		if (!defined($self->{_nat_control_dynamic_nat_interfaces}))
		{
			$self->{_nat_control_dynamic_nat_interfaces} = {};
		}
		if (!defined($self->{_nat_control_dynamic_nat_interfaces}->{$realIfc}))
		{
			$self->{_nat_control_dynamic_nat_interfaces}->{$realIfc} = {};
		}

		$self->{_nat_control_dynamic_nat_interfaces}->{$realIfc}->{all} = 1;
	}
	elsif ($isDynamicNat)
	{
		if (!defined($self->{_nat_control_dynamic_nat_interfaces}))
		{
			$self->{_nat_control_dynamic_nat_interfaces} = {};
		}
		if (!defined($self->{_nat_control_dynamic_nat_interfaces}->{$realIfc}))
		{
			$self->{_nat_control_dynamic_nat_interfaces}->{$realIfc} = {};
		}

		$self->{_nat_control_dynamic_nat_interfaces}->{$realIfc}->{same} = 1;
	}

	my $key = $realIfc . '_' . $mappedIfc;

	if (!defined($self->{natControlMappedAddrs}))
	{
		$self->{natControlMappedAddrs} = {};
	}
	if (!defined($self->{natControlMappedAddrs}->{$key}))
	{
		$self->{natControlMappedAddrs}->{$key} = [];
	}

	foreach my $addressMapping (@{$addressMappings})
	{
		my $mappedAddress = {};

		$mappedAddress->{protocol} = $protocol;

		# Source range
		if (defined($addressMapping->{sourceAddrEndMapped}))
		{
			$mappedAddress->{sourceIpAddr}->{range}->{endAddress} = $addressMapping->{sourceAddrEndMapped};
			$mappedAddress->{sourceIpAddr}->{range}->{startAddress} = $addressMapping->{sourceAddrMapped};
		}
		# Source network
		elsif (defined($addressMapping->{sourceAddrMaskMapped}))
		{
			$mappedAddress->{sourceIpAddr}->{network}->{address} = $addressMapping->{sourceAddrMapped};
			$mappedAddress->{sourceIpAddr}->{network}->{mask} = $addressMapping->{sourceAddrMaskMapped}; 
		}
		# Source host
		elsif (defined($addressMapping->{sourceAddrMapped}))
		{
			$mappedAddress->{sourceIpAddr}->{host} = $addressMapping->{sourceAddrMapped};
		}
		# Source objectGroup
		elsif (defined($addressMapping->{sourceAddrNameMapped}))
		{
			$mappedAddress->{sourceIpAddr}->{objectGroupReference} = $addressMapping->{sourceAddrNameMapped};
		}

		# Destination range
		if (defined($addressMapping->{destAddrEndMapped}))
		{
			$mappedAddress->{destinationIpAddr}->{range}->{endAddress} = $addressMapping->{destAddrEndMapped};
			$mappedAddress->{destinationIpAddr}->{range}->{startAddress} = $addressMapping->{destAddrMapped};
		}
		# Destination network
		elsif (defined($addressMapping->{destAddrMaskMapped}))
		{
			$mappedAddress->{destinationIpAddr}->{network}->{address} = $addressMapping->{destAddrMapped};
			$mappedAddress->{destinationIpAddr}->{network}->{mask} = $addressMapping->{destAddrMaskMapped}; 
		}
		# Destination host
		elsif (defined($addressMapping->{destAddrMapped}))
		{
			$mappedAddress->{destinationIpAddr}->{host} = $addressMapping->{destAddrMapped};
		}
		# Destination objectGroup
		elsif (defined($addressMapping->{destAddrNameMapped}))
		{
			$mappedAddress->{destinationIpAddr}->{objectGroupReference} = $addressMapping->{destAddrNameMapped};
		}

		# Source port range
		if (defined($addressMapping->{sourcePortEndMapped}))
		{
			$mappedAddress->{sourceService}->{portRange}->{endPort} = $addressMapping->{sourcePortEndMapped};
			$mappedAddress->{sourceService}->{portRange}->{startPort} = $addressMapping->{sourcePortMapped}; 
		}
		# Source port
		elsif (defined($addressMapping->{sourcePortMapped}))
		{
			$mappedAddress->{sourceService}->{portExpression}->{operator} = 'eq';
			$mappedAddress->{sourceService}->{portExpression}->{port} = $addressMapping->{sourcePortMapped};
		}
		# Source port objectGroupReference
		elsif (defined($addressMapping->{sourcePortNameMapped}))
		{
			$mappedAddress->{sourceService}->{objectGroupReference} = $addressMapping->{sourcePortNameMapped};
		}
		# Destination port range
		elsif (defined($addressMapping->{destPortEndMapped}))
		{
			$mappedAddress->{destinationService}->{portRange}->{endPort} = $addressMapping->{destPortEndMapped};
			$mappedAddress->{destinationService}->{portRange}->{startPort} = $addressMapping->{destPortMapped}; 
		}
		# Destination port
		elsif (defined($addressMapping->{destPortMapped}))
		{
			$mappedAddress->{destinationService}->{portExpression}->{operator} = 'eq';
			$mappedAddress->{destinationService}->{portExpression}->{port} = $addressMapping->{destPortMapped};
		}
		# Destination port objectGroupReference
		elsif (defined($addressMapping->{destPortNameMapped}))
		{
			$mappedAddress->{destinationService}->{objectGroupReference} = $addressMapping->{destPortNameMapped};
		}

		push(@{$self->{natControlMappedAddrs}->{$key}}, $mappedAddress);
	}
}

# With NAT exemption there will be no mapped addresses, just real addresses NAT doesn't apply to.
# realIfc is required, but mappedIfc is optional.
# This subroutine is a wrapper for _recordMappedAddressesForNatControl, it does the following:
#  1) Converts the addressCriterion arguments syntax into equvilant addressTranslation arguments.
#  2) Determines all the possible mapped interfaces (for NAT control probably all interfaces).
#  3) For each realIfc to mappedIfc combirnation calls _recordMappedAddressesForNatControl with the
#     mappings generated in step 1.
sub _recordMappedAddressesForNatControlFromCriterion
{
	my ($self, $realIfc, $mappedIfc, $criterionArguments, $protocol) = @_;

	# Don't do anything if we are not under nat control
	if (!$self->{_is_under_nat_control})
	{
		return;
	}

	my @mappedInterfaces;

	# the zone mapper hash already has an array of every interface, use it.
	if (!defined($mappedIfc))
	{
		push(@mappedInterfaces, @{$self->{_zoneMapperHash}->{any}});
	}
	else
	{
		push(@mappedInterfaces, $mappedIfc);
	}

	# Convert the criterionArguments syntax into addressMapping syntax as the _recordMappedAddressesForNatControl
	# subroutine expects.
	my $addressMapping = {};
	$addressMapping->{sourceAddrMapped} = $criterionArguments->{sourceIp} if defined($criterionArguments->{sourceIp});
	$addressMapping->{sourceAddrEndMapped} = $criterionArguments->{sourceIpEnd} if defined($criterionArguments->{sourceIpEnd});
	$addressMapping->{sourceAddrMaskMapped} = $criterionArguments->{sourcePrefix} if defined($criterionArguments->{sourcePrefix});
	$addressMapping->{sourceAddrNameMapped} = $criterionArguments->{sourceIpObjectGroup} if defined($criterionArguments->{sourceIpObjectGroup});
	$addressMapping->{sourcePortMapped} = $criterionArguments->{sourceServicePort} if defined($criterionArguments->{sourceServicePort});
	$addressMapping->{sourcePortEndMapped} = $criterionArguments->{sourceServicePortEnd} if defined($criterionArguments->{sourceServicePortEnd});
	$addressMapping->{sourcePortNameMapped} = $criterionArguments->{sourceServiceObjectName} if defined($criterionArguments->{sourceServiceObjectName});

	$addressMapping->{destAddrMapped} = $criterionArguments->{destIp} if defined($criterionArguments->{destIp});
	$addressMapping->{destAddrEndMapped} = $criterionArguments->{destIpEnd} if defined($criterionArguments->{destIpEnd});
	$addressMapping->{destAddrMaskMapped} = $criterionArguments->{destPrefix} if defined($criterionArguments->{destPrefix});
	$addressMapping->{destAddrNameMapped} = $criterionArguments->{destIpObjectGroup} if defined($criterionArguments->{destIpObjectGroup});
	$addressMapping->{destPortMapped} = $criterionArguments->{destServicePort} if defined($criterionArguments->{destServicePort});
	$addressMapping->{destPortEndMapped} = $criterionArguments->{destServicePortEnd} if defined($criterionArguments->{destServicePortEnd});
	$addressMapping->{destPortNameMapped} = $criterionArguments->{destServiceObjectName} if defined($criterionArguments->{destServiceObjectName});

	foreach my $mappedIfc (@mappedInterfaces)
	{
		_recordMappedAddressesForNatControl($self, $realIfc, $mappedIfc, [$addressMapping], $protocol);
	}
}

# Static NAT is bi-direcitonal, so the criterion addresses provided by the rule is also the mapped addresses
# for the reverse rule.  This subroutine takes in a creterion mapping, inverts everything (source becomes dest, dest
# becomes source), and converts it into address mapping notation.  It returns an array of one element containing
# the new mappings.
sub _reverseCriterionArgumentsForReverseNatControlMapping
{
	my ($self, $criterionArguments) = @_;

	my $addressMapping = {};
	$addressMapping->{sourceAddrMapped} = $criterionArguments->{destIp} if defined($criterionArguments->{destIp});
	$addressMapping->{sourceAddrEndMapped} = $criterionArguments->{destIpEnd} if defined($criterionArguments->{destIpEnd});
	$addressMapping->{sourceAddrMaskMapped} = $criterionArguments->{destPrefix} if defined($criterionArguments->{destPrefix});
	$addressMapping->{sourceAddrNameMapped} = $criterionArguments->{destIpObjectGroup} if defined($criterionArguments->{destIpObjectGroup});
	$addressMapping->{sourcePortMapped} = $criterionArguments->{destServicePort} if defined($criterionArguments->{destServicePort});
	$addressMapping->{sourcePortEndMapped} = $criterionArguments->{destServicePortEnd} if defined($criterionArguments->{destServicePortEnd});
	$addressMapping->{sourcePortNameMapped} = $criterionArguments->{destServiceObjectName} if defined($criterionArguments->{destServiceObjectName});

	$addressMapping->{destAddrMapped} = $criterionArguments->{sourceIp} if defined($criterionArguments->{sourceIp});
	$addressMapping->{destAddrEndMapped} = $criterionArguments->{sourceIpEnd} if defined($criterionArguments->{sourceIpEnd});
	$addressMapping->{destAddrMaskMapped} = $criterionArguments->{sourcePrefix} if defined($criterionArguments->{sourcePrefix});
	$addressMapping->{destAddrNameMapped} = $criterionArguments->{sourceIpObjectGroup} if defined($criterionArguments->{sourceIpObjectGroup});
	$addressMapping->{destPortMapped} = $criterionArguments->{sourceServicePort} if defined($criterionArguments->{sourceServicePort});
	$addressMapping->{destPortEndMapped} = $criterionArguments->{sourceServicePortEnd} if defined($criterionArguments->{sourceServicePortEnd});
	$addressMapping->{destPortNameMapped} = $criterionArguments->{sourceServiceObjectName} if defined($criterionArguments->{sourceServiceObjectName});

	return [$addressMapping];
}

# This function is designed to obtain a mask, based on an input IP's address class.
#
# It follows the guidelines set out here:
#
# http://www.cisco.com/en/US/docs/security/asa/asa82/command/reference/no.html#wp1769584
# 
# And here:
#
# http://www.cisco.com/en/US/docs/security/asa/asa82/command/reference/s8.html#wp1512538
#
# The second document contains the following exception, that doesn't exist in the first document:
#
# 	If a host-bit is non-zero after masking, a host mask of 255.255.255.255 is used.
#
# This functionality in not enabled by default, but can be enabled by setting $hostBitNonZeroCheck to true.
sub _getMaskFromAddressClass
{
	my ($ip, $hostBitNonZeroCheck) = @_;

	my $mask;

	# If mask isn't defined then mask is set based on the class the IP address belongs to.
	$ip =~ /^(\d+)/;
	# Class A
	if ($1 <= 127)
	{
		$mask = 8;
	}
	# Class B
	elsif ($1 <= 191)
	{
		$mask = 16;
	}
	# Class C
	else
	{
		$mask = 24;
	}

	# If any host bit is non zero, return 32.
	if ($hostBitNonZeroCheck)
	{
		my $parsedIp = Net::IP->new($ip)->intip();
		my $firstIp = _buildIpFromCidr("$ip/$mask")->intip();

		# If $parsedIp doesn't equal $firstIp then a host bit is set in $parsedIp, return mask of 32.
		if ($parsedIp->bcmp($firstIp))
		{
			$mask = 32;
		}
	}

	return $mask;
}

# This function is creates a new address translation rule in $addressTranslation.  $addRuleArguments and
# $criterionArguments are both references to a hash of arguments to be passed into addRule and
# addRuleAddressCriterion methods.  $addressMappings is a reference to an array, that contains references to
# hashes that contains argumnets to be passed into addRuleAddressTranslationMapping.  addRuleAddressTranslationMapping
# is called for each set of arguments in $addressMappings.
#
# For the passed in argument sets for addRuleAddressCriterion and addRuleAddressTranslationMapping the rule argument
# does not need to be set.  This is set automatically after the call to addRule is completed.
sub _createAddressTranslationRule
{
	my ($addressTranslation, $addRuleArguments, $criterionArguments, $addressMappings) = @_;

	my $rule = $addressTranslation->addRule($addRuleArguments);

	$criterionArguments->{rule} = $rule;
	$addressTranslation->addRuleAddressCriterion($criterionArguments);

	foreach my $addressMapping (@{$addressMappings})
	{
		$addressMapping->{rule} = $rule;
		$addressTranslation->addRuleAddressTranslationMapping($addressMapping);
	}
}

# The new constructor for Net::IP doesn't accept a CIDR that has any host bits set, such as the following:
#     192.168.5.5/16
# Some functionaly in the ASA NAT parser relies on this to work.  This subroutine was created to allow this.
sub _buildIpFromCidr
{
    my ($cidr) = @_;

    if ($cidr =~ /^(\d+\.\d+\.\d+\.\d+)\/(\d+)$/)
    {  
        my $ip = $1;
        my $mask = $2;

		# Obtain the first IP address in CIDR and create the IP object with that.
        $ip = _getFirstIpInCidr($ip, $mask);
        return Net::IP->new("$ip/$mask");
    }
    else
    {  
	ZipTie::Logger::get_logger()->debug(qq{AddressTranslationParser::_buildIpFromCidr: Cannot regex on cidr="$cidr"});
        return;
    }
}

# Takes in the address and a mask portion of a CIDR and returns the first address in the CIDR.
# This subroutine accepts CIDRs that have host bits set, such as 192.168.5.5/16.
sub _getFirstIpInCidr
{
    my ($address, $mask) = @_;

	# Convert the mask to a bitmask (bigint).
    $mask = _generateBitmaskFromPrefix($mask);

	# Turn the address into a bigint.
    $address = Net::IP->new($address)->intip();

	# Perform bitwise AND between the address and mask, and then convert the result back to a string IP address.
	# This will return the first IP address in the CIDR.
    return _bigintToIpString($address->band($mask));
}

# Takes in the mask portion of a CIDR (an integer between 0 and 32), and returns a bitmask of length 32 for the
# corrisponding CIDR.
sub _generateBitmaskFromPrefix
{
    my ($mask) = @_;

    my $bitmask = Math::BigInt->bzero();

    for (my $i = 0;$i < $mask;$i++)
    {  
        $bitmask->brsft(1);
        $bitmask->badd(2147483648); # Add the most significant bit of an IP.
    }

    return $bitmask;
}

# Converts a bigint into an IP string, such as '192.168.0.0'.
sub _bigintToIpString
{
    my ($originalBigint) = @_;

	my $bigint = $originalBigint->copy();

    my $mask = Math::BigInt->new(0b11111111);

    my $ipStr = '';

    for (my $i = 0;$i < 4;$i++)
    {  
        my $quad = $bigint->copy();
        $quad = $quad->band($mask)->numify();
        $bigint->brsft(8);

        $ipStr = $quad . ".$ipStr";
    }

    # Remove the trailing '.'
    $ipStr =~ s/\.$//;

    return $ipStr;
}

# Returns the protocol as a string of a splitServiceGroup.  If more than one
# protocol, or no protocol,  is defined inside the service group 'ip' is returned.
sub _protocol_from_split_service_group
{
	my ($splitServiceGroup, $objectGroups) = @_;

	my $protocol = undef;

	foreach my $srcDest ('source', 'destination')
	{
		if (defined($splitServiceGroup->{$srcDest}) &&
				defined($objectGroups->{$splitServiceGroup->{$srcDest}}))
		{
			my $foundProtocol = _protocol_from_service_group_recursive(
					$objectGroups->{$splitServiceGroup->{$srcDest}}, $objectGroups);

			if (!defined($foundProtocol))
			{
				return 'ip';
			}
			elsif (!defined($protocol))
			{
				$protocol = $foundProtocol;
			}
			else
			{
				if ($protocol ne $foundProtocol)
				{
					return 'ip';
				}
			}
		}
	}

	if (!defined($protocol))
	{
		return 'ip';
	}

	return $protocol;
}

# This is a recursive helper subroutine for _protocol_from_split_service_group.  It returns
# the protocol of a service group, crawlwing recursively through objectGroupReferences.
# If more than one, or no protocol is defined, returns 'ip'.
sub _protocol_from_service_group_recursive
{
	my ($currentGroup, $objectGroups, $seen) = @_;

	if (!defined($seen))
	{
		$seen = {};
	}
	if (defined($seen->{$currentGroup->{id}}))
	{
		return '';
	}

	$seen->{$currentGroup->{id}} = 1;

	my $protocol = undef;

	foreach my $portExpression (@{$currentGroup->{portExpression}}, @{$currentGroup->{portRange}})
	{
		my $foundProtocol = $portExpression->{protocol};

		if (!defined($foundProtocol))
		{
			return;
		}
		elsif (!defined($protocol))
		{
			$protocol = $foundProtocol;
		}
		elsif ($protocol ne $foundProtocol)
		{
			return;
		}
	}

	foreach my $objectGroupReference (@{$currentGroup->{objectGroupReference}})
	{
		my $group = $objectGroups->{$objectGroupReference};

		my $foundProtocol = _protocol_from_service_group_recursive($group, $objectGroups, $seen);

		if (!defined($foundProtocol))
		{
			return;
		}
		elsif ($foundProtocol eq '')
		{
			next;
		}
		elsif (!defined($protocol))
		{
			$protocol = $foundProtocol;
		}
		elsif ($protocol ne $foundProtocol)
		{
			return;
		}
	}

	if (defined($protocol))
	{
		return $protocol;
	}

	return 'ip';
}

# checks to see if the incoming port is an integer.  If not
# it will do a lookup of the service name to convert it to an integer
sub _int_port
{
	my ($port) = @_;
	if ( $port =~ /^\d+$/ )
	{
		return $port;
	}
	else
	{
		return get_port_number($port);
	}
}

=head1 NAME

	ZipTie::Adapters::Cisco::SecurityAppliance::AddressTranslationParser -
	Parse ASA NAT and PAT config.

=head1 DESCRIPTION

	This module impliments the NAT and PAT parsers for both ASA 8.2 (and before)
	and ASA 8.3 (and later).  It uses ZipTie::Model::AddressTranslation for the
	low level XML generation.

=head1 SEE ALSO

	http://wiki.q1labs.inc/display/QRM/Address+Translation+Support+in+ZipTie+and+CSM

=head1 AUTHOR(S)

	David Payne
	david.payne@ca.ibm.com

=head1 COPYRIGHT

	Copyright (c) 2012, IBM Corporation. All Rights Reserved.
	This module is free software. It may be used, redistributed
	and/or modified under the same terms as Perl itself.

=cut

1;
