package ZipTie::Adapters::Cisco::SecurityAppliance::GetConfig;

use 5.010_000;
use strict;
use warnings;

use ZipTie::CLIProtocol;
use ZipTie::Response;
use ZipTie::Recording;
use ZipTie::Recording::Interaction;
use ZipTie::TransferProtocolFactory;
use ZipTie::TransferProtocol;
use ZipTie::ConnectionPath;
use ZipTie::ConnectionPath::Protocol;
use ZipTie::ConnectionPath::FileServer;
use ZipTie::Logger;
use ZipTie::Adapters::Utils qw(create_empty_file create_unique_filename escape_filename);

use Exporter 'import';
our @EXPORT_OK = qw(get_config);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

# Get the instance of the ZipTie::Recording module
my $RECORDING = ZipTie::Recording::get_recording();

sub get_config
{
	# Grab our ZipTie::CLIProtocol and optional ZipTie::ConnectionPath
	my $cli_protocol    = shift;
	my $connection_path = shift;
	my $filename        = shift;
	my $altPrompt       = shift;  # OPTIONAL

	# Create an undef reference that can eventually hold the startup configuration contents that are found
	my $response = undef;

	# Check to see if either TFTP or SCP are supported	
	my $tftp_protocol = $connection_path->get_protocol_by_name("TFTP") if ( defined($connection_path) );
	my $scp_protocol = $connection_path->get_protocol_by_name("SCP") if ( defined($connection_path) );

	# Check to see if TFTP is supported.  If so, a combination of a CLI Protocol AND TFTP will be used
	# to retrieve the startup configuration
	if ( defined($tftp_protocol) )
	{
		_touch_file_for_tftp($cli_protocol, $connection_path, $filename);
		$response = _get_config_tftp($cli_protocol, $connection_path, $filename);
	}
	
	# Check to see if SCP is supported.  If so, a SCP client will be used to retrieve the startup configuration
	elsif ( defined($scp_protocol) )
	{
		$response = _get_config_scp($connection_path, $filename);
	}
	
	# Otherwise, fall back to CLI protocol only
	else
	{
		$response = _get_config_cli($cli_protocol, $filename, $altPrompt);
	}

	my $prompt = $altPrompt // $cli_protocol->get_prompt_by_name("enablePrompt") // '[>#]\s*$';
	$response =~ s/$prompt$//; # Strip prompt
	$response =~ s/^.*?(?=^(!|\S+\s+Version\s+\d))//mis; # Strip heading lines
	$response =~ s/<---\s+More\s+--->\s*//mig; # Strip more prompt

	# Return the startup configuration found
	return $response;
}

sub _get_config_cli
{
	# Grab our ZipTie::CLIProtocol object
	my $cli_protocol = shift;
	my $filename     = shift;
	my $altPrompt    = shift;  # OPTIONAL

	# Sending "show startup-config" command
	my $command = "show $filename";
	my $t = time();
	$LOGGER->debug("Command to get $filename started at: " .  scalar localtime($t));
	$cli_protocol->send( $command );
	
	# Grab the enable prompt that was retrieved by the auto-login of the IOS device.
	my $enable_prompt = (defined $altPrompt) ? $altPrompt : $cli_protocol->get_prompt_by_name("enablePrompt");
	
	# use enable prompt if we have it, deal if we don't.
	
	my $regex = ($enable_prompt) ? $enable_prompt : '[>#]\s*$';

	# +++wayne response to Defect 74923 : Raise timeout to ridiculous levels for ginormous running configs.
	my $response = $cli_protocol->wait_for($regex, 900);
	
	if ($response =~ /% Invalid input detected|for a list of available commands/)
	{
		$LOGGER->fatal_error_code($INSUFFICIENT_PRIVILEGE, $cli_protocol->get_ip_address(), "Unable to issue \"$command\"");
	}
	$t = time();
	$LOGGER->debug("Command to get $filename finished at: " .  scalar localtime($t));
	
	# remove any --More-- prompt lines from older (1900) devices
	$response =~ s/^--More--\s*$//m;
	
	# remove leading cruft from the 'show' command output
	$response =~ s/^.*?(?=^!)//ms;
	$LOGGER->debug("SecurityAppliance::get_config_cli: response was: $response");
	
	# Return the startup configuration found
	return $response;
}

# +++wayne: What I've discovered on the support for SCP on ASA as of about 9.1....
#
#  *  The main issue is that show [running|startup]-config is by default an enable level operation.
#      To that end, the ASA needs to be accessed by an enable level account (which really isn't
#      operationally secure or otherwise appropriate for QRM).
#  *   There is a documented config level "ssh scopy enable" command, but I can't find it supported
#      as of 9.1.
#  *   Another alternative is to de-priv the show config command, but that's a bit sticky itself.
#  *   Cisco actually goes so far as to recommend outbound TFTP or SCP from the firewall to do
#      config archiving.  We don't have a server resident on the managed host or elsewhere to support
#      the former, and this routine isn't set up for the latter (to have the device SCP back to the
#      QRM MH I mean).
#  With that in mind, while SCP the way we want to do it would be very attractive for certain difficult
#  circumstances (like ginormous configs that take a ton of time to transfer over the SSH pty for 
#  get_config_cli), the configuration regime isn't well set up to allow us to do that.  This routine
#  sets up the MH as a SCP client, which is what we'd like to support.
#
sub _get_config_scp
{
	# Grab our ZipTie::ConnectionPath object
	my $connection_path = shift;
	my $filename	 	= shift;

	# Grab the ZipTie::ConnectionPath::Protocol object representing SCP from the ZipTie::ConnectionPath object
	my $scp_protocol = $connection_path->get_protocol_by_name("SCP");

	# Retrieve the startup configuration file from the device
	my $xfer_client = ZipTie::TransferProtocolFactory::create( $scp_protocol->get_name() );
	$xfer_client->connect(	$connection_path->get_ip_address(),
							$scp_protocol->get_port(),
							$connection_path->get_credential_by_name("username"),
							$connection_path->get_credential_by_name("password") );

	my $temp_filename = create_unique_filename();
	$xfer_client->get( $filename, $temp_filename );

	# Open up the startup configuration file and read it into memory
	open(STARTUP_CONFIG, $temp_filename) || $LOGGER->fatal("[$SCP_ERROR]\nCould not open the retrieved startup configuration file stored in '$temp_filename'");
	my @entire_file = <STARTUP_CONFIG>;
	close(STARTUP_CONFIG);
	my $config_contents = join( "", @entire_file );

	# Record the file transfer of the startup config
    # Arguments: protocol name, file name, response/contents, whether or not ZipTie acted as the file transfer server
    $RECORDING->create_xfer_interaction($scp_protocol->get_name(), $filename, $config_contents, 0);

	# Return the contents of the startup configuration
	return $config_contents;
}

sub _get_config_tftp
{
	# Grab our ZipTie::CLIProtocol and ZipTie::ConnectionPath objects
	my $cli_protocol	= shift;
	my $connection_path = shift;
	my $filename	 	= shift;

	# Specify the responses to check for
	my @responses = ();
	push(@responses, ZipTie::Response->new('memory invalid', undef, $DEVICE_MEMORY_ERROR));
	push(@responses, ZipTie::Response->new('Error opening nvram', undef, $NVRAM_CORRUPTION_ERROR));
	push(@responses, ZipTie::Response->new('Error|Unrecognized command', undef, $TFTP_ERROR));
	push(@responses, ZipTie::Response->new('% Invalid input detected at', undef, $INSUFFICIENT_PRIVILEGE));
	push(@responses, ZipTie::Response->new('[Uu]sage:', \&_write_net));
	push(@responses, ZipTie::Response->new('[Ss]ource [Ff]ilename', \&_specify_source_file));
	push(@responses, ZipTie::Response->new('[Rr]emote host', \&_specify_tftp_address));
	push(@responses, ZipTie::Response->new('[Dd]estination [Ff]ilename', \&_specify_destination_file));
	push(@responses, ZipTie::Response->new('bytes copied in [\d\.]+ secs|OK|#\s*', \&_finish));

	# Sending "copy $filename tftp" command
	my $command = "copy $filename tftp";
	$cli_protocol->send( $command );
	my $response = $cli_protocol->wait_for_responses(\@responses);

	# Based on the response of the device, determine the next interaction that should be executed.
	my $next_interaction = undef;
	if ($response)
	{
		$next_interaction = $response->get_next_interaction();
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}
	
	# Call the next interaction if there is one to call
	if ($next_interaction)
	{
		# Return the startup configuration found
		return &$next_interaction( $cli_protocol, $connection_path, $filename );
	}
}

sub _write_net()
{
	my $cli_protocol	= shift;
	my $connection_path = shift;
	my $filename	 	= shift;

	# Specify the responses to check for
	my @responses = ();
	push(@responses, ZipTie::Response->new('memory invalid', undef, $DEVICE_MEMORY_ERROR));
	push(@responses, ZipTie::Response->new('Error opening nvram', undef, $NVRAM_CORRUPTION_ERROR));
	push(@responses, ZipTie::Response->new('Error|Unrecognized command', undef, $TFTP_ERROR));
	push(@responses, ZipTie::Response->new('% Invalid input detected at', undef, $INSUFFICIENT_PRIVILEGE));
	push(@responses, ZipTie::Response->new('Timed out', undef, $TFTP_ERROR));
	push(@responses, ZipTie::Response->new('OK|#\s*', \&_finish));

	# Build destination filename
	my $full_filename = escape_filename( $cli_protocol->get_ip_address() ) . ".$filename";

	# Grab the ZipTie::Connection::FileServer object representing TFTP file server
	my $tftp_file_server	= $connection_path->get_file_server_by_name("TFTP");
	my $tftp_server_address	= $tftp_file_server->get_ip_address();

	# Sending "write net <tftp_server_address>:filename" command
	my $command = "write net $tftp_server_address:$full_filename";
	$cli_protocol->send( $command );
	my $response = $cli_protocol->wait_for_responses( \@responses, 120 );

	# Based on the response of the device, determine the next interaction that should be executed.
	my $next_interaction = undef;
	if ($response)
	{
		$next_interaction = $response->get_next_interaction();
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}
	
	# Call the next interaction if there is one to call
	if ($next_interaction)
	{
		# Return the startup configuration found
		return &$next_interaction( $cli_protocol, $connection_path, $filename );
	}
}

sub _specify_destination_file()
{
	# Grab our ZipTie::CLIProtocol and ZipTie::ConnectionPath objects
	my $cli_protocol	= shift;
	my $connection_path = shift;
	my $filename	 	= shift;
	
	# Specify the responses to check for
	my @responses = ();
	push(@responses, ZipTie::Response->new('Error|Unrecognized command|[Rr]emote host|[Dd]estination [Ff]ilename', undef, $TFTP_ERROR));
	push(@responses, ZipTie::Response->new('bytes copied in [\d\.]+ secs|OK|#\s*', \&_finish));

	# Sending "<ipAddress>.$fielname" as the destination file name for the configuration.
	my $command = escape_filename( $cli_protocol->get_ip_address() ) . ".$filename";
	$cli_protocol->send( $command );
	my $response = $cli_protocol->wait_for_responses( \@responses, 120 );

	# Based on the response of the device, determine the next interaction that should be executed.
	my $next_interaction = undef;
	if ($response)
	{
		$next_interaction = $response->get_next_interaction();
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}
	
	# Call the next interaction if there is one to call
	if ($next_interaction)
	{
		# Return the startup configuration found
		return &$next_interaction( $cli_protocol, $connection_path, $filename );
	}
}

sub _specify_tftp_address
{
	# Grab our ZipTie::CLIProtocol and ZipTie::ConnectionPath objects
	my $cli_protocol	= shift;
	my $connection_path = shift;
	my $filename	 	= shift;
	
	# Grab the ZipTie::Connection::FileServer object representing TFTP file server
	my $tftp_file_server = $connection_path->get_file_server_by_name("TFTP");
	
	# Specify the responses to check for
	my @responses = ();
	push(@responses, ZipTie::Response->new('Error|Unrecognized command|[Ss]ource [Ff]ilename|[Rr]emote host', undef, $TFTP_ERROR));
	push(@responses, ZipTie::Response->new('[Dd]estination [Ff]ilename', \&_specify_destination_file));
	push(@responses, ZipTie::Response->new('bytes copied in [\d\.]+ secs|OK|#\s*', \&_finish));

	# Sending the TFTP server address
	$cli_protocol->send( $tftp_file_server->get_ip_address() );
	my $response = $cli_protocol->wait_for_responses( \@responses );
	
	# Based on the response of the device, determine the next interaction that should be executed.
	my $next_interaction = undef;
	if ($response)
	{
		$next_interaction = $response->get_next_interaction();
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}
	
	# Call the next interaction if there is one to call
	if ($next_interaction)
	{
		# Return the startup configuration found
		return &$next_interaction( $cli_protocol, $connection_path, $filename );
	}
}


sub _specify_source_file
{
	# Grab our ZipTie::CLIProtocol and ZipTie::ConnectionPath objects
	my $cli_protocol	= shift;
	my $connection_path = shift;
	my $filename	 	= shift;

	# Specify the responses to check for
	my @responses = ();
	push(@responses, ZipTie::Response->new('Error|Unrecognized command|[Ss]ource [Ff]ilename', undef, $TFTP_ERROR));
	push(@responses, ZipTie::Response->new('[Dd]estination [Ff]ilename', \&_specify_destination_file));
	push(@responses, ZipTie::Response->new('[Rr]emote host', \&_specify_tftp_address));
	push(@responses, ZipTie::Response->new('bytes copied in [\d\.]+ secs|OK|#\s*', \&_finish));
 
	# Sending the TFTP server address
	$cli_protocol->send( $filename );
	my $response = $cli_protocol->wait_for_responses( \@responses );
	
	# Based on the response of the device, determine the next interaction that should be executed.
	my $next_interaction = undef;
	if ($response)
	{
		$next_interaction = $response->get_next_interaction();
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}
	
	# Call the next interaction if there is one to call
	if ($next_interaction)
	{
		# Return the startup configuration found
		return &$next_interaction( $cli_protocol, $connection_path, $filename );
	}
}

sub _finish
{
	# Grab our ZipTie::CLIProtocol and ZipTie::ConnectionPath objects
	my $cli_protocol	= shift;
	my $connection_path = shift;
	my $filename	 	= shift;

	$cli_protocol->get_response(0.25); # flush the response

	# Grab the ZipTie::Connection::FileServer object representing TFTP file server
	my $tftp_file_server = $connection_path->get_file_server_by_name("TFTP");
	
	# Retrieve the startup configuration file from the TFTP server
	my $config_name = escape_filename( $cli_protocol->get_ip_address() ) . ".$filename";
	my $config_file = $tftp_file_server->get_root_dir() . "/$config_name";
	
	# Open up the startup configuration file and read it into memory
	open(CONFIG_FILE, $config_file) || $LOGGER->fatal("[$TFTP_ERROR]\nCould not open the retrieved configuration file stored in '$config_file'");
	my @entire_file = <CONFIG_FILE>;
	close(CONFIG_FILE);
	my $config_contents = join("", @entire_file);
	
	# Clean up our tracks by deleteing the configuration file that was sent to the TFTP server
	unlink ($config_file);
	
	# Record the file transfer of the startup config
    # Arguments: protocol name, file name, response/contents
    $RECORDING->create_xfer_interaction($tftp_file_server->get_protocol(), $config_name, $config_contents);
	
	# Return the contents of the startup configuration
	return $config_contents;
}

sub _touch_file_for_tftp
{
	my $cli_protocol	= shift;
	my $connection_path = shift;
	my $filename	 	= shift;
	
	my $tftp_file_server = $connection_path->get_file_server_by_name("TFTP");
	my $config_name = escape_filename( $cli_protocol->get_ip_address() ) . ".$filename";
	create_empty_file($tftp_file_server->get_root_dir() . "/$config_name");
}

1;
__END__

=head1 Parsers

ZipTie::Adapters::Cisco::SecurityAppliance::GetConfig

=head1 SYNOPSIS

use ZipTie::Adapters::Cisco::SecurityAppliance::GetConfig qw(get_config);
get_config(cli_protocol, connection_path, filename, altPrompt);	

=head1 DESCRIPTION

Module to handle the interactions to retrieve and return one of the named configurations on the Cisco
Security Appliance.

=head2 Methods

=over 12

=item C<get_config(cli_protocol, connection_path, filename, altPrompt)>

Retrieves and returns a named configuration using the CLI attributes and connection data as passed.

=over -6

PARAMS:

=item C<cli_protocol> - ref to ZipTie::CLIProtocol object which wraps the communication channel to the device

=item C<connection_path> - object returned from ZipTie::Typer::translate_document which encapsulates the device session endpoint
 
=item C<filename> - well-known filename being retrieved, generally running-config or startup-config

=item C<altPrompt> - optional parameter, if provided, will be scanned for instead of current enable prompt as end of response in any cli context

=back

=over -6

RETURNS:

=item C<response> - "raw" response from configuration fetch (i.e, prompts, interior formatting not stripped).	

=back

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/
  
Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.
 
The Original Code is Ziptie Client Framework.
  
The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

Contributor(s): rkruse, Wayne Tackabury (Q1 Labs)

Date: Sept. 22, 2011

=cut