package ZipTie::Adapters::Cisco::SecurityAppliance::Parsers;

use strict;
use warnings;
use 5.010_000;

use Data::Dumper;
$Data::Dumper::Terse = 1; # Do not print $VAR1 before each dumped variable
$Data::Dumper::Sortkeys = 1; # Sort hash keys in dumped hashes

use ZipTie::Adapters::Utils qw(mask_to_bits strip_mac seconds_since_epoch get_mask get_port_number get_port_expression get_protocol_name trim get_interface_type get_crep);
use ZipTie::Adapters::Cisco::SecurityAppliance::AddressTranslationParser;
use MIME::Base64 qw(encode_base64);
use ZipTie::Adapters::Cisco::SecurityAppliance::NameMapper;
use ZipTie::Adapters::Cisco::SecurityAppliance::AclToRoute;
use ZipTie::RemoteGatewayResolution;

use Exporter qw(import);
our @EXPORT_OK =
  qw(parse_names parse_arp parse_telemetry_interfaces parse_mac_table parse_routing_neighbors parse_routing create_config parse_local_accounts parse_chassis parse_object_groups parse_filters parse_snmp parse_system parse_interfaces parse_dynamic_routes parse_static_routes parse_vlans parse_stp parse_subelements parse_subelement_data);

# Class constants.
# The reasonable memory sizes in MBs: 2 to the n-th power.
our @VALID_MEMSIZES = map { 2**$_ } qw(2 3 4 5 6 7 8 9 10 11 12 13 14 15 16);

# The string which collects the sizes from 4 (2**2) MBs to 65536 (2**16) MBs.
# Format the array into a space-padded string.
our $SIZEPATTERNS_STRING = " @VALID_MEMSIZES ";

# The number of KBs in 80 percent of 1 MB.
our $KBS_80PERCENTMB = int( my $volatile = 1024 * 0.8 );

my $cipm			= get_crep('cipm');
my $cidr			= get_crep('cidr');

# This at one point WAS where all mappings came from, but is now only for the contingency where the ASA redefines a 
# given port to a different value
# Current reference: http://www.cisco.com/en/US/docs/security/asa/asa84/configuration/guide/ref_ports.html#wp1007738

my $PIX_PREDEFINED_SERVICES = [
						{ serviceName => "aol", mapping =>  [{protocol => 'tcp', port => 5190}]},  # IANA has it as a range
						{ serviceName => "bootpc", mapping =>  [{protocol => 'udp', port => 68}]}, # ASA mapping ignores TCP port
						{ serviceName => "bootps", mapping =>  [{protocol => 'udp', port => 67}]}, # ditto
						{ serviceName => "chargen", mapping =>  [{protocol => 'tcp', port => 19}]},  # ASA ignores UDP mapping
						{ serviceName => "daytime", mapping =>  [{protocol => 'tcp', port => 13}]},  # ditto
						{ serviceName => "dnsix", mapping =>  [{protocol => 'udp', port => 195}]}, # somewhat obscure mapping
						{ serviceName => "h323", mapping => [{protocol => 'tcp', port => 1720}]}, # ASA ignores call control port
						{ serviceName => "hostname", mapping => [{protocol => 'tcp', port => 101}]}, # ASA ignores UDP mappings
						{ serviceName => "irc", mapping =>  [{protocol => 'tcp', port => 194}]}, # ASA ignores UDP
						{ serviceName => "isakmp", mapping =>  [{protocol => 'udp', port => 500}]}, 
						{ serviceName => "klogin", mapping =>  [{protocol => 'tcp', port => 543}]},
						{ serviceName => "kshell", mapping =>  [{protocol => 'tcp', port => 544}]},
						{ serviceName => "ldap", mapping =>  [{protocol => 'tcp', port => 389}]},
						{ serviceName => "lpd", mapping =>  [{protocol => 'tcp', port => 515}]},
						{ serviceName => "login", mapping =>  [{protocol => 'tcp', port => 513}]},  # aka rlogin
						{ serviceName => "lotusnotes", mapping =>  [{protocol => 'tcp', port => 1352}]},
						{ serviceName => "mobile-ip", mapping => [{protocol => 'udp', port => 496}]}, # ASA unique designation
						{ serviceName => "radius", mapping => [{protocol => 'udp', port => 1645}]},  # This is an ASA *bug*
						{ serviceName => "radius-acct", mapping => [{protocol => 'udp', port => 1646}]}, # ditto
						{ serviceName => "rpc", mapping => [{protocol => 'tcp', port =>135}]},						
						{ serviceName => "talk", mapping => [{protocol => 'tcp', port => 517}, {protocol => 'udp', port => 517}]},
						{ serviceName => "time", mapping => [{protocol => 'udp', port => 37}]},  
						{ serviceName => "uucp", mapping => [{protocol => 'tcp', port =>540}]},						
						{ serviceName => "whois", mapping => [{protocol => 'tcp', port =>43}]},						
						{ serviceName => "xdmcp", mapping => [{protocol => 'udp', port =>177}]},
				];

my $MAX_NAT_SIZE = 20000000;
my $MAX_ACL_SIZE = 20000000;

# As of sometime around ASA 8.3, you can use for the longer syntax of "object group", just 
# the text "object", for cases of a single entry, e.g.,
#
#   object network LabNet
#   subnet 172.16.30.0 255.255.255.0
#   description Allow ftp backups of cisco devices
#   object network EIGRPRouters
#   range 172.16.57.2 172.16.57.14
#   description Allow EIGRP Router Backups
#     . . .
# access-list outside_access_in extended permit tcp 172.16.56.0 255.255.255.0 object LabNet eq ftp
# 
# This is shorthand for "object" with optional "-group" following it, that won't muck up any
# surrounding regexp capture grouping.
#
my $OG_KWD = qr/object(?:-group)?/;

# The following is the list of behaviours to apply to a BVI interface (transparent firewalls)
# The switching behavior is by request of Rory Bray for topology
# Setting the interface behavior to switching so topology knows this is more than just an
# administrative only interface, and topology will have to pay attention to it.
my @BVI_BEHAVIORS = ('administrative', 'switching');
 
my $LOGGER = ZipTie::Logger::get_logger();

sub parse_routing_neighbors
{
	my ( $in, $out ) = @_;
	my $opened = 0;

	while ( $in->{ospf} =~ /^\s*([\da-f.:]+)\s+\d+\s+\S+(?:\s+\S+)?\s+\S+\s+([\da-f.:]+)\s+(\S+)\s*$/gm )
	{
		$out->open_element('routingNeighbors') if (!$opened);
		$opened = 1;
		my $neighbor =
		{
			protocol => 'OSPF',	
			routerId => $1,	
			ipAddress => $2,	
			interface => $3,
		};
		$out->print_element('routingNeighbor', $neighbor);
	}
	my $route_neighbors;
	while ( $in->{routes} =~ /^(I|R|B|D|i|S)\*?\s+([\da-f:.]+)\s+(?:[\da-f:.]+)\s+\S+\s+via\s+([\da-f:.]+),.*?(\S+)\s*$/gm )
	{
		if ( $1 eq 'I' )
		{
			push @{$route_neighbors->{igrp}}, { protocol => 'IGRP', ipAddress => $3, interface => $4 };
		}
		elsif ( $1 eq 'I' )
		{
			push @{$route_neighbors->{rip}}, { protocol => 'RIP', ipAddress => $3, interface => $4 };
		}
		elsif ( $1 eq 'B' )
		{
			push @{$route_neighbors->{bgp}}, { protocol => 'BGP', ipAddress => $3, interface => $4 };
		}
		elsif ( $1 eq 'D' )
		{
			push @{$route_neighbors->{eigrp}}, { protocol => 'EIGRP', ipAddress => $3, interface => $4 };
		}
		elsif ( $1 eq 'i' )
		{
			push @{$route_neighbors->{isis}}, { protocol => 'ISIS', ipAddress => $3, interface => $4 };
		}
		elsif ( $1 eq 'S' )
		{
			push @{$route_neighbors->{static}}, { protocol => 'STATIC', ipAddress => $3, interface => $4 };
		}
	}
	while ( ( my $key, my $value ) = each( %{$route_neighbors} ) )
	{
		if ( ref( $route_neighbors->{$key} ) eq 'ARRAY' )
		{
			$out->open_element('routingNeighbors') if (!$opened);
			$opened = 1;
			foreach my $neighbor ( @{$route_neighbors->{$key}} )
			{
				$out->print_element( 'routingNeighbor', $neighbor );
			}
		}
	}

	$out->close_element('routingNeighbors') if ($opened);
}

sub parse_arp
{
	my ( $in, $out ) = @_;

	$out->open_element('arpTable');
	while ( $in->{arp} =~ /^\s*(\S+)\s+($cipm+)\s+(\S+)/mg )
	{
		my $arp = {
			ipAddress  => $2,
			macAddress => strip_mac($3),
			interface  => $1,
		};
		$out->print_element( 'arpEntry', $arp );
	}
	$out->close_element('arpTable');
}

sub parse_telemetry_interfaces
{
	my ( $in, $out ) = @_;

	my @interfaces;

	$out->open_element('interfaces');
	while ( $in->{interfaces} =~ /(\S+)\s+"([^"]+)",?\s+(is.+?)(?=^\S)/msg )
	{
		my $blob = $3;
		my $name = $2;
		my $interface =
		{ 
			name => $name,
			type => _get_interface_type($1, $blob, $in, $name),
			inputBytes => 0,
		};
		if ( $blob =~ /line protocol is (down|up)/mi )
		{
			$interface->{operStatus}= ucfirst($1);
		}
		if ( $blob =~ /(\d+) packets input,\s+(\d+)\s*bytes/mi )
		{
			$interface->{inputBytes}= $2;
		}
		while ( $blob =~ /IP\s+address\s+([\d\.:a-fA-F]+),\s+subnet\s+mask\s+([\d\.:a-fA-F]+)/mgi )
		{
			my $ipEntry =
			{
				ipAddress => $1,	
				mask => mask_to_bits ( $2 ),
			};
			push (@{$interface->{ipEntry}}, $ipEntry);
		}
		$out->print_element( 'interface', $interface );
		push(@interfaces, $interface);
	}
	$out->close_element( 'interfaces' );

	# Return the interfaces hash for choose_admin_ip
	return { interface => \@interfaces };
}

sub parse_mac_table
{
	my ( $in, $out ) = @_;

	my $macTableOpened = 0;

	# Skip past the column headers
	$in->{mac} =~ /^\s*-{10,}\s*$/mg;

	# Example mac table entry taken from the Cisco documentation:
	# outside                 0009.7cbe.2100                 static               -

	# Example mac table entry taken from a device with bridge groups
	# inside                     0050.5691.664c          dynamic      4         1
	while ($in->{mac} =~ /^\s*(\S+)\s+(\S+)/mg)
	{
		my $interface = $1;
		my $mac = strip_mac($2);

		if (!$macTableOpened)
		{
			$out->open_element('macTable');
			$macTableOpened = 1;
		}

		$out->print_element('macEntry', { macAddress => $mac, interface => $interface });
	}

	if ($macTableOpened)
	{
		$out->close_element('macTable');
	}
}
##################################################

sub parse_chassis
{
	my ( $in, $out, $connPtr, $connType ) = @_;
	my $serialNumber;
	my $modelNumber;
	my $hwNumber;
	my $processorBoard;
	my $cpuid         = 0;
	my @memories      = ();
	my @deviceStorage = ();

	foreach my $line ( split( /\n/, $in->{version} ) )
	{
		if ( $line =~ /^Hardware:\s+([^\s,]+)/i )
		{
			$hwNumber = $1;
		}
		if ( $line =~ /^Serial\s+Number:\s+(\S+)/i )
		{
			$serialNumber = $1;
		}
		if ( $line =~ /\bCPU\s+(\S{1}.+)$/i )
		{
			$processorBoard = $1;
		}

		#if ( $line =~ /^Hardware:\s+([^\,\s]+)/i )
		#{
		#	$modelNumber = $1;
		#}
		if ( $line =~ /\b(\d+)\s+(\S+)\s+RAM/i )
		{
			my $ramsize    = $1;
			my $sizeprefix = $2;
			if ( $sizeprefix =~ /MB/i )
			{
				$ramsize = $ramsize * 1024 * 1024;
			}
			elsif ( $sizeprefix =~ /KB/i )
			{
				$ramsize = $ramsize * 1024;
			}
			push @memories, { kind => 'RAM', size => $ramsize };
		}
		if ( $line =~ /^Flash([^\,]+),\s+(\d+)(\S+)/ )
		{
			my $flashsize  = $2;
			my $sizeprefix = $3;
			if ( $sizeprefix =~ /MB/i )
			{
				$flashsize = $flashsize * 1024 * 1024;
			}
			elsif ( $sizeprefix =~ /KB/i )
			{
				$flashsize = $flashsize * 1024;
			}
			push @memories, { kind => 'Flash', size => $flashsize };
			push @deviceStorage, { name => 'flash', storageType => 'flash', size => $flashsize };
		}
		if ( $line =~ /^BIOS\s+Flash([^\,]+),\s+(\d+)(\S+)/ )
		{
			my $configmemsize = $2;
			my $sizeprefix    = $3;
			if ( $sizeprefix =~ /MB/i )
			{
				$configmemsize = $configmemsize * 1024 * 1024;
			}
			elsif ( $sizeprefix =~ /KB/i )
			{
				$configmemsize = $configmemsize * 1024;
			}
			push @memories, { kind => 'ConfigurationMemory', size => $configmemsize };
		}
	}

	if ( $in->{version} =~ /(?<!:\/)\b(PIX[^\s,]+)/mi )
	{
		$modelNumber = $1;
	}
	elsif ( $in->{version} =~ /(?<!:\/)\b(ASA\d+)/mi )
	{
		$modelNumber = $1;
	}
	elsif ( $in->{version} =~ /(?<!:\/)\b(WS[^\s,]+)/mi )
	{
		$modelNumber = $1;
	}
	elsif ( $in->{version} =~ /(FWSM)/ )
	{
		$modelNumber = $1;
	}
	else
	{
		$modelNumber = 'PIX-520';
	}

	$out->open_element("chassis");

	my $chassisAsset = { "core:assetType" => "Chassis", };
	$chassisAsset->{'core:factoryinfo'}->{'core:hardwareVersion'} = $hwNumber     if ( defined $hwNumber );
	$chassisAsset->{'core:factoryinfo'}->{'core:serialNumber'}    = $serialNumber if ( defined $serialNumber );
	$chassisAsset->{'core:factoryinfo'}->{'core:make'}            = "Cisco";
	$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'}     = $modelNumber  if ( defined $modelNumber );

	$out->print_element( "core:asset", $chassisAsset );

	my $cpu;
	$cpu->{cpuType} = $processorBoard if $cpuid != -1 and $processorBoard;
	$out->print_element( "cpu", $cpu ) if (defined $cpu);

	foreach my $ds (@deviceStorage)
	{
		$out->print_element( "deviceStorage", $ds );
	}

	foreach my $memory (@memories)
	{
		$out->print_element( "memory", $memory );
	}

	$out->close_element("chassis");
}

sub parse_names
{
	my ($in) = @_;
	
	my $ipRe = get_crep('cipm');

	my $nameMapper = ZipTie::Adapters::Cisco::SecurityAppliance::NameMapper->new();

	while ($in->{names} && $in->{names} =~ /^name\s+($ipRe)\s+(\S+)/mg)
	{
		my $ip = $1;
		my $name = $2;

		$nameMapper->add_ip_mapping($name, $ip);
	}

	$in->{name_mapper} = $nameMapper;
}

sub parse_system
{
	my ( $in, $out ) = @_;
	my ($hostname) = $in->{running_config} =~ /^hostname\s+(\S+)/im;
	$out->print_element( 'core:systemName', $hostname );

	$out->open_element('core:osInfo');
	if ( $in->{version} =~ /^System image file is "([^"]+)/m )
	{
		$out->print_element( 'core:fileName', $1 );
	}

	# Try to grab OS name, OS type and Version using 1 pattern
	$out->print_element( 'core:make', 'Cisco' );
	if ( $in->{version} =~ /^Cisco (\S+)\s.*(?=Version)Version\s+(\S+)/mi )
	{
		$out->print_element( 'core:name',    $1 );
		$out->print_element( 'core:version', $2 );
	}
	elsif ( $in->{version} =~ /Adaptive Security Appliance Software Version\s+(\d+\.\d+\(\d+\))/mi )
	{
		$out->print_element( 'core:name',    "ASDM" );
		$out->print_element( 'core:version', $1 );
	}
	else
	{
		my $osType = "";
		if ( $in->{version} =~ /^(?:Cisco )?(\S+)\s+(\S+)/mi )
		{
			$out->print_element( 'core:name', $1 );
			$osType = $1;
		}
		if ( $in->{version} =~ /^(?:Cisco )?\S+\s+[\S]+\s+Version (\S[^\s,]+)/mi )
		{
			$out->print_element( 'core:version', $1 );
		}
		elsif ( $in->{version} =~ /^Software Version (\S+)/mi )
		{
			$_ = $1;
			s/^[Vv]//;
			$out->print_element( 'core:version', $_ );
		}
		elsif ( $in->{version} =~ /^Version\s+(\S+)/mi )
		{
			$_ = $1;
			s/^[Vv]//;
			$out->print_element( 'core:version', $_ );
		}
	}
	$out->print_element( 'core:osType', 'PIX' );

	# Print order of operations for NAT.
	$out->open_element('core:packetProcessingOrdering');
	$out->open_element('core:packetProcessingElement');
	$out->print_element('core:packetProcessingPhase', 'packet-filtering');
	$out->print_element('core:processingPoint', 'source');
	$out->close_element('core:packetProcessingElement');
	$out->open_element('core:packetProcessingElement');
	$out->print_element('core:packetProcessingPhase', 'address-translation');
	$out->close_element('core:packetProcessingElement');
	$out->open_element('core:packetProcessingElement');
	$out->print_element('core:packetProcessingPhase', 'packet-filtering');
	$out->print_element('core:processingPoint', 'destination');
	$out->close_element('core:packetProcessingElement');
	$out->close_element('core:packetProcessingOrdering');

	$out->close_element('core:osInfo');

	if (   ( $in->{version} =~ /ROM:\s+System\s+Bootstrap[,]\s+Version\s+([^\s,]+)/i )
		or ( $in->{version} =~ /ROM:\s+Bootstrap\s+program\s+is\s+(.*)/i )
		or ( $in->{version} =~ /BOOTLDR:\s+\S+\s+Boot\s+Loader.*Version\s+([^\s,]+)/i )
		or ( $in->{version} =~ /^ROM:\s+TinyROM\s+version\s+([^\s]+)/mi )
		or ( $in->{version} =~ /^ROM:\s+([^\s]+)/mi ) )
	{
		$out->print_element( 'core:biosVersion', $1 );
	}

	if ( $in->{version} =~ /\bWS-C\d{4}|catalyst/i )
	{
		$out->print_element( 'core:deviceType', 'Switch' );
	}
	else
	{
		$out->print_element( 'core:deviceType', 'Firewall' );
	}

	if ($in->{running_config} =~ /^snmp-server contact (.+)/m)
	{
		$out->print_element( 'core:contact', $1);
	}

	# System restarted at 18:00:01 CST Sun Feb 28 1993
	if ( $in->{version} =~ /^System restarted (?:by \S+\s)?at\s+(\d{1,2}:\d{1,2}:\d{1,2})\s+(\S+)\s+\S+\s+(\S+)\s+(\d{1,2})\s+(\d{4})/mi )
	{
		my $year     = $5;
		my $month    = $3;
		my $day      = $4;
		my $time     = $1;
		my $timezone = $2;

		my ( $hour, $min, $sec ) = $time =~ /(\d+):(\d+):(\d+)/;

		$out->print_element( "core:lastReboot", seconds_since_epoch( $sec, $min, $hour, $day, $month, $year, $timezone ) );
	}
	elsif (( $in->{version} =~ /uptime is\s+(.+)/i )
		or ( $in->{version} =~ /\bup\s+([\w\s]+)$/mi ) )
	{
		$_ = $1;
		my ($years)   = /(\d+)\s*years?/;
		my ($weeks)   = /(\d+)\s*weeks?/;
		my ($days)    = /(\d+)\s*days?/;
		my ($hours)   = /(\d+)\s*hours?/;
		my ($minutes) = /(\d+)\s*minutes?/;

		# subract the last reboot from the current time
		my $lastReboot = time();
		$lastReboot -= $years * 52 * 7 * 24 * 60 * 60 if ($years);
		$lastReboot -= $weeks * 7 * 24 * 60 * 60      if ($weeks);
		$lastReboot -= $days * 24 * 60 * 60           if ($days);
		$lastReboot -= $hours * 60 * 60               if ($hours);
		$lastReboot -= $minutes * 60                  if ($minutes);
		$out->print_element( "core:lastReboot", $lastReboot );
	}
}

# parse_dynamic_routes
#
# Parse any supported dynamic routes.  Currently supported dynamic routes:
#	- eigrp
#	- ospf
#
# PARAMS: $in - Input context.  See the actual parsing subroutines for requirements. (ex, _parse_eigrp_routes).
#         $out - $out - output XML printer context.
#
# RETURNS: No useful return value.  If dynamic routes was parsed they will have been printed to $out.  If no
#          dynamic routes was parsed nothing will be printed to $out.
sub parse_dynamic_routes {
	my ($in, $out) = @_;
	my @dynamicRoutes = ();
	
	$in->{_remoteGatewayResolution} = ZipTie::RemoteGatewayResolution->new();

	$LOGGER->debug("Parsing dynamic routes");

	my %supported_protocols = ( eigrp => 'D|EX', ospf => 'O' );

	foreach my $name ( keys %supported_protocols ) {
		my $code = $supported_protocols{$name};
		my $routes_found = 0;

		$LOGGER->debug("Parsing $name routes");

		# If the routing process is not in the running config then ospf is not configured.
		if( $in->{running_config} =~ /^router\s+$name\s+(\d+)/m ) {
			# When we support multiple RIBs this will be the advertisingRoutingProcess
			#my $routing_process = $1;

			# Parse the actual routes from the show route command
			while( $in->{show_route} =~ /^(($code)\s*\S{0,2}\s+($cipm)\s+($cipm)\s+\[\d+\/(\d+)\]\s+via\s+($cipm),\s+[\d:]+,\s+(\S+))/mg) {
				my $line = $2;

				my $dynamicRoute = {
					advertisingRoutingProcess => $name,
					routeNetwork => {
						address => $3,
						mask => mask_to_bits($4)
					},
					# The schema supports multiple routeNextHop elements.
					routeNextHop => [{
						adjacencyMetric => $5,
						adjacencyNode => $6,
						adjacencyInterface => $7
					}]
				};

				$LOGGER->verbose("Parsed $name route: '$line'\nroute #" . ++$routes_found . " =\n" . Dumper($dynamicRoute));

				push @dynamicRoutes, $dynamicRoute;
			}
		} else {
			$LOGGER->debug("No $name instance declared in running-config");
		}

		$LOGGER->debug("Parsed $routes_found $name routes");
	}

	# We only want the dynamicRoutes element to exist if we actually have dynamicRoute element(s) to print.
	my $route_count = scalar @dynamicRoutes;
	if( $route_count ) {
		$out->open_element('dynamicRoutes');
		foreach my $dynamicRoute (@dynamicRoutes) {

			$out->print_element('dynamicRoute', $dynamicRoute);

			# Add the dynamic routes to the remoteGatewayResolution object to help resolve static routes with remote gateways.
			foreach my $routeNextHop (@{$dynamicRoute->{routeNextHop}}) {
				$in->{_remoteGatewayResolution}->add_route($dynamicRoute->{routeNetwork}->{address},
						$dynamicRoute->{routeNetwork}->{mask}, $routeNextHop->{adjacencyNode},
						$routeNextHop->{adjacencyMetric}, 0, { dynamicRoute => 1 });
			}
		}
		$out->close_element('dynamicRoutes');
	}

	$LOGGER->debug("Found $route_count dynamic routes");

	return;
}

sub parse_static_routes
{
	my ( $in, $out ) = @_;

	my @route_array = $in->{running_config} =~ m/^route\s.*/mig;
	foreach my $line (@route_array){
		# route interface_name ip_address netmask gateway_ip [[ metric ] [ track number ] | tunneled ]
		if ($line =~ /^route\s+(?<interface>\S+)\s+(?<dstIp>\S+)\s+(?<mask>$cipm)\s+(?<gateway>(\S+))\s?(?<metric>\d+)?/mig){
			
			my $parsedInterface = $+{interface};
			my $destinationAddress = $in->{name_mapper}->get_ip($+{dstIp});
			my $gatewayAddress = $in->{name_mapper}->get_ip($+{gateway});
			my $destinationMask = mask_to_bits($+{mask});
			my $routeMetric = $+{metric} ? $+{metric} : undef;
			unless ($gatewayAddress =~ /$cipm/){ # handle the case where gateway_ip is optinal
				$gatewayAddress = undef;
			}
		
			my $defaultGateway = 0;
			if (($destinationAddress eq '0.0.0.0') && ($destinationMask eq '0'))
			{
				$defaultGateway = 1;
			}
			my $additionalDetails = {};
			$additionalDetails->{defaultGateway} = $defaultGateway ? 'true' : 'false';
			$in->{_remoteGatewayResolution}->add_route($destinationAddress, $destinationMask, $gatewayAddress, $routeMetric, $defaultGateway, $additionalDetails);
			
			# Add additional attributes to route object not set in add_route for null0 "discard" case
			if ( $parsedInterface =~ /null0/i) {
				my @routes = @{$in->{_remoteGatewayResolution}->{_routes}};
				$routes[$#routes]->{interface} = $parsedInterface;
				$routes[$#routes]->{discard} = "true";
			}
		}
	}

	$in->{_remoteGatewayResolution}->resolve_remote_gateways();

	my @routes = @{ $in->{_remoteGatewayResolution}->get_routes() };
	my $count = scalar @routes;
	$LOGGER->debug("Found $count static routes");

	my $staticRoutesOpen = 0;
	foreach my $route ( @routes )
	{
		# Don't print out dynamic routes.
		if (defined($route->{otherDetails}->{dynamicRoute}))
		{
			next;
		}

		my $printableRoute = {};
		$printableRoute->{gatewayAddress} = $route->{gatewayAddress} if defined $route->{gatewayAddress}; # optional
		$printableRoute->{destinationAddress} = $route->{destinationAddress};
		$printableRoute->{destinationMask} = $route->{destinationMask};
		$printableRoute->{interface} = $route->{interface};
		$printableRoute->{defaultGateway} = $route->{otherDetails}->{defaultGateway};
		$printableRoute->{routeMetric} = $route->{routeMetric} if (defined($route->{routeMetric})); # optional
		$printableRoute->{discard} = $route->{discard} if (defined($route->{discard})); # optional

		if (!$staticRoutesOpen)
		{
			$staticRoutesOpen = 1;
			$out->open_element('staticRoutes');
		}

		$out->print_element('staticRoute', $printableRoute);
	}

	# Print the crypto map routes
	my $crypto_count = 0;
	if (defined($in->{_crypto_map_routes}))
	{
		my @crypto_routes = @{ $in->{_crypto_map_routes} };
		$crypto_count = scalar @crypto_routes;

		foreach my $route (@crypto_routes)
		{
			if (!$staticRoutesOpen)
			{
				$staticRoutesOpen = 1;
				$out->open_element('staticRoutes');
			}

			$out->print_element('staticRoute', $route);
		}
	}
	$LOGGER->debug("Found $crypto_count crypto map routes (modelled as static routes)");

	if ($staticRoutesOpen)
	{
		$out->close_element('staticRoutes');
	}

	if (scalar(@{$in->{_remoteGatewayResolution}->get_problem_routes()}))
	{
		$LOGGER->debug('parse_static_routes: Couln\'t resolve remote gateway addresses for ' . scalar(@{$in->{_remoteGatewayResolution}->get_problem_routes()}) . ' routes.  Routes dropped.');
	}

	delete $in->{_remoteGatewayResolution};
}

sub parse_object_groups
{
	my ( $in, $out ) = @_;

	my ($og_blob)		= $in->{running_config} =~ /^($OG_KWD.+?)access-list/mis;
	my $group_opened	= 0;
	my $og_type			= "";
	my $og_name			= "";
	my $og_proto;

	my $network_groups;
	my $network_groups_hash = {};  # Store them in a hash referenced by ID for easier retreval.
	my $protocol_groups;
	my $service_groups = [];
	my $icmp_groups;
	my $group_params;

	# Save the object groups for later retreaval
	$in->{object_groups} = {};

	if ( $og_blob )
	{
		while ( $og_blob =~ /^(.+)$/mig )
		{
			$_ = $1;
			if ( /^$OG_KWD (\S+) (\S+)(?: (\S+))?\s*$/i )
			{
				if ( $og_type eq 'network' )
				{
					push @{$network_groups}, $group_params;
					$network_groups_hash->{$group_params->{id}} = $group_params;
				}
				elsif ( $og_type eq 'service' )
				{
					# Because a single 'object service' could create more than one service group, special handling is required.
					_add_service_groups($in, $service_groups, $group_params);
				}
				elsif ( $og_type eq 'protocol' )
				{
					push @{$protocol_groups}, $group_params;
				}
				elsif ( $og_type eq 'icmp-type' )
				{
					push @{$icmp_groups}, $group_params;
				}
				$group_params	= undef;
				$og_type		= lc($1);
				$og_name		= $2;

				if (defined($3))
				{
					$og_proto = $3;

					# This service group will not be broken into source and destination components
					if ($og_type eq 'service')
					{
						$group_params->{serviceComponents} = 0;
					}
				}
				else
				{
					$og_proto = undef;

					# This service group will be broken into source and destination components
					if ($og_type eq 'service')
					{
						$group_params->{serviceComponents} = 1;
					}
				}

				$group_params->{id} = $og_name;
				if ( $og_type eq 'network' )
				{
					$group_opened = 1;
				}
				elsif ( $og_type eq 'service' )
				{
					$group_opened = 1;
				}
				elsif ( $og_type eq 'protocol' )
				{
					$group_opened = 1;
				}
				elsif ( $og_type eq 'icmp-type' )
				{
					$group_opened = 1;
				}
				else
				{
					$group_opened = 0;
				}
			}
			else
			{
				if ( /^\s*description (\S.+)$/i && $group_opened )
				{
					$group_params->{description} = $1;
				}
				elsif ( /^\s*group-object (\S+)\s*$/i && $group_opened )
				{
					push @{$group_params->{objectGroupReference}}, $1;
				}
				elsif ( /^\s*network-object\s+object\s+(\S+)\s*$/i && $group_opened )
				{
					# this is an object group reference inside of a "object network"
					# notation -- presumed ASA >= 8.3 but it doesn't really matter
					push @{$group_params->{objectGroupReference}}, $1;
				}
				elsif ( /^\s*network-object(?: host)? (\S+)( $cipm)?\s*$/i && $group_opened )
				{
					my $ip = $in->{name_mapper}->get_ip($1);
					tagNonIpv4Address ($group_params, $ip);	
					# This is for ASA 8.2 and earlier
					push @{$group_params->{host}}, $ip if ( !$2 );
					if ( $2 )
					{
						my $sn = _create_subnet($ip, trim($2));
						if ( trim($2) eq mask_to_bits( trim($2) ) ) {
                        				$LOGGER->debug(qq{parse_object_groups: mask is non-contiguous or invalid: $ip } . trim ($2));
                				} else {	
							push @{$group_params->{subnet}}, $sn;
						}
					}
				}
				elsif ( /^\s*protocol-object (\S+)\s*$/i && $group_opened )
				{
					push @{$group_params->{protocol}}, lc( $1 );
				}
				elsif ( /^\s*port-object\s+(\S+)\s+(\S+)( \S+)?\s*$/i && $group_opened )
				{
					# TODO: should use "_svc_int_port()" to map PIX service references
					if ( $1 eq 'range' )
					{
						my $portStart;
						my $portEnd;
						if (defined($og_proto) && ($og_proto eq 'tcp' || $og_proto eq 'udp'))
						{
							$portStart = _int_port($2, $og_proto);
							$portEnd = _int_port(trim($3), $og_proto);
						}
						else
						{
							$portStart = _int_port($2);
							$portEnd = _int_port(trim($3));
						}

						if (defined($og_proto))
						{
							if ($og_proto eq 'tcp-udp')
							{
								push @{$group_params->{portRange}}, { portStart => $portStart, portEnd => $portEnd, protocol => 'tcp' };
								push @{$group_params->{portRange}}, { portStart => $portStart, portEnd => $portEnd, protocol => 'udp' };
							}
							else
							{
								push @{$group_params->{portRange}}, { portStart => $portStart, portEnd => $portEnd, protocol => $og_proto };
							}
						}
						else
						{
							push @{$group_params->{portRange}}, { portStart => $portStart, portEnd => $portEnd };
						}
					}
					else
					{
						my $operator = $1;
						my $port;
						if (defined($og_proto) && ($og_proto eq 'tcp' || $og_proto eq 'udp'))
						{
							$port = _int_port($2, $og_proto);
						}
						else
						{
							$port = _int_port($2);
						}

						if (defined($og_proto))
						{
							if ($og_proto eq 'tcp-udp')
							{
								push @{$group_params->{portExpression}}, { operator => $operator, port => $port, protocol => 'tcp' };
								push @{$group_params->{portExpression}}, { operator => $operator, port => $port, protocol => 'udp' };
							}
							else
							{
								push @{$group_params->{portExpression}}, { operator => $operator, port => $port, protocol => $og_proto };
							}
						}
						else
						{
							push @{$group_params->{portExpression}}, { operator => $operator, port => $port };
						}
					}
				}
				# The ASA 8.3 "object network...." subgroup notation
				elsif ( /^\s*host\s+(\S+)/ && $group_opened )
				{
					my $ip = $in->{name_mapper}->get_ip($1);
				
					tagNonIpv4Address ($group_params, $ip);					
					push @{$group_params->{host}}, $ip;
				}
				elsif ( /^\s*subnet\s+(\S+)\s+($cipm)/ && $group_opened )
				{
					my $ip = $in->{name_mapper}->get_ip($1);
					tagNonIpv4Address ($group_params, $ip);	
					my $snMask = trim($2);
					if ($snMask && ! ( $snMask eq mask_to_bits( $snMask ) ) )
					{
						push @{$group_params->{subnet}}, _create_subnet($ip, $snMask);
					}
					else
					{
						$LOGGER->debug("parse_object_groups: Could not derive subnet mask for 8.3 address group beginning with $ip Or netmask is non-contiguous $snMask subnet dropped.");
					}
				}
				elsif ( /^\s*subnet\s+(\S+)/ && $group_opened )
				{
					tagNonIpv4Address ($group_params, $1); # chech if a non-IPv4 address
				}	
				elsif ( /^\s*range\s+(\S+)\s+(\S+)/ && $group_opened )
				{
					my $startAddress = $in->{name_mapper}->get_ip($1);
					tagNonIpv4Address ($group_params, $startAddress);	
					my $endAddress = $in->{name_mapper}->get_ip($2);
					push @{$group_params->{range}}, { startAddress => $startAddress, endAddress => $endAddress};
				}				
				elsif ( /^\s*icmp-object (\S+)\s*$/i && $group_opened )
				{
					push @{$group_params->{icmpObject}}, $1;
				}
				elsif ( /^\s*service(?:-object)?\s+(tcp-udp|udp|tcp)(.*)/ && $group_opened )
				{
					my $protocol = $1;
					my $remainder = $2;

					my @capture = $remainder =~ /\s*(source|destination)?\s*(\S+)\s*(\S+)\s*(\S+)?/gc;

					my $i = 0;
					while ($i < scalar(@capture))
					{
						my $srcDest = 'destination';
						if (defined($capture[$i]))
						{
							$srcDest = $capture[$i];
						}
						++$i;

						my $operator = $capture[$i];
						++$i;

						my $port = _int_port($capture[$i]);
						++$i;

						my $portEnd = undef;
						if (defined($capture[$i]))
						{
							$portEnd = _int_port($capture[$i]);
						}
						++$i;

						if ($operator eq 'range')
						{
							if ($protocol eq 'tcp-udp')
							{
								push @{$group_params->{$srcDest}->{portRange}},
										{ portStart => $port, portEnd => $portEnd, protocol => 'tcp' };
								push @{$group_params->{$srcDest}->{portRange}},
										{ portStart => $port, portEnd => $portEnd, protocol => 'udp' };
							}
							else
							{
								push @{$group_params->{$srcDest}->{portRange}},
										{ portStart => $port, portEnd => $portEnd, protocol => $protocol };
							}
						}
						else
						{
							if ($protocol eq 'tcp-udp')
							{
								push @{$group_params->{$srcDest}->{portExpression}},
										{ operator => $operator, port => $port, protocol => 'tcp' };
								push @{$group_params->{$srcDest}->{portExpression}},
										{ operator => $operator, port => $port, protocol => 'udp' };
							}
							else
							{
								push @{$group_params->{$srcDest}->{portExpression}},
										{ operator => $operator, port => $port, protocol => $protocol };
							}
						}
					}
				}
				elsif ( /^\s*service(?:-object)?\s+(icmp|icmp6)\s*(\S+)?/ && $group_opened )
				{
					my $protocol = $1;
					my $icmpType = $2;

					push @{$group_params->{destination}->{portRange}}, { portStart => 0, portEnd => 65535, protocol => $protocol };
				}
				elsif ( /^\s*service(?:-object)?\s+object\s+(\S+)/ && $group_opened )
				{
					my $objectName = $1;

					push @{$group_params->{objectGroupReference}}, $objectName;
				}
				elsif ( /^\s*service(?:-object)?\s+(\S+)/ && $group_opened )
				{
					my $protocol = _name_protocol($1);

					push @{$group_params->{destination}->{portRange}}, { portStart => 0, portEnd => 65535, protocol => $protocol };
				}
			}
		}
	}

	if ( $og_type eq 'network' )
	{
		push @{$network_groups}, $group_params;
		$network_groups_hash->{$group_params->{id}} = $group_params;
	}
	elsif ( $og_type eq 'service' )
	{
		# Because a single 'object service' could create more than one service group, special handling is required.
		_add_service_groups($in, $service_groups,$group_params);
	}
	elsif ( $og_type eq 'protocol' )
	{
		push @{$protocol_groups}, $group_params;
	}
	elsif ( $og_type eq 'icmp-type' )
	{
		push @{$icmp_groups}, $group_params;
	}
	
	if ( $network_groups || $protocol_groups || $service_groups )
	{
		$out->open_element("objectGroups");

		if ( $icmp_groups )
		{
			$group_params = undef;
			foreach $group_params ( @{$icmp_groups} )
			{
				$in->{object_groups}->{$group_params->{id}} = $group_params;
				$in->{object_groups}->{$group_params->{id}}->{type} = 'icmp';

				$out->open_element("icmpGroup");
				$out->print_element( "description", $group_params->{description} ) if ($group_params->{description});
				$out->print_element( "id", $group_params->{id} );
				if ( $group_params->{objectGroupReference} )
				{
					foreach my $objectGroupReference ( @{$group_params->{objectGroupReference}} )
					{
						$out->print_element( "objectGroupReference", $objectGroupReference ) ;
					}
				}
				if ( $group_params->{icmpObject} )
				{
					foreach my $icmpObject ( @{$group_params->{icmpObject}} )
					{
						$out->print_element( "icmpObject", $icmpObject ) ;
					}
				}
				$out->close_element("icmpGroup");
			}
		}
		if ( $network_groups )
		{
			$group_params = undef;
			NETWORK_GROUP: foreach $group_params ( @{$network_groups} )
			{
				if ($group_params->{notIpv4}){
					$LOGGER->debug("parse_object_groups: Dropping network group $group_params->{id} as contains non-IPv4 address type(s)");
					next; # skip any groups containing non-IPv4 addresses (likely IPv6 which QRM does NOT support)
				}
				if ( $group_params->{objectGroupReference} )
				{
					foreach my $objectGroupReference ( @{$group_params->{objectGroupReference}} )
					{
						if (not exists $in->{object_groups}->{$objectGroupReference}){
							$LOGGER->debug("parse_object_groups: Dropping network group $group_params->{id} as nested group $objectGroupReference not found");
							next NETWORK_GROUP;
						}
					}
				}
				
				$in->{object_groups}->{$group_params->{id}} = $group_params;
				$in->{object_groups}->{$group_params->{id}}->{type} = 'network';

				$out->open_element("networkGroup");
				$out->print_element( "description", $group_params->{description} ) if ($group_params->{description});
				$out->print_element( "id", $group_params->{id} );
				if ( $group_params->{objectGroupReference} )
				{
					foreach my $objectGroupReference ( @{$group_params->{objectGroupReference}} )
					{
						$out->print_element( "objectGroupReference", $objectGroupReference ) ;
					}
				}
				if ( $group_params->{host} )
				{
					foreach my $host ( @{$group_params->{host}} )
					{
						$out->print_element( "host", $host ) ;
					}
				}
				if ( $group_params->{range} )
				{
					foreach my $range ( @{$group_params->{range}} )
					{
						$out->print_element( "range", $range ) ;
					}
				}
				if ( $group_params->{subnet} )
				{
					foreach my $subnet ( @{$group_params->{subnet}} )
					{
						$out->print_element( "subnet", $subnet ) ;
					}
				}
				$out->close_element("networkGroup");
			}
		}
		if ( $protocol_groups )
		{
			$group_params = undef;
			foreach $group_params ( @{$protocol_groups} )
			{
				$in->{object_groups}->{$group_params->{id}} = $group_params;
				$in->{object_groups}->{$group_params->{id}}->{type} = 'protocol';

				$out->open_element("protocolGroup");
				$out->print_element( "description", $group_params->{description} ) if ($group_params->{description});
				$out->print_element( "id", $group_params->{id} );
				if ( $group_params->{objectGroupReference} )
				{
					foreach my $objectGroupReference ( @{$group_params->{objectGroupReference}} )
					{
						$out->print_element( "objectGroupReference", $objectGroupReference ) ;
					}
				}
				if ( $group_params->{protocol} )
				{
					foreach my $protocol ( @{$group_params->{protocol}} )
					{
						$out->print_element( "protocol", $protocol ) ;
					}
				}
				$out->close_element("protocolGroup");

				# Save protocol groups for use in NAT.
				$in->{protocolGroups}->{$group_params->{id}} = $group_params;
			}
		}
		if ( $service_groups )
		{
			$group_params = undef;
			foreach $group_params ( @{$service_groups} )
			{
				# Skip bad service groups
				if (_bad_service_group($group_params->{id}, $in->{object_groups}))
				{
					# Since this object group will not be printed out remove it from the tracking hashes
					if (defined($in->{object_groups}->{$group_params->{id}}))
					{
						delete($in->{object_groups}->{$group_params->{id}});
					}
					if ($group_params->{id} =~ /(.*)_(source|destination)/ &&
							defined($in->{splitServiceGroups}->{$1}->{$2}))
					{
						delete($in->{splitServiceGroups}->{$1}->{$2});
					}
					next;
				}

				$out->open_element("serviceGroup");
				$out->print_element( "description", $group_params->{description} ) if ($group_params->{description});
				$out->print_element( "id", $group_params->{id} );
				if ( $group_params->{objectGroupReference} )
				{
					foreach my $objectGroupReference ( @{$group_params->{objectGroupReference}} )
					{
						if (!_bad_service_group($group_params->{id}, $in->{object_groups}))
						{
							$out->print_element( "objectGroupReference", $objectGroupReference ) ;
						}
					}
				}
				if ( $group_params->{portExpression} )
				{
					foreach my $portExpression ( @{$group_params->{portExpression}} )
					{
						$out->print_element( "portExpression", $portExpression ) ;
					}
				}
				if ( $group_params->{portRange} )
				{
					foreach my $portRange ( @{$group_params->{portRange}} )
					{
						$out->print_element( "portRange", $portRange ) ;
					}
				}
				$out->close_element("serviceGroup");
			}
		}

		$out->close_element("objectGroups");

		# Store the network groups for later use by NAT.
		$in->{_network_groups} = $network_groups_hash;
	}
}

# 'object service' groups has the possibility of creating two groups, one for source services, one for destination services.
# Because of this it requires special treatment.
sub _add_service_groups
{
	my ($in, $serviceGroups, $groupParams) = @_;

	if ($groupParams->{serviceComponents})
	{
		foreach my $srcDest ('source', 'destination')
		{
			my $newGroup = {};
			$newGroup->{id} = $groupParams->{id} . "_$srcDest";

			foreach my $key (keys(%{$groupParams->{$srcDest}}))
			{
				$newGroup->{$key} = $groupParams->{$srcDest}->{$key};
			}

			foreach my $objectGroupReference (@{$groupParams->{objectGroupReference}})
			{
				push @{$newGroup->{objectGroupReference}}, $objectGroupReference . "_$srcDest";
			}

			if (defined($groupParams->{description}))
			{
				$newGroup->{description} = $groupParams->{description};
			}

			push @{$serviceGroups}, $newGroup;
			_track_service_groups($in, $newGroup);
		}
	}
	else
	{
		push @{$serviceGroups}, $groupParams;
		_track_service_groups($in, $groupParams);
	}
}

# If we split a service group into source and destination service groups, the ID of the service group will
# no longer match the original ID used on the device.  We track this in a hash $in->{splitServiceGroups},
# so users of the object groups will be able to use the correct ID.
# Record all service groups inside $in->{object_groups} with the correct type.
sub _track_service_groups
{
	my ($in, $newGroup) = @_;

	$in->{object_groups}->{$newGroup->{id}} = $newGroup;
	$in->{object_groups}->{$newGroup->{id}}->{type} = 'service';

	if ($newGroup->{id} =~ /^(.+)_(source|destination)$/)
	{
		$in->{splitServiceGroups}->{$1}->{$2} = $newGroup->{id};
	}
}

# Recursive subroutine.  Takes in a groupId and a reference to $in->{object_groups}.  Returns true if
# the object group doesn't exist or if this object group and any object group referenced by it doesn't
# contain any actual ports or protocols, otherwise returns false.  $checkedGroups should not be set
# when being called from outside _bad_service_group, it only contains state information when called
# recursivly.
sub _bad_service_group
{
	my ($currentGroupId, $allGroups, $checkedGroups) = @_;

	my $currentGroup;
	if (defined($allGroups->{$currentGroupId}))
	{
		$currentGroup = $allGroups->{$currentGroupId};
	}
	else
	{
		return 1;
	}

	if ((defined($currentGroup->{portExpression}) && scalar($currentGroup->{portExpression})) ||
			(defined($currentGroup->{portRange}) && scalar($currentGroup->{portRange})))
	{
		return 0;
	}

	if (!defined($checkedGroups))
	{
		$checkedGroups = {};
	}
	if (defined($checkedGroups->{$currentGroupId}))
	{
		return 1;
	}

	$checkedGroups->{$currentGroupId} = 1;

	if (defined($currentGroup->{objectGroupReference}))
	{
		foreach my $objectGroupReference (@{$currentGroup->{objectGroupReference}})
		{
			unless (_bad_service_group($objectGroupReference, $allGroups, $checkedGroups))
			{
				return 0;
			}
		}
	}

	return 1;
}

sub _parse_security_levels
{ 
	my ($in) = @_;
	my $security_levels;
	pos($in->{running_config}) = undef;
	while( $in->{running_config} =~ /^interface\s+(\S+)\s+([^!]+)!/mig )
	{
		my $blob = $2;
		my $interface = $1;
		if( $blob =~ /no nameif/ )
		{
			next;
		}
		if( $blob =~ /nameif\s+(\S+)/ )
		{
			$interface = $1;
			$security_levels->{$interface} = 0;
			$security_levels->{$interface} = 100 if( $interface =~ /^inside$/ );
			if( $blob =~ /security-level\s+(\d+)/ )
			{
				$security_levels->{$interface} = int($1);	
			}
		}
	}
	return $security_levels;
}

sub parse_filters
{
	my ($in, $out) = @_;
	my $term_process_order = 1;
	my $name = " ";
	my $old_name = " ";

	$out->open_element("filterLists");

	$LOGGER->debug("parse_filters:_get_access_groups");
	my $access_groups = _get_access_groups($in);
	$LOGGER->debug("parse_filters:_parse_security_levels");
	my $security_levels = _parse_security_levels($in);
	$LOGGER->debug("parse_filters:parse_time_range");
	my $time_range = parse_time_range( $in );
	$LOGGER->debug("parse_filters:_generate_security_levels_acl_references");
	my $securityLevelsAclReferences = _generate_security_levels_acl_references($in, $access_groups,
			$security_levels);

	# Parse the ACLs, doesn't actually print them out yet.
	$LOGGER->debug("parse_filters:_parse_acls");
	my $acls = _parse_acls($in, $time_range);

	$LOGGER->debug("parse_filters:_parse_shun");
	my $shunFilterLists = _parse_shun($in);

	# Add the deny all for the security levels deny.
	if ($securityLevelsAclReferences ne '')
	{
		push(@{$acls}, {
			name => 'Security_Levels_Deny',
			entries => "<filterEntry>\n<log>false</log>\n<primaryAction>deny</primaryAction>\n" .
					"<processOrder>1</processOrder>\n</filterEntry>\n"
		});
	}

	# Parse the interfaces.  This will create the $in->{_interface_ips} hash, but will not print out any XML
	# because the second argument is undefined.
	$LOGGER->debug("parse_filters:parse_interfaces");
	parse_interfaces($in, undef);

	# Parse the NAT rules, doesn't actually print them out yet.
	my $addressTranslationParser = ZipTie::Adapters::Cisco::SecurityAppliance::AddressTranslationParser->
			new($in, $security_levels);
	$LOGGER->debug("parse_filters:parse82Nat");
	$addressTranslationParser->parse82Nat();
	$LOGGER->debug("parse_filters:parseTwiceNat");
	$addressTranslationParser->parseTwiceNat();
	$LOGGER->debug("parse_filters:parseObjectNat");
	$addressTranslationParser->parseObjectNat();
	$LOGGER->debug("parse_filters:generateXml");
	$addressTranslationParser->generateXml();

	# Undef is an expected value of returnNatControlFilterListAndReference
	my ($natControlFilterListReferences, $natControlFilterLists) =
			$addressTranslationParser->returnNatControlFilterListAndReference();

	# Print pre-routing ACL filterListReferences.
	_print_acl_references($out, $acls, $access_groups, 'in', $shunFilterLists);

	# Print pre-routing NAT.
	my $preRoutingNAT = $addressTranslationParser->getPreRoutingFilterLists();

	$preRoutingNAT = _trim_NAT($preRoutingNAT);
	unless ($preRoutingNAT eq '')
	{
		$out->open_element('filterListsSection');
		$out->print_string($preRoutingNAT);
		$out->open_element('phaseDiscriminator');
		$out->open_element('processingPhase');
		$out->print_element('core:packetProcessingPhase', 'routing');
		$out->close_element('processingPhase');
		$out->print_element('processingPhasePosition', 'before');
		$out->close_element('phaseDiscriminator');
		$out->close_element('filterListsSection');
	}

	# Print post-routing NAT.
	my $postRoutingNAT = $addressTranslationParser->getPostRoutingFilterLists();

	$postRoutingNAT = _trim_NAT($postRoutingNAT);
	unless ($postRoutingNAT eq '')
	{
		$out->open_element('filterListsSection');
		$out->print_string($postRoutingNAT);
		$out->open_element('phaseDiscriminator');
		$out->open_element('processingPhase');
		$out->print_element('core:packetProcessingPhase', 'routing');
		$out->close_element('processingPhase');
		$out->print_element('processingPhasePosition', 'after');
		$out->close_element('phaseDiscriminator');
		$out->close_element('filterListsSection');
	}

	# Print post-routing filterListReferences.
	_print_acl_references($out, $acls, $access_groups, 'out', $shunFilterLists,
			$securityLevelsAclReferences, $natControlFilterListReferences);

	# Print all ACLs.
	_print_acls($out, $acls, $natControlFilterLists, $shunFilterLists);

	$out->close_element("filterLists");
}

sub _trim_NAT{
	my ($nat) = @_;
	my $current_size = 0;
	my $trimmed_nat = '';

	my $original_size = length($nat);
	if($original_size <= $MAX_NAT_SIZE){
		return($nat);
	}

	$LOGGER->debug("NAT section size, $original_size, is greater than the allowed max, $MAX_NAT_SIZE\nStarting to trim NAT section...");

	my @filter_lists = split(/(?=<filterList>)/, $nat);
	foreach my $filter_list(@filter_lists){
		my $filter_list_size = length($filter_list);
		$current_size += $filter_list_size;

		if($current_size > $MAX_NAT_SIZE){

			$LOGGER->debug("NAT section too large, skipping");
			$current_size -= $filter_list_size;

			last;
		}
		$trimmed_nat .= $filter_list;
	}

	$LOGGER->debug("Finished trimming NAT section\nNew NAT section size is $current_size bytes");
	return($trimmed_nat);
}

# Prints either input ACL references (post-routing) or output ACL references (pre-routing).
sub _print_acl_references
{
	my ($out, $acls, $access_groups, $inOut, $shunFilterLists,
			$securityLevelsAclReferences, $natControlFilterListReferences) = @_;

	# Keeps track, if the filterListsSection has been open or not.
	my $filterListsSectionOpen = 0;

	# $appliesToSourceOrDest is set to destination if this is an output ACL,
	# and source if this is an input ACL.  Used for the appliesTo element.
	my $appliesToSourceOrDest = 'destination';
	# $direction is set to egress if this is an output ACL, and ingress if
	# this is an input ACL.
	my $direction = 'egress';
	# $beforeOrAfterRouting is set to after if this is a post routing ACL,
	# and before if this is a pre routing ACL.  All input ACLs are pre-routing,
	# and all output ACLs are post-routing. 
	my $beforeOrAfterRouting = 'after';
	if ($inOut eq 'in')
	{
		$appliesToSourceOrDest = 'source';
		$beforeOrAfterRouting = 'before';
		$direction = 'ingress';
	}

	# Output any shun filter list references if they exist.
	foreach my $shunFilterList (@{$shunFilterLists})
	{
		my $filterListReference = "filterListReference_$inOut";

		if (defined($shunFilterList->{$filterListReference}))
		{
			if (!$filterListsSectionOpen)
			{
				$out->open_element('filterListsSection');
				$filterListsSectionOpen = 1;
			}

			$out->print_string($shunFilterList->{$filterListReference});
		}
	}

	if (defined($natControlFilterListReferences) && $natControlFilterListReferences ne '')
	{
		if (!$filterListsSectionOpen)
		{
			$out->open_element('filterListsSection');
			$filterListsSectionOpen = 1;
		}

		$out->print_string($natControlFilterListReferences);
	}

	# Print the security levels acl if if it exists.
	if (defined($securityLevelsAclReferences) && $securityLevelsAclReferences ne '')
	{
		if (!$filterListsSectionOpen)
		{
			$out->open_element('filterListsSection');
			$filterListsSectionOpen = 1;
		}

		$out->print_string($securityLevelsAclReferences);
	}

	# We loop through every ACL and determine if it's applied to the interface as input or output.
	foreach my $acl (@{$acls})
	{
		my $name = $acl->{name};

		# If this ACL isn't applied to an interface with the matching $inOut, then we don't process it here.
		if (defined($access_groups->{$inOut}->{$name}))
		{
			if (!$filterListsSectionOpen)
			{
				$out->open_element('filterListsSection');
				$filterListsSectionOpen = 1;
			}

			# Print out the ACL.
			$out->open_element('filterListReference');
			foreach my $interface (@{$access_groups->{$inOut}->{$name}})
			{
				$out->open_element('appliesTo');
				$out->print_element($appliesToSourceOrDest, $interface);
				$out->close_element('appliesTo');
			}
			$out->print_element('direction', $direction);
			$out->print_element('name', $name);
			$out->close_element('filterListReference');
		}
	}

	if ($filterListsSectionOpen)
	{
		$out->open_element('phaseDiscriminator');
		$out->open_element('processingPhase');
		$out->print_element('core:packetProcessingPhase', 'routing');
		$out->close_element('processingPhase');
		$out->print_element('processingPhasePosition', $beforeOrAfterRouting);
		$out->close_element('phaseDiscriminator');
		$out->close_element('filterListsSection');
	}
}

# This prints out all the ACLs as filterLists, without any appliesTo or direction.  appliesTo and direction
# is provided by the filterListReferences.
sub _print_acls
{
	my ($out, $acls, $natControlFilterLists, $shunFilterLists) = @_;

	my $filterListsSectionOpen = 0;

	if (defined($natControlFilterLists) && $natControlFilterLists ne '')
	{
		$out->open_element('filterListsSection');
		$filterListsSectionOpen = 1;

		$out->print_string($natControlFilterLists);
	}

	# Print any shun acls if they exist.
	foreach my $shunFilterList (@{$shunFilterLists})
	{
		if (defined($shunFilterList->{filterListXml}))
		{
			if (!$filterListsSectionOpen)
			{
				$out->open_element('filterListsSection');
				$filterListsSectionOpen = 1;
			}

			$out->print_string($shunFilterList->{filterListXml});
		}
	}

	$LOGGER->debug(q{_print_acls: There are } . scalar (@{$acls}) . q{ acls to output to xml});
	foreach my $acl (@{$acls})
	{
		my $name = $acl->{name};
		my $acl_size = length($acl->{entries});
		if($acl_size > $MAX_ACL_SIZE){
			$LOGGER->debug("ACL '$name' is too large to print, $acl_size, not printing it");
			next;
		}
		if($acl_size == 0){
			$LOGGER->debug("ACL '$name' has no entries found, not printing it.");
			next; # skip empty ACL
		}

		my $filterEntries = $acl->{entries};

		if (!$filterListsSectionOpen)
		{
			$out->open_element('filterListsSection');
			$filterListsSectionOpen = 1;
		}

		$out->open_element('filterList');
		$out->print_string($filterEntries);
		$out->print_element('mode', 'stateful');
		$out->print_element('name', $name);
		$out->open_element('processingPhase');
		$out->print_element('core:packetProcessingPhase', 'packet-filtering');
		$out->close_element('processingPhase');
		$out->close_element('filterList');
	}
	if($filterListsSectionOpen)
	{
		$out->close_element('filterListsSection');
	}
}

# This parses the acls.  It will return an array of ACLs (in the order they appear in the configuration).
# Each element of the array will be a hash, consisting of two named elements, name and entries.
# Name will be the name of the ACL, entries will be all of the filterEntries in the ACL.
sub _parse_acls
{
	my ($in, $time_range) = @_;
	
	my @acls;

	my $currentAcl;
	my $aclFh;
	my $out;
	my $openedFilterLists = 0;
	my $old_name = '';
	my $name = '';
	my $term_process_order;
	my @acl_array = $in->{running_config} =~ m/(^access-list\s.*)/mig;
	$in->{_acl_terms} = {};
	foreach my $acl_line (@acl_array){
		if ($acl_line =~ /^access-list\s+(\S+)(?:\s+line\s+(\d+))?(\s+extended|\s+standard|\sremark)?\s+(\S.+)$/mi ){
			# Increment the process order if this is a continuation of the last ACL.
			if ( $name eq $1 )
			{
				$term_process_order++;
			}
			else
			{
				$term_process_order = 1;
			}
		
			$name		= $1;
			my $line	= $2 if ( $2 );
			my $format = "extended";
			$format = trim($3) if $3;
			my $details	= trim ( $4 );
			my $thisterm;
		
			# We don't process access-list remarks
			if ($format eq 'remark')
			{
				next;
			}
		
			if ($old_name ne $name)
			{
				if ($old_name ne '')
				{
					close($aclFh);
					push(@acls, {name => $old_name, entries => $currentAcl});
				}
		
				$old_name = $name;
				$currentAcl = '';
				open($aclFh, '>', \$currentAcl);
				$out = ZipTie::Model::XmlPrint->new(\*$aclFh, 'core');
			}
		
			my $action;
			my $protocol;
			my $src_address;
			my $src_service;
			my $dst_address;
			my $dst_service;
			$_ = $details;
			if ( $format eq "extended" )
			{
				unless ($_ =~ /^(permit|deny|no permit)\s+($OG_KWD\s+\S+|\S+)(.+)$/)
				{
					$LOGGER->debug("Cisco SecurityAppliance parsing: Unable to process extended ACE: $_");
					next;
				}
		
				$action = $1;
				$protocol = $2;
				my $remainder = $3;
		
				# The following anonymous subs sets either the source/destination address/service based on what has already been parsed.
				my $setAddress = sub {
					my $value = shift;
		
					# Both a source address and a destination address is required, and only one can exist of each.  Checks if the source
					# address is set, if not sets it, otherwise sets the destination address.
					if (!defined($src_address))
					{
						$src_address = $value;
					}
					else
					{
						$dst_address = $value;
					}
				};
				my $setService = sub {
					my $value = shift;
		
					# Both services are optional.  A source service if defined will always appear before the destination address, and a
					# destination service if defined will always appear after the destination address.  Checks to see if the destination
					# address has been defined, if not sets the source service, otherwise sets the destination service.
					if (!defined($dst_address))
					{
						$src_service = $value;
					}
					else
					{
						$dst_service = $value;
					}
				};
		
				while ($remainder =~ /\s*(?:
							(any\d?)|                                                                                 # any address
							(host\s+\S+)|                                                                             # host address
							($OG_KWD\s+\S+)|                                                                          # object group, could be address or service
							((?:neq|eq|gt|lt)\s+\S+)|                                                                 # operator service
							(range\s+\S+\s+\S+)|									  # range service
							(interface\s+\S+)|                                                                        # interface as the source or destination
							(\S+(?:\s+[\d\.]+)?)                                                                      # IP and mask address
						)/gxi)
				{
					# any address definition
					if (defined($1))
					{
						$setAddress->($1);
					}
					# host address definition
					elsif (defined($2))
					{
						$setAddress->($2);
					}
					# object-group decleration, can be address or service
					elsif (defined($3))
					{
						my $ogDecleration = $3;
						$ogDecleration =~ /$OG_KWD\s+(\S+)/;
						my $ogId = $1;
		
						if (defined($in->{object_groups}->{$ogId}))
						{
							my $type = $in->{object_groups}->{$ogId}->{type};
		
							if ($type eq 'network')
							{
								$setAddress->($ogDecleration);
							}
							else
							{
								$setService->($ogDecleration);
							}
						}
						else
						{
							$LOGGER->debug("_parse_acls: Object group not defined: $ogId, dropping rule");
		
							# If action is undef then this rule will not be processed.
							$action = undef;
						}
					}
					# Operator service decleration
					elsif (defined($4))
					{
						$setService->($4);
					}
					# Range service decleration
					elsif (defined($5))
					{
						$setService->($5);
					}
					# Interface source or destination decleration
					elsif (defined($6))
					{
						$setAddress->($6);
					}
					# IP/mask address decleration
					elsif (defined($7) && !defined($dst_address))
					{
						$setAddress->($7);
					}
				}
			}
			elsif (/                                     # this regex parses the important part of the acl
				^(permit|deny|no permit)                 # grab the action
				(\s+(?:ip|tcp|udp|icmp|gre))?            # grab the protocol
				\s+(any\d?|host\s+\S+|\S+(?:\s+[\d\.]+)?)   # grab the source address
				/ix	)
			{
				$action		 = $1;
				$protocol	 = 'ip';
				$protocol	 = trim($2) if( $2 );
				$dst_address = $3;
			}
		
			# This documentation is a good description of how the protocol argument is handled.
			# http://www.cisco.com/en/US/docs/security/asa/asa84/command/reference/a1.html#wp1598407
			if ( $protocol =~ /$OG_KWD\s+(\S+)/i )
			{
				my $ogGroupId = $1;
		
				# The object group should either be a protocl object group, or a sevice object group that has been
				# broken into source and destination components.
				if (defined($in->{object_groups}->{$ogGroupId}) && $in->{object_groups}->{$ogGroupId}->{type} eq 'protocol')
				{
					$thisterm->{"protocolObjectGroupReference"} = $ogGroupId;
				}
				elsif (defined($in->{splitServiceGroups}->{$ogGroupId}))
				{
					# Protocol_argument is actually a service group that has been broken into source and destination componentns,
					# add the components as source and destination services.
					if (defined($in->{splitServiceGroups}->{$ogGroupId}->{source}))
					{
						push @{$thisterm->{sourceService}},
							{objectGroupReference => $in->{splitServiceGroups}->{$ogGroupId}->{source}}
					}
					if (defined($in->{splitServiceGroups}->{$ogGroupId}->{destination}))
					{
						push @{$thisterm->{destinationService}},
							{objectGroupReference => $in->{splitServiceGroups}->{$ogGroupId}->{destination}};
					}
				}
				elsif (defined($in->{object_groups}->{$ogGroupId}) && $in->{object_groups}->{$ogGroupId}->{type} eq 'service')
				{
					# I don't expect this to ever happen, but if it does print debug log and then proceed assuming
					# it's a destination service.
					$LOGGER->debug('Cisco SecurityAppliance parsing: ACL parsing, service group provided as protocol_argument ' .
							"not broken into source/destination components: $ogGroupId");
					push @{$thisterm->{destinationService}}, {objectGroupReference => $ogGroupId};
				}
				else
				{
					$LOGGER->debug("Cisco SecurityAppliance parsing: ACL parsing, object group not defined: $ogGroupId, dropping rule.");
		
					# Setting action to undef will cause this rule to be dropped.
					$action = undef;
				}
			}
			else
			{
				$thisterm->{"protocol"} = _name_protocol($protocol);
			}
		
			if( $action )
			{
				$action = "deny" if ( $action eq 'no permit' );
				$thisterm->{"primaryAction"} = $action;
				
				my $printFilterEntry = 1;
				if ($src_address) {
					$printFilterEntry = print_filter_address ( trim ( $src_address ), $thisterm, "sourceIpAddr", $in->{name_mapper} );
				}

				if ( $src_service ) {
					print_filter_service ( trim ( $src_service ), $thisterm, "sourceService", $protocol );
				}

				if( $dst_address && $printFilterEntry == 1) {
					$printFilterEntry = print_filter_address ( trim ( $dst_address ), $thisterm, "destinationIpAddr", $in->{name_mapper} );
				}

				if ( $dst_service ) {
					print_filter_service ( trim ( $dst_service ), $thisterm, "destinationService", $protocol );
				}

				# look for the log instruction
				if ( /\slog(?:\s|$)/i )
				{
					$thisterm->{log} = "true";
				}
				else
				{
					$thisterm->{log} = "false";
				}
		
				# look for the time-range instruction
				if ( /\btime-range\s+(\S+)/i )
				{
					my $time_range_name = $1;
					if($time_range->{$time_range_name}){
						$thisterm->{timeAllowed} = $time_range->{$time_range_name};
					}
				}
		
				# store process order ( line number )
				$thisterm->{processOrder} = $term_process_order;
		
				if (!defined($in->{_acl_terms}->{$name}))
				{
					$in->{_acl_terms}->{$name} = [];
				}
		
				push(@{$in->{_acl_terms}->{$name}}, $thisterm);
		
				if ( defined ($printFilterEntry) && $printFilterEntry == 1) {
					$out->print_element( "filterEntry", $thisterm );
				}
			}
		}
	}
	if ($name ne '')
	{
		close($aclFh);
		push(@acls, {name => $name, entries => $currentAcl});
	}
	
	return \@acls;
}

# Parses the output of the the 'show shun' command stored in $in->{show_shun}.
# Returns an array of hash references.  The keys of each hash are:
# 	* filterListXml - A string containing the filter list XML.
# 	* filterListReference_in - A string containing the input filter list reference XML.
#	* filterListReference_out - A string containing the output filter list reference XML.
sub _parse_shun
{
	my ($in) = @_;

	if( not $in->{show_shun} ) { return; }

	my %shunFilterEntries;

	# Example of a shun command:
	# shun (inside) 39.84.124.22 192.168.254.254 56323 32423 17
	while ($in->{show_shun} && $in->{show_shun} =~ /^shun \((\S+)\) (\S+)/mg)
	{
		my $interface = $1;
		my $srcIp = $in->{name_mapper}->get_ip($2);

		# Use a seperate XmlPrint object for each interface.
		if (!defined($shunFilterEntries{$interface}))
		{
			$shunFilterEntries{$interface}->{filterListXml} = '';
			open($shunFilterEntries{$interface}->{fh}, '>', \$shunFilterEntries{$interface}->{filterListXml});
			$shunFilterEntries{$interface}->{out} =  ZipTie::Model::XmlPrint->new($shunFilterEntries{$interface}->{fh}, 'core');
			$shunFilterEntries{$interface}->{processOrder} = 1;

			$shunFilterEntries{$interface}->{out}->open_element('filterList');
			$shunFilterEntries{$interface}->{out}->print_element('defaultAction', 'none');
		}

		# Generate source deny rule
		my $sourceFilterEntry = {
				'log' => 'true',
				primaryAction => 'drop',
				processOrder => $shunFilterEntries{$interface}->{processOrder}++,
				sourceIpAddr => {
						host => $srcIp
					}
			};

		# Generate destination deny rule
		my $destinationFilterEntry= {
				destinationIpAddr => {
						host => $srcIp
					},
				'log' => 'true',
				primaryAction => 'drop',
				processOrder => $shunFilterEntries{$interface}->{processOrder}++
			};

		$shunFilterEntries{$interface}->{out}->print_element('filterEntry', $sourceFilterEntry);
		$shunFilterEntries{$interface}->{out}->print_element('filterEntry', $destinationFilterEntry);
	}

	foreach my $interface (keys(%shunFilterEntries))
	{
		# The name of the shun filter list for this interface.
		my $filterListName = "shun_$interface";

		$shunFilterEntries{$interface}->{out}->print_element('mode', 'stateful');
		$shunFilterEntries{$interface}->{out}->print_element('name', $filterListName);
		$shunFilterEntries{$interface}->{out}->open_element('processingPhase');
		$shunFilterEntries{$interface}->{out}->print_element('core:packetProcessingPhase', 'packet-filtering');
		$shunFilterEntries{$interface}->{out}->close_element('processingPhase');
		$shunFilterEntries{$interface}->{out}->close_element('filterList');

		close($shunFilterEntries{$interface}->{fh});

		# Remove any data that is no longer required.
		delete($shunFilterEntries{$interface}->{fh});
		delete($shunFilterEntries{$interface}->{out});
		delete($shunFilterEntries{$interface}->{processOrder});

		# Create the filter list references
		foreach my $direction ('ingress', 'egress')
		{
			my $filterListReference = 'filterListReference_in';
			my $sourceDestination = 'source';
			if ($direction eq 'egress')
			{
				$filterListReference = 'filterListReference_out';
				$sourceDestination = 'destination';
			}

			$shunFilterEntries{$interface}->{$filterListReference} = '';
			open(my $fh, '>', \$shunFilterEntries{$interface}->{$filterListReference});
			my $out = ZipTie::Model::XmlPrint->new($fh, 'core');
			$out->print_element('filterListReference', {
					appliesTo => {
							$sourceDestination => $interface
						},
					direction => $direction,
					name => $filterListName
				});
			close($fh);
		}
	}

	# Return the values as an array.
	my @returnArray = values(%shunFilterEntries);
	return \@returnArray;
}

sub _generate_security_levels_acl_references
{
	my ($in, $access_groups, $security_levels) = @_;

	my $sameSecurityLevelPermit = 0;
	my $intraInterfacePermit = 0;

	if ($in->{running_config} =~ /^same-security-traffic permit inter-interface/m)
	{
		$sameSecurityLevelPermit = 1;
	}

	if ($in->{running_config} =~ /^same-security-traffic permit intra-interface/m)
	{
		$intraInterfacePermit = 1;
	}

	# Create the filterListReference for the security levels ACL.
	my $securityLevelsAclReferences = '';
	open(my $securityLevelsAclFh, '>', \$securityLevelsAclReferences);
	my $securityLevelsAclPrinter = ZipTie::Model::XmlPrint->new(\*$securityLevelsAclFh, 'core');

	$securityLevelsAclPrinter->open_element('filterListReference');

	my $atLeastOneAppliesTo = 0;

	foreach my $sourceInterface (keys(%{$security_levels}))
	{
		# By default trafic entering and exiting on the same interface is denied.
		# http://www.cisco.com/en/US/docs/security/asa/asa84/command/reference/s1.html#wp1444448
		if (!$intraInterfacePermit)
		{
			$securityLevelsAclPrinter->open_element('appliesTo');
			$securityLevelsAclPrinter->print_element('source', $sourceInterface);
			$securityLevelsAclPrinter->print_element('destination', $sourceInterface);
			$securityLevelsAclPrinter->close_element('appliesTo');
			$atLeastOneAppliesTo = 1;
		}

		# If the source interface alreday has an input ACL then we don't need to block traffic here.
		if (defined($access_groups->{interfaces_in}->{$sourceInterface}))
		{
			next;
		}

		my $sourceSecurityLevel = $security_levels->{$sourceInterface};

		foreach my $destInterface (keys(%{$security_levels}))
		{
			# The interfaces are equal.  We already handled same security level permit above.
			if ($sourceInterface eq $destInterface)
			{
				next;
			}

			# If the destination interface alreday has an output ACL then we don't need to block traffic here.
			if (defined($access_groups->{interfaces_out}->{$destInterface}))
			{
				next;
			}

			my $destSecurityLevel = $security_levels->{$destInterface};

			if (($sourceSecurityLevel < $destSecurityLevel) ||
					(!$sameSecurityLevelPermit && ($sourceSecurityLevel == $destSecurityLevel)))
			{
				$securityLevelsAclPrinter->open_element('appliesTo');
				$securityLevelsAclPrinter->print_element('source', $sourceInterface);
				$securityLevelsAclPrinter->print_element('destination', $destInterface);
				$securityLevelsAclPrinter->close_element('appliesTo');
				$atLeastOneAppliesTo = 1;
			}
		}
	}

	$securityLevelsAclPrinter->print_element('direction', 'egress');
	$securityLevelsAclPrinter->print_element('name', 'Security_Levels_Deny');
	$securityLevelsAclPrinter->close_element('filterListReference');

	close($securityLevelsAclFh);

	if (!$atLeastOneAppliesTo)
	{
		$securityLevelsAclReferences = '';
	}

	return $securityLevelsAclReferences;
}

=comment out ipv6 ACL parsing - NOT supported in QRM
sub parse_ipv6_filters
{
	my ( $in, $out, $openedFilterLists, $time_range ) = @_;
	my $term_process_order = 1;
	my $name		= " ";
	my $old_name	= " ";

	while ( $in->{running_config} =~ /^ipv6\s+access-list\s+(\S+)(?:\s+line\s+(\d+))?\s+(\S.+)$/mig )
	{
		if ( $name eq $1 )
		{
			$term_process_order ++;
		}
		else
		{
			$term_process_order = 1;
		}
		$name		= $1;
		my $line	= $2 if ( $2 );
		my $details	= trim ( $3 );
		my $thisterm;

		$_ = $details;
		if (	# this regex parses the important part of the acl
			/
			^(permit|deny|no permit)							# grab the action
			\s+(\S+|$OG_KWD\s+\S+)								# grab the protocol
			\s+(any\d?|host\s+\S+|$OG_KWD\s+\S+|[\da-f\.\:\/]+(?:\s+[\da-f\.\:\/]+)?)		# grab the source address
			(\s+(?:neq|eq|gt|lt)\s+\S+|\s+range\s+\S+\s+\S+|\s+$OG_KWD\s+\S+)?	# grab the source service
			\s+(any\d?|host\s+\S+|$OG_KWD\s+\S+|[\da-f\.\:\/]+(?:\s+[\da-f\.\:\/]+)?)		# grab the destination address
			(\s+(?:neq|eq|gt|lt)\s+\S+|\s+range\s+\S+\s+\S+|\s+$OG_KWD\s+\S+)?	# grab the destination service
			/ix
		)
		{
			my $action	= $1;
			my $protocol	= $2;
			my $src_address = $3;
			my $src_service = $4 if ( $4 );
			my $dst_address = $5;
			my $dst_service	= $6 if ( $6 );

			$action = "deny" if ( $action eq 'no permit' );
			$thisterm->{"primaryAction"} = $action;

			if ( $protocol =~ /$OG_KWD\s+(\S+)/i )
			{
				$thisterm->{"protocolObjectGroupReference"} = $1;
			}
			else
			{
				$thisterm->{"protocol"} = $protocol;
			}

			print_filter_address ( trim ( $src_address ), $thisterm, "sourceIpAddr", 6, $in->{name_mapper} );

			print_filter_service ( trim ( $src_service ), $thisterm, "sourceService", $protocol) if ( $src_service );

			print_filter_address ( trim ( $dst_address ), $thisterm, "destinationIpAddr", 6, $in->{name_mapper} );

			print_filter_service ( trim ( $dst_service ), $thisterm, "destinationService", $protocol ) if ( $dst_service );

			# look for the log instruction
			if ( /\slog(?:\s|$)/i )
			{
				$thisterm->{log} = "true";
			}
			else
			{
				$thisterm->{log} = "false";
			}

			# look for the time-range instruction
			if ( /\btime-range\s+(\S+)/i )
			{
				my $time_range_name = $1;
				if($time_range->{$time_range_name}){
					$thisterm->{timeAllowed} = $time_range->{$time_range_name};
				}
			}

			# store process order ( line number )
			$thisterm->{processOrder} = $term_process_order;

			$out->open_element("filterLists") if ( !$openedFilterLists );
			$openedFilterLists	= 1;

			if ( $old_name ne $name && $old_name ne " ")
			{
				$out->print_element( "mode", 'stateful');
				$out->print_element( "name", $old_name );
				$out->close_element("filterList");
			}
			if ( $old_name ne $name)
			{
				$old_name = $name;
				$out->open_element("filterList");
			}
			$out->print_element( "filterEntry", $thisterm );
		}
	}
	if ( $old_name ne " ")
	{
		$out->print_element( "mode", 'stateful');
		$out->print_element( "name", $old_name );
		$out->close_element("filterList");
	}
}
=cut

sub _parse_firewall_modes {
	my ($in) = @_;

	my @contexts = @{ $in->{userContexts} }; 
	for my $context ( @contexts ) {

		my $name = $context->{name};
		if($in->{show_firewall} =~ /^$name\s+(\S+)\s*$/m ) {
			my $mode = $1;
			$LOGGER->debug("Firewall mode for context '$name' is '$mode'");

			my $transparent = 0;
			if( $mode eq 'Transparent' ) {
				$transparent = 1;
			} elsif( $mode ne 'Router') {
				$LOGGER->debug("Firewall mode '$mode' for context '$name' is not recognised");
			}
			$context->{firewallTransparent} = $transparent;

		# Handle versions with per context firewall modes by setting to global firewall mode
		} else {
			$context->{firewallTransparent} = $in->{firewallTransparent};
		}
	}

	return;
}

sub parse_time_range
{
	my ( $in ) = @_;

	my $months = 
	{
		JAN => '01',
		FEB => '02',
		MAR => '03',
		APR => '04',
		MAY => '05',
		JUN => '06',
		JUL => '07',
		AUG => '08',
		SEP => '09',
		OCT => '10',
		NOV => '11',
		DEC => '12',
	};
	my $time_range;
	while ( $in->{running_config} =~ /^time-range\s+(\S+)(.+?)(?:!|: end)/migs )
	{
		my $name	= $1;
		$_		= $2;

		if ( /absolute\s+start\s+(\d+:\d+)\s+(\d+)\s+(\S+)\s+(\d+)\s+end\s+(\d+:\d+)\s+(\d+)\s+(\S+)\s+(\d+)/mi )
		{
			$time_range->{$name}->{startTime}	= $1.':00';
			$time_range->{$name}->{startDate}	= $4.'-'.$months->{uc( substr ($3,0,3) )}.'-'.$2;
			$time_range->{$name}->{endTime}		= $5.':00';
			$time_range->{$name}->{endDate}		= $8.'-'.$months->{uc( substr ($7,0,3) )}.'-'.$6;
			$time_range->{$name}->{startTime}	= '0'.$time_range->{$name}->{startTime} if ( $time_range->{$name}->{startTime} =~ /^\d:/ );
			$time_range->{$name}->{endTime}		= '0'.$time_range->{$name}->{endTime} if ( $time_range->{$name}->{endTime} =~ /^\d:/ );
			$time_range->{$name}->{days} = "All";
		}
		elsif ( /periodic\s+(\D+)(\d+:\d+)\s+to\s+(\d+:\d+)/mi )
		{
			my $days				= $1;
			$time_range->{$name}->{startTime}	= $2.':00';
			$time_range->{$name}->{endTime}		= $3.':00';
			$time_range->{$name}->{startTime}	= '0'.$time_range->{$name}->{startTime} if ( $time_range->{$name}->{startTime} =~ /^\d:/ );
			$time_range->{$name}->{endTime}		= '0'.$time_range->{$name}->{endTime} if ( $time_range->{$name}->{endTime} =~ /^\d:/ );
			$days = "Mon Tue Wed Thursday Fri" if $days =~ /weekdays/;
			$days = "Sat Sun" if $days =~ /weekend/;
			$days = "All" if $days =~ /daily/;
			$days =~ s{ (\w+) }{ ( lc ( $1 ) ne 'thursday' ? substr ( $1, 0, 3 ) : substr ( $1, 0, 4 ) ) }gex;
			push @{$time_range->{$name}->{days}}, split ( /\s+/, $days );
		}
	}

	return $time_range;
}

#
#  print_filter_address()
#
#  This routine, given an input to scan and and output SED hash element context, will create within
#  that output the addressing information associated with the input configuration fragment.
#
#   Specifically, this is intended to fill in the subtree of a FilterIpAddress complex type instance.
#   This therefore deals with things like whether an address is singular or a range, whether it is
#   and network group reference instead or a subnet.
#
#    PARAMS:
#      $in - input context
#      $out - output context to create a common:FilterIpAddress type instance within
#	   $parentElement - parent element over which to iterate.
#      $v = version indicator; valid values 4 or 6.  If 6, then addresses interpreted/printed using ipv6
#              syntax/subnet semantics.
#      $nameMapper - Name mapper object, used for mapping names to IP addresses.
#
#    RETURNS:
#       none
#
sub print_filter_address
{
	my ( $in, $out, $parentElement, $nameMapper ) = @_;
	#$returnVal of 1 means a successful run, 0 something went wrong
	my $returnVal = 1;

	my $any_network = { "address" => "0.0.0.0", "mask" => "0" };
	if ( $in =~ /^any(4)?$/i )
	{
		push @{ $out->{$parentElement} }, { network => $any_network};
	}
	elsif ( $in =~ /^$cidr$/i )
	{
		my ( $address, $mask ) = split ( /\//, $in );
		push(@{$out->{$parentElement}}, {network => _create_subnet($address, $mask)});
		if ( $mask eq mask_to_bits( $mask ) ) {
			$LOGGER->debug(qq{print_filter_address: mask is non-contiguous or invalid: $address $mask, dropping rule});
			$returnVal = 0;
		}
	}
	elsif ( $in =~ /^host\s+(\S+)$/i )
	{
		
		my $ip = $nameMapper->get_ip($1);
		push @{ $out->{$parentElement} }, {host => $ip};
		unless ( $ip =~ $cipm ) {
			$LOGGER->debug(qq{print_filter_address: host is a non-IPv4 address type(s): $ip, dropping rule"});
			$returnVal = 0;
		}
	}
	elsif ( $in =~ /^$OG_KWD\s+(\S+)$/i )
	{
		push @{ $out->{$parentElement}} , {objectGroupReference => $1};
	}
	elsif ( $in =~ /^(\S+)\s+($cipm)$/i )
	{
		my $ip = $nameMapper->get_ip($1);
		push(@{$out->{$parentElement}}, {network => _create_subnet($ip, $2)});
		if ( $2 eq mask_to_bits( $2 ) )	{
			$LOGGER->debug(qq{print_filter_address: mask is non-contiguous or invalid: $ip $2, dropping rule});	
			$returnVal = 0;
		}
	}
	else{
		$LOGGER->debug(qq{print_filter_address: address $in could not be parsed, dropping rule});	
		$returnVal = 0;
	}
	return $returnVal;
}

#
#  print_filter_service
#
#  This prints a service in the context of a  filter.  
#
#  So, it turns out that there are two different elements here--one for the object group-based definition
#  of a service and its port/etc. resources (common:ServiceGroup complex type in the Ziptie schema), the 
#  other being the definition of a service as a filtering element (common:Service -- this has usage 
#  elsewhere in the ZED schema, but the one we care about is the filter).  This routine deals with the latter.
#  SHOULD a service (by name) portend a mapping to multiple ports, the two types force different types to 
#  iterate -- the  the Service requires multiple containing "sourceService" or "destinationService" elements
#  each allowing a single portExpression/portRange element.
#  Then there is the other possibility (objectGroupReference subelement), which thankfully we
#  don't have to worry about multiples of, in this config regime.
#
#  Since the named service lookup in _svc_int_port is a tad complex because of this iteration, then hide
#  the details of how to lay down that iteration in this routine.
#
#   PARAMS:
#     $in - input context for configuration
#     $out - output context for ZED element generation
#     $parentElement - an element on which to iterate each instance of service mapping found
#             (see above)
#     $filterProtocol - string - provides ability to throttle on just matching entries for tcp, udp
#
#    RETURNS:
#        nothing
#
sub print_filter_service
{
	my ( $in, $out, $parentElement, $filterProtocol ) = @_;

	if ( $in =~ /^(lt|gt|eq|neq)\s+(\S+)$/i )
	{
		my $svcPortSet = _svc_int_port($2);
		
		if (defined $svcPortSet)
		{
			foreach my $portMapping (@{$svcPortSet})
			{
				my $portExp = { "port" => $portMapping->{port}, "operator" => ( $1 ne 'neq' ? $1 : 'ne' )};
				# See if we can suppress a term based on prevailing protocol
				if (defined $portMapping->{protocol})
				{
					if (($filterProtocol =~ /^[ut][dc]p/i) and
						($filterProtocol ne lc($portMapping->{protocol})))
					{
						next;  # it's for a specific IP bearer protocol, just not the one for this filter					
					}
					# but we know it's defined
					$portExp->{protocol} = $portMapping->{protocol}; 
				} 
										
				push ( @{$out->{$parentElement}}, { portExpression => $portExp});
			}
			
		}		
		# otherwise push nothing, since we're dealing with undefined content		
	}
	elsif ( $in =~ /^range\s+(\S+)\s+(\S+)$/i )
	{
		# using service names in ranges is a bit more exotic, but syntactically allowed.  	We'll take the 
		# first mapping here, otherwise this makes no sense.
		my $nPStart = $1; my $nPEnd = $2; 
		my $svcPortSet = _svc_int_port($nPStart);

		if (defined $svcPortSet)
		{
			my $i = 0;

			if ($filterProtocol =~ /^[ut][dc]p/i)
			{
				# See if we can suppress a term based on prevailing protocol

				for ($i = 0; $i < scalar( @{$svcPortSet}); $i++)
				{
					if ((defined $svcPortSet->[$i]->{protocol}) and
						($svcPortSet->[$i]->{protocol} eq lc($filterProtocol)))
					{
						last;		
					}					
				}
				if ($i == scalar( @{$svcPortSet}))
				{
					# back it down to zero I guess to have something to generate
					$i = 0;					
				}
			}
						
			my $rPortExp = 	{ "portStart" => $svcPortSet->[$i]->{port}};		
			$rPortExp->{protocol} = $svcPortSet->[$i]->{protocol} if (defined $svcPortSet->[$i]->{protocol});		

			$svcPortSet = _svc_int_port($nPEnd);
			if (defined $svcPortSet->[0]->{port})
			{
				$rPortExp->{portEnd} = $svcPortSet->[0]->{port};	
				push ( @{$out->{$parentElement}}, { portRange => $rPortExp} );			
			}		
		}
	}
	elsif ( $in =~ /^$OG_KWD\s+(\S+)$/i )
	{
		push ( @{$out->{$parentElement}}, { objectGroupReference => $1});
	}
}

sub create_config
{
	my ( $in, $out ) = @_;

	# the name of the repository
	my $repository;
	$repository->{'core:name'} = '/';
	my $limit_length = 10000000;

	# build the simple text configuration
	my $running;
	my $runningLength = length( $in->{"running_config"} );
	$LOGGER->debug(qq{create_config: show running_config length: $runningLength });
	$running->{"core:name"}       = "running-config";
	if ($runningLength < $limit_length) {
		$running->{"core:textBlob"}  = encode_base64( $in->{"running_config"} );
	} else {
		$LOGGER->debug(qq{create_config: show running_config length too large: $runningLength Greater than $limit_length });
		$running->{"core:textBlob"}  = encode_base64( qq{show running_config, too large to show} );
	}
	$running->{"core:mediaType"}  = "text/plain";
	$running->{"core:context"}    = "active";
	$running->{'core:promotable'} = 'false';

	# push the configuration into the repository
	push( @{ $repository->{'core:config'} }, $running );

	my $startup;
	my $startupLength = length( $in->{"startup_config"} );
	$LOGGER->debug(qq{create_config: show startup_config length: $startupLength });
	$startup->{"core:name"}      = "startup-config";
	if ($startupLength < $limit_length) {
		$startup->{"core:textBlob"}  = encode_base64( $in->{"startup_config"} );
	} else {
		$LOGGER->debug(qq{create_config: show startup_config length too large: $startupLength  Greater than $limit_length});
		$startup->{"core:textBlob"}  = encode_base64( qq{show startup_config, too large to show} );
	}
	$startup->{"core:mediaType"} = "text/plain";
	$startup->{"core:context"}   = "boot";
	
	if ($in->{version} =~ /(Adaptive\s+Security\s+Appliance|ASA)/)
	{
		$startup->{'core:promotable'} = 'true';
	}
	else
	{
		$startup->{'core:promotable'} = 'false';
	}

	# push the configuration into the repository
	push( @{ $repository->{'core:config'} }, $startup );

	
	if (   ( defined $in->{"na_running_config"} )
		|| ( defined $in->{"na_startup_config"} ) )
	{
		my $cFolder;
        $cFolder->{'core:name'} = 'system space';
        
		my $na_running;
		$na_running->{"core:name"}       = "running-config";
		$na_running->{"core:textBlob"}   = encode_base64( $in->{"na_running_config"} );
		$na_running->{"core:mediaType"}  = "text/plain";
		$na_running->{"core:context"}    = "active";
		$na_running->{'core:promotable'} = 'false';

		# push the configuration into the repository
		push (@{$cFolder->{'core:config'}}, $na_running);

		my $na_startup;
		$na_startup->{"core:name"}       = "startup-config";
		$na_startup->{"core:textBlob"}   = encode_base64( $in->{"na_startup_config"} );
		$na_startup->{"core:mediaType"}  = "text/plain";
		$na_startup->{"core:context"}    = "boot";
		$na_startup->{'core:promotable'} = 'true';

		# push the configuration into the repository
		push( @{ $cFolder->{'core:config'} }, $na_startup );
		push( @{ $repository->{'core:folder'} }, $cFolder );
	}

	# Extract and insert per-context names
	for my $ucName (keys(%{$in->{"context_configs"}}))
	{
		my $rc_context;
		my $rc_folder;
        $rc_folder->{'core:name'} = $ucName;
		
		$rc_context->{"core:name"} =  "running-config";
		$rc_context->{"core:textBlob"}   = encode_base64( $in->{"context_configs"}->{$ucName});
		$rc_context->{"core:mediaType"}  = "text/plain";
		$rc_context->{"core:context"}    = "active";
		$rc_context->{'core:promotable'} = 'false';
		
		# now, push the config		
		push( @{ $rc_folder->{'core:config'} }, $rc_context );
		push( @{ $repository->{'core:folder'} }, $rc_folder );
	}
		
	# print the repository
	$out->print_element( 'core:configRepository', $repository );
}

sub create_nadmin_config
{
	my ( $in, $out ) = @_;

	#$modelNumber
}

sub parse_routing
{
	my ( $in, $out ) = @_;
}

sub parse_local_accounts
{

	# local accounts - local usernames
	my ( $in, $out ) = @_;
	my $c = $in->{running_config};
	$out->open_element("localAccounts");

	while ( $c =~ /^username\s+(\S+)\s+(.+)/mig )
	{
		my $account = { accountName => $1, };
		my $userconfig = $2;
		if ( $userconfig =~ /privilege\s+(\S+)/ )
		{
			$account->{accessLevel} = $1;
		}
		if ( $userconfig =~ /password\s+(\S+)\s+(\S+)/ )
		{
			$account->{password} = $1;
		}
		$out->print_element( "localAccount", $account );
	}

	# local accounts - enable and enable secret
	while ( $c =~ /^enable\s+(\S+)\s+(\S+)\s+(level\s+(\d+))?/mig )
	{
		my $enabletype = $1;
		my $username   = "enable";
		my $encryption = $4;
		if ( ! defined $encryption || $encryption eq "" )
		{
			$encryption = "0";
		}

		if ( $enabletype eq "secret" ) { $username = "enablesecret"; }
		my $account = {
			accountName => $username,
			password    => $2,
			accessLevel => 15,
		};
		$out->print_element( "localAccount", $account );
	}
	$out->close_element("localAccounts");

}

sub parse_snmp
{
	my ( $in, $out ) = @_;
	my $c = $in->{running_config};
	$out->open_element("snmp");

	while ( $c =~ /^(snmp-server\s+community\s+([^\s]+).*)$/mig )
	{
		my $line      = $1;
		my $community = {
			communityString => $2,
			accessType      => 'RO',
		};
		if ( $line =~ /\s+(RO|RW)/i )
		{
			$community->{accessType} = uc($1);
		}
		if ( $line =~ /\s+view\s+([^\s]+)/ )
		{
			$community->{mibView} = $1;
		}
		if ( $line =~ /\s+(RO|RW)\s+(\d+)/i )
		{
			$community->{filter} = $2;
		}
		$out->print_element( "community", $community );
	}

	if ( $c =~ /^snmp-server\s+contact\s+(.*[^\s])\s*$/mi )
	{
		$out->print_element( "sysContact", $1 );
	}
	if ( $c =~ /^snmp-server\s+location\s+(.*[^\s])\s*$/mi )
	{
		$out->print_element( "sysLocation", $1 );
	}
	if ( $c =~ /^hostname\s+(\S+)/mi )
	{
		$out->print_element( "sysName", $1 );
	}
	if ( $c =~ /^snmp-server\s+system-shutdown\s*$/mi )
	{
		$out->print_element( "systemShutdownViaSNMP", "true" );
	}

	while ( $c =~ /^(snmp-server\s+host\s+\"?(\d+\.\d+\.\d+\.\d+)\"?.*$)/mig )
	{
		my $trapHost = { ipAddress => $2, };
		my $line     = $1;

		if ( $line =~ /\"?\d+\.\d+\.\d+\.\d+\"?\s+([^\s]+.*)$/ )
		{
			my $extraInformation = $1;
			my $communityString;

			if ( $extraInformation =~ /^trap\s+version\s+\S+\s+(\S+)/i )
			{
				$communityString = $1;
			}
			elsif ( $extraInformation =~ /^trap\s+(\S+)/i )
			{
				$communityString = $1;
			}
			elsif ( $extraInformation =~ /^inform\s+version\s+\S+\s+(\S+)/i )
			{
				$communityString = $1;
			}
			elsif ( $extraInformation =~ /^inform\s+(\S+)/i )
			{
				$communityString = $1;
			}
			elsif ( $extraInformation =~ /^(\S+)/ )
			{
				$communityString = $1;
			}

			if ($communityString)
			{
				$trapHost->{communityString} = $communityString;
			}
		}
		$out->print_element( "trapHosts", $trapHost );
	}
	if ( $c =~ /^snmp-server\s+trap-source\s+(.*[^\s])\s*$/mi )
	{
		$out->print_element( "trapSource", $1 );
	}
	if ( $c =~ /^snmp-server\s+trap-timeout\s+(\d+)/mi )
	{
		$out->print_element( "trapTimeout", $1 );
	}
	$out->close_element("snmp");
}

sub parse_interfaces
{
	my ( $in, $out ) = @_;

	# If $out is not defined, then we don't want to print the XML.  However, since the configuration will be parsed,
	# generate the XML and save it for future use.
	# This function may be called to generate the $in->{_interface_ips} hash before actually being called to print
	# something to the document.
	my $outFh = undef;
	if (!defined($out))
	{
		# The XML has already been generated, but we have no place to print it to, just return.
		if (defined($in->{_interface_xml}))
		{
			return;
		}

		open($outFh, '>', \$in->{_interface_xml});
		$out = ZipTie::Model::XmlPrint->new(\*$outFh, 'core');
	}
	elsif (defined($in->{_interface_xml}))
	{
		$out->print_string($in->{_interface_xml});
		delete $in->{_interface_xml}; # we should never print this twice
		return;
	}

	# Determine if any bridge groups are defined.
	my $bridgeGroupsDefined = 0;
	if ($in->{running_config} =~ /^\s*bridge-group/m)
	{
		$bridgeGroupsDefined = 1;
	}

	# Parse all of the vlan information for all interfaces.
	my $portVlans = _parse_vlan_information($in);

	my $order = 0;
	my @interfaces;

	# parse_dynamic_routes should always be called before parse_interfaces so the following shold not be needed.
	# However, some of our older unit tests dosen't call parse_dynamic_routes at all, so this has been added for
	# compatibility.
	if (!defined($in->{_remoteGatewayResolution}))
	{
		$in->{_remoteGatewayResolution} = ZipTie::RemoteGatewayResolution->new();
	}

	# Because of the specific stuff we're after, turns out $in->{interfaces}
	# has some interesting parser challenges.  Namely, we start with the
	# parse cursor on "interface ...", and want to parse, over new lines, to
	# the first line without white space before start of line.
	#
	# There's a good regexp here...parse up to the first newline then nonwhitespace,
	# but don't make that newline and nonwhitespace part of the pattern match.
	# That's what (?=I_dont_want_to_match_this) does, effectively "lookahead".
	# But we want to anchor on something, and if we're matching over newlines,
	# after the first iteration, '^' won't match anything really.  So what we 
	# want to have as a "start of line", is that newline before "interface".
	# Since our starting buffer doesn't have that...let's provide it by hack-
	# prepending it.
	#
	my $interfaceDataBuffer = "\n";
	$interfaceDataBuffer .= $in->{interfaces} if $in->{interfaces};
	# This addresses a multitude of sins from the ASA terminal driver, who may choose to
	# represent a trailing prompt with characters that will be tough to latch onto.  The idea
	# here is: find the last real EOL sequence in the whole buffer, and regarless of what follows
	# it (it could simply be the device prompt), put a real character which will match on /S after
	# it.  We don't actually care about that trailing prompt, except as something crucial to latch
	# onto with a pattern.  The following is intentionally made "greedy" as a search.
	$interfaceDataBuffer =~ s/(.+\n)(.+)/$1X$2/sg;
	
	while ( $interfaceDataBuffer =~ /\n(interface.+?)(?=\n[^\s#])/sig )
	{
		my $iline = $1;
		my $ifName = "";

		my $interface = {};
		my $interfaceEthernet = {};

		my $bridgeGroupMember;

		if ( $iline =~ /^interface\s+([\S]+)(?:\s+\"([^\"]*)\")?(\s+|\W+)/i )
		{						
			# as we get here, we are *always* starting the creation of a new interface definition
			$ifName = $1;
			
			# see if there's a "nameif" in the configuration
			if ($2)
			{
				$interface->{name} = $2;
				$interface->{description} = "$2 ($ifName)" ;
			}
			else
			{
				$interface->{name} = $ifName; # if not, just use the port name here				
			}
			
			$interface->{physical}      = _is_physical($ifName);
			$interface->{interfaceType} = _get_interface_type($ifName, $iline, $in, $interface->{name});

			# For physical interfaces we need to determine 802.1Q trunking information.
			if ($interface->{physical} eq 'true')
			{
				foreach my $port (keys(%{$portVlans}))
				{
					# This will determine if $port is a subinterface of $ifName, if it is we need to record
					# the vlan of $port as an allowed vlan on $ifName as described in the following Cisco
					# documentation:
					#
					# http://www.cisco.com/en/US/docs/security/asa/asa81/config/guide/intrface.html#wp1044006
					if ($port =~ /^\Q$ifName\E\.\d+$/)
					{
						if (!defined($interface->{interfaceVlanTrunks}))
						{
							$interface->{interfaceVlanTrunks} = [];
						}
						push(@{$interface->{interfaceVlanTrunks}}, {startVlan => $portVlans->{$port}});
					}
				}
			}
			
			if ($in->{running_config} =~ /^interface\s+$ifName\s+([^!]+)!/mig)
			{
				my $ifCfgBlob = $1;
				foreach my $ifBehavior (@{_get_interface_behaviors($ifName, $iline, $ifCfgBlob, $in, $bridgeGroupsDefined)})
				{
					push ( @{ $interface->{'interfaceBehavior'}}, $ifBehavior);	
				}

				# Set $bridgeGroupMember to true if this is an ASA that supports bridge groups and this interface
				# is a member of a bridge group.
				$bridgeGroupMember = 0;
				if ($ifCfgBlob =~ /^\s*bridge-group \d+/m)
				{
					$bridgeGroupMember = 1;
				}

				if ($ifCfgBlob =~ /^\s*switchport trunk allowed vlan (\S+)/m)
				{
					my $allowedVlans = $1;
					foreach my $vlanToken (split(/,/, $allowedVlans))
					{
						if ($vlanToken =~ /(\d+)(?:-(\d+))?/)
						{
							my $vlanTrunk = {};
							$vlanTrunk->{startVlan} = $1;
							if (defined($2))
							{
								$vlanTrunk->{endVlan} = $2;
							}

							if (!defined($interface->{interfaceVlanTrunks}))
							{
								$interface->{interfaceVlanTrunks} = [];
							}
							push(@{$interface->{interfaceVlanTrunks}}, $vlanTrunk);
						}
						else
						{
							$LOGGER->debug("vlan token \"$vlanToken\" is not in the expected format, not parsed.");
						}
					}
				}

				# Parse the bridge group from a bridge-group statement, or from a bvi interface.
				if (($ifCfgBlob =~ /bridge-group (\d+)/) || ($ifName =~ /bvi(\d+)/i))
				{
					my $bridgeGroupNumber = $1;

					push(@{$interface->{interfaceGrouping}},
							{ ifGroupName => "BridgeGroup$bridgeGroupNumber", ifGroupingCharacteristic => 'proxyAffiliation' });
				}
				elsif ($in->{firewallTransparent} && !$bridgeGroupsDefined && ($ifCfgBlob =~ /^\s*nameif\s+\S+/m))
				{
					push(@{$interface->{interfaceGrouping}}, {ifGroupName => "BridgeGroup-simulated", ifGroupingCharacteristic => 'proxyAffiliation'});
				}
			}
		}
		if ( defined $interface->{name} )
		{
			if ( $iline =~ /\bis(\s+\S+)?\s+(up|down)/i )
			{
				$interface->{adminStatus} = $2;
			}
			# Add the interfaceIp information if it exists.
			# If this is a pre 8.4 ASA in transparent firewall mode we don't want to add the admin IP address to every interface.
			if (($iline =~ /\bIP\s+address\s+(\S+)\s*,\s+subnet\s+mask\s+(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})(\W+|$)/i) &&
				(!$in->{firewallTransparent} || ($bridgeGroupsDefined && !$bridgeGroupMember)))
			{
				my $ip = $in->{name_mapper}->get_ip($1);
				my $ipConfiguration = {
					ipAddress  => $ip,
					mask       => mask_to_bits($2),
					precedence => ++$order,
				};
				push( @{ $interface->{"interfaceIp"}->{"ipConfiguration"} }, $ipConfiguration );

				# Save the interface IP for future use.
				if (!defined($in->{_interface_ips}->{$interface->{name}}))
				{
					# Use a hash to dedup
					$in->{_interface_ips}->{$interface->{name}} = {};
				}
				$in->{_interface_ips}->{$interface->{name}}->{$1} = 1;
			}
			if ( $iline =~ /\bMTU\s+(\d+)(\s+bytes)?(\W+|$)/i )
			{
				$interface->{mtu} = $1;
			}
			if ( $iline =~ /\bBW\s+(\d+)\s+(\S+)(\s+(half|full|auto)\s+duplex)?(\W+|$)/i )
			{
				my $speed = $1;
				my $units = $2;
				$interfaceEthernet->{operationalDuplex} = lc($4) if ( defined $4 );
				if ( lc($1) eq 'auto' )
				{
					$interfaceEthernet->{autoDuplex} = 'true';
				}
				else
				{
					$interfaceEthernet->{autoDuplex} = 'false';
				}

				if ( $units =~ /Mb/i ) { $speed = $speed * 1000 * 1000; }
				if ( $units =~ /Kb/i ) { $speed = $speed * 1000; }
				$interface->{speed} = $speed;
			}
			if ( $iline =~ /\b(Auto-Duplex)?\((Auto|Half|Full)-duplex\)/i
				&& !defined $interfaceEthernet->{operationalDuplex} )
			{
				$interfaceEthernet->{operationalDuplex} = lc($2) if ( defined $2 );
				if ( defined $1 && lc($1) eq 'auto-duplex' )
				{
					$interfaceEthernet->{autoDuplex} = 'true';
				}
				else
				{
					$interfaceEthernet->{autoDuplex} = 'false';
				}
			}
			if ( $iline =~ /\bAuto-Speed/i )
			{
				$interfaceEthernet->{autoSpeed} = 'true';
			}
			if ( $iline =~ /\bHardware\s+is(.+)\n/i )
			{
				my $hwtext = $1;
				if ( $hwtext =~ /\baddress\s+is\s+([a-f0-9]{4})\.([a-f0-9]{4})\.([a-f0-9]{4})/i )
				{
					my $macaddress = uc( $1 . $2 . $3 );
					$interfaceEthernet->{macAddress} = $macaddress;
				}
			}
			if ( $iline =~ /\bMAC\s+address\s+([a-f0-9]{4})\.([a-f0-9]{4})\.([a-f0-9]{4})/i )
			{
				my $macaddress = uc( $1 . $2 . $3 );
				$interfaceEthernet->{macAddress} = $macaddress;
			}
			if ( $iline =~ /\bInterface number is (\d+)/i )
			{
				$interface->{ifIndex} = $1;
			}
		}
		# Now assign MAC, Ethernet, IPv6 properties
		if ( defined $interface->{name} )
		{
			if ( defined $interfaceEthernet->{macAddress} )
			{
				$interface->{interfaceEthernet}->{macAddress} = $interfaceEthernet->{macAddress};
			}
			if ( defined $interfaceEthernet->{operationalDuplex} )
			{
				$interface->{interfaceEthernet}->{operationalDuplex} = $interfaceEthernet->{operationalDuplex};
			}
			if ( defined $interfaceEthernet->{autoSpeed} )
			{
				$interface->{interfaceEthernet}->{autoSpeed} = $interfaceEthernet->{autoSpeed};
			}
			# Add the interfaces to the remoteGatewayResolution object to help resolve static routes with remote gateways.
			if( defined $interface->{interfaceIp} and defined $interface->{interfaceIp}->{ipConfiguration} ) {
				foreach my $ipConfiguration (@{$interface->{interfaceIp}->{ipConfiguration}})
				{
					$in->{_remoteGatewayResolution}->add_local_subnet($ipConfiguration->{ipAddress}, $ipConfiguration->{mask}, $interface->{name});
				}
			}
		}

		push(@interfaces, $interface);
	}

	# Create the simulated bridge group interface for pre 8.4 ASA devices using the admin IP address
	if ($in->{firewallTransparent} && !$bridgeGroupsDefined && ($in->{running_config} =~ /^ip\s+address\s+(\S+)\s+(\S+)/m))
	{
		my $ipAddress = $in->{name_mapper}->get_ip($1);
		my $mask = mask_to_bits($2);
		my $name = 'SimulatedBridgeGroup';

		push(@interfaces, {
			adminStatus => 'up',
			description => 'Pre ASA 8.4 simulated BVI interface',
			interfaceGrouping => {
				ifGroupName => 'BridgeGroup-simulated',
				ifGroupingCharacteristic => 'proxyAffiliation' },
			interfaceIp => {
				ipConfiguration => {
					ipAddress => $ipAddress,
					mask => $mask,
					precedence => ++$order }},
			interfaceBehavior => \@BVI_BEHAVIORS,
			interfaceType => 'virtual',
			name => $name,
			physical => 'false' });

		$in->{_remoteGatewayResolution}->add_local_subnet($ipAddress, $mask, $name);
	}

	_parse_crypto_map_tunnels($in);

	if (defined($in->{_crypto_map_interfaces}))
	{
		push(@interfaces, @{$in->{_crypto_map_interfaces}});
	}

	# This is where we print out the interface.  Unfortunately, owing to (someone :) making
	# a current definition of at least one element out of sort order, the fields must be individually
	# printed
	$out->open_element("interfaces");
	foreach my $interface (@interfaces)
	{
		$out->open_element("interface");
		for (qw(adminStatus description failover ifIndex interfaceEthernet interfaceGrouping interfaceIp interfaceTunnel interfaceBehavior interfaceType interfaceVlanTrunks mtu name physical speed))
		{
			# startVlan and endVlan are out of order in the schema, they must also be printed individually.
			if ($_ eq 'interfaceVlanTrunks')
			{
				foreach my $interfaceVlanTrunk (@{$interface->{interfaceVlanTrunks}})
				{
					$out->open_element('interfaceVlanTrunks');
					$out->print_element('startVlan', $interfaceVlanTrunk->{startVlan});
					$out->print_element('endVlan', $interfaceVlanTrunk->{endVlan}) if defined($interfaceVlanTrunk->{endVlan});
					$out->close_element('interfaceVlanTrunks');
				}
			}
			else
			{
				$out->print_element( "$_", $interface->{$_}) if defined ($interface->{$_});
			}
		}
		$out->close_element("interface");
	}
	$out->close_element("interfaces");

	if (defined($outFh))
	{
		close($outFh);
	}
}

# This subroutine is called in parse_interfaces.  It parses vlan membership information from
# $in->{running_config}.  It creates two hashes that are used in two different contexts.
#
# One hash, $in->{_vlans} is stored for later use by parse_vlans.  $in->{_vlans} uses each
# vlan as the key, and each value is an array of interface names that is a member of the vlan.
# $in->{_vlans} will prefer ifname names over the actual port name.
#
# The second hash is returned by the subroutine.  It uses the port name for the key.  It will
# always use the port name and never the ifname.  This is done so parse_interfaces can determine
# the vlans of any subinterfaces of a physical interface.  The value will be the vlan of the
# subinterface.
sub _parse_vlan_information
{
	my ($in) = @_;

	# This has will be used by parse_interfaces itself
	my $portNameVlans = {};

	# This parses ASA 5505 switch port vlan information.
	while ($in->{running_config} =~ /^interface\s+(\S+)\s+([^!]+)!/mig)
	{
		my $portName = $1;
		my $content = $2;

		my $interfaceName = $portName;
		if ($content =~ /^\s*nameif\s+(\S+)/m)
		{
			$interfaceName = $1;
		}

		my $vlan = undef;
		if ($portName =~ /^Vlan(\d+)$/)
		{
			$vlan = $1;
		}
		elsif ($content =~ /^\s*(?:switchport trunk native )?vlan (\d+)/m)
		{
			$vlan = $1;
		}

		if (defined($vlan))
		{
			# Set $in->{_vlans} for later use by parse_vlans
			if (!defined($in->{_vlans}->{$vlan}))
			{
				$in->{_vlans}->{$vlan} = [];
			}
			push(@{$in->{_vlans}->{$vlan}}, $interfaceName);

			# Set $portNameVlans for use by parse_interfaces
			$portNameVlans->{$portName} = $vlan;
		}
	}

	# The following parses all other ASA switch port vlan information.
	while ($in->{interfaces} && $in->{interfaces} =~ /(?:\n|^)(Interface (\S+) "(\S+)"(?:.*\n(?!Interface))+)/g)
	{
		my $configuration = $1;
		my $portName = $2;
		my $interfaceName = $3;

		if ($configuration =~ /Control Point Vlan(\d+)/)
		{
			my $vlan = $1;

			if (!defined($in->{_vlans}->{$vlan}))
			{
				$in->{_vlans}->{$vlan} = [];
			}
			push(@{$in->{_vlans}->{$vlan}}, $interfaceName);

			# Set $portNameVlans for use by parse_interfaces
			$portNameVlans->{$portName} = $vlan;
		}
	}

	return $portNameVlans;
}

sub _get_access_groups
{
	my ($in) = @_;

	my $result;

	# eg: access-group acl-name in interface outside
	while ($in->{running_config} =~ /access-group\s+(\S+)\s+(in|out)\s+interface\s+(\S+)/mig)
	{
		$result->{"interfaces_$2"}->{$3} = '';

		if (!defined($result->{$2}->{$1}))
		{
			$result->{$2}->{$1} = [];
		}

		push(@{$result->{$2}->{$1}}, $3);
	}

	return $result;
}

sub _is_physical
{
	my $name = shift;
	if ( $name =~ /\.\d+$/ )
	{
		return "false";
	}
	elsif ( $name =~ /seri|eth|gig|^fe|^fa|token|[bp]ri/i )
	{
		return "true";
	}
	else
	{
		return "false";
	}
}

# This subroutene generates the interfaces and static routes required for crypto map tunnels.
# Stores the interfaces in $in->{_crypto_map_interfaces} and routes in $in->{_crypto_map_routes}.
sub _parse_crypto_map_tunnels
{
	my ($in) = @_;

	my $cryptoMapWorkHash = {};

	# Dynamic crypto maps are not applied directly to an interface, instead they are applied to a
	# static crypto map that in turn is applied to an interface.
	#
	# http://www.cisco.com/en/US/docs/security/asa/asa84/command/reference/c5.html#wp2265844
	#
	# This hash is intended to map dynamic crypto maps to static crypto maps to help determine the
	# correct interface.
	my $dynamicCryptoMappings = {};

	while ($in->{running_config} =~ /^crypto (?:dynamic-)?map (\S+) (\d+) (.+)\s*$/gm)
	{
		my $mapName = $1;
		my $sequenceNumber = $2;
		my $line = $3;

		if ($line =~ /^match address (\S+)/)
		{
			$cryptoMapWorkHash->{$mapName}->{$sequenceNumber}->{acl} = $1;
		}
		elsif ($line =~ /^set peer (\S+)/)
		{
			$cryptoMapWorkHash->{$mapName}->{$sequenceNumber}->{peer} = $1;
		}
		elsif ($line =~ /^ipsec-isakmp dynamic (\S+)/)
		{
			$dynamicCryptoMappings->{$1} = $mapName;
		}
	}

	my $localIpAddressEnding = 1;

	foreach my $mapName (keys(%{$cryptoMapWorkHash}))
	{
		my $interface = undef;
		if ($in->{running_config} =~ /^crypto map \Q$mapName\E interface (\S+)/m)
		{
			$interface = $1;
		}
		elsif (defined($dynamicCryptoMappings->{$mapName}))
		{
			my $dynamicMappedName = $dynamicCryptoMappings->{$mapName};
			if ($in->{running_config} =~ /^crypto map \Q$dynamicMappedName\E interface (\S+)/m)
			{
				$interface = $1;
			}
		}

		if (!defined($interface))
		{
			next;
		}

		foreach my $sequenceNumber (keys(%{$cryptoMapWorkHash->{$mapName}}))
		{
			# Not enough information to coninue if acl is not defined.
			if (!defined($cryptoMapWorkHash->{$mapName}->{$sequenceNumber}->{acl}))
			{
				next;
			}

			my $peerIpAddress = undef;
			if (defined($cryptoMapWorkHash->{$mapName}->{$sequenceNumber}->{peer}))
			{
				$peerIpAddress = $cryptoMapWorkHash->{$mapName}->{$sequenceNumber}->{peer};
			}
			# If the peer is not defined we stil have a chance to obtain the peer address from the output of show crypto ipsec sa command.
			else
			{
				if (defined($in->{crypto_ipsec_sa}))
				{
					my @peers;

					my $inCryptoMap = 0;
					while ($in->{crypto_ipsec_sa} =~ /^(.*?)\s*$/gm)
					{
						my $line = $1;

						if ($line =~ /Crypto map tag: (\S+), seq num: (\d+), local addr:/g)
						{
							if (($mapName eq $1) && ($sequenceNumber eq $2))
							{
								$inCryptoMap = 1;
							}
							else
							{
								$inCryptoMap = 0;
							}
						}
						elsif ($inCryptoMap)
						{
							if ($line =~ /current_peer: ([^\s,]+)/)
							{
								push(@peers, $1);
							}
						}
					}

					# We will only accept this peer if we found exactly 1.
					if (scalar(@peers) == 1)
					{
						$peerIpAddress = $peers[0];
					}
				}
			}

			if (!defined($peerIpAddress))
			{
				next;
			}

			my $simulatedInterfaceName = "SimL2LTunnelInterface-$mapName-$sequenceNumber";

			my $aclToRoute = ZipTie::Adapters::Cisco::SecurityAppliance::AclToRoute->new();
			$aclToRoute->convertAclToRoute($cryptoMapWorkHash->{$mapName}->{$sequenceNumber}->{acl}, $in->{_acl_terms}, $in->{_network_groups});
			my $routes = $aclToRoute->getRoutes($simulatedInterfaceName);

			# Only create the interface and routes if there is at least one route to use the interface
			if (scalar(@{$routes}))
			{
				# Add simulated routes
				if (!defined($in->{_crypto_map_routes}))
				{
					$in->{_crypto_map_routes} = [];
				}
				push(@{$in->{_crypto_map_routes}}, @{$routes});

				# Create the simulated interface
				my $interface = {
						description => 'Simulated interface to support ASA LAN-to-LAN tunnels',
						interfaceIp => {
								ipConfiguration => {
										# Choose addresses from rfc5737 TEST-NET-1, should not be used anywhere else on this device.
										ipAddress => '192.0.2.' . $localIpAddressEnding++,
									}
							},
						interfaceTunnel => {
								encapsulationProtocol => 'ipSec',
								tunnelLocalPeering => {
										tunnelLocalPeeringInterface => $interface,
									},
								tunnelRemotePeerIPAddress => $peerIpAddress
							},
						interfaceType => 'other',
						name => $simulatedInterfaceName,
						physical => 'false'
					};

				if (!defined($in->{_crypto_map_interfaces}))
				{
					$in->{_crypto_map_interfaces} = [];
				}
				push(@{$in->{_crypto_map_interfaces}}, $interface);
			}
		}
	}
}

# The following acts as a front-end, of sorts, to Utils::get_interface_type.
# it determines if there's something PIX/ASA specific to trigger a typing
# that we'd want to see here.
#
# PARAMS:
#   $type - "type" of interface, more likely interface name such as ethernet0, 
#        serial0.0/0, those sorts of things.
#   $interaceDetail - section of 'get interfaces full' that gives full device
#         definition to this interface.
#   $confiParams - output] from startup/running config 
#
# RETURNS:
#      returns a string, with an exact SED-assignable interface type value
#          from ONM 'interfaceType' enumeration.
#
sub _get_interface_type
{
	my ($type, $interfaceDetail, $configParams, $name) = @_;

	# First, let's see if there's anything in the interface detail that's a 
	# giveaway here...
	if (defined $interfaceDetail)
	{
		# Scan for: start of line, followed by white space, followed by 
		# "management-only interface".	
		if ($interfaceDetail =~ /^\s+management\-only\sinterface/mi)
		{
			return "softwareLoopback";		
		}
		# other cases go here, otherwise fall thorough.
	}
	
	# special cases derivable from the configuration section for this
	# interface	config
	# _get_interface_type is called both during telemetry collection and
	# during a backup.  When collecting telemetry the running_config is
	# not available.  This means the failover link will not be set to
	# softwareLoopback during telemetry collection, but this should be
	# ok because topology does not use telemetry to build its model.
	if (defined $configParams->{running_config})
	{
		# Check for failover link.  This can be either cable-based or LAN-based.  The statements 
		# we're looking for assign a name which is logical in the context of the failover config
		# to the *physical* interface (name passed) in the context of subsequent failover
		# configuration.  That logical name is the "\S+" below
		if ($configParams->{running_config} =~ /^failover (link\s\S+|lan\sinterface\s\S+) $type/m)
		{
			# This is a failover link, which we're currently using as a 
			# "loopback" inteface, to mitigate ambitious attempts to use
			# for routing/bridged traffic egress
			return "softwareLoopback";					
		}
	}

	if ($type =~ /^Internal-Control/i)
	{
		return 'softwareLoopback';
	}

	if (defined($name) && $name eq '_internal_loopback')
	{
		return 'softwareLoopback';
	}

	# bvi is a type specific to the ASA (so it shouldn't be in Utils.pm
	if ($type =~ /^bvi/i)
	{
		return 'virtual';
	}
	
	return get_interface_type($type);
	
}

# This routine enforces the current policy for set of interface behavior.,  Behaviors
# can occur more than once (often will) and are not mutually exclusive.. .  
#:
#  PARAMS::
#   $ifIdentifier - interface name such as ethernet0, 
#        serial0.0/2, those sorts of things.
#   $interaceDetail - section of 'get interfaces full' that gives full device
#         definition to this interface.
#   $ifCfg - Configuration of interface from running config
#   $confiParams - output] from startup/running config 
#
#  RETURNS:
#  Reference to a list of strings which defines the behaviors for thie list.  This
#  list may be empty, but the reference returned won't be undefined.
#
sub _get_interface_behaviors {
	my ($ifIdentifier, $interfaceDetail, $ifCfg, $configParams, $bridgeGroupsDefined) = @_;
	my @behaviorList = ();
	my $suppressTrafficProcessingOptions = 0;

	# First, let's see if there's anything in the interface detail that's a 
	# giveaway here...
	if (defined $interfaceDetail)
	{
		# Scan for: start of line, followed by white space, followed by 
		# "management-only interface". We don't want to be considering this
		# type of interface for traffic forwarding, then.
		if ($interfaceDetail =~ /^\s+management\-only interface/mi)
		{
			push @behaviorList, "administrative";
			$suppressTrafficProcessingOptions = 1;
		}
		# other cases go here, otherwise fall thorough.
	}
	
	# special cases derivable from the configuration section for this interface config
	if (defined $configParams)
	{
		#Check for failover link
		if ($configParams->{running_config} =~ /^failover\s+(link\s+\S+|lan\s+interface\s+\S+)\s+$ifIdentifier/m)
		{
			# This is a failover link, which we're currently using as a 
			# "loopback" inteface, to mitigate ambitious attempts to use
			# for routing/bridged traffic egress
			push @behaviorList, "failover";					
		}
	}
        
	if (not $suppressTrafficProcessingOptions)
	{
		# TODO: put together a set of heruristics to determine "switching", "routing", "firewall", etc.
		# "promiscuous-mode" if IPS features are on
	}

	# If the interface is a member of a bridge-group then set an interfaceBehavior to 'bridging'.
	if ($ifCfg =~ /bridge-group/)
	{
		push(@behaviorList, 'bridging');
	}
	# If this is a transparent firewall and nameif is set then add bridging as a behaviour list.
	# An interface only passes traffic if nameif is set.
	elsif ($configParams->{firewallTransparent} && !$bridgeGroupsDefined && ($ifCfg =~ /^\s*nameif\s+\S+/m))
	{
		push(@behaviorList, 'bridging');
	}

	# bvi interfaces are the administrative interface for a bridge-group.
	if ($ifIdentifier =~ /^bvi/i)
	{
		push(@behaviorList, @BVI_BEHAVIORS);
	}

	# return reference to list
	return \@behaviorList;
}

sub parse_vlans
{
	my ($in, $out) = @_;

	# Vlan information has been already parsed in parse_interfaces, if they exist the
	# information will be stored in $in->{_vlans}.  Print it out here if vlans exist.
	if (defined($in->{_vlans}))
	{
		$out->open_element('vlans');
		foreach my $vlan (keys(%{$in->{_vlans}}))
		{
			$out->open_element('vlan');
			$out->print_element('enabled', 'true');
			foreach my $interface (@{$in->{_vlans}->{$vlan}})
			{
				$out->print_element('interfaceMember', $interface);
			}
			$out->print_element('number', $vlan);
			$out->close_element('vlan');
		}
		$out->close_element('vlans');
	}
}

sub parse_stp
{

	# parses Spanning Tree information
	my ( $in, $out ) = @_;
}

# Parses out user contexts from multicontext config
sub parse_subelements {
	my ($in, $out) = @_;
	
	while( $in->{na_running_config} =~ /^context\s+(\S+)\s*\n([^!]+)!/mig ) {
		my ($name , $blob) = ($1, $2);

		$LOGGER->debug("Parsing ASA/FWSM context '$name'");
		my $context = { name => $name }; 

		# Look for interfaces, which *can* come as a range, and can have effectively aliases by
		# which they're referred in the context-specific running-config
		while ($blob =~ /^\s+allocate-interface\s+(\S+)\s*(\S*)\s*$/mig ) {
			my ($interface_name, $interface_alias) = ($1, $2);
			my $interface = {
				name => $interface_name
			};
			my $log_message = "Found allocated interface(s) '$interface_name'";

			if( $interface_alias ) {
				$interface->{alias} = $interface_alias;
				$log_message .= " aliased as $interface_alias";
			}

			$LOGGER->debug($log_message);

			# TODO Add support for context interface aliasing
			# RTC#107950
			#push @{$context->{interfaceAliases}}, $interface;
		}

		# Not using other possibles: description, allocate-ips, service class membership, config-url

		push @{$in->{userContexts}}, $context;
	}

	return;
}

# Parses all single user contexts by name, establishes its entry as an "instance" in the SED
sub parse_subelement_data {
	my ($in, $out) = @_;

	# Make sure we have security contexts...
	if( defined $in->{userContexts} ) {
		_parse_firewall_modes($in);

		$out->open_element('subelements');
		for my $context ( @{$in->{userContexts}} ) {
			if( $context->{running_config} ) {
				$out->open_element('subelement');

				# parse the entire per-context visible data.
				_parse_subelement($context, $out);

				$out->close_element('subelement');
			} else {
				$LOGGER->debug("No running-config for context '$context->{name}', dropping context.");
			}
		}
		$out->close_element('subelements');
	}

	return;
}

sub _parse_subelement {
	my ($in, $out) = @_;

	if( $in->{running_config} =~ /^hostname\s+(\S+)/im ) {
		$out->print_element( 'core:systemName', $1 );
	}
	
	# There are two subelement-scope nodal attributes indicating gestalt of this virtual device/context.
	$out->print_element( 'core:deviceType', 'Firewall' );
	
	# call the litany of subsystem parsers which might be relevant to the user-context-scoped
	# $in configuration data and sections.  Opening and closing section for the subelement are
	# taken care of by the caller.
	parse_names($in, $out);
	parse_object_groups( $in, $out );
	parse_filters( $in, $out );
	parse_interfaces($in, $out);
	parse_local_accounts($in, $out);
	parse_snmp($in, $out);
	parse_static_routes($in, $out);
	parse_vlans($in, $out);
	
	# The second subelement scope attribute is roughly an indicator as to the "type" of element
	# subordination if you will...is this a proxy to another system, or is it virtualized within
	# a chassis.  That's right, you got it.
	$out->print_element( 'subelementType', 'virtual' );
	return;
}

# checks to see if the incoming port is an integer.  If not
# it will do a lookup of the service name to convert it to an integer
# $protocol is optional
sub _int_port
{
	my ($port, $protocol) = @_;
	if ( $port =~ /^\d+$/ )
	{
		return $port;
	}
	else
	{
		return get_port_number($port, $protocol);
	}
}

# Checks if the incoming IP protocol is an integer.  If it is it will
# do a lookup to convert the protocol name, otherwise it returns the name
sub _name_protocol
{
	my ($protocol) = @_;
	if ( $protocol =~ /^\d+$/ )
	{
		return get_protocol_name($protocol);
	}
	else
	{
		return $protocol;
	}
}

# Creates a network schema object and returns it as a hash
# If $mask is non continuous the ipV4Mask element is used, otherwise the mask element is used.
sub _create_subnet
{
	my ($ip, $mask) = @_;

	my $bits = mask_to_bits($mask);

	if ($bits =~ /^$cipm$/)
	{
		return {address => $ip, ipV4Mask => $mask};
	}
	else
	{
		return {address => $ip, mask => $bits};
	}
}

# Extends _int_port to do a service lookup where possible. Return a protocol where
# possible. 
#
# PARAMS:
#  $portInput - numeric or service-name identifier of port
#
# RETURNS:
#  REF of ARRAY of hash refs: (port => $portNumber, protocol => $protocol) - 
#					$portNumber is numeric identifier of protocol,
#									$protocol is protocol string.
#  return undef if no mapping. 

sub _svc_int_port
{
	my $portInput = shift;
	my $portProtocol = undef;
	
	# attempt lookup in our own table of PIX-known services.
	foreach my $pixPortMapping (@{$PIX_PREDEFINED_SERVICES})
	{
		if ($pixPortMapping->{serviceName} eq $portInput)
		{
		# port mapping is an array... of port/protocol hash key values  
			return ($pixPortMapping->{mapping});
		}
	}
	
	my @pMappingList;
	
	# if we get here, there's nothing PIX specific here, so look to our general pool
	my $pExp = get_port_expression($portInput); 
	if (defined ($pExp))
	{
		foreach my $pExpTerm (@{$pExp})
		{
			if (defined $pExpTerm->{portExpression})
			{
				push (@pMappingList, 
						{ port => $pExpTerm->{portExpression}->{port},
							protocol => $pExpTerm->{portExpression}->{protocol}
						}	);							
			}
			elsif (defined $pExpTerm->{portRange})
			{
				# This interface doesn't really accomodate "range", but this is a start
				push (@pMappingList,
					{ port => $pExpTerm->{portRange}->{portStart},
							protocol => $pExpTerm->{portRange}->{protocol}
						});
			}
		}
	}
	
	if ( scalar (@pMappingList))
	{
		return \@pMappingList;
	}
	
	if (_int_port($portInput) =~ /^(\d+)$/)
	{
		my $pMapping = { port => $1, protocol => $portProtocol};
		@pMappingList = ($pMapping);
		# MUST return list ref.
		return \@pMappingList;
	}
	
	# if we get here, we couldn't map
	$LOGGER->debug("Cisco SecurityAppliance parsing: could not map service to port: $portInput");
	return undef;
}

sub tagNonIpv4Address {
	my $group_params = shift;
	my $ip           = shift;

	unless ( $ip =~ $cipm ) {
		$group_params->{notIpv4} = 1;
	}
}

1;

__END__

=head1 Parsers

ZipTie::Adapters::Cisco::SecurityAppliance::Parsers

=head1 SYNOPSIS

	use ZipTie::Adapters::Cisco::SecurityAppliance::Parsers qw(parse_routing create_config parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_interfaces parse_static_routes parse_vlans parse_stp);
	parse_filters( $cliResponses, $xmlPrinter);	

=head1 DESCRIPTION

Module with many static methods that take in a hash of
device responses and print out ZipTie model elements.

=head2 Methods

=over 12

=item C<parse_names( $in )>

Parse the output of show names.  Doesn't output any XML, but creates name to IP mappings and stores it inside $in.  This is used by subsequent parsers.

=item C<create_config( $in, $out )>

Uses the device configs to put together a ZipTie configurationRepository element.

=item C<parse_local_accounts( $in, $out )>

Parse the local accounts.

=item C<parse_chassis( $in, $out )>

Parse out the chassis level elements.

=item C<parse_filters( $in, $out )>

Parse out firewall rules.

=item C<parse_routing( $in, $out )>

Parse out BGP and OSPF configuration

=item C<parse_snmp( $in, $out )>

Parse out the SNMP elements.

=item C<parse_system( $in, $out )>

Parses out some top level core attributes of the ZipTie model ZiptieElementDocument

=item C<parse_vlan( $in, $out )>

Parses out the VLAN elements.

=item C<parse_stp( $in, $out )>

Parses out the spannint tree elements.

=item C<parse_user_contexts($in, $out)>

Parses and creates entries and subinterface associations, etc., for user contexts based on collected running config data

=item C<parse_user_context($in, $out, $contextName)>

Parses and generates ZED elements for a given user context (virtualized firewall).

=back

=head1 LICENSE

  The contents of this file are subject to the Mozilla Public License
  Version 1.1 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/
  
  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
  License for the specific language governing rights and limitations
  under the License.
 
  The Original Code is Ziptie Client Framework.
  
  The Initial Developer of the Original Code is AlterPoint.
  Portions created by AlterPoint are Copyright (C) 2006,
  AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

  Contributor(s): rkruse
  Date: Apr 23, 2007

=cut
