package ZipTie::Adapters::Cisco::SecurityAppliance::Telemetry;

use 5.010_000;
use strict;
use warnings;

use ZipTie::Logger;
use ZipTie::Typer;
use ZipTie::CLIProtocolFactory;
use ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin;
use ZipTie::Adapters::Cisco::SecurityAppliance::Disconnect qw(disconnect);
use ZipTie::Adapters::Cisco::SecurityAppliance::Parsers qw(parse_arp parse_telemetry_interfaces parse_mac_table parse_routing_neighbors);
use ZipTie::Adapters::Utils qw(choose_admin_ip get_model_filehandle close_model_filehandle);
use ZipTie::Model::XmlPrint;

my $LOGGER = ZipTie::Logger::get_logger();

sub invoke
{
	my $pkg               = shift;
	my $connectionPathDoc = shift;
	
	# setup work
	my ($connectionPath, $discoveryParams) = ZipTie::Typer::translate_document( $connectionPathDoc, 'connectionPath' );
	my $filehandle = get_model_filehandle( "SecurityAppliance", $connectionPath->get_ip_address() );
	my $cliProtocol = ZipTie::CLIProtocolFactory::create($connectionPath);
	my $responses = {};
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'telemetry');
	$printer->attributes(1);
	$printer->open_discovery_event();

	# Make a connection to device and get the data necessary to fill out the discovery event document 
	my $promptRegex = ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin::execute( $cliProtocol, $connectionPath );

	# Attempt to turn off paging and if that fails set the more prompt
	my $tpResp = $cliProtocol->send_and_wait_for('terminal pager 0', $promptRegex);       
	if( $tpResp =~ /% Invalid input detected|for a list of available commands/ ) {
		$cliProtocol->set_more_prompt( '<--- More --->\s*$', '20');
	}
                                        	
	# We will now start collecting multi-context data if avaiable.  The data is parsed at a later time.  We collect the data now
	# to ensure we issue the commands in the same order the commands are issued by the backup interaction, to ensure recording compatibility.
	
	$printer->open_element('neighbors');
	
	# process the ARP table
	$responses->{arp} = $cliProtocol->send_and_wait_for('show arp', $promptRegex);
	parse_arp($responses, $printer);
	delete $responses->{arp};
	
	# process the MAC address table 
	$responses->{mac} = $cliProtocol->send_and_wait_for('show mac-address-table', $promptRegex);
	parse_mac_table($responses, $printer);
	delete $responses->{mac};
	
	# process the routing neighbors
	$responses->{ospf}		= $cliProtocol->send_and_wait_for('show ospf neighbor', $promptRegex);
	$responses->{routes}	= $cliProtocol->send_and_wait_for('show route', $promptRegex);
	#$responses->{eigrp} = $cliProtocol->send_and_wait_for('', $promptRegex);
	#$responses->{bgp} = $cliProtocol->send_and_wait_for('', $promptRegex);
	parse_routing_neighbors($responses, $printer);
	delete $responses->{ospf};
	delete $responses->{routes};
	#delete $responses->{eigrp};
	#delete $responses->{bgp};
	
	$printer->close_element('neighbors');

	my $changeSystem = $cliProtocol->send_and_wait_for('changeto system', $promptRegex . '|[#>]\s*$');

	# If the prompt changed from what it was, then we made something happen.

	my @contexts;
	my $contextResponses = {};

	if ($changeSystem !~ /$promptRegex/)
	{
		my $systemPrompt = ZipTie::Adapters::Cisco::SecurityAppliance::AutoLogin::_get_prompt($cliProtocol);

		# Determine the non-admin contexts on the system
		my $showContext = $cliProtocol->send_and_wait_for('show context', $systemPrompt);
		while ($showContext =~ /^\s{1,4}([^\*\s#]+)\s(?!\s*active Security Contexts)(?!\s*Name\s+Class)/img)
		{
			push(@contexts, $1);
		}

		my $adminContext;
		if ($showContext =~ /^\s*\*(\S+)/m)
		{
			$adminContext = $1;
		}
		else
		{
			$LOGGER->fatal('Failed to determine admin context.');
		}

		foreach my $context (@contexts)
		{
			my $contextPrompt = "\\/\Q$context\E[#>]\\s*\$";
			$cliProtocol->send_and_wait_for("changeto context $context", $contextPrompt);

			# Issue all the telemetry commands that can be issued in a subcontext.
			$contextResponses->{$context} = {};

			$contextResponses->{$context}->{arp} = $cliProtocol->send_and_wait_for('show arp', $contextPrompt);

			$contextResponses->{$context}->{ndp} = $cliProtocol->send_and_wait_for('show ipv6 neighbor', $contextPrompt);

			$contextResponses->{$context}->{mac} = $cliProtocol->send_and_wait_for('show mac-address-table', $contextPrompt);

			$contextResponses->{$context}->{routes} = $cliProtocol->send_and_wait_for('show route', $contextPrompt);

			$contextResponses->{$context}->{ospf} = '';	# ospf is required by parse_routing_neighbors, but not avaiable in a context.

			$contextResponses->{$context}->{interfaces_ipv6} = $cliProtocol->send_and_wait_for( "show ipv6 interface", $contextPrompt);

			$contextResponses->{$context}->{interfaces} = $cliProtocol->send_and_wait_for('show interface', $contextPrompt);
		}

		$cliProtocol->send_and_wait_for("changeto context $adminContext", $promptRegex);
	}
	
	# process the interfaces and their status
	$responses->{interfaces_ipv6} = $cliProtocol->send_and_wait_for( "show ipv6 interface", $promptRegex);
	$responses->{interfaces} = $cliProtocol->send_and_wait_for('show interface', $promptRegex);
	my $interfacesHash = parse_telemetry_interfaces($responses, $printer);
	delete $responses->{interfaces};
	
	my $adminIp    = $connectionPath->get_ip_address();
	if ( $discoveryParams->{calculateAdminIp} eq 'true' )
	{
		$adminIp = choose_admin_ip($connectionPath->get_ip_address(), $interfacesHash);
	}
	$printer->print_element( 'adminIp', $adminIp );

	# Parse the interaction data previously collected.
	if (scalar(@contexts))
	{
		$printer->open_element('subdevices');

		foreach my $context (@contexts)
		{
			$printer->open_element('subdevice');
			$printer->print_element('subdeviceName', $context);

			$printer->open_element('neighbors');

			parse_arp($contextResponses->{$context}, $printer);
			delete($contextResponses->{$context}->{arp});

			parse_cdp($contextResponses->{$context}, $printer);
			delete($contextResponses->{$context}->{ndp});

			parse_mac_table($contextResponses->{$context}, $printer);
			delete($contextResponses->{$context}->{mac});

			parse_routing_neighbors($contextResponses->{$context}, $printer);
			delete($contextResponses->{$context}->{routes});
			delete($contextResponses->{$context}->{ospf});

			$printer->close_element('neighbors');

			parse_telemetry_interfaces($contextResponses->{$context}, $printer);
			delete($contextResponses->{$context}->{interfaces});
			delete($contextResponses->{$context}->{interfaces_ipv6});

			$printer->close_element('subdevice');
		}

		$printer->close_element('subdevices');
	}
	
	# tear down work
	$printer->close_element('DiscoveryEvent');
	close_model_filehandle($filehandle);
	disconnect($cliProtocol);

	return;
}

1;
