package ZipTie::Adapters::F5::BigIP;

use strict;
use warnings;

use ZipTie::Adapters::F5::BigIP::AutoLogin qw(execute);
use ZipTie::Adapters::F5::BigIP::Parsers qw(parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_interfaces parse_static_routes parse_vlans parse_stp parse_filter_enabled parse_default_filter get_bigip_version);
use ZipTie::Adapters::F5::BigIP::TMSHParsers qw{parse_tmsh_system parse_tmsh_chassis parse_tmsh_interfaces parse_tmsh_stp parse_tmsh_static_routes parse_tmsh_vlans parse_tmsh_filter_enabled parse_tmsh_default_filter parse_tmsh_filters};
use ZipTie::Adapters::F5::BigIP::Disconnect qw(disconnect);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::Adapters::GenericAdapter;
use ZipTie::Model::XmlPrint;
use ZipTie::CLIProtocolFactory;
use ZipTie::CLIProtocol;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use Exporter qw(import);
our @EXPORT_OK = qw(get_version);

# Specifies that this adapter is a subclass of ZipTie::Adapters::BaseAdapter
use ZipTie::Adapters::BaseAdapter;
our @ISA = qw(ZipTie::Adapters::BaseAdapter);

my $LOGGER = ZipTie::Logger::get_logger();

sub backup{
	my $package_name = shift;

	# Retrieve the operation XML document that contains all of the IP, protocol, credential, and file server information
	# that is needed to successfully backup a device.
	my $backup_doc = shift;

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ($connection_path) = ZipTie::Typer::translate_document($backup_doc, 'connectionPath');

	# Connect to the device and capture the ZipTie::CLIProtocol that is created as a result of the connection.
	# Also be sure to capture the device prompt that is returned as a result of a successful connection.
	my ($cli_protocol, $device_prompt_regex) = _connect($connection_path);

	# Grab an output filehandle for the model.  This usually points to STDOUT
	my $filehandle = get_model_filehandle('BigIP', $connection_path->get_ip_address());
	
	# Initialize the model printer
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, 'common');
	$printer->open_model();

	my $responses;

	my $version = get_version($cli_protocol, $device_prompt_regex);
	if($version == 11){
		$responses = _get_tmsh_responses($cli_protocol, $device_prompt_regex);
		$responses->{packetFilterEnabled} = parse_tmsh_filter_enabled($responses, $printer);
		$responses->{defaultFilterAction} = parse_tmsh_default_filter($responses, $printer);
		parse_tmsh_system($responses, $printer);
		parse_tmsh_chassis($responses, $printer);
		parse_tmsh_filters($responses, $printer);
		parse_tmsh_interfaces($responses, $printer);
		parse_local_accounts($responses, $printer);
		parse_snmp($responses, $printer);
		parse_tmsh_stp($responses, $printer);
		parse_tmsh_static_routes($responses, $printer);
		parse_tmsh_vlans($responses, $printer);
	}else{
		$responses =_get_responses($cli_protocol, $device_prompt_regex);
		$responses->{packetFilterEnabled} = parse_filter_enabled($responses, $printer);
		$responses->{defaultFilterAction} = parse_default_filter($responses, $printer);
		parse_system($responses, $printer);
		parse_chassis($responses, $printer);
		parse_filters($responses, $printer);
		parse_interfaces($responses, $printer);
		parse_local_accounts($responses, $printer);
		parse_snmp($responses, $printer);
		parse_stp($responses, $printer);
		parse_static_routes($responses, $printer);
		parse_vlans($responses, $printer);
	}

	$printer->close_model();
	close_model_filehandle($filehandle);
	disconnect($cli_protocol);
}

sub _connect{
	my $connection_path = shift;
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	# Make a connection to and successfully authenticate with the BigIP device
	my $device_prompt_regex = execute($cli_protocol, $connection_path);

	# Store the found prompt as "prompt" on the specified CLI protocol.
	$cli_protocol->set_prompt_by_name("prompt", $device_prompt_regex);

	# Return the created ZipTie::CLIProtocol object and the device prompt encountered after successfully connecting to a device.
	return ($cli_protocol, $device_prompt_regex);
}

sub get_version{
	my ($cli_protocol, $device_prompt_regex) = @_;

	my $bigpipe_response = $cli_protocol->send_and_wait_for(q{bigpipe version show},$device_prompt_regex);
	my $tmsh_response = $cli_protocol->send_and_wait_for(q{tmsh -q show sys version},$device_prompt_regex);

	if($bigpipe_response =~ /VERSION\s+((\d+)[\d\.]+)/i || $tmsh_response =~ /VERSION\s+((\d+)[\d\.]+)/i){
		$LOGGER->debug(qq{Version found: '$1'});
		return $2;
	}
	
	$LOGGER->fatal(qq{No version found!});
}

# Returns a hash of responses from the device
sub _get_responses{
	my $cli_protocol = shift;
	my $device_prompt_regex = shift;
	my $responses;
	$responses->{dmesg} = $cli_protocol->send_and_wait_for("dmesg",$device_prompt_regex);
	$responses->{global} = $cli_protocol->send_and_wait_for("bigpipe global",$device_prompt_regex);
	$responses->{hostname} = $cli_protocol->send_and_wait_for("bigpipe system hostname",$device_prompt_regex);
	$responses->{platform} = $cli_protocol->send_and_wait_for("bigpipe platform",$device_prompt_regex);
	$responses->{uptime} = $cli_protocol->send_and_wait_for("uptime",$device_prompt_regex);
	$responses->{version}  = $cli_protocol->send_and_wait_for("bigpipe version show",$device_prompt_regex);
	$responses->{license} = $cli_protocol->send_and_wait_for("cat /config/bigip.license",$device_prompt_regex);
	$responses->{snmp} = $cli_protocol->send_and_wait_for("cat /config/snmp/snmpd.conf",$device_prompt_regex);
	$responses->{users} = $cli_protocol->send_and_wait_for("cat /etc/passwd",$device_prompt_regex);
	$responses->{filter_enabled} = $cli_protocol->send_and_wait_for("bigpipe db packetfilter",$device_prompt_regex);
	$responses->{default_filter} = $cli_protocol->send_and_wait_for("bigpipe db packetfilter.defaultaction",$device_prompt_regex);
	$responses->{filters} = $cli_protocol->send_and_wait_for("bigpipe packet filter list",$device_prompt_regex);
	$responses->{nats} = $cli_protocol->send_and_wait_for("bigpipe nat list all",$device_prompt_regex);
	$responses->{vlans} = $cli_protocol->send_and_wait_for("bigpipe vlan show all",$device_prompt_regex);
	$responses->{vlan_groups_list} = $cli_protocol->send_and_wait_for("bigpipe vlangroup list all",$device_prompt_regex);
	$responses->{vlan_groups} = $cli_protocol->send_and_wait_for("bigpipe vlangroup",$device_prompt_regex);
	$responses->{interfaces} = $cli_protocol->send_and_wait_for("ip addr list",$device_prompt_regex);
	$responses->{physical_interfaces} = $cli_protocol->send_and_wait_for("bigpipe interface show all",$device_prompt_regex);
	$responses->{interface_snmp_index} = $cli_protocol->send_and_wait_for("snmpwalk -c public localhost 1.3.6.1.4.1.3375.2.1.2.4.3.2.1.1",$device_prompt_regex);
	$responses->{interface_snmp_names} = $cli_protocol->send_and_wait_for("snmpwalk -c public localhost 1.3.6.1.4.1.3375.2.1.2.4.3.2.1.2",$device_prompt_regex);
	$responses->{interface_speeds} = $cli_protocol->send_and_wait_for("bigpipe interface all media speed",$device_prompt_regex);
	$responses->{trunks} = $cli_protocol->send_and_wait_for(q{bigpipe trunk all interfaces},$device_prompt_regex);
	$responses->{stp} = $cli_protocol->send_and_wait_for("bigpipe stp show all",$device_prompt_regex);
	$responses->{static_routes} = $cli_protocol->send_and_wait_for("route -n",$device_prompt_regex);
	$responses->{static_routes_list} = $cli_protocol->send_and_wait_for("bigpipe route all list all",$device_prompt_regex);
	$responses->{mgmt} = $cli_protocol->send_and_wait_for("bigpipe mgmt show all",$device_prompt_regex);
	$responses->{mgmt_route} = $cli_protocol->send_and_wait_for("bigpipe mgmt route show all",$device_prompt_regex);
	$responses->{pools} = $cli_protocol->send_and_wait_for("bigpipe pool",$device_prompt_regex);
	$responses->{selfs} = $cli_protocol->send_and_wait_for("bigpipe self",$device_prompt_regex);
	$responses->{virtuals} = $cli_protocol->send_and_wait_for("bigpipe virtual list all",$device_prompt_regex);
	$responses->{snats} = $cli_protocol->send_and_wait_for("bigpipe snat list all",$device_prompt_regex);
	$responses->{snat_pools} = $cli_protocol->send_and_wait_for("bigpipe snatpool list all",$device_prompt_regex);
	$responses->{snats_any_protocol} = $cli_protocol->send_and_wait_for(q{b db snat.anyipprotocol},$device_prompt_regex);
	return $responses;
}

sub _get_tmsh_responses{
	my ($cli_protocol, $device_prompt_regex) = @_;
	my $responses;
	$responses->{hostname} = $cli_protocol->send_and_wait_for(q{tmsh -q list sys global-settings hostname},$device_prompt_regex);
	$responses->{version} = $cli_protocol->send_and_wait_for(q{tmsh -q show sys version},$device_prompt_regex);
	$responses->{hardware} = $cli_protocol->send_and_wait_for(q{tmsh -q show sys hardware},$device_prompt_regex);
	$responses->{contact} = $cli_protocol->send_and_wait_for(q{tmsh -q list sys snmp sys-contact},$device_prompt_regex);
	$responses->{dmesg} = $cli_protocol->send_and_wait_for(q{dmesg},$device_prompt_regex);
	$responses->{memory} = $cli_protocol->send_and_wait_for(q{tmsh -q show sys memory},$device_prompt_regex);
	$responses->{snmp} = $cli_protocol->send_and_wait_for(q{cat /config/snmp/snmpd.conf},$device_prompt_regex);
	$responses->{users} = $cli_protocol->send_and_wait_for(q{cat /etc/passwd},$device_prompt_regex);
	$responses->{interfaces} = $cli_protocol->send_and_wait_for(q{ip addr list},$device_prompt_regex);
	$responses->{physical_interfaces} = $cli_protocol->send_and_wait_for(q{tmsh -q list /net interface all-properties},$device_prompt_regex);
	$responses->{trunks} = $cli_protocol->send_and_wait_for(q{tmsh -q list net trunk},$device_prompt_regex);
	$responses->{interface_snmp_index} = $cli_protocol->send_and_wait_for(q{snmpwalk -c public localhost 1.3.6.1.4.1.3375.2.1.2.4.3.2.1.1},$device_prompt_regex);
	$responses->{interface_snmp_names} = $cli_protocol->send_and_wait_for(q{snmpwalk -c public localhost 1.3.6.1.4.1.3375.2.1.2.4.3.2.1.2},$device_prompt_regex);
	$responses->{filter_enabled} = $cli_protocol->send_and_wait_for(q{tmsh -q list /sys db packetfilter},$device_prompt_regex);
	$responses->{default_filter} = $cli_protocol->send_and_wait_for(q{tmsh -q list /sys db packetfilter.defaultaction},$device_prompt_regex);
	$responses->{filters} = $cli_protocol->send_and_wait_for(q{tmsh -q list /net packet-filter},$device_prompt_regex);
	$responses->{vlans} = $cli_protocol->send_and_wait_for(q{tmsh -q list /net vlan all-properties},$device_prompt_regex);
	$responses->{vlans_show} = $cli_protocol->send_and_wait_for(q{tmsh -q show /net vlan},$device_prompt_regex);
	$responses->{vlan_groups_list} = $cli_protocol->send_and_wait_for(q{tmsh -q list /net vlan-group all all-properties},$device_prompt_regex);
	$responses->{vlan_groups_show} = $cli_protocol->send_and_wait_for(q{tmsh -q show /net vlan-group},$device_prompt_regex);
	$responses->{virtuals} = $cli_protocol->send_and_wait_for(q{tmsh -q list ltm virtual},$device_prompt_regex);
	$responses->{nats} = $cli_protocol->send_and_wait_for(q{tmsh -q list ltm nat},$device_prompt_regex);
	$responses->{snat_pools} = $cli_protocol->send_and_wait_for(q{tmsh -q list ltm snatpool},$device_prompt_regex);
	$responses->{snats} = $cli_protocol->send_and_wait_for(q{tmsh -q list ltm snat},$device_prompt_regex);
	$responses->{snats_any_protocol} = $cli_protocol->send_and_wait_for(q{tmsh -q list sys db snat.anyipprotocol},$device_prompt_regex);
	$responses->{stp_global} = $cli_protocol->send_and_wait_for(q{tmsh -q list net stp-globals all-properties},$device_prompt_regex);
	$responses->{stp_priority} = $cli_protocol->send_and_wait_for(q{tmsh -q list net stp priority},$device_prompt_regex);
	$responses->{stp} = $cli_protocol->send_and_wait_for(q{tmsh -q list net stp all-properties},$device_prompt_regex);
	$responses->{static_routes} = $cli_protocol->send_and_wait_for(q{route -n},$device_prompt_regex);
	$responses->{static_routes_list} = $cli_protocol->send_and_wait_for(q{tmsh -q list net route},$device_prompt_regex);
	$responses->{mgmt} = $cli_protocol->send_and_wait_for(q{tmsh -q list sys management-ip},$device_prompt_regex);
	$responses->{mgmt_route} = $cli_protocol->send_and_wait_for(q{tmsh -q list sys management-route},$device_prompt_regex);
	$responses->{pools} = $cli_protocol->send_and_wait_for(q{tmsh -q list ltm pool},$device_prompt_regex);
	$responses->{selfs} = $cli_protocol->send_and_wait_for(q{tmsh -q list net self},$device_prompt_regex);
	$responses->{ipsec} = $cli_protocol->send_and_wait_for(q{tmsh -q list net ipsec},$device_prompt_regex);
	$responses->{tunnels} = $cli_protocol->send_and_wait_for(q{tmsh -q list net tunnels},$device_prompt_regex);
	return $responses;
}

1;
