package ZipTie::Adapters::F5::BigIP::Parsers;

use strict;
use warnings;
use ZipTie::Adapters::Utils qw{mask_to_bits get_crep strip_mac trim get_interface_type get_port_expression};
use ZipTie::Logger;
use Data::Dumper;
use Exporter qw{import};
our @EXPORT_OK = qw{parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_interfaces parse_static_routes parse_vlans parse_stp parse_filter_enabled parse_default_filter generate_packet_processing_declaration parse_file_storage get_bigip_version full_int_name get_host_or_network open_nat_list print_nat_sections get_vlan_groups print_packet_filter_virtual_servers};
my $LOGGER = ZipTie::Logger::get_logger();

sub parse_chassis{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_chassis; Parsing chassis});
	my $mac_regex = get_crep(qq{mac1}) . qq{|} . get_crep(qq{mac2}) . qq{|} .  get_crep(qq{mac3}) . qq{|} .  get_crep(qq{mac4});
	$out->open_element(qq{chassis});

	my $chassisAsset = {qq{core:assetType} => qq{Chassis}};
	$chassisAsset->{qq{core:factoryinfo}}->{qq{core:make}} = qq{F5 Networks};
	if($in->{platform} =~ /CHASSIS\s+SERIAL\s*:\s*(\S+)/i){
		my $serial = $1;
		my $hex = qq{[0-9a-f]};
		if($serial =~ /^\s*$hex{8}-($hex{4}-)+$hex{12}\s*$/){
			$chassisAsset->{qq{core:factoryinfo}}->{qq{core:serialNumber}} = $serial;
		}else{
			$chassisAsset->{qq{core:factoryinfo}}->{qq{core:serialNumber}} = $serial . qq{_} . $in->{hostname_value};
		}
	}
	if($in->{license} =~ /^(Platform\s+)?ID\s*:\s+(\S+)\s*$/mi){
		$chassisAsset->{qq{core:factoryinfo}}->{qq{core:modelNumber}} = qq{BIG-IP }.$2;
	}
	$out->print_element(qq{core:asset}, $chassisAsset);

	my $cpu;
	if($in->{dmesg} =~ /^CPU\d+:\s+(\S.+HZ)/mi){
		$cpu->{qq{core:description}} = $1;
		my $cputRE = _createCPUTypesRE();
		if ($cpu->{qq{core:description}} =~ /($cputRE)/i){
			$cpu->{cpuType} = $1;
		}
	}
	$out->print_element(qq{cpu}, $cpu ) if defined $cpu;

	parse_file_storage($in, $out);

	if($in->{platform} =~ /BASE\s+MAC\s*:\s+($mac_regex)/i){
		$out->print_element(qq{macAddress}, strip_mac($1));
	}

	my @memories = ();
	# under BigIP only common RAM memory is available.
	if($in->{global} =~ /MEMORY\s+\(TOTAL,\s*USED\)\s*=\s*\(([\d\.]+)([MGK]),\s*([\d\.]+)([MGK])\)\s*\n\s*\+/mi){
		my $memSize = $1;
		if(uc($2) eq qq{K}){
			$memSize *= 1024;
		}elsif(uc($2) eq qq{M}){
			$memSize *= 1024 * 1024;
		}elsif(uc($2) eq qq{G}){
			$memSize *= 1024 * 1024 * 1024;
		}
		#trim decimal point
		my $roundedMemSize = int($memSize + 0.5);
		push(@memories, {qq{core:description} => qq{RAM}, kind => qq{RAM}, size => $roundedMemSize});
	}
	foreach (@memories){
		$out->print_element(qq{memory}, $_);
	}

	$out->close_element(qq{chassis});
}

sub parse_file_storage{
	my ($in, $out) = @_;

	# populate the deviceStorage elements of the chassis
	while($in->{dmesg} =~ /^HD(\S+):\s*\d+\s+SECTORS\s+\((\d+)\s+([MGK])B\)/mig){
		my $storage = {
			name => qq{hd} . $1,
			storageType => qq{disk},
			size => $2,
		};
		if(uc($3) eq qq{K}){
			$storage->{size} *= 1024;
		}elsif(uc($3) eq qq{M}){
			$storage->{size} *= 1024 * 1024;
		}elsif(uc($3) eq qq{G}){
			$storage->{size} *= 1024 * 1024 * 1024;
		}

		# we won't populate FileDirectory node
		$out->print_element(qq{deviceStorage}, $storage);
	}
}

sub parse_system{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_system; Parsing system});	
	if($in->{hostname} =~ /LOCAL\s+HOST\s+NAME\s*:\s+(\S+)\s*$/mi){
		$out->print_element(qq{core:systemName}, $1);
		$in->{hostname_value} = $1;
	}

	$out->open_element(qq{core:osInfo});
	$out->print_element(qq{core:make}, qq{F5} );
	$out->print_element(qq{core:name}, qq{BIG-IP});
	if($in->{version} =~ /VERSION\s*([\d\.]+)/i){
		$out->print_element(qq{core:version}, $1);
	}
	$out->print_element(qq{core:osType}, qq{BIG-IP});
	generate_packet_processing_declaration($out);
	$out->close_element(qq{core:osInfo});

	if($in->{platform} =~ /VER:\s+(\S+)\s+.*(REV\S*\s*\S+)\s*$/mi){
		$out->print_element(qq{core:biosVersion}, $1 . qq{, } . $2);
	}elsif($in->{platform} =~ /\b(REV\S*\s*\S+)\s*$/mi){
		$out->print_element(qq{core:biosVersion}, $1);
	}elsif($in->{platform} =~ /\s+LICENSE\s+.*VE\s+/mi){
		$out->print_element(qq{core:biosVersion}, qq{Virtual Machine});
	}else{
		$out->print_element(qq{core:biosVersion}, qq{Unknown});
	}

	$out->print_element(qq{core:deviceType}, qq{Load Balancer});

	if($in->{snmp} =~ /^SYSCONTACT\s+(.+)<(.+)>/mi){
		$out->print_element(qq{core:contact}, $1 . $2 );
	}

	# Prints a blank line when it detects a core:lastReboot element for some reason
	#_parse_uptime($in, $out);
}

sub _parse_uptime{
	my ($in, $out) = @_;
	if($in->{uptime} =~ /\bUP\s+(.+)$/mi){
		$_ = $1;
		my ($years) = /(\d+)\s*YEARS?/i;
		my ($weeks) = /(\d+)\s*WEEKS?/i;
		my ($days) = /(\d+)\s*DAYS?/i;
		my ($hours) = /(\d+)\s*HOURS?/i;
		my ($minutes) = /(\d+)\s*MINUTES?/i;

		if(!defined $hours){
			my $temp = $_;
			if($temp =~ /(\d+)\:(\d+)/){
				$hours = $1;
				$minutes = $2;
			}
		}

		# subract the last reboot from the current time
		my $lastReboot = time();
		$lastReboot -= $years * 52 * 7 * 24 * 60 * 60 if ($years);
		$lastReboot -= $weeks * 7 * 24 * 60 * 60 if ($weeks);
		$lastReboot -= $days * 24 * 60 * 60 if ($days);
		$lastReboot -= $hours * 60 * 60 if ($hours);
		$lastReboot -= $minutes * 60 if ($minutes);
		$out->print_element(qq{core:lastReboot}, $lastReboot);
	}
}

sub parse_local_accounts{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_local_accounts; Parsing local accounts});
	my $account_found = 0;

	while(($in->{users} =~ /^([^\s:]+):([^:]+):(\d+):(\d+):([^:]+):([^:]+):([^:]+)$/mig)){
		my $account = {
			accountName => $1,
			accessLevel => $4,
			password    => $2,
			fullName    => $5
		};
		if(not $account_found){
			$account_found = 1;
			$out->open_element(qq{localAccounts});
		}
		$out->print_element(qq{localAccount}, $account);
	}
	if ($account_found){
		$out->close_element(qq{localAccounts});
	}
}

sub parse_snmp{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_snmp; Parsing SNMP});
	my $someSysPrinted = 0;
	my $snmp_found = 0;
	my $sysContact; my $sysLocation; my $sysName;
	my @communities; my @traps;
	
	if($in->{snmp} =~ /^\s*SYSCONTACT\s+(.+)<(.+)>\s*$/mi){
		$sysContact = $1 . $2;
		$someSysPrinted = 1;
	}

	if($in->{snmp} =~ /^\s*SYSLOCATION\s+(.+?)\s*$/mi){
		$sysLocation = $1;
		$someSysPrinted = 1;
	}

	if($in->{hostname} =~ /LOCAL\s+HOST\s+NAME\s*:\s+(\S+)\s*$/mi
	|| $in->{hostname} =~ /^\s*HOSTNAME\s+(\S+)\s*$/mi){
		$sysName = $1;
		$someSysPrinted = 1;
	}
	
	# parse communities and trap hosts
	while($in->{snmp} =~ /^(.+)$/mig){
		my $commline = $1;
		if ($commline =~ /^(R[OW])COMMUNITY\s+"(\S+)"\s*(\S*)\s*(\S*)$/i){
			if(not $snmp_found){
				$out->open_element(qq{snmp});
				$snmp_found = 1;
			}
			my $community;
			$community->{accessType} = uc($1);
			$community->{communityString} = $2;
			if($4){
				$community->{mibView} = $4;
			}
			push(@communities, $community);
		}elsif($commline =~ /^TRAPSESS.+\-C\s+"(\S+)"\s+(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})(\%(\d+))?:(\d+)\s*$/i){
			if($3){
				$LOGGER->debug(qq{::parse_snmp; Ignoring SNMP TRAP: route domains not supported});
			}else{
				if(not $snmp_found){
					$out->open_element(qq{snmp});
					$snmp_found = 1;
				}
				my $itsatrap;
				$itsatrap->{ipAddress} = $2;
				$itsatrap->{communityString} = $1;
				push(@traps, $itsatrap);
			}
		}
	}
	
	foreach my $community(@communities){
		$out->print_element(qq{community}, $community);
	}
	if($someSysPrinted){
		if(not $snmp_found){
			$out->open_element(qq{snmp});
			$snmp_found = 1;
		}
		$out->print_element(qq{sysContact}, $sysContact) if defined $sysContact;
		$out->print_element(qq{sysDescr}, qq{BIG-IP});
		$out->print_element(qq{sysLocation}, $sysLocation) if defined $sysLocation;
		$out->print_element(qq{sysName}, $sysName) if defined $sysName;
	}
	foreach my $itsatrap(@traps){
		$out->print_element(qq{trapHosts}, $itsatrap);
	}
	if($snmp_found){
		$out->close_element(qq{snmp});
	}
}

sub parse_static_routes{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{Parsing static routes});
	my $ip_regex = get_crep(qq{cipm});
	my $pools = _parse_pools($in);
	my $staticRoutes;

	foreach my $route(split(/^ROUTE\s*/mi, $in->{static_routes_list})){
		my $static_route;
		my $type;
		foreach my $route_line(split(/^/, $route)){
			if($route_line =~ /^\s*(DEFAULT|(($ip_regex)(\/(\S+))?)(\%(\d+))?)\s+/i){
				if($6){
					$LOGGER->debug(qq{::parse_static_routes; Ignoring static route: route domains not supported});
					$static_route = undef;
					last;
				}

				if(lc($1) eq qq{default}){
					$static_route->{destinationAddress} = qq{0.0.0.0};
					$static_route->{destinationMask} = qq{0};
					$static_route->{defaultGateway} = qq{true};
				}else{
					$static_route->{destinationAddress} = $3;
					if($4){
						$static_route->{destinationMask} = $5;
					}else{
						$static_route->{destinationMask} = qq{32};
					}
					$static_route->{defaultGateway} = qq{false};
				}
			}elsif($route_line =~ /^\s*(GATEWAY|REJECT|POOL|VLAN)\s*$/i){
				$type = lc($1);
				if($type eq qq{reject}){
					$static_route->{discard} = qq{true};
				}else{
					$static_route->{discard} = qq{false};
				}
			}elsif($route_line =~ /^\s*VLAN\s+(\S+)\s*$/i){
				if($type eq qq{vlan}){
					$static_route->{interface} = $1;
				}
			}elsif($route_line =~ /^\s*GATEWAY\s+($ip_regex)(\%(\d+))?\s*$/i){
				if($2){
					$LOGGER->debug(qq{::parse_static_routes; Ignoring static route: route domains not supported});
					$static_route = undef;
					last;
				}
				if($type eq qq{gateway}){
					$static_route->{gatewayAddress} = $1;
				}
			}elsif($route_line =~ /^\s*POOL\s+(\S+)\s*$/i){
				if($type eq qq{pool}){
					$static_route->{gatewayAddress} = $1;
				}
			}
		}
		if(defined $static_route){ 
			if(not $static_route->{interface}){
				while($in->{static_routes} =~ /^([\d\.]+)\s+([\d\.]+)\s+([\d\.]+)\s+(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\S+)\s*$/mig){
					if($1 eq q{0.0.0.0}){
						if($static_route->{gatewayAddress} && $2 eq $static_route->{gatewayAddress}){
							$static_route->{interface} = $8;
						}
					}elsif($static_route->{destinationAddress} eq $1){
						$static_route->{interface} = $8;
					}
				}
			}
			if($type eq qq{pool}){
				foreach my $pool_member (@{$pools->{$static_route->{gatewayAddress}}->{member}}){
					if ($pool_member->{enabled}){
						my $pool_route = {
							destinationAddress => $static_route->{destinationAddress},
							destinationMask => $static_route->{destinationMask},
							defaultGateway => $static_route->{defaultGateway},
							discard => $static_route->{discard},
							interface => $static_route->{interface},
							gatewayAddress => $pool_member->{ip}
						};
						push( @{ $staticRoutes->{staticRoute} }, $pool_route );
					}
				}
			}else{
				push( @{ $staticRoutes->{staticRoute} }, $static_route );
			}
		}
	}

	# Get the management route
	if($in->{mgmt} =~ /^\s*MGMT\s+($ip_regex)\s+NETMASK\s+($ip_regex)\s*$/mi){
		my $static_route = {
			destinationAddress => $1,
			destinationMask => mask_to_bits($2),
			defaultGateway => qq{true},
			interface => qq{mgmt},
			discard => qq{false}
		};
		if($in->{mgmt_route} =~ /^\s*\|\s+GATEWAY\s+($ip_regex)\s+MTU\s+(\d+)\s*$/mi ){
			$static_route->{gatewayAddress} = $1;
		}
		push(@{$staticRoutes->{staticRoute}}, $static_route);
	}
	
	$out->print_element(qq{staticRoutes}, $staticRoutes) if defined $staticRoutes;
}

sub _parse_self_ips{
	my ($in) = @_;
	my $ip_regex = get_crep(qq{cipm});
	my $selfs;
	foreach my $self_seg (split(/^SELF\s+/mi, $in->{selfs})){
		my $self;
		foreach my $line(split(/^/, $self_seg)){
			if($line =~ /($ip_regex)(\%(\d+))?\s+MASK\s+($ip_regex)\s*$/i){
				if($2){
					$LOGGER->debug(qq{::_parse_self_ips; Ignoring self ip: route domains not supported});
					last;
				}
				$self->{network} = $1;
				$self->{mask} = $4;
			}elsif($line =~ /\|\s*VLAN\s+(\S+)\s+FLOATING\s+(\S+)\s*$/i){
				push(@{$selfs->{$1}}, $self);
			}
		}
	}
	if($selfs){
		return $selfs;
	}
	return undef;
}

sub _parse_pools{
	my ($in) = @_;
	my $ip_regex = get_crep(qq{cipm});	
	my $pools;

	foreach my $pool (split(/^POOL\s+/mi, $in->{pools})){
		my $pool_name = qq{};
		foreach my $pool_member(split(/^\+\->\s+POOL\s+MEMBER\s+/mi, $pool)){
			my $member;
			foreach my $line(split(/^/, $pool_member)){
				if ($line =~ /\A(\S+)\s+LB\s+METHOD\s+/i){
					$pool_name = $1;
				}elsif ($line =~ /^$pool_name\/($ip_regex)(\%(\d+))?\:(\S+)\s+/i){
					if($2){
						$LOGGER->debug(qq{::_parse_pools; Ignoring pool member: route domains not supported});
						$member = undef;
						last;
					}
					$member->{ip} = $1;
					$member->{service} = lc($4);
				}elsif ($line =~ /^\|?\s*\|\s*SESSION\s+(ENABLED|DISABLED)\s+/i){
					$member->{enabled} = lc($1) eq qq{enabled} ? 1 : 0;
				}
			}
			push(@{$pools->{$pool_name}->{member}}, $member) if (defined $member);
		}
	}
	return $pools;
}

sub _parse_physical_interfaces{
	my ($in, $out, $interface_found) = @_;
	my $mac_regex = get_crep(qq{mac2});
	
	foreach my $if_part(split(/^INTERFACE\s+/mi, $in->{physical_interfaces})){
		my $interface;
		foreach my $if_line(split(/^/, $if_part)){
			if($if_line =~ /^(\S+)\s+($mac_regex)\s+(ENABLED|DISABLED)\s+(\S+)\s+MTU\s+(\d+)\s*$/i){
				$interface->{name} = $1;
				$interface->{description} = $1;
				$interface->{interfaceType} = qq{ethernet};
				$interface->{physical} = qq{true};
				$interface->{interfaceEthernet}->{autoSpeed} = qq{true};
				$interface->{interfaceEthernet}->{autoDuplex} = qq{true};
				$interface->{interfaceEthernet}->{macAddress} = strip_mac($2);
				$interface->{adminStatus} = lc($4);
				$interface->{mtu} = $5;
				if($interface->{adminStatus} ne qq{up}){
					$interface->{adminStatus} = qq{down};
				}
			}elsif($if_line =~ /^\|\s+MEDIA\s+(\S+)\s+\((NONE|((\S+)\s+(HALF|FULL)))\)\s+/i){
				if($3){
					$interface->{interfaceEthernet}->{mediaType} = $4;
					$interface->{interfaceEthernet}->{operationalDuplex} = lc($5);
				}
			}
		}
		if(defined $interface){
			# get ifIndex
			my $count = 0;
			my @ifIndex = split(/^/, $in->{interface_snmp_index});
			foreach my $if_name(split(/^/, $in->{interface_snmp_names})){
				if($if_name =~ /STRING\s*:\s+(\S+)\s*$/i){
				 	if($interface->{name} eq $1){
						if($ifIndex[$count] =~ /INTEGER\s*:\s+(\d+)\s*$/i){
							$interface->{ifIndex} = $1;
						}
					}
				}
				++$count;
			}

			# get speed
			foreach my $speed_part(split(/^\s*INTERFACE\s+/mi, $in->{interface_speeds})){
				my $name = $interface->{name};
				if($speed_part =~ /^$name.*SPEED\s*:\s+(\d+)/mi){
					$interface->{speed} = $1.qq{000000};
				}
			}

			# get vlan trunking
			foreach my $vlan(split(/^\s*VLAN\s+/mi, $in->{vlans})){
				if($vlan =~ /VLAN\s+MEMBER\s+(\S+)\s+/mi){
					my $mapString = $1;
					my @map = split(/\//, $mapString);
					if($map[1] eq $interface->{name}){
						my $vlanTrunk;
						$vlanTrunk->{startVlan} = $map[0];
						push(@{$interface->{interfaceVlanTrunks}}, $vlanTrunk);
					}
				}
			}

			foreach my $trunk(split(/^TRUNK\s+/mi, $in->{trunks})){
				my $trunk_name;
				if($trunk =~ /\A(\S+)\s+\-/m){
					$trunk_name = $1;
				}else{ next; }

				foreach my $line(split(/^/, $trunk)){
					if($line =~ /\s+INTERFACE\s+(\S+)\s*$/i){
						if($interface->{name} eq $1){
							$interface->{interfaceGrouping}->{ifGroupName} = $trunk_name;
							$interface->{interfaceGrouping}->{ifGroupingCharacteristic} = q{bondingGroupMember};
						}
					}
				}
			}
			
			if(not $interface_found){
				$interface_found = 1;
				$out->open_element(qq{interfaces});
			}
			$out->print_element(qq{interface}, $interface);
		}
	}
	return $interface_found;
}

sub parse_interfaces{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_interfaces; Parsing interfaces});
	my $ip_regex = get_crep(qq{cipm});
	my $interface_found = 0;
	
	# split on lines that don't start with a space, ie each interface
	foreach my $if_part(split(/^(?!\s)/m, $in->{interfaces})){
		my $interface;

		# split on new lines to parse one at a time
		foreach my $if_line(split(/^/, $if_part)){
			if($if_line =~ /^(\d+):\s+(\S+):\s+<(\S+)>\s+MTU\s+(\d+)/mi ){
				$interface->{name} = $2;
		    		$interface->{description} = $2;
				$interface->{physical} = qq{false};
				$interface->{adminStatus} = lc($3);
				$interface->{mtu} = $4;
				if($interface->{adminStatus} !~ /,UP,/i){
			       		$interface->{adminStatus} = qq{down};
				}else{
					$interface->{adminStatus} = qq{up};
				}
				$interface->{physical} = qq{false};
			}elsif($if_line =~ /\s*LINK\/(\S+)\s+(\S+)/mi){
				$interface->{interfaceType} = get_interface_type($1);
				if($1 eq qq{ether}){
					my $mac = strip_mac($2);
					if($mac ne qq{000000000000}){
						$interface->{interfaceEthernet}->{macAddress} = $mac;
					}
				}
			}elsif ($if_line =~ /\s*INET\s+($ip_regex)\/(\d+)\s+(BRD\s+($ip_regex))?/mi){
				my $ipConfiguration;
				$ipConfiguration->{ipAddress} = $1;
				$ipConfiguration->{mask} = $2;
				if($4){ $ipConfiguration->{broadcast} = $4; }
				push(@{$interface->{interfaceIp}->{ipConfiguration}}, $ipConfiguration);
			}
		}
		if(defined $interface){
			if(not $interface_found){
				$interface_found = 1;
				$out->open_element(qq{interfaces});
			}

			if($interface->{interfaceIp}){
				my $all_loopbacks = 1;
				foreach my $ip_config(@{$interface->{interfaceIp}->{ipConfiguration}}){
					if(substr($ip_config->{ipAddress}, 0, 4) ne q{127.}){
						$all_loopbacks = 0; last;
					}
				}
				if($all_loopbacks){
					$interface->{interfaceType} = q{softwareLoopback};
				}
			}

			my @vlan_names = @{_get_vlan_full_names($in)};
			foreach my $vlan_name(@vlan_names){
				if($interface->{name} =~ /\~\d+/){
					if($vlan_name->{shortname} && $vlan_name->{shortname} eq $interface->{name}){
						$interface->{name} = $vlan_name->{fullname};
						redo;
					}
				}elsif($vlan_name->{fullname} eq $interface->{name}){
					$interface->{interfaceType} = q{virtual};
					if($vlan_name->{tag}){
						$interface->{interfaceVlanTrunks}->{startVlan} = $vlan_name->{tag};
					}
				}
			}

			$out->print_element(qq{interface}, $interface);
		}
	}
	$interface_found = _parse_physical_interfaces($in, $out, $interface_found);
	if($interface_found){
		$out->close_element(qq{interfaces});
	}
}

sub parse_vlans{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_vlans; Parsing vlans});
	my $vlan_found = 0;

	foreach my $vlan_part(split(/^VLAN/mi, $in->{vlans})){
		if($vlan_part =~ /\s*(\S+)\s+(\((\S+)\)\s+)?TAG\s+(\d+)\s+(?:(?:[0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2})?\s*MTU\s+(\d+)/mi){
			my $vlan;
			if($2){
				$vlan->{name} = $3;
			}else{
				$vlan->{name} = $1;
			}
			$vlan->{number} = $4;
			$vlan->{mtu} = $5;
			$vlan->{enabled} = qq{true};

			while($vlan_part =~ /^\s*\+\->\s+INTERFACE\s+([\d\.]+)/mig){
				push(@{$vlan->{interfaceMember}}, full_int_name($1));
			}
			
			if(not $vlan_found){
				$vlan_found = 1;
			       	$out->open_element(qq{vlans});
			}
			$out->print_element(qq{vlan}, $vlan);
		}
	}

	$out->close_element(qq{vlans}) if $vlan_found;
}

sub parse_filter_enabled{
	my ($in, $out) = @_;
	my $returnVal = 0;
	if($in->{filter_enabled} =~ /PACKETFILTER\ =\ ENABLE/i){
		$returnVal = 1;
	}
	return $returnVal;
}

sub parse_default_filter{
	my ($in, $out) = @_;
	my $returnVal = qq{discard};
	if($in->{default_filter} =~ /PACKETFILTER\.DEFAULTACTION\ =\ (\S+)\s*/i){
		$returnVal = $1;
	}
	return $returnVal;
}

sub parse_filters{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_filters; Parsing filters});
	my @vlans = @{_get_vlan_names($in)};
	#add an empty element to represent all vlans
	push (@vlans, q{});
	
	my $filters_found = 0;
	foreach my $vlan (@vlans) {
		$filters_found = print_filterLists ($vlan, $in, $out, \@vlans, $filters_found);
	}

	my $virtual_servers = _get_vs_array_ref($in);
	if($virtual_servers){
		my $vlan_names = _get_vlan_names($in);
		$filters_found = print_packet_filter_virtual_servers($in, $out, $virtual_servers, $vlan_names);
	}

	if($filters_found){
		my $phaseDiscriminator;
		$phaseDiscriminator->{processingPhase}->{qq{core:packetProcessingPhase}} = qq{address-translation};
		$phaseDiscriminator->{processingPhasePosition} = qq{before};
		$out->print_element(qq{phaseDiscriminator}, $phaseDiscriminator);
		$out->close_element("filterListsSection");
	}

	_parse_nat($in, $out, $filters_found);
}

sub print_filterLists {
	my ( $vlan, $in, $out, $arrayVlans, $filters_found ) = @_;
	my $filter_found = 0;	
	foreach my $filter_part ( split( /^packet\ filter\ /mi, $in->{"filters"} ) )
	{
		#logit ("bla:$filter_part");
		if ( $filter_part =~ /^(.*)\ {/mi )
		{	
			my $filterElementName = $1;	
			#filter has a name and not the default filter
			if ($filterElementName && ($filterElementName !~ /^ / ) ) {
				#logit ("filterelementname:$filterElementName:") ;
				my $filterElementLog = q{false};
				my $filterElementPrimaryAction = q{deny};
				my $filterElementOrder = q{0};
				my $filterVlan = q{};
				my $filterLine = q{};

				my @filterparts = split( /\n/, $filter_part );
				@filterparts = grep (/^ / ,@filterparts);
			#	$out->open_element("filterEntry");
				foreach my $line ( @filterparts ) {
					#logit ("LINE: $line");
					if ( $line =~ /^\ *log\ (\S+)\s*$/mi ) {
						$filterElementLog = "true";
					}	
					if ( $line =~ /^\ *action\ (\S+)\s*$/mi ) {
						my $action = $1;
						if ($action =~ /accept/ ) {	
							$filterElementPrimaryAction = "permit";
						}elsif ($action =~ /continue/ ) {
							$filterElementPrimaryAction = "next";
						} 
					}
					if ( $line =~ /^\ *order\ (\S+)\s*$/mi ) {
						$filterElementOrder = $1;
					}	
					if ( $line =~ /^\ *vlan\ (\S+)\s*$/mi ) {
						$filterVlan = $1;
						#logit (qq{vlan: $filterVlan});
					}	
					if ( $line =~ /^\ *filter\ (.*)$/mi ) {
						$filterLine = $1;
					}
				}
				
				if ($filterVlan eq $vlan) {
					#logit (qq{vlan: found});
#					filter element
					if ( not $filter_found) {
						if(not $filters_found){
							$out->open_element("filterLists");
							$out->open_element("filterListsSection");
							$filters_found = 1;
						}
						$filter_found = 1;
						$out->open_element("filterList");
						my $defaultFilterAction = $in->{"defaultFilterAction"};
						if ($defaultFilterAction =~ /discard|reject/) {
							$out->print_element( "defaultAction", "deny" );
						} else {
							$out->print_element( "defaultAction", "permit" );
						}
						if ( $in->{"packetFilterEnabled"} ) {
							#if filter applies to all vlans lets print them out
							if ($filterVlan eq q{})	{
								my @allVlans = @$arrayVlans;
								pop (@allVlans);
								foreach my $iter (@allVlans) {
									$out->open_element("appliesTo");
									$out->print_element("source", $iter );
									$out->close_element("appliesTo");
								}
							} else {
								$out->open_element("appliesTo");
								$out->print_element("source", $vlan );
								$out->close_element("appliesTo");
							}
						}
					}
					parse_filter_line($filterElementName, $filterElementLog, $filterElementPrimaryAction, $filterElementOrder, $filterLine, $out);
				}
			}
		}
	}
	if ($filter_found) {
		$out->print_element( "mode", "stateful" );
		if ( $vlan eq q{} ) {
			$out->print_element( "name", q{ALL VLANS} );
		} else { 
			$out->print_element( "name", $vlan );
		}
		$out->close_element("filterList");
	}
	return $filters_found;
}

sub _is_unparseable_filter {
	my $line = shift;
	my $returnVal = 0;
	#things the zed cant handle
	#dst or src negated
	if ($line =~ m/host\ not|net\ not|proto\ not/i) {
		$LOGGER->debug("_is_unparseable_filter; ignoring packetfilter, " .
					" negating host, network or protocol $line");
		$returnVal = 1;
	} 
	#multiple levels of parenthesis
	if ($line =~ m/\([^\)]*\(/i) {
		$LOGGER->debug("_is_unparseable_filter; ignoring packetfilter, " .
					"nested brackets $line");
		$returnVal = 1;
	}
	#and in componennts ( src host 2.2.2.2  and src host 3.3.3.3 )
	foreach my $chunk ( split ( /\)/i, $line ) ) {
		if ($chunk =~ /\(.*and/i) {
			$LOGGER->debug("::_is_unparseable_filter; ignoring packetfilter, " .
					"Using AND within a component $line");
			$returnVal = 1;
		}
	} 
	#or between components ( src host 2.2.2.2 ) or ( dst host 3.3.3.3 )     
	if ($line =~ m/\)\ *or\ *\(/i) {
		$LOGGER->debug("::_is_unparseable_filter; ignoring packetfilter, " .
					"OR between components $line");
		$returnVal = 1;
	} 
	#pcap specific keywords  (less 100) and (ip multicast)
	return $returnVal;
}

sub parse_filter_line {
	my ($filterElementName, $filterElementLog, $filterElementPrimaryAction, $filterElementOrder, $line, $out) = @_;
	my $returnVal;
	my @dstHost;
	my @dstNetwork;
	my @dstNotPort;
	my @dstPort;
	my @srcHost;
	my @srcNetwork;
	my @protocol;
	if (! _is_unparseable_filter ($line) ) {
		foreach my $chunk ( split ( /\)/i, $line ) ) {
			#logit ("chunk:$chunk:");
			if ( $chunk =~ /src\ (.*)$/i ) {
				my $src = $1;
				if ($src =~ /or/) {
					#logit ("MULTIBALL SRC");
					foreach my $or ( split ( /or/i, $src ) ) {
						if ($or =~ /host(.*)/ ) {
							#logit ("src host:$1:");
							push (@srcHost, $1);
						}
						if ($or =~ /net(.*)/ ) {
							#logit ("src network:$1:");
							push (@srcNetwork, $1);
						}
					}
				} else {
					if ($src =~ /host(.*)/ ) {
						#logit ("src host:$1:");
						push (@srcHost, $1);
					} 
					if ($src =~ /net(.*)/ ) {
						#logit ("src network:$1:");
						push (@srcNetwork, $1);
					}
				}
			}
			if ( $chunk =~ /dst\ (.*)$/i ) {
				my $dst = $1;
				if ($dst =~ /or\ /) {
					#logit ("MULTIBALL DST");
					foreach my $or ( split ( /or\ /i, $dst ) ) {
						if ($or =~ /host(.*)/ ) {
							#logit ("dst host:$1:");
							push (@dstHost, $1);
						} 
						if ($or =~ /net(.*)/ ) {
							#logit ("dst net:$1:");
							push (@dstNetwork, $1);
						}
						if ($or =~ /port(.*)/ ) {
							my $checkForNot = $1;
							if ($checkForNot =~ /not(.*)/ ) {
								push (@dstNotPort, $1);
							#	logit ("dst port not:$1:");
							} else {
								push (@dstPort, $1);
							 #       logit ("dst port:$1:");
							}
						}
					}
				} else {
					if ($dst =~ /host(.*)/ ) {
						#logit ("dst host:$1:");
						push (@dstHost, $1);
					} 
					if ($dst =~ /net(.*)/ ) {
						#logit ("dst net:$1:");
						push (@dstNetwork, $1);
					}
					if ($dst =~ /port(.*)/ ) {
						my $checkForNot = $1;
						if ($checkForNot =~ /not(.*)/ ) {
							push (@dstNotPort, $1);
						} else {
							push (@dstPort, $1);
						}
					}
				}
			}
			if ( $chunk =~ /proto\ (.*)$/i ) {
				my $dst = $1;
				if ($dst =~ /tcp/i ) {
					#logit ("proto tcp");
					push (@protocol, "tcp");
				}
				if ($dst =~ /udp/i ) {
					#logit ("proto udp");
					push (@protocol, "udp");
				}
				if ($dst =~ /icmp/i ) {
					#logit ("proto icmp");
					push (@protocol, "icmp");
				}
			}
		}
		printFilterEntry  ($filterElementName, $filterElementLog, $filterElementPrimaryAction, $filterElementOrder, \@srcHost, \@srcNetwork, \@dstHost, \@dstNetwork, \@dstPort, \@dstNotPort, \@protocol, $out);
	}
}

sub printFilterEntry {
	my ($filterElementName, $filterElementLog, $filterElementPrimaryAction, $filterElementOrder, $srcHost, $srcNetwork, $dstHost, $dstNetwork, $dstPort, $dstNotPort, $protocol, $out) = @_;
	$out->open_element("filterEntry");
	my $numDstHosts =  scalar @$dstHost;
	my $numDstNetworks =  scalar @$dstNetwork;
	my $dstExists = $numDstHosts + $numDstNetworks;
	if ($dstExists > 0) {
		if ($numDstHosts > 0) {
			foreach my $host ( @$dstHost ) {
				$out->open_element("destinationIpAddr");
				$host = trim($host);
				$out->print_element( "host", $host );
				$out->close_element("destinationIpAddr");
			}
		}
		if ($numDstNetworks > 0) {
			foreach my $network ( @$dstNetwork ) {
				$network =~ /(.*)\/(.*)/;
			$out->open_element("destinationIpAddr");
			$out->open_element("network");
			$out->print_element( "address", trim($1) );
			$out->print_element( "mask", trim($2) );
			$out->close_element("network");
			$out->close_element("destinationIpAddr");
			}
		}
	}
	foreach my $port ( @$dstPort ) {
		$out->open_element("destinationService");
		$out->open_element("portExpression");
		$out->print_element( "operator", "eq" );
		$port = trim($port);
		$out->print_element( "port", $port );
		$out->close_element("portExpression");
		$out->close_element("destinationService");
	}
	foreach my $port ( @$dstNotPort ) {
		$out->open_element("destinationService");
		$out->open_element("portExpression");
		$out->print_element( "operator", "ne" );
		$port = trim($port);
		$out->print_element( "port", $port );
		$out->close_element("portExpression");
		$out->close_element("destinationService");
	}
	$out->print_element( "log", $filterElementLog );
	$out->print_element( "name", $filterElementName );				
	$out->print_element( "primaryAction", $filterElementPrimaryAction );	
	$out->print_element( "processOrder", $filterElementOrder );
	foreach my $protocol ( @$protocol ) {
		$out->print_element( "protocol", $protocol );
	}
	my $numSrcHosts =  scalar @$srcHost;
	my $numSrcNetworks =  scalar @$srcNetwork;
	my $srcExists = $numSrcHosts + $numSrcNetworks;
	if ($srcExists > 0) {
		foreach my $host ( @$srcHost ) {
			$out->open_element("sourceIpAddr");
			$host = trim($host);
			$out->print_element( "host", $host );
			$out->close_element("sourceIpAddr");
		}
		if ($numSrcNetworks > 0) {
			$out->open_element("sourceIpAddr");
			$out->open_element("network");
			foreach my $network ( @$srcNetwork ) {
				$network =~ /(.*)\/(.*)/;
				$out->print_element( "address", trim ($1) );
				$out->print_element( "mask", trim ($2) );
			}
			$out->close_element("network");
			$out->close_element("sourceIpAddr");
		}
	}
	$out->close_element("filterEntry");
}

sub print_packet_filter_virtual_servers{
	my ($in, $out, $virtual_servers, $vlan_names) = @_;
	my $vlan_lists = get_vlan_groups($in, $virtual_servers, $vlan_names);
	my @vlan_lists = @$vlan_lists;
	my $list_count = 0;

	for(my $i = 0; $i <= $#vlan_lists; ++$i){
		my $list_opened = 0;
		my $entry_count = 0;
		foreach my $virtual_server(@$virtual_servers){
			if($virtual_server->{list} == $i){
				if($virtual_server->{type} eq qq{reject}){
					if(not $list_opened){
						$out->open_element(qq{filterList});
						$out->print_element(qq{defaultAction}, qq{permit});
						foreach my $vlan(@{$vlan_lists[$i]}){
							my $appliesTo = {source => $vlan};
							$out->print_element(qq{appliesTo}, $appliesTo);
						}
						$list_opened = 1;
					}
					$out->open_element(qq{filterEntry});
					$out->print_element(qq{destinationIpAddr}, get_host_or_network($virtual_server->{destination}, $virtual_server->{mask}));

					if($virtual_server->{service} ne qq{any}){
						$out->print_element(qq{destinationService}, get_port_expression($virtual_server->{service}));
					}

					$out->print_element(qq{log}, qq{false});
					$out->print_element(qq{name}, $virtual_server->{name});
					$out->print_element(qq{primaryAction}, qq{deny});
					$out->print_element(qq{processOrder}, ++$entry_count);

					if($virtual_server->{protocol} && $virtual_server->{protocol} ne qq{any}){
						$out->print_element(qq{protocol}, $virtual_server->{protocol});
					}
					$out->close_element(qq{filterEntry});
				}
			}
		}
		if($list_opened){
			$out->print_element(qq{mode}, qq{stateful});
			$out->print_element(qq{name}, qq{Reject_Virtual_Servers_} . ++$list_count);
			$out->print_element(qq{processingPhase}, {qq{core:packetProcessingPhase} => qq{packet-filtering}});
			$out->close_element(qq{filterList});
		}
	}

	return $list_count;
}

sub parse_stp{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_stp; Parsing STP});
	my $spanningTree;
	my ($helloTime, $maxAge, $forwardDelay, $holdTime) = (0, 0, 0, 0);

	if($in->{stp} =~ /^STP\s+MODE\s+(\S+)/mi){
		$spanningTree->{mode} = $1;
	}

	if($in->{stp} =~ /^\|\s*FORWARD\s+DELAY\s+(\d+)\s+HELLO\s+TIME\s+(\d+)\s+MAX\s+AGE\s+(\d+)\s+TRANSMIT\s+HOLD\s+(\d+)/mi){
		$helloTime = $2;
		$maxAge = $3;
		$forwardDelay = $1;
		$holdTime = $4;
	}

	foreach my $stpins_part(split(/^\s*\+\->\s*STP\s+INSTANCE/mi, $in->{stp})){
		if($stpins_part =~ /^\s*(\d+)\s+PRIORITY\s+(\d+)\s+ROOT\s+BRIDGE\s+(\S+)/mi){
			if($spanningTree->{mode} eq qq{mstp}){
				my $instance = {
					helloTime => $helloTime,
					maxAge => $maxAge,
					forwardDelay => $forwardDelay,
					holdTime => $holdTime,
					priority => $2
				};
				if($3 ne qq{none}){
					$instance->{systemMacAddress} = strip_mac($3);
				}
				if($stpins_part =~ /^\s*\+\->\s+STP\s+VLAN\s+\d+\/(\S+)/mi){
					$instance->{vlan} = $1;
				}
				push(@{$spanningTree->{spanningTreeInstance}}, $instance);
			}else{
				$spanningTree->{helloTime} = $helloTime;
				$spanningTree->{maxAge} = $maxAge;
				$spanningTree->{forwardDelay} = $forwardDelay;
				$spanningTree->{priority} = $2;
				if($3 ne qq{none}){
					$spanningTree->{systemMacAddress} = strip_mac($3);
				}
			}
		}
	}
	$out->print_element(qq{spanningTree}, $spanningTree) if $spanningTree;
}

sub _parse_nat{
	my ($in, $out, $filters_found) = @_;
	my $virtual_servers = _get_vs_array_ref($in);
	my $nats = _get_nat_array_ref($in);
	my $snats = _get_snat_array_ref($in);
	
	my @big_array;
	push(@big_array, @$virtual_servers) if $virtual_servers;
	push(@big_array, @$nats) if $nats;
	push(@big_array, @$snats) if $snats;

	if(@big_array){
		if(not $filters_found){
			$out->open_element(qq{filterLists});
			$filters_found = 1;
		}
		my $vlan_names = _get_vlan_names($in);	
		my $vlan_lists = get_vlan_groups($in, \@big_array, $vlan_names);
		print_nat_sections($out, $virtual_servers, $nats, $snats, $vlan_lists);
	}
	if($filters_found){
		$out->close_element(qq{filterLists});
	}
}

sub _get_vs_array_ref{
	my ($in) = @_;
	$LOGGER->debug(qq{::_get_vs_array_ref; Parsing virtual servers});
	my $pools = _parse_pools($in);
	my $snatpools = _parse_snat_pools($in);
	my $self_ips = _parse_self_ips($in);
	my $ip_regex = get_crep(qq{cipm});
	my @virtuals;

	foreach my $virtual(split(/^VIRTUAL\s+/mi, $in->{virtuals})){
		if($virtual =~ /^\s*DISABLE\s*$/mi){
			next;
		}
		my $virtual_hash;
		if($virtual =~ /\A(\S+)\s*\{\s*$/mi){
			$virtual_hash->{name} = $1;
		}else{ next; }
		my $multiple_vlans = 0;
		foreach my $line(split(/^/, $virtual)){
			if($line =~ /^\s*(POOL|REJECT|FASTHTTP|(IP\s+FORWARD)|(L2\s+FORWARD)|FASTL4)\s*$/i){
				$virtual_hash->{type} = lc($1);
				if($virtual_hash->{type} eq q{l2 forward}){
					$LOGGER->debug(qq{::_get_vs_array_ref; Ignoring virtual server: layer 2 forwarding virutal servers not supported});
				}
			}elsif($multiple_vlans){
				if($line =~ /^\s*\}\s*(ENABLE|DISABLE)\s*$/i){
					$virtual_hash->{vlans_enabled} = lc($1) eq qq{enable} ? 1 : 0;
					$multiple_vlans = 0;
				}elsif($line =~ /^\s*(\S+)\s*$/){
					if($virtual_hash->{vlans}){
						$virtual_hash->{vlans} .= qq{ } . $1;
					}else{
						$virtual_hash->{vlans} = $1;
					}
				}
			}elsif($line =~ /^\s*DESTINATION\s+($ip_regex|ANY)(\%(\d+))?:(\S+)\s*$/i){
				if($2){
					$LOGGER->debug(qq{::_get_vs_array_ref; Ignoring virtual server: route domains not supported});
					$virtual_hash = undef;
					last;
				}
				if(lc($1) ne qq{any}){
					$virtual_hash->{destination} = $1;
				}else{
					$virtual_hash->{destination} = qq{0.0.0.0};
				}
				if($virtual_hash->{destination} eq qq{0.0.0.0}){
					$virtual_hash->{no_reverse} = 1;
				}
				if(not $virtual_hash->{service}){
					$virtual_hash->{service} = lc($4);
				}
			}elsif($line =~ /^\s*MASK\s+($ip_regex)\s*$/i){
				$virtual_hash->{mask} = mask_to_bits($1);
			}elsif($line =~ /^\s*IP\s+PROTOCOL\s+(\S+)\s*$/i){
				$virtual_hash->{protocol} = $1;
			}elsif($line =~ /^\s*POOL\s+(\S+)\s*$/i){
				$virtual_hash->{pool} = $1;
			}elsif($line =~ /^\s*VLANS\s+((\S+)\s+(ENABLE|DISABLE)|\{)\s*$/i){
				if($1 eq qq{\{}){
					$multiple_vlans = 1;
				}else{
					$virtual_hash->{vlans} = $2;
					$virtual_hash->{vlans_enabled} = lc($3) eq qq{enable} ? 1 : 0;
				}
			}elsif($line =~ /^\s*SNATPOOL\s+(\S+)\s*$/i){
				if(lc($1) ne q{none}){
					$virtual_hash->{snatpool} = $1;
				}
			}elsif($line =~ /^\s*SNAT\s+AUTOMAP\s*$/i){
				$virtual_hash->{automap} = 1;
			}elsif($line =~ /^\s*TRANSLATE\s+SERVICE\s+DISABLE\s*$/i){
				$virtual_hash->{service} = q{any};
			}
		}
		if ($virtual_hash){
			if($virtual_hash->{type} eq qq{pool} || $virtual_hash->{type} eq qq{fastl4} || $virtual_hash->{type} eq qq{fasthttp}){
				if($pools && $pools->{$virtual_hash->{pool}}){
					my $pool = $pools->{$virtual_hash->{pool}};
					foreach my $node(@{$pool->{member}}){
						if($node->{enabled}){
							if(($virtual_hash->{service} eq qq{any}) or ($node->{service} eq qq{any}) or ($virtual_hash->{service} eq $node->{service})){
								my $already;
								foreach my $hash_node(@{$virtual_hash->{nodes}}){
									if($hash_node eq $node->{ip}){
										$already = 1;
									}
								}
								if(not $already){
									push(@{$virtual_hash->{nodes}}, $node->{ip});
								}
							}
						}
					}
				}
			}
			if($virtual_hash->{snatpool}){
				if($snatpools && $snatpools->{$virtual_hash->{snatpool}}){
					my $snatpool = $snatpools->{$virtual_hash->{snatpool}};
					foreach my $member(@{$snatpool->{members}}){
						push(@{$virtual_hash->{snat_nodes}}, $member);
					}
				}
			}
			if($virtual_hash->{automap}){
				if($self_ips){
					my @keys = keys(%$self_ips);
					foreach my $key(@keys){
						foreach my $address(@{$self_ips->{$key}}){
							push(@{$virtual_hash->{snat_nodes}}, $address->{network} . qq{/} . mask_to_bits($address->{mask}));
						}
					}
				}
			}
			if($virtual_hash->{snat_nodes}){
				$virtual_hash->{source}->{ip} = q{0.0.0.0};
				$virtual_hash->{source}->{mask} = q{0};
				$virtual_hash->{no_reverse} = 1;
			}
			push(@virtuals, $virtual_hash) if($virtual_hash);
		}
	}
	if(@virtuals){
		return \@virtuals;
	}
	return undef;
}

sub _get_nat_array_ref{
	my ($in) = @_;
	$LOGGER->debug(qq{::_get_nat_array_ref; Parsing NAT});
	my $ip_regex = get_crep(qq{cipm});
	my @nats;
	foreach my $nat(split(/^NAT\s+/mi, $in->{nats})){
		if($nat =~ /^\s*DISABLE\s*$/mi){
			next;
		}
		my $nat_hash;
		my $multiple_VLANS = 0;
		foreach my $line(split(/^/, $nat)){
			if($multiple_VLANS){
				if($line =~ /^\s*(\S+)\s*$/){
					if($nat_hash->{vlans}){
						$nat_hash->{vlans} .= qq{ } . $1;
					}else{
						$nat_hash->{vlans} = $1;
					}
				}elsif($line =~ /^\s*\}\s+(ENABLE|DISABLE)\s*$/i){
					$multiple_VLANS = 0;
					$nat_hash->{vlans_enabled} = lc($1) eq qq{enable} ? 1 : 0;
				}
			}elsif($line =~ /^\s*($ip_regex)(\%(\d+))?\s+TO\s+($ip_regex)(\%(\d+))?\s+\{\s*$/i){
				if($2 || $5){
					$LOGGER->debug(qq{::_get_nat_array_ref; Ignoring NAT: route domains not supported});
					$nat_hash = undef;
					last;
				}
				$nat_hash->{source} = $1;
				$nat_hash->{destination} = $4;
			}elsif($line =~ /^\s*VLANS\s+(\{|(\S+)\s+(ENABLE|DISABLE))\s*$/i){
				if($1 eq qq{\{}){
					$multiple_VLANS = 1;
				}else{
					$nat_hash->{vlans} = $2;
					$nat_hash->{vlans_enabled} = lc($3) eq qq{enable} ? 1 : 0;
				}
			}
		}
		push(@nats, $nat_hash) if($nat_hash);
	}

	if(@nats){
		return \@nats;
	}
	return undef;
}

sub _get_snat_array_ref{
	my ($in) = @_;
	$LOGGER->debug(qq{::_get_snat_array_ref; Parsing SNAT});
	my $ip_regex = get_crep(qq{cipm});
	my $snat_pools = _parse_snat_pools($in);
	my $self_ips =  _parse_self_ips($in);

	my $anyProtocol = 0;
	if($in->{snats_any_protocol} =~ /ENABLE/i){
		$anyProtocol = 1;
	}

	my @snats;
	foreach my $snat(split(/^SNAT\s+/mi, $in->{snats})){
		my $snat_hash;
		if($snat =~ /\A(\S+)\s*\{\s*$/mi){
			$snat_hash->{name} = $1;
		}else{ next; }
		my $multiple_origins = 0;
		my $multiple_vlans = 0;
		foreach my $line(split(/^/, $snat)){
			if($multiple_origins){
				if($line =~ /^\s*\}\s*$/i){
					$multiple_origins = 0;
				}elsif($line =~ /^\s*(\S+)\s*$/i){
					push(@{$snat_hash->{origins}}, $1);
					if($1 =~ /\%\d+/){
				 		$LOGGER->debug(qq{::_get_snat_array_ref; Ignoring SNAT: route domains not supported});
						$snat_hash = undef;
						last;
					}
				}
			}elsif($multiple_vlans){
				if($line =~ /^\s*\}\s*(ENABLE|DISABLE)\s*$/i){
					$snat_hash->{vlans_enabled} = lc($1) eq qq{enable} ? 1 : 0;
					$multiple_vlans = 0;
				}elsif($line =~ /^\s*(\S+)\s*$/i){
					if($snat_hash->{vlans}){
						$snat_hash->{vlans} .= qq{ } . $1;
					}else{
						$snat_hash->{vlans} = $1;
					}
				}
			}elsif($line =~ /^\s*(AUTOMAP|SNATPOOL|TRANSLATION)\s*$/i){
				$snat_hash->{type} = lc($1);
			}elsif($line =~ /^\s*TRANSLATION\s+($ip_regex)(\%(\d+))?\s*$/i){
				if($2){
					$LOGGER->debug(qq{::_get_snat_array_ref; Ignoring SNAT: route domains not supported});
					$snat_hash = undef;
					last;
				}
				push(@{$snat_hash->{translations}}, $1);
			}elsif($line =~ /^\s*SNATPOOL\s+(\S+)\s*$/i){
				if(lc($1) ne qq{none}){
					$snat_hash->{snatpool} = $1;
				}
			}elsif($line =~ /^\s*ORIGINS\s+((\S+)|\{|DEFAULT\s+INET)\s*$/i){
				if($1 eq qq{\{}){
					$multiple_origins = 1;
				}elsif(lc($1) eq qq{default inet}){
					$snat_hash->{no_reverse} = 1;
					push(@{$snat_hash->{origins}}, qq{0.0.0.0/0});
				}else{
					if($2 eq qq{0.0.0.0/0}){
						$snat_hash->{no_reverse} = 1;
					}
					push(@{$snat_hash->{origins}}, $2);
				}
				if($1 =~ /\%\d+/){
					$LOGGER->debug(qq{::_get_snat_array_ref; Ignoring SNAT: route domains not supported});
					$snat_hash = undef;
					last;
				}
			}elsif($line =~ /^\s*VLANS\s+((\S+)\s+(ENABLE|DISABLE)|\{)\s*$/i){
				if($1 eq qq{\{}){
					$multiple_vlans = 1;	
				}else{
					$snat_hash->{vlans} = $2;
					$snat_hash->{vlans_enabled} = lc($3) eq qq{enable} ? 1 : 0;
				}
			}
		}

		if(defined $snat_hash){
			if($anyProtocol){
				$snat_hash->{protocol} = 1;	
			}

			if($snat_hash->{type} eq qq{snatpool}){
				if($snat_pools->{$snat_hash->{snatpool}}){
					$snat_hash->{translations} = $snat_pools->{$snat_hash->{snatpool}}->{members};
				}else{
					$snat_hash = undef;
				}
			}elsif($snat_hash->{type} eq qq{automap}){
				if($self_ips){
					my @keys = keys(%$self_ips);
					my @translations;
					foreach my $key(@keys){
						foreach my $address(@{$self_ips->{$key}}){
							push(@translations, $address->{network} . qq{/} . mask_to_bits($address->{mask}));
						}
					}
					$snat_hash->{translations} = \@translations;
				}else{
					$snat_hash = undef;
				}
			}

			push(@snats, $snat_hash) if(defined $snat_hash);
		}
	}
	if(@snats){
		return \@snats;
	}
	return undef;
}

sub print_nat_sections{
	my ($out, $vs_ref, $nats_ref, $snats_ref, $vlan_lists) = @_;
	my @vlan_lists = @$vlan_lists;

	foreach my $direction(qq{source}, qq{destination}){
		my $listCount = 0;
		$out->open_element(qq{filterListsSection});
		for(my $i = 0; $i <= $#vlan_lists; ++$i){
			my $list_opened = 0;
			my $entryCount = 0;
			if($vs_ref){
			foreach my $virtual_server(@$vs_ref){
				if($virtual_server->{list} == $i){
					if((not $virtual_server->{nodes} and (not $virtual_server->{type} =~ /forward/i)) or $virtual_server->{type} eq qq{reject}){
						next;
					}
					if($direction eq qq{source}){
						if($virtual_server->{no_reverse}){
							$LOGGER->debug(qq{::print_nat_sections; Ignoring reverse VS: Any, 0.0.0.0/0, address not supported for translation});
							next;
						}
					}
					if(not $list_opened){
						open_nat_list($out, $vlan_lists, $i);
						$list_opened = 1;
					}
					$out->open_element(qq{filterEntry});
					my $translation;
					$translation->{translateAppliedAddressing} = qq{true};
					if($direction eq qq{destination}){
						if($virtual_server->{type} !~ /forward/i){
							my @nodes = @{$virtual_server->{nodes}};
							foreach my $node(@nodes){
								my $translationIP;
								$translationIP->{translationIP}->{host} = $node;
								push(@{$translation->{translationAddress}}, $translationIP);
							}
							$out->print_element(qq{destAddressTranslation}, $translation);
						}
						$out->print_element(qq{destinationIpAddr}, get_host_or_network($virtual_server->{destination}, $virtual_server->{mask}));
						if($virtual_server->{service} ne qq{any}){
							$out->print_element(qq{destinationService}, get_port_expression($virtual_server->{service}));
						}
					}elsif($direction eq qq{source}){
						if($virtual_server->{snat_nodes}){
							my @nodes = @{$virtual_server->{snat_nodes}};
							foreach my $node(@nodes){
								my ($ip, $mask) = split(q{/}, $node);
								$out->print_element(qq{destinationIpAddr}, get_host_or_network($ip, $mask));
							}
						}
						if($virtual_server->{source}){
							my $translation;
							$translation->{translationAddress}->{translationIP} = get_host_or_network($virtual_server->{source}->{ip}, $virtual_server->{source}->{mask});
							$out->print_element(qq{destAddressTranslation}, $translation);
						}
					}

					$out->print_element(qq{log}, qq{false});
					$out->print_element(qq{name}, $virtual_server->{name});
					if($virtual_server->{type} =~ /forward/i){
						$out->print_element(qq{primaryAction}, qq{forward});
					}else{
						$out->print_element(qq{primaryAction}, qq{permit});
					}
					$out->print_element(qq{processOrder}, ++$entryCount);
					if ($virtual_server->{protocol} && $virtual_server->{protocol} ne qq{any}){
						$out->print_element(qq{protocol}, $virtual_server->{protocol});
					}

					my $sourceTranslation;
					$sourceTranslation->{translateAppliedAddressing} = qq{true};
					if($direction eq qq{source}){
						if($virtual_server->{service} ne qq{any}){
							$sourceTranslation->{translationAddress}->{translationServicePort}->{translationPort} = get_port_expression($virtual_server->{service});
							if($virtual_server->{protocol}){
								if($virtual_server->{protocol} eq qq{any}){
									$sourceTranslation->{translationAddress}->{translationServicePort}->{translationProtocol} = q{ip};
								}else{
									$sourceTranslation->{translationAddress}->{translationServicePort}->{translationProtocol} = $virtual_server->{protocol};
								}
							}
						}
						if($virtual_server->{type} !~ /forward/i){
							$sourceTranslation->{translationAddress}->{translationIP} = get_host_or_network($virtual_server->{destination}, $virtual_server->{mask});
							$out->print_element(qq{sourceAddressTranslation}, $sourceTranslation);
							my @nodes = @{$virtual_server->{nodes}};
							foreach my $node(@nodes){
								my $sourceIpAddr;
								$sourceIpAddr->{host} = $node;
								$out->print_element(qq{sourceIpAddr}, $sourceIpAddr);
							}
						}else{
							if($virtual_server->{service} ne qq{any}){
								$out->print_element(qq{sourceAddressTranslation}, $sourceTranslation);
							}
							$out->print_element(qq{sourceIpAddr}, get_host_or_network($virtual_server->{destination}, $virtual_server->{mask}));
						}
					}elsif($direction eq qq{destination}){
						if($virtual_server->{snat_nodes}){
							my @nodes = @{$virtual_server->{snat_nodes}};
							foreach my $node(@nodes){
								my ($ip, $mask) = split(q{/}, $node);
								my $translationIP;
								$translationIP->{translationIP} = get_host_or_network($ip, $mask);
								push(@{$sourceTranslation->{translationAddress}}, $translationIP);
							}
							$out->print_element(qq{sourceAddressTranslation}, $sourceTranslation);
						}
						if($virtual_server->{source}){
							$out->print_element(qq{sourceIpAddr}, get_host_or_network($virtual_server->{source}->{ip}, $virtual_server->{source}->{mask}));
						}
					}
					$out->close_element(qq{filterEntry});
				}
			}}
			if($list_opened){
				$out->print_element(qq{mode}, qq{stateful});
				$out->print_element(qq{name}, qq{Virtual_Server_List_} . ++$listCount . qq{_} . $direction);
				$out->print_element(qq{processingPhase}, {qq{core:packetProcessingPhase} => qq{address-translation}});
				$out->close_element(qq{filterList});
			}
		}
		
		$listCount = 0;
		for(my $i = 0; $i <= $#vlan_lists; ++$i){
			my $list_opened = 0;
			my $entryCount = 0;
			if($snats_ref){
			foreach my $snat(@$snats_ref){
				if($snat->{list} == $i){
					if($direction eq qq{destination}){
						if($snat->{no_reverse}){
							$LOGGER->debug(qq{::print_nat_sections; Ignoring reverse SNAT: Any, 0.0.0.0/0, address not supported for translation});
							next;
						}
					}
					if(not $list_opened){
						open_nat_list($out, $vlan_lists, $i);
						$list_opened = 1;
					}
					$out->open_element(qq{filterEntry});

					my @translations; my @origins;
					if($direction eq qq{source}){
						@translations = @{$snat->{translations}} if $snat->{translations};
						@origins = @{$snat->{origins}} if $snat->{origins};
					}elsif($direction eq qq{destination}){
						@translations = @{$snat->{origins}} if $snat->{origins};
						@origins = @{$snat->{translations}} if $snat->{translations};
					}

					my $translationElement;
					$translationElement->{translateAppliedAddressing} = qq{true};
					foreach my $translation(@translations){
						my $translationIP;
						my ($ip, $mask) = split(/\//, $translation);
						$translationIP->{translationIP} = get_host_or_network($ip, $mask);
						push(@{$translationElement->{translationAddress}}, $translationIP);
					}

					if($direction eq qq{destination}){
						$out->print_element(qq{destAddressTranslation}, $translationElement);
						foreach my $origin(@origins){
							my ($ip, $mask) = split(/\//, $origin);
							$out->print_element(qq{destinationIpAddr}, get_host_or_network($ip, $mask));
						}
					}

					$out->print_element(qq{log}, qq{false});
					$out->print_element(qq{primaryAction}, qq{permit});
					$out->print_element(qq{processOrder}, ++$entryCount);

					if(not $snat->{protocol}){
						$out->print_element(qq{protocol}, qq{tcp});
						$out->print_element(qq{protocol}, qq{udp});
					}

					if($direction eq qq{source}){
						$out->print_element(qq{sourceAddressTranslation}, $translationElement);
						foreach my $origin(@origins){
							my ($ip, $mask) = split(/\//, $origin);
							$out->print_element(qq{sourceIpAddr}, get_host_or_network($ip, $mask));
						}
					}

					$out->close_element(qq{filterEntry});
				}
			}}
			if($list_opened){
				$out->print_element(qq{mode}, qq{stateful});
				$out->print_element(qq{name}, qq{SNAT_List_} . ++$listCount . qq{_} . $direction);
				$out->print_element(qq{processingPhase}, {qq{core:packetProcessingPhase} => qq{address-translation}});
				$out->close_element(qq{filterList});
			}
		}
		$listCount = 0;
		for(my $i = 0; $i <= $#vlan_lists; ++$i){
			my $list_opened = 0;
			my $entryCount = 0;
			if($nats_ref){
			foreach my $nat(@$nats_ref){
				if($nat->{list} == $i){
					if(not $list_opened){
						open_nat_list($out, $vlan_lists, $i);
						$list_opened = 1;
					}
					$out->open_element(qq{filterEntry});
					my $original; my $translation;
					$translation->{translateAppliedAddressing} = qq{true};
					if($direction eq qq{destination}){
						$translation->{translationAddress}->{translationIP}->{host} = $nat->{destination};
						$out->print_element(qq{destAddressTranslation}, $translation);
						$original->{host} = $nat->{source};
						$out->print_element(qq{destinationIpAddr}, $original);
					}

					$out->print_element(qq{log}, qq{false});
					$out->print_element(qq{primaryAction}, qq{permit});
					$out->print_element(qq{processOrder}, ++$entryCount);

					if($direction eq qq{source}){
						$translation->{translationAddress}->{translationIP}->{host} = $nat->{source};
						$out->print_element(qq{sourceAddressTranslation}, $translation);
						$original->{host} = $nat->{destination};
						$out->print_element(qq{sourceIpAddr}, $original);
					}

					$out->close_element(qq{filterEntry});
				}
			}}
			if($list_opened){
				$out->print_element(qq{mode}, qq{stateful});
				$out->print_element(qq{name}, qq{NAT_List_} . ++$listCount . qq{_} . $direction);
				$out->print_element(qq{processingPhase}, {qq{core:packetProcessingPhase} => qq{address-translation}});
				$out->close_element(qq{filterList});
			}
		}

		my $phaseDiscriminator;
		$phaseDiscriminator->{processingPhase}->{qq{core:packetProcessingPhase}} = qq{routing};
		$phaseDiscriminator->{processingPhasePosition} = qq{before};
		$out->print_element(qq{phaseDiscriminator}, $phaseDiscriminator);
		$out->close_element(qq{filterListsSection});
	}
}

sub open_nat_list{
	my ($out, $vlan_lists, $i) = @_;
	my @vlan_lists = @$vlan_lists;
	$out->open_element(qq{filterList});
	$out->print_element(qq{defaultAction}, qq{none});
	foreach my $vlan(@{$vlan_lists[$i]}){
		foreach my $direction(qq{source}, qq{destination}){
			my $appliesTo = {$direction => $vlan};
			$out->print_element(qq{appliesTo}, $appliesTo);
		}
	}
}

sub get_vlan_groups{
	my ($in, $hashes, $vlan_names) = @_;
	my @hashes = @$hashes;
	my @vlan_names = @$vlan_names;
	my @lists;

	# CREATE LIST OF LISTS OF VLAN AND ASSIGN INDEX OF RELEVENT LIST TO EACH HASH
	foreach my $hash(@hashes){
		
		# WORK OUT VLANS TO BE ENABLED / CONVERT DISABLED LIST TO ENABLED LIST
		my @vlans;
		if($hash->{vlans_enabled} && $hash->{vlans}){
			foreach my $vlan(split(/\s/, $hash->{vlans})){
				push(@vlans, $vlan);
			}
		}else{
			foreach my $vlan(@vlan_names){
				my $disabled = 0;
				if($hash->{vlans}){
					foreach my $disabled_vlan(split(/\s/, $hash->{vlans})){
						if($disabled_vlan eq $vlan){
							$disabled = 1;
						}
					}
				}
				push(@vlans, $vlan) if (not $disabled);
			}
		}

		# COMPARING HASH->VLANS TO LISTS OF VLANS TO SEE IF A LIST EXISTS FOR THAT COMBINATION OF VLANS OR NEED TO CREATE ONE
		if(@vlans){
			if(@lists){
				my $any_same_list = 0;
				my $list_count = -1;
				foreach my $list_ref(@lists){
					++$list_count;
					my @list = @{$list_ref};
					if($#list == $#vlans){
						my $same_list = 1;
						foreach my $list_vlan(@list){
							my $found = 0;
							foreach my $vlan(@vlans){
								if($vlan eq $list_vlan){
									$found = 1;
								}
							}
							if(not $found){
								$same_list = 0;
							}
						}
						if($same_list){
							$any_same_list = 1;
							$hash->{list} = $list_count;
						}
					}
				}
				if(not $any_same_list){
					push(@lists, \@vlans);
					$hash->{list} = $#lists;
				}
			}else{
				push(@lists, \@vlans);
				$hash->{list} = 0;
			}
		}else{
			$hash->{list} = -1;
		}
	}
	return (\@lists);
}

sub _parse_snat_pools{
	my ($in) = @_;
	my $ip_regex = get_crep(qq{cipm});
	my $snat_pools;

	foreach my $snat_pool(split(/^SNATPOOL\s+/mi, $in->{snat_pools})){
		my $snat_pool_name;
		if($snat_pool =~ /\A(\S+)\s*\{\s*$/mi){
			$snat_pool_name = $1;
		}else{ next; }
		my $multiple_members = 0;
		foreach my $line(split(/^/, $snat_pool)){
			if($multiple_members){
				if($line =~ /^\s*\}\s*$/){
					$multiple_members = 0;
				}elsif($line =~ /^\s*($ip_regex)(\%(\d+))?\s*$/){
					if($2){
						$LOGGER->debug(qq{::_parse_snat_pools; Ignoring SNAT pool member: route domains not supported});
						last;
					}
					push(@{$snat_pools->{$snat_pool_name}->{members}}, $1);
				}
			}elsif($line =~ /^\s*MEMBERS\s+($ip_regex|\{)(\%(\d+))?\s*$/i){
				if($2){
					$LOGGER->debug(qq{::_parse_snat_pools; Ignoring SNAT pool member: route domains not supported});
					last;
				}
				if($1 eq qq{\{}){
					$multiple_members = 1;
				}else{
					push(@{$snat_pools->{$snat_pool_name}->{members}}, $1);
				}
			}
		}
	}

	return $snat_pools;
}

sub get_host_or_network{
	my ($ip, $mask) = @_;
	my $ipHash;
	if(not defined $mask or $mask eq qq{32}){
		$ipHash->{host} = $ip;
	}else{
		 $ipHash->{network}->{address} = $ip;
		 $ipHash->{network}->{mask} = $mask;
	}
	return $ipHash;
}

sub _get_vlan_full_names{
	my $in = shift;
	my @vlan_full_names;
	foreach my $vlan(split(/^\s*VLAN\s+/mi, $in->{vlans})){
		if($vlan =~ /^(\S+)\s+(\((\S+)\)\s+)?TAG\s+(\d+)/mi){
			my $vlan_full_name = {
				fullname => $1,
				shortname => $3,
				tag => $4
			};
			push(@vlan_full_names, $vlan_full_name);
		}
	}
	my $command = 0;
	foreach my $vlan_group(split(/^\s*VLANGROUP\s+/mi, $in->{vlan_groups})){
		if($command eq 0){
			$command = 1;
			next;
		}
		if($vlan_group =~ /\A(\S+)\s+(\((\S+)\)\s+)?/mi){
			if(lc($1) ne q{all}){
				my $vlan_full_name = {
					fullname => $1,
					shortname => $3
				};
				push(@vlan_full_names, $vlan_full_name);
			}
		}
	}
	return \@vlan_full_names;
}

sub _get_vlan_names{
	my $in = shift;
	my @vlan_names;
	foreach my $vlan(split(/^\s*VLAN\s+/mi, $in->{vlans})){
		if($vlan =~ /^(\S+)\s+(\((\S+)\)\s+)?TAG\s+/mi){
			push(@vlan_names, $1);
		}
	}
	foreach my $vlan_group(split(/^VLANGROUP\s+/mi, $in->{vlan_groups_list})){
		if($vlan_group =~ /\A(\S+)\s+\{\s*$/mi){
			if($1 ne 'all'){
				push(@vlan_names, $1);
			}
		}
	}
	return \@vlan_names;
}

sub generate_packet_processing_declaration{
	my $out = shift;
	$out->open_element(qq{core:packetProcessingOrdering});
	foreach my $pRPhase(qq{packet-filtering}, qq{address-translation}, qq{routing}){
		my ($pPhase, $pPosition) = split(qq{/}, $pRPhase);
		$out->open_element(qq{core:packetProcessingElement});
		$out->print_element(qq{core:packetProcessingPhase}, $pPhase);
		if (defined $pPosition){
			$out->print_element(qq{core:processingPoint}, $pPosition);
		}
		$out->close_element(qq{core:packetProcessingElement});
	}
	$out->close_element(qq{core:packetProcessingOrdering});
}

sub full_int_name{
	# given a short name like "Fa3/15" return "FastEthernet3/15"
	my $name = trim(shift);
	for($name){
		s/^Fa(?=\d)/FastEthernet/;
		s/^Eth(?=\d)/Ethernet/;
		s/^Gig(?=\d)/GigabitEthernet/;
	}
	return $name;
}

sub _createCPUTypesRE{
	my @cpuTypes = qw(Pentium Celeron Xeon XeonMP Itanium Athlon AthlonFX Opteron OpteronMP Duron Sempron);
	'\b(' . join( '|', @cpuTypes ) . ')';
}

sub logit{
	my $s = shift;
	my ($logsec,$logmin,$loghour,$logmday,$logmon,$logyear,$logwday,$logyday,$logisdst)=localtime(time);
	my $logtimestamp = sprintf("%4d-%02d-%02d %02d:%02d:%02d",$logyear+1900,$logmon+1,$logmday,$loghour,$logmin,$logsec);
	$logmon++;
	my $logfile="/tmp/logfile.log";
	my $fh;
	open($fh, '>>', "$logfile") or die "$logfile: $!";
	print $fh "$logtimestamp $s\n";
	close($fh);
}

1;
