package ZipTie::Adapters::F5::BigIP::TMSHParsers;

use strict;
use warnings;
use ZipTie::Adapters::Utils qw{mask_to_bits get_crep strip_mac trim get_interface_type get_port_expression};
use ZipTie::Adapters::F5::BigIP::Parsers qw{generate_packet_processing_declaration parse_file_storage full_int_name get_host_or_network open_nat_list print_nat_sections get_vlan_groups print_packet_filter_virtual_servers};
use ZipTie::Logger;
use Data::Dumper;
use Exporter qw{import};
our @EXPORT_OK = qw{parse_tmsh_system parse_tmsh_chassis parse_tmsh_interfaces parse_tmsh_stp parse_tmsh_static_routes parse_tmsh_vlans parse_tmsh_filter_enabled parse_tmsh_default_filter parse_tmsh_filters};
my $LOGGER = ZipTie::Logger::get_logger();

sub parse_tmsh_system{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_tmsh_system; Parsing system});
	if($in->{hostname} =~ /^\s*HOSTNAME\s+(\S+)\s*$/mi){
		$out->print_element(qq{core:systemName}, $1);
		$in->{hostname_value} = $1;
	}

	$out->open_element(qq{core:osInfo});
	$out->print_element(qq{core:make}, qq{F5} );
	$out->print_element(qq{core:name}, qq{BIG-IP});
	if($in->{version} =~ /^\s*VERSION\s+(\S+)\s*$/mi){
		$out->print_element(qq{core:version}, $1);
	}
	$out->print_element(qq{core:osType}, qq{BIG-IP});
	generate_packet_processing_declaration($out);
	$out->close_element(qq{core:osInfo});

	my $biosVersion = qq{Virtual Machine};
	if($in->{hardware} =~ /^\s*BIOS\s+REVISION\s+(.+)\s*$/mi){
		$biosVersion = $1;
		if($biosVersion =~ /BASE\s+MAC/i){
			$biosVersion = qq{Virtual Machine};
		}
	}
	$out->print_element(qq{core:biosVersion}, $biosVersion);
	$out->print_element(qq{core:deviceType}, qq{Load Balancer});

	if($in->{contact} =~ /^\s*SYS\-CONTACT\s+"(.+)<(.+)>"\s*$/mi){
		$out->print_element(qq{core:contact}, $1 . $2);
	}
}

sub parse_tmsh_chassis{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_tmsh_chassis; Parsing chassis});
	my $mac_regex = get_crep(qq{mac1}) . qq{|} . get_crep(qq{mac2}) . qq{|} .  get_crep(qq{mac3}) . qq{|} .  get_crep(qq{mac4});
	$out->open_element(qq{chassis});

	my $chassisAsset = {qq{core:assetType} => qq{Chassis}};
	$chassisAsset->{qq{core:factoryinfo}}->{qq{core:make}} = qq{F5 Networks};
	if($in->{hardware} =~ /CHASSIS\s+SERIAL\s+(\S+)/i){
		my $serial = $1;
		my $hex = qq{[0-9a-f]};
		if($serial =~ /^\s*$hex{8}-($hex{4}-)+$hex{12}\s*$/){
			$chassisAsset->{qq{core:factoryinfo}}->{qq{core:serialNumber}} = $serial;
		}else{
			$chassisAsset->{qq{core:factoryinfo}}->{qq{core:serialNumber}} = $serial . qq{_} . $in->{hostname_value};
		}
	}
	if($in->{hardware} =~ /^\s*SYSTEM\s+INFORMATION\s*\n?\s*TYPE\s+(\S+)\s*$/mi){
		$chassisAsset->{qq{core:factoryinfo}}->{qq{core:modelNumber}} = qq{BIG-IP }.$1;
	}
	$out->print_element(qq{core:asset}, $chassisAsset);

	my $cpu;
	if($in->{dmesg} =~ /^CPU\d+:\s+(\S.+HZ)/mi){
		$cpu->{qq{core:description}} = $1;
		my $cputRE = _createCPUTypesRE();
		if ($cpu->{qq{core:description}} =~ /($cputRE)/i){
			$cpu->{cpuType} = $1;
		}
	}
	$out->print_element(qq{cpu}, $cpu ) if defined $cpu;

	parse_file_storage($in, $out);

	if($in->{hardware} =~ /BASE\s+MAC\s+($mac_regex)/i){
		my $baseMac = strip_mac($1);
		my $padding = 12 - length($baseMac);
		for(my $i = 0; $i < $padding; ++$i){
			$baseMac = q{0} . $baseMac;
		}
		$out->print_element(qq{macAddress}, $baseMac);
	}

	my @memories = ();
	if($in->{memory} =~ /TOTAL\s+PHYS\s+MEMORY\s+([\d\.]+)([MGK])/i){
		my $memSize = $1;
		if(uc($2) eq qq{K}){
			$memSize *= 1024;
		}elsif(uc($2) eq qq{M}){
			$memSize *= 1024 * 1024;
		}elsif(uc($2) eq qq{G}){
			$memSize *= 1024 * 1024 * 1024;
		}
		my $roundedMemSize = int($memSize + 0.5);
		push(@memories, {qq{core:description} => qq{RAM}, kind => qq{RAM}, size => $roundedMemSize});
	}
	foreach (@memories){
		$out->print_element(qq{memory}, $_);
	}

	$out->close_element(qq{chassis});
}

sub parse_tmsh_interfaces{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_tmsh_interfaces; Parsing interfaces});
	my $ip_regex = get_crep(qq{cipm});
	my $interface_found = 0;

	foreach my $if_part(split(/^(?!\s)/m, $in->{interfaces})){
		my $interface;

		foreach my $if_line(split(/^/, $if_part)){
			if($if_line =~ /^(\d+):\s+(\S+):\s+<(\S+)>\s+MTU\s+(\d+)/mi ){
				$interface->{name} = $2;
				$interface->{description} = $2;
				$interface->{physical} = qq{false};
				$interface->{interfaceType} = lc($3);
				$interface->{adminStatus} = lc($3);
				$interface->{mtu} = $4;
				if($interface->{adminStatus} !~ /,UP,/i){
					$interface->{adminStatus} = qq{down};
				}else{
					$interface->{adminStatus} = qq{up};
				}
				$interface->{physical} = qq{false};
			}elsif($if_line =~ /\s*LINK\/(\S+)\s+(\S+)/mi){
				my $link_type = lc($1);
				if($link_type eq qq{ether}){
					my $mac = strip_mac($2);
					if($mac ne qq{000000000000}){
						$interface->{interfaceEthernet}->{macAddress} = $mac;
					}
				}
				if($interface->{interfaceType} =~ /POINTOPOINT,/i){
					$interface->{interfaceType} = q{tunnel};
				}else{
					$interface->{interfaceType} = get_interface_type($link_type);
				}
			}elsif ($if_line =~ /\s*INET\s+($ip_regex)\/(\d+)\s+(BRD\s+($ip_regex))?/mi){
				my $ipConfiguration;
				$ipConfiguration->{ipAddress} = $1;
				$ipConfiguration->{mask} = $2;
				if($4){ $ipConfiguration->{broadcast} = $4; }
				push(@{$interface->{interfaceIp}->{ipConfiguration}}, $ipConfiguration);
			}
		}
		if(defined $interface && $interface->{interfaceType} eq q{tunnel}){
			$interface = undef;
			# TUNNELS DONE LATER
		}
		if(defined $interface){
			if(not $interface_found){
				$interface_found = 1;
				$out->open_element(qq{interfaces});
			}

			if($interface->{interfaceIp}){
				my $all_loopbacks = 1;
				foreach my $ip_config(@{$interface->{interfaceIp}->{ipConfiguration}}){
					if(substr($ip_config->{ipAddress}, 0, 4) ne q{127.}){
						$all_loopbacks = 0; last;
					}
				}
				if($all_loopbacks){
					$interface->{interfaceType} = q{softwareLoopback};
				}
			}

			my @vlan_names = @{_get_tmsh_vlan_full_names($in)};
			foreach my $vlan_name(@vlan_names){
				if($interface->{name} =~ /\~\d+/){
					if($vlan_name->{shortname} eq $interface->{name}){
						$interface->{name} = $vlan_name->{fullname};
						redo;
					}
				}elsif($vlan_name->{fullname} eq $interface->{name}){
					$interface->{interfaceType} = q{virtual};
					if($vlan_name->{tag}){
						$interface->{interfaceVlanTrunks}->{startVlan} = $vlan_name->{tag};
					}
				}
			}
			$out->print_element(qq{interface}, $interface);
		}
	}
	$interface_found = _parse_tmsh_tunnels($in, $out, $interface_found);
	$interface_found = _parse_tmsh_ipsec($in, $out, $interface_found);
	$interface_found = _parse_tmsh_physical_interfaces($in, $out, $interface_found);
	if($interface_found){
		$out->close_element(qq{interfaces});
	}
}

sub _parse_tmsh_tunnels{
	my ($in, $out, $interface_found) = @_;
	my $ip_regex = get_crep(q{cipm});
	my $self_ips = _parse_tmsh_self_ips($in);

	foreach my $tunnel(split(/^NET\s+TUNNELS\s+TUNNEL\s+/mi, $in->{tunnels})){
		my $tunnel_hash;
		if($tunnel =~ /\A(\S+)\s+\{\s*$/mi){
			$tunnel_hash->{name} = $1;
			$tunnel_hash->{description} = $1;
			$tunnel_hash->{physical} = q{false};
			$tunnel_hash->{interfaceType} = q{tunnel};
			$tunnel_hash->{adminStatus} = q{up};
		}
		foreach my $line(split(/^/, $tunnel)){
			if($line =~ /^\s*LOCAL\-ADDRESS\s+(($ip_regex)(\%(\d+))?|ANY)\s*$/i){
				if($3){
					$LOGGER->debug(qq{Ignoring tunnel: route domains not supported});
					$tunnel_hash = undef;
					last;
				}
				if(lc($1) eq q{any}){
					$tunnel_hash->{interfaceTunnel}->{tunnelLocalPeering}->{tunnelLocalPeeringIPAddress} = q{0.0.0.0}
				}else{
					$tunnel_hash->{interfaceTunnel}->{tunnelLocalPeering}->{tunnelLocalPeeringIPAddress} = $2;
				}
			}elsif($line =~ /^\s*REMOTE\-ADDRESS\s+(($ip_regex)(\%(\d+))?|ANY)\s*$/i){
				if($3){
					$LOGGER->debug(qq{Ignoring tunnel: route domains not supported});
					$tunnel_hash = undef;
					last;
				}
				if(lc($1) eq q{any}){
					$tunnel_hash->{interfaceTunnel}->{tunnelRemotePeerIPAddress} = q{0.0.0.0}
				}else{
					$tunnel_hash->{interfaceTunnel}->{tunnelRemotePeerIPAddress} = $2;
				}
			}elsif($line =~ /^\s*PROFILE\s+(\S+)\s*$/i){
				my $protocol = lc($1);
				if($protocol eq q{ipip} || $protocol eq q{ip4ip4}){
					$protocol = q{ipv4-in-ipv4};
				}elsif($protocol eq q{ip6ip4}){
					$protocol = q{ipv6-in-ipv4};
				}elsif($protocol eq q{ip4ip6}){
					$protocol = q{ipv4-in-ipv6};
				}elsif($protocol eq q{ppp}){
					$protocol = q{pptp};
					$tunnel_hash = undef;
					last;
				}elsif($protocol ne q{gre}){
					$protocol = q{other};
				}
				$tunnel_hash->{interfaceTunnel}->{encapsulationProtocol} = $protocol;
			}
		}
		if(defined $tunnel_hash){
			if(not $interface_found){
				$interface_found = 1;
				$out->open_element(qq{interfaces});
			}

			if($self_ips && $self_ips->{$tunnel_hash->{name}}){
				foreach my $self_ip(@{$self_ips->{$tunnel_hash->{name}}}){
					my $ipConfiguration = {
						ipAddress => $self_ip->{network},
						mask => $self_ip->{mask}
					};
					push(@{$tunnel_hash->{interfaceIp}->{ipConfiguration}}, $ipConfiguration);
				}
			}
			$out->print_element(qq{interface}, $tunnel_hash);
		}
	}

	return $interface_found;
}

sub _parse_tmsh_ipsec{
	my ($in, $out, $interface_found) = @_;
	my $ip_regex = get_crep(q{cipm});

	foreach my $ipsec(split(/^NET\s+IPSEC\s+/mi, $in->{ipsec})){
		my $ipsec_hash;
		if($ipsec =~ /^\s*MODE\s+TUNNEL\s*$/mi 
		&& $ipsec =~ /\AIPSEC\-POLICY\s+(\S+)\s+\{\s*$/mi){
			$ipsec_hash->{name} = $1;
			$ipsec_hash->{description} = $1;
			$ipsec_hash->{physical} = q{false};
			$ipsec_hash->{interfaceType} = q{tunnel};
			$ipsec_hash->{adminStatus} = q{up};
			$ipsec_hash->{interfaceTunnel}->{encapsulationProtocol} = q{ipSec};
		}else{ next; }

		foreach my $line(split(/^/, $ipsec)){
			if($line =~ /^\s*TUNNEL\-LOCAL\-ADDRESS\s+($ip_regex)(\%(\d+))?\s*$/i){
				if($2){
					$LOGGER->debug(qq{Ignoring IPSEC tunnel: route domains not supported});
					$ipsec_hash = undef;
					last;
				}
				$ipsec_hash->{interfaceTunnel}->{tunnelLocalPeering}->{tunnelLocalPeeringIPAddress} = $1;
			}elsif($line =~ /^\s*TUNNEL\-REMOTE\-ADDRESS\s+($ip_regex)(\%(\d+))?\s*$/i){
				if($2){
					$LOGGER->debug(qq{Ignoring IPSEC tunnel: route domains not supported});
					$ipsec_hash = undef;
					last;
				}
				$ipsec_hash->{interfaceTunnel}->{tunnelRemotePeerIPAddress} = $1;
			}
		}

		if(defined $ipsec_hash){
			if(not $interface_found){
				$interface_found = 1;
				$out->open_element(qq{interfaces});
			}
			$out->print_element(qq{interface}, $ipsec_hash);
		}
	}
	return $interface_found;
}

sub _parse_tmsh_physical_interfaces{
	my ($in, $out, $interface_found) = @_;
	my $mac_regex = get_crep(qq{mac2});

	foreach my $if_part(split(/^NET\s+INTERFACE\s+/mi, $in->{physical_interfaces})){
		my $interface;
		if($if_part =~ /\A(\S+)\s+\{/m){
			$interface->{name} = $1;
			$interface->{description} = $1;
			$interface->{interfaceType} = qq{ethernet};
			$interface->{physical} = qq{true};
			$interface->{interfaceEthernet}->{autoSpeed} = qq{true};
			$interface->{interfaceEthernet}->{autoDuplex} = qq{true};
		}else{ next; }
		foreach my $if_line(split(/^/, $if_part)){
			if($if_line =~ /^\s*(ENABLED|DISABLED)\s*$/i){
				$interface->{adminStatus} = lc($1) eq q{enabled} ? q{up} : q{down};
			}elsif($if_line =~ /^\s*MAC\-ADDRESS\s+($mac_regex)\s*$/i){
				my $macAddress = strip_mac($1);
				my $padding = 12 - length($macAddress);
				for(my $i = 0; $i < $padding; ++$i){
					$macAddress = q{0} . $macAddress;
				}
				$interface->{interfaceEthernet}->{macAddress} = $macAddress;
			}elsif($if_line =~ /^\s*MTU\s+(\d+)\s*$/i){
				$interface->{mtu} = $1;
			}elsif($if_line =~ /^\s*MEDIA\-ACTIVE\s+(\d+)(\S+)\-(\S+)\s*$/i){
				$interface->{speed} = $1.qq{000000};
				$interface->{interfaceEthernet}->{mediaType} = $1 . q{base} . $2;
				$interface->{interfaceEthernet}->{operationalDuplex} = lc($3) eq q{fd} ? q{full} : q{half};
			}
		}
		if(defined $interface){
			my $count = 0;
			my @ifIndex = split(/^/, $in->{interface_snmp_index});
			foreach my $if_name(split(/^/, $in->{interface_snmp_names})){
				if($if_name =~ /STRING\s*:\s+(\S+)\s*$/i){
					if($interface->{name} eq $1){
						if($ifIndex[$count] =~ /INTEGER\s*:\s+(\d+)\s*$/i){
							$interface->{ifIndex} = $1;
						}
					}
				}
				++$count;
			}

			foreach my $vlan(split(/^NET\s+VLAN\s+/mi, $in->{vlans})){
				 if($vlan =~ /\A(\S+)\s+\{/mi){
					my $vlanName = $1;
					while($vlan =~ /^\s*(\S+)\s+\{/mig){
						if($1 ne $vlanName && lc($1) ne q{interfaces}){
							if($interface->{name} eq $1){
								my $vlanTrunk = { startVlan => $vlanName };
								push(@{$interface->{interfaceVlanTrunks}}, $vlanTrunk);
							}
						}
					}
				}
			}

			foreach my $trunk(split(/^NET\s+TRUNK\s+/mi, $in->{trunks})){
				my $trunk_name;
				if($trunk =~ /\A(\S+)\s*\{\s*$/m){
					$trunk_name = $1;
				}
				my $interfaces_opened;
				foreach my $line(split(/^/, $trunk)){
					if($interfaces_opened){
						if($line =~ /^\s*\}\s*$/){
							$interfaces_opened = 0;
						}elsif($line =~ /^\s*(\S+)\s*$/){
							if($interface->{name} eq $1){
								$interface->{interfaceGrouping}->{ifGroupName} = $trunk_name;
								$interface->{interfaceGrouping}->{ifGroupingCharacteristic} = q{bondingGroupMember};
							}
						}
					}elsif($line =~ /^\s*INTERFACES\s*\{\s*$/i){
						$interfaces_opened = 1;
					}
				}
			}

			if(not $interface_found){
				$interface_found = 1;
				$out->open_element(qq{interfaces});
			}
			$out->print_element(qq{interface}, $interface);
		}
	}
	return $interface_found;
}

sub parse_tmsh_stp{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_tmsh_stp; Parsing STP});
	my $spanningTree;
	my ($helloTime, $maxAge, $forwardDelay, $holdTime, $priority) = (0, 0, 0, 0);

	if($in->{stp_global} =~ /^\s*MODE\s+(\S+)/mi){
		$spanningTree->{mode} = $1;
	}
	if($in->{stp_global} =~ /^\s*FWD\-DELAY\s+(\d+)/mi){
		$forwardDelay = $1;
	}
	if($in->{stp_global} =~ /^\s*HELLO\-TIME\s+(\d+)/mi){
		$helloTime = $1;
	}
	if($in->{stp_global} =~ /^\s*MAX\-AGE\s+(\d+)/mi){
		$maxAge = $1;
	}
	if($in->{stp_global} =~ /^\s*TRANSMIT\-HOLD\s+(\d+)/mi){
		$holdTime = $1;
	}
	if($in->{stp_priority} =~ /^\s*PRIORITY\s+(\S+)/mi){
		$priority = $1;
	}

	foreach my $stpins_part(split(/^\s*NET\s+STP\s+(\d+)/mi, $in->{stp})){
		if($spanningTree->{mode} eq qq{mstp}){
			my $instance = {
				helloTime => $helloTime,
				maxAge => $maxAge,
				forwardDelay => $forwardDelay,
				holdTime => $holdTime,
				priority => $priority
			};
			if($stpins_part =~ /^\s*VLANS\s+(\{\s+(\S+)|(\S+))/mi){
				$instance->{vlan} = $2;
			}
			push(@{$spanningTree->{spanningTreeInstance}}, $instance);
		}else{
			$spanningTree->{helloTime} = $helloTime;
			$spanningTree->{maxAge} = $maxAge;
			$spanningTree->{forwardDelay} = $forwardDelay;
			$spanningTree->{priority} = $priority;
		}
	}
	$out->print_element(qq{spanningTree}, $spanningTree) if $spanningTree;
}

sub parse_tmsh_static_routes{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_tmsh_static_routes; Parsing static routes});
	my $ip_regex = get_crep(qq{cipm});
	my $pools = _parse_tmsh_pools($in);
	my $staticRoutes;

	foreach my $route(split(/^NET\s+ROUTE\s*/mi, $in->{static_routes_list})){
		my $static_route;
		my $type;
		foreach my $route_line(split(/^/, $route)){
			if($route_line =~ /^\s*NETWORK\s+(DEFAULT|(($ip_regex)(\%(\d+))?\/(\S+)))\s*$/i){
				if(lc($1) eq qq{default}){
					$static_route->{destinationAddress} = qq{0.0.0.0};
					$static_route->{destinationMask} = qq{0};
					$static_route->{defaultGateway} = qq{true};
				}elsif($4){
					$LOGGER->debug(qq{::parse_tmsh_static_routes; Ignoring static route: route domains not supported});
					$static_route = undef;
					last;
				}else{
					$static_route->{destinationAddress} = $3;
					$static_route->{destinationMask} = $6;
					$static_route->{defaultGateway} = qq{false};
				}
			}elsif($route_line =~ /^\s*INTERFACE\s+\S+\/(\S+)\s*$/i){
				$static_route->{interface} = $1;
				$static_route->{discard} = qq{false};
				$type = q{vlan};
			}elsif($route_line =~ /^\s*GW\s+($ip_regex)(\%(\d+))?\s*$/i){
				if($2){
					$LOGGER->debug(qq{::parse_tmsh_static_routes; Ignoring static route: route domains not supported});
					$static_route = undef;
					last;
				}
				$static_route->{gatewayAddress} = $1;
				$static_route->{discard} = qq{false};
				$type = q{gateway};
			}elsif($route_line =~ /^\s*POOL\s+\S+\/(\S+)\s*$/i){
				$static_route->{gatewayAddress} = $1;
				$static_route->{discard} = qq{false};
				$type = q{pool};
			}elsif($route_line =~ /^\s*BLACKHOLE\s*$/i){
				$static_route->{discard} = qq{true};
				$type = q{reject};
			}
		}
		if(defined $static_route && defined $type){
			if(not $static_route->{interface}){
				while($in->{static_routes} =~ /^([\d\.]+)\s+([\d\.]+)\s+([\d\.]+)\s+(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\S+)\s*$/mig){
					if($1 eq q{0.0.0.0}){
						if($static_route->{gatewayAddress} && $2 eq $static_route->{gatewayAddress}){
							$static_route->{interface} = $8;
						}
					}elsif($static_route->{destinationAddress} eq $1){
						$static_route->{interface} = $8;
					}
				}
			}
			if($type eq qq{pool}){
				foreach my $pool_member (@{$pools->{$static_route->{gatewayAddress}}->{member}}){
					if ($pool_member->{enabled}){
						my $pool_route = {
							destinationAddress => $static_route->{destinationAddress},
							destinationMask => $static_route->{destinationMask},
							defaultGateway => $static_route->{defaultGateway},
							discard => $static_route->{discard},
							interface => $static_route->{interface},
							gatewayAddress => $pool_member->{ip}
						};
						push( @{ $staticRoutes->{staticRoute} }, $pool_route );
					}
				}
			}else{
				push( @{ $staticRoutes->{staticRoute} }, $static_route );
			}
		}
	}

	if($in->{mgmt} =~ /MANAGEMENT\-IP\s+($ip_regex)\/(\d+)/i){
		my $static_route = {
			destinationAddress => $1,
			destinationMask => $2,
			defaultGateway => qq{true},
			interface => qq{mgmt},
			discard => qq{false}
		};
		if($in->{mgmt_route} =~ /GATEWAY\s+($ip_regex)/i){
			$static_route->{gatewayAddress} = $1;
		}
		push(@{$staticRoutes->{staticRoute}}, $static_route);
	}
	
	$out->print_element(qq{staticRoutes}, $staticRoutes) if defined $staticRoutes;
}

sub _parse_tmsh_self_ips{
	my ($in) = @_;
	my $ip_regex = get_crep(qq{cipm});
	my $selfs;
	foreach my $self_seg (split(/^NET\s+SELF\s+/mi, $in->{selfs})){
		my $self;
		foreach my $line(split(/^/, $self_seg)){
			if($line =~ /ADDRESS\s+($ip_regex)(\%(\d+))?\/(\d+)/i){
				if($2){
					$LOGGER->debug(qq{::_parse_tmsh_self_ips; Ignoring self ip: route domains not supported});
					last;
				}
				$self->{network} = $1;
				$self->{mask} = $4;
			}elsif($line =~ /VLAN\s+(\S+)/i){
				push(@{$selfs->{$1}}, $self) if defined $self;
			}
		}
	}
	if($selfs){
		return $selfs;
	}
	return undef;
}

sub _parse_tmsh_pools{
	my ($in) = @_;
	my $ip_regex = get_crep(qq{cipm});	
	my $pools;

	foreach my $pool (split(/^LTM\s+POOL\s+/mi, $in->{pools})){
		my $pool_name = q{};
		if($pool =~ /\A(\S+)\s+\{/m){
			$pool_name = $1;
		}
		foreach my $pool_member(split(qq{\}}, $pool)){
			my $member;
			if($pool_member =~ /($ip_regex)(\%(\d+))?:(\S+)/){
				if($2){
					$LOGGER->debug(qq{::_parse_tmsh_pools; Ignoring pool member: route domains not supported});
				}else{
					$member->{ip} = $1;
					$member->{service} = lc($4);
					$member->{enabled} = 1;
					if($pool_member =~ /SESSION\s+(ENABLED|DISABLED)/i){
			 			$member->{enabled} = lc($1) eq qq{enabled} ? 1 : 0;
					}
				}
			}
			push(@{$pools->{$pool_name}->{member}}, $member) if (defined $member);
		}
	}
	return $pools;
}

sub parse_tmsh_vlans{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_tmsh_vlans; Parsing vlans});
	my $vlan_found = 0;

	foreach my $vlan_part(split(/^NET\s+VLAN\s+/mi, $in->{vlans})){
		if($vlan_part =~ /\A(\S+)\s+\{/mi){
			my $vlan;
			$vlan->{name} = $1;
			$vlan->{enabled} = qq{true};
			if($vlan_part =~ /^\s*TAG\s+(\d+)\s*$/mi){
				$vlan->{number} = $1;
			}
			if($vlan_part =~ /^\s*MTU\s+(\S+)\s*$/mi){
				$vlan->{mtu} = $1;
			}

			while($vlan_part =~ /^\s*(\S+)\s+\{/mig){
				if($1 ne $vlan->{name} && lc($1) ne q{interfaces} && lc($1) ne q{sflow}){
					push(@{$vlan->{interfaceMember}}, full_int_name($1));
				}
			}
			
			if(not $vlan_found){
				$vlan_found = 1;
			       	$out->open_element(qq{vlans});
			}
			$out->print_element(qq{vlan}, $vlan);
		}
	}
	
	$out->close_element(qq{vlans}) if $vlan_found;
}

sub parse_tmsh_filter_enabled{
	my ($in, $out) = @_;
	my $returnVal = 0;
	if($in->{filter_enabled} =~ /value\ \"enable\"/i){
		$returnVal = 1;
	}
	return $returnVal;
}

sub parse_tmsh_default_filter{
	my ($in, $out) = @_;
	my $returnVal = qq{discard};
	if($in->{default_filter} =~ /value\ \"(\S+)\"/i){
		$returnVal = $1;
	}
	return $returnVal;
}

sub parse_tmsh_filters{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{::parse_tmsh_filters; Parsing filters});
	my @vlans = @{_get_tmsh_vlan_names($in)};
	#add an empty element to represent all vlans
	push (@vlans, q{});

	my $filters_found = 0;
	foreach my $vlan (@vlans) {
		logit (qq{vlan: $vlan});
		$filters_found = print_filterLists ($vlan, $in, $out, \@vlans, $filters_found);
	}
	my $virtual_servers = _get_tmsh_vs_array_ref($in);
	if($virtual_servers){
		if (! $filters_found) {
			$out->open_element("filterLists");
			$out->open_element("filterListsSection");
		}
		$filters_found = 1;
		my $vlan_names = _get_tmsh_vlan_names($in);
		print_packet_filter_virtual_servers($in, $out, $virtual_servers, $vlan_names);
	}
	if($filters_found){
		my $phaseDiscriminator;
		$phaseDiscriminator->{processingPhase}->{qq{core:packetProcessingPhase}} = qq{address-translation};
		$phaseDiscriminator->{processingPhasePosition} = qq{before};
		$out->print_element(qq{phaseDiscriminator}, $phaseDiscriminator);
		$out->close_element("filterListsSection");
	}

	_parse_tmsh_nat($in, $out, $filters_found);
}

sub _get_tmsh_vlan_full_names{
	my $in = shift;
	my @vlan_full_names;
	foreach my $vlan(split(/^NET::VLAN:\s+/mi, $in->{vlans_show})){
		my $vlan_full_name;
		if($vlan =~ /\A(\S+)\s*$/m){
			$vlan_full_name->{fullname} = $1;
		}
		if($vlan =~ /^\s*INTERFACE\s+NAME\s+(\S+)\s*$/mi){
			$vlan_full_name->{shortname} = $1;
		}
		if($vlan =~ /^\s*TAG\s+(\d+)\s*$/mi){
			$vlan_full_name->{tag} = $1;
		}
		push(@vlan_full_names, $vlan_full_name) if defined $vlan_full_name;
	}
	foreach my $vlan_group(split(/^NET::Vlangroup:\s+/mi, $in->{vlan_groups_show})){
		if($vlan_group =~ /\A(\S+)\s+\(IFNAME:\s+(\S+)\)\s*$/mi){
			my $vlan_full_name = {
				fullname => $1,
				shortname => $2
				};
			push(@vlan_full_names, $vlan_full_name);
		}
	}
	return \@vlan_full_names;
}

sub _get_tmsh_vlan_names{
	my $in = shift;
	my @vlan_names;
	foreach my $vlan(split(/^net vlan\s+/mi, $in->{vlans})){
		if($vlan =~ /^(\S+)\s+{/mi){
			push(@vlan_names, $1);
		}
	}
	foreach my $vlan_group(split(/^net vlan-group\s+/mi, $in->{vlan_groups_list})){
		if($vlan_group =~ /\A(\S+)\s+\{\s*$/mi){
			push(@vlan_names, $1);
		}
	}
	foreach my $tunnel(split(/^NET\s+TUNNELS\s+TUNNEL\s+/mi, $in->{tunnels})){
		if($tunnel =~ /\A(\S+)\s+\{\s*$/mi){
			push(@vlan_names, $1);
		}
	}
	return \@vlan_names;
}

sub print_filterLists{
	my ( $vlan, $in, $out, $arrayVlans, $filters_found ) = @_;
	my $filter_found = 0;	
	foreach my $filter_part ( split( /^net.packet.filter\s+/mi, $in->{"filters"} ) )
	{
		if ( $filter_part =~ /^(.*)\ {/mi )
		{	
			my $filterElementName = $1;	
			#filter has a name and not the default filter
			if ($filterElementName && ($filterElementName !~ /^ / ) ) {
				#logit ("filterelementname:$filterElementName:") ;
				my $filterElementLog = q{false};
				my $filterElementPrimaryAction = q{deny};
				my $filterElementOrder = q{0};
				my $filterVlan = q{};
				my $filterLine = q{};

				my @filterparts = split( /\n/, $filter_part );
				@filterparts = grep (/^ / ,@filterparts);
				foreach my $line ( @filterparts ) {
					if ( $line =~ /^\ *logging\ (\S+)\s*$/mi ) {
						$filterElementLog = "true";
					}	
					if ( $line =~ /^\ *action\ (\S+)\s*$/mi ) {
						my $action = $1;
						if ($action =~ /accept/ ) {	
							$filterElementPrimaryAction = "permit";
						} elsif ($action =~ /continue/ ) {
							$filterElementPrimaryAction = "next";
						}
					}
					if ( $line =~ /^\ *order\ (\S+)\s*$/mi ) {
						$filterElementOrder = $1;
					}	
					if ( $line =~ /^\ *vlan\ (\S+)\s*$/mi ) {
						$filterVlan = $1;
					}	
					if ( $line =~ /^\ *rule\ (.*)$/mi ) {
						$filterLine = $1;
					}
				}
				
				if ($filterVlan eq $vlan) {
					logit (qq{filterElementName $filterVlan eq $vlan});
					if ( not $filter_found) {
						if(not $filters_found){
							$out->open_element("filterLists");
							$out->open_element("filterListsSection");
							$filters_found = 1;
						}
						$filter_found = 1;
						$out->open_element("filterList");
						my $defaultFilterAction = $in->{"defaultFilterAction"};
						if ($defaultFilterAction =~ /discard|reject/) {
							$out->print_element( "defaultAction", "deny" );
						} else {
							$out->print_element( "defaultAction", "permit" );
						}
						if ( $in->{"packetFilterEnabled"} ) {
							#if filter applies to all vlans lets print them out
							if ($filterVlan eq q{})	{
								my @allVlans = @$arrayVlans;
								pop (@allVlans);
								foreach my $iter (@allVlans) {
									$out->open_element("appliesTo");
									$out->print_element("source", $iter );
									$out->close_element("appliesTo");
								}
							} else {
								$out->open_element("appliesTo");
								$out->print_element("source", $vlan );
								$out->close_element("appliesTo");
							}
						}
					}
					parse_filter_line($filterElementName, $filterElementLog, $filterElementPrimaryAction, $filterElementOrder, $filterLine, $out);
				}
			}
		}
	}
	if ($filter_found) {
		$out->print_element( "mode", "stateful" );
		if ( $vlan eq q{} ) {
			$out->print_element( "name", q{ALL VLANS} );
		} else { 
			$out->print_element( "name", $vlan );
		}
		$out->close_element("filterList");
	}
	return $filters_found;
}

sub _is_unparseable_filter {
	my $line = shift;
	my $returnVal = 0;
	#things the zed cant handle
	#dst or src negated
	if ($line =~ m/host\ not|net\ not|proto\ not|portrange\ not/i) {
		$LOGGER->debug("_is_unparseable_filter; ignoring packetfilter, " .
					" negating host, network, portrange or protocol $line");
		$returnVal = 1;
	} 
	#multiple levels of parenthesis
	if ($line =~ m/\([^\)]*\(/i) {
		$LOGGER->debug("_is_unparseable_filter; ignoring packetfilter, " .
					"nested brackets $line");
		$returnVal = 1;
	}
	#and in componennts ( src host 2.2.2.2  and src host 3.3.3.3 )
	foreach my $chunk ( split ( /\)/i, $line ) ) {
		if ($chunk =~ /\(.*and/i) {
			$LOGGER->debug("::_is_unparseable_filter; ignoring packetfilter, " .
					"Using AND within a component $line");
			$returnVal = 1;
		}
	} 
	#or between components ( src host 2.2.2.2 ) or ( dst host 3.3.3.3 )     
	if ($line =~ m/\)\ *or\ *\(/i) {
		$LOGGER->debug("::_is_unparseable_filter; ignoring packetfilter, " .
					"OR between components $line");
		$returnVal = 1;
	} 
	#pcap specific keywords  (less 100) and (ip multicast)
	return $returnVal;
}

sub parse_filter_line {
	my ($filterElementName, $filterElementLog, $filterElementPrimaryAction, $filterElementOrder, $line, $out) = @_;
	my $returnVal;
	my @dstHost;
	my @dstNetwork;
	my @dstNotPort;
	my @dstPort;
	my @dstPortRangeLower;
	my @dstPortRangeUpper;
	my @srcHost;
	my @srcNetwork;
	my @protocol;
	if ( ! _is_unparseable_filter ($line) ) {
		foreach my $chunk ( split ( /\)/i, $line ) ) {
			logit ("chunk:$filterElementName:$chunk:");
			if ( $chunk =~ /src\ (.*)$/i ) {
				my $src = $1;
				if ($src =~ /or/) {
					#logit ("MULTIBALL SRC");
					foreach my $or ( split ( /or/i, $src ) ) {
						if ($or =~ /host(.*)/ ) {
							#logit ("src host:$1:");
							push (@srcHost, $1);
						}
						if ($or =~ /net(.*)/ ) {
							#logit ("src network:$1:");
							push (@srcNetwork, $1);
						}
					}
				} else {
					if ($src =~ /host(.*)/ ) {
						#logit ("src host:$1:");
						push (@srcHost, $1);
					} 
					if ($src =~ /net(.*)/ ) {
						#logit ("src network:$1:");
						push (@srcNetwork, $1);
					}
				}
			}
			if ( $chunk =~ /dst\ (.*)$/i ) {
				my $dst = $1;
				if ($dst =~ /or\ /) {
					#logit ("MULTIBALL DST");
					foreach my $or ( split ( /or\ /i, $dst ) ) {
						if ($or =~ /host(.*)/ ) {
							#logit ("dst host:$1:");
							push (@dstHost, $1);
						} 
						if ($or =~ /net(.*)/ ) {
							#logit ("dst net:$1:");
							push (@dstNetwork, $1);
						}
						if ($or =~ /port\ (.*)/ ) {
							my $checkForNot = $1;
							if ($checkForNot =~ /not(.*)/ ) {
								push (@dstNotPort, $1);
							#       logit ("dst port not:$1:");
							} else {
								push (@dstPort, $1);
							#       logit ("dst port:$1:");
							}
						}
						if ($or =~ /portrange(.*)/ ) {
							my $range = $1;
							if ($range =~ /(\d*)-(\d*)/ ) {
								push (@dstPortRangeLower, $1);
								push (@dstPortRangeUpper, $2);
							}
						}
					}
				} else {
					if ($dst =~ /host(.*)/ ) {
						#logit ("dst host:$1:");
						push (@dstHost, $1);
					} 
					if ($dst =~ /net(.*)/ ) {
						#logit ("dst net:$1:");
						push (@dstNetwork, $1);
					}
					if ($dst =~ /port\ (.*)/ ) {
						#logit ("dst port:$1:");
						my $checkForNot = $1;
						if ($checkForNot =~ /not(.*)/ ) {
							push (@dstNotPort, $1);
						} else {
							push (@dstPort, $1);
						}
					}
					if ($dst =~ /portrange(.*)/ ) {
						#logit ("dst portrange:$1:");
						my $range = $1;
						if ($range =~ /(\d*)-(\d*)/ ) {
							push (@dstPortRangeLower, $1);
							push (@dstPortRangeUpper, $2);
						}
					}
				}
			}
			if ( $chunk =~ /proto\ (.*)$/i ) {
				my $dst = $1;
				if ($dst =~ /tcp/i ) {
					#logit ("proto tcp");
					push (@protocol, "tcp");
				}
				if ($dst =~ /udp/i ) {
					#logit ("proto udp");
					push (@protocol, "udp");
				}
				if ($dst =~ /icmp/i ) {
					#logit ("proto icmp");
					push (@protocol, "icmp");
				}
			}
		}
		printFilterEntry  ($filterElementName, $filterElementLog, $filterElementPrimaryAction, $filterElementOrder, \@srcHost, \@srcNetwork, \@dstHost, \@dstNetwork, \@dstPort, \@dstNotPort, \@dstPortRangeUpper, \@dstPortRangeLower, \@protocol, $out);
	}
}

sub printFilterEntry {
	my ($filterElementName, $filterElementLog, $filterElementPrimaryAction, $filterElementOrder, $srcHost, $srcNetwork, $dstHost, $dstNetwork, $dstPort, $dstNotPort, $dstPortRangeUpper, $dstPortRangeLower, $protocol, $out) = @_;
	$out->open_element("filterEntry");
	my $numDstHosts =  scalar @$dstHost;
	my $numDstNetworks =  scalar @$dstNetwork;
	my $dstExists = $numDstHosts + $numDstNetworks;
	if ($dstExists > 0) {
		if ($numDstHosts > 0) {
			foreach my $host ( @$dstHost ) {
				$out->open_element("destinationIpAddr");
				$host = trim($host);
				$out->print_element( "host", $host );
				$out->close_element("destinationIpAddr");
			}
		}
		if ($numDstNetworks > 0) {
			foreach my $network ( @$dstNetwork ) {
				$network =~ /(.*)\/(.*)/;
			$out->open_element("destinationIpAddr");
			$out->open_element("network");
			$out->print_element( "address", trim($1) );
			$out->print_element( "mask", trim($2) );
			$out->close_element("network");
			$out->close_element("destinationIpAddr");
			}
		}
	}
	foreach my $port ( @$dstPort ) {
		$out->open_element("destinationService");
		$out->open_element("portExpression");
		$out->print_element( "operator", "eq" );
		$port = trim($port);
		$out->print_element( "port", $port );
		$out->close_element("portExpression");
		$out->close_element("destinationService");
	}
	foreach my $port ( @$dstNotPort ) {
		$out->open_element("destinationService");
		$out->open_element("portExpression");
		$out->print_element( "operator", "ne" );
		$port = trim($port);
		$out->print_element( "port", $port );
		$out->close_element("portExpression");
		$out->close_element("destinationService");
	}
	my $numberOfRanges = scalar @$dstPortRangeUpper;
	for(my $i=0;$i<$numberOfRanges;++$i) {
		$out->open_element("destinationService");
		$out->open_element("portRange");
		$out->print_element( "portEnd", @$dstPortRangeUpper[$i] );
		$out->print_element( "portStart", @$dstPortRangeLower[$i] );
		$out->close_element("portRange");
		$out->close_element("destinationService");
	}
	$out->print_element( "log", $filterElementLog );
	$out->print_element( "name", $filterElementName );				
	$out->print_element( "primaryAction", $filterElementPrimaryAction );	
	$out->print_element( "processOrder", $filterElementOrder );
	foreach my $protocol ( @$protocol ) {
		$out->print_element( "protocol", $protocol );
	}
	my $numSrcHosts =  scalar @$srcHost;
	my $numSrcNetworks =  scalar @$srcNetwork;
	my $srcExists = $numSrcHosts + $numSrcNetworks;
	if ($srcExists > 0) {
		foreach my $host ( @$srcHost ) {
			$out->open_element("sourceIpAddr");
			$host = trim($host);
			$out->print_element( "host", $host );
			$out->close_element("sourceIpAddr");
		}
		if ($numSrcNetworks > 0) {
			$out->open_element("sourceIpAddr");
			$out->open_element("network");
			foreach my $network ( @$srcNetwork ) {
				$network =~ /(.*)\/(.*)/;
				$out->print_element( "address", trim ($1) );
				$out->print_element( "mask", trim ($2) );
			}
			$out->close_element("network");
			$out->close_element("sourceIpAddr");
		}
	}
	$out->close_element("filterEntry");
}

sub _parse_tmsh_nat{
	my ($in, $out, $filters_found) = @_;
	$LOGGER->debug(qq{::_parse_tmsh_nat; Parsing NAT});
	my $virtual_servers = _get_tmsh_vs_array_ref($in);
	my $nats = _get_tmsh_nat_array_ref($in);
	my $snats = _get_tmsh_snat_array_ref($in);

	my @big_array;
	push(@big_array, @$virtual_servers) if $virtual_servers;
	push(@big_array, @$nats) if $nats;
	push(@big_array, @$snats) if $snats;

	if(@big_array){
		if(not $filters_found){
			$out->open_element(qq{filterLists});
			$filters_found = 1;
		}
		my $vlan_names = _get_tmsh_vlan_names($in);
		my $vlan_lists = get_vlan_groups($in, \@big_array, $vlan_names);
		print_nat_sections($out, $virtual_servers, $nats, $snats, $vlan_lists);
	}
	if($filters_found){
		$out->close_element(qq{filterLists});
	}
}

sub _get_tmsh_vs_array_ref{
	my ($in) = @_;
	my $pools = _parse_tmsh_pools($in);
	my $snatpools = _parse_tmsh_snat_pools($in);
	my $self_ips = _parse_tmsh_self_ips($in);
	my $ip_regex = get_crep(qq{cipm});
	my @virtuals;
	my $virtualsCount = 0;
	my $virtualsUnprocessed = 0;

	foreach my $virtual(split(/^LTM\s+VIRTUAL\s+/mi, $in->{virtuals})){
		if($virtual =~ /^\s*DISABLED\s*$/mi){
			next;
		}
		my $virtual_hash;
		if($virtual =~ /\A(\S+)\s*\{\s*$/mi){
			$virtual_hash->{name} = $1;
		}else{ next; }
		$virtualsCount++;
		my $multiple_vlans = 0;
		my $source_address_translation = 0;
		foreach my $line(split(/^/, $virtual)){
			if($multiple_vlans){
				if($line =~ /^\s*\}\s*$/i){
					$multiple_vlans = 0;
				}elsif($line =~ /^\s*(\S+)\s*$/){
					if($virtual_hash->{vlans}){
						$virtual_hash->{vlans} .= qq{ } . $1;
					}else{
						$virtual_hash->{vlans} = $1;
					}
				}
			}elsif($source_address_translation){
				if($line =~ /^\s*\}\s*$/i){
					$source_address_translation = 0;
				}elsif($line =~ /^\s*POOL\s+(\S+)\s*$/i){
					$virtual_hash->{snatpool} = $1;
				}elsif($line =~ /^\s*TYPE\s+AUTOMAP\s*$/i){
					$virtual_hash->{automap} = 1;
				}
			}elsif($line =~ /^\s*(REJECT|(IP\-FORWARD)|(L2\-FORWARD)|INTERNAL)\s*$/i){
				$virtual_hash->{type} = lc($1);
				if($virtual_hash->{type} eq q{internal}){
					$LOGGER->debug(qq{::_get_tmsh_vs_array_ref; Ignoring virtual server: internal type not supported});
					$virtual_hash = undef;
					last;
				}
			}elsif($line =~ /^\s*DESTINATION\s+($ip_regex(\%(\d+))?|ANY):(\S+)\s*$/i){
				if($2){
					$LOGGER->debug(qq{::_get_tmsh_vs_array_ref; Ignoring virtual server: route domains not supported});
					$virtual_hash = undef;
					last;
				}
				if(lc($1) ne qq{any}){
					$virtual_hash->{destination} = $1;
				}else{
					$virtual_hash->{destination} = qq{0.0.0.0};
				}
				if($virtual_hash->{destination} eq qq{0.0.0.0}){
					$virtual_hash->{no_reverse} = 1;
				}
				$virtual_hash->{service} = lc($4);
			}elsif($line =~ /^\s*MASK\s+($ip_regex|ANY)\s*$/i){
				if(lc($1) ne qq{any}){
					$virtual_hash->{mask} = mask_to_bits($1);
				}else{
					$virtual_hash->{mask} = 0;
				}
			}elsif($line =~ /^\s*IP\-PROTOCOL\s+(\S+)\s*$/i){
				$virtual_hash->{protocol} = $1;
			}elsif($line =~ /^\s*POOL\s+(\S+)\s*$/i){
				$virtual_hash->{pool} = $1;
				$virtual_hash->{type} = q{pool};
			}elsif($line =~ /^\s*VLANS\s+\{\s*$/i){
				$multiple_vlans = 1;
			}elsif($line =~ /^\s*VLANS\-ENABLED\s*$/i){
				$virtual_hash->{vlans_enabled} = 1;
			}elsif($line =~ /^\s*SOURCE\-ADDRESS\-TRANSLATION\s+\{\s*$/i){
				$source_address_translation = 1;
				if($virtual_hash->{source_translated}){
					$virtual_hash->{no_reverse} = 1;
				}
			}elsif($line =~ /^\s*SOURCE\s+($ip_regex)(\%(\d+))?\/(\d+)\s*$/i){
				if($2){
					$LOGGER->debug(qq{::_get_tmsh_vs_array_ref; Ignoring virtual server: route domains not supported});
					$virtual_hash = undef;
					last;
				}
				if($1 ne q{0.0.0.0}){
					$virtual_hash->{source}->{ip} = $1;
					$virtual_hash->{source}->{mask} = $4;
				}else{
					$virtual_hash->{source_translated} = 1;
				}
			}elsif($line =~ /^\s*TRANSLATE\-PORT\s+DISABLED\s*$/i){
				$virtual_hash->{service} = q{any};
			}
		}
		if(! $virtual_hash->{type}) {
			$LOGGER->debug(qq{::_get_tmsh_vs_array_ref; type not set, this is not supported,Name = } . ($virtual_hash->{name} // q{N/A}));
			$virtualsUnprocessed++;
			$virtual_hash = undef;
			next;	
		}
		if ($virtual_hash){
			if($virtual_hash->{type} eq qq{pool}){
				my $pool = $pools->{$virtual_hash->{pool}};
				foreach my $node(@{$pool->{member}}){
					if($node->{enabled}){
						if(($virtual_hash->{service} eq qq{any}) or ($node->{service} eq qq{any}) or ($virtual_hash->{service} eq $node->{service})){
							my $already;
							foreach my $hash_node(@{$virtual_hash->{nodes}}){
								if($hash_node eq $node->{ip}){
									$already = 1;
								}
							}
							if(not $already){
								push(@{$virtual_hash->{nodes}}, $node->{ip});
							}
						}
					}
				}
			}

			if($virtual_hash->{snatpool}){
				if($snatpools && $snatpools->{$virtual_hash->{snatpool}}){
					my $snatpool = $snatpools->{$virtual_hash->{snatpool}};
					foreach my $member(@{$snatpool->{members}}){
						push(@{$virtual_hash->{snat_nodes}}, $member);
					}
				}
			}

			if($virtual_hash->{automap}){
				if($self_ips){
					my @keys = keys(%$self_ips);
					foreach my $key(@keys){
						foreach my $address(@{$self_ips->{$key}}){
							push(@{$virtual_hash->{snat_nodes}}, $address->{network} . qq{/} . mask_to_bits($address->{mask}));
						}
					}
				}
			}

			if($virtual_hash->{snat_nodes} && not $virtual_hash->{source}){
				$virtual_hash->{source}->{ip} = q{0.0.0.0};
				$virtual_hash->{source}->{mask} = q{0};
			}

			push(@virtuals, $virtual_hash) if($virtual_hash);
		}
	}

	$LOGGER->debug(qq{::_get_tmsh_vs_array_ref; $virtualsUnprocessed unprocessable out of $virtualsCount});
	if(@virtuals){
		return \@virtuals;
	}
	return undef;
}

sub _get_tmsh_nat_array_ref{
	my ($in) = @_;
	my $ip_regex = get_crep(qq{cipm});
	my @nats;
	foreach my $nat(split(/^LTM\s+NAT\s+/mi, $in->{nats})){
		if($nat =~ /^\s*DISABLED\s*$/mi){
			next;
		}
		my $nat_hash;
		my $multiple_VLANS = 0;
		foreach my $line(split(/^/, $nat)){
			if($multiple_VLANS){
				if($line =~ /^\s*\}\s*$/){
					$multiple_VLANS = 0;
				}elsif($line =~ /^\s*(\S+)\s*$/){
					if($nat_hash->{vlans}){
						$nat_hash->{vlans} .= qq{ } . $1;
					}else{
						$nat_hash->{vlans} = $1;
					}
				}
			}elsif($line =~ /^\s*TRANSLATION\-ADDRESS\s+($ip_regex)(\%(\d+))?\s*$/i){
				if($2){
					$LOGGER->debug(qq{::_get_tmsh_nat_array_ref; Ignoring NAT: route domains not supported});
					$nat_hash = undef;
					last;
				}
				$nat_hash->{destination} = $1;
			}elsif($line =~ /^\s*ORIGINATING\-ADDRESS\s+($ip_regex)(\%(\d+))?\s*$/i){
				if($2){
					$LOGGER->debug(qq{::_get_tmsh_nat_array_ref; Ignoring NAT: route domains not supported});
					$nat_hash = undef;
					last;
				}
				$nat_hash->{source} = $1;
			}elsif($line =~ /^\s*VLANS\-ENABLED\s*$/i){
				$nat_hash->{vlans_enabled} = 1;
			}elsif($line =~ /^\s*VLANS\s+\{\s*$/i){
				$multiple_VLANS = 1;
			}
		}
		push(@nats, $nat_hash) if($nat_hash);
	}

	if(@nats){
		return \@nats;
	}
	return undef;
}

sub _get_tmsh_snat_array_ref{
	my ($in) = @_;
	my $ip_regex = get_crep(qq{cipm});
	my $snat_pools = _parse_tmsh_snat_pools($in);
	my $self_ips =  _parse_tmsh_self_ips($in);

	my $anyProtocol = 0;
	if($in->{snats_any_protocol} =~ /ENABLE/i){
		$anyProtocol = 1;
	}

	my @snats;
	foreach my $snat(split(/^LTM\s+SNAT\s+/mi, $in->{snats})){
		my $snat_hash;
		if($snat =~ /\A(\S+)\s*\{\s*$/mi){
			$snat_hash->{name} = $1;
		}else{ next; }
		my $multiple_origins = 0;
		my $multiple_vlans = 0;
		foreach my $line(split(/^/, $snat)){
			if($multiple_origins){
				if($line =~ /^\s*\}\s*$/i){
					$multiple_origins = 0;
				}elsif($line =~ /^\s*(\S+)\s+/){
					my $string = $1;
					if($string =~ /\%\d+/){
						$LOGGER->debug(qq{::_get_tmsh_snat_array_ref; Ignoring SNAT: route domains not supported});
						$snat_hash = undef;
						last;
					}
					if($string =~ /$ip_regex/){
						if($string =~ /0\.0\.0\.0\/0/){
							$snat_hash->{no_reverse} = 1;
						}
						push(@{$snat_hash->{origins}}, $string);
					}
				}
			}elsif($multiple_vlans){
				if($line =~ /^\s*\}\s*$/i){
					$multiple_vlans = 0;
				}elsif($line =~ /^\s*(\S+)\s*$/i){
					if($snat_hash->{vlans}){
						$snat_hash->{vlans} .= qq{ } . $1;
					}else{
						$snat_hash->{vlans} = $1;
					}
				}
			}elsif($line =~ /^\s*AUTOMAP\s*$/i){
				$snat_hash->{type} = q{automap};
			}elsif($line =~ /^\s*TRANSLATION\s+\S+($ip_regex)(\%(\d+))?\s*$/i){
				push(@{$snat_hash->{translations}}, $1);
				if($2){
					$LOGGER->debug(qq{::_get_tmsh_snat_array_ref; Ignoring SNAT: route domains not supported});
					$snat_hash = undef;
					last;
				}
				$snat_hash->{type} = q{translation};
			}elsif($line =~ /^\s*SNATPOOL\s+\S+\/(\S+)\s*$/i){
				$snat_hash->{snatpool} = $1;
				$snat_hash->{type} = q{snatpool};
			}elsif($line =~ /^\s*ORIGINS\s+\{\s*$/i){
				$multiple_origins = 1;
			}elsif($line =~ /^\s*VLANS\s+\{\s*$/i){
				$multiple_vlans = 1;
			}elsif($line =~ /^\s*VLANS\-ENABLED\s*$/i){
				$snat_hash->{vlans_enabled} = 1;
			}
		}

		if(defined $snat_hash){
			if($anyProtocol){
				$snat_hash->{protocol} = 1;
			}

			if($snat_hash->{type}){
				if($snat_hash->{type} eq qq{snatpool}){
					if($snat_pools->{$snat_hash->{snatpool}}){
						$snat_hash->{translations} = $snat_pools->{$snat_hash->{snatpool}}->{members};
					}else{
						$snat_hash = undef;
					}
				}elsif($snat_hash->{type} eq qq{automap}){
					if($self_ips){
						my @keys = keys(%$self_ips);
						my @translations;
						foreach my $key(@keys){
							foreach my $address(@{$self_ips->{$key}}){
								my $translation = $address->{network} . qq{/};
								if($address->{mask}){
									$translation .= mask_to_bits($address->{mask});
								}
								push(@translations, $translation);
							}
						}
						$snat_hash->{translations} = \@translations;
					}else{
						$snat_hash = undef;
					}
				}
			}

			push(@snats, $snat_hash) if($snat_hash);
		}
	}

	if(@snats){
		return \@snats;
	}
	return undef;
}

sub _parse_tmsh_snat_pools{
	my ($in) = @_;
	my $ip_regex = get_crep(qq{cipm});
	my $snat_pools;

	foreach my $snat_pool(split(/^LTM\s+SNATPOOL\s+/mi, $in->{snat_pools})){
		my $snat_pool_name;
		if($snat_pool =~ /\A(\S+)\s*\{\s*$/mi){
			$snat_pool_name = $1;
		}else{ next; }
		my $multiple_members = 0;
		foreach my $line(split(/^/, $snat_pool)){
			if($multiple_members){
				if($line =~ /^\s*\}\s*$/){
					$multiple_members = 0;
				}elsif($line =~ /^\s*(\S+)\s*$/){
					my $string = $1;
					if($string !~ /\%/){
						push(@{$snat_pools->{$snat_pool_name}->{members}}, $string);
					}else{
						$LOGGER->debug(qq{::_parse_tmsh_snat_pools; Ignoring SNAT pool member: route domains not supported});
					}
				}
			}elsif($line =~ /^\s*MEMBERS\s+\{\s*$/i){
				$multiple_members = 1;
			}
		}
	}

	return $snat_pools;
}

sub _createCPUTypesRE{
	my @cpuTypes = qw(Pentium Celeron Xeon XeonMP Itanium Athlon AthlonFX Opteron OpteronMP Duron Sempron);
	'\b(' . join( '|', @cpuTypes ) . ')';
}

sub logit{
	my $s = shift;
	my ($logsec,$logmin,$loghour,$logmday,$logmon,$logyear,$logwday,$logyday,$logisdst)=localtime(time);
	my $logtimestamp = sprintf("%4d-%02d-%02d %02d:%02d:%02d",$logyear+1900,$logmon+1,$logmday,$loghour,$logmin,$logsec);
	$logmon++;
	my $logfile="/tmp/logfile.log";
	my $fh;
	open($fh, '>>', "$logfile") or die "$logfile: $!";
	print $fh "$logtimestamp $s\n";
	close($fh);
}

1;
