package ZipTie::Adapters::F5::BigIP::Telemetry;

use strict;
use warnings;

use Data::Dumper;
use ZipTie::Logger;
use ZipTie::Typer;
use ZipTie::CLIProtocolFactory;
use ZipTie::Model::XmlPrint;
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle get_crep strip_mac get_interface_type choose_admin_ip);
use ZipTie::Adapters::F5::BigIP::AutoLogin qw(execute);
use ZipTie::Adapters::F5::BigIP::Disconnect qw(disconnect);
use ZipTie::Adapters::F5::BigIP qw{get_version};
use Exporter qw{import};
our @EXPORT_OK = qw{open_neighbors close_neighbors parse_system parse_arp parse_mac_table parse_telemetry_interfaces};
my $LOGGER = ZipTie::Logger::get_logger();

sub invoke{
	my ($package, $connectionPathDoc) = @_;
	my ($connectionPath, $discoveryParams) = ZipTie::Typer::translate_document($connectionPathDoc, 'connectionPath');
	my $filehandle = get_model_filehandle('BigIP', $connectionPath->get_ip_address());
	my $cliProtocol = ZipTie::CLIProtocolFactory::create($connectionPath);
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'telemetry');
	my $promptRegex = execute($cliProtocol, $connectionPath);

	$LOGGER->debug('Getting version...');
	my $version = get_version($cliProtocol, $promptRegex);

	my $responses;
	if($version == 11){
		$responses = _fill_tmsh_responses($cliProtocol, $promptRegex);
	}else{
		$responses = _fill_responses($cliProtocol, $promptRegex);
	}

	$printer->attributes(1);
	$printer->open_discovery_event();

	parse_system($responses, $printer);

	open_neighbors($responses, $printer);
	parse_arp($responses, $printer);
	parse_mac_table($responses, $printer);
	close_neighbors($responses, $printer);

	my $interfacesHash = parse_telemetry_interfaces($responses, $printer);
	my $adminIp = $connectionPath->get_ip_address();
	if ($discoveryParams->{calculateAdminIp} eq 'true'){
		$adminIp = choose_admin_ip($adminIp, $interfacesHash);
	}
	$printer->print_element('adminIp', $adminIp);
	
	$printer->close_element('DiscoveryEvent');
	close_model_filehandle($filehandle);
	disconnect($cliProtocol);
}

sub _fill_responses{
	my ($cli_protocol, $device_prompt_regex) = @_;

	my @key_to_commands = (
			'hostname->bigpipe system hostname',
			'arp->arp -n',
			'arp_static->bigpipe arp',
			'mac->bigpipe vlan all fdb all interface',
			'interfaces_physical->bigpipe interface',
			'interfaces_s_link->ip -s link',
			'interfaces->ip addr list',
			#'static_routes_list->bigpipe route all list all'
	);

	my $responses;
	foreach my $key_to_command(@key_to_commands){
		my ($key, $command) = split('->', $key_to_command);
		$responses->{$key} = $cli_protocol->send_and_wait_for($command, $device_prompt_regex);
	}

	return($responses);
}

sub _fill_tmsh_responses{
	my ($cli_protocol, $device_prompt_regex) = @_;

	my @key_to_commands = (
			'hostname->tmsh -q list sys global-settings hostname',
			'arp->arp -n',
			'arp_static->tmsh -q list net arp',
			'mac->tmsh -q show net fdb vlan all',
			'interfaces_physical->tmsh -q show net interface all-properties',
			'interfaces_s_link->ip -s link',
			'interfaces->ip addr list',
			#'static_routes_list->tmsh -q list net route'
	);

	my $responses;
	foreach my $key_to_command(@key_to_commands){
		my ($key, $command) = split('->',$key_to_command);
		$responses->{$key} = $cli_protocol->send_and_wait_for($command, $device_prompt_regex);
	}

	return($responses);
}

sub parse_system{
	my ($in, $out) = @_;
	$LOGGER->debug(q{Parsing system});
	
	if($in->{hostname} =~ /Local\s+Host\s+Name:\s+(\S+)\s*$/mi 
	|| $in->{hostname} =~ /^\s*HOSTNAME\s+(\S+)\s*$/mi){
		$out->print_element('sysName', $1);
	}

	$out->print_element(q{sysObjectId}, q{1.3.6.1.4.1.3375.2});
	$out->print_element(q{sysDescr}, q{F5 Networks BIG-IP});
}

sub parse_arp{
	my ($in, $out) = @_;
	$LOGGER->debug(q{Parsing arp});

	my $ip_regex = get_crep('cipm');
	my $mac_regex = get_crep('mac1') .'|' . get_crep('mac2') . '|' .  get_crep('mac3') . '|' .  get_crep('mac4');
	my $arps;	

	foreach my $arp_line(split(/^/, $in->{arp})){
		if ($arp_line =~ /^($ip_regex)\s+\S+\s+($mac_regex)\s+\S+\s+(\S+)\s*$/mig ){
			my $arp = {
				ipAddress  => $1,
				macAddress => _pad_mac(strip_mac($2)),
				interface => $3
			};
			push(@{$arps->{arpEntry}}, $arp);
		}
	}

	my $tmsh_arp;
	foreach my $arp_line(split(/^/, $in->{arp_static})){
		if($arp_line =~ /^ARP\s+($ip_regex)(\%(\d+))?\s+\-\s+($mac_regex).*$/mi){
			if($2){
				$LOGGER->debug(q{Ignoring static arp: route domains not supported});
			}else{
				my $arp = {
					ipAddress  => $1,
					macAddress => _pad_mac(strip_mac($4))
				};
				push(@{$arps->{arpEntry}}, $arp);
			}
		}elsif($arp_line =~ /^\s*IP\-ADDRESS\s+($ip_regex)(\%(\d+))?\s*$/i){
			if($2){
				$LOGGER->debug(q{Ignoring static arp: route domains not supported});
			}else{
				$tmsh_arp = {
					ipAddress  => $1
				}
			}
		}elsif($arp_line =~ /^\s*MAC\-ADDRESS\s+($mac_regex)\s*$/i){
			if($tmsh_arp){
				$tmsh_arp->{macAddress}  = _pad_mac(strip_mac($1));
				push(@{$arps->{arpEntry}}, $tmsh_arp);
				$tmsh_arp = undef;
			}
		}
	}

	if($arps){
		$out->print_element('arpTable', $arps);
	}
}

sub parse_mac_table{
	my ( $in, $out ) = @_;
	$LOGGER->debug(q{Parsing mac table});

	my $ip_regex = get_crep('cipm');
	my $mac_regex = get_crep('mac1') .'|' . get_crep('mac2') . '|' .  get_crep('mac3') . '|' .  get_crep('mac4');
	my $macs;
	
	foreach my $mac_line(split(/^/, $in->{mac})){
		if($mac_line =~ /^(\S+)\s+($mac_regex)\s+\S+:(\S+)\s+(YES|NO)\s*$/mi
		|| $mac_line =~ /^\s*FDB\s+(\S+)\/(\S+)\s+\-\s+INTERFACE\s*:\s+(\S+)\s*$/mi){
			my $mac = {
				vlan       => $1,
				macAddress => _pad_mac(strip_mac($2)),
				interface  => $3
			};
			push(@{$macs->{macEntry}}, $mac);
		}
	}

	if($macs){
		$out->print_element('macTable', $macs);
	}
}

#TODO
sub parse_routing_neighbors{
	my ($in, $out) = @_;
	$LOGGER->debug(q{Parsing routing neighbours});

	my $routes;

	# No license for advanced routing modules, so just static atm...
	foreach my $route(split(/^NET\sROUTE\s*/mi, $in->{static_routes_list})){
		if($route =~ /^$/mig ){
			my $route = {
				ipAddress  => $1,
				protocol => 'STATIC',
				interface => $3
			};
			push(@{$routes->{routingNeighbor}}, $route);
		}
	}

	if($routes){
		$out->print_element('routingNeighbors', $routes);
	}	
}

sub parse_telemetry_interfaces{
	my ($in, $out) = @_;
	$LOGGER->debug(q{Parsing interfaces});

	my $interfaces;

	foreach my $interface_seg(split(/^(?!\s)/m, $in->{interfaces_s_link})){
		my $interface;
		my $next_line = 0;
		foreach my $interface_line(split(/^/, $interface_seg)){
			if ($interface_line =~ /^(\d+):\s+(\S+):\s+<(\S+)>\s+mtu\s+(\d+)/mig ){
				$interface->{name} = $2;
				$interface->{operStatus} = lc($3);
				if ( $interface->{operStatus} !~ /,up,/i ) {
					$interface->{operStatus} = q{Down};
				}else{
					$interface->{operStatus} = q{Up};
				}
			}elsif ($interface_line =~ /^\s*link\/(\S+)\s+/mig ){
				$interface->{type} = get_interface_type($1);
			}elsif ($interface_line =~ /^\s*RX:\s+/mig ){
				$next_line = 1;
			}elsif ($next_line){
				if ($interface_line =~ /^\s*(\d+)\s+/mi ){
					$interface->{inputBytes} = $1;
				}
				$next_line = 0;
			}
		}
		
		if($interface){
			foreach my $interface_seg(split(/^(?!\s)/m, $in->{interfaces})){
				my $this_interface = 0;
	 			foreach my $interface_line(split(/^/, $interface_seg)){
					if ($interface_line =~ /^(\d+):\s+(\S+):\s+<(\S+)>\s+mtu\s+(\d+)/mi ){
						if ($interface->{name} eq $2){
							$this_interface = 1;
						}
					}elsif ($this_interface){
						if ($interface_line =~ /\s*inet\s+(\S+)\/(\d+)\s+(.*:(\S+)\s+)?/mi){
							my $ipEntry = {
								ipAddress => $1,
								mask => $2
							};  
							push( @{ $interface->{ipEntry} }, $ipEntry);
						}
					}
				}
			}
			push( @{ $interfaces->{interface} }, $interface);
		}
	}

	foreach my $interface_line(split(/^/, $in->{interfaces_physical})){
		if($interface_line =~ /^(\S+)\s+(\S+)\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+\S+\s+(\S+)\s+\S+\s+\S+\s*$/mig
		|| $interface_line =~ /^(\S+)\s+(\S+)\s+\S+\s+\S+\s+([\d\.]+[MGK])/mig){
#					  1.1     UP     1000  FD    1.34M 226   0     0     12.88G  13918 0
#					  1.1     up     MAC   9198  945.1M  452.5K  866.4K    842      0      0      0      0     0     0      0      none  none   none  detached
			my $interface = {
				name => $1,
				operStatus => ucfirst(lc($2)),
				type => 'ethernet'
			};
			
			if ($interface->{operStatus} ne 'Up'){
				$interface->{operStatus} = 'Down';
			}
			
			my $inputBytes = $3;
			my $size = chop($inputBytes);
			$size = lc ($size);
			if($size eq 'k'){
				$inputBytes *= 1024;
			}elsif($size eq 'm'){
				$inputBytes *= 1024 * 1024;
			}elsif($size eq 'g'){
				$inputBytes *= 1024 * 1024 * 1024;
			}else{
				$inputBytes = $inputBytes . $size;
			}
			$interface->{inputBytes} = int($inputBytes + 0.5);	
			
			push( @{ $interfaces->{interface} }, $interface);
		}
	}

	$out->attributes(1);
	$out->print_element('interfaces', $interfaces) if (defined $interfaces);
	return $interfaces;
}

sub open_neighbors{
	my ($in, $xmlOut) = @_;
	$xmlOut->attributes(1);
	$xmlOut->open_element('neighbors');
}

sub close_neighbors{
	my ($in, $xmlOut) = @_;
	$xmlOut->close_element('neighbors');
}

sub _pad_mac{
	my ($mac) = @_;

	my $padding = 12 - length($mac);
	if ($padding > 0){
		for(my $i = 0; $i < $padding; ++$i){
			$mac = "0$mac";
		}
	}

	return($mac);
}

1;
