package ZipTie::Adapters::Generic::SNMP;

use strict;
use warnings;
use 5.010_000;

use Data::Dumper;
use ZipTie::Adapters::BaseAdapter;
use ZipTie::Adapters::Generic::SNMP::Parsers qw(parse_system parse_chassis parse_static_routes);
use ZipTie::Adapters::GenericAdapter;
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::CLIProtocolFactory;
use ZipTie::CLIProtocol;
use ZipTie::ConnectionPath;
use ZipTie::Credentials;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::SnmpSessionFactory;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

# Specifies that this adapter is a subclass of ZipTie::Adapters::BaseAdapter
our @ISA = qw(ZipTie::Adapters::BaseAdapter);

sub backup {
	my ($package_name, $backup_doc) = @_;

	$LOGGER->debug("\nStarting backup");

	# Translate the backup operation XML document into ZipTie::ConnectionPath
	my ($connection_path) = ZipTie::Typer::translate_document( $backup_doc, 'connectionPath' );

	# Grab the ZipTie::Credentials object from the connection path
	my $credentials = $connection_path->get_credentials();

	# Grab an output filehandle for the model. This usually points to STDOUT
	my $filehandle = get_model_filehandle( 'Generic', $connection_path->get_ip_address() );

	# Initialize the model printer
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

	# Create a Net::SNMP session
	$LOGGER->debug('Creating Net::SNMP session');
	my $snmp_session = ZipTie::SnmpSessionFactory->create( $connection_path, $credentials );
	$LOGGER->debug('Net::SNMP session created');

	my $responses = {};

	# Getting SNMP information
	$LOGGER->debug('Getting SNMP information');
	$responses->{snmp} = ZipTie::Adapters::GenericAdapter::get_snmp($snmp_session);
	$LOGGER->debug('SNMP information retrieved');

	# Gather the system uptime
	$LOGGER->debug('Getting system uptime');
	$snmp_session->translate(['-timeticks' => 0,]);
	my $sys_uptime_oid = '.1.3.6.1.2.1.1.3.0';
	my $get_result = ZipTie::SNMP::get($snmp_session, [$sys_uptime_oid]);
	$LOGGER->debug( sprintf ( '%s => %s', $sys_uptime_oid, $get_result->{$sys_uptime_oid} ) );
	$responses->{uptime} = $get_result->{$sys_uptime_oid};
	$LOGGER->debug('Systen uptime retrieved');

	$LOGGER->debug('Parsing system information');
	parse_system($responses, $printer);
	$LOGGER->debug('System information complete');

	$LOGGER->debug('Parsing chassis information');
	parse_chassis($responses, $printer);
	$LOGGER->debug('Chassis information complete');

	$LOGGER->debug('Parsing interface information');

	# Keep the interfaces for matching interface name to ifindex in parse_static_routes
	$responses->{interfaces} = ZipTie::Adapters::GenericAdapter::get_interfaces( $snmp_session, 0, 1 );
	$printer->print_element( 'interfaces', $responses->{interfaces} );

	$LOGGER->debug('Interface information complete');

	# Static routing not collected as it gives QRM incomplete and practically useless data
	#$LOGGER->debug('Parsing static route information');

	#$responses->{static_routes} = ZipTie::Adapters::GenericAdapter::get_static_routes( $snmp_session );
	#parse_static_routes( $responses, $printer);

	#$LOGGER->debug('Static route information complete');

	$LOGGER->debug('Parsing SNMP information');
	$printer->print_element( 'snmp', $responses->{snmp} );
	$LOGGER->debug('SNMP information complete');

	$printer->close_model();

	# Make sure to close the model file handle
	close_model_filehandle($filehandle);

	$LOGGER->debug('Backup complete');
}

1;
