package ZipTie::Adapters::Generic::SNMP::Parsers;

use strict;
use warnings;
use 5.010_000;

use Digest::MD5 qw(md5_hex);
use MIME::Base64 qw(encode_base64);
use ZipTie::Adapters::EnterpriseNumbers;
use ZipTie::Logger;

use Exporter qw(import);
our @EXPORT_OK = qw(parse_system parse_chassis parse_static_routes);

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

sub parse_system {
	my ( $in, $out ) = @_;

	# The device's domain name
	$out->print_element( 'core:systemName', $in->{snmp}->{sysName} );

	# We cannot find these values, 'Unknown' is better than leaving UI blank
	$out->open_element( 'core:osInfo' );
	$out->print_element( 'core:make', 'Unknown' );

	# Was originally 'Undefined', not sure if needed, changed for conformity
	$out->print_element( 'core:version', 'Unknown' );

	$out->print_element( 'core:osType', 'Unknown' );
	$out->close_element( 'core:osInfo' );

	# Default type to 'Router'
	$out->print_element( 'core:deviceType', 'Router' );
	$out->print_element( 'core:contact', $in->{snmp}->{sysContact} );
	
	# Gather the system uptime
	if ( defined( $in->{uptime} ) ) {
		my $last_reboot = time - ( $in->{uptime} / 100 );
		$out->print_element( 'core:lastReboot', int( $last_reboot ) );
	}
}

sub parse_chassis {
	my ( $in, $out ) = @_;

	$out->open_element( 'chassis' );

	my $chassis_asset = {
		'core:assetType' => 'Chassis',
	};

	# Get the vendor from a list of vendor specific oids
	my $make = ZipTie::Adapters::EnterpriseNumbers::get_enterprise_name( $in->{snmp}->{sysObjectId} );
	$chassis_asset->{'core:factoryinfo'}->{'core:make'} = $make;
	$chassis_asset->{'core:factoryinfo'}->{'core:modelNumber'} = 'Unknown';

	# Create a serial number by getting the hex characters from the domain name's md5 digest
	my $serial_number = Digest::MD5::md5_hex( $in->{snmp}->{sysName} );
	$chassis_asset->{'core:factoryinfo'}->{'core:serialNumber'} = $serial_number;

	$out->print_element( 'core:asset', $chassis_asset );
	$out->print_element( 'core:description', $in->{snmp}->{sysDescr} );
	$out->close_element( 'chassis' );
}


sub parse_static_routes {
	my ( $in, $out ) = @_;
	my $static_routes = $in->{static_routes};
	my $interfaces = $in->{interfaces};

	# Create a hash where interfaces' ifindexes are the keys and names are the values
	my %index_to_name = ();
	foreach my $interface ( @{ $interfaces->{interface} } ) {
		if ( defined ( $interface->{ifIndex} ) && defined ( $interface->{name} ) ) {
			$index_to_name{ $interface->{ifIndex} } = $interface->{name};
		} else {
			$LOGGER->debug('WARNING: Could not map interface ifindex to name.');
		}
	}

	# Convert the static routes' interface element from ifindex to name
	foreach my $static_route ( @{ $static_routes->{staticRoute} } ) {
		if ( defined ( $static_route->{interface} ) ) {
			if ( $static_route->{interface} =~ /^(\d+)$/ ) {
				if ( defined ( $index_to_name{$1} ) ) {
					$static_route->{interface} = $index_to_name{$1};
				}
			}
		}
	}

	if ( scalar ( @{ $static_routes->{staticRoute} } ) ) {
		$out->print_element( 'staticRoutes', $static_routes );
	} else {
		$LOGGER->debug('No static routes found');
	}
}

1;
