package ZipTie::Adapters::Generic::XML;

use strict;

use ZipTie::Adapters::Generic::XML::AutoLogin;
use ZipTie::Adapters::Generic::XML::Parsers qw(parse_routing create_config parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_interfaces parse_static_routes parse_vlans parse_stp);

use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::Adapters::GenericAdapter;
use ZipTie::CLIProtocol;
use ZipTie::CLIProtocolFactory;
use ZipTie::ConnectionPath;
use ZipTie::Credentials;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::SNMP;
use ZipTie::SnmpSessionFactory;
use ZipTie::Typer;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

sub backup
{
	my $packageName = shift;
	my $backupDoc   = shift;    # how to backup this device
	my $responses    = {};       # will contain device responses to be handed to the Parsers module

	$LOGGER->debug("starting\n");
	
	my $pretend_ip;
	if( $backupDoc =~ m/host=\"([^\"]+)\"/ )
	{
		$pretend_ip=$1;
	}
	$LOGGER->debug("given ".$pretend_ip."\n");

        my $start_time = [Time::HiRes::gettimeofday()];

	# first, check for the mapping file created by the tool
	my $repo = undef;
	my $remote_path = undef;
	my $remote_file_pattern = $pretend_ip."_sed.xml"; # by default, the file content is expected to contain the IP
	if( -f "ziptie.adapters.generic.xml.mapping.txt" )
	{
		$LOGGER->debug("reading ziptie.adapters.generic.xml.mapping.txt\n");
		my $stat = open(MAPPING,"ziptie.adapters.generic.xml.mapping.txt");
		# this file is expected to exist and contain at least one line
		# of the form:
		# [repository_ip:path]
		# where repository_ip is the name of the fileserver where ZED xmls are stored
		# and path is a directory. Path is optional
		# any following lines of the form:
		# device_ip=file_pattern
		# will be checked for device_ip == the connection ip handed to the adapter
		# if there's a match, the remote_file_pattern will be reset to file_prefix 
		if( $stat )
		{
			$LOGGER->debug("file openned\n");
			
			while(<MAPPING>)
			{
				my $line = $_;
				if( $line =~ /\[(\S+)\]/ )
				{
					$repo = $1;
					if( $repo =~ /(.*?):(.*)/ )
					{
						$repo = $1;
						$remote_path = $2;
						$LOGGER->debug("found repo $repo and path $remote_path\n");
					}
					else
					{
						$LOGGER->debug("found repo $repo\n");
					}
				}
#				elsif( $line =~ /^(.*?)=(.*)$/ )
#				{
#					if( $1 == $pretend_ip )
#					{
#						$LOGGER->debug("found mapping for $pretend_ip\n");
#						$remote_file_pattern = $2;
#						if( defined($repo) )
#						{
#							last;
#						}
#					}
#				}
			}
			close(MAPPING);
		}
	}
	else
	{
		die "NO FILE!!\n";
	}

	if( defined($repo) )
	{
		$backupDoc =~ s/host=\"([^\"]+)\"/host=\"$repo\"/;
	}
	else
	{
		die "No configuration, please run discovery first.";
	}

        my $diff = Time::HiRes::tv_interval($start_time);
        $LOGGER->debug("TIME TO PROCESS ziptie.adapters.generic.xml.mapping.txt $diff");

        $start_time = [Time::HiRes::gettimeofday()];

        # Translate the backup operation XML document into ZipTie::ConnectionPath
	my ($connectionPath) = ZipTie::Typer::translate_document( $backupDoc, 'connectionPath' );

        $diff = Time::HiRes::tv_interval($start_time);
        $LOGGER->debug("TIME TO EXTRACT CONNECTION PATH FROM THE BACKUP XML $diff");


	# Set up the XmlPrint object for printing the ZiptieElementDocument (ZED)
	my $filehandle = get_model_filehandle( 'Generic XML', $pretend_ip ); #$connectionPath->get_ip_address() );
	#my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	#$printer->open_model();
	

        $start_time = [Time::HiRes::gettimeofday()];

	# Make an FTP or SFTP connection
	my $xferProtocol = _connect($connectionPath);

        

	# now list the files, and look for those that match the prefix
#	if( defined($remote_path) )
#	{
#		$responses->{file_list} = $xferProtocol->list($remote_path);
#	}
#	else
#	{
#		$responses->{file_list} = $xferProtocol->list();
#	}
#	$LOGGER->debug(scalar(@{$responses->{file_list}})." files listed\n");
#	my $remote_file = undef;
#	foreach my $candidate ( @{$responses->{file_list}} )
#	{
#		$LOGGER->debug($candidate."?\n");
#		if( $candidate =~ /^(.*$remote_file_pattern)$/ )
#		{
#			$remote_file = $1;
#		}
#	}
	my $path_separator = "/";
	my $remote_file = join "", $remote_path, $path_separator, $remote_file_pattern;

        $diff = Time::HiRes::tv_interval($start_time);
        $LOGGER->debug("TIME TO LIST AND FIND THE FILE USING SFTP $diff");


        $start_time = [Time::HiRes::gettimeofday()];

	if( defined($remote_file) )
	{
		my $local_file = $remote_file;
		if( defined($remote_path) )
		{
			$local_file =~ /$remote_path.(.*)/;
			$local_file = $1;
		}
		if( $xferProtocol->get($remote_file,$local_file) )
		{
			open THE_FILE, "<$local_file";
			my @local_lines = <THE_FILE>;
			close THE_FILE;
			my $local_content = join "", @local_lines;
			
			$LOGGER->debug( "Local content size is " . length($local_content) );
			$local_content =~ s/<sigs>.*<\/verification>//msg; # wipe out the invalid ZED bits
			$LOGGER->debug( "Local content size after replace is " . length($local_content) );

			print $filehandle $local_content;
			print $filehandle "\n";
			unlink $local_file;
		}
		else
		{
			$LOGGER->debug("FAILED TO FETCH $remote_file FROM SERVER");
		}
	}
	else
	{
		$LOGGER->debug( "XML file ($remote_file_pattern) is missing from remote server.");
	}

        $diff = Time::HiRes::tv_interval($start_time);
        $LOGGER->debug("TIME TO READ AND PROCESS THE REMOTE FILE AND WRITE IT LOCALLY $diff");

        $start_time = [Time::HiRes::gettimeofday()];

	_disconnect($xferProtocol);
	#$printer->close_model();                # close out the ZiptieElementDocument
	close_model_filehandle($filehandle);    # Make sure to close the model file handle

        $diff = Time::HiRes::tv_interval($start_time);
        $LOGGER->debug("TIME TO CLEANUP HANDLES $diff");

}

sub commands
{
	my $packageName = shift;
	my $commandDoc  = shift;

	my ( $connectionPath, $commands ) = ZipTie::Typer::translate_document( $commandDoc, 'connectionPath' );
	my ( $cliProtocol, $devicePromptRegex ) = _connect($connectionPath);

	my $result = ZipTie::Adapters::GenericAdapter::execute_cli_commands( 'Generic XML', $cliProtocol, $commands, $devicePromptRegex . '|(#|\$|>)\s*$' );
	_disconnect($cliProtocol);
	return $result;
}

sub _connect
{

	# Grab our arguments
	my $connectionPath = shift;

	# Create a new Transfer protocol object by using the ZipTie::TransferProtocolFactory::create sub-routine
	# to examine the ZipTie::ConnectionPath argument for any command line interface protocols
	# that may be specified.
	my $cliProtocol = ZipTie::TransferProtocolFactory::create($connectionPath);

	# Make a connection to and successfully authenticate with the device
	ZipTie::Adapters::Generic::XML::AutoLogin::execute( $cliProtocol, $connectionPath );

	# Return the created ZipTie::CLIProtocol object and the device prompt encountered after successfully connecting to a device.
	return $cliProtocol;
}

sub _disconnect
{

	# Grab the ZipTie::TransferProtocol object passed in
	my $xferProtocol = shift;

	# Close this session and exit
	$xferProtocol->disconnect();
}

sub _get_uptime
{

	# retrieve the sysUpTime via SNMP
	my $snmpSession = shift;

	$snmpSession->translate( [ '-timeticks' => 0, ] );    # turn off Net::SNMP translation of timeticks
	my $sysUpTimeOid = '.1.3.6.1.2.1.1.3.0';                              # the OID for sysUpTime
	my $getResult = ZipTie::SNMP::get( $snmpSession, [$sysUpTimeOid] );
	return $getResult->{$sysUpTimeOid};
}

1;
