package ZipTie::Adapters::Generic::XML::Applications;

use strict;
use warnings;

use ZipTie::ApplicationsClient;
use ZipTie::Adapters::Generic::XML::AutoLogin;
use ZipTie::Logger;
use ZipTie::Typer;
use ZipTie::ConnectionPath;
use ZipTie::TransferProtocolFactory;
use ZipTie::CLIProtocolFactory;
use ZipTie::Adapters::Generic::XML::Parsers qw(parse_applications);

my $LOGGER = ZipTie::Logger::get_logger();

sub invoke
{
	my ($connectionPathDoc) = @_;

	$LOGGER->debug('In ZipTie::Adapters::Generic::XML::Applications->invoke().');

	# Check for the mapping file created by the discovery script
	my $repo = undef;
	my $remotePath = undef;
	if (-f 'ziptie.adapters.generic.xml.mapping.txt')
	{
		$LOGGER->debug('Found ziptie.adapters.generic.xml.mapping.txt');

		open(my $mappingFileHandle, '<', 'ziptie.adapters.generic.xml.mapping.txt');
		if ($mappingFileHandle)
		{
			while (my $mappingFileLine = <$mappingFileHandle>)
			{
				# The mapping file should be in the following form:
				# [repository_ip:path]
				# The path is optional.
				if ($mappingFileLine =~ /\[([^:\]]+)(?::([^\]]+))?/)
				{
					$repo = $1;
					$LOGGER->debug("Found reop $repo.");

					if (defined($2))
					{
						$remotePath = $2;
						$remotePath =~ s/\/$//;
						$LOGGER->debug("Found remote path $remotePath.");
					}
					else
					{
						$LOGGER->debug('No remote path.');
					}
				}
			}
			close($mappingFileHandle);
		}
	}
	else
	{
		$LOGGER->fatal('Unable to find ziptie.adapters.generic.xml.mapping.txt, make sure to run discovery first.');
	}

	if (!defined($repo))
	{
		$LOGGER->fatal('Failed to obtain repo from ziptie.adapters.generic.xml.mapping.txt.');
	}

	# Connect to the remote repository
	$connectionPathDoc =~ s/host="[^"]+"/host="$repo"/;
	my ($connectionPath) = ZipTie::ConnectionPath::from_xml($connectionPathDoc);

	my $xferProtocol = ZipTie::TransferProtocolFactory::create($connectionPath);
	ZipTie::Adapters::Generic::XML::AutoLogin::execute($xferProtocol, $connectionPath);

	# <> reads until end of file
	local($/);

	# List the files on the remote repository, and find any that match the form <adapterId>_applications.xml
	my $remoteFileList = $xferProtocol->list($remotePath);
	foreach my $remoteFile (@{$remoteFileList})
	{
		my $localFile = $remoteFile;
		$localFile =~ s/^.*?\/([^\/]+)$/$1/;
		$localFile =~ s/__/::/g;
		if ($localFile =~ /^(\S+)_applications\.xml$/)
		{
			my $adapterId = $1;
			if ($adapterId =~ /^ZipTie::Adapters::[^\s\]]+::[^\s\]]+$/)
			{
				$xferProtocol->get($remoteFile, $localFile);
				open(my $fh, '<', $localFile);
				my $applicationsXml = <$fh>;
				close($fh);
				unlink($localFile);

				$LOGGER->debug("Parsing applications for adapter ID $adapterId");

				parse_applications($adapterId, $applicationsXml);
			}
		}
	}

	$xferProtocol->disconnect();
}

1;
