package ZipTie::Adapters::Generic::XML::Parsers;

use strict;
use warnings;
use ZipTie::Adapters::Utils qw(seconds_since_epoch get_mask get_port_number trim get_interface_type);
use MIME::Base64 'encode_base64';

use XML::Twig::XPath;
use ZipTie::ApplicationsClient;
use ZipTie::ApplicationsClient::Application;
use ZipTie::ApplicationsClient::ApplicationClassification;
use ZipTie::ApplicationsClient::ApplicationService;
use ZipTie::Logger;

use Exporter 'import';
our @EXPORT_OK =
  qw(parse_routing create_config parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_interfaces parse_static_routes parse_vlans parse_stp parse_applications);

my $LOGGER = ZipTie::Logger::get_logger();

sub parse_chassis
{
	my ( $in, $out ) = @_;
	my $chassis;

	$chassis->{'core:asset'}->{'core:assetType'} = 'Chassis';

	# $chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:make'}        = 'TBD';
	# $chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = 'TBD';
	# $chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = 'TBD';
	# $chassis->{'core:description'}                                       = 'TBD';
	
	$out->print_element( "chassis", $chassis );
}

sub parse_system
{
	my ( $in, $out ) = @_;

	$out->print_element( 'core:systemName', $in->{snmp}->{sysName} );
	
	$out->open_element('core:osInfo');
	# $out->print_element( 'core:make',    'TBD' );
	# $out->print_element( 'core:version', 'TBD' );
	# $out->print_element( 'core:osType',  'TBD' );
	$out->close_element('core:osInfo');
		
	# $out->print_element( 'core:deviceType', 'TBD' );
	$out->print_element( 'core:contact',    $in->{snmp}->{sysContact} );
	
	if (defined $in->{uptime})
	{
		my $now = time();
		my $lastReboot = $now - ($in->{uptime} / 100);
		$out->print_element('core:lastReboot', int($lastReboot));
	}
}

sub create_config
{
	my ( $in, $out ) = @_;

	# the name of the repository
	my $repository;
	$repository->{'core:name'} = '/';

	# build the simple text configuration
	my $config;
	$config->{'core:name'}       = 'config';
	$config->{'core:textBlob'}   = encode_base64( $in->{'config'} );
	$config->{'core:mediaType'}  = 'text/plain';
	$config->{'core:context'}    = 'active';
	$config->{'core:promotable'} = 'true';

	# push the configuration into the repository
	push( @{ $repository->{'core:config'} }, $config );

	# print the repository
	$out->print_element( 'core:configRepository', $repository );
}

sub parse_routing
{
	my ( $in, $out ) = @_;
}

sub parse_local_accounts
{
	my ( $in, $out ) = @_;
}

sub parse_filters
{
	my ( $in, $out ) = @_;
}

sub parse_snmp
{
	my ( $in, $out ) = @_;
	$out->print_element('snmp', $in->{snmp});
}

sub parse_static_routes
{
	my ( $in, $out ) = @_;
}

sub parse_interfaces
{
	my ( $in, $out ) = @_;
	$out->print_element('interfaces', $in->{interfaces});
}

sub parse_vlans
{
	my ( $in, $out ) = @_;
}

sub parse_stp
{
	my ( $in, $out ) = @_;
}

sub parse_applications
{
	my ($adapterId, $applicationsXml) = @_;

	eval
	{
		my $applications_client = ZipTie::ApplicationsClient->new();

		my $twig = XML::Twig->new(twig_roots => {
			application => sub {
				my ($twig, $element) = @_;

				my $app_name = $element->first_child_text('appName');

				if (defined($app_name) && $app_name ne '')
				{
					my $application_paramiters = {adapterId => $adapterId, applicationName => $app_name};

					my $app_description = $element->first_child_text('appDescription');
					if (defined($app_description) && $app_description ne '')
					{
						$application_paramiters->{applicationDescription} = $app_description;
					}

					my $app_risk_severity = $element->first_child_text('appRiskSeverity');
					if (defined($app_risk_severity) && $app_risk_severity =~ /^\d+$/)
					{
						$application_paramiters->{riskSeverity} = $app_risk_severity;
					}

					my $application = ZipTie::ApplicationsClient::Application->new($application_paramiters);

					foreach my $app_network_port_element ($element->children('appNetworkPort'))
					{
						my $protocol = $app_network_port_element->first_child_text('protocol');
						if (defined($protocol) && $protocol ne '')
						{
							foreach my $port_element ($app_network_port_element->children('port'))
							{
								my $port_low = $port_element->first_child_text('portValue');
								if (defined($port_low) && $port_low =~ /^\d+$/)
								{
									$application->add_application_service(ZipTie::ApplicationsClient::ApplicationService->new(
										{protocol => $protocol, portLow => $port_low}));
								}
								else
								{
									my $port_range_element = $port_element->first_child('portRange');
									if (defined($port_range_element))
									{
										$port_low = $port_range_element->first_child_text('portLow');
										my $port_high = $port_range_element->first_child_text('portHigh');

										if (defined($port_low) && defined($port_high) && $port_low =~ /^\d+$/ && $port_high =~ /^\d+$/)
										{
											$application->add_application_service(ZipTie::ApplicationsClient::ApplicationService->new(
												{protocol => $protocol, portLow => $port_low, portHigh => $port_high}));
										}
									}
								}
							}
						}
					}

					foreach my $app_classification_element ($element->children('appClassification'))
					{
						my $classifier_name = $app_classification_element->first_child_text('classifierName');
						if (defined($classifier_name) && $classifier_name ne '')
						{
							my $classifier_value = $app_classification_element->first_child_text('classifierValue');
							if (!defined($classifier_value))
							{
								$classifier_value = '';
							}

							$application->add_application_classification(ZipTie::ApplicationsClient::ApplicationClassification->new(
								{classifierName => $classifier_name, classifierValue => $classifier_value}));
						}
					}

					$applications_client->add_application($application);
				}

				return 1;
			}});
		$twig->parse($applicationsXml);

		$applications_client->commit();
	};
	if ($@)
	{
		$LOGGER->debug($@);
	}
}

1;
