package ZipTie::Adapters::Generic::XML::Telemetry;

use strict;
use warnings;

use ZipTie::Logger;
use ZipTie::Typer;
use ZipTie::CLIProtocolFactory;
use ZipTie::Adapters::Generic::XML::AutoLogin;
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);

my $LOGGER = ZipTie::Logger::get_logger();

sub invoke
{
	my $packageName = shift;
	my $connectionPathDoc = shift;

	my $pretendIp;
	if ($connectionPathDoc =~ /host="([^"]+)"/)
	{
		$pretendIp = $1;
	}
	else
	{
		$LOGGER->fatal('Failed to obtain IP from connection doc.');
	}

	$LOGGER->debug("Obtaining telemetry for $pretendIp.");

	# Check for the mapping file created by the discovery script
	my $repo = undef;
	my $remotePath = undef;
	if (-f 'ziptie.adapters.generic.xml.mapping.txt')
	{
		open(my $mappingFileHandle, '<', 'ziptie.adapters.generic.xml.mapping.txt');

		if ($mappingFileHandle)
		{
			while (my $mappingFileLine = <$mappingFileHandle>)
			{
				# The mapping file should be in the following form:
				# [repository_ip:path]
				# The path is optional.
				if ($mappingFileLine =~ /\[([^:\]]+)(?::([^\]]+))?/)
				{
					$repo = $1;
					$LOGGER->debug("Found repo $repo.");

					if (defined($2))
					{
						$remotePath = $2;
						$LOGGER->debug("Found remote path $remotePath.");
					}
					else
					{
						$LOGGER->debug('No remote path.');
					}
				}
			}
			close($mappingFileHandle);
		}
	}

	if (!defined($repo))
	{
		$LOGGER->fatal('Failed to obtain repo, make sure to run discovery first.');
	}

	$connectionPathDoc =~ s/host="[^"]+"/host="$repo"/;

	my ($connectionPath) = ZipTie::Typer::translate_document($connectionPathDoc, 'connectionPath');
	my $filehandle = get_model_filehandle('Generic XML', $pretendIp);

	my $xferProtocol = ZipTie::TransferProtocolFactory::create($connectionPath);
	ZipTie::Adapters::Generic::XML::AutoLogin::execute($xferProtocol, $connectionPath);

	# The file should be in the form <ip>_telemetry.xml.
	my $localFile = $pretendIp . '_telemetry.xml';
	my $remoteFile = $localFile;
	# Append the remote path to the remote file name.
	if (defined($remotePath))
	{
		# Strip any trailing / if it exists, because we are about to add it back.
		$remotePath =~ s/\/$//;

		$remoteFile = "$remotePath/$remoteFile";
	}

	$LOGGER->debug("Copying remote file $remoteFile to local file $localFile.");

	# We don't check to make suer the remote file exists before obtaining it, so this may fail.
	$xferProtocol->get($remoteFile, $localFile);

	$xferProtocol->disconnect();

	# Make sure the transfer was successful by checking for existance of the local file.
	unless (-f $localFile)
	{
		$LOGGER->fatal("Failed to obtain $remoteFile.");
	}

	# Slurp in the content local file and remove it from disk.
	my $content = undef;
	{
		local($/);
		open(my $localFileHandle, '<', $localFile);
		if ($localFileHandle)
		{
			$content = <$localFileHandle>;
			close($localFileHandle);
		}
		unlink($localFile);
	}

	if (!defined($content))
	{
		$LOGGER->fatal("Failed to read file $localFile.");
	}

	print $filehandle $content;
	close_model_filehandle($filehandle);

	$LOGGER->debug('Finished.');
}

1;
