package ZipTie::Adapters::HP::ProVision::AutoLogin;

use strict;

use ZipTie::CLIProtocol;
use ZipTie::CLIProtocolFactory;
use ZipTie::Response;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Adapters::HP::ProVision::Interactions qw(strip_control_chars);
use ZipTie::Adapters::Utils qw(trim);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

###########################################
#
# PUBLIC METHODS
#
###########################################

sub connect {

	# Grab our arguments
	my $connection_path = shift;

	# Create a new CLI protocol object
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	# Make a connection to and successfully authenticate with the ProVision device
	my $enable_prompt_regex = execute( $cli_protocol, $connection_path );

	# Store the found prompt as "enablePrompt" on the specified CLI protocol.
	$cli_protocol->set_prompt_by_name( "enablePrompt", $enable_prompt_regex );

	# Get rid of the more prompt
	$LOGGER->debug("Sending 'no page' command");
	my $termLen = $cli_protocol->send_and_wait_for( "no page", $enable_prompt_regex );

	# if for some reason the model that we are backing up doesn't support the 'no page'
	# command we need to set the more prompt so that it can be removed later
	if ( $termLen =~ /Invalid input/i ) {
		$LOGGER->debug("Paging was not disabled. Setting the more prompt.");
		
		$cli_protocol->enable_send_chars_separately();
		$cli_protocol->put("show ?");
		$cli_protocol->disable_send_chars_separately();
		my $response = strip_control_chars($cli_protocol->wait_for('(-- MORE --.*?)$'));
		
		my ($more) = $response =~ /(-- MORE --.*?)$/mi;
		
		$LOGGER->debug("Settting the More Prompt as: $more");
		
		$cli_protocol->set_more_prompt( $more, '20' );
		
		$cli_protocol->send_as_bytes( '20', 0 );
		$cli_protocol->wait_for($enable_prompt_regex);
	}


	# Return the created ZipTie::CLIProtocol object and the enable prompt encountered after successfully connecting to a device.
	return ( $cli_protocol, $enable_prompt_regex );
}




# Connects to and authenticates with a ProVision device that is being connected to by a specified CLIProtocol implementation.
# Attempt to handle all of the edge cases that are encountered and successfully put the ProVision device into
# enable mode.  If all is successful, the device prompt is parsed out and is returned to be used as the primary
# regular expression to wait for when sending commands.
#
# Input:	$cli_protocol -	The CLIProtocol object that is being used to communicate with the device.
#			$connection_path -	Reference to a ZipTie::ConnectionPath object that contains the IP address
#								and any credentials that would need to be used for the current connection.
#
# Returns:	The regular expression that will match the device's enable mode prompt that was found.
sub execute
{
	my $cli_protocol    = shift;
	my $connection_path = shift;

	# Pull off the ZipTie::Credentials instance from the ZipTie::ConnectionPath object
	my $credentials = $connection_path->get_credentials();

	# Store all the information needed to connect to the device
	my $ip_address = $connection_path->get_ip_address();
	my $port       = $cli_protocol->get_port();
	my $username   = $credentials->{username};
	my $password   = $credentials->{password};
		

	my $protocol_name = $cli_protocol->get_protocol_name();
	$LOGGER->debug("Protocol in use: $protocol_name");
	#my $cred_str = $credentials->to_string();
	#$LOGGER->debug("Credentials in use: $cred_str");

	# Attempt to connect to the device
	$cli_protocol->connect( $ip_address, $port, $username , $password );
	$LOGGER->debug("Initial connection complete ...");
		
	# change terminal type
	#$cli_protocol->turn_vt102_on(80,50); # set terminal size
	#$LOGGER->debug("Enabling vt102 emulation ...");
	
	
	# If all goes well during the login process, the last method to be called will be "_calculate_prompt_regex"
	# and its return value will be returned all the way to here	
	my $enable_prompt_regex = _initial_connection( $cli_protocol, $credentials );

	# Print that the login process has been completed
	$LOGGER->debug("Login has successfully completed!\n");

	# Return the prompt regular expression that was found
	return $enable_prompt_regex;
}

###########################################
#
# PRIVATE SUB-ROUTINES
#
###########################################

sub _initial_connection
{
	my $cli_protocol = shift;
	my $credentials  = shift;

	# Specify the responses to check for
	
	my @responses = (ZipTie::Response->new( 'any key to continue',  \&_send_press_any_key ) );
	
	push( @responses, ZipTie::Response->new( 'assword:',  		\&_send_password ) );

	#At the moment if we receive an operator level prompt we send the enable command too increase or privileges
	#This may not ultimately be necessary
	push( @responses, ZipTie::Response->new( '\S+>',  		\&_send_enable ) );
	
	#If we get a prompt with a '#' then we are all the way logged in. Get the prompt and return it.
	push( @responses, ZipTie::Response->new( '\S+#',  		\&_calculate_prompt_regex ) );
	
#	#if we are logging in using Telnet we will need to send the username
#	if ($cli_protocol->get_protocol_name() eq 'Telnet' ) {
		push( @responses, ZipTie::Response->new( 'sername:\s*\Z',  \&_send_username ) );
		push( @responses, ZipTie::Response->new( 'ogin as:',  		\&_send_username ) );
#	}

	#wait for one of the responses and execute the next command
	#If all goes well this will eventually return the prompt regex
	return _next_response($cli_protocol, $credentials, \@responses);
}

sub _send_username
{
	my $cli_protocol = shift;
	my $credentials = shift;

	my @responses = (ZipTie::Response->new( 'any key to continue',  \&_send_press_any_key ) );
	push( @responses, ZipTie::Response->new( 'assword:',  		\&_send_password ) );
	
	#At the moment if we receive an operator level prompt we send the enable command to increase or privileges
	#This may not ultimately be necessary
	push( @responses, ZipTie::Response->new( '\S+>',  		\&_send_enable ) );

	#If we get a prompt with a '#' then we are all the way logged in. Get the prompt and return it.	
	push( @responses, ZipTie::Response->new( '\S+#',  		\&_calculate_prompt_regex ) );

	# Send username credential
	my $username = $credentials->{username};
	$LOGGER->debug("Sending username credential ...");

	#Send $username, wait for one of the responses and execute the next command
	#If all goes well this will eventually return the prompt regex
	return _next_response($cli_protocol, $credentials, \@responses, $username);
}

sub _send_password
{
	my $cli_protocol = shift;
	my $credentials = shift;
	
	my @responses = (ZipTie::Response->new( 'any key to continue',  \&_send_press_any_key ) );
	
	#At the moment if we receive an operator level prompt we send the enable command too increase or privileges
	#This may not ultimately be necessary
	push( @responses, ZipTie::Response->new( '\S+>',  		\&_send_enable ) );
	
	#If we get a prompt with a '#' then we are all the way logged in. Get the prompt and return it.
	push( @responses, ZipTie::Response->new( '\S+#',  		\&_calculate_prompt_regex ) );
	
	
	# Send password credential
	my $password = $credentials->{password};
	$LOGGER->debug("Sending password credential ...");	

	#Send $password, wait for one of the responses and execute the next command
	#If all goes well this will eventually return the prompt regex.  Make sure to obfuscate the data.
	return _next_response($cli_protocol, $credentials, \@responses, $password, 1);
}


sub _send_enable
{
	my $cli_protocol = shift;
	my $credentials  = shift;

	if ( $credentials->{enablePassword} )
	{
		my @responses = ();
		push( @responses, ZipTie::Response->new( 'sername:',  	\&_send_enable_username ) );
		push( @responses, ZipTie::Response->new( 'assword:',  	\&_send_enable_password ) );
		
		#If we get a prompt with a '#' then we are all the way logged in. Get the prompt and return it.
		push( @responses, ZipTie::Response->new( '\S+#', 	\&_calculate_prompt_regex ) );
		push( @responses, ZipTie::Response->new( 'any key to continue',  \&_send_press_any_key ) );

		# Send enable command
		$LOGGER->debug("Sending the 'enable' command");
		my $command = "enable";

		#Send command, wait for one of the responses and execute the next command
		#If all goes well this will eventually return the prompt regex		
		return _next_response($cli_protocol, $credentials, \@responses, $command);
	}
	else
	{
		return _calculate_prompt_regex($cli_protocol);
	}    
}


sub _send_enable_username
{
	my $cli_protocol = shift;
	my $credentials  = shift;

	# Specify the responses to check for
	my @responses = ();
	push( @responses, ZipTie::Response->new( 'assword:',  \&_send_enable_password ) );
	push( @responses, ZipTie::Response->new( 'any key to continue',  \&_send_press_any_key ) );

	# Send username
	my $enable_username = $credentials->{enableUsername};
	
	if ( $enable_username eq "" )
	{
		$LOGGER->fatal("[$INVALID_CREDENTIALS]\nAuthentication refused with an empty username");
	}
	
	$LOGGER->debug("Entering the enable username credential for enable mode authentication ....");
	
	#Send $username, wait for one of the responses and execute the next command
	#If all goes well this will eventually return the prompt regex	
	return _next_response($cli_protocol, $credentials, \@responses, $enable_username);
}


sub _send_enable_password
{
	my $cli_protocol = shift;
	my $credentials  = shift;

	# Specify the responses to check for
	my @responses = ();
	push( @responses, ZipTie::Response->new( 'Access denied',		undef, $INVALID_CREDENTIALS ) );
	push( @responses, ZipTie::Response->new( 'nvalid password',                    	undef, $INVALID_CREDENTIALS ) );
	push( @responses, ZipTie::Response->new( 'Authentication failed',      	undef, $INVALID_CREDENTIALS ) );
	
	#If we get a prompt with a '#' then we are all the way logged in. Get the prompt and return it.
	push( @responses, ZipTie::Response->new( '\S+#', 			\&_calculate_prompt_regex ) );

	# Send the enablePassword
	my $enable_password = $credentials->{enablePassword};
	$LOGGER->debug("Sending the enable password credential '$enable_password'\n");

	#Send $enable_password, wait for one of the responses and execute the next command
	#If all goes well this will eventually return the prompt regex	
	return _next_response($cli_protocol, $credentials, \@responses, $enable_password);
}

sub _calculate_prompt_regex
{
	my $cli_protocol = shift;

	# Grab the last match that the CLIProtocol implementation captured
	my $last_match = $cli_protocol->last_match();
	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[LAST MATCH CAPTURED]");
	$LOGGER->debug($last_match);

	#strip ansii control sequences out of the string containing the prompt so
	#that we can get a good look at it
	
	$last_match = strip_control_chars($last_match);
	
	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[ESCAPE SEQUENCES REMOVED]");
	$LOGGER->debug($last_match);
	
	# Use a regular expression to extract the prompt from the match
#	my ($device_prompt) = $last_match =~ /^([\S][^#\>]+[#\>])\s*$/m;
	my ($device_prompt) = $last_match =~ /^(.*[#\>])\s*$/m;
	$device_prompt = trim($device_prompt);
    
    # Register the device prompt with the ZipTie::Recording singleton instance module
    use ZipTie::Recording;
    my $recording = ZipTie::Recording::get_recording();
    $recording->set_device_prompt($device_prompt);
    
    # Clean up the device prompt for regular expression use
    $device_prompt = quotemeta($device_prompt);

	# Construct the regular expression for the prompt so that it will only match the prompt that is displayed
	# as the very last thing.  This will get around the issue of a device possibly echoing back it's prompt when
	# a command is sent.
	my $device_prompt_regex = $device_prompt;

	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[REGULAR EXPRESSION TO MATCH DEVICE PROMPT]");
	$LOGGER->debug($device_prompt_regex);
	$LOGGER->debug("---------------------------------------------------------");

	return $device_prompt_regex;
}


sub _send_press_any_key
{
	my $cli_protocol = shift;
	my $credentials  = shift;

	# Specify the responses to check for
	my @responses = ();
	
	push( @responses, ZipTie::Response->new( 'sername:',	\&_send_username ) );
	push( @responses, ZipTie::Response->new( 'ogin as:',	\&_send_username ) );
	push( @responses, ZipTie::Response->new( 'assword:',  	\&_send_password ) );
	
	#At the moment if we receive an operator level prompt we send the enable command too increase or privileges
	#This may not ultimately be necessary
	push( @responses, ZipTie::Response->new( '\S+>',  		\&_send_enable ) );

	#If we get a prompt with a '#' then we are all the way logged in. Get the prompt and return it.	
	push( @responses, ZipTie::Response->new( '\S+#',  		\&_calculate_prompt_regex ) );
	
	
	# Send a CRLF to the device
	$LOGGER->debug("Sending a newline to respond to Any Key ...");
	my $command = "";

	#Send $command, wait for one of the responses and execute the next command
	#If all goes well this will eventually return the prompt regex	
	return _next_response($cli_protocol, $credentials, \@responses, $command);
}



###################################################
#This function takes a cli_protocol object, a set of credentials for that object, a list of possible responses
#and optionally a command to be issued.
#It issues the command then waits for a response from the cli_protocol that matches a response from the list.
#Once it recieves a valid response it executes the function associated with the response.
###################################################
sub _next_response
{
	my $cli_protocol = shift;
	my $credentials = shift;
	
	# $responses is a pointer to a list of response objects
	my $responses = shift;
	
	# $command is a string representing the command to send. This is optional.
	my $command = shift;
	
	# obfuscation flag indicates whether send() should know that the data is to be obscured for
	# purposes of logging.
	my $obfuscateResponse = shift;

	if (defined($command)) {
#		$LOGGER->debug("Sending command: $command"); Removed because this can cause the enable password to be displayed in plaintext in the log
#														when the first set of credentials is operator level and the second is manager level.
		$cli_protocol->send($command, $obfuscateResponse);
	}
	
	my $response = $cli_protocol->wait_for_responses( $responses );

	# Based on the response of the device, determine the next interaction that should be executed.
	my $next_interaction = undef;
	if ($response)
	{
		my $regex = $response->get_regex();
		$LOGGER->debug("Response matched: $regex");
		$next_interaction = $response->get_next_interaction();
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}

	# Call the next interaction if there is one to call
	if ($next_interaction)
	{

		# Return the enable mode prompt found
		return &$next_interaction( $cli_protocol, $credentials );
	}	
}

1;

__END__

=head1 NAME

ZipTie::Adapters::HP::ProVision::AutoLogin - Automates the connection and authentication with an HP ProVision device.

=head1 SYNOPSIS

    use ZipTie::Adapters::HP::ProVision::AutoLogin;
	my $cli_protocol = ZipTie::CLIProtocolFactory::create( $connection_path );
	my $prompt_regex = ZipTie::Adapters::HP::ProVision::AutoLogin::execute( $cli_protocol, $connection_path );

=head1 DESCRIPTION

The autologin module for an adapter provides an abstract sequence of commands and responses
that allow the user to login to the device.  Usually this is via a CLI protocol such as Telnet
or SSH.  Once fully connected and authenticated with the device, the autologin module must be able
to calculate and return the a regular expression that will be leveraged by users of the module
to match the primary prompt for the device.  This is crucial in being able to send commands and wait
for their responses where leveraging the command line interface (CLI) clients provided by the ZipTie
Perl framework, such as C<ZipTie::Telnet> and C<ZipTie::SSH>.

This module should be able to handle all the possible scenarios that might be encountered when
connecting to and authenticating with a device.

The autologin module should be abstract enough that scripts outside of the adapters can use it
as well.  For example, a script to make mass changes to a device should be able to leverage
the autologin sequence just like an adapter does during backup.

=head1 PUBLIC SUB-ROUTINES

=over 12


=item C<connect($connection_path)>

Creates the initial CLI connection to the device.  A list is returned containing two elements: the first is a
C<ZipTie::CLIProtocol> object that is a CLI client that can communitcate with the device, and the second is a
regular expression that can be used to match the primary prompt of the device.  This is useful when sending
commands to and receiving responses from the device through the C<ZipTie::CLIProtocol> object and being able to
know when a command has generated all the output and returned back to the primary prompt.
This Function issues the no page command and sets the more prompt if necessary.


=item C<execute($cli_protocol, $connection_path)>

The main entry point for connecting to and authenticating with a network device.  This sub-routine
will connect to the device using the specified C<ZipTie::CLIProtocol> object as the command line interface (CLI)
client and using the hostname or IP address specified within the C<ZipTie::ConnectionPath> object.

This method must return a regular expression that can be used to match the primary prompt of the device that
has been connected to.  This is required so that users of the autologin module can have a way to match the
prompt after sending a command to a device and waiting for the response.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_inital_connection($cli_protocol, $connection_path)>

Once an inital connection has been established with the network device, the first text response over the CLI is
examined.  Depending on the response, various sub-routines may be called.

=item C<_send_press_any_key($cli_protocol, $connection_path)>

This sub-routine is called if a response from the device indicates that any input should be sent to the
device.  This will cause newline to be sent to the device.

=item C<_send_username($cli_protocol, $connection_path)>

This sub-routine is called if a response from the device indicates that the "username" credential should be sent to the
device.

=item C<_send_enable_username($cli_protocol, $connection_path)>

This sub-routine is called if a response from the device indicates that the "enableUsername" credential is required after the "enable" command
has been sent to the device.

=item C<_send_password($cli_protocol, $connection_path)>

This sub-routine is called if a response from the device indicates that the "password" credential should be sent to the
device.

=item C<_send_enable($cli_protocol, $connection_path)>

This sub-routine is called if we have reached the normal device prompt.  Since the ProVision adapter may use commands 
that can only be called from enabled mode, the "enable" command will be sent to the device so that we may attempt to
enter the enabled mode.

=item C<_send_enable_password($cli_protocol, $connection_path)>

This sub-routine is called if a response from the device indicates that the "enablePassword" credential should be sent to the
device so that we may attempt to enter the enabled mode.  This most likely to happen after the "enable" command has been
sent to the device.

=item C<_calculate_prompt_regex($cli_protocol, $connection_path)>

This sub-routine is called once the device has been connected to and authentication has completed successfully.  The
last match from the device will be examined the primary prompt will be parsed from it and converted into a regular
expression that can be used to match the prompt of the device.  This regular expression will be returned so that it can
be used by it's caller.

=item C<_next_response ($cli_protocol, $credentials, $responses, $command)>

This function sends the command specified in $command, to the given cli Protocol $cli_protocol it then waits for a response that is
specified in the list passed in $responses. If the response that is found is associated with a function, that function is then run.
$credentials may be needed depending on what function will be run.

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

Contributor(s): rkruse, Dylan White (dylamite@ziptie.org), Brent Gerig (brgerig@taylor.edu), Daniel Badilla (dbadilla@isthmusit.com)
Date: December 10, 2007

=cut
