package ZipTie::Adapters::HP::ProVision::Disconnect;

use strict;
use ZipTie::CLIProtocol;
use ZipTie::Response;
use ZipTie::Logger;

use Exporter 'import';
our @EXPORT_OK = qw(disconnect);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

# Attempts to disconnect from an ProVision-based device that a CLIProtocol object is currently connected to.
# If all is well, the device will be successfully disconnected from and no error will occur.
#
# Input:	$cliProtocol -	A valid ZipTie::CLIProtocol object that is already connected to a device.
#			$debug -		Whether or not to display debug information.
#
# Returns:	Nothing.
sub disconnect
{
	# Grab the CLI protocol
	my $cli_protocol = shift;
	
	$LOGGER->debug("Disconnecting from the device");
	
	# Keep sending the "logout" command until we have successfully exited the device
	_send_exit($cli_protocol);
	
	# Finally, disconnect from our CLIProtocol
	$cli_protocol->disconnect();
	
	$LOGGER->debug("Successfully disconnected from the device");

}

# send the "logout" command to a device that a CLIProtocol is already connected to until it
# has successfully closed out its connection.
#
# Input:	$cliProtocol -	A valid ZipTie::CLIProtocol object that is already connected to a device.
#
# Returns:	Nothing.
sub _send_exit
{
	# Grab the CLI protocol
	my $cli_protocol = shift;
	
	# Specify the responses to handle
	my @responses = ();
	push(@responses, ZipTie::Response->new('log\s+out\s+\[y\/n\]', \&_send_yes));
	push(@responses, ZipTie::Response->new('save', \&_send_no));
	push(@responses, ZipTie::Response->new('#|>', \&_send_exit));
	#push(@responses, ZipTie::Response->new('.*'));
	
	# Send the "logout" command
	my $command = "logout";
	$LOGGER->debug("Sending logout command.");
	
	_next_response($cli_protocol, \@responses, $command);

}

sub _send_yes
{
	# Grab the CLI protocol
	my $cli_protocol = shift;
	
	# Specify the responses to handle
	my @responses = ();
	push(@responses, ZipTie::Response->new('save', \&_send_no));
	push(@responses, ZipTie::Response->new('.*'));
	
	# Send a "y"
	my $command = 'y';
	
	$LOGGER->debug("Sending y in response to a log out prompt.");
	_next_response($cli_protocol, \@responses, $command);

}

sub _send_no
{
	# Grab the CLI protocol
	my $cli_protocol = shift;
	
	# Specify the responses to handle
	my @responses = ();
	push(@responses, ZipTie::Response->new('.*'));
	
	# Send an "n"
	my $command = 'n';
	$LOGGER->debug("Sending n in response to a save prompt.");
	
	_next_response($cli_protocol, \@responses, $command);

}

###################################################
#This function takes a cli_protocol object, a list of possible responses, and a command to be issued
#If there is a command it issues the command via put as some commands my not need a newline character
#It then waits for a response from the cli_protocol that matches a response from the list.
#Once it recieves a valid response it executes the function associated with the response.
###################################################
sub _next_response
{
	my $cli_protocol = shift;
	
	# $responses is a pointer to a list of response objects
	my $responses = shift;
	
	# $command is a string representing the command to send. This is optional.
	my $command = shift;

	#if the command is defined, send it
	if (defined($command)) {
		#use put because the final prompt may break if we send a newline character
		$cli_protocol->send($command);
	}
	
	#wait for an interaction from the device. $response is one of the response objects in the $responses list.
	my $response = $cli_protocol->wait_for_responses( $responses );

	# Based on the response of the device, determine the next interaction that should be executed.
	my $next_interaction = undef;
	if ($response)
	{
		my $regex = $response->get_regex();
		$LOGGER->debug("Response matched: $regex");
		$next_interaction = $response->get_next_interaction();
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}

	# Call the next interaction if there is one to call
	if ($next_interaction)
	{
		# Execute the next function defined in $responses
		&$next_interaction( $cli_protocol );
	}	
}


1;
