package ZipTie::Adapters::HP::ProVision::Interactions;

use strict;
use warnings;

use ZipTie::CLIProtocol;
use ZipTie::Logger;
use ZipTie::Adapters::Utils qw(trim);


use Exporter 'import';
our @EXPORT_OK = qw(issue_commands strip_control_chars);


my $LOGGER = ZipTie::Logger::get_logger();


##
# This function takes a list of commands in priority order and issues them one after another until a valid
# response is recieved. If none of the commands are valid then undef is returned.
# Before the response is returned the enable prompt and the echoed command are stripped as is all 
# leading and trailing whitespace. 
##
sub issue_commands {
	
	my $command_list = shift;
	my $cli_protocol = shift;
	my $enable_prompt_regex = shift;
	
	
	my $response;
	my $command;
	my @tried_commands = ();
	do {
		$command = shift @$command_list;
		if (defined($command)) {
			$response = strip_control_chars($cli_protocol->send_and_wait_for( "$command", $enable_prompt_regex ));
			push(@tried_commands, $command);
		} else {
			$response = undef;
		}
	} while (_invalid_response($response));
	
	if (defined($response)) {
		$response = trim($response);
	
		#strip the prompt from the response
		$response =~ s/\s*$enable_prompt_regex\s*//mi;
	
		#strip the echoed command from the response
		$response =~ s/\s*$command\s*//mi;
	
		#strip the more prompt and any associated whitespace from the response
		#might consider replacing this with two different snips: one for the more prompt line and one for blank lines
		if (defined($cli_protocol->{more_prompt})) {
			my $more = $cli_protocol->{'more_prompt'}->{'prompt'};
			$response =~ s/($more)//mi;
		}
	}
	
	if (defined($response)) {
		$LOGGER->debug("Tried these commands: @tried_commands\n");
		$LOGGER->debug("Response to command ($command):\n$response");
	} else {
		$LOGGER->debug("Command(s) Failed: @tried_commands\n");
	}

	return $response;
}


#This function checks to see if the response returned from a device is invalid.
#It checks the response against a list of known error messages and returns the matching message if found.
#Note that undef is a valid response that indicates that the adapter has run out of commands to try
sub _invalid_response {
	my $response = shift;
	
	if (defined($response)) {
	
		#this list declaration could be moved to a more prominent position so
		#that new items can be added easily
		my @invalid_responses = ('invalid input', 'ambiguous input', 'incomplete command', 'module not present');
	
		foreach my $item (@invalid_responses) {
			if ($response =~ /$item/mi) {
				return $item;
			}
		}
	}
	return undef;	
}

#This function strips out ansi control sequences from the device output
#Most sequences start with an escape character and a '[' character, and end with a character in the range [@-~]
#characters between the start and end include numbers, ';', and '?'
#non greedy matching is used to only match one sequence at a time.

#A few sequences are composed of only an escape character and a single following character.
sub strip_control_chars {
	my $string = shift;
	
	$string =~ s/\e(\[(.*?)[@-~])//mig;
	$string =~ s/\e(.)//mig;
	
	return $string;
}


1;

__END__

=head1 NAME

ZipTie::Adapters::HP::ProVision::Interactions - Module for issuing commands to a ProVision Device

=head1 SYNOPSIS

    use ZipTie::Adapters::HP::ProVision;
	use ZipTie::Adapters::HP::ProVision::Parsers;
	issue commands($command_list, $cli_protocol, $enable_prompt_regex);

=head1 DESCRIPTION

This module allows CLI commands to be issued to an HP ProVision Device in a consistent way.

=head1 PUBLIC SUB-ROUTINES

=over 12

=item C<issue commands($command_list, $cli_protocol, $enable_prompt_regex)>
Issues a series of commands specified in the list reference $command_list intil a valid response is received from the device
or until there are no more commands to try. The purpose of this function is to allow for slight variations in command syntax
between device models. If no valid response is received this function returns undef.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_invalid_response($response)>
This function checks the string $response against a list of known error messages to determin if the response was valid.
Any thing that does not match, including undef, is considered a valid response.

=item C<_strip_control_chars($string)>
Responses from HP networking devices often include ansi control sequences used to move the cursor, erase lines, etc.
This function strips those characters from the output.

=back

=head1 AUTHOR

Contributor(s):  David Tickner
Date: August 23, 2013

=cut
