package ZipTie::Adapters::HP::ProVision::Parsers;

use strict;
use warnings;
use ZipTie::Adapters::Utils qw(strip_mac trim mask_to_bits create_network_element invert_mask get_mask get_protocol_name);
use MIME::Base64 'encode_base64';
use ZipTie::Logger;

use Exporter 'import';
our @EXPORT_OK = qw(parse_chassis print_chassis create_config print_config parse_system print_system
    			 	parse_interfaces print_interfaces parse_stp parse_vlans print_vlans vlan_list
    				interface_list parse_snmp print_snmp list_acls parse_acls print_acls list_static_filters
    				parse_routes print_static_routes print_dynamic_routes print_routing_protocols parse_routing_protocols
    				parse_stp print_stp list_mst_instances);

my $LOGGER = ZipTie::Logger::get_logger();


#this is a hash of IANAifType codes with their associated SED ifType
#using walkmib ifType will return a list of this type of number
#add to this list if more types are found.
my %interface_types = (	'6' => "ethernet",
						'9' => "tokenRing",
						'23' => "ppp",
						'24' => "softwareLoopback",
						'32' => "frameRelay",
						'37' => "atm",
						'39' => "sonet",
						'53' => "virtual", #53 is proprietary virtual - HP uses this for vlans
						'54' => "virtual", #54 is proprietary multiplexing - HP's code for port trunks
						'161' => "virtual",	); #161 is IEEE 802.3ad Link Aggregate

sub parse_system {
	my ( $in, $out ) = @_;

	#Parse the system name. Note that HP networking devices often come
	#with a default system name that contains spaces
	#This element must exist
	my ($systemName) = trim($in->{system_information} =~ /^\s*System Name\s*:\s*(.+)$/mi);
	$in->{parsed}->{'core:systemName'} = $systemName;
		
	#parse the software name. It is not clear what this element really should be
	#so I am leaving the original interpretation
	if ( $in->{version} =~ /^Image stamp:\s+(.*?)\s*$/mi ) {
		$in->{parsed}->{'core:name'} = $1;
	}

	#parse the software image location
	if ( $in->{version} =~ /^Boot Image:\s+(.*?)\s*$/mi ) {
		$in->{parsed}->{'core:softwareImage'}->{'core:image'} = $1;
	}

	#parse the software/firmware version. 
	if ( $in->{system_information} =~ /(Software|Firmware) revision\s*:*\s+([A-Z]{1,2}\.\d+\.\d+(\.\d+)?)/mi ) {
		$in->{parsed}->{'core:version'} = $2;
	}	

	#Parse the bios version
	my ($biosVersion) = trim($in->{system_information} =~ /ROM Version\s*:\s*(\S+)/mi);
	$in->{parsed}->{'core:biosVersion'} = $biosVersion;

	#parse the device type. At the present time only switches are supported
	#This element must exist
	if ( (defined($in->{sysDescr})) and (my ($deviceType) = $in->{sysDescr} =~ /(Switch)/mi) ){
	$in->{parsed}->{'core:deviceType'} = $deviceType;
	} else {
		$in->{parsed}->{'core:deviceType'} = "Switch";
	}


	#parse the system contact. Note that the contact can be blank and we should not
	#create the element if the data does not exist
	my ($contact) = trim($in->{system_information} =~ /System Contact\s+:\s*?(.*?)\s*$/mi);
	if($contact ne "") {
		$in->{parsed}->{'core:contact'} = $contact;
	}

	#parse the uptime and use it to calculate the time since the last reboot
	if ( $in->{system_information} =~ /Up Time\s+:\s+(.+)\s+Memory/i ) {
		$_ = $1;
		my ($years)   = /(\d+)\s*years?/;
		my ($weeks)   = /(\d+)\s*weeks?/;
		my ($days)    = /(\d+)\s*days?/;
		my ($hours)   = /(\d+)\s*hours?/;
		my ($minutes) = /(\d+)\s*minutes?/;

		# subract the last reboot from the current time
		my $lastReboot = time();
		$lastReboot -= $years * 52 * 7 * 24 * 60 * 60 if ($years);
		$lastReboot -= $weeks * 7 * 24 * 60 * 60      if ($weeks);
		$lastReboot -= $days * 24 * 60 * 60           if ($days);
		$lastReboot -= $hours * 60 * 60               if ($hours);
		$lastReboot -= $minutes * 60                  if ($minutes);
		$in->{parsed}->{'core:lastReboot'} = $lastReboot;
	}
} #end sub parse_system


sub print_system {
	my ( $in, $out ) = @_;
	
	_print_if_defined($in->{parsed}, 'core:systemName', $out);
	
	$out->open_element('core:osInfo');
	
		#Print out the os make. This is default to HP
		$out->print_element( 'core:make', 'HP' );
	
		_print_if_defined($in->{parsed}, 'core:name', $out);
		
		_print_if_defined($in->{parsed}, 'core:softwareImage', $out);
	
		_print_if_defined($in->{parsed}, 'core:version', $out);
	
		#Print out the os type. This is default to ProVision.
		$out->print_element( 'core:osType', 'ProVision' );
	
	$out->close_element('core:osInfo');
	
	_print_if_defined($in->{parsed}, 'core:biosVersion', $out);
	
	_print_if_defined($in->{parsed}, 'core:deviceType', $out);
	
	_print_if_defined($in->{parsed}, 'core:contact', $out);
	
	_print_if_defined($in->{parsed}, 'core:lastReboot', $out);
	
}


sub parse_chassis {
	my ( $in, $out ) = @_;
	
	#The asset we are creating is a Chassis so this can be hard coded
	$in->{parsed}->{chassis}->{'core:asset'}->{'core:assetType'} = 'Chassis';
	
	#Parse the firmware version. Note that some switches use software rather
	#than firmware so this will not always be populated
	if ( $in->{system_information} =~ /Firmware revision\s*:\s*(\S+)/mi ) {
		$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:firmwareVersion'} = $1;
	}	
	
	#This adapter will only handle HP switches
	$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:make'} = 'HP';
	
	#Parse the part number and the model number
	#Even though HP's use of sysDescr is inconsistent, it is the prefered way to find the model number and the part number
	if (defined($in->{sysDescr})) {
		
		my ($data) = $in->{sysDescr} =~ /sysDescr\.0\s*=\s*([\w\s\-]+),/mi;
		my ($part) = $data =~ /(j\d+\S)/mi;
		$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:partNumber'} = $part;
		
		$data =~ s/(hp)|(switch)|$part|(procurve)//mig;
		$data = trim($data);
		$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $data;
	
	#the results of the show module command will likely include the part number and may include the model number
	} elsif ( defined($in->{module}) ) {
		
		#extract the part of the output that should contain the information we need
		my ($data) = $in->{module} =~ /Chassis\s*\:\s*(.*)Serial Number/mi;
		
		#only try to parse the part number and model number if we have data to use
		if (defined($data)) {
			#the part number should be in a very regular format
			my ($part) = $data =~ /(j\d+\S)/mi;
			$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:partNumber'} = $part;
			
			#chop the part number out of the data. The result should be the model number if it is present
			$data =~ s/$part//;
			$data = trim($data);
			
			#if the model number isn't present then default to unknown
			if ($data eq "") {
				$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = "Unknown ProVision Device";
			} else {
				$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $data;
			}
		} else {
			#model number is a required element, but if we have no way to find it we need a default value
			$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = "Unknown ProVision Device";
		}
		
	#if we don't have a way to determine the model number then default it to Unknown ProVision Device
	} else {
		$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = "Unknown ProVision Device";
	}
	

	#parse the device serial number
	if ( $in->{system_information} =~ /Serial Number\s*:\s*(\S+)/mi ) {
		$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = $1;
	}
		
	#parse the system location
	#use the system information rather than the sysLocation.0 mib as this
	#only requires operator level access
	my ($location) = trim($in->{system_information} =~ /System Location\s+:\s*?(.*?)\s*$/mi);
	if($location ne "") {
		$in->{parsed}->{chassis}->{'core:asset'}->{'core:location'}->{'core:description'} = $location;
	}

	#parse the Base MAC Address
	if ( $in->{system_information} =~ /Base MAC Addr\s*:\s*(\S+)/mi ) {
		$in->{parsed}->{chassis}->{'macAddress'} = strip_mac($1);
	}
	
	#Parse the system memory
	if ( $in->{system_information} =~ /Memory\s*-\s*Total\s*:\s*([\d,]+)/mi ) {
		
		my $size = $1;
		$size =~ s/[^\d]//g;
		
		$in->{parsed}->{chassis}->{'memory'}->{'kind'} = 'RAM';
		$in->{parsed}->{chassis}->{'memory'}->{'size'} = $size;
	}

	#parse the list of power supplies omitting any that are listed as not present
	if ( (defined($in->{system_power})) and ( $in->{system_power} =~ /Power Supply Status:/mi )) {
		my %psus = ($in->{system_power} =~ /\s*(\d+)\s*\|\s*(.+?)\s*\|/mig);
		
		while ( my ($number, $state) = each %psus) {
			if (!($state =~ /not present/mi)) {
				$in->{parsed}->{chassis}->{'powersupply'}->{'number'} = $number;
				$in->{parsed}->{chassis}->{'powersupply'}->{'status'} = $state;
			}
		}
	}
} #end sub parse_chassis


sub print_chassis {
	my ( $in, $out ) = @_;
	$out->open_element('chassis');

		_print_if_defined($in->{parsed}->{chassis}, 'core:asset', $out);
	
		_print_if_defined($in->{parsed}->{chassis}, 'macAddress', $out);
	
		_print_if_defined($in->{parsed}->{chassis}, 'memory', $out);
	
		_print_if_defined($in->{parsed}->{chassis}, 'powersupply', $out);
	
	$out->close_element('chassis');
}


#returns a list of interfaces that can be used to issue commands on a per-interface basis
sub interface_list {
	
	my ( $in ) = @_;
	
	# We only want to issue commands for interfaces where we can receive valid results
	my @names = $in->{interfaces} =~ /^\s*(\S+?)(?:\-Trk\d+)?\s*\S+\s*\|\s*(?:no|yes)\s*(?:no|yes).*$/mig;
	return @names;

} #end sub interface_list


sub parse_interfaces {
	my ( $in, $out ) = @_;
	
	#if we have enable access this will be defined
	#and all other mibs should be available
	if (defined($in->{ifIndex})) {
		#parse the ids of all interfaces
		#since we are using walkmib ifIndex this will include all interfaces
		#on the device, physical interfaces, virtual interfaces, and management interfaces
		my @indecies = $in->{ifIndex} =~ /^ifIndex\.\d+ = (\S+)\s*$/mig;
		foreach my $index (@indecies) {
			_parse_interface($in, $out, $index);
		}
	} else {
		#parse the ids of all interfaces
		#since we are using show interfaces brief this will not include all interfaces
		#on the device, only standard physical interfaces. Virtual interfaces and management interfaces
		#will not be present
		my @indecies = $in->{interfaces} =~ /^\s*(\S+?)(?:\-Trk\d+)?\s*\S+\s*\|\s*(?:no|yes)\s*(?:no|yes).*$/mig;
		foreach my $index (@indecies) {
			_parse_interface($in, $out, $index);
		}
		
		#VLANs are interfaces and should appear in the list of interfaces
		#if we don't have enable level access we will not be able to parse
		#detailed information, but the basics should be filled in
		my @vlans = keys %{$in->{parsed}->{vlan}};
		foreach my $vlan (@vlans) {
			_parse_vlan_as_interface($in, $out, $vlan);
		}
		
		_parse_trunks_as_interfaces($in, $out);
		
	}
	
	#add trunk information to the interfaces
	_parse_trunks($in, $out);
	
} #end sub parse_interfaces


sub print_interfaces {
	my ( $in, $out ) = @_;
	
	$out->open_element('interfaces');
	
	foreach my $index (_sort_unknown_data(keys %{$in->{parsed}->{interface}})) {
		my $interface_data = $in->{parsed}->{interface}->{$index};
		
		$out->open_element("interface");
			_print_if_defined($interface_data, 'adminStatus', $out);
			_print_if_defined($interface_data, 'description', $out);
			_print_if_defined($interface_data, 'ifIndex', $out);
			_print_if_defined($interface_data, 'interfaceEthernet', $out);
		
			##### print interface ip information
			if (defined($interface_data->{interfaceIp})) {
				$out->open_element("interfaceIp");
				foreach my $ip (_sort_unknown_data(keys %{$interface_data->{interfaceIp}->{'ipConfiguration'}})) {
					my $ip_data = $interface_data->{'interfaceIp'}->{'ipConfiguration'}->{$ip};
					$out->open_element("ipConfiguration");
						_print_if_defined($ip_data, 'ipAddress', $out);
						_print_if_defined($ip_data, 'mask', $out);
					$out->close_element("ipConfiguration");
				}
				_print_if_defined($interface_data->{'interfaceIp'}, 'localProxyArp', $out);
				$out->close_element("interfaceIp");
			}# end if
			
			
			if (defined($interface_data->{interfaceGrouping})) {
				$out->open_element("interfaceGrouping");
				_print_if_defined($interface_data->{'interfaceGrouping'}, 'ifGroupName', $out);
				_print_if_defined($interface_data->{'interfaceGrouping'}, 'ifGroupingCharacteristic', $out);
				$out->close_element("interfaceGrouping");
			}
			
			my @behaviors = @{$in->{parsed}->{'interface'}->{$index}->{'behavior'}};
			
			foreach my $behavior (@behaviors) {
				$out->print_element("interfaceBehavior", $behavior);
			}
			
			
			_print_if_defined($interface_data, 'interfaceType', $out);
			
			##### print vlan trunk information
			if ( defined($interface_data->{'vlan_trunk'}) ) {
				foreach my $vlan ( _sort_unknown_data(keys %{$interface_data->{vlan_trunk}}) ) {
					$out->open_element("interfaceVlanTrunks");
					$out->print_element( "startVlan", $vlan );
					$out->close_element("interfaceVlanTrunks");
				}
			}
		
			_print_if_defined($interface_data, 'mtu', $out);
			_print_if_defined($interface_data, 'name', $out);
			_print_if_defined($interface_data, 'physical', $out);
			_print_if_defined($interface_data, 'speed', $out);
		$out->close_element('interface');
	}
	
	$out->close_element('interfaces');
}
	

sub _parse_interface {
	my ( $in, $out, $index ) = @_;
	
	#extract the port number, type, status, and mode (?:\-Trk\d+)?
	$in->{interfaces} =~ /^\s*($index)(?:\-Trk\d+)?\s+(\S+)\s+\|\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+).*$/mi;
	my ($type, $enabled, $status, $mode) = ($2, $4, lc($5), $6);
	
	$in->{parsed}->{'interface'}->{$index}->{'ifIndex'} = $index;

########################Parse The status##############################	
	#parse the admin status of the interface.
	#in the Schema adminStatus appears to be the operational status of the device (up or down)
	#this may not be correct. There are 2 different MIBs, ifOperStatus and ifAdminStatus.
	#	ifOperStatus indicates the current operational status of the interface
	#	ifAdminStatus indicates the desired state of the device
	if (defined($in->{ifOperStatus})) {
		$in->{ifOperStatus} =~ /ifOperStatus\.$index = (\d+)/mi;
		$status = _decode_status($1);
	}
	
	if (defined($status)) {
		$in->{parsed}->{'interface'}->{$index}->{'adminStatus'} = $status;
	}
########################End Parse The Status##############################


#########################Parse The Name#############################	
#according to the SED, name is the vendor defined name of this interface
	my $name;
	if (defined($in->{ifName})) {
		$in->{ifName} =~ /ifName\.$index = (.+)/mi;
		$name = trim($1);
#	} elsif  (defined($in->{interface}->{$index})) {
#		($name) = $in->{interface}->{$index} =~ /^\s*Name\s+:\s*(\S+)\s*$/mi;
#		if ($name eq "") {
#			$name = $index;
#		}
	} else {
		$name = $index;
	}
	$in->{parsed}->{'interface'}->{$index}->{'name'} = $name;
########################End Parse The Name##############################


########################Parse The Description##############################	
#according to the SED description is the user defined name or alias of the interface
	
	my $descr;
	
	#if this is a physical port we should have a response that contains the description
	if  (defined($in->{interface}->{$index})) {

		($descr) = $in->{interface}->{$index} =~ /^\s*Name\s*:([^\n]*)$/mi;
		$descr = trim($descr);

		if ($descr eq "") {
			$descr = $index;
		}
		
	#if this is not a physical port
	} else {
		my ($id, $value);
		#search through the VLANs to see if there is a VLAN with the same
		#vendor defined name as this interface. If there is then
		#use the description from that VLAN
		while ( ($id, $value) = each(%{$in->{parsed}->{vlan}}) ) {

			if ( (defined($value->{name})) and ($value->{name} eq $name) ) {
				$descr = $value->{'descr'};
				last;
			}
		}
		
		#if all else fails, use the ifDescr MIB entry which is meaningful only in some cases.
		#this does not appear to be available without enable access
		if(  (!(defined($descr))) and (defined($in->{ifDescr})) ) {
			$in->{ifDescr} =~ /ifDescr\.$index = (.+)($)/mi;
			$descr = trim($1);
			
		}
	}

	$in->{parsed}->{'interface'}->{$index}->{'description'} = $descr;
########################End Parse The Description##############################


########################Parse The Type##############################
	#default to ethernet if we can't find the type another way
	#these are ethernet switches, so this is probably reasonable
	if (defined($in->{ifType})) {
		$in->{ifType} =~ /ifType\.$index = (\d+)/mi;
		$type = _decode_type($1);
	} else {
		$type = 'ethernet';
	}
	$in->{parsed}->{'interface'}->{$index}->{'interfaceType'} = $type;
########################End Parse The Type##############################


########################Parse Ethernet Interface##############################
	#if the interface is an ethernet interface then parse its characteristics
	if ($type eq "ethernet") {
		_parse_interface_ethernet($in, $out, $index);
	}
#######################End Parse Ethernet Interface###############################

	
########################Parse IP Interface################
# we parse ip information using the description because that is what will appear
# in the 'show ip' output 
	_parse_interface_ip($in, $out, $descr, $index);
########################End Parse IP Interface################

	
#######################Parse The MTU###############################
	#this does not appear to be available on physical interfaces without enable access
	
	#could default this to 1500 as that seems to be standard
	if (defined($in->{ifMtu})) {
		$in->{ifMtu} =~ /ifMtu\.$index = (\S+)/mi;
		my $mtu = trim($1);
		$mtu =~ s/[^\d]//g;
		$in->{parsed}->{'interface'}->{$index}->{'mtu'} = $mtu;
	}
######################End Parse The MTU################################


#######################Parse Pysical###############################
	#default to virtual unless the interface is of a physical type we recognise
	#add more types to the regex as more types are recognized
	my $physical = 'false';
	if ($type =~ /ethernet/i) {
		$physical = 'true';
	}
	$in->{parsed}->{'interface'}->{$index}->{'physical'} = $physical;
########################End Parse Physical##############################


########################Parse The Speed##############################

	my $speed;
	if (defined($in->{ifSpeed})) {
		#if the interface is down this will report the lowest possible speed
		#if the interface is virtual, the speed will be zero
		$in->{ifSpeed} =~ /ifSpeed\.$index = (\S+)/mi;
		$speed = trim($1);
		$speed =~ s/[^\d]//g;
	} else {
		#calculate the speed of the given interface
		#note that for some interfaces that are down this calculated speed may differ
		#from the speed returned from walkmib ifSpeed
		if ($mode =~ /(\d+)Gig/i) {
			$speed = $1 * 1000 * 1000 *1000;
		} elsif ($mode =~ /(\d+)\S+/i) {
			$speed = $1 * 1000 * 1000;
		}
	}
	if (defined($speed)) {
		$in->{parsed}->{'interface'}->{$index}->{'speed'} = $speed;
	}
#######################End Parse The Speed###############################


########################Parse The Behavior##############################

	my @behaviors = ();

	my $behavior = "switching";

	if ($in->{ip} =~ /IP Routing : enabled/mi) {
		$behavior = "routing";
	}
	
	if ($descr =~ /Out Of Band Management/mi) {
		$behavior = "administrative";
	}
	
	push @behaviors, $behavior;
	
	if ($physical eq 'false') {
		push @behaviors, "virtual";
	}
		
	$in->{parsed}->{'interface'}->{$index}->{'behavior'} = \@behaviors;

#######################End Parse The Behavior###############################

} #end _parse_interface


# parse_vlans needs to be run before this to ensure that vlan information
# is available to be used here. Therefore parse vlan must be run befor parse_interfaces
sub _parse_vlan_as_interface {
	my ( $in, $out, $index ) = @_;
	
	my $name = $in->{parsed}->{vlan}->{$index}->{'name'};
	
	$in->{parsed}->{'interface'}->{$name}->{'name'} = $name;
	
	if ($in->{parsed}->{vlan}->{$index}->{'enabled'} eq "true") {
		$in->{parsed}->{'interface'}->{$name}->{'adminStatus'} = 'up';
	} else {
		$in->{parsed}->{'interface'}->{$name}->{'adminStatus'} = 'down';
	}
	
	my $descr = $in->{parsed}->{vlan}->{$index}->{'descr'};
	
	#Parse ip information related to this vlan
	_parse_interface_ip($in, $out, $descr, $name);
		
	$in->{parsed}->{'interface'}->{$name}->{'interfaceType'} = 'virtual';
	
	$in->{parsed}->{'interface'}->{$name}->{'physical'} = 'false';
	
	$in->{parsed}->{'interface'}->{$name}->{'description'} = $descr;
	
	$in->{parsed}->{'interface'}->{$name}->{'mtu'} = $in->{parsed}->{vlan}->{$index}->{'mtu'};
	
	my $behavior = "switching";

	if ($in->{ip} =~ /IP Routing : enabled/mi) {
		$behavior = "routing";
	}
	
	my @behaviors = ('virtual', $behavior);
	$in->{parsed}->{'interface'}->{$name}->{'behavior'} = \@behaviors;

	
} #end sub _parse_vlan_as_interface


#port trunks are interfaces. we want to represent them as virtual interfaces
#this function creates interface entries for each port trunk
sub _parse_trunks_as_interfaces {
	my ( $in, $out ) = @_;
	
	if ( !(defined($in->{trunk})) ) {
		return undef;
	}
	
	my @lines = split /\n/, $in->{trunk};
		
		foreach my $line (@lines) {
	
			my ($index, $name, $type, $trunk_name, $trunk_type) =  $line =~ /^\s*(\d+)\s+\|\s+(.*)\s+(\S+)\s+\|\s+(\S+)\s+(\S+)\s*$/mig;
			
			
			if ( (defined($trunk_type)) and ($trunk_type =~ /trunk|lacp/i) ) {
				$in->{parsed}->{'interface'}->{$trunk_name}->{'name'} = $trunk_name;
				
				$in->{parsed}->{'interface'}->{$trunk_name}->{'description'} = $trunk_name;
				
				$in->{parsed}->{'interface'}->{$trunk_name}->{'interfaceType'} = "virtual";
				
				$in->{parsed}->{'interface'}->{$trunk_name}->{'physical'} = 'false';
				
				$in->{parsed}->{'interface'}->{$trunk_name}->{'mtu'} = "1500";
				
				if ( (!(defined($in->{parsed}->{'interface'}->{$trunk_name}->{'adminStatus'})) ) or ($in->{parsed}->{'interface'}->{$trunk_name}->{'adminStatus'} eq "down") ) {
					$in->{parsed}->{'interface'}->{$trunk_name}->{'adminStatus'} = $in->{parsed}->{'interface'}->{$index}->{'adminStatus'};
				}
				
				if  (!(defined($in->{parsed}->{'interface'}->{$trunk_name}->{'speed'})) ) {
					$in->{parsed}->{'interface'}->{$trunk_name}->{'speed'} = 0;
				}
				$in->{parsed}->{'interface'}->{$trunk_name}->{'speed'} += $in->{parsed}->{'interface'}->{$index}->{'speed'};
				
				
				my $behavior = "switching";
	
				if ($in->{ip} =~ /IP Routing : enabled/mi) {
					$behavior = "routing";
				}
				
				my @behaviors = ('virtual', $behavior);
				$in->{parsed}->{'interface'}->{$trunk_name}->{'behavior'} = \@behaviors;
			}
			
		}
	
	
}



sub _parse_interface_ethernet {
	my $in = shift;
	my $out = shift;
	my $index = shift;
	
	#if the given interface is present in the standard interface list we can parse some information from there
	if( $in->{interfaces} =~ /^\s*($index)(?:\-Trk\d+)?\s+(\S+)\s+\|\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+).*$/mi) {

		#we are interested in the type and the mode
		my ($type, $mode) = ($2, $6);
		
		#from the mode determin if the interface is set to auto-negotiate duplex and speed settings
		my $autoDuplex = 'false';
		my $autoSpeed = 'false';
		if ($mode =~ /auto/i) {
			$autoDuplex ='true';
			$autoSpeed = 'true';
		}
		$in->{parsed}->{'interface'}->{$index}->{'interfaceEthernet'}->{'autoDuplex'} = $autoDuplex;
		$in->{parsed}->{'interface'}->{$index}->{'interfaceEthernet'}->{'autoSpeed'} = $autoSpeed;
		
		#from the mode, determin if the device is set to run at full duplex or half duplex
		my $opDuplex = undef;
		if ($mode =~ /fd/i) {
			$opDuplex = 'full';
		} elsif ($mode =~ /hd/i){
			$opDuplex = 'half';
		}
		$in->{parsed}->{'interface'}->{$index}->{'interfaceEthernet'}->{'operationalDuplex'} = $opDuplex;
		
		#save the media type
		$in->{parsed}->{'interface'}->{$index}->{'interfaceEthernet'}->{'mediaType'} = $type;
		
	}
	
	#the interface may be an ethernet interface and not appear on the standard list
	#if it is an out-of-band management interface. in that case it is likely that we can
	#only obtain the mac address.
	
	#get the mac address
	my $mac;
	if (defined($in->{ifPhysAddress})) {
		$in->{ifPhysAddress} =~ /^ifPhysAddress\.$index = (.+)$/mi;
		$mac = strip_mac($1);
	} elsif (defined($in->{interface}->{$index})){
		if($in->{interface}->{$index} =~ /^\s*MAC Address\s+:\s*(\S+)\s*$/mi) {
			$mac = strip_mac($1);
		} else {
			$mac = undef;
		}
	}
	if (defined($mac)) {
		$in->{parsed}->{'interface'}->{$index}->{'interfaceEthernet'}->{'macAddress'} = $mac;
	}

} #end sub _parse_interface_ethernet


sub _parse_interface_ip {
	my $in = shift;
	my $out = shift;
	my $descr = shift;
	my $ifIndex = shift;
	
	#only try to parse ip information if the show ip command succeeded
	if (defined($in->{ip})) {
		
		#if we can capture a list of ip addresses for this interface
		#we can extract the ip address, the mask, and the localProxyArp state
		if ($in->{ip} =~ /^\s*(?:$descr)([\s\S]+?)(?:(?:^\s*[^\s\|])|\Z)/mi) {
			my $entry = trim($1);

			my @lines = split /\n/, $entry;
			
			#it is possible for a vlan to have many ip addresses assigned to it.			
			foreach my $line (@lines) {		
				if ($line =~ /\|\s*(\S+)\s+(\S+)\s+(\S+)\s*(\S+)?\s*(\S+)?\s*$/mi) {
					my $ip = $2;
					my $mask = $3;
					
					if ( defined($ifIndex) ) {
						$in->{parsed}->{'interface'}->{$ifIndex}->{'interfaceIp'}->{'ipConfiguration'}->{$ip}->{'ipAddress'} = $ip;
						$in->{parsed}->{'interface'}->{$ifIndex}->{'interfaceIp'}->{'ipConfiguration'}->{$ip}->{'mask'} =  mask_to_bits("$mask");
					}# else {
					#	$in->{parsed}->{'interface'}->{$descr}->{'interfaceIp'}->{'ipConfiguration'}->{$ip}->{'ipAddress'} = $ip;
					#	$in->{parsed}->{'interface'}->{$descr}->{'interfaceIp'}->{'ipConfiguration'}->{$ip}->{'mask'} =  mask_to_bits("$mask");
					#}
					
					if (defined($in->{parsed}->{lookup}->{vlan_descr_to_name}->{$descr})) {
						$in->{parsed}->{lookup}->{ip_to_interface}->{$ip} = $in->{parsed}->{lookup}->{vlan_descr_to_name}->{$descr};
					} else {
						$in->{parsed}->{lookup}->{ip_to_interface}->{$ip} = $descr;
					}
				}
			} #end foreach
			
			#the first line of the ip address entries for a vlan should contain proxy arp information
			if ($entry =~ /\|\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*$/mi) {
				my $local_arp = "false";
				if ($5 =~ /yes/i) {
					$local_arp = "true";
				}
				if ( defined($ifIndex) ) {
					$in->{parsed}->{'interface'}->{$ifIndex}->{'interfaceIp'}->{'localProxyArp'} = $local_arp;
				}# else {
				#	$in->{parsed}->{'interface'}->{$descr}->{'interfaceIp'}->{'localProxyArp'} = $local_arp;
				#}
			} elsif ($entry =~ /\|\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*$/mi) {
				my $local_arp = "false";
				if ($4 =~ /yes/i) {
					$local_arp = "true";
				}
				if ( defined($ifIndex) ) {
					$in->{parsed}->{'interface'}->{$ifIndex}->{'interfaceIp'}->{'localProxyArp'} = $local_arp;
				}
			}
			#end inner if
			
		
		
		#if we can not capture this interface from the output of show ip
		#then we can check the ipAddressIfIndex mib, but the ip mask is not readily available
		} elsif ( defined($ifIndex) and (defined($in->{ipAddressTable})) ){
			
			#only capture ipv4 adresses here
			my @ips = $in->{ipAddressTable} =~ /^ipAddressIfIndex\.\d\.\d\.(\d+\.\d+\.\d+\.\d+) = $ifIndex\s*$/mig;
			
			foreach my $ip (@ips) {
				$in->{parsed}->{'interface'}->{$ifIndex}->{'interfaceIp'}->{'ipConfiguration'}->{$ip}->{'ipAddress'} = $ip;
				
				if (defined($in->{parsed}->{lookup}->{vlan_descr_to_name}->{$descr})) {
					$in->{parsed}->{lookup}->{ip_to_interface}->{$ip} = $in->{parsed}->{lookup}->{vlan_descr_to_name}->{$descr};
				} else {
					$in->{parsed}->{lookup}->{ip_to_interface}->{$ip} = $descr;
				}
			}
		}
		
	} #end defined if
	
} #end sub _parse_interface_ip


#port trunks contain several physical interfaces. We want to indicat in each of these interfaces
#that it is part of a trunk.
#this function creates interface grouping elements for each interface that is part of a trunk
sub _parse_trunks {
	my ( $in, $out ) = @_;
	
	#if we didn't get a valid response from the 'show trunk' command skip parsing trunks
	if ( !(defined($in->{trunk})) ) {
		return undef;
	}
	
	#check for the table header. Add else clauses to this if statement to add different table layouts
	if ($in->{trunk} =~ /^\s*Port.+Name.+Type.+Group.+Type\s*$/mi) {
		
		my @lines = split /\n/, $in->{trunk};
		
		foreach my $line (@lines) {
			#parse the attributes from each line of the table
			my ($index, $name, $type, $trunk_name, $trunk_type) =  $line =~ /^\s*(\d+)\s+\|\s+(.*)\s+(\S+)\s+\|\s+(\S+)\s+(\S+)\s*$/mig;
			
			#for now only deal with trunk and lacp entries
			if ( (defined($trunk_type)) and ($trunk_type =~ /trunk|lacp/i) ) {
				
				#since each port can only be a member of one trunk group we don't beed to maintain an indexed list at this time
				$in->{parsed}->{'interface'}->{$index}->{'interfaceGrouping'}->{'ifGroupName'} = $trunk_name;
				
				#we can default to bondingGroupMember as that is the type used for trunks
				$in->{parsed}->{'interface'}->{$index}->{'interfaceGrouping'}->{'ifGroupingCharacteristic'} = "bondingGroupMember";
			}
		}
	}
	
} # end sub _parse_trunks


sub create_config {

	my ( $in, $out ) = @_;

	# the name of the repository
	my $repository;
	$repository->{"core:name"} = "/";

	if (defined($in->{"running_config"})) {
		# build the simple text configuration
		my $running;
		$running->{"core:name"}       = "running-config";
		$running->{"core:textBlob"}   = encode_base64( $in->{"running_config"} );
		$running->{"core:mediaType"}  = "text/plain";
		$running->{"core:context"}    = "active";
		$running->{"core:promotable"} = "false";
	
		# push the configuration into the repository
		push( @{ $repository->{'core:config'} }, $running );
	}

	if (defined($in->{"startup_config"})) {
		my $startup;
		$startup->{"core:name"}       = "startup-config";
		$startup->{"core:textBlob"}   = encode_base64( $in->{"startup_config"} );
		$startup->{"core:mediaType"}  = "text/plain";
		$startup->{"core:context"}    = "boot";
		$startup->{"core:promotable"} = "true";                                                     
	
		# push the configuration into the repository
		push( @{ $repository->{'core:config'} }, $startup );
	}

	if ( (defined($in->{"running_config"})) or (defined($in->{"startup_config"})) ) {
		# print the repository
		$in->{parsed}->{'core:configRepository'} = $repository;
	}
	
} #end sub create_config

sub print_config {
	my ( $in, $out ) = @_;
	_print_if_defined($in->{parsed}, 'core:configRepository', $out);	
}

sub parse_stp {
	my ( $in, $out ) = @_;
	

	#If spanning tree is off we will not be able to find any information
	if ( (!(defined($in->{stp})) ) or ($in->{stp} =~ /STP\s+Enabled\s*:\s*No/mi) ){
		return undef;
	}
	
	#parse global spanning tree configuration information
	my ($forward_delay) = $in->{stp} =~ /Forward\s+Delay\s*\:\s*(\d+)/mi;
	
	my $hello_time;
	if ($in->{stp} =~ /Hello\s+Time.*?\:\s*(\d+)/mi) {
		($hello_time) = $1;
	} else {
		($hello_time) = $in->{stp_config} =~ /Hello\s+Time.*?\:\s*(\d+)/mi;
	}
	my ($max_age) = $in->{stp} =~ /Max\s+Age\s*\:\s*(\d+)/mi;
	my ($priority) = $in->{stp} =~ /Switch\s+Priority\s*\:\s*(\d+)/mi;
	my $system_mac_address;
	if ($in->{stp} =~ /Switch\s+Mac\s+Address\s*\:\s*(\S+)/mi) {
		$system_mac_address = $1;
	} else {
		$system_mac_address = $in->{parsed}->{chassis}->{'macAddress'};
	}
		 
	my ($mode) = $in->{stp} =~ /Force\s+Version\s*\:\s*(MST|STP|RSTP)/mi;
	
	$in->{parsed}->{spanningTree}->{forwardDelay} = $forward_delay;
	$in->{parsed}->{spanningTree}->{helloTime} = $hello_time;
	$in->{parsed}->{spanningTree}->{maxAge} = $max_age;
	$in->{parsed}->{spanningTree}->{priority} = $priority;
	$in->{parsed}->{spanningTree}->{systemMacAddress} = strip_mac($system_mac_address);
	$in->{parsed}->{spanningTree}->{mode} = $mode;
	
	my @instances = ();
	
	my $cst = undef;
	
	#if there is root information in the main output then we are probably using RSTP or STP
	#we need to add this information to an instance because it is not representable in the SED otherwise
	if ($in->{stp} =~ /^\s*Root MAC Address\s*\:\s*(\S+)/mi) {
		$cst->{designatedRootMacAddress} = strip_mac($1);
	}
	if ($in->{stp} =~ /^\s*Root Path Cost\s*\:\s*(\S+)/mi) {
		$cst->{designatedRootCost} = strip_mac($1);
	}
	if ($in->{stp} =~ /^\s*Root Port\s*\:\s*(\S+)/mi) {
		$cst->{designatedRootPort} = strip_mac($1);
	}
	if ($in->{stp} =~ /^\s*Root Priority\s*\:\s*(\S+)/mi) {
		$cst->{designatedRootPriority} = strip_mac($1);
	}
	
	if (defined($cst)) {
		$cst->{forwardDelay} = $in->{parsed}->{spanningTree}->{forwardDelay};
		$cst->{helloTime} = $in->{parsed}->{spanningTree}->{helloTime};
		$cst->{maxAge} = $in->{parsed}->{spanningTree}->{maxAge};
		$cst->{systemMacAddress} = $in->{parsed}->{spanningTree}->{systemMacAddress};
		$cst->{priority} = $in->{parsed}->{spanningTree}->{priority};
		push(@instances, $cst);
	}
	
	#if there are spanning tree instances extract as much information as we can
	foreach my $key (keys %{$in->{stp_instance}}) {
		
		my $stp_data = $in->{stp_instance}->{$key};

		my ($designatedRootCost) = $stp_data =~ /Regional Root Path Cost\s*\:\s*(\S+)/mi;
		my ($designatedRootMacAddress) = $stp_data =~ /Regional Root MAC Address\s*\:\s*(\S+)/mi;
		my ($designatedRootPort) = $stp_data =~ /Regional Root Port\s*\:\s*(\S+)/mi;
		my ($designatedRootPriority) = $stp_data =~ /Regional Root Priority\s*\:\s*(\S+)/mi;
		my ($forwardDelay) = $in->{parsed}->{spanningTree}->{forwardDelay};
		my ($helloTime) = $in->{parsed}->{spanningTree}->{helloTime};
		my ($maxAge) = $in->{parsed}->{spanningTree}->{maxAge};
		my ($systemMacAddress) = $in->{parsed}->{spanningTree}->{systemMacAddress};
		my ($priority) = $stp_data =~ /Switch Priority\s*\:\s*(\S+)/mi;
		my ($vlan) = $stp_data =~ /Mapped VLANs\s*\:\s*(\S+)/mi;
		
		my $instance;
		
		$instance->{designatedRootCost} = $designatedRootCost if (defined($designatedRootCost));
		$instance->{designatedRootMacAddress} = strip_mac($designatedRootMacAddress) if (defined($designatedRootMacAddress));
		$instance->{designatedRootPort} = $designatedRootPort if (defined($designatedRootPort));
		$instance->{designatedRootPriority} = $designatedRootPriority if (defined($designatedRootPriority));
		$instance->{forwardDelay} = $forwardDelay if (defined($forwardDelay));
		$instance->{helloTime} = $helloTime; #helloTime is required
		$instance->{maxAge} = $maxAge if (defined($maxAge));
		$instance->{systemMacAddress} = $systemMacAddress if (defined($systemMacAddress));
		$instance->{priority} = $priority; #priority is required
		$instance->{vlan} = $vlan if (defined($vlan));
		
		#save this instance to the list
		push (@instances, $instance);
	}
	
	$in->{parsed}->{spanningTree}->{spanningTreeInstance} = \@instances;

} # end sub parse_stp

#extract the mst instances that are configured
sub list_mst_instances {
	my $in = shift;
	
	if (!(defined($in->{mst_config}))) {
		return ();
	}
	
	my @list = $in->{mst_config} =~ /^\s*(\d+)/mig;
	return @list;
} # end sub list_mst_instances


#print out the spanningTree SED element and all sub elements that exist
sub print_stp {
	my ( $in, $out ) = @_;
	
	_print_if_defined($in->{parsed}, 'spanningTree', $out);
} # end sub print_stp


sub parse_snmp {
	#Parses the SNMP Configuration
	my ( $in, $out ) = @_;
	
	#If for some reason we dont have snmp information
	#don't try and parse it
	if(!(defined($in->{snmp}))) {
		return undef;
	}
	
	#parse the SNMP communities
	$in->{snmp} =~ /Write Access\s*^[\s\-]*$([\s\S]*?)^\s*$/mi;
	my $communities = trim($1);
		
	my @communities_list = $communities =~ /(^.*$)/mig;
	
	foreach my $line (@communities_list) {
				
		#some version of the ProVision operating system have more columns than this
		#these columns should appear in this order in all versions
		if ($line =~ /\s*(\S+)\s+(\S+)\s+(\S+)\s*/) {
			my ($communityString, $mibView, $accessType ) = ($1, $2, $3);
		
			#ProVision refers to access type as 'restricted' which mean read only, and 'unrestricted'
			#which mean read/write.
			if ($accessType eq "Unrestricted") {
				$accessType = "RW"
			} else {
				$accessType = "RO"
			}
			
			$in->{parsed}->{snmp}->{community}->{$communityString}->{'accessType'} = $accessType;
			
			$in->{parsed}->{snmp}->{community}->{$communityString}->{'communityString'} = $communityString;
			
			$in->{parsed}->{snmp}->{community}->{$communityString}->{'mibView'} = $mibView;
		}
	}
	
	#parse sysContact. Note that the contact can be blank and we should not
	#create the element if the data does not exist
	if ($in->{system_information} =~ /System Contact\s+:\s*?(.*?)\s*$/mi) {
		my $contact = trim($1);
		if ($contact ne "") {
			$in->{parsed}->{snmp}->{'sysContact'} = $contact;
		}
	}
	
	#parse sysDescr
	#if we don't have enable access this data will be undef and we should not print the element
	if (defined($in->{sysDescr})) {
		my $descr = $in->{sysDescr};
		$descr =~ s/\s*\n\s*//;
		$descr =~ s/sysDescr.0 = //i;
		$descr = trim($descr);
		$in->{parsed}->{snmp}->{'sysDescr'} = $descr;
	}
	
	#parse sysLocation. Note that the contact can be blank and we should not
	#create the element if the data does not exist
	if ($in->{system_information} =~ /System Location\s+:\s*?(.*?)\s*$/mi) {
		my ($location) = trim($1);
		if ($location ne "") {
			$in->{parsed}->{snmp}->{'sysLocation'} = $location;
		}
	}
	
	#parse sysName. This element is required, and the device must have a hostname.
	my ($systemName) = trim($in->{system_information} =~ /^\s*System Name\s*:\s*(.+)$/mi);
	$in->{parsed}->{snmp}->{'sysName'} = $systemName;
	
	#parse sysObjectId. Note that the contact can be blank and we should not
	#create the element if the data does not exist
	if (defined($in->{sysObjectId})) {
		my $objId = $in->{sysObjectId};
		$objId =~ s/sysObjectId.0 = //i;
		$objId = trim($objId);
		$in->{parsed}->{snmp}->{'sysObjectId'} = $objId;
	}
	
	#parse the trap hosts
		
	#this RE matches the header lines of the trap hosts table and the data in that table.
	#It captures the trap host data up until the first blank line, which indicates the end of the table.
	if ($in->{snmp} =~ /(?:^\s*Address.*Community.*$)\s*(?:^[\s\-]*$)([\s\S]*?)(?:^\s*$)/mi) {
		my $trap_hosts = trim($1);
		
		my @hosts = $trap_hosts =~ /^[\s\S]*?$/mig;
		my $id = 1;
		foreach my $host (@hosts) {
			if ($host =~ /\s*(\d+\.\d+\.\d+\.\d+)\s*(\S+)/i) {
				my ($ip, $community) = ($1, $2);
				
				$in->{parsed}->{snmp}->{'trapHosts'}->{$id}->{'communityString'} = $community;
				$in->{parsed}->{snmp}->{'trapHosts'}->{$id}->{'ipAddress'} = $ip;
				
				$id++;
			}
		}
	}
	
	#parse trapSource
	#trap source can be configured in several ways. An IP address will appear if the 'Configured IP' policy is used.
	if ($in->{snmp} =~ /\s*Selection Policy\s*\:\s*Configured IP\s*Ip Address\s*\:\s*(\d+\.\d+\.\d+\.\d+)/mi) {
		$in->{parsed}->{snmp}->{'trapSource'} = trim($1);
	}
	
} #end parse_snmp


sub print_snmp {
	my ( $in, $out ) = @_;
	
	#if there is no snmp information
	#we don't want to open any tags
	if(!(defined($in->{parsed}->{snmp}))) {
		return undef;
	}
	
	$out->open_element("snmp");
	
	foreach my $communityName (_sort_unknown_data(keys %{$in->{parsed}->{snmp}->{community}} ) ) {
		
		my $community_data = $in->{parsed}->{snmp}->{community}->{$communityName};
		$out->open_element("community");
			_print_if_defined($community_data, 'accessType', $out);
			_print_if_defined($community_data, 'communityString', $out);
			_print_if_defined($community_data, 'mibView', $out);
		$out->close_element("community");
	}
	
	_print_if_defined($in->{parsed}->{snmp}, 'sysContact', $out);
	
	_print_if_defined($in->{parsed}->{snmp}, 'sysDescr', $out);
	
	_print_if_defined($in->{parsed}->{snmp}, 'sysLocation', $out);
	
	_print_if_defined($in->{parsed}->{snmp}, 'sysName', $out);
	
	_print_if_defined($in->{parsed}->{snmp}, 'sysObjectId', $out);
	
	
	foreach my $trap_host_id (_sort_unknown_data(keys %{$in->{parsed}->{snmp}->{trapHosts}} ) ) {
		my $trap_host_data = $in->{parsed}->{snmp}->{trapHosts}->{$trap_host_id};
		$out->open_element("trapHosts");
			_print_if_defined($trap_host_data, 'communityString', $out);
			_print_if_defined($trap_host_data, 'ipAddress', $out);
		$out->close_element("trapHosts");
	}
	
	_print_if_defined($in->{parsed}->{snmp}, 'trapSource', $out);
	
	$out->close_element("snmp");
}

sub vlan_list {
	#Parses the VLAN ID's for later
	#retrieval of detailed info
	my $in = shift;
	
	#If the "show vlans" command failed then there are no VLANs. Return an empty list.
	if(!(defined($in->{vlans}))) {
		return ();
	}

	#if there are no vlans configured this pattern will not match and $ids will be empty
	my @ids = $in->{vlans} =~ /^\s*(\d+)\s+\S+[\s\|]+\S+.*$/mig;
	return @ids;
}


sub parse_vlans {
	#Parse information for each vlan
	my ($in, $out) = @_;

	#If no VLAN exists, $in->{vlan} will not be defined
	if(!(defined($in->{vlan}))) {
		return undef;
	}

	my $vlans = $in->{vlan};
	
	#Iterate over the list of VLANs
	while (my ($vlan_number, $vlan_data) = each(%{$vlans})) {
		
		#Parse the VLAN ID which is needed later
		$vlan_data =~ /VLAN ID\s*\:\s*(\d+)/mi;
		my $id = $1;
		
		#parse the configuration source
		$vlan_data =~ /^\s*Status\s*\:\s*(\S+)\s*$/mi;
		my $source = $1;
		
		#The default source is local		
		my $config_source = "local";
		
		#In provision devices, VLAN 1 is the default VLAN
		if ($id eq "1") {
			$config_source = "default";
			
		#Learned VLANs may be indicated by the learned or the dynamic keyword
		} elsif ($source =~ /dynamic|learned/i) {
			$config_source = "learned";
		}
		
		$in->{parsed}->{vlan}->{$vlan_number}->{'configSource'} = $config_source;
		
		#parse the enabled status of the vlan
		#A VLAN is enabled if at least one of its interfaces is up
		#status could also be parsed form the result of "walkmib ifOperStatus", but that would
		#require enable level access. 
		if ($vlan_data =~ /\s*\S+\s+(?:untagged|tagged|auto)\s+\S+\s+(Up)/mi) {
			$in->{parsed}->{vlan}->{$vlan_number}->{'enabled'} = "true";
		} else {
			$in->{parsed}->{vlan}->{$vlan_number}->{'enabled'} = "false";
		}
		
		
		#parse the interface members of this VLAN
		my @interfaces = $vlan_data =~ /^\s*(\S+)\s+(?:Untagged|tagged|auto)\s+\S+\s+(?:Up|down)\s*$/mig;
		foreach my $interface (@interfaces) {
			$in->{parsed}->{vlan}->{$vlan_number}->{'interfaceMembers'}->{$interface}->{'interfaceMember'} = $interface;
			
			
			if ($vlan_data =~ /^\s*(?:$interface)\s+(tagged)\s+\S+\s+(?:Up|down)\s*$/mi) {
				$in->{parsed}->{'interface'}->{$interface}->{'vlan_trunk'}->{$vlan_number} = $vlan_number;
			}
			#$in->{parsed}->{vlan}->{$vlan_number}->{'interfaceMembers'}->{$interface}->{'tagged'} = $1;
			
		}
	
		
		# Build the Internal VLAN name which is needed later
		my $name;
		if ($vlan_number eq "1") {
			$name = "DEFAULT_VLAN";
		} else {
			$name = "VLAN$vlan_number";
		}
		
		#parse the friendly name (description) of this VLAN 
		$vlan_data =~ /^\s*Name\s*\:\s*(.+)$/mi;
		my $descr = trim($1);
		
		#parse the MTU of this VLAN
		my $mtu = 1500; #default to 1500 if we can't get anything else
		
		#if we have enable level acces we can read the MTU directly from the MIB
		if (defined($in->{ifMtu})) {
			
			#first we need to figure out the interface index based on the VLAN name
			$in->{ifName} =~/^\s*ifName\.(\d+) = $name\s*$/mi;
			my $ifIndex = $1;
			
			#then we can look up the MTU of that interface index
			$in->{ifMtu} =~ /ifMtu\.$ifIndex = (\S+)/mi;
			$mtu = trim($1);
			$mtu =~ s/[^\d]//g;
		
		#if we don't have enable access we may be able to determin the MTU by checking if
		#jumbo frames are enable. If they are not, then the default of 1500 is appropriate.				
		} elsif( ($vlan_data =~ /^\s*Jumbo\s*\:\s*Yes\s*$/mi) and (defined($in->{jumbos})) ){
			$in->{jumbos} =~ /^\s*In Use.*Ip-MTU\s*\:\s*(\S+)\s*$/mi;
			$mtu = trim($1);
			$mtu =~ s/[^\d]//g;
		}
		$in->{parsed}->{vlan}->{$vlan_number}->{'mtu'} = $mtu;
		
		#save the name of this vlan	
		$in->{parsed}->{vlan}->{$vlan_number}->{'name'} = $name;
		
		#save the friendly name of this vlan
		$in->{parsed}->{vlan}->{$vlan_number}->{'descr'} = $descr;
		
		#save the number of this vlan		
		$in->{parsed}->{vlan}->{$vlan_number}->{'number'} = $id;
		
		#store the description and name in a lookup table so we can convert the description to the name later
		$in->{parsed}->{lookup}->{vlan_descr_to_name}->{$descr} = $name;
		
	}
}

sub print_vlans {
	my ($in, $out) = @_;
	
	#if there are no VLANs parsed, we don't want to print any tags
	if(!(defined($in->{parsed}->{vlan}))) {
		return undef;
	}
	
	$out->open_element("vlans");
	
	foreach my $id (_sort_unknown_data(keys %{$in->{parsed}->{vlan}})) {
		my $vlan_data = $in->{parsed}->{vlan}->{$id};
		
		$out->open_element("vlan");
			_print_if_defined($vlan_data, 'configSource', $out);
			_print_if_defined($vlan_data, 'enabled', $out);
			foreach my $key (_sort_unknown_data( keys %{$vlan_data->{'interfaceMembers'}} ) ){
				#if ($vlan_data->{'interfaceMembers'}->{$key}->{'tagged'} =~/untagged/i) {
					_print_if_defined($vlan_data->{'interfaceMembers'}->{$key}, 'interfaceMember', $out);
				#}
			}
			_print_if_defined($vlan_data, 'mtu', $out);
			_print_if_defined($vlan_data, 'name', $out);
			_print_if_defined($vlan_data, 'number', $out);
		$out->close_element("vlan");	
	}
	
	$out->close_element("vlans");
}


#This function produces a list of ACLs configured on a device so that individual commands can be issued for each ACL
sub list_acls {
	my $in = shift;
	
	#we are only interested in the lines below the table header
	my ($acls) = $in->{ACLs} =~ /^[\s\-]+\-[\s\-]+$([\s\S]*)/mi;
	
	my @acl_names = $acls =~ /^\s*\S+\s+\S+\s+(\S+)\s*$/mig;
	
	return @acl_names;
}

#This function produces a list of Static filters configured on the device so that individual commands can be issued for each filter
sub list_static_filters {
	my $in = shift;
	
	#we only need to identify the filter index
	my (@filters) = $in->{static_filter} =~ /^\s*(\d+)\s*\|?\s*.+\s*\|/mig;
	
	return @filters;
}

sub parse_acls {
	my ($in, $out) = @_;
	
	#if ACL data does not exist, then we have nothing to do
	if ( !(defined($in->{ACL})) ) {
		return undef;
	}

	#ACL data is arranged in a hash, indexed by acl name
	#go through each ACL and parse it
	while (my ($acl_name, $acl_data) = each(%{$in->{ACL}})) {
		$acl_data = trim($acl_data);
		
		#ipV6 ACLs are not supported at the moment, they will be added later
		if( !($acl_data =~ /Type\s*\:\s*ipv6/mi) ){
			
			$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'name'} = "$acl_name";
			$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'mode'} = "stateful";
			
			#in HP networking devices there is an implicit drop all rule at the end of each ACL
			$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'defaultAction'} = "deny";
	
			#break up the data into blocks representing each filter entry
			#each entry starts with a line number and a defined action and ends with a blank line
			#or the end of the data
			my @entries = $acl_data =~ /(^\s*\d+\s*Action\:[\s\S]*?)(?:(?=^(?:\s*)$)|\Z)/mig;
			
			
			#boolean value indicating if this is a standard or extended ACL
			#standard ACLs have certain default values that are configurable in 
			#extended ACLs
			my $standard = $acl_data =~ /type\s*\:\s*standard/mi;
			
			#ACL entries are listed in the order that they are processed
			my $processOrder = 1;
			
			foreach my $entry (@entries) {
			
				#parse the line number and the action for this entry. We need the line number first so that we can index the entry in our hash
				my ($lineNumber, $action) = $entry =~ /(\d+)\s+Action\:\s+(\S+)\s+/mi;#
				$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'primaryAction'} = $action;
				
				$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'processOrder'} = $processOrder;
				$processOrder++;
				
################Parse Source Information################
				#parse the source ip, the source mask, the source port, the protocol, and if this entry is only for established connections
				#standard ACLs have no port information and apply only to ipV4 traffic, thus they can't apply only to established traffic.
				my ($sourceIP, $sourceMask, $sourcePortString);
				my ($proto, $established);
				
				if ($standard) {
					($sourceIP, $sourceMask) = $entry =~ /IP\s*\:\s*(\d+\.\d+\.\d+\.\d+)\s*Mask\s*\:\s*(\d+\.\d+\.\d+\.\d+)\s*/;
					$proto = "ip";
				} else {
					($sourceIP, $sourceMask, $sourcePortString) = $entry =~ /Src IP\s*\:\s*(\d+\.\d+\.\d+\.\d+)\s+Mask\:\s*(\d+\.\d+\.\d+\.\d+)\s+Port\(s\)\:(.*)$/mi;
					($proto, $established) = $entry =~ /Proto\s*\:\s*(\S+)(.*)$/mi;
				}
				
				#if there is source port information for this entry, parse it
				my ($sourceOperator, $sourcePort);
				if (defined($sourcePortString)) {
					($sourceOperator, $sourcePort) = $sourcePortString =~ /(\S+)\s+(\d+)/;
				}
				
				#HP Provision uses neq, while the SED uses ne
				if ( (defined($sourceOperator)) and ($sourceOperator eq "neq") ) {
					$sourceOperator = "ne";
				}
				
				#handle the special case of a 'range' operator
				if ( (defined($sourceOperator)) and ($sourcePortString =~ /(\d+) - (\d+)/i) ) {
					my ($startport, $endport) = ($1, $2);
					$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'sourceService'}->{'portRange'}->{'portStart'} = $startport if defined($startport);
					$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'sourceService'}->{'portRange'}->{'portEnd'} = $endport if defined($endport);
				} else {
					$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'sourceService'}->{'portExpression'}->{'port'} = $sourcePort if defined($sourcePort);
					$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'sourceService'}->{'portExpression'}->{'operator'} = $sourceOperator if defined($sourceOperator);
				}
				
				$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'sourceIpAddr'}->{'network'} = _HP_network_element($sourceIP, $sourceMask, 1);
				$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'protocol'} = _translate_HP_protocol_names($proto, $entry); #entry passed for logging purposes
################End Parse Source Information################


################Parse Destination Information################				
				#parse the destination ip, mask, and port information
				#standard ACLs don't have destination information and this pattern will not match
				my ($destIP, $destMask, $destPortString) = $entry =~ /Dst IP\s*\:\s*(\d+\.\d+\.\d+\.\d+)\s+Mask\:\s*(\d+\.\d+\.\d+\.\d+)\s+Port\(s\)\:(.*)$/mi;
				
				#if there is destination port information for this entry, parse it
				my ($destPort, $destOperator);
				if (defined($destPortString)) {
					($destOperator, $destPort) = $destPortString =~ /(\S+)\s+(\d+)/;
				}
				
				#HP Provision uses neq, while the SED uses ne
				if ( (defined($destOperator)) and ($destOperator eq "neq") ) {
					$destOperator = "ne";
				}
				
				#handle the special case of a 'range' operator
				if ( (defined($destOperator)) and ($destPortString =~ /(\d+) - (\d+)/i) ) {
					my ($startport, $endport) = ($1, $2);
					$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'destinationService'}->{'portRange'}->{'portStart'} = $startport if defined($startport);
					$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'destinationService'}->{'portRange'}->{'portEnd'} = $endport if defined($endport);
					
				} else {
					$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'destinationService'}->{'portExpression'}->{'port'} = $destPort if defined($destPort);
					$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'destinationService'}->{'portExpression'}->{'operator'} = $destOperator if defined($destOperator);
					
				}
				
				$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'destinationIpAddr'}->{'network'} = _HP_network_element($destIP, $destMask, 1) if defined($destIP);
				################End Parse Destination Information################					
				
				
				#if this entry is marked as established then make that value true. Default to false
				if ( (defined($established)) and ($established =~ /Established/) ) {
					$established = "true";					
				} else {
					$established = "false";
				}
				$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'established'} = $established;
				

				#parse the precedence information
				my ($precedence) = $entry =~ /Precedence\s*\:\s*(\S+)/mi;
				$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'precedence'} = _decode_precedence($precedence);
				
				
				#parse the tos information
				my ($tos) = $entry =~/TOS\s*\:\s*(\S+)/mi;
				$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'tos'} = _decode_TOS($tos);
				
				# if this entry logs hits, then set log to true. Default to false.
				my ($log) = $entry =~ /(\(log\))/mi;
				if (defined($log)) {
					$log = "true";
				} else {
					$log = "false";
				}
				$in->{parsed}->{acls}->{filterList}->{$acl_name}->{'filterEntry'}->{$lineNumber}->{'log'} = $log;
			
			
			} #end entry for loop
	
		} #end ipV6 if statement
		
	} #end ACL while loop
	
	#static filters are treated as a form of ACL so should be parsed here.
	_parse_filter_lists($in, $out);
	
	#parse information about where ACLs are applied
	_acl_application($in, $out);
	
} #end sub parse_acls


# $proto_in is the protocol to be decoded. This may be an iana protocol number or a text string of the protocol name
# $entry is the text of the entry from which this protocol was parsed. This is only needed so that we can log a failure
#    in case we get an unsupported protocol number.
sub _translate_HP_protocol_names {
	
	my $proto_in = shift;
	my $entry = shift;
	
	my $proto_out;
	
	#if we get undef, do nothing
	if ( !(defined($proto_in)) ) {
		return undef;
	}
	
	
	##### NOTE
	#
	# At the moment the adapter treats ip-in-ip and ip as the same and produces a SED with "ip" in the 'protocol' element.
	# This is not how the schema is written at this time, but it appears to be the way that core handles ip protocol elements.
	# Thus there is no way to differentiate 'all ip protocols' and 'ip-in-ip protocols'. If either core or the Schema changes
	# to be able to represent ip-in-ip then this section should change to match. 
	
	if ($proto_in =~ /ip-in-ip/i) { #case for ip-in-ip encapsulation
		$proto_out = "ip";
	} elsif ($proto_in =~ /ipv6/i) { #case for ipv6-in-ip encapsulation, may dispaly as IPV6 or IPV6-IN-IP
		$proto_out = "ipv6"
	} elsif ($proto_in =~ /ip/i) { #case for ip, which really means all ip protocols
		$proto_out = "ip";
	} elsif ($proto_in =~ /\d+/i) { #protocol may be specified as a port number
		$proto_out = get_protocol_name ($proto_in);
		
		#get_protocol_name returns "" if it has no match. Return undef so we don't print a blank element
		if ($proto_out eq "") {
			$proto_out = undef;
			if (defined($entry)) {
				$LOGGER->debug("Unable to translate ip protocol number '$proto_in' while parsing rule: $entry\n");
			} else {
				$LOGGER->debug("Unable to translate ip protocol number '$proto_in' while parsing an unknown rule");
			}
		}
	} else { # if nothing above matches then we should have a recognised protocol name that we can use directly
		$proto_out = lc($proto_in);
	}
	#####
	
	return $proto_out;
	
}



#We are representing a source port static filter as a drop all with a set of appliesTo conditions
#from the source port to each destination port that has a drop entry in the filter list
#we are representing a protocol static filter as a drop all with a set of appliesTo conditions
#for each destination port that has an entry in the filter list. The protocol entry is set to IP or IPv6
sub _parse_filter_lists {
	
	my ($in, $out) = @_;
	
	while (my ($filter_index, $filter_data) = each(%{$in->{filter}})) {
		
		#Parse source port filters
		if ( ($filter_data =~ /Filter Type\s*\:\s*Source[\s\-]Port/mi) ) {
		
			$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'defaultAction'} = "none";
			
			
			$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'filterEntry'}->{'1'}->{'log'} = "false";
			$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'filterEntry'}->{'1'}->{'primaryAction'} = "drop";
			$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'filterEntry'}->{'1'}->{'processOrder'} = "1";
			
			
			my ($source_port) = $filter_data =~ /Source Port\s*\:\s*(\d+)/mi;
			
			#this filter should only be applied to source/destination combinations where the action is drop
			my @applied_ports = $filter_data =~ /^\s*(\d+)\s*.*drop/mig;
			
			foreach my $destination (@applied_ports) {
				$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'appliesTo'}->{$destination}->{'source'} = $source_port;
				$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'appliesTo'}->{$destination}->{'destination'} = $destination;
			}
			
			$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'mode'} = "stateful";
			
			#Create a default name 'Static Filter <index>' for this filter so the namespace is less likely to collide
			$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'name'} = "Static Filter $filter_index";
			
			#default to layer2-bridging for the moment. This may not be the best phase to use.
			$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'processingPhase'}->{'core:packetProcessingPhase'} = "layer2-bridging";
		
		#parse protocol based filters
		} elsif ( ($filter_data =~ /Filter Type\s*\:\s*Protocol/mi) ) {
			
			my ($protocol) = $filter_data =~ /Frame Type\s*\:\s*(\S+)/mi;
			
			#at the moment only IP is relevant. Also including IPv6 as it is a trivial addition.
			if ( (lc($protocol) eq "ip") or (lc($protocol) eq "ipv6") ){
				$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'filterEntry'}->{'1'}->{'log'} = "false";
				$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'filterEntry'}->{'1'}->{'primaryAction'} = "drop";
				$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'filterEntry'}->{'1'}->{'processOrder'} = "1";
				$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'filterEntry'}->{'1'}->{'protocol'} = lc($protocol);
				
				#The filter is applied to destination ports where the action is drop
				my @applied_ports = $filter_data =~ /^\s*(\d+)\s*.*drop/mig;
				
				foreach my $destination (@applied_ports) {
					$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'appliesTo'}->{$destination}->{'destination'} = $destination;
				}
				
				$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'defaultAction'} = "none";
				
				$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'mode'} = "stateful";
				
				#Create a default name 'Static Filter <index>' for this filter so the namespace is less likely to collide
				$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'name'} = "Static Filter $filter_index";
				
				#default to layer2-bridging for the moment. This may not be the best phase to use.
				$in->{parsed}->{acls}->{filterList}->{"Static Filter $filter_index"}->{'processingPhase'}->{'core:packetProcessingPhase'} = "layer2-bridging";
			}
		}
		
	}
	
} #end sub _parse_filter_lists



#HP provision devices record precedence information for ACL entries as text
#decode the tect into the IP header precedence value
sub _decode_precedence {
	my $precedence_in = shift;
	
	#if precedence is undefined, leave it undefined
	if (!(defined($precedence_in))) {
		return undef;
	}
	
	#this hash related text descriptions of priority to numeric values
	my %precedence_table = ("routine" 			=> 0,
							"priority" 			=> 1,
							"immediate" 		=> 2,
							"flash" 			=> 3,
							"flash-override" 	=> 4,
							"critical" 			=> 5,
							"internet" 			=> 6,
							"network" 			=> 7,	);
	
	#if precedence is already a valid number, leave it alone
	if ($precedence_in =~ /[0-7]/) {
		return $precedence_in;
	#if precedence matches in our table, convert it to a number
	} elsif (defined($precedence_table{$precedence_in})) {
		return $precedence_table{$precedence_in};
	#otherwise precedence is undefined
	} else {
		return undef;
	}
} # end sub _decode_precedence


sub _decode_TOS {
	my $tos_in = shift;
	
	#if TOS is undefined, leave it undefined
	if (!(defined($tos_in))) {
		return undef;
	}
	
	#this hash related text descriptions of TOS to numeric values
	my %tos_table = (	"normal" 			=> 0,
						"max-reliability" 	=> 2,
						"max-throughput" 	=> 4,
						"minimize-delay" 	=> 8,	);

	#if TOS is already a valid number, leave it alone
	if ($tos_in =~ /\d+/) {
		return $tos_in;
	#if TOS matches in our table, convert it to a number
	} elsif (defined($tos_table{$tos_in})) {
		return $tos_table{$tos_in};
	#otherwise TOS is undefined
	} else {
		return undef;
	}
	
} #end sub _decode_TOS


#HP networking devices allow ACL entries in the format 1.2.3.4/0
#This should be converted to 0.0.0.0/0 for compatability reasons
sub _HP_network_element {
	my ($ip, $mask, $invert) = @_;
	
	my $result = create_network_element($ip, $mask, $invert);
	
	#in the case where we have a non-zero ip address and a zero mask, create network element
	#returns an ipV4 element of '0.0.0.0'. Confert this to a mask element with value '0'
	if (defined($result->{ipV4Mask}) and ($result->{ipV4Mask} eq "0.0.0.0")) {
		$result->{mask} = 0;
		delete $result->{ipV4Mask};
	}
	
	#If there is a numeric mask defined convert it to an ip mask
	#if there is a non-continuous mask defined use it directly
	my $temp_mask;
	if (defined($result->{mask})) {
		$temp_mask = get_mask($result->{mask});
	} elsif (defined($result->{ipV4Mask})) {
		$temp_mask = $result->{ipV4Mask};
	}
	#if for some reason there is no mask defined then skip the masking operation
	
	
	#HP devices allow mask elements that overlap with non-zero portions of an ip address.
	#This may cause problems, so we should zero out any masked portions of the ip address
	if (defined($temp_mask)) {
		
		#convert the ip address to a 32 bit integer. (Effectively binary representation)
		my $packed_address = pack('C4', split(/\./, $result->{address}));
		my $ip_num = unpack('N', $packed_address);

		#convert the mask to a matching 32 bit integer
		#my $temp_mask = get_mask($result->{mask});
		my $packed_mask = pack('C4', split(/\./, $temp_mask));
		my $mask_num = unpack('N', $packed_mask);

		#bitwise and between the ip and the mask will set bits in the ip address to zero if the matching
		#bits in the mask are zero. When bits in the mask are one the bits in the ip will be unchanged.
		$ip_num =  $ip_num & $mask_num;

		#convert the masked integer back into an ip address
		my $temp = pack('N', $ip_num);
		$result->{address} = join( '.', unpack('C4', $temp));	
	}
	
	#return the result
	return $result;
	
} # end sub _HP_network_element


# HP Provision devices associate ACLs with VLANs and interfaces while the SED
# assiciates VLANs and interfaces with ACLs. This function parses appliesTo data
# and reverses it so that it can be placed in the SED properly.
sub _acl_application {
	my ($in, $out) = @_;
	
	#ACL application to VLANs and interfaces are parsed seperately because they are
	#displayed by different commands and with different output syntax.
	#VLANs are indexed by VLAN name and interfaces are indexed by interface number to reduce the
	#likelyhood of namespace collision.
	
	#go through the ACL application information we have for each vlan
	while (my ($vlan_number, $vlan_data) = each(%{$in->{ACL_application}->{vlan}})) {
		
		# Check if this vlan has a VACL applied to it
		if ( ($vlan_data =~ /IPv4 VLAN\s*\:\s*(\S+)\s+Type/mi) or ($vlan_data =~ /^\s*VACL Access List\s*\:\s*(\S+)\s+Type/mi) ) {
			my $acl_name = $1;
			
			if(my $vlan_name = $in->{parsed}->{vlan}->{$vlan_number}->{name}) {

				my $VLAN_index = $vlan_name."VLAN";
				
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{filterListReference}->{name} = $acl_name;
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{filterListReference}->{direction} = "ingress";
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{filterListReference}->{appliesTo}->{source} = $vlan_name;
				#VLAN filters are applied as the packet enters the interface
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{phaseDiscriminator}->{'processingPhase'}->{'core:packetProcessingPhase'} = "layer2-bridging";
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{phaseDiscriminator}->{processingPhasePosition} = "before";
			}
		}
		
		# Check if this vlan has an inbound RACL applied to it
		if ( ($vlan_data =~ /IPv4 Inbound\s*\:\s*(\S+)\s+Type/mi) or ($vlan_data =~ /Inbound Access List\s*\:\s*(\S+)\s+Type/mi) ) {
			my $acl_name = $1;
			
			if(my $vlan_name = $in->{parsed}->{vlan}->{$vlan_number}->{name}) {

				my $VLAN_index = $vlan_name."Inbound";
				
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{filterListReference}->{name} = $acl_name;
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{filterListReference}->{direction} = "ingress";
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{filterListReference}->{appliesTo}->{source} = $vlan_name;
				#inbound RACLS are applied after the packet is routed
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{phaseDiscriminator}->{'processingPhase'}->{'core:packetProcessingPhase'} = "routing";
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{phaseDiscriminator}->{processingPhasePosition} = "after";
			}
		}
		
		# Check if this vlan has an outbound RACL applied to it
		if ( ($vlan_data =~ /IPv4 Outbound\s*\:\s*(\S+)\s+Type/mi) or ($vlan_data =~ /Outbound Access List\s*\:\s*(\S+)\s+Type/mi) ) {
			my $acl_name = $1;
			
			if(my $vlan_name = $in->{parsed}->{vlan}->{$vlan_number}->{name}) {

				my $VLAN_index = $vlan_name."Outbound";
				
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{filterListReference}->{name} = $acl_name;
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{filterListReference}->{direction} = "egress";
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{filterListReference}->{appliesTo}->{destination} = $vlan_name;
				#outbound RACLS are applied after the packet is routed
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{phaseDiscriminator}->{'processingPhase'}->{'core:packetProcessingPhase'} = "routing";
				$in->{parsed}->{acls}->{filterListsSection}->{$VLAN_index}->{phaseDiscriminator}->{processingPhasePosition} = "after";
			}
		}
	}
	
	#go through the ACL application information we have for each port
	while (my ($interface_name, $interface_data) = each(%{$in->{ACL_application}->{port}})) {
		
		# Check if this port has na ACL applied to it
		if ( (defined($interface_data)) and ($interface_data =~ /Inbound \s*\:\s*(\S+)\s*$/mi) ){
			my $acl_name = $1;

			$in->{parsed}->{acls}->{filterListsSection}->{$interface_name}->{filterListReference}->{name} = $acl_name;
			$in->{parsed}->{acls}->{filterListsSection}->{$interface_name}->{filterListReference}->{direction} = "ingress";
			$in->{parsed}->{acls}->{filterListsSection}->{$interface_name}->{filterListReference}->{appliesTo}->{source} = $interface_name;
			#interface filters are applied as the packet enters the interface
			$in->{parsed}->{acls}->{filterListsSection}->{$interface_name}->{phaseDiscriminator}->{'processingPhase'}->{'core:packetProcessingPhase'} = "layer2-bridging";
			$in->{parsed}->{acls}->{filterListsSection}->{$interface_name}->{phaseDiscriminator}->{processingPhasePosition} = "before";
		}
	}
	
} # end sub _acl_application



sub print_acls {
	my ($in, $out) = @_;
	
	#if no ACLs are defined there is nothing to do
	if ( !(defined($in->{ACL})) ) {
		return undef;
	}
	
	$out->open_element("filterLists");
	
	#print each filterlist inside of a filterListsSection
	foreach my $acl_name (_sort_unknown_data(keys %{$in->{parsed}->{acls}->{filterList}})) {
		$out->open_element("filterListsSection");
		$out->open_element("filterList");
		
		my $acl_data = $in->{parsed}->{acls}->{filterList}->{$acl_name};
		
		_print_if_defined($acl_data, 'defaultAction', $out);
		
		#print the interfaces/vlans that this list applies to
		while (my ($index, $value) = each(%{$acl_data->{appliesTo}}) ) {
			$out->open_element("appliesTo");
			if (ref($value) eq "HASH") {
				#need to print these manually as they are not in alphabetical order
				$out->print_element("source", $value->{source});
				$out->print_element("destination", $value->{destination});
			} else {
				$out->print_element("source", $index);
			}
			$out->close_element("appliesTo");
		}
		
		#print each filter entry
		foreach my $lineNumber (_sort_unknown_data( keys %{$acl_data->{'filterEntry'}} ) ){
						
			$out->open_element("filterEntry lineStart=\'" . "$lineNumber" . "\'");
			my $acl_entry = $acl_data->{'filterEntry'}->{$lineNumber};
			
			_print_if_defined($acl_entry, 'destinationIpAddr', $out);
			_print_if_defined($acl_entry, 'destinationService', $out);
			_print_if_defined($acl_entry, 'established', $out);
			_print_if_defined($acl_entry, 'log', $out);
			$out->print_element("name", $lineNumber);
			_print_if_defined($acl_entry, 'precedence', $out);
			_print_if_defined($acl_entry, 'primaryAction', $out);
			_print_if_defined($acl_entry, 'processOrder', $out);
			_print_if_defined($acl_entry, 'protocol', $out);
			_print_if_defined($acl_entry, 'sourceIpAddr', $out);
			_print_if_defined($acl_entry, 'sourceService', $out);
			_print_if_defined($acl_entry, 'tos', $out);
			
			$out->close_element("filterEntry");	
		}
		
		_print_if_defined($acl_data, 'mode', $out);
		_print_if_defined($acl_data, 'name', $out);
		
		_print_if_defined($acl_data, 'processingPhase', $out);
		
		$out->close_element("filterList");
		$out->close_element("filterListsSection");
	}


	# print filterlist application information as filterListReferences inside of filterListsSections
	foreach my $ifName (_sort_unknown_data( keys %{$in->{parsed}->{acls}->{filterListsSection}} ) ){
		$out->open_element("filterListsSection");
		
		my $acl_application = $in->{parsed}->{acls}->{filterListsSection}->{$ifName};
		
		_print_if_defined($acl_application, 'filterListReference', $out);
		_print_if_defined($acl_application, 'phaseDiscriminator', $out);
		
		$out->close_element("filterListsSection");
		
	}
	
	$out->close_element("filterLists");
	
} #end sub print_acls



# This function takes a hash, a key and an output printer and prints the SED element identified by the key
# if that element is defined in the hash.
# Inputs:
# $element is a hash of parsed data where the keys are names of SED elements
# $key is the name of a SED element of be printed
# $out is the output printer where the SED element will printed

sub _print_if_defined {
	my ( $element, $key, $out ) = @_;
	
	if (defined($element->{$key})) {
		$out->print_element($key, $element->{$key});
	}
}

# This function takes a list of data and checks to see if it is numeric.
# if the data is all numeric it is sorted numerically and returned.
# if there is any non-numeric data then the data is sorted alphabetically and returned
sub _sort_unknown_data {
	my @data = @_;
	my $numeric = 'true';
	
	foreach my $element (@data) {
		if ($element =~ /\D/) {
			$numeric = undef;
			last;
		}
	}
	if (defined($numeric)) {
		return sort {$a <=> $b} @data;
	} else {
		return sort @data;
	}	
}

#This subroutine takes a numeric code representing the device status and returns a string representation of the status.
#these numeris codes are returned from walkmib ifAdminStatus and walkmib ifOperStatus
#only codes for up and down are implemented at the momment as only those codes can be represented in the SED

#ifAdminStatus codes are:
# 1 = up
# 2 = down
# 3 = testing

#ifOperStatus codes are the same plus:
# 4 = unknown
# 5 = dormant
# 6 = notPresent
# 7 = lowerLayerDown

sub _decode_status {
	
	my $status = shift;
	my $result = undef;
	
	if ($status eq "1") {
		$result = "up";
	} elsif ($status eq "2") {
		$result = "down";
	}
	return $result;
} #end sub _decode_status


#This subroutine decodes IANAifType codes into their associated SED interfaceTypeType
sub _decode_type {
	
	my $type = shift;
	my $result = undef;
	
	if(defined($interface_types{$type})) {
		return $interface_types{$type};
	} else {
		return "unknown";
	}
} #end sub _decode_type


sub parse_routes {
	my ($in, $out) = @_;
	
	#each route occupies one line in the table. Lines also includes the table headers, but they will not match later
	my @lines = split /\n/, $in->{ip_route};
	
	#route number is a counter used to index routes stored in $in. it is not used in the SED
	my $route_number = 0;
	
	
	my %dynamic_route_reverse_lookup;
	
	
	foreach my $line (@lines) {
		my ($destination, $mask, $gateway, $vlan_id, $metric, $preference, $discard, $interface, $type, $subType) = (undef, undef, undef, undef, undef, undef, "false", undef, undef, undef);
		
		#match each line against against a set of pattersn that define different ways that the data may be layed out
		if ($line =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\/(\d{1,2})\s*(.+?)\s+(\d*)\s+(static|rip|ospf|connected)\s+(\S+|\s*)\s+(\d+)\s+(\d+)\s*$/mi) {
			#old /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\/(\d{1,2})\s*(.+?)\s+(\d*)\s+static\s+(\d+)\s+(\d+)/mi
			($destination, $mask, $gateway, $vlan_id, $type, $subType, $metric, $preference) = ($1, $2, $3, $4, $5, $6, $7, $8);
			
			#if there is no vlan listed under VLAN ID then the interface is null
			if ( !($vlan_id =~ /\d+/mi) ) {
				$vlan_id = "null";
			}
			
			#if the vlan_id matches an existing VLAN we can use that VLANs name as the interface name
			#if not, then use whatever we found before (null at this time)
			if ( defined($in->{parsed}->{vlan}->{$vlan_id}) ) {
				$interface = $in->{parsed}->{vlan}->{$vlan_id}->{'name'};
			} else {
				$interface = $vlan_id;
			}
			
		} elsif ($line =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s+(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s+(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s+(static|rip|ospf|connected)\s+(\d+)/mi) {
			($destination, $mask, $gateway, $type, $metric) = ($1, $2, $3, $4, $5);
			
			#the mask is listed in ip address form, convert it to bit form
			$mask = mask_to_bits($mask);
			
			#look up the gateway address in the arp table to see if we can find an interface associated with it
			if ($in->{arp_table} =~ /$gateway.*(\d+)\s*$/mi) {
				$interface = $1;
				
			#if we can't find an interface, set it to null
			} else {
				$interface = "null";
			}
		} else {
			#skip this line if it doesn't match a format we can parse
			next;
		}
		
		#HP Provision defines 'reject' and 'blackhole' as gateways that indicate the packet is dropped
		if ( !(defined($gateway)) ) {
			$discard = undef
		} elsif ($gateway =~ /reject|blackhole/mi) {
			$discard = "true";
			$gateway = undef;
		
		} 
		#special case for gateway addresses that are loopback interfaces
		elsif ( ($gateway =~ /lo\d+/i) and ($vlan_id eq "null") ){
			$interface = $gateway;
			$gateway = undef;
		}
		#if the gatway is not reject or blackhole and is not an ip address, drop it for now.
		elsif ( !($gateway =~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/mi) ) {
			$gateway = undef;
		}
		
		if ($type =~ /static|connected/) {
			$in->{parsed}->{static_routes}->{$route_number}->{'gatewayAddress'} = $gateway;
			
			#if the destination is 0.0.0.0/0 then this is the default gateway				
			if ( !(defined($destination)) ) {
				$in->{parsed}->{static_routes}->{$route_number}->{'defaultGateway'} = undef;
			} elsif ( ($destination eq "0.0.0.0") and ($mask eq "0")) {
				$in->{parsed}->{static_routes}->{$route_number}->{'defaultGateway'} = "true";
			} else {
				$in->{parsed}->{static_routes}->{$route_number}->{'defaultGateway'} = "false";
			}
			
			$in->{parsed}->{static_routes}->{$route_number}->{'destinationAddress'} = $destination;
			$in->{parsed}->{static_routes}->{$route_number}->{'destinationMask'} = $mask;
			$in->{parsed}->{static_routes}->{$route_number}->{'discard'} = $discard;
			$in->{parsed}->{static_routes}->{$route_number}->{'interface'} = $interface;
			$in->{parsed}->{static_routes}->{$route_number}->{'routeMetric'} = $metric;
			$in->{parsed}->{static_routes}->{$route_number}->{'routePreference'} = $preference;
			
			#If there is no destination then this line was not a valid route entry (it could be a blank line or a table header)
			#delete the hash entry so that it doesn't mess up the SED
			if ( !(defined($in->{parsed}->{static_routes}->{$route_number}->{'destinationAddress'})) ) {
				delete $in->{parsed}->{static_routes}->{$route_number};
			}
			
		} # end if for static routes
		else {
			
			#dynamic routes can be configured to have multiple paths to the same destination and the SED is
			#capable of representing this. The $dynamic_route_reverse_lookup Hash stores the unique destinations that we have
			#seen so far so that we can append any identical destination using the same routing protocol to the list of
			#next Hops
			
			if (defined($dynamic_route_reverse_lookup{"$destination\/$mask\/$type"})) {
				
				my $old_number = $dynamic_route_reverse_lookup{"$destination\/$mask\/$type"};
				
				$in->{parsed}->{dynamic_routes}->{$old_number}->{'routeNextHop'}->{$route_number}->{'adjacencyInterface'} = $interface;
				$in->{parsed}->{dynamic_routes}->{$old_number}->{'routeNextHop'}->{$route_number}->{'adjacencyMetric'} = $metric;
				$in->{parsed}->{dynamic_routes}->{$old_number}->{'routeNextHop'}->{$route_number}->{'adjacencyNode'} = $gateway;
				
				
			} else {
			
				$dynamic_route_reverse_lookup{"$destination\/$mask\/$type"} = $route_number;
				
				$in->{parsed}->{dynamic_routes}->{$route_number}->{'advertisingRoutingProcess'} = $type;
				$in->{parsed}->{dynamic_routes}->{$route_number}->{'routeNetwork'}->{'address'} = $destination;
				$in->{parsed}->{dynamic_routes}->{$route_number}->{'routeNetwork'}->{'mask'} = $mask;
				$in->{parsed}->{dynamic_routes}->{$route_number}->{'routeNextHop'}->{$route_number}->{'adjacencyInterface'} = $interface;
				$in->{parsed}->{dynamic_routes}->{$route_number}->{'routeNextHop'}->{$route_number}->{'adjacencyMetric'} = $metric;
				$in->{parsed}->{dynamic_routes}->{$route_number}->{'routeNextHop'}->{$route_number}->{'adjacencyNode'} = $gateway;
			}
		} # end if for dynamic routes
		
		$route_number++;
	}

} #end sub parse_routes

sub print_static_routes {
	my ($in, $out) = @_;
	
	if (defined($in->{parsed}->{static_routes})) {
		$out->open_element("staticRoutes");
		while (my ($index, $route_data) = each(%{$in->{parsed}->{static_routes}}) ) {
			$out->open_element("staticRoute");
			
			_print_if_defined($route_data, 'defaultGateway', $out);
			_print_if_defined($route_data, 'destinationAddress', $out);
			_print_if_defined($route_data, 'destinationMask', $out);
			_print_if_defined($route_data, 'discard', $out);
			_print_if_defined($route_data, 'gatewayAddress', $out);
			_print_if_defined($route_data, 'interface', $out);
			_print_if_defined($route_data, 'routeMetric', $out);
			_print_if_defined($route_data, 'routePreference', $out);
			
			$out->close_element("staticRoute");
		}
		$out->close_element("staticRoutes");
	}
} #end sub print_static_routes


sub print_dynamic_routes {
	my ($in, $out) = @_;
	
	if (defined($in->{parsed}->{dynamic_routes})) {
		$out->open_element("dynamicRoutes");
		while (my ($index, $route_data) = each(%{$in->{parsed}->{dynamic_routes}}) ) {
			$out->open_element("dynamicRoute");
			_print_if_defined($route_data, 'advertisingRoutingProcess', $out);
			_print_if_defined($route_data, 'routeNetwork', $out);
			
			while (my ($hop_index, $hop_data) = each(%{$route_data->{routeNextHop}}) ) {
				
				$out->open_element("routeNextHop");
				_print_if_defined($hop_data, 'adjacencyInterface', $out);
				_print_if_defined($hop_data, 'adjacencyMetric', $out);
				_print_if_defined($hop_data, 'adjacencyNode', $out);
				$out->close_element("routeNextHop");
			}
			$out->close_element("dynamicRoute");
		}
		$out->close_element("dynamicRoutes");
		
	}
	
} #end sub print_dynamic_routes


# Parse information about routing protocols such as RIP, OSPF, and BGP
# 
sub parse_routing_protocols {
	my ($in, $out) = @_;
	
	#Parse RIP information only if RIP is enabled
	if ( defined($in->{ip_rip}) and ($in->{ip_rip} =~ /RIP protocol\s*\:\s*enabled/mi) ) {
		
		if ($in->{ip_rip} =~ /Auto\-summary\s*\:\s*enabled/mi) {
			$in->{parsed}->{routing}->{rip}->{autoSummarization} = "true";
		} else {
			$in->{parsed}->{routing}->{rip}->{autoSummarization} = "false";
		}
		
		
		my @enabled_interfaces = ();
		my @disabled_interfaces = ();
		my @participating_networks = ();
		
		# Separate the portion of the output that represents RIP interfaces
		my ($networks) = $in->{ip_rip} =~ /RIP interface information\s+([\s\S]*)\s+RIP peer information/mi;
		my @networks = split /\n/, $networks;
		
		# lookup each ip address in the address to interface name lookup table
		# add the interface to the appropriate list
		foreach my $network (@networks) {
			if ($network =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s+(\S+)/i) {
				my ($ip, $status) = ($1, $2);
				
				if ( defined($in->{parsed}->{lookup}->{ip_to_interface}->{$ip}) ) {
					if ($status =~ /enabled/i) {
						push (@enabled_interfaces, $in->{parsed}->{lookup}->{ip_to_interface}->{$ip});
					} else {
						push (@disabled_interfaces, $in->{parsed}->{lookup}->{ip_to_interface}->{$ip});
					}
					
				}
			}
		} # end foreach
		
		
		# Store the lists only if they contain elements
		if (@enabled_interfaces > 0) {
			$in->{parsed}->{routing}->{rip}->{enabledInterface} = \@enabled_interfaces;
		}
		if (@disabled_interfaces > 0) {
			$in->{parsed}->{routing}->{rip}->{disabledInterface} = \@disabled_interfaces;
		}
		
		# separate the portion of the output that represents RIP peer information
		my ($peers) = $in->{ip_rip} =~ /RIP peer information\s+([\s\S]*)\s+/mi;
		my @peers = split /\n/, $peers;
		
		#subnet mask is default to 32 as this information is not available
		foreach my $peer (@peers) {
			if ( $peer =~ /(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})/i ) {
				my $ip = $1;
				
				my $participating_network = {address => $ip, mask => 32};
						
				push (@participating_networks, $participating_network);
			}
		} #end foreach
		
		
		# save the list only if it has elements
		if (@participating_networks > 0) {
			$in->{parsed}->{routing}->{rip}->{network} = \@participating_networks;
		}
		
		
		# find the routing protocols whose routes RIP is redistributing
		if (defined($in->{ip_rip_redistribute})) {
			my @types;
			
			if ($in->{ip_rip_redistribute} =~ /Route type\s+Status/mi) {
				@types = $in->{ip_rip_redistribute} =~ /(connected|static|ospf)\s+enabled/mig;
			} elsif ($in->{ip_rip_redistribute} =~ /Route type\s+RouteMap\s+Options/mi) {
				@types = $in->{ip_rip_redistribute} =~ /(connected|static|ospf)/mig;
			}
			
			# save the list only if it has elements
			if (@types > 0) {
				
				# alter the structure of the list so that is is easier to print.
				my @redistribution;
				foreach my $type (@types) {
					push (@redistribution, {"targetProtocol" => $type});
				}
				
				$in->{parsed}->{routing}->{rip}->{redistribution} = \@redistribution;
			}
		}
		
		#default to version 2 as that is what the device likely supports. Individual interfaces my be set to use
		#lower versions, but this is not representable in the SED.
		$in->{parsed}->{routing}->{rip}->{version} = "2";
	}
	
} # end sub parse_routing_protocols

sub print_routing_protocols {
	my ($in, $out) = @_;
	
	if (defined($in->{parsed}->{routing})) {
		$out->open_element("routing");
		_print_if_defined($in->{parsed}->{routing}, 'rip', $out);
		$out->close_element("routing");

	}
} # end sub print_routing_protocols



1;

__END__

=head1 Parsers

ZipTie::Adapters::HP::ProVision::Parsers

=head1 SYNOPSIS

	use ZipTie::Adapters::HP::ProVision::Parsers;
	create_config( $cliResponses, $xmlPrinter );
	print_config( $cliResponses, $xmlPrinter );	
	parse_system( $cliResponses, $xmlPrinter );	
	print_system( $cliResponses, $xmlPrinter );
	parse_chassis( $cliResponses, $xmlPrinter );
	print_chassis( $cliResponses, $xmlPrinter );
	parse_interfaces( $cliResponses, $xmlPrinter );	
	print_interfaces( $cliResponses, $xmlPrinter );
	parse_vlans( $cliResponses, $xmlPrinter );
	print_vlans( $cliResponses, $xmlPrinter );
	parse_stp( $cliResponses, $xmlPrinter );
	print_stp( $cliResponses, $xmlPrinter );
	list_mst_instances( $cliResponses, $xmlPrinter );
	list_acls( $cliResponses);
	list_static_filters( $cliResponses);
	parse_acls( $cliResponses, $xmlPrinter );
	print_acls( $cliResponses, $xmlPrinter );
	parse_routes( $cliResponses, $xmlPrinter );
	print_static_routes( $cliResponses, $xmlPrinter );
	print_dynamic_routes( $cliResponses, $xmlPrinter );
	parse_routing_protocols( $cliResponses, $xmlPrinter );
	print_routing_protocols( $cliResponses, $xmlPrinter );
	parse_supporting_info_telemetry( $cliResponses, $xmlPrinter );
	parse_interfaces_telemetry( $cliResponses, $xmlPrinter );
	parse_mac_telemetry( $cliResponses, $xmlPrinter );
	parse_arp_telemetry( $cliResponses, $xmlPrinter );
	parse_discovery_protocol_telemetry( $cliResponses, $xmlPrinter );
	parse_routing_telemetry( $cliResponses, $xmlPrinter );
	print_telemetry( $cliResponses, $xmlPrinter );

=head1 DESCRIPTION

Module with many static methods that take in a hash of
ProVision device responses and print out ZipTie model elements.

=head2 Methods

=over 12

=item C<create_config>

Uses the HP ProVision startup and running configs to put together a ZipTie
configurationRepository element. This element can later be printed into a SED
using print_config

=item C<print_config>

Given a that a configurationRepository element has already been parsed using create_config
this function prints out that SED element into a SED.

=item C<parse_system>

This function parses some top level SED elements such as model number, etc.

=item C<print_system>

This function prints the SED elements parsed by the parse_system function.

=item C<parse_chassis>

Parses information about the switch's physical box. This information includes
make, model, serial number, physical location, amount of memory, state of the
power supply, etc.

=item C<print_chassis>

This function prints the SED elements parsed by the parse_chassis function.

=item C<parse_interfaces>

Populate the ZipTie interfaces model. If enable level access is available this uses
the responses to various 'walkmib' commands. If enable level access is not available
this uses the responses from 'show interfaces brief' and show interfaces <$id>.
Note that these two methods may produce different results as some information must be
calculated or inferred instead of read directly when enable access is not available.

=item C<print_interfaces>

This function prints the SED elements parsed by the parse_interfaces function.

=item C<interface_list>

This function returns a list of the device's interfaces so that commands can be issued on a
per-interface basis. If enable level access is available this function uses the result of
'walkmib ifIndex'. If it is not available it uses the result of 'show interfaces brief'.
The former is preferable as it lists all interfaces including virtual interfaces and
out-of-band management interfaces.

=item C<parse_snmp>
Populate the ZipTie SNMP model. Some elements require enable level access to retrieve the
required data.

=item C<print_snmp>

This function prints the SED elements parsed by the parse_snmp function.

=item C<parse_vlans>
Populate the ZipTie VLAN model.

=item C<print_vlans>

This function prints the SED elements parsed by the parse_vlans function.

=item C<vlan_list>
This function returns a list of the VLANs currently configuted on the device so that commands
can be issued on a per-VLAN basis. This function uses the result of the "show vlans" command.
If VLANs are disabled then this command returns an empty list.

=item C<list_acls>
This function returns a list of ACLs configured on the device so that commands can be issued on
a per-ACL basis

=item C<parse_acls>
This function uses the responses from 'show access-list <acl name>' to extract information
about ACLs configured on the device

=item C<list_static_filters>
This function takes the hash of cli responses and returns a list of static filters configured on the device.
It is used so that further commands can be issued for each static filter.

=item C<print_acls>
This function prints the SED elements parsed by the parse_acls function, including static filters parsed
in the _parse_filter_lists function.

=item C<parse_routes>
This function parses static route information obtained from the 'show ip route' command. It may cross-reference
this information with data obtained from 'show vlans' and 'show arp' in order to determin interface names.
This function must be called before print_static_routes and print_dynamic_routes.

=item C<print_static_routes>
This function prints the static route information parsed in parse_routes.

=item C<print_dynamic_routes>
This function prints the dynamic route information parsed in parse_routes.

=item C<parse_routing_protocols>
This function parses information about routing protocols such as RIP, OSPF, and BGP.
Currently only RIP is supported.

=item C<print_routing_protocols>
This function prints the routing protocol information parsed in parse_routes.

=item C<parse_stp>
This function extracts STP information from device output

=item C<print_stp>
This function prints the spanningTree element of the SED

=item C<list_mst_instances>
This function returns a list of MST instances that are configured on the device so that commands may be issued for each one.

=back

=head2 Private Methods

=over 12

=item C<_decode_status>

This subroutine takes a numeric code representing the device status and returns a string representation of the status.
these numeris codes are returned from walkmib ifAdminStatus and walkmib ifOperStatus
only codes for up and down are implemented at the momment as only those codes can be represented in the SED

ifAdminStatus codes are:
 1 = up
 2 = down
 3 = testing

ifOperStatus codes are the same plus:
 4 = unknown
 5 = dormant
 6 = notPresent
 7 = lowerLayerDown

=item C<_decode_type>

This subroutine decodes IANAifType codes into their associated SED interfaceTypeType

=item C<_parse_interface>

This is a helper function for parse_interfaces that parses information about a single interface
given an interface ID. it is refined out in order to improve code readability.

=item C<_parse_interface_ethernet>

This helper function parses ethernet information for a single interface given an index.
This is refined out in order to improve code readability.

=item C<_parse_interface_ip>

This helper function parses IPv4 address information for a single interface.
This is refined out to improve code readability.

=item C<_parse_vlan_as_interface>

In the event that we do not have enable level access on the device we need to 
seperately parse interface information for vlans as they do not appear on the
standard interface list.

=item C<_print_if_defined>

This function is used to print an element into the SED if it is defined.
It takes a hash and the name of a SED element that must be a key in the hash.
It checks to see if the hash item is defined, and if it is it prints the SED element.
This is useful because it avoids printing empty SED tags.

=item C<_sort_unknown_data>

This function takes a list of variables and sorts them. If the list is entirely numeric
data then a numeric sort is used. If it contains alphabetic characters then it is sorted alphabetically.
This is usefull because we are storing indecies in a hash and we dont know if these indecies
are numeric or alphabetic. However we do want them to be printed in a consistent order.

=item C<_decode_precedence>

This function takes a textual representation of packet priority and turns it into a numeric representation.
If the text is undefined it leaves it undefined.
If the text is already a valid precedence number it returns it as is.
If the text does not mach a valid precedence it returns undef.

=item C<_HP_network_element>

This function takes an ip address, a subnet mask, and a boolean indicating if the mash is in inverse form.
It returns a hash containing an address and a mask suitable for printing in a SED.
This function wraps the ZipTie::Adapters::Util::create_network_element to solve the following issues:

non-zero ip addresses with zero masks were returning an ipV4Mask element of '0.0.0.0' where they
should return a mask element of '0'

non-zero bits in an ip address that are masked by the ip mask should be converted to zero bits so that
those addresses can be handled by core.

=item C<_acl_application>

This function uses previously parsed VLAN name information and interface index information to determine which
ACLs are applied to which interfaces and VLANs.

=item C<_parse_filter_lists>

This function transforms static filters configured on an HP ProVision Networking device into SED filterLists.
A port based static filter is represented as a drop all rule that is applied to combinations of source port and destination port.
A protocol based static filter is represented by a drop all rule that is applied to a set of destination ports and a specified protocol.
Static filters do not have a name, only an index, so they are given a default name of 'Static Filter' plus their index.
Since static filters are a form of ACL, their SED elements are printed in print_acls.

=item C<_parse_trunks>

This function identifies trunk groups and the interfaces that are assigned to them and adds the appropriate group membership
information to the parsed interface data.


=back

=head1 LICENSE
  Copyright IBM 2013

  The contents of this file are subject to the Mozilla Public License
  Version 1.1 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/
  
  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
  License for the specific language governing rights and limitations
  under the License.

=head1 AUTHOR

  Contributor(s): rkruse, Brent Gerig (brgerig@taylor.edu), David Tickner (djtickner@ca.ibm.com)
  Date: November 20, 2007
  Updated: July 5, 2013

=cut
