package ZipTie::Adapters::HP::ProVision::TelemetryParsers;

use strict;
use warnings;
use ZipTie::Adapters::Utils qw(strip_mac trim mask_to_bits);
use ZipTie::Adapters::HP::ProVision::Parsers qw(interface_list vlan_list);
#use MIME::Base64 'encode_base64';
use ZipTie::Logger;

use Exporter 'import';
our @EXPORT_OK = qw(parse_supporting_info_telemetry parse_mac_telemetry print_telemetry parse_interfaces_telemetry parse_arp_telemetry
    				list_lldp_neighbors list_cdp_neighbors parse_discovery_protocol_telemetry parse_routes_telemetry);


#This function parses basic information needed for a DiscoveryEvent. IP address is provided by Telemetry.pm which gets it from
#the connection document.

sub parse_supporting_info_telemetry {
	my ($in, $out) = @_;
	
	
	$in->{parsed}->{adminIp} = $in->{ip_address};
	
	#parse the system name
	my ($systemName) = trim($in->{system_information} =~ /^\s*System Name\s*:\s*(.+)$/mi);
	$in->{parsed}->{sysName} = $systemName;
	
	#parse the system object it. This needs enable access
	if (defined($in->{sysObjectId})) {
		my $objId = $in->{sysObjectId};
		$objId =~ s/sysObjectId.0 = //i;
		$objId = trim($objId);
		$in->{parsed}->{sysObjectId} = $objId;
	}
	
	#parse the system description. This needs enable access
	if (defined($in->{sysDescr})) {
		my $descr = $in->{sysDescr};
		$descr =~ s/\s*\n\s*//;
		$descr =~ s/sysDescr.0 = //i;

		$in->{parsed}->{sysDescr} = $descr;
	}

} # end sub parse_supporting_info_telemetry


#this function parses interface information needed for a DiscoveryEvent
#This function preferes to use mib information if it is available since this information
#is more likely to be present and formatted consistently accross many different device types

sub parse_interfaces_telemetry {
	my ($in, $out) = @_;
	
	my @interfaces = ();

	#prefer to use the mibs if we have enable access as they will be reliably formatted and present accross all systems
	if (defined($in->{ifName})) {

		#get a list of all interface indecies
		my @ifIndex = $in->{ifName} =~ /ifName\.(\d+) = \S+/mig;
		
		foreach my $ifIndex (@ifIndex) {
			my $interface;
			
			#parse the vendor specified interface name
			my ($name) = $in->{ifName} =~ /ifName\.$ifIndex = (\S+)/mi;
			$interface->{name} = $name;
			
			#parse the interface type
			my ($type) = $in->{ifType} =~ /ifType\.$ifIndex = (\S+)/mi;
			$type = _decode_type_telemetry($type);
			$interface->{type} = $type;
			
			#parse the operational status of the interface
			my ($operStatus) = $in->{ifOperStatus} =~ /ifOperStatus\.$ifIndex = (\S+)/mi;
			$operStatus = _decode_status_telemetry($operStatus);
			$interface->{operStatus} = $operStatus;
			
			#parse the interface's input bytes
			my ($inputBytes) = $in->{ifInOctets} =~ /ifInOctets\.$ifIndex = (\S+)/mi;
			$inputBytes =~ s/[\D]//mig;
			$interface->{inputBytes} = $inputBytes;
			
			
			my @ipEntries = ();
			
			#capture the ip addresses associated with this interface (if any)
			my @ips = $in->{ipAddressIfIndex} =~ /^\s*ipAddressIfIndex\.1.*\.(\d+\.\d+\.\d+\.\d+) = $ifIndex\s*$/mig;
			foreach my $ip (@ips) {
				#parse the subnet mask associated with this ip address
				my ($mask) = $in->{ipAddressPrefixOrigin} =~ /ipAddressPrefixOrigin.*$ip\.(\d+) = \d+/mi;
				my $entry = {'ipAddress' => $ip, 'mask' => $mask};
				push @ipEntries, $entry;
			}
			
			#if we found any ip addresses add then to the interface
			if (@ipEntries) {
				$interface->{ipEntry} = \@ipEntries;
			}
			
			#add this interface to the list of interfaces
			push @interfaces, $interface;
			
		}
		
		if (@interfaces) {
			$in->{parsed}->{interfaces} = \@interfaces;
		}
		
	}
	
	#if we have no enable access
	else {
		
		#get a list of interfaces (physical)
		my @names = interface_list($in);
		
		
		foreach my $name (@names){
			my $interface;
			
			$interface->{name} = $name;
			#default to ethernet 
			$interface->{type} = "ethernet";
			
			#parse the operational status of the interface
			my ($operStatus) = $in->{interfaces} =~ /^\s*$name(?:\-Trk\d+)?\s*\S+\s*\|\s*(?:no|yes)\s*(?:no|yes)\s*(up|down).*$/mig;
			$interface->{operStatus} = $operStatus;
			
			#parse the input bytes for this interface
			my ($inputBytes) = $in->{interfaces_detail} =~ /^\s*$name(?:\-Trk\d+)?\s+(\S+)/mig;
			$inputBytes =~ s/[\D]//mig;
			$interface->{inputBytes} = $inputBytes;
			
			#add the interface to the list
			push @interfaces, $interface;
		}
		
		#vlans are also interfaces, and they are the only interfaces likely to have ip information
		my @vlans = vlan_list($in);
		$in->{vlans_custom} = trim($in->{vlans_custom});
		
		
		foreach my $vlan (@vlans) {
			my $interface;
			my $name;
			
			#set the name to HP's naming standard
			if ($vlan eq "1") {
				$name = "DEFAULT_VLAN";
			} else {
				$name = "VLAN$vlan";
			}
			
			$interface->{name} = $name;
			$interface->{type} = "other";
			
			#parse the operational status of the interface
			my ($operStatus) = $in->{vlans_custom} =~ /^\s*$vlan\s+(up|down)/mi;
			$interface->{operStatus} = $operStatus;
			#vlans report input bytes of 0 in the MIBs, so use the same here
			$interface->{inputBytes} = 0;
		
		
			my @lines = split '\n', $in->{vlans_custom};
			my $line = "";
		
			my @ipEntries = ();
		
			#loop through the lines in the output until we find one that matches this vlan number
			until ($line =~ /^\s*$vlan\s+\S+\s+/i) {
				if (!@lines) {
					$line = "";
					last;
				}
				$line = shift @lines;
			}
		
		
		
			do {
				#if the current line has an ip address and a subnet mask then parse them
				if ($line =~ /(\d+\.\d+\.\d+\.\d+)\s+(\d+\.\d+\.\d+\.\d+)/i) {
					my ($ip, $mask) = ($1, $2);
					
					$mask = mask_to_bits($mask);
					my $entry = {'ipAddress' => $ip, 'mask' => $mask};
					push @ipEntries, $entry;
				}
				
				#if there are more lines in the output get the next one
				#otherwise set a flag to exit the loop
				if (@lines) {
					$line = shift @lines;
				} else {
					$line = "";
				}
				
			#if we are out of lines or we reach the start of entries for the next vlan
			#then stop processing ip entries
			} until ( ($line eq "") or ($line =~ /^\s*\d+\s+\S+\s+/i) );

			#if we found any ip addresses add them to the interface
			if (@ipEntries) {
				$interface->{ipEntry} = \@ipEntries;
			}
			
			#add the interface to the list
			push @interfaces, $interface;
		}
		
		if (@interfaces) {
			$in->{parsed}->{interfaces} = \@interfaces;
		}
		
	}

} #end sub parse_interfaces_telemetry



sub parse_mac_telemetry {
	my ($in, $out) = @_;
	
	my @macs = ();
	
	#break up the input into lines
	my @lines = split '\n', $in->{mac_table};
	
	foreach my $line (@lines) {
		
		#if the current line matches a mac table entry
		if ( my ($mac, $port, $vlan) = $line =~ /([0-9a-f\-]{12,13})\s+(\d+)\s+(\d+)?/i ) {
			
			#standardize the mac address
			$mac = strip_mac($mac);
			
			#set the VLAN name based on HP's naming convention
			if (defined($vlan)) {
				if ($vlan eq "1")  {
					$vlan = "DEFAULT_VLAN";
				} else {
					$vlan = "VLAN$vlan";
				}
			}
			
			#store the information in an entry
			my $entry = {'macAddress' => $mac, 'interface' => $port};
			if (defined($vlan)) {
				$entry->{'vlan'} = $vlan;
			}
			push @macs, $entry;
			
		}
	}
	
	#only store the mac addresses if there is at least one entry
	if (@macs) {
		$in->{parsed}->{neighbors}->{macTable} = \@macs;
	}

} #end sub parse_mac_telemetry


sub parse_arp_telemetry {
	my ($in, $out) = @_;
	
	my @arps = ();
	
	#break up the input into lines
	my @lines = split '\n', $in->{arp_table};
	
	foreach my $line (@lines) {
		my ($ip, $mac, $port) = (undef, undef, undef);
		#if the current line matches a arp table entry
		if ( $line =~ /^\s*(\d+\.\d+\.\d+\.\d+)\s+(\S+)\s+\S+(?:\s+(\d+))?/i ) {
			($ip, $mac, $port) = ($1, $2, $3);
			
			#standardize the mac address
			$mac = strip_mac($mac);
			
			#store the information in an entry
			my $entry = {'ipAddress' => $ip, 'macAddress' => $mac};
			if (defined($port)) {
				 $entry->{'interface'} = $port;
			}
			push @arps, $entry;
			
		}
	}
	
	#only store the arpTable if there is at least one entry
	if (@arps) {
		$in->{parsed}->{neighbors}->{arpTable} = \@arps;
	}
}



sub parse_discovery_protocol_telemetry {
	my ($in, $out) = @_;
	
	my @neighbours;
	
	#for each port where there is LLDP data
	foreach my $port (keys %{$in->{lldp_detail}}) {
		my $lldp_data = $in->{lldp_detail}->{$port};
		
		#default the protocol name
		my ($protocol) = "LLDP";
		
		#extract the system name. default to undef if there is no sysname entry
		my ($sysName) = $lldp_data =~ /SysName\s+:\s*?(.*?)\s*$/mi;
		if ( (!defined($sysName)) or (trim($sysName) eq "" ) ) {
			$sysName = undef;
		} else {
			$sysName = trim($sysName);
		}
		
		#extract the system description. default to undef if there is no sysDescr entry
		my ($sysDescr) = $lldp_data =~ /System Descr\s*\:\s*(.+)\s*$/mi;
		if ( (!defined($sysDescr)) or (trim($sysDescr) eq "" ) ) {
			$sysDescr = undef;
		} else {
			$sysDescr = trim($sysDescr);
		}
		
		#extract the remote interface name. There should be an entry, but default to undef if there is not
		my ($remoteInterface) = $lldp_data =~ /PortId\s*\:\s*(.+)\s*$/mi;
		if ( (!defined($remoteInterface)) or (trim($remoteInterface) eq "" ) ) {
			$remoteInterface = undef;
		} else {
			$remoteInterface = trim($remoteInterface);
		}
		
		#extract the local interface name. There should be an entry, but default to undef if there is not
		my ($localInterface) = $lldp_data =~ /Local Port\s*\:\s*(.+)\s*$/mi;
		if ( (!defined($localInterface)) or (trim($localInterface) eq "" ) ) {
			$localInterface = undef;
		} else {
			$localInterface = trim($localInterface);
		}
		
		#extract the ip Address there must be an entry here
		my ($ipAddress) = $lldp_data =~ /Address\s*\:\s*(\d+\.\d+\.\d+\.\d+)\s*$/mi;

		#populate a discovery protocol entry
		my $discoveryProtocolNeighbor = { 	'protocol' => $protocol,
											'sysDescr' => $sysDescr,
											'remoteInterface' => $remoteInterface,
											'localInterface' => $localInterface,
											'ipAddress' => $ipAddress,
		};
		if (defined($sysName)) {
			$discoveryProtocolNeighbor->{sysName} = $sysName;
		}

		# add the current neighbor to the list of neighbors
		push (@neighbours, $discoveryProtocolNeighbor);
	}
	
	#for each port where there is CDP data
	foreach my $port (keys %{$in->{cdp_detail}}) {
		my $cdp_data = $in->{cdp_detail}->{$port};
		
		#default the protocol name
		my $protocol = "CDP";
		
		#extract the system description. default to undef if there is no sysDescr entry
		my ($sysDescr) = $cdp_data =~ /Version\s*\:\s*(.+)\s*$/mi;
		if ( (!defined($sysDescr)) or (trim($sysDescr) eq "" ) ) {
			$sysDescr = undef;
		} else {
			$sysDescr = trim($sysDescr);
		}
		
		#extract the remote interface name. There should be an entry, but default to undef if there is not
		my ($remoteInterface) = $cdp_data =~ /Device Port\s*\:\s*(.+)\s*$/mi;
		if ( (!defined($remoteInterface)) or (trim($remoteInterface) eq "" ) ) {
			$remoteInterface = undef;
		} else {
			$remoteInterface = trim($remoteInterface);
		}
		
		#extract the local interface name. There should be an entry, but default to undef if there is not
		my ($localInterface) = $cdp_data =~ /^\s*Port\s*\:\s*(.+)\s*$/mi;
		if ( (!defined($localInterface)) or (trim($localInterface) eq "" ) ) {
			$localInterface = undef;
		} else {
			$localInterface = trim($localInterface);
		}
		
		#extract the platform. default to undef if there is no platform entry
		my ($platform) = $cdp_data =~ /^\s*Platform\s*\:\s*(.+)\s*$/mi;
		if ( (!defined($platform)) or (trim($platform) eq "" ) ) {
			$platform = undef;
		} else {
			$platform = trim($platform);
		}
		
		#extract the ip Address there must be an entry here
		my ($ipAddress) = $cdp_data =~ /Address\s*\:\s*(\d+\.\d+\.\d+\.\d+)\s*$/mi;

		#populate a discovery protocol entry
		my $discoveryProtocolNeighbor = { 	'protocol' => $protocol,
											'sysDescr' => $sysDescr,
											'remoteInterface' => $remoteInterface,
											'localInterface' => $localInterface,
											'ipAddress' => $ipAddress,
											'platform' => $platform,
		};

		# add the current neighbor to the list of neighbors
		push (@neighbours, $discoveryProtocolNeighbor);
		
	}
	
	#only store the discoveryProtocolNeighbors if there is at least one entry
	if (@neighbours) {
		$in->{parsed}->{neighbors}->{discoveryProtocolNeighbors} = \@neighbours;
	}
	
}

sub parse_routes_telemetry {
	my ($in, $out) = @_;
	
	my $neighbours;
	
	my @lines = split /\n/, $in->{ip_route};
	
	foreach my $line (@lines) {
		
		#reset this information each time through the loop
		my ($protocol, $ipAddress, $interface) = ("UNKNOWN", undef, undef);
		
		#IP routing information can come in several different formats. If a new format is discovered add it here
		
		if ($line =~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\/\d{1,2}\s*(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s+(\d*)\s+(static|rip|ospf)/mi) {
			($ipAddress, $interface, $protocol) = ($1, $2, uc($3));
			
			# this format contains vlan interface information. translate the vlan number into HP's VLAN naming format
			if ($interface eq "1") {
				$interface = "DEFAULT_VLAN";
			} else {
				$interface = "VLAN$interface"
			}

			
		} elsif ($line =~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\s*\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\s+(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s+(static|rip|ospf)/mi) {
				($ipAddress, $protocol) = ($1, uc($2));
		}
		
		
		#The protocol type 'STATIC' exists in the schema at this time, but it is not respected in ziptie server before 72 layer 2
		#Leaving it as static causes an illegalArguement exception in DiscoveryEventParser.java in the startElement method
		# when 'RoutingProtocol.valueOf(protocol)' is called since the RoutingProtocol enum does not contain the value 'STATIC'
		if ($protocol eq 'STATIC') {
			$protocol = 'UNKNOWN';
		}
		
		
		#we are only interested in an entry if we have a new ip address, otherwise we are just adding duplicate informating
		if ( (defined($ipAddress)) and (!defined($neighbours->{$ipAddress}))  ){		
			$neighbours->{$ipAddress}->{protocol} = $protocol;
			if (defined($interface)) {		
				$neighbours->{$ipAddress}->{interface} = $interface;
			}
			
		}
		
	
	} #end foreach
	
	my @routingNeighbors = ();
	
	#go through the table of information we constructed
	foreach my $ip (keys %{$neighbours}) {
	
		#create a routingNeighbor entry
		
		#routerId is set to the same thing as ipAddress since it is manditory.
		#it is not required in the schema, but ziptie server throws a null pointer exception if it is not present
		my $routingNeighbour = { 	'ipAddress' => $ip,
									'protocol' => $neighbours->{$ip}->{protocol},
									'routerId' => $ip,
	
		};
		if (defined($neighbours->{$ip}->{interface})) {
			$routingNeighbour->{interface} = $neighbours->{$ip}->{interface};
		}
		
		#add the routingNeighbor to the list
		push (@routingNeighbors, $routingNeighbour);
	}
	
	#only store the routingNeighbors if there is at least one entry
	if (@routingNeighbors) {
		$in->{parsed}->{neighbors}->{routingNeighbours} = \@routingNeighbors;
	}

} # end sub parse_routes_telemetry


#The discovery event must be opened and closed outside of this function.
#this is due to the unit test framework's inability to deal with a full discovery document
sub print_telemetry {
	my ($in, $out) = @_;
	
	#print basic information
	_print_if_defined($in->{parsed}, 'adminIp', $out);
	_print_if_defined($in->{parsed}, 'sysName', $out);
	_print_if_defined($in->{parsed}, 'sysObjectId', $out);
	_print_if_defined($in->{parsed}, 'sysDescr', $out);
	
	$out->open_element('neighbors');
	
	if (defined($in->{parsed}->{neighbors}->{routingNeighbours}) ) {
		$out->open_element('routingNeighbors');
		#routingNeighbours entries are elements with attributes rather than with sub elements, so set the
		#printer accordingly
		$out->attributes(1);
		foreach my $routing_neighbor_entry ( @{$in->{parsed}->{neighbors}->{routingNeighbours}} ) {
			$out->print_element("routingNeighbor", $routing_neighbor_entry);
		}
		$out->attributes(0);
		$out->close_element('routingNeighbors');
	}
	
	
	if (defined($in->{parsed}->{neighbors}->{discoveryProtocolNeighbors}) ) {
		$out->open_element('discoveryProtocolNeighbors');
		#discoveryProtocolNeighbors entries are elements with attributes rather than with sub elements, so set the
		#printer accordingly
		$out->attributes(1);
		foreach my $discovery_protocol_entry ( @{$in->{parsed}->{neighbors}->{discoveryProtocolNeighbors}} ) {
			$out->print_element("discoveryProtocolNeighbor", $discovery_protocol_entry);
		}
		$out->attributes(0);
		$out->close_element('discoveryProtocolNeighbors');
	}
	
	if (defined($in->{parsed}->{neighbors}->{arpTable}) ) {
		$out->open_element('arpTable');
		#arp entries are elements with attributes rather than with sub elements, so set the
		#printer accordingly
		$out->attributes(1);
		foreach my $arp_entry ( @{$in->{parsed}->{neighbors}->{arpTable}} ) {
			$out->print_element("arpEntry", $arp_entry);
		}
		$out->attributes(0);
		$out->close_element('arpTable');
	}
	
	if (defined($in->{parsed}->{neighbors}->{macTable}) ) {
		$out->open_element('macTable');
		
		#mac entries are elements with attributes rather than with sub elements, so set the
		#printer accordingly
		$out->attributes(1);
		foreach my $mac_entry ( @{$in->{parsed}->{neighbors}->{macTable}} ) {
			$out->print_element("macEntry", $mac_entry);
		}
		$out->attributes(0);
		
		
		$out->close_element('macTable');
	}
	
	$out->close_element('neighbors');
	
	
	$out->open_element('interfaces');
	foreach my $interface ( @{$in->{parsed}->{interfaces}} ) {
		
		#interfaces need to be printed differently if they have ip addresses
		#since the interface information is stored as attributes but the ip elements are sub elements.
		if (defined($interface->{ipEntry})) {
			
			my $type = $interface->{type};
			my $operStatus = $interface->{operStatus};
			my $name = $interface->{name};
			my $inputBytes = $interface->{inputBytes};
			
			#need to print the interface with attributes
			$out->open_element("interface type=\'" . "$type" . "\' operStatus=\'" . "$operStatus" . "\' name=\'" . "$name" . "\' inputBytes=\'" . "$inputBytes" . "\'");
			
			#and need to print the ipEntries as sub elements
			$out->attributes(1);
			foreach my $ipEntry (@{$interface->{ipEntry}}) {
				$out->print_element("ipEntry", $ipEntry);
			}
			$out->attributes(0);
			
			$out->close_element("interface");
		} else {
			#there are no sub elements, so just print the interfaces with attributes
			$out->attributes(1);
			$out->print_element("interface", $interface);
			$out->attributes(0);
			
		}
		
	}
	
	$out->close_element('interfaces');

} # end sub print_telemetry



#This function parses the operational status of an interface based on a numeric code in the MIB
#This function is seperate from the _decode_status function because the enumerations in
#ziptie-common.xsd and telemetry.xsd use different case for the first letter of each element

#ifOperStatus codes are 
# 1 = Up
# 2 = Down
# 3 = Testing
# 4 = Unknown
# 5 = Dormant
# 6 = NotPresent
# 7 = LowerLayerDown
sub _decode_status_telemetry {
	my $status = shift;
	
	my $codes = {
		 1 => "Up",
		 2 => "Down",
		 3 => "Testing",
		 4 => "Unknown",
		 5 => "Dormant",
		 6 => "NotPresent",
		 7 => "LowerLayerDown",
	};
	
	return $codes->{$status};
	
	
} #end sub _decode_status_telemetry


#this function parses the interface type based on a numeric code from the MIB
#this function is seperate from _decode_type because the enumerations in
#ziptie-common.xsd and telemetry.xsd contain different elements and thus need to be decoded differently
sub _decode_type_telemetry {
	my $type = shift;
	
	my %types = (	'6' => "ethernet",
					'9' => "tokenRing",
					'23' => "ppp",
					'24' => "softwareLoopback",
					'32' => "frameRelay",
					'37' => "atm",
					'39' => "sonet",
					'48' => "modem",
					'53' => "other", #53 is proprietary virtual - HP uses this for vlans
					'63' => "isdn", 
					'54' => "other", #54 is proprietary multiplexing - HP's code for port trunks
					'161' => "other",	); #161 is IEEE 802.3ad Link Aggregate
					
	if(defined($types{$type})) {
		return $types{$type};
	} else {
		return "unknown";
	}
} # end sub _decode_type_telemetry


# This function takes a hash, a key and an output printer and prints the SED element identified by the key
# if that element is defined in the hash.
# Inputs:
# $element is a hash of parsed data where the keys are names of SED elements
# $key is the name of a SED element of be printed
# $out is the output printer where the SED element will printed

sub _print_if_defined {
	my ( $element, $key, $out ) = @_;
	
	if (defined($element->{$key})) {
		$out->print_element($key, $element->{$key});
	}
}

1;


__END__

=head1 Parsers

ZipTie::Adapters::HP::ProVision::Parsers

=head1 SYNOPSIS

	parse_supporting_info_telemetry( $cliResponses, $xmlPrinter );
	parse_interfaces_telemetry( $cliResponses, $xmlPrinter );
	parse_mac_telemetry( $cliResponses, $xmlPrinter );
	parse_arp_telemetry( $cliResponses, $xmlPrinter );
	parse_discovery_protocol_telemetry( $cliResponses, $xmlPrinter );
	parse_routes_telemetry( $cliResponses, $xmlPrinter );
	print_telemetry( $cliResponses, $xmlPrinter );

=head1 DESCRIPTION

Module with many static methods that take in a hash of
ProVision device responses and print out ZipTie model elements.

=head2 Methods

=over 12

=item C<parse_supporting_info_telemetry>

This function parses the basic information needed for a DiscoveryEvent

=item C<parse_interfaces_telemetry>

This function parses the interface information needed for a DiscoveryEvent. This includes VLANs parsed
as interfaces. This function operates differently depending on if enable level access is present. This allows
a more reliable mthod of parsing to be used if enable level access is present.

=item C<parse_mac_telemetry>

This function parses the mac address table to be included in a DiscoveryEvent

=item C<parse_arp_telemetry>

This function parses the arp table to be included in a DiscoveryEvent

=item C<parse_discovery_protocol_telemetry>

This function parses the CDP and LLDP information to be included in a DiscoveryEvent

=item C<parse_routes_telemetry>

This function parses information in the routing table to be included in a DiscoveryEvent

=item C<print_telemetry>

This function prints a complete DiscoveryEvent document based on the information parsed in: parse_supporting_info_telemetry,
parse_interfaces_telemetry, parse_mac_telemetry, parse_arp_telemetry, parse_discovery_protocol_telemetry, parse_routing_telemetry.


=back

=head2 Private Methods

=over 12


=item C<_decode_status_telemetry>

This function decodes a status number into a text representation of interface status. This function operates in the same
way as _decode_status except that the enumeration of values that it targets is slightly different.

=item C<_decode_type_telemetry>

This function decodes an interface type number into a text representation of interface status. This function operates in the same
way as _decode_type except that the enumeration of values that it targets is slightly different.

=back

=head1 LICENSE
  Copyright IBM 2013
  
  The contents of this file are subject to the Mozilla Public License
  Version 1.1 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/
  
  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
  License for the specific language governing rights and limitations
  under the License.


=head1 AUTHOR

  Contributor(s): David Tickner (djtickner@ca.ibm.com)
  Date: August 30, 2013


=cut
