package ZipTie::Adapters::IBM::ISS_GX;

=pod

=head1 NAME

ZipTie::Adapters::IBM::ISS_GX - Backup operation for the IBM ISS_GX adapter.

=head1 SUBROUTINES

=cut

use strict;
use warnings;

use ZipTie::Adapters::IBM::ISS_SiteProtector::AgentDocumentCollector;
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::DBI;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

use ZipTie::Adapters::IBM::ISS_GX::Parsers qw(parse_tuning_parameters parse_content_analyzer parse_open_signatures parse_snort parse_applications parse_object_groups parse_system parse_chassis parse_interfaces parse_static_routes parse_filters parse_snmp create_config);

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

# An array of name/value pairs.  response_name will be the name of the policy in the $responses hash.  Mapps response_names to namespaces.
my @POLICIES = (
		{response_name => 'NetworkDefender_network', namespace => 'http://www.iss.net/cml/NetworkDefender/network'},
		{response_name => 'NetworkDefender_adapter', namespace => 'http://www.iss.net/cml/NetworkDefender/adapter'},
		{response_name => 'NetworkDefender_sensorCapabilities', namespace => 'http://www.iss.net/cml/NetworkDefender/sensorCapabilities'},
		{response_name => 'NetworkDefender_snmp', namespace => 'http://www.iss.net/cml/NetworkDefender/snmp'},
		{response_name => 'NetworkDefender_snort', namespace => 'http://www.iss.net/cml/NetworkDefender/snort'},
		{response_name => 'CommonDocuments_AgentStatusDetails', namespace => 'http://www.iss.net/cml/CommonDocuments/AgentStatusDetails'},
		{response_name => 'NetProtectModule_virtualSensors', namespace => 'http://www.iss.net/cml/NetProtectModule/virtualSensors'},
		{response_name => 'NetProtectModule_firewall', namespace => 'http://www.iss.net/cml/NetProtectModule/firewall'},
		{response_name => 'NetProtectModule_connectionEvents', namespace => 'http://www.iss.net/cml/NetProtectModule/connectionEvents'},
		{response_name => 'NetProtectModule_contentAnalyzer', namespace => 'http://www.iss.net/cml/NetProtectModule/contentAnalyzer'},
		{response_name => 'NetProtectModule_trons', namespace => 'http://www.iss.net/cml/NetProtectModule/trons'},
		{response_name => 'NetProtectModule_eventFilters', namespace => 'http://www.iss.net/cml/NetProtectModule/eventFilters'},
		{response_name => 'NetProtectModule_securityEvents', namespace => 'http://www.iss.net/cml/NetProtectModule/securityEvents'},
		{response_name => 'issues_csv', namespace => 'issues.csv'},
		{response_name => 'issue_responses_csv', namespace => 'issue_responses2.csv'},
		{response_name => 'NetProtectModule_tuningParameters', namespace => 'http://www.iss.net/cml/NetProtectModule/tuningParameters'}
	);

=pod

B<backup($package_name, $backup_doc)>

Adapter entry point for a backup operation.

=cut
sub backup
{
	my ($package_name, $backup_doc) = @_;

	my ($connection_path) = ZipTie::Typer::translate_document($backup_doc, 'connectionPath');

	my $filehandle = get_model_filehandle('ISS_GX', $connection_path->get_ip_address());

	# initialize the model printer
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, 'common');
	$printer->open_model();

	my $dbh = ZipTie::DBI->create_sybase_connection($connection_path, 'RealSecureDB');

	# Store the results of the SQL statments executed on the device.
	my $responses = {};

	# Required for create_config
	$responses->{policies} = \@POLICIES;

	my $agentDocumentCollector = ZipTie::Adapters::IBM::ISS_SiteProtector::AgentDocumentCollector->new($dbh,
		$connection_path->get_ip_address(), $connection_path->get_management_server_ip_address());

	# Collect all of the policy documents and populate $responses.
	foreach my $policy (@POLICIES)
	{
		$responses->{$policy->{response_name}} = $agentDocumentCollector->get_agent_documents($policy->{namespace});
	}

	$responses->{model} = $agentDocumentCollector->get_model();
	$responses->{guid} = $agentDocumentCollector->get_guid();

	$dbh->disconnect();

	parse_tuning_parameters($responses);

	parse_content_analyzer($responses);

	parse_snort($responses);

	parse_open_signatures($responses);

	parse_system($responses, $printer);

	parse_chassis($responses, $printer);

	create_config($responses, $printer);

	parse_applications($responses, $printer);

	parse_object_groups($responses, $printer);

	parse_filters($responses, $printer);

	parse_interfaces($responses, $printer);

	parse_snmp($responses, $printer);

	parse_static_routes($responses, $printer);

	# close out the ZiptieElementDocument
	$printer->close_model();

	close_model_filehandle($filehandle);
}

1;

__END__

=pod

=head1 AUTHOR(S)

David Payne
david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2013, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
