package ZipTie::Adapters::IBM::ISS_GX::Parsers;

=pod

=head1 NAME

ZipTie::Adapters::IBM::ISS_GX::Parsers - Parsers for the IBM ISS_GX adapter.

=head1 SUBROUTINES

=cut

use strict;
use warnings;

use Exporter 'import';
our @EXPORT_OK = qw(parse_tuning_parameters parse_content_analyzer parse_open_signatures parse_snort parse_applications parse_object_groups parse_system parse_chassis parse_interfaces parse_static_routes parse_filters parse_snmp create_config);

use ZipTie::Adapters::Utils qw(seconds_since_epoch mask_to_bits get_protocol_name get_crep);

use Data::Dumper;
use MIME::Base64 'encode_base64';
use Text::CSV;
use XML::Twig;
use ZipTie::Snort;

my $LOGGER = ZipTie::Logger::get_logger();

=pod

B<parse_tuning_parameters($in)>

Parses the devices tuning parameters.  During parsing the following data structures will be created:

$in->{_tuning_parameters}:
This is a hash that contain the name/value pairs of all tuning parameters for future parser use.

$in->{_open_signature_responses}:
This is a hash that contains any responses for Open Signature rules.  The key of the hash will be
the Open Signature ID, the value of the hash will be the responses.

This subroutine doesn't print any XML.

=cut
sub parse_tuning_parameters
{
	my ($in) = @_;

	$in->{_tuning_parameters} = {};
	$in->{_open_signature_responses} = {};

	my $twig = XML::Twig->new(twig_roots => {
			tuningParametersList => sub {
					my ($twig, $element) = @_;

					if (defined($element->{att}->{Enabled}) && $element->{att}->{Enabled} eq 'true' &&
						defined($element->{att}->{Name}))
					{
						my $name = $element->{att}->{Name};

						if (defined($element->{att}->{Value}))
						{
							my $value = $element->{att}->{Value};

							# Parse Open Signature responses.
							if ($name =~ /^np\.vs\.0\.issue\.(\d+)\.response$/)
							{
								my $issue_id = $1;

								# The open signature id can be derrived from the issue ID by
								# subtracting 6000000.  See the Open Signatures User Guide for more
								# information.
								my $open_signature_id = $issue_id - 6000000;

								$in->{_open_signature_responses}->{$open_signature_id} = $value;
							}

							$in->{_tuning_parameters}->{$name} = $value;
						}
						else
						{
							$in->{_tuning_parameters}->{$name} = '';
						}
					}

					return 1;
				}
		});
	foreach my $netProtectModule_tuning_parameters_policy (@{$in->{NetProtectModule_tuningParameters}})
	{
		$twig->parse($netProtectModule_tuning_parameters_policy);
	}
}

=pod

B<parse_content_analyzer($in)>

Content analyzer is called Data Loss Prevention in the SiteProtector UI, and is sometimes also
known as pam.

Parses content analyzer information from $in->{NetProtectModule_contentAnalyzer}.  Parsed data will
be stored in $in->{_content_analyzer_applications}, $in->{_content_analyzer_application_groups} and
$in->{_content_analyzer_rules}.  The data is stored for later use, this subroutine doesn't print any
XML.

=cut
sub parse_content_analyzer
{
	my ($in) = @_;

	$in->{_content_analyzer_applications} = [];
	$in->{_content_analyzer_application_groups} = [];
	$in->{_content_analyzer_rules} = {};

	foreach my $netProtectModule_content_analyzer_policy (@{$in->{NetProtectModule_contentAnalyzer}})
	{
		my @global_classifiers;
		my @applications;
		my @application_groups;
		my %rules;

		my $add_rule = sub {
				my ($element, $name, $type) = @_;

				foreach my $virtual_sensor ($element->children('VirtualSensor'))
				{
					unless (defined($virtual_sensor->{att}->{ignore}) && $virtual_sensor->{att}->{ignore} eq 'true' &&
						defined($virtual_sensor->{att}->{value}))
					{
						my $rule = {
								application => {$type => $name},
								'log' => 'false',
								'primaryAction' => 'next'
							};

						if (defined($virtual_sensor->{att}->{block}) && $virtual_sensor->{att}->{block} eq 'true')
						{
							$rule->{primaryAction} = 'deny';
						}

						if (defined($virtual_sensor->{att}->{logEvidence}) && $virtual_sensor->{att}->{logEvidence} ne 'None')
						{
							$rule->{'log'} = 'true';
						}

						push(@{$rules{$virtual_sensor->{att}->{value}}}, $rule);
					}
				}
			};

		# This is used to parse PredefinedEvents, UserDefinedEvents and UserCombinedEvents.
		my $event_parser = sub {
				my ($twig, $element) = @_;

				if (defined($element->{att}->{Name}))
				{
					my $application = {
							'ztApp:appName' => $element->{att}->{Name},
							'ztApp:appClassification' => []
						};

					foreach my $attribute ('MinMatch', 'Expression')
					{
						if (defined($element->{att}->{$attribute}))
						{
							push(@{$application->{'ztApp:appClassification'}}, {'ztApp:classifierName' => $attribute, 'ztApp:classifierValue' => $element->{att}->{$attribute}});
						}
					}

					if (defined($element->{att}->{Enabled}) && $element->{att}->{Enabled} eq 'true')
					{
						$add_rule->($element, $application->{'ztApp:appName'}, 'applicationID');
					}

					push(@applications, $application);
				}

				return 1;
			};

		my $twig = XML::Twig->new(twig_roots => {
				Tuning => sub {
						my ($twig, $element) = @_;

						if (defined($element->{att}->{enabled}) && $element->{att}->{enabled} eq 'true' &&
							defined($element->{att}->{Name}))
						{
							push(@global_classifiers, {'ztApp:classifierName' => 'TuningConfig', 'ztApp:classifierValue' => $element->{att}->{Name}});
						}
	
						return 1;
					},
				Protocol => sub {
						my ($twig, $element) = @_;
	
						if (defined($element->{att}->{enabled}) && $element->{att}->{enabled} eq 'true' &&
							defined($element->{att}->{Name}))
						{
							push(@global_classifiers, {'ztApp:classifierName' => 'Protocol', 'ztApp:classifierValue' => $element->{att}->{Name}});
						}

						return 1;
					},
				'PredefinedEvents/PredefinedEvent' => $event_parser,
				'UserDefinedEvents/UserDefinedEvent' => $event_parser,
				'UserCombinedEvents/UserCombinedEvent' => sub {
						my ($twig, $element) = @_;

						if (defined($element->{att}->{Name}))
						{
							my $application_group = {
									id => $element->{att}->{Name},
									application => []
								};

							foreach my $event ($element->children('PredefinedEvent'))
							{
								if (defined($event->{att}->{value}))
								{
									push(@{$application_group->{application}}, $event->{att}->{value});
								}
							}

							# NOTE: We do not support the "exclue" attribute.

							if (defined($element->{att}->{Enabled}) && $element->{att}->{Enabled} eq 'true')
							{
								$add_rule->($element, $element->{att}->{Name}, 'groupReference');
							}

							push(@application_groups, $application_group);
						}

						return 1;
					}
			});
		$twig->parse($netProtectModule_content_analyzer_policy);

		# Copy the global classifiers (tuning configs and applicaiton protocols) into each application.
		foreach my $application (@applications)
		{
			foreach my $global_classifier (@global_classifiers)
			{
				my %global_classifier = %{$global_classifier};
				push(@{$application->{'ztApp:appClassification'}}, \%global_classifier);
			}
		}
		push(@{$in->{_content_analyzer_applications}}, @applications);

		push(@{$in->{_content_analyzer_application_groups}}, @application_groups);

		foreach my $key (keys(%rules))
		{
			push(@{$in->{_content_analyzer_rules}->{$key}}, @{$rules{$key}});
		}
	}
}

=pod

B<parse_snort($in)>

Parses snort inforamtion from $in->{NetworkDefender_snort}.  Creates an array of
ZipTie::Snort objects stored in $in->{_snort} for later use.  Doesn't print
any XML.

=cut
sub parse_snort
{
	my ($in) = @_;

	# parse_interfaces must be called before parse_snort.  Calling parse_interfaces with $out set
	# to undef will cause parse_interfaces to store any parsed XML in a scalar for later output to
	# the SED.
	parse_interfaces($in, undef);

	$in->{_snort} = [];

	my %previous_snort_config_names;
	my $snort_config_number = 1;
	foreach my $networkDefender_snort_policy (@{$in->{NetworkDefender_snort}})
	{
		my $snort;

		my $snort_config_name = $snort_config_number++;

		my $snort_execution_enabled = 0;
		my $alert_before_pass = 0;

		my $twig = XML::Twig->new(start_tag_handlers => {
				SnortExec => sub {
					my ($twig, $element) = @_;

					if ($element->{att}->{SnortExecutionEnabled} eq 'true')
					{
						$snort_execution_enabled = 1;
					}

					return 1;
				},
				CommandLine => sub {
					my ($twig, $element) = @_;

					if (defined($element->{att}->{'dashdash-alert-before-pass'}) && $element->{att}->{'dashdash-alert-before-pass'} eq 'true')
					{
						$alert_before_pass = 1;
					}

					return 1;
				},
				policy => sub {
					my ($twig, $element) = @_;

					if (defined($element->{att}->{name}))
					{
						my $name = $element->{att}->{name};
						if (defined($previous_snort_config_names{$name}))
						{
							# Before the next statement $snort_config_name will be set to $snort_config_number.
							$snort_config_name = "$name:$snort_config_name";
						}
						else
						{
							$snort_config_name = $name;
							$previous_snort_config_names{$name} = 1;
						}
					}

					return 1;
				},
				Interfaces => sub {
					my ($twig, $element) = @_;

					if (defined($snort) && defined($element->{att}->{Adapter}))
					{
						my $interfaces = _map_letters_to_interfaces($in, $element->{att}->{Adapter});

						if (scalar(@{$interfaces}))
						{
							my $applies_to = [];

							foreach my $interface (@{$interfaces})
							{
								push(@{$applies_to}, {source => $interface});
							}

							$snort->set_applies_to($applies_to);
						}
					}

					return 1;
				},
				ConfFile => sub {
					my ($twig, $element) = @_;

					if (!$snort_execution_enabled)
					{
						return 0;
					}

					my $config = $element->{att}->{Contents};

					my $get_config = sub {
							my ($config_name) = @_;

							if ($config_name ne 'INITIAL_CONFIG')
							{
								$LOGGER->debug("Include keyword found in SNORT configuration not supported by the GX adapter, config '$config_name' not parsed.");
								return undef;
							}

							return $config;
						};

					$snort = ZipTie::Snort->new($snort_config_name, $get_config);

					if ($alert_before_pass)
					{
						$snort->alert_before_pass();
					}

					$snort->parse_config('INITIAL_CONFIG');

					return 1;
				},
				Rule => sub {
					my ($twig, $element) = @_;

					if (defined($snort))
					{
						$snort->parse_rules($element->{att}->{RuleName}, $element->{att}->{RuleString});
					}
					else
					{
						return 0;
					}

					return 1;
				}
			});
		$twig->parse($networkDefender_snort_policy);

		if (defined($snort) && $snort->has_rules())
		{
			push(@{$in->{_snort}}, $snort);
		}
	}
}

=pod

B<parse_open_signatures($in)>

Parses Open Signatures inforamtion from $in->{NetProtectModule_trons}.  Appends the Snort objects to
$in->{_snort} created by the Snort parser.  Doesn't print any XML.

=cut
sub parse_open_signatures
{
	my ($in) = @_;

	# First verify if Open Signatures are enabled.
	unless ((defined($in->{_tuning_parameters}->{'engine.opensignature.enabled'}) && $in->{_tuning_parameters}->{'engine.opensignature.enabled'} eq 'true') ||
		(defined($in->{_tuning_parameters}->{'pam.trons.enabled'}) && $in->{_tuning_parameters}->{'pam.trons.enabled'} eq 'true'))
	{
		$LOGGER->debug('Open Signatures not enabled, policies not parsed.');
		return;
	}

	my $open_signature_action_callback = sub {
			my ($sid) = @_;

			if (defined($in->{_open_signature_responses}->{$sid}))
			{
				if ($in->{_open_signature_responses}->{$sid} =~ /reset/i)
				{
					return 'deny';
				}
				elsif ($in->{_open_signature_responses}->{$sid} =~ /drop/i)
				{
					return 'drop';
				}
				else
				{
					return 'next';
				}
			}

			return undef;
		};

	# Tom Dechman from the ISS team mentioned that it's possible for users to define vairables in
	# the RuleString attribute.  If a call to parse_rules returns a value of 0, meaning no rules
	# found, then $rule_string_config will be set to the value of the RuleString attribute, and
	# parse_config will be called with 'INITIAL_CONFIG'.  This will allow any vairables to be
	# parsed by the config parser.
	my $rule_string_config = '';

	my $get_config = sub {
			my ($config_name) = @_;

			if ($config_name eq 'INITIAL_CONFIG')
			{
				# Return a standard configuration with the default vairable values.
				# See Variables section from the Open Signatures Author's Guide.
				return "ipvar HOME_NET any\n" .
					"ipvar EXTERNAL_NET any\n" .
					"ipvar DNS_SERVERS \$HOME_NET\n" .
					"ipvar SMTP_SERVERS \$HOME_NET\n" .
					"ipvar HTTP_SERVERS \$HOME_NET\n" .
					"ipvar SQL_SERVERS \$HOME_NET\n" .
					"ipvar TELNET_SERVERS \$HOME_NET\n" .
					"ipvar SNMP_SERVERS \$HOME_NET\n" .
					"portvar SSH_PORTS 22\n" .
					"portvar HTTP_PORTS 80\n" .
					"portvar SHELLCODE_PORTS !80\n" .
					"portvar ORACLE_PORTS 1521\n" .
					"portvar FILE_DATA_PORTS [\$HTTP_PORTS,110,143]\n";
			}
			elsif ($config_name eq 'RULE_STRING_CONFIG')
			{
				return $rule_string_config;
			}
			else
			{
				$LOGGER->debug("Include keyword found in Open Signature configuration not supported by the GX adapter, config '$config_name' not parsed.");
				return undef;
			}
		};

	my $config_number = 1;
	foreach my $netProtectModule_trons_policy (@{$in->{NetProtectModule_trons}})
	{
		my $snort = undef;
		my $config_name = $config_number++;
		my $rule_number = 1;

		my $twig = XML::Twig->new(start_tag_handlers => {
				policy => sub {
					my ($twig, $element) = @_;

					if (defined($element->{att}->{name}))
					{
						$config_name = $element->{att}->{name};
					}

					$snort = ZipTie::Snort->new($config_name, $get_config, 'Open_Signatures', $open_signature_action_callback);
					$snort->parse_config('INITIAL_CONFIG');

					# Open Signature policies apply to all interfaces.
					my $applies_to = [];
					foreach my $interface (@{_map_letters_to_interfaces($in, 'A,B,C,D,E,F,G,H,I,J,K,L,M,N,O,P')})
					{
						push(@{$applies_to}, {source => $interface});
					}
					$snort->set_applies_to($applies_to);

					return 1;
				},
				TronsRule => sub {
					my ($twig, $element) = @_;

					if (defined($element->{att}->{Enabled}) && $element->{att}->{Enabled} eq 'true' &&
						defined($element->{att}->{RuleString}) && defined($snort))
					{
						my $rules_parsed = $snort->parse_rules($config_name . ' ' . $rule_number++, $element->{att}->{RuleString});

						# If no rules were found then this may be a vairable decleration, attepmt
						# to parse it as part of a configuration.
						if ($rules_parsed == 0)
						{
							$rule_string_config = $element->{att}->{RuleString};
							$snort->parse_config('RULE_STRING_CONFIG');
						}
					}

					return 1;
				}
			});
		$twig->parse($netProtectModule_trons_policy);

		if (defined($snort) && $snort->has_rules())
		{
			push(@{$in->{_snort}}, $snort);
		}
	}

	# Remove duplicate applications on this device.
	ZipTie::Snort::application_dedup($in->{_snort});
}

=pod

B<parse_system($in, $out)>

Parses system information from $in->{CommonDocuments_AgentStatusDetails},
$in->{NetworkDefender_network} and $in->{NetworkDefender_snmp}.  Requires $in->{model} to be set.

Prints resulting XML to $out.

=cut
sub parse_system
{
	my ($in, $out) = @_;

	my $policies = _check_single_policy_existance($in, 'CommonDocuments_AgentStatusDetails', 'NetworkDefender_network', 'NetworkDefender_snmp');

	# Parse the following values from CommonDocuments_AgentStatusDetails
	my $version = undef;
	my $last_reboot = undef;
	my $twig = XML::Twig->new(twig_roots => {
				 section => sub {
					my ($twig, $element) = @_;

					if ($element->{att}->{name} eq 'agentinfo')
					{
						foreach my $param_element ($element->children('param'))
						{
							if ($param_element->{att}->{name} eq 'Base Version Number')
							{
								$version = $param_element->{att}->{value};
							}
							elsif ($param_element->{att}->{name} eq 'Last Restart')
							{
								# 2013-09-24 15:41:23
								if ($param_element->{att}->{value} =~ /(\d{4})-(\d{2})-(\d{2}) (\d{2}):(\d{2}):(\d{2})/)
								{
									$last_reboot = seconds_since_epoch($6, $5, $4, $3, $2, $1);
								}
							}
						}
					}

					return 1;
			}
		});
	$twig->parse($policies->{CommonDocuments_AgentStatusDetails});

	if (!defined($version))
	{
		$LOGGER->fatal('Failed to parse version from CommonDocuments_AgentStatusDetails.');
	}

	if (!defined($last_reboot))
	{
		$LOGGER->debug('Faild to parse lastReboot from CommonDocuments_AgentStatusDetails.  Setting to current time.');
		$last_reboot = time();
	}

	# Parse the following values from NetworkDefender_network
	my $hostname = undef;
	$twig = XML::Twig->new(twig_roots => {
				Config => sub {
					my ($twig, $element) = @_;

					if (defined($element->{att}->{HostName}))
					{
						if ($element->{att}->{HostName} =~ /^([^\.]+)/)
						{
							$hostname = $1;
						}
					}

					return 1;
			}
		});
	$twig->parse($policies->{NetworkDefender_network});

	if (!defined($hostname))
	{
		$LOGGER->fatal('Failed to parse hostname from CommonDocuments_AgentStatusDetails.');
	}

	# Store the SNMP information for later use by parse_snmp.
	$in->{_snmp} = {};

	# Parse the following values from NetworkDefender_snmp
	my $contact = undef;
	$twig = XML::Twig->new(twig_roots => {
				Config => sub {
					my ($twig, $element) = @_;

					if (defined($element->{att}->{SystemContact}) && $element->{att}->{SystemContact} ne 'Unknown')
					{
						$contact = $element->{att}->{SystemContact};
					}

					$in->{_snmp}->{SystemLocation} = $element->{att}->{SystemLocation};
					$in->{_snmp}->{SystemContact} = $element->{att}->{SystemContact};
					$in->{_snmp}->{SystemName} = $element->{att}->{SystemName};
			}
		});
	$twig->parse($policies->{NetworkDefender_snmp});

	# Print the XML
	$out->print_element('core:systemName', $hostname);

	$out->open_element('core:osInfo');
	$out->print_element('core:make', 'IBM');
	$out->print_element('core:name', $in->{model});
	$out->print_element('core:version', $version);
	$out->print_element('core:osType', 'Linux');
	$out->close_element('core:osInfo');

	$out->print_element('core:deviceType', 'IDS/IPS');
	$out->print_element('core:contact', $contact) if (defined($contact));
	$out->print_element('core:lastReboot', $last_reboot);
}

=pod

B<parse_chassis($in, $out)>

Requires $in->{model} and $in->{guid} to be set.

Prints chassis information to $out.

=cut
sub parse_chassis
{
	my ($in, $out) = @_;

	$out->open_element('chassis');
	$out->open_element('core:asset');
	$out->print_element('core:assetType', 'Chassis');
	$out->open_element('core:factoryinfo');
	$out->print_element('core:make', 'IBM');
	$out->print_element('core:modelNumber', $in->{model});
	$out->print_element('core:serialNumber', $in->{guid});
	$out->close_element('core:factoryinfo');
	$out->close_element('core:asset');
	$out->close_element('chassis');
}

=pod

B<create_config($in, $out)>

Stores the base64 encoded XML configurations in the configurationRepository element.

Requires $in->{policies} to be set by ISS_GX.pm, listing the text configurations to store.

=cut
sub create_config
{
	my ($in, $out) = @_;

	my $repository = {};
	$repository->{'core:name'} = '/'; # XmlPrint.pm has special handling making this an attribute.
	$repository->{'core:config'} = [];
	
	foreach my $policy (@{$in->{policies}})
	{
		# The csv policies are very large and not created by the users, skip them.
		if ($policy->{namespace} =~ /\.csv$/)
		{
			next;
		}

		# Sometimes there is more than one policy document with the same name.  If this is the case append each
		# config name with a number.
		my $append_policy_number = 0;
		my $policy_number = 1;
		if (scalar(@{$in->{$policy->{response_name}}}) != 1)
		{
			$append_policy_number = 1;
		}

		foreach my $text (@{$in->{$policy->{response_name}}})
		{
			my $config_text = '';

			# If this is a snort configuration we want to display the actual snort configuration, not the XML.
			# This is much more human readable.
			if ($policy->{namespace} =~ /snort/)
			{
				my $twig = XML::Twig->new(twig_roots => {
					ConfFile => sub {
							my ($twig, $element) = @_;

							if (defined($element->{att}->{Contents}))
							{
								$config_text = $element->{att}->{Contents};
							}

							return 1;
						}
					});
				$twig->parse($text);

				if (!defined($config_text))
				{
					next;
				}
			}
			# Format the XML to make it easier to view.
			else
			{
				eval
				{
					my $twig = XML::Twig->new(pretty_print => 'indented');
					$twig->set_indent("\t");
					$twig->parse($text);
					open(my $fh, '>', \$config_text);
					$twig->print($fh);
					close($fh);
				};
				# Catch any problem formatting the XML and display the original.
				if ($@)
				{
					$config_text = $text;
				}
			}

			push(@{$repository->{'core:config'}}, {'core:context' => 'active',
					'core:mediaType' => 'text/plain', # There is no XML media type in the schema.
					'core:name' => $policy->{response_name} . ($append_policy_number ? '_' . $policy_number++ : ''),
					'core:promotable' => 'false',
					'core:textBlob' => encode_base64($config_text)
				});
		}
	}

	# print the repository
	$out->print_element('core:configRepository', $repository);
}

=pod

B<parse_applications($in, $out)>

Prints layer7 applications to $out.

=cut
sub parse_applications
{
	my ($in, $out) = @_;

	my $applications_open = 0;

	foreach my $snort (@{$in->{_snort}})
	{
		my $snort_applications = $snort->get_applications();
		if (scalar(@{$snort_applications}))
		{
			if (!$applications_open)
			{
				$applications_open = 1;
				$out->open_element('applications');
			}

			$out->print_element('application', $snort_applications);
		}
	}

	if (defined($in->{_content_analyzer_applications}) && scalar(@{$in->{_content_analyzer_applications}}))
	{
		if (!$applications_open)
		{
			$applications_open = 1;
			$out->open_element('applications');
		}

		$out->print_element('application', $in->{_content_analyzer_applications});
	}

	if ($applications_open)
	{
		$out->close_element('applications');
	}
}

=pod

B<parse_object_groups($in, $out)>

Parses object groups and prints them to $out.

=cut
sub parse_object_groups
{
	my ($in, $out) = @_;

	$out->open_element('objectGroups');

	# Output content analyzer application groups
	foreach my $application_group (@{$in->{_content_analyzer_application_groups}})
	{
		$out->open_element('applicationGroup');
		$out->print_element('id', $application_group->{id});
		foreach my $application (@{$application_group->{application}})
		{
			$out->print_element('application', $application);
		}
		$out->close_element('applicationGroup');
	}

	# Output snort network groups
	foreach my $snort (@{$in->{_snort}})
	{
		$out->print_element('networkGroup', $snort->get_network_groups());
	}

	# Output snort service groups
	foreach my $snort (@{$in->{_snort}})
	{
		$out->print_element('serviceGroup', $snort->get_service_groups());
	}

	$out->close_element('objectGroups');
}

=pod

B<parse_filters($in, $out)>

Prints filter information to $out.

=cut
sub parse_filters
{
	my ($in, $out) = @_;

	$out->open_element('filterLists');

	_parse_NetProtectModule_connectionEvents($in, $out);
	_parse_NetProtectModule_firewall($in, $out);

	$in->{_virtual_sensors} = _parse_NetProtectModule_virtualSensors($in);

	_output_content_analyzer_rules($in, $out);

	_output_snort_rules($in, $out);

	_parse_idp_policies($in, $out);

	$out->close_element('filterLists');
}

sub _output_content_analyzer_rules
{
	my ($in, $out) = @_;

	if (!defined($in->{_virtual_sensors}))
	{
		$LOGGER->debug('$in->{_virtual_sensors} not defined, content analyzer rules not parsed..');
		return;
	}

	foreach my $virtual_sensor (keys(%{$in->{_content_analyzer_rules}}))
	{
		my $interface_list = $in->{_virtual_sensors}->{$virtual_sensor};

		$out->open_element('filterList');
		$out->print_element('defaultAction', 'none');
		foreach my $interface (@{$interface_list})
		{
			$out->open_element('appliesTo');
			$out->print_element('source', $interface);
			$out->close_element('appliesTo');
		}

		my $process_order = 1;
		foreach my $content_analyzer_rule (@{$in->{_content_analyzer_rules}->{$virtual_sensor}})
		{
			$content_analyzer_rule->{processOrder} = $process_order++;
			$out->print_element('filterEntry', $content_analyzer_rule);
		}

		$out->print_element('mode', 'stateful');
		$out->print_element('name', "$virtual_sensor Content Analyzer");

		$out->close_element('filterList');
	}
}

sub _output_snort_rules
{
	my ($in, $out) = @_;

	foreach my $snort (@{$in->{_snort}})
	{
		# Preprocessor rules
		my $preprocessor_rules = $snort->get_preprocessor_rules();
		if (scalar(@{$preprocessor_rules}))
		{
			$out->open_element('filterList');

			$out->print_element('defaultAction', 'none');

			if (defined($snort->get_applies_to()))
			{
				$out->print_element('appliesTo', $snort->get_applies_to());
			}

			$out->print_element('filterEntry', $preprocessor_rules);

			$out->print_element('mode', 'stateful');
			$out->print_element('name', $snort->get_config_name() . ' preprocessor');

			$out->close_element('filterList');
		}

		# Snort rules
		$out->open_element('filterList');

		$out->print_element('defaultAction', 'none');

		if (defined($snort->get_applies_to()))
		{
			$out->print_element('appliesTo', $snort->get_applies_to());
		}

		$out->print_element('filterEntry', $snort->get_rules());

		$out->print_element('mode', 'stateful');
		$out->print_element('name', $snort->get_config_name());

		$out->close_element('filterList');
	}
}

sub _parse_idp_policies
{
	my ($in, $out) = @_;

	if (!defined($in->{_virtual_sensors}))
	{
		$LOGGER->debug('$in->{_virtual_sensors} not defined, security events not parsed.');
		return;
	}

	my $responses = _parse_issue_responses_csv($in);

	if (!defined($responses))
	{
		$LOGGER->debug('$responses not defined, security events not parsed.');
		return;
	}

	my $stock_security_events = _parse_issues_csv($in, $responses);

	if (!defined($stock_security_events))
	{
		$LOGGER->debug('$stock_security_events not defined, security events not parsed.');
		return;
	}

	my $user_security_events = _parse_NetProtectModule_securityEvents($in, $stock_security_events);
	_print_idp_policies($in->{_virtual_sensors}, $stock_security_events, $user_security_events, $out);
}

sub _print_idp_policies
{
	my ($virtual_sensors, $stock_security_events, $user_security_events, $out) = @_;

	my @global_unmodified_events;
	my @global_modified_events;
	foreach my $stock_event (values(%{$stock_security_events}))
	{
		my $event_list;
		if (defined($stock_event->{modified}))
		{
			$event_list = \@global_modified_events;
		}
		else
		{
			$event_list = \@global_unmodified_events;
		}

		if (defined($user_security_events->{Global}) && defined($user_security_events->{Global}->{$stock_event->{id}}))
		{
			push(@{$event_list}, $user_security_events->{Global}->{$stock_event->{id}});
		}
		else
		{
			push(@{$event_list}, $stock_event);
		}
	}

	# We must ensure unmodified Global events are printed first so they apply to all trafic, and
	# Global_modified appears last so it's only evaluated if no user defined policies are evaluated.
	foreach my $virtual_sensor ('Global_unmodified', keys(%{$virtual_sensors}), 'Global_modified')
	{
		my $default_action;
		my $security_events;
		my $interface_list;
		if ($virtual_sensor eq 'Global_unmodified')
		{
			$default_action = 'none';
			$interface_list = $virtual_sensors->{Global};
			$security_events = \@global_unmodified_events;
		}
		elsif ($virtual_sensor eq 'Global_modified')
		{
			$default_action = 'permit';
			$interface_list = $virtual_sensors->{Global};
			$security_events = \@global_modified_events;
		}
		elsif ($virtual_sensor eq 'Global')
		{
			# The global sensors are printed in Global_modified, Global_unmodified and at the end of all
			# the user defined virtual sensors.
			next;
		}
		else
		{
			# Handling of security events contained inside user created protection domains.

			if (!defined($user_security_events->{$virtual_sensor}))
			{
				$LOGGER->debug("No security events defined for virtual sensor $virtual_sensor.");
				next;
			}

			$default_action = 'permit';
			$interface_list = $virtual_sensors->{$virtual_sensor};

			my @virtual_sensor_events;
			foreach my $event (values(%{$user_security_events->{$virtual_sensor}}))
			{
				# If the security event is disabled or is ignored don't print it.
				if ($event->{enabled} eq 'false')
				{
					next;
				}

				push(@virtual_sensor_events, $event);
			}

			# Append the modified Global security events to the end of every user created virtual sensor.
			foreach my $event (@global_modified_events)
			{
				# Don't print out the global version of the security event if the event was
				# over written by this virtual sensor.
				if (defined($user_security_events->{$virtual_sensor}->{$event->{id}}))
				{
					next;
				}

				push(@virtual_sensor_events, $event);
			}
			$security_events = \@virtual_sensor_events;
		}

		$out->open_element('filterList');
		$out->print_element('defaultAction', $default_action);
		foreach my $interface (@{$interface_list})
		{
			$out->open_element('appliesTo');
			$out->print_element('source', $interface);
			$out->close_element('appliesTo');
		}

		my $process_order = 1;

		foreach my $security_event (@{$security_events})
		{
			# If the security event is disabled or is ignored don't print it.
			if ($security_event->{enabled} eq 'false')
			{
				next;
			}
			
			_print_filter_entry($security_event->{name},  $security_event->{action},
					$process_order++, $security_event->{signature}, $out);
		}

		$out->print_element('mode', 'stateful');
		$out->print_element('name', $virtual_sensor);
		$out->close_element('filterList');
	}
}

sub _print_filter_entry
{
	my ($name, $action, $process_order, $signature, $out) = @_;

	$out->open_element('filterEntry');
	$out->print_element('log', 'true');
	$out->print_element('name', $name);
	$out->print_element('primaryAction', $action);
	$out->print_element('processOrder', $process_order);
	$out->open_element('vulnerability');
	$out->print_element('name', $signature);
	$out->print_element('type', 'Proventia');
	$out->close_element('vulnerability');
	$out->close_element('filterEntry');
}

sub _parse_NetProtectModule_securityEvents
{
	my ($in, $stock_security_events) = @_;

	my $user_security_events = {};

	my $twig = XML::Twig->new(twig_roots => {
				policy => sub {
					my ($twig, $element) = @_;

					foreach my $securityEventsList_element ($element->children('securityEventsList'))
					{
						my $issue_id = $securityEventsList_element->{att}->{issueID};
						my $virtual_sensor = $securityEventsList_element->{att}->{virtualSensor};

						if (defined($in->{disabled_virtual_sensors}->{$virtual_sensor}))
						{
							$LOGGER->debug("Skipping security event id $issue_id assigned to disabled virtual sensor $virtual_sensor.");
							next;
						}

						my $new_security_event = {id => $issue_id,
								virtual_sensor => $virtual_sensor,
								signature => $securityEventsList_element->{att}->{checkName},
								protocol => $securityEventsList_element->{att}->{protocol}};

						if ($securityEventsList_element->{att}->{block} eq 'true')
						{
							$new_security_event->{action} = 'deny';
						}
						else
						{
							$new_security_event->{action} = 'next';
						}

						if ($securityEventsList_element->{att}->{Enabled} eq 'false' || $securityEventsList_element->{att}->{ignore} eq 'true')
						{
							$new_security_event->{enabled} = 'false';
						}
						else
						{
							$new_security_event->{enabled} = 'true';
						}

						if (defined($stock_security_events->{$issue_id}))
						{
							$new_security_event->{description} = $stock_security_events->{$issue_id}->{description};
							$new_security_event->{name} = $stock_security_events->{$issue_id}->{name};
						}
						else
						{
							$LOGGER->debug("ERROR: Stock security event with ID $issue_id not found, skipping.");
							next;
						}

						if ($virtual_sensor ne 'Global')
						{
							$stock_security_events->{$issue_id}->{modified} = 1;
						}

						if (!defined($user_security_events->{$virtual_sensor}))
						{
							$user_security_events->{$virtual_sensor} = {};
						}

						$user_security_events->{$virtual_sensor}->{$issue_id} = $new_security_event;
					}

					return 1;
			}
		});
	foreach my $netProtectModule_securityEvents_policy (@{$in->{NetProtectModule_securityEvents}})
	{
		$twig->parse($netProtectModule_securityEvents_policy);
	}

	return $user_security_events;
}

# Parse the default issues_csv
sub _parse_issues_csv
{
	my ($in, $responses) = @_;

	if (scalar(@{$in->{issues_csv}}) != 1)
	{
		$LOGGER->debug('ERROR: ' . scalar(@{$in->{issues_csv}}) . ' elements in $in->{issues_csv} found, only 1 expected.');
		return undef;
	}

	my $csv = Text::CSV->new({binary => 1});

	my $stock_security_events = {};

	open(my $issues_csv_io_ref, '<', \$in->{issues_csv}->[0]);
	while (my $values = $csv->getline($issues_csv_io_ref))
	{
		if (scalar(@{$values}) != 14)
		{
			$LOGGER->debug('Expected 14 values, found: ' . Dumper($values));
			next;
		}

		# Issue Id,Report Name,Module Name,Sec Chk Id,Algorithm,False Positive,False Negative,Protocol,Is Attack?,Is Audit?,Old Cumberland Name,Decription,Priority,Is Status?
		my ($id, $signature, $module_name, $sec_chk_id, $description, $false_positive, $false_negative, $protocol, $is_attack,
				$is_audit, $old_cumberland_name, $name, $priority, $is_status) = @{$values};

		my $security_event = {id => $id, signature => $signature, protocol => $protocol,
				description => $description, name => $name};

		if (defined($responses->{$id}))
		{
			foreach my $key (keys(%{$responses->{$id}}))
			{
				$security_event->{$key} = $responses->{$id}->{$key};
			}
		}
		else
		{
			$security_event->{enabled} = 'false';
		}

		$stock_security_events->{$id} = $security_event;
	}
	close($issues_csv_io_ref);

	return $stock_security_events;
}

sub _parse_issue_responses_csv
{
	my ($in) = @_;

	if (scalar(@{$in->{issue_responses_csv}}) != 1)
	{
		$LOGGER->debug('ERROR: ' . scalar(@{$in->{issue_responses_csv}}) . ' elements in $in->{issue_responses_csv} found, only 1 expected.');
		return undef;
	}

	my $csv = Text::CSV->new({binary => 1});

	my $responses = {};
	open(my $issue_responses_csv_io_ref, '<', \$in->{issue_responses_csv}->[0]);
	while (my $values = $csv->getline($issue_responses_csv_io_ref))
	{
		if (scalar(@{$values}) != 11)
		{
			$LOGGER->debug('Expected 11 values, found: ' . Dumper($values));
			next;
		}

		# Issue Id,Drop Packet,Block Connection,Block Traffic,Block ICMP Traffic,Reset Victim,Reset Intruder,Reject Intruder,Rewrite Victim,Rewrite Intruder,Display
		my ($id, $drop, $block_connection, $block_traffic, $block_icmp_traffic, $reset_victim, $reset_intruder, $reject_intruder, $rewrite_victim, $rewrite_intruder, $display) = @{$values};
		my $action;
		if ($block_connection eq 'Y' || $block_traffic eq 'Y' || $block_icmp_traffic eq 'Y' || $reset_victim eq 'Y' || $reset_intruder eq 'Y' ||
			$reject_intruder eq 'Y' || $rewrite_victim eq 'Y' || $rewrite_intruder eq 'Y')
		{
			$action = 'deny';
		}
		elsif ($drop eq 'Y')
		{
			$action = 'drop';
		}
		else
		{
			$action = 'next';
		}

		# If the security event exists in the issues_response file then the event is enabled, otherwise it's disabled.
		$responses->{$id} = {id => $id, enabled => 'true', action => $action};
	}
	close($issue_responses_csv_io_ref);

	return $responses;
}

# Parse virtualSensor policies.  This creates $in->{_virtual_sensors}.
sub _parse_NetProtectModule_virtualSensors
{
	my ($in) = @_;

	my $virtual_sensors = {};

	# Keeps track of any virtual sensor that has been disabled.  Used when parsing security events.
	# Security events assigned to disabled virtual sensors are ignored.
	$in->{disabled_virtual_sensors} = {};

	my $twig = XML::Twig->new(twig_roots => {
				policy => sub {
					my ($twig, $element) = @_;

					my $schema_version = $element->{att}->{'schema-version'};
					if (!defined($schema_version) || $schema_version ne '5.0.0')
					{
						return 1;
					}

					foreach my $virtual_sensor_element ($element->children('virtualSensorElement'))
					{
						my $virtual_sensor_name = $virtual_sensor_element->{att}->{VSName};
						if (!defined($virtual_sensor_name))
						{
							$LOGGER->debug('Virtual attribute VSName not found, skipping.');
							next;
						}

						if (defined($virtual_sensors->{$virtual_sensor_name}))
						{
							$LOGGER->debug("Virtual Sensor alreday parsed $virtual_sensor_name, skipping virtual sensor.");
							next;
						}

						my $enabled = $virtual_sensor_element->{att}->{Enabled};
						if (defined($enabled) && $enabled eq 'false')
						{
							$in->{disabled_virtual_sensors}->{$virtual_sensor_name} = 1;
							$LOGGER->debug("Virtual sensor $virtual_sensor_name is disabled.");
							next;
						}
	
						my $interface = $virtual_sensor_element->{att}->{Interface};
						if (defined($interface))
						{
							my $interfaces = _map_letters_to_interfaces($in, $interface);
							if (scalar(@{$interfaces}))
							{
								$virtual_sensors->{$virtual_sensor_name} = $interfaces;
							}
							else
							{
								$LOGGER->debug("No valid interfaces found in Interface attribute for virtual sensor $virtual_sensor_name, skipping virtual sensor.");
							}
						}
						else
						{
							$LOGGER->debug("Interface attribute not found for virtual sensor $virtual_sensor_name, skipping virtual sensor.");
						}
					}

					return 1;
			}
		});
	foreach my $netProtectModule_vurtualSensors_policy (@{$in->{NetProtectModule_virtualSensors}})
	{
		$twig->parse($netProtectModule_vurtualSensors_policy);
	}

	return $virtual_sensors;
}

# Parse any NetProtectModule ConnectionEvents policies as filterLists.  parse_interfaces needs to be called
# before this subroutine.  Prints the XML to $out.
sub _parse_NetProtectModule_connectionEvents
{
	my ($in, $out) = @_;

	if (!defined($in->{NetProtectModule_connectionEvents}))
	{
		return;
	}

	my $filter_lists = {};

	my $process_orders = {};

	my $twig = XML::Twig->new(twig_roots => {
				policy => sub {
					my ($twig, $element) = @_;

                    my $current_policy_name;
                    if (defined($element->{att}->{name}))
                    {  
                        $current_policy_name = $element->{att}->{name};
                    }
                    else
                    {  
                        $current_policy_name = 'no_policy_name_connectionEvent';
                    }

					# If the filterList doesn't exist, create it.
					if (!defined($filter_lists->{$current_policy_name}))
					{
						$filter_lists->{$current_policy_name} = {
								defaultAction => 'none',
								filterEntry => [],
								appliesTo => [],
								name => $current_policy_name,
								mode => 'stateful'
							};

						# Connection Events apply to all policies that are non-blocking
						foreach my $interface (values(%{$in->{_interface_letter_name_mapping}}))
						{
							if (defined($in->{_non_blocking_ports}->{$interface}))
							{
								next;
							}

							push(@{$filter_lists->{$current_policy_name}->{appliesTo}}, {source => $interface});
						}
						$process_orders->{$current_policy_name} = 1;
					}

					foreach my $connection_event_element ($element->children('connectionEventElement'))
					{
						my $filter_entry = {
								sourceIpAddr => [],
								destinationIpAddr => [],
								sourceService => [],
								destinationService => [],
								'log' => 'true'            # All Connection Events log
							};

						my $event_name = _get_FirewallRule_attribute($connection_event_element, 'Event', $current_policy_name, undef);
						if (!defined($event_name))
						{
							next;
						}
						$filter_entry->{name} = $event_name;
						
						my $enabled = _parse_firewall_enabled($connection_event_element, $current_policy_name, $event_name);
						if (!defined($enabled))
						{
							next;
						}

						my $block = _get_FirewallRule_attribute($connection_event_element, 'block', $current_policy_name, $event_name);
						if (!defined($block))
						{
							next;
						}
						$filter_entry->{primaryAction} = ($block eq 'true') ? 'deny' : 'next';

						my $protocol = _parse_firewall_protocol($connection_event_element, $current_policy_name, $event_name);
						if (!defined($protocol))
						{
							next;
						}
						if ($protocol ne 'all')
						{
							$filter_entry->{protocol} = $protocol;
						}

						# AddressAndPort is of type firewall:SourceTargetType, schema version 5 and later.
						if (!_parse_address_and_port($connection_event_element, $filter_entry, 'addressPort', $current_policy_name, $event_name))
						{
							next;
						}

						$filter_entry->{processOrder} = $process_orders->{$current_policy_name}++;

						push(@{$filter_lists->{$current_policy_name}->{filterEntry}}, $filter_entry);
					}

					return 1;
				}
		});
	foreach my $netProtectModule_connectionEvents_policy (@{$in->{NetProtectModule_connectionEvents}})
	{
		$twig->parse($netProtectModule_connectionEvents_policy);
	}

	_print_filter_lists($filter_lists, $out);
}

# Parse any NetProtectModule firewall policies as filterLists.  parse_interfaces needs to be called
# before this subroutine.  Prints the XML to $out.
sub _parse_NetProtectModule_firewall
{
	my ($in, $out) = @_;

	# No firewall to parse, return imedeatly.
	if (!defined($in->{NetProtectModule_firewall}))
	{
		return;
	}

	# Hash of filterLists, the key is the name of the filter list.
	my $filter_lists = {};

	# The current process order, the key is the name of the filter list.
	my $process_orders = {};

	my $twig = XML::Twig->new(twig_roots => {
				policy => sub {
					my ($twig, $element) = @_;

					my $current_policy_name;
					if (defined($element->{att}->{name}))
					{
						$current_policy_name = $element->{att}->{name};
					}
					else
					{
						$current_policy_name = 'no_policy_name';
					}

					foreach my $firewall_rule_element ($element->children('FirewallRule'))
					{
						my $filter_entry = {
								sourceIpAddr => [],
								destinationIpAddr => [],
								sourceService => [],
								destinationService => []
							};

						my $rule_id = _get_FirewallRule_attribute($firewall_rule_element, 'RuleId', $current_policy_name, undef);
						if (!defined($rule_id))
						{
							next;
						}
						$filter_entry->{name} = $rule_id;

						my $enabled = _parse_firewall_enabled($firewall_rule_element, $current_policy_name, $rule_id);
						if (!defined($enabled))
						{
							next;
						}

						# Log defaults to true if not defined.
						my $log =  $firewall_rule_element->{att}->{Log};
						if (!defined($log))
						{
							$log = 'true';
						}
						$filter_entry->{'log'} = $log;

						my $primary_action = _parse_firewall_primary_action($firewall_rule_element, $current_policy_name, $rule_id);
						if (!defined($primary_action))
						{
							next;
						}
						$filter_entry->{primaryAction} = $primary_action;

						# Parse protocol
						my $protocol = _parse_firewall_protocol($firewall_rule_element, $current_policy_name, $rule_id);
						if (!defined($protocol))
						{
							next;
						}
						if ($protocol ne 'all')
						{
							$filter_entry->{protocol} = $protocol;
						}

						my $applies_to_interfaces = _parse_firewall_applies_to($firewall_rule_element, $current_policy_name, $rule_id, $in);
						if (!defined($applies_to_interfaces))
						{
							next;
						}

						# Determine the filterList name
						my $filter_list_name = join('_', $current_policy_name, @{$applies_to_interfaces});
						$filter_list_name =~ s/ /-/g;

						# If the filterList doesn't exist, create it.
						if (!defined($filter_lists->{$filter_list_name}))
						{
							$filter_lists->{$filter_list_name} = {
									defaultAction => 'none',
									filterEntry => [],
									appliesTo => [],
									name => $filter_list_name,
									mode => 'stateful'
								};
							foreach my $applies_to_interface (@{$applies_to_interfaces})
							{
								push(@{$filter_lists->{$filter_list_name}->{appliesTo}}, {source => $applies_to_interface});
							}
							$process_orders->{$filter_list_name} = 1;
						}

						# addressPort is of type firewall:SourceDestinationType, schema version 4 and before.
						foreach my $address_port_element ($firewall_rule_element->children('addressPort'))
						{
							# TODO: Are we going to support this?
						}
						# TODO: Investigate DeviceList element, do we want to support it?

						# AddressAndPort is of type firewall:SourceTargetType, schema version 5 and later.
						if (!_parse_address_and_port($firewall_rule_element, $filter_entry, 'AddressAndPort', $current_policy_name, $rule_id))
						{
							next;
						}

						$filter_entry->{processOrder} = $process_orders->{$filter_list_name}++;

						push(@{$filter_lists->{$filter_list_name}->{filterEntry}}, $filter_entry);
					}

					return 1;
				}
		});
	foreach my $netProtectModule_firewall_policy (@{$in->{NetProtectModule_firewall}})
	{
		$twig->parse($netProtectModule_firewall_policy);
	}

	_print_filter_lists($filter_lists, $out);
}

# Print filterLists
sub _print_filter_lists
{
	my ($filter_lists, $out) = @_;

	foreach my $key (keys(%{$filter_lists}))
	{
		if (scalar(@{$filter_lists->{$key}->{filterEntry}}))
		{
			$out->open_element('filterList');
			foreach my $element ('defaultAction', 'appliesTo', 'filterEntry', 'mode', 'name')
			{
				$out->print_element($element, $filter_lists->{$key}->{$element});
			}
			$out->close_element('filterList');
		}
		else
		{
			$LOGGER->debug("No filterEntries found for $key, not printing filterList.");
		}
	}
}

# Parse addresses and ports from either NetProtectModule firewall or NetProtectModule ConnectionEvents policies.
sub _parse_address_and_port
{
	my ($firewall_rule_element, $filter_entry, $element_name, $current_policy_name, $rule_id) = @_;

	# AddressAndPort is of type firewall:SourceTargetType, schema version 5 and later.
	foreach my $address_and_port_element ($firewall_rule_element->children($element_name))
	{
		my $ip_version = _parse_address_and_port_ip_version($address_and_port_element, $current_policy_name, $rule_id);
		if (!defined($ip_version))
		{
			next;
		}

		my $source_addresses = _parse_address_and_port_address($address_and_port_element, 'src', $current_policy_name, $rule_id);
		if (defined($source_addresses))
		{
			push(@{$filter_entry->{sourceIpAddr}}, @{$source_addresses});
		}

		my $destination_addresses = _parse_address_and_port_address($address_and_port_element, 'dest', $current_policy_name, $rule_id);
		if (defined($destination_addresses))
		{
			push(@{$filter_entry->{destinationIpAddr}}, @{$destination_addresses});
		}

		if (defined($filter_entry->{sourceService}))
		{
			my $source_ports = _parse_address_and_port_ports($address_and_port_element, 'src', $current_policy_name, $rule_id);
			if (defined($source_ports))
			{
				if ($source_ports->[0] eq 'all')
				{
					delete($filter_entry->{sourceService});
				}
				else
				{
					push(@{$filter_entry->{sourceService}}, @{$source_ports});
				}
			}
		}

		if (defined($filter_entry->{destinationService}))
		{
			my $destination_ports = _parse_address_and_port_ports($address_and_port_element, 'dest', $current_policy_name, $rule_id);
			if (defined($destination_ports))
			{
				if ($destination_ports->[0] eq 'all')
				{
					delete($filter_entry->{destinationService});
				}
				else
				{
					push(@{$filter_entry->{destinationService}}, @{$destination_ports});
				}
			}
		}
	}

	# Skip rule unless each component is defined in the rule.
	unless (scalar(@{$filter_entry->{sourceIpAddr}}))
	{
		$LOGGER->debug("No valid source address component for rule $rule_id in policy $current_policy_name, skipping.");
		return 0;
	}
	unless (scalar(@{$filter_entry->{destinationIpAddr}}))
	{
		$LOGGER->debug("No valid destination address component for rule $rule_id in policy $current_policy_name, skipping.");
		return 0;
	}
	unless (!defined($filter_entry->{sourceService}) || scalar(@{$filter_entry->{sourceService}}))
	{
		$LOGGER->debug("No valid source port component for rule $rule_id in policy $current_policy_name, skipping.");
		return 0;
	}
	unless (!defined($filter_entry->{destinationService}) || scalar(@{$filter_entry->{destinationService}}))
	{
		$LOGGER->debug("No valid destination port component for rule $rule_id in policy $current_policy_name, skipping.");
		return 0;
	}

	return 1;
}

# Parse SourcePort or TargetPort attributes on the AddressAndPort element.  $src_dest should be 'src' or
# 'dest', depending on what attribute you want to parse.  Returns undef if we were unable to parse the ports.
sub _parse_address_and_port_ports
{
	my ($address_and_port_element, $src_dest, $current_policy_name, $rule_id) = @_;

	my $ports = [];

	my $port_text = $address_and_port_element->{att}->{$src_dest eq 'src' ? 'SourcePort' : 'TargetPort'};

	if ($port_text eq 'ANY')
	{
		$ports = ['all'];
	}
	else
	{
		# Determine if the port set is inverted
		# Currently we only support invert if only a single port is inverted.
		my $invert = 0;
		if ($port_text =~ s/^!//)
		{
			$invert = 1;
		}

		my @port_expressions = split(/,/, $port_text);

		if ($invert && (scalar(@port_expressions) != 1))
		{
			$LOGGER->debug("Except flag not supported with more than a single port on rule $rule_id in policy $current_policy_name, skipping.");
			return undef;
		}

		foreach my $port_expression (@port_expressions)
		{
			if ($port_expression =~ /^(\d+)-(\d+)$/)
			{
				my $port_start = $1;
				my $port_end = $2;

				push(@{$ports}, {portRange => {portEnd => $port_end, portStart => $port_start}});

				if ($invert)
				{
					$LOGGER->debug("Except flag not supported with port ranges on rule $rule_id in policy $current_policy_name, skipping.");
					return undef;
				}
			}
			elsif ($port_expression =~ /^(\d+)$/)
			{
				push(@{$ports}, {portExpression => {port => $port_expression, operator => ($invert ? 'ne' : 'eq')}});
			}
			else
			{
				$LOGGER->debug("Port not supported \"$port_expression\" in rule $rule_id in policy $current_policy_name, skipping.");
				next;
			}
		}

		if (!scalar(@{$ports}))
		{
			$LOGGER->debug("No valid ports parsed from rule $rule_id in policy $current_policy_name, skipping.");
			return undef;
		}
	}

	return $ports;
}

# Parse SourceIP or TargetIP attribute on the AddressAndPort element.  $src_dest should be 'src' or
# 'dest', depending on what attribute you want to parse.  Returns undef if we were unable to parse the IP.
sub _parse_address_and_port_address
{
	my ($address_and_port_element, $src_dest, $current_policy_name, $rule_id) = @_;

	my $ipre = get_crep('cipm');

	my $addresses = [];

	# Some example addresses
	# SourceIP="192.168.50.60-192.168.50.67"
	# SourceIP="ANY"
	# TargetIP="!192.168.60.60"

	my $address_text = $address_and_port_element->{att}->{$src_dest eq 'src' ? 'SourceIP' : 'TargetIP'};

	if ($address_text eq 'ANY')
	{
		$addresses = [{network => {address => '0.0.0.0', mask => '0'}}];
	}
	else
	{
		# Determine if the address set is inverted
		my $invert = 0;
		if ($address_text =~ s/^!//)
		{
			$invert = 1;
		}

		my @expressions = split(/,/, $address_text);
		foreach my $expression (@expressions)
		{
			# Range
			if ($expression =~ /^($ipre)-($ipre)$/)
			{
				my $range_start = $1;
				my $range_end = $2;

				push(@{$addresses}, {range => {endAddress => $range_end, startAddress => $range_start}});
			}
			# Subnet
			elsif ($expression =~ /^($ipre)\/(\d+)$/)
			{
				my $address = $1;
				my $mask = $2;

				push(@{$addresses}, {network => {address => $address, mask => $mask}});
			}
			# Host
			elsif ($expression =~ /^$ipre$/)
			{
				push(@{$addresses}, {host => $expression});
			}
			# Unspported
			else
			{
				$LOGGER->debug("Address not supported \"$expression\" in rule $rule_id in policy $current_policy_name, skipping.");
			}
		}

		if (!scalar(@{$addresses}))
		{
			$LOGGER->debug("No valid addresses parsed from rule $rule_id in policy $current_policy_name, skipping.");
			return undef;
		}

		if ($invert)
		{
			$LOGGER->debug("Except flag not supported for addresses on rule $rule_id in policy $current_policy_name, skipping.");
			return undef;
		}
	}

	return $addresses;
}

# Parse the IPVersion attribute from the firewall policy.  Returns either IPv4 or undef, since no other IP protocols are supported.
sub _parse_address_and_port_ip_version
{
	my ($address_and_port_element, $current_policy_name, $rule_id) = @_;

	my $ip_version = $address_and_port_element->{att}->{IPVersion};
	if (defined($ip_version) && $ip_version ne 'IPv4')
	{
		$LOGGER->debug("Ip version \"$ip_version\" not supported in rule $rule_id in policy $current_policy_name, skipping.");
		return undef;
	}

	return 'IPv4';
}

# Parse the enable attribute from the firewall policy.  Returns true if the rule is enabled or undef
# if the rule is disabled..
sub _parse_firewall_enabled
{
	my ($firewall_rule_element, $current_policy_name, $rule_id) = @_;

	my $enabled = $firewall_rule_element->{att}->{Enabled};

	# Enabled defaults to true if it's not defined.
	if (defined($enabled) && $enabled eq 'false')
	{
		$LOGGER->debug("Disabled rule $rule_id in policy $current_policy_name, skipping.");
		return undef;
	}

	return 'true';
}

# Parse the primary action from the firewall policies FirewallAction attribute.  Returns the
# primaryAction value for the SED, or undef when it fails to parse the primaryAction.
sub _parse_firewall_primary_action
{
	my ($firewall_rule_element, $current_policy_name, $rule_id) = @_;

	# FirewallAction defaults to drop.
	# Possible values are drop, drop-and-reset, protect, monitor or ignore.
	#
	# Here are the explinations on what each of the actions mean, taken from the
	# Proventia Networks IPS G/GX User Guide:
	# Ignore - Allows the matching packet to pass through, so that no further
	#          actions or responses are taken on the packet.
	# # Protect - Packets that match this rule are processed by PAM. Enables
	#             matching packets to be processed by normal responses, such as
	#             (but not limited to) logging, the block response, and
	#             quarantine response.
	# Monitor - Functions as an IP whitelist. Allows to packets that match the
	#           statements bypass the quarantine response and bypass the block
	#           response. However, all other responses still apply to the
	#           packet.
	# Drop - Drops the packets as they pass through the firewall. Because the
	#        firewall is inline, this action prevents the packets from reaching
	#        the target system. To the person whose packet is dropped, it
	#        appears as if the target system does not respond. The connection
	#        most likely makes several retry attempts, and then the connection
	#        eventually times out.
	# Drop and Reset - Functions in the same manner as the drop action, but
	#                  sends a TCP reset to the source system. The connection
	#                  terminates more quickly (because it is automatically
	#                  reset) than with the drop action.
	my $primary_action = $firewall_rule_element->{att}->{FirewallAction};
	if (!defined($primary_action) || $primary_action eq 'drop')
	{
		$primary_action = 'drop';
	}
	elsif ($primary_action eq 'drop-and-reset')
	{
		$primary_action = 'deny';
	}
	elsif ($primary_action eq 'ignore')
	{
		$primary_action = 'permit';
	}
	elsif ($primary_action eq 'protect')
	{
		$primary_action = 'none';
	}
	elsif ($primary_action eq 'monitor')
	{
		$primary_action = 'permit';
	}
	else
	{
		$LOGGER->debug("Unsupported FirewallAction detected \"$primary_action\" on rule $rule_id in policy $current_policy_name, skipping.");
		return undef;
	}

	return $primary_action;
}

# Parse the protocol from the firewall policies ProtocolSimple and ProtocolNumber attributes.
# Returns the protocl for the SED, or undef if it fails to parse the protocol.
sub _parse_firewall_protocol
{
	my ($firewall_rule_element, $current_policy_name, $rule_id) = @_;

	my $protocol; # Stores the end result
	my $protocol_simple = $firewall_rule_element->{att}->{ProtocolSimple};
	# According to the schema, the possible values for ProtocolSimple are Any, TCP, UDP, ICMP, IP and Number
	if (!defined($protocol_simple))
	{
		# Schema defaults to TCP
		$protocol = 'tcp';
	}
	elsif ($protocol_simple eq 'Any')
	{
		# The caller will not set the protocol element if this subroutine returns 'all'.
		$protocol = 'all';
	}
	elsif ($protocol_simple eq 'Number')
	{
		# If ProtocolSimple is set to Number, then we must parse the protocol from ProtocolNumber
		# ProtocolNumber will be an int
		my $protocol_number = _get_FirewallRule_attribute($firewall_rule_element, 'ProtocolNumber', $current_policy_name, $rule_id);
		if (!defined($protocol_number))
		{
			return undef;
		}
		$protocol = get_protocol_name($protocol_number);
	}
	elsif ($protocol_simple eq 'ICMPv6')
	{
		# _parse_firewall_protocol is also used for parsing the connection event policies.  ICMPv6 is a valid protocol in the
		# conenction event policies, and must be parsed as icmp6.
		$protocol = 'icmp6';
	}
	else
	{
		$protocol = lc($protocol_simple);
	}

	return $protocol;
}

# Parse the firewalls Adapter attribute, and return a reference to an array containing the list of
# interfaces the firewall rule applies to or undef if we failed to parse the firewall rules.
sub _parse_firewall_applies_to
{
	my ($firewall_rule_element, $current_policy_name, $rule_id, $in) = @_;

	my $adapter = _get_FirewallRule_attribute($firewall_rule_element, 'Adapter', $current_policy_name, $rule_id);
	if (!defined($adapter))
	{
		return undef;
	}

	return _map_letters_to_interfaces($in, $adapter);
}

# The Adapter element contains only a letter, A - P.  We need to maps the letters to actual interface names.
# A lot of the interfaces will not even be defined.
sub _map_letters_to_interfaces
{
	my ($in, $letters) = @_;

	my $interfaces = [];

	foreach my $interface_letter (split(/,/, $letters))
	{
		# Only add the interface to the applies to if it's not a non-blocking port
		my $interface_name = $in->{_interface_letter_name_mapping}->{$interface_letter};
		if (defined($interface_name) && !defined($in->{_non_blocking_ports}->{$interface_name}))
		{
			push(@{$interfaces}, $in->{_interface_letter_name_mapping}->{$interface_letter});
		}
	}

	return $interfaces;
}

# Checks for the existance of $attribute_name attribute in $element.  If it exists then the value of
# the attribute is returned, otherwise an error message is printed to $LOGGER->debug() and undef is
# returned.
sub _get_FirewallRule_attribute
{
	my ($element, $attribute_name, $current_policy_name, $rule_id) = @_;

	if (defined($element->{att}->{$attribute_name}))
	{
		return $element->{att}->{$attribute_name};
	}

	if (defined($rule_id))
	{
		$LOGGER->debug("$attribute_name attribute not defined on rule $rule_id in policy $current_policy_name, skipping.");
	}
	else
	{
		$LOGGER->debug("$attribute_name attribute not defined for rule in policy $current_policy_name, skipping.");
	}

	return undef;
}

=pod

B<parse_interfaces($in, $out)>

Requires $in->{NetworkDefender_network} and $in->{NetworkDefender_adapter} to be set.

$out is optional.  If $out is not defined the XML will be generated and stored in
$in->{_interface_xml}.  If $out is defined and $in->{_interface_xml} is populated then
$in->{_interface_xml} will be printed to $out and the subroutine will return imedeatly.

=cut
sub parse_interfaces
{
	my ($in, $out) = @_;

	# If $out is not defined, then we don't want to print the XML.  However, since the configuration will be parsed,
	# generate the XML and save it for future use.
	# This function may be called to generate the $in->{_interface_ips} hash before actually being called to print
	# something to the document.
	my $outFh = undef;
	if (!defined($out))
	{
		# The XML has already been generated, but we have no place to print it to, just return.
		if (defined($in->{_interface_xml}))
		{
			return;
		}

		open($outFh, '>', \$in->{_interface_xml});
		$out = ZipTie::Model::XmlPrint->new(\*$outFh, 'core');
	}
	elsif (defined($in->{_interface_xml}))
	{
		$out->print_string($in->{_interface_xml});
		return;
	}

	if (!scalar(@{$in->{NetworkDefender_adapter}}))
	{
		$LOGGER->fatal('No NetworkDefender_adapter policies found.');
	}

	my $policies = _check_single_policy_existance($in, 'NetworkDefender_network');

	$out->open_element('interfaces');

	# Used to assist in the interface letter name mapping
	my @card_ports;

	# Used to keep track of ports that does not block any traffic.  If a port doesn't block any traffic
	# then no filterLists will be applied to it.
	$in->{_non_blocking_ports} = {};

	# Parse the filtering interfaces.
	my $twig = XML::Twig->new(twig_roots => {
				CardManagementParameters => sub {
					my ($twig, $element) = @_;

					if (!defined($element->{att}->{CardID}))
					{
						$LOGGER->debug('CardID not defined, skipping.');
						return 1;
					}
					my $card_id = $element->{att}->{CardID};

					# $element->{att}->{AdapterMode} defaults to 'InlineProtection' in the schema, so not defined is also valid here.
					my $non_blocking_port = 0;
					if (defined($element->{att}->{AdapterMode}) && $element->{att}->{AdapterMode} ne 'InlineProtection')
					{
						$non_blocking_port = 1;
					}

					foreach my $port ('Port1', 'Port2')
					{
						if (!defined($element->{att}->{$port}))
						{
							$LOGGER->debug("$port not defined for card id $card_id, skipping.");
							return 1;
						}

						my $interface_name = $element->{att}->{$port};

						# Record this as a non-blocking port for future use
						if ($non_blocking_port)
						{
							$in->{_non_blocking_ports}->{$interface_name} = 1;
						}

						# We store the interface names for later processing, so it still works if we read the configurations out of order.
						$card_ports[$card_id]->{$port} = $interface_name;

						$out->open_element('interface');
						$out->print_element('adminStatus', 'up');

						# The following section dosen't work in topology for pre-
						#$out->open_element('interfaceGrouping');
						#$out->print_element('ifGroupName', "card-id-$card_id");
						#$out->print_element('ifGroupingCharacteristic', 'proxyAffiliation');
						#$out->close_element('interfaceGrouping');
						#$out->print_element('interfaceBehavior', 'bridging');

						$out->print_element('interfaceType', 'other');
						$out->print_element('name', $interface_name);
						$out->print_element('physical', 'true');
						$out->close_element('interface');
					}

					return 1;
			}
	});
	foreach my $network_defender_adapter_policy (@{$in->{NetworkDefender_adapter}})
	{
		$twig->parse($network_defender_adapter_policy);
	}

	# When applying firewall policies to interfaces the user selects the interface based on a letter,
	# A - P.  A maps to the first security interface, B to the second security interface, etc.  This
	# maps the letters to the named security interfaces.
	my @interface_letters = qw(A B C D E F G H I J K L M N O P Q R S T U V W X Y Z);
	my $interface_letter_index = 0;
	$in->{_interface_letter_name_mapping} = {};
	foreach my $card (@card_ports)
	{
		if (defined($card))
		{
			foreach my $port ('Port1', 'Port2')
			{
				# Create the interface letter name mapping.
				$in->{_interface_letter_name_mapping}->{$interface_letters[$interface_letter_index++]} = $card->{$port};
			}
		}
	}

	# Parse the admin interface.
	my $ip = undef;
	my $mask = undef;
	$twig = XML::Twig->new(twig_roots => {
				MgmtInterface => sub {
					my ($twig, $element) = @_;

					if (defined($ip))
					{
						$LOGGER->debug('More than one admin interface found, this was unexpected, skipping.');
						return 1;
					}

					# We currently only support IPv4
					my $ipv4_element = $element->first_child('IPv4');
					if (!defined($ipv4_element))
					{
						$LOGGER->debug('IPv4 element not found for management interface, skipping.');
						return 1;
					}

					my $static_element = $ipv4_element->first_child('Static');
					if (!defined($static_element))
					{
						$LOGGER->debug('IPv4/Static element not found for management interface, skipping.');
						return 1;
					}

					if (!defined($static_element->{att}->{IpAddr}))
					{
						$LOGGER->debug('IPv4/Static element doesn\'t contain IpAddr attribute for management interface, skipping.');
						return 1;
					}
					if (!defined($static_element->{att}->{SubnetMask}))
					{
						$LOGGER->debug('IPv4/Static element doesn\'t contain SubnetMask attribute for management interface, skipping.');
						return 1;
					}

					$ip = $static_element->{att}->{IpAddr};
					$mask = mask_to_bits($static_element->{att}->{SubnetMask});

					# This information is used in parse_static_routes.
					if (defined($static_element->{att}->{GatewayIpAddr}))
					{
						$in->{_static_route} = {
								gatewayAddress => $static_element->{att}->{GatewayIpAddr},
								destinationAddress => '0.0.0.0',
								destinationMask => '0',
								interface => 'administrative',	
								defaultGateway => 'true'
							};
					}

					return 1;
			}
	});
	$twig->parse($policies->{NetworkDefender_network});

	if (defined($ip))
	{
		$out->open_element('interface');
		$out->print_element('adminStatus', 'up');
		$out->open_element('interfaceIp');
		$out->open_element('ipConfiguration');
		$out->print_element('ipAddress', $ip);
		$out->print_element('mask', $mask);
		$out->close_element('ipConfiguration');
		$out->close_element('interfaceIp');
		# Commented out because of defect in core.  See defect 63835 and story 64939.
		#$out->print_element('interfaceBehavior', 'administrative');
		$out->print_element('interfaceType', 'other');
		$out->print_element('name', 'administrative');
		$out->print_element('physical', 'true');
		$out->close_element('interface');
	}

	$out->close_element('interfaces');

	if (defined($outFh))
	{  
		close($outFh);
	}
}

=pod

B<parse_static_routes($in, $out)>

If $in->{_static_route} has been set by parse_interfaces then it prints the static route to the SED.
Doesn't print anything if $in->{_static_route} isn't set.

=cut
sub parse_static_routes
{
	my ($in, $out) = @_;

	if (defined($in->{_static_route}))
	{
		$out->open_element('staticRoutes');
		$out->print_element('staticRoute', $in->{_static_route});
		$out->close_element('staticRoutes');
	}
}

=pod

B<parse_snmp($in, $out)>

Prints SNMP information to the SED if any exists in the NetworkDefender_snmp policy.

NetworkDefender_snmp is parsed in parse_system.  This method expects parse_system to be called first.
parse_system will set $in->{_snmp}, expected by this subroutine.

If the system name is not set in the NetworkDefender_snmp policy then no SNMP information will be printed
in the SED.

=cut
sub parse_snmp
{
	my ($in, $out) = @_;

	my $system_contact = undef;
	my $system_location = undef;
	my $system_name = undef;

	if (defined($in->{_snmp}))
	{

		if (defined($in->{_snmp}->{SystemContact}) && $in->{_snmp}->{SystemContact} ne 'Unknown')
		{
			$system_contact =  $in->{_snmp}->{SystemContact};
		}

		if (defined($in->{_snmp}->{SystemLocation}) && $in->{_snmp}->{SystemLocation} ne 'Unknown')
		{
			$system_location = $in->{_snmp}->{SystemLocation};
		}

		if (defined($in->{_snmp}->{SystemName}) && $in->{_snmp}->{SystemName} ne 'Unknown')
		{
			$system_name = $in->{_snmp}->{SystemName};
		}

	}

	# System name is a required element
	if (defined($system_name))
	{
		$out->open_element('snmp');

		if (defined($system_contact))
		{
			$out->print_element('sysContact', $system_contact);
		}

		if (defined($system_location))
		{
			$out->print_element('sysLocation', $system_location);
		}

		if (defined($system_name))
		{
			$out->print_element('sysName', $system_name);
		}

		$out->close_element('snmp');
	}
}

# Private subroutine used to check for the existence of policies, and verifiesthat only one instance
# of that policy exists.  This subroutine should not be used if it's expected that more than one
# policy will exist for a type of policy.
#
# This subroutine calls $LOGGER->fatal() if no policies exist for a given policy type.  If more
# than one policy exists for a policy type this subroutine will print a debug log and return the
# first policy.  If this occurs it indicates a problem with the adapter that should be fixed.
#
# The first argument is a reference to $in.  All other arguments will be interpreted as policy
# names.
#
# Returns a hash of the policies.
sub _check_single_policy_existance
{
	my $in = shift;

	my $return_policies = {};

	foreach my $policy_name (@_)
	{
		if (!defined($in->{$policy_name}) || scalar(@{$in->{$policy_name}}) == 0)
		{
			$LOGGER->fatal("No $policy_name policy found.");
		}

		if (scalar(@{$in->{$policy_name}}) != 1)
		{
			$LOGGER->debug("More than one $policy_name policy found, parsing first policy.");
		}

		$return_policies->{$policy_name} = $in->{$policy_name}->[0];
	}

	return $return_policies;
}

1;

__END__

=pod

=head1 AUTHOR(S)

David Payne
david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2013, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
