package ZipTie::Adapters::IBM::ISS_SiteProtector::AgentDocumentCollector;

=pod

=head1 NAME

ZipTie::Adapters::IBM::ISS_SiteProtector::AgentDocumentCollector -- Obtain agent documents from a
SiteProtector server.

=head1 SYNOPSIS

	my $agentDocumentCollector =
		ZipTie::Adapters::IBM::ISS_SiteProtector::AgentDocumentCollector->new($dbh, $ip,
		$management_server_ip);
	my $namespaceDocuments = $agentDocumentCollector->get_agent_documents($namespace);

=head1 DESCRIPTION

This class collects agent documents from a ISS SiteProtector server based on document namespaces.

=head1 METHODS

=cut

use strict;
use warnings;

use ZipTie::DBI;
use ZipTie::Logger;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

=pod

B<new($$dbh, $ip, $management_server_ip)> (constructor)

Create a new AgentDocumentCollector for the supplied IP address and obtain the list of agent
documents from the SiteProtector server.

Returns: ZipTie::Adapters::IBM::ISS_SiteProtector::AgentDocumentCollector object.

=cut
sub new
{
	my ($class, $dbh, $ip, $management_server_ip) = @_;

	# We will later be setting TEXTSIZE.  This is to make sure it applies to all database transactions.
	$dbh->set_syb_no_child_con();

	my ($guid, $model) = _get_guid_model($dbh, $ip, $management_server_ip);

	my $self = {
			_dbh => $dbh,
			_ip => $ip,
			_management_server_ip => $management_server_ip,
			_guid => $guid,
			_model => $model,
			_global_issues_ids => {}
		};

	$self->{_agent_document_list} = _get_agent_document_list($self, $guid);

	bless($self, $class);

	return $self;
}

=pod

B<get_agent_documents($namespace)>

Returns: An array reference containing all the agent documents for the provided namespace.

=cut
sub get_agent_documents
{
	my ($self, $namespace) = @_;

	my $agent_documents = [];

	my $binary_data_ids = undef;

	if (!defined($self->{_agent_document_list}->{$namespace}))
	{
		if ($namespace =~ /issue.*csv/)
		{
			my $issues_binary_data_id = _get_global_issues($self, $namespace);

			if (defined($issues_binary_data_id))
			{
				$binary_data_ids = [{binary_data_id => $issues_binary_data_id}];
			}
		}
	}
	else
	{
		$binary_data_ids = $self->{_agent_document_list}->{$namespace};
	}

	if (!defined($binary_data_ids))
	{
		$LOGGER->debug("No agent documents found for namespace: $namespace.");
		return [];
	}

	my $query = $self->{_dbh}->prepare('SELECT CAST(CAST(Value AS varbinary(max)) AS varchar(max)) FROM BinaryData ' .
		'WHERE BinaryDataID = ?') or $LOGGER->fatal('Failed to prepare BinaryData SQL: ' . $self->{_dbh}->errstr());

	my $document_number = 1;
	foreach my $document (@{$binary_data_ids})
	{
		$query->set_next_interaction_label($namespace . ' #' . $document_number++);
		if (!$query->execute($document->{binary_data_id}))
		{
			$LOGGER->debug('Failed to execute BinaryData SQL: ' . $self->{_dbh}->errstr());
			next;
		}

		my $rows = $query->fetchall_arrayref();
		$query->finish();

		if (scalar(@{$rows}) != 1)
		{
			$LOGGER->debug('Expected 1 agent document, received ' . scalar(@{$rows}));
			next;
		}

		push(@{$agent_documents}, $rows->[0]->[0]);
	}

	return $agent_documents;
}

=pod

B<get_model()>

Returns: The ModelDesc defined in the Component table for the device.

=cut
sub get_model
{
	my ($self) = @_;

	return $self->{_model};
}

=pod

B<get_guid()>

Returns: The SiteProtector GUID for the device.

=cut
sub get_guid
{
	my ($self) = @_;

	return $self->{_guid};
}

# Used by this class to obtain a list of agent documents for the provided guid.
sub _get_agent_document_list
{
	my ($self, $guid) = @_;

	my $documents = {};

	# After getting a list of documents we will need to use the binary_data_id's to set the textsize.
	my $binary_data_ids = [];

	my $query = $self->{_dbh}->prepare('EXEC iss_GetAgentDocuments ?') or $LOGGER->fatal('Failed to prepare stored procedure iss_GetAgentDocuments.');
	$query->set_next_interaction_label('EXEC iss_GetAgentDocuments');
	$query->execute($guid) or $LOGGER->fatal('Failed to execute stored procedure iss_GetAgentDocuments.');
	my $rows = $query->fetchall_arrayref();
	$query->finish();
	foreach my $row (@{$rows})
	{
		my ($namespace, $version, $binary_data_id, $checksum, $is_local_document) = @{$row};

		if (!defined($documents->{$namespace}))
		{
			$documents->{$namespace} = [];
		}

		push(@{$documents->{$namespace}}, {namespace => $namespace, version => $version,
			binary_data_id => $binary_data_id, checksum => $checksum, is_local_document => $is_local_document});

		push(@{$binary_data_ids}, $binary_data_id);
	}

	_set_textsize($self, $binary_data_ids);

	return $documents;
}

# Set the TEXTSIZE.  Without setting TEXTSIZE the agent documents returned will be truncated.
sub _set_textsize
{   
	my ($self, $binary_data_ids) = @_;

	# Add the global issues.csv and issue_responses2.csv binary data ids to the list of binary data ids.
	my @all_binary_data_ids;
	push(@all_binary_data_ids, @{$binary_data_ids});
	foreach my $issues ('issues.csv', 'issue_responses2.csv', 'paranoid_issue_responses2.csv', 'aggressive_issue_responses2.csv', 'moderate_issue_responses2.csv')
	{
		my $issues_binary_data_id = _get_global_issues($self, $issues);
		if (defined($issues_binary_data_id))
		{
			push(@all_binary_data_ids, $issues_binary_data_id);
		}
	}

	my $sql = sprintf('SELECT MAX(DATALENGTH(Value)) FROM BinaryData WHERE BinaryDataID IN (%s)', join(',', @all_binary_data_ids));
	my $query = $self->{_dbh}->prepare($sql) or $LOGGER->fatal('Failed to prepare SQL to determine maximum agent document size.');
	$query->set_next_interaction_label('Find maximum agent document size.');
	$query->execute() or $LOGGER->fatal('Failed to execute SQL to determine maximum agent document size.');

	my $rows = $query->fetchall_arrayref();
	$query->finish();
	my ($max_length) = @{$rows->[0]};

	$max_length += 2;

	$self->{_dbh}->do("SET TEXTSIZE $max_length") or $LOGGER->fatal('Failed to set TEXTSIZE');
}

# Determines the guid and model of the provided IP address.  Fails with an error if the IP address isn't managed by the SiteProtector server.
sub _get_guid_model
{
	my ($dbh, $ip, $management_server_ip) = @_;

	my $sql = 'SELECT c.GUID, c.ModelDesc FROM Component c INNER JOIN SensorHost sh ON c.ComponentID = sh.SensorID WHERE c.DELETED = 0 AND sh.SensorIPDisplay = ?';
	my $query = $dbh->prepare($sql) or $LOGGER->fatal("Failed to prepare SQL to determine GUID for IP $ip: " . $dbh->errstr());
	$query->set_next_interaction_label('Get GUID and ModelDesc');
	$query->execute($ip) or $LOGGER->fatal("Failed to execute SQL to determine GUID for IP $ip: " . $dbh->errstr());

	my $rows = $query->fetchall_arrayref();
	$query->finish();

	# If we did not get a GUID then this device isn't managed by this SiteProtector.
	if (scalar(@{$rows}) == 0)
	{
		$LOGGER->fatal("Device $ip not managed by SiteProtector $management_server_ip.");
	}

	# This should never happen, but add a check anyway
	if (scalar(@{$rows}) != 1)
	{
		$LOGGER->fatal("Expected 1 GUID for IP $ip but received " . scalar(@{$rows}) . '.');
	}

	# First column is the GUID, second column is the ModelDesc
	return ($rows->[0]->[0], $rows->[0]->[1]);
}

# Obtains the binary data id for issues.csv, issue_responses2.csv, or other related csv files.
sub _get_global_issues
{
	my ($self, $issues) = @_;

	if (defined($self->{_global_issues_ids}->{$issues}))
	{
		return $self->{_global_issues_ids}->{$issues};
	}

	my $query = $self->{_dbh}->prepare('SELECT prv.BinaryDataID ' .
							'FROM PolicyRepository pr ' .
							'INNER JOIN Namespace n ON pr.NamespaceID = n.NamespaceID ' .
							'INNER JOIN PolicyRepositoryVersion prv ON pr.PolicyRepositoryID = prv.PolicyRepositoryID ' .
							'INNER JOIN BinaryData b ON prv.BinaryDataID = b.BinaryDataID ' .
							'WHERE pr.SchemaVersion = \'\' AND n.Namespace = ? ' .
							'AND prv.Version = (SELECT MAX(Version) ' .
								'FROM PolicyRepositoryVersion ' .
								'WHERE PolicyRepositoryVersion.PolicyRepositoryID = prv.PolicyRepositoryID ' .
								'GROUP BY PolicyRepositoryID)');

	$query->set_next_interaction_label("Get global $issues BinaryDataID.");

	if (!$query->execute($issues))
	{
		$LOGGER->debug("Failed to execute SQL to obtain $issues.");
		return undef;
	}

	my $rows = $query->fetchall_arrayref();
	$query->finish();

	if (scalar(@{$rows}) != 1)
	{
		$LOGGER->debug("Expected 1 BinaryDataID for $issues, received " . scalar(@{$rows}));
		return undef;
	}

	$self->{_global_issues_ids}->{$issues} = $rows->[0]->[0];

	return $rows->[0]->[0];
}

1;

__END__

=pod

=head1 AUTHOR(S)

David Payne
david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2013, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
