#!/usr/bin/perl

use strict;
use warnings;

use ZipTie::Logger;
use ZipTie::Client;
use ZipTie::Adapters::Utils qw(get_crep);
use ZipTie::DBI;
use ZipTie::ConnectionPath;
use ZipTie::ConnectionPath::Protocol;
use ZipTie::Credentials;
use ZipTie::Recording;

# A list of role/adapter pairs to discover.
my @ROLES_TO_DISCOVER = (
		{role_id => 24, adapter => 'ZipTie::Adapters::IBM::ISS_GX'}
	);

# Check input
my $ipre = get_crep('cipm');

if (scalar(@ARGV) != 1)
{
	print "ERROR: Wrong number of arguments passed to spdiscovery.pl\n";
	exit(1);
}

my $management_server_ip = $ARGV[0];

if ($management_server_ip !~ /$ipre/)
{
	print "ERROR: Invalid IP address: $management_server_ip\n";
}

# Set up logging
my $LOGGER = ZipTie::Logger::get_logger();
if ($ENV{ADAPTER_LOG_TO_FILE})
{
	$LOGGER->enable_logging_to_file('ISS_SiteProtector', 'spdiscovery', $management_server_ip);
}

$LOGGER->debug("Performing Discover From SiteProtector on $management_server_ip.");

my $ziptie_client = ZipTie::Client->new();

my $existing_devices = get_devices($ziptie_client);

my ($username, $password, $port) = get_credentials($ziptie_client, $management_server_ip);

$LOGGER->debug("Connecting to SiteProtector database with username '$username' and port $port.");

# ZipTie::DBI's create_sybase_connection expects a connection path objcet.  Create a connection path object.
my $connection_path = ZipTie::ConnectionPath->new();
$connection_path->set_management_server_ip_address($management_server_ip);
$connection_path->set_credentials(ZipTie::Credentials->new(username => $username, password => $password));
$connection_path->set_protocols([ZipTie::ConnectionPath::Protocol->new({name => 'SQL', port => $port})]);

my $dbh = ZipTie::DBI->create_sybase_connection($connection_path, 'RealSecureDB');

my $device_service = $ziptie_client->devices();

my $sql = 'SELECT sh.SensorIPDisplay FROM Component c INNER JOIN SensorHost sh ON c.ComponentID = ' .
	'sh.SensorID WHERE c.DELETED = 0 AND c.RoleID = ?';
my $query = $dbh->prepare($sql);
foreach my $role_adapter_pair (@ROLES_TO_DISCOVER)
{
	my $role_id = $role_adapter_pair->{role_id};
	my $adapter = $role_adapter_pair->{adapter};

	$LOGGER->debug("Discovering $adapter devices.");

	$query->set_next_interaction_label('Discover Devices');
	$query->execute($role_id);
	my $rows = $query->fetchall_arrayref();

	foreach my $row (@{$rows})
	{
		my ($device_ip) = @{$row};
		add_or_update_device($existing_devices, $device_service, $device_ip, $management_server_ip, $adapter);
	}
}

$dbh->disconnect();

if ($ENV{ENABLE_RECORDING})
{
	my $recording_connection_path = ZipTie::ConnectionPath->new();
	$recording_connection_path->set_ip_address($management_server_ip);
        ZipTie::Recording::get_recording()->to_file('ISS_SiteProtector', 'spdiscovery', $recording_connection_path);
}

sub get_credentials
{
	my ($ziptie_client, $management_server_ip) = @_;

	my $credentials_service = $ziptie_client->credentials();

	my @credentials = $credentials_service->getCredentialSetsByIpAddress(ipAddress => $management_server_ip);

	if (!scalar(@credentials))
	{
		_fatal("ERROR: No credentials found.");
	}

	my $username = undef;
	my $password = undef;
	if (ref($credentials[0]->{credentials}) ne 'ARRAY')
	{
		$credentials[0]->{credentials} = [$credentials[0]->{credentials}];
	}
	foreach my $credential (@{$credentials[0]->{credentials}})
	{
		if ($credential->{name} eq 'username')
		{
			$username = $credential->{value};
		}
		elsif ($credential->{name} eq 'password')
		{
			$password = $credential->{value};
		}
	}

	if (!defined($username) || $username eq '')
	{
		_fatal("ERROR: Authentication refused with an empty username.");
	}

	if (!defined($password) || $password eq '')
	{
		_fatal("ERROR: Authentication refused with an empty password.");
	}

	my @protocols = $credentials_service->getProtocolConfigByIpAddress(ipAddress => $management_server_ip);

	if (!scalar(@protocols))
	{
		_fatal("ERROR: No protocols found.");
	}

	my $port = undef;
	foreach my $protocol (@{$protocols[0]->{protocols}})
	{
		if ($protocol->{name} eq 'SQL' && $protocol->{enabled} eq 'true')
		{
			$port = $protocol->{port};
			last;
		}
	}

	if (!defined($port) || $port eq '')
	{
		_fatal("ERROR: Required protocol either disabled or undefined.");
	}

	return ($username, $password, $port);
}

sub add_or_update_device
{
	my ($existing_devices, $device_service, $device_ip, $management_server_ip, $adapter) = @_;

	my %device = (ipAddress => $device_ip, adapterId => $adapter,
		managedNetwork => 'Default', managementServerIp => $management_server_ip);

	# Device already exists.
	if (defined($existing_devices->{$device_ip}))
	{
		$device_service->updateDevice(ipAddress => $device_ip, managedNetwork => 'Default', device => \%device);
		$existing_devices->{$device_ip} = \%device;

		$LOGGER->debug("Updating existing device $device_ip.");

		print "$device_ip,Updated\n";
	}
	# Add new device.
	else
	{
		$device_service->createManagedDevice(%device);
		$existing_devices->{$device_ip} = \%device;

		$LOGGER->debug("Adding new device $device_ip.");

		print "$device_ip,Added\n";
	}
}

sub get_devices
{
	my ($ziptie_client) = @_;
	my %deviceHash;
	my $pageSize = 500;
	my $offset = 0;
	my $total = undef;

	my $device_search_service = $ziptie_client->devicesearch();

	while (!defined($total) || ($offset < $total))
	{
		my $page = $device_search_service->search(
				scheme => 'ipAddress', query => '',
				pageData => {pageSize => $pageSize, offset => $offset}
			);
		$offset += $pageSize;

		my $devices = [];
		if (ref($page->{devices}) eq 'ARRAY')
		{
			$devices = $page->{devices};
		}
		elsif (defined($page->{devices}))
		{
			$devices = [$page->{devices}];
		}

		foreach my $device (@{$devices})
		{
			$deviceHash{$device->{ipAddress}} = $device;
		}

		# Total is only returned for the first search
		if (!defined($total))
		{
			$total = $page->{total};
		}
	}
	return \%deviceHash;
}

sub _fatal
{
	my ($message) = @_;

	print "$message\n";
	$LOGGER->fatal($message);
	exit(1);
}

1;

__END__

=head1 NAME

spdiscovery.pl

=head1 DESCRIPTION

ZipTie plugin used to discover supported devices from a IBM ISS SiteProtector.

=head1 AUTHOR(S)

David Payne
david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2013, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
