package ZipTie::Adapters::Juniper::JUNOS;

use strict;
use warnings;
use 5.010_000;

use Data::Dumper;
use ZipTie::Adapters::BaseAdapter;
use ZipTie::Adapters::GenericAdapter;
use ZipTie::Adapters::Juniper::JUNOS::AutoLogin;
use ZipTie::Adapters::Juniper::JUNOS::Disconnect qw(disconnect);
use ZipTie::Adapters::Juniper::JUNOS::GetActiveConfig qw(get_scp_active_config);
use ZipTie::Adapters::Juniper::JUNOS::ParseInterfaces qw(parse_interfaces);
use ZipTie::Adapters::Juniper::JUNOS::Parsers qw(parse_static_routes parse_routing create_config parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_object_groups clear_parsing_state);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::CLIProtocolFactory;
use ZipTie::CLIProtocol;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

our @ISA = qw(ZipTie::Adapters::BaseAdapter);

sub backup{
	my ($package_name, $backup_doc) = @_;
	my $responses   = {};

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ($connection_path) = ZipTie::Typer::translate_document($backup_doc, 'connectionPath');

	# Connect to the device and capture the ZipTie::CLIProtocol that is created as a result of the connection.
	# Also be sure to capture the device prompt that is returned as a result of a successful connection.
	my ($cli_protocol, $device_prompt_regex) = _connect($connection_path);

	my $configure_prompt_regex = $device_prompt_regex;
	$configure_prompt_regex =~ s/\\>\\/\\#\\/;

	# Store the found prompt as "prompt" on the CLI protocol.
	$cli_protocol->set_prompt_by_name('prompt', $device_prompt_regex);

	# Grab an output filehandle for the model
	my $filehandle = get_model_filehandle('JUNOS', $connection_path->get_ip_address());
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, 'common');
	$printer->open_model();

	# Gather inputs for the model
	$cli_protocol->send_and_wait_for('set cli screen-length 0', $device_prompt_regex);
	$responses->{showVer}              = _send_and_check($cli_protocol, 'show version',                                     $device_prompt_regex, 1);
        $responses->{showHardware}         = _send_and_check($cli_protocol, 'show chassis hardware | display xml',              $device_prompt_regex, 1);
	$responses->{showMAC}              = _send_and_check($cli_protocol, 'show chassis mac-addresses',                       $device_prompt_regex, 1);
	$responses->{showConfigInterfaces} = _send_and_check($cli_protocol, 'show configuration interfaces | display xml',      $device_prompt_regex, 1);
	$responses->{interfaces}           = _send_and_check($cli_protocol, 'show interfaces | display xml',                    $device_prompt_regex, 1);
	$responses->{showUptime}           = _send_and_check($cli_protocol, 'show system uptime',                               $device_prompt_regex);
	$responses->{showFirmware}         = _send_and_check($cli_protocol, 'show chassis firmware',                            $device_prompt_regex);
	$responses->{showRen}              = _send_and_check($cli_protocol, 'show chassis routing-engine',                      $device_prompt_regex);
	$responses->{snmp}                 = _send_and_check($cli_protocol, 'show configuration snmp | display xml',            $device_prompt_regex);
	$responses->{showFirewall}         = _send_and_check($cli_protocol, 'show configuration firewall | display xml',        $device_prompt_regex);
	$responses->{showFirewallInet6}    = _send_and_check($cli_protocol, 'show configuration firewall family inet6',         $device_prompt_regex);
	$responses->{showPolicy}           = _send_and_check($cli_protocol, 'show configuration security | display xml',        $device_prompt_regex);
	$responses->{showZones}            = _send_and_check($cli_protocol, 'show configuration security zones | display xml',  $device_prompt_regex);
	$responses->{showApplications}     = _send_and_check($cli_protocol, 'show configuration applications | display xml',    $device_prompt_regex);
	$responses->{showFirewallPlain}    = _send_and_check($cli_protocol, 'show configuration firewall',                      $device_prompt_regex);
	$responses->{interfacesFilters}    = _send_and_check($cli_protocol, 'show interfaces filters | display xml',            $device_prompt_regex);
	$responses->{showOspfInterface}    = _send_and_check($cli_protocol, 'show ospf interface detail | display xml',         $device_prompt_regex);
	$responses->{showBgp}              = _send_and_check($cli_protocol, 'show bgp neighbor | display xml',                  $device_prompt_regex);
	$responses->{showRoute}            = _send_and_check($cli_protocol, 'show configuration routing-options | display xml', $device_prompt_regex);
	$responses->{showMibSystem}        = _send_and_check($cli_protocol, 'show snmp mib walk system',                        $device_prompt_regex);

	# See if routing options are placed indirectly through a VRF
	if($responses->{showRoute}){
		while($responses->{showRoute} =~ /^\s*<next-table>(\S+?)<\/next-table>/mg){
			my ($vrf) = $1 =~ /^(\w+)\.?/;
			if ($vrf){
				# We have a table to search for
				$responses->{showNamedVrfs}->{$vrf} = $cli_protocol->send_and_wait_for("show configuration routing-instances $vrf | display xml", $device_prompt_regex);
			}
		}
	}

	# Retrieve the active config, which can be very large and might be causing ssh timeout issues.
	my $scp_protocol;
	if(defined($connection_path)){
		$scp_protocol = $connection_path->get_protocol_by_name('SCP');
	}
	
	my $activeResponse = undef;
	if(defined($scp_protocol)){
		$activeResponse = get_scp_active_config($connection_path);
	}
	
	if(!defined($scp_protocol) || !defined($activeResponse)){
		# We wanted to realign processes to use similar methods for retrieving data
		my $showConfigResponse = $cli_protocol->send_and_wait_for('show configuration', $device_prompt_regex);
		$showConfigResponse =~ s/^show configuration\s*$//ms;
		$showConfigResponse =~ s/^$device_prompt_regex\s*$//ms;
		$showConfigResponse =~ s/^\s*$//msg;
		$activeResponse = $showConfigResponse;
	}
	if(!defined($activeResponse)){
		die 'Couldn\'t retrieve running config.';
	}		
	$responses->{active} = $activeResponse;

	# Parse configuration and print SED
	parse_system($responses, $printer);
	parse_chassis($responses, $printer);
	create_config($responses, $printer);
	parse_object_groups($responses,$printer);
	parse_filters($responses, $printer);
	parse_interfaces($responses, $printer);
	parse_routing($responses, $printer);
	parse_snmp($responses, $printer);
	parse_static_routes($responses, $printer);

	disconnect($cli_protocol);
	clear_parsing_state($responses);

	$printer->close_model();

	# Close the model output file handle
	close_model_filehandle($filehandle)
}

sub _connect{
	# Grab our arguments
	my $connection_path = shift;

	# Create a new CLI protocol object
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	# Make a connection to and successfully authenticate with the JUNOS device
	my $device_prompt_regex = ZipTie::Adapters::Juniper::JUNOS::AutoLogin::execute($cli_protocol, $connection_path);
	
	# Return the created ZipTie::CLIProtocol object and the device prompt encountered after successfully connecting to a device.
	return($cli_protocol, $device_prompt_regex);
}

sub _send_and_check{
	my ($cli_protocol, $command, $device_prompt_regex, $compulsory, $timeout) = @_;

	my $response;
	if($timeout){
		$response = $cli_protocol->send_and_wait_for($command, $device_prompt_regex, $timeout);
	}else{
		$response = $cli_protocol->send_and_wait_for($command, $device_prompt_regex);
	}

	if($compulsory){
		if($response =~ /syntax\s+error/i || $response =~ /unknown\s+command/i || $response =~ /Interaction\s+not\s+found/i){
			die("Access to the command '$command' is compulsory for a backup to succeed");
		}
	}

	return($response);
}

1;

__END__

=head1 NAME

ZipTie::Adapters::Juniper::JUNOS - Adapter for performing various operations against Juniper JUNOS devices.

=head1 SYNOPSIS

	use ZipTie::Adapters::Juniper::JUNOS;
	ZipTie::Adapters::Juniper::JUNOS::backup( $backup_document );

=head1 DESCRIPTION

This module represents an adapter that can be used to perform various operations against against Juniper JUNOS devices.

Generally you would run this module through the ZipTie C<invoke.pl> script.

=head1 PUBLIC SUB-ROUTINES

=over 12

=item C<backup($backup_document)>

Performs the backup of the device that is described in the specified XML document.  This XML document contains
a "connectionPath" element that contains the IP/hostname, protocol, credential, and file server information that
may be needed to connect to, authenticate with, and back up the device in question.  This XML will be parsed into
a C<ZipTie::ConnectionPath> object to help easily access this vital information.

The act of backing up a device means that information about a device will be retreived, as well as the configuration
data/files for that device.  All of this collected information will be used to populate the ZipTie Element Document,
which is an open and extensible format for describing a network device.

The return value for this method will be a string containing the successfully populated ZipTie Element Document.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_connect($connection_path)>

Creates the initial CLI connection to the device.  A list is returned containing two elements: the first is a
C<ZipTie::CLIProtocol> object that is a CLI client that can communitcate with the device, and the second is a
regular expression that can be used to match the primary prompt of the device.  This is useful when sending
commands to and receiving responses from the device through the C<ZipTie::CLIProtocol> object and being able to
know when a command has generated all the output and returned back to the primary prompt.

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

Contributor(s): rkruse, Dylan White (dylamite@ziptie.org), Jak Devaney (john.devaney@uk.ibm.com)
Date: August 10, 2007

=cut
