package ZipTie::Adapters::Juniper::JUNOS::ParseInterfaces;

use strict;
use warnings;

use Data::Dumper;
use MIME::Base64 qw(encode_base64);
use Net::IP;
use XML::Simple qw(XMLin);
use XML::Twig;
use ZipTie::Logger;

use Exporter qw(import);
our @EXPORT_OK = qw(parse_interfaces);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

sub _getRawInterfaces{
	my ($in) = @_;
	
	my $forceArray = [ 'physical-interface', 'logical-interface' ];
	my ($interfaceData) = $in->{interfaces} =~ /(<rpc-reply.+?<\/rpc-reply>)/sm;

	my $rawInterfaces = XMLin( $interfaceData, ForceArray => $forceArray, KeyAttr => ['name'] );
	$rawInterfaces = $rawInterfaces->{'interface-information'}->{'physical-interface'};
	
	return $rawInterfaces;
}

# Needs to now return a hash of physical names to array of units
sub _getConfInterfaces{
	my ($in) = @_;

	my $forceArray = ['input-list', 'output-list', 'address', 'unit'];
	
	my ($interfaceData) = $in->{showConfigInterfaces} =~ /(<interfaces.+?<\/interfaces>)/sm;
	
	my $confInterfaces = XMLin($interfaceData, ForceArray => $forceArray, KeyAttr => []);
	$confInterfaces = $confInterfaces->{interface};

	# Now this is an array, I'd rather this be a hash so I can just call hey hash give me this element or does this element exist
	my %hash = ();
	foreach my $confInterface(@$confInterfaces){
		$hash{ $confInterface->{name} } = [$confInterface];
	}

	return \%hash;
}

sub _getConfiguredInterfaceArray{
	my ($phyInterfaceName, $rawInterface, $confInterfaces) = @_;
	my $configuredInterfaceArray = undef;
	
	# Good chance this is an array as well?
	if($rawInterface->{'logical-interface'}){
		my $logicalInterfaceHash = $rawInterface->{'logical-interface'};
		
		if(ref $logicalInterfaceHash eq 'HASH'){
			if($confInterfaces->{$phyInterfaceName}){
				$configuredInterfaceArray = $confInterfaces->{$phyInterfaceName};
			}
		}else{
			foreach my $logInterface(@$logicalInterfaceHash){
				if($logInterface->{name} && $logInterface->{name} =~ /$phyInterfaceName/i){
					$configuredInterfaceArray = $confInterfaces->{$phyInterfaceName};
					last;
				}
			}
		}
	}
	if($configuredInterfaceArray && ref $configuredInterfaceArray ne 'ARRAY')
	{
		die('Type isn\'t ARRAY but ' . ref($configuredInterfaceArray));
	}

	return $configuredInterfaceArray;
}

sub _getInterfaceStatus{
	my ($raw) = @_;
	return $raw->{'oper-status'} // 'n/a';
}

sub _getEgressFilter
{
	my ($unit, $firewallEnabled) = @_;
	my $efilter = undef;
	
	if( $unit->{family} && $unit->{family}->{inet} ) 
	{
		my $inetConfig = $unit->{family}->{inet};
		
		# Output filters
		if( $inetConfig->{filter} && $firewallEnabled ) {
			if( $inetConfig->{filter}->{output} ) {

				# There are two different forms of input/output filter encoding, one with a 'filter-name'
				# subelement off of 'output', and one with it as a direct element value
				my $output = $inetConfig->{filter}->{output};
				$efilter = ( ref $output eq 'HASH' ) ? $output->{'filter-name'} : $output;
			}
		}
	}
				
	return $efilter;
}

sub _getIngressFilter{
	my ($unit, $firewallEnabled) = @_;
	my $ifilter = undef;
	
	if($firewallEnabled && $unit->{family} && $unit->{family}->{inet}){
		my $inetConfig = $unit->{family}->{inet};

		# Addresses and filters
		if($inetConfig->{filter}){
			if($inetConfig->{filter}->{input}){
				# There are two different forms of input/output filter encoding, one with a 'filter-name'
				# subelement off of 'output', and one with it as a direct element value
				my $input = $inetConfig->{filter}->{input};
				$ifilter = ref($input) eq 'HASH' ? $input->{'filter-name'} : $input;
			}
		}
	}

	return $ifilter;
}

sub _getPhysicalIfIndex{
	my ($raw) = @_;	
	return($raw->{'snmp-index'} // undef);
}

sub _getIfIndex{
	my ($raw, $unitKey) = @_;
	my $ifIndex = undef;
	
	if($raw){
		$ifIndex = $raw->{'snmp-index'};

		# See if overridden with index at logical interface level.  This could be a
		# hash or an array, depending on whether this is, for example, a serial
		# subinterface or allocated GigE subinterface.

		my $logicalInterface;

		if(ref($raw->{'logical-interface'}) eq 'ARRAY'){
			for $logicalInterface (@{$raw->{'logical-interface'}}){
				if($logicalInterface->{name} eq $unitKey){
					$ifIndex = $logicalInterface->{'snmp-index'};
					last;
				}
			}
		}else{
			$logicalInterface = $raw->{'logical-interface'};

			if(defined $logicalInterface and defined $logicalInterface->{$unitKey}){
				# Reassign ifIndex attr from logical i/f it it exists here
				$ifIndex = $logicalInterface->{$unitKey}->{'snmp-index'};
			}
		}
	}

	return $ifIndex;
}

sub _getMac{
	my ($raw, $iType) = @_;
	my $content = undef;		
	
	# We are looking in live data to suppliment config data
	if($iType && $iType =~ /ethernet/i && $raw->{'hardware-physical-address'}){
		my $addy = $raw->{'hardware-physical-address'};
			
		if(ref($addy) eq 'HASH'){
			$content = $addy->{content};
		}else{
			$content = $addy;
		}
		$content =~ s/://g;
	}

	return $content;
}

sub _getIpConfigurationList{
	my ($conf, $iType) = @_;
	
	my $addyList = undef;
	if($conf->{family}){
		if($conf->{family}->{inet}){
			my $inetConfig = $conf->{family}->{inet};
			if($inetConfig->{address}){
				$addyList = $inetConfig->{address};
			}
		}
		if($conf->{family}->{inet6}){
			my $inet6Config = $conf->{family}->{inet6};
			if($inet6Config->{address}){
				my $skipped;
				foreach my $ipv6_address ( @{$inet6Config->{address}} ) {
					$skipped .= $ipv6_address->{name} . ', ';
				}
				$skipped = substr($skipped, 0, -2);
				$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping interface IP addresses: $skipped");
			}
		}
	}

	return $addyList;
}

sub _writePhysicalInterface{
	my ($out,$raw,$interfaceStatus,$iType,$phyInterfaceName) = @_;
	
	$out->open_element('interface');
	$out->print_element('adminStatus', $interfaceStatus);
	
	# show chassis cluster status redundancy-group 0  only failover I found is at chassis level, not interface.
	
	my $physIfIndex = _getPhysicalIfIndex($raw);
	if($physIfIndex){
		$out->print_element('ifIndex', $physIfIndex);
	}
	
	my $mac = _getMac($raw, $iType);			  
	if($mac){
		$out->open_element('interfaceEthernet');
		$out->print_element('macAddress', $mac);
		$out->close_element('interfaceEthernet');
	}

	if($iType){
		$out->print_element('interfaceType',
			  $iType =~ /ethernet/i ? 'ethernet'
			: $iType =~ /loopback/i ? 'softwareLoopback'
			: $iType =~ /frame/i    ? 'frameRelay'
			: $iType =~ /sonet/i    ? 'sonet'
			: $iType =~ /token/i    ? 'tokenRing'
			: $iType =~ /atm/i      ? 'atm'
			: $iType =~ /ppp/i      ? 'ppp'
			:                         'other');
	}else{
		$out->print_element('interfaceType', 'unknown');
	}

	my $mtu = _getMtu($raw);
	if($mtu){
		$out->print_element('mtu', $mtu);
	}
	
	$out->print_element('name', $phyInterfaceName);
	
	$out->print_element('physical', 'true');
	
	if($raw->{speed}){
		$out->print_element('speed', _process_speed_rating($raw->{speed}));
	}
	$out->close_element('interface');
}

sub _processLogicalInterfaces{
	my ($out,$in,$raw,$interfaceStatus,$confInterfaceArray,$firewallEnabled,$iType,$phyInterfaceName) = @_;

	if($confInterfaceArray){
		if(ref $confInterfaceArray ne 'ARRAY'){
			die "Not an unit array '$phyInterfaceName', '" . ref($confInterfaceArray) . "'\n" . Dumper($confInterfaceArray);
		}
	} else {
		return;
	}

	foreach my $confInterface (@$confInterfaceArray){
		my $unitArray = $confInterface->{unit};

		foreach my $unit (@$unitArray){
			my $interfaceName = $phyInterfaceName . '.' . $unit->{name};

			$out->open_element('interface');
			$out->print_element('adminStatus', $interfaceStatus); 

			# Filters only at logical level
			my $efilter = _getEgressFilter($unit, $firewallEnabled);
			if($efilter){
				$out->print_element('egressFilter', $efilter);
			}
			
			# show chassis cluster status redundancy-group 0
			# The only failover I found is at chassis level, not interface.
			my $ifIndex = _getIfIndex($raw, $interfaceName);
			if($ifIndex){
				$out->print_element('ifIndex', $ifIndex);
			}
			
			my $ifFilter = _getIngressFilter($unit, $firewallEnabled);			  
			if($ifFilter){
				$out->print_element('ingressFilter', $ifFilter);
			}
			
			if($raw->{'if-type'}){
				$iType=$raw->{'if-type'};
			}

			my $ipConfigurationList = _getIpConfigurationList($unit, $iType);
			if($ipConfigurationList){
				$out->open_element('interfaceIp');
				foreach my $addy (@$ipConfigurationList){
					$out->open_element('ipConfiguration');

					my ($ip, $mask) = split(/\//, $addy->{name});
					$out->print_element('ipAddress', $ip);
					$out->print_element('mask', $mask);

					$out->close_element('ipConfiguration');

					# This is the requirement for $in if we were to return something we could avoid this. but meh.
					my $subnet_junos = new ZipTie::Adapters::Juniper::JUNOS::Addressing::SubnetJunos($ip, $mask);
					push(@{ $in->{IFIpForwarding}->{$interfaceName}}, $subnet_junos);
				}
				$out->close_element('interfaceIp');
			}

			if($in->{parsed} && $in->{parsed}->{zones}){
				my $zones = $in->{parsed}->{zones};
				foreach my $zone_name(keys(%{$zones})){
					my $zone = $zones->{$zone_name};
					foreach my $zone_interface(keys(%{$zone->{interfaces}})){
						if($interfaceName eq $zone_interface){
							$out->print_element('interfaceGrouping', {
								ifGroupName => $zone_name,
								ifGroupingCharacteristic => 'securityZoneMember'
							});
						}
					}
				}
			}


			if($iType){
				$out->print_element('interfaceType',
					  $iType =~ /ethernet/i ? 'ethernet'
					: $iType =~ /loopback/i ? 'softwareLoopback'
					: $iType =~ /frame/i    ? 'frameRelay'
					: $iType =~ /sonet/i    ? 'sonet'
					: $iType =~ /token/i    ? 'tokenRing'
					: $iType =~ /atm/i      ? 'atm'
					: $iType =~ /ppp/i      ? 'ppp'
					:                         'other');
			}else{
				$out->print_element('interfaceType', 'unknown');
			}

			my $mtu = _getMtu($raw);
			if($mtu){
				$out->print_element('mtu', $mtu);
			}

			$out->print_element('name', $interfaceName);
			$out->print_element('physical', 'false'); 
			if($unit->{bandwidth}){
				$out->print_element('speed', _process_speed_rating( $unit->{bandwidth}));
			}

			$out->close_element('interface');
		}
	}

	return;
}

sub _processInterface{
	my ($in, $out, $raw, $confInterfaceArray, $firewallEnabled, $phyInterfaceName) = @_;
	
	my $interfaceStatus = _getInterfaceStatus($raw);
	my $iType = $raw->{'link-level-type'};
	
	# Process physical interface
	_writePhysicalInterface($out, $raw, $interfaceStatus, $iType, $phyInterfaceName);
	
	#process logical interfaces
	_processLogicalInterfaces($out, $in, $raw, $interfaceStatus, $confInterfaceArray, $firewallEnabled, $iType, $phyInterfaceName);

 
	# Not sure what to do here, everything in the config that also appears in show interfaces is physical

	# Now after we get this all working we have to check
	# a. we have both if and else conditions covered from orignal method
	# b. explore the parse Interfaces for telemmetry and see if that can make use of this class.
	# c. we are writing both physical and logical interfaces
	# d. verify the filterlist applys to contains the correct interface name reported here
	# e. static routes contain interface element to the new name.
	
	# Should I try to be a little more cute adding processing for each known junos product classification?
	# perhaps its better to differentiate if idp firewall or routing features are found and what if there are combinations of features 
}

# $out is printer class XmlPrint
# we are retooling with a mind to operate over the show iterfaces data instead of the show configuration interfaces;
# with this in mind, the processing will look like 
# for each physical interface
#  do we have logical interface and is it defined in config interfaces if not continue
#  write phyiscal interface
#  write logical interface 
sub parse_interfaces{
	my ($in, $out) = @_;

	# Raw interfaces the xml parsed from the show interfaces command
	my $rawInterfaces =  _getRawInterfaces($in);

	# Conf interfaces the xml parsed from the show configuration interfaces command
	my $confInterfaces = _getConfInterfaces($in);

	my $firewallEnabled = 1;
	if($in->{showFirewallPlain} && $in->{showFirewallPlain} =~ /Warning: configuration block ignored/){
		$firewallEnabled = 0;
	}

	my $opened = 0;
	foreach my $phyInterfaceName (keys(%{$rawInterfaces})){
		if(!$opened){
			$out->open_element('interfaces');
			$opened = 1;
		}

		my $rawInterface = $rawInterfaces->{$phyInterfaceName};
		my $confInterfaceArray = _getConfiguredInterfaceArray($phyInterfaceName, $rawInterface, $confInterfaces);

		# TODO - It would be great to get $in out of the argument list; we need it to store data for later, perhaps the data to be set can be returned.	
		_processInterface($in, $out, $rawInterface, $confInterfaceArray, $firewallEnabled, $phyInterfaceName);
	}

	if($opened){
		$out->close_element('interfaces');
	}

	return;
}

#T1= 1544kbps
#T2=6312 kbps 
#T3= 44736 kbps 
#Relevant to Australia are:
#E1 = 2,048 Mbit/s
#E2 = 4 x E1 = 8,448 Mbit/s
#E3 = 4 x E2 = 34,368 Mbit/s
#E4 = 4 x E3 = 139,264 Mbit/s
#E5 = 4 x E4 = 565 Mbit/

# This just takes a text value, not a tree node
sub _process_speed_rating{
	my $speed = shift;

	if($speed =~ /^\d+$/){
		return $speed;
	}elsif($speed =~ /(\d+)m/i){
		return($1 * 1000 * 1000);
	}elsif($speed =~ /(\d+)k/i){
		return($1 * 1000);
	}elsif($speed =~ /T3/i){
		return(44736000);
	}elsif($speed =~ /T2/i){
		return(6312000);
	}elsif($speed =~ /T1/i){
		return(1544000);
	}elsif($speed =~ /E1/i){
		return(2048000);
	}elsif($speed =~ /E2/i){
		return(8448000);
	}elsif($speed =~ /E3/i){
		return(34368000);
	}elsif($speed =~ /E4/i){
		return(139264000);
	}elsif($speed =~ /E5/i){
		return(565000000);
	}elsif($speed =~ /(\d+)/i){
		return($1);
	}elsif($speed eq 'Unlimited'){
		# SNMP values Unlimited at 2147483647
		return('2147483647');
	}else{
		return(0);
	}
}

sub _getMtu{
	my ($raw) = @_;
	my $mtu = $raw->{mtu} // return;
	return($mtu eq q{Unlimited} ? 2147483647 : $mtu);
}

1;

__END__

=head1 NAME

ZipTie::Adapters::Juniper::JUNOS::GetActiveConfig - Retrieves the active configuration from an JUNOS-based device.

=head1 SYNOPSIS

    use ZipTie::Adapters::Juniper::JUNOS::GetActiveConfig qw(get_active_config);
	my $active_config = get_active_config($cli_protocol_obj);
	my $active_config = get_active_config($cli_protocol_obj, $connection_path);

=head1 DESCRIPTION

C<ZipTie::Adapters::Juniper::JUNOS::GetActiveConfig> allows for the retrieve of the active configuration
from an JUNOS-based device by using either a C<ZipTie::CLIProtocol> object to retrieve the active configuration
via the command line using the "file show /config/juniper.conf.gz" command, or by using a C<ZipTie::ConnectionPath>
object to retrieve the active config via a file transfer protocol client/agent. The only file transfer protocol currently
supported is SCP.

The only subroutine a user of C<ZipTie::Adapters::Juniper::JUNOS::GetActiveConfig> should be concerned with is 
C<get_active_config($cli_protocol_obj, $connection_path)>; it provides an abstraction layer that hides the mechanism
for retrieving the active configuration.

=head1 EXPORTED SUBROUTINES

=over 12

=item C<get_active_config($cli_protocol_obj, $connection_path)>

Main entry point into the functionality for retrieveing a active configuration from a JUNOS-based device.
It provides an abstraction layer that hides the mechanism for retrieving the configuration file.

First, it is examined to see if a file transfer protocol has been specified to use as the transfer mechanism.  If this is
the case, then the corresponding private subroutine that utilizes that type of file transfer protocol will be called.
The only file transfer protocol that is supported is currently is SCP and this is accomplished through the 
C<_get_active_config_scp($cli_protocol_obj, $connection_path)> private subroutine.

If no file transfer protocol has been specified, then the specified C<ZipTie::CLIProtocol> object will be used to
retrieve the active configuration via CLI through the C<_get_active_config_cli($cli_protocol_obj)> private
subroutine.

Regardless of the mechanism chosen, the contents of the active configuration will be returned.

Input:		$cli_protocol_obj -	A valid C<ZipTie::CLIProtocol> object that is already connected to a JUNOS-based device.
			$connection_path -	Optional.  C<A valid ZipTie::ConnectionPath> object that contains all of the IP, credential,
								protocol, and file server information needed to correctly retrieve the configuration
								from the the JUNOS-based device.

=back

=head1 PRIVATE SUBROUTINES

=over 12

=item C<_get_active_config_cli($cli_protocol_obj)>

Retrieves the active configuration for an JUNOS-based device by executing the "file show /config/juniper.conf.gz"
command via a C<ZipTie::CLIProtocol> object that has been previously connected to a JUNOS-based device.

This subroutine will be called by C<get_active_config($cli_protocol_obj, $connection_path)> if no valid
C<ZipTie::ConnectionPath> object was specified.

=item C<_get_active_config_scp($cli_protocol_obj, $connection_path)>

Retrieves the active configuration from the JUNOS-based device using an SCP client to retrieve the
"/config/juniper.conf.gz" gzip file that stores the configuration file.  The contents of this gzip file is then decompressed
using C<IO::Zlib> and the active configuration is parsed from it and returned.

This subroutine will be called by C<get_active_config($cli_protocol_obj, $connection_path)> if a
valid C<ZipTie::ConnectionPath> object containing protocol information representing SCP was specified.

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

Contributor(s): dwhite (dylamite@ziptie.org), Jak Devaney (john.devaney@uk.ibm.com)
Date: May 23, 2007

=cut
