package ZipTie::Adapters::Juniper::JUNOS::Parsers;

use strict;
use warnings;
use 5.010_000;

use Data::Dumper;
use MIME::Base64 qw(encode_base64);
use Net::IP;
use XML::Simple qw(XMLin);
use XML::Twig;

use ZipTie::Adapters::Utils qw(seconds_since_epoch strip_mac get_mask mask_to_bits get_port_number trim get_interface_type getUnitFreeNumber parseCIDR get_port_expression get_crep ncname_from_string get_protocol_name);
use ZipTie::Adapters::Juniper::JUNOS::Addressing::SubnetJunos;
use ZipTie::Logger;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

use Exporter qw(import);
our @EXPORT_OK = qw(parse_routing_neighbors parse_arp parse_telemetry_interfaces parse_routing create_config parse_local_accounts parse_chassis parse_filters parse_static_routes parse_snmp parse_system parse_object_groups clear_parsing_state);

my $ip_regex = get_crep('cipm');
my $ipv6_regex = get_crep('ipv6');

my $securityApps = {
	# ICMP-replated
	'ICMP-ANY'                           => [ { protocol => 'icmp' }, ],
	'ICMP-ADDRESS-MASK'                  => [ { protocol => 'icmp' }, ],
	'ICMP-DEST-UNREACH'                  => [ { protocol => 'icmp' }, ],
	'ICMP Fragment Needed'               => [ { protocol => 'icmp' }, ],
	'ICMP FragmentReassembly'            => [ { protocol => 'icmp' }, ],
	'ICMP-HOST-UNREACH'                  => [ { protocol => 'icmp' }, ],
	'ICMP-INFO'                          => [ { protocol => 'icmp' }, ],
	'ICMP-PARAMETER-PROBLEM'             => [ { protocol => 'icmp' }, ],
	'ICMP-PORT-UNREACH'                  => [ { protocol => 'icmp' }, ],
	'ICMP-PROTOCOL-UNREACH'              => [ { protocol => 'icmp' }, ],
	'ICMP-REDIRECT'                      => [ { protocol => 'icmp' }, ],
	'ICMP-REDIRECT-HOST'                 => [ { protocol => 'icmp' }, ],
	'ICMP-REDIRECT-TOS-HOST'             => [ { protocol => 'icmp' }, ],
	'ICMP-REDIRECT-TOS-NET'              => [ { protocol => 'icmp' }, ],
	'ICMP-SOURCE-QUENCH'                 => [ { protocol => 'icmp' }, ],
	'ICMP-SOURCE-ROUTE-FAIL'             => [ { protocol => 'icmp' }, ],
	'ICMP-TIME-EXCEEDED'                 => [ { protocol => 'icmp' }, ],
	'ICMP-TIMESTAMP'                     => [ { protocol => 'icmp' }, ],
	'ICMP-ECHO-FRAGMENT-ASSEMBLY-EXPIRE' => [ { protocol => 'icmp' }, ],
	'Traceroute'                         => [ { protocol => 'icmp' }, ],

	# Internet related
	'AOL' => [
		{
			portRange =>
			  { portStart => 5190, portEnd => 5193, protocol => 'tcp', },
		},
	],
	'DHCP relay' => [
		{
			portExpression =>
			  { operator => 'eq', port => 67, protocol => 'udp', },
		},
	],
	'DHCP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 67, protocol => 'udp', },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 68, protocol => 'udp', },
		},
	],
	'DNS' => [
		{
			portExpression =>
			  { operator => 'eq', port => 53, protocol => 'udp', },
		},
	],
	'FTP-Get' => [
		{
			portExpression =>
			  { operator => 'eq', port => 20, protocol => 'tcp', },
		},
	],
	'FTP-Put' => [
		{
			portExpression =>
			  { operator => 'eq', port => 21, protocol => 'tcp', },
		},
	],
	'Gopher' => [
		{
			portExpression =>
			  { operator => 'eq', port => 70, protocol => 'tcp', },
		},
	],
	'HTTP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 8080, protocol => 'tcp', },
		},
	],
	'HTTPS' => [
		{
			portExpression =>
			  { operator => 'eq', port => 443, protocol => 'tcp', },
		},
	],
	'IRC' => [
		{
			portRange =>
			  { portStart => 6665, portEnd => 6669, protocol => 'tcp', },
		},
	],
	'LDAP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 389, protocol => 'tcp', },
		},
	],
	'TFTP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 69, protocol => 'udp', },
		},
	],

	# M$ related, many missing because they are based of off content, not ports
	'MS-RPC-EPM' => [
		{
			portExpression =>
			  { operator => 'eq', port => 135, protocol => 'tcp' },
		},
	],

	# Dynamic routing related
	'RIP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 520, protocol => 'udp' },
		},
	],
	'OSPF' => [ { protocol => 'osfp' } ],

	# streaming related
	'H.323' => [
		{
			portExpression =>
			  { operator => 'eq', port => 1720, protocol => 'tcp', },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 1503, protocol => 'tcp', },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 389, protocol => 'tcp', },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 522, protocol => 'tcp', },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 1731, protocol => 'tcp', },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 1719, protocol => 'udp', },
		},
	],
	'NetMeeting' => [
		{
			portExpression =>
			  { operator => 'eq', port => 1720, protocol => 'tcp', },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 1503, protocol => 'tcp', },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 389, protocol => 'tcp', },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 522, protocol => 'tcp', },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 1731, protocol => 'tcp', },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 1719, protocol => 'udp', },
		},
	],
	'Real Media' => [
		{
			portExpression =>
			  { operator => 'eq', port => 7070, protocol => 'tcp' },
		},
	],
	'RTSP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 554, protocol => 'tcp' },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 554, protocol => 'udp' },
		},
	],
	'SIP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 5056, protocol => 'udp' },
		},
	],
	'VDO Live' => [
		{
			portRange =>
			  { portStart => 7000, portEnd => 7010, protocol => 'tcp', },
		},
	],

	# SUN RPC related are all left out because they are protocol based.
	# IPSEC/Tunnel related
	'IKE' => [
		{
			portExpression =>
			  { operator => 'eq', port => 500, protocol => 'udp' },
		},
	],
	'L2TP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 1723, protocol => 'tcp' },
		},
	],
	'PPTP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 1723, protocol => 'tcp' },
		},
	],

	# IM related
	'Gnutella' => [
		{
			portExpression =>
			  { operator => 'eq', port => 6346, protocol => 'tcp' },
		},
	],
	'MSN' => [
		{
			portExpression =>
			  { operator => 'eq', port => 1863, protocol => 'tcp' },
		},
	],
	'NNTP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 119, protocol => 'tcp' },
		},
	],
	'SMB' => [
		{
			portExpression =>
			  { operator => 'eq', port => 445, protocol => 'tcp' },
		},
	],
	'YMSG' => [
		{
			portExpression =>
			  { operator => 'eq', port => 5010, protocol => 'tcp' },
		},
	],

	# management related
	'NBNAME' => [
		{
			portExpression =>
			  { operator => 'eq', port => 137, protocol => 'udp' },
		},
	],
	'NDBDS' => [
		{
			portExpression =>
			  { operator => 'eq', port => 138, protocol => 'udp' },
		},
	],
	'NTP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 123, protocol => 'udp' },
		},
	],
	'RLOGIN' => [
		{
			portExpression =>
			  { operator => 'eq', port => 513, protocol => 'tcp' },
		},
	],
	'RSH' => [
		{
			portExpression =>
			  { operator => 'eq', port => 514, protocol => 'tcp' },
		},
	],
	'SNMP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 161, protocol => 'udp' },
		},
	],
	'SQL*Net V1' => [
		{
			portExpression =>
			  { operator => 'eq', port => 66, protocol => 'tcp' },
		},
	],
	'SQL*Net V2' => [
		{
			portExpression =>
			  { operator => 'eq', port => 66, protocol => 'tcp' },
		},
	],
	'MSSQL' => [
		{
			portExpression =>
			  { operator => 'eq', port => 1433, protocol => 'tcp' },
		},
	],
	'SSH' => [
		{
			portExpression =>
			  { operator => 'eq', port => 22, protocol => 'tcp' },
		},
	],
	'Talk' => [
		{
			portRange =>
			  { portStart => 517, portEnd => 518, protocol => 'tcp', },
		},
	],
	'Telnet' => [
		{
			portExpression =>
			  { operator => 'eq', port => 23, protocol => 'tcp' },
		},
	],

	# Mail related
	'IMAP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 143, protocol => 'tcp' },
		},
	],
	'SMTP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 25, protocol => 'tcp' },
		},
	],
	'POP3' => [
		{
			portExpression =>
			  { operator => 'eq', port => 110, protocol => 'tcp' },
		},
	],

	# UNIX related
	'FINGER' => [
		{
			portExpression =>
			  { operator => 'eq', port => 79, protocol => 'tcp' },
		},
	],
	'UUCP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 117, protocol => 'tcp' },
		},
	],

	# Misc services
	'CHARGEN' => [
		{
			portExpression =>
			  { operator => 'eq', port => 19, protocol => 'tcp' },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 19, protocol => 'udp' },
		},
	],
	'DISCARD' => [
		{
			portExpression =>
			  { operator => 'eq', port => 9, protocol => 'tcp' },
		},
		{
			portExpression =>
			  { operator => 'eq', port => 9, protocol => 'udp' },
		},
	],
	'LPR' => [
		{
			portExpression =>
			  { operator => 'eq', port => 515, protocol => 'tcp' },
		},
	],
	'RADIUS' => [
		{
			portExpression =>
			  { operator => 'eq', port => 1812, protocol => 'tcp' },
		},
	],
	'SQLMON' => [
		{
			portExpression =>
			  { operator => 'eq', port => 1434, protocol => 'udp' },
		},
	],
	'VNC' => [
		{
			portExpression =>
			  { operator => 'eq', port => 5800, protocol => 'tcp' },
		},
	],
	'WHOIS' => [
		{
			portExpression =>
			  { operator => 'eq', port => 43, protocol => 'tcp' },
		},
	],
	'SCCP' => [
		{
			portExpression =>
			  { operator => 'eq', port => 2000, protocol => 'tcp' },
		},
	],
};

sub parse_routing_neighbors{
	my ($in, $out) = @_;

	my %protocols = (
		ospf => {
			regex => q{^([\da-f.:]+)\s+(\S+)\s+\S+\s+([\da-f.:]+)\s+\d+\s+\d+\s*$},
			a => 'ipAddress',
			b => 'interface',
			c => 'routerId'
		},
		bgp => {
			regex => q{^\s*Peer\s+ID:\s+([\da-f.:]+)\s+Local\s+ID:\s+([\da-f.:]+)},
			a => 'ipAddress',
			b => 'routerId'
		},
		rip => {
			regex => q{^\s*(\S+)\s+\S+\s+([\da-f.:]+)\s+([\da-f.:]+)},
			a => 'interface',
			b => 'routerId',
			c => 'ipAddress'
		}
	);

	my $neighbors;
	foreach my $protocol ( keys( %protocols ) ){
		if($in->{$protocol}){

			my $template = $protocols{$protocol};
			NEIGHBOR: while ($in->{$protocol} =~ /$template->{regex}/gmi){
				my $neighbor = {
					protocol => uc($protocol), 
					$template->{a} => $1,
					$template->{b} => $2
				};

				if($template->{c}){
					$neighbor->{$template->{c}} = $3;
				}

				foreach my $possible_ipv6 ( 'ipAddress', 'routerId' ) {
					if($neighbor->{$possible_ipv6} && $neighbor->{$possible_ipv6} =~ /$ipv6_regex/i ){
						$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping $neighbor->{protocol} route: $possible_ipv6 = $neighbor->{$possible_ipv6}");
						next(NEIGHBOR);
					}
				}

				push(@{$neighbors->{routingNeighbor}}, $neighbor);
			}
		}
	}

	if(exists $neighbors->{routingNeighbor}){
		$out->print_element('routingNeighbors', $neighbors);
	}
}

sub parse_arp{
	my ($in, $out) = @_;
	my $opened = 0;

	if($in->{arp}){
		while ($in->{arp} =~ /^([\da-f:]+)\s+([\da-f.:]+)\s+(\S+)/mgi){
			my ($mac, $ip, $interface) = ($1, $2, $3);

			if( $ip =~ /$ip_regex/ ) {
				if( !$opened ) {
					$out->open_element('arpTable');
					$opened = 1;
				}
				my $arp = {
					ipAddress  => $ip,
					macAddress => strip_mac($mac),
					interface  => $interface
				};
				$out->print_element('arpEntry', $arp);
			} elsif( $ip =~ /$ipv6_regex/i ) {
				$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping ARP entry: $ip : $mac on $interface");
			}
		}
		if( $opened ) {
			$out->close_element('arpTable');
		}
	}
}

# Creates the ZipTie interfaces model by reading the Juniper interface-information
# XML with XML::Twig and then traversing the twigs.  This could be streamlined
# by reading only physical-interface blocks at a time and using a twig callback
# to a method that parses that single interface
sub parse_telemetry_interfaces{
	my ($in, $out) = @_;
	my $interfaces;

	my ($intSection) = $in->{interfaces} =~ /(<interface-information.+?<\/interface-information>)/s;
	my $interfacesTwig = XML::Twig->new();
	$interfacesTwig->parse($intSection);

	# TODO JAK - Is this needed?
	# This undef pattern is all through the code, I believe it is meant to stop a memory leak with XML::Twig
	undef($intSection);

	foreach my $physicalInt ($interfacesTwig->descendants('physical-interface')){
		my $name       = $physicalInt->first_child('name')->text();
		my $operStatus = lc($physicalInt->first_child('oper-status')->text());

		my $type       = $name;
		if($physicalInt->first_child('if-type')){
			$type = $physicalInt->first_child('if-type')->text();
		}
		$type = get_interface_type($type);

		my $trafficStatsElement = $physicalInt->first_child('traffic-statistics');
		my $inputBytes = 0;
		if(defined($trafficStatsElement)){
			my $inputPacketsElement = $trafficStatsElement->first_child('input-packets');
			if(defined $inputPacketsElement){
				$inputBytes = $inputPacketsElement->text();
			}
		}

		my $majorInt = {
			name       => $name,
			operStatus => ucfirst($operStatus),
			type       => $type,
			inputBytes => $inputBytes,
		};

		push(@{$interfaces->{interface}}, $majorInt);

		foreach my $logicalInt ($physicalInt->descendants('logical-interface')){

			my $subName;
			if($logicalInt->first_child('name')){
				$subName = $logicalInt->first_child('name')->text();
			}else{
				next;
			}

			my $subInt = {
				name       => $subName,
				operStatus => ucfirst($operStatus),
				type       => $type,
				inputBytes => 0,
			};

			foreach my $fam ($logicalInt->descendants('address-family')){
				foreach my $addr ($fam->descendants('interface-address')){
					my $child = $addr->first_child('ifa-local');
					if($child){
						my $net_ip = new Net::IP($addr->first_child('ifa-local')->text());
						if(defined($net_ip)){
							if($net_ip->ip() =~ /$ipv6_regex/i){
								$LOGGER->debug("WARNING: IPv6 not supported by QRM, skipping sub interface $subName");
								next;
							}
							my $ipConfiguration = {ipAddress => $net_ip->ip()};
							if(defined($addr->first_child('ifa-destination'))){
								my ($mask) = $addr->first_child('ifa-destination')->text() =~ /\/(\d+)/;
								$ipConfiguration->{mask} = $mask;
							}
							push(@{$subInt->{ipEntry}}, $ipConfiguration);
						}
					}
				}
			}
			push(@{$interfaces->{interface}}, $subInt);
		}
	}

	$out->attributes(1);
	$out->print_element('interfaces', $interfaces);
	return $interfaces;
}

# Process BGP configuration data
sub parse_routing{
	my ($in, $out) = @_;
	my $routing;
	my ($bgpSection) = $in->{showBgp} =~ /(<bgp-information.+?<\/bgp-information>)/s;

	if($bgpSection){
		my $twig = XML::Twig->new();
		$twig->parse($bgpSection);
		undef($bgpSection);

		my $bgp;
		foreach my $peerElement ($twig->descendants('bgp-peer')){
			if(!$bgp->{routerId}){
				my $routerId = $peerElement->first_child('local-address')->text();
				$routerId =~ s/\+\d+$//;

				if($routerId ne 'unspecified'){
					$bgp->{routerId} = $routerId;
				}
			}

			if(!$bgp->{asNumber}){
				$bgp->{asNumber} = $peerElement->first_child('local-as')->text();
			}

			my $neighbor = {
				address  => $peerElement->first_child('peer-address')->text(),
				asNumber => $peerElement->first_child('peer-as')->text(),
			};

			# Strip trailing digits after the +
			$neighbor->{address} =~ s/\+\d+$//;

			if( $neighbor->{address} !~ /$ipv6_regex/ ) {
				push(@{$bgp->{neighbor}}, $neighbor);
			} else {
				$LOGGER->warning($ROUTE_DROPPED, "IPv6 peer address not supported: '$neighbor->{address}'");
			}
		}

		if($bgp){
			$routing->{bgp} = $bgp;
		}
	}

	if($routing){
		$out->print_element('routing', $routing);
	}
}

sub clear_parsing_state{
	my ($in) = @_;

	if(defined $in->{IFIpForwarding}){
		delete($in->{IFIpForwarding});
	}
}

sub parse_chassis{
	my ($in, $out) = @_;

	$out->open_element('chassis');

	my $chassisAsset = {
			'core:assetType' => 'Chassis',
			'core:factoryinfo' => {
				'core:make' => 'Juniper'
			}
	};

	if($in->{showVer} =~ /^\s*Model:\s+(\S+)/im){
		$chassisAsset->{'core:factoryinfo'}->{'core:modelNumber'} = $1;
	}

	my (@powerSupplies, @cards, @cpus);
	my ($hardwareSection) = $in->{showHardware} =~ /(<chassis[\s|>].+?<\/chassis>)/s;

	if($hardwareSection){
		my $twig = XML::Twig->new();
		$twig->parse($hardwareSection);
		undef($hardwareSection);

		my $rootElement = $twig->root();

		if( $rootElement->first_child('serial-number') ) {
			$chassisAsset->{'core:factoryinfo'}->{'core:serialNumber'} = $rootElement->first_child('serial-number')->text();
		} elsif( $in->{showVer} =~ /^\s*Serial\s+Number:\s+(\S+)\s*$/im ){
			$chassisAsset->{'core:factoryinfo'}->{'core:serialNumber'} = $1;
		}

		$out->print_element('core:asset', $chassisAsset);
		$out->print_element('core:description', $rootElement->first_child('description')->text());

		foreach my $chassisModule ($rootElement->descendants('chassis-module')){
			my $module = _parse_module($chassisModule);

			my $name   = $chassisModule->first_child('name')->text();
			if($name =~ /Power\s+Supply\s+(\S)/i){
				my $number = ord($1) - 64;
				$module->{'core:asset'}->{'core:assetType'} = 'PowerSupply';
				$module->{number} = $number;
				push(@powerSupplies, $module);
			}elsif($name =~ /^FPC\s+(\d+)/ ){
				$module->{'core:asset'}->{'core:assetType'} = 'Card';
				$module->{'slotNumber'} = $1;

				# Daughter cards
				foreach my $daughterCard ($chassisModule->descendants('chassis-sub-module')){
					my $dc = _parse_module($daughterCard);
					$dc->{'core:asset'}->{'core:assetType'} = 'Card';

					my $name = $daughterCard->first_child('name')->text();
					if($name =~ /(\d+)$/){
						$dc->{slotNumber} = $1;
					}

					push(@{$module->{daughterCard}}, $dc);
				}
				push(@cards, $module);
			}elsif($name =~ /^FEB/){
				$module->{'core:asset'}->{'core:assetType'} = 'CPU';
				push(@cpus, $module);
			}
		}
	}

	foreach my $card (@cards){
		$out->print_element('card', $card);
	}
	foreach my $cpu (@cpus) {
		$out->print_element('cpu', $cpu);
	}

	if($in->{showMAC} && $in->{showMAC} =~ /Public base address\s+((?:\w{2}\:){5}\w{2})/){
		$out->print_element('macAddress', strip_mac($1));
	}

	if($in->{showRen} =~ /^\s*DRAM\s+(\d+)\s+(\S+)/mi){
		my $memory = {
			'core:description' => 'RAM',
			kind               => 'RAM',
			size               => getUnitFreeNumber($1, $2, 'byte'),
		};
		$out->print_element('memory', $memory);
	}

	foreach my $power_supply (@powerSupplies){
		$out->print_element('powersupply', $power_supply);
	}

	$out->close_element('chassis');

	return;
}

sub parse_system{
	my ($in, $out) = @_;

	if($in->{showVer} =~ /^\s*Hostname:\s+(\S+)/im){
		$out->print_element('core:systemName', $1);
	}elsif($in->{snmp} =~ /<system-name>(.*)<\/system-name>/im){
		$out->print_element('core:systemName', $1);
	}

	$out->open_element('core:osInfo');

	if(defined $in->{showFirmware} and $in->{showFirmware} =~ /Juniper ROM Monitor Version\s(\d+\.\w+)/s){
		$out->print_element('core:fileName', $1);
	}

	$out->print_element('core:make', 'Juniper');

	my $version;
	if($in->{showVer} =~ /^\s*JUNOS (Base OS Software Suite|Software Release|Base Version) \[(.*)\]/im){
		$version = $2;
		$out->print_element('core:version', $version);
	}

	$out->print_element('core:osType', 'JUNOS');
	$out->close_element('core:osInfo');

	if($version){
		$out->print_element('core:biosVersion', $version);
	}

	my $applianceType = 'Firewall';
	my $verString = $in->{showVer};
	if($verString =~ /^\s*Model:\s+(\S+)/im){
		my $modelStr = $1;
		my $patHash = _get_patterns();

		foreach my $patKey (keys(%{$patHash})){
			my $pat = $patHash->{$patKey};

			if($modelStr =~ /$pat/i){
				$applianceType = $patKey;
				last;
			}
		}
	}
	$out->print_element('core:deviceType', $applianceType);

	# System booted: 2007-03-08 10:09:09 CST (6w6d 03:03 ago)
	if($in->{showUptime} =~ /^System booted:\s+(\d+)-(\d+)-(\d+)\s+(\d{1,2}:\d{1,2}:\d{1,2})\s+(\S+)/mi){
		my ($year, $month, $day, $time, $timezone) = ($1, $2, $3, $4, $5);
		my ($hour, $min, $sec) = $time =~ /(\d+):(\d+):(\d+)/;
		$out->print_element('core:lastReboot', seconds_since_epoch($sec, $min, $hour, $day, $month, $year, $timezone));
	}

	return;
}

# Populates the configuration entity for the main Juniper config
sub create_config{
	my ($in, $out) = @_;

	# The name of the repository
	my $repository = {'core:name' => '/'};

	my $active = {
		'core:name'       => 'active-config',
		'core:textBlob'   => encode_base64($in->{active}),
		'core:mediaType'  => 'text/plain',
		'core:context'    => 'active',
		'core:promotable' => 'false',
	};

	# Push the configuration into the repository
	push(@{$repository->{'core:config'}}, $active);

	# Print the repository
	$out->print_element('core:configRepository', $repository);

	return;
}

sub parse_static_routes{
	my ($in, $out) = @_;
	return unless ($in->{showRoute} =~ /rpc-reply/);

	my $staticRoutes;
	my $vrfRoutes;

	my ($show_route) = $in->{showRoute} =~ /(<rpc-reply.+?<\/rpc-reply>)/s;

	my $arrays = ['route'];
	my $reply = XMLin($show_route, ForceArray => $arrays, KeyAttr => []);

	my ($defaultMetric, $defaultPreference);
	my $defaultRouteBlock = $reply->{configuration}->{'routing-options'}->{static}->{defaults};
	if($defaultRouteBlock){
		if(defined($defaultRouteBlock->{metric}) && defined($defaultRouteBlock->{metric}->{'metric-value'})){
			$defaultMetric = $defaultRouteBlock->{metric}->{'metric-value'};
		}

		if(defined($defaultRouteBlock->{preference}) && defined($defaultRouteBlock->{preference}->{'metric-value'})){
			$defaultPreference = $defaultRouteBlock->{preference}->{"metric-value"};
		}
	}

	my $routes = $reply->{configuration}->{'routing-options'}->{static}->{route};
	foreach my $route (ref($routes) eq 'ARRAY' ? @$routes : ($routes)){
		if(!defined($route->{name})){ next; }

		my $cidr = $route->{name};
		if($cidr eq 'default'){
			$cidr = '0.0.0.0/0';
		}

		# Skip routes with a destination with zero as the first byte (but not 0.0.0.0)
		if($cidr =~ /^0\./ and $cidr !~ /^0\.0\.0\.0/){
			next;
		}

		if(defined($route->{discard})){
			my $static_route = _create_route($cidr);
			$static_route->{discard} = 'true';
			$static_route->{interface} = 'n/a';
			push(@{$staticRoutes->{staticRoute}}, $static_route);
		}else{
			my $efRoute = $route;

			# See if we have to chase this to a VRF table
			if(defined($efRoute->{'next-table'})){

				# We want to chase the first in a potentially dot-delimited path
				my ($tblName) = $efRoute->{'next-table'} =~ /^([^\s\.]+)\.?.*/;

				# Look for named VRF which matches what we can find here
				if(exists($in->{showNamedVrfs}->{$tblName})){

					# We're trying to cache the XML serialization of the VRFs, since there tends to be
					# many references to a relatively small number of tables
					if(!defined($vrfRoutes->{$tblName})){
						$vrfRoutes->{$tblName} = XMLin(
							$in->{showNamedVrfs}->{$tblName} =~ /(<instance.+?<\/instance>)/s,
							ForceArray => $arrays,
							KeyAttr    => {route => 'name'}
						); # 'routes' is the key in the VRF as well
					}

					# Now, reassign from the routing def in the VRF
					if(defined $vrfRoutes->{$tblName}->{'routing-options'}->{static}->{route}->{$cidr}){
						$efRoute = $vrfRoutes->{$tblName}->{'routing-options'}->{static}->{route}->{$cidr};

						# Provide the interface from this VRF routing options def, ifpresent
						$efRoute->{interfaceName} = $vrfRoutes->{$tblName}->{interface}->{name};
					}else{
						$LOGGER->debug("JUNOS: parse_static_routes : unable to find VRF def for route $cidr in vrtable '$tblName'");
						next;
					}
				}else{
					$LOGGER->debug("JUNOS: parse_static_routes : cannot find vrTable '$tblName' in routing config");
					next;
				}
			} # else, $efRoute is fully defined in line, no need to chase a table ref.

			my $next_hops = $efRoute->{'next-hop'};

			my $index = 0;
			foreach my $next_hop (ref($next_hops) eq 'ARRAY' ? @$next_hops : ($next_hops)){
				my $metric_key = 'metric';
				my $preference_key = 'preference';

				if(++$index > 1){
					$metric_key .= $index;
					$preference_key .= $index;
				}

				my $static_route = _create_route($cidr);
				if(defined($next_hop)){

					# When next hop is an interface name, replace with the first ip associated with that interface
					if ($next_hop !~ /^$ip_regex$/) {

						if ( $in->{IFIpForwarding}->{$next_hop} ) {
							$next_hop = $in->{IFIpForwarding}->{$next_hop}[0]->get_original_host();
						} else {
							$LOGGER->debug("JUNOS: parse_static_routes : unable to find resolve interface name '$next_hop' to get the gateway address for static route $cidr");
							next;
						}
					}
					$static_route->{gatewayAddress} = $next_hop;

					# Determining  interface is a matter of finding the interface definition whose direct
					# forwarding matches this next hop.  For starters, we may have forced the presence of the
					# i/f from knowing it when the VRF was parsed.
					my $ifNameFound = $efRoute->{interfaceName};

					if(!defined($ifNameFound)){
						while (my ($ifName, $ifSNArrayRef) = each(%{$in->{IFIpForwarding}})){
							foreach my $ifSNRef (@{$ifSNArrayRef}){
								if($ifSNRef->contains($next_hop)){
									$ifNameFound = $ifName;
									last;
								}
							}
							if(defined($ifNameFound)){

								# This is necessary to reset the internal iterator for 'each'.
								# see http://perldoc.perl.org/functions/each.html.
								keys %{$in->{IFIpForwarding}};
								last;
							}
						}
					}

					if(defined($ifNameFound)){
						$static_route->{interface} = $ifNameFound;
					}else{
						# If we can't find an interface, we're creating a nonconforming static route
						$LOGGER->debug("JUNOS: parse_static_routes : unable to find forwarding interface for static route gateway $next_hop");
					}
				}else{
					$LOGGER->debug("JUNOS: parse_static_routes : unable to find gateway for static route $cidr");
					next;
				}

				if($metric_key){
					# Use metric from block OR preference from block OR the default metric OR the default preference
					if(defined($efRoute->{$metric_key}) and defined($efRoute->{$metric_key}->{'metric-value'})){
						$static_route->{routeMetric} = $efRoute->{$metric_key}->{'metric-value'};
					}elsif(defined($efRoute->{$preference_key}) and defined($efRoute->{$preference_key}->{'metric-value'})){
						$static_route->{routeMetric} = $efRoute->{$preference_key}->{'metric-value'};
					}elsif($defaultMetric){
						$static_route->{routeMetric} = $defaultMetric;
					}elsif($defaultPreference){
						$static_route->{routeMetric} = $defaultPreference;
					}
				}

				push(@{$staticRoutes->{staticRoute}}, $static_route);
			}
		}
	}

	if(defined($staticRoutes)){
		$out->print_element('staticRoutes', $staticRoutes);
	}

	return;
}

sub _create_route{
	my ($cidr) = @_;

	my ($addr, $mask) = split(/\//, $cidr);

	my $static_route = {
		destinationAddress => $addr,
		destinationMask    => $mask,
		defaultGateway     => $cidr eq '0.0.0.0/0' ? 'true' : 'false',
	};

	return $static_route;
}

# Using the configuration, populate the ZipTie localAccounts model
sub parse_local_accounts {
	my ($in, $out) = @_;

	$out->open_element('localAccounts');

	my $login_ref = _get_section('login', $in->{active});
	my $login = pop(@{$login_ref});
	my $users = _get_section('user', $login);
	foreach my $user_data (@{$users}){
		my ($user)  = $user_data =~ /^\s+user\s*(\S+)/mi;
		my ($class) = $user_data =~ /^\s+class\s*(\S+\b)/mi;

		my $account = {
			accountName => $user,
			accessGroup => $class,
		};
		$out->print_element('localAccount', $account);
	}
	$out->close_element('localAccounts');

	return;
}

sub _get_applies_to {
	my ($in) = @_;

	# Get the interfaces that have lists of filters defined
	my ($payload) = $in->{showConfigInterfaces} =~ /(<interfaces.+?<\/interfaces>)/sm;

	my $arrays = ['input-list', 'output-list', 'address', 'unit'];
	my $interfaces    = XMLin( $payload, ForceArray => $arrays, KeyAttr => [] );
	my $interfaceList = $interfaces->{interface};
	my $appliesTo     = {};
	foreach my $int (@$interfaceList){
		if($int->{unit}){
			foreach my $unit (@{$int->{unit}}){
				if($unit->{family} && $unit->{family}->{inet} && $unit->{family}->{inet}->{filter}){

					my $outputList = $unit->{family}->{inet}->{filter}->{'output-list'};
					if($outputList){
						foreach my $list (@$outputList){
							my $lists = $appliesTo->{$list};
							if(!$lists){
								$lists = {
									out => [],
									in  => [],
								};
								$appliesTo->{$list} = $lists;
							}
							my $out = $lists->{out};
							push(@$out, $int->{name} . '.' . $unit->{name});
						}
					}

					my $inputList = $unit->{family}->{inet}->{filter}->{'input-list'};
					if($inputList){
						foreach my $list (@$inputList){
							my $lists = $appliesTo->{$list};
							if(!$lists){
								$lists = {
									out => [],
									in  => [],
								};
								$appliesTo->{$list} = $lists;
							}
							my $input = $lists->{in};
							push(@$input, $int->{name} . '.' . $unit->{name});
						}
					}
				}
			}
		}
	}
	return($appliesTo);
}

# Uses partial loading of the firewall XML output to process all filters
sub parse_filters{
	my ($in, $out) = @_;

	$out->open_element('filterLists');

	my ($firewallSection) = $in->{showFirewall} =~ /(<firewall.+?<\/firewall>)/s;

	# If the inet6 block is ignore then lets strip it out so that it isn't parsed
	if($in->{showFirewallInet6} && $in->{showFirewallInet6} =~ /Warning: configuration block ignored/){
		$firewallSection =~ s/<inet6>.+?<\/inet6>//s;
	}

	if($in->{showFirewallPlain} && $in->{showFirewallPlain} =~ /Warning: configuration block ignored/m){
		$firewallSection =~ s/<firewall>.+?<\/firewall>//s;
	}

	if($firewallSection){
		my $filterListHash = {};
		my $twig = XML::Twig->new(
			twig_roots => {
				'filter' => _process_filter($in, $out, _get_applies_to($in), $filterListHash)
			}
		);
		$twig->parse($firewallSection);
		undef $firewallSection;
	}

	if(defined($in->{showPolicy}) and $in->{showPolicy} =~ /configuration/){
		_parse_srx_filters($in, $out);
	}

	$out->close_element('filterLists');

	return;
}

sub _parse_zone_interfaces{
	my ($in, $out) = @_;

	# TODO this should all be assigned to objectGroups
	return unless ($in->{showZones} =~ /rpc-reply/);
	my ($zoneSec) = $in->{showZones} =~ /(<rpc\-reply[\s|>].+?<\/rpc\-reply>)/s;

	my $result;

	my $arrays = ['address', 'security-zone', 'interfaces', 'address-set'];
	my $reply = XMLin($zoneSec, ForceArray => $arrays);

	my $zones = $reply->{configuration}->{security}->{zones}->{'security-zone'};
	while (my ($name, $zone) = each(%{$zones})){
		my $interfaces = $zone->{interfaces};
		$result->{$name}->{interfaces} = $interfaces;
	}

	$in->{parsed}->{zones} = $result;

	return($result);
}

# SRX IPS configuration...
sub _parse_srx_filters{
	my ($in, $out) = @_;

	my $arrays = ['destination-address', 'source-address', 'application'];
	my ($policySec) = $in->{showPolicy} =~ /(<rpc\-reply[\s|>].+?<\/rpc\-reply>)/s;

	my $reply = XMLin($policySec, ForceArray => $arrays, KeyAttr => []);
	unless (defined($reply->{configuration}->{security}->{policies})){
		# SRX policy rules aren't applicable.
		return;
	}

	my $zones_interfaces = _parse_zone_interfaces($in, $out);
	my $applications = _parse_applications($in, $out);
	my $defaultPolicy = $reply->{configuration}->{security}->{policies}->{'default-policy'};
	if($defaultPolicy->{'permit-all'}){
		$defaultPolicy = 'permit';
	}else{
		$defaultPolicy = 'deny';
	}

	my $policies = $reply->{configuration}->{security}->{policies}->{policy};
	if( $policies ) {
		foreach my $policy_set (ref($policies) eq 'ARRAY' ? @$policies : ($policies)){
			$out->open_element('filterList');

			# Only set the applies to if there are interfaces assigned to both the from-zone
			# and the to-zone.  In SRX the appliesTo must be complete with source and destination
			# interfaces to be valid
			my $fromZoneInterfaces = $zones_interfaces->{$policy_set->{'from-zone-name'}};
			my $toZoneInterfaces = $zones_interfaces->{$policy_set->{'to-zone-name'}};
			if($fromZoneInterfaces) {
				if(scalar(keys(%{$fromZoneInterfaces->{interfaces}}))){
					if($toZoneInterfaces){
						if(scalar(keys(%{$toZoneInterfaces->{interfaces}}))){
							$out->open_element('appliesTo');
							$out->print_element('sourceGroup', $policy_set->{'from-zone-name'});
							$out->print_element('destinationGroup', $policy_set->{'to-zone-name'});
							$out->close_element('appliesTo');
						}
					}
				}
			}

			my $currentProcessOrder = 0;
			my $ps = $policy_set->{policy};

			POLICY_LOOP:
			foreach my $policy (ref($ps) eq 'ARRAY' ? @$ps : ($ps)){
				my $name  = $policy->{name};
				my $match = $policy->{match};
				my $then  = $policy->{then};

				my $action = 'permit';
				if(exists $then->{'accept'}){
					$action = 'permit';
				}elsif(exists $then->{discard}){
					$action = 'drop';
				}elsif(exists $then->{reject}){
					$action = 'deny';
				}elsif(exists $then->{deny}){
					$action = 'deny';
				}

				my $log = (exists $then->{'log'} or exists $then->{syslog}) ? 'true' : 'false';

				my $fe = {
					'log'           => $log,
					'primaryAction' => $action,
					'name'          => $name,
				};

				foreach my $direction('source', 'destination'){
					foreach my $address(@{$match->{"$direction-address"}}){
						if($address eq 'any'){ next; }
						if( $address =~ /$ipv6_regex/ ) {
							$LOGGER->warning($RULE_DROPPED, "IPv6 address '$address' not supported for policy '$name'");
							next POLICY_LOOP;
						}

						my $zone_name = $direction eq 'source' ? $policy_set->{'from-zone-name'} : $policy_set->{'to-zone-name'};
						my $ref = '_ENTRY_' . ncname_from_string($address);
						my $zone_ref = $zone_name . $ref;
						my $global_ref = 'global' . $ref;

						if($in->{parsed}->{object_groups}->{$zone_ref}){
							push(@{ $fe->{$direction . 'IpAddr'} }, { objectGroupReference => $zone_ref });
						}elsif($in->{parsed}->{object_groups}->{$global_ref}){
							push(@{ $fe->{$direction . 'IpAddr'} }, { objectGroupReference => $global_ref });
						}else{
							$LOGGER->warning($RULE_DROPPED, "Object group '$zone_ref' or '$global_ref' not recognised in policy '$name'");
							next POLICY_LOOP;
						}
					}
				}

				my $app_failed = 0;
				foreach my $app (@{$match->{application}}){
					if($app eq 'any'){ next; }

					if($app =~ /^junos-(.+)$/){
						my $port = $1;

						my $portExpressions = get_port_expression($port);
						if($portExpressions){
							foreach my $expression (@$portExpressions){
								push(@{$fe->{destinationService}}, $expression);
							}
						}else{
							$app_failed = 1;
						}
					}elsif($applications->{$app}){
						$fe->{destinationService} = $applications->{$app};
					}
					else {
						$app_failed = 1;
					}
				}

				# Only add (and increment) the processOrder ifthe filterEntry is actually going to be written out.
				# Don't want to skip a processOrder value in the final SED.
				if($app_failed == 0){
					$fe->{processOrder} = ++$currentProcessOrder;
					$out->print_element('filterEntry', $fe);
				}
			}

			$out->print_element('filterEntry', {
					log           => 'false',
					primaryAction => $defaultPolicy,
					processOrder  => ++$currentProcessOrder,
					name          => 'default-policy',
				}
			);

			$out->print_element('mode', 'stateless');
			$out->print_element('name', 'fromZone_' . $policy_set->{'from-zone-name'} . '-toZone_' . $policy_set->{'to-zone-name'});
			$out->close_element('filterList');
		}
	}

	return;
}

sub _parse_applications{
	my ($in, $out) = @_;

	my $arrays = ['application', 'application-set'];
	if($in->{showApplications}){
		my ($appSec) = $in->{showApplications} =~ /(<rpc\-reply[\s|>].+?<\/rpc\-reply>)/s;

		if( $appSec ) {
			my $result = XMLin($appSec, ForceArray => $arrays);
			if($result->{configuration}->{applications}){
				my $appresult = $result->{configuration}->{applications}->{application};
				while (my ($name, $application) = each(%$appresult)){
					$securityApps->{$name} = _parse_application($name, $application);
				}

				while (my ($name, $application_set) = each(%{$result->{configuration}->{applications}->{'application-set'}})){
					my @services;
					foreach my $appname (keys(%{$application_set->{application}})){
						foreach my $lookup (@{$securityApps->{$appname}}){
							push(@services, $lookup);
						}
					}
					$securityApps->{$name} = \@services;
				}
			}
		}
	}
	return($securityApps);
}

sub _parse_application{
	my ($name, $application) = @_;
	my @services;

	my $port = $application->{'destination-port'};
	if($port && $port =~ /(\d+)\-(\d+)/){
		push(@services,	{
			portRange => {
				portStart => $1,
				portEnd   => $2,
				protocol  => $application->{protocol},
			},
		});
	}elsif($port && $port =~ /,/){
		my @ports = split(/,/, $port);
		foreach my $portNo (@ports){
			push(@services,	{
				portExpression => {
					operator => 'eq',
					port     => get_port_number($portNo),
					protocol => $application->{protocol},
				},
			});
		}
	}elsif(defined($port)){
		my $lookup = get_port_number($port);
		if($lookup =~ /^\d+/){
			push(@services,	{
				portExpression => {
					operator => 'eq',
					port     => get_port_number($port),
					protocol => $application->{protocol},
				},
			});
		}
	}elsif($application->{protocol}){

		# Service group cannot be created without a port number so this is a failure of the ZED
		# We'll 'fake' it with the full port range, clients will have to be aware.
		push(@services,	{
			portExpression => {
				operator => 'gt',
				port     => 0,
				protocol => $application->{protocol},
			},
		});
	}
	return \@services;
}

sub parse_object_groups{
	my ($in, $out) = @_;
	my %network_groups = ();

	# Parse each zones address book
	my @arrays = qw(address security-zone interfaces address-set);
	if( $in->{showZones} and $in->{showZones} =~ /rpc-reply/ ) {
		my ($zone_section) = $in->{showZones} =~ /(<rpc\-reply[\s|>].+?<\/rpc\-reply>)/s;
		my $reply = XMLin( $zone_section, ForceArray => \@arrays );
		my $zones = $reply->{configuration}->{security}->{zones}->{'security-zone'};
		for my $zone_name ( keys %{ $zones } ) {
			my $zone_groups = _parse_address_book($zone_name, $zones->{$zone_name}->{'address-book'});
			if( $zone_groups ) {
				%network_groups = ( %network_groups, %{ $zone_groups } );
			}
		}
	}

	# Parse the global address book
	if( $in->{showPolicy} and $in->{showPolicy} =~ /rpc-reply/ ) {
		my ($policy_section) = $in->{showPolicy} =~ /(<rpc\-reply[\s|>].+?<\/rpc\-reply>)/s;
		my $reply = XMLin($policy_section, ForceArray => \@arrays);

		my $global_address_book = $reply->{configuration}->{security}->{'address-book'};
		my $global_groups = _parse_address_book('global', $global_address_book);
		if( $global_groups ) {
			%network_groups = ( %network_groups, %{ $global_groups } );
		}
	}

	# Parse the global prefix lists
	if($in->{active} && $in->{active} =~ /^policy-options/m){
		my $prefix_groups = _parse_prefix_lists( $in->{active} );
		if( $prefix_groups ) {
			%network_groups = ( %network_groups, %{ $prefix_groups } );
		}
	}

	if( %network_groups ) {
		$out->open_element('objectGroups');
		for my $id ( keys %network_groups ) {
			_print_network_group($out, $network_groups{$id});
		}
		$out->close_element('objectGroups');
	}

	$in->{parsed}->{object_groups} = \%network_groups;

	return;
}

sub _parse_address_book{
	my ($name, $address_book) = @_;
	my $addresses = $address_book->{address};
	my $network_groups;

	# Parse single addresses that have been given a name
	for my $address_name ( keys %{ $addresses } ) {
		my $id = $name . '_ENTRY_' . ncname_from_string($address_name);
		my $address = $addresses->{$address_name}->{'ip-prefix'};

		if( $address !~ /$ipv6_regex/ ) {
			my $net = parseCIDR( $address );
			my $host = $net->{host};
			my $network = mask_to_bits( $net->{network} );

			if( $host ){
				my $network_group = { id => $id };

				if( $network eq 32 ){
					$network_group->{host} = $host;
				}else{
					$network_group->{subnet} = {address => $host, mask => $network};
				}

				$network_groups->{$id} = $network_group;
			} else {
				$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Host IP address not found for address '$id'");
			}
		} else {
			$LOGGER->warning($NETWORK_OBJECT_DROPPED, "IPv6 address not supported: '$address'");
		}
	}

	my $count = scalar keys %{ $network_groups };
	$LOGGER->debug("Parsed $count addresses to network groups in address book '$name'");

	# Parse address sets
	my %unresolved;
	my $address_sets = $address_book->{'address-set'};
	if( $address_sets ) {
		for my $group_name ( keys %{ $address_sets } ) {
			my $id = $name . '_ENTRY_' . ncname_from_string($group_name);
			my $network_group = { id => $id };

			my $address_set = $address_sets->{$group_name};
			my @child_names = keys %{ $address_set->{address} };

			for my $child_name ( @child_names ) {
				my $child_id = $name . '_ENTRY_' . ncname_from_string($child_name);

				if( $network_groups->{$child_id} ) {
						push @{ $network_group->{objectGroupReference} }, $child_id;;
				} else {
					push @{ $unresolved{$child_id} }, $id;
				}
			}

			$network_groups->{$id} = $network_group;
		}
	}

	# Any address references dropped eariler because they had not yet been parsed are retrospectively added
	for my $id ( keys %unresolved ) {
		if( $network_groups->{$id} ) {
			for my $set_id ( @{ $unresolved{ $id } } ) {
				$LOGGER->verbose("Retrospectively adding reference '$id' to address set '$set_id'");
				push @{ $network_groups->{$set_id}->{objectGroupReference} }, $id;
			}
		} else {
			my $unresolved_sets = join ', ', @{ $unresolved{ $id } };
			$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Unrecognised address object '$id' in address sets: $unresolved_sets");
		}
	}

	$count -= scalar keys %{ $network_groups }; $count *= -1;
	$LOGGER->debug("Parsed $count address sets to network objects in address book '$name'");

	return $network_groups;
}

# Parses prefix lists defined in the running config
sub _parse_prefix_lists {
	my ($config) = @_;
	my $network_groups;

	my ($policy_options) = $config =~ /(^policy-options\s+.+?^})/sm;
	my @prefix_lists = $policy_options =~ /(^\s+prefix-list\s+\S+\s+{.+?;\s*^\s+})/smg;
	for my $prefix_list ( @prefix_lists ) {
		if( $prefix_list =~ /prefix-list\s+(\S+)/ ) {
			my $id = $1;
			my $network_group = { id => $id };
			while( $prefix_list =~ /^\s+($ip_regex)\/(\d+);\s*$/gm ) {
				my ($address, $mask) = ($1, $2);
				if($mask eq 32){
					push @{ $network_group->{host} }, $address;
				}else{
					push @{ $network_group->{subnet} }, {address => $address, mask => $mask};
				}
			}
			$network_groups->{$id} = $network_group;
		}
	}

	my $count = scalar keys %{ $network_groups };
        $LOGGER->debug("Parsed $count prefix-lists to network objects");

	return $network_groups;
}

sub _print_network_group{
	my ($out, $network_group) = @_;

	$out->open_element('networkGroup');
	$out->print_element('id', $network_group->{id});

	for my $element ( 'host', 'objectGroupReference', 'subnet' ) {
		if(defined $network_group->{$element}){
			$out->print_element($element, $network_group->{$element});
		}
	}

	$out->close_element('networkGroup');

	return;
}

sub _parse_module{
	my ($raw_module) = @_;
	my $module;

	$module->{'core:asset'}->{'core:factoryinfo'}->{'core:make'} = 'Juniper';
	$module->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = 'Unknown';

	if(defined($raw_module->first_child('version'))){
		$module->{'core:asset'}->{'core:factoryinfo'}->{'core:hardwareVersion'} = $raw_module->first_child('version')->text();
	}

	if(defined($raw_module->first_child('part-number'))){
		$module->{'core:asset'}->{'core:factoryinfo'}->{'core:partNumber'} = $raw_module->first_child('part-number')->text();
	}

	if(defined($raw_module->first_child('serial-number'))){
		$module->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = $raw_module->first_child('serial-number')->text();
	}

	if(defined($raw_module->first_child('description'))){
		$module->{'core:description'} = $raw_module->first_child('description')->text();
	}

	return($module);
}

sub _process_filter{
	my ($in, $out, $appliesTo, $filterListHash) = @_;
	return sub {

		# Processes a single firewall filter as an XML::Twig
		my ($twig, $element) = @_;
		my $filterName = _get_element_text($element, 'name');
		if($filterName){
			$out->open_element('filterList');

			my $lists = $appliesTo->{$filterName};
			if($lists && $lists->{in}){
				my $list = $lists->{in};
				foreach my $int (@$list){
					$out->open_element('appliesTo');
					$out->print_element('source', $int);
					$out->close_element('appliesTo');
				}
			}
			if($lists && $lists->{out}){
				my $list = $lists->{out};
				foreach my $int (@$list){
					$out->open_element('appliesTo');
					$out->print_element('destination', $int);
					$out->close_element('appliesTo');
				}
			}
		}

		my $currentProcessOrder = 0;
		foreach my $filterEntryElement ($element->descendants('term')){
			my $filterEntry = {
				name          => _get_element_text($filterEntryElement, 'name'),
				log           => 'false',
				primaryAction => 'permit',
			};

			# Process filter references
			my $filterRef = _get_element_text($filterEntryElement, 'filter');
			if($filterRef){
				my $inlineEntries = $filterListHash->{$filterRef};
				foreach my $inlineEntry (@$inlineEntries){
					$out->print_element('filterEntry', $inlineEntry);
				}
				next;
			}

			# Process the address information
			my $from = $filterEntryElement->first_child('from');
			if($from){
				my $address = $from->first_child('address');
				if($address){
					my $parsed_address = _process_address($address);
					if( $parsed_address ) {
						push @{$filterEntry->{sourceIpAddr}}, $parsed_address;
					}
				}

				for my $direction ( 'destination', 'source' ) {
					my $element = $direction . 'IpAddr';

					my @addresses = $from->children("$direction-address");
					if(@addresses){
						my $parsed_addresses = _process_addresses(@addresses);
						if( $parsed_addresses ) {
							push @{$filterEntry->{$element}}, $parsed_addresses;
						}
					}

					my @ranges = $from->children("$direction-address-range");
					if(@ranges){
						my $parsed_ranges = _process_address_ranges(@ranges);
						if( $parsed_ranges ) {
							push @{$filterEntry->{$element}}, $parsed_ranges;
						}
					}

					my @lists = $from->children("$direction-prefix-list");
					if(@lists){
						my $parsed_lists = _process_address_lists($in, @lists);
						if( $parsed_lists ) {
							push @{$filterEntry->{$element}}, $parsed_lists;
						}
					}
				}

				# Process the port information
				my @ports = $from->children('port');
				if(@ports){
					$filterEntry->{sourceService} = _process_port(@ports);
				}

				my @dstPorts = $from->children('destination-port');
				if(@dstPorts){
					$filterEntry->{destinationService} = _process_port(@dstPorts);
				}

				my @srcPorts = $from->children('source-port');
				if(@srcPorts){
					$filterEntry->{sourceService} = _process_port(@srcPorts)
				}

				my @exceptSrcPorts = $from->children('source-port-except');
				if(@exceptSrcPorts){
					$filterEntry->{sourceService} = _process_port_negate(@exceptSrcPorts);
				}

				my @exceptDestPorts = $from->children('destination-port-except');
				if(@exceptDestPorts){
					$filterEntry->{destinationService} = _process_port_negate(@exceptDestPorts);
				}

				foreach my $prot ($from->children('protocol')){
					my $protocol = $prot->text();

					if($protocol =~ /^\d+$/){
						$protocol = get_protocol_name($protocol);

						if(!$protocol){
							$LOGGER->debug("[WARNING] Unable to resolve protocol number '" . $prot->text() . "'");
							next;
						}
					}

					push(@{$filterEntry->{protocol}}, $protocol);
				}
			}

			my $then = $filterEntryElement->first_child('then');
			if($then){
				if($then->first_child('accept')){
					$filterEntry->{primaryAction} = 'permit';
				}elsif($then->first_child('reject')){
					$filterEntry->{primaryAction} = 'deny';
				}elsif($then->first_child('discard')){
					$filterEntry->{primaryAction} = 'drop';
				}elsif($then->first_child('next')){
					$filterEntry->{primaryAction} = 'next';
				}

				if($then->first_child('log') || $then->first_child('syslog')){
					$filterEntry->{log} = 'true';
				}
			}

			$filterEntry->{processOrder} = ++$currentProcessOrder;

			if($filterName){
				push(@{$filterListHash->{$filterName}}, $filterEntry);
			}
			$out->print_element('filterEntry', $filterEntry);
		}

		if($filterName){
			$out->print_element('mode', 'stateless');
			$out->print_element('name', $filterName);
			$out->close_element('filterList');
		}
	}
}

# If the element is defined this returns the text value of the element.
# If the element is not defined this returns undef;
sub _get_element_text{
	my ($twig, $elementName) = @_;
	my $child = $twig->first_child($elementName);
	if($child){
		return $child->text();
	}else{
		return undef;
	}
}

sub _process_addresses{
	my @twigs = @_;

	my @addresses;
	foreach my $twig (@twigs){
		my $parsed_address = _process_address($twig);
		if( $parsed_address ) {
			push @addresses, $parsed_address;
		}
	}
	return \@addresses;
}

sub _process_address{
	my $twig = shift;

	my $addresses;
	foreach my $child ($twig->children('name')){
		my ($addr, $mask) = split(/\//, $child->text());

		if($addr !~ /$ipv6_regex/){
			if($mask eq 32){
				push(@{$addresses}, {host => $addr});
			}else{
				push(@{$addresses}, {network => {address => $addr, mask => $mask,}});
			}
		} else {
			$LOGGER->warning($RULE_DROPPED, "IPv6 addresses are not supported: $addr");
		}
	}
	return $addresses;
}

sub _process_address_ranges{
	my @twigs = @_;

	my @ranges;
	foreach my $twig (@twigs){
		my ($low) = $twig->children('low');
		$low = $low->text();
		my ($high) = $twig->children('high');
		$high = $high->text();

		if( $low !~ /$ipv6_regex/ and $high !~ /$ipv6_regex/ ) { 
			push @ranges, {range => {startAddress => $low, endAddress => $high}};
		} else {
			$LOGGER->warning($RULE_DROPPED, "IPv6 ranges are not supported: $low - $high");
		}
	}

	return \@ranges;
}

sub _process_address_lists{
	my ($in, @twigs) = @_;

	my @lists;
	for my $twig (@twigs){
		for my $twig_ref ( $twig->children('name') ) {
			my $ref = $twig_ref->text();
			if( $in->{parsed}->{object_groups}->{$ref} ) {
				push @lists, {objectGroupReference => $ref};
			} else {
				$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Reference '$ref' is not known, dropping from rule");
			}
		}
	}

	return \@lists;
}

sub _process_port_negate{
	my $portArray = _process_port(@_);

	foreach my $port (@$portArray){
		$port->{negate} = 'true';
	}

	return $portArray;
}

sub _process_port{
	my $portArray;
	foreach my $port (@_){
		my $p = $port->text();
		if($p =~ /^(\d+)\-(\d+)$/){
			push(@{$portArray}, {portRange => {portStart => $1, portEnd => $2,}});
		}elsif($p =~ /^\d+$/){
			push(@{$portArray}, {portExpression => {operator => 'eq', port => $p,}});
		}elsif($p =~ /\S+/){
			push(@{$portArray}, {portExpression => {operator => 'eq', port => get_port_number($p),}});
		}
	}
	return $portArray;
}

# Using the configuration, populate the ZipTie SNMP model
sub parse_snmp{
	my ( $in, $out ) = @_;
	my $snmp = {};

	if($in->{showMibSystem}	&& $in->{showMibSystem} =~ /^sysName\.0\s+=\s*(.+?)\s*$/mi){
		$snmp->{sysName} = $1;
	} else {
		$snmp->{sysName} = '';
	}

	my ($snmpSection) = $in->{snmp} =~ /(<snmp>.+?<\/snmp>)/s;

	if($snmpSection){
		my $twig = XML::Twig->new();
		$twig->parse($snmpSection);
		undef $snmpSection;

		my $root = $twig->root;
		if($root->first_child('contact')){
			$snmp->{sysContact} = $root->first_child('contact')->text();
		}

		if($root->first_child('location')){
			$snmp->{sysLocation} = $root->first_child('location')->text();
		}

		foreach my $communityElement ($twig->descendants('community')){
			if($communityElement->first_child('name')){
				my $communityHash = {
					communityString => $communityElement->first_child('name')->text(),
					accessType => 'RO',
				};

				if($communityElement->first_child('authorization')){
					if($communityElement->first_child('authorization')->text() eq 'read-write'){
						$communityHash->{accessType} = 'RW';
					}
				}

				foreach my $clientElement ($communityElement->descendants('clients')){
					my $fullClient = $clientElement->first_child('name')->text();

					if($fullClient =~ /(\d+\.\d+\.\d+\.\d+)\/(\d+)/mi){
						my $address = {network => {address => $1, mask => $2,}};
						push(@{$communityHash->{embeddedAccessFilter}}, $address);
					}
				}

				push(@{$snmp->{community}}, $communityHash);
			}
		}

		foreach my $trapGroupElement ($twig->descendants('trap-group')){
			foreach my $trapHostElement ($trapGroupElement->descendants('targets')){
				my $trapHost = {ipAddress => $trapHostElement->first_child('name')->text(),};
				push(@{$snmp->{trapHosts}}, $trapHost);
			}
		}
	}

	$out->print_element('snmp', $snmp);

	return;
}

# TODO JAK - This isn't used, it needs re written
# The command it is parsing shows any ingress or egress firewall filters applied to an interface
sub _get_interface_filters{
	my ($in) = @_;

	my ($payload) = $in->{interfacesFilters} =~ /(<rpc-reply.+?<\/rpc-reply>)/sm;

	my $arrays = ['physical-interface', 'filter-information', 'logical-interface'];
	my $reply = XMLin($payload, ForceArray => $arrays, KeyAttr => []);
	my $ifi = $reply->{'interface-filter-information'};

	unless(defined($ifi->{'physical-interface'})){
		return({});
	}

	my $filters;
	foreach my $physical_interface(@{$ifi->{'physical-interface'}}){
		next unless (defined($physical_interface->{'logical-interface'}));

		foreach my $logical_interface (@{$physical_interface->{'logical-interface'}}){
			next unless (defined( $logical_interface->{'filter-information'}));

			my $name = $logical_interface->{name};
			foreach my $filter_information (@{$logical_interface->{'filter-information'}}){
				next unless (defined($filter_information->{'filter-family'}) and $filter_information->{'filter-family'} eq 'inet');

				$filters->{$name}->{out} = $filter_information->{'filter-output'};
				$filters->{$name}->{in} = $filter_information->{'filter-input'};
			}
		}
	}
	return($filters);
}

# Given a TWIG of the 'show ospf interfaces detail' command and the name
# of a real interface, parse out the ospf details for the specific interface
#
# This code was orignally called like this:
#######################################################################################################################
# Check to see if the response from the "show ospf interface detail | display xml" command contains valid info for OSPF
# my ($ospfIntSection) = $in->{showOspfInterface} =~ /(<ospf-interface-information.+?<\/ospf-interface-information>)/s;

# If any OSPF interface information section was found, the create an XML::Twig object to store it
# my $ospfInts = undef;
# if($ospfIntSection) {
#	 $ospfInts = XML::Twig->new();
#	 $ospfInts->parse($ospfIntSection);
#	 undef $ospfIntSection;
# }
# ...
# my $ospfInterface = _get_ospf_interface($physInterface->{name}, $ospfInts);
# if($ospfInterface){
# 	$physInterface->{interfaceOspf} = $ospfInterface;
# }
#
# Was done the same for logical interfaces too
######################################################################################################################
sub _get_ospf_interface{
	my ($name, $twig) = @_;

	# Only continue ifthe XML::Twig object representing the OSPF interface information is valid
	if(defined($twig)){
		foreach my $int ($twig->descendants('ospf-interface')){
			my $ospfIntName = $int->first_child('interface-name')->text();
			if($name eq $ospfIntName){
				my $routerPriority;
				if($int->first_child('router-priority')){
					$routerPriority = $int->first_child('router-priority')->text();
				}

				my $ospfInt = {
					area               => $int->first_child('ospf-area')->text(),
					networkType        => _normalize_ospf_type($int->first_child('interface-type')->text()),
					routerState        => $int->first_child('ospf-interface-state')->text(),
					helloInterval      => $int->first_child('hello-interval')->text(),
					cost               => $int->first_child('interface-cost')->text(),
					retransmitInterval => $int->first_child('retransmit-interval')->text(),
				};

				if($routerPriority){
					$ospfInt->{routerPriority} = $routerPriority;
				}
				return($ospfInt);
			}
		}
	}
	return(0);
}

# Only used by the unused sub above
sub _normalize_ospf_type{
	my $networkType = shift;
	for ($networkType){
		return
		    /LAN/i    ? "BROADCAST"
		  : /NBMA/i   ? "NON_BROADCAST"
		  : /P2P/i    ? "POINT_TO_POINT"
		  : /P2MP/i   ? "POINT_TO_MULTIPOINT"
		  : /Virtual/ ? "VIRTUAL_LINK"
		  :             $networkType;
	}
}

sub _get_patterns{
	return({
#		Router => '^(acx|ctp|e|jcs|ln|m|ptx|srx|t)',
		Switch => '^(ex|qfx|wl)',
		IDP => '^(idp)',
	});
}

# Given the name of a juniper configuration section, returns an array of each section
sub _get_section{
	my ($section, $input) = @_;
	my @results;

	while ($input =~ /^((\s+)$section[^\n]*?{(.+?)^\2})/msig){
		push(@results, $1);
	}

	return(\@results);
}

1;

__END__

=head1 Parsers

ZipTie::Adapters::Juniper::JUNOS::Parsers

=head1 SYNOPSIS

	use ZipTie::Adapters::Juniper::JUNOS::Parsers;
	parse_filters( $xmlPrinter, $cliResponses );

=head1 DESCRIPTION

Module with many static methods that take in a hash of
JUNOS device responses and print out ZipTie model elements.

Ideally the methods in this module will deal with XML output
from the JUNOS devices.  From the command line you can get XML
formatted output using "<command> | display xml".  See the C<parse_interfaces>
method for an example of this.

=head2 Methods

=over 12

=item C<create_config>

Uses the Juniper JUNOS active and candidate configs to put together a ZipTie
configurationRepository element.

=item C<parse_local_accounts>

Uses the Juniper JUNOS configuration to put together the ZipTie
localAccounts portion of the model.

=item C<clear_parsing_state>

Clear all state and structures latched over the course of a complete configuration parsing.
Should be called at the end of a backup operation.

=item C<parse_chassis>

Given the JUNOS output from 'show system memory' and 'show host os' parses out
some chassis level details from the core ZipTie model

=item C<parse_filters>

Uses the Juniper JUNOS configuration to put together the ZipTie
localAccounts portion of the model.

=item C<parse_routing>

Parses out global BGP and OSPF parameters.

=item C<parse_snmp>

Uses the Juniper JUNOS configuration to put together the ZipTie
SNMP portion of the model.

=item C<parse_system>

Given the JUNOS output from 'show version' and 'show host name' parses out
some top level core attributes of the ZipTie model ZiptieElementDocument

=back

=head1 LICENSE

  The contents of this file are subject to the Mozilla Public License
  Version 1.1 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/

  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
  License for the specific language governing rights and limitations
  under the License.

  The Original Code is Ziptie Client Framework.

  The Initial Developer of the Original Code is AlterPoint.
  Portions created by AlterPoint are Copyright (C) 2006,
  AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

  Contributor(s): rkruse
  Date: Apr 23, 2007

G=cut
