package ZipTie::Adapters::Juniper::NSM;

use strict;

use Cwd;

use XML::Simple qw(XMLin);

use ZipTie::Adapters::Juniper::NSM::API;
use ZipTie::Adapters::Juniper::NSM::AutoLogin;
use ZipTie::Adapters::Juniper::NSM::Parsers qw(create_config parse_chassis parse_object_groups parse_filters parse_system parse_interfaces);
use ZipTie::Adapters::Juniper::NSM::ServerMappings qw(get_nsm_address_for_ip add_nsm_mappings);
use ZipTie::Adapters::Juniper::NSM::NsmElf qw(get_array get_attack_group_refs);

use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::Adapters::GenericAdapter;
use ZipTie::CLIProtocol;
use ZipTie::CLIProtocolFactory;
use ZipTie::ConnectionPath;
use ZipTie::Credentials;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::SNMP;
use ZipTie::SnmpSessionFactory;
use ZipTie::Typer;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

sub backup
{
	my $packageName = shift;
	my $backupDoc   = shift;    # how to backup this device
	my $responses    = {};       # will contain device responses to be handed to the Parsers module

	# Translate the backup operation XML document into ZipTie::ConnectionPath
	my ($connection_path) = ZipTie::Typer::translate_document( $backupDoc, 'connectionPath' );

    my $admin_ip = $connection_path->get_ip_address();
    my $credentials = $connection_path->get_credentials();
    my $username  = $credentials->{username};
    my $password  = $credentials->{password};

	# Set up the XmlPrint object for printing the ZiptieElementDocument (ZED)
	my $filehandle = get_model_filehandle( 'Juniper NSM', $admin_ip );
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

    my $address = _get_nsm_address($admin_ip);
    my $api = ZipTie::Adapters::Juniper::NSM::API->new( server => $address );
    $api->login($username, $password);

    $responses->{xml} = $api->get_device_info($admin_ip);

    $responses->{device} = XMLin($responses->{xml}, ForceArray => ['_1', 'interfaces',]);

	parse_system( $responses, $printer );
	parse_chassis( $responses, $printer );
	create_config( $responses, $printer );

    my $policy = $api->get_object_by_ref($responses->{device}->{'nsmpolicy-id'});

    if ($responses->{device}->{header}->{platform} =~ /IDP/)
    {
        $responses->{idp_policy} = $api->get_object_by_ref($policy->{rulebases}->{idp});
    }
    elsif (defined $policy->{rulebases}->{firewall})
    {
        $responses->{fw_policy} = $api->get_object_by_ref($policy->{rulebases}->{firewall});
    }

    undef $policy;

    my $callback = sub
    {
        my $value = shift or die;

        return if ($value =~ /^[^&]/);

        $responses->{addresses}->{$value} = undef;
    };

    my @attack_group_refs;
    foreach my $rule (get_array($responses, 'idp_policy', 'rules_collection', 'rules'))
    {
        _handle_array($callback, $rule, 'src_addr_collection', 'src_addr');
        _handle_array($callback, $rule, 'dst_addr_collection', 'dst_addr');

        push(@attack_group_refs, get_attack_group_refs($rule));
    }

    my $attack_groups = $api->get_object_by_ref(\@attack_group_refs);
    $responses->{objects} = $attack_groups;

    foreach my $attack_grp (values %$attack_groups)
    {
        foreach my $member (get_array($attack_grp, 'members_collection', 'members'))
        {
            $responses->{attacks}->{$member} = undef;
        }
    }

    my @addr_keys = keys %{$responses->{addresses}};
    $responses->{addresses} = $api->get_object_by_ref(\@addr_keys);

    my @keys = keys(%{$responses->{attacks}});
    my $attacks = $api->get_object_by_ref(\@keys, '<attack><name_ /></attack>');

    while (my ($attack_id, $attack) = each(%$attacks))
    {
        $responses->{attacks}->{$attack_id} = $attack->{name_};
    }

    $responses->{policy_name} = $responses->{idp_policy}->{name_};

	parse_object_groups($responses, $printer);
	parse_filters($responses, $printer);

    delete $responses->{idp_policy};
    delete $responses->{fw_policy};
    delete $responses->{attacks};
    delete $responses->{addresses};
    delete $responses->{objects};

	parse_interfaces( $responses, $printer );

    $api->logout();

	$printer->close_model();                # close out the ZiptieElementDocument
	close_model_filehandle($filehandle);    # Make sure to close the model file handle
}

sub _handle_array
{
    my $callback = shift or die;
    my $obj = shift or die;

    my @array = get_array($obj, @_);
    foreach my $entry (@array)
    {
        &$callback($entry);
    }    
}

sub _get_nsm_address
{
    my ($admin_ip) = @_;

    my $addr = get_nsm_address_for_ip($admin_ip);
    unless ($addr)
    {
        $LOGGER->fatal("No NSM is configured for the device at '$admin_ip'.");
    }

    return $addr;
}


1;
