package ZipTie::Adapters::Juniper::NSM::API;

use strict;
use warnings;

#use SOAP::Lite +trace => 'all';
use SOAP::Lite;
use MIME::Base64 qw(encode_base64);

use XML::Simple qw(XMLin);

use ZipTie::Logger;

use constant DEBUG => 0;

# The old NSM server in our lab fails to switch to SSL v3 after receiving an error with SSL v2/v3.
#
#    SSL_connect:SSLv2/v3 write client hello A
#    SSL3 alert read:fatal:unexpected_message
#    SSL_connect:error in SSLv2/v3 read server hello A
#    SSL_connect:before/connect initialization
#
# We addressed this issue by forcing SSL v3.
$ENV{HTTPS_VERSION} = 3;

delete $ENV{'https_proxy'};

my $LOGGER = ZipTie::Logger::get_logger();

sub new
{
    my ( $proto, %params ) = @_;

    my $package = ref($proto) || $proto;

    my $self = {};

    foreach my $key ( keys %params )
    {
        $self->{$key} = $params{$key};
    }

    bless($self, $package);
}

sub soap_call
{
    my ($self, $service, $method, $input) = @_;

    my $server = $self->{server};

    my $soap_service = SOAP::Lite
            ->proxy("https://${server}/axis2/services/$service")
            ->uri("http://juniper.net/webproxy/" . lc($service))
            ->ns('http://juniper.net/core', 'core')
            ->ns('http://juniper.net/nbiservice/datacentricservice', 'dat')
        ;

    my $soap_method = SOAP::Data->name($method)->attr( {'xmlns' => "http://juniper.net/webproxy/" . lc($service)} );

    $LOGGER->debug("SOAP call: $service#$method");

    # Execute and grab response
    my $response;
    if ( defined $input )
     {
        $response = $soap_service->call($soap_method, @{$input});
    }
    else
    {
        $response = $soap_service->call($soap_method);
    }

    if ($response && $response->fault)
    {
        $LOGGER->fatal("$service#$method: " . $response->faultcode . ": " . $response->faultstring . "\n");
    }

    return $response;
}

sub login
{
    my ($self, $username, $password) = @_;

    my @apiLogin = (
        SOAP::Data->name('domainName')->value('global'),
        SOAP::Data->name('userName')->value($username),
        SOAP::Data->name('password')->value($password),
    );

    my $response = $self->soap_call("SystemService","LoginRequest",\@apiLogin);
    my $loginStatus = $response->valueof("//Body/LoginResponse/loginStatus/status");
    if ($loginStatus eq "Success" )
    {
        $self->{token} = $response->valueof("//Body/LoginResponse/authToken/Token");
    }
    else
    {
        $LOGGER->fatal_error_code( $INVALID_CREDENTIALS, $self->{server}, $loginStatus );
    }
}

sub logout
{
    my ($self) = @_;

    return $self->soap_call("SystemService","LogoutRequest", [$self->_token()]);
}

sub _get_view_objects
{
    my ($self, $category, $filter, $domain_id) = @_;

    $domain_id = 1 unless (defined $domain_id);

    my @args = (
        $self->_token()
            ,
        SOAP::Data
            ->prefix('dat')
            ->name('category')
            ->value($category)
            ,
        SOAP::Data
            ->prefix('dat')
            ->name('domainId')
            ->value($domain_id)
            ,
    );

    if ($filter)
    {
        my $obj_filter;

        if ($filter eq 'metadata')
        {
            $obj_filter = SOAP::Data
                    ->prefix('dat')
                    ->name('metadataOnly')
                    ->value('true')
                ;     
        }
        else
        {
            my $encoded = encode_base64($filter);
            $obj_filter = SOAP::Data
                    ->prefix('dat')
                    ->name('filter')
                    ->type('core:OpaqueDataType')
                    ->value($encoded)
                ;
        }

        push(@args, SOAP::Data
                ->prefix('dat')
                ->name('objectFilter')
                ->type('dat:ViewFilterType')
                ->value(\$obj_filter)
        );
    }

    my $envelope = $self->soap_call("DataCentricService", "GetObjectViewByCategoryRequest", \@args);
    return $envelope->valueof('//Body/GetObjectViewByCategoryResponse/object');
}

sub _create_filter
{
    my ($self, $filter) = @_;

    my $encoded = encode_base64($filter);
    my $obj_filter = SOAP::Data
            ->prefix('dat')
            ->name('filter')
            ->type('core:OpaqueDataType')
            ->value($encoded)
        ;

    return SOAP::Data
            ->prefix('dat')
            ->name('objectFilter')
            ->type('dat:ViewFilterType')
            ->value(\$obj_filter)
        ;
}

sub get_all_devices
{
    my ($self) = @_;

    my $filter = '<deviceobj xmlns="http://juniper.net/nsm/ADMSchema">'
            . " <header>"
            . " <ip/>"
            . " <platform />"
            . " </header>"
            . " <startup>"
            . " <devicefamily />"
            . " </startup>"
            . " </deviceobj>";

    my @obj_xmls = $self->_get_view_objects('deviceobj', $filter);

    my @devices;

    foreach my $obj (@obj_xmls)
    {
        my $xml = $obj->{objectData}->{data}->{Include};
        my $id = $obj->{objectMetadata}->{objectId};
        
        my ($ip) = $xml =~ /<ip>(.+?)<\/ip>/s;
        my ($platform) = $xml =~ /<platform>(.+?)<\/platform>/s;
        my ($family) = $xml =~ /<devicefamily>(.+?)<\/devicefamily>/s;

        push(@devices, {
            id => $id,
            ip => $ip,
            family => $family,
            platform => $platform,
        });
    }

    return @devices;
}

sub get_object_by_ref
{
    my ($self, $ref, $metadata) = @_;

    return if (ref($ref) eq 'ARRAY' && !@$ref);

    $LOGGER->debug("get_object_by_ref: ($ref)");
    my @ret = $self->_get_object_by_id($ref, $metadata);


    my $result;

    foreach my $obj (@ret) 
    {
        my $id = $obj->{objectMetadata}->{objectId};
        my $cat = $obj->{objectMetadata}->{category};
        my $domain_id = $obj->{objectMetadata}->{domainId};
        
        my $refid = "&$domain_id.$cat.$id";
        if ($metadata and $metadata eq '1')
        {
            $result->{$refid} = $obj;
        }
        else
        {
            my $xml = $obj->{objectData}->{data}->{Include};
            my $parsed = XMLin($xml, ForceArray => ['rules', qr/_list$/, 'members', qr/_addr$/,]);
            $result->{$refid} = $parsed;
        }
    }

    return (ref($ref) eq 'ARRAY') ? $result : $result->{$ref};

}

sub _get_device_id
{
    my ($self, $ip) = @_;
   
    my @all = $self->get_all_devices();
    foreach my $meta (@all)
    {
        return $meta->{id} if ($meta->{ip} eq $ip);
    }

    $LOGGER->fatal("No device in NSM with the IP '$ip'");
}

sub _get_object_by_id
{
    my ($self, $refs, $metadata) = @_;

    my @args = (
        $self->_token()
        ,
    );

    my $last_category;

    foreach my $ref ((ref($refs) eq 'ARRAY') ? @$refs : ($refs))
    {
        my ($domain, $cat, $id) = $ref =~ /&(\d+)\.([^.]+)\.(\d+)/;
        $last_category = $cat;

        push(@args, SOAP::Data
            ->prefix('dat')
            ->name('objectIdentifier')
            ->type('core:ObjectIdentifierType')
            ->value([
                SOAP::Data
                    ->prefix('core')
                    ->name('domainId')
                    ->value($domain)
                    ,
                SOAP::Data
                    ->prefix('core')
                    ->name('category')
                    ->value($cat)
                    ,
                SOAP::Data
                    ->prefix('core')
                    ->name('objectIdOrName')
                    ->type('core:ObjectIdOrNameType')
                    ->value([
                        SOAP::Data
                            ->prefix('core')
                            ->name('objectId')
                            ->value($id)
                            ,
                    ])
                    ,
            ])
        );
    }

    if ($metadata)
    {
        my $obj_filter;

        if ($metadata ne '1')
        {
            my $encoded = encode_base64($metadata);
            $obj_filter = SOAP::Data
                    ->prefix('dat')
                    ->name('filter')
                    ->type('core:OpaqueDataType')
                    ->value($encoded)
                ;
        }
        else
        {
            $obj_filter = SOAP::Data
                    ->prefix('dat')
                    ->name('metadataOnly')
                    ->value('true')
                ;                 
        }


        my $view_filter = SOAP::Data
                ->prefix('dat')
                ->name('objectFilter')
                ->type('dat:ObjectViewFilterType')
                ->value([
                    SOAP::Data
                        ->prefix('dat')
                        ->name('category')
                        ->value($last_category)
                        ,
                    SOAP::Data
                        ->prefix('dat')
                        ->name('viewFilter')
                        ->type('dat:ViewFilterType')
                        ->value(\$obj_filter)
                        ,
                ]);
        push(@args, $view_filter);
    }

    my $envelope = $self->soap_call("DataCentricService", "GetObjectViewByIdRequest", \@args);
    my $status = $envelope->valueof('//Body/GetObjectViewByIdResponse/Status');
    if ($status ne 'Success')
    {
        my $error = $envelope->valueof('//Body/GetObjectViewByIdResponse/Errors');
        $LOGGER->fatal($error->{ErrorNumber} . ':' . $error->{ErrorMessage} . "\n" . $error->{ErrorDetails});
    }

    return $envelope->valueof('//Body/GetObjectViewByIdResponse/object');
}

sub get_device_info
{
    my ($self, $admin_ip) = @_;

    my $id = $self->_get_device_id($admin_ip);
    my $obj = $self->_get_object_by_id('&1.deviceobj.' . $id);

    my $include = $obj->{objectData}->{data}->{Include};
    return $include;
}

sub _token
{
    my ($self) = @_;

    my $token = SOAP::Data
            ->prefix('core')
            ->name('Token')
            ->value($self->{token})
        ;

    return SOAP::Data
            ->prefix('core')
            ->name('AuthToken')
            ->type('core:AuthTokenType')
            ->value(\$token)
        ;
}

1;
