package ZipTie::Adapters::Juniper::NSM::Parsers;

use strict;
use warnings;

use ZipTie::Adapters::Juniper::NSM::NsmElf qw(get_array get_attack_group_refs);
use ZipTie::Adapters::Utils qw(seconds_since_epoch get_mask get_port_number trim get_interface_type);
use MIME::Base64 'encode_base64';

use ZipTie::Logger;

use Exporter 'import';
our @EXPORT_OK =
  qw(create_config parse_chassis parse_object_groups parse_filters parse_system parse_interfaces);


my $LOGGER = ZipTie::Logger::get_logger();

sub parse_chassis
{
	my ( $in, $out ) = @_;
	my $chassis;

    my $platform = $in->{device}->{header}->{platform};
    my $serial = $in->{device}->{header}->{serial};
    $serial =~ s/^0+//;

	$chassis->{'core:asset'}->{'core:assetType'} = 'Chassis';

	$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:make'}        = 'Juniper';
	$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $platform;
	$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = $serial;
	
	$out->print_element( "chassis", $chassis );
}

sub parse_system
{
	my ( $in, $out ) = @_;

    my $os_ver = $in->{device}->{header}->{'os-version'};
    my $osfamily = $in->{device}->{startup}->{devicefamily};

	$out->print_element( 'core:systemName', $in->{device}->{name_} );
	
	$out->open_element('core:osInfo');
	$out->print_element( 'core:make',    'Juniper' );
	$out->print_element( 'core:version', $os_ver );
	$out->print_element( 'core:osType',  $osfamily);
	$out->close_element('core:osInfo');

    my $platform = $in->{device}->{header}->{platform};

    my $type = 'Unknown';
    if ($platform =~ /(NS-IDP|ns)/)
    {
    	$type = 'IDS/IPS';
    }
    elsif ($platform =~ /(SRX|M\/MX)/)
    {
        $type = 'Router';
    }
    elsif ($platform =~ /EX-/)
    {
        $type = 'Switch';
    }
    elsif ($platform =~ /(SSG|ISG)/)
    {
        $type = 'Firewall';
    }
    elsif ($platform =~ /SA/)
    {
        $type = 'VPN Concentrator';
    }

    $out->print_element( 'core:deviceType', $type );

	$out->print_element('core:lastReboot', 0);
}

sub create_config
{
	my ( $in, $out ) = @_;

	# the name of the repository
	my $repository;
	$repository->{'core:name'} = '/';

	# build the simple text configuration
	my $config;
	$config->{'core:name'}       = 'deviceobj';
	$config->{'core:textBlob'}   = encode_base64( $in->{xml} );
	$config->{'core:mediaType'}  = 'text/plain';
	$config->{'core:context'}    = 'active';
	$config->{'core:promotable'} = 'false';

	# push the configuration into the repository
	push( @{ $repository->{'core:config'} }, $config );

	# print the repository
	$out->print_element( 'core:configRepository', $repository );
}

sub parse_local_accounts
{
	my ( $in, $out ) = @_;
}

sub parse_object_groups
{
    my ($in, $out) = @_;

    $out->open_element('objectGroups');
    if ($in->{addresses})
    {
        while ( my ($id, $addr) = each(%{$in->{addresses}}) ) 
        {
            if ($addr->{type} eq 'network')
            {
                my $addr_array = $addr->{address}->{address};

                my @subnets;
                
                foreach my $a (ref($addr_array) eq 'ARRAY' ? @$addr_array : ($addr_array))
                {
                    my $subnet = $a->{subnet};
                    push(@subnets, {
                        address => $subnet->{ip},
                        mask => $subnet->{netmask},
                    });
                }

                $id =~ s/^&//;

                # prefix id with 'a' so that it is a valid NCName.
                $id = 'a' . $id;
                
                my $ng = {
                    description => $addr->{name_},
                    id => $id,
                    subnet => \@subnets,
                };

                $out->print_element('networkGroup', $ng);
            }
        }
    }

    $out->close_element('objectGroups');
}

sub parse_filters
{
	my ( $in, $out ) = @_;

    my $platform = $in->{device}->{header}->{platform};
    if ($platform =~ /IDP/)
    {
        _parse_idp_filters($in, $out);
    }
}

sub _parse_idp_filters
{
    my ($in, $out) = @_;

    my $policy = $in->{idp_policy};

    $out->open_element('filterLists');
    $out->open_element('filterList');
    my $processOrder = 1;  # starts at 1

    foreach my $rule (@{$policy->{rules_collection}->{rules}})
    {
        if (defined $rule->{ape_rule} and $rule->{ape_rule} eq 'true')
        {
            # skip application policy enforcement rule for now.
            next;
        }

        $out->open_element('filterEntry');
        my $log = 'false';
        if (defined $rule->{'log'} and defined $rule->{'log'}->{enabled})
        {
            $log = 'true';
        }

#        qw(none ignore drop-packet drop-connection close-client close-server closeClientAndServer);
        my $action = $rule->{action};
        if ($action =~ /close/)
        {
            $action = 'deny';
        }
        elsif ($action =~ /drop/)
        {
            $action = 'drop';
        }
        else
        {
        	$action = 'permit';
        }

        my @srcs = get_array($rule, 'src_addr_collection', 'src_addr');
        my @dsts = get_array($rule, 'dst_addr_collection', 'dst_addr');

        _parse_addresses($in, $out, 'sourceIpAddr', $rule, \@srcs);
        _parse_addresses($in, $out, 'destinationIpAddr', $rule, \@dsts);

        $LOGGER->debug("Parsing attack groups...");

        foreach my $refid (get_attack_group_refs($rule))
        {
            my $attack_grp = $in->{objects}->{$refid};
            if( not $attack_grp )
            {
            	$LOGGER->debug("undef'd attack_grp");
            	next;
            }
            foreach my $member (get_array($attack_grp, 'members_collection', 'members'))
            {
                my $vulnid = $in->{attacks}->{$member};

                $out->print_element('vulnerability', {
                            'name' => $vulnid,
                            'type' => 'NetScreenIDP',
                    });
            }
        }


        $out->print_element('log', $log);
        $out->print_element('primaryAction', $action);
        $out->print_element('processOrder', $processOrder);
        $out->close_element('filterEntry');
        $processOrder++;
    }

    $out->print_element('mode', 'stateful');
    $out->print_element('name', $policy->{name_});
    $out->close_element('filterList');
    $out->close_element('filterLists');
}

sub _parse_addresses
{
    my ($in, $out, $tag, $rule, $addrs) = @_;


    foreach my $addr (@$addrs)
    {
        if ($addr =~ s/^&//)
        {
            $out->print_element($tag, {objectGroupReference => 'a' . $addr});
        }
    }
}

sub parse_interfaces
{
	my ( $in, $out ) = @_;

    $out->open_element('interfaces');
    if (exists $in->{device}->{interface_collection})
    {
        _parse_standard_interfaces($in, $out, $in->{device}->{interface_collection});
    }
    elsif (exists $in->{device}->{idpobj})
    {
        _parse_idp_interfaces($in, $out, $in->{device}->{idpobj}->{interfaces_collection});
    }
    else
    {
        _print_admin_ip($in, $out);
    }

	$out->close_element('interfaces');
}

sub _print_admin_ip
{
    my ($in, $out) = @_;

    my $ip = $in->{device}->{header}->{ip};

    my $iface = {
        interfaceIp => {
            ipConfiguration => {
                ipAddress => $ip,
            },
        },
        interfaceType => 'ethernet',
        name => 'admin',
        physical => 'true',
    };
    $out->print_element('interface', $iface);
}

sub _parse_standard_interfaces
{
    my ($in, $out, $iface_collection) = @_;

    if (defined $iface_collection->{'_1'})
    {
        foreach my $interface (@{$iface_collection->{'_1'}})
        {
            next unless (exists $interface->{ip});

            my $ip = $interface->{ip}->{numbered}->{'ip-address'};
            my $mask = $interface->{ip}->{numbered}->{netmask};
           
            my $iface = {
                interfaceIp => {
                    ipConfiguration => {
                        ipAddress => $ip,
                        mask => $mask,
                    },
                },
                interfaceType => 'ethernet',
                name => $interface->{name_},
                physical => 'true',
            };
            $out->print_element('interface', $iface);
        }
    }
}

sub _parse_idp_interfaces
{
    my ($in, $out, $iface_collection) = @_;

    if (defined $iface_collection->{interfaces})
    {
        foreach my $interface (@{$iface_collection->{interfaces}})
        {
            my $mac = $interface->{mac};
            $mac =~ s/://g;
            my $iface = {
                interfaceEthernet => {
                    macAddress => $mac,
                },
                interfaceType => 'ethernet',
                name => $interface->{name_},
                physical => 'true',
                ingressFilter => $in->{policy_name},
            };
            $out->print_element('interface', $iface);
        }
    }

    _print_admin_ip($in, $out);
}

1;
