#!/usr/bin/perl
# Connects to an NSM server and attempts to add all of the devices in the 
# inventory to the ziptie inventory as Juniper NSM devices.

use strict;
use warnings;

use ZipTie::Adapters::Juniper::NSM::API;
use ZipTie::Adapters::Juniper::NSM::ServerMappings qw(add_nsm_mappings);
use ZipTie::Client;

my ($server, $username, $password, $ziptie_user, $ziptie_password) = @ARGV;

unless ($server =~ /:\d+$/)
{
    $server = $server . ':8443';
}

my $api = ZipTie::Adapters::Juniper::NSM::API->new( server => $server );
$api->login($username, $password);

my @devices = $api->get_all_devices();

$api->logout();
delete $ENV{HTTPS_VERSION};

my $ziptie;
if ($ziptie_user)
{
    $ziptie = ZipTie::Client->new(username => $ziptie_user, password => $ziptie_password);
}
else
{
    $ziptie = ZipTie::Client->new();
}
my $device_service = $ziptie->devices();

my @added;

foreach my $deviceMeta (@devices)
{
    my $ip = $deviceMeta->{ip};
    my $platform = $deviceMeta->{platform};
    my $family = $deviceMeta->{family};

    if (!defined($ip))
    {
        print(",Skipped device with no IP defined.\n");
        next;
    }

    if (!defined($platform))
    {
        print("$ip,Skipped.  Unable to determine device platform.\n");
        next;
    }

    my $device = $device_service->getDevice(ipAddress => $ip, managedNetwork => 'Default');
    if ($device)
    {
        print("$ip,Skipped.  Device already in inventory.\n");
    }
    else
    {
        # IDPs often have 'screenos' for the family so first we check if the platform has IDP in it. 
        if ($platform =~ /IDP/)
        {
            $device_service->createDevice(
                ipAddress => $ip,
                adapterId => 'ZipTie::Adapters::Juniper::NSM',
                managedNetwork => 'Default'
            );

            push(@added, $ip);
            print("$ip,Added to inventory.\n");
        }
        elsif (defined($family) && ($family eq 'screenos'))
        {
           $device_service->createDevice(
                ipAddress => $ip,
                adapterId => 'ZipTie::Adapters::Juniper::ScreenOS',
                managedNetwork => 'Default'
            );

            print("$ip,Added to inventory.\n");
        }
        elsif (defined($family) && ($family =~ /junos/))
        {
            $device_service->createDevice(
                ipAddress => $ip,
                adapterId => 'ZipTie::Adapters::Juniper::JUNOS',
                managedNetwork => 'Default'
            );

            print("$ip,Added to inventory.\n");
        }
        else
        {
            print("$ip,Unsupported platform: $platform\n");
        }
    }
}

if (@added)
{
    add_nsm_mappings($server, \@added);
}
