package ZipTie::Adapters::Juniper::ScreenOS;

use 5.010_000;
use strict;
use warnings;

use ZipTie::CLIProtocolFactory;
use ZipTie::CLIProtocol;
use ZipTie::ConnectionPath;
use ZipTie::Adapters::Juniper::ScreenOS::AutoLogin;
use ZipTie::Adapters::Juniper::ScreenOS::Disconnect qw(disconnect);
use ZipTie::Adapters::Juniper::ScreenOS::Parsers qw(create_config parse_local_accounts parse_chassis parse_filters parse_snmp parse_object_groups parse_system parse_interfaces parse_static_routes parse_services);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle ncname_from_string);
use ZipTie::Model::XmlPrint;
use ZipTie::Logger;
use ZipTie::Typer;

use Exporter 'import';
our @EXPORT_OK = qw(clean_responses);

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

# Specifies that this adapter is a subclass of ZipTie::Adapters::BaseAdapter
use ZipTie::Adapters::BaseAdapter;
our @ISA = qw(ZipTie::Adapters::BaseAdapter);

sub backup {
	my ( $package_name, $backup_doc ) = @_;

	# Translate the backup operation XML document into ZipTie::ConnectionPath
	my ( $connection_path, $credentials ) = ZipTie::Typer::translate_document( $backup_doc, 'connectionPath' );
	my ( $cli_protocol, $prompt_regex ) = _connect( $connection_path, $credentials );

	# Grab an output filehandle for the model. This usually points to STDOUT
	my $filehandle = get_model_filehandle( 'ScreenOS', $connection_path->get_ip_address() );

	# Initialize the model printer
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

	# Begin executing commands on the device. The results of each command will
	# be stored in a single hashtable ($parsers) and fed into each parsing method
	my $responses = {};

	# Attempt to disable paging - default is off
	$cli_protocol->send_and_wait_for( 'set console page 0', $prompt_regex );
	
	$responses = get_command_responses( $cli_protocol, $prompt_regex );
	
	#clean responses
	clean_responses ($responses);
	
	parse_system( $responses, $printer );
	parse_chassis( $responses, $printer );
	create_config( $responses, $printer );
	parse_services($responses, $printer);
	parse_object_groups($responses, $printer);
	parse_filters( $responses, $printer );
	filters_mm($responses);
	parse_interfaces( $responses, $printer );
	interfaces_mm ($responses);
	parse_local_accounts( $responses, $printer );
	parse_snmp( $responses, $printer );
	parse_static_routes( $responses, $printer );
	# Close out the ZiptieElementDocument
	$printer->close_model();
	# Make sure to close the model file handle
	close_model_filehandle($filehandle);
}

sub interfaces_mm {
	my $responses = shift;

	delete $responses->{interfaces};
	while ( ( my $key, my $value ) = each( %{$responses} ) ) {
		delete $responses->{$key} if ( $key =~ /^if_\S+$/i );
	}
}

sub filters_mm {
	my $responses = shift;

	delete $responses->{policy};
	delete $responses->{addresses};
	while ( ( my $key, my $value ) = each( %{$responses} ) ) {
		delete $responses->{$key} if ( $key =~ /^acl_\S+$/i );
	}
}

sub get_command_responses {
	my ($cli_protocol, $prompt_regex) = @_;
	
	my $responses = {};
	
	$responses->{system} = $cli_protocol->send_and_wait_for( 'get system', $prompt_regex );
	$responses->{config} = $cli_protocol->send_and_wait_for( 'get config', $prompt_regex, 600 );
	$responses->{snmp}   = $cli_protocol->send_and_wait_for( 'get snmp',   $prompt_regex );
	$responses->{memory}    = $cli_protocol->send_and_wait_for( 'get memory',    $prompt_regex );
	$responses->{file_info} = $cli_protocol->send_and_wait_for( 'get file info', $prompt_regex );
	$responses->{files}     = $cli_protocol->send_and_wait_for( 'get file',      $prompt_regex );
	$responses->{services} = $cli_protocol->send_and_wait_for( 'get service', $prompt_regex );
	$responses->{service_groups} = $cli_protocol->send_and_wait_for( 'get service group', $prompt_regex );
	get_service_group_responses ($prompt_regex, $cli_protocol, $responses);
	$responses->{addresses} = $cli_protocol->send_and_wait_for( "get address", $prompt_regex);
	resolve_dns_names ($prompt_regex, $cli_protocol, $responses);
	get_address_group_responses ($prompt_regex, $cli_protocol, $responses);
	# Need to get the responses for zones overall, then dive in to get per-zone data.
	$responses->{zones} = $cli_protocol->send_and_wait_for("get zone", $prompt_regex);
	get_zone_responses ($prompt_regex, $cli_protocol, $responses);
	$responses->{interfaces} = $cli_protocol->send_and_wait_for( "get interface", $prompt_regex );
	get_interface_responses ($prompt_regex, $cli_protocol, $responses);
	$responses->{policy} = $cli_protocol->send_and_wait_for( "get policy all",  $prompt_regex );
	get_policy_responses ($prompt_regex, $cli_protocol, $responses);
	$responses->{users}    = $cli_protocol->send_and_wait_for( 'get admin user',  $prompt_regex );
	$responses->{routes} = $cli_protocol->send_and_wait_for( 'get route', $prompt_regex );
	
	disconnect($cli_protocol);
	return $responses;
}

sub _connect {
	# Grab our arguments
	my $connection_path = shift;

	# Create a new CLI protocol object by using the ZipTie::CLIProtocolFactory::create sub-routine
	# to examine the ZipTie::ConnectionPath argument for any command line interface (CLI) protocols
	# that may be specified.
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);

	# Make a connection to and successfully authenticate with the device
	my $device_prompt_regex = ZipTie::Adapters::Juniper::ScreenOS::AutoLogin::execute( $cli_protocol, $connection_path );

	# Store the regular expression that matches the primary prompt of the device under the key "prompt"
	# on the ZipTie::CLIProtocol object
	$cli_protocol->set_prompt_by_name( 'prompt', $device_prompt_regex );

	# Return the created ZipTie::CLIProtocol object and the device prompt encountered after successfully connecting to a device.
	return ( $cli_protocol, $device_prompt_regex );
}

sub get_service_group_responses {
	my $prompt_regex = shift;
	my $cli_protocol = shift;
	my $responses    = shift;

	while ( $responses->{service_groups} =~ /^(.+?)\s+\d+\s+.*defined\s*$/mig )
	{
		my $serviceGroupName = $1;
		$LOGGER->debug("Getting data for service group $serviceGroupName");

		$responses->{"service_group_$serviceGroupName"} =
		  $cli_protocol->send_and_wait_for(
			"get service group $serviceGroupName",
			$prompt_regex );

		# Now, we dig one layer deeper...turns out that a number of the member services in the
		# service groups need one further level of elaboration, since they\'re not defined anywhere else.
		# We only want to be looking at individual services though, not the entire list, which includes
		# service groups, which were individually queried above.
		if ( $responses->{"service_group_$serviceGroupName"} =~
			/\s+Members:\s+(.*)\n/ )
		{
			my $memberSvcs = $1;
			$memberSvcs =~ s/[\\'\"]//g;

			foreach my $memberSvc ( split( /\s+/, $memberSvcs ) ) {

				# Try to avoid getting data on service groups as individual services, it just junks things up a bit.
				if (   defined( $responses->{"service_group_$memberSvc"} )
					|| defined( $responses->{"service_$memberSvc"} ) )
				{
					next;
				}

				$LOGGER->debug(
					"Getting data for service group dependent group $memberSvc"
				);
				$responses->{"service_$memberSvc"} =
				  $cli_protocol->send_and_wait_for( "get service $memberSvc",
					$prompt_regex );
			}
		}
		else {
			$LOGGER->debug("Screen OS backup: Unable to find 'members' section for service group $serviceGroupName");
		}
	}
}



sub get_address_group_responses {
	my $prompt_regex = shift;
	my $cli_protocol = shift;
	my $responses    = shift;

	while ( $responses->{config} =~	/set group address \"(.*?)\" \"(.*?)\"( add)?/mig )
	{
		my ( $address_zone, $address_group ) = ( $1, $2 );
		if ( !defined($3) ) {

			my $ncname = ncname_from_string($address_group);
			if ( defined( $responses->{"address_group_$ncname"} ) ) {
				next;
			}

			$LOGGER->debug("Getting data for address group $address_group and address zone $address_zone");
			$responses->{"address_group_$ncname"} = $cli_protocol->send_and_wait_for("get group address \"$address_zone\" \"$address_group\"", $prompt_regex );
		}
	}
}


sub resolve_dns_names {
	my $prompt_regex = shift;
	my $cli_protocol = shift;
	my $responses    = shift;

	while ( $responses->{addresses} =~ /^(.+?)\s+(\S+?)\s+\d{1,4}[\w\s]*$/mig )
	{
		my $addr_name = ncname_from_string($1);
		my $host_name = $2;

		#...and skip over those that already contains straight IP addresses
		if ( $host_name !~ /\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}/m ) {
			$LOGGER->debug("Get DNS address for host name $2");

			# resolve DNS name from ScreenOS device
			$responses->{'dns_resolver'}->{$addr_name} = $cli_protocol->send_and_wait_for( "get dns name $host_name",  $prompt_regex );
		}
	}
}


sub get_zone_responses {
	my $prompt_regex = shift;
	my $cli_protocol = shift;
	my $responses    = shift;

	while ( $responses->{zones} =~ /^\s*\d+\s+(\S+)\s+\S+\s.+$/mig ) {
		my $ncname = ncname_from_string($1);
		$responses->{"zone_$ncname"} = $cli_protocol->send_and_wait_for( "get zone $1", $prompt_regex );
	}
}


sub get_interface_responses {
	my $prompt_regex = shift;
	my $cli_protocol = shift;
	my $responses    = shift;

	while ($responses->{interfaces} =~ /^(\S+)\s+((unnumbered)|(\d+\.)).+$/mig )
	{
		$responses->{"if_$1"} = $cli_protocol->send_and_wait_for( "get interface $1", $prompt_regex );
	}
}


sub get_policy_responses {
	my $prompt_regex = shift;
	my $cli_protocol = shift;
	my $responses    = shift;

	foreach my $line ( split( "\n", $responses->{policy} ) ) {
		if ( $line =~ /^\s*(\d+)\s+\S+\s+\S+\s.+$/ ) {
			$responses->{"acl_$1"} = $cli_protocol->send_and_wait_for( "get policy id $1", $prompt_regex );
		}
	}
}


sub clean_responses {
	my $responses = shift;

	foreach my $command ( keys %$responses ) {
		$responses->{$command} =~ s/[\b]\s?//mig;
		$responses->{$command} =~ s/^--- more ---(\s+)?//mg;
	}
}

1;

__END__

=head1 NAME

ZipTie::Adapters::Juniper::ScreenOS

=head1 SYNOPSIS

    use ZipTie::Adapters::Juniper::ScreenOS;
	ZipTie::Adapters::Juniper::ScreenOS::backup( $backup_document );

=head1 DESCRIPTION


=head1 PUBLIC SUB-ROUTINES

=over 12

=item C<backup($backup_document)>

Performs the backup of the device that is described in the specified XML document.  This XML document contains
a "connectionPath" element that contains the IP/hostname, protocol, credential, and file server information that
may be needed to connect to, authenticate with, and back up the device in question.  This XML will be parsed into
a C<ZipTie::ConnectionPath> object to help easily access this vital information.

The act of backing up a device means that information about a device will be retreived, as well as the configuration
data/files for that device.  All of this collected information will be used to populate the ZipTie Element Document,
which is an open and extensible format for describing a network device.

The return value for this method will be a string containing the successfully populated ZipTie Element Document.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_connect($connection_path)>

Creates the initial CLI connection to the device.  A list is returned containing two elements: the first is a
C<ZipTie::CLIProtocol> object that is a CLI client that can communitcate with the device, and the second is a
regular expression that can be used to match the primary prompt of the device.  This is useful when sending
commands to and receiving responses from the device through the C<ZipTie::CLIProtocol> object and being able to
know when a command has generated all the output and returned back to the primary prompt.

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

Contributor(s): rkruse, Dylan White (dylamite@ziptie.org)
Date: August 10, 2007

=cut
