package ZipTie::Adapters::Juniper::ScreenOS::Telemetry;

use strict;
use warnings;

use ZipTie::Adapters::Juniper::ScreenOS::AutoLogin;
use ZipTie::Adapters::Juniper::ScreenOS::Disconnect qw(disconnect);
use ZipTie::Adapters::Utils qw(mask_to_bits strip_mac get_crep choose_admin_ip get_model_filehandle close_model_filehandle);
use ZipTie::CLIProtocolFactory;
use ZipTie::Model::XmlPrint;
use ZipTie::Logger;
use ZipTie::Typer;

use Exporter qw(import);
our @EXPORT_OK = qw(parse_arp parse_mac_table parse_telemetry_interfaces);

my $LOGGER = ZipTie::Logger::get_logger();

my $CIPM_RE = get_crep('cipm');
my $MAC_RE1 = get_crep('mac1');
my $MAC_RE2 = get_crep('mac2');
my $MAC_RE3 = get_crep('mac3');

sub invoke {
	my ( $pkg, $connectionPathDoc ) = @_;
	
	# Setup work
	my ($connectionPath, $discoveryParams) = ZipTie::Typer::translate_document( $connectionPathDoc, 'connectionPath' );
	my $filehandle = get_model_filehandle( 'ScreenOS', $connectionPath->get_ip_address() );
	my $cliProtocol = ZipTie::CLIProtocolFactory::create($connectionPath);
	my $responses = {};
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'telemetry');
	$printer->attributes(1);
	$printer->open_discovery_event();

	# Make a connection to device and get the data necessary to fill out the discovery event document 
	my $promptRegex = ZipTie::Adapters::Juniper::ScreenOS::AutoLogin::execute( $cliProtocol, $connectionPath );
	$cliProtocol->send_and_wait_for('set console page 0', $promptRegex);
	$printer->open_element('neighbors');
	
	# Process the ARP table
	$responses->{arp} = $cliProtocol->send_and_wait_for('get arp', $promptRegex);
	parse_arp($responses, $printer);
	delete $responses->{arp};

	# Process the MAC address table 
	$responses->{mac} = $cliProtocol->send_and_wait_for('get mac-learn', $promptRegex);
	parse_mac_table($responses, $printer);
	delete $responses->{mac};

	$printer->close_element('neighbors');

	# Process the interfaces and their status
	$responses->{interfaces} = $cliProtocol->send_and_wait_for('get interface', $promptRegex);
	while ( $responses->{interfaces} =~ /^\s*(\S+)\s+(?:[\da-f\.\:]+)\/\d+\s+\S+\s+(?:[\da-f\.\:\-]+)\s+\S+\s+(?:\S+)\s+\S+(\s+\S+)?\s*$/mig )
	{
		$responses->{"if_counter_$1"} = $cliProtocol->send_and_wait_for("get counter statistics interface $1", $promptRegex);
	}
	my $interfacesHash = parse_telemetry_interfaces($responses, $printer);
	undef( %{$responses} );

	my $adminIp    = $connectionPath->get_ip_address();
	if ( $discoveryParams->{calculateAdminIp} eq 'true' ) {
		$adminIp = choose_admin_ip($connectionPath->get_ip_address(), $interfacesHash);
	}
	$printer->print_element( 'adminIp', $adminIp );
	$interfacesHash = 0;

	# Tear down work
	$printer->close_element('DiscoveryEvent');
	close_model_filehandle($filehandle);
	disconnect($cliProtocol);
}

sub parse_arp {
	my ( $in, $out ) = @_;
	my ( $arps, $arp_seg, $arp_asic_seg );
	( $arp_seg, $arp_asic_seg ) = split( /^\s*Arp entries on ASIC chip\(s\)\s*$/im, $in->{arp} );

	while ( $arp_seg =~ /^\s*($CIPM_RE)\s+($MAC_RE1)\s+\S+?\/(\S+)\s+\S+\s+\d+\s*/mig )
	{
		my $arp = {
			ipAddress  => $1,
			macAddress => strip_mac($2),
			interface  => $3,
		};
		push( @{ $arps->{arpEntry} }, $arp );
	}
	if ( $arp_asic_seg ) {
		while ( $arp_asic_seg =~ /^\s*\d+\s*($CIPM_RE)\s+($MAC_RE1)\s+(\S+)\s+\S+\s+\d+\s*/mig )
		{
			my $arp = {
				ipAddress  => $1,
				macAddress => strip_mac($2),
				interface  => $3,
			};
			push( @{ $arps->{arpEntry} }, $arp );
		}
	}
	if ( $arps ) {
		$out->print_element('arpTable', $arps);
	} else {
		$LOGGER->debug('WARNING: No arp entries found');
	}
}

sub parse_mac_table {
	my ( $in, $out ) = @_;
	my $macs;

	while ( $in->{mac} =~ /^\s*(\S+)\s+($MAC_RE1|$MAC_RE2|$MAC_RE3)\s+\S+\s+(\S+)\s*$/mig ) {
		my $macEntry = {
			interface => _full_int_name($1),
			macAddress => strip_mac($2),
			vlan => $3,
		};
		push( @{ $macs->{macEntry} }, $macEntry );
	}
	if ( $macs ) {
		$out->print_element('macTable', $macs);
	} else {
		$LOGGER->debug('WARNING: No mac entries found');
	}
}

sub parse_telemetry_interfaces {
	my ( $in, $out ) = @_;
	my $interfaces;

	while ( $in->{interfaces} =~ /^\s*(\S+)\s+($CIPM_RE)\/(\d+)\s+\S+\s+(N\/A|$MAC_RE1|$MAC_RE2|$MAC_RE3)\s+\S+\s+(\S+)\s+\S+(\s+\S+)?\s*$/mig ) {	
		my $name		= $1;
		my $ipAddress	   = $2;
		my $mask		= $3;
		my $status	      = $5;
		my $interface = {
			name       => $name,
			type       => 'other',
			inputBytes => 0,
			operStatus => ( $status eq 'U' ? 'Up' : 'Down' ),
		};
		if ( $in->{"if_counter_$name"} && $in->{"if_counter_$name"} =~ /in\s+bytes\s+(\d+)/mi ) {
			$interface->{inputBytes} = $1;
		}
		my $ipEntry = {
			ipAddress => $ipAddress,
			mask      => mask_to_bits ( $mask ),
		};
		push( @{ $interface->{ipEntry} }, $ipEntry );
		push( @{ $interfaces->{interface} }, $interface );
	}
	if ( $interfaces ) {
		$out->print_element( 'interfaces', $interfaces );
	} else {
		$LOGGER->fatal('FATAL: No interfaces found, there must be at least one');
	}

	return $interfaces;
}

1;
