package ZipTie::Adapters::McAfee::Sidewinder;

use v5.10;
use strict;
use warnings;
use ZipTie::Adapters::McAfee::Sidewinder::AutoLogin qw(cli_connect);
use ZipTie::Adapters::McAfee::Sidewinder::Parsers qw(parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_interfaces parse_static_routes parse_object_groups parse_vlans parse_stp);
use ZipTie::Adapters::McAfee::Sidewinder::Disconnect qw(disconnect);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::Adapters::GenericAdapter;
use ZipTie::CLIProtocol;
use ZipTie::CLIProtocolFactory;
use ZipTie::ConnectionPath;
use ZipTie::Credentials;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::SNMP;
use ZipTie::SnmpSessionFactory;
use ZipTie::Typer;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

sub backup
{
	my $packageName = shift;
	my $backupDoc   = shift;    # how to backup this device
	my $responses    = {};      # will contain device responses to be handed to the Parsers module

	# Translate the backup operation XML document into ZipTie::ConnectionPath
	my ($connectionPath) = ZipTie::Typer::translate_document( $backupDoc, 'connectionPath' );

	# Set up the XmlPrint object for printing the ZiptieElementDocument (ZED)
	my $filehandle = get_model_filehandle( 'McAfee Sidewinder', $connectionPath->get_ip_address() );
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

		# Make a Telnet or SSH connection
	my ( $cliProtocol, $promptRegex ) = cli_connect($connectionPath);
	
	# escalate from user to admin
	$promptRegex =~ s/User/(Admn|AdRO)/;
	$responses->{srole} = $cliProtocol->send_and_wait_for( 'srole', $promptRegex );

	# send commands to the system and gather the responses
	$responses = get_command_responses($cliProtocol, $promptRegex);

	# call methods to parse the data from the device responses
	if(!parse_system( $responses, $printer )){ return; } # exit backup if hostname wasn't found
	parse_chassis( $responses, $printer );
	parse_object_groups( $responses, $printer );
	parse_filters($responses, $printer);
	parse_interfaces( $responses, $printer );
	parse_vlans( $responses, $printer );
	parse_static_routes( $responses, $printer );
	
	clear_command_responses($responses); 	# memory management

	$printer->close_model();                # close out the ZiptieElementDocument
	close_model_filehandle($filehandle);    # Make sure to close the model file handle
}

sub get_command_responses
{
	my ($cliProtocol, $promptRegex) = @_;
	my $responses = {};
	
	$responses->{hostname} = $cliProtocol->send_and_wait_for( 'hostname', $promptRegex );
	$responses->{version} = $cliProtocol->send_and_wait_for( 'uname -r', $promptRegex );
	$responses->{uptime} = $cliProtocol->send_and_wait_for( 'uptime', $promptRegex );
	$responses->{license} = $cliProtocol->send_and_wait_for( 'cf license q', $promptRegex );
	$responses->{routingTable} = $cliProtocol->send_and_wait_for( 'cf route status', $promptRegex );
	$responses->{addresses} = $cliProtocol->send_and_wait_for( 'cf ipaddr q', $promptRegex );
	$responses->{ipranges} = $cliProtocol->send_and_wait_for( 'cf iprange q', $promptRegex );
	$responses->{subnets} = $cliProtocol->send_and_wait_for( 'cf subnet q', $promptRegex );
	$responses->{domainList} = $cliProtocol->send_and_wait_for( 'cf domain q', $promptRegex );
	# get responses for each domain
	my @domainList = get_domain_list($responses);
	foreach my $domain(@domainList){
		my $name = $domain->{name};
		my $address = $domain->{address};
		$responses->{domain}{$name} = $cliProtocol->send_and_wait_for("dig $address +noall +answer", $promptRegex);
	}
	$responses->{hosts} = $cliProtocol->send_and_wait_for( 'cf host q', $promptRegex );
	$responses->{networkMaps} = $cliProtocol->send_and_wait_for( 'cf netmap q', $promptRegex );
	$responses->{networkGroups} = $cliProtocol->send_and_wait_for( 'cf netgroup q', $promptRegex );
	$responses->{predefinedServices} = $cliProtocol->send_and_wait_for( 'cf appdb list verbose=on', $promptRegex );
	$responses->{customServices} = $cliProtocol->send_and_wait_for( 'cf application q', $promptRegex );
	$responses->{serviceGroups} = $cliProtocol->send_and_wait_for( 'cf appgroup q', $promptRegex );
	$responses->{policies} = $cliProtocol->send_and_wait_for( 'cf policy q', $promptRegex );
	$responses->{interfaces} = $cliProtocol->send_and_wait_for( 'cf interface q', $promptRegex );
	$responses->{zones} = $cliProtocol->send_and_wait_for( 'cf zone q', $promptRegex );
	
	disconnect($cliProtocol);
	return $responses;
}

sub get_domain_list
{
	my $in = shift;
	my @domainList;
	my $domainResponse = $in->{domainList};
	my %domain;

	$domainResponse =~ s/\\\n//g;
	$domainResponse =~ s/\\\r\n//g;

	foreach my $line(split("\n", $domainResponse)){
		$line =~ s/(^\s+|\s+$)//g;
		my $name;
		my $address;
		if($line =~ /name=\'(.*?)\'/i){
			$domain{name} = $1;
		}elsif($line =~ /name=(\S+)/i){
			$domain{name} = $1;
		}
		if($line =~ /domain=(https?\:\/\/)?(www\.)?(\S+)/i){
			$domain{address} = $3;
		}

		if(defined $domain{name} && defined $domain{address}){
			push (@domainList, {%domain});
		}
	}

	return @domainList;
}

sub clear_command_responses
{	
	my $responses = shift;
	
	foreach my $command (keys %$responses){
		delete($responses->{$command});
	}
}

1;