package ZipTie::Adapters::McAfee::Sidewinder::Parsers;

use 5.010_000;
use strict;
use warnings;

use ZipTie::Adapters::Utils qw(seconds_since_epoch get_mask get_port_number trim get_interface_type get_crep mask_to_bits getUnitFreeNumber);
use MIME::Base64 'encode_base64';
use Data::Dumper;
use Storable 'dclone';

use Exporter qw(import);
our @EXPORT_OK = qw(parse_routing parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_interfaces parse_static_routes parse_object_groups parse_vlans parse_stp);

use ZipTie::Logger;
my $LOGGER = ZipTie::Logger::get_logger();

my $ip_regex = get_crep("cipm");
my $unknown = "Unknown";
my $make = "McAfee";
my %serviceGroupsNAT;

sub parse_chassis
{
	my ( $in, $out ) = @_;
	my $chassis;

	$chassis->{'core:asset'}->{'core:assetType'} = 'Chassis';
	$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:make'} = $make;
	
	if (defined $in->{version}){
		if($in->{version} =~ /^uname -r\n*\s*(\S+)\n*/mi){
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:firmwareVersion'} = $1;
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = 'Sidewinder ' . $1;
		}else{
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:firmwareVersion'} = $unknown;
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $unknown;
		}
	}
	
	if(defined $in->{license}){
		if($in->{license} =~ /serial_number=(\S+)/mi){
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = $1;
		}
	}
	
	$out->print_element( "chassis", $chassis );
}

sub parse_system
{
	my ( $in, $out ) = @_;

	# parse and print host/system name
	if (defined $in->{hostname}) {
		if($in->{hostname} =~ /^hostname\n*\s*(\S+)\n*/mi){
			$out->print_element('core:systemName', $1);
		}else{
			$LOGGER->fatal("System Name could not be found - This is required for a successful backup!");
			return 0; # else cause backup to fail
		}
	}
	
	$out->open_element('core:osInfo');
	$out->print_element( 'core:make',    $make . ' Sidewinder' );
	
	# parse and print software version
	if (defined $in->{version}){
		if($in->{version} =~ /^uname -r\n*\s*(\S+)\n*/mi){
			$out->print_element('core:version', $1);
		}else{
			$out->print_element('core:version', $unknown);
		}
	}

	$out->print_element( 'core:osType',  'SecureOS' );
	_generate_packet_processing_declaration($out);
	$out->close_element('core:osInfo');
	
	$out->print_element( 'core:deviceType', 'Firewall' );
		
	# parse, calculate and print the last reboot time
	if (defined $in->{uptime}){
	    if ($in->{uptime} =~ /up\s(.+)$/mi){
	    	my $lastReboot = parse_reboot($1);
	        $out->print_element("core:lastReboot", $lastReboot);
	    }
    }
    
    return 1; # hostname was found - continue backup
}

sub parse_reboot
{
	my $_ = @_;
	
	my ($years)   = /(\d+)\s*(years|year)?/;
    my ($weeks)   = /(\d+)\s*(weeks|week)?/;
    my ($days)    = /(\d+)\s*(days|day)?/;
    my ($hours)   = /(\d+)\s*(hours|hour)?/;
    my ($minutes) = /(\d+)\s*(mins|min)?/;

    # subract the last reboot from the current time
    my $lastReboot = time();
    $lastReboot -= $years * 52 * 7 * 24 * 60 * 60 if ($years);
    $lastReboot -= $weeks * 7 * 24 * 60 * 60      if ($weeks);
    $lastReboot -= $days * 24 * 60 * 60           if ($days);
    $lastReboot -= $hours * 60 * 60               if ($hours);
    $lastReboot -= $minutes * 60                  if ($minutes);
    
    return $lastReboot;
}

sub parse_routing
{
	my ( $in, $out ) = @_;
}

sub parse_local_accounts
{
	my ( $in, $out ) = @_;
}

sub get_value {
	my ($key, $text) = @_;
	my $value;
	
	if($text =~ /$key=\'(.*?)\'/i || $text =~ /$key=(\S+)/i){
		$value = $1;
	}
	
	return $value;
}

sub parse_firewall_policies {
	my ($in, $policyFilter) = @_;
	my $cleanedPolicyResponse = clean_response($in->{policies});
	
	# check for static NAT
	my $staticNatOnly = undef;
	if (defined $policyFilter){
		if ($policyFilter =~ /staticNatOnly/) {
			$staticNatOnly = 1;
		}
	}

	# process policies and generate filters
	my %filterEntry;
	my @filtersArray;
	my $processOrder;
	foreach my $policy ( split( '\n', $cleanedPolicyResponse ) ) {
		if ( $policy =~ /^policy add table=rule / ) {
			$filterEntry{processOrder} = ++$processOrder;
			$filterEntry{name} = get_value('name', $policy);
			if ( $policy =~ /disable=yes/ ) {
				$filterEntry{disabled} = 1;
			}
			if ( $policy =~ /authgroups=/ ) {
				if ( $policy !~ /authgroups=(\'\'|\'\*\')/ ){
					$filterEntry{identityBased} = 1;
				}
			}
			if ($policy =~ /description='(\w+(\s+\w+){0,})'/) {
	        	$filterEntry{description} = $1;
	        }
			my $ruleGroup = get_value('rulegroup', $policy);
			$filterEntry{ruleGroup} = defined $ruleGroup ? $ruleGroup : 'default';
			
			# get NAT related info
			if ( $policy =~ /nat_mode=(normal|src_port|none)/ ) {
				$filterEntry{natMode} = $1;
			}
			if ( $policy =~ /nat_addr=\'(.*?)\'/ || $policy =~ /nat_addr=(\S+)/ ){
				my ($key, $value) = split(/\:/, $1, 2);
				$filterEntry{natAddr} = $value;
			}
			if ( $policy =~ /redir=\'(.*?)\'/ || $policy =~ /redir=(\S+)/ ){
				my ($key, $value) = split(/\:/, $1, 2);
				$filterEntry{redirAddr} = $value;
			}
			if ( $policy =~ /redir_port=(\d+)/ ){
				$filterEntry{redirPort} = $1;
			}
			if ( defined $filterEntry{natAddr} || defined $filterEntry{redirAddr} ) {
				$filterEntry{nat} = $1;
			}
			
			# get addresses and zones
			%filterEntry = get_policy_addresses(\%filterEntry, 'source', $policy);
			%filterEntry = get_policy_addresses(\%filterEntry, 'dest', $policy);
			%filterEntry = get_policy_zones(\%filterEntry, 'source_zones', $policy);
			%filterEntry = get_policy_zones(\%filterEntry, 'dest_zones', $policy);
			
			if( $policy =~ /action=(\S+)/ ){
				$filterEntry{primaryAction} = $1 eq 'allow' ? 'permit' : $1;
			}
			
			# get services
			my @services;
			if($policy =~ /application=\'(.*?)\'/ || $policy =~ /application=(\S+)/){
				@services = split(/\,/, $1);
			}
			foreach my $service (@services){
				my ($key, $value) = split(/\:/, $service, 2);
				push @{ $filterEntry{services} }, $value
			}
			
			# collect and store port over-rides
			my $tcp_ports = get_value('tcp_ports', $policy);
			my $ssl_ports = get_value('ssl_ports', $policy);
			my $udp_ports = get_value('udp_ports', $policy);
			if($tcp_ports ne '' || $ssl_ports ne '' || $udp_ports ne ''){
				%filterEntry = get_port_lists(\%filterEntry, $tcp_ports, 'tcp');
				%filterEntry = get_port_lists(\%filterEntry, $ssl_ports, 'tcp');
				%filterEntry = get_port_lists(\%filterEntry, $udp_ports, 'udp');
			}else{
				# store default ports for NAT rules if ports were not overriden
				%filterEntry = remove_source_duplicates(\%filterEntry);
			}	
		
			if (defined $filterEntry{disabled}){ # don't store disabled polices
				%filterEntry = ();                    
				next;
			}
			
			# set various default values if not parsed
			$filterEntry{log} = 'true' unless defined $filterEntry{log};
			$filterEntry{primaryAction} = 'deny' unless defined $filterEntry{primaryAction};
			$filterEntry{identityBased} = 0 unless defined $filterEntry{identityBased};
			$filterEntry{geolocation} = 0 unless defined $filterEntry{geolocation};
			$filterEntry{ipv6} = 0 unless defined $filterEntry{ipv6};
			
			if (defined $staticNatOnly) {
				if (defined $filterEntry{nat}) {
					push @filtersArray, {%filterEntry};
				}
			}else{
				push @filtersArray, {%filterEntry};	
			}
					
			%filterEntry = ();
			next;
		}
	}
	
	return @filtersArray;
}

sub custom_compare{
	my ($portRef, $existingPortRef) = @_;
	my %port = %$portRef;
	my %existingPort = %$existingPortRef;
	my $existingPortFlag = 1;
	# compare two hashes
	for my $key ( keys %port ) {
    	if($port{$key} ne $existingPort{$key}){
        	$existingPortFlag = 0;
	    }
	}
	
	return $existingPortFlag;
}

sub remove_source_duplicates {
	my ($filterEntryRef) = @_;
	my %filterEntry = %$filterEntryRef;
	
	foreach my $service(@{$filterEntry{services}}){
		if(defined $serviceGroupsNAT{$service}{portExpressions}){
			foreach my $portExp(@{$serviceGroupsNAT{$service}{portExpressions}}){
				my $existingPortFlag = 0;
				foreach my $existingPort(@{$filterEntry{portExpressions}}){
					if (custom_compare($portExp, $existingPort)){
						$existingPortFlag = 1;
					}
				}
				if($existingPortFlag == 0){
					push @{$filterEntry{portExpressions}}, $portExp;
				}
			}
		}
		if(defined $serviceGroupsNAT{$service}{portRanges}){
			foreach my $portRange(@{$serviceGroupsNAT{$service}{portRanges}}){
				my $existingPortRangeFlag = 0;
				foreach my $existingPortRange(@{$filterEntry{portRanges}}){
					if (custom_compare($portRange, $existingPortRange)){
						$existingPortRangeFlag = 1;
					}
				}
				if($existingPortRangeFlag == 0){
					push @{$filterEntry{portRanges}}, $portRange;
				}
			}
		}
	}
	return %filterEntry;
}

sub remove_duplicate_ports {
	my ($filterEntry, $type, $out) = @_;
	my $newFilter;
	
	foreach my $service(@{$filterEntry->{services}}){
		if(defined $filterEntry->{portExpressions}){
			foreach my $portExp(@{$filterEntry->{portExpressions}}){
				my $existingPortFlag = 0;
				foreach my $existingPort(@{$newFilter->{portExpressions}}){
					if (custom_compare($portExp, $existingPort)){
						$existingPortFlag = 1;
					}
				}
				if($existingPortFlag == 0){
					push @{$newFilter->{portExpressions}}, $portExp;
				}
			}
		}
		if(defined $filterEntry->{portRanges}){
			foreach my $portRange(@{$filterEntry->{portRanges}}){
				my $existingPortRangeFlag = 0;
				foreach my $existingPortRange(@{$newFilter->{portRanges}}){
					if (custom_compare($portRange, $existingPortRange)){
						$existingPortRangeFlag = 1;
					}
				}
				if($existingPortRangeFlag == 0){
					push @{$newFilter->{portRanges}}, $portRange;
				}
			}
		}
	}
	
	print_ports_and_ranges($newFilter, 'destinationService', $out);
}

sub get_policy_addresses {
	my ($filterEntryRef, $type, $line) = @_;
	my %filterEntry = %$filterEntryRef;
	
	my $addresses = get_value($type, $line);
	if(defined $addresses){
		my @addresses = split /\,/, $addresses;
		my @supportedAddresses;
		foreach my $address(@addresses){
			if($address =~ /(\w+):(.+)/){
				my $key = $1;
				my $value = $2;
				if($key eq 'all'){
					if($value eq 'v6'){
						$filterEntry{ipv6} = 1;
					}elsif($value eq 'v4'){
						push @supportedAddresses, 'any';
					}
				}elsif ($key eq 'geolocation'){
					$filterEntry{geolocation} = 1;
				}else{
					push @supportedAddresses, $value;
				}
			}elsif ($address =~ /\*/){
				push @supportedAddresses, 'any';
			}
		}
		if($type eq 'source'){
			push @{ $filterEntry{srcAddresses} }, @supportedAddresses;
		}elsif($type eq 'dest'){
			push @{ $filterEntry{dstAddresses} }, @supportedAddresses;
		}
	}
	return %filterEntry;
}

sub get_policy_zones {
	my ($filterEntryRef, $type, $line) = @_;
	my %filterEntry = %$filterEntryRef;
	
	my $zones = get_value($type, $line);
	if(defined $zones){
		$zones =~ s/\w*://g;
		$zones =~ s/v4|\*/any/g;
		my @zones = split /\,/, $zones;
		@zones = sort @zones;
		if($type eq 'source_zones'){
			push @{ $filterEntry{srcZones} }, @zones;
		}elsif($type eq 'dest_zones'){
			push @{ $filterEntry{dstZones} }, @zones;
		}
	}else{
		if($type eq 'source_zones'){
			push @{ $filterEntry{srcZones} }, 'any';
		}elsif($type eq 'dest_zones'){
			push @{ $filterEntry{dstZones} }, 'any';	
		}
	}
	return %filterEntry;
}

sub create_filter_lists {
	my ($filtersArray, $natList) = @_;
	my $filterList;
	my $orderCount = 0;
	
	foreach my $filterEntry (@$filtersArray) {
		my $filterListName;
		if($natList){
			if(defined $filterEntry->{nat}){
				$filterListName = $filterEntry->{name} . "_NATMapping_forward";
			}elsif(defined $filterEntry->{natReverse}){
				$filterListName = $filterEntry->{name} . "_NATMapping_reverse";
			}else{
				$filterListName = $filterEntry->{name} . "_NATMapping_forward";
			}
		}else{
			$filterListName = $filterEntry->{name};
		}
		$filterList->{$filterListName}->{identityBased} = $filterEntry->{identityBased};
		$filterList->{$filterListName}->{ipv6} = $filterEntry->{ipv6};
		$filterList->{$filterListName}->{geolocation} = $filterEntry->{geolocation};
		$filterList->{$filterListName}->{description} = $filterEntry->{description};
		$filterList->{$filterListName}->{srcZones} = $filterEntry->{srcZones};
		$filterList->{$filterListName}->{dstZones} = $filterEntry->{dstZones};
		$filterList->{$filterListName}->{order} = $orderCount++;
		
		# if SDNAT then split into src and dst filterEntrys
		if($natList && defined $filterEntry->{srcAddrTranslation} && defined $filterEntry->{dstAddrTranslation}){
			@{ $filterList->{$filterListName}->{filters} } = split_filters($filterEntry);
		}else{
			push @{ $filterList->{$filterListName}->{filters} }, $filterEntry;
		}
	}
	
	return $filterList;
}

sub split_filters {
	my $filterEntry = shift;
	my @filtersArray;
	my $redirFilter = dclone($filterEntry);
	
	$filterEntry->{name} = $filterEntry->{name} . "_SRC";
	delete $filterEntry->{dstAddrTranslation};
	push @filtersArray, $filterEntry;
	
	$redirFilter->{name} = $redirFilter->{name} . "_DST";
	delete $redirFilter->{srcAddrTranslation};
	push @filtersArray, $redirFilter;
	
	return @filtersArray;
}

sub print_implicit_deny_filter {
	my $out = shift;

	# print implied deny all rule
	$out->open_element("filterList");
	$out->print_element( "defaultAction", "deny" );
	$out->print_element( "appliesTo", { source => "any" } );
	$out->print_element( "mode", "stateful" );
	$out->print_element( "name", "Implicit" );
	$out->print_element( "processingPhase", {'core:packetProcessingPhase' => "packet-filtering"});
	$out->close_element("filterList");
}

sub print_filter_lists {
	my ($out, $key, $filterList, $isNatList) = @_;
	my @dstZones = @{$filterList->{$key}->{dstZones}} if $filterList->{$key}->{dstZones};
	my @srcZones = @{$filterList->{$key}->{srcZones}} if $filterList->{$key}->{srcZones};
	
	if ($filterList->{$key}->{identityBased} || $filterList->{$key}->{ipv6} || $filterList->{$key}->{geolocation}){
		if ($filterList->{$key}->{identityBased}){
			$LOGGER->debug("Dropping policy id $key - identity-based policies currently not supported by QRM.");
		}
		if ($filterList->{$key}->{ipv6}){
			$LOGGER->debug("Dropping policy id $key - IPv6 policies currently not supported by QRM.");
		}
		if ($filterList->{$key}->{geolocation}){
			$LOGGER->debug("Dropping policy id $key - geolocation policies currently not supported by QRM.");
		}
	}else{
		$out->open_element("filterList");
		$out->print_element( "defaultAction", "none" ); # so core moves on to next list if no match
		if(defined $filterList->{$key}->{description}){
			$out->print_element("description", $filterList->{$key}->{description});
		}
		foreach my $source(@srcZones){
			foreach my $dest(@dstZones){
				$out->open_element("appliesTo");
				$out->print_element( "sourceGroup", $source );
				$out->print_element( "destinationGroup", $dest );
				$out->close_element("appliesTo");
			}
		}
		foreach my $filterEntry ( @{ $filterList->{$key}->{filters} } ) {
			print_filter_entry( $out, $filterEntry );
		}
		$out->print_element( "mode", "stateful" );
		$out->print_element( "name", $key);
		if ($isNatList){
			$out->print_element( "processingPhase", {'core:packetProcessingPhase' => "address-translation"});
		} else {
			$out->print_element( "processingPhase", {'core:packetProcessingPhase' => "packet-filtering"});
		}
		$out->close_element("filterList");
	}
}

sub print_filter_entry {
	my ($out, $filterEntry) = @_;
	
	$out->open_element("filterEntry");
	print_nat_addresses($filterEntry, $out, "dst", "destination");
	
	if(defined $filterEntry->{portExpressions} || defined $filterEntry->{portRanges} && defined $filterEntry->{nat}){
			remove_duplicate_ports($filterEntry, 'destinationService', $out);
	}else{
		foreach my $service (@{$filterEntry->{services}}){
			unless ($service eq 'any'){
				if(defined $filterEntry->{portExpressions} || defined $filterEntry->{portRanges}){
					print_ports_and_ranges($filterEntry, 'destinationService', $out);
				}else{
					$out->print_element('destinationService', {objectGroupReference => $service});
				}
			}
		}
	}
	
	$out->print_element("log", $filterEntry->{log});
	$out->print_element("name", $filterEntry->{name});
	$out->print_element("primaryAction", $filterEntry->{primaryAction});
	$out->print_element("processOrder", $filterEntry->{processOrder});
	
	print_nat_addresses($filterEntry, $out, "src", "source");
	
	if (defined $filterEntry->{sourceServices}){
		foreach my $sourceService ($filterEntry->{sourceServices}){
			if (defined $sourceService->{translationAddress}->{translationServicePort}->{translationPort}){
				$out->print_element('sourceService',  $sourceService->{translationAddress}->{translationServicePort}->{translationPort});
			}
		}
	}
	
	if(defined $filterEntry->{ztSignature}){
		$out->print_element("ztSignature", $filterEntry->{ztSignature});
	}
	$out->close_element("filterEntry");
}

sub print_nat_addresses {
	my ($filterEntry, $out, $type, $printType) = @_;
	
	if (defined $filterEntry->{$type."AddrTranslation"}){
		my $tempType = $type eq 'dst' ? 'dest' : $printType;
		$out->print_element($tempType."AddressTranslation", $filterEntry->{$type."AddrTranslation"});
	}
	foreach my $address (@{$filterEntry->{$type."Addresses"}}){
		unless ($address eq 'any'){
			if (defined $filterEntry->{natReverse}){
				$out->print_element($printType."IpAddr", $address);
			} else{
				$out->print_element($printType."IpAddr", {objectGroupReference => $address});
			}
		}
	}
}

sub print_ports_and_ranges {
	my ($filterEntry, $type, $out) = @_;
		
	foreach my $portExp (@{$filterEntry->{portExpressions}}){
		if($type eq 'destinationService'){	
			$out->open_element($type);
		}
		$out->open_element('portExpression');
		$out->print_element('port', $portExp->{port});
		$out->print_element('operator', 'eq');
		if (defined $portExp->{protocol}){
			$out->print_element('protocol', $portExp->{protocol});
		}
		$out->close_element('portExpression');
		if($type eq 'destinationService'){	
			$out->close_element($type);
		}
	}
	foreach my $portRange (@{$filterEntry->{portRanges}}){	
		if($type eq 'destinationService'){	
			$out->open_element($type);
		}
		$out->open_element('portRange');
		$out->print_element('portStart', $portRange->{portStart});
		$out->print_element('portEnd', $portRange->{portEnd});
		if (defined $portRange->{protocol}){
			$out->print_element('protocol', $portRange->{protocol});
		}
		$out->close_element('portRange');
		if($type eq 'destinationService'){	
			$out->close_element($type);
		}
	}
}

sub parse_filters
{
	my ( $in, $out ) = @_;
	
	my @filters = parse_firewall_policies($in);
	my @staticNat = parse_static_nat($in);
	my $staticNatList = create_filter_lists(\@staticNat, 1);
	
	$out->open_element("filterLists");
	
	my $filterList = create_filter_lists(\@filters);
	# print only the implicit filter if no policies defined
	unless (defined $filterList){
		print_implicit_deny_filter($out);
		$out->close_element("filterLists");
		return;
	}
	$out->open_element("filterListsSection");
	
	foreach my $key ( sort {$filterList->{$a}->{order} <=> $filterList->{$b}->{order}} keys %{$filterList}){ # iterate filterlists in order
		print_filter_lists($out, $key, $filterList);
	}
	
	print_implicit_deny_filter($out); # needs to be last list if no matches
	$out->close_element("filterListsSection");
	
	# nat section after packet filters
	if (defined $staticNatList){
		$out->open_element("filterListsSection");
		foreach my $key ( sort {$staticNatList->{$a}->{order} <=> $staticNatList->{$b}->{order}} keys %{$staticNatList}){ # iterate filterlists in order
			print_filter_lists($out, $key, $staticNatList, 1);
		}
		$out->print_element("phaseDiscriminator", {processingPhase => {'core:packetProcessingPhase' => "routing"}, processingPhasePosition => "after"});
		$out->close_element("filterListsSection");
	}
	
	$out->close_element("filterLists");
}

sub build_NAT_structure {
	my ($filterEntry, $translationType, $addressType) = @_;
	
	$filterEntry->{$translationType} = {translateAppliedAddressing => 'true', translationAddress => {translationIP => {objectGroupReference => $filterEntry->{$addressType}}}};
	if (defined $filterEntry->{nat} && defined $filterEntry->{redirPort} && $addressType eq 'redirAddr'){ # if dst PAT
		$filterEntry->{$translationType}->{translationAddress}->{translationServicePort} = {translationPort => {portExpression => {port => $filterEntry->{redirPort}, operator => 'eq'}}, translationProtocol => 'tcp'};
	}else{
		if ($filterEntry->{natMode} ne 'src_port'){ # else if not set to preserve source port
			$filterEntry->{$translationType}->{translationAddress}->{translationServicePort} = {translationPort => {portRange => {portStart => 1024, portEnd => 65535}}, translationProtocol => 'tcp'};
		}
	}
		
	return $filterEntry;
}

sub parse_static_nat {
	my $in = shift;
	my @staticNatArray = parse_firewall_policies( $in, "staticNatOnly" );

	my @staticNatReverseArray;
	foreach my $filterEntry (@staticNatArray) {
		if(defined $filterEntry->{natAddr} && $filterEntry->{natAddr} ne 'any'){ # src nat translation
			$filterEntry = build_NAT_structure($filterEntry, "srcAddrTranslation", "natAddr");
		}
		if(defined $filterEntry->{redirAddr} && $filterEntry->{redirAddr} ne 'any'){ # dst nat translation
			$filterEntry = build_NAT_structure($filterEntry, "dstAddrTranslation", "redirAddr");
		}
		# create reverse filter
		my $reverseFilter = get_reverse_nat_filter($filterEntry, 1);
		if (defined $reverseFilter){
			if (defined $filterEntry->{redirAddr}){
				@{ $reverseFilter->{srcAddresses} } = {objectGroupReference => $filterEntry->{redirAddr}};
			}
			if (defined $filterEntry->{natAddr}){
				@{ $reverseFilter->{dstAddresses} } = {objectGroupReference => $filterEntry->{natAddr}};
			}
			
			push @staticNatReverseArray, $reverseFilter;
		}	
	}
	
	my @mergedStaticNat = (@staticNatArray, @staticNatReverseArray);
	return @mergedStaticNat;
}

sub get_reverse_nat_filter {
	my ($filterEntry, $includePAT) = @_;
	my $reverseFilter;
	
	$reverseFilter = get_nat_addresses($filterEntry, $reverseFilter, "src", "dst");
	$reverseFilter = get_nat_addresses($filterEntry, $reverseFilter, "dst", "src");
	
	if (defined $reverseFilter){
		$reverseFilter->{sourceServices}        = $filterEntry->{dstAddrTranslation};
		$reverseFilter->{srcZones}              = $filterEntry->{dstZones};
		$reverseFilter->{dstZones}              = $filterEntry->{srcZones};
		$reverseFilter->{natReverse}            = 1;
		$reverseFilter->{log}                   = $filterEntry->{log};
		$reverseFilter->{name}                  = $filterEntry->{name};
		$reverseFilter->{primaryAction}         = $filterEntry->{primaryAction};
		$reverseFilter->{processOrder}          = $filterEntry->{processOrder};
		$reverseFilter->{description}           = $filterEntry->{description};
	}
	
	return $reverseFilter;
}

sub get_nat_addresses {
	my ($filterEntry, $reverseFilter, $addressType, $reverseAddressType) = @_;
	
	if (defined $filterEntry->{$addressType."AddrTranslation"}){
		foreach my $address (@{$filterEntry->{$reverseAddressType."Addresses"}}){
			if ($address eq 'any' || $address eq 'all'){
				return undef;
			}
			push @{ $reverseFilter->{$addressType."Addresses"} }, {objectGroupReference => $address};
		}
		foreach my $address (@{$filterEntry->{$addressType."Addresses"}}){
			if ($address eq 'any' || $address eq 'all'){
				return undef;
			}
			my $translationIP = {objectGroupReference => $address};
			push @{ $reverseFilter->{$reverseAddressType."AddrTranslation"}->{translationAddress}->{translationIP} }, $translationIP;
		}
		if (defined $filterEntry->{nat} && defined $filterEntry->{natAddr} || (defined $filterEntry->{redirPort} && not defined $filterEntry->{redirAddr})){ # if PAT
			$reverseFilter->{$reverseAddressType."AddrTranslation"}->{translationAddress}->{translationServicePort} = {translationPort => {portRange => {portStart => 1024, portEnd => 65535}}, translationProtocol => 'tcp'};
		}elsif (defined $filterEntry->{nat}){
			foreach my $portExp(@{$filterEntry->{portExpressions}}){
				my $translationPort = {translationPort => {portExpression => {port => $portExp->{port}, protocol => $portExp->{protocol}, operator => $portExp->{operator}}}, translationProtocol => 'tcp'};
				push @{ $reverseFilter->{$reverseAddressType."AddrTranslation"}->{translationAddress}->{translationServicePort} }, $translationPort;
			}
			foreach my $portRange(@{$filterEntry->{portRanges}}){
				my $translationPort = {translationPort => {portRange => {portStart => $portRange->{portStart}, portEnd => $portRange->{portEnd}}}, translationProtocol => 'tcp'};
				push @{ $reverseFilter->{$reverseAddressType."AddrTranslation"}->{translationAddress}->{translationServicePort} }, $translationPort;
			}
		}
		$reverseFilter->{$reverseAddressType."AddrTranslation"}->{translateAppliedAddressing} = 'true';
	}
	
	return $reverseFilter;
}

sub parse_snmp
{
	my ( $in, $out ) = @_;
}

sub parse_static_routes
{
	my ( $in, $out ) = @_;
	
	my $staticRoutes;
	my $routeRegex = qr/^\s*($ip_regex)\/(\d+)\s+($ip_regex)\s+\S+\s+\S+\s+([\S ]+\S+)/;
	my $defaultRouteRegex = qr/^\s*default\s+($ip_regex)\s+\S+\s+\S+\s+([\S ]+\S+)/;

	foreach my $line (split("\n", $in->{routingTable})) {
		if ($line =~ $routeRegex){
			my $route = {};
			$route->{'defaultGateway'} = 'false';
			$route->{'destinationAddress'} = $1;
			$route->{'destinationMask'} = $2;
			$route->{'gatewayAddress'} = $3;
			$route->{'interface'} = $4;
			# Routing table output includes interfaces - this check avoids adding interfaces as routes.
			if ($route->{'destinationAddress'} ne $route->{'gatewayAddress'}) {
				push( @{ $staticRoutes->{staticRoute} }, $route );
			}
		}elsif ($line =~ $defaultRouteRegex){
			my $route = {};
			$route->{'defaultGateway'} = 'true';
			$route->{'destinationAddress'} = '0.0.0.0';
			$route->{'destinationMask'} = '0';
			$route->{'gatewayAddress'} = $1;
			$route->{'interface'} = $2;
			push( @{ $staticRoutes->{staticRoute} }, $route );
		}
	}
	
	$out->print_element('staticRoutes', $staticRoutes);
}

sub parse_object_groups{
	my ( $in, $out ) = @_;
	
	# collect and parse network objects
	my @addresses = parse_addresses($in);
	my @ranges = parse_address_ranges($in);
	my @subnets = parse_subnets($in);
	my @domains = parse_domains($in);
	my @hosts = parse_hosts($in);
	my @networkGroups = (@addresses, @ranges, @subnets, @domains, @hosts);
	parse_firewall_object_groups($in, "networkGroups", \@networkGroups);
	parse_firewall_object_groups($in, "networkMaps", \@networkGroups);

	# parse all service objects and generate service groups
	my @serviceGroups = parse_predefined_services($in);
	parse_custom_services($in, \@serviceGroups);
	parse_firewall_object_groups($in, "serviceGroups", \@serviceGroups);
	
	# print all generated object groups
	$out->open_element('objectGroups');
	print_network_groups($out, \@networkGroups);
	print_service_groups ($out, \@serviceGroups);
	$out->close_element('objectGroups');
	
	# populate filter required for NAT later (sourceService and destService ports)
	foreach my $service (@serviceGroups){
		foreach my $portExp (@{$service->{portExpressions}}){
			if (defined $portExp){
				push @{$serviceGroupsNAT{$service->{id}}{portExpressions}}, $portExp;
			}
		}
		foreach my $portRange (@{$service->{portRanges}}){
			if (defined $portRange){
				push @{$serviceGroupsNAT{$service->{id}}{portRanges}}, $portRange;
			}
		}
	}
}

sub print_network_groups {
	my $out               = shift;
	my $networkGroupArray = shift;

	foreach my $networkObject (@$networkGroupArray){
		$out->open_element('networkGroup');
		if (defined $networkObject->{description}){
			$out->print_element('description', $networkObject->{description});
		}
		$out->print_element('id', $networkObject->{id});
		if (defined $networkObject->{subnet}){
			$out->print_element('subnet', $networkObject->{subnet});
		}elsif (defined $networkObject->{range}){
			$out->print_element('range', $networkObject->{range});
		}elsif (defined $networkObject->{host}){
			$out->print_element('host', $networkObject->{host});
		}elsif (defined $networkObject->{domain}){
			foreach my $domain (@{$networkObject->{domain}}){
				$out->print_element('host', $domain);
			}
		}elsif (defined $networkObject->{hosts}){
			foreach my $host (@{$networkObject->{hosts}}){
				$out->print_element('host', $host);
			}
		}elsif (defined $networkObject->{members}){
			foreach my $member (@{$networkObject->{members}}){
				$out->print_element('objectGroupReference', $member);
			}
		}
		$out->close_element('networkGroup');
	}
	return $out;
}

sub parse_addresses {
	my ($in) = @_;
	my %networkGroup;
	my @networkGroupArray;
	
	foreach my $address(split("\n", $in->{addresses})){
		$address =~ s/(^\s+|\s+$)//g;
		if($address =~ /^ipaddr/i){
			$networkGroup{id} = get_value("name", $address);
			if($address =~ /ipaddr=($ip_regex)/i){
				$networkGroup{host} = $1;
			}
			if ($address =~ /description='(\w+(\s+\w+){0,})'/) {
	        	$networkGroup{description} = $1;
	        }
		}
		if (exists $networkGroup{id}) {
			push @networkGroupArray, {%networkGroup};
			%networkGroup = ();
		}
	}

    return @networkGroupArray;
}

sub parse_address_ranges {
	my ($in) = @_;
	my %networkGroup;
	my @networkGroupArray;
	
	foreach my $iprange(split("\n", $in->{ipranges})){
		$iprange =~ s/(^\s+|\s+$)//g;
		if($iprange =~ /^iprange/i){
			$networkGroup{id} = get_value("name", $iprange);
			if($iprange =~ /begin=($ip_regex)/i){
				$networkGroup{range}{startAddress} = $1;
			}
			if($iprange =~ /end=($ip_regex)/i){
				$networkGroup{range}{endAddress} = $1;
			}
			if ($iprange =~ /description='(\w+(\s+\w+){0,})'/) {
	        	$networkGroup{description} = $1;
	        }
		}
		if (exists $networkGroup{id}) {
			push @networkGroupArray, {%networkGroup};
			%networkGroup = ();
		}
	}

    return @networkGroupArray;
}

sub parse_subnets {
	my ($in) = @_;
	my %networkGroup;
	my @networkGroupArray;
	
	foreach my $subnet(split("\n", $in->{subnets})){
		$subnet =~ s/(^\s+|\s+$)//g;
		if($subnet =~ /^subnet/i){
			$networkGroup{id} = get_value("name", $subnet);
			if($subnet =~ /subnet=($ip_regex)/i){
				$networkGroup{subnet}{address} = $1;
			}
			if($subnet =~ /bits=(\d+)/i){
				$networkGroup{subnet}{mask} = $1;
			}
			if ($subnet =~ /description='(\w+(\s+\w+){0,})'/) {
	        	$networkGroup{description} = $1;
	        }
		}
		if (exists $networkGroup{id}) {
			push @networkGroupArray, {%networkGroup};
			%networkGroup = ();
		}
	}

    return @networkGroupArray;
}

sub parse_domains {
	my ($in) = @_;
	my %networkGroup;
	my @domainList;
	my @networkGroupArray;
	
	foreach my $domain ( keys %{ $in->{domain} } ) {
		my $domainResponse = $in->{domain}{$domain};
		my @addressList;
		$networkGroup{id} = $domain; 
		if ($domainResponse =~ /description='(\w+(\s+\w+){0,})'/) {
        	$networkGroup{description} = $1;
        }
		foreach my $line(split("\n", $domainResponse)){
			$line =~ s/(^\s+|\s+$)//g;
			if($line =~ /A\s+($ip_regex)/i){
				push @addressList, $1;
			}
		}
		if (exists $networkGroup{id}) {
			push @{ $networkGroup{domain} }, @addressList;
			push @networkGroupArray, {%networkGroup};
			%networkGroup = ();
		}
	}
	
    return @networkGroupArray;
}

sub parse_hosts {
	my ($in) = @_;
	my %networkGroup;
	my @networkGroupArray;
	my $cleanedNetworkResponse = clean_response($in->{hosts});
	
	foreach my $host(split("\n", $cleanedNetworkResponse)){
		my @addressList;
		$host =~ s/(^\s+|\s+$)//g;
		if($host =~ /^host/i){
			$networkGroup{id} = get_value("name", $host);
			if ($host =~ /description='(\w+(\s+\w+){0,})'/) {
	        	$networkGroup{description} = $1;
	        }
			if($host =~ /ipaddrs=(\S+)/i){
				@addressList = split(/\,/, $1);
			}
		}
		if (exists $networkGroup{id}) {
			push @{ $networkGroup{hosts} }, @addressList;
			push @networkGroupArray, {%networkGroup};
			%networkGroup = ();
		}
	}
	
	# populate pre-defined host addresses
	push @networkGroupArray, {id => 'localhost', host => '127.0.0.1'};
	push @networkGroupArray, {id => 'ePO server', host => '127.0.0.1'};
	push @networkGroupArray, {id => 'CommandCenter servers', host => '127.0.0.1'};
	push @networkGroupArray, {id => 'Firewall', host => '127.0.0.1'};
	push @networkGroupArray, {id => 'SmartFilter server', host => '127.0.0.1'};
	
	return @networkGroupArray;
}

sub parse_predefined_services {
	my ( $in, $out ) = @_;
	my @serviceGroupArray;
	my $cleanedServiceResponse = clean_response($in->{predefinedServices});
	my @predefinedServices = split ("\n\n", $cleanedServiceResponse);
	
	foreach my $preService(@predefinedServices){
		my %service;
		foreach my $line (split("\n", $preService)){
			$line =~ s/(^\s+|\s+$)//g;
			if($line =~ /^(.+)\:\s*(.+)/i){
				my ($key, $val) = split(": ", $line, 2);
				$service{$key} = $val;
			}
		}
		
		unless (defined $service{Name}) {
			next;
		}
		
		my %serviceGroup;
		$serviceGroup{id} = $service{Name};
		
		if(exists $service{'TCP ports'}){
			%serviceGroup = get_port_lists(\%serviceGroup, $service{'TCP ports'}, 'tcp')
		}
		if(exists $service{'SSL (TCP) ports'}){
			%serviceGroup = get_port_lists(\%serviceGroup, $service{'SSL (TCP) ports'}, 'tcp')
		}
		if(exists $service{'UDP ports'}){
			%serviceGroup = get_port_lists(\%serviceGroup, $service{'UDP ports'}, 'udp')
		}
		if(exists $service{'Other IP protos'}){
			my $portExp = {protocol => 'ip', operator => 'eq', port => $service{'Other IP protos'}};
			push @{$serviceGroup{portExpressions}}, $portExp;
		}
		
		if (exists $service{description}) {
        	$serviceGroup{description} = $service{description};
        }
		
		push @serviceGroupArray, \%serviceGroup;
	}
	
	return @serviceGroupArray;
}

sub get_ports {
	my $ports = shift;
	my $protocol = shift;
	my @portExpressions;
	my @portRanges;
	
	# parse list of ports/port ranges and push to array
	if($ports =~ /\d+\,+/){
		my @portList = split(",", $ports);
		foreach my $port (@portList){
			if($port =~ /(\d+)-(\d+)/){
				my $portRange = {protocol => $protocol, portStart => $1, portEnd => $2};
				if($portRange->{portStart} =~ /^\d+?$/ && $portRange->{portEnd} =~ /^\d+?$/){
					push @portRanges, $portRange;
				}
			}else{
				my $portExp = {protocol => $protocol, operator => 'eq', port => $port};
				if($portExp->{port} =~ /^\d+?$/){
					push @portExpressions, $portExp;
				}
			}
		}
	# else parse single port or port range and push to array
	}elsif($ports =~ /\d+/){
		if($ports =~ /(\d+)-(\d+)/){
			my $portRange = {protocol => $protocol, portStart => $1, portEnd => $2};
			if($portRange->{portStart} =~ /^\d+?$/ && $portRange->{portEnd} =~ /^\d+?$/){
				push @portRanges, $portRange;
			}
		}else{
			my $portExp = {protocol => $protocol, operator => 'eq', port => $ports};
			if($portExp->{port} =~ /^\d+?$/){
				push @portExpressions, $portExp;
			}
		}
	}
	
	return (\@portExpressions, \@portRanges);
}

sub get_port_lists {
	my ($serviceGroupRef, $ports, $protocol) = @_;
	my %serviceGroup = %$serviceGroupRef;
	
	my ($portExpList, $portRangeList) = get_ports($ports, $protocol);
	foreach my $portExp(@$portExpList){
		push @{$serviceGroup{portExpressions}}, $portExp;
	}
	foreach my $portRange(@$portRangeList){
		push @{$serviceGroup{portRanges}}, $portRange;
	}
	
	return %serviceGroup;
}

sub parse_custom_services {
	my $in = shift;
	my $serviceGroupsArray = shift;
	my $cleanedServiceResponse = clean_response($in->{customServices});
	
	my %serviceGroup;
	foreach my $service ( split( "\n", $cleanedServiceResponse ) ) {
		$service =~ s/(^\s+|\s+$)//g;
		if ($service =~ /^application\s(modify|add)/i){
			$serviceGroup{id} = get_value("name", $service);
			if ($service =~ /description='(\w+(\s+\w+){0,})'/) {
	        	$serviceGroup{description} = $1;
	        }
			if ( $service =~ /tcp_ports=([\d,]+)/ ) {
				%serviceGroup = get_port_lists(\%serviceGroup, $1, 'tcp');
			}
			if ( $service =~ /ssl_ports=([\d,]+)/ ){
				%serviceGroup = get_port_lists(\%serviceGroup, $1, 'tcp');
			}
			if ( $service =~ /udp_ports=([\d,]+)/ ){
				%serviceGroup = get_port_lists(\%serviceGroup, $1, 'udp');
			}
			push @{$serviceGroupsArray}, {%serviceGroup};
			%serviceGroup = ();
			next;
		}
	}
}

sub print_service_groups {
	my $out               = shift;
	my $serviceGroupArray = shift;

	foreach my $serviceGroup (@$serviceGroupArray) {
		$out->open_element('serviceGroup');
		if (defined $serviceGroup->{description}){
			$out->print_element('description', $serviceGroup->{description});
		}
		$out->print_element('id', $serviceGroup->{id});
		if (defined $serviceGroup->{members}){
			foreach my $member (@{$serviceGroup->{members}}){
				$out->print_element('objectGroupReference', $member);
			}
		}
		print_ports_and_ranges($serviceGroup, 'serviceGroup', $out);
		$out->close_element('serviceGroup');
	}
}

sub parse_firewall_object_groups {
	my ($in, $response, $fwObjectGroupArray) = @_;
	my %fwObjectGroup;
	my $netmap = 0;
	my $cleanedResponse = clean_response($in->{$response});
	
	foreach my $fwObject (split( "\n", $cleanedResponse)) {
		$fwObject =~ s/(^\s+|\s+$)//g;
		if ($fwObject =~ /^(app|net)(group|map)/i) {
			$fwObjectGroup{id} = get_value("name", $fwObject);
			if($fwObject =~ /^netmap/i){
				$netmap = 1;
			}
		}
		if ($fwObject =~ /description='(\w+(\s+\w+){0,})'/) {
        	$fwObjectGroup{description} = $1;
        }
		
		my @members;
		my @memberNames;
		
		if($netmap == 0){
			if($fwObject =~ /members=\'(.*?)\'/ || $fwObject =~ /members=(\S+)/){
				@members = split(/\,/, $1);
			}
			foreach my $member (@members){
				my ($key, $value) = split(/\:/, $member);
				push (@memberNames, $value);
			}
		}else{
			if($fwObject =~ /members=\'(.*?)\'/ || $fwObject =~ /members=(\S+)/){
				@members = split(/\:/, $1);
			}
			foreach my $value(@members){
				if ($value !~ /subnet|ipaddr/){
					push (@memberNames, $value);
				}
			}
		}
		
		push @{ $fwObjectGroup{members} }, @memberNames;
		
		if (defined $fwObjectGroup{id}) {
			push @{$fwObjectGroupArray}, {%fwObjectGroup};
			%fwObjectGroup = ();
		}
	}
}

sub parse_zones 
{
	my $in = shift;
	unless (defined  $in->{zones} ){
		return;
	}
	
	my @parsed_zones;
	my $cleaned_zones = clean_response($in->{zones});
	while ($cleaned_zones =~ /^zone.+\s+\S+='.*'/gm) {
		my $raw_zone = $&;
		my $zone_name;
		my $zone_groups;
		
		if ($raw_zone =~ /name=(\S+)/) {
			$zone_name = $1;
		}
		if ($raw_zone =~ /zonegroups=(\S+)/) {
			$zone_groups = $1;
		}
		
		$in->{parsed_zones}->{$zone_name}{zone_groups} = $zone_groups;
	}
}

sub parse_interfaces
{
	my ( $in, $out ) = @_;
	
	parse_zones($in);
	my @modelled_interfaces;
	
	# for each interface
	my $cleanedInterfaces = clean_response($in->{interfaces});
	while ($cleanedInterfaces =~ /^interface.+(\s+\S+=(')?.*(')?)*/gm) {
		my $raw_interface = $&;
		my $modelled_interface;
		my $interface_name;

        # gather info
        if ($raw_interface =~ /entrytype=(\w+)/) {
        	my $entrytype = $1;
        	$modelled_interface->{entrytype} = $entrytype;
        	if ($entrytype eq 'nic_group') {
        		next;
        	}
        	elsif ($entrytype eq 'vlan') {
        		$modelled_interface->{physical} = 'false';
        		$modelled_interface->{interfaceType} = 'virtual';
        	} else {
        		$modelled_interface->{physical} = 'true';
        		$modelled_interface->{interfaceType} = 'ethernet';
        	}
        }
        if ($raw_interface =~ /name=(\w+)/) {
        	$interface_name = $1;
        	$modelled_interface->{name} = $interface_name;
        }
        if ($raw_interface =~ /hwdevice=(\w+)/) {
        	$modelled_interface->{hwdevice} = $1;
        }
        if ($raw_interface =~ /mac_addr=(\w+)/) {
        	$modelled_interface->{interfaceEthernet}->{macAddress} = $1;
        }
        
        if ($raw_interface =~ /addresses=(\d{1,3}(?:\.\d{1,3}){0,3})\/(\d+)/) {
        	my ($ip, $mask) = ($1, $2);
            my $ip_cfg = {ipAddress => $ip,};
            $ip_cfg->{mask} = mask_to_bits($mask);
            push(@{ $modelled_interface->{interfaceIp}->{ipConfiguration} }, $ip_cfg);
        }
        # Only 'enabled=yes' results in 'adminStatus=up'
        if ($raw_interface =~ /enabled=(\w+)/) {
        	if ($1 eq 'yes') { $modelled_interface->{adminStatus} = 'up'; } 
        } else { 
        	$modelled_interface->{adminStatus} = 'down'; 
        }
        if ($raw_interface =~ /mtu=(\d+)/) {
        	$modelled_interface->{mtu} = $1;
        }
        if ($raw_interface =~ /description='(.+)'/) {
        	$modelled_interface->{description} = $1;
        }
        if ($raw_interface =~ /zone=(\S+)/) {
        	my $zone_name = $1;
        	my @interface_groupings = _get_interface_groupings($in, $zone_name);
			$modelled_interface->{interfaceGrouping} = \@interface_groupings;			
        }
        if ($raw_interface =~ /vlan_id=(\d+)/) {
        	my $vlan_id = $1;
        	$modelled_interface->{interfaceVlanTrunks}->{startVlan} = $vlan_id;
        	
        	# Store vlans for later
        	if ($in->{vlans}->{$vlan_id}->{interfaces}) {
        		$in->{vlans}->{$vlan_id}->{interfaces} .= ',' . $interface_name;
        	} else {
        		$in->{vlans}->{$vlan_id}->{interfaces} = $interface_name;
        	}
        }
        
        # Determine speed
        if ($raw_interface =~ /iftype=(\w+)/) {
        	my $mediaType = $1;
			if ($mediaType ne 'autoselect') {
				if ($mediaType =~ /(^\d+)/) {
					$modelled_interface->{speed} = getUnitFreeNumber($1, 'M', 'bit');							
				}
			}
        }
        
        push(@modelled_interfaces, $modelled_interface);
	}
	
	# Check for duplicates
	my @final_interfaces;
	# Compare all interface elements to each other
  	foreach my $interfaceX (@modelled_interfaces) {
  		my $isMatch = 0;
  		foreach my $interfaceY (@modelled_interfaces) {
  			# Check for duplicate references to NICs
  			if (defined $interfaceY->{hwdevice}) {
				if (($interfaceX->{name} eq $interfaceY->{hwdevice}) && ($interfaceY->{entrytype} ne 'vlan')) {
					$interfaceY->{interfaceEthernet}->{macAddress} = $interfaceX->{interfaceEthernet}->{macAddress};
					if ($interfaceX->{speed}) { $interfaceY->{speed} = $interfaceX->{speed}; }
					$isMatch = 1;
				}
  			}
  		}
  		
  		delete $interfaceX->{entrytype};
  		delete $interfaceX->{hwdevice};
  		# Only add NICs not already in use therefore avoiding duplicates
		if (!$isMatch) { push(@final_interfaces, $interfaceX); }
  	}
	$out->print_element('interfaces', {interface=> \@final_interfaces});
}

sub _get_interface_groupings {
	
	my $in = shift;
	my $zone_name = shift;
	my @interface_groupings;
	
	foreach my $zone ( keys %{ $in->{parsed_zones} } ) {
		# Verify zone referenced in interface and add it
    	if ($zone_name eq $zone) {
       		my $interface_grouping;
       		$interface_grouping->{ifGroupName} = $zone_name;
			$interface_grouping->{ifGroupingCharacteristic} = 'securityZoneMember';
			push (@interface_groupings, $interface_grouping);
			
			# Add any zone groups that contain this zone
			if ($in->{parsed_zones}->{$zone}->{zone_groups}){
				my @zone_groups = split(',', $in->{parsed_zones}->{$zone}->{zone_groups});
				foreach my $zone_group (@zone_groups) {
					my $interface_grouping;
					$interface_grouping->{ifGroupName} = $zone_group;
					$interface_grouping->{ifGroupingCharacteristic} = 'securityZoneMember';
					push (@interface_groupings, $interface_grouping);
				}	 
	    	}
       	}
    }
    return @interface_groupings;
}

sub parse_vlans
{
	my ( $in, $out ) = @_;
	unless (defined $in->{vlans}){
		return;
	}
	
	$out->open_element('vlans');
	foreach my $vlan (keys %{ $in->{vlans} }) {
		$out->open_element('vlan');
		$out->print_element('enabled', 'true');
		foreach my $interface (split(',' , $in->{vlans}->{$vlan}->{interfaces})){
			$out->print_element('interfaceMember', $interface);
		}
		$out->print_element('number', $vlan);
		$out->close_element('vlan');
	}
	$out->close_element('vlans');
}

sub parse_stp
{
	my ( $in, $out ) = @_;
}

sub clean_response
{
	my ( $response ) = @_;
	
		$response =~ s/\\\n//g;
		$response =~ s/\\\r\n//g;
		$response =~ s/\r\n/\n/g;
	
	return $response;
}

sub _generate_packet_processing_declaration
{
	my $out = shift;
	
	$out->open_element("core:packetProcessingOrdering");
	
	foreach my $pRPhase ('packet-filtering/source','routing','packet-filtering/destination','address-translation' )
	{
		my ($pPhase, $pPosition) = split ('/', $pRPhase);
		$out->open_element( 'core:packetProcessingElement');
		$out->print_element( 'core:packetProcessingPhase', $pPhase );
		if (defined $pPosition)
		{
			$out->print_element( 'core:processingPoint', $pPosition);
		}		
		$out->close_element( 'core:packetProcessingElement');
	}
	$out->close_element("core:packetProcessingOrdering");
}

1;
