#!/usr/bin/perl

use strict;
use warnings;

use English qw( -no_match_vars );    #Friendly names for punctuation variables
use Data::Dumper;

#use ZipTie::Client;
use ZipTie::ConnectionPath;
use ZipTie::ConnectionPath::Protocol;
use ZipTie::Credentials;
use ZipTie::Recording;
use ZipTie::Logger;

use ZipTie::Adapters::PaloAlto::PANOS::ParsingUtils qw(get_value_from_xml);
use ZipTie::Adapters::PaloAlto::PANOS::AutoLogin;
my $adapter_name = 'ZipTie::Adapters::PaloAlto::PANOS';

# Set up logging
my $LOGGER = ZipTie::Logger::get_logger();

execute_discovery( $ARGV[0] );

sub execute_discovery {
	my $management_server_ip = shift;

	#Enable logging to file
	if ( $ENV{ADAPTER_LOG_TO_FILE} ) {
		$LOGGER->enable_logging_to_file( 'PANOS_Panorama', 'PanoramaDiscovery', $management_server_ip );
	}

	#check to see if the ip address is valid
	if ( $management_server_ip !~ /\d+[.]\d+[.]\d+[.]\d+/ ) {
		$LOGGER->fatal("ERROR: Invalid IP address: $management_server_ip");
	}
	$LOGGER->debug("Using IP: $management_server_ip");

	#obtain a connection to the Ziptie API
	my $ziptie_client = ZipTie::Client->new();

	### SEARCH FOR EXISTING DEVICES IN THE ZIPTIE DB ###

	#initialize the devicesearch service
	my $device_search_service = $ziptie_client->devicesearch();

############################################ DO THIS BETTER ###################################
	my $total    = undef;
	my $page_size = 500;
	my $offset   = 0;
	my $existing_devices;

	while ( !defined($total) || ( $offset < $total ) ) {
		my $page = $device_search_service->search(
			scheme   => 'ipAddress',
			query    => q{},
			pageData => { pageSize => $page_size, offset => $offset }
		);
		$offset += $page_size;

		my $devices = [];
		if ( ref( $page->{devices} ) eq 'ARRAY' ) {
			$devices = $page->{devices};
		}
		elsif ( defined( $page->{devices} ) ) {
			$devices = [ $page->{devices} ];
		}

		foreach my $device ( @{$devices} ) {
			$existing_devices->{ $device->{ipAddress} } = $device;
		}

		# Total is only returned for the first search
		if ( !defined $total ) {
			$total = $page->{total};
		}
	}
############################################ DO THIS BETTER ###################################

	### CONNECT TO THE MANAGEMENT SERVER AND COLLECT INFORMATION ON MANAGED DEVICES ###

	#create a connection path object to use when connecting to the device
	my $connection_path = get_connection_path( $ziptie_client, $management_server_ip );

	#my $test = $connection_path->to_xml_string(1);
	#$LOGGER->debug("==================\n\n$test\n\n==================\n\n");

	#Log in to the device
	my $connection = ZipTie::Adapters::PaloAlto::PANOS::AutoLogin->new($connection_path);
	$connection->establish_connection();

	#issue the show device connected command
	my $eval_success = 1;
	my ($show_device_connected, $show_system_info);

	$eval_success = eval {
		$show_device_connected = $connection->issue_command( '<show><device><connected></connected></device></show>', 'show device connected' );
		1;
	};
	if ( !$eval_success ) {
		$LOGGER->fatal("Failed to issue 'show device all' command: $EVAL_ERROR");
	}

    $eval_success = eval {
        $show_system_info = $connection->issue_command( '<show><system><info></info></system></show>', 'show system info' );
        1;
    };
    if ( !$eval_success ) {
        $LOGGER->fatal("Failed to issue 'show system info' command: $EVAL_ERROR");
    }

	#parse the ip addresses of the managed devices from the output of the show device connected command
	my @device_ips = get_device_ips($show_device_connected);

	my $host_name = get_value_from_xml( '/response/result/system/hostname', $show_system_info );

	### ADD NEW DEVICES TO THE DB OR UPDATE EXISTING DEVICES ###

	#initialize the devices service
	my $device_service = $ziptie_client->devices();

	#go through the list of devices we found
	foreach my $device_ip (@device_ips) {

		#make a representation of the device
		my $device = {
			ipAddress          => $device_ip,
			adapterId          => $adapter_name,
			managedNetwork     => $host_name,
			managementServerIp => $management_server_ip
		};

		# Update the device if it laready exists
		if ( defined( $existing_devices->{$device_ip} ) ) {

			my $existing_ip = $existing_devices->{$device_ip}->{ipAddress};
			my $existing_network = $existing_devices->{$device_ip}->{managedNetwork};

			$device_service->updateDevice( ipAddress => $existing_ip, managedNetwork => $existing_network, device => $device );
			$existing_devices->{$device_ip} = $device;

			$LOGGER->debug("Updating existing device $device_ip.");
		}

		# Add a new device if it does not already exist
		else {
			$device_service->createManagedDevice($device);
			$existing_devices->{$device_ip} = $device;

			$LOGGER->debug("Adding new device $device_ip.");
		}
	}    #end foreach device_ip

	$ziptie_client->logout();

	return;
}    # End execute_discovery

sub get_device_ips {
	my $show_device_connected = shift;

	return;
}

sub get_connection_path {
	my ( $ziptie_client, $ip_address ) = @_;

	#initialize the credentials service
	my $credentials_service = $ziptie_client->credentials();

	#get the credentials for the device
	my @credentials = $credentials_service->getCredentialSetsByIpAddress( ipAddress => $ip_address );

	if ( !scalar(@credentials) ) {
		$LOGGER->fatal('ERROR: No credentials found.');
	}

	#extract the username and password
	my $username = undef;
	my $password = undef;
	if ( ref( $credentials[0]->{credentials} ) ne 'ARRAY' ) {
		$credentials[0]->{credentials} = [ $credentials[0]->{credentials} ];
	}
	foreach my $credential ( @{ $credentials[0]->{credentials} } ) {
		if ( $credential->{name} eq 'username' ) {
			$username = $credential->{value};
		}
		elsif ( $credential->{name} eq 'password' ) {
			$password = $credential->{value};
		}
	}

	if ( !defined($username) || $username eq q{} ) {
		$LOGGER->fatal('ERROR: Username is empty.');
	}

	if ( !defined($password) || $password eq q{} ) {
		$LOGGER->fatal('ERROR: Password is empty.');
	}

	#extract the port number to be used
	my @protocols = $credentials_service->getProtocolConfigByIpAddress( ipAddress => $ip_address );

	if ( !scalar(@protocols) ) {
		$LOGGER->fatal('ERROR: No protocols found.');
	}

	my $port = undef;
	foreach my $protocol ( @{ $protocols[0]->{protocols} } ) {
		if ( $protocol->{name} eq 'HTTPS' && $protocol->{enabled} eq 'true' ) {
			$port = $protocol->{port};
			last;
		}
	}

	if ( !defined($port) || $port eq q{} ) {
		$LOGGER->fatal('ERROR: Required protocol either disabled or undefined.');
	}

	my $connection_path = ZipTie::ConnectionPath->new();

	$connection_path->set_credentials( { username => $username, password => $password } );
	$connection_path->set_ip_address($ip_address);

	my $protocol = ZipTie::ConnectionPath::Protocol->new();
	$protocol->set_name('HTTPS');
	$protocol->set_port($port);

	my @protocol_list;
	push @protocols, $protocol;

	$connection_path->set_protocols( \@protocol_list );

	return $connection_path;
}

1;

#
#my $connection_path;
#
#my $connection = ZipTie::Adapters::PaloAlto::PANOS::AutoLogin::establish_connection($connection_path);
