package ZipTie::Adapters::PaloAlto::PANOS::AutoLogin;

use strict;
use warnings;

use 5.0100000;

use Data::Dumper;
use Carp;
use English qw( -no_match_vars );	#Friendly names for punctuation variables
use Exporter 'import';

use XML::Twig::XPath;
use ZipTie::Adapters::PaloAlto::PANOS::Parsers qw(get_serial_number);
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::RESTAPI;

our @EXPORT_OK = qw(issue_command);

our $VERSION = 1.0;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

sub new {
	my $class_name	  = shift;
	my $connection_path = shift;

	my $this;

	$this->{connection_path}		  = $connection_path;
	$this->{use_management_server_ip} = 0;
	$this->{target_query_string}	  = q{};
	$this->{target_name_string}	   = q{};

	bless $this, $class_name;
	return $this;
}

sub establish_connection {
	my $this			= shift;
	my $connection_path = $this->{connection_path};

	my $management_server_ip = $connection_path->get_management_server_ip_address();
	if ( defined $management_server_ip ) {
		$this->{use_management_server_ip} = 1;
	}

	# Create a new REST API Object
	my $connection = ZipTie::RESTAPI->new( $connection_path, $this->{use_management_server_ip} );

	# Store all the information needed to connect to the device
	my $credentials = $connection_path->get_credentials();
	my $username	= $credentials->{username};
	my $password	= $credentials->{password};

	#set up the query string to retrieve a security token
	my $query_string = "api/?type=keygen&user=$username&password=$password";

	# make the first connection to the device to obtain a connection token.
	# provide a parser to transform the response content into a valid security token
	$connection->obtain_connection_token( $query_string, \&_parse_connection_key );

	#set the base URL for future commands
	$connection->set_api_location('api/?type=op&cmd=');

	#set up the target device if we are communicating through a management server
	if ( $this->{use_management_server_ip} ) {
	   $this->set_target_string();
	}

	$this->{connection} = $connection;

	#	return $connection;
	return;
}

sub set_target_string {
	my $this = shift;

	my $target_query_string = q{};
	my $target_name_string  = q{};
	my $eval_success		= 1;

	#issue the show devices command, and find the serial number of the target given the ip address of the target

	my $show_device_connected;

	$eval_success = eval {
		$show_device_connected = $this->issue_command( '<show><device><connected></connected></device></show>', 'show device connected' );
		1;	#Magic 'true' so that $eval_success is always set properly if the eval block finishes
	};
	if ( !$eval_success ) {
		$LOGGER->fatal( '"Show Device Connected" command failed. The device at IP "' . $this->{connection_path}->get_management_server_ip_address() . '" is not a recognized Panorama Management Server.' );
	}

	my $target_ip = $this->{connection_path}->get_ip_address();

	my $serial_number = get_serial_number( $show_device_connected, $target_ip );

	if ( defined $serial_number ) {
		$target_query_string = '&target=' . $serial_number;
		$target_name_string  = ' for target=' . $serial_number;
	}
	else {
		$LOGGER->fatal( "A device with IP Address '$target_ip' is not currently connected to the management server at '" . $this->{connection_path}->get_management_server_ip_address() . "'" );
	}

	$this->{target_query_string} = $target_query_string;
	$this->{target_name_string} = $target_name_string;

	return;
}

sub _parse_connection_key {
	my ($xml) = @_;

	# Check we got a valid response
	if($xml !~ /</){
		$LOGGER->fatal("Could not reach device: $xml\nCheck your credentials and protocol settings.\n");
	}

	#parse the response into an XML twig
	my $twig = XML::Twig::XPath->new();
	$twig->parse($xml);

	#the response XML document contains a 'status' attribute that should equal 'success'
	#if the request did not succeed die and provide the error message contained in the response
	my $attributes = $twig->root()->atts();
	if ( !( $attributes->{'status'} eq 'success' ) ) {
		my $msg = $twig->findvalue('/response/result/msg');
		$LOGGER->fatal("REST API CALL FAILED: $msg\n");
	}

	#the REST request may succeed, but the command issued may be invalid in some way
	#if this occurs then the response will contain an 'error' element.
	#die and provide the error message
	if ( $twig->findnodes('/response/result/error') ) {
		my $msg = $twig->findvalue('/response/result/error');
		$LOGGER->fatal("REST API CALL FAILED: $msg\n");
	}

	#format the security token which will be appended to future requests. Also format the fake authentication token that will be included in the recording
	return '&key=' . $twig->findvalue('/response/result/key');
}

sub issue_command {
	my $this	= shift;
	my $command = shift;
	my $name = shift;

	$command = $command . $this->{target_query_string};
	$name	= $name . $this->{target_name_string};

	my $eval_success = 1;
	my $response;
	$eval_success = eval {
		$response = $this->{connection}->get( $command, $name );
		1;	#Magic 'true' so that $eval_success is always set properly if the eval block finishes
	};
	if ( !$eval_success ) {
		croak("REST API CALL '$command' FAILED. Communication with the device failed: $EVAL_ERROR");
	}

	return _validate_response( $response, $command );

} #end sub issue_command

sub get_custom_url {
	my $this = shift;
	my $url  = shift;
	my $name = shift;

	$url  = $url . $this->{target_query_string};
	$name = $name . $this->{target_name_string};

	######need to validate responses here
	my $eval_success = 1;
	my $response;
	$eval_success = eval {
	   $response = $this->{connection}->get_custom_url( $url, $name );
		1;	#Magic 'true' so that $eval_success is always set properly if the eval block finishes
	};
	if ( !$eval_success ) {
		croak("REST API CALL '$url' FAILED. Communication with the device failed: $EVAL_ERROR");
	}

	return _validate_response( $response, $url );

} #end sub get_custom_url

sub _validate_response {
	my ($xml, $command) = @_;

	my $eval_success = 1;
	my $twig;
	$eval_success = eval {
		$twig = XML::Twig::XPath->new()->parse($xml);
		1;	#Magic 'true' so that $eval_success is always set properly if the eval block finishes
	};

	if ( !$eval_success ) {
		croak("REST API CALL '$command' FAILED due to XML parsing error: $EVAL_ERROR");
	}

	#the response XML document contains a 'status' attribute that should equal 'success'
	#if the request did not succeed die and provide the error message contained in the response
	my $attributes = $twig->root()->atts();
	if ( !( $attributes->{'status'} eq 'success' ) ) {
		my $msg = $twig->findvalue('/response/result/msg');
		croak("REST API CALL '$command' FAILED: $msg");

		#the REST request may succeed, but the command issued may be invalid in some way
		#if this occurs then the response will contain an 'error' element.
		#die and provide the error message
	}
	elsif ( $twig->findnodes('/response/result/error') ) {
		my $msg = $twig->findvalue('/response/result/error');
		croak("REST API CALL '$command' FAILED: $msg");
	}

	return $twig;
}

1;

__END__

=head1 NAME

ZipTie::Adapters::PaloAlto::PANOS::AutoLogin - Automates the connection and authentication with a PaloAlto PANOS Firewall.

=head1 VERSION

This document describes ZipTie::Adapters::PaloAlto::PANOS::AutoLogin version 1.0

=head1 SYNOPSIS

	use ZipTie::Adapters::PaloAlto::PANOS::AutoLogin;
	establish_connection ($connection_path);

=head1 DESCRIPTION

This module provides the functionallity to authenticate against a Palo Alto Firewall. It prepares a RESTAPI object using a connectionPath
object that allows future REST API commands to be issued.

=head1 SUBROUTINES/METHODS

=over 12


=item C<establish_connection($connection_path)>

This function takes a ZipTie::ConnectionPath object and uses it to create a ZipTie::RESTAPI object. It then sends an initial
authentication request and sets up RESTAPI object so that further commands can be issued.
It returns the ZipTie::RESTAPI object to the caller.

=item C<issue_command($command, $connection)>

This function makes a call to ZipTie::RESTAPI::get() and parses the result into an XML::Twig::XPath
object. It exists to make the issuing of commands to a device cleaner in the backup operation.

It also checks the response to see if the command succeeded. If the command failed for some reason there are two options:
If the command is required for the backup to proceed then a fatal error is issued. If the command is only used for optional
elements then a debug warning is issued and the response is set to undef to indicate that it should not be used.


=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_parse_connection_key($xml, $fake_authentication_token)>

This function is passed by reference to ZipTie::RESTAPI::connect() so that the it can parse the authentication response from the device.
It checks to see if the XML response was valid, and if it was it constructs the authentication key query string for the RESTAPI to use.
It also constructs a fake authentication token based on the fake token passed by RESTAPI.pm. This is used when recordings are enabled
in order to prevent the real token from appearing in the recording.


=back

=head1 DIAGNOSTICS

=over 12

=item C<REST API CALL '$command' FAILED: $msg\n>

This message indicates that the command '$command' issued to the device failed for some reason. Any message returned by the device
is included as '$msg'. If this message relates to a B<REQUIRED> command then it will result in termination of the backup operation.
If it is not a required command then the backup can continue, although reduced functionallity may be experienced.

=back

=head1 CONFIGURATION AND ENVIRONMENT

N/A

=head1 DEPENDENCIES

This module requires a specific version (which one?) of ziptie-common and ziptie-server.

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

N/A

=head1 LICENSE AND COPYRIGHT

Copyright IBM 2013

=head1 AUTHOR

Contributor(s): David Tickner dtickner@ca.ibm.com

Date: October 2013

=cut