package ZipTie::Adapters::PaloAlto::PANOS::Parsers;

use strict;
use warnings;

use 5.0100000;

use XML::Twig::XPath;
use MIME::Base64 'encode_base64';

use ZipTie::Adapters::Utils qw(strip_mac trim mask_to_bits create_network_element invert_mask get_mask get_port_expression get_crep);
use ZipTie::Logger;
use ZipTie::Adapters::PaloAlto::PANOS::Rulebase qw(parse_acls print_filter_lists parse_nat);
use ZipTie::Adapters::PaloAlto::PANOS::TrimServiceGroups qw(trim_service_groups);
use ZipTie::Adapters::PaloAlto::PANOS::ParsingUtils qw(get_value_from_xml print_if_defined unwind_port_expression parse_cidr_panos get_interface_type_panos);

use Exporter 'import';
our @EXPORT_OK = qw(parse_system print_system create_config print_config parse_chassis print_chassis
  parse_interfaces print_interfaces parse_vsystems print_sub_elements parse_vlans print_vlans parse_virtual_routers
  parse_snmp print_snmp get_serial_number parse_applications parse_predefined_and_shared_applications parse_shared_elements print_shared_object_groups);

our $VERSION = 1.0;

my $LOGGER = ZipTie::Logger::get_logger();

sub parse_system {
	my ( $in, $out ) = @_;

	#grab the XML Twigs that we need
	my $show_system_info	= $in->{'show_system_info'};
	my $show_config_running = $in->{'show_config_running'};

	#find the system name
	my $system_name = get_value_from_xml( '/response/result/system/devicename', $show_system_info );
	$in->{parsed}->{'core:systemName'} = $system_name;

	#Set the type - currently default to firewall
	my $device_type = 'Firewall';
	$in->{parsed}->{'core:deviceType'} = $device_type;

	#Find the contact
	my $contact = get_value_from_xml( '/response/result/config/devices/entry/deviceconfig/system/snmp-setting/snmp-system/contact', $show_config_running );
	$in->{parsed}->{'core:contact'} = $contact;

	#calculate the lastreboot based on the uptime
	my $uptime = get_value_from_xml( '/response/result/system/uptime', $show_system_info );

	my ($years) = $uptime =~ /(\d+)\s*years?/misx;
	my ($weeks) = $uptime =~ /(\d+)\s*weeks?/misx;
	my ($days)  = $uptime =~ /(\d+)\s*days?/misx;
	my ( $hours, $minutes, $seconds ) = $uptime =~ /(\d+)\:(\d+)\:(\d+)/misx;

	# subract the last reboot from the current time
	my $last_reboot = time;
	$last_reboot -= $years * 52 * 7 * 24 * 60 * 60 if ($years);
	$last_reboot -= $weeks * 7 * 24 * 60 * 60	   if ($weeks);
	$last_reboot -= $days * 24 * 60 * 60		   if ($days);
	$last_reboot -= $hours * 60 * 60			   if ($hours);
	$last_reboot -= $minutes * 60				   if ($minutes);
	$last_reboot -= $seconds					   if ($seconds);
	$in->{parsed}->{'core:lastReboot'} = $last_reboot;

	#Set the os Name, currently default to PAN-OS
	$in->{parsed}->{'core:osInfo'}->{'core:name'} = 'PAN-OS';

	#Find the software version
	my $osVersion = get_value_from_xml( '/response/result/system/sw-version', $show_system_info );
	$in->{parsed}->{'core:osInfo'}->{'core:version'} = $osVersion;
	$in->{lookup_tables}->set_os_version($osVersion); # store for later lookup

	return;
}

sub print_system {
	my ( $in, $out ) = @_;

	print_if_defined( $in->{parsed}, 'core:systemName', $out );

	$out->open_element('core:osInfo');

	#Print out the os make. This is default to PAN-OS
	$out->print_element( 'core:make', 'Palo Alto' );

	print_if_defined( $in->{parsed}->{'core:osInfo'}, 'core:name',			$out );
	print_if_defined( $in->{parsed}->{'core:osInfo'}, 'core:softwareImage', $out );
	print_if_defined( $in->{parsed}->{'core:osInfo'}, 'core:version',		$out );

	#Print out the os type. This is default to PAN-OS.
	$out->print_element( 'core:osType', 'PAN-OS' );

	_print_packet_process_ordering($out);

	$out->close_element('core:osInfo');

	print_if_defined( $in->{parsed}, 'core:biosVersion', $out );
	print_if_defined( $in->{parsed}, 'core:deviceType',  $out );
	print_if_defined( $in->{parsed}, 'core:contact',	 $out );
	print_if_defined( $in->{parsed}, 'core:lastReboot',  $out );
	return;
}

sub parse_chassis {
	my ( $in, $out ) = @_;

	#Grab the XML Twigs that we need
	my $show_system_info	= $in->{'show_system_info'};
	my $show_config_running = $in->{'show_config_running'};

	#The asset we are creating is a Chassis so this can be hard coded
	$in->{parsed}->{chassis}->{'core:asset'}->{'core:assetType'} = 'Chassis';

	#This adapter will only handle Palo Alto devices
	$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:make'} = 'Palo Alto';

	#Find the model number
	my $model_number = get_value_from_xml( '/response/result/system/model', $show_system_info );
	if( $model_number ) {
		$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $model_number;
	}

	#Find the Serial Number
	#Do not allow 'unknown', will not allow more than one PANOS VM to appear in topology
	my $serial_number = get_value_from_xml( '/response/result/system/serial', $show_system_info );
	if( $serial_number ) {
		if( $serial_number !~ /^unknown$/i ) {
			$in->{parsed}->{chassis}->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = $serial_number;
		} else {
			$LOGGER->warning($INVALID_CONFIGURATION, "Serial number '$serial_number' will not be modelled because it can be identical for different virtual PANOS instances.");
		}
	}

	#Find the device's location
	my $location = get_value_from_xml( '/response/result/config/devices/entry/deviceconfig/system/snmp-setting/snmp-system/location', $show_config_running );
	if( $location ) {
		$in->{parsed}->{chassis}->{'core:asset'}->{'core:location'}->{'core:description'} = $location;
	}

	#Find the System level mac address
	my $mac_address = get_value_from_xml( '/response/result/system/mac-address', $show_system_info );
	if( $mac_address ) {
		$in->{parsed}->{chassis}->{'macAddress'} = strip_mac($mac_address);
	}

	#Find the system memory
	if( defined $in->{show_system_resources} ) {
		my $system_resources = get_value_from_xml( '/response/result', $in->{show_system_resources} );
		if( defined $system_resources ) {
			my ($size) = $system_resources =~ /Mem:\s*(\d+)k\s*total/i;

			$size *= 1024;

			$in->{parsed}->{chassis}->{memory}->{kind} = 'RAM';
			$in->{parsed}->{chassis}->{memory}->{size} = $size;
		}
	}

	return;
}

sub print_chassis {
	my ( $in, $out ) = @_;
	$out->open_element('chassis');
	print_if_defined( $in->{parsed}->{chassis}, 'core:asset',  $out );
	print_if_defined( $in->{parsed}->{chassis}, 'macAddress',  $out );
	print_if_defined( $in->{parsed}->{chassis}, 'memory',	   $out );
	print_if_defined( $in->{parsed}->{chassis}, 'powersupply', $out );
	$out->close_element('chassis');
	return;
}

sub create_config {
	my ( $in, $out ) = @_;

	# the name of the repository
	my $repository;
	$repository->{'core:name'} = '/';

	# build the simple text configuration
	$in->{'show_config_running'}->set_pretty_print('indented');

	{
		my $running;
		$running->{'core:name'}		  = 'running-config';
		#we encode to utf8, because we have no assurance that there are not crazy characters, encode_base64 will barf with the following "Wide character in subroutine entry"
		$running->{'core:textBlob'}   = encode_base64( Encode::encode_utf8( $in->{'show_config_running'}->first_elt('config')->sprint() ));
		$running->{'core:mediaType'}  = 'text/xml';
		$running->{'core:context'}	  = 'active';
		$running->{'core:promotable'} = 'false';

		# push the configuration into the repository
		my $system_folder;
		$system_folder->{'name'} = 'system space';
		push @{$system_folder->{'core:config'}}, $running;
		push @{$repository->{'core:folder'}}, $system_folder;

		#also push it at the base level
		my $test_folder;
		$test_folder->{'name'} = 'admin';
		push @{$test_folder->{'core:config'}}, $running;
		push @{$repository->{'core:folder'}}, $test_folder;
	}



	#Extract vsys entries from the running config
	my @virtual_systems = $in->{'show_config_running'}->findnodes('/response/result/config/devices/entry/vsys/entry');
	foreach my $vsys (@virtual_systems) {
		my $running;
		my $folder;
		$folder->{name} = $vsys->att('name');

		$running->{'core:name'}		  = 'running-config';
		#we encode to utf8, because we have no assurance that there are not crazy characters, encode_base64 will barf with the following "Wide character in subroutine entry"
		$running->{'core:textBlob'}   =  encode_base64( Encode::encode_utf8( $vsys->sprint() ));
		$running->{'core:mediaType'}  = 'text/xml';
		$running->{'core:context'}	  = 'active';
		$running->{'core:promotable'} = 'false';

		# push the configuration into the repository
		push @{ $folder->{'core:config'} }, $running;
		push @{ $repository->{'core:folder'} }, $folder;
	}

#	#Extract virtual router entries from the running config
#	my @virtual_routers = $in->{'show_config_running'}->findnodes('/response/result/config/devices/entry/network/virtual-router/entry');
#	foreach my $router (@virtual_routers) {
#		my $running;
#		my $folder;
#		$folder->{name} = $router->att('name');
#
#		$running->{'core:name'}		  = 'running-config';
#		$running->{'core:textBlob'}   = encode_base64( $router->sprint() );
#		$running->{'core:mediaType'}  = 'text/xml';
#		$running->{'core:context'}	  = 'active';
#		$running->{'core:promotable'} = 'false';
#
#		# push the configuration into the repository
#		push @{ $folder->{'core:config'} }, $running;
#		push @{ $repository->{'core:folder'} }, $folder;
#	}

	# save the repository
	$in->{parsed}->{'core:configRepository'} = $repository;
	return;
}

sub print_config {
	my ( $in, $out ) = @_;

	$out->open_element('core:configRepository name="'.$in->{parsed}->{'core:configRepository'}->{'core:name'}.'"');

	#print the configs at the base level, ie the complete running config
	foreach my $config (@{$in->{parsed}->{'core:configRepository'}->{'core:config'}}) {
		$out->print_element('core:config', $config);
	}

	#print the sub-configs, ie the configs for virtual systems and virtual routers
	foreach my $folder (@{$in->{parsed}->{'core:configRepository'}->{'core:folder'}}) {
		$out->open_element('core:folder name="'.$folder->{name}.'"');
		foreach my $config (@{$folder->{'core:config'}}) {
			$out->print_element('core:config', $config);
		}
		$out->close_element('core:folder');
	}

	$out->close_element('core:configRepository');

	return;
}

sub parse_snmp {
	my ( $in, $out ) = @_;

	my $show_system_info = $in->{show_system_info};
	my $show_config_running = $in->{show_config_running};

	my @snmp_settings = $in->{show_config_running}->find_nodes('/response/result/config/devices/entry/deviceconfig/system/snmp-setting');
	#should not be more than one of these on a device
	if (@snmp_settings) {
		my $snmp_settings = shift @snmp_settings;

		$in->{parsed}->{snmp}->{sysLocation} = get_value_from_xml('snmp-system/location', $snmp_settings);
		$in->{parsed}->{snmp}->{sysContact} = get_value_from_xml('snmp-system/contact', $snmp_settings);

		my @community = $snmp_settings->find_nodes('access-setting/version//snmp-community-string');
		foreach my $community (@community) {
			$in->{parsed}->{snmp}->{community}->{communityString} = get_value_from_xml('.', $community);
		}
	}

	#find any trap hosts configured and extract the ip address and the community string (if it exists)
	my @trap_hosts = $in->{show_config_running}->find_nodes('/response/result/config/shared/log-settings/snmptrap/entry/version//server/entry');
	foreach my $host (@trap_hosts) {

		my $parsed_host;
		$parsed_host->{ipAddress} = $host->att('name');
		$parsed_host->{communityString} = get_value_from_xml('community', $host);
		if (!defined $parsed_host->{communityString}) {
			delete $parsed_host->{communityString};
		}

		push @{$in->{parsed}->{snmp}->{trapHosts}}, $parsed_host;
	}

	#the device name is not listed in the snmp sections, but we can get it from the show shsytem info command
	$in->{parsed}->{snmp}->{sysName} = get_value_from_xml('/response/result/system/devicename', $show_system_info);

	return;
}

sub print_snmp {
	my ( $in, $out ) = @_;

	print_if_defined($in->{parsed}, 'snmp', $out);

	return;
}

sub parse_interfaces {
	my ( $in, $out ) = @_;

	my $show_interface_all  = $in->{'show_interface_all'};
	my $show_config_running = $in->{'show_config_running'};

	#set up virtual wire information in the lookup table
	my @virtual_wires = $show_config_running->find_nodes('/response/result/config/devices/entry/network/virtual-wire/entry');
	_parse_virtual_wires(\@virtual_wires, $in->{lookup_tables});

	# Create interface management profile information in the lookup table
	my @management_profiles = $show_config_running->find_nodes('/response/result/config/devices/entry/network/profiles/interface-management-profile/entry');
	_parse_management_profile(\@management_profiles, $in->{lookup_tables});

	# Get the global mtu value
	my $global_mtu = get_value_from_xml('/response/result/config/devices/entry/deviceconfig/setting/jumbo-frame/mtu', $show_config_running);
	if( defined $global_mtu ) {
		$in->{lookup_tables}->set_global_mtu($global_mtu);
	}

	#parse the configured ipsec tunnels so the information is available when parsing tunnel interfaces
	my @tunnel_entrys = $show_config_running->findnodes('/response/result/config/devices/entry/network/tunnel/ipsec/entry');
	my @gateway_entrys = $show_config_running->findnodes('/response/result/config/devices/entry/network/ike/gateway/entry');
	_parse_ipsec_tunnels(\@tunnel_entrys, \@gateway_entrys, $in->{lookup_tables});

	my @types = qw(ethernet/entry vlan loopback tunnel);
	foreach my $type (@types) {
		my @interfaces = $show_config_running->find_nodes("/response/result/config/devices/entry/network/interface/$type");

		$type =~ s/\/.+//; # remove '/entry' for ethernet

		#foreach interface call _parse_interface, passing hash of needed elements
		#receive hash of data in SED format representing one interface
		foreach my $interface (@interfaces) {

			my $parsed_interface = _parse_interface(
				{
					'show_interface_all'   => $show_interface_all,
					'show_config_running'  => $interface,
					'type'  => $type,
					'group_characteristic' => 'primaryInterface',
					'lookup' => $in->{lookup_tables}
				}
			);

			#register this interface in the lookup table for future use
			my $if_name = $parsed_interface->{name};
			$in->{lookup_tables}->add_interface( $if_name, $parsed_interface );

			my @sub_interfaces = $interface->findnodes('.//units/entry');
			foreach my $sub_interface (@sub_interfaces) {
				my $parsed_sub_interface = _parse_interface(
					{
						'show_interface_all'   => $show_interface_all,
						'show_config_running'  => $sub_interface,
						'type'  => $type,
						'group_characteristic' => 'subInterface',
						'lookup' => $in->{lookup_tables}
					}
				);
				my $sub_if_name = $parsed_sub_interface->{name};
				$in->{lookup_tables}->add_interface( $sub_if_name, $parsed_sub_interface );
			}
		} #end foreach interface
	} #end foreach type

	my @ifnet_interfaces = $show_interface_all->find_nodes('/response/result/ifnet/entry');
	my (@primary_interfaces, $vsys);

	foreach my $ifnet_interface(@ifnet_interfaces){

		my $name = get_value_from_xml('name', $ifnet_interface);
		my $interface = $in->{lookup_tables}->get_interface($name);

		if( not $interface ){
			$LOGGER->debug("Interface '$name' does not exist, creating a new interface");

			$interface = {
				name => $name
			};

			$interface->{interfaceType} = get_interface_type_panos($name);
			$interface->{physical} = $interface->{interfaceType} eq 'ethernet' ? 'true' : 'false';

			push(@{$interface->{interfaceGrouping}}, {
				ifGroupName => 'any',
				ifGroupingCharacteristic => 'securityZoneMember'
			});

			$vsys->{$name} = get_value_from_xml('vsys', $ifnet_interface);
			push(@{$in->{parsed}->{vsys_interfaces}->{$vsys->{$name}}}, $interface);

			$in->{lookup_tables}->add_interface($name, $interface);
		}

		if($name =~ /^(.*?)\.\d+$/){

			my $parent_name = $1;

			my $add = 1;
			foreach my $interface_grouping(@{$interface->{interfaceGrouping}}){
				if($interface_grouping->{ifGroupingCharacteristic} eq 'subInterface'){
					$add = 0; last;
				}
			}

			if($add){
				push(@{$interface->{interfaceGrouping}}, {
					ifGroupName => $parent_name,
					ifGroupingCharacteristic => 'subInterface'
				});
			}

			# Remeber to add the primary interface later
			push(@primary_interfaces, $parent_name);
		}


		my $zone = get_value_from_xml('zone', $ifnet_interface);
		if($zone){

			my $add = 1;
			foreach my $interface_grouping(@{$interface->{interfaceGrouping}}){
				if($interface_grouping->{ifGroupingCharacteristic} eq 'securityZoneMember'){
					$add = 0; last;
				}
			}
			if($add){
				push(@{$interface->{interfaceGrouping}}, {ifGroupName => $zone, ifGroupingCharacteristic => 'securityZoneMember'});
			}
		}

		# Add any IP addresses we did not previously add
		# These include address objects, that are resolved by PANOS in "show interface all"
		my $ip = get_value_from_xml('ip', $ifnet_interface);
		my @addresses = ($ip);
		my @addr_members = $ifnet_interface->findnodes('addr/member');
		for my $addr_member ( @addr_members ) {
			push @addresses, get_value_from_xml('.', $addr_member);
		}

		for my $address (@addresses) {
			if($address =~ /^((\d+\.)+\d+)\/(\d+)$/){
				my $ip_address = $1;
				my $mask = $3 // 32;

				my $ip_configuration = { ipAddress => $ip_address };
				if($mask != 32){
					$ip_configuration->{mask} = $mask;
				}

				my $add = 1;
				foreach my $ip_config(@{$interface->{interfaceIp}->{ipConfiguration}}){
					if($ip_config->{ipAddress} eq $ip_address){
						$add = 0; last;
					}
				}
				if($add){
					push(@{$interface->{interfaceIp}->{ipConfiguration}}, $ip_configuration);
				}
			}
		}
	}

	my @hardware_interfaces = $show_interface_all->find_nodes('/response/result/hw/entry');
	my $aggregate_interfaces;

	foreach my $hardware_interface(@hardware_interfaces){
		my $name = get_value_from_xml('name', $hardware_interface);
		my $interface = $in->{lookup_tables}->get_interface($name);

		if( not $interface ){
			$LOGGER->debug("Interface '$name' does not exist, creating a new interface");

			$interface = { 
				name => $name
			};
			
			# Deduce interface type and physicality
			$interface->{interfaceType} = get_interface_type_panos($name);
			$interface->{physical} = $interface->{interfaceType} eq 'ethernet' ? 'true' : 'false';

			$in->{lookup_tables}->add_interface($name, $interface);
		}

		# Get status of the interface, is it active or not
		my $admin_status = 'down';

		# If this is a sub interface, use parent's state
		if($name =~ /^(.*?)\.\d+$/){

			my $parent_name = $1;
			$admin_status = get_value_from_xml("/response/result/hw/entry[name=\"$parent_name\"]/state", $show_interface_all);

			my $add = 1;
			foreach my $interface_grouping(@{$interface->{interfaceGrouping}}){
				if($interface_grouping->{ifGroupingCharacteristic} eq 'subInterface'){
					$add = 0; last;
				}
			}

			if($add){
				# Also add an interface grouping while we know
				push(@{$interface->{interfaceGrouping}}, {
					ifGroupName => $parent_name,
					ifGroupingCharacteristic => 'subInterface'
				});
			}

			# Remeber to add the primary interface later
			push(@primary_interfaces, $parent_name);

		}else{
			$admin_status = get_value_from_xml('state', $hardware_interface);
		}

		if(!$interface->{adminStatus} and $admin_status =~ /^up|down$/i){
			$interface->{adminStatus} = $admin_status;
		}

		# Get the ethernet data
		if(!$interface->{interfaceEthernet}){
			my $interface_ethernet = _parse_interface_ethernet($hardware_interface);
			$interface->{interfaceEthernet} = $interface_ethernet;
		}

		# Get the speed, if it isn't a number throw it away
		my $speed = get_value_from_xml('speed', $hardware_interface);

		if(!defined($interface->{speed}) and defined($speed) and $speed =~ /^\d+$/){
			$interface->{speed} = $speed * 1000;
		}
	
		my @aggregation_members = $hardware_interface->find_nodes('ae_member/member');
		foreach my $aggregation_member(@aggregation_members){

			my $member = get_value_from_xml('.', $aggregation_member);
			push(@{$aggregate_interfaces->{$name}}, $member);
		}
	}

	foreach my $primary_interface(@primary_interfaces){
		my $interface = $in->{lookup_tables}->get_interface($primary_interface);

		if($interface){

			my $add = 1;

			if($interface->{interfaceGrouping}){
				foreach my $interface_grouping(@{$interface->{interfaceGrouping}}){
					if($interface_grouping->{ifGroupName} eq $primary_interface
					and $interface_grouping->{ifGroupingCharacteristic} eq 'primaryInterface'){
						$add = 0; last;
					}
				}
			}

			if($add){
				push(@{$interface->{interfaceGrouping}}, {
					ifGroupName => $primary_interface,
					ifGroupingCharacteristic => 'primaryInterface'
				});
			}
		}else{
			$LOGGER->debug("Interface '$primary_interface' not found, cannot add primary interface");
		}
	}

	foreach my $aggregate_key(keys(%{$aggregate_interfaces})){
		my $interface_names = $aggregate_interfaces->{$aggregate_key};

		foreach my $interface_name(@{$interface_names}){
			my $interface = $in->{lookup_tables}->get_interface($interface_name);

			if($interface){

				push(@{$interface->{interfaceGrouping}}, {
					ifGroupName => $aggregate_key,
					ifGroupingCharacteristic => 'bondingGroupMember'
				});

				push(@{$in->{parsed}->{vsys_interfaces}->{$vsys->{$aggregate_key}}}, $interface);

			}else{
				$LOGGER->debug("Interface '$interface_name' not found, cannot add aggregate interface");
			}
		}
	}

	#set up a stub management interface to be the interface on the admin context of the device
	my $mgmt_ip;

	#the management interface is up, or we couldn't be accessing the device
	$mgmt_ip->{adminStatus} = 'up';

	#default this to be a virtual interface for now
	$mgmt_ip->{interfaceType} = 'virtual';
#	$mgmt_ip->{interfaceBehavior} = 'administrative';
	$mgmt_ip->{name} = 'Management Interface';
	$mgmt_ip->{physical} = 'false';

	#get the management ip address and mask from the show system info command
	my $mgmt_ip_address = get_value_from_xml('/response/result/system/ip-address', $in->{show_system_info});
	my $mgmt_mask = mask_to_bits( get_value_from_xml('/response/result/system/netmask', $in->{show_system_info}) );

	$mgmt_ip->{interfaceIp}->{ipConfiguration}->{ipAddress} = $mgmt_ip_address;
	$mgmt_ip->{interfaceIp}->{ipConfiguration}->{mask} = $mgmt_mask;
	$in->{lookup_tables}->add_management_ip_addr($mgmt_ip_address, $mgmt_mask);

	push @{$in->{parsed}->{interfaces}}, $mgmt_ip;

	return;

}	#end sub parse_interfaces


#sub parse_interfaces {
#	my ( $in, $out ) = @_;
#
#	my $show_interface_all  = $in->{'show_interface_all'};
#	my $show_config_running = $in->{'show_config_running'};
#
#	#get list of interfaces
#	my @interfaces = $show_interface_all->find_nodes('/response/result/ifnet/entry');
#
#	my @parsed_interfaces;
#
#	#foreach interface call _parse_interface, passing hash of needed elements
#	#receive hash of data in SED format representing one interface
#	foreach my $interface (@interfaces) {
#
#		my $interface_name = get_value_from_xml( 'name', $interface );
#
#		$LOGGER->debug("==Parsing interface: $interface_name");
#		my $parsed_interface = _parse_interface(
#			{
#				'show_interface_all'   => $show_interface_all,
#				'show_config_running'  => $show_config_running,
#				'show_interface_ifnet' => $interface,
#				'lookup' => $in->{lookup_tables}
#			}
#		);
#		push @parsed_interfaces, $parsed_interface;
#
#		#register this interface in the lookup table for future use
#		my $if_name = $parsed_interface->{name};
#		$in->{lookup_tables}->add_interface( $if_name, $parsed_interface );
#	}
#
#
#	#set up a stub management interface to be the interface on the admin context of the device
#	my $mgmt_ip;
#
#	#the management interface is up, or we couldn't be accessing the device
#	$mgmt_ip->{adminStatus} = 'up';
#
#	#default this to be a virtual interface for now
#	$mgmt_ip->{interfaceType} = 'virtual';
##   $mgmt_ip->{interfaceBehavior} = 'administrative';
#	$mgmt_ip->{name} = 'Management Interface';
#	$mgmt_ip->{physical} = 'false';
#
#	#get the management ip address and mask from the show system info command
#	$mgmt_ip->{interfaceIp}->{ipConfiguration}->{ipAddress} = get_value_from_xml('/response/result/system/ip-address', $in->{show_system_info});
#	$mgmt_ip->{interfaceIp}->{ipConfiguration}->{mask} = mask_to_bits( get_value_from_xml('/response/result/system/netmask', $in->{show_system_info}) );
#
#
##   $in->{parsed}->{interfaces} = \@parsed_interfaces;
#	push @{$in->{parsed}->{interfaces}}, $mgmt_ip;
#
#	return;
#
#}	#end sub parse_interfaces


# this function takes a hash of XML Twig elements and other data relevant to one interface,
# parses the needed information and returns a hash of data in SED format suitable for xml printing
sub _parse_interface {
	my $parameters = shift;

	my $show_interface_all   = $parameters->{'show_interface_all'};
	my $show_config_entry  = $parameters->{'show_config_running'};
	my $interface_type = $parameters->{'type'};
	my $lookup = $parameters->{'lookup'};

	my $name = $show_config_entry->att('name');

	#special case for 'simulated' interfaces : 'vlan', 'loopback', and 'tunnel'
	if (!defined $name) {
		$name = $interface_type;
	}

	$LOGGER->debug("==Parsing interface: $name");

	my $physical  = 'false';

	#palo alto uses a naming convention for interfaces where a sub-interface has the same name as the
	#primary interface, with a dot and a number tacked on the end.
	my $group_name;
	my $admin_status;
	my $group_characteristic = $parameters->{group_characteristic};
	if ($group_characteristic eq 'primaryInterface') {
		$group_name = $name;
		$admin_status = get_value_from_xml( '/response/result/hw/entry[name="' . $name . '"]/state', $show_interface_all );
	} else {
		($group_name) = $name =~ /(.*)[.]\d+/;
		
		#if this interface is a sub interface, use the status of the primary interface
		$admin_status = get_value_from_xml( '/response/result/hw/entry[name="' . $group_name . '"]/state', $show_interface_all );
	}


	my $interface_ethernet;
	my @show_interface_hw_info = $show_interface_all->findnodes( '/response/result/hw/entry[name="' . $name . '"]' );
	#will only be at most one hardware entry for a given interface.
	if (@show_interface_hw_info) {

		#parse the interfaceEthernet information
		$interface_ethernet = _parse_interface_ethernet(shift @show_interface_hw_info);

	} else {
		$LOGGER->debug("No hardware information found for interface '$name'. Ethernet information not parsed. ");
	}

	#parse the interfaceIp information
	my $interface_ip = _parse_interface_ip($show_config_entry);

	my $parsed_tunnel = _parse_pppoe_tunnel($name, $show_config_entry);

	#palo alto uses a standard interface naming convention
	if (($interface_type eq 'ethernet') and ($group_characteristic eq 'primaryInterface')) {
		$physical = 'true';
	}
	elsif ($interface_type eq 'loopback') {
		$interface_type = 'softwareLoopback';
	}
	elsif ($interface_type eq 'tunnel'){
		$parsed_tunnel = $lookup->get_ipsec_tunnel($name);
	}
	elsif ($interface_type eq 'vlan'){
		$interface_type = 'virtual';
	}

	#parse the interface description
	my $description = get_value_from_xml( 'comment', $show_config_entry );

	#parse the interface speed
	my $speed = get_value_from_xml( '/response/result/hw/entry[name="' . $name . '"]/speed', $show_interface_all );

	if ( ( !defined $speed ) or ( $speed eq 'ukn' ) or ( $speed =~ /n\/a/i ) ) {
		#default to undef so we don't print a SED element. defaulting to zero in some or all cases would also be valid
		$speed = undef;
	} else {
		$speed *= 1000;
	}

	#use the interface ID as the ifIndex as it is unique
	my $if_index = get_value_from_xml( '/response/result/ifnet/entry[name="' . $name . '"]/id', $show_interface_all );

	my $mtu = get_value_from_xml('layer3/mtu', $show_config_entry);

        # Subinterfaces <mtu> element are not under a <layer3> element
	if( not defined $mtu ) {
		$mtu = get_value_from_xml('mtu', $show_config_entry);
	}

	#my $vlan_tag = get_value_from_xml('tag', $show_config_running) if defined $show_config_running;

	#Virtual wire information is parsed in the function _parse_virtual_wires and stored in the lookup table
	my $virtual_wire = $lookup->get_virtual_wire($name);

	my @vlan_membership = $lookup->get_interface_vlan_membership($name);
	my @formatted_vlans;
	foreach my $vlan (@vlan_membership) {
		push @formatted_vlans, {startVlan => $vlan};
	}

	my $zone_name = get_value_from_xml( '/response/result/ifnet/entry[name="' . $name . '"]/zone', $show_interface_all );

	#construct the hash representing the interface
	my $interface = {
		'ifIndex'		   => $if_index,
		'name'			  => $name,
		'physical'		  => $physical,
		'interfaceType'	 => $interface_type,
		'adminStatus'	   => $admin_status,
		'description'	   => $description,
		'interfaceEthernet' => $interface_ethernet,
		'interfaceIp'	   => $interface_ip,
		'interfaceTunnel'   => $parsed_tunnel,
		'speed'			 => $speed,
		'mtu'			   => $mtu // $lookup->get_global_mtu(),
	};

	if (defined $group_name) {
		push @{$interface->{'interfaceGrouping'}}, {
			'ifGroupName'			   => $group_name,
			'ifGroupingCharacteristic' => $group_characteristic
		};
	}
	
	# For 7.2 MR1: make sure to not allow the virtual wire interface grouping to be generated if interface is down--will
	# cause topology to go haywire.
	if (defined $virtual_wire and $admin_status ne 'down') {
		push @{$interface->{'interfaceGrouping'}}, {
			'ifGroupName'			   => $virtual_wire,
			'ifGroupingCharacteristic' => 'proxyAffiliation'
		};
		push @{$interface->{'interfaceBehavior'}}, 'bridging';
	}

	if (defined $zone_name) {
		push @{$interface->{'interfaceGrouping'}}, {'ifGroupName' => $zone_name, 'ifGroupingCharacteristic' => 'securityZoneMember' };
	}

	if (@formatted_vlans) {
		$interface->{interfaceVlanTrunks} = \@formatted_vlans;
	}

	push(@{$interface->{interfaceGrouping}}, {
		ifGroupName => 'any',
		ifGroupingCharacteristic => 'securityZoneMember'
	});

	my $management_profile = get_value_from_xml( 'layer3/interface-management-profile', $show_config_entry );
	if ($management_profile) {
		$lookup->add_interface_to_management_profile($management_profile, $name);
	}

	return $interface;
}	#end sub _parse_interface


#sub _parse_interface {
#
#	my $parameters = shift;
#
#	my $show_interface_all   = $parameters->{'show_interface_all'};
#	my $show_config_running  = $parameters->{'show_config_running'};
#	my $show_interface_ifnet = $parameters->{'show_interface_ifnet'};
#	my $lookup = $parameters->{'lookup'};
#
#	my $name = get_value_from_xml( 'name', $show_interface_ifnet );
#
#	my $physical  = 'true';
#
#	#palo alto uses a naming convention for interfaces where a sub-interface has the same name as the
#	#primary interface, with a dot and a number tacked on the end.
#	my $group_name;
#	my $group_characteristic;
#	if ( $name =~ /(.*)\.\d+/ ) {
#		$physical = 'false';
#		$group_name		   = $1;
#		$group_characteristic = 'subInterface';
#	} else {
#		$group_name		   = $name;
#		$group_characteristic = 'primaryInterface';
#	}
#
#	my $admin_status;
#	my $state = get_value_from_xml( '/response/result/hw/entry[name="' . $name . '"]/state', $show_interface_all );
#
#	#if this interface has a state value in the hardware section, use it. If not then it is a sub interface and we can use the state of
#	#the parent interface
#	if ( defined $state ) {
#		$admin_status = $state;
#	} else {
#		$admin_status = get_value_from_xml( '/response/result/hw/entry[name="' . $group_name . '"]/state', $show_interface_all );
#	}
#
#
#	my $interface_ethernet;
#	my @show_interface_hw_info = $show_interface_all->findnodes( '/response/result/hw/entry[name="' . $name . '"]' );
#	#will only be at most one hardware entry for a given interface.
#	if (@show_interface_hw_info) {
#
#		#parse the interfaceEthernet information
#		my $show_interface_hw = shift @show_interface_hw_info;
#		$interface_ethernet = _parse_interface_ethernet($show_interface_hw);
#
#	} else {
#		$LOGGER->debug("No hardware information found for interface '$name'. Ethernet information not parsed. ");
#	}
#
#	#parse the interfaceIp information
#	my $interface_ip = _parse_interface_ip($show_interface_ifnet);
#
#	#   my $mtu;
#	my $parsed_tunnel;
#
#	#palo alto uses a standard interfave naming convention
#	my $interface_type;
#	if ( $name =~ /ethernet/ ) {
#		$interface_type = 'ethernet';
#	} elsif ( $name =~ /vlan/ ) {
#		$interface_type = 'virtual';
#		$physical = 'false';
#	} elsif ( $name =~ /tunnel/ ) {
#		$interface_type = 'tunnel';
#
#		$parsed_tunnel = _parse_ipsec_tunnel($name, $show_config_running);
#
#		$physical = 'false';
#	} elsif ( $name =~ /loopback/ ) {
#		$interface_type = 'softwareLoopback';
#		$physical = 'false';
#	} else {
#		$interface_type = 'unknown';
#		$physical = 'false';
#	}
#
#
#	#parse the interface description
#	my $description;
#	if (defined $show_config_running) {
#		$description = get_value_from_xml( '/response/result/config/devices/entry/network/interface//entry[@name="' . $name .'"]/comment', $show_config_running );
#	}
#
#	#parse the interface speed
#	my $speed = get_value_from_xml( '/response/result/hw/entry[name="' . $name . '"]/speed', $show_interface_all );
#
#	if ( ( !defined $speed ) or ( $speed eq 'ukn' ) or ( $speed =~ /n\/a/i ) ) {
#		#default to undef so we don't print a SED element. defaulting to zero in some or all cases would also be valid
#		$speed = undef;
#	} else {
#		$speed *= 1000;
#	}
#
#	#use the interface ID as the ifIndex as it is unique
#	my $ifIndex = get_value_from_xml( 'id', $show_interface_ifnet );
#
#	#   my $vlan_tag = get_value_from_xml('tag', $show_config_running) if defined $show_config_running;
#
#	my @vlan_membership = $lookup->get_interface_vlan_membership($name);
#	my @formatted_vlans;
#	foreach my $vlan (@vlan_membership) {
#		push @formatted_vlans, {startVlan => $vlan};
#	}
#
#
#	#construct the hash representing the interface
#	my $interface = {
#		'ifIndex'		   => $ifIndex,
#		'name'			  => $name,
#		'physical'		  => $physical,
#		'interfaceType'	 => $interface_type,
#	};
#	
#	if (defined $group_name) {
#		push @{$interface->{'interfaceGrouping'}},  {   'ifGroupName'			  => $group_name,
#														'ifGroupingCharacteristic' => $group_characteristic};
#	}
#
#	if ( defined $admin_status ) {
#		$interface->{adminStatus} = $admin_status;
#	}
#
#	if (defined $description ) {
#		$interface->{description} = $description;
#	}
#
##   if ( ( defined $interface_type ) and ( $interface_type ne q{} ) ) {
##	   $interface->{interfaceType} = $interface_type;
##   }
#
#	if ( defined $interface_ethernet ) {
#		$interface->{interfaceEthernet} = $interface_ethernet;
#	}
#
#	if ( defined $interface_ip ) {
#		$interface->{interfaceIp} = $interface_ip;
#	}
#
#	if (defined $parsed_tunnel) {
#		$interface->{interfaceTunnel} = $parsed_tunnel;
#	}
#
#	if (@formatted_vlans) {
#		$interface->{interfaceVlanTrunks} = \@formatted_vlans;
#	}
#	
#	if ( defined $speed ) {
#		$interface->{speed} = $speed;
#	}
#
#	return $interface;
#}	#end sub _parse_interface




sub _parse_interface_ethernet {
	my $hw_entry = shift;

	#parse duplex information from the mode value.
	my $mode = get_value_from_xml( 'mode', $hw_entry );
	my $auto_duplex;
	my $auto_speed;

	if ( $mode && $mode =~ /autoneg/ ) {
		$auto_duplex = 'true';
		$auto_speed  = 'true';
	} else {
		$auto_duplex = 'false';
		$auto_speed  = 'false';
	}

	#parse the mac address
	my $mac = strip_mac( get_value_from_xml( 'mac', $hw_entry ) );

	#	my $media_type;

	my $operational_duplex = get_value_from_xml( 'duplex', $hw_entry );
	if ( $operational_duplex =~ /(ukn)|(n\/a)|(unknown)/ ) {
		$operational_duplex = undef;
	}

	my $interface_ethernet = {
		'autoDuplex' => $auto_duplex,
		'autoSpeed'  => $auto_speed,
		'macAddress' => $mac,
	};

	if ( defined $operational_duplex ) {
		$interface_ethernet->{operationalDuplex} = $operational_duplex;
	}

	return $interface_ethernet;
}	#end sub _parse_interface_ethernet


sub _parse_interface_ip {
	my ($show_config_entry) = @_;
	my $ip_regex = get_crep('cipm');
	my @ip_configs;

	# IP information can be in one of several different places depending on the interface type. Check them all
	my @entries = $show_config_entry->findnodes('layer3/ip/entry');
	push @entries, $show_config_entry->findnodes('ip/entry');

	foreach my $entry (@entries) {
		my $address = $entry->att('name');

		if( $address =~ /^$ip_regex(\/\d+)?$/ ) {
			my $element = parse_cidr_panos($address);
			my $ip_configuration = {
				ipAddress => $element->{host},
				mask	  => $element->{prefix}
			};
			push @ip_configs, $ip_configuration;
		} # if the address is an object reference, PANOS will have resolved it for us in "show interface all", add this address when parsing there
	}

	my $interface_ip;

	# Only attach the list if there are elements
	if ( scalar @ip_configs ) {
		$interface_ip->{ipConfiguration} = \@ip_configs;
	}

	return $interface_ip;
}

sub _parse_pppoe_tunnel{
	my ($interface_name, $interface_entry) = @_;
	my $interface_tunnel;

	my $pppoe = get_value_from_xml('layer3/pppoe/enable', $interface_entry);
	if($pppoe && $pppoe eq 'yes'){

		$interface_tunnel = {
			encapsulationProtocol => 'other',
			tunnelLocalPeering => {tunnelLocalPeeringInterface => $interface_name}
		};

		my $static_address = get_value_from_xml('layer3/pppoe/static-address/ip', $interface_entry);
		if($static_address){
			$interface_tunnel->{tunnelLocalPeering} = {tunnelLocalPeeringIPAddress => $static_address};
		}
	}

	return($interface_tunnel);
}

sub _parse_ipsec_tunnels {
	my ( $tunnel_entries, $gateway_entries, $lookup ) = @_;
	my $parsed_gateways;

	foreach my $entry (@{$gateway_entries}) {
		my $gateway_name = $entry->att('name');
		$parsed_gateways->{$gateway_name}->{tunnelLocalPeeringInterface} = get_value_from_xml('local-address/interface', $entry);
		$parsed_gateways->{$gateway_name}->{tunnelRemotePeerIPAddress} = get_value_from_xml('peer-address/ip', $entry);
	}

	for my $tunnel (@{$tunnel_entries}) {
		my $parsed_tunnel;

		#this function deals exclusively with ipSec tunnels, so default the type accordingly
		$parsed_tunnel->{encapsulationProtocol} = 'ipSec';

		# this function uses 'tunnelLocalPeeringInterface' rather than 'tunnelLocalPeeringIpAddress' as the latter will not always be present

		#ipSec tunnels can be manually configured
		if ($tunnel->first_child_matches('manual-key')) {
			$parsed_tunnel->{tunnelLocalPeering}->{tunnelLocalPeeringInterface} = get_value_from_xml('manual-key/local-address/interface', $tunnel);
			$parsed_tunnel->{tunnelRemotePeerIPAddress} = get_value_from_xml('manual-key/peer-address/ip', $tunnel);
		}

		#ipSec tunnels can be auto configured
		elsif ($tunnel->first_child_matches('auto-key')) {
			if ($tunnel->first_child()->first_child_matches('ike-gateway')) {

				# need to extract the ike gateway name and look up the gateway information
				my $gateway = shift @{$tunnel->findnodes('auto-key/ike-gateway/entry')};
				my $gateway_name = $gateway->att('name');

#				my $gateway_entry = $lookup->get_ike_gateway($gateway_name);
#				my $gateway_entry = shift @{$show_config_running->findnodes('/response/result/config/devices/entry/network/ike/gateway/entry[@name="' .$gateway_name. '"]')};

				$parsed_tunnel->{tunnelLocalPeering}->{tunnelLocalPeeringInterface} = $parsed_gateways->{$gateway_name}->{tunnelLocalPeeringInterface};
				$parsed_tunnel->{tunnelRemotePeerIPAddress} = $parsed_gateways->{$gateway_name}->{tunnelRemotePeerIPAddress};
			}
		}

		#ipSec tunnels can be configured through a global-protect satellite
		elsif ($tunnel->first_child_matches('global-protect-satellite')) {
				$parsed_tunnel->{tunnelLocalPeering}->{tunnelLocalPeeringInterface} = get_value_from_xml('global-protect-satellite/local-address/interface', $tunnel);
				$parsed_tunnel->{tunnelRemotePeerIPAddress} = get_value_from_xml('global-protect-satellite/portal-address', $tunnel);
		}

		#will cause a failed to mergre interfaces parse warning if printed empty
		if( exists( $parsed_tunnel->{tunnelRemotePeerIPAddress} ) && not( $parsed_tunnel->{tunnelRemotePeerIPAddress} ) ) {
			delete( $parsed_tunnel->{tunnelRemotePeerIPAddress} );
		}

		my $tunnel_interface = get_value_from_xml('tunnel-interface', $tunnel);
		$lookup->add_ipsec_tunnel($tunnel_interface, $parsed_tunnel);

	} #end outer for loop
	return;
}

sub _parse_virtual_wires {
	my ( $virtual_wires, $lookup ) = @_;

#	my @entries = $show_virtual_wire->findnodes('/response/result/entries/entry');

	foreach my $entry (@{$virtual_wires}) {

		my $name = $entry->att('name');
		my $interface_one = get_value_from_xml('interface1', $entry);
		my $interface_two = get_value_from_xml('interface2', $entry);

		$lookup->add_virtual_wire($name, $interface_one, $interface_two);
	}
	return;
}

sub _parse_management_profile{
	my ($profiles, $lookup) = @_;

	if( defined $profiles ) {
		foreach my $profile(@{$profiles}){

			my $name = $profile->att('name');

			my $network_group;
			my @permitted_ips = $profile->findnodes('permitted-ip/entry');

			foreach my $permitted_ip(@permitted_ips){

				my $name = $permitted_ip->att('name');
				my ($ip, $mask) = split(/\//, $name);

				if(!defined($mask) || $mask == 32){
					push(@{$network_group->{host}}, $ip);
				}else{
					my $subnet = {
						address => $ip,
						mask => $mask
					};
					push(@{$network_group->{subnet}}, $subnet);
				}
			}

			my $service_group;

			my @services = ( qw(http https ssh snmp ping telnet) );
			foreach my $service(@services){

				my $service_permitted = get_value_from_xml($service, $profile);
				if($service_permitted && $service_permitted eq 'yes'){
					
					my $port_expressions = get_port_expression($service);
					foreach my $port_expression(@{$port_expressions}){
						foreach my $expression_type(keys(%{$port_expression})){
							push(@{$service_group->{$expression_type}}, $port_expression->{$expression_type});
						}
					}
				}
			}

			my $response_pages = get_value_from_xml('response-pages', $profile);
			if(defined ($response_pages) && $response_pages eq 'yes'){
				push(@{$service_group->{portRange}}, {
					portStart => 6080,
					portEnd => 6081
				})
			}

			$lookup->add_interface_management_profile($name, $network_group, $service_group);
		}
	}
	return;
}

sub print_interfaces {
	my ( $in, $out ) = @_;

	$out->open_element('interfaces');
	foreach my $interface ( @{ $in->{parsed}->{interfaces} } ) {
		_print_interface($interface, $out);
	}
	$out->close_element('interfaces');

	return;
}

#interface elements are broken out and printed separatly due to the fact that
# the 'interfaceBehavior' element is out of order
sub _print_interface {
	my ($interface, $out) = @_;

	$out->open_element('interface');

	print_if_defined( $interface, 'adminStatus', $out );
	print_if_defined( $interface, 'description', $out );
	print_if_defined( $interface, 'ifIndex', $out );
	print_if_defined( $interface, 'interfaceEthernet', $out );
	print_if_defined( $interface, 'interfaceGrouping', $out );
	print_if_defined( $interface, 'interfaceIp', $out );
	print_if_defined( $interface, 'interfaceTunnel', $out );
	print_if_defined( $interface, 'interfaceBehavior', $out );
	print_if_defined( $interface, 'interfaceType', $out );
	print_if_defined( $interface, 'interfaceVlanTrunks', $out );
	print_if_defined( $interface, 'mtu', $out );
	print_if_defined( $interface, 'name', $out );
	print_if_defined( $interface, 'physical', $out );
	print_if_defined( $interface, 'speed', $out );

	$out->close_element('interface');

	return;
}


sub parse_vlans {
	my ( $in, $out ) = @_;

	my $show_config_running = $in->{'show_config_running'};
	my $lookup = $in->{'lookup_tables'};

	my @parsed_vlans;

	my @vlans = $show_config_running->findnodes('/response/result/config/devices/entry/network/vlan/entry');

	foreach my $vlan (@vlans) {

		my $parsed_vlan;

		$parsed_vlan->{configSource} = 'local';
#		$parsed_vlan->{name} = get_value_from_xml( 'name', $vlan );
		$parsed_vlan->{name} = $vlan->att('name');

		###Vlans in Palo alto don't have numbers directly associated with them, however a number is required by the sed.
		#pull the number from the associated interface if it exists. Otherwise the vlan will be dropped.
		my $virtual_interface = get_value_from_xml( 'virtual-interface/interface', $vlan );
		if (defined $virtual_interface) {
		  ( $parsed_vlan->{number} ) = $virtual_interface =~ /vlan\.(\d+)/;
		  $lookup->add_vlan_member($parsed_vlan->{name}, $virtual_interface);
		}

		$parsed_vlan->{enabled} = 'true';

		#parse the interface members of this vlan
		my @members		= $vlan->findnodes('interface/member');
		my @parsed_members = ();
		foreach my $member (@members) {
			my $interface_member = get_value_from_xml( '.', $member );
			push @parsed_members, $interface_member;
			$lookup->add_vlan_member($parsed_vlan->{name}, $interface_member);
		}

		if (@parsed_members) {
			$parsed_vlan->{interfaceMember} = \@parsed_members;
		}

		#Only save the vlan if it has a valid number
		if ( defined $parsed_vlan->{number} ) {
			push @parsed_vlans, $parsed_vlan;
		} else {
			$LOGGER->debug( "Parsing of VLAN '" . $parsed_vlan->{name} . "' failed. VLAN has no number. This is likely due to it having no associated vlan interface." );
		}
	}

	if (@parsed_vlans) {
		$in->{parsed}->{vlans}->{vlan} = \@parsed_vlans;
	}

	return;
}

sub print_vlans {
	my ( $in, $out ) = @_;

	print_if_defined($in->{parsed}, 'vlans', $out);
#	$out->print_element( 'vlans', $in->{parsed}->{vlans} );

	return;
}

# This sets up parsing of the <shared> definitions of the configuration XML from the device.
sub parse_shared_elements 
{
	my ($in, $out) = @_;
	my $show_config_running = $in->{'show_config_running'};

	# There is only one of these
	my $sharedSystemElements = $show_config_running->findnodes('/response/result/config/shared');
	if ($sharedSystemElements)
	{
		my $parsedSharedElementSet->{objectGroups} = _parse_object_groups($in, $sharedSystemElements->[0]);
		if ($parsedSharedElementSet) {
			$in->{parsed}->{sharedElements} = $parsedSharedElementSet; 
		}
	} else {
		$LOGGER->debug("Parsing of shared definitions bypassed: no applicable shared elements found");	
	}

	return;
}

# This function extracts the VSystem information from the running config (and anywhere else we find it)
# and calls sub parsers to create subelement entities to include in the SED.
sub parse_vsystems {
	my ($in, $out) = @_;

	my $show_config_running = $in->{'show_config_running'};
	my @virtual_systems = $show_config_running->findnodes('/response/result/config/devices/entry/vsys/entry');
	my @parsed_vsystems = ();
	my $count = 0;

	foreach my $vsys (@virtual_systems) {

		my $parsed_vsys;
		$parsed_vsys->{'core:systemName'} = $vsys->att('name');

		#virtual systems are essentially virtual firewalls
		$parsed_vsys->{'core:deviceType'} = 'Firewall';
		$parsed_vsys->{'subelementType'} = 'virtual';

		#lookup the interfaces that are in this virtual system
		$LOGGER->debug( "Importing interface information for virtual system '" . $parsed_vsys->{'core:systemName'} . "'" );
		my @imported_interfaces = $vsys->findnodes('import/network/interface/member');
		my $interfaces = _import_interfaces( \@imported_interfaces, $in->{lookup_tables} );
		if($in->{parsed}->{vsys_interfaces}->{++$count}){
			push(@{$interfaces->{interface}}, @{$in->{parsed}->{vsys_interfaces}->{$count}});
		}

		# TODO RTC#106410 Tracks this quick-fix which should be removed once QRM can handle virutal systems with no ip address
		my $ip = 0;
		foreach my $interface (@{$interfaces->{interface}}) {
			if ($interface->{interfaceIp}) {
				$ip = 1; last;
			}
		}
		if (!$ip) {
			$LOGGER->debug('WARNING: Virtual system has no interfaces with an IP, dropping.');
			next;
		}

		$parsed_vsys->{interfaces} = $interfaces;

		#Parse the applications defined on this vsys
		$parsed_vsys->{applications} = parse_applications($vsys);

		#parse the object groups in this vsys
		$LOGGER->debug( "Parsing object groups for virtual system '" . $parsed_vsys->{'core:systemName'} . "'" );
		$parsed_vsys->{objectGroups} = _parse_object_groups( $in, $vsys);

		#parse the rules in this vsys
		$parsed_vsys->{filterLists} = parse_acls( $vsys, $in->{lookup_tables});
		
		# Add any needed application-default service groups to object groups
		my $adp_service_groups = $in->{lookup_tables}->get_default_port_service_groups();
		push(@{$parsed_vsys->{objectGroups}->{serviceGroup}}, @{$adp_service_groups});

		push @parsed_vsystems, $parsed_vsys;

		my @natRules = $vsys->findnodes('rulebase/nat/rules/entry');
		$parsed_vsys->{natFilterLists} = parse_nat(\@natRules, $in->{lookup_tables});		
	}

	#There is always at least one vsys
	$in->{parsed}->{subelements}->{subelement} = \@parsed_vsystems;

	return;
}

#this function takes a vsys entity and parses the object group information it contains
sub _parse_object_groups {
	my ($in, $vsys) = @_;

	my $lookup = $in->{lookup_tables};
	my $shared_objects_groups = $in->{parsed}->{sharedElements};
	my $show_predefined_service_groups = $in->{show_predefined_service_groups};
	my $show_dynamic_address_groups = $in->{show_dynamic_address_groups};
	my $request_dynamic_block_list = $in->{request_dynamic_block_list};

	my $parsed_object_groups;

	$LOGGER->debug('==Parsing network group information');
	$parsed_object_groups->{networkGroup} = _parse_network_groups( $vsys, $lookup, $shared_objects_groups, $show_dynamic_address_groups, $request_dynamic_block_list);

	$LOGGER->debug('==Parsing service group information');
	$parsed_object_groups->{serviceGroup} = _parse_service_groups( $vsys, $lookup, $shared_objects_groups, $show_predefined_service_groups );

	$LOGGER->debug('==Parsing application group information');
	my @application_groups = $vsys->findnodes('application-group/entry');
	$LOGGER->debug(@application_groups . ' application groups found');
	$parsed_object_groups->{applicationGroup} = _parse_application_groups($lookup, @application_groups);

	$LOGGER->debug('==Parsing application filter information');
	my @application_filters = $vsys->findnodes('application-filter/entry');	
	$LOGGER->debug(@application_filters . ' application filters found');
	my $parsed_application_filters = _parse_application_filters($lookup, @application_filters);
	push(@{$parsed_object_groups->{applicationGroup}}, @$parsed_application_filters);

	#_parse_security_zones
	#don't add these to the object groups hash as they are not displayed at present. Also, not present for shared elements.
	my ($zones) = $vsys->findnodes('zone');
	if($zones){
		$LOGGER->debug('==Parsing security zone information');
		_parse_security_zones( $zones, $lookup );
	}
	
	# parse user groups to be used in future
	#$LOGGER->debug('==Parsing user group information');
	#$parsed_object_groups->{userGroup} = parse_user_groups ($vsys, $lookup, $show_user_group_name);
	
	return $parsed_object_groups;
}

#this function takes an address entity and an address group entity parses them and returns
#a list of hashes structured as SED networkGroup elements
sub _parse_network_groups {
	my ($vsys, $lookup, $shared_network_groups, $show_dynamic_address_groups, $request_dynamic_block_list) = @_;
	my $ip_regex = get_crep('cipm');

	my @parsed_network_groups = ();

	# We collect any FQDNs and later drop them if in this hash
	my %fqdns;
	my @address_entries = $vsys->findnodes('address/entry');
	foreach my $entry (@address_entries) {

		my $id = $entry->att('name');
		my $parsed_network_group = { 'id' => $id, };

		my $description = get_value_from_xml( 'description', $entry );
		if ( defined $description ) {	#Only save the description if it is not empty
			$parsed_network_group->{description} = $description;
		}

		#network groups can be defined with netmasks or ranges
		if ( $entry->exists('ip-netmask') ) {
			my $cidr = get_value_from_xml( 'ip-netmask', $entry );

			my $element = parse_cidr_panos($cidr);
			$parsed_network_group->{subnet}->{address} = $element->{host};
			$parsed_network_group->{subnet}->{mask}	= $element->{prefix};

		} elsif ( $entry->exists('ip-range') ) {

			my $range = get_value_from_xml( 'ip-range', $entry );

			my ( $start, $end ) = $range =~ /(.*)-(.*)/;

			$parsed_network_group->{range}->{startAddress} = $start;
			$parsed_network_group->{range}->{endAddress}   = $end;
		} elsif( $entry->exists('fqdn') ) {
			my $fqdn = get_value_from_xml('fqdn', $entry);
			$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Fully qualified domain names (FQDNs) are not supported. Dropping FQDN '$fqdn'.");
			$fqdns{$fqdn} = 1;
		}

		push @parsed_network_groups, $parsed_network_group;
	}

	# We collect dynamic address group names and descs from vsys running config, then populate with addresses after this loop
	my %dynamic_address_groups;
	my @address_group_entries = $vsys->findnodes('address-group/entry');
	foreach my $entry (@address_group_entries) {
		my $id = $entry->att('name');
		my $parsed_network_group = { 'id' => $id, };

		my $description = get_value_from_xml( 'description', $entry );
		if ( defined $description ) {
			$parsed_network_group->{description} = $description;
		}

		$parsed_network_group->{objectGroupReference} = ();

		# Dynamic address groups addresses are parsed later
		if( $entry->exists('dynamic') ) {
			$dynamic_address_groups{$id} = $parsed_network_group;
		} else {

			# Two XPaths used for pre and post PAN-OS 6.0
			my @members = ();
			foreach my $xpath('member', 'static/member'){
				my $member_nodes = $entry->findnodes($xpath);
				while($member_nodes->size() > 0){
					push @members, $member_nodes->shift();
				}
			}

			foreach my $member (@members) {
				my $reference = get_value_from_xml( '.', $member );
				if( defined $fqdns{$reference} ) {
					$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Full qualified domain names (FQDNs) are not supported. Dropping nested FQDN '$reference'.");
				} else {
					push @{ $parsed_network_group->{objectGroupReference} }, $reference;
				}
			}

			push @parsed_network_groups, $parsed_network_group;
		}
	}

	# Only do this if we have the response with the dynamic address group members
	# and if this is not the shared parser, but the vsys parser
	my $parsed_dynamic_address_groups = _parse_dynamic_address_groups($vsys, $show_dynamic_address_groups, \%dynamic_address_groups);
	push @parsed_network_groups, @{ $parsed_dynamic_address_groups };

	my $profiles = $lookup->get_interface_management_profiles();
	foreach my $profile(@{$profiles}){

		my $parsed_network_group = $profile->{network_group};

		my $id = $profile->{name};
		$parsed_network_group->{id} = "Interface management profile $id network group";
		
		push(@parsed_network_groups, $parsed_network_group);
	}

	my @external_lists = $vsys->findnodes('external-list/entry');
	foreach my $external_list(@external_lists){
		my $name = $external_list->att('name');
		my $parsed_network_group = {'id' => $name};

		my $description = get_value_from_xml('description', $external_list);
		if(!defined($description)){
			$description = "Dynamic Blocking List for $name";
		}
		$parsed_network_group->{description} = $description;

		# Request the addresses in this dynamic block list
		my $dbl = &$request_dynamic_block_list($name);

		my $line = get_value_from_xml('response/msg/line', $dbl);
		while($line =~ /^\s*($ip_regex)(\/(\d{1,2}))?(-($ip_regex)(\/(\d{1,2}))?)?/mg){
			my $ip = $1;
			my $mask = $3 // 32;
			my $range = $4 // undef;

			if($range){
				my $range_ip = $5;
				my $range_mask = $7 // 32;
				my $skip;

				if($range_mask != 32){

					my $subnet_size = 2 ** (32 - $range_mask);
					my @octals = reverse(split(/\./, $range_ip));

			   FIND_SUBNET_END: while(--$subnet_size > 0){
						my $octal_count = 0;
						foreach my $octal(@octals){
							++$octal_count;
							if(++$octal > 255){
								if($octal_count > 3){
									@octals = (255, 255, 255, 255);
									last(FIND_SUBNET_END);
								}
								$octal = 0;
							}else{
								last;
							}
						}
					}
					$range_ip = join('.', reverse(@octals));
				}

				if($ip eq $range_ip){
					push(@{$parsed_network_group->{host}}, $ip);
				}else{
					my @start_octals = split(/\./, $ip);
					my @end_octals = split(/\./, $range_ip);

					my $loop;
					for(my $i = 0; $i < 4; ++$i){
						if($start_octals[$i] > $end_octals[$i]){
							$loop = 1;
						}
						if($start_octals[$i] != $end_octals[$i]){
							last;
						}
					}

					if($loop){
					
						push(@{$parsed_network_group->{range}}, (
							{
								startAddress => $ip,
								endAddress => '255.255.255.255'
							},{
								startAddress => '0.0.0.0',
								endAddress => $range_ip
							}
						));

					}else{

						push(@{$parsed_network_group->{range}}, {
							startAddress => $ip,
							endAddress => $range_ip
						});
					}
				}
			}elsif($mask == 32){
				push(@{$parsed_network_group->{host}}, $ip);
			}else{
				push(@{$parsed_network_group->{subnet}}, {
					address => $ip,
					mask => $mask
				});
			}
		}

		push(@parsed_network_groups, $parsed_network_group);
	}

	# Find all referenced network groups in this virtual context's filters
	my %referenced_network_groups;
	my @rulebase = $vsys->findnodes('rulebase/security/rules/entry');
	foreach my $rule(@rulebase){
		my @addresses = ($rule->findnodes('source/member'), $rule->findnodes('destination/member'));
		foreach my $address(@addresses){
			my $address_value = get_value_from_xml('.', $address);
			if($address_value ne 'any' && $address_value !~ /^\d+\.\d+\.\d+\.\d+(\/\d+)?$/){
				$referenced_network_groups{$address_value} = 1;
			}
		}
	}

	# Find all referenced network groups in this virtual context's NAT filters
	my @nat_rulebase = $vsys->findnodes('rulebase/nat/rules/entry');
	foreach my $nat_rule(@nat_rulebase){
		my @addresses = (
			$nat_rule->findnodes('source/member'),
			$nat_rule->findnodes('destination/member'),
			$nat_rule->findnodes('destination-translation/translated-address'),
			$nat_rule->findnodes('source-translation/static-ip/translated-address'),
			$nat_rule->findnodes('source-translation/dynamic-ip/translated-address/member'),
			$nat_rule->findnodes('source-translation/dynamic-ip-and-port/translated-address/member')
		);
		foreach my $address(@addresses){
			my $address_value = get_value_from_xml('.', $address);
			if($address_value ne 'any' && $address_value !~ /^\d+\.\d+\.\d+\.\d+(\/\d+)?$/){
				$referenced_network_groups{$address_value} = 1;
			}
		}
	}

	# Find any nested shared network groups
	foreach my $shared_network_group(@{$shared_network_groups->{objectGroups}->{networkGroup}}){
		if($shared_network_group->{objectGroupReference}){
			foreach my $object_group_ref(@{$shared_network_group->{objectGroupReference}}){
				$referenced_network_groups{$object_group_ref} = 1;
			}
		}
	}

	# Same as above but look inside network groups defined for this virtual context
	foreach my $parsed_network_group(@parsed_network_groups){
		if($parsed_network_group->{objectGroupReference}){
			foreach my $object_group_ref(@{$parsed_network_group->{objectGroupReference}}){
				$referenced_network_groups{$object_group_ref} = 1;
			}
		}
	}

	# Add any we found that are used
	foreach my $shared_network_group(@{$shared_network_groups->{objectGroups}->{networkGroup}}){
		if($referenced_network_groups{$shared_network_group->{id}}){
			push(@parsed_network_groups, $shared_network_group);
		}
	}
	
	# Add parsed network groups to lookup
	foreach my $group (@parsed_network_groups) {
		$lookup->add_network_group( $group->{id}, $group );
	}

	return \@parsed_network_groups;
}

sub _parse_dynamic_address_groups {
	my ($vsys, $show_dynamic_address_groups, $dynamic_address_groups_ref) = @_;
	my %dynamic_address_groups = %{ $dynamic_address_groups_ref };
	my $vsys_id = $vsys->att('name');

	my @parsed_dynamic_address_groups = ();

	if( defined $show_dynamic_address_groups && defined $vsys_id ) {
		my @dynamic_address_groups = $show_dynamic_address_groups->findnodes('response/result/dyn-addr-grp/entry');
		foreach my $entry (@dynamic_address_groups) {
			my $group_id = get_value_from_xml( 'group-name', $entry );
			my $group_vsys_id = get_value_from_xml( 'vsys', $entry );
 
			if( $group_vsys_id eq $vsys_id ) {
				if( defined $dynamic_address_groups{$group_id} ) {
					my @members = $entry->findnodes('member-list/entry');
					foreach my $member (@members) {
						my $name = $member->att('name');
						my $type = $member->att('type');
						if( $type eq 'registered-ip' ) {
							push @{ $dynamic_address_groups{$group_id}->{subnet} }, { address => $name, mask => 32 };
						} elsif( $type eq 'address-object' ) {
							push @{ $dynamic_address_groups{$group_id}->{objectGroupReference} }, $name;
						} else {
							$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Dynamic address group '$group_id' member '$name' has an unknown type '$type'");
						}
					}

					push @parsed_dynamic_address_groups, $dynamic_address_groups{$group_id};
				} else {
					$LOGGER->warning($NETWORK_OBJECT_DROPPED, "Dynamic address group '$group_id' was not in the running-config but was in the dynamic address groups response");
				}
			}
		}
	} else {
		foreach my $id ( keys %dynamic_address_groups ) {
			$LOGGER->warning($NETWORK_OBJECT_DROPPED, "No response for dynamic address groups. Dropping dynamic address group '$id' found in the running-config");
		}
	}

	return \@parsed_dynamic_address_groups;
}

sub _parse_service_groups {
	my ($vsys, $lookup, $shared_service_groups, $predefined_services ) = @_;

	my @parsed_services;

	# The list of all services includes predefined services and user defined services
	my @service_entries = $vsys->findnodes('service/entry');
	
	if (defined $predefined_services) {
	   push @service_entries, $predefined_services->findnodes('/response/result/service/entry');
	}else{
		$LOGGER->debug('Predefined service groups not parsed beacuse the command failed.');
	}

	# Palo Alto uses the term service to mean a set of source/destination ports on a particular protocol
	# We translate this as a service group
	# this should not be confused with the Palo Alto usage of service group which is dealt with later
	foreach my $entry (@service_entries) {
		my $id = $entry->att('name');
		my $description = get_value_from_xml( 'description', $entry );

		my $protocol;
		if ( $entry->exists('protocol/tcp') ) {
			$protocol = 'tcp';
		} elsif ( $entry->exists('protocol/udp') ) {
			$protocol = 'udp';
		}

		### Since the SED can not represent a service group with a destinction between source and destination ports,
		### We have to split each service into source and destination components.

		# This section parses the destination portion of a service group
		if ( $entry->exists("protocol/$protocol/port") ) {
			my $parsed_service_destination = { 'id' => $id . '_destination', };

			# Only save the description if it is not blank
			if ( defined $description ) {
				$parsed_service_destination->{description} = $description;
			}

			my $ports = get_value_from_xml( "protocol/$protocol/port", $entry );

			# Break the port element into a set of ports and port ranges
			my ( $port_expressions, $port_ranges ) = unwind_port_expression($ports);

			foreach my $expression ( @{$port_expressions} ) {
				push @{ $parsed_service_destination->{portExpression} }, { 'port' => $expression, 'operator' => 'eq', 'protocol' => $protocol };
			}
			foreach my $range ( @{$port_ranges} ) {
				my ( $start, $end ) = $range =~ /(.*)-(.*)/;
				push @{ $parsed_service_destination->{portRange} }, { 'portStart' => $start, 'portEnd' => $end, 'protocol' => $protocol };
			}

			push @parsed_services, $parsed_service_destination;
		}

		# This section parses the source section of a service group
		if ( $entry->exists("protocol/$protocol/source-port") ) {
			my $parsed_service_source = { 'id' => $id . '_source', };

			# Only save the description if it is not blank
			if ( defined $description ) {
				$parsed_service_source->{description} = $description;
			}

			my $ports = get_value_from_xml( "protocol/$protocol/source-port", $entry );

			# Break the port element into a set of ports and port ranges
			my ( $port_expressions, $port_ranges ) = unwind_port_expression($ports);

			foreach my $expression ( @{$port_expressions} ) {
				push @{ $parsed_service_source->{portExpression} }, { 'port' => $expression, 'operator' => 'eq', 'protocol' => $protocol };
			}

			foreach my $range ( @{$port_ranges} ) {
				my ( $start, $end ) = $range =~ /(.*)-(.*)/;
				push @{ $parsed_service_source->{portRange} }, { 'portStart' => $start, 'portEnd' => $end, 'protocol' => $protocol };
			}

			push @parsed_services, $parsed_service_source;
		}

	}

	# Palo Alto uses the term service-group to mean a set of references to 'services' or other service groups
	# We create source and destination versions of each of these groups and then trim all of the empty groups
	# and references to empty groups.
	my @service_group_entries = $vsys->findnodes('service-group/entry');

	my @parsed_service_groups;
	foreach my $entry (@service_group_entries) {
		my $id = $entry->att('name');

		# Break the service group up into source and destination components by default. We will trim the list later
		my $parsed_service_group_source	     = { 'id' => $id . '_source', };
		my $parsed_service_group_destination = { 'id' => $id . '_destination', };

		$parsed_service_group_source->{objectGroupReference}	  = ();
		$parsed_service_group_destination->{objectGroupReference} = ();

		# Set up the objectGroupReferences
		# Two XPaths for pre and post PAN-OS 6.0
		my @members = ();
		foreach my $xpath('member', 'members/member'){
			my $member_nodes = $entry->findnodes($xpath);
			while($member_nodes->size() > 0){
				push(@members, $member_nodes->shift());
			}
		}

		foreach my $member (@members) {

			my $reference = get_value_from_xml( '.', $member );
			push @{ $parsed_service_group_source->{objectGroupReference} },	  $reference . '_source';
			push @{ $parsed_service_group_destination->{objectGroupReference} }, $reference . '_destination';
		}

		push @parsed_service_groups, $parsed_service_group_source;
		push @parsed_service_groups, $parsed_service_group_destination;
	}

	# Trim the service group list. See TrimServiceGroups.pm for details. +++wayne note: I'm not buying this 
	# is entirely necessary, but if so, it needs to take into account the multi-vsys/shared scope defs.
	my $parsed_service_groups = trim_service_groups( \@parsed_services, \@parsed_service_groups, $shared_service_groups );

	# Register the service groups in the lookup table to they can be used later
	foreach my $group ( @{$parsed_service_groups} ) {
		$lookup->add_service_group( $group->{id}, $group );
	}

	my $profiles = $lookup->get_interface_management_profiles();
	foreach my $profile(@{$profiles}){

		my $parsed_service_group = $profile->{service_group};

		my $id = $profile->{name};
		$parsed_service_group->{id} = "Interface management profile $id service group";

		push(@{$parsed_service_groups}, $parsed_service_group);
	}

	foreach my $shared_service_group(@{$shared_service_groups->{objectGroups}->{serviceGroup}}){
		push(@{$parsed_service_groups}, $shared_service_group);
	}

	#return the list of all parsed and decomposed services and all parsed service groups
	return $parsed_service_groups;
}

# Palo Alto uses the concept of security zones when creating ACLs. A security zone is a named set of interfaces.
# In order to be able to propery parse filter lists we need to create a mapping between each zone name and the
# list of interfaces that it represents. When filterlist parsing encounters a zone name it will look up the name
# to find out which interfaces it needs to apply the rule to.
sub _parse_security_zones {

	# an xml entity containing the zone information
	my $zones = shift;

	# a lookup table in which to store the parsed zone information
	my $lookup = shift;

	my @entries = $zones->findnodes('entry');

	foreach my $entry (@entries) {

		my $name = $entry->att('name');

		my @interfaces = ();

		my @members = ();

		#collect the members of this zone
		foreach my $type ( 'tap', 'virtual-wire', 'layer2', 'layer3' ) {
			push @members, $entry->findnodes("network/$type/member");
		}

		#extract the interface names from the member xml
		foreach my $member (@members) {
			push @interfaces, get_value_from_xml( '.', $member );
		}

		#if the zone contains no interfaces, drop it for now.
		if ( @interfaces == 0 ) {
			$LOGGER->debug("Security Zone '$name' contained no interface information and was dropped. Rules using this zone may also be dropped");
		} else {
			$lookup->add_security_zone( $name, \@interfaces );
		}

	}
	return;
}	# end _parse_security_zones

sub _parse_application_groups {
	my $lookup = shift;
	my @application_groups = @_;

	my $group_names;

	my @parsed_application_groups = ();

	foreach my $application_group (@application_groups) {

		my $name = $application_group->att('name');
		my $parsed_application_group = {
			id => $name,
			description => "Application group for application group $name"
		};

		my @members = $application_group->findnodes('member');
		foreach my $member (@members) {

			my $member_name = get_value_from_xml('.', $member);

			if ($lookup->is_application_group_defined($member_name)) {
				push @{$parsed_application_group->{objectGroupReference}}, $member_name;
			}else{
				push @{$parsed_application_group->{application}}, $member_name;
			}
		} #end foreach member

		$lookup->add_application_group($name, $parsed_application_group);

		push @parsed_application_groups, $parsed_application_group;
		
	} #end foreach application group


	return \@parsed_application_groups;
	
} #end sub _parse_application_groups

sub _parse_application_filters{
	my ($lookup, @application_filters) = @_;

	my @parsed_application_filters = ();

	my @classifiers = qw(
		risk
		category
		subcategory
		technology
		evasive-behavior
		consume-big-bandwidth
		used-by-malware
		able-to-transfer-file
		has-known-vulnerability
		tunnel-other-application
		prone-to-misuse
	);

	foreach my $application_filter(@application_filters){

		my $name = $application_filter->att('name');

		my $search_criteria;

		foreach my $filter_type(@classifiers){

			my @members = $application_filter->findnodes("$filter_type/member");

			foreach my $member(@members){

				my $member_value = get_value_from_xml('.', $member);

				push(@{$search_criteria->{$filter_type}}, $member_value);
			}
		}

		my $parsed_application_filter = _parse_application_filter_group($name, $search_criteria, $lookup);

		$lookup->add_application_filter($name, $parsed_application_filter);

		push(@parsed_application_filters, $parsed_application_filter);
	}

	return(\@parsed_application_filters);
}

sub _parse_application_filter_group{
	my ($name, $search_criteria, $lookup) = @_;
	my $parsed_application_filter = {
		id => $name,
		description => "Application group for application filter $name"
	};

	my $applications = $lookup->get_applications();
	my @application_keys = keys(%{$applications});

	foreach my $search_criterion_key(keys(%{$search_criteria})){
		my $search_criterion = $search_criteria->{$search_criterion_key};
		foreach my $search_criterion_value(@$search_criterion){

			my @new_application_keys = ();

			foreach my $application_key(@application_keys){
				my $application = $applications->{$application_key};

				foreach my $criteria_key(keys(%{$application->{criteria}})){

					if($criteria_key eq $search_criterion_key){
						my $criteria = $application->{criteria}->{$criteria_key};

						foreach my $criteria_value(@$criteria){
							if($criteria_value eq $search_criterion_value){
								push(@new_application_keys, $application_key);
							}
						}
					}
				}
			}
			@application_keys = @new_application_keys;
		}
	}

	$parsed_application_filter->{application} = \@application_keys;

	return($parsed_application_filter);
}

sub _print_application_groups {
	my ( $system, $out ) = @_;
	
	if (!defined $system->{applicationGroup}) {
		return;
	}
	
	foreach my $group (@{$system->{applicationGroup}}) {
		$out->open_element('applicationGroup');
		
		print_if_defined($group, 'description', $out);
		print_if_defined($group, 'id', $out);
		print_if_defined($group, 'objectGroupReference', $out);
		print_if_defined($group, 'application', $out);
		
		$out->close_element('applicationGroup');
	}

	return;
}



sub _print_object_groups {
	my ( $system, $out ) = @_;
	
	if (!defined $system->{objectGroups}) {
		return;
	}
	
	$out->open_element('objectGroups');
	
	#print application groups
	_print_application_groups($system->{objectGroups}, $out);
	
	#print network groups
	if($system->{objectGroups}->{networkGroup}){

		foreach my $network_group(@{$system->{objectGroups}->{networkGroup}}){
			$out->open_element('networkGroup');

			print_if_defined($network_group, 'description', $out);
			print_if_defined($network_group, 'id', $out);
			print_if_defined($network_group, 'objectGroupReference', $out);
			print_if_defined($network_group, 'host', $out);
			print_if_defined($network_group, 'range', $out);
			print_if_defined($network_group, 'subnet', $out);

			$out->close_element('networkGroup');
		}
	}
	
	#print service groups
	print_if_defined($system->{objectGroups}, 'serviceGroup', $out);
	
	# print user groups
	if($system->{objectGroups}->{userGroup}){

		foreach my $user_group (@{$system->{objectGroups}->{userGroup}}){
			$out->open_element('userGroup');

			print_if_defined($user_group, 'id', $out);
			foreach my $user (@{$user_group->{user}}){
				$out->print_element('user', $user);
			}
			$out->close_element('userGroup');
		}
	}
	
	
	$out->close_element('objectGroups');
	
	return;
}

sub parse_virtual_routers {
	my ( $in, $out ) = @_;

	my $show_config_running = $in->{'show_config_running'};
	my $lookup = $in->{lookup_tables};

	#get a list of all virtual routers on the device
	my @routers = $show_config_running->findnodes('/response/result/config/devices/entry/network/virtual-router/entry');

	foreach my $router (@routers) {
		my $parsed_router;

		my $name = $router->att('name');
		$parsed_router->{'core:systemName'} = $name;

		#default to Router since we are parsing a virtual router
		$parsed_router->{'core:deviceType'} = 'Router';

		#default to virtual since we are parsing a virtual router
		$parsed_router->{'subelementType'} = 'virtual';

		$LOGGER->debug( "Importing interface information for virtual router '$name'" );

		#the names of the interfaces imported to this virtual device are stored in interface/member elements
		my @imported_interfaces = $router->findnodes('interface/member');
		$parsed_router->{interfaces} = _import_interfaces( \@imported_interfaces, $lookup );

		#Extract static route elements
		$LOGGER->debug( "Parsing static route information for virtual router '$name'" );
		my @static_routes = $router->findnodes('routing-table/ip/static-route/entry');

		#parse route information for each route
		foreach my $static_route (@static_routes) {
			push @{ $parsed_router->{staticRoutes}->{staticRoute} }, _parse_static_route($static_route);
		}

		#$in->{parsed}->{subelements}->{subelement} should already be populated with virtual systems
		#just add the virtual routers to the list
		push @{ $in->{parsed}->{subelements}->{subelement} }, $parsed_router;
	}

	return;
}

sub _import_interfaces {
	my $nodes  = shift;
	my $lookup = shift;

	my $interfaces->{interface} = ();

	foreach my $member ( @{$nodes} ) {

		#extract the name of the interface, which is stored in a 'member' element
		my $if_name = get_value_from_xml( '.', $member );

		#Lookup the name in the table to see if it exists
		my $interface = $lookup->get_interface($if_name);

		if ( defined $interface ) {
			push @{ $interfaces->{interface} }, $interface;
		} else {
			$LOGGER->debug("WARNING: Failed to import interface '$if_name' into a virtual device. This interface does not exist in the lookup table.");
		}
	}

	return $interfaces;
}

sub _parse_static_route {
	my $route = shift;

	my $parsed_static_route;

	$parsed_static_route->{defaultGateway} = 'false';

	my $destination = get_value_from_xml( 'destination', $route );

	my ( $address, $mask ) = $destination =~ /(\d+\.\d+\.\d+\.\d+)\/(\d+)/;

	$parsed_static_route->{destinationAddress} = $address;
	$parsed_static_route->{destinationMask}	= $mask;

	my @discard = $route->findnodes('nexthop/discard');
	my @next_vr = $route->findnodes('nexthop/next-vr');
	if (@discard) {
		$parsed_static_route->{discard} = 'true';
	} elsif (@next_vr) {
		$parsed_static_route->{discard} = 'false';

		#how to define the next hop as another virtual router?
	} else {
		$parsed_static_route->{discard} = 'false';

#		my $gateway_address = get_value_from_xml( 'nexthop/ip-address', $route );
#		if ( defined $gateway_address ) {
#			$gateway_address =~ s/\/\d+//;
#			$parsed_static_route->{gatewayAddress} = $gateway_address;
#		}
	}

	my $interface = get_value_from_xml( 'interface', $route );
	if ( defined $interface ) {
		$parsed_static_route->{interface} = $interface;
	} else {
		#interface is required in the SED, but routes that discard packets don't have interfaces
		$parsed_static_route->{interface} = 'null';
	}

	#get the route metric
	my $metric = get_value_from_xml( 'metric', $route );
	if ( defined $metric ) {
		$parsed_static_route->{routeMetric} = $metric;
	}

	#get the admin distance
	my $preference = get_value_from_xml( 'admin-dist', $route );
	if ( defined $preference ) {
		$parsed_static_route->{routePreference} = $preference;
	}

	return $parsed_static_route;

}

sub print_shared_object_groups
{
	my ($in, $out) = @_;
	print_if_defined($in->{parsed}->{sharedElements}, 'objectGroups', $out);
	return;
}

sub print_sub_elements {
	my ( $in, $out, $qradarVersion ) = @_;

	$out->open_element('subelements');

	foreach my $system (@{ $in->{parsed}->{subelements}->{subelement} }) {

		$out->open_element('subelement');

		print_if_defined( $system, 'core:systemName', $out );

		print_if_defined( $system, 'core:deviceType', $out );

		print_if_defined( $system, 'applications', $out );

#		print_if_defined( $system, 'objectGroups', $out );
		_print_object_groups( $system, $out );

		print_filter_lists( $system, $out, $in->{'qradarVersion'} );

		$out->open_element('interfaces');
		foreach my $interface ( @{ $system->{interfaces}->{interface} } ) {
			_print_interface($interface, $out);
		}
		$out->close_element('interfaces');

		print_if_defined( $system, 'staticRoutes', $out );

		print_if_defined( $system, 'subelementType', $out );

		$out->close_element('subelement');
	}

	$out->close_element('subelements');

	return;
}

# Print the manifest of packet processing for PANOS
sub _print_packet_process_ordering
{
	my $out = shift;

	$out->open_element('core:packetProcessingOrdering');

	foreach my $mPhase ('address-translation/destination', 'routing', 'address-translation/source', 'layer2-bridging',
				'packet-filtering' )
	{
		my ($pPhase, $pPosition) = split('/', $mPhase);
		$out->open_element('core:packetProcessingElement');
		$out->print_element('core:packetProcessingPhase', $pPhase);
		if (defined $pPosition)
		{
			$out->print_element('core:processingPoint', $pPosition);
		}
		$out->close_element('core:packetProcessingElement');
	}
	$out->close_element('core:packetProcessingOrdering');

	return;
}

sub parse_applications {
	my ($vsys, $lookup) = @_;

	my @applications = $vsys->findnodes('application/entry');
	my @parsed_applications = ();

	foreach my $application (@applications) {
		my $parsed_application;
		my $lookup_application;

		my $name = $application->att('name');
		$parsed_application->{'ztApp:appName'} = $name;
		
		my $description = get_value_from_xml('description', $application);
		if ($description) {
			$parsed_application->{'ztApp:appDescription'} = $description;
		}

		my $risk = get_value_from_xml('risk', $application);
		$parsed_application->{'ztApp:appRiskSeverity'} = $risk;

		if(defined($risk)){
			push(@{$lookup_application->{criteria}->{risk}}, $risk);
		}

		# This is a list of known categories used for classification
		my @classifiers = qw(
			category
			subcategory
			technology
			evasive-behavior
			consume-big-bandwidth
			used-by-malware
			able-to-transfer-file
			has-known-vulnerability
			tunnel-other-application
			prone-to-misuse
			pervasive-use
			file-type-ident
			virus-ident
			spyware-ident
			data-ident
			parent-app
		);

		# Each application has a set of classifiers
		foreach my $classifier (@classifiers) {

			my $app_classification;
			$app_classification->{'ztApp:classifierName'} = $classifier;

			my $classifier_value = get_value_from_xml($classifier, $application);
			$app_classification->{'ztApp:classifierValue'} = $classifier_value;

			if(defined($classifier_value)){
				push(@{$lookup_application->{criteria}->{$classifier}}, $classifier_value);
			}

			# Adapter level logging to detect when these strings are longer than the database can handle.
			if (length $app_classification->{'ztApp:classifierName'} > 128) {
				$LOGGER->debug('Clasifier name: ' . $app_classification->{'ztApp:classifierName'} . ' is bigger than 128 characters for application ' . $application->att('name'));
			}
			if (defined $app_classification->{'ztApp:classifierValue'} and length $app_classification->{'ztApp:classifierValue'} > 128) {
				$LOGGER->debug('Clasifier Value: ' . $app_classification->{'ztApp:classifierValue'} . ' is bigger than 128 characters for application ' . $application->att('name'));
			}
			if (defined $app_classification->{'ztApp:classifierValue'}) {
				push @{$parsed_application->{'ztApp:appClassification'}}, $app_classification;
			}
		}


		# Find the protocol/port entries for this application (if any)
		my @ports = $application->findnodes('default/port/member');
		my $first = 1;
		my $lookup_port;
		foreach my $port_entry (@ports) {
			my $parsed_port;

			# Get the entry and break it up into port and protocol
			my $value = get_value_from_xml('.', $port_entry);
			my ($protocol, $port) = $value =~ /(tcp|udp)\/(\S+)/i;

			if (defined $protocol) {

				# Upcase the protocol to fit it in the SED
				$parsed_port->{'ztApp:protocol'} = uc $protocol;

				# Represent a port entry of 'dynamic' as all possible ports
				if ($port =~ /dynamic|any/i) {
					$parsed_port->{'ztApp:port'}->{'ztApp:portRange'}->{'ztApp:portHigh'} = 65535;
					$parsed_port->{'ztApp:port'}->{'ztApp:portRange'}->{'ztApp:portLow'} = 1;

					push(@{$lookup_port->{portRange}}, {
						portStart => 1,
						portEnd => 65535,
						protocol => lc($protocol)
					});
				}else{
					my ($ports, $port_ranges) = unwind_port_expression($port);
					if (@{$ports}) {
						my $port_value = shift(@{$ports});

						$parsed_port->{'ztApp:port'}->{'ztApp:portValue'} = $port_value;

						push(@{$lookup_port->{portExpression}}, {
							port => $port_value,
							operator => 'eq',
							protocol => lc($protocol)
						});
					}
					elsif (@{$port_ranges}) {
						my ($low, $high) = split '-', shift @{$port_ranges};
						$parsed_port->{'ztApp:port'}->{'ztApp:portRange'}->{'ztApp:portHigh'} = $high;
						$parsed_port->{'ztApp:port'}->{'ztApp:portRange'}->{'ztApp:portLow'} = $low;

						push(@{$lookup_port->{portRange}}, {
							portStart => $low,
							portEnd => $high,
							protocol => lc($protocol)
						});
					}
				}

				if($first){
					push @{$parsed_application->{'ztApp:appNetworkPort'}}, $parsed_port;
				}
			}else{
				next;
			}

			# For now only parse the first port as the schema only supports one protocol definition
			$first = 0;
		}

		if($lookup_port){
			$lookup_application->{service_group} = $lookup_port;
		}

		# Store in the lookup table for finding it later
		if($lookup){
			$lookup->add_application($name, $lookup_application);
		}

		push @parsed_applications, $parsed_application;
	}

	if(@parsed_applications){
		return( { application => \@parsed_applications } );
	}else{
		return;
	}
}

sub parse_predefined_and_shared_applications{
	my ($in) = @_;

	my $lookup = $in->{lookup_tables};
	my $predefined = $in->{show_predefined_applications};
	my $show_config_running = $in->{show_config_running};

	my $predefined_nodes = $predefined->findnodes('/response/result');
	my $predefined_applications = shift(@{$predefined_nodes});

	$LOGGER->debug("Parsing predefined applications");
	parse_applications($predefined_applications, $lookup);

	my $shared_nodes = $show_config_running->findnodes('/response/result/config/shared');
	my $shared_applications = shift(@{$shared_nodes});

	$LOGGER->debug("Parsing shared applications");
	parse_applications($shared_applications, $lookup);
}

sub get_serial_number {
	my ($show_device_connected, $ip_address) = @_;

	if (defined $show_device_connected) {
		#parse the list of serial numbers
	}

	return;
}

=begin to be used for future user group modelling
sub parse_user_groups {
	my $vsys      = shift;
	my $lookup = shift;
	my $show_user_group_name = shift;
	
	my @parsedUserGroups;
	if (!$show_user_group_name){
		my @user_groups = $vsys->findnodes('local-user-database/user-group/entry');
		foreach my $entry (@user_groups){
			my $name = $entry->att('name');
			my $userGroup->{id} = $name;
			# get users
			my @userGroupMembers = $entry->findnodes('user/member');
			foreach my $member (@userGroupMembers){
				my $user = get_value_from_xml( '.', $member );
				push @{$userGroup->{user}}, $user;
			}
			$lookup->add_user_group($name, $userGroup);
			push @parsedUserGroups, $userGroup;
		}
	} else { # must be a vsystem
		foreach my $userGroupName (keys %{$show_user_group_name}){
			my $userGroupResult = $show_user_group_name->{$userGroupName}->findvalue('/response/result');
			my $userGroup;
			foreach my $line (split ("\n", $userGroupResult)){
				if ($line =~ /^short\sname:\s+(.*)/){
					my $userShortName = lc ($1);
					$userGroup->{id} = $userShortName;
					$lookup->add_user_group($userGroupName, $userGroup, $userShortName);
				} elsif ($line =~ /^\[\d+\s+\]\s(.*)/){
					my $user = $1;
					push @{$userGroup->{user}}, lc($user) if $user ne "";
				}
			}
			push @parsedUserGroups, $userGroup if defined $userGroup->{id};
		}
	}
	
	return \@parsedUserGroups;
}
=cut


__END__

=head1 NAME

ZipTie::Adapters::PaloAlto::PANOS::Parsers - Parsers to transform Palo Alto configuration data in to Standard Element Document format.

=head1 VERSION

This document describes ZipTie::Adapters::PaloAlto::PANOS::Parsers version 1.1

=head1 SYNOPSIS

	use ZipTie::Adapters::PaloAlto::PANOS::Parsers;
	create_config ($in, $out);
	get_serial_number ($show_device_connected, $ip_address);
	parse_applications($in, $out);
	parse_chassis ($in, $out);
	parse_interfaces ($in, $out);
	parse_shared_elements($in, $out);
	parse_snmp ($in, $out)
	parse_system ($in, $out);
	parse_virtual_routers ($in, $out);
	parse_vlans ($in, $out);
	parse_vsystems ($in, $out);
	print_chassis ($in, $out);
	print_config ($in, $out);
	print_interfaces ($in, $out);
	print_snmp ($in, $out)
	print_sub_elements ($in, $out);
	print_system ($in, $out);
	print_vlans ($in, $out);

=head1 DESCRIPTION

 This module contains various functions to allow the parsing of Palo Alto device information into Standard Element Document (SED) format.
 For each relevent section of the SED there is a parser function and a matching printer function. This allows printing to be done in one order,
 while parsing is done in another. The only stipulation is that the parse function must be called before the matching print function.

 Each parser and printer function takes two parameters:
 1) A hash of XML Twig objects known as $in. This contains the raw responses to various commands that were issued to the device.
 2) An XMLPrint object known as $out. This is the printer that will produce the SED.

 Some parsed information is stored in the $in hash in the $in->{parsed} element. This allows easy transfer of information from one parsing
 function to the next. Other information is stored in a LookupTable object stored in $in-{lookup_tables}


=head1 SUBROUTINES/METHODS

=over 12


=item C<create_config($in, $out)>

This function constructs the configRepository element. It includes the base 64 encoded system configuration from the following command:

	show config running

	
=item C<get_serial_number($show_device_connected, $ip_address)>

This function is a stub for Panarama integration


=item C<parse_application>

This function parsed palo alto application information. It takes an xml element containing application entries and
returns a list of hashes containing application information suitable for XML printing. This function can be used to 
parse either user defined applications or vendor defined application as they have the same format in Palo Alto's XML.
At present this function parses a significant number of application categorization elements, it is unlikely that they
are all needed.

LIMITATION - at present this function only parses one set of protocol information although palo alto can define many. This
is a schema limitation.


=item C<parse_chassis($in, $out)>

This function parses chassis information. It uses information from the following commands:

	show system info
	show config running
	show system resources
	

=item C<parse_interfaces($in, $out)>

This function parses interface information. It uses information from the following commands:

	show interfaces
	show config running

This function relies on helper functions to perform the actual parsing. Its role is to determine the list of interfaces present
on the device and select the correct XML needed for the parsing.

This function also stores each interface in a lookup table, indexed by name so that other functions that need interface
information can find it easily.


=item C<parse_snmp($in, $out)>

This function parses a device's snmp settings. It uses information from the following commands:

	show system info
	show config running


=item C<parse_system($in, $out)>

This function parses basic system information. It uses information from the following commands:

	show system info
	show config running
	

=item C<parse_virtual_routers($in, $out)>

This function parses information about the virtual routers configured on this device.

It uses information from the following commands:

	show config running
	
It also uses information previously parsed in parse_interfaces and stored in a LookupTable object.


=item C<parse_vlans($in, $out)>

This function parses vlan information from the device. On a Palo Alto device vlans are defined at the device level rather than
at the vsys level. This function uses output from the "show vlan all" command.

Note that vlans on Palo Alto devices do not appear to have vlan numbers. Vlan number is a required element in the SED.
Currently the adapter reads the number of the Vlan interface associated with the vlan and uses that as the vlan number.
In general this should be filled in, as Palo alto documentation suggests that it be included and the device issues
a warning if it is not. However, this is not actually mandatory so Vlans that do not have an associated vif are currently
dropped.


=item C<parse_vsystems($in, $out)>

This function is a top level parser that extracts information about the virtual systems configured on a device and parses it
into SED subelements. It relys on helper functions to populate interface and objectGroup information for each virtual system.
This function also fills in high level information for the subelement SED.

This function uses information from the following commands:

	show config running

It also uses information previously parsed in parse_interfaces and stored in a LookupTable object.

=item C<sub parse_shared_elements($in, $out)>

This function is a top level parser that extracts information about the "shared" area of the top-level configuration
(i.e, that which is not contained within a subelement/vsys) and picks up the applicable definitions.  It works as 
something of a "parallel" function to parse_vsystems.

This function uses information from the following commands:

	show config running

It also uses information previously parsed in parse_interfaces and stored in a LookupTable object.

=item C<print_chassis($in, $out)>

This function prints the SED elements parsed in parse_chassis. The following elements are printed if present:

	chassis/core:asset
	chassis/macAddress
	chassis/memory
	chassis/powersupply
	

=item C<print_config($in, $out)>

This function prints the SED elements parsed in create_config. The following elements are printed if present:

	core:configRepository
	

=item C<print_interfaces($in, $out)>

This function prints interface the information parsed in parse_interfaces.


=item C<print_sub_elements($in, $out)>

This function prints the subelements parsed by parse_vsystems and parse_virtual_routers.

Sub elements represent virtual devices configured on the main device. In the case of Palo Alto these represent virtual
firewalls and virtual routers.

This function prints the following SED elements in a subElement in the SED if they are defined:

	core:systemName
	core:deviceType
	objectGroups
	filterLists
	interfaces
	staticRoutes
	subelementType


=item C<print_snmp($in, $out)>

This function prints the SED elements parsed by parse_snmp. It prints the snmp section of the SED.


=item C<print_system($in, $out)>

This function prints the SED elements parsed in parse_system. The following elements are printed if present:

	core:systemName
	core:osInfo/core:make
	core:osInfo/core:name
	core:osInfo/core:softwareImage
	core:osInfo/core:version
	core:osInfo/core:osType
	core:biosVersion
	core:deviceType
	core:contact
	core:lastReboot
	

=item C<print_vlans($in, $out)>

This function prints the elements parsed by parse_vlans.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12


=item C<_import_interfaces($nodes, $lookup)>

This function takes a reference to a list of XML elements representing interfaces to be imported into a vsys or virtual router 
along with a lookup table containing parsed interface information.
It retrieves parsed interface information for the listed interfaces from the table. If the interface exists
in the lookup table the parsed SED hash is saved, otherwise the interface is ignored.
This function returns a list of hashes representing the interfaces that the virtual device imports.


=item C<_parse_applications($vsys)>

This function takes an XML element representing a virtual system and parses the user defined applications associated
with that application. It returns a list of SED application elements attached to a hash, suitable for XMLPrinting.


=item C<_parse_interface($parameters)>

This function takes a hash of XML elements and parses the XML into a hash of SED elements suitable for printing. It parses the following elements:

	adminStatus
	ifIndex
	interfaceEthernet - using a helper function.
	interfaceIp - using a helper function.
	interfaceTunnel - using a helper function.
	interfaceType
	name
	physical
	speed
	description


=item C<_parse_interface_ethernet($show_interface_all)>

This function takes an xml element representing the output of the show interfaces all command for a particular interface
and parses the EthernetInterface information it contains. It returns a hash of SED elements suitable for printing. It parses the following elements:

	autoDuplex
	autoSpeed
	macAddress
	operationalDuplex
	

=item C<_parse_interface_ip($in, $out)>

This function takes an xml element representing the output of the show interfaces all command for a particular interface
and parses the IpInterface information it contains. It returns a hash of SED elements suitable for printing. It parses the following elements:

	ipConfiguration/address
	ipConfiguration/mask


=item C<_parse_ipsec_tunnel($name, $show_config_running)>

This function takes the name of an ipSec tunnel configured on the device and the device's running config. It returns a hash representing
an interfaceTunnel element. It parses the following elements:

	encapsulationProtocol - Defaults to ipSec
	tunnelLocalPeering/tunnelLocalPeeringInterface
	tunnelRemotePeerIPAddress

tunnelLocalPeeringInterface is used rather than tunnelLocalPeeringIpAddress as it will always be present. tunnelLocalPeeringIpAddress will not
always be available.

This function parses information about manual-key, auto-key, and global-protect-satellite tunnels.

=item C<_parse_management_profile($profiles, $lookup)>

This function takes XML elements representing interface management profiles and creates a network group and service group for each of them.
It then puts a management profile hash in the lookup table for later use.


=item C<_parse_network_groups($addresses, $address_groups)>

This function takes XML elements representing named addresses and named address groups and parses them into networkGroup SED elements.
It returns a list of hashes, each hash representing one networkGroup element.
It parses the following networkGroup elements:

	id
	description
	subnet/address
	subnet/mask
	range/startaddress
	range/endaddress
	member

It also parses interface management network groups from the lookup table.
It also uses a callback to make an API call requesting for any dynamic block lists needed, then it creates a network group for each list.


=item C<_parse_object_groups($in, $vsys)>

This function takes an XML element representing a virtual system and separates out the subelements relevent to object groups defined in
that virtual system. It then passes those subelements to other functions for parsing. It takes the responses from those other functions
and constructs a hash representing the objectGroups SED element. This hash is returned.


=item C<_parse_pppoe_tunnel()>

This function takes an XML element representing an interface and searches for PPPoE sub-elements.
It then creates an interfaceTunnel element with the following elements:

	encapsulationProtocol - Defaults to other
		tunnelLocalPeering/tunnelLocalPeeringInterface - Depending on whether a static-ip was assigned


=item C<_parse_security_zones($zones, $lookup)>

This function takes an xml element representing Palo Alto security zone information and a lookup table (hash reference) to store the
parsed security information in. It extracts a list of all of the interfaces associated with the each security zone and stores that list in
the lookup table references by the zone's name. The SED does not support the idea of security zones or interface groups, so this information
is not output directly into the SED. It is intended to be used with the B<_get_security_zone> function when ACL information, NAT information,
and any other information referencing security zones is parsed.


=item C<_parse_service_groups($services, $service_groups, $predefined_services, $lookup)>

This function takes XML elements representing named services, named service groups, predefined services and parses them into
serviceGroup SED elements. It returns a list of hashes, each hash representing one serviceGroup element.
It parses the following serviceGroup elements

	id
	description
	portExpression/port
	portExpression/operator
	portExpression/protocol
	portRange/startPort
	portRange/endPort
	portRange/protocol
	member

Note that the SED schema currently does not support the idea of a service group with source and destination components. Therefor
Palo Alto 'services' are broken up into two sub-serviceGroups, one ending in _source and one ending in _destination.
When these sub-serviceGroups are created references to them are stored in a lookup table so that references to the original
service can be resolved to the sub-serviceGroup(s) that were created.
Palo Alto 'service-group' elements do not contain direct port and protocol information like 'services', so they are not included in
this lookup table.


=item C<_parse_static_route($route)>

This function takes an Palo Alto XML element representing a static route and converts it into a staticRoute SED element.
It parses the following elements:

	defaultGateway
	destinationAddress
	destinationMask
	discard
	interface - hardcoded to 'null' when it does not exist, ie when the route is set to discard
	routeMetric
	routePreference


=back

=head1 DIAGNOSTICS

Functions in this module are tested by unit tests in the ../../../../tests directory in the file 01_VM300.t

These warnings may appear in the adapter log file in /usr/share/ziptie-server/osgi-config/scratch/logs

=over 12

=item C<WARNING: Failed to import interface '$if_name' into a virtual device. This interface does not exist in the lookup table.>

This warning means that a virtual device (either a virtual system or a virtual router) is configured with and interface that was not
parsed by the 'parse_interfaces' function.

=back

=head1 CONFIGURATION AND ENVIRONMENT

N/A

=head1 DEPENDENCIES

This module requires a specific version B<(which one?)> of ziptie-common and ziptie-server.

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

This software is in active development. It should not be used in a production environmnent.

At present the parse_applications function only parses one set of protocol information although palo alto can
define many. This is a schema limitation.

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2013 IBM. All rights reserved.

=head1 AUTHOR

Contributor(s): David Tickner dtickner@ca.ibm.com, 
    Jak Devaney john.devaney@uk.ibm.com,
    Wayne Tackabury wtackabury@us.ibm.com

Date: October 2013

=cut
