package ZipTie::Adapters::PaloAlto::PANOS::Rulebase;

use strict;
use warnings;

use 5.0100000;    # use perl version 5.10
use XML::Twig::XPath;
use ZipTie::Adapters::PaloAlto::PANOS::LookupTable;
use ZipTie::Adapters::PaloAlto::PANOS::ParsingUtils qw(get_value_from_xml print_if_defined convert_rule_action);
use ZipTie::Adapters::Utils qw(get_crep parseCIDR);
use ZipTie::Model::AddressTranslation;
use ZipTie::Logger;

our @EXPORT_OK = qw(parse_acls print_filter_lists parse_nat);
use Exporter 'import';

our $VERSION = 1.0;

my $LOGGER = ZipTie::Logger::get_logger();

# This holds a long term expansion of the interfaces by zone, coupled with those for the "all" 
# pseudo-zone
# my %NATZoneHash = ();

#Each palo alto security policy needs to be represented as a separate ACL so that it can be applied to the correct interfaces
sub parse_acls {
	my $vsys  = shift;    #list reference to rules entries
	my $lookup = shift;    #Lookup table object

	my $parsed_filter_lists;
	my @parsed_rules = ();

	my $countDroppedRules = 0;
	my $rules = $vsys->findnodes('rulebase/security/rules/entry');
	foreach my $rule ( @{$rules} ) {
		my $parsed_rule;

		$parsed_rule->{name}          = $rule->att('name');
		$parsed_rule->{mode}          = 'stateful';
		$parsed_rule->{defaultAction} = 'none';

		#palo alto can define when logging happens, for our purposes we only care if logging happens at all
		my $start_log = get_value_from_xml( 'log-start', $rule );
		my $end_log   = get_value_from_xml( 'log-end',   $rule );

		$parsed_rule->{filterEntry}->{'log'} = 'false';
		if ($start_log)
		{
			if ( ( $start_log eq 'yes' ) or ( $end_log eq 'yes' ) ) {
				$parsed_rule->{filterEntry}->{'log'} = 'true';
			}
		}

		my $action = get_value_from_xml( 'action', $rule );
		$parsed_rule->{filterEntry}->{primaryAction} = convert_rule_action($action);

		$parsed_rule->{filterEntry}->{processOrder} = 1;

		#determine which interfaces this rule applies to
		$parsed_rule->{appliesTo} = _parse_acl_application( $rule, $lookup );

		#Parse information about source addresses this rule applies to
		my @source_addresses = $rule->findnodes('source/member');
		$parsed_rule->{filterEntry}->{sourceIpAddr} = _parse_filter_ip_addresses(\@source_addresses, $lookup);

		#Parse information about destination addresses this rule applies to
		my @destination_addresses = $rule->findnodes('destination/member');
		$parsed_rule->{filterEntry}->{destinationIpAddr} = _parse_filter_ip_addresses(\@destination_addresses, $lookup);
		
		next if missing_network_object_references ($lookup, $parsed_rule, $LOGGER);

		my @users = $rule->findnodes('source-user/member');
		my $isUserBasedFilter = 0;
		foreach my $user(@users){
			my $user_name = get_value_from_xml('.', $user);
			if ( defined ($user_name) && $user_name ne "any" ) {
				$isUserBasedFilter = 1;
				$parsed_rule->{userFilterEntry} = $parsed_rule->{filterEntry};
				if ($user_name =~ /^cn=(.*),cn=/){
					$user_name = $1;
				}	
				push @{ $parsed_rule->{userFilterEntry}->{sourceUser}}, {userName => $user_name};
			}
		}

		my @applications = $rule->findnodes('application/member');

		my @services = $rule->findnodes('service/member');
		foreach my $service (@services) {

			my $servicename = get_value_from_xml( q{.}, $service ); #get the value in the service element

			#if service name is 'any' then just skip to the end. dont fill in any service information
			if ( $servicename eq 'any' ) {
				last;
			}elsif($servicename eq 'application-default'){

				foreach my $application(@applications){
					my $application_name = get_value_from_xml('.', $application);

					$lookup->add_default_port_service_group($application_name);

					push(@{$parsed_rule->{filterEntry}->{destinationService}}, {objectGroupReference => $application_name});
				}
			}else{
				#Lookup the service group in the table
				my $decoded_group = $lookup->get_service_group($servicename);

				#place the source and destination service groups in the SED if they exist
				if ( defined $decoded_group->{destination} ) {
					push @{ $parsed_rule->{filterEntry}->{destinationService} }, { objectGroupReference => $decoded_group->{destination} };
				}

				if ( defined $decoded_group->{source} ) {
					push @{ $parsed_rule->{filterEntry}->{sourceService} }, { objectGroupReference => $decoded_group->{source} };
				}
			}
		} #end foreach service
		
		foreach my $member (@applications) {
			my $application = get_value_from_xml( '.', $member );
			
			if ( $application eq 'any' ) {
				last;
			}

			if($lookup->is_application_group_defined($application)
			|| $lookup->get_application_filter($application)) {
				push @{ $parsed_rule->{filterEntry}->{application} }, { groupReference => $application };
			} else {
				push @{ $parsed_rule->{filterEntry}->{application} }, { applicationID => $application };
			}

		} #end foreach member applications

		$parsed_rule->{description} = get_value_from_xml( 'description', $rule );

		if ($isUserBasedFilter) {
			delete $parsed_rule->{filterEntry};
		} 
		push @parsed_rules, $parsed_rule;
	}
	
	my @defaultRules = $vsys->findnodes('rulebase/default-security-rules/rules/entry');
	# parse default rules
	my $defaultOverrides = parse_default_rules (\@defaultRules);

	#this rule is added to represent the fact that traffic inside a security zone is allowed by default
	my $default_allow = create_default_rule($defaultOverrides, "intrazone-default");

	@{ $default_allow->{appliesTo} } = $lookup->applies_to_within_zones();

	#This rule is added to represent the fact that all traffic between security zones is blocked by default
	my $default_deny = create_default_rule($defaultOverrides, "interzone-default");

	@{ $default_deny->{appliesTo} } = $lookup->applies_to_across_zones();

	push @parsed_rules, $default_allow, $default_deny;

	if (@parsed_rules) {
		$parsed_filter_lists->{filterListsSection}->{filterList} = \@parsed_rules;
	}

	return $parsed_filter_lists;
}

sub _parse_filter_ip_addresses {
	my ($addresses, $lookup) = @_;
	my $ip_regex = get_crep('cipm');

	my @filter_ip_entry;

	foreach my $address (@{$addresses}) {
		my $address_value = get_value_from_xml( '.', $address );

		# if the ip address is any, skip strait to the end
		if ( $address_value eq 'any' ) {
		    last;
		}

		#match the IP address, and potentially the mask or range end
		if($address_value =~ /^($ip_regex)(\/(\d+)|-($ip_regex))?$/){
			my ($ip, $mask, $range_end) = ($1, $3, $4);

			if (defined $mask) {
				my $network = {address => $ip, mask => $mask};
				push @filter_ip_entry, { network => $network };
			} elsif( defined $range_end ) {
				my $range = {startAddress => $ip, endAddress => $range_end};
				push @filter_ip_entry, { range => $range };
			} else {
				push @filter_ip_entry, { host => $ip };
			}
		} else {
			push @filter_ip_entry, { objectGroupReference => $address_value };
		}
	}

	return \@filter_ip_entry;
}

sub _parse_acl_application {
	my $rule   = shift;
	my $lookup = shift;    #Lookup table object

	my @applies_to = ();

	my @source_zones;
	my @destination_zones;

	#get the source and destination zones that this policy applies to
	my @source_zone_members = $rule->findnodes('from/member');
	my @destination_zone_members = $rule->findnodes('to/member');

	#extract the zone names
	foreach my $destination_zone (@destination_zone_members) {
		push @destination_zones, get_value_from_xml('.', $destination_zone);
	}

	#extract the zone names
	foreach my $source_zone (@source_zone_members) {
		push @source_zones, get_value_from_xml('.', $source_zone);
	}
	
	# if 6.1 or > add new rule type support
	if ($lookup->get_os_version() && $lookup->get_os_version() >= 6.1){
		my $ruleType = get_value_from_xml( 'rule-type', $rule );
		if (!$ruleType || $ruleType eq "universal"){
			my $universal_pairs = create_interzone_universal_pairs (\@source_zones, \@destination_zones, 1);
			push @applies_to, @{$universal_pairs};	
		}elsif ($ruleType eq "intrazone"){
			my $intrazone_pairs = create_intrazone_pairs (\@source_zones);
			push @applies_to, @{$intrazone_pairs};
		}elsif ($ruleType eq "interzone"){
			my $interzone_pairs = create_interzone_universal_pairs (\@source_zones, \@destination_zones);
			push @applies_to, @{$interzone_pairs};
		}
	} else{
		#fill in all combinations of sources and destinations
		foreach my $destination_zone (@destination_zones) {
			foreach my $source_zone (@source_zones) {
				my $applies_to;
				$applies_to->{sourceGroup}       = $source_zone;
				$applies_to->{destinationGroup}  = $destination_zone;
				push @applies_to, $applies_to;
			}
		}
	}
	return \@applies_to;
}

#we have to break up filter list printing because source and destination in appliesTo are out of order
sub print_filter_lists {
	my ( $system, $out, $qradarVersion ) = @_;

	if ( !defined $system->{filterLists} && !defined $system->{natFilterLists} ) {
		return;
	}

	$out->open_element('filterLists');

	# we start with the destination NAT filters
	if (defined ($system->{natFilterLists}))
	{
		_generate_nat_translation($out, $system->{natFilterLists}, 'destination');
	}

	if (defined ($system->{filterLists}))
	{
		$out->open_element('filterListsSection');

		foreach my $list ( @{ $system->{filterLists}->{filterListsSection}->{filterList} } ) {

			$out->open_element('filterList');
			print_if_defined( $list, 'defaultAction', $out );
			print_if_defined( $list, 'description',   $out );

			if ( defined $list->{appliesTo} ) {
				foreach my $application ( @{ $list->{appliesTo} } ) {
					$out->open_element('appliesTo');
					print_if_defined( $application, 'source',      $out );
					print_if_defined( $application, 'sourceGroup',      $out );
					print_if_defined( $application, 'destination', $out );
					print_if_defined( $application, 'destinationGroup',      $out );
					$out->close_element('appliesTo');
				}
			}

			print_if_defined( $list, 'filterEntry', $out );
			
			if (defined $qradarVersion && int($qradarVersion) >= 726){
				print_if_defined( $list, 'userFilterEntry', $out ); # only print userFilters from 7.2.6 onwards
			}
			print_if_defined( $list, 'mode',        $out );
			print_if_defined( $list, 'name',        $out );
			$out->close_element('filterList');
		}

		$out->close_element('filterListsSection');
	}
	# we start with the destination NAT filters
	if (defined ($system->{natFilterLists}))
	{
		_generate_nat_translation($out, $system->{natFilterLists}, 'source');
		# Remove current contents of NAT-level Zone/interfaces hash so as to accurately
		# build up for next backup request
#		%NATZoneHash = ();
	}

	$out->close_element('filterLists');

	return;
}

sub parse_nat
{
	my ($rules, $lookup_context) = @_;
	my $ztTranslationObjects = ();  
	my @zoneNATRules; # Each entry will receive a hash ent with  zone "key" of "<srcZone>|<dstZone>", constituting a separate source/dest address translation
					# An array of addr rules, and constraints therein, will be present on each entry.
	my %zoneNATRuleIHash;	  # A secondary index hash to the above
	my $ipAddrRe = get_crep('cipm'); # const
	
	# We have scoped down to the "rules" node in the NAT-specific part of the rulebase
	
	my $ruleIdx = 1; # diagnostic purposes only
	
	foreach my $natRule (@{$rules})
	{		
		my $nrName = $natRule->att('name') // 'nat-rule-' . $ruleIdx;
		++$ruleIdx;
		my $tTranslation;
		my $srcZones;
		my $dstZones;
		
		# Reject any which specify ipv6 translation 
		my $nrTransType = get_value_from_xml( 'nat-type',   $natRule ) // 'ipv4';
		if ($nrTransType =~ /nat64/i)
		{
			$LOGGER->debug("ZipTie::Adapters::PaloAlto::PANOS::Rulebase::parse_nat : bypassing presumed ipv6 translation rule $nrName");
			next;
		}
		# similarly, since the disable is at the 'entry' level, just bypass if found disabled
		if (get_value_from_xml('disabled', $natRule))
		{
			$LOGGER->debug("ZipTie::Adapters::PaloAlto::PANOS::Rulebase::parse_nat : bypassing disabled translation rule $nrName");
			next;
		}
				
		# Determine source, dest zone expansion.  Each one is in a <member> encoding.  See if we need to allocate an Address Translation object here
		# for the bucketing.  Either way, the zone references here can be list based throughout, so assume that worse case and iterate yer tail
		# off.
		for (@{$natRule->findnodes('from/member')})
		{
			push (@{$srcZones}, get_value_from_xml(q(.), $_));
		}
		for (@{$natRule->findnodes('to/member')})
		{
			push (@{$dstZones}, get_value_from_xml(q(.), $_));
		}

		# Default the "all" here.
		# +++wayne it might be cleaner in the SED to just omit the appliesTo source/dest altogether if 'all' is specified or implied, rather than needing
		# to dump the entire interface list here.  Might also make things a bit tougher diagnostically, though.
		for ($srcZones, $dstZones)
		{
			if (not defined($_))
			{
				push (@{$_}, 'all');
			}
		}

		# Now, we are looking for a new rule to add here.  Build it up so that we can then distribute it across the <src_zones | dst_zones > 
		# cross product.

		foreach my $sZone (@{$srcZones})
		{
			# we localize this in case we need to "decorate" it to take on source/dest zones.
			my $localSZone = $sZone;

			for (@{$natRule->findnodes('from-interface')})  # NB!  There's only one of these if any at all.
			{
				my $fIfVal = get_value_from_xml(q(.), $_);

				# if this is a non-default value, then let's "decorate" the source zone
				if ($fIfVal ne 'any')
				{
					$localSZone .= "::$fIfVal";
				}
			}

			foreach my $dZone (@{$dstZones})  
			{
				my $localDZone = $dZone;

				for (@{$natRule->findnodes('to-interface')})  # Note, only one if any (that would make a huge difference on dZone key formulation otherwise)
				{
					my $tIfVal = get_value_from_xml(q(.), $_);

	
					if ($tIfVal ne 'any')
					{
						# "decorate" the dest zone to be "qualified" by this interface
						$localDZone .= "::$tIfVal";
					}
				}

				my $zoneKey = "$localSZone|$localDZone";
				my $tXLAddr;  # Address translations as set into addressTranslation
				my $rDynAddr;  # Address translation as found in device config
				my $tRule;
				
				if (not defined $zoneNATRuleIHash{$zoneKey})
				{
					$zoneNATRuleIHash{$zoneKey} = { zone => $zoneKey};
					push (@zoneNATRules, $zoneNATRuleIHash{$zoneKey} );
				}
				$tTranslation = $zoneNATRuleIHash{$zoneKey};
	
				#+++wayne this can get combined to a single for loop for 'source', 'dest', etc.....
				
				# Look for source address translation.  3 options:
				#    static-ip
				#    dynamic-ip
				#    dynamic-ip-and-port
				#
				#  The management Web U/I only allows you to set one, but you can goose more than one
				#    of each in using the CLI (it's unclear if that's not a CLI validation bug)
				$tRule->{isStaticNATMapping} = 0; # Initialize
				
				undef $tXLAddr;
				
				for ($natRule->findnodes('source-translation'))  # there will only be one, if any
				{
					# static-ip			
					for ($_->findnodes('static-ip'))
					{
						my $rStaticAddr = get_value_from_xml('translated-address', $_);
						
						if (not defined $rStaticAddr)
						{
							$LOGGER->debug("ZipTie::Adapters::PaloAlto::PANOS::Rulebase::parse_nat : cannot find static addr mapping for zone key $zoneKey");
							next;
						}
							
						$tRule->{isStaticNATMapping} = 1;
							
						# if present and 'yes', bi-directional is true.  Otherwise, don't allow reverse connections.
						my $bdr = get_value_from_xml('bi-directional', $_) // 'no';
						$tRule->{establishedReverseConnectionTrafficOnly} = $bdr =~ /yes/i ? 0 : 1;

						# translated address here is always an IP, or subnet or IP address object maybe.  
						if ($rStaticAddr =~ /^($ipAddrRe)(\/(\d{1,2}))?$/)
						{
							$tRule->{ sourceAddrMapped } = $1;
							if ($2)
							{
								# It's more than a bit odd to be speaking of static NAT converting to a subnet, but strictly speaking, 
								# the PA config editor allows it.  I'd think this is a config error.
								# +++wayne should see what the PA tries to map this to in this case, if anything.
								$tRule->{ sourceAddrMaskMapped } = $3;
							}
						}
						else
						{
							# hope this is a valid address object
							$tRule->{sourceAddrNameMapped} = $rStaticAddr;
						}
					}
					for ($_->findnodes('dynamic-ip'))
					{
						foreach ($_->findnodes('translated-address/member', $_))
						{
							$rDynAddr = get_value_from_xml(q(.), $_);
							if (not defined $rDynAddr)
							{
								$LOGGER->debug("ZipTie::Adapters::PaloAlto::PANOS::Rulebase::parse_nat : cannot find primary dynamic addr mapping for zone key $zoneKey");
								next;
							}
	
							# translated address here is always an IP, or IP address object maybe
							if ($rDynAddr =~ /^($ipAddrRe)(\/(\d{1,2}))?$/)
							{
								$tXLAddr->{ sourceAddrMapped } = $1;
								if ($2)
								{
									$tXLAddr->{ sourceAddrMaskMapped } = $3;
								}
							}
							elsif ($rDynAddr =~ /^($ipAddrRe)\s*\-\s*($ipAddrRe)/)
							{
								$tXLAddr->{sourceAddrMapped} = $1;
								$tXLAddr->{sourceAddrEndMapped} = $2;
							}
							else
							{
								# hope this is a valid address object
								$tXLAddr->{sourceAddrNameMapped} = $rDynAddr;
							}
							# Put this in the list of address mappings to be set on rule
							if ($tXLAddr)
							{
								push (@{$tRule->{_paAddrTranslations}}, $tXLAddr);
								undef $tXLAddr;
							}
						}
						
						# fallback.
						# +++wayne The current version of the web configuration f/e does
						# not properly save this--it can only be entered via CLI.  The
						# following is consistent with the CLI spec. 
						for ($_->findnodes('fallback'))
						{
							# Two sections can be present here
							for ($_->findnodes('interface-address'))
							{
								my $fallbackInterfaceValue = get_value_from_xml('interface', $_);
								# this will be an interface name
								if ($fallbackInterfaceValue)
								{
									my $fallbackInterfaceParams = $lookup_context->get_interface($fallbackInterfaceValue);
									# Now, dig out the primary interface IP value
									given ($fallbackInterfaceParams->{'interfaceIp'}->{'ipConfiguration'}->[0])
									{
										push (@{$tRule->{_paAddrTranslations}}, {sourceAddrMapped => $_->{'ipAddress'}});										
									}
								}
								else
								{
									for $fallbackInterfaceValue (get_value_from_xml('floating-ip', $_), get_value_from_xml('ip', $_))
									{
										push (@{$tRule->{_paAddrTranslations}}, {sourceAddrMapped => $fallbackInterfaceValue});
									}
								}
							}
						}
							
					}
					for ($_->findnodes('dynamic-ip-and-port'))
					{
						# aside from the port address translation, it's the same translated-address/member structure found for dynamic-ip						
						foreach ($_->findnodes('translated-address/member', $_))
						{
							$rDynAddr = get_value_from_xml(q(.), $_);
							if (not defined $rDynAddr)
							{
								$LOGGER->debug("ZipTie::Adapters::PaloAlto::PANOS::Rulebase::parse_nat : cannot find primary dynamic addr mapping for zone key $zoneKey");
								next;
							}

							$tXLAddr->{'usePATOverload'} = 1;
	
							# translated address here is always an IP, or IP address object maybe
							if ($rDynAddr =~ /^($ipAddrRe)(\/(\d{1,2}))?$/)
							{
								$tXLAddr->{ sourceAddrMapped } = $1;
								if ($2)
								{
									$tXLAddr->{ sourceAddrMaskMapped } = $3;						
								}
							}
							elsif ($rDynAddr =~ /^($ipAddrRe)\s*\-\s*($ipAddrRe)/)
							{
								$tXLAddr->{sourceAddrMapped} = $1;
								$tXLAddr->{sourceAddrEndMapped} = $2;
							}
							else
							{
								# hope this is a valid address object
								$tXLAddr->{sourceAddrNameMapped} = $rDynAddr;								
							}							
							# Put this in the list of address mappings to be set on rule
							if ($tXLAddr)
							{
								push (@{$tRule->{_paAddrTranslations}}, $tXLAddr);											
								undef $tXLAddr;					
							}
						}		
					}					
				}
	
				# Look for dest address translation								
				for ($natRule->findnodes('destination-translation'))  # there will only be one, if any
				{
					my $destTranslationAddr;
					
					# static-ip			
					{
						my $destTranslationAddr = get_value_from_xml('translated-address', $_);

						if ($destTranslationAddr)
						{												
							# translated address here is always an IP, or IP address object maybe
							if ($destTranslationAddr =~ /^($ipAddrRe)(\/(\d{1,2}))?$/)
							{
								$tRule->{ destAddrMapped } = $1;
								if ($2)
								{
									$tRule->{ destAddrMaskMapped } = $3;								
								}
							}
							else
							{
								# hope this is a valid address object
								$tRule->{destAddrNameMapped} = $destTranslationAddr;																
							}
						}
						
						# Port can also be specified here
						$destTranslationAddr = get_value_from_xml('translated-port', $_);
						
						if ($destTranslationAddr)
						{
							$tRule->{destPortMapped} = $destTranslationAddr;
						}											
					}
				}
					
				# evaluate constraints on rule....wait, what rule?
				if (not defined $tRule)
				{
					$LOGGER->debug("ZipTie::Adapters::PaloAlto::PANOS::Rulebase::parse_nat : did not evaluate mapping for zone key $zoneKey");
					next;
				}
				
				$tRule->{name} = $nrName;

				my $sConstraintHash;
				
				# service port or group
				# PANOS has a flaw in how it exposes this, since there is a single "service" entry to a NAT rule.  I'm not sure
				# if it tests on both source and dest for the service port, although logically it *has* to test on destination
				# port.  That does post a limitation though, one that doesn't exist on other configuration regimes.  I can, for 
				# example, on IOS say, "do NAT mapping if the destination service is LDAP and the source is a dynamic port..."
				#
				# +++waynet move this to after the source/dest constraint stuff so we can test on presence at least of source
				# or dest constraint recorded.  This should resolve the "source|dest" arg for port/service constraint maping, at least.
				for ($natRule->findnodes('service'))
				{
					my $sPort = get_value_from_xml(q(.), $_);
					
					if ($sPort && $sPort ne 'any')  # any means no constraint at all, so don't hassle it
					{
						if ($sPort =~ /^\d+$/)
						{
							$sConstraintHash->{destServicePort} = $sPort;
						}
						else
						{
							# Need to make a specific check to avoid the SED that will result in issue #66420
							my $hiPort = _check_hi_range_port_bug($sPort, $lookup_context);
							if ($hiPort)
							{
								$LOGGER->debug("ZipTie::Adapters::PaloAlto::PANOS::parse_nat : high range port constraint reference to $hiPort " .
											"suppressed to avoid QRM topology port issue : this is fixed in product release 7.2MR2");
							}
							else
							{											
								# All other values including range need be defined as an object group.  In fact, the web management
								# U/I won't allow hardcoded port values, only the CLI.
								$sConstraintHash->{destServiceObjectName} = _retrieve_service_group_by_direction($sPort, 'source|destination', $lookup_context);
							}
						}
					}
				}
				
				foreach my $direction('source', 'destination'){

					# These can be multivalued....
					for ($natRule->findnodes("$direction/member"))
					{
						my $cons = get_value_from_xml(q(.), $_);

						my $short_direction = $direction eq 'destination' ? 'dest' : 'source';

						if ($cons eq 'any'){
							# Assign 'any' as an ip here
							$sConstraintHash->{$short_direction . 'Ip'} = $cons;
						}
						# prefix mask, range, object group
						elsif ($cons =~ /^($ipAddrRe)(\/(\d{1,2}))?/)
						{
							$sConstraintHash->{$short_direction . 'Ip'} = $1;
							if($2){
								$sConstraintHash->{$short_direction . 'Prefix'} = $3;
							}
						}					
						elsif ($cons =~ /^($ipAddrRe)\s*\-\s*($ipAddrRe)/)
						{
							$sConstraintHash->{$short_direction . 'Ip'} = $1;
							$sConstraintHash->{$short_direction . 'IpEnd'} = $2;
						}
						else
						{
							# presumed object group
							$sConstraintHash->{$short_direction . 'IpObjectGroup'} = $cons;
						}
					}
				}

				# see if we had constraints
				if ($sConstraintHash)
				{
					push ( @{$tRule->{_paAddrConstraints}} , $sConstraintHash);
				}										

				push (@{$tTranslation->{_paNATRules}}, $tRule); 
			}
		}
	}
	
			
	# We have traversed all of the NAT rules at this point, and can iterate through them to convert them to 
	# AddressTranslation objects
	for my $paAddrTransObj (@zoneNATRules)
	{
		my $ztTranslation = _createZiptieAddressTranslationFrom($paAddrTransObj, $lookup_context);
		
		if ($ztTranslation)
		{
			push (@{$ztTranslationObjects}, $ztTranslation);			
		}
	} 		
	return $ztTranslationObjects;
}

# The input to this routine is from the format maintained by parse_nat().  Namely
#  $paAddressTranslationObject ---->  {zone} --> ( '|' delimited source/dest zone)
#                                   {_paNATRules) --> [ rule #1, with hash keys consistent with ruleArgs param to AddressTranslation::addRule()
#                                                   |
#                                                   |
#                                                   --> {_paAddrConstraints} --> [ constraint #1, with hash keys consistent with ruleArgs param to AddressTranslation::addRuleAddressCriterion ]
#                                                                                [ constraint #2 ]
#                                                                                   ....
#                                                                                [ constraint n ]
#                                                      ]
#                                                      [ rule #2 (with its own constraints, if any)
#                                                               ....
#                                                      [ rule n ]   
   
   
sub _createZiptieAddressTranslationFrom
{
	my ($paAddrTransObj, $ifLookupContext) = @_;
	
	my ($sZone, $dZone) = split( /\|/, $paAddrTransObj->{zone});
	
	my $ztATObj = new ZipTie::Model::AddressTranslation($sZone, $dZone);
	if ($paAddrTransObj->{name})
	{
		$ztATObj->setTranslationName($paAddrTransObj->{name});
	}
	# We will set a mapping on the dynamic port range...
	$ztATObj->addPortTranslationRange( (1024,65535) );

	$ztATObj->useZones();

	for my $nPARule (@{$paAddrTransObj->{_paNATRules}})
	{
		my $ztATRule = $ztATObj->addRule($nPARule);
		
		for my $nPACons (@{$nPARule->{_paAddrConstraints}})
		{
			$nPACons->{rule} = $ztATRule;
			$ztATObj->addRuleAddressCriterion($nPACons);
		}
		# see if there are additional or multiple address mappings to set up
		for my $nPAXlat (@{$nPARule->{_paAddrTranslations}})
		{
			$nPAXlat->{rule} = $ztATRule;
			# Substitute down any rule-level destination params to the individual translation mapping about to be added
			for ('destAddrMapped', 'destAddrMaskMapped', 'destAddrNameMapped', 'destAddrPortMapped')
			{
				if ($ztATRule->{$_})
				{
					$nPAXlat->{$_} = $ztATRule->{$_};
				}
			}
			$ztATObj->addRuleAddressTranslationMapping($nPAXlat);
		}
	}
	
	return $ztATObj;
}

sub _generate_nat_translation
{
	my ($out, $rules, $direction) = @_; # direction is 'source' or 'destination'
	my $flReset = 0;
	
	# We generate the filter list section
	$out->open_element('filterListsSection');
	for my $nAddressTranslation (@{$rules})
	{
		if (!$flReset)
		{
			$nAddressTranslation->resetForNewFiltersSection();
			$flReset = 1;
		}
		$nAddressTranslation->generateTranslationList($out, $direction);
	}
	# need to set phase here...
	$out->open_element("phaseDiscriminator");
	$out->print_element('processingPhase', { 'core:packetProcessingPhase' => "packet-filtering"});
	$out->print_element('processingPhasePosition', ($direction eq 'destination') ? 'before' : 'after');
	$out->close_element("phaseDiscriminator");
	$out->close_element('filterListsSection');
	
	return;
}

# Convenience routine to allow a lookup of object group and expand it into multiple groups (as stowed
# away in the lookupContext object), and find the one indicated by direction as passed.
sub _retrieve_service_group_by_direction
{
	my ($groupName, $direction, $groupLookupContext) = @_;
	my $lookupDirectedGroup;

	my $direction_set = $groupLookupContext->get_service_group_direction_set($groupName);
	if( defined $direction_set ) {
		for (@{ $direction_set })
		{
			$lookupDirectedGroup = $_;
			if ($lookupDirectedGroup =~ /\S+$direction/)
			{
				return $lookupDirectedGroup;
			}
		}	
	}

	# if we get here, we've failed the lookup.  Flag it as a warning (sounds like it will show up as an incorrectly 
	# identified group, but seek to continue for the sake of the rest of the NAT configuration/SED.
	$LOGGER->debug("ZipTie::Rulebase::parse_nat : unable to find direction ($direction) in object group $groupName");
	return $lookupDirectedGroup // $groupName;
}

# Routine to chase down port issues which could raise a bug in 7.2MR1 core.  This should be removed for the adapter released with
# 7.2MR2 or later.  A 0 return is "good" otherwise return the value of the out of range port.  This routine will call itself 
# recursively

use constant MAX_PRE_722_PORT_VALUE => 32767;

sub _check_hi_range_port_bug
{
	my ($sPortRef, $lookupContext) = @_;
	my $portRtn = 0;
	
	# for now, if the port is numeric or a straight range, we're deeming the port "not offending"
	if ($sPortRef !~ /^[\d\-]+$/)
	{	
		given ($lookupContext->get_service_group_object($sPortRef))
		{
			my $portElem = $_;
			
			for (@{$portElem->{portExpression}})
			{
				if ($_->{port} > MAX_PRE_722_PORT_VALUE )
				{
					$portRtn = $_->{port};
					last;
				}
			}
			for (@{$portElem->{portRange}})
			{
				for ($_->{portStart}, $_->{portEnd})
				{
					if ($_ =~ /^\d+$/)
					{
						if ($_ > MAX_PRE_722_PORT_VALUE )
						{
							$portRtn = $_;
						}
					}
					else
					{
						$portRtn = _check_hi_range_port_bug($_, $lookupContext);
					}
					if ($portRtn > MAX_PRE_722_PORT_VALUE )
					{
						last;
					}
				}
			}
			
			for (@{$portElem->{objectGroupReference}})
			{
				# need to recursively dive to get value
				$portRtn = _check_hi_range_port_bug($_, $lookupContext);				
				if ($portRtn)
				{
					last;
				} 
			}

		}
	}
	return $portRtn;
}

sub _parse_profile_filter_list{
	my ($profile, $management_ip_addr) = @_;
	my $profile_list;

	if($profile->{interfaces}){

		my $name = $profile->{name};
		$profile_list->{defaultAction} = 'drop';
		$profile_list->{description} = "Interface management profile for $name";

		foreach my $interface(@{$profile->{interfaces}}){
			push(@{$profile_list->{appliesTo}}, {destination => $interface});
		}

		push(@{$profile_list->{filterEntry}}, {
			destinationIpAddr => $management_ip_addr,
			destinationService => { objectGroupReference => "Interface management profile $name service group" },
			log => 'false',
			name => "Interface management profile $name filter entry",
			primaryAction => 'none',
			processOrder => 1,
			sourceIpAddr => { objectGroupReference => "Interface management profile $name network group" }
		});

		$profile_list->{mode} = 'stateful';
		$profile_list->{name} = "Interface management profile $name acl";
	}

	return($profile_list);
}

1;


sub missing_network_object_references {
	my $lookup      = shift;
	my $parsed_rule = shift;
	my $LOGGER      = shift;
	
	my @allAddresses = (@{ $parsed_rule->{filterEntry}->{sourceIpAddr} }, @{$parsed_rule->{filterEntry}->{destinationIpAddr}});

	foreach my $address ( @allAddresses ) {
		if ( exists $address->{objectGroupReference} ) {
			unless ($lookup->get_network_group($address->{objectGroupReference})){
				$LOGGER->debug("ZipTie::Adapters::PaloAlto::PANOS::Rulebase::parse_acls - network object $address->{objectGroupReference} noy found, dropping rule $parsed_rule->{name}");
				return 1;
			}
		}
	}
	return 0;
}


sub create_intrazone_pairs {
	my $source_zones = shift;
	
	my @applies_to_pairs;
	foreach my $source_zone (@$source_zones) {    # build intrazone pairs
		my $applies_to;
		$applies_to->{sourceGroup}      = $source_zone;
		$applies_to->{destinationGroup} = $source_zone;
		push @applies_to_pairs, $applies_to;
	}
	return \@applies_to_pairs;
}

sub create_interzone_universal_pairs {
	my $source_zones      = shift;
	my $destination_zones = shift;
	my $isUniversal = shift;

	my @applies_to_pairs;
	foreach my $source_zone (@$source_zones) {
		foreach my $dest_zone (@$destination_zones) {
			if ($source_zone eq $dest_zone){
				next unless $isUniversal; # universal includes intrazone
			}
			my $applies_to;
			$applies_to->{sourceGroup}      = $source_zone;
			$applies_to->{destinationGroup} = $dest_zone;
			push @applies_to_pairs, $applies_to;
		}
	}
	return \@applies_to_pairs;
}

sub parse_default_rules {
	my $defaultRules = shift;

	my $defaultOverrides = {};
	foreach my $rule (@$defaultRules) {
		my $ruleName = $rule->att('name');
		if ($ruleName eq "intrazone-default" || $ruleName eq "interzone-default"){
			my $action = get_value_from_xml( 'action', $rule );
			$defaultOverrides->{$ruleName} = convert_rule_action($action);
		}
	}
    return $defaultOverrides;
}


sub create_default_rule {
	my $defaultOverrides = shift;
	my $defaultRuleName = shift;

	my $default_allow;
	$default_allow->{name} = $defaultRuleName;
	$default_allow->{mode} = 'stateful';
	if ($defaultRuleName eq "intrazone-default"){
		$default_allow->{defaultAction} = $defaultOverrides->{'intrazone-default'} // "permit";
	}elsif ($defaultRuleName eq "interzone-default"){
		$default_allow->{defaultAction} = $defaultOverrides->{'interzone-default'} // "deny";
	}
    return $default_allow;
}

__END__

=head1 NAME

ZipTie::Adapters::PaloAlto::PANOS::Rulebase - Top level routines to parse
elements of filtering and address translation.

=head1 VERSION

This document describes ZipTie::Adapters::PaloAlto::PANOS::Rulebase version 1.1

=head1 SYNOPSIS

    use ZipTie::Adapters::PaloAlto::PANOS::Rulebase;
    parse_acls($rules, $lookup);
    parse_nat($rules, $lookup);
    print_filter_lists($system, $out);

=head1 DESCRIPTION

These are the parsers and generators for the pieces of  the PANOS rulebase:
the security and NAT policies.

=head1 SUBROUTINES/METHODS

=over 12

=item C<parse_acls($rules, $lookup)>

This function takes a reference to a list of XML elements representing Palo Alto security policies and a LookupTable object.
It returns a list of parsed filterLists SED elements. Palo Alto security policies don't map very well to filterList SED elements.
Each policy is turned into an independant filter list with a single filter entry and a default action of 'NONE'.

This function also creates two simulated filter lists representing the default rules for trafic on the firewall. By default all
traffic between interfaces in the same security zone is allowed, while all traffic between interfaces in different security zones
is forbidden. This behavior is simulated by a rule with a default action of drop that applies to all combinations of all interfaces and
another rule right above it with a default action of permit that applies to all combinations of all interfaces in each security zone.
All other rules are above these two default rules, and so override them.

This function will create an acl for every interface management profile applied to an interface and these will be the first alcs hit.

=item C<parse_nat($rules, $lookup)>

This function takes a reference to a list of XML elements representing Palo Alto Nat rules and a LookupTable object.
It returns a list of parsed filterLists SED elements.

=item C<print_filter_lists($system, $out)>

This function takes a hash representing a parsed subelement and an XML printer. It prints out the filter lists contained in the
subelement. This function is needed because filter lists and their sub-elements can not be printed in alphbetical order. This
is due to the fact that the appliesTo element of a filter list contains source and destination elements in a sequence that is not
in alphabetical order.

=back

=head1 PRIVATE SUB-ROUTINES

=over 12

=item C<_createZiptieAddressTranslationFrom($paAddrTransObj, $ifLookupContext)>

UNKNOWN - This needs documentation



=item C<_generate_nat_translation($out, $rules, $direction)>

UNKNOWN - This needs documentation

=item C<_parse_acl_application($rule, $lookup)>

This function takes an xml element representing a Palo Alto security policy and retuns a list of appliesTo entries that describe
which interfaces the policy is applied to. It creates these entries by determining what security zones the policy applies to and
looking them up in the lookup table.

=item C<_parse_filter_ip_addresses(@addresses)>

This function takes a list of addresses, determins if they are hosts, networks, or network group references, and returns a reference
to a list of filterIpEntries filled in appropriatly.

=item C<_parse_profile_filter_list($profile, $management_ip_addr)>

This function creates a filter list from an interface management profile.
The list has one filter entry, described below, and a default action to block as interface management profiles define permitted traffic.
It uses the system management address as the destination address, the profile's previously created service group as its destination
service and the profile's previously created network group as its source address.

=back

=head1 DIAGNOSTICS

N/A

=head1 CONFIGURATION AND ENVIRONMENT

N/A

=head1 DEPENDENCIES

In order for acl application information to be parsed correctly, security zone information must have already been parsed and
stored in the lookup table.

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

N/A

=head1 LICENSE AND COPYRIGHT

    Copyright (c) 2013, IBM Corporation. All Rights Reserved.
    This module is free software. It may be used, redistributed
    and/or modified under the same terms as Perl itself.

=head1 AUTHOR

    Wayne Tackabury (wtackabury@us.ibm.com), David Tickner (dtickner@ca.ibm.com), Jak Devaney (john.devaney@uk.ibm.com)

=cut
