package ZipTie::Adapters::PaloAlto::PANOS::Telemetry;

use strict;
use warnings;

use 5.0100000;
use XML::Twig::XPath;
use English qw( -no_match_vars ); #Friendly names for punctuation variables

use ZipTie::ConnectionPath;
use ZipTie::Model::XmlPrint;
use ZipTie::Adapters::PaloAlto::PANOS::AutoLogin qw(issue_command);
use ZipTie::Adapters::PaloAlto::PANOS::TelemetryParsers qw(parse_system_telemetry parse_interfaces_telemetry parse_mac_telemetry parse_arp_telemetry parse_routes_telemetry print_telemetry);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle trim);
use ZipTie::Adapters::PaloAlto::PANOS::LookupTable;
use ZipTie::Typer;
use ZipTie::Logger;

#use Data::Dumper;

our $VERSION = 1.0;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

sub invoke {
	my $package_name = shift;
	my $telemetry_doc   = shift;

	#	$LOGGER->debug("==================\n\n$backup_doc\n\n==================\n\n");

	$LOGGER->debug('Starting telemetry operation.');

########Create Initial Connection################
    my ($connection_path) = ZipTie::Typer::translate_document( $telemetry_doc, 'connectionPath' );

#        my $test = $connection_path->to_xml_string(1);
#        $LOGGER->debug("==================\n\n$test\n\n==================\n\n");

    my $connection = ZipTie::Adapters::PaloAlto::PANOS::AutoLogin->new($connection_path);
    $connection->establish_connection();

########End Create Initial Connection################

########Gather Device Responses################

    my $responses = {};
    my $eval_success = 1;

    $eval_success = eval {
        $responses = _commands($connection);
        1; #Magic 'true' so that $eval_success is always set properly if the eval block finishes
    };
    if (!$eval_success) {
        $LOGGER->fatal("Parsing FAILED while issuing commands: $EVAL_ERROR");
    }

########End Gather Device Responses################

	#Instantiate the LookupTable object that will be used to pass information between parsers
	$responses->{'lookup_tables'} = ZipTie::Adapters::PaloAlto::PANOS::LookupTable->new();

########Parse Inputs################
	# Grab an output filehandle for the model
	$LOGGER->debug('Opening Filehandle...');
	my $filehandle = get_model_filehandle( 'PANOS', $connection_path->get_ip_address() );

	# initialize the model printer
	$LOGGER->debug('Opening Printer...');
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'telemetry' );

	$LOGGER->debug('Parsing Telemetry...');

	$LOGGER->debug('Parsing Supporting Information...');
	parse_system_telemetry( $responses, $printer );

	$LOGGER->debug('Parsing Interfaces...');
	parse_interfaces_telemetry( $responses, $printer );

	$LOGGER->debug('Parsing Mac Table...');
	parse_mac_telemetry( $responses, $printer );

	$LOGGER->debug('Parsing Arp Table...');
	parse_arp_telemetry( $responses, $printer );

	#	$LOGGER->debug("Parsing Discovery protocols...");
	#	parse_discovery_protocol_telemetry($responses, $printer);

	$LOGGER->debug('Parsing Routes...');
	parse_routes_telemetry( $responses, $printer );

	$LOGGER->debug('Successfully Parsed Telemetry Information.');

	$printer->attributes(1);
	$printer->open_discovery_event();

	print_telemetry( $responses, $printer );
	$LOGGER->debug('DiscoveryEvent Document Printed.');

	$printer->close_element('DiscoveryEvent');

	close_model_filehandle($filehandle);
	$LOGGER->debug('Filehandle closed');

	foreach my $key ( keys %{$responses} ) {
		delete $responses->{$key};
	}

########End Parse Inputs################
	$LOGGER->debug('Telemetry operation complete.');
}

sub _commands {
    my ($connection) = @_;

    my $responses;

### REQUIRED COMMANDS ###
    $responses->{'show_system_info'} = $connection->issue_command( '<show><system><info></info></system></show>', 'show system information' );

    $responses->{'show_interface_all'} = $connection->issue_command( '<show><interface>all</interface></show>', 'show interfaces' );

    $responses->{'show_routing_interface'} = $connection->issue_command( '<show><routing><interface></interface></routing></show>', 'show routing interface' );


### OPTIONAL COMMANDS ###
    my $eval_success = eval {
        $responses->{'show_interface_counters'} = $connection->issue_command( '<show><counter><interface>all</interface></counter></show>', 'show interface counters' );
        1;
    };
    if (!$eval_success) {
        $LOGGER->debug('Retrieval of interface counters failed. This is not a required command so the operation can proceed.');
        $LOGGER->debug($EVAL_ERROR);
    }

    $eval_success = eval {
       $responses->{'show_arp_all'} = $connection->issue_command( '<show><arp>all</arp></show>', 'show arp table' );
       1;
    };
    if (!$eval_success) {
        $LOGGER->debug('Retrieval of arp table failed. This is not a required command so the operation can proceed.');
        $LOGGER->debug($EVAL_ERROR);
    }

    $eval_success = eval {
       $responses->{'show_mac_all'} = $connection->issue_command( '<show><mac>all</mac></show>', 'show mac table' );
       1;
    };
    if (!$eval_success) {
        $LOGGER->debug('Retrieval of mac table failed. This is not a required command so the operation can proceed.');
        $LOGGER->debug($EVAL_ERROR);
    }

    $eval_success = eval {
       $responses->{'show_routing_route'} = $connection->issue_command( '<show><routing><route></route></routing></show>', 'show routes' );
       1;
    };
    if (!$eval_success) {
        $LOGGER->debug('Retrieval of routing table failed. This is not a required command so the operation can proceed.');
        $LOGGER->debug($EVAL_ERROR);
    }

    return $responses;
}

1;

__END__

=head1 NAME

ZipTie::Adapters::PaloAlto::PANOS::Telemetry - ZipTie adapter for Palo Alto PANOS Firewalls and the Panorama management server

=head1 VERSION

This document describes ZipTie::Adapters::PaloAlto::PANOS::Telemetry version 1.0

=head1 SYNOPSIS

    use ZipTie::Adapters::PaloAlto::PANOS::Telemetry;
    invoke ($backup_doc);

=head1 DESCRIPTION

This module is an adapter that allows the ZipTie framework to connect to Palo Alto firewalls and management servers.
It retrieves neighbor data from a specified device and transforms it into a Discovery Event Document.
It uses ZipTie::RESTAPI::get() to issue commands to the device.
This module issues the following commands to the device:

'<show><system><info></info></system></show>' - B<REQUIRED>

'<show><interface>all</interface></show>' - B<REQUIRED>

'<show><mac>all</mac></show>'

'<show><arp>all</arp></show>'

'<show><routing><route></route></routing></show>'

'<show><counter><interface>all</interface></counter></show>'


This adapter operation requires access privilages to issue the listed commands. In PAN-OS 4.0.0 and earlier SuperUser access is required.
In PAN-OS 4.1.0 either SuperUser or SuperReader can be used. In PAN-OS 5.0.0 and later all role types can be used, however limited roles
may not result in a complete view of the device. SuperReader access is the recommended level of access.


=head1 SUBROUTINES/METHODS

=over 12


=item C<invoke($backup_doc)>

This function takes a backup doccument, uses the information that it contains to connect to the specified device, and reads configuration
information from that device. It then transforms that information into Discovery Event Document format and prints it into an xml document.
This document provides information about the device's neighbors in a standard format to the caller.

=back

=head1 PRIVATE SUB-ROUTINES

N/A

=head1 DIAGNOSTICS

N/A

=head1 CONFIGURATION AND ENVIRONMENT

This module uses XML::Twig and LWP::UserAgent. These modules must be present on a system for this module to run.
In particular the version of XML::Twig must be the custom version available from IBM.

=head1 DEPENDENCIES

This module requires a version of adapters-common that includes RESTAPI.pm

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

This software is in active development. It should not be used in a production environmnent.

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2013 IBM. All rights reserved.

=head1 AUTHOR

Contributor(s): David Tickner dtickner@ca.ibm.com

Date: November 2013

=cut

