package ZipTie::Adapters::PaloAlto::PANOS::TelemetryParsers;

use strict;
use warnings;

use 5.0100000;
use Data::Dumper;
use XML::Twig::XPath;
use ZipTie::Adapters::Utils qw(strip_mac);
use ZipTie::Adapters::PaloAlto::PANOS::ParsingUtils qw(get_value_from_xml print_if_defined);
use ZipTie::Logger;
use Exporter 'import';
our @EXPORT_OK = qw(parse_system_telemetry parse_interfaces_telemetry parse_mac_telemetry parse_arp_telemetry parse_routes_telemetry print_telemetry);

our $VERSION = 1.0;

my $LOGGER = ZipTie::Logger::get_logger();

sub parse_system_telemetry {
	my ( $in, $out ) = @_;
	my $show_system_info = $in->{'show_system_info'};

	$in->{parsed}->{adminIp} = get_value_from_xml( '/response/result/system/ip-address', $show_system_info );
	$in->{parsed}->{sysName} = get_value_from_xml( '/response/result/system/hostname',   $show_system_info );

	return;
}

sub parse_interfaces_telemetry {
	my ( $in, $out ) = @_;

	my $show_interface_counters = $in->{'show_interface_counters'};
	my $show_routing_interface  = $in->{'show_routing_interface'};
	my $show_interface_all      = $in->{'show_interface_all'};

	my @entries = $show_routing_interface->findnodes('/response/result/interface');

	foreach my $entry (@entries) {

		my $vr_name = get_value_from_xml( 'vr', $entry );

		my $parsed_entry;
		my $if_name = get_value_from_xml( 'name', $entry );

		#If the router name is longer than 13 characters the virtual interface that passes traffic to
		#other virtual routers may have its name truncated. The name for this interface is always
		# the name of the router + '/i3' as far as I know.
		# so check to see if the router name got truncated and fix this.
		if ( ( ( length $if_name ) == 16 ) && ( $vr_name =~ /$if_name/ ) ) {
			$if_name = $vr_name . '/i3';
		}
		$parsed_entry->{name} = $if_name;

		#Palo Alto interfaces are named by the system according to their type
		my $type;
		if ( $parsed_entry->{name} =~ /ethernet/i ) {
			$type = 'ethernet';
		} elsif ( $parsed_entry->{name} =~ /vlan/i ) {
			$type = 'other';
		} elsif ( $parsed_entry->{name} =~ /tunnel/i ) {
			$type = 'other';
		} elsif ( $parsed_entry->{name} =~ /$vr_name/i ) {
			$type = 'other';
		} elsif ( $parsed_entry->{name} =~ /loopback/i ) {
			$type = 'softwareLoopback';
		} else {
			$type = 'unknown';
		}
		$parsed_entry->{type} = $type;

		#Status is represented as a '*' character or a blank in the output of 'show routing interfaces'
		my $status = get_value_from_xml( 'status', $entry );
		if ( $status =~ /\*/ ) {
			$parsed_entry->{operStatus} = 'Up';
		} else {
			$parsed_entry->{operStatus} = 'Down';
		}

		if ($show_interface_counters)
		{
			$parsed_entry->{inputBytes} = get_value_from_xml( '/response/result/ifnet/entry[name="' . $parsed_entry->{name} . '"]/ibytes', $show_interface_counters );
		}
		if ( ( !defined $parsed_entry->{inputBytes} ) || ( !( $parsed_entry->{inputBytes} =~ /\d+/ ) ) ) {
			$parsed_entry->{inputBytes} = '0';
		}

		my @addresses = $entry->findnodes('address');
		foreach my $address_entry (@addresses) {
			my $cidr = get_value_from_xml( '.', $address_entry );
			if ( ( defined $cidr ) and ( $cidr =~ m{(\d+\.\d+\.\d+\.\d+)(?:/)(\d+)} ) ) {
				my ( $address, $mask ) = $cidr =~ m{(\d+\.\d+\.\d+\.\d+)(?:/)(\d+)};
				my $ip_entry = { ipAddress => $address, mask => $mask };
					push @{ $parsed_entry->{ipEntry} }, $ip_entry;
			}
		}    # end foreach

		$in->{'lookup_tables'}->add_interface( $parsed_entry->{name}, $parsed_entry );
		push @{ $in->{parsed}->{interfaces}->{interface} }, $parsed_entry;
	}

	#get all the network interfaces
	@entries = $show_interface_all->findnodes('/response/result/ifnet/entry');

	foreach my $entry (@entries) {

		my $parsed_entry;

		$parsed_entry->{name} = get_value_from_xml( 'name', $entry );

		if ( !defined $in->{'lookup_tables'}->get_interface( $parsed_entry->{name} ) ) {

			#Palo Alto interfaces are named by the system according to their type
			my $type;
			if ( $parsed_entry->{name} =~ /ethernet/i ) {
				$type = 'ethernet';
			} elsif ( $parsed_entry->{name} =~ /vlan/i ) {
				$type = 'other';
			} elsif ( $parsed_entry->{name} =~ /tunnel/i ) {
				$type = 'other';
			} elsif ( $parsed_entry->{name} =~ /loopback/i ) {
				$type = 'softwareLoopback';
			} else {
				$type = 'unknown';
			}
			$parsed_entry->{type} = $type;

			#sub interfaces have a name that is the name of the master interface plus a dot and a number
			my ($basename) = $parsed_entry->{name} =~ m{([^\.]*)(?:\.\d+)?};

			#use the base name to extract the state of the interface. Assume that if the master is down then all the subs are down
			$parsed_entry->{operStatus} = ucfirst( get_value_from_xml( "/response/result/hw/entry[name=\"$basename\"]/state", $show_interface_all ) );

			#find this interface in the output of show counters and extract the ibytes entry...if the counter exists (optional command elicited it)
			$parsed_entry->{inputBytes} = ($show_interface_counters) ? 
				get_value_from_xml( '/response/result/ifnet/entry[name="' . $parsed_entry->{name} . '"]/ibytes', $show_interface_counters ) 
				: 0;

			#extract the first ip address information if it exists
			my $cidr = get_value_from_xml( 'ip', $entry );

			if ( ( defined $cidr ) and ( $cidr =~ m{(\d+\.\d+\.\d+\.\d+)(?:/)(\d+)} ) ) {

				my ( $address, $mask ) = $cidr =~ m{(\d+\.\d+\.\d+\.\d+)(?:/)(\d+)};
				my $ip_entry = { ipAddress => $address, mask => $mask };

				push @{ $parsed_entry->{ipEntry} }, $ip_entry;

				#interfaces can also have more ip addresses stored in a different place in the XML. Grab them as well
				my @alt_addresses = $entry->findnodes('addr/member');

				foreach my $alt_address (@alt_addresses) {

					$cidr = get_value_from_xml( '.', $alt_address );

					if ( ( defined $cidr ) and ( $cidr =~ m{(\d+\.\d+\.\d+\.\d+)(?:/)(\d+)} ) ) {
						( $address, $mask ) = $cidr =~ m{(\d+\.\d+\.\d+\.\d+)(?:/)(\d+)};
						$ip_entry = { ipAddress => $address, mask => $mask };
						push @{ $parsed_entry->{ipEntry} }, $ip_entry;
					}
				}    # end foreach

			}

			$in->{'lookup_tables'}->add_interface( $parsed_entry->{name}, $parsed_entry );
			push @{ $in->{parsed}->{interfaces}->{interface} }, $parsed_entry;
		}
	}

	return;
}

sub parse_mac_telemetry {
	my ( $in, $out ) = @_;

	my $show_mac_all = $in->{show_mac_all} || return;  # mac entries were optional as a command

	#grab the MAC table entries
	my @entries = $show_mac_all->findnodes('/response/result/entries/entry');

	foreach my $entry (@entries) {

		my $parsed_entry;

		#extract the parts we need
		$parsed_entry->{macAddress} = strip_mac( get_value_from_xml( 'mac', $entry ) );
		$parsed_entry->{interface} = get_value_from_xml( 'interface', $entry );
		$parsed_entry->{vlan}      = get_value_from_xml( 'vlan',      $entry );

		push @{ $in->{parsed}->{neighbors}->{macTable}->{macEntry} }, $parsed_entry;
	}
	return;
}

sub parse_arp_telemetry {
	my ( $in, $out ) = @_;

	my $show_arp_all = $in->{show_arp_all} || return;  # ARP retrieval was an optional command

	# Grab the ARP table entries
	my @entries = $show_arp_all->findnodes('/response/result/entries/entry');
	foreach my $entry (@entries) {

		my $parsed_entry;

		#extract the parts we need
		$parsed_entry->{macAddress} = strip_mac( get_value_from_xml( 'mac', $entry ) );
		$parsed_entry->{interface} = get_value_from_xml( 'interface', $entry );
		$parsed_entry->{ipAddress} = get_value_from_xml( 'ip',        $entry );

		push @{ $in->{parsed}->{neighbors}->{arpTable}->{arpEntry} }, $parsed_entry;
	}
	return;
}

sub parse_routes_telemetry {
	my ($in, $out) = @_;

	my $show_routing_route = $in->{show_routing_route} || return;

	my $entries = $show_routing_route->findnodes('/response/result/entry') || return;
	while($entries->size() > 0){
		my $entry = $entries->shift();
		my $parsed_entry;

		# Routing protocol is represented by letter flags in the entry
		my $flags = get_value_from_xml('flags', $entry);

		# Default to unknown
		my $protocol = 'UNKNOWN';
		if ($flags =~ /S/){
			####Disabled due to inability of 71 mr2 to handle 'STATIC'
			# $protocol = 'STATIC';
		}elsif($flags =~ /R/){
			$protocol = 'RIP';
		}elsif($flags =~ /O/){
			$protocol = 'OSPF';
		}elsif($flags =~ /B/){
			$protocol = 'BGP';
		}

		$parsed_entry->{protocol} = $protocol;

		# Get the destination IP address that this route applies to
		my $destination = get_value_from_xml('destination', $entry);
		( $parsed_entry->{ipAddress} ) = $destination =~ m{(\d+\.\d+\.\d+\.\d+)(?:/d+)?};

		# In 71_mr2 router ID is optional in the schema, but Ziptie Server breaks with a null pointer exception
		# if it is not included, setting the routerID to be the same as the destination does not make any sense
		# but it is the only value I have. Hopefully it is not actually used in ZT server or QRM core.
		$parsed_entry->{routerId} = $parsed_entry->{ipAddress};

		# Get the interface that this route exits through
		$parsed_entry->{interface} = get_value_from_xml('interface', $entry);

		# Make sure we don't have an empty attribute if there is no interface
		if(!defined($parsed_entry->{interface})){
			delete $parsed_entry->{interface};
		}elsif(!defined($in->{'lookup_tables'}->get_interface($parsed_entry->{interface}))){
			$LOGGER->debug(
				'Routing entry for '.$parsed_entry->{ipAddress}.' has an unknown interface entry. Removing interface '.$parsed_entry->{interface}.' from the route entry.'
			);
			delete $parsed_entry->{interface};
		}

		push @{ $in->{parsed}->{neighbors}->{routingNeighbors}->{routingNeighbor} }, $parsed_entry;
	}
	return;
}

sub print_telemetry {
	my ( $in, $out ) = @_;

	print_if_defined( $in->{parsed}, 'adminIp', $out );
	print_if_defined( $in->{parsed}, 'sysName', $out );

	$out->attributes(1);
	print_if_defined( $in->{parsed}, 'neighbors', $out );

	print_if_defined( $in->{parsed}, 'interfaces', $out );

	return;
}

1;

__END__

=head1 NAME

ZipTie::Adapters::PaloAlto::PANOS::TelemetryParsers - Parsers to transform Palo Alto neighbor data in to Discovery Event Document format.

=head1 VERSION

This document describes ZipTie::Adapters::PaloAlto::PANOS::TelemetryParsers version 1.0

=head1 SYNOPSIS

    use ZipTie::Adapters::PaloAlto::PANOS::TelemetryParsers;
    parse_arp_telemetry ($in, $out);
    parse_interfaces_telemetry ($in, $out);    
    parse_mac_telemetry ($in, $out);
    parse_routes_telemetry ($in, $out);
    parse_system_telemetry ($in, $out);
    print_telemetry ($in, $out);


=head1 DESCRIPTION

 This module contains various functions to allow the parsing of Palo Alto neighbor information into Discovery Event Document format.
 For each relevent section of the Document there is a parser function. All printing is done via print_telementry as the Discovery Event
 format is properly ordered.

 Each parser and printer function takes two parameters:
 1) A hash of XML Twig objects known as $in. This contains the raw responses to various commands that were issued to the device.
 2) An XMLPrint object known as $out. This is the printer that will produce the Discovery Event Document.

 Some parsed information is stored in the $in hash in the $in->{parsed} element. This allows easy transfer of information from one parsing
 function to the next. Other information is stored in a LookupTable object stored in $in-{lookup_tables}


=head1 SUBROUTINES/METHODS

=over 12


=item C<parse_arp_telemetry($in, $out)>

This function parses arp table information. It uses information from the following command:

    show arp all
    

=item C<parse_interfaces_telemetry($in, $out)>

This function parses interface information. It uses information from the following commands:

    show interfaces all
    show counter interface all
    

=item C<parse_mac_telemetry($in, $out)>

This function parses mac table information. It uses information from the following commands:

    show mac all


=item C<parse_routes_telemetry($in, $out)>

This function parses routing table information. It uses information from the following commands:

    show routing route


=item C<parse_system_telemetry($in, $out)>

This function parses basic system information. It uses information from the following commands:

    show system info
    

=item C<print_telemetry($in, $out)>

This function prints the entire Discovery Event Document.


=back


=head1 PRIVATE SUB-ROUTINES

N/A

=head1 DIAGNOSTICS

Functions in this module are tested by unit tests in the ../../../../tests directory in the file 02_Telemetry.t


=head1 CONFIGURATION AND ENVIRONMENT

N/A

=head1 DEPENDENCIES

This module requires a specific version B<(which one?)> of ziptie-common and ziptie-server.

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

This software is in active development. It should not be used in a production environmnent.

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2013 IBM. All rights reserved.

=head1 AUTHOR

Contributor(s): David Tickner dtickner@ca.ibm.com

Date: November 2013

=cut
