package ZipTie::Adapters::PaloAlto::PANOS::TrimServiceGroups;

# This function detects service groups that are empty or contain only references to other empty
# service groups and removes them. This needs to be done since QRM core can not handle empty object groups.
#
# Since service groups on Palo Alto devices can be configured with cycles we need a solution that fully removes
# all redundant groups without running into problems with cycles.
#
# This solution constructs a (virtual) directed graph with nodes representing service groups and edges representing
# membership relationships between groups. A subset of the service groups must be known to be valid. That is to say
# we must know in advance that they are not empty. In the case of Palo Alto this subset includes all defined 'services'
# since when they are originally parsed a service is only created if there is port information for it. All other groups
# (what Palo Alto calls 'service groups' as opposed to 'services') are considered to be invalid at the start of the algorithm.
#
# The algorithm loops through the set of edges looking for any that end at a valid node. This condition indicates that
# the group at the start of the edge contains at least one reference to a group that contains valid service information.
# Therefore that group is also non-empty and the node at the start of the edge is marked as valid. This edge connecting
# two valid nodes is then discarded as future iterations do not need to consider it.
#
# The algorithm continues to loop through the set of remaining edges until it passes through the entire set without finding any
# edges that end in valid nodes. This indicates either that the list of edges is empty or that all remaining edges link only to
# empty service groups.
#
# When this portion of the algorithm terminates any edges still in the edge list represent a membership relationship to an
# empty service group. These membership relationships are removed.
# Any nodes that remain marked invalid represent empty service groups. These service groups are removed.
#
# The result of this algorithm is a set of service groups. All objectGroupReferences in each service group
# reference service groups that exist. All service groups that exists contain either a portExpression/portRange or
# an objectGroupReference that eventually leads to an service group containing a portExpression/portRange.

use strict;
use warnings;

use 5.0100000;

use Exporter 'import';
our @EXPORT_OK = qw(trim_service_groups);

our $VERSION = 1.0;

# this hash is keyed by service group id. The values are boolean. True indicates that the service group identified
# by the key is non-empty. False indicates unknown. At the end of the algorithm false indicates either an empty
# group or a group containing references (recursive) only to other empty groups.
my $nodes;

# This hash is keyed by a combination of service group id and objectGroupReference name. The values are hashes
# containing 'from' and 'to' keys. The values of these sub hashes map to nodes in the $node hash.
# These directed edges represent membership relationships between service groups. The group referenced in 'from'
# contains an objectGroupReference to the group referenced in 'to'.
my $edges;

sub trim_service_groups {

	# This list should contain all service groups that have either portExpression or portRange information.
	# They may also contain objectGroupReferences.
	my $services_in = shift;    #services come in as a list of hashes

	# This list should contain all servivce groups that have neither portExpressions nor portRanges.
	my $service_groups_in = shift;    #service groups come in as a list of hashes.
	
	my $shared_objects = shift;  # +++waynet these are servicees defined at an "admin level"/shared scope

######## Set up the hashs and the graph ########

	# Clear the node/edge hashes since they can be called between virtual contexts.
	%{$nodes} = ();
	%{$edges} = ();
	
	my $service_groups;
	
	# +++waynet these are hashes to be able to index by shared service.  This could all theoretically be improved
	# by having these created once by the calling context, but I don't want to pollute the PAN parsing code since
	# I'm unclear that any of this is actually necessary.  Also, there shouldn't be a *huge* number (say < 1000)
	# shared service elements in the most ambitious multicontext config.
	my $indexed_shared_services = ();
	
	if ($shared_objects)
	{
		for my $sharedService (@{$shared_objects->{objectGroups}->{serviceGroup}})
		{
			# I'm not going to go all recursive and figure out if every reference here points to am 
			# empty service.  Maybe call this recursively to create a secondary "trim_service_groups" on
			# the shared defs. if that's a thing.  And yes, I'm using the same namespace for services and
			# service groups at the shared level...  
			$indexed_shared_services->{$sharedService->{id}} = $sharedService;
		}
	}

	# Restructure the $service_groups_in hash so that we can easily delete unwanted items later.
	# Also set up the graph components
	foreach my $service_group ( @{$service_groups_in} ) {

		my @list = @{ $service_group->{objectGroupReference} };
		delete $service_group->{objectGroupReference};

		foreach my $member (@list) {

			# Change the objectGroupReference list into a hash so that we can delete from it by index later
			$service_group->{objectGroupReference}->{$member} = 1;

			# Add edges to the graph. Edges are directed from this service group to the members of the group
			$edges->{ $service_group->{'id'} . $member } = { 'from' => $service_group->{'id'}, 'to' => $member };
		}

		# service groups in the $service_groups_in list have no portExpression/portRange information therefore
		# we don't know at this point if they contain any references to valid groups. Set the valid bit to false.
		$nodes->{ $service_group->{'id'} } = 0;

		$service_groups->{ $service_group->{'id'} } = $service_group;
	}

	#all services are by definition valid nodes since we only created a service if it had ports associated with it
	foreach my $service ( @{$services_in} ) {

		# service groups in the $services_in list have portExpression/portRange information therefore
		# we know that they are themselves valid groups. Set the valid bit to true
		$nodes->{ $service->{'id'} } = 1;

		# on some devices these services may also contain objectGroupReferences. (+++waynet wtf does that mean, "some devices" ?!) 
		# If they do we need to convert the list to a hash
		# and creat edges to represent the membership relationships.
		if ( defined $service->{objectGroupReference} ) {
			my @list = @{ $service->{objectGroupReference} };
			delete $service->{objectGroupReference};
			foreach my $member (@list) {
				$service->{objectGroupReference}->{$member} = 1;
				$edges->{ $service->{'id'} . $member } = { 'from' => $service->{'id'}, 'to' => $member };
			}
		}

		$service_groups->{ $service->{'id'} } = $service;
	}

######## Find which nodes and edges are invalid ########

	#we want to do at least one pass
	my $changes = 1;

	#if changes were made last time we want to do another pass
	while ($changes) {

		#no changes have been made yet.
		$changes = 0;

		#go through all of the edges still in the list of edges.
		foreach my $edge ( keys %{$edges} ) {

			#if this edge points at a valid node
			if ( $nodes->{ $edges->{$edge}->{'to'} } ) {

				#Then the node this edge points from is also valid
				$nodes->{ $edges->{$edge}->{'from'} } = 1;

				#This edge is valid so we don't have to check it again and we don't want to remove it from the service group at the end
				delete $edges->{$edge};

				#changes have occured to the node list and the edge list so we need to loop again.
				$changes = 1;

			}    # end if
			# see if this key is in the shared service indexed names,ce we set up
			elsif ($indexed_shared_services->{$edges->{$edge}->{from}} or $indexed_shared_services->{$edges->{$edge}->{to}}) 
			{
				#The node this edge points from is  valid
				$nodes->{ $edges->{$edge}->{'from'} } = 1;
				# This edge is valid so we don't have to check it again and we don't want to remove it from the service group at the end
				delete $edges->{$edge}; # keep it from being checked for nuking later

				#changes have occured to the node list and the edge list so we need to loop again. 
				$changes = 1;								
			} # end elsif
		}    # end foreach

	}    #end while

######## Delete the invalid nodes and edges from the hash ########

	#any edges that are left represent references to empty service groups. Delete that member from the service group.
	foreach my $edge ( values %{$edges} ) {
		delete $service_groups->{ $edge->{'from'} }->{objectGroupReference}->{ $edge->{'to'} };
	}

	#any node that is still marked as invalid represents a service group with no valid members. Delete that service group.
	foreach my $node ( keys %{$nodes} ) {
		if ( !( $nodes->{$node} ) ) {
			delete $service_groups->{$node};
		}
	}

######## Convert the hash back to a printable format  ########

	foreach my $g ( values %{$service_groups} ) {
		my @list = keys %{ $g->{objectGroupReference} };
		$g->{objectGroupReference} = \@list;
	}
	my @service_groups_out = values %{$service_groups};

	return \@service_groups_out;

}

1;

__END__

=head1 NAME

ZipTie::Adapters::PaloAlto::PANOS::TrimServiceGroups - Utility function to ensure service groups are not empty and contain only valid references.

=head1 VERSION

This document describes ZipTie::Adapters::PaloAlto::PANOS::TrimServiceGroups version 1.0

=head1 SYNOPSIS

    use ZipTie::Adapters::PaloAlto::PANOS::TrimServiceGroups;
    trim_service_groups ($services_in, $service_groups_in);

=head1 DESCRIPTION

Service groups on Palo Alto devices have both source and destination components, thus they need to be broken in half
to be properly represented in the SED. This can result in service groups that are empty. Empty service groups can be
problematic so they should not be included in the SED. Furthermore Palo Alto service groups can contain cycles,
that is to say group A can reference group B while group B references group A.

This module provides a general solution to this problem that takes a set of hashes suitable for xml printing
and returns a trimed set with problematic serviceGroups and objectGroupReferences removed.


=head1 SUBROUTINES/METHODS

=over 12


=item C<trim_service_groups($services_in, $service_groups_in)>

This function detects service groups that are empty or contain only references to other empty
service groups and removes them. This needs to be done since QRM core can not handle empty object groups.

Since service groups on Palo Alto devices can be configured with cycles we need a solution that fully removes
all redundant groups without running into problems with cycles.

This solution constructs a (virtual) directed graph with nodes representing service groups and edges representing
membership relationships between groups. A subset of the service groups must be known to be valid. That is to say
we must know in advance that they are not empty. In the case of Palo Alto this subset includes all defined 'services'
since when they are originally parsed a service is only created if there is port information for it. All other groups
(what Palo Alto calls 'service groups' as opposed to 'services') are considered to be invalid at the start of the algorithm.

The algorithm loops through the set of edges looking for any that end at a valid node. This condition indicates that
the group at the start of the edge contains at least one reference to a group that contains valid service information.
Therefore that group is also non-empty and the node at the start of the edge is marked as valid. This edge connecting
two valid nodes is then discarded as future iterations do not need to consider it.

The algorithm continues to loop through the set of remaining edges until it passes through the entire set without finding any
edges that end in valid nodes. This indicates either that the list of edges is empty or that all remaining edges link only to
empty service groups.

When this portion of the algorithm terminates any edges still in the edge list represent a membership relationship to an
empty service group. These membership relationships are removed.
Any nodes that remain marked invalid represent empty service groups. These service groups are removed.

The result of this algorithm is a set of service groups. All objectGroupReferences in each service group
reference service groups that exist. All service groups that exists contain either a portExpression/portRange or
an objectGroupReference that eventually leads to an service group containing a portExpression/portRange.


=back

=head1 DIAGNOSTICS

N/A

=head1 CONFIGURATION AND ENVIRONMENT

N/A

=head1 DEPENDENCIES

This module has no dependencies.

=head1 INCOMPATIBILITIES

N/A

=head1 BUGS AND LIMITATIONS

This software is in active development. It should not be used in a production environmnent.

=head1 LICENSE AND COPYRIGHT

Copyright (c) 2013 IBM. All rights reserved.

=head1 AUTHOR

Contributor(s): David Tickner dtickner@ca.ibm.com

Date: October 2013

=cut

