# Auth: John "Jak" Devaney
# Purp: Base code for the backup operation of Sourcefire 3D adapter
# Date: 24.02.14

package ZipTie::Adapters::Sourcefire::3D;

use strict;
use warnings;

use ZipTie::Adapters::BaseAdapter;
use ZipTie::Adapters::GenericAdapter;
use ZipTie::Adapters::Sourcefire::3D::AutoLogin qw{login};
use ZipTie::Adapters::Sourcefire::3D::Parsers qw{parse_system parse_chassis parse_interfaces get_vuln_ids parse_object_groups parse_filters print_interfaces parse_local_accounts parse_static_routes};
use ZipTie::Adapters::Sourcefire::3D::Disconnect qw{disconnect};
use ZipTie::Adapters::Utils qw{get_model_filehandle close_model_filehandle};
use ZipTie::Model::XmlPrint;
use ZipTie::CLIProtocolFactory;
use ZipTie::ConnectionPath;
use ZipTie::Logger;
use ZipTie::Snort;
use Data::Dumper;
use Exporter qw{import};

our @ISA = qw{ZipTie::Adapters::BaseAdapter};
my $LOGGER = ZipTie::Logger::get_logger();

sub backup{
	my ($package_name, $backup_doc) = @_;
	$LOGGER->debug(q{Getting connection, protocol and prompt});

	# Parse the backup operation XML document and extract a ZipTie::ConnectionPath object from it
	my ($connection_path) = ZipTie::Typer::translate_document($backup_doc, q{connectionPath});

	# Connect to the device and capture the ZipTie::CLIProtocol that is created as a result of the connection.
	# Also be sure to capture the device prompt that is returned as a result of a successful connection.
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);	
	my $device_prompt_regex = login($cli_protocol, $connection_path);

	# Store the found prompt as "prompt" on the specified CLI protocol.
	$cli_protocol->set_prompt_by_name(q{prompt}, $device_prompt_regex);

	# Get the model set up
	$LOGGER->debug(q{Getting model});
	my $filehandle = get_model_filehandle(q{3D}, $connection_path->get_ip_address());
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, q{common});
	$printer->open_model();

	# Get the command responses and feed into the parser
	my $responses = _get_responses($cli_protocol, $device_prompt_regex, $connection_path);
	parse_system($responses, $printer);
	parse_chassis($responses, $printer);
	parse_interfaces($responses, $printer);
	get_vuln_ids($responses, $printer);
	parse_object_groups($responses, $printer);
	parse_filters($responses, $printer);
	print_interfaces($responses, $printer);
	parse_local_accounts($responses, $printer);
	parse_static_routes($responses, $printer);

	# Tidy up. Job done.
	$printer->close_model();
	close_model_filehandle($filehandle);
	disconnect($cli_protocol);
}

# Returns a reference to responses from the device
sub _get_responses{
	my ($cli_protocol, $device_prompt_regex, $connection_path) = @_;
	
	# Hash keys -> commands for the Sourcefire shell
	my @sourcefire_shell = (
		$cli_protocol->{ors},
		q{version->show version},
		q{memory->show memory},
		q{network->show network},
		q{interfaces->show interfaces}
	);

	# Hash keys -> commands for standard bash shell	
	my @bash_shell = (
		q{disk->df},
		q{hostname->hostname},
		q{ip_addr->ip addr},
		q{route_n->route -n},
		q{users->cat /etc/passwd}
	);

	# Send the commands and fill the responses
	my $responses;
	$LOGGER->debug(q{Gathering responses});

	foreach my $key_command(@sourcefire_shell){
		my ($key, $command) = split(/-\>/, $key_command);
		$responses->{$key} = $cli_protocol->send_and_wait_for($command, $device_prompt_regex);

		# Magic			^[[7m--More--^[[27m
		$responses->{$key} =~ s/\e\[\d+M-+MORE-+\e\[\d+M//gi;
	}

	# Change to expert shell
	my $successful = q{};
	my $removed_first_prompt = q{(}.substr($device_prompt_regex, index($device_prompt_regex, q{|}) + 1);
	while($successful !~ /$removed_first_prompt/){
		$successful = $cli_protocol->send_and_wait_for(q{expert}, $device_prompt_regex);
	}

	# Switch user to root
	# TODO: Remove this and find privleges needed, update User Docs
	my $removed_second_prompt = q{(}.substr($removed_first_prompt, index($removed_first_prompt, q{|}) + 1);
	my $password_prompt_regex = q{([Pp]assword\s*\:)|}.$removed_second_prompt;

	$successful = $cli_protocol->send_and_wait_for(q{sudo su -}, $password_prompt_regex);
	if($successful !~ /$removed_second_prompt/){
		my $password = $connection_path->get_credentials()->{password};

		$LOGGER->debug(q{Sending root password quietly});
		$LOGGER->set_logging_level(1);
		$cli_protocol->send_and_wait_for($password, $removed_second_prompt);
		$LOGGER->set_logging_level(0);
	}

	# Send expert shell commands and collect their responses
	foreach my $key_command(@bash_shell){
		my ($key, $command) = split(/-\>/, $key_command);
		$responses->{$key} = $cli_protocol->send_and_wait_for($command, $device_prompt_regex);
	}

	_get_snort_responses($cli_protocol, $device_prompt_regex, $responses);
	_get_db_responses($cli_protocol, $device_prompt_regex, $responses);
	
	# get QRadar version - (to determine if we model userFilters)
	$responses->{'qradarVersion'} = get_qradar_version();

	return($responses);
}

sub _get_snort_responses{
	my ($cli_protocol, $device_prompt_regex, $responses) = @_;

	# Get ROOT from /etc/sf/ims.conf this will point to where the detection engines are kept
	my $snort_root = q{/usr/local/sf};
	my $snort_ims_conf = $cli_protocol->send_and_wait_for(q{cat /etc/sf/ims.conf}, $device_prompt_regex);
	if($snort_ims_conf =~ /^ROOT="(\S+)"\s*$/mi){
		$snort_root = $1;
		while($snort_root =~ /\$\{(\S+)\}/g){
			my $variable = $1;
			my $value = q{};
			if($snort_ims_conf =~ /^$variable="(\S+)"\s*$/mi){
				$value = $1;
			}
			$snort_root =~ s/\$\{$variable\}/$value/g;
		}
	}

	# Get the detection engine uuid
	my $snort_de_path;
	my $snort_affinity_conf = $cli_protocol->send_and_wait_for(qq{cat $snort_root/fwcfg/affinity.conf}, $device_prompt_regex);
	if($snort_affinity_conf =~ /DES\s*{\s*(\S+)\s*{\s*INSTANCES/ig){
		$snort_de_path = qq{$snort_root/detection_engines/$1};
	}else{
		$LOGGER->debug(q{Could not find detection engine path for master snort});
		return;
	}
	$responses->{policy_text} = $cli_protocol->send_and_wait_for(qq{cat $snort_de_path/policyText_full.yaml}, $device_prompt_regex);

	# Create callback for getting files
	my $get_config_callback = 
		sub{
			my ($file) = @_;

			if($file !~ /^\//){
				$file = qq{$snort_de_path/$file};
			}

			my $response = $cli_protocol->send_and_wait_for(qq{cat $file}, $device_prompt_regex);
			return($response);
		};

	my @snort_confs = (qq{$snort_de_path/snort.conf});
	while(scalar(@snort_confs)){
		my $snort_conf_path = shift(@snort_confs);

		# Create SNORT object
		my $snort_name = q{};
		my $snort_head = $cli_protocol->send_and_wait_for(qq{head $snort_conf_path}, $device_prompt_regex);
		if($snort_head =~ /^#\s*(AC)?\s*NAME\s*\:\s*(.*?)\s*$/mi){
			$snort_name = $2;
			$snort_name =~ s/\s+/_/g;
		}
		my $snort = ZipTie::Snort->new($snort_name, $get_config_callback);

		# Parse snort.conf and any files it refrences
		$snort->parse_config($snort_conf_path);
		
		# Get SNORT uuid and put the SNORT object in the responses, support multiple snorts
		my $snort_uuid = q{MASTER};
		if($snort_conf_path =~ /SNORT\.CONF\.([0-9A-F-]+)/i){
			$snort_uuid = $1;
		}
		$responses->{snorts}->{$snort_uuid} = $snort;
		
		push(@snort_confs, @{$snort->get_bound_snorts()});
	}

	# Get securtiy intelligence lists and feeds
	my $si_files = $cli_protocol->send_and_wait_for(qq{find $snort_root/iprep_download -name "*lf"}, $device_prompt_regex);
	while($si_files =~ /^($snort_root\/iprep_download\/([0-9a-f-]+)\.[wb]lf)\s*$/img){
		my ($file, $uuid) = ($1, $2);
		$responses->{lists}->{$uuid} = $cli_protocol->send_and_wait_for(qq{cat $file}, $device_prompt_regex);
	}
}

sub _get_db_responses{
	my ($cli_protocol, $device_prompt_regex, $responses) = @_;

	# Get database credentials
	my $snort_ims_data = $cli_protocol->send_and_wait_for(q{cat /etc/sf/ims-data.conf}, $device_prompt_regex);
	while($snort_ims_data =~ /^\s*SF_MYSQL_(\S+)\s+(\S+)\;\s*$/img){
		my $key = lc($1);
		$responses->{snort_db}->{$key} = $2;
	}

	# Set up credentials with defaults if none where found
	my $snort_db_name = $responses->{snort_db}->{database} // q{sfsnort};
	my $snort_db_username = $responses->{snort_db}->{name} // q{barnyard};
	my $snort_db_password = $responses->{snort_db}->{pw} // q{barnyard};
	my $snort_db_command = qq{mysql -u$snort_db_username -p$snort_db_password $snort_db_name};
	my $snort_db_prompt = q{\S+\>\s*$};

	# Quietly send the command to access the database
	$LOGGER->debug(q{Entering database quietly});
	$LOGGER->set_logging_level(1);
	my $successful = $cli_protocol->send_and_wait_for($snort_db_command, $device_prompt_regex);
	$LOGGER->set_logging_level(0);

	# If the prompt is not the expected database prompt fail
	# WARNING: This has misfired once
	if($successful !~ /$snort_db_prompt/){
		$LOGGER->fatal(qq{Access denied from mysql db: $snort_db_name});
	}

	# Ignore apps with an id greater than 2000000000. These are "client applications", we are only interested in their "server application" counterpart.
	my $client_start = 2000000000;
	my $custom_start = 0;#1000000; 
	my @snort_db_queries = (
		qq{db_app_id->select appId, appName, appDescription, risk_index, productivity_index, (isServer + (isClient * 2) + (isWebApp * 4)) as typeMask from appIdInfo where appId>=$custom_start and appId<$client_start and deleted=0;},
		qq{db_tag->select appId, tagName from application_tag_view where appId>=$custom_start and appId<$client_start;},
		qq{db_category->select appId, categoryName from application_category_view where appId>=$custom_start and appId<$client_start;}
	);

	# Send queries and collect responses
	foreach my $key_query(@snort_db_queries){
		my ($key, $command) = split(/-\>/, $key_query);
		$responses->{$key} = $cli_protocol->send_and_wait_for($command, $device_prompt_regex);
	}
}

sub get_qradar_version {

	my $version = `/opt/qradar/bin/myver`;
	$version =~ s/\.//g;
	$version =~ /(\d\d\d)/;
	return $1;
}

1;
