# Auth: John "Jak" Devaney
# Purp: Get a connection to the 3D device and return a regex of expected prompts
# Date: 24.02.14

package ZipTie::Adapters::Sourcefire::3D::AutoLogin;

use strict;
use warnings;

use ZipTie::CLIProtocol;
use ZipTie::Recording;
use ZipTie::Response;
use ZipTie::Logger;
use Exporter qw{import};
our @EXPORT_OK = qw{login};

my $LOGGER = ZipTie::Logger::get_logger();

sub login{
	my ($cli_protocol, $connection_path) = @_;

	# Store all the information needed to connect to the device
	my $credentials = $connection_path->get_credentials();
	my $ip_address = $connection_path->get_ip_address();
	my $port = $cli_protocol->get_port();
	my $username = $credentials->{username};
	my $password = $credentials->{password};

	my $protocol_name = $cli_protocol->get_protocol_name();
	$LOGGER->debug(qq{Protocol in use: $protocol_name, port: $port});
	$LOGGER->debug(qq{Credentials in use: username: $username, ipAddr: $ip_address});

	# Attempt to connect to the device
	$cli_protocol->connect($ip_address, $port, $username, $password);    
	$LOGGER->debug(q{Verifying the initial connection ...});

	# Now we need to go through the device's shells and collect prompts for a master prompt regex
	my $prompt_regex = _get_master_prompt($cli_protocol, $credentials);
	$LOGGER->debug(qq{Login has successfully completed!\n});
	
	return($prompt_regex);
}

sub _get_master_prompt{
	my ($cli_protocol, $credentials) = @_;

	# Make an array of possible prompts and the subroutines to call if they are matched
	my @responses = (
		ZipTie::Response->new(q{(#|\$|\>)\s*$}),
		ZipTie::Response->new(q{(\>\s*expert\a?)\s*$}),
		ZipTie::Response->new(q{([Pp]assword)\s*\:\s*$})
	);

	# Get the first prompt - the sourcefire shell
	$cli_protocol->wait_for_responses(\@responses);
	my $prompt_regex = q{[^-]} . _get_prompt($cli_protocol);

	# Attempt to enter expert mode
	$cli_protocol->send(q{expert});
	$cli_protocol->wait_for_responses(\@responses);
	my $expert_prompt = _get_prompt($cli_protocol);

	if($expert_prompt =~ /expert/i){
		# Fail the backup as expert mode cannot be entered
		$LOGGER->fatal(q{Expert mode is disabled on target device. It must be enabled, however it is usually irreversible without a hotfix from Sourcefire Support.});
	}else{
		# Add the new expert mode prompt to the master prompt
		$prompt_regex .= q{|}.$expert_prompt;
	}
	
	# Attempt to switch user to root
	my $password = $credentials->{password};
	$cli_protocol->send(q{sudo su -});
	$cli_protocol->wait_for_responses(\@responses);
	my $root_prompt = _get_prompt($cli_protocol);

	if($root_prompt =~ /assword/i){

		# Send root password quietly
		$LOGGER->debug(q{Sending root password quietly});
		$LOGGER->set_logging_level(1);
		$cli_protocol->send($password);
		$LOGGER->set_logging_level(0);
		$cli_protocol->wait_for_responses(\@responses);
		$root_prompt = _get_prompt($cli_protocol);

	}elsif($root_prompt =~ /try again/i){
		$LOGGER->fatal(q{Switching user to root failed.});
	}

	# Add the new root prompt to the master prompt
	$prompt_regex .= q{|}.$root_prompt;

	# Back out to first prompt
	$prompt_regex = q{(}.$prompt_regex.q{)\s*$};
	$cli_protocol->send(q{exit});
	$cli_protocol->send(q{exit});

	_print(q{MASTER REGULAR EXPRESSION TO MATCH DEVICE PROMPT}, $prompt_regex);

	$cli_protocol->set_more_prompt(q{\e\[\d+m-+More-+\e\[\d+m\s*$}, q{20});
	return($prompt_regex);
}

sub _get_prompt{
	my ($cli_protocol) = @_;

	# Grab the last match that the CLIProtocol implementation captured
	my $last_match = $cli_protocol->last_match();
	_print(q{LAST MATCH CAPTURED}, $last_match);

	# Use a regular expression to strip the prompt from the match
	$last_match =~ /([^\s][^\r\n]*)$/;
	my $device_prompt = $1;

	# Register the device prompt with the ZipTie::Recording singleton instance module
	my $recording = ZipTie::Recording::get_recording();
	$recording->set_device_prompt($device_prompt);

	# Clean up the device prompt for regular expression use
	$device_prompt =~ s/^\s+//;
	$device_prompt =~ s/\s+$//;
	$device_prompt = quotemeta($device_prompt);

	_print(q{REGULAR EXPRESSION TO MATCH DEVICE PROMPT}, $device_prompt);
	return($device_prompt);
}

sub _print{
	my ($title, $value) = @_;
	my $delimiter = q{-} x 57;

	$LOGGER->debug($delimiter);
	$LOGGER->debug(qq{[$title]});
	$LOGGER->debug($value);
	$LOGGER->debug($delimiter);
}

1;
