# Auth: John "Jak" Devaney
# Purp: Parses passed responses from a Sourcefire 3D device to xml
# Date: 24.02.14

package ZipTie::Adapters::Sourcefire::3D::Parsers;

use strict;
use warnings;

use ZipTie::Adapters::Utils qw{get_crep strip_mac get_interface_type mask_to_bits};
use ZipTie::Snort;
use ZipTie::Logger;
use Data::Dumper;

use Exporter qw{import};
our @EXPORT_OK = qw{parse_system parse_chassis parse_interfaces get_vuln_ids parse_object_groups parse_filters print_interfaces parse_local_accounts parse_static_routes};

my $LOGGER = ZipTie::Logger::get_logger();
my $vuln_mapping = {};

sub parse_system{
	my ($in, $out) = @_;
	$LOGGER->debug(q{[SYSTEM] Started});

	if($in->{hostname} =~ /(HOSTNAME\s*)+(\S+)/i){
		$out->print_element(q{core:systemName}, $2);
	}

	$LOGGER->debug(q{[SYSTEM] OS Information});
	$out->open_element(q{core:osInfo});
	$out->print_element(q{core:make}, q{Sourcefire} );
	$out->print_element(q{core:name}, q{3D});
	if($in->{version} =~ /^MODEL.+VERSION\s+([\d\.]+)\s+/mi){
		$out->print_element(q{core:version}, $1);
	}
	$out->print_element(q{core:osType}, q{3D});

	$LOGGER->debug(q{[SYSTEM] Packet Processing Order});
	$out->open_element(q{core:packetProcessingOrdering});
	foreach my $phase(q{packet-filtering}, q{address-translation}, q{routing}){
		my ($name, $position) = split(q{/}, $phase);
		$out->open_element(q{core:packetProcessingElement});
		$out->print_element(q{core:packetProcessingPhase}, $name);
		if($position){
			$out->print_element(q{core:processingPoint}, $position);
		}
		$out->close_element(q{core:packetProcessingElement});
	}
	$out->close_element(q{core:packetProcessingOrdering});

	$out->close_element(q{core:osInfo});

	$out->print_element(q{core:deviceType}, q{IDS/IPS});

	$LOGGER->debug(q{[SYSTEM] Finished});
}

sub parse_chassis{
	my ($in, $out) = @_;
	my $mac_regex = get_crep(q{mac1}) . q{|} . get_crep(q{mac2}) . q{|} .  get_crep(q{mac3}) . q{|} .  get_crep(q{mac4});
	$LOGGER->debug(q{[CHASSIS] Started});

	$out->open_element(q{chassis});

	$LOGGER->debug(q{[CHASSIS] Assets});
	my $asset = {q{core:assetType} => q{Chassis}};
	$asset->{q{core:factoryinfo}}->{q{core:make}} = q{Sourcefire};
	if($in->{version} =~ /^MODEL\s*\:\s*(.*?)\s*(\(\d+\)|VERSION)/mi){
		$asset->{q{core:factoryinfo}}->{q{core:modelNumber}} = $1;
	}
	if($in->{version} =~ /^UUID\s*\:\s*(\S+)\s*$/mi){
		$asset->{q{core:factoryinfo}}->{q{core:serialNumber}} = $1;
	}
	$out->print_element(q{core:asset}, $asset);
	
	$LOGGER->debug(q{[CHASSIS] Disks});
	my $BLOCK_SIZE = 1024;
	while($in->{disk} =~ /^(\S+)\s+(\d+)\s+\d+\s+(\d+)\s+/mg){
		my $storage = {
			name => substr($1, rindex($1, q{/}) + 1),
			storageType => q{disk},
			size => $2 * $BLOCK_SIZE,
			freeSpace => $3 * $BLOCK_SIZE
		};
		$out->print_element(q{deviceStorage}, $storage);
	}

	if($in->{network} =~ /MAC\s+ADDRESS\s+\:\s*($mac_regex)/i){
		$out->print_element(q{macAddress}, strip_mac($1));
	}

	$LOGGER->debug(q{[CHASSIS] Memory});
	if($in->{memory} =~ /^MEM\:\s+(\d+)\s+/mi){
		$out->print_element(q{memory}, {q{core:description} => q{RAM}, kind => q{RAM}, size => $1 * $BLOCK_SIZE});
	}
	if($in->{memory} =~ /^SWAP\:\s+(\d+)\s+/mi){
		$out->print_element(q{memory}, {q{core:description} => q{SWAP}, kind => q{ROM}, size => $1 * $BLOCK_SIZE});
	}

	$out->close_element(q{chassis});
	$LOGGER->debug(q{[CHASSIS] Finished});
}

sub get_vuln_ids{
	my ($in, $out) = @_;
	my $applications;

	# Get SNORT applications
	foreach my $key(keys(%{$in->{snorts}})){
		my $snort = $in->{snorts}->{$key};
		my $snort_name = $snort->get_config_name();
		my $snort_applications = $snort->get_applications();
		my $n_snort_applications = scalar(@$snort_applications);
		if($n_snort_applications){
			foreach my $application(@$snort_applications){
				my $msg_classifier;
				foreach my $classifier(@{$application->{q{ztApp:appClassification}}}){
					if($classifier->{q{ztApp:classifierName}}){
						if($classifier->{q{ztApp:classifierName}} eq q{reference}){
							if($classifier->{q{ztApp:classifierValue}} =~ /(^\d+\-\d+$)/){
								$vuln_mapping->{$application->{q{ztApp:appName}}}->{cve} = $1;
							}
						}elsif($classifier->{q{ztApp:classifierName}} eq q{sid}){
							$vuln_mapping->{$application->{q{ztApp:appName}}}->{sid} = $classifier->{q{ztApp:classifierValue}};
						}
					}
				}
			}
		}
	}
}

sub parse_object_groups{
	my ($in, $out) = @_;
	my $ip_regex = get_crep(q{cipm});
	my $object_groups;
	$LOGGER->debug(q{[OBJECT GROUPS] Started});

	# Get SNORT object groups, network and service
	foreach my $key (keys(%{$in->{snorts}})){
		my $snort = $in->{snorts}->{$key};
		my $snort_name = $snort->get_config_name();

		my $network_groups = $snort->get_network_groups();
		my $n_snort_network_groups = scalar(@$network_groups);
		if($n_snort_network_groups){
			$LOGGER->debug(qq{[OBJECT GROUPS] Found $n_snort_network_groups SNORT network groups in $snort_name});
			$object_groups->{networkGroup} = $network_groups;
		}
		
		my $service_groups = $snort->get_service_groups();
		my $n_snort_service_groups = scalar(@$service_groups);
		if($n_snort_service_groups){
			$LOGGER->debug(qq{[OBJECT GROUPS] Found $n_snort_service_groups SNORT service groups in $snort_name});
			$object_groups->{serviceGroup} = $service_groups;
		}
	}

	if($in->{policy_text} =~ /^OBJECTS\s*\:\s*(.*)\s*SECURITY\s+INTELLIGENCE\s*\:\s*$/sim){
		my $objects = $1;
		foreach my $object(split(/^\-/m, $objects)){
			my $name;
			if($object =~ /^\s*NAME\s*\:\s*(.*?)\s*$/mi){
				$name = $1;
			}else{
				next;
			}

			if($object =~ /^\s*TYPE\s*\:\s*(\S+)\s*$/mi){
				my $type = lc($1);

				# Convert objectgroups to their more specific group type
				if($type eq q{objectgroup}){
					if($object =~ /^\s*GROUP\s+TYPE\s*\:\s*(\S+)\s*$/mi){
						$type = lc($1);
					}
				}

				my $object_group = {
					description => $name
				};

				if($object =~ /^\s*ID\s*\:\s*(\S+)\s*$/mi){
					$object_group->{id} = $1;
				}

				if($object =~ /^\s*OBJECTS\s*\:\s*(.*?)\s*\S+\s*\:/sim){
					my $object_refs = $1;
					foreach my $object_ref(split(/\s*\-\s+/, $object_refs)){
						if($object_ref =~ /([0-9a-f\-]+)/i){
							push(@{$object_group->{objectGroupReference}}, $1);
						}
					}
				}

				if($type eq q{serviceobject}){
					my $port;
					if($object =~ /^\s*PORT\s*\:\s*(.*)\s*$/mi){
						$port = $1;
					}

					my $protocol;
					if($object =~ /^\s*PROTOCOL\s*\:\s*(.*?)\s*$/mi){
						$protocol = $1;
						$protocol =~ s/'//g;
						if($protocol == 6){
							$protocol = q{tcp};
						}elsif($protocol == 17){
							$protocol = q{udp};
						}else{
							$protocol = undef;
							$LOGGER->debug(qq{[OBJECT GROUPS] Unsupported protocol: $protocol});
						}
					}

					if($port){
						if($port =~ /(\d+)\-(\d+)/){
							$object_group->{portRange} = {
								portStart => $1,
								portEnd => $2
							};
							if($protocol){
								$object_group->{portRange}->{protocol} = $protocol;
							}
						}elsif($port =~ /'?(\d+)'?/){
							$object_group->{portExpression} = {
								port => $1,
								operator => q{eq}
							};
							if($protocol){
								$object_group->{portExpression}->{protocol} = $protocol;
							}
						}else{
							$LOGGER->debug(qq{[OBJECT GROUPS] Unsupported port expression: $port});
						}
					}

					push(@{$object_groups->{serviceGroup}}, $object_group);

				}elsif($type eq q{networkobject}){
					if($object =~ /^\s*IP\s*\:\s*(.*?)\s*\S+\s*\:/sim){
						my $addresses = $1;
						foreach my $address(split(/\s*\-\s+/, $addresses)){
							if($address =~ /($ip_regex)(\/(\d+))?/){
								my $ip = $1;
								my $mask = $3 // 32;

								if($mask == 32){
									push(@{$object_group->{host}}, $ip);
								}else{
									my $subnet = {
										address => $ip,
										mask => $mask
									};
									push(@{$object_group->{subnet}}, $subnet);
								}
							}
						}
					}

					push(@{$object_groups->{networkGroup}}, $object_group);
				}elsif($type eq q{iplistobject}){
					# IP list / feed objects not done this way, should they be?
				}
			}
		}
	}	

	# Create network groups for white and black, lists and feeds
	foreach my $uuid(keys(%{$in->{lists}})){
		_parse_ip_file($in, $uuid);
	}
	if($in->{parsed_lists}){
		foreach my $parsed_list(keys(%{$in->{parsed_lists}})){
			push(@{$object_groups->{networkGroup}}, $in->{parsed_lists}->{$parsed_list});
		}
	}

	if($in->{parsed_application_groups}){
		foreach my $application_group_key(keys(%{$in->{parsed_application_groups}})){
			my $application_group = $in->{parsed_application_groups}->{$application_group_key};
			push(@{$object_groups->{applicationGroup}}, $application_group);
		}
	}

	if($object_groups){
		$out->open_element(q{objectGroups});
		if($object_groups->{applicationGroup}){
				$out->print_element(q{applicationGroup}, $object_groups->{applicationGroup});
		}
		if($object_groups->{networkGroup}){
				foreach my $network_group(@{$object_groups->{networkGroup}}){
					$out->open_element(q{networkGroup});
					foreach my $element(q{description}, q{id}, q{objectGroupReference}, q{host}, q{range}, q{subnet}){
						if($network_group->{$element}){
							$out->print_element($element, $network_group->{$element});
						}
					}
					$out->close_element(q{networkGroup});
				}
		}
		if($object_groups->{serviceGroup}){
				$out->print_element(q{serviceGroup}, $object_groups->{serviceGroup});
		}
		$out->close_element(q{objectGroups});
	}
	$LOGGER->debug(q{[OBJECT GROUPS] Finished});
}

sub parse_filters{
	my ($in, $out) = @_;
	my $filters;
	$LOGGER->debug(q{[FILTERS] Started});
	$out->open_element(q{filterLists});

	# Firewall Section
	$out->open_element(q{filterListsSection});
	parse_ips_policy($in, $out);
	_print_section_phasing($out, q{address-translation}, q{before});
	$out->close_element(q{filterListsSection});

	$out->close_element(q{filterLists});
	$LOGGER->debug(q{[FILTERS] Finished});
}

sub parse_interfaces{
	my ($in, $out) = @_;
	my $ip_regex = get_crep(q{cipm});
	my $mac_regex = get_crep(q{mac1}) . q{|} . get_crep(q{mac2}) . q{|} .  get_crep(q{mac3}) . q{|} .  get_crep(q{mac4});
	my $interfaces;
	$LOGGER->debug(q{[INTERFACES] Started});

	foreach my $if_section(split(/^-/m, $in->{interfaces})){
		my $interface;
		if($if_section =~ /^-+\[\s+(.+)\s+\]-+/mi){
			$interface->{name} = $1;
			$interface->{interfaceType} = get_interface_type($1);
		}
		if($if_section =~ /^PHYSICAL\s+INTERFACE\s+\:\s*(.+?)\s*$/mi){
			if($interface->{name}){
				$interface->{physical} = $interface->{name} eq $1 ? q{true} : q{false};
			}
		}
		if($if_section =~ /^TYPE\s+\:\s*(.+?)\s*$/mi){
			$interface->{description} = $1;
		}
		if($if_section =~ /^SECURITY\s+ZONE\s+\:\s*(.+?)\s*$/mi){
			push(@{$interface->{interfaceGrouping}}, {
				ifGroupName => $1,
				ifGroupingCharacteristic => q{securityZoneMember}
			});
		}
		if($if_section =~ /^MTU\s+\:\s*(\d+)/mi){
			$interface->{mtu} = $1;
		}
		if($if_section =~ /^MAC\s+ADDRESS\s+\:\s*($mac_regex)/mi){
			$interface->{interfaceEthernet}->{macAddress} = strip_mac($1);	
		}
		if($interface){
			push(@{$interfaces->{interface}}, $interface);
			$LOGGER->debug(q{[INTERFACES] Added interface: }.$interface->{name});
		}
	}

	foreach my $if_section(split(/^(?!\s)/m, $in->{ip_addr})){
		foreach my $interface(@{$interfaces->{interface}}){
			my $name = $interface->{name};
			if($if_section =~ /^\d+:\s+$name:\s+<(\S+)>/){
				$interface->{adminStatus} = $1 =~ /,UP,/i ? q{up} : q{down};
				if($if_section =~ /INET\s+($ip_regex)\/(\d+)\s+(BRD\s+($ip_regex))?/i){
					my $ipConfiguration = {
						ipAddress => $1,
						mask => $2
					};
					if($4){
						$ipConfiguration->{broadcast} = $4;
					}
					push(@{$interface->{interfaceIp}->{ipConfiguration}}, $ipConfiguration);
				}
			}
		}
	}

	if($interfaces){
		my $n_interfaces = scalar(@{$interfaces->{interface}});
		$LOGGER->debug(qq{[INTERFACES] Found $n_interfaces interfaces});

		# Hold off until later to print the interfaces
		$in->{parsed_interfaces} = $interfaces;
	}
	$LOGGER->debug(q{[INTERFACES] Parsed});
}

sub print_interfaces{
	my ($in, $out) = @_;
	$LOGGER->debug(q{[INTERFACES] Printing});

	# If we have any parse interfaces print them
	if($in->{parsed_interfaces}){
		$out->print_element(q{interfaces}, $in->{parsed_interfaces});
	}
	$LOGGER->debug(q{[INTERFACES] Finished});
}

sub parse_local_accounts{
	my ($in, $out) = @_;
	my $accounts;
	$LOGGER->debug(q{[LOCAL ACCOUNTS] Started});

	# Fill array with names, password and access level
	while($in->{users} =~ /^([^\s:]+):([^:]+):\d+:(\d+):([^:]+):.+$/mg){
		my $account = {
			accountName => $1,
			password => $2,
			accessLevel => $3,
			fullName => $4
		};
		push(@{$accounts->{localAccount}}, $account);
	}

	# Print the array of accounts
	if($accounts){
		my $n_accounts = scalar(@{$accounts->{localAccount}});
		$LOGGER->debug(qq{[LOCAL ACCOUNTS] Found $n_accounts local accounts});
		$out->print_element(q{localAccounts}, $accounts);
	}
	$LOGGER->debug(q{[LOCAL ACCOUNTS] Finished});
}

sub parse_static_routes{
	my ($in, $out) = @_;
	my $ip_regex = get_crep(q{cipm});
	my $static_routes;
	my $default_interface = q{eth0};
	$LOGGER->debug(q{[STATIC ROUTES] Started});

	# Collect routes from route table and add to an array
	while($in->{route_n} =~ /^($ip_regex)\s+($ip_regex|-)\s+($ip_regex)\s+\S+\s+\d+\s+\S+\s+\d+\s+(\S+)\s*$/mg){
		my $static_route = {
			defaultGateway => $1 eq q{0.0.0.0} ? q{true} : q{false},
			destinationAddress => $1,
			destinationMask => mask_to_bits($3),
			interface => $4
		};
		my $gateway_address = $2;
		if($gateway_address eq q{-}){
			$static_route->{discard} = q{true};
		}else{
			$static_route->{discard} = q{false};
			if($gateway_address ne q{0.0.0.0}){
				$static_route->{gatewayAddress} = $gateway_address;
			}
		}
		if($static_route->{defaultGateway} eq q{true}){
			$default_interface = $static_route->{interface};
		}
		push(@{$static_routes->{staticRoute}}, $static_route);
	}

	# Give discard static routes the default interface
	foreach my $static_route(@{$static_routes->{staticRoute}}){
		if($static_route->{interface} eq q{-}){
			$static_route->{interface} = $default_interface;
		}
	}

	# Print static route array
	if($static_routes){
		my $n_static_routes = scalar(@{$static_routes->{staticRoute}});
		$LOGGER->debug(qq{[STATIC ROUTES] Found $n_static_routes static routes});
		$out->print_element(q{staticRoutes}, $static_routes);
	}
	$LOGGER->debug(q{[STATIC ROUTES] Finished});
}

sub parse_ips_policy{
	my ($in, $out) = @_;
	$LOGGER->debug(qq{[DEFAULT ACTION] Started});

	# Get the type of default policy and the IPS policy uuid if it exists
	my $action = q{permit};
	my $logging = q{false};
	my $default_policy_uuid;
	if($in->{policy_text} =~ /^DEFAULT\s+ACTION\s*\:\s*(.*?)\s*DESCRIPTION/ims){
		my $default_action = $1;
		if($default_action =~ /ACTION\s*\:\s*(\S+)\s*/i){
			my $default_action_action = $1;
			if($default_action_action eq q{allow}){
				$action = q{none};
				if($default_action =~ /INTRUSION\s+POLICY\s*\:\s*(\S+)\s*/i){
					$default_policy_uuid = q{MASTER};
				}
			}elsif($default_action_action eq q{fast-path}
			    || $default_action_action eq q{discover}){
				$action = q{permit};
			}elsif($default_action_action eq q{block}){
				$action = q{drop};
			}else{
				$LOGGER->debug(qq{[DEFAULT ACTION] Unrecognized default action: $default_action_action});
			}
		}
		if($default_action =~ /^\s*BEGINNING:\s*'?(\d+)'?\s*DC:\s*'?\d+'?\s*END:\s*'?(\d+)'?/mi){
			my $log_start = $1;
			my $log_end = $2;
			if($log_start || $log_end){
				$logging = q{true};
			}
		}
	}

	# Get the IPS policy if the default action is an IPS policy
	my @snort_rules;
	my $snort_name = q{};
	if($default_policy_uuid){
		if($in->{snorts}->{$default_policy_uuid}){
			my $snort = $in->{snorts}->{$default_policy_uuid};
			$snort_name = $snort->get_config_name();

			my $rules = $snort->get_rules();
			my $name_mapping = $snort->get_generated_names_to_msg();
			if(scalar(@$rules)){

				# Add vulnerability element to each rule
				foreach my $rule(@$rules){
					if($rule->{application} && $rule->{application}->{applicationID}){
						my $vuln_name = $rule->{application}->{applicationID};
						delete $rule->{application}->{applicationID};
						my $vuln_type = "QRM_SID";
						if(int($in->{'qradarVersion'}) >= 726){
							if($vuln_mapping->{$vuln_name}->{cve}){
								$vuln_name = $vuln_mapping->{$vuln_name}->{cve};
								$vuln_type = "QRM_CVE";
							}elsif($vuln_mapping->{$vuln_name}->{sid}){
								$vuln_name = $vuln_mapping->{$vuln_name}->{sid};
							}
						}
						$rule->{vulnerability} = {
							name => $vuln_name,
							type => $vuln_type
						};
					}
					
					push(@snort_rules, $rule);
				}
			}
		}else{
			$LOGGER->debug(qq{[DEFAULT ACTION] Could not find SNORT instance with uuid: $default_policy_uuid});
		}
	}

	foreach my $all(q{inline}, q{passive}){
		my $name = q{Default action};

		# Find all interfaces of type $all
		my @interfaces;
		foreach my $interface(@{$in->{parsed_interfaces}->{interface}}){
			if(lc($interface->{description}) eq $all){
				push(@interfaces, $interface->{name});
			}
		}
		if(!@interfaces){
			next;
		}

		$out->open_element(q{filterList});

		if($all eq q{passive} && $action eq q{drop}){
			$action = q{none}
		}
		$out->print_element(q{defaultAction}, $action);
		$out->print_element(q{description}, q{The default action rules});

		# Apply to all interfaces of the type $all
		foreach my $interface(@interfaces){
			foreach my $direction(q{source}, q{destination}){
				$out->open_element(q{appliesTo});
				$out->print_element($direction, $interface);
				$out->close_element(q{appliesTo});
			}
		}

		# Print the IPS policy if the default action is an IPS policy
		if(@snort_rules){
			my $n_snort_rules = scalar(@snort_rules);

			if($n_snort_rules){
				$LOGGER->debug(qq{[DEFAULT ACTION] Found $n_snort_rules SNORT rules in $snort_name for its }.lc($name));

				# Print filters
				foreach my $snort_rule(@snort_rules){
					if($all eq q{passive} && ($snort_rule->{primaryAction} eq q{drop} || $snort_rule->{primaryAction} eq q{deny})){
						$snort_rule->{primaryAction} = q{none};
					}
					$out->print_element(q{filterEntry}, $snort_rule);
				}
			}
		}else{
			$out->print_element(q{filterEntry}, {
				q{log} => $logging,
				name => qq{Default action to $action for $all interfaces},
				primaryAction => $action,
				processOrder => 1
			});
		}

		$out->print_element(q{mode}, q{stateful});
		$out->print_element(q{name}, $name);
		$out->print_element(q{processingPhase}, {q{core:packetProcessingPhase} => q{packet-filtering}});
		$out->close_element(q{filterList});
	}

	$LOGGER->debug(qq{[DEFAULT ACTION] Finished});
}

sub _parse_ip_file{
	my ($in, $uuid) = @_;
	my $file = $in->{lists}->{$uuid};
	my $ip_regex = get_crep(q{cipm});

	# Attempt to get the name of the file
	my $name = q{};
	if($file =~ /^\s*\#(.*\:)?\s*(.*?)\s*$/m){
		$name = qq{ $2};
	}

	my $network_group = {
		id => $uuid,
		description => qq{Security Intelligence List / Feed$name}
	};

	# While there are CIDRS add to the network group
	while($file =~ /^\s*($ip_regex)(\/(\d+))?.*$/mg){
		my $ip = $1;
		my $mask = $3 // 32;

		if($mask == 32){
			push(@{$network_group->{host}}, $ip);
		}else{
			my $subnet = {
				address => $ip,
				mask => $mask
			};
			push(@{$network_group->{subnet}}, $subnet);
		}
	}

	if($network_group->{host} || $network_group->{subnet}){
		$in->{parsed_lists}->{$uuid} = $network_group;
	}
}

sub _print_section_phasing{
	my ($out, $phase, $when) = @_;

	# Print a filter list section phase discriminator
	my $phaseDiscriminator;
	$phaseDiscriminator->{processingPhase}->{q{core:packetProcessingPhase}} = $phase;
	$phaseDiscriminator->{processingPhasePosition} = $when;
	$out->print_element(q{phaseDiscriminator}, $phaseDiscriminator);
}

sub _get_zone_objects{
	my ($in) = @_;

	# If this suubroutine has not already been called before
	if(!$in->{parsed_zones}){
		# Get all the zones and store their uuid, name and type
		foreach my $object(split(/^-\s*/m, $in->{policy_text})){
			if($object =~ /ID\s*\:\s*([0-9A-F-]+)\s*NAME\s*\:\s*(.*?)\s*TYPE\s*\:\s*SECURITYZONE/i){
				my $uuid = $1;
				my $name = $2;
				$in->{parsed_zones}->{$uuid}->{name} = $name;
			}
		}
	}

	return($in->{parsed_zones});
}
1;
