# Auth: John "Jak" Devaney
# Purp: Collects layer 2 data from a Sourcefire 3D device
# Date: 21.05.14

package ZipTie::Adapters::Sourcefire::3D::Telemetry;

use strict;
use warnings;

use Data::Dumper;
use ZipTie::Logger;
use ZipTie::Typer;
use ZipTie::CLIProtocolFactory;
use ZipTie::Model::XmlPrint;
use ZipTie::Adapters::Utils qw{get_model_filehandle close_model_filehandle get_crep strip_mac get_interface_type choose_admin_ip};
use ZipTie::Adapters::Sourcefire::3D::AutoLogin qw{login};
use ZipTie::Adapters::Sourcefire::3D::Disconnect qw{disconnect};

use Exporter qw{import};
our @EXPORT_OK = qw{parse_system parse_neighbors parse_telemetry_interfaces};

my $LOGGER = ZipTie::Logger::get_logger();

sub invoke{
	my ($package, $connection_path_doc) = @_;
	
	my ($connection_path, $discovery_params) = ZipTie::Typer::translate_document($connection_path_doc, q{connectionPath});
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);	

	my $filehandle = get_model_filehandle(q{3D}, $connection_path->get_ip_address());
	my $printer = ZipTie::Model::XmlPrint->new($filehandle, q{telemetry});

	my $prompt_regex = login($cli_protocol, $connection_path);
	my $responses = _get_responses($cli_protocol, $prompt_regex);

	$printer->attributes(1);
	$printer->open_discovery_event();

	parse_system($responses, $printer);
	parse_neighbors($responses, $printer);

	my $interfaces = parse_telemetry_interfaces($responses, $printer);
	my $admin_ip = $connection_path->get_ip_address();
	if($discovery_params->{calculateAdminIp} eq q{true}){
		$admin_ip = choose_admin_ip($admin_ip, $interfaces);
	}
	$printer->print_element(q{adminIp}, $admin_ip);
	
	$printer->close_element(q{DiscoveryEvent});
	close_model_filehandle($filehandle);
	disconnect($cli_protocol);
}

sub _get_responses{
	my ($cli_protocol, $device_prompt_regex) = @_;

	my @key_to_commands = (
			q{hostname->hostname},
			q{arp->arp -n},
			q{ip_link->ip -s link},
			q{ip_addr->ip addr}
	);

	$cli_protocol->send_and_wait_for(q{expert}, $device_prompt_regex);

	my $responses;
	foreach my $key_to_command(@key_to_commands){
		my ($key, $command) = split(q{->}, $key_to_command);
		$responses->{$key} = $cli_protocol->send_and_wait_for($command, $device_prompt_regex);
	}

	return($responses);
}

sub parse_system{
	my ($in, $out) = @_;
	$LOGGER->debug(q{[TELEMETRY SYSTEM] Started});

	if($in->{hostname} =~ /\s*HOSTNAME\s*(\S+)\s*/i){
		$out->print_element(q{sysName}, $1);
	}

	$out->print_element(q{sysObjectId}, q{1.3.6.1.4.1.14223});
	$out->print_element(q{sysDescr}, q{Sourcefire 3D sensor});

	$LOGGER->debug(q{[TELEMETRY SYSTEM] Finished});
}

sub parse_neighbors{
	my ($in, $out) = @_;
	$LOGGER->debug(q{[TELEMETRY NEIGHBORS] Started});

	$out->attributes(1);
	$out->open_element(q{neighbors});

	parse_arp($in, $out);

	$out->close_element(q{neighbors});

	$LOGGER->debug(q{[TELEMETRY NEIGHBORS] Finished});
}

sub parse_arp{
	my ($in, $out) = @_;
	$LOGGER->debug(q{[TELEMETRY ARP] Started});

	my $ip_regex = get_crep(q{cipm});
	my $mac_regex = get_crep(q{mac1}) . q{|} . get_crep(q{mac2}) . q{|} .  get_crep(q{mac3}) . q{|} .  get_crep(q{mac4});

	my $arp_table;
	while($in->{arp} =~ /^($ip_regex)\s+\S+\s+($mac_regex)\s+\S+\s+(\S+)\s*$/mg){
		push(@{$arp_table->{arpEntry}}, 
			{
				ipAddress  => $1,
				macAddress => strip_mac($2),
				interface => $3
			}
		);		
	}

	if($arp_table){
		$out->print_element(q{arpTable}, $arp_table);
	}

	$LOGGER->debug(q{[TELEMETRY ARP] Finished});
}

sub parse_telemetry_interfaces{
	my ($in, $out) = @_;
	$LOGGER->debug(q{[TELEMETRY INTERFACES] Started});

	my $ip_regex = get_crep(q{cipm});
	my $interfaces;

	while($in->{ip_link} =~ /^\d+:\s+(\S+):\s+.*?\s+STATE\s+(\S+).*?\s+LINK\/(\S+)\s+.*?RX:.*?\s+(\d+)/imgs){

		push(@{$interfaces->{interface}},
			{
				name => $1,
				operStatus => lc($2) eq q{up} ? q{Up} : q{Down},
				type => get_interface_type($3),
				inputBytes => $4
			}
		);
	}

	$LOGGER->debug(q{[TELEMETRY INTERFACES] Basic interfaces built});

	foreach my $section(split(/^(?!\s)/m, $in->{ip_addr})){
		my $interface;

		if($section =~ /^\d+:\s+(\S+):\s+/){
			my $name = $1;

			foreach my $if(@{$interfaces->{interface}}){
				if($if->{name} eq $name){
					$interface = $if;
					last;
				}
			}
		}
		if(!$interface){
			next;
		}

		while($section =~ /^\s+INET\s+($ip_regex)\/(\d+)/img){
			push(@{$interface->{ipEntry}},
				{
					ipAddress => $1,
					mask => $2
				}
			);
			
		}
	}

	if($interfaces){
		$out->attributes(1);
		$out->print_element(q{interfaces}, $interfaces);
	}
	$LOGGER->debug(q{[TELEMETRY INTERFACES] Finished});

	return($interfaces);
}

1;
