# Auth: John "Jak" Devaney
# Purp: Populate ZTS with managed sourcefire devices
# Date: 24.02.14

#!/usr/bin/perl

use strict;
use warnings;

use ZipTie::Logger;
use ZipTie::Client;
use ZipTie::Adapters::Utils qw{get_crep};
use ZipTie::Adapters::Sourcefire::3D::Disconnect qw{disconnect};
use ZipTie::CLIProtocolFactory;
use ZipTie::ConnectionPath;
use ZipTie::ConnectionPath::Protocol;
use ZipTie::Credentials;
use ZipTie::Recording;
use Data::Dumper;

# A list of role/adapter pairs to discover
my @ROLES_TO_DISCOVER = (
	{role_id => 2, adapter => q{ZipTie::Adapters::Sourcefire::3D}}
);

if(scalar(@ARGV) != 1){
	_fatal(q{Wrong number of arguments passed to dcdiscovery.pl});
}
my $ui_ip = $ARGV[0];

# Start logging
my $LOGGER = ZipTie::Logger::get_logger();
if($ENV{ADAPTER_LOG_TO_FILE}){
	$LOGGER->enable_logging_to_file(q{Sourcefire_Defense_Center}, q{dcdiscovery}, $ui_ip);
}
$LOGGER->debug(qq{Performing Discover From Defense Center on $ui_ip.});
main($ui_ip);

sub main{
	my ($management_server_ip) = @_;

	# Check parameter is a valid ip
	my $ip_regex = get_crep(q{cipm});
	if($management_server_ip !~ /$ip_regex/){
		_fatal(qq{Invalid IP address: $management_server_ip});
	}

	# Get credentials for management device
	my $ziptie_client = ZipTie::Client->new();
	my ($username, $password, $port) = _get_credentials($ziptie_client, $management_server_ip);

	# Connect to the management device and get managed device IPs, also get a hash of any pre existing devices in ZTS
	$LOGGER->debug(qq{Connecting to Defense Center with username '$username' and port '$port'.});
	my $found_devices = _get_managed_devices($management_server_ip, $username, $password, $port);
	$LOGGER->debug(q{Found } . scalar(@$found_devices) . q{ device(s) on Defense Center});

	$LOGGER->debug(q{Collecting devices already in CSM});
	my $existing_devices = _get_devices($ziptie_client);
	$LOGGER->debug(q{Found } . scalar(keys(%{$existing_devices})) . q{ device(s) in CSM});

	my $device_service = $ziptie_client->devices();
	foreach my $role_adapter_pair(@ROLES_TO_DISCOVER){
		my $role_id = $role_adapter_pair->{role_id};
		my $adapter = $role_adapter_pair->{adapter};

		$LOGGER->debug(qq{Discovering $adapter devices.});
		foreach my $found_device(@$found_devices){
			if($found_device->{role} == $role_id){
				_add_or_update_device($existing_devices, $device_service, $found_device->{ip}, $adapter);
			}
		}
	}

	if($ENV{ENABLE_RECORDING}){
		my $recording_connection_path = ZipTie::ConnectionPath->new();
		$recording_connection_path->set_ip_address($management_server_ip);
		ZipTie::Recording::get_recording()->to_file(q{Sourcefire_Defense_Center}, q{dcdiscovery}, $recording_connection_path);
	}
}

#######################################################################
# Get username, password and port for connecting to the defense center
#######################################################################
sub _get_credentials{
	my ($ziptie_client, $management_server_ip) = @_;
	
	# Get credentials, fail if none are found
	my $credentials_service = $ziptie_client->credentials();
	my @credentials = $credentials_service->getCredentialSetsByIpAddress(ipAddress => $management_server_ip);
	if(!scalar(@credentials)){
		_fatal(q{No credentials found.});
	}

	# Get username and password for the management ip, fail if none are found
	my $username;
	my $password;
	if(ref($credentials[0]->{credentials}) ne q{ARRAY}){
		$credentials[0]->{credentials} = [$credentials[0]->{credentials}];
	}
	foreach my $credential(@{$credentials[0]->{credentials}}){
		if($credential->{name} eq q{username}){
			$username = $credential->{value};
		}elsif ($credential->{name} eq q{password}){
			$password = $credential->{value};
		}
	}
	if(!$username || $username eq q{}){
		_fatal(q{Authentication refused with an empty username.});
	}
	if(!$password || $password eq q{}){
		_fatal(q{Authentication refused with an empty password.});
	}

	# Get port for SSH, if it is enabled, else fail
	my @protocols = $credentials_service->getProtocolConfigByIpAddress(ipAddress => $management_server_ip);
	if(!scalar(@protocols)){
		_fatal(q{No protocols found.});
	}
	my $port;
	foreach my $protocol (@{$protocols[0]->{protocols}}){
		if($protocol->{name} eq q{SSH} && $protocol->{enabled} eq q{true}){
			$port = $protocol->{port};
			last;
		}
	}
	if(!$port || $port eq q{}){
		_fatal(q{Required protocol either disabled or undefined.});
	}

	return($username, $password, $port);
}

#######################################################################
# If the device is already added update it, otherwise add a new device
#######################################################################
sub _add_or_update_device{
	my ($existing_devices, $device_service, $device_ip, $adapter) = @_;
	$LOGGER->debug(qq{Analyzing device $device_ip...});

	# Device already exists else add device
	if($existing_devices->{$device_ip}){
		$LOGGER->debug(qq{Skipping existing device: $device_ip.});
		print(qq{$device_ip,Skipped\n});
	}else{
		my %device = (ipAddress => $device_ip, adapterId => $adapter, managedNetwork => q{Default});
		$device_service->createDevice(%device);
		$existing_devices->{$device_ip} = \%device;

		$LOGGER->debug(qq{Adding new device $device_ip.});
		print(qq{$device_ip,Added\n});
	}
}

###########################################################################
# Get devices currently in CSM. Returns a hash where keys are devices' ips
###########################################################################
sub _get_devices{
	my ($ziptie_client) = @_;
	my %device_hash;

	# Variables for getting page of devices
	my $PAGE_SIZE = 500;
	my $offset = 0;
	my $total;
	my $device_search_service = $ziptie_client->devicesearch();
	
	# While there are still devices to get
	while(!defined($total) || $offset < $total){

		# Get devices, 500 at a time from the current offset
		my $page = $device_search_service->search(scheme => q{ipAddress}, query => q{},	pageData => {pageSize => $PAGE_SIZE, offset => $offset});
		$offset += $PAGE_SIZE;
		$total = $total // $page->{total};

		# Get an array reference of the devices
		my $devices;
		if($page->{devices}){
			$devices = $page->{devices};
			if(ref($devices) ne q{ARRAY}){
				$devices = [$devices];
			}
		}

		# For each device in the array add it to the hash with its ip as the key
		foreach my $device(@$devices){
			$device_hash{$device->{ipAddress}} = $device;
		}
	}
	return(\%device_hash);
}

##########################################################################
# Logs into the Defense Center and returns the ips of the managed devices
##########################################################################
sub _get_managed_devices{
	my ($management_server_ip, $username, $password, $port) = @_;
	my $ip_regex = get_crep(q{cipm});

	# Create a connection path to the Defense Center
	my $connection_path = ZipTie::ConnectionPath->new();
	$connection_path->set_management_server_ip_address($management_server_ip);
	$connection_path->set_credentials(ZipTie::Credentials->new(username => $username, password => $password));
	$connection_path->set_protocols([ZipTie::ConnectionPath::Protocol->new({name => q{SSH}, port => $port})]);

	# Connect to the Defense Center
	my $cli_protocol = ZipTie::CLIProtocolFactory::create($connection_path);
	$cli_protocol->connect($management_server_ip, $port, $username, $password);

	# Get the prompt
	my @responses;
	push(@responses, ZipTie::Response->new(q{(#|\$)\s*$}, \&_get_prompt));
	my $response = $cli_protocol->wait_for_responses(\@responses);
	my $device_prompt_regex;
	if($response){
		my $next_interaction = $response->get_next_interaction();
		$device_prompt_regex = &$next_interaction($cli_protocol, $connection_path->get_credentials());
	}else{
		_fatal(q{Invalid response from device encountered!});
	}
	$cli_protocol->set_prompt_by_name(q{prompt}, $device_prompt_regex);
	
	# Run the command to look in the file with the managed devices
	my $sftunnel = $cli_protocol->send_and_wait_for(q{cat /etc/sf/sftunnel.conf}, $device_prompt_regex);
	disconnect($cli_protocol);

	# Parse through the returned file and return an array of devices found, that we care about
	my @managed_devices;
	my $device;
	foreach my $line(split(/^/, $sftunnel)){
		if($line =~ /^\s+ROLE\s+(\d+)\;\s*$/i){
			$device->{role} = $1;
		}elsif($device){
			# Roles: SERVER=1,CLIENT=2,HA PRIME=3,HA SEC=4,MDC=5,DC=6
			if($line =~ /^\s+IP\s+($ip_regex)\;\s*$/i){
				if($device->{role} == 2){
					$device->{ip} = $1;
					push(@managed_devices, $device);
				}elsif($device->{role} == 6){
					$LOGGER->debug(qq{Discovering off managed Defense Center $1});
					main($1);
				}
				$device = undef;
			}
		}
	}

	if(@managed_devices){
		$LOGGER->debug(qq{Found managed devices:\n}.Dumper(@managed_devices));
	}else{
		$LOGGER->debug(q{No devices found});
	}

	return(\@managed_devices);
}

#####################################################################################
# Gets the device prompt on the Defense Center to look out for when running commands
#####################################################################################
sub _get_prompt{
	my $cli_protocol = shift;
	my $last_match = $cli_protocol->last_match();
	$last_match =~ /([^\s][^\r\n]*)$/;

	# Clean up prompt and regi-fy
	my $device_prompt = $1;
	$device_prompt =~ s/^\s+//;
	$device_prompt =~ s/\s+$//;
	$device_prompt = quotemeta($device_prompt);

	return($device_prompt);
}

#############################
# Used to print fatal errors 
#############################
sub _fatal{
	my ($message) = @_;
	print(qq{ERROR: $message\n});
	$LOGGER->fatal($message);
	exit(1);
}

1;
