package ZipTie::Adapters::TippingPoint::IPS;

use 5.010_000;
use strict;
use warnings;

use Data::Dumper;
use XML::Twig::XPath;

use ZipTie::Adapters::TippingPoint::IPS::AutoLogin qw(cli_connect rest_sms_connection_setup);
use ZipTie::Adapters::TippingPoint::IPS::Parsers qw(parse_sms create_config parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_interfaces parse_ips_rules);
use ZipTie::Adapters::TippingPoint::IPS::Disconnect qw(disconnect);
use ZipTie::Adapters::Utils qw(get_model_filehandle close_model_filehandle);
use ZipTie::CLIProtocol;
use ZipTie::CLIProtocolFactory;
use ZipTie::ConnectionPath;
use ZipTie::Credentials;
use ZipTie::Logger;
use ZipTie::Model::XmlPrint;
use ZipTie::Typer;

# Grab a reference to the ZipTie::Logger
my $LOGGER = ZipTie::Logger::get_logger();

sub backup {
	my ($packageName, $backupDoc) = @_;

	# Will contain device responses to be handed to the Parsers module
	my $responses = {};

	# Translate the backup operation XML document into ZipTie::ConnectionPath
	my ($connectionPath) = ZipTie::Typer::translate_document( $backupDoc, 'connectionPath' );

	# Track the IP address of this device for use in identifying it out of the SMS's managed devices
	my $ip_address = $connectionPath->get_ip_address();
	$responses->{ip_address} = $ip_address;

	# Set up the XmlPrint object for printing the ZiptieElementDocument (ZED)
	my $filehandle = get_model_filehandle( 'TippingPoint IPS', $ip_address );
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common' );
	$printer->open_model();

	# Make a Telnet or SSH connection
	my ( $cliProtocol, $promptRegex ) = cli_connect($connectionPath);

	# Send IPS commands
	$responses = _get_command_responses($cliProtocol, $promptRegex, $responses);

	# Parse sms ip
	my $sms_ip = parse_sms($responses);
	unless( $sms_ip ) {
		$LOGGER->fatal("SMS IP could not be found - This IPS must be under SMS control for successful backup!");
	}
	$LOGGER->debug("SMS IP is $sms_ip");
	$connectionPath->set_management_server_ip_address($sms_ip);
	
	# Make SMS connection via RESTAPI
	my $connection = ZipTie::Adapters::TippingPoint::IPS::AutoLogin::rest_sms_connection_setup($connectionPath);

	# Send SMS comannds
	$responses = _get_sms_command_responses($connection, $responses);
	
	if( !parse_system( $responses, $printer ) ) {
		$LOGGER->fatal('System name is required and was not found');
		return;
	}

	parse_chassis( $responses, $printer );
	create_config( $responses, $printer );
	parse_interfaces( $responses );
	parse_ips_rules( $responses, $printer );
	parse_interfaces( $responses, $printer );

	# Disconnect the IPS connection, finished with callbacks by here
	disconnect( $cliProtocol );

	undef( $responses );

	$printer->close_model();                # close out the ZiptieElementDocument
	close_model_filehandle($filehandle);    # Make sure to close the model file handle

	return(1);
}

sub _get_command_responses {
	my ($cliProtocol, $promptRegex, $responses) = @_;
	
	$responses->{config} = $cliProtocol->send_and_wait_for( 'show config', $promptRegex );
	$responses->{config} =~ s/^.*\b//; # remove the command
	$responses->{config} =~ s/\b.*$//; # remove the prompt
	$responses->{version} = $cliProtocol->send_and_wait_for( 'show version', $promptRegex );
	$responses->{interfaces} = $cliProtocol->send_and_wait_for( 'show interface', $promptRegex );
	$responses->{host} = $cliProtocol->send_and_wait_for( 'show host', $promptRegex );
	$responses->{sms} = $cliProtocol->send_and_wait_for( 'show sms', $promptRegex );

	$responses->{get_filter_callback} =
		sub {
			my ($filter_number) = @_;

			my $response = $cliProtocol->send_and_wait_for( "show filter $filter_number", $promptRegex );

			return($response);
		};

	return( $responses );
}

sub _get_sms_command_responses {
	my ($connection, $responses) = @_;

	# Check the SMS web services are available
	my $sms_status = $connection->get( 'method=Status', 'get status' ); chomp($sms_status);
	$LOGGER->debug("SMS Web Service Status = $sms_status");

	if( $sms_status !~ /OK/i ) {
		$LOGGER->fatal('SMS web services are not available and are needed.');
	}

	my $sms_responses->{signature} = $connection->get( 'method=DataDictionary&table=SIGNATURE&format=xml', 'get signature table' );

	# Check we got valid responses and parse them into XML::Twig objects
	foreach my $key ( keys( %{ $sms_responses } ) ) {
		if( $sms_responses->{$key} =~ /connect failed/i ) {
			my $table = uc($key);
			$LOGGER->fatal( "Could not collect the $table table." );
		}

		my $twig;
		my $eval_success = 1;
		$eval_success = eval {
			$twig = XML::Twig::XPath->new()->parse($sms_responses->{$key}); 1;
		};

		if( !$eval_success ) {
			$LOGGER->fatal("Failed to process SIGNATURE table due to XML parsing error: $@");
		}

		# Add to the responses hash ref
		$responses->{$key} = $twig;
	}

	return( $responses );
}

1;
