package ZipTie::Adapters::TippingPoint::IPS::AutoLogin;

use 5.010_000;
use strict;
use warnings;

use ZipTie::CLIProtocolFactory;
use ZipTie::CLIProtocol;
use ZipTie::Recording;
use ZipTie::RESTAPI;
use ZipTie::Response;
use ZipTie::Logger;

use Exporter qw(import);
our @EXPORT_OK = qw(cli_connect rest_sms_connection_setup);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

sub cli_connect {
    # Grab our arguments
    my $connectionPath = shift;

    # Create a new CLI protocol object by using the ZipTie::CLIProtocolFactory::create sub-routine
    # to examine the ZipTie::ConnectionPath argument for any command line interface (CLI) protocols
    # that may be specified.
    my $cliProtocol = ZipTie::CLIProtocolFactory::create($connectionPath);

    # Make a connection to and successfully authenticate with the device
    my $devicePromptRegex = _execute( $cliProtocol, $connectionPath );

    # Store the regular expression that matches the primary prompt of the device under the key "prompt"
    # on the ZipTie::CLIProtocol object
    $cliProtocol->set_prompt_by_name( 'prompt', $devicePromptRegex );

    # Return the created ZipTie::CLIProtocol object and the device prompt encountered after successfully connecting to a device.
    return ( $cliProtocol, $devicePromptRegex );
}

sub rest_sms_connection_setup {
	my ($connection_path) = @_;

	my $management_server_ip = $connection_path->get_management_server_ip_address();
	my $connection = ZipTie::RESTAPI->new( $connection_path, $management_server_ip );

	my $credentials	= $connection_path->get_credentials();
	my $username = $credentials->{enableUsername};
	my $password = $credentials->{enablePassword};

	$connection->set_api_location("dbAccess/tptDBServlet?");
	$connection->set_connection_token("&smsuser=$username&smspass=$password");

	return( $connection );
}

sub _execute {
	# login to a device via the CLI (SSH or Telnet)
	my $cli_protocol = shift;
	my $connection_path = shift;
	
	# Pull off the ZipTie::Credentials instance from the ZipTie::ConnectionPath object
	my $credentials = $connection_path->get_credentials();

	# Store all the information needed to connect to the device
	my $ip_address = $connection_path->get_ip_address();
	my $port      = $cli_protocol->get_port();

	my $username = $credentials->{username};
	my $password = $credentials->{password};

	my $protocolName = $cli_protocol->get_protocol_name();
	$LOGGER->debug("Protocol in use: $protocolName");

	# TippingPoint needs to be run in binary mode because for some reason it uses CR as command terminators.
	$cli_protocol->binary_mode(1);
	$cli_protocol->output_record_separator("\r");

	# Attempt to connect to the device
	$cli_protocol->connect( $ip_address, $port, $username, $password );    
	$LOGGER->debug("Verifying the initial connection ...");

	# If all goes well during the login process, the last method to be called will be "get_prompt"
	# and its return value will be returned all the way to here
	my $prompt_regex = _initial_connection( $cli_protocol, $credentials );

	# Print that the login process has been completed
	$LOGGER->debug("Login has successfully completed!\n");

	return $prompt_regex;
}

sub _initial_connection
{
	my $cli_protocol = shift;
	my $credentials = shift;
	
	# Specify the responses to check for
	my @responses = ();
	push(@responses, ZipTie::Response->new('sername:|ogin:', \&_send_username));
	push(@responses, ZipTie::Response->new('assword:', \&_send_password));
	push(@responses, ZipTie::Response->new('(#|\$|>)\s*$', \&_get_prompt));
	
	my $response = $cli_protocol->wait_for_responses(\@responses);
	
	# Based on the response of the device, determine the next interaction that should be executed.
	if ($response)
	{
		my $next_interaction = $response->get_next_interaction();
		return &$next_interaction($cli_protocol, $credentials);
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}
}

sub _send_username
{
	my $cli_protocol = shift;
	my $credentials = shift;
	
	# Specify the responses to check for
	my @responses = ();
	push(@responses, ZipTie::Response->new('[Pp]assword:', \&_send_password));
	push(@responses, ZipTie::Response->new('incorrect', undef, $INVALID_CREDENTIALS));
	
	# Send username credential
	my $username = $credentials->{username};
	$LOGGER->debug("Sending username credential ...");
	$cli_protocol->send($username, 1);
	my $response = $cli_protocol->wait_for_responses(\@responses);
	
	if ($response)
	{
		my $next_interaction = $response->get_next_interaction();
		return &$next_interaction($cli_protocol, $credentials);
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}
}

sub _send_password
{
	my $cli_protocol = shift;
	my $credentials = shift;
	
	# Specify the responses to check for
	my @responses = ();
	push(@responses, ZipTie::Response->new('Access denied', undef, $INVALID_CREDENTIALS));
	push(@responses, ZipTie::Response->new('(#|\$|>)\s*$', \&_get_prompt));
	
	# Send password credential
	my $password = $credentials->{password};
	$LOGGER->debug("Sending password credential ...");
	$cli_protocol->send($password, 1);
	my $response = $cli_protocol->wait_for_responses(\@responses);
	
	if ($response)
	{
		my $next_interaction = $response->get_next_interaction();
		return &$next_interaction($cli_protocol, $credentials);
	}
	else
	{
		$LOGGER->fatal("Invalid response from device encountered!");
	}
}

sub _get_prompt
{
	my $cli_protocol = shift;

	# Grab the last match that the CLIProtocol implementation captured
	my $last_match = $cli_protocol->last_match();
	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[LAST MATCH CAPTURED]");
	$LOGGER->debug($last_match);

	# Use a regular expression to strip the prompt from the match
	$last_match =~ /([^\s][^\r\n]*)$/;

	# Grab the extracted device prompt
	my $device_prompt = $1;
	
	# Register the device prompt with the ZipTie::Recording singleton instance module
	my $recording = ZipTie::Recording::get_recording();
	$recording->set_device_prompt($device_prompt);

	# Clean up the device prompt for regular expression use
	$device_prompt =~ s/^\s+//;
	$device_prompt =~ s/\s+$//;
	$device_prompt = quotemeta($device_prompt);

	# Construct the regular expression for the prompt so that it will only match the prompt that is displayed
	# as the very last thing.  This will get around the issue of a device possibly echoing back it's prompt when
	# a command is sent.
	my $device_prompt_regex = $device_prompt . "\\s*\$";
	
	$LOGGER->debug("---------------------------------------------------------");
	$LOGGER->debug("[REGULAR EXPRESSION TO MATCH DEVICE PROMPT]");
	$LOGGER->debug($device_prompt_regex);
	$LOGGER->debug("---------------------------------------------------------");
	
	return $device_prompt_regex;
}

1;
