package ZipTie::Adapters::TippingPoint::IPS::Parsers;

use 5.010_000;
use strict;
use warnings;

use Data::Dumper;
use MIME::Base64 qw(encode_base64);
use ZipTie::Adapters::Utils qw(mask_to_bits get_port_number trim get_interface_type get_crep getUnitFreeNumber);

use Exporter qw(import);
our @EXPORT_OK = qw(parse_sms create_config parse_local_accounts parse_chassis parse_filters parse_snmp parse_system parse_interfaces parse_ips_rules);

my $LOGGER = ZipTie::Logger::get_logger();
my $unknown = "Unknown";
my $make = "HP TippingPoint";

sub parse_chassis {
	my ( $in, $out ) = @_;	
	my $chassis;
	
	$chassis->{'core:asset'}->{'core:assetType'} = 'Chassis';
	$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:make'} = $make;

	if (defined $in->{version}) {
		# parse and store firmware version
		if($in->{version} =~ /\s*Software:\s*(\S+)/){
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:firmwareVersion'} = $1;
		}else{
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:firmwareVersion'} = $unknown;
		}
		# parse and store model number
		if($in->{version} =~ /\s*Model:\s*(\S+)/){
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = 'TippingPoint-' . $1;
		}else{
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:modelNumber'} = $unknown;
		}
		# parse and store serial number
		if($in->{version} =~ /\s*Serial:\s*(\S+)/){
			$chassis->{'core:asset'}->{'core:factoryinfo'}->{'core:serialNumber'} = $1;
		}
	}

	# print out the chassis information
	$out->print_element("chassis", $chassis);

	return;
}

sub parse_system {
	my ( $in, $out ) = @_;

	# parse and print host/system name
	my $system_name_found;
	if (defined $in->{host}) {
		if($in->{host} =~ /Host Name\s*=\s*(.+)/){
			$out->print_element('core:systemName', $1);
			$system_name_found = 1;
		}
	}
	
	$out->open_element('core:osInfo');
	$out->print_element('core:make', $make);
	
	# parse and print software version
	if (defined $in->{version}){
		if($in->{version} =~ /\s*Software:\s*(\S*)/){
			$out->print_element('core:version', $1);
		}else{
			$out->print_element('core:version', $unknown);
		}
	}
	
	$out->print_element('core:osType', 'TOS');
	$out->close_element('core:osInfo');
	$out->print_element('core:deviceType', 'IDS/IPS');

	# parse, calculate and print the last reboot time
	if (defined $in->{version}){
		if ($in->{version} =~ /Uptime is (.+)$/mi){
			my $lastReboot = parse_reboot($1);
			$out->print_element("core:lastReboot", $lastReboot);
		}else{
			$out->print_element("core:lastReboot", $unknown);
		}
	}
	
	return( $system_name_found );
}

sub parse_reboot {
	my $_ = @_;
	
	my ($years) = /(\d+)\s*years?/;
	my ($weeks) = /(\d+)\s*weeks?/;
	my ($days) = /(\d+)\s*days?/;
	my ($hours) = /(\d+)\s*hours?/;
	my ($minutes) = /(\d+)\s*minutes?/;
	my ($seconds) = /(\d+)\s*seconds?/;

	# subract the last reboot from the current time
	my $lastReboot = time();
	$lastReboot -= $years * 52 * 7 * 24 * 60 * 60 if ($years);
	$lastReboot -= $weeks * 7 * 24 * 60 * 60 if ($weeks);
	$lastReboot -= $days * 24 * 60 * 60 if ($days);
	$lastReboot -= $hours * 60 * 60 if ($hours);
	$lastReboot -= $minutes * 60 if ($minutes);
	$lastReboot -= $seconds if ($seconds);

	return( $lastReboot );
}

sub create_config {
	my ( $in, $out ) = @_;

	# the name of the repository
	my $repository = { 'core:name' => '/' };
	
	# remove the CR's from the beginning of each line
	my $raw_config = ($in->{config});
	$raw_config =~ s/[\r]+//g;

	# build the simple text configuration
	my $config;
	$config->{'core:context'}	= 'active';
	$config->{'core:mediaType'}  = 'text/plain';
	$config->{'core:name'}	   = 'config';
	$config->{'core:promotable'} = 'false';
	$config->{'core:textBlob'}   = encode_base64( $raw_config );

	# push the configuration into the repository
	push( @{ $repository->{'core:config'} }, $config );

	# print the repository
	$out->print_element( 'core:configRepository', $repository );

	return( 1 );
}


sub parse_local_accounts
{
	my ( $in, $out ) = @_;
}

sub parse_filters
{
	my ( $in, $out ) = @_;
}

sub parse_snmp
{
	my ( $in, $out ) = @_;
}

sub parse_interfaces {
	my ( $in, $out ) = @_;
	my @interfaces;

	if( $in->{parsed_interfaces} ) {
		@interfaces = @{ $in->{parsed_interfaces} };
	} else {
		# for each interface
		while ($in->{interfaces} =~ /Interface.+?TX\sTotal\sBytes\s+\d+/sg)
		{
			my $raw_interface = $&;
			my $modelled_interface = {
				physical => 'true',
				interfaceType => 'ethernet'
			};

			# Management Interface
			my $ipRegex = get_crep("cipm");
			if ($raw_interface =~ /Host\sIP\s+($ipRegex)\/(\d+)/) {
				$modelled_interface->{name} = "mgmtEthernet";

				my ($ip, $mask) = ($1, $2);
				my $ip_cfg = {ipAddress => $ip};
				$ip_cfg->{mask} = mask_to_bits($mask);
				push(@{ $modelled_interface->{interfaceIp}->{ipConfiguration} }, $ip_cfg);
				
				if ($raw_interface =~ /MAC\s+Address\s*(\S+)/) {
					my $macAddress = $1;
					# strip macAddress of colon chars
					$macAddress =~ s/://g;
					$modelled_interface->{interfaceEthernet}->{macAddress} = $macAddress;
				}
			} # Regular (Segment) Interface
			elsif ($raw_interface =~ /^Interface\s+(.+)/) {
				# Interface names can be in two formats [slot-port] or just [port].
				$modelled_interface->{name} = $1;

				# Add to the 'any' security zone for easy referenceing of all segment interfaces
				$modelled_interface->{interfaceGrouping} = {
					ifGroupName => 'any',
					ifGroupingCharacteristic => 'securityZoneMember'
				};

				# Keep track of the names of parsed interfaces
				$in->{interface_names}->{$1} = 1;
			}

			# Common values		
			if ($raw_interface =~ /MTU\s+(\d+)/) {
				$modelled_interface->{mtu} = $1;
			}
			if ($raw_interface =~ /Speed\s+(\d+)(\D)/) {
				my $speed = getUnitFreeNumber($1, $2, 'bit');
				$modelled_interface->{speed} = $speed;
			}
			if ($raw_interface =~ /Link\s+(\w+)/) {
				my $adminStatus = lc($1);
				$modelled_interface->{adminStatus} = $adminStatus;
			}

			push(@interfaces, $modelled_interface);
		}

		$in->{parsed_interfaces} = \@interfaces;
		$in->{interface_names}->{any} = 1;
		delete ($in->{interfaces}); # memory management
	}

	if( $out ) {
		$out->print_element('interfaces', {interface => \@interfaces});
	}

	return( 1 );
}

sub parse_sms {
	my $responses = shift;
	my $ip_regex = get_crep('cipm');

	my $sms_ip;
	if ( $responses->{sms} =~ /SMS IP:\s($ip_regex)/ ) {
		$sms_ip = $1;
	} 

	return( $sms_ip );
}

sub parse_ips_rules {
	my ( $in, $out ) = @_;
	
	# parse DV recommended actions per deployment type
	my $recommendedActions = {};
	my @actionFileList = qw( default core edge perimeter aggressive hyper-aggressive );
	foreach my $file (@actionFileList){
		parse_recommended_actions_file ($recommendedActions, $file, $LOGGER);
	}

	my $ports = parse_virtual_ports ($in, $LOGGER);

	my $profile_order = parse_profiles ($in, $ports, $LOGGER);

	my $action_sets = parse_action_sets ($in, $LOGGER);

	my $profiles = create_profiles_category_action_set_mapping ($in, $action_sets, $LOGGER);
	
	# add profile deployment type
	find_profile_deployment_type ($profiles, $in);

	# TODO: protection-settings: Addresses or objects the IPS rules only apply to
	# Does this matter to QRM for vuln path searching?

	my @signatures = @{ _parse_digital_vaccine($in) };
	$LOGGER->debug('Found ' . scalar( @signatures ) . ' filters in the Digital Vaccine');

	# Get filter for each signature in the digital vaccine
	my $callback_responses;
	foreach my $signature ( @signatures ) {
		my $filter_number = $signature->{num};
		$callback_responses->{$filter_number} = $in->{get_filter_callback}->($filter_number);
	}

	foreach my $profile_key ( keys( %{ $profiles } ) ) {
		my $profile = $profiles->{$profile_key};

		foreach my $signature ( @signatures ) {

			my $filter = {
				name => $signature->{name},
				cve => $signature->{cve_id},
				protocol => $signature->{protocol},
				number => $signature->{num},
			};

			# Get the signature cateogry and map to equivalent
			my $filter_category = lc( $signature->{class} );
			$filter_category =~ s/_/-/g;
			$filter_category =~ s/ anomaly/-normal/;
			$filter_category =~ s/vulnerability/vulnerabilities/;
			$filter_category =~ s/equip/equipment/;
			$filter_category =~ s/exploit/exploits/;
			$filter_category =~ s/policy/security-policy/;
			
			if( $profile->{categories}->{$filter_category} ) {
				$filter->{action} = $profile->{categories}->{$filter_category}->{action};
				$filter->{log} = $profile->{categories}->{$filter_category}->{log};
			}

			# Find enabled filters
			if( $callback_responses->{ $signature->{num} } ) {
				my $filter_response = $callback_responses->{ $signature->{num} };
				parse_filter_response ($signature, $filter, $filter_response, $profile_key, $profile, $recommendedActions);
			}
		}
	}

	my $filter_lists_opened;
	foreach my $profile_segment (sort {$a <=> $b} keys( %{$profile_order} ) ) {
		my $name = $profile_order->{$profile_segment}->{name};
		my $profile = $profiles->{$name};

		$name .= ':' . $profile_order->{$profile_segment}->{segment};

		$profile->{applies_to} = $profile_order->{$profile_segment}->{applies_to};

		unless( exists( $profile->{vulnerabilities} ) && defined( $profile->{applies_to} ) ) {
			next;
		}

		if( !$filter_lists_opened ) {
			$out->open_element('filterLists');
			$filter_lists_opened = 1;
		}

		my $count = 0;

		$out->open_element('filterList');
		$out->print_element( 'defaultAction', 'none' );

		foreach my $applies_to ( @{ $profile->{applies_to} } ) {
			$out->open_element( 'appliesTo' );

			foreach my $direction ( 'source', 'destination' ) {
				if( $applies_to->{$direction} ) {
					$out->print_element( $direction, $applies_to->{$direction} );
				}

				my $direction_group = $direction . 'Group';
				if( $applies_to->{$direction_group } ) {
					$out->print_element( $direction_group, $applies_to->{$direction_group} );
				}
			}

			$out->close_element( 'appliesTo' );
		}

		foreach my $vuln_key (sort {$a <=> $b} keys( %{ $profile->{vulnerabilities} } ) ) {
			my $vuln = $profile->{vulnerabilities}->{$vuln_key};

			if( $vuln->{cve} ) {
				$out->open_element('filterEntry');

				my $log = $vuln->{log} || 'false';
				$out->print_element('log', $log);

				my $vuln_name;
				if( $vuln->{name} && $vuln->{name} ne '' ) {
					$vuln_name = $vuln->{name};
					$out->print_element('name', $vuln_name);
				}

				my $action = $vuln->{action} || 'drop';
				$out->print_element('primaryAction', $action);
				$out->print_element('processOrder', $vuln->{number});

				if( $vuln->{protocol} && $vuln->{protocol} ne '' ) {
					if( $vuln->{protocol} =~ /^(tcp|udp)$/i ) {
						$out->print_element( 'protocol', $vuln->{protocol} );
					}
				}

				foreach my $cve ( split( /,/, $vuln->{cve} ) ) {
					$out->print_element('vulnerability', {
						name => $cve,
						type => 'QRM_CVE',
					});
				}

				if( $vuln_name && $vuln_name =~ /^\d+:\s(.+?)\s*$/ ) {
					$out->print_element('ztSignature', $1);
				}

				$out->close_element('filterEntry');
			} else {
				$LOGGER->debug('No CVE ID for \'' . $vuln->{name} . '\', skipping');
			}
		}

		$out->print_element('mode', 'stateful');
		$out->print_element('name', $name);
		$out->close_element('filterList');
	}

	if( $filter_lists_opened ) {
		$out->close_element('filterLists');
	}

	return( 1 );
}

sub _parse_digital_vaccine {
	my ($in) = @_;
	my @signatures;

	if( $in->{signature} ) {
		@signatures = @{ _parse_xml_to_array_of_hashes( $in->{signature} ) };

		my @cve_signatures;
		foreach my $signature ( @signatures ) { 
			if( $signature->{cve_id} ) {
				push( @cve_signatures, $signature );
			}
		}

		@signatures = @cve_signatures;
	} else {
		$LOGGER->fatal( 'Digital Vaccine needed and missing from responses' );
	}

	return \@signatures;
}

# The sms returns database entries as xml
# This subroutine returns each database row as an element in an array
# Each element is a hash, where keys are the database column names
#
# OPTIONAL ARG: $columns_wanted is for specifying what columns we want to put in the hash
#		$columns_wanted should be a hash reference with keys that point to a TRUE value
#		e.g. if we only wanted the name and ip_address from a table:
#			$columns_wanted = { name => 1, ip_address => 1 };
sub _parse_xml_to_array_of_hashes {
	my ($xml, $columns_wanted) = @_;
	my $hash;

	my $table_name = $xml->findnodes('resultset/table')->[0]->att('name');
	$LOGGER->debug("Parsing database table $table_name from xml");

	my $start_time = time();

	# Get each of the columns names in order
	my @column_names;
	foreach my $column ( $xml->findnodes('resultset/table/column') ) {
		my $column_name = lc( $column->att('name') );
		push( @column_names, $column_name );
		$LOGGER->debug("Found column $column_name in $table_name");
	}

	# Add each row to the array, with each row as a hash
	my @data;
	foreach my $row ( $xml->findnodes('resultset/table/data/r') ) {
		my $datum;

		my @columns = $row->findnodes('c');
		for( my $i = 0; $i < scalar( @columns ); ++$i ) {

			# Only collect columns specified if we have specified we want a subset
			if( !defined( $columns_wanted ) || $columns_wanted->{$column_names[$i]} ) {
				$datum->{$column_names[$i]} = _get_value_from_xml( '.', $columns[$i] );
			}
		}

		if( $datum ) {
			push( @data, $datum );
		}
	}

	my $time = time() - $start_time;
	$LOGGER->debug("Parsing took $time seconds for table $table_name");

	return( \@data );
}

sub _get_value_from_xml {
	my ($path, $xml) = @_;

	my $value = $xml->findvalue($path);
	if ( ref($value) eq 'XML::XPath::Literal' ) {
		$value = $value->value();
	}

	if ( $value eq '' ) {
		$value = undef;
	}

	return( $value );
}

sub parse_filter_response {
	my $signature       = shift;
	my $filter          = shift;
	my $filter_response = shift;
	my $profile_key     = shift;
	my $profile = shift;
	my $recommendedActions = shift;
	
	my $filterEnabled = 1;
	my $profileDeployment = lc $profile->{deployment};
	my $filterNumber = int $filter->{number};
	foreach my $line (split ("\n", $filter_response)){
		if ($line =~ /\s*(.*\S)\s+Config:\s+(Enabled|by-deployment|Disabled)\s+AFC:\s+\S+\s+(Category|Override):\s+(.+\n?\S+)/){
			my ( $profile_name, $config, $override, $category ) = ( $1, $2, lc($3), lc($4) );
			
			unless ( $profile_name eq $profile_key || $profile_name eq "(default policy)" && (defined $filter->{action} && $filter->{action} eq "Recommended")) {
				next;
			}
			
			if ($config eq "Disabled"){
				$filterEnabled = 0;
			}elsif ($config eq "by-deployment" && (defined $filter->{action} && $filter->{action} eq "Recommended")){
				# lookup recommended actions
				unless (exists $recommendedActions->{$profileDeployment}->{$filterNumber}){
					$filterEnabled = 0; # if filter doesn't exist in recommended actions it must be disabled for this deployment
				}
			}
			
			if ($override eq 'override' ) { # look for custom filter actions "override"
				if( $category =~ /blo/ ) {
					$filter->{action} = "deny";
				}elsif( $category =~ /per|tru/ ) {
					$filter->{action} = "permit";
				}else{
					next;
				}
				$filterEnabled = 1;
				last;
			}
		}
	}
	if ($filterEnabled && defined $filter->{action} && $filter->{action} eq "Recommended"){
		# lookup action
		$filter->{action} = $recommendedActions->{$profileDeployment}->{$filterNumber};
	}
	
	if ($filterEnabled){
		$profile->{vulnerabilities}->{ $signature->{num} } = $filter;
	}
}

sub parse_virtual_ports {
	my $in     = shift;
	my $LOGGER = shift;

	my $ports;
	while ( $in->{config} =~ /^\s*virtual-port\s+"([^"]+)"\s+add-row\s+-port-list\s+"([^"]+)"\s+-vlan-list\s+"([^"]+)"\s+-cidr-list\s+"([^"]+)"\s*$/img) {
		my ( $name, $port, $vlan, $cidr ) = ( $1, $2, $3, $4 );
		$LOGGER->debug("Parsing virtual port: $&");

		foreach my $single_port ( split( /\s*,\s*/, $port ) ) {
			if ( $in->{interface_names}->{$single_port} ) {
				push( @{ $ports->{$name}->{interface} }, $single_port );
			}
			else {
				$LOGGER->debug("Virtual port '$name' refrences an unknown interface '$single_port'"
				);
				next;
			}
		}

		# We do not use vlan or cidr, no schema support
		if (
			$vlan ne 'any' ) {
			$ports->{$name}->{vlan} = $vlan;
		}

		if( $cidr ne 'any' ) {
			foreach my $single_cidr ( split( /\s*,\s*/, $cidr ) ) {
				push( @{ $ports->{$name}->{cidr} }, $cidr );
			}
		}
	}
	return $ports;
}

sub parse_profiles {
	my $in = shift;
	my $ports         = shift;
	my $LOGGER        = shift;
	
	my $profile_order;
	while ( $in->{config} =~ /^\s*profile\s+"([^"]+)"\s+add-pair\s+"([^"]+)"\s+"([^"]+)"\s*$/img )
	{
		my ( $profile, $ingress, $egress ) = ( $1, $2, $3 );
		$LOGGER->debug("Parsing profile: $&");

		# Create applies to section of each profile
		my @applies_to;
		if ( $ports->{$ingress} && $ports->{$egress} ) {
			foreach my $source ( @{ $ports->{$ingress}->{interface} } ) {
				foreach my $destination ( @{ $ports->{$egress}->{interface} } )
				{
					my $applies_to;

					if (
						$source eq 'any' ) {
						$applies_to->{sourceGroup} = $source;
					} else {
						$applies_to->{source} = $source;
					}

					if( $destination eq 'any' ) {
						$applies_to->{destinationGroup} = $destination;
					} else {
						$applies_to->{destination} = $destination;
					}

					push( @applies_to, $applies_to );
				}
			}
		}

		if( @applies_to ) {

			$LOGGER->debug( "Finding order of virtual segment $ingress -> $egress" );
			if( $in->{config} =~ /^\s*virtual-segment\s+"$ingress"\s+"$egress"\s+position\s+(\d+)\s*$/im ) {
				$profile_order->{$1} = {
					name => $profile,
					applies_to => \@applies_to,
					segment => $ingress.'to'.$egress
				};
			}
		}
	}
	return $profile_order;
}


sub parse_action_sets {
	my $in     = shift;
	my $LOGGER = shift;

	my $action_sets = {
		Recommended => {
			action => 'Recommended',
			log => 'true'
		}
	};

	# Create action sets with QRM equivalent action
	while( $in->{config} =~ /^\s*action-set\s+"([^"]+)"\s+(permit|block|trust)\s*$/img ) {
		my ($name, $action) = ($1, $2);
		$LOGGER->debug("Parsing action-set: $&");

		if( $action eq 'block' ) {
			$action = 'deny';
		} elsif( $action eq 'trust' ) {
			$action = 'permit';
		}

		$action_sets->{$name} = {
				action => $action,
				log => 'false'
		};
	}
	
	# Find if the action set is loggged to some degree
	while( $in->{config} =~ /^\s*action-set\s+"([^"]+)"\s+notify-contact\s+add\s+"[^"]+"\s*$/img){
		my $name = $1;
		$action_sets->{$name}->{log} = 'true';
	}
	
    return $action_sets;
}

sub create_profiles_category_action_set_mapping {
	my $in          = shift;
	my $action_sets = shift;
	my $LOGGER      = shift;

	my $profiles;
	while ( $in->{config} =~ /^\s*category-settings\s+-profile\s+"([^"]+)"\s+(\S+)\s+enable\s+-\s*action-set\s+"([^"]+)"\s*$/img){
		my ( $profile, $category, $action_set ) = ( $1, $2, $3 );
		$LOGGER->debug("Parsing categories: $&");

		my $action = $action_sets->{$action_set}->{action};
		$profiles->{$profile}->{categories}->{$category}->{action} = $action;

		my $log = $action_sets->{$action_set}->{log};
		$profiles->{$profile}->{categories}->{$category}->{log} = $log;
	}
	return $profiles;
}

sub find_profile_deployment_type {
	my $profiles = shift;
	my $in       = shift;

	while ( $in->{config} =~ /^\s*profile\s+"([^"]+)"\s+deployment\s+\"(\S+)\"$/img )
	{
		my ( $profileName, $deploymentType ) = ( $1, $2 );
		if ( exists $profiles->{$profileName} ) {
			$profiles->{$profileName}->{deployment} = $deploymentType;
		}
	}
}

sub parse_recommended_actions_file {
	my $recommendedActions = shift;
	my $deployment = shift;
	my $LOGGER           = shift;
	
	# get package's path 
	( my $pkgDir = __PACKAGE__ ) =~ s!::!/!g;
	my ($pkgPath) = $INC{"$pkgDir.pm"} =~ m!^(.+)/!;

	my $actionsFile = "$pkgPath/../../../../../dv-actions/$deployment.txt";
	open ACTIONSFILE, '<', $actionsFile or $LOGGER->fatal("Can't open 'Recommended' actions file $actionsFile!");
	while (<ACTIONSFILE>){
		my $line = $_;
		unless ($line =~ /CVE-/){
			next; # we ignore filters with no CVEs
		}
		if ($line =~ /(^\d+)/){
			my $filterId = int($1);
			if ($line =~ /Block \/ Notify/i){
				$recommendedActions->{$deployment}->{$filterId} = "deny";
			}
		}
	}
	close ACTIONSFILE or $LOGGER->fatal("Can't close 'Recommended' actions file $actionsFile!");
}
