package Credentials;

use strict;

use Switch;
use XML::Simple;

use CliHelpers qw( get_number get_ip get_yes_no get_credential_string );

sub new()
{
	my $className = shift;
	
	my $self = {
		host_ip  => "127.0.0.1",
		protocol => undef,
		username => undef,
		password => undef,
		enableUsername => undef,
		enablePassword => undef,
		snmpCommunity => undef,
		snmpUsername => undef,
		snmpAdminPass => undef,
		snmpPrivPass => undef,
		valid_protocols => undef,
	};
	
	bless( $self, $className );
	
	my @valid_protocols = ('SSH','Telnet','FTP','SFTP','SCP','SNMP','HTTP','HTTPS');
	
	$self->{valid_protocols} = \@valid_protocols;
	
	return $self;
}

sub from_xml
{
	my $filename = shift;
	
	my $self = Credentials->new();

	open(INPUT_XML,"<$filename");
	my $input_xml_content = "";
	while(<INPUT_XML>)
	{
		$input_xml_content .= $_;
	}
	close(INPUT_XML);

	if( $input_xml_content =~ /.*protocol name=\"(.*?)\".*/g )
	{
		$self->{protocol} = $1;
	}
	if( $input_xml_content =~ /.* host=\"(\d+\.\d+\.\d+\.\d+)\".*/g )
	{
		$self->{host_ip} = $1;
	}
	while( $input_xml_content =~ /.*credential name=\"(.*?)\" value=\"(.*?)\".*/g )
	{
		switch($1)
		{
			case "roCommunityString" { $self->{snmp_community}  = $2 }
			case "snmpPrivPassword"  { $self->{snmp_privacy}   = $2 }
			case "enableUsername"    { $self->{enable_user} = $2 }
			case "snmpAuthPassword"  { $self->{snmp_password}  = $2 }
			case "enablePassword"    { $self->{enable_password} = $2 }
			case "snmpUsername"      { $self->{snmp_username}   = $2 }
			case "password"          { $self->{password}       = $2 }
			case "username"          { $self->{username}       = $2 }
			else {}
		}
	}
	
	return $self;
}

sub ask_for_protocol
{
	my $self = shift;
	my $valid_protocols = $self->{valid_protocols};
	my $protocol_counter = 0;
	foreach my $valid_protocol (@$valid_protocols)
	{
		print $protocol_counter++."\t".$valid_protocol."\n";
	}
	my $protocol_index = get_number("Choose the protocol by number",0,scalar(@$valid_protocols)-1);
	$self->{protocol} = @$valid_protocols[$protocol_index];	
	print " --> ".$self->{protocol}."\n";
}

sub ask_for_credential_set
{
	my $self = shift;
	
	$self->ask_for_protocol();
	
	$self->{username}        = get_credential_string("Username");
	$self->{password}        = get_credential_string("Password");
	$self->{enable_user}     = get_credential_string("Enable Username");
	$self->{enable_password} = get_credential_string("Enable Password");
	$self->{snmp_community}  = get_credential_string("SNMP Community");
	$self->{snmp_username}   = get_credential_string("SNMP Username");
	$self->{snmp_password}   = get_credential_string("SNMP Auth Password");
	$self->{snmp_privacy}    = get_credential_string("SNMP Privacy Password");
}

sub get_host_ip
{
	my $self = shift;
	
	return $self->{host_ip};
}

sub set_host_ip
{
	my $self = shift;
	
	$self->{host_ip} = shift;
}

sub get_credential_xml
{
	my $self = shift;

	my $connection_string = <<__CONNECTION_STRING__;
<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<operationInputXML>
	<connectionPath host="$self->{host_ip}">
		<protocols>
			<protocol name="$self->{protocol}" port="-1" validatedOnDevice="false">
				<properties/>
			</protocol>
		</protocols>
		<credentials>
			<credential name="roCommunityString" value="$self->{snmp_community}"/>
			<credential name="snmpPrivPassword" value="$self->{snmp_privacy}"/>
			<credential name="enableUsername" value="$self->{enable_user}"/>
			<credential name="snmpAuthPassword" value="$self->{snmp_password}"/>
			<credential name="enablePassword" value="$self->{enable_password}"/>
			<credential name="snmpUsername" value="$self->{snmp_username}"/>
			<credential name="password" value="$self->{password}"/>
			<credential name="username" value="$self->{username}"/>
		</credentials>
	</connectionPath>
</operationInputXML>
__CONNECTION_STRING__

	return $connection_string;
}

sub store_credential_xml
{
	my $self = shift;
	my $filename = shift;

	if( defined($filename) && length($filename) > 0 )
	{	
		open(INPUT_XML,">$filename");
		print INPUT_XML $self->get_credential_xml();
		close(INPUT_XML);
		chmod 0600, "$filename";
		return 1;
	}
	return 0;
}

sub validate_xml
{
	my $self = shift;
	my $xml_content = shift;
	
	my $result = 1;
	
	my $obj = XMLin($xml_content, ForceArray=>['protocol','credential'], KeyAttr=>[]);
	
	if( not defined $obj->{connectionPath} or not defined $obj->{connectionPath}->{host})
	{
		print "invalid connection XML, element 'connectionPath' must be present, and must have a host attribute\n";
		$result = 0;
	}
	else
	{
		if( not defined $obj->{connectionPath}->{protocols} or not defined $obj->{connectionPath}->{protocols}->{protocol})
		{
			print "invalid connection XML, no protocols defined\n";
			$result = 0;
		}
		else
		{
			foreach my $conn_proto (@{$obj->{connectionPath}->{protocols}->{protocol}})
			{
				my $matched = 0;
				foreach my $proto (@{$self->{valid_protocols}})
				{
					if( defined $conn_proto->{name} and $proto eq $conn_proto->{name} )
					{
						$matched = 1;
					}
				}
				if( not $matched )
				{
					$result = 0;
					print "invalid connection XML, \"".(defined $conn_proto->{name}?$conn_proto->{name}:"")."\" is not a valid protocol name\n";
				}
			}
		}
		if( not defined $obj->{connectionPath}->{credentials} or not defined $obj->{connectionPath}->{credentials}->{credential})
		{
			print "invalid connection XML, no credentials defined\n";
			$result = 0;
		}
	}
	return $result;
}

1;
