package ZipTie::Adapters::Unix::GetConfig;

use strict;

use File::Temp;
use ZipTie::Adapters::Utils qw(create_empty_file create_unique_filename escape_filename);
use ZipTie::Logger;
use ZipTie::Recording;
use ZipTie::TransferProtocol;
use ZipTie::TransferProtocolFactory;

use Exporter 'import';

our @EXPORT_OK = qw(get_config);
my $LOGGER = ZipTie::Logger::get_logger();

my $RECORDING = ZipTie::Recording::get_recording();

sub get_config
{
    my $cli_protocol = shift or die;
    my $connection_path = shift or die;
    my $path = shift or die;
    my $scp_needs_root = shift;

    my $scp_protocol = $connection_path->get_protocol_by_name('SCP');
    if (defined($scp_protocol))
    {
	return _get_config_scp($connection_path, $scp_protocol, $path, $scp_needs_root);
    }

    # Fall back to CLI protocol only.
    return _get_config_cli($cli_protocol, $path);
}

sub _get_config_scp
{
    my ($connection_path, $scp_protocol, $config_name, $needs_enable) = @_;

    my $user;
    my $password;
    if ($needs_enable)
    {
        # If only root has permission to see the config.
        $user = $connection_path->get_credential_by_name('enableUsername') || 'root';
        $password =$connection_path->get_credential_by_name('enablePassword'); 
    }
    else
    {
        $user = $connection_path->get_credential_by_name('username');
        $password = $connection_path->get_credential_by_name('password');
    }

    my $xfer_client = ZipTie::TransferProtocolFactory::create($scp_protocol->get_name());
    $xfer_client->connect(
    	$connection_path->get_ip_address(),
        $scp_protocol->get_port(),
        $user,
	    $password,
    );

    my $temp_file = create_unique_filename();
    $xfer_client->get($config_name, $temp_file);
    
    # Open up the configuration file and read it into memory
    open(CONFIG, $temp_file) || $LOGGER->fatal("[$SCP_ERROR]\nCould not open the retrieved configuration file stored in '$temp_file'");
    my @entire_file = <CONFIG>;
    close(CONFIG);
    my $config_contents = join( "", @entire_file );
    unlink ($temp_file) if (-e $temp_file); # delete the temp file 

    # Record the file transfer of the config
    # Arguments: protocol name, file name, response/contents, whether or not ZipTie acted as the file transfer server
    $RECORDING->create_xfer_interaction($scp_protocol->get_name(), $config_name, $config_contents, 0);

    # Return the contents of the configuration
    return $config_contents;
}

sub _get_config_cli
{
    my ($cli_protocol, $path ) = @_;

    my $prompt = $cli_protocol->get_prompt_by_name('prompt');
    my $cmd = 'cat ' . $path;
    my $response = $cli_protocol->send_and_wait_for($cmd, $prompt, 300);

    #remove the prompt from the response
    $response =~ s/$prompt$//;
    $response =~ s/^$cmd\n//s;

    return $response;
}

1;
