package ZipTie::Adapters::Unix::ParsingUtils;

use strict;
use warnings;

use ZipTie::Adapters::Utils qw(get_interface_type mask_to_bits);

sub parse_last_reboot
{
	my ( $uptime, $out ) = @_;
	if ( $uptime =~ /\bup\s+(.+)$/mi )
	{
		$_ = $1;
		my ($years)   = /(\d+)\s*years?/;
		my ($weeks)   = /(\d+)\s*weeks?/;
		my ($days)    = /(\d+)\s*days?/;
		my ($hours)   = /(\d+)\s*hours?/;
		my ($minutes) = /(\d+)\s*minutes?/;

		if ( !defined $hours )
		{
			my $temp = $_;
			if ( $temp =~ /(\d+)\:(\d+)/ )
			{
				$hours   = $1;
				$minutes = $2;
			}
		}

		# subract the last reboot from the current time
		my $lastReboot = time();
		$lastReboot -= $years * 52 * 7 * 24 * 60 * 60 if ($years);
		$lastReboot -= $weeks * 7 * 24 * 60 * 60      if ($weeks);
		$lastReboot -= $days * 24 * 60 * 60           if ($days);
		$lastReboot -= $hours * 60 * 60               if ($hours);
		$lastReboot -= $minutes * 60                  if ($minutes);
		$out->print_element( "core:lastReboot", $lastReboot );
	}
}

sub parse_local_accounts
{
    my ( $users, $out ) = @_;
    $out->open_element("localAccounts");

    while ( ( $users =~ /^([^\s:]+):([^:]+):(\d+):(\d+):([^:]*):([^:]+):([^:]+)$/mig ) )
    {
        my $account = {
            accountName => $1,
            accessLevel => $4,
            password    => $2,
            fullName    => $5
        };
        $out->print_element( "localAccount", $account );
    }

    $out->close_element("localAccounts");
}


sub parse_snmp
{
    my ( $in, $out ) = @_;

    $out->open_element("snmp");
    my @communities = ();
    my @traps       = ();

    # parse communities and trap hosts
    while ( $in->{snmp} =~ /^(.+)$/mig )
    {
        my $commline = $1;
        if ( $commline =~ /^(r[ow])community\S*\s+(\S+)\s+\S+\s*(\S*)$/i )
        {
            push( @communities, { accessType => uc($1), communityString => $2, mibView => $4 } );
        }
        elsif ( $commline =~ /^trapsess.+\-c\s+(\S+)\s+(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}).+$/i )
        {
            push( @traps, { ipAddress => $2, communityString => $1 } );
        }
    }

    # parse other snmp data and store the configuration
    foreach (@communities)
    {
        $out->print_element( "community", $_ );
    }

    my $someSysPrinted = 0;
    if ( $in->{snmp} =~ /^syscontact\s+(.+)$/mi )
    {
        $out->print_element( "sysContact", $1 );
        $someSysPrinted = 1;
    }

    if ( $in->{snmp} =~ /^syslocation\s+(.+)$/mi )
    {
        $out->print_element( "sysLocation", $1 );
        $someSysPrinted = 1;
    }

    if ( $in->{hostname} =~ /^hostname\s*(\S+)$/mi )
    {
        $out->print_element( "sysName", $1 );
        $someSysPrinted = 1;
    }

    if ($someSysPrinted)
    {
        foreach (@traps)
        {
            $out->print_element( "trapHosts", $_ );
        }
    }

    $out->close_element("snmp");
}

sub parse_static_routes
{
    my ( $in, $out ) = @_;
    my $staticRoutes;

    while ( $in->{static_routes} =~ /^(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|default)\s+(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\s+\S+\s+\d+\s+(\S+)/mig )
    {
        my $defaultGw = 'false';
        my $destIP    = $1;
        my $destMask  = '32';
        my $gateway   = $2;
        my $iface     = $3;

        if ( lc($destIP) eq 'default' )
        {
            $defaultGw = 'true';
            $destIP    = '0.0.0.0';
        }

        my $route = {
            defaultGateway     => $defaultGw,
            destinationAddress => $destIP,
            destinationMask    => $destMask,
            gatewayAddress     => $gateway,
            interface          => $iface
        };

        push( @{ $staticRoutes->{staticRoute} }, $route );
    }

    $out->print_element( 'staticRoutes', $staticRoutes ) if ( defined $staticRoutes );
}

sub parse_df_output
{
    # populate the deviceStorage elements of the chassis
    my ( $df, $out ) = @_;

    while ( $df =~ /^(\S+)\s+(?:\d+)\s+(\d+)\s+(\d+)\s+(\d+)%\s+\S.*$/mig )
    {
        my $prt_name   = $1;
        my $total_size = $2 + $3;
        my $storage    = {
            name        => $prt_name,
            storageType => 'disk',
            size        => $total_size,
        };

        # we won't populate FileDirectory node
        $out->print_element( "deviceStorage", $storage );
    }
    
}

sub parse_interfaces
{
    my ($in, $out, $extensions) = @_;
    my $interface_data;

    $out->open_element('interfaces');
    while ( $in->{interfaces} =~ /^(\S+)\s+Link(.+?)(?=^\s*$)/msg )
    {
        my $name = $1;
        my $blob = $2;

        my $interfaceType = get_interface_type($1);
        if ($interfaceType eq 'other' and $blob =~ /^\s*encap:Ethernet/)
        {
            $interfaceType = 'ethernet';
        }

        my $interface = {
            name          => $name,
            adminStatus   => 'down',
            interfaceType => $interfaceType,
        };
        if ( $blob =~ /^\s*UP/m )
        {
            $interface->{adminStatus} = 'up';
        }
        $interface->{physical} = ( $interface->{interfaceType} eq 'softwareLoopback' ) ? 'false' : 'true';
        ( $interface->{mtu} ) = $blob =~ /MTU:(\d+)/;
        if ( $blob =~ /HWaddr\s+([a-z\d:-]+)/i )
        {
            my $mac = $1;
            $mac =~ s/[:-]//g;

            unless ($mac =~ /^0+$/)
            {
                $interface->{interfaceEthernet}->{macAddress} = $mac;
            }
        }
	
	    my @local_ip_addresses;
        while ( $blob =~ /^\s+inet\saddr:(\S+)\s+(.+)/mg )
        {
            my $ipLine = $2;
            my $ipConfiguration = { ipAddress => $1, };
            if ( $ipLine =~ /Mask:(\S+)/i )
            {
                $ipConfiguration->{mask} = mask_to_bits($1);
            }
            if ( $ipLine =~ /Bcast:(\S+)/i )
            {
                $ipConfiguration->{broadcast} = $1;
            }
            push( @{ $interface->{interfaceIp}->{ipConfiguration} }, $ipConfiguration );
            push( @local_ip_addresses, $ipConfiguration );
        }
        
        $interface_data = {$name => \@local_ip_addresses};

        if ($extensions && defined $extensions->{$interface->{name}})
        {
            my %exts = %{$extensions->{$interface->{name}}};
            while (my ($key, $value) = each %exts)
            {
                $interface->{$key} = $value;
            }
        }

        $out->print_element( 'interface', $interface );
    }
    $in->{local_ip_addresses} = $interface_data;
    $out->close_element('interfaces');
}
1;

__END__

=head1 NAME

ZipTie::Adapters::Unix::ParsingUtils

=head1 SYNOPSIS

    use ZipTie::Adapters::Unix::ParsingUtils qw(parse_ifconfig);
	parse_ifconfig($responses->{ifconfig}, $printer);

=head1 DESCRIPTION

Offers parsing mechanisms to turn common Unix/Linux command output into the ZED XML format.

Results of each method are printed using the $printer arguement.

=head1 METHODS

=over 12

=item C<parse_last_reboot($uptime, $printer)>

Takes the output of an 'uptime' command and prints the core:lastReboot element
of the ZED.

=back

=head1 LICENSE

 The contents of this file are subject to the Mozilla Public License
 Version 1.1 (the "License"); you may not use this file except in
 compliance with the License. You may obtain a copy of the License at
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS"
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 License for the specific language governing rights and limitations
 under the License.

=head1 AUTHOR

  Contributor(s): rkruse
  Date: Sep 17, 2008

=cut
