package ZipTie::Addressing::RouteFilter;

use strict;
use warnings;
use v5.10;
use Net::CIDR::Set;

use constant HOST_PREFIX_LENGTH => 32;

sub new
{
    my ($className, $defaultPassOrDeny) = @_;

    my $this = 
    {
		defaultDisp => $defaultPassOrDeny,
		passFilterRouteSet => Net::CIDR::Set->new(),
		denyFilterRouteSet => Net::CIDR::Set->new()
    };

    bless ($this, $className);
    return $this;
}

sub addConstraint
{
    my ($this, $ip, $prefixLen, $passOrDeny) = @_;

    if ($passOrDeny) # pass
    {
		$this->{passFilterRouteSet}->add(($ip . "/" . $prefixLen));
    }
    else
    {
		$this->{denyFilterRouteSet}->add(($ip . "/" . $prefixLen));
    }
}

sub contains
{
    my ($this, $host) = @_;
    my $hostNet = $host . "/" . HOST_PREFIX_LENGTH;


	unless ($this->{denyFilterRouteSet}->is_empty)
	{
	    return 0 if ($this->{denyFilterRouteSet}->contains($hostNet));	
	}

	unless ($this->{passFilterRouteSet}->is_empty)
	{
	    return 1 if ($this->{passFilterRouteSet}->contains($hostNet));
	}
    return -1;
}

sub exerciseFilterOn 
{
    my ($this, $ip, $prefixLen) = @_;
    my $ipNet = $ip . "/" . $prefixLen;
    
	unless ($this->{denyFilterRouteSet}->is_empty)
	{
    	return 0 if ($this->{denyFilterRouteSet}->contains($ipNet));
    	
	}
	
	unless ($this->{passFilterRouteSet}->is_empty)
	{
    	return 1 if ($this->{passFilterRouteSet}->contains($ipNet));
	}
	
    return $this->{defaultDisp};
}

1;

=head1 DESCRIPTION

Object to aggregate multiple routes, conditions governing their inclusion, and then to be able to individually test another route as to wiether it would pass the tests of the "filter.

As it stands now, this is a simple reflection of constraints as destination
nets.

Requires Net::CIDR::Set.

=head1 CONSTRUCTOR

=over 12

=item C<new($defaultPassOrDeny)>

Builds a ZipTie::Addressing::RouteFilter, with a "default disposition
as indicated.  If none of the subsequently added constraint nets don't
match, then this is the forward-or-drop indication that will be exposed
from the routine.  Like other such pass-or-deny indications, 0 indicates
deny/drop, nonzero indicates pass/forward.

=head1 METHODS

=over 12

=item C<addConstraint( $ip, $prefixLengrth, $passOrDeny)>

Adds a network to the set which are being tested.  The tests are applied 
iteratively relative to the nets with respect to the sequence being established
by consecutive calls to addConstraint.

=item C<contains($host)>

Returns nonzero if the host would be passed based on explicit rules, zero
if it would be denied, and -1 if there would be no explicit route provision
for the route based upon the rules added to the filter.

=item C<exerciseFilterOn($ip, $prefix?Length)>

Returns a pass-or-drop indication on the overall route pass-or-drop based upon
the overall filter rules and defaulting logic.

=back

=head1 LICENSE

 The contents of this file are subject to the Mozilla Public License
 Version 1.1 (the "License"); you may not use this file except in
 compliance with the License. You may obtain a copy of the License at
 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS"
 basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
 License for the specific language governing rights and limitations
 under the License.

=head1 AUTHOR

  Contributor(s): Wayne Tackabury <wtackabury@us.ibm.com>
  Date: March 15, 2013

=cut
