package ZipTie::ApplicationsClient;

=pod

=head1 NAME

ZipTie::ApplicationsClient -- Provide applicaiton information to ziptie-server.

=head1 DESCRIPTION

This class acts as a layer between an adapter and ziptie-server's applicaitons API.  It's used to
collect all application information and then provide it to ziptie-server through the API.

In the future this class could optionally be used to write out a ziptie-applications.xsd XML
document and provide the application data to the Generic XML adapter to import the data.  That
would allow applicaiton data to be used with the disconnected configuration collection toolkit.

=head1 METHODS

=cut

use strict;
use warnings;

use ZipTie::Client;
use ZipTie::Logger;
use Carp;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

=pod

B<new()> (constructor)

Returns a new ZipTie::ApplicationsClient object.

=cut
sub new
{
	my ($class) = @_;

	my $self = {applications => []};

	bless($self, $class);

	return $self;
}

=pod

B<add_application($application)>

Add a new ZipTie::ApplicationsClient::Application object to the list of appliactions to be provided
to ziptie-server.

$application - The ZipTie::ApplicationsClient::Application object to add to the list of
applications.

=cut
sub add_application
{
	my ($self, $application) = @_;

	push(@{$self->{applications}}, $application->get_soap_object());
}

=pod

B<commit()>

Once all of the applications have been added using the add_application() method the applications
must be committed to ziptie-server using this method.

If no applications have been added then the API will not be called.

=cut
sub commit
{
	my ($self) = @_;

	my $numberOfApplications = scalar(@{$self->{applications}});

	# No need to call the API if no applications have been added to the client.
	if (!$numberOfApplications)
	{
		$LOGGER->debug('ZipTie::ApplicationsClient->commit() called with 0 applications, not calling API.');
		return;
	}
	else
	{
		$LOGGER->debug("ZipTie::Applications->commit() called with $numberOfApplications, calling API.");
	}

	if (!defined($ENV{ZIPTIE_AUTHENTICATION}))
	{
		$LOGGER->fatal('Environment variable ZIPTIE_AUTHENTICATION not set');
	}

	my $ziptieClient = ZipTie::Client->new();
	my $applications = $ziptieClient->applications();
	$applications->addApplications(applications => $self->{applications});
}

1;

__END__

=pod

=head1 AUTHOR(S)

David Payne
david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2014, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
