package ZipTie::ApplicationsClient::Application;

=pod

=head1 NAME

ZipTie::ApplicationsClient::Application -- Contains details for a layer7 application to be provided
to ziptie-server.

=head1 DESCRIPTION

This class contains the details for a layer7 application that will be provided to ziptie-server.
It performs error checking on the data provided to ensure it's valid.

=head1 METHODS

=cut

use strict;
use warnings;

use Carp;

=pod

B<new($paramiters)> (constructor)

Creates a new ZipTie::ApplicationsClient::Application object with the paramiters provided.

$paramiters is expected to be a hash reference, with the following keys:

adapterId (Required) - The adapter ID for the applicaiton.
applicaitonName (Required) - The name of the application.
applicationDescription (Optional) - A description of the application.
riskSeverity (Optional) - An int representation of the severity of the risk for this application.

Example:
	my $testApplication = ZipTie::ApplicationsClient::Application->new(
		{adapterId => 'ZipTie::Adapters::Cisco::SecurityAppliance',
		applicationName => 'Example Application Name',
		applicationDescription => 'Example applicaiton description.', riskSeverity => 5});

The constructor performs validation on the paramiters passed in, and will croak if it encounteres a
problem.

Returns a new ZipTie::ApplicationsClient::Application object.

=cut
sub new
{
	my ($class, $paramiters) = @_;

	if (!defined($paramiters) || (ref($paramiters) ne 'HASH'))
	{
		croak 'constructor expects paramiters to be passed via hash reference';
	}

	my $self = {
			adapterId => undef,
			applicationName => undef,
			applicationDescription => undef,
			riskSeverity => undef,
			applicationServices => [],
			applicationClassifications => []
		};

	foreach my $key (keys(%{$paramiters}))
	{
		if (exists($self->{$key}))
		{
			$self->{$key} = $paramiters->{$key};
		}
		else
		{
			croak "$key is not a valid paramiter to constructor";
		}
	}

	if (!defined($self->{adapterId}))
	{
		croak 'adapterId is a required paramiter to constructor';
	}

	if (!defined($self->{applicationName}))
	{
		croak 'applicationName is a required paramiter to constructor';
	}

	if (defined($self->{riskSeverity}))
	{
		if ($self->{riskSeverity} !~ /^\d+$/)
		{
			croak $self->{riskSeverity} . ' is not a valid value for riskSeverity';
		}
	}

	bless($self, $class);

	return $self;
}

=pod

B<add_application_service($applicationService)>

Adds a new ZipTie::ApplicationsClient::ApplicationService object to the application.

=cut
sub add_application_service
{
	my ($self, $applicationService) = @_;

	if (!defined($applicationService))
	{
		croak 'applicationService is a required paramiter';
	}

	push(@{$self->{applicationServices}}, $applicationService->get_soap_object());
}

=pod

B<add_application_classification($applicationClassification)>

Adds a new ZipTie::ApplicationsClient::ApplicationClassification object to the application.

=cut
sub add_application_classification
{
	my ($self, $applicationClassification) = @_;

	if (!defined($applicationClassification))
	{
		croak 'applicationClassification is a required paramiter';
	}

	push(@{$self->{applicationClassifications}}, $applicationClassification->get_soap_object());
}

=pod

B<get_soap_object()>

Return a hash that is a valid representation of the Application SOAP object.

=cut
sub get_soap_object
{
	my ($self) = @_;

	my $soap_object = {%{$self}};

	# Remove any hash element that isn't defined, this prevents
	# default values beirg used in the soap call.
	foreach my $key (keys(%{$soap_object}))
	{
		if (!defined($soap_object->{$key}))
		{
			delete($soap_object->{$key});
		}
	}

	return $soap_object;
}

1;

__END__

=pod

=head1 AUTHOR(S)

David Payne
david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2014, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
