package ZipTie::ApplicationsClient::ApplicationService;

=pod

=head1 NAME

ZipTie::ApplicationsClient::ApplicationService -- Contains details for a layer7 application service
to be provided to the ziptie-server.

=head1 DESCRIPTION

This class contains the datails for a layer7 applicaiton service that will be provided to
ziptie-server.  It performs error checking on the data provided to ensure it's valid.

=head1 METHODS

=cut

use strict;
use warnings;

use Carp;

=pot

B<new($paramiters)> (constructor)

Creates a new ZipTie::ApplicaitonsClient::ApplicationService object with the paramiters provided.

$paramiters is expected to be a hash reference, with the following keys:

protocol (Required) - The protocol of the application service, TCP or UDP.
portLow (Required) - The low port of the applicaiton service if the service uses a port range, or
the port if the service is on a single port.
portHigh (Optional) - The high port of the application service if the service uses a port range.

Example:
	my $testService = ZipTie::ApplicationsClient::ApplicationService->new(
		{protocol => 'TCP', portLow => 5, portHigh => 10})

The constructor performs validation on the paramiters passed in, and will croak if it encounteres a
problem.

Returns a new ZipTie::ApplicationsClient::ApplicationService object.

=cut
sub new
{
	my ($class, $paramiters) = @_;

	if (!defined($paramiters) || (ref($paramiters) ne 'HASH'))
	{
		croak 'constructor expects paramiters to be passed via hash reference';
	}

	my $self = {
			protocol => undef,
			portLow => undef,
			portHigh => undef 
		};

	foreach my $key (keys(%{$paramiters}))
	{
		if (exists($self->{$key}))
		{
			$self->{$key} = $paramiters->{$key};
		}
		else
		{
			croak "$key is not a valid paramiter to constructor";
		}
	}

	if (!defined($self->{protocol}))
	{
		croak 'protocol is a required paramiter to constructor';
	}

	if (($self->{protocol} ne 'TCP') && ($self->{protocol} ne 'UDP'))
	{
		croak $self->{protocol} . ' is not a valid value for protocol';
	}

	if (!defined($self->{portLow}))
	{
		croak 'portLow is a required paramiter to constructor';
	}

	if ($self->{portLow} !~ /^\d+$/ || ($self->{portLow} > 65535))
	{
		croak $self->{portLow} . ' is not a valid value for portLow';
	}

	if (defined($self->{portHigh}))
	{
		if ($self->{portHigh} !~ /^\d+$/ ||
			($self->{portHigh} > 65535) ||
			($self->{portHigh} <= $self->{portLow}))
		{
			croak $self->{portHigh} . ' is not a valid value for portHigh';
		}
	}

	bless($self, $class);

	return $self;
}

=pod

B<get_soap_object()>

Return a hash that is a valid representation of the ApplicationService SOAP object.

=cut
sub get_soap_object
{
	my ($self) = @_;

	my $soap_object = {%{$self}};

	# Remove any hash element that isn't defined, this prevents
	# default values beirg used in the soap call.
	foreach my $key (keys(%{$soap_object}))
	{
		if (!defined($soap_object->{$key}))
		{
			delete($soap_object->{$key});
		}
	}

	return $soap_object;
}

1;

__END__

=pod

=head1 AUTHOR(S)

David Payne
david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2014, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
