package ZipTie::DBI;

=pod

=head1 NAME

ZipTie::DBI -- ZipTie layer between the adapter an the acutal DBI protocol.

=head1 DESCRIPTION

This class acts as a layer between an adapter and Perl's DBI interface.  It parses the connection
path document and connects to the database server.  It performs adapter logging, and creates
recordings if recordings are enabled.  When simulating devices it is able to simulate a DBI
interface.

=head1 METHODS

=cut

use strict;
use warnings;

use ZipTie::ConnectionPath;
use ZipTie::ConnectionPath::Protocol;
use ZipTie::DBI::Query;
use ZipTie::DBI::SqlInteractionManager;
use ZipTie::DBI::SqlInteractionManager::SqlInteraction;
use ZipTie::Logger;
use ZipTie::Recording;

use Data::Dumper;
use DBI;
use XML::Twig;
use MIME::Base64;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

our $AUTOLOAD;

# This is the file that will indicate we are supposed to simulate a device.
my $RECORDING_PLAYBACK_FILE = '/tmp/dbiRecordingPlaybackFile';

=pod

B<connect($data_source, $host, $username, $password, $dbi_protocol)> (constructor)

Connect to a database and create a ZipTie::DBI object.  Although it's possible for an adapter to
use this method directly, usually adapters should not call this method, instead adapters should use
a method like C<create_sybase_connection>.

Returns: ZipTie::DBI object.

=cut
sub connect
{
	my ($class, $data_source, $host, $username, $password, $dbi_protocol) = @_;

	my $dbh;
	my $playback_interactions = undef;
	if (-f $RECORDING_PLAYBACK_FILE)
	{
		open(my $fh, '<', $RECORDING_PLAYBACK_FILE) or
			$LOGGER->fatal("ERROR: Can't open recording playback file $RECORDING_PLAYBACK_FILE");
		my $recording_file = undef;
		while (my $line = <$fh>)
		{
			$line =~ s/\n//;
			if (-f $line)
			{
				$recording_file = $line;
				last;
			}
		}
		close($fh);

		if (!defined($recording_file))
		{
			# If the backup from recording is failing on the following line, make sure the ziptie user has read/execute
			# permissions on the directory that contains the recording file.
			$LOGGER->fatal("ERROR: Unable to find recording filename in recording playback file $RECORDING_PLAYBACK_FILE");
		}

		$LOGGER->debug("Using DBI simulator with recording file $recording_file");

		$playback_interactions = _parse_recording($recording_file);
	}
	else
	{
		$dbh = DBI->connect($data_source, $username, $password);
		if (!$dbh)
		{
			$LOGGER->debug("ERROR: Error connecting to database: $DBI::errstr");

			# Some sybase specific errors
			if ($DBI::errstr =~ /Login failed for user/)
			{
				$LOGGER->fatal_error_code($INVALID_CREDENTIALS, $host, "Authentication refused");
			}
			if ($DBI::errstr =~ /Adaptive Server is unavailable or does not exist/)
			{
				$LOGGER->fatal('ERROR: Database server is unavailable or does not exist');
			}

			$LOGGER->fatal("ERROR: Can't connect to $data_source: $DBI::errstr.");
		}

		$LOGGER->debug("Connected successfully using DBI.");
	}

	my $recording = undef;
	my $interaction_manager = undef;
	if ($ENV{ENABLE_RECORDING} && !defined($playback_interactions))
	{
		# Get the instance of the ZipTie::Recording module
		$recording = ZipTie::Recording::get_recording();
		$interaction_manager =  ZipTie::DBI::SqlInteractionManager->new();
	}

	my $self = {
			_dbh => $dbh,
			_interaction_manager => $interaction_manager,
			_recording => $recording,
			_playback_interactions => $playback_interactions,
			_dbi_protocol => $dbi_protocol
		};

	bless($self, $class);

	return $self;
}

=pod

B<create_sybase_connection($connection_path, $database_name)> (constructor)

Connect to a database using the sybase database driver and create a ZipTie::DBI object.  The sybase
database driver can also be used to connect to Microsoft SQL Server databases.  This method parses
the username, password, protocol, port and management server IP from the connection path document.

Returns: ZipTie::DBI object.

=cut
sub create_sybase_connection
{
	my ($class, $connection_path, $database_name) = @_;

	my $management_server_ip = $connection_path->get_management_server_ip_address();
	my $host = $connection_path->get_host();
	my $protocols = $connection_path->get_protocols();

	my $port;
	if (defined($protocols->{SQL}))
	{
		$port = $protocols->{SQL}->get_port();
	}
	else
	{
		$LOGGER->fatal("ERROR: No supported protocols to establish database connection.");
	}

	my $username = $connection_path->get_credential_by_name('username');
	my $password = $connection_path->get_credential_by_name('password');
	if (!defined($username) || !defined($password))
	{
		$LOGGER->fatal_error_code($INVALID_CREDENTIALS, $host, "Authentication refused with an empty username or password\nAborting login...");
	}

	my $tdsver = '7.1';

	$ENV{TDSHOST} = $management_server_ip;
	$ENV{TDSPORT} = $port;
	$ENV{TDSVER} = $tdsver;

	my $data_source = "dbi:Sybase:;database=$database_name";

	$LOGGER->debug("Connecting to $data_source using TDSHOST: $management_server_ip, TDSPORT: $port, TDSVER: $tdsver.");

	return ZipTie::DBI::connect($class, "dbi:Sybase:;database=$database_name", $management_server_ip, $username, $password, 'sybase');
}

=pod

B<set_syb_no_child_con()>

Set the syb_no_child_con attribute to 1.  Only use if the DBI protocol is sybase.  Calls
$LOGGER->fatal() if the DBI protocol is not sybase.

=cut
sub set_syb_no_child_con
{
	my ($self) = @_;

	if ($self->{_dbi_protocol} ne 'sybase')
	{
		$LOGGER->fatal('set_syb_no_child_con() called when DBI protocol set to ' . $self->{_dbi_protocol} .
			'.  Only call this method if the DBI protocol is \'sybase\'.');
	}

	if (defined($self->{_playback_interactions}))
	{
		# We don't need to set this value if we are playing back an interaction.
		return;
	}

	$LOGGER->debug('Setting syb_no_child_con to 1.');

	$self->{_dbh}->{syb_no_child_con} = 1;
}

=pod

B<disconnect()>

Disconnects from the database server.  If recordings are enabled this populates the
ZipTie::Recording object with interaction data.

=cut
sub disconnect
{
	my ($self) = @_;

	if (defined($self->{_playback_interactions}))
	{
		# Do noting
		return;
	}

	# Populate the recording if recordings are enabled.
	if (defined($self->{_recording}))
	{
		my $recording = $self->{_recording};

		my $sql_interactions = $self->{_interaction_manager}->get_interactions();
		foreach my $sql_interaction (@{$sql_interactions})
		{
			my $interaction = ZipTie::Recording::Interaction->new(
				richCommand => _create_rich_command($sql_interaction->get_prepare_args(), $sql_interaction->get_execute_args()),
				name => $sql_interaction->get_interaction_label(),
				cliProtocol => $self->{_dbi_protocol});
			$recording->start_current_interaction($interaction);
			$interaction->cliResponse(_serialize($sql_interaction->get_rows(), 1));
			$recording->finish_current_interaction();
		}
	}

	$self->{_dbh}->disconnect() or $LOGGER->debug("DBI disconnection failed: $DBI::errstr");
}

=pod

B<prepare($args)>

ZipTie::DBI equivalent of DBI's prepare method.  Performs adapter logging.  If recordings are
enabeled records the SQL for the recording.  Calls DBI's prepare method.

Returns: The return value of DBI's prepare method.

=cut
sub prepare
{
	my $self = shift;

	# Create a copy of the prepare arguments
	my @prepare_args = @_;

	$LOGGER->debug('-----------------------------------------------------------------------------');
	$LOGGER->debug('[PREPARE]');
	$LOGGER->debug('prepare args: ' . _serialize(\@prepare_args));
	$LOGGER->debug('-----------------------------------------------------------------------------');

	if ($self->{_playback_interactions})
	{
		my $ziptie_query = ZipTie::DBI::Query->new_playback_query(\@prepare_args, $self->{_playback_interactions});

		return $ziptie_query;
	}
	else
	{
		my $query = $self->{_dbh}->prepare(@prepare_args);

		my $ziptie_query = ZipTie::DBI::Query->new($self->{_interaction_manager}, $query, \@prepare_args);

		return $ziptie_query;
	}
}

=pod

B<do($args)>

ZipTie::DBI equivalent of DBI's do method.  Performs adapter logging then calls DBI's prepare
method.

This method does not cause an interaction to be created since no data is returned by this method.

Returns: The return value of DBI's do method.

=cut
sub do
{
	my $self = shift;

	$LOGGER->debug('-----------------------------------------------------------------------------');
	$LOGGER->debug('[DO]');
	$LOGGER->debug('do args: ' . _serialize(\@_));
	$LOGGER->debug('-----------------------------------------------------------------------------');

	if ($self->{_playback_interactions})
	{
		# do nothing
		return 1;
	}
	else
	{
		return $self->{_dbh}->do(@_);
	}
}

=pod

B<AUTOLOAD($args)>

This method should not be called directly.  If a method isn't implimented by ZipTie::DBI AUTOLOAD
will perform adapter logging and then attempt to call that method using Perl's DBI.

Returns: The return value of DBI's method.

=cut
sub AUTOLOAD
{
	my $self = shift;
	my $method = substr($AUTOLOAD, rindex($AUTOLOAD, '::') + 2);

	# Ignore the DESTROY method call
	if ($method eq 'DESTROY')
	{
		return;
	}

	$LOGGER->debug('-----------------------------------------------------------------------------');
	$LOGGER->debug("ZipTie::DBI AUTOLOAD calling method $method");
	$LOGGER->debug('args: ' . _serialize(\@_));
	$LOGGER->debug('-----------------------------------------------------------------------------');

	return $self->{_dbh}->$method(@_);
}

# Uses Data::Dumper to serialize $to_serialize.
# $multi_line is optional, defaults to single line.
sub _serialize
{
	my ($to_serialize, $multi_line) = @_;

	local($Data::Dumper::Indent);
	if ($multi_line)
	{
		$Data::Dumper::Indent = 1;
	}
	else
	{
		$Data::Dumper::Indent = 0;
	}

	local($Data::Dumper::Terse);
	$Data::Dumper::Terse = 1;

	return Dumper($to_serialize)
}

sub _create_rich_command
{
	my ($prepare_args, $execute_args) = @_;

	return 'Prepare Arguments: ' . _serialize($prepare_args) . ', Execute Arguments: ' . _serialize($execute_args);
}

sub _parse_recording
{
	my ($recording_file) = @_;

	my $interactions = {};

	my $twig = XML::Twig->new(twig_roots => {
			interaction => sub {
					my ($twig, $element) = @_;

					my $rich_command_element = $element->first_child('richCommand');
					my $response_element = $element->first_child('cliResponse');
					if (defined($rich_command_element) && defined($response_element))
					{
						my $rich_command = decode_base64($rich_command_element->text());
						my $response = decode_base64($response_element->text());
						$response =~ s/\n$//;

						if (!defined($interactions->{$rich_command}))
						{
							$interactions->{$rich_command} = {
									responses => [],
									current_response => 0
								};
						}

						push(@{$interactions->{$rich_command}->{responses}}, $response);
					}

					return 1;
				}
		});
	$twig->safe_parsefile($recording_file) or
		$LOGGER->fatal("ERROR: Unable to parse recording file $recording_file");

	return $interactions;
}

1;

__END__

=pod

=head1 AUTHOR(S)

David Payne
david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2013, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
