package ZipTie::DBI::Query;

=pod

=head1 NAME

ZipTie::DBI::Query -- ZipTie layer between the adapter and the actual DBI protocol.

=head1 DESCRIPTION

This class acts as a layer between an adapter and Perl's DBI interface.  It performs adapter logging
and contributes to recordfings if recordings are enabled.  When simulating devices it is able to
simulate DBI's Query object.

=head1 METHODS

=cut

use strict;
use warnings;

use ZipTie::DBI::SqlInteractionManager;
use ZipTie::DBI::SqlInteractionManager::SqlInteraction;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

our $AUTOLOAD;

=pod

B<new($interaction_manager, $query)> (constructor)

Creates a new ZipTie::DBI::Query object.

$interaction_manager is optional, only needs to be defined when creating a recording.

$query should be a DBI Query object.

$sql is the aruments passed into DBI's prepare method.

Returns: ZipTie::DBI::Query object.

=cut
sub new
{
	my ($class, $interaction_manager, $query, $sql) = @_;

	my $self = {
			_interaction_manager => $interaction_manager,
			_query => $query,
			_bind_columns_array => undef,
			_sql => $sql,
			_current_interaction => undef,
			_current_interaction_time => 0
		};

	bless($self, $class);
	return $self;
}

=pod

B<new($sql, $interactions)> (constructor)

Creates a new ZipTie::DBI::Query object for recording playback.  $sql is the arguments that wer
passed into DBI's prepare method, $playback_interactions is the interactions parsed from the
recording.

Returns: ZipTie::DBI::Query object.

=cut
sub new_playback_query
{
	my ($class, $sql, $playback_interactions) = @_;

	my $self = {
			_sql => $sql,
			_playback_interactions => $playback_interactions
		};

	bless($self, $class);
	return $self;
}

=pod

B<execute($args)>

ZipTie::DBI equivalent of DBI's execute method.  Performs adapter logging.  If recordings are
enabeled records the arguments for the recording.  Calls DBI's execute method.

Returns: The return value of DBI's execute method.

=cut
sub execute
{
	my $self = shift;

	# Create a copy of the execute args
	my @execute_args = @_;

	$self->{_execute_args} = \@execute_args;

	$LOGGER->debug('-----------------------------------------------------------------------------');
	$LOGGER->debug('[EXECUTE]');
	$LOGGER->debug('On statement: ' . ZipTie::DBI::_serialize($self->{_sql}));
	$LOGGER->debug('Execute args: ' . ZipTie::DBI::_serialize(\@execute_args));
	$LOGGER->debug('-----------------------------------------------------------------------------');

	$self->{_current_interaction_time} = time;

	if (defined($self->{_playback_interactions}))
	{
		my $dbiCommand = ZipTie::DBI::_create_rich_command($self->{_sql}, \@execute_args);

		if (defined($self->{_playback_interactions}->{$dbiCommand}))
		{
			my $current_response_number = $self->{_playback_interactions}->{$dbiCommand}->{current_response};

			my $current_rows = undef;
			eval('$current_rows = ' . $self->{_playback_interactions}->{$dbiCommand}->{responses}->[$current_response_number] . ';');

			if (!defined($current_rows) || ref($current_rows) ne 'ARRAY')
			{
				$LOGGER->fatal("Unable to eval response for dbiCommand \"$dbiCommand\".  Invalid recording?");
			}

			$self->{_current_playback_rows} = $current_rows;
			$self->{_current_playback_row_number} = 0;

			if ($current_response_number < $#{$self->{_playback_interactions}->{$dbiCommand}->{responses}})
			{
				++$self->{_playback_interactions}->{$dbiCommand}->{current_response};
			}

			return 1;
		}
		else
		{
			$LOGGER->fatal("No interaction found for dbiCommand \"$dbiCommand\".  Interaction missing from recording.");
		}
	}
	else
	{
		if (defined($self->{_interaction_manager}))
		{
			$self->{_current_interaction} = $self->{_interaction_manager}->create_interaction($self->{_sql}, \@execute_args);
		}
		else
		{
			$self->{_current_interaction} = undef;
		}

		return $self->{_query}->execute(@execute_args);
	}
}

=pod

B<bind_columns($args)>

ZipTie::DBI equivalent of DBI's bind_columns method.  Saves the column references so we are able to
obtain the data when fetch is called.  Calls DBI's bind_columns method.

Returns: The return value of DBI's bind_columns method.

=cut
sub bind_columns
{
	my $self = shift;

	my @bind_columns_array = @_;

	$self->{_bind_columns_array} = \@bind_columns_array;

	return $self->{_query}->bind_columns(@_);
}

=pod

B<fetch($args)>

ZipTie::DBI equivalent of DBI's fetch method.  Performs adapter logging.  If recordings are enabeled
records the rows for the recording.  Calls DBI's fetch method.

Returns: The return value of DBI's fetch method.

=cut
sub fetch
{
	my $self = shift;

	$LOGGER->debug('-----------------------------------------------------------------------------');
	$LOGGER->debug('[FETCH]');
	$LOGGER->debug('On statement: ' . ZipTie::DBI::_serialize($self->{_sql}));
	$LOGGER->debug('Execute args: ' . ZipTie::DBI::_serialize($self->{_execute_args}));
	$LOGGER->debug('fetch args: ' . ZipTie::DBI::_serialize(\@_));
	$LOGGER->debug('-----------------------------------------------------------------------------');

	my $row = undef;
	my $ret = 1;
	if (defined($self->{_playback_interactions}))
	{
		my $row = $self->{_current_playback_rows}->[$self->{_current_playback_row_number}];

		if ($self->{_current_playback_row_number} < $#{$self->{_current_playback_rows}})
		{
			++$self->{_current_playback_row_number};
		}
	}
	else
	{
		$ret = $self->{_query}->fetch(@_);

		foreach my $value_ref (@{$self->{_bind_columns_array}})
		{
			push(@{$row}, ${$value_ref});
		}
	}

	$LOGGER->debug('-----------------------------------------------------------------------------');
	$LOGGER->debug('[RESPONSE]');
	$LOGGER->debug(ZipTie::DBI::_serialize($row, 1));
	$LOGGER->debug('-----------------------------------------------------------------------------');

	if (defined($self->{_current_interaction}))
	{
		$self->{_current_interaction}->add_row($row);
	}

	return $ret;
}

=pod

B<fetchall_arrayref($args)>

ZipTie::DBI equivalent of DBI's fetchall_arrayref method.  Performs adapter logging.  If recordings
are enabeled records the rows for the recording.  Calls DBI's fetchall_arrayref method.

Returns: The return value of DBI's fetchall_arrayref method.

=cut
sub fetchall_arrayref
{
	my $self = shift;

	$LOGGER->debug('-----------------------------------------------------------------------------');
	$LOGGER->debug('[FETCHALL_ARRAYREF]');
	$LOGGER->debug('On statement: ' . ZipTie::DBI::_serialize($self->{_sql}));
	$LOGGER->debug('Execute args: ' . ZipTie::DBI::_serialize($self->{_execute_args}));
	$LOGGER->debug('fetchall_arrayref args: ' . ZipTie::DBI::_serialize(\@_));
	$LOGGER->debug('-----------------------------------------------------------------------------');

	my $rows = undef;
	if (defined($self->{_playback_interactions}))
	{
		$rows = $self->{_current_playback_rows};
	}
	else
	{
		$rows = $self->{_query}->fetchall_arrayref(@_);
	}

	$LOGGER->debug('-----------------------------------------------------------------------------');
	$LOGGER->debug('[RESPONSE]');
	$LOGGER->debug(ZipTie::DBI::_serialize($rows, 1));
	$LOGGER->debug('-----------------------------------------------------------------------------');

	if (defined($self->{_current_interaction}))
	{
		foreach my $row (@{$rows})
		{
			$self->{_current_interaction}->add_row($row);
		}
	}

	return $rows;
}

=pod

B<finish($args)>

ZipTie::DBI equivalent of DBI's finish method.  Performs adapter logging then calls DBI's finish
method.

Returns: The return value of DBI's do method.

=cut
sub finish
{
	my $self = shift;

	$LOGGER->debug('-----------------------------------------------------------------------------');
	$LOGGER->debug('[FINISH]');
	$LOGGER->debug('finish args: ' . ZipTie::DBI::_serialize(\@_));
	$LOGGER->debug('-----------------------------------------------------------------------------');

	my $response_time = time - $self->{_current_interaction_time};
	$self->{_current_interaction_time} = 0;
	$LOGGER->debug('[RESPONSE TIME]');
	$LOGGER->debug("$response_time second(s)");
	$LOGGER->debug('-----------------------------------------------------------------------------');

	if (defined($self->{_playback_interactions}))
	{
		# do nothing
		return 1;
	}
	else
	{
		return $self->{_query}->finish(@_);
	}
}

=pod

B<set_next_interaction_label($interaction_label)>

Sets the next interaction label (for recordings) to $interaction_label.

=cut
sub set_next_interaction_label
{
	my ($self, $interaction_label) = @_;

	if (defined($self->{_interaction_manager}))
	{
		$self->{_interaction_manager}->set_next_interaction_label($interaction_label);
	}
}

=pod

B<AUTOLOAD($args)>

This method should not be called directly.  If a method isn't implimented by ZipTie::DBI::Query
AUTOLOAD will perform adapter logging and then attempt to call that method using Perl's DBI.

Returns: The return value of DBI's method.

=cut
sub AUTOLOAD
{
	my $self = shift;
	my $method = substr($AUTOLOAD, rindex($AUTOLOAD, '::') + 2);

	# Ignore the DESTROY method call
	if ($method eq 'DESTROY')
	{
		return;
	}

	$LOGGER->debug('-----------------------------------------------------------------------------');
	$LOGGER->debug("ZipTie::DBI::Query AUTOLOAD calling method $method");
	$LOGGER->debug('Args: ' . ZipTie::DBI::_serialize(\@_, 1));
	$LOGGER->debug('-----------------------------------------------------------------------------');

	return $self->{_query}->$method(@_);
}

1;

__END__

=pod

=head1 AUTHOR(S)

David Payne
david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2013, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
