package ZipTie::DiscoveryParams;

use strict;
use warnings;

use ZipTie::Logger;
use XML::Twig;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

sub from_xml
{
	my $xml = shift or $LOGGER->fatal("ERROR - No XML specified to convert into a ZipTie::DiscoveryParams object!");

	# The discoveryParams object is just a simple hash, that's what all the telemetry operations already written expects.
	my $discoveryParams = {};

	my $twig = XML::Twig->new(twig_roots => { discoveryParams => sub {
				my ($twig, $element) = @_;

				$discoveryParams = $element->atts();

				return 1;
			}
		});
	$twig->parse($xml);

	return $discoveryParams;
}

1;

__END__

=head1 NAME

ZipTie::DiscoveryParams - Stores paramiters related to the telemetry adapter operation.

=head1 SYNOPSIS

	use ZipTie::DiscoveryParams;
	my $discoveryParams = ZipTie::DiscoveryParams::from_xml($xml_doc);
	if ($discoveryParams->{calculateAdminIp} eq 'true')
	{
		# Calculate the admin IP address
	}

=head1 DESCRIPTION

The C<ZipTie::DiscoveryParams> module provides a simple way to store and retreive paramiters related to the discovery adapter operation.


=head1 PUBLIC SUB-ROUTINES

=over 12

=item C<from_xml($input_xml_element)>

C<$input_xml_element> is a discoveryParam XML element.  Parses the discoveryParam XML element and returns a hash.  The keys of the hash
will be the name of each discovery paramiter, and the values of the hash will be the values of the paramiter.

Currently the only supported paramiter is calculateAdminIp.  If calculateAdminIp is set to the text string 'true' then the adapter should
calculate the admin IP address, otherwise it shouldn't have to calculate the admin IP address.

An example of a C<discoveyrParams> XML element:

	<discoveryParams calculateAdminIp="false"/>

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is AlterPoint.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

Portions created by IBM are Copyright (C) 2013, IBM Corp.

=head1 AUTHOR

Contributor(s): David Payne (david.payne@ca.ibm.com)
Date: Feb 25, 2013

=cut
