#!/usr/bin/perl

package ZipTie::Model::AddressTranslation;

use strict;
use warnings;
use v5.10;
use Storable qw(dclone);

use Data::Dumper;
use ZipTie::Logger;
use ZipTie::Addressing::Subnet;
use ZipTie::Adapters::Utils qw(get_any_address_encoding create_network_element get_any_address_address);

my $LOGGER = ZipTie::Logger::get_logger();

sub new 
{
   my ($class, $sourceZone, $destZone) = @_;
   my $self = {
	_atTranslateRules => [],
	_atSrcZone => $sourceZone,
	_atDstZone => $destZone,
	_atSrcTranslate => 1,
	_atDstTranslate => 1,
	_atProtocol => "tcp",
	_atEntryIndex => 1,
	# The following will, if necessary, acquire sets of ranges with which to MAP like ports
	_atPATMappingPortRanges => [ ],
	_addrTranslationLedgerMap => {},
	_useZones => 0,
    };
   bless $self, $class;
   return $self;
}

sub generateTranslationList
{
    my ($self, $outXml, $direction, $nonDefaultListName, $correctReverseName) = @_;
    my $fe = undef;
    my $feDisplayName;
    my $reorderedRuleList;
    my $genRule;
	my $critRule;
	my $expandedEntryAddrListSet;

	# As a matter of nicety, we don't generate if there aren't any rules 
	return unless (scalar(@{$self->{_atTranslateRules}}));

	# we need to reorder the rules based upon the exemption status...exemptions go first.
	# This is probably just as fast as executing this as a custom "sort()" sort criterion
	# algorithm.
	foreach $genRule (@{$self->{_atTranslateRules}})
	{
		if ($genRule->{isExemption})
		{
			unshift (@{$reorderedRuleList}, $genRule);
		}
		else # normal rule
		{
			push (@{$reorderedRuleList}, $genRule);			
		}
	}
	
	if ($self->_shouldGenerateList("source", $direction)) 
	{	
		### Generate forward traffic list
		$self->resetForNewFiltersSection();
		$outXml->open_element('filterList');

		$outXml->print_element('defaultAction', 'none');
	
		# we usually want to move on to the next filter list here
		$self->_generateAppliesTo($outXml, "forward");
		
		
		foreach $genRule (@{$reorderedRuleList}) 
		{
			$self->generateSingleTranslationRule($outXml, $direction, $genRule);
		}
		
		$outXml->print_element("mode", "stateful");
		$feDisplayName = $self->{_atName} // $self->{_atSrcZone} . "_" . $self->{_atDstZone} .
			"_NATMapping";
		$outXml->print_element('name', 
						$nonDefaultListName //  $feDisplayName . "_forward" );
		# This is about NAT
		$outXml->print_element("processingPhase", { 'core:packetProcessingPhase' => "address-translation"});
		$outXml->close_element('filterList');

		undef $fe; # clear filter entry
	}

	if ($self->_shouldGenerateList("destination", $direction))
	{
		# Generate reverse traffic list, only override fields which we need to
		$self->resetForNewFiltersSection();
		
		$outXml->open_element('filterList');
		$outXml->print_element('defaultAction', 'none');
		$self->_generateAppliesTo($outXml, "reverse");
		
		
		foreach $genRule (@{$reorderedRuleList}) 
		{
			$self->generateSingleTranslationRule($outXml, $direction, $genRule);
		}
		
		$outXml->print_element("mode", "stateful");
		if ($correctReverseName){
			$feDisplayName = $self->{_atName} //  $self->{_atSrcZone} . "_" . $self->{_atDstZone} . 
			"_NATMapping";
		} else {
			$feDisplayName = $self->{_atName} //  $self->{_atDstZone} . "_" . $self->{_atSrcZone} . 
			"_NATMapping";
		}	
		$outXml->print_element('name', 
				$nonDefaultListName // $feDisplayName . "_reverse" );
		# This is about NAT
		$outXml->print_element("processingPhase",  { 'core:packetProcessingPhase' => "address-translation"});
		$outXml->close_element('filterList');
	}	
}

# Generate a single rule in one specified direction (forward or reverse)
sub generateSingleTranslationRule
{
	my ($self, $outXml, $direction, $atRule) = @_;
	
	my $fe = undef;
    my $feDisplayName;
    my $expandedEntryAddrListSet;
    my $critRule;
    
 	$fe->{log} = "false";
 	# The action we will cause to take here, if there is a match on the filter, is to "accept", i.e., 
 	# move on to next containing filter section.
	$fe->{primaryAction} = "permit";
	
	if ($atRule->{name})
	{
		$fe->{name} = $atRule->{name};
	}
	


	# If the protocol is TCP or IP, then the reverse mapping should be for 
	# established session traffic only.
	# +++waynet Nope. What if this is a single-address bidirectional static exemption allowing
	# bidirectional address mapping?
	
	if ($direction ~~ "destination")
	{
		if ($atRule->{establishedReverseOnly} )
		{
			if(!$atRule->{protocol} or ($atRule->{protocol} and $atRule->{protocol} ne "udp")){
				$fe->{established} = "true";
			}
		}		
	}

	# See if we can generate the set of criteria as a linear set of src/dst mappings, or if]
	# we need distinct rules to capture the semantics
	$expandedEntryAddrListSet = $self->_doExpansionOfRuleEntryAddrs($atRule);
	
	# if we are doing port mapping, we need to split this into as many separate generateEntryRuleAddr invocations as
	# there are predefined port ranges.
	#
	my $lPATAddrTranslations;
	my $lnPATAddrTranslations;
	my $lAddrTranslation;
	my $lAddrTranslations = $atRule->{addrTranslations};

	foreach my $addrTranslation (@{$atRule->{addrTranslations}})
	{
		
		if ($addrTranslation->{doPortRangeMapping})
		{
			push ( @{$lPATAddrTranslations}, $addrTranslation);
		}
		else
		{
			push ( @{$lnPATAddrTranslations}, $addrTranslation);			
		}
	}
			
	if ($lPATAddrTranslations)
	{
		if ($direction eq "destination")
		{
			# Generate once -> rules, "normally", for all addrTranslations which are not tagged with PAT
			
			if ($lnPATAddrTranslations)
			{
				foreach $critRule (@{$expandedEntryAddrListSet})
				{
					if (defined ($atRule->{protocol}))
					{
						$fe->{protocol} = $atRule->{protocol};				
					}
	
					foreach my $critAddr (@{$critRule})
					{
						$self->_generateEntryRuleAddrs($fe, $critAddr, $lnPATAddrTranslations, ($direction eq "source") ? "forward" : "reverse");			
					}
	
					# output the reverse element only if all translations weren't suppressed for some reason (e.g., 'deny' clauses on reverse)
					if ($self->_filterEntryHasTranslations($fe) or $direction eq "source")
					{
						$fe->{processOrder} = $self->{_atEntryIndex}++;
						$outXml->print_element ("filterEntry", $fe) ;								
					}
					delete $fe->{sourceIpAddr};				
					$self->_clearRuleAddrFields($fe);				
				}
				
			} 
		}		
		# Now for PAT. We have to turn this into multiple calls to _generateEntryRuleAddrs, with different logic for each.		
		for my $patPortRange (@{$self->{_atPATMappingPortRanges}})
		{
			if (defined ($atRule->{protocol}))
			{
				$fe->{protocol} = $atRule->{protocol};				
			}

			# regenerate the 'establshed' logic
			if ($direction ~~ "destination")
			{
				if ($atRule->{establishedReverseOnly} )
				{
					if(!$atRule->{protocol} or ($atRule->{protocol} and $atRule->{protocol} ne "udp")){
						$fe->{established} = "true";
					}
				}		
			}

			foreach $critRule (@{$expandedEntryAddrListSet})
			{
				# If we're working with multiple discrete criterion or translation addrs, then we only call with the
				# criterion extra port range ONCE if we're doing this on source (forward), OR only once with the
				# destination criterion addrs once if we're doing this on destination (reverse)
				my $fwdPatPortRange = $patPortRange;
				my $reversePatPortRange = $patPortRange;
				
				foreach my $critAddr (@{$critRule})
				{
					if ($direction eq "source")
					{
						$self->_generateEntryRuleAddrs($fe, $critAddr, 
							 $lAddrTranslations, "forward", 
								$fwdPatPortRange, $reversePatPortRange);
						# undef the fwd port range for the rest of the $critAddr/$critRule iterations
						undef $fwdPatPortRange;							
					}
					else
					{
						$self->_generateEntryRuleAddrs($fe, $critAddr, 
							 $lPATAddrTranslations, "reverse", 
								$fwdPatPortRange, $reversePatPortRange);							
						# undef the reverse port range for the rest of the $critAddr/$critRule iterations
						undef $reversePatPortRange;													
					}
				}
				# output the reverse element only if all translations weren't suppressed for some reason (e.g., 'deny' 
				# clauses on reverse)
				if ($self->_filterEntryHasTranslations($fe) or $direction eq "source")
				{
					$fe->{processOrder} = $self->{_atEntryIndex}++;
					$outXml->print_element ("filterEntry", $fe) ;				
				}
				delete $fe->{sourceIpAddr};				
				$self->_clearRuleAddrFields($fe);
			}				
		}
	}
	else
	{ 

		foreach $critRule (@{$expandedEntryAddrListSet})
		{
			if (defined ($atRule->{protocol}))
			{
				$fe->{protocol} = $atRule->{protocol};				
			}


			foreach my $critAddr (@{$critRule})
			{
				$self->_generateEntryRuleAddrs($fe, $critAddr, $atRule->{addrTranslations}, ($direction eq "source") ? "forward" : "reverse");			
			}
			# output the reverse element only if all translations weren't suppressed for some reason (e.g., 'deny' clauses on reverse)
			if ($self->_filterEntryHasTranslations($fe) or $direction eq "source")
			{
				$fe->{processOrder} = $self->{_atEntryIndex}++;
				$outXml->print_element ("filterEntry", $fe) ;								
			}
			delete $fe->{sourceIpAddr};				
			$self->_clearRuleAddrFields($fe);
		}
	}
}

# You Choose one of these or the other,  The context object group was added later.

sub setInterfaceZoneMapper
{
    my ($self, $zoneMapper) = @_;

    $self->{_atZoneMapper} = $zoneMapper;
}

sub setInterfaceContextEnabledZoneMapper
{
    my ($self, $contextZoneMapper, $czMapperContext) = @_;

    $self->{_atZoneMapper} = $contextZoneMapper;
    $self->{_atZoneMapperContext} = $czMapperContext;
    	
}
# the following is optional
sub setObjectGroupMapper
{
    my ($self, $ogMapper) = @_;

    $self->{_atObjectGroupMapper} = $ogMapper;
}

sub setSourceTranslationOnly 
{ 
    my $self = shift;
    $self->{_atDstTranslate} = 0;
}

sub setDstTranslationOnly 
{ 
    my $self = shift;
    $self->{_atSrcTranslate} = 0;
}

# If a given mapping has or wants a name....

sub setTranslationName 
{ 
    my ($self, $name) = @_;
    $self->{_atName} = $name;
}

sub getTranslationName
{
	my $self = shift;
	return $self->{_atName};
}

sub addPortTranslationRange
{
	# argument is a list of ranges contiguously
	my ($self, $ptrLow, $ptrHigh) = @_;
	
	if ($ptrLow >= $ptrHigh)
	{
		$LOGGER->debug("AddressTranslation: addPortTranslationRange; cannot add mapping range from $ptrLow TO $ptrHigh contiguously");
	}
	else
	{
		push (@{$self->{_atPATMappingPortRanges}}, [$ptrLow, $ptrHigh]);	
	}
}

# Add a rule then return it.  Provide "set" vectors for things you might
# want to reset on/append to a rule.

sub addRule
{
	my ($self, $ruleArgs) = @_;

	my $newRule;
	$newRule->{protocol} = lc($ruleArgs->{protocol}) if ($ruleArgs->{protocol});
	$newRule->{name} = $ruleArgs->{name} if ($ruleArgs->{name});

	$ruleArgs->{rule} = $newRule;
	$newRule->{isExemption} = 0; # initialize exemption status

	# We track the static notion of a rule from the static NAT mapping flag
	$newRule->{isStatic} = $ruleArgs->{isStaticNATMapping} // 0;
	$self->addRuleAddressTranslationMapping($ruleArgs); # add the arguments defining this "initial" address translation	    
	
	# We set this to only allow established connection traffic on the reverse channel IF this is a dynamic NAT
	# rule (=> not a static NAT rule) or it's been explicitly requested that way
	if ($ruleArgs->{"establishedReverseConnectionTrafficOnly"} or not $ruleArgs->{"isStaticNATMapping"})
	{
		$self->setEstablishedReverseConnectionTrafficOnly($newRule, "true");
	} 

	push ( @{$self->{_atTranslateRules}}, $newRule);
	return $newRule;
}

sub setExemptionRule
{
	my ($self, $rule) = @_;
	$rule->{isExemption} = 1;
	return;
}

sub setEstablishedReverseConnectionTrafficOnly
{
	my ($self, $rule, $value) = @_;
	
	# Setting this up so it can be passed in 1/0 or "true"/"false"
	if ($value eq "" . 1 or $value ~~ "true")
	{
		$rule->{establishedReverseOnly} = "true" ;	
	}
	elsif ($value eq "" . 0 or $value ~~ "false")
	{
		undef $rule->{establishedReverseOnly};
	}
	
	return $rule->{establishedReverseOnly};
}

sub addRuleAddressTranslationMapping
{
	my ($self, $ruleArgs) = @_;
	my $atM = undef;
	my $rule = $ruleArgs->{rule};
	my $atArg = 	{ sourceAddrMapped => $ruleArgs->{"sourceAddrMapped"}, 
					sourceAddrEndMapped => $ruleArgs->{"sourceAddrEndMapped"},
					sourceAddrMaskMapped => $ruleArgs->{"sourceAddrMaskMapped"}, # this can also take a # of mask bits
					sourceAddrNameMapped => $ruleArgs->{"sourceAddrNameMapped"},
					sourcePortMapped => $ruleArgs->{"sourcePortMapped"},
					sourcePortEndMapped => $ruleArgs->{"sourcePortEndMapped"},
					sourcePortNameMapped => $ruleArgs->{"sourcePortNameMapped"},
					destAddrMapped => $ruleArgs->{"destAddrMapped"},
					destAddrEndMapped => $ruleArgs->{"destAddrEndMapped"},
					destAddrMaskMapped => $ruleArgs->{"destAddrMaskMapped"}, 
					destAddrNameMapped => $ruleArgs->{"destAddrNameMapped"}, 
					destPortMapped => $ruleArgs->{"destPortMapped"},
					destPortEndMapped => $ruleArgs->{"destPortEndMapped"},
					destPortNameMapped => $ruleArgs->{"destPortNameMapped"},
					doPortRangeMapping => $ruleArgs->{"usePATOverload"}};

	while ((my $atK, my $atV) = each %{$atArg})
	{
		if (defined $atV)
		{
			$atM->{$atK} = $atV;
		}
	}

   	if (defined $atM)
   	{
		# Per decisions by Rory, DPayne and Wayne, 9/18/12, "appliedAddressTranslation" now tracks
		# the notion of static--overall--in the rule.
		$atM->{'appliedAddressTranslation'} = $ruleArgs->{'useAppliedAddressRules'} //
			( ($rule->{isStatic}) ? "true" : undef); 

	   	push (@{$rule->{addrTranslations}}, $atM);	
   	}
}

sub addRuleAddressCriterion
{
	my ($self, $ruleArgs) = @_;
	my $rule = $ruleArgs->{rule};
	my $addrCriterion;
	my $addrArg;

	for $addrArg  ("sourceIp", "sourcePrefix", "sourceIpEnd", "sourceIpObjectGroup", 
						"sourceServicePort", "sourceServicePortEnd", "sourceServiceObjectName",
						"destIp", "destPrefix", "destIpEnd", "destIpObjectGroup", 
						"destServicePort", "destServicePortEnd", "destServiceObjectName",
						"ipProtocol")
	{
		if (defined $ruleArgs->{$addrArg})
		{
			$addrCriterion->{$addrArg} = $ruleArgs->{$addrArg};
		}
	}
	
	# Map "any"
	for $addrArg ("sourceIp","destIp" )
	{
		if ($ruleArgs->{$addrArg} ~~ "any")
		{
			#+++wayne should have this notion as 	
		}
	}
	
	# provide a definitive value for "negateMatch"
	$addrCriterion->{"negateMatch"} = $ruleArgs->{"negateMatch"} // 0;
	
	push (@{$rule->{addrCriteria}}, $addrCriterion);
	return $rule;		
}

sub resetForNewFiltersSection
{
	my $self = shift;
	$self->{_atEntryIndex} = 1;
}

sub useZones
{
	my ($self) = @_;
	$self->{_useZones} = 1;
}

sub _generateAppliesTo
{
	my ($self, $outXml, $direction) = @_;

	# What we need to do is to figure out our sourceAppliesTo and our sourceDest.  For each of them
	# pair sourceAppliesTo[0..n] and destAppliesTo[0..n]

	my $sourceAppliesZone = ($direction ~~ "reverse") ? $self->{_atDstZone} : $self->{_atSrcZone};
	my $destAppliesZone = ($direction ~~ "reverse") ? $self->{_atSrcZone} : $self->{_atDstZone};

	if($self->{_useZones}){
		$outXml->open_element("appliesTo");
		if($sourceAppliesZone =~ /^.*?::(.*)$/){
			$outXml->print_element("source", $1);
		}else{
			$outXml->print_element("sourceGroup", $sourceAppliesZone);
		}
		if($destAppliesZone =~ /^.*?::(.*)$/){
			$outXml->print_element("destination", $1);
		}else{
			$outXml->print_element("destinationGroup", $destAppliesZone);
		}
		$outXml->close_element("appliesTo");
	}else{

		my $zoneHash = $self->{_atZoneMapper}->($self->{_atZoneMapperContext});

		foreach my $sourceAppliesIf (@{$zoneHash->{$sourceAppliesZone}})
		{
			foreach my $destAppliesIf (@{$zoneHash->{$destAppliesZone}})
			{
				$outXml->open_element("appliesTo");
				$outXml->print_element("source", $sourceAppliesIf);
				$outXml->print_element("destination", $destAppliesIf);
				$outXml->close_element("appliesTo");
			}
		}
	}
}
#  This inner routine just abstracts out a somewhat clunky and complicated test from
#  generateTranslationList -- whether to emit a set of the implied rules based on 
#  the way the AddressTranslation was set up, how the generate is being asked for
#  and...that's about it.
 
sub _shouldGenerateList
{
	my ($self, $proposedDirection, $specifiedDirection) = @_;
	
	if (defined $specifiedDirection)
	{
		if ($proposedDirection ne $specifiedDirection)
		{
			return 0;
		}
	}
	 
	if ($proposedDirection ~~ "source") { return $self->{_atSrcTranslate}; }
	if ($proposedDirection ~~ "destination") { return $self->{_atDstTranslate}; }	 
	return 0;	 	
}	

#   _generateEntryRuleAddrs
#
#  This is the main driver routine to generate the addressing within a passed (and partially initialized) 
#  hash corresponding to a <filterEntry>
#
#  PARAMETERS:
#    $self
#    $fEntry - hash to be used to set up addressing "for"
#    $ruleAddrTuple - criterion address rule *as added originally when passed to addRuleAddressCriterion*
#    $translationParams - set of translation vectors for rule - effectively the entire set of {addrTranslations}
#         within the rule
#    $genDirection - either "forward" or "reverse", indicates the "sense" of mapping being generated at this point for
#           the rule.  Ultimately, both directions will be called on rule (if so enabled)
#    $extraDynamicPortCriterionRange - OPTIONAL, if specified, a <high, low> tuple, will be used as PAT mapping to augment $ruleAddrTuple and
#           $translationParams for criterion mappings, *replacing any port definition in the rule otherwise* if present)
#    $extraDynamicPortTranslationRange - OPTIONAL, if specified, a <high, low> tuple, will be used as PAT mapping to augment $ruleAddrTuple and
#           $translationParams for translation mappings, *replacing any port definition in the rule otherwise* if present)

sub _generateEntryRuleAddrs
{
	my ($self, $fEntry, $ruleAddrTuple, $translationParams, $genDirection, $extraDynamicPortCriterionRange, $extraDynamicPortTranslationRange) = @_;
	my $feIP;
	my $mappedAddrTransforms = [];
	my $mappedAddrTransform;
	my $mappedTranslationTransform;
	my $translationParam;
	
	##### ---> CASE 1 : 'forward' generation (relative to filter list marked zones).
	if ($genDirection ~~ "forward")
	{
		###### -> CASE 1a: 'forward' generation, for 'source' focused NAT
		if (defined $ruleAddrTuple->{"sourceIp"})
		{
			# See if it's a range 
			if (defined $ruleAddrTuple->{"sourceIpEnd"})
			{
				push (@{$fEntry->{"sourceIpAddr"}},
				_nrEnc($ruleAddrTuple->{"sourceIp"}, $ruleAddrTuple->{"sourceIpEnd"}));
			}
			elsif (defined $ruleAddrTuple->{"sourcePrefix"})
			{				
				push (@{$fEntry->{"sourceIpAddr"}}, _nNetEnc($ruleAddrTuple->{"sourceIp"}, $ruleAddrTuple->{"sourcePrefix"}));
			}
			else
			{
				push (@{$fEntry->{"sourceIpAddr"}}, _hostEnc($ruleAddrTuple->{"sourceIp"})); 
			}
		}
		elsif (defined $ruleAddrTuple->{"sourceIpObjectGroup"})
		{
			push (@{$fEntry->{"sourceIpAddr"}}, _ogEnc($ruleAddrTuple->{"sourceIpObjectGroup"}));
		}
		
		# if we just generated something, then use the extra port range if present
		if (defined $extraDynamicPortCriterionRange and defined $fEntry->{"sourceIpAddr"})
		{
			push (@{$fEntry->{"sourceService"}}, _pRangeEnc( $extraDynamicPortCriterionRange->[0], 
														$extraDynamicPortCriterionRange->[1])); 			
		}
		elsif (defined $ruleAddrTuple->{"sourceServicePort"})
		{
			if (defined $ruleAddrTuple->{"sourceServicePortEnd"})
			{
				push (@{$fEntry->{"sourceService"}}, _pRangeEnc( $ruleAddrTuple->{"sourceServicePort"}, $ruleAddrTuple->{"sourceServicePortEnd"})); 
			}
			elsif (defined $ruleAddrTuple->{"sourceServicePortOperator"})
			{
				push (@{$fEntry->{"sourceService"}}, _portEnc($ruleAddrTuple->{"sourceServicePort"},$ruleAddrTuple->{"sourceServicePortOperator"}));
			}
			else
			{
				push (@{$fEntry->{"sourceService"}}, _portEnc($ruleAddrTuple->{"sourceServicePort"}));	
			}
		}
		elsif (defined $ruleAddrTuple->{"sourceServiceObjectName"})
		{
			push (@{$fEntry->{"sourceService"}}, _ogEnc($ruleAddrTuple->{"sourceServiceObjectName"}));
		}
		
		# and the mapping. If this is a "deny" rule, then we forego this altogether, we're just trying to hook
		# up the source params we just set up with a disposition of "none".  Check the ledger to make sure we 
		# haven't already generated this.
		
		if ($ruleAddrTuple->{"negateMatch"})
		{
			# reset entry
			$fEntry->{primaryAction} = "none";
		}
		else
		{
			for $translationParam (@{$translationParams})
			{
				undef $mappedAddrTransform;
				if (defined $translationParam->{"sourceAddrMapped"})
				{
					if (defined $translationParam->{"sourceAddrEndMapped"})
					{
						$mappedAddrTransform = _nrEnc($translationParam->{"sourceAddrMapped"},  $translationParam->{"sourceAddrEndMapped"});
					}
					elsif (defined $translationParam->{"sourceAddrMaskMapped"})
					{
						$mappedAddrTransform = _nNetEnc($translationParam->{"sourceAddrMapped"}, $translationParam->{"sourceAddrMaskMapped"});						
					}
					else
					{
						$mappedAddrTransform = _hostEnc($translationParam->{"sourceAddrMapped"});					
					}
					if (not $self->_isInTranslationLedger($translationParam, "sourceAddrMapped"))
					{
						push (@{$mappedTranslationTransform->{translationIP}}, $mappedAddrTransform);								
						$self->_addToTranslationLedger($translationParam, "sourceAddrMapped", $mappedAddrTransform);
					}
				}	
				elsif (defined $translationParam->{"sourceAddrNameMapped"})
				{
					$mappedAddrTransform = _ogEnc($translationParam->{"sourceAddrNameMapped"});
					if (not $self->_isInTranslationLedger($translationParam, "sourceAddrNameMapped"))
					{
						push (@{$mappedTranslationTransform->{translationIP}}, $mappedAddrTransform);
						$self->_addToTranslationLedger($translationParam, "sourceAddrNameMapped", $mappedAddrTransform);						
					}
				}				

				if (defined $mappedAddrTransform)
				{
					push (@{$mappedAddrTransforms}, $mappedAddrTransform);
				}
				# is it port-qualified? Or called up by name?  Make sure it was assigned as an addr and not 
				# bypassed by already being in the ledger.
				if (defined $extraDynamicPortTranslationRange and $translationParam->{doPortRangeMapping} and 
					(defined $mappedTranslationTransform))
				{
					push (@{$mappedTranslationTransform->{translationServicePort}}, 
					{translationPort => _pRangeEnc( $extraDynamicPortTranslationRange->[0], $extraDynamicPortTranslationRange->[1]),
						translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)}); 			
				}
				elsif (defined $translationParam->{"sourcePortMapped"})
				{
					if (defined $translationParam->{"sourcePortEndMapped"})
					{
						push (@{$mappedTranslationTransform->{translationServicePort}}, 
							{translationPort => _pRangeEnc($translationParam->{sourcePortMapped}, $translationParam->{sourcePortEndMapped}), 
								translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});
					}
					else
					{
						push (@{$mappedTranslationTransform->{translationServicePort}}, 
							{translationPort => _portEnc($translationParam->{"sourcePortMapped"}),
								translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});
					}
				}
				elsif (defined $translationParam->{"sourcePortNameMapped"})
				{
					push (@{$mappedTranslationTransform->{translationServicePort}},
						{translationPort =>	_ogEnc($translationParam->{"sourcePortNameMapped"}), 
							translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});			
				}

				if (defined $mappedTranslationTransform)
				{
					push (@{$fEntry->{sourceAddressTranslation}->{translationAddress}}, $mappedTranslationTransform);
					undef $mappedTranslationTransform;
				}				

				# Push out applied addressing assertion iff source translation is defined
				if (defined $translationParam->{"appliedAddressTranslation"} and defined $fEntry->{sourceAddressTranslation})
				{
					$fEntry->{sourceAddressTranslation}->{translateAppliedAddressing} =
						$translationParam->{"appliedAddressTranslation"};
				}	

				# add to translationAddress if we've accumulated enough to do so
				if (defined $mappedTranslationTransform)
				{
					push(@{$fEntry->{sourceAddressTranslation}->{translationAddress}}, $mappedTranslationTransform);
					undef $mappedTranslationTransform;
				}
			}			
		}	

		###### -> CASE 1b: 'forward' generation, for 'destination' focused NAT

		if (defined $ruleAddrTuple->{"destIp"})
		{
			# See if it's a range 
			if (defined $ruleAddrTuple->{"destIpEnd"})
			{
				push (@{$fEntry->{"destinationIpAddr"}},
				_nrEnc($ruleAddrTuple->{"destIp"}, $ruleAddrTuple->{"destIpEnd"}));
			}
			elsif (defined $ruleAddrTuple->{"destPrefix"})
			{
				push (@{$fEntry->{"destinationIpAddr"}}, _nNetEnc($ruleAddrTuple->{"destIp"}, $ruleAddrTuple->{"destPrefix"}));
			}
			else
			{
				push (@{$fEntry->{"destinationIpAddr"}}, _hostEnc($ruleAddrTuple->{"destIp"})); 
			}
		}
		elsif (defined $ruleAddrTuple->{"destIpObjectGroup"})
		{
			push (@{$fEntry->{"destinationIpAddr"}}, _ogEnc($ruleAddrTuple->{"destIpObjectGroup"}));
		}
		
		# if we have an "extra port" override then use it here if we just generated something.  Let's make
		# sure it wasn't just the extra dest arm of a src + dst qualified ("twice NAT") condition.
		if (defined $extraDynamicPortCriterionRange and defined $fEntry->{"destinationIpAddr"}
			and not defined $fEntry->{"sourceIpAddr"} )
		{
			push (@{$fEntry->{"destinationService"}},_pRangeEnc( $extraDynamicPortCriterionRange->[0], 
										$extraDynamicPortCriterionRange->[1]));
		}
		elsif (defined $ruleAddrTuple->{"destServicePort"})
		{
			if (defined $ruleAddrTuple->{"destServicePortEnd"})
			{
				push (@{$fEntry->{"destinationService"}}, 
					_pRangeEnc($ruleAddrTuple->{"destServicePort"}, $ruleAddrTuple->{"destServicePortEnd"}));
			}
                        elsif (defined $ruleAddrTuple->{"destServicePortOperator"})
                        {
                                push (@{$fEntry->{"destService"}}, _portEnc($ruleAddrTuple->{"destServicePort"},$ruleAddrTuple->{"destServicePortOperator"}));
                        }
			else
			{
				push (@{$fEntry->{"destinationService"}}, _portEnc($ruleAddrTuple->{"destServicePort"}));
			}
		}
		elsif (defined $ruleAddrTuple->{"destServiceObjectName"})
		{
				push (@{$fEntry->{"destinationService"}}, _ogEnc($ruleAddrTuple->{"destServiceObjectName"}));
		}

		# and the mapping for the address translation. If this is a "deny" rule, then we forego this altogether, we're just trying to hook
		# up the source params we just set up with a disposition of "none".  Also, note here that destAddressTranslation is unique
		# in its specification of support for "appliedAddressing"
		
		if ($ruleAddrTuple->{"negateMatch"})
		{
			# reset entry primary action
			$fEntry->{primaryAction} = "none";
		}
		else
		{
			for $translationParam (@{$translationParams})
			{
				undef $mappedAddrTransform;
				if (defined $translationParam->{"destAddrMapped"})
				{
					if (defined $translationParam->{"destAddrEndMapped"})
					{
						$mappedAddrTransform = _nrEnc($translationParam->{"destAddrMapped"},  $translationParam->{"destAddrEndMapped"});
					}
					elsif (defined  $translationParam->{"destAddrMaskMapped"})
					{
						$mappedAddrTransform = _nNetEnc($translationParam->{"destAddrMapped"}, $translationParam->{"destAddrMaskMapped"});						
					}
					else
					{
						$mappedAddrTransform = _hostEnc($translationParam->{"destAddrMapped"});					
					}
					
					if (not $self->_isInTranslationLedger($translationParam, "destAddrMapped"))
					{
						push (@{$mappedTranslationTransform->{translationIP}}, 
							$mappedAddrTransform);			
						$self->_addToTranslationLedger($translationParam,"destAddrMapped", $mappedAddrTransform);					
					}				
				}
				elsif (defined $translationParam->{"destAddrNameMapped"})
				{
					$mappedAddrTransform = _ogEnc($translationParam->{"destAddrNameMapped"});
	
					if (not $self->_isInTranslationLedger($translationParam, "destAddrNameMapped"))
					{				
						push (@{$mappedTranslationTransform->{translationIP}}, 
							$mappedAddrTransform);
						$self->_addToTranslationLedger($translationParam, "destAddrNameMapped", $mappedAddrTransform);					
					}						
				}	
					
				# is the mapping port-qualified? Make sure it was assigned as an addr and not 
				# bypassed by already being in the ledger.
				if (defined $extraDynamicPortTranslationRange and $translationParam->{doPortRangeMapping} and defined $mappedTranslationTransform)
				{
					push (@{$mappedTranslationTransform->{translationServicePort}},  
						{translationPort => _pRangeEnc( $extraDynamicPortTranslationRange->[0], $extraDynamicPortTranslationRange->[1]),
						translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});	
				}
				elsif (defined $translationParam->{"destPortMapped"})
				{
					if (defined $translationParam->{"destPortEndMapped"})
					{
						push (@{$mappedTranslationTransform->{translationServicePort}}, 
						{translationPort => _pRangeEnc($translationParam->{"destPortMapped"}, $translationParam->{"destPortEndMapped"}), 
							translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});						
					}
					else
					{
						push(@{$mappedTranslationTransform->{translationServicePort}}, 
						{translationPort => _portEnc($translationParam->{"destPortMapped"}), 
							translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});						
					}
				}
				elsif (defined $translationParam->{"destPortNameMapped"})
				{
					push(@{$mappedTranslationTransform->{translationServicePort}}, 
						{translationPort =>	_ogEnc($translationParam->{"destPortNameMapped"}), 
							translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});			
				}

								
				# push out the set of addrs, ports for translation into one value of the translation address, if we have it
				if (defined $mappedTranslationTransform)
				{
					# Carry through translation iff we have already asserted a destination translation
					if (defined $translationParam->{"appliedAddressTranslation"})
					{
						$fEntry->{destAddressTranslation}->{translateAppliedAddressing} =
							$translationParam->{"appliedAddressTranslation"};
					}	
					push(@{$fEntry->{destAddressTranslation}->{translationAddress}}, $mappedTranslationTransform);
					undef $mappedTranslationTransform;
				}
				if (defined $mappedAddrTransform)
				{
					push (@{$mappedAddrTransforms}, $mappedAddrTransform);
				}
			}
		}
	}
	# There are two cases by which we bypass generation on the reverse side: one is if the rule was explicitly marked as
	# 'negate' (since there was no forward mapping effected, there's nothing to check for on the reverse), or if the
	# source AND dest here is 'any'.
	elsif (not $ruleAddrTuple->{"negateMatch"})
	{
		my $preMappedNetworkElt;
			##### ---> CASE 2 : 'reverse' generation (relative to filter list marked zones).  Note that we can
				# 				forego all of this for a "deny" based address rule, since there will have been
				#				no mapped address to have to change out of.

		###### -> CASE 2a: 'reverse' generation, for 'source' focused NAT

		for $translationParam (@{$translationParams})
		{
			undef $mappedAddrTransform;
			# We will pull up the translation address, for "reverse" mapping as a source/dest IP.  If
			# we find it{ has already been generated in this entry, however, we can suppress it.
			if (defined $translationParam->{"sourceAddrMapped"} and
				not $self->_isInTranslationLedger($translationParam, "sourceAddrMapped"))
			{
				if (defined $translationParam->{"sourceAddrEndMapped"})
				{
					$mappedAddrTransform = { ipAddr => _nrEnc($translationParam->{"sourceAddrMapped"},  $translationParam->{"sourceAddrEndMapped"})};
				}
				elsif (defined $translationParam->{"sourceAddrMaskMapped"})
				{
					$mappedAddrTransform = { ipAddr => _nNetEnc($translationParam->{"sourceAddrMapped"}, $translationParam->{"sourceAddrMaskMapped"})}; 					
				}
				else
				{
					$mappedAddrTransform =  { ipAddr => _hostEnc($translationParam->{"sourceAddrMapped"})}; 				
				}
				$self->_addToTranslationLedger($translationParam,"sourceAddrMapped", $mappedAddrTransform->{ipAddr});						
			}
			elsif (defined $translationParam->{"sourceAddrNameMapped"} and 
				not $self->_isInTranslationLedger($translationParam,"sourceAddrNameMapped"))
			{
				$mappedAddrTransform =  { ipAddr => _ogEnc($translationParam->{"sourceAddrNameMapped"})}; 									
				$self->_addToTranslationLedger($translationParam, "sourceAddrNameMapped", $mappedAddrTransform->{ipAddr});						
			}
			
			# See if this reverse mapping involves a PAT translation range
			if ((defined $extraDynamicPortCriterionRange) and ($translationParam->{doPortRangeMapping}) and 
					(defined $mappedAddrTransform))
			{
				$mappedAddrTransform->{service} =	_pRangeEnc( $extraDynamicPortCriterionRange->[0], 
										$extraDynamicPortCriterionRange->[1]);				
			}

			
			if (defined $mappedAddrTransform)
			{
				# Push out applied addressing assertion iff source translation is defined
				# +++wayne ugh....this is actually misplaced.  appliedAddressTranslation is defined at the
				# per-translation-address-mapping level, and the translateAppliedAddressing is at the overall
				# addressTranslation level, single instanced within the filterEntry.  It is theoretically possible
				# (you can't do this in IOS) that this could be defined on "some" addresses and not others.  What that
				# would suggest is that we'd need to segreate the translationParams (which list the formal definition of
				# the addressTranslation address set) into two lists for two filterEntries at the caller level--one list
				# for those with appliedAddressTranslation, one without. This is identical to how we deal with 
				# mixed "withPAT"/"use PAT"/"don't use PAT" specs in the same filter now.
				if (defined $translationParam->{"appliedAddressTranslation"})
				{
					$fEntry->{destAddressTranslation}->{translateAppliedAddressing} =
						$translationParam->{"appliedAddressTranslation"};
				}	

				push (@{$mappedAddrTransforms}, $mappedAddrTransform);  # for later generation as a source/dest
			}

		}
		
 		if (defined $ruleAddrTuple->{"sourceIp"})
		{
			# *if the source is 'any', then bypass the rest of this reverse generation, 
			if (_addrIsAny($ruleAddrTuple->{"sourceIp"}))
			{
				# Bypass the rest of this case,
				my $x = 0; 
			}
			else
			{
			
				# we're interested in the mapping for the address		
				# See if it's a range 
				if (defined $ruleAddrTuple->{"sourceIpEnd"})
				{
					$preMappedNetworkElt = _nrEnc($ruleAddrTuple->{"sourceIp"}, $ruleAddrTuple->{"sourceIpEnd"});				
				}
				elsif (defined $ruleAddrTuple->{"sourcePrefix"})
				{	
					$preMappedNetworkElt = _nNetEnc($ruleAddrTuple->{"sourceIp"}, $ruleAddrTuple->{"sourcePrefix"});
				}
				else
				{
					# +++wayne need to handle the "any" address case
					$preMappedNetworkElt = _hostEnc($ruleAddrTuple->{"sourceIp"});
				}
	
				# did we get anything out of that?			
				if (defined $preMappedNetworkElt)
				{
					push (@{$mappedTranslationTransform->{translationIP}}, $preMappedNetworkElt);
				}
			}
		}
		elsif (defined $ruleAddrTuple->{"sourceIpObjectGroup"})
		{
			push (@{$mappedTranslationTransform->{translationIP}}, 
				_ogEnc($ruleAddrTuple->{"sourceIpObjectGroup"}));
			
		}	

		# If we need to make this reverse map to a PAT range, we were passed it as a parameter, and it overrides anything
		# we'd find elsewhere in the ruleAddrTuple definition
		 
		if (defined $extraDynamicPortCriterionRange and defined ($mappedTranslationTransform))
		{
			push(@{$mappedTranslationTransform->{translationServicePort}},  
					{ translationPort => _pRangeEnc( $extraDynamicPortCriterionRange->[0], $extraDynamicPortCriterionRange->[1]), 
						translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});			
		} 
		elsif (defined $ruleAddrTuple->{"sourceServicePort"})
		{
			if (defined $ruleAddrTuple->{"sourceServicePortEnd"})
			{
				push(@{$mappedTranslationTransform->{translationServicePort}},  
					{ translationPort => _pRangeEnc($ruleAddrTuple->{"sourceServicePort"}, $ruleAddrTuple->{"sourceServicePortEnd"}), 
						translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});
			}
			elsif (defined $ruleAddrTuple->{sourceServicePortOperator})
			{
				push(@{$mappedTranslationTransform->{translationServicePort}},  
                                        { translationPort => _portEnc($ruleAddrTuple->{"sourceServicePort"},$ruleAddrTuple->{sourceServicePortOperator}), 
                                                translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});
			}
			else
			{
				push(@{$mappedTranslationTransform->{translationServicePort}},  
					{ translationPort => _portEnc($ruleAddrTuple->{"sourceServicePort"}), 
						translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});
			}
		}
		elsif (defined $ruleAddrTuple->{"sourceServiceObjectName"})
		{
			push(@{$mappedTranslationTransform->{translationServicePort}}, 
				{ translationPort => _ogEnc($ruleAddrTuple->{"sourceServiceObjectName"}),
					translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});
		}
		
		
		if (defined $mappedTranslationTransform)
		{
			# We have to test whether we've accumulated an addr transform, or are already stuffing $fEntry with
			# dest address translations (and maybe didn't accumulate an addr transform on this call owing to the
			# ledger state)
			if (scalar ( @{$mappedAddrTransforms}) or defined ($fEntry->{destAddressTranslation}))
			{
				push (@{$fEntry->{destAddressTranslation}->{translationAddress}}, $mappedTranslationTransform);			
			
				# Now we have to check for whether any *destination criteria* were established in the rule.
				# If they were, we can just turn them around as source criteria for this reverse mapping.
				# This is the "twice NAT" case, where there's a source *and* dest constraint, and in this 
				# case, only source NAT mapping/Translation went on on the forward direction.
				my $reverseDestMapping = undef;
				
				if ( (defined $ruleAddrTuple->{"destIp"}) and (not _addrIsAny($ruleAddrTuple->{"destIp"})) )
				{
					# So, was the dest a range
					if (defined $ruleAddrTuple->{"destIpEnd"})
					{
						$reverseDestMapping = _nrEnc($ruleAddrTuple->{"destIp"}, $ruleAddrTuple->{"destIpEnd"});		
					}	
					# or a prefix...
					elsif (defined $ruleAddrTuple->{"destPrefix"})
					{
						$reverseDestMapping =  _nNetEnc($ruleAddrTuple->{"destIp"}, 
							$ruleAddrTuple->{"destPrefix"});
					}
					# else it's just an addr			
					else
					{
						$reverseDestMapping = _hostEnc($ruleAddrTuple->{"destIp"});
					}
				}
				elsif (defined $ruleAddrTuple->{"destIpObjectGroup"})
				{
					$reverseDestMapping = _ogEnc($ruleAddrTuple->{"destIpObjectGroup"});
				}				
				
				if (defined $reverseDestMapping)
				{
					push (@{$fEntry->{"sourceIpAddr"}}, $reverseDestMapping);
				}
			} # end of check for #mappedAddrTransforms or destAddrTranslations in $fEntry
			undef $mappedTranslationTransform;					
		}
		
		
		for $mappedAddrTransform (@{$mappedAddrTransforms})
		{
			push (@{$fEntry->{"destinationIpAddr"}}, $mappedAddrTransform->{ipAddr});
			push (@{$fEntry->{"destinationService"}}, $mappedAddrTransform->{service}) if ($mappedAddrTransform->{service});	
		}
		
		for $translationParam (@{$translationParams})
		{
			# We defined the usage of the mapped address above.  See if it was port-qualified as well.
			# +++wayne I think this is breaking the address-port relationship we set up in the source/dest
			# address translation a little.  I'm hoping this doesn't have any breakage in a practical way
			# (which I'd read as "translate to this address using this port, or this other address using this 
			# other port", I don't think that sounds like a rational use case.)
			if (defined $translationParam->{"sourcePortMapped"})
			{
				if (defined $translationParam->{"sourcePortEndMapped"})
				{
					push (@{$fEntry->{destinationService}},_pRangeEnc($translationParam->{sourcePortMapped},
						$translationParam->{"sourcePortEndMapped"}));					
				}
				else
				{
					push (@{$fEntry->{destinationService}},_portEnc($translationParam->{sourcePortMapped}));
				}
			}
			elsif (defined $translationParam->{"sourcePortNameMapped"})
			{
				push (@{$fEntry->{destinationService}},_ogEnc($translationParam->{"sourcePortNameMapped"}));
			}			
		}

		# and having pushed this, let's clear it.
		$mappedAddrTransforms = [];

		if (defined $preMappedNetworkElt)
		{
			my $destFocusBit = undef;
			
			# Need to see if we have destination focused translations to worry about as well.
			
			for $translationParam (@{$translationParams})
			{
				if (defined $translationParam->{"destAddrMapped"} or
					defined $translationParam->{"destAddrNameMapped"} or
					defined $translationParam->{"destPortMapped"} or
					defined $translationParam->{"destPortNameMapped"})
				{
					$destFocusBit = $translationParam;
					last;  # leave, with translationParam defined since loop didn't iterate to end		
				}
			}
			
			if (not defined $destFocusBit)
			{
				# we're done, return here so we don't fall into 2b where we could muck things up
				return $fEntry;				
			}
		}			
		
		# if we get here, we either only or additoinally have destination focused NAT mappings, so sally forth to
		#  case 2b.
		
		# Clear out the accumulator for a locally generated network elt derived from a criterion address
		undef $preMappedNetworkElt;
			
		###### -> CASE 2b: 'reverse' generation, for 'destination' focused NAT

		# if the reverse mapping (as a source/dest criterion) has already been emitted (as reflected in the
		# ledger) then suppress its duplicate generation
		
		for $translationParam (@{$translationParams})
		{
			undef $mappedAddrTransform;
			if (defined $translationParam->{"destAddrMapped"} and
				not $self->_isInTranslationLedger($translationParam,"destAddrMapped"))
			{							
				if (defined $translationParam->{"destAddrEndMapped"})
				{
					$mappedAddrTransform = {ipAddr => _nrEnc($translationParam->{"destAddrMapped"},  
						$translationParam->{"destAddrEndMapped"})};
				}
				elsif (defined $translationParam->{"destAddrMaskMapped"})
				{
					$mappedAddrTransform = {ipAddr => _nNetEnc($translationParam->{"destAddrMapped"}, 
						$translationParam->{"destAddrMaskMapped"})};
				}
				else
				{
					$mappedAddrTransform =  { ipAddr =>_hostEnc($translationParam->{"destAddrMapped"})};				
				}
				$self->_addToTranslationLedger($translationParam, "destAddrMapped", $mappedAddrTransform->{ipAddr});									 						
			}
			elsif (defined $translationParam->{"destAddrNameMapped"}and
				not $self->_isInTranslationLedger($translationParam,"destAddrNameMapped"))
			{
				$mappedAddrTransform =  { ipAddr => _ogEnc($translationParam->{"destAddrNameMapped"})}; 									
				$self->_addToTranslationLedger($translationParam, "destAddrNameMapped", $mappedAddrTransform->{ipAddr});	
			}

			
			if (defined $mappedAddrTransform)
			{


				# Push out applied addressing assertion iff source translation is defined
				# +++wayne ugh....this is actually misplaced.  appliedAddressTranslation is defined at the
				# per-translation-address-mapping level, and the translateAppliedAddressing is at the overall
				# addressTranslation level, single instanced within the filterEntry.  It is theoretically possible
				# (you can't do this in IOS) that this could be defined on "some" addresses and not others.  What that
				# would suggest is that we'd need to segreate the translationParams (which list the formal definition of
				# the addressTranslation address set) into two lists for two filterEntries at the caller level--one list
				# for those with appliedAddressTranslation, one without. This is identical to how we deal with 
				# mixed "withPAT"/"use PAT"/"don't use PAT" specs in the same filter now.
				if (defined $translationParam->{"appliedAddressTranslation"})
				{
					$fEntry->{sourceAddressTranslation}->{translateAppliedAddressing} =
						$translationParam->{"appliedAddressTranslation"};
				}	

				push (@{$mappedAddrTransforms}, $mappedAddrTransform);
			}
		}
				
 		if (defined $ruleAddrTuple->{"destIp"})
		{
			# *if the dest is 'any', then bypass the rest of this reverse generation
			if (_addrIsAny($ruleAddrTuple->{"destIp"}))
			{
				# Return with a fairly denuded filterEntry object, indicating that no translation was 
				# brought about
				return $fEntry;
			}
			# we're interested in the mapping for the address		
			# See if it's a range 
			if (defined $ruleAddrTuple->{"destIpEnd"})
			{
				$preMappedNetworkElt = _nrEnc($ruleAddrTuple->{"destIp"}, $ruleAddrTuple->{"destIpEnd"});
			}
			elsif (defined $ruleAddrTuple->{"destPrefix"})
			{
				$preMappedNetworkElt = _nNetEnc($ruleAddrTuple->{"destIp"}, 
						$ruleAddrTuple->{"destPrefix"});
			}			
			else
			{
				$preMappedNetworkElt = _hostEnc($ruleAddrTuple->{"destIp"});
			}
			
			if (defined $preMappedNetworkElt)
			{
				push (@{$mappedTranslationTransform->{translationIP}},	$preMappedNetworkElt);
				
			}
		}
		elsif (defined $ruleAddrTuple->{"destIpObjectGroup"})
		{
			# +++wayne going to have to think more about what this means
			push (@{$mappedTranslationTransform->{translationIP}},	_ogEnc($ruleAddrTuple->{"destIpObjectGroup"}));						
		}	

		# If we need to make this reverse map to a PAT range, we were passed it as a parameter, and it overrides anything
		# we'd find elsewhere in the ruleAddrTuple definition.  Make sure we just actually generated something for it to
		# augment as well.
		if (defined $extraDynamicPortCriterionRange and defined ($mappedTranslationTransform))
		{
			push(@{$mappedTranslationTransform->{translationServicePort}}, 
					{ translationPort => _pRangeEnc( $extraDynamicPortCriterionRange->[0], $extraDynamicPortCriterionRange->[1]),
				translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});					
		}
		elsif (defined $ruleAddrTuple->{"destServicePort"})
		{
			if (defined $ruleAddrTuple->{"destServicePortEnd"})
			{
				push (@{$mappedTranslationTransform->{translationServicePort}}, 
				 {translationPort => _pRangeEnc($ruleAddrTuple->{"destServicePort"},  $ruleAddrTuple->{"destServicePortEnd"} ), 
				 	translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});
			}
			elsif (defined $ruleAddrTuple->{destServicePortOperator})
			{
				push (@{$mappedTranslationTransform->{translationServicePort}}, 
                                {translationPort => _portEnc($ruleAddrTuple->{"destServicePort"},$ruleAddrTuple->{destServicePortOperator}), 
                                        translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});
			}
			else
			{
				push (@{$mappedTranslationTransform->{translationServicePort}}, 
				{translationPort => _portEnc($ruleAddrTuple->{"destServicePort"}), 
					translationProtocol => $self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});
			}
		}
		elsif (defined $ruleAddrTuple->{"destServiceObjectName"})
		{
			push (@{$mappedTranslationTransform->{translationServicePort}}, 
			{translationPort => _ogEnc($ruleAddrTuple->{"destServiceObjectName"}), 
				translationProtocol =>$self->_effectivePortProtocol($fEntry, $ruleAddrTuple)});
		}

		for $translationParam (@{$translationParams})
		{
			# We defined the usage of the mapped address above.  See if it was dest port-qualified as well.
			if (defined $translationParam->{"destPortMapped"})
			{
				if (defined $translationParam->{"destPortEndMapped"})
				{
					push (@{$fEntry->{sourceService}},_pRangeEnc($translationParam->{"destPortMapped"}, 
						$translationParam->{"destPortEndMapped"}));
				}
				else
				{
					push (@{$fEntry->{sourceService}},_portEnc($translationParam->{"destPortMapped"}));				
				}
			}
			elsif (defined $translationParam->{"destPortNameMapped"})
			{
				push (@{$fEntry->{sourceService}},_ogEnc($translationParam->{"destPortNameMapped"}));			
			}			
		}
		
		# Generate another source address if we've accumulated an addr transform (because we acted on the translation params)
		# or we've found we're already accumulating source addresses (and maybe didn't accumulate a new addr transform because)
		# it was in the ledger
		
		if ( scalar ( @{$mappedAddrTransforms}) or defined $fEntry->{sourceAddressTranslation})
		{
			if (defined $mappedTranslationTransform)
			{
				push (@{$fEntry->{sourceAddressTranslation}->{translationAddress}}, $mappedTranslationTransform);
				undef $mappedTranslationTransform;
			}			
		}
		
		if ( scalar ( @{$mappedAddrTransforms}))
		{
			# If we've accumulated a source mapping, let's generate it into the translationAddress array
			for $mappedAddrTransform (@{$mappedAddrTransforms})
			{
				push (@{$fEntry->{"sourceIpAddr"}}, $mappedAddrTransform->{ipAddr});
				push (@{$fEntry->{"sourceService"}}, $mappedAddrTransform->{ipAddr}->{service}) if ($mappedAddrTransform->{service});			
			}

			# Now we have to check for whether any *source criteria* were established in the rule.
			# If they were, we can just turn them around as dest criteria for this reverse mapping.
			# This is the "twice NAT" case, where there's a source *and* dest constraint, and in this 
			# case, only destiination NAT mapping/Translation went on on the forward direction.
			my $reverseSrcMapping = undef;
			
			if ( (defined $ruleAddrTuple->{"sourceIp"}) and (not _addrIsAny($ruleAddrTuple->{"sourceIp"})) )
			{
				# So, was the dest a range
				if (defined $ruleAddrTuple->{"sourceIpEnd"})
				{
					$reverseSrcMapping = _nrEnc($ruleAddrTuple->{"sourceIp"}, $ruleAddrTuple->{"sourceIpEnd"});		
				}	
				# or a prefix...
				elsif (defined $ruleAddrTuple->{"sourcePrefix"})
				{
					$reverseSrcMapping =  _nNetEnc($ruleAddrTuple->{"sourceIp"}, 
						$ruleAddrTuple->{"sourcePrefix"});
				}
				# else it's just an addr			
				else
				{
					$reverseSrcMapping = _hostEnc($ruleAddrTuple->{"sourceIp"});
				}
			}
			elsif (defined $ruleAddrTuple->{"sourceIpObjectGroup"})
			{
				$reverseSrcMapping = _ogEnc($ruleAddrTuple->{"sourceIpObjectGroup"});
			}				
			
			if (defined $reverseSrcMapping)
			{
				push (@{$fEntry->{"destinationIpAddr"}}, $reverseSrcMapping);
				
				# Check for port here....starting with the "extra" mapping
				
				if (defined $extraDynamicPortCriterionRange)
				{
					push (@{$fEntry->{destinationService}},_pRangeEnc( $extraDynamicPortCriterionRange->[0], 
										$extraDynamicPortCriterionRange->[1]));													
				}
				elsif (defined $ruleAddrTuple->{"sourceServicePort"})
				{
					if (defined $ruleAddrTuple->{"sourceServicePortEnd"})
					{
						push (@{$fEntry->{destinationService}},_pRangeEnc($ruleAddrTuple->{"sourceServicePort"},
							$ruleAddrTuple->{"sourceServicePortEnd"}));
					}
					elsif (defined $ruleAddrTuple->{"sourceServicePortOperator"})
					{
						push (@{$fEntry->{destinationService}},_portEnc( $ruleAddrTuple->{"sourceServicePort"},
							$ruleAddrTuple->{"sourceServicePortOperator"}));
					}
					else # simple port spec
					{
						push (@{$fEntry->{destinationService}},_portEnc( $ruleAddrTuple->{"sourceServicePort"}));
					}
				}
				elsif (defined $ruleAddrTuple->{"sourceServiceObjectName"})
				{
					push (@{$fEntry->{destinationService}},_ogEnc($ruleAddrTuple->{"sourceServiceObjectName"}));	
				}
			}
		}
	}
	return $fEntry;
}

sub _filterEntryHasTranslations
{
	my ($self, $filterEntry) = @_;
	if (defined $filterEntry->{sourceAddressTranslation})
	{
		return (defined $filterEntry->{sourceAddressTranslation}->{translationAddress});
	}
	elsif (defined $filterEntry->{destAddressTranslation})
	{
		return (defined $filterEntry->{destAddressTranslation}->{translationAddress});
	}
	else { return 0; }
}

sub _addrIsAny
{
	my $addr = shift;
	return ($addr ~~ "any" or $addr ~~ get_any_address_address());
}

sub _clearRuleAddrFields
{
	my ($self, $fEntry) = @_;
	
	delete $fEntry->{"sourceIpAddr"};
	delete $fEntry->{"destinationIpAddr"};
	delete $fEntry->{"destinationService"};
	delete $fEntry->{"destAddressTranslation"};
	delete $fEntry->{"protocol"};
	delete $fEntry->{"sourceAddressTranslation"};
	delete $fEntry->{"sourceService"};
	delete $fEntry->{"established"};
	$self->{"_addrTranslationLedgerMap"} = {};	
}

# Routines to simply do enveloping of frequent scalars into tagged hashes based
# on well-known keys.  Also handle 'any' host translation here.
sub _hostEnc
{
	my $hAddr = shift;
	# if the host is encoded as 'any', deal wit
	if ($hAddr ~~ "any")
	{
		my $anyElt = get_any_address_encoding();
		return ( {network => $anyElt} );
	}

	return { host => $hAddr};
}

sub _nNetEnc
{
	my ($nAddr, $nMask) = @_;
	return {network => create_network_element($nAddr, $nMask)};	
}

sub _portEnc
{
	my $sPort = shift;
	my $sOper = shift;
	return { portExpression => {port => $sPort, operator => $sOper // 'eq'}};
}

sub _pRangeEnc
{
	my ($sPortStart, $sPortEnd) = @_;
	return ( { portRange => { portStart => $sPortStart, portEnd => $sPortEnd }} );
}	

sub _ogEnc
{
	my $ogName = shift;
	return { objectGroupReference => $ogName };
}

sub _nrEnc
{
	my ($aStart, $aEnd) = @_;
	return { range =>
			{ startAddress => $aStart, endAddress => $aEnd}
		
	};
}

# This returns a list of sets of entry addrs--each one corresponding to another set which need to be uniquely
# generated with another instance of the rule params in genRule.
sub _doExpansionOfRuleEntryAddrs
{
	my ($self, $rule) = @_;
	
	# if this for some reason has no address criteria, then scram
	if (not $rule->{addrCriteria})
	{
		return 0;
	}
	
	my $runningAddrDisposition = $rule->{addrCriteria}->[0]->{negateMatch};
	my $startingAddrList = $rule->{addrCriteria};
	my $runningSPort = $self->_effectivePort($startingAddrList->[0], "source");
	my $runningDPort = $self->_effectivePort($startingAddrList->[0], "dest");
	my $reformattedAddrList;
	my $reformattedAddrListSet = undef;
	
	for my $thisEntryAddr (@{$startingAddrList})
	{
		# see if we can just copy this across
		if (($thisEntryAddr->{negateMatch} == $runningAddrDisposition ) &&
			( $self->_effectivePort($thisEntryAddr, "source") ~~ $runningSPort) &&
			( $self->_effectivePort($thisEntryAddr, "dest") ~~ $runningDPort))
		{
			push (@{$reformattedAddrList}, $thisEntryAddr);
		}
		else
		{
			# store this list created to this point
			if ($reformattedAddrList)
			{
				push (@{$reformattedAddrListSet}, $reformattedAddrList);
				$reformattedAddrList = [];
				$runningAddrDisposition = $thisEntryAddr->{negateMatch};
				$runningSPort = $self->_effectivePort($thisEntryAddr, "source");
	 			$runningDPort = $self->_effectivePort($thisEntryAddr, "dest");
	 			push (@{$reformattedAddrList}, $thisEntryAddr);
	 			
			}
		}
		
	}
	push(@{$reformattedAddrListSet},  $reformattedAddrList);
	return $reformattedAddrListSet;
}

# This returns either a port range or port object name from a criterion address attached to (presumably)
# a rule
sub _effectivePort
{
	my ($self, $addrCrit, $direction) = @_;
	
	if (exists $addrCrit->{$direction . "ServicePortEnd"})
	{
		# return an encoding fo source/dest
		return $addrCrit->{$direction . "ServicePort"} . "_" .$addrCrit->{$direction . "ServicePortEnd"}; 
	}
	else
	{
		return ( $addrCrit->{$direction . "ServicePort"} // ($addrCrit->{$direction . "ServiceObjectName"} // 0));
	}	
}

# This dredges up or defaults a protocol to assign to the addressTranslation port value.
# This has to be tcp or udp, you see....
#

sub _effectivePortProtocol
{
	my ($self, $filterEntry, $critRule) = @_;
	
	if ((defined $critRule->{ipProtocol}) && ($critRule->{ipProtocol} =~ /^tcp|udp|ip$/))
	{
		return $critRule->{ipProtocol};
	}
	elsif ((defined $filterEntry->{protocol}) && ($filterEntry->{protocol} =~ /^tcp|udp|ip$/))
	{
		return $filterEntry->{protocol};
	}
	else
	{
		# A basic default, for now, particularly for those NAT rules which only target TCP
		# in key stakeholder applications anyways.  Still, this is a crummy coercion.
		return "tcp";
	}
}

# We internally build up, over the course of a filter entry generation, state for the translation addresses
# (either source or dest, only one of those will be in play for any filter entry) already generated into 
# the entry, so as not to generate them multiple times across the evaluation of multiple rule addressing
# criteria.
#
#  The idea here is that a single key can map to a simple value (say, an object group reference pointing to 
#  itself) or a more complex hash mapping (say, and IP address pointing to a range).  They won't ever be too
#  complex, just addr/mask networks or range encodings.

sub _addToTranslationLedger
{
	my ($self, $translationParam, $translationKey, $translationMapping) = @_;
	my $valOfTranslKey = $translationParam->{"$translationKey"};
	
	$self->{"_addrTranslationLedgerMap"}->{"$translationKey"}->{"$valOfTranslKey"} = $translationMapping;
}

sub _isInTranslationLedger
{
	my ($self, $translationParam, $translationKey, $comparisonValue) = @_;
	my $foundInLedger = 0;
	my $ledgerValue;
	my $valOfTranslKey = $translationParam->{"$translationKey"};
	
	if (defined $self->{"_addrTranslationLedgerMap"})
	{
		$ledgerValue = $self->{"_addrTranslationLedgerMap"}->{"$translationKey"}->{"$valOfTranslKey"} // 0;
	
		if ($ledgerValue and defined $comparisonValue) # they want us to do a comparison
		{
			# This is a good Perl 5.10 technique for a lightweight hash comparison.  It won't handle,
			# e.g., assignment of values to different keys, but the comparisons here are simple across
			# a single value.  For a deeper compare, compare the results of a Data::Dumper or use a 
			# great CPAN module called FreezeThaw that I don't feel like going through the legal issues
			# to use right now.
			#
			$foundInLedger = ( %{$ledgerValue} ~~ %{$comparisonValue} && 
				[ sort values %{$ledgerValue}] ~~ %{$comparisonValue});
		}	
		else
		{
			$foundInLedger = $ledgerValue; # something nonzero, or not
		}		
	}
	
	return $foundInLedger;
}

=head1 NAME

    ZipTie::Model::AddressTranslation - Perform Ziptie translation for NAT and PAT 

=head1 VERSION

    This document refers to version 1.0 of ZipTie::Model::AddressTranslation

=head1 SYNOPSIS

  # Short examples of Perl code that illustrate the use of the class

=head1 DESCRIPTION

=head2 Overview

    This module provides a means of incrementally constructing and then 
    generating the FilterList corresponding to a discrete set of address
    translation rules.  The current implementation will create up to two
    <filterList> elements denoting mapping; one for the "forward" sense of
    the NAT rule ('source' or 'destination') across a set of interfaces comprising
    a zone, the other for "reverse" of that packet flow.  That symmetry can be
    overridden using SetSrc/DstTranslationOnly() and for TCP protocol traffic 
    flow, setting SetEstablishedOnly().

=head2 Constructor and initialization

=over

=item ZipTie::Model::AddressTranslation(sourceZone, destZone)

    sourceZone : zone name for source of translation, or "all"
    destZone : zone name for destination of translation, or "all"

=back

=head2 Class and object methods

=over

=item  setInterfaceZoneMapper(izMapper)

     izMapper : map which keys a "zone", by name, to an array of interfaces comprising that zone.
     
=item   setSourceTranslationOnly()

      sets the object to only be doing "source" translation ZED element generation    

=item   setDstTranslationOnly()

      sets the object to only be doing "destination" translation ZED element generation    

=item  setTranslationName(name)

      name - something to be set which will be used to override naming field of generated filterLists.

=item  generateTranslationList(outXML, direction, nonDefaultListName)

	  outXML - XMLPrinter instance which should be invoked for element generation
	  direction - MUST be "source" or "destination", will tailor filterList(s) output
	        to just that direction of translation.  Considered in combination with source and dest translation
	        properties of object.  The general model is: the caller calls to generate the list(s) in the context
	        of a filterListSection which generates for "destination", caller then generates other filters 
	        (most usually, packet filtering rules), then generates for "source".  ***THIS USAGE PATTERN 
	        DEPENDS ENTIRELY ON THE PACKET PROCESSING PHASE MODEL ASSOCIATED WITH THE CONFIGURATION REGIME
	        WHOSE SED IS BEING GENERATED** however.
	  nonDefaultListName - optional, and only to be used in a specific circumstance where the caller wishes to
	  	cirumvent the typical "two call" model described in the description of the preceding parameter.  This
	  	will differentiate the typical "_forward" and "_reverse" suffixes put on the generated list (depending
	  	on 'direction'). 

=item  generateTranslationList(outXML, direction)

	  outXML - XMLPrinter instance which should be invoked for element generation
	  direction - (optional) if specified, MUST be "source" or "destination", will tailor filterList(s) output
	        to just that direction of translation.  Considered in combination with source and dest translation
	        properties of object.  

=item  addPortTranslationRange ( $mappingRange )

       $mappingRange : a (low,high) tuple, creating a subset for like-port mapping in PAT/"overload" situations to
          ensure that ports of this range on a translation-source gets set to the same range on a translation-dest.
          Lacking any of these, the entire port map range will be mapped.  If the port range (0,65535) is not 
          completely mapped by subsequent calls to this, ports *outside* the mapped range will be dropped for 
          translation ZED generation.
          
          As the interface has evolved, you may see strange things if you *don't* put together a port translation 
          range.  In particular, reverse rules with PAT enabled *won't* be generated.  Moreover, consider the 
          topology semantics--if I have destination NAT being imposed at the firewall, and I have an outgoing 
          connection attempt from (dynamic) port 5186, do you really want to have topology think it's possible
          to have mapped the destination port for the return SYN-ACK packet to port 80?
          	  
=item  addRule({ param_set })
      where param_set has key/value pairs:

      sourceAddrMapped - address mapping for source translation
      sourcePortMapped - port specific mapping for source translation
      sourcePortNameMapped - an object group comprising source/port reference for mapping in source translation
      destAddrMapped - address mapping for destination translation
      destPortMapped - port specific mapping for destination translation
      destPortNameMapped - an object group comprising destination/port reference for mapping in destination translation
      protocol - IP bearer protocol (e.g., 'tcp', 'udp', 'rtp') of IP-borne protocol.
      doPortRangeMapping - if "1", enforce rules for port range mapping
      establishedReverseConnectionTrafficOnly - Only set translation for traffic over which established port connections 
             are already in place.
      isStaticNATMapping - Indicates that this is a static NAT mapping.  The end result of that, as of this implementation,
      	is to set 'establishedReverseConnectionTrafficOnly' on the rule IF this isn't set (i.e., it is a dynamic NAT mapping).
      
	  RETURNS: source "handle", which can be passed subsequently into addRuleAddressCriterion(),  setExemptionRule(), et al.

=item  setExemptionRule{rule)
	
		where 'rule' is a rule returned from addRule.  Sets the rule to be an "exemptiom" rule, to logically override
		conditions set in any other rule established within the AddressTranslation governing broader conditions/
		mappings.
		
		value - if 1 or "true", will enable for this rule, if 0 or "false" will *disable* for this rule
		
=item  setEstablishedReverseConnectionTrafficOnly(rule, value)

		where 'rule' is a rule returned from addRule.  This allows a rule, after being allocated, to only allow translation on
		traffic in the "reverse" direction relative to the main <sourceZone> -> <destZone> packet path on packets on "established"
		connections--in other words, no translation would occur on TCP packets sent with the SYN bit set in the TCP
		header.  If 'value' is 1 or 'true', this asymmetry will be marked on the traffic path.  If it is zero or 'false', it will 
		be cleared regardless of current setting.  Unless this is set using this interfce, a setting in addRule, or indirectly via
		the 'isStaticNATMapping' flag of addRule, then the default state will be symmetric capability for transport connection 
		establishment viability to the (reverse-)translated address.
				
=item  addRuleAddressCriterion({ param_set })

		Except for cases where all packets generated from a 'sourceZone' (from the time of addressTranslation
		object instance creation) to optional 'destZone' are to be mapped, generally, some address criteria will
		need to be set. some address criteria need to be established on each rule.  Established by 
		semi-mutally-exclusive sets of named parameters in 'param_set', those would be:
		
		rule - rule handle, as returned from addRule() .  REQUIRED.
		sourceIp - source IP address of primary packet flow
		sourcePrefix - prefix governing network, with sourceIp (REQUIRED if this is specified) as a starting address.
		sourceIpEnd - end of range whose start being sourceIp (REQUIRED if this is specified) as a start.  Would
		     be used independently of 'sourcePrefix'.
		sourceIpObjectGroup - named object group for source address attributes.  Would be used indepedently of 
			sourceIp, sourcePrefix, or sourceIpEnd.
		sourceServicePort - port within governing protocol for scoping of 'protocol' provided at addRule() time.
		sourceServicePortEnd - end of range whose start being sourceServicePort (REQUIRED if this is specified).
		sourceServiceObjectName - named object group for source service/port attributes.  Would be used indepedently of 
			sourceServicePort or sourceServicePortEnd.
		destIp - source IP address of primary packet flow
		destPrefix - prefix governing network, with destIp (REQUIRED if this is specified) as a starting address.
		destIpEnd - end of range whose start being destIp (REQUIRED if this is specified) as a start.  Would
		     be used independently of 'destPrefix'.
		destIpObjectGroup - named object group for destination address attributes.  Would be used indepedently of 
			destIp, destPrefix, or destIpEnd.
		destServicePort - port within governing protocol for scoping of 'protocol' provided at addRule() time.
		destServicePortEnd - end of range whose start being destServicePort (REQUIRED if this is specified).
		destServiceObjectName - named object group for destination service/port attributes.  Would be used indepedently of 
			destServicePort or destServicePortEnd.
		negateMatch - if set with "1", this denotes that the rest of the parameters here are specifying a criterion
			to trigger on the "non-matching" of (e.g., a "deny" rule sense)
	 
=item  resetForNewFiltersSection()

	   Indicates that upon the next call to generateTranslationList, the generation within a new containing 
	   <filterLists> or <filterListsSection> should be construed.
	  	  
=back   

=head2 Lifecycle

=over

=item * Construct object
=item * Set zone translation vector
=item * Set attributes on object if desired
=item * Add rules
=item * Set attributes/overrides on rules if desired
=item   Call addRuleAddressCriterion() to establish address criteria for the translation.
=item * call generateTranslationList to do ZED generation of <filterList>'s for translation

=back

=head1 ENVIRONMENT

    List of environment variables and other O/S related information
    on which the class relies

=head1 DIAGNOSTICS

  None, really.  As a utility object, this module does not assume the 
  access to a logger, etc.

=head1 BUGS

    Description of known bugs (and any work-arounds).
    Usually also includes an invitation to send the author bug reports.

=head1 FILES

    List of any files or other Perl modules needed by the class and a brief
    explanation why.

=head1 SEE ALSO

    http://wiki.q1labs.inc/display/QRM/Address+Translation+Support+in+ZipTie+and+CSM

=head1 AUTHOR(S)
    Wayne Tackabury
    wtackabury@us.ibm.com

=head1 COPYRIGHT
    Copyright (c) 2012, IBM Corporation. All Rights Reserved.
    This module is free software. It may be used, redistributed
    and/or modified under the same terms as Perl itself.

=cut

1;
