package ZipTie::Model::XmlPrint;

use strict;
use warnings;
use Carp;
use Data::Dumper;
use XML::Simple qw(XMLout);
use MIME::Base64 qw(encode_base64);
use ZipTie::Model::ZiptieSignatureAssembly;

# Create the object
sub new
{
	my ( $class, $filehandle, $modeltype, $otherXsd ) = @_;
	my $self = {
		_filehandle => $filehandle,
		_modeltype  => $modeltype,
		_attributes => 0,
		_otherXsd   => ( $otherXsd or '' ),
		_signedTagStack => [],
		_isBeingSigned => 0,
	};
	bless $self, $class;
	return $self;
}

#------------------------ Subroutines -------------------------

sub attributes
{
	my $self     = shift;
	my $newValue = shift;
	if ( defined $newValue )
	{
		$self->{_attributes} = $newValue;
	}
	return $self->{_attributes};
}

sub open_discovery_event
{
	my ($self) = @_;
	my $FILEHANDLE = $self->{_filehandle};
	print $FILEHANDLE <<'END';
<?xml version="1.0" encoding="UTF-8"?>
<DiscoveryEvent xmlns="http://www.ziptie.org/model/telemetry/1.0"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
    xmlns:telemetry="http://www.ziptie.org/model/telemetry/1.0"
    xsi:schemaLocation="http://www.ziptie.org/model/telemetry/1.0 telemetry.xsd">
END
}

# Starts the XML document with the appropriate tags
sub open_model
{
	my ($self)     = @_;
	my $FILEHANDLE = $self->{_filehandle};
	my $modeltype  = $self->{_modeltype};
	my $otherXsd   = $self->{_otherXsd};
	print $FILEHANDLE '<?xml version="1.0" encoding="UTF-8"?>';

	if ( $modeltype eq 'core' )
	{
		print $FILEHANDLE '<ZiptieElementDocument xmlns="http://www.ziptie.org/model/core/1.0"';
	}
	elsif ( $modeltype ne 'common' )
	{
		print $FILEHANDLE '<' . $modeltype . ':ZiptieElementDocument xmlns="http://www.ziptie.org/model/common/1.0"';
	}
	else
	{
		print $FILEHANDLE '<ZiptieElementDocument xmlns="http://www.ziptie.org/model/common/1.0"';
	}

	print $FILEHANDLE ' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"';

	if ( $otherXsd =~ /^(\S+)\s+(\S+)\.xsd$/ )
	{
		print $FILEHANDLE ' xmlns:' . $2 . '="' . $1 . '"';
	}

	print $FILEHANDLE ' xmlns:core="http://www.ziptie.org/model/core/1.0"';
	print $FILEHANDLE ' xmlns:common="http://www.ziptie.org/model/common/1.0"' if ( $modeltype ne "core" );
	print $FILEHANDLE ' xmlns:ztApp="http://www.ziptie.org/model/ziptieApplication/1.0"' if ( $modeltype ne "core" );
	print $FILEHANDLE ' xsi:schemaLocation="http://www.ziptie.org/model/core/1.0 ziptie-core.xsd';
	print $FILEHANDLE ' http://www.ziptie.org/model/common/1.0 ziptie-common.xsd' if ( $modeltype ne "core" );
	print $FILEHANDLE ' http://www.ziptie.org/model/ziptieApplication/1.0 ziptie-applications.xsd' if ( $modeltype ne "core" );
	printf $FILEHANDLE ( ' %s', $otherXsd ) if ($otherXsd);
	print $FILEHANDLE '">';
	print $FILEHANDLE "\n";
}

# Closes the XML document tags
sub close_model
{
	my ($self)     = @_;
	my $FILEHANDLE = $self->{_filehandle};
	my $modeltype  = $self->{_modeltype};
	if ( $modeltype eq 'core' || $modeltype eq 'common' )
	{
		print $FILEHANDLE "</ZiptieElementDocument>\n";
	}
	else
	{
		print $FILEHANDLE "</$modeltype:ZiptieElementDocument>\n";
	}
}

# prints out just the opening tag of an element
sub open_element
{
	my ( $self, $tag ) = @_;
	my $FILEHANDLE = $self->{_filehandle};
	print $FILEHANDLE "<$tag>\n";
}

# attaches a signature to a (presumably, just opened) element
sub attach_signature_assembly
{
	my ($self, $tag, $sigAttr) = @_;
	$self->{_isBeingSigned}++;  # Bump "depth" of signature "stack"
	my $newSa = ZipTie::Model::ZiptieSignatureAssembly->new();
	
	push ( @{$self->{_signedTagStack}}, 
		{ tag => $tag,
		  sigAttr => $sigAttr,   # The value that is supposed to hold the signature at close_element time
		  sa => $newSa});
		  
	# Need to generate the tag into the signature assembly--it's already in the document
	$newSa->AddPayload("<$tag>\n");
}

# prints out just the closing tag of an element
sub close_element
{
	my ( $self, $tag ) = @_;

	my $FILEHANDLE = $self->{_filehandle};
	
	if ($self->{_isBeingSigned})
	{
		my $xpSaNode = $self->_peek_current_signature();
		if ($xpSaNode->{tag} eq $tag)
		{
			# Add closing tag into payload and document.  This *just* wants to go into 
			# the signature buffer, prior to its actual generation into the document.
			$xpSaNode->{sa}->AddPayload("</$tag>\n");
			
			# As it is now closed, we can generate the signature element.  Don't put it into 
			# signatures, however.
			$self->_print_to_xml_targets($xpSaNode->{sa}->GenerateSignature(),
				$xpSaNode->{sigAttr},
				{suppressSignaturePrint => 1} );
			$xpSaNode = pop(@{$self->{_signedTagStack}});
			
			# force destruct on sigassembly object
			$xpSaNode->{sa} = undef;
			$self->{_isBeingSigned}--;  # Decrement signing count			
		}
	}
	
	# Now, put the closing tag into the remaining stream
	print $FILEHANDLE "</$tag>\n";
}

# Print out a generic element.  This should only be used when there isn't a
# specific method to print out the element.  E.g. ->printChassis()
sub print_element
{
	my ( $self, $tag, $element, $args ) = @_;
	my $suppressFromSignatureAssembly = $args->{suppressFromSignatureAssembly};
	my $type = ref($element);
	if ( $tag eq "cisco:interface" )
	{
		# I haven't put this into the signature assembly logic yet, but we don't
		# sign interfaces anyways.
		$self->_print_cisco_interface( $element, $tag );
	}
	elsif ($type eq 'HASH' && !$self->{_attributes})
	{
		if ($tag eq 'core:configRepository' || $tag eq 'core:folder')
		{
			my $repoName = $element->{'core:name'};
			if ($repoName)
			{
				$element->{'-attributes'} = { 'name'=>$repoName, };
				delete $element->{'core:name'};
			}
		}
		
		my $openElement = $tag;
		if ($element->{'-attributes'})
		{
			foreach my $attribute (keys %{$element->{'-attributes'}})	
			{
				$openElement .= ' '.$attribute.'="'.$element->{'-attributes'}->{$attribute}.'"';
			}
		}
		$self->open_element($openElement);
		foreach my $key (sort keys %$element)		
		{
			if ($key !~ /^-/)
			{
				$self->print_element($key, $element->{$key}, 
					{suppressFromSignatureAssembly => $suppressFromSignatureAssembly});
			}
		}
		$self->close_element($tag);
	}
	elsif ($type eq 'ARRAY' && !$self->{_attributes})
	{
		my $openElement = $tag;
		my @array = @{ $element };
		foreach my $subElement (@array)		
		{
			$self->print_element($tag, $subElement, 
				{suppressFromSignatureAssembly => $suppressFromSignatureAssembly});
		}
	}
	else
	{
		$self->_print_to_xml_targets( $element, $tag, 
			{suppressSignaturePrint => $suppressFromSignatureAssembly});
	}
}

# This is used to print pre-generated XML stored in a string to the XML document.
sub print_string
{
	my ($self, $string) =@_;

	if (!defined($string))
	{
		carp 'String not defined when calling print_string';
		return;
	}

	my $FILEHANDLE = $self->{_filehandle};

	print $FILEHANDLE $string;
}

# +++wayne do we need this anymore??
sub _print_cisco_interface
{

	# prints out a cisco interface, which may have an eigrp property
	# that needs to come last to match the XSD
	my ( $self, $element, $tag ) = @_;
	my $FILEHANDLE = $self->{_filehandle};

	if ( defined $element->{"cisco:eigrp"} )
	{
		print $FILEHANDLE "<" . $tag . ">";
		foreach my $key ( sort keys %{$element} )
		{
			if ( $key ne "cisco:eigrp" )
			{
				if(ref($element->{$key}) eq 'ARRAY'){
					foreach my $array_element(@{$element->{$key}}){
						_print_to_xml_targets($self, $array_element, $key);
					}
				}else{
					_print_to_xml_targets( $self, $element->{$key}, $key );
				}
			}
		}
		_print_to_xml_targets( $self, $element->{"cisco:eigrp"}, "cisco:eigrp" );
		print $FILEHANDLE "</" . $tag . ">";
	}
	else
	{
		_print_to_xml_targets( $self, $element, $tag );
	}
}

# +++waynet these comments predate me, they're totally wrong.
#
# Inputs:
#	1 - the filehandle reference
#	2 - the hash element
#   3 - the root name
#
# Uses XMLout to print out pretty XML will fully escaped special characters.
sub _print_xml_element
{
	my ( $self, $element, $rootName, $keyattr ) = @_;
	my $filehandle = $self->{_filehandle};

	if (defined $element)
	{
		if ( !defined $keyattr )
		{
			$keyattr = { 'core:configRepository' => '+core:name', 'core:folder' => '+core:name', };
		}
		print $filehandle XMLout( $element, RootName => $rootName, keyattr => $keyattr, noattr => ($self->{_attributes}) ? 0 : 1,);		
	}
	else  # Otherwise, it's just an empty tag
	{
		print $filehandle XMLout( undef, RootName => $rootName );
	}
}

# Simply provide a reference to the "current" signature node.  Since it's in an array stack, it's at
# end of array.
sub _peek_current_signature
{
	my $self = shift;
	return ($self->{_isBeingSigned}) ? $self->{_signedTagStack}->[$self->{_isBeingSigned} - 1]  : undef;
}

# Print to all targets in the current set
#
sub _print_to_xml_targets
{
	my ( $self, $element, $rootName,$args ) = @_;
	my $suppressSignaturePrint = $args->{suppressSignaturePrint};
	my $keyattr = $args->{keyattr};  # +++waynet couldn't find a place this was actually used.

	# We wish to print through to all current signature sources, plus the actual XML
	# Printer filehandle
	if (not $suppressSignaturePrint)
	{
		for my $nextSigDest (@{$self->{_signedTagStack}})
		{
			# See if we have an actual tag/element pairing	
			if (defined $element)
			{
				# Check for suppression in the ZiptieSignatureAssembly itself
				if (not $nextSigDest->{sa}->IsSubtreeException($rootName))
				{
					$nextSigDest->{sa}->AddPayload(
						XMLout( $element, RootName => $rootName, keyattr => $keyattr, 
						noattr => ($self->{_attributes}) ? 0 : 1,)
						);													
				}
			}
			else
			{
				$nextSigDest->{sa}->AddPayload($rootName . "\n");
			}
		}			
	}
	# Now for the real print
	_print_xml_element($self, $element, $rootName, $keyattr);
}
1;

__END__

=head1 NAME

ZipTie::Model::XmlPrint - Utility Module to Print XML Elements from Perl Scalars

=head1 SYNOPSIS

	use ZipTie::Model::XmlPrint;
	my $printer = ZipTie::Model::XmlPrint::new(\*STDOUT, "common");
	$printer->open_model();
	$printer->print_element( "snmp", $snmpHash );
	$printer->close_model();

=head1 DESCRIPTION

Create the XmlPrint class with a given FILEHANDLE.  Then use the class to help
print elements from the ZipTie model.  Your elements should be small but well formed
hashtables that match the ZipTie model schema.

This modules uses XML::Simple for printing elements so the values all will have any
special XML characters escaped properly.

If you want to mix attributes with elements, you can do one of two things.

1) Any element should be pushed onto an array and any attribute should be simply set.
   You must also set $printer->attributes(1) for this to work.
   
2) Put a hash onto a hash element at '-attributes' and those elements will be
   printed out as attributes instead of elements.  For example:
   
   my $xml = {};
   $xml->{log} = 'true';
   $xml->{'-attributes'} = {name=>'entry1', state=>'active',};
   $printer->print_element('entry', $xml);
   
   will yeild....
   
   <entry name="entry1" state="active">
   		<log>true</log>
   </entry>


=head2 Methods

=over 12

=item C<new($filehandle, $targetNamespace, $extraSchema)>

Create an instance of this object with a provided filehandle.  This filehandle can be
an open filehandle for a real file, or simply \*STDERR or \*STDOUT.

INPUTS:
	$filehandle - an open filehandle reference, such as \*STDOUT
	$targetNamespace - this is usually 'commom'
	$extraSchema - optional argument if you are writing to an extended schema
			
EXAMPLES:
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, 'common', "http://www.ziptie.org/model/cisco/1.0 cisco.xsd" );
	my $printer = ZipTie::Model::XmlPrint->new( $filehandle, "common" );

=item C<attributes($bool)>

If no arguments are given, this method simply returns the current value of 'attributes'.  By passing a '1' to this method
you will turn on printing of elements with attributes for values instead of nested children, which is the default.

=item C<open_model()>

Prints out the root XML element of the ZipTie model.  After all printing is done you
should also call the C<close_model> method;

=item C<open_discovery_event()>

Prints out the root XML element of the DiscoveryEvent (telemetry.xsd) model.  Be sure to manually close the element
for the </DiscoveryEvent>.

=item C<open_element($name)>

Print out just the beginning of an element.
e.g. $printer->open_element("chassis");

which will print out "<chassis>"

=item C<print_element($name, $elementHash)>

Prints out an XML element, deep or shallow, given a hashtable.
$printer->print_element($filterEntry);

=item C<close_element($name)>

Print out just the end of an element.
e.g. $printer->close_element("chassis");

which will print out "</chassis>"

=item C<close_model()>

Prints out just the closing tag of the ZipTie model.
e.g. "</ZiptieElementDocument>"

=back

=head1 LICENSE

  The contents of this file are subject to the Mozilla Public License
  Version 1.1 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/
  
  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
  License for the specific language governing rights and limitations
  under the License.
 
  The Original Code is Ziptie Client Framework.
  
  The Initial Developer of the Original Code is AlterPoint.
  Portions created by AlterPoint are Copyright (C) 2006,
  AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

  Contributor(s): rkruse
  Date: Apr 23, 2007

=cut
