package ZipTie::Model::XmlUtils;

use strict;
use warnings;

use Exporter 'import';

use ZipTie::Logger;
use Digest::MD5 qw(md5_hex);

our @EXPORT_OK = qw(escapeXMLString generateSignatureData);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

our @XML_ESCAPE_TABLE = (
	[ "\&", "&amp;"],
	[ "<", "&lt;"],
	[ ">", "&gt;"],
	[ "\"", "&quot;"],
	[ "'", "&apos;"],
);

sub escapeXMLString {
	# Note that anything which wishes to take adavantage of this has to be in 
	# a XML 1.1 document, where escape sequences beyond &amp and &lt and the like
	# were defined.
	
	my $escSource = shift;
	my $escPost = $escSource;
	my $xfEscSource = "";
	my $xfEscDest = "";
	
	foreach my $xTransform (@XML_ESCAPE_TABLE) {
		my @xTransformTuple = @{$xTransform};
		# capture text in case it's matched in the output
		$escPost =~ s/($xTransformTuple[0])/$xTransformTuple[1]/g;
	}	
	
	# Transform escapes.  I tried putting this into @XML_ESCAPE_TABLE, couldn't get it to be quite
	# interpolated properly with the embedded function even with the 'e' diretive.  Oh well....I'll
	# figure that out someday as a trivia question, for now, a second transformation...
	$escPost =~ s/([\x01-\x08\x0b\x0c\x0e-\x1f\x7f-\xff])/"&#x0" .  unpack("H2", pack("C", ord($1))) . ";"/eg;
	
	return $escPost;		
}

sub generateSignatureData {
	my $sigSource = shift;
	my $rawSignatureContent = md5_hex($sigSource);
	my $sigBuf = "";
	
	# Now, we have to combine the first half and the second half of the raw signature via xor,
	# byte by byte to create a 64-bit result from the 128-bit raw signature.
	for (my $i = 0; $i < length($rawSignatureContent)/2; $i += 2) {
   		my $sigByte = hex(substr($rawSignatureContent, $i, 2)) ^  
   			hex(substr($rawSignatureContent, $i + (length($rawSignatureContent)/2), 2));
   		$sigBuf .= sprintf("%02x", $sigByte);
	}
	return $sigBuf;
}
	
1;

__END__

=head1 NAME

ZipTie::Model::XmlUtils - Transformational and processing utilities for payload destined for ZipTie

=head1 SYNOPSIS

    use ZipTie::Model::XmlUtils;
	my $escapedString = escapeXMLString("this string & the horse it \x1B escaped on");
	my $ztSig = generateSignatureData("<signedFields>Payload of signed fields</signedFields>")

=head1 Description
   
   This module exposes certain utilities for, mostly transforming payload, in ZED or other ZipTie 
   XML documents.  It can be used for any routine which can exist at non-object namespace scope
   which requires some sort of processing on its XML context or content.
   
   
=head2 Methods

=over 12

=item C<escapeXMLString>   
	Does escaping of an input string, based on fields which require XML 1.1 escaping sequences.

=item C<generateSignatureData>   
	Creates a 64-bit "signature", usable as an sparse index across fields in the ZED.

=back

=head1 LICENSE

  The contents of this file are subject to the Mozilla Public License
  Version 1.1 (the "License"); you may not use this file except in
  compliance with the License. You may obtain a copy of the License at
  http://www.mozilla.org/MPL/
  
  Software distributed under the License is distributed on an "AS IS"
  basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
  License for the specific language governing rights and limitations
  under the License.
 
  The Original Code is Ziptie Client Framework.
  
  The Initial Developer of the Original Code is Q1 Labs.
  Portions created by AlterPoint are Copyright (c) 2011,
  Q1 Labs, Inc. All Rights Reserved.

=head1 AUTHOR

  Contributor(s): wayne.tackabury, Q1Labs Inc.
  Date: 29 October 2011

=cut


