package ZipTie::Netconf;

use strict;
use warnings;

use v5.10;
use ZipTie::CLIProtocol;
use ZipTie::SSH;
use ZipTie::Recording;
use ZipTie::Recording::Interaction;
use ZipTie::Logger;
use ZipTie::Adapters::Utils qw(get_port_number);

use IPC::Run qw( start pump finish timeout binary );
use Time::HiRes qw(sleep);

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

# Get the instance of the ZipTie::Recording module
my $RECORDING = ZipTie::Recording::get_recording();

# Specifies that ZipTie::SSH is a subclass of ZipTie::CLIProtocol
our @ISA = qw(ZipTie::SSH);
our @EXPORT_OK = qw(get_netconf_message_termination_seq);

use constant NETCONF_FRAMING_SEQ => "]]>]]>";
use constant XML_HEADER => "<?xml version=\"1.0\" encoding=\"UTF-8\"?>";

sub new
{
	my $className = shift;

	# Initialize this instance of the class
	my $this = ZipTie::SSH->new();

	# Turn $this into a ZipTie::SSH object
	bless( $this, $className );

        # These are all restatements of what is in CLIProtocol::SSH 
        # currently.

	# Set the protocol defaults
	$this->{protocol_name}            = "Netconf";
	$this->{version}                  = undef;
	$this->{opened}                   = 0;
	$this->{handle}                   = undef;
	$this->{timeout}                  = 300;
	$this->{compression}              = 0;
	$this->{continue_with_connection} = 1;
	$this->{cache_key}                = 0;
	$this->{update_cached_key}        = 0;
	$this->{ors}	        		  = "\n";
	$this->{binary_mode}              = 0;
	$this->{session_id}               = undef;
	$this->{command_hdr_tag}           = undef;
	$this->{exec_hdr_tag}              = undef;
	$this->{next_msg_id}			  = 100;
	$this->{element_ns_list}          = ();

        # Set this SSH subtype to require a rich cli request type
        # in recordings
        $this->{useRichCliRecRequest}     = 1;

	# Return the object
	return $this;
}


sub connect
{
	my $this       = shift;
	my $ip_address = shift;
	my $port       = shift;
	my $username   = shift;
	my $password   = shift;

	$this->_set_ip_address($ip_address);

	# Make sure that if the port is not defined or less than 1, to default it to 22
	if ( !defined($port) || $port < get_port_number("ssh") )
	{
            $LOGGER->debug("Invalid port '$port' defined. Defaulting to port" . get_port_number("ssh"));
            $port = get_port_number("ssh");
	}
	$this->_set_port($port);

	if ( !$this->{opened} )
	{
		if ( $username eq "" )
		{
			$LOGGER->debug("Found empty user name");
			$LOGGER->fatal_error_code( $INVALID_CREDENTIALS, $ip_address, "Authentication refused with an empty username\nAborting login..." );
		}

		my @cmd = ();
		push( @cmd, 'plink' );
		push( @cmd, '-ssh', '-v', '-a', '-x' );    # Verbose output, disable agent forwarding, and disable X11 forwarding
		push (@cmd, '-2');  # force SSH2 for subsystem select
		push( @cmd, ( '-l',  $username ) ) if $username;
		push( @cmd, ( '-pw', $password ? $password : '') );
		push( @cmd, '-' . $this->{version} ) if $this->{version} =~ /^[12]$/;
		push( @cmd, '-C' ) if $this->{compression};
		push( @cmd, ( '-P', $this->{port} ) );
		push( @cmd, $ip_address );
		# This is where the subsystem wants to reside
		push (@cmd, ( '-s', 'xmlagent'));

		# Connect to the device
		$LOGGER->debug("About to connect to netconf XML agent at $ip_address over port $port ...");
		eval
		{
		    if ($this->{binary_mode})
		    {
                $this->{handle} = start \@cmd, '<', binary, \$this->{in}, '>', binary, \$this->{out}, '2>', \$this->{err}, ( $this->{timer} = timeout( $this->{timeout} ) );		        
		    }
		    else
		    {
    		    $this->{handle} = start \@cmd, \$this->{in}, \$this->{out}, \$this->{err}, ( $this->{timer} = timeout( $this->{timeout} ) );
		    }
		};
        if ($@)
        {
        	$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Could not make netconf connection to device because plink process couldn't be started!\n[DETAILED ERROR MESSAGE]\n$@" );
        }
        
		my ( $have_version, $started, $store_key_prompt, $update_cached_key_prompt, $continue_with_connection_prompt ) = ( 0, 0, 0, 0, 0 );
		my $errorBuffer = '';
		while ( !$have_version || !$started )
		{
			if (length($this->{err}) <= 0)
			{
				eval { $this->{handle}->pump until length($this->{err}) > 0; };
				if ($@)
				{
					$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Could not make netconf connection to device because plink produced an unexpected response!\n[DETAILED ERROR MESSAGE]\n$@" );
				}
			}
			$errorBuffer .= $this->{err};
			$LOGGER->debug( $this->{err} );
			$this->{err} = '';

			if ( !$have_version && $errorBuffer =~ /Using SSH protocol version (\d+)/ )
			{
				$have_version = 1;
				$this->{version} = $1;
			}
			if ( !$store_key_prompt && $errorBuffer =~ /Store key in cache/ )
			{
				$store_key_prompt = 1;
				$this->send( $this->{cache_key} ? "y" : "n" );
			}
			if ( !$update_cached_key_prompt && $errorBuffer =~ /Update cached key/ )
			{
				$update_cached_key_prompt = 1;
				$this->send( $this->{update_cached_key} ? "y" : "n" );
			}
			if ( !$continue_with_connection_prompt && $errorBuffer =~ /Continue with connection/ )
			{
				$continue_with_connection_prompt = 1;
				$this->send( $this->{continue_with_connection} ? "y" : "n" );
			}
			if ( !$started && $errorBuffer =~ /Started session|Started a shell/i )
			{
				$started = 1;
				$this->{err} = '';
			}
			elsif ( !$started && $errorBuffer =~ /Authentication refused/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::Netconf::connect] Authentication refused";
				$LOGGER->fatal_error_code( $INVALID_CREDENTIALS, $ip_address, "Authentication refused" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Connection refused/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::Netconf::connect] Connection refused";
				$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Connection refused" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /'plink' is not recognized|plink:\s+command not found/ )
			{
				$this->{opened} = 0;
				$! =
				    "[ZipTie::Netconf::connect] plink executable was not found in the system path!\n"
				  . "The system path contains the following locations: '"
				  . $ENV{PATH} . "'";
				$LOGGER->fatal(
					"plink executable was not found in the system path!\n" . "The system path contains the following locations: '" . $ENV{PATH} . "'" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Connection timed out/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::Netconf::connect] Connection timed out";
				$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Connection timed out" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Unable to authenticate|password:\s*$/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::Netconf::connect] Unable to authenticate";
				$LOGGER->fatal_error_code( $INVALID_CREDENTIALS, $ip_address, "Unable to authenticate" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Fatal:|Fatal error:/i )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::Netconf::connect] Fatal error encountered in plink!";
				$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Fatal error encounted!\n" . $errorBuffer );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Host does not exist/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::Netconf::connect] Host does not exist";
				$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Host does not exist" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Disconnected/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::Netconf::connect] Disconnected!";
				$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Disconnected prematurely" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /unknown option/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::Netconf::connect] Unknown option passed into plink!";
				$LOGGER->fatal( "Unknown option passed into plink!\n" . $errorBuffer );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Server refused to start a shell command/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::Netconf::connect] Device SSH server would not start xmlagent subsystem shell";
				$LOGGER->fatal( "Device SSH server would not start xmlagent subsystem shell\n" . $errorBuffer );
				last;
			}
		}

		# For debugging purposes, display the version of SSH that is being used to connect to the device
		$LOGGER->debug( "Connected successfully to xmlagent at  $ip_address over port $port using the Netconf over SSH version " . $this->{version} );
		$this->{opened} = 1;
	}
}

sub disconnect 
{
	my $this = shift;
	$LOGGER->debug("Closing Netconf session; sending close-session message ...");
	
	$this->_send_netconf_rpc_request("<nc:close-session/>", \&_send_nc_message);
	$this->ZipTie::SSH::disconnect();	
}

# copied from RFC 6242
my $nc_hello_resp_text = <<'__END_NC_HELLO_RESP';
<nc:hello xmlns:nc="urn:ietf:params:xml:ns:netconf:base:1.0">
  <nc:capabilities>
    <nc:capability>urn:ietf:params:xml:ns:netconf:base:1.0</nc:capability>
  </nc:capabilities>
</nc:hello>
__END_NC_HELLO_RESP

sub send_netconf_hello_response
{
	my $this = shift;
	
	my $nc_hello_resp = XML_HEADER . "\n" . $nc_hello_resp_text .   
		"\n" . NETCONF_FRAMING_SEQ;
	
	# for some reason, if this is sent as a complete message with a blocking wait, which 
	# is what send() does, you'll end up waiting a ton of apparently unnecessary time.  
	# doing this with put() eliminates that.  Don't know why yet.
	$this->put($nc_hello_resp);
}

# Global scope routine
sub get_netconf_message_termination_seq
{	
	return NETCONF_FRAMING_SEQ;
}


sub set_netconf_session_id 
{
	my $this = shift;
	$this->{session_id} = shift;	
}

sub set_netconf_message_headers
{
	my ($this, $execHdr, $commandHdr) = @_;
	
	$this->{exec_hdr_tag} = $execHdr;
	$this->{command_hdr_tag} = $commandHdr if ($commandHdr);
}

sub set_netconf_local_namespace
{
	my ($this, $nsTag, $nsValue) = @_;
	$this->{config_namespace} = $nsTag . "=\"" . $nsValue . '"';
}

sub add_netconf_element_namespace
{
	my ($this, $nsID) = @_;
	push ( @{$this->{element_ns_list}}, $nsID . ':');	
}

sub strip_netconf_elemment_namespaces
{
	my ($this, $stripDoc) = @_;
	my $lStripDoc = $stripDoc;
	my $nsID;
	
	foreach $nsID (@{$this->{element_ns_list}})
	{
		# remove from leading XML tags
		$lStripDoc =~ s/<$nsID/</g;
		$lStripDoc =~ s/<\/$nsID/<\//g;
	}
	
	return $lStripDoc;	
}

# this is used for any kind of show request
sub send_netconf_exec_msg
{
	my ($this, $cmdText) = @_;
	
	my $exMsg = "";
	
	if ($this->{command_hdr_tag}) { $exMsg .= "\n     <" . $this->{command_hdr_tag} . '>'; }
	$exMsg .= $cmdText;
	if ($this->{command_hdr_tag}) { $exMsg .= "</" . $this->{command_hdr_tag} . ">\n" ; }
	
	return $this->_send_netconf_rpc_request($exMsg, \&_send_nc_message_and_wait_for_response);
}

sub _send_netconf_rpc_request
{
	my ($this, $rReqText, $sendVector) = @_;
	my $rMsg =  XML_HEADER . "\n";
	$rMsg .= "<nf:rpc xmlns:nf=\"urn:ietf:params:xml:ns:netconf:base:1.0\"\n";
	$rMsg .=  $this->{config_namespace} .  " message-id=\"" . $this->_get_next_message_id() . "\">\n";
	$rMsg .= "   <" .  $this->{exec_hdr_tag} .">\n";
	$rMsg .= "   $rReqText" . "\n";
	$rMsg .= "   </" .  $this->{exec_hdr_tag} .">\n";
	# and finally...
	$rMsg .= "</nf:rpc>" . NETCONF_FRAMING_SEQ ;
	
	return &$sendVector($this, $rMsg);
}

sub _send_nc_message
{
	my ($this, $msg) = @_;
	$this->send($msg);
}

sub _send_nc_message_and_wait_for_response
{
	my ($this, $msg) = @_;

	return $this->send_and_wait_for($msg, $this->get_netconf_message_termination_seq(), $this->get_timeout() );
}


sub _get_next_message_id
{
	my $this = shift;
	
	# +++wayne multithread-protect?
	return $this->{next_msg_id}++;
}

sub DESTROY
{
	# Preserve any pre-existing error message stored in the $@ sign, since
	# disconnect might over-ride what's already stored in there
	my $err = $@;

	my $this = shift;

	# Disconnect from the device
	$this->disconnect();

	# Restore any pre-existing error message
	$@ = $err;
}


1;

__END__

=head1 NAME

ZipTie::Netconf - Implementation of the ZipTie::CLIProtocol interface/abstract class, supporting netconf over SSH.

=head1 SYNOPSIS

	use ZipTie::CLIProtocol;
	use ZipTie::CLIProtocolFactory;

	my $nconf_inst = ZipTie::CLIProtocolFactory::create("Netconf", "10.10.10.10", 22);
	$Netconf->connect($Netconf->get_ip_address(), $Netconf->get_port(), "joebob", "thisismypassword");

=head1 DESCRIPTION

C<ZipTie::Netconf> provides an Netconf-overSSH implementation of the C<ZipTie::CLIProtocol> interface/abstract class.
It leverages the implemented methods from C<ZipTie::CLIProtocol> and utilizes IPC::Run to interact with
PuTTy's plink program.

=head1 METHODS

Please refer to the "METHOD" section of the C<ZipTie::CLIProtocol> documentation to know what methods
there are and how they are used.  Any additional documentation to methods that have their own unique
implementation for Netconf purposes will be documented below.

=over 12

=item C<new()>

Creates a new instance of the C<ZipTie::Netconf> class with a default timeout value of 30 seconds.
While this constructor method can be called on its own, it is recommend that the C<ZipTie::CLIProtocolFactory>
and its C<create> method be used for creating objects based off of the C<ZipTie::CLIProtocol> class.

=item C<connect($ip_address, $port, $username, $password)>

Connects to a device at a specified IP address and port.  A username and password may also
be specified since some SSH implementations may have a need for it when connecting to a device.

$ip_address -	The administrative IP address of the device.

$port -			The port to connect to on the device.

$username -		The username credential to authenticate with the device.

$password -		The password credential to authenticate with the device.

=item C<disconnect()>

Disconnects from the device this C<ZipTie::Netconf> object is already connected to.

=item C<send($input)>

Sends an input string to the device.  In order to retrieve the output, it is expected that the user
will call C<wait_for()> method with a valid regular expression after the call to C<send()>.

=item C<put($input)>

Same as C<send> except that no line separator, e.g. a CRLF, is appended to the $input.

=item C<send_as_bytes($input)>

Converts an input string into a hexadecimal byte representation and sends it to the device.
Unlike C<send()>, C<send_as_bytes()> will not append any special CRLF character/bytes to the end of the input.
The input will simply be converted into hexadecimal and will be sent as is.  It is expected that the
command interpreter on the device should know how to handle the byte information coming in.

In order to retrieve the output, it is expected that the user will call C<wait_for()> with a valid regular 
expression after the call to C<send_as_bytes()>.

=item C<set_timeout($timeout)>

Set the length (in seconds) for how long the C<wait_for()> method should wait on a response from a
device before timing out.  This time out value will only be used if the the C<wait_for()> method is used
with the matching regular expression as its only parameter.

=item C<wait_for($regex, $timeout)>

Waits for the response from a device and checks to see if the response matches a specified regular expression.
If a match is found, all of the response preceeding AND including the match will be returned.  It is assumed
that this function will be called directly after a call to send().

$regex -	A valid regular expression to match the response from the device.

$timeout -	Optional.  Specifies the amount of time (in seconds) to wait for a response from the device.
			If this parameter is specified, the timeout specified by a call to C<set_timeout()> will be ignored
			for this function call only.

=item C<get_response($timeout)>

Waits a specified window of time for a response to come back from the device.  Given a window of time, if any data
is sent from the device, it will be collected and stored as part of the overall response from the device.  If the window
of time goes by without any response from the device, it is then assumed that the device is done responding and the
overall response will be returned.

The use of C<get_response($timeout)> might be desireable over C<wait_for($regex, $timeout)> if there is not a particular
part of the response from a device you are waiting for, or if the response from a device comes back in an unexpected
format/flow.

$timeout -	Optional; 5 seconds by default.  Specifies the a window of time (in seconds) to wait for a response
			from the device.

=item C<version($version)>

Get or set the version for SSH to use.  If the $version arg is not provided, this method simply returns the current version.
Valid version values are 'auto', '1' or '2'.

This is a no-op; the version is always set to 2 out of requirements of plink connecting to a SSH subsystem (in this case, 
'xmlagent').

=item C<output_record_separator($charSequence)>

Set or get the output record separator.  The $charSequence argument is optional.  If no argument is given, this method
simply returns the current value of the output record separator, such as a CR or CRLF.

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is IBM.
Portions created by AlterPoint are Copyright (C) 2006,
AlterPoint, Inc. All Rights Reserved.

=head1 AUTHOR

Contributor(s): Wayne Tackabury (wtackabury@us.ibm.com)
Date: October 4, 2013

=cut
