package ZipTie::Ranges;

use strict;
use warnings;

sub new
{
	my ($class) = @_;

	my $self = { _rangeList => [] };

	bless($self, $class);

	return $self;
}

sub add_range
{
	my ($self, $lowInt, $highInt) = @_;

	# $highInt is optional, if it's not provided we are adding a single integer to the range.
	if (!defined($highInt))
	{
		$highInt = $lowInt;
	}

	# The $rangeList itself is a reference to an array that must be maintained as a sorted, non-overlaping list of ranges.
	# Because of this, this subroutine may actually end up removing items from the list.  Take for example the existing ranges:
	# 
	# 10-20, 30-40
	#
	# If add_range is called with lowInt equal to 15, and highInt equal to 35, the following would be the resulting ranges:
	#
	# 10-40
	my $rangeList = $self->{_rangeList};
	
	# If the list is empty simply add the range and return.
	if ($#{$rangeList} == -1)
	{
		$rangeList->[0] = { lowInt => $lowInt, highInt => $highInt };
		return;
	}
	
	# The rest of this subroutine works by replacing a 0 length or more section of the range list with a new range.  To do this
	# we need to find the high index and low index of the new range, along with the lowInt and highInt of the new range.
	# The lowInt and/or highInt values actually used in the range may be the values actually passed in, or could be taken from
	# one of the ranges replaced, if the replaced range starts before or after the range we are inserting.

	# Search the range list based on $highInt.  This is the "middle" index.  We will count up from here to find the high index,
	# and count down from here to find the lower index.
	my $index = _bin_search($rangeList, 0, $#{$rangeList}, $highInt);

	# Find low and high index.
	my $lowIndex = $index;
	while ($lowIndex > 0 && ($lowInt - 1) <= $rangeList->[$lowIndex - 1]->{highInt})
	{
		$lowIndex--;
	}
	my $highIndex = $index;
	while ($highIndex < $#{$rangeList} && ($highInt + 1) >= $rangeList->[$highIndex + 1]->{lowInt})
	{
		$highIndex++;
	}

	# Determine the correct lowInt and highInt values.
	if (defined($rangeList->[$lowIndex]))
	{
		$lowInt = $lowInt < $rangeList->[$lowIndex]->{lowInt} ? $lowInt : $rangeList->[$lowIndex]->{lowInt};
	}
	if ($lowIndex != $highIndex && defined($rangeList->[$highIndex - 1]))
	{
		$highInt = $highInt > $rangeList->[$highIndex - 1]->{highInt} ? $highInt : $rangeList->[$highIndex - 1]->{highInt};
	}

	# Determine if we need to replace the range under highIndex
	my $add = 0;
	if (($highIndex <= $#{$rangeList}) && (($highInt + 1) == $rangeList->[$highIndex]->{lowInt}))
	{
		$add = 1;
		$highInt = $rangeList->[$highIndex]->{highInt};
	}

	# Insert the new range.
	splice(@{$rangeList}, $lowIndex, ($highIndex - $lowIndex + $add), { lowInt => $lowInt, highInt => $highInt });
}

# Given the start of the possible range, and the end of the possible range, return the negated list of ranges.
sub get_negated_ranges
{
	my ($self, $startRange, $endRange) = @_;

	my $rangeList = $self->{_rangeList};
	
	my $lowInt = $startRange;
	my @negatedList;
	
	# Loop through the rangeList and negate as we go.  Since it's in order, this will be ok.
	foreach my $range (@{$rangeList})
	{
		if ($lowInt != $range->{lowInt})
		{
			my $highInt = $range->{lowInt} - 1;
			push(@negatedList, { lowInt => $lowInt, highInt => $highInt });
		}
		
		$lowInt = $range->{highInt} + 1;
	}
	
	if ($lowInt <= $endRange)
	{
		push(@negatedList, { lowInt => $lowInt, highInt => $endRange });
	}
	
	return \@negatedList;
}

sub get_ranges
{
	my ($self) = @_;

	return $self->{_rangeList};
}

# Performs a binary search based on the lowInt of the ranges already in the list.
sub _bin_search
{
	my ($list, $min, $max, $value) = @_;

	if ($max < $min)
	{
		return $min;
	}

	my $mid = int(($min + $max) / 2);

	if ($list->[$mid]->{lowInt} > $value)
	{
		return _bin_search($list, $min, $mid - 1, $value);
	}

	if ($list->[$mid]->{lowInt} < $value)
	{
		return _bin_search($list, $mid + 1, $max, $value);
	}

	return $mid;
}

=head1 NAME

ZipTie::Ranges - Used to combine multilpe values and ranges into the smallest amount of ranges.

=head1 SYNOPSIS

	use ZipTie::Ranges;

	my $ranges = ZipTie::Ranges->new();

    $ranges->add_range(1, 5);
    $ranges->add_range(22);
    $ranges->add_range(21);
    $ranges->add_range(20);
    $ranges->add_range(23);
	$ranges->add_range(22, 26);

	use Data::Dumper;
	print Dumper($ranges->get_ranges());

output:

	$VAR1 = [ 
	          { 
	            'lowInt' => 1,
	            'highInt' => 5
	          },
	          { 
	            'lowInt' => 20,
	            'highInt' => 26
	          }
	        ];

=head1 DESCRIPTION

Used to combine multiple values and ranges into the smallest amount of ranges.  Handles overlapping ranges, keeps the ranges sorted.

=head1 METHODS

=head2 Constructor

=over

=item new()

    my $ranges = ZipTie::Ranges->new();

Creates a Ranges object.

=back

=head2 Object Methods

=over

=item add_range($lowInt, $highInt)

	$range->add_range(1, 5);
	$range->add_range(22);

Adds a range to the list of ranges.  $lowInt is a required value, highInt is optional.  If $highInt is not provided it is assumed to be equal
to the $lowInt (ie, add a single value).  The ranges are inclusive.

=back

=item get_ranges()

	my $rangeList = $ranges->get_ranges();

Returns the list of ranges as a reference to an array.

=back

=item get_negated_ranges($startRange, $endRange)

	my $negatedRangeList = $ranges->get_negated_ranges(0, 65535);

Given the start and end of the possible range of integers, return a negated list of integers.

=back

=head1 AUTHOR(S)

    David Payne
    david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2013, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut

1;
