package ZipTie::RemoteGatewayResolution;

use strict;
use warnings;

use Net::CIDR::Set;

sub new
{
	my ($class) = @_;

	my $self = {
			# This is a list of local subnets with corrisponding interfaces, populated using add_route.
			_local_subnets => [],

			# This is a list of routes, populated using add_route.
			_routes => [],

			# After calling resolved_remote_gateways, this will contain a list of the routes resolved
			# to local gateways.
			_resolved_routes => [],

			# After calling resolve_remote_gateways this will contain a list of routes that
			# didn't resolve to a local interface.
			_problem_routes => [],

			_default_route => undef,

			_routes_resolved => 0
		};

	bless($self, $class);

	return $self;
}

sub add_local_subnet
{
	my ($self, $localIp, $localMask, $interface) = @_;

	push(@{$self->{_local_subnets}}, {subnet => Net::CIDR::Set->new("$localIp/$localMask"), interface => $interface});
}

my $routeUniqueId = 0;
sub add_route
{
	my ($self, $destinationAddress, $destinationMask, $gatewayAddress, $routeMetric, $defaultRoute, $otherDetails) = @_;

	my $route = {destinationAddress => $destinationAddress,
			destinationMask => $destinationMask,
			gatewayAddress => $gatewayAddress,
			routeMetric => $routeMetric,
			otherDetails => $otherDetails,
			destinationSubnet => Net::CIDR::Set->new("$destinationAddress/$destinationMask"),
			uniqueId => ++$routeUniqueId # Give each route a unique id to help detect infinate recursion in sub _resolve_remote_gateway
		};

	push(@{$self->{_routes}}, $route);

	# The default route will be set to the route last route marked as the default route.
	# If no route is marked as the default route, but a route with a mask of 0 exists it will be marked as the default route.
	# If no route is marked as the default route, and no route has a mask of 0, then no route will be the default route.
	if ($defaultRoute || ($destinationMask eq '0' && !defined($self->{_default_route})))
	{
		$self->{_default_route} = $route;
	}
}

sub resolve_remote_gateways
{
	my ($self) = @_;

	# Sort the routes based on prefix size
	my @sortedRoutes = sort _compare_routes (@{$self->{_routes}});

	# This loop resolves each route to either:
	# 	1) A local subnet if the gateawy is local, in which case the interface attribute on the route is set.
	#   2) If the gateway is not local the route is resolved to another 
	FIRST_PASS: foreach my $route (@sortedRoutes)
	{
		# Determine if the gateway is local, and if so, set the interface
		foreach my $localSubnet (@{$self->{_local_subnets}})
		{
			if (defined $route->{gatewayAddress}) {
				if ($localSubnet->{subnet}->contains($route->{gatewayAddress}))
				{
					$route->{interface} = $localSubnet->{interface}; # If interface is set on the route then we know it's local
					next FIRST_PASS;
				}
			}
		}

		# The gateway is not local, see if the gateway is reachable through another route
		foreach my $compareRoute (@sortedRoutes)
		{
			if (defined $route->{gatewayAddress}) {
				if ($compareRoute->{destinationSubnet}->contains($route->{gatewayAddress}))
				{
					$route->{nextRoute} = $compareRoute;
					next FIRST_PASS;
				}
			}
		}
	}

	# This loop finilizes the process, it's the second pass.  Any route that resolves to some other route instead of a local gateway
	# will now have to have the correct interface and gateway address set.
	foreach my $route (@sortedRoutes)
	{
		my ($interface, $gatewayAddress) = _resolve_remote_gateway($self, $route);

		if (defined($interface))
		{
			$route->{interface} = $interface;
			$route->{gatewayAddress} = $gatewayAddress;
		}
	}

	$self->{_routes_resolved} = 1;
}

sub get_routes
{
	my ($self) = @_;

	# sub resolve_remote_gateways must be called before calling this method
	if (!$self->{_routes_resolved})
	{
		return undef;
	}

	return _return_routes($self->{_routes});
}

sub get_problem_routes
{
	my ($self) = @_;

	# sub resolve_remote_gateways must be called before calling this method
	if (!$self->{_routes_resolved})
	{
		return undef;
	}

	return _return_routes($self->{_problem_routes}, 1);
}

sub _resolve_remote_gateway
{
	my ($self, $route, $seen) = @_;

	# Check to see if this route is already resolved to a local interface.
	if (defined($route->{interface}))
	{
		return ($route->{interface}, $route->{gatewayAddress});
	}

	# Check to see if this is already a known problem route.
	if (defined($route->{problemRoute}))
	{
		return (undef, undef);
	}

	if (!defined($seen))
	{
		$seen = {};
	}

	# Make sure we don't enter infinate recursion, using the uniqueId on each route
	if (!defined($seen->{$route->{uniqueId}}))
	{
		# Mark the route as seen
		$seen->{$route->{uniqueId}} = 1;
	}
	else
	{
		# We have hit a loop, fall back to the default route if it exists, otherwise return undef.
		if (defined($self->{_default_route}->{interface}))
		{
			return ($self->{_default_route}->{interface}, $self->{_default_route}->{gatewayAddress});
		}
		else
		{
			return (undef, undef);
		}
	}

	# If this route has a valid nextRoute set then determine the interface and gateway address of the next route.
	# If this route doesn't have a valid nextRoute, then use the default gateway if possible, otherwise it's a problem route.
	my $interface = undef;
	my $gatewayAddress = undef;
	if (defined($route->{nextRoute}))
	{
		($interface, $gatewayAddress) = _resolve_remote_gateway($self, $route->{nextRoute}, $seen);
	}
	else
	{
		if (defined($self->{_default_route}->{interface}))
		{
			$interface = $self->{_default_route}->{interface};
			$gatewayAddress = $self->{_default_route}->{gatewayAddress};
		}
	}

	if (defined($interface))
	{
		$route->{interface} = $interface;
		$route->{gatewayAddress} = $gatewayAddress;
	}
	else
	{
		$route->{problemRoute} = 1;
		push(@{$self->{_problem_routes}}, $route);
		return (undef, undef);
	}

	return ($interface, $gatewayAddress);
}

# Used to sort routes into the order they should be used for resolving remote gateway addresses.
sub _compare_routes
{
	# Always use routes with a more narrow mask before routes with a wider mask.
	my $maskCmp = $b->{destinationMask} <=> $a->{destinationMask};
	if ($maskCmp)
	{
		return $maskCmp;
	}

	# If the masks are the same use the routeMetric to break ties.
	if (defined($a->{routeMetric}) && defined($b->{routeMetric}))
	{
		return $a->{routeMetric} <=> $b->{routeMetric};
	}

	# The routes are equal.
	return 0;
}

sub _return_routes
{
	my ($routes, $problemRoutes) = @_;

	my @routesToReturn;
	foreach my $route (@{$routes})
	{
		# Unless $problemRoutes is set to true, then ignore any route with a problem.
		if (!$problemRoutes)
		{
			if (defined($route->{problemRoute}))
			{
				next;
			}
		}

		push(@routesToReturn, {
			destinationAddress => $route->{destinationAddress},
			destinationMask => $route->{destinationMask},
			gatewayAddress => $route->{gatewayAddress},
			routeMetric => $route->{routeMetric},
			otherDetails => $route->{otherDetails},
			interface => $route->{interface},
			discard => $route->{discard},
		});
	}

	return \@routesToReturn;
}

1;

=head1 NAME

ZipTie::RemoteGatewayResolution - Used to resolve remote gateways (aka next-hops) to routes with locally occuring gateway addresses.

=head1 SYNOPSIS

use ZipTie::RemoteGatewayResolution;

	my $rgr = ZipTie::RemoteGatewayResolution->new();

	$rgr->add_local_subnet('172.16.50.1', '24', 'outside');
	$rgr->add_local_subnet('172.16.100.1', '24', 'inside');

	# The 4th argument to add_route is optional, and it can be anything of the users choosing.
	# It's useful to carry extra details with the route.
	$rgr->add_route('0.0.0.0', '0', '172.16.50.2', 1, 1, {anythingYouWant => 1});

	# They can be passed in any order
	$rgr->add_route('172.16.30.0', '24', '172.16.20.3', 1, 0);              # This next-hop is remote
	$rgr->add_route('172.16.20.0', '24', '172.16.100.2', 4, 0, {id => 5});  # This next-hop is local

	$rgr->resolve_remote_gateways();

	use Data::Dumper;

	print "Resolved routes:\n";
	print Dumper($rgr->get_routes());

output:

	Resolved routes:
	$VAR1 = [
	          {
	            'otherDetails' => {
	                                'anythingYouWant' => 1
	                              },
	            'gatewayAddress' => '172.16.50.2',
	            'destinationAddress' => '0.0.0.0',
	            'destinationMask' => 0,
				'routeMetric' => 1,
	            'interface' => 'outside'
	          },
	          {
	            'otherDetails' => undef,
	            'gatewayAddress' => '172.16.100.2',
	            'destinationAddress' => '172.16.30.0',
	            'destinationMask' => 24,
				'routeMetric' => 1,
	            'interface' => 'inside'
	          },
	          {
	            'otherDetails' => {
	                                'id' => 5
	                              },
	            'gatewayAddress' => '172.16.100.2',
	            'destinationAddress' => '172.16.20.0',
	            'destinationMask' => 24,
				'routeMetric' => 4,
	            'interface' => 'inside'
	          }
	        ];

=head1 DESCRIPTION

Used to resolve remote gateways (aka next-hops) to routes with locally occuring gateway addresses.

=head1 METHODS

=head2 Constructor

=over

=item new()

	my $rgr = ZipTie::RemoteGatewayResolution->new();

Creates a RemoteGatewayResolution object.

=back

=head2 Object Methods

=over

=item add_local_subnet($localIp, $localMask, $interface)

	$rgr->add_local_subnet('172.16.50.1', '24', 'outside');

Adds a local subnet to the collection of local subnets.

=item add_route($destinationAddress, $destinationMask, $gatewayAddress, $routeMetric, $defaultRoute, $otherDetails)

	$rgr->add_route('172.16.30.0', '24', '172.16.20.3', 1, 0);

$routeMetric, $defaultRoute and $otherDetails are optional attributes.  Any of them can be set to undef.

$defaultRoute is treated as a boolean.

$otherDetails can be anything the user desires, it's not processed by RemoteGatewayResolution at all, but
is provided back to the caller in the output of get_routes and get_problem_routes.  It can be used to store
additional route details.

=item resolve_remote_gateways()

	$rgr->resolve_remote_gateways();

Resolves all the routes in the route collection (added by add_route) into local gatewayAddresses and interfaces.
All the resolved routes can then be returned by get_routes, and any routes that could not be resolved can be
returned by get_problem_routes.

=item get_routes()

	$rgr->get_routes();

Returns a list of routes that have been resolved to a local gateway address and interafce.

=item get_problem_routes()

	$rgr->get_problem_routes();

Returns a list of routes that could not be resolved into a local gateway address and interface.

=back

=head1 AUTHOR(S)

	David Payne
	david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2013, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
