package ZipTie::SFTP;

use strict;
use warnings;
use IPC::Run qw( start run pump timeout );
use File::Basename qw(fileparse);

use ZipTie::TransferProtocol;
use ZipTie::Logger;

# Get the instance of the ZipTie::Logger module
my $LOGGER = ZipTie::Logger::get_logger();

# Specifies that ZipTie::SFTP is a subclass of ZipTie::TransferProtocol
our @ISA = qw(ZipTie::TransferProtocol);

sub new
{
	my $class_name = shift;

	# Initialize this instance of the class as a ZipTie::TransferProtocol
	# since at its core, that is what it is.
	my $this = ZipTie::TransferProtocol->new();

	# Turn $this into a ZipTie::FTP object
	bless( $this, $class_name );

	# Set the protocol name
	$this->{protocol_name} = "SFTP";
	$this->{opened}        = 0;
	$this->{handle}        = undef;
	$this->{ors}	       = "\n";
	# Default the timeout to 30 seconds
	$this->{timeout}       = 30;

	# Return the object
	return $this;
}


sub connect
{
	# Grab a reference to our self/this instance
	my $this = shift;

	# Store the parameters
	my $ip_address = shift;
	my $port       = shift;
	my $username   = shift;
	my $password   = shift;
	

	# Set the IP address and port for this client
	$this->_set_ip_address($ip_address);
	if ( !defined($port) || $port <= 0 )
	{
		$LOGGER->debug("Invalid port '$port' defined. Defaulting to port '22'.");
		$port = 22;
	}
	$this->_set_port($port);
	
	if( ! $this->{opened} )
	{	
		if ( $username eq "" )
		{
			$LOGGER->debug("Found empty user name");
			$LOGGER->fatal_error_code( $INVALID_CREDENTIALS, $ip_address, "Authentication refused with an empty username\nAborting login..." );
		}
		# build up the cmd line args for psftp. be as explicit as possible
		my @cmd = ();
		push( @cmd, 'psftp' );
		push( @cmd, '-C' );
		push( @cmd, '-v' ); # verbose
		push( @cmd, ('-l', $username) ); # login user
		push( @cmd, ('-P', $port) ); # always specify port
		push( @cmd, ('-pw', $password) ) if($password); # conditional, assuming it was provided
		push( @cmd, '-2' ); # force version 2, can't think of legit reason to use 1
		push( @cmd, $ip_address); # remote host
		
		$LOGGER->debug("connecting via psftp ...");
		$this->{handle} = start \@cmd, \$this->{in}, \$this->{out}, \$this->{err}, ( $this->{timer} = timeout( $this->{timeout} ) );
		 if( $@ )
		 {
		 	$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Connection failed to $ip_address as $username\n[DETAILED ERROR MESSAGE]\n$@");
		 }

		my ( $have_version, $started, $store_key_prompt, $update_cached_key_prompt, $continue_with_connection_prompt ) = ( 0, 0, 0, 0, 0 );
		my $errorBuffer = '';
		while ( !$have_version || !$started )
		{
			if (length($this->{err}) <= 0)
			{
				eval { $this->{handle}->pump until length($this->{err}) > 0; };
				if ($@)
				{
					$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Could not make connection to device because plink produced an unexpected response!\n[DETAILED ERROR MESSAGE]\n$@" );
				}
			}
			$errorBuffer .= $this->{err};
			$LOGGER->debug( $this->{err} );
			$this->{err} = '';

			if ( !$have_version && $errorBuffer =~ /Using SSH protocol version (\d+)/ )
			{
				$have_version = 1;
				$this->{version} = $1;
			}
			if ( !$store_key_prompt && $errorBuffer =~ /Store key in cache/ )
			{
				$store_key_prompt = 1;
				$this->send( $this->{cache_key} ? "y" : "n" );
			}
			if ( !$update_cached_key_prompt && $errorBuffer =~ /Update cached key/ )
			{
				$update_cached_key_prompt = 1;
				$this->send( $this->{update_cached_key} ? "y" : "n" );
			}
			if ( !$continue_with_connection_prompt && $errorBuffer =~ /Continue with connection/ )
			{
				$continue_with_connection_prompt = 1;
				$this->send( $this->{continue_with_connection} ? "y" : "n" );
			}
			if ( !$started && $errorBuffer =~ /Started session|Started a shell/i )
			{
				$started = 1;
				$this->{err} = '';
			}
			elsif ( !$started && $errorBuffer =~ /Authentication refused/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::SFTP::connect] Authentication refused";
				$LOGGER->fatal_error_code( $INVALID_CREDENTIALS, $ip_address, "Authentication refused" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Connection refused/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::SFTP::connect] Connection refused";
				$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Connection refused" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /'psftp' is not recognized|psftp:\s+command not found/ )
			{
				$this->{opened} = 0;
				$! =
				    "[ZipTie::SFTP::connect] psftp executable was not found in the system path!\n"
				  . "The system path contains the following locations: '"
				  . $ENV{PATH} . "'";
				$LOGGER->fatal(
					"plink executable was not found in the system path!\n" . "The system path contains the following locations: '" . $ENV{PATH} . "'" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Connection timed out/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::SFTP::connect] Connection timed out";
				$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Connection timed out" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Unable to authenticate|password:\s*$/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::SFTP::connect] Unable to authenticate";
				$LOGGER->fatal_error_code( $INVALID_CREDENTIALS, $ip_address, "Unable to authenticate" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Fatal:|Fatal error:/i )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::SFTP::connect] Fatal error encountered in psftp!";
				$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Fatal error encounted!\n" . $errorBuffer );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Host does not exist/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::SFTP::connect] Host does not exist";
				$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Host does not exist" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /Disconnected/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::SFTP::connect] Disconnected!";
				$LOGGER->fatal_error_code( $SSH_ERROR, $ip_address, "Disconnected prematurely" );
				last;
			}
			elsif ( !$started && $errorBuffer =~ /unknown option/ )
			{
				$this->{opened} = 0;
				$! = "[ZipTie::SFTP::connect] Unknown option passed into psftp!";
				$LOGGER->fatal( "Unknown option passed into psftp!\n" . $errorBuffer );
				last;
			}
		}
		
		$this->{handle}->pump until $this->{out} =~ /psftp> /;
		
		# For debugging purposes, display the version of SSH that is being used to connect to the device
		$LOGGER->debug( "Connected successfully to $ip_address over port $port using the SSH version " . $this->{version} );
		$this->{opened} = 1;
	}	
}


sub disconnect
{
    my $this = shift;

    # Tear down the IPC::Run harness
    if ( $this->{handle} )
    {
        $LOGGER->debug("Tearing down IPC::Run harness used to communicate with psftp.");

        # Pump out all of the output that may be waiting around for 1 second
        eval {
            $this->{timer}->start(1);
            $this->{handle}->pump until $this->{out} =~ /.*/;
        };
        if ($@)
        {
            $LOGGER->debug("Unable to read the rest of the output from psftp.  Ignoring ...");
        }
        
        eval {
            $LOGGER->debug("Finishing the IPC::Run harness ...");
            $this->{handle}->finish;
            $LOGGER->debug("Killing off any child processes associated with the IPC::Run harness ...");
            $this->{handle}->kill_kill;
        };
        if ($@)
        {
            $LOGGER->debug("Error on finishing off the IPC::Run harness, so killing everything instead ...");

            # Kill this process with only a 1 second grace period
            $this->{handle}->kill_kill( grace => 1 );
        }

        # Mark that the IPC::Run harness has been shutdown and the connection to the device closed
        $this->{opened} = 0;
    }
}

sub send
{
	my $this  = shift;
	my $input = shift;

	# Clear out the buffer so that no previous messages pollute it
	$this->_empty_buffers;

	# Store the current input on the object.  This will make the current input be registered as the last
	# input that was sent
	$this->{input} = $input;

	# Send the input to the device and append any necessary CRLF data to the end of the input.
	$LOGGER->debug("-----------------------------------------------------------------------------");
	$LOGGER->debug("[SENDING]");
	$LOGGER->debug($input);
	$LOGGER->debug("-----------------------------------------------------------------------------");

	$this->{in} .= $input . $this->{ors};
	while ( $this->{in} )
	{
		$this->{handle}->pump;
	}
}

sub recv
{
	my $this = shift;
	
	$LOGGER->debug("-----------------------------------------------------------------------------");
	$LOGGER->debug( "[WAITING " . $this->{timeout} . " SECOND(S) FOR]" );
	$LOGGER->debug("psftp> ");
	$LOGGER->debug("-----------------------------------------------------------------------------");
	
	my $startTime = time();
	my $start     = "Started waiting at: " . scalar localtime($startTime);

	# Wait for the response from the device
	$this->{timer}->start($this->{timeout});
	$this->{handle}->pump until $this->{out} =~ /psftp> /;

	# Handle any potential timeout
	if ($@)
	{
		$this->disconnect();
		$start .= " -- Ended waiting at: " . scalar localtime( time() ) . " -- ";
		my $ip_address      = $this->get_ip_address();
		my $actual_command  = $this->last_input();
		my $actual_response = $this->{out};
		my $error_message .= "timed-out after ".$this->{timeout}." seconds ($start) while waiting to match the prompt 'psftp> '.\n";
		$error_message .= "[INPUT]\n$actual_command\n" unless ( !defined($actual_command) || !$actual_command );
		$error_message .= "[ACTUAL RESPONSE]\n$actual_response";
		$LOGGER->fatal_error_code( $UNEXPECTED_RESPONSE, $ip_address, $error_message );
	}

	# We want to know the entire response from the device for a given input, not just what was matched
	my $response = $this->{out};

	return $response;
}

sub get
{
	# Grab a reference to our self/this instance
	my $this = shift;
	
	# Grab the parameters
	my $remote_file = shift;
	my $local_file = shift;
	
	# Get the remote file and store it in the local file
	my $ip_address = $this->get_ip_address();
	my $port       = $this->get_port();

	$LOGGER->debug("Retrieving '$remote_file' from sftp://$ip_address:$port/ and storing it as '$local_file'");
	$this->send("get $remote_file $local_file");
	
	my $response = $this->recv();
	
	$LOGGER->debug($response);
	
	if( -f $local_file )
	{
		$LOGGER->debug("Successfully retrieved '$remote_file' from sftp://$ip_address:$port/ and stored it as '$local_file'");
	}
	else
	{
		$LOGGER->fatal_error_code( $UNEXPECTED_RESPONSE, $ip_address, "file retrieval failed, file doesn't exist");
		return 0;
	}
	
	return 1;
}

sub put
{
	# Grab a reference to our self/this instance
	my $this = shift;
	
	# Grab the parameters
	my $remote_file = shift;
	my $local_file = shift;
	
	# Send the local file and store it in the remote file
	my $ip_address = $this->get_ip_address();
	my $port       = $this->get_port();

	$LOGGER->debug("Sending '$local_file' to sftp://$ip_address:$port/ and storing it as '$remote_file'");
	$this->send("put $local_file $remote_file\n");
	
	my $response = $this->recv();
	
	$LOGGER->debug($response);
	
	$LOGGER->debug("Successfully sent '$local_file' to sftp://$ip_address:$port/ and stored it as '$remote_file'");

	return 1;	
}

sub list
{
	my $this = shift;
	
	my $remote_dir = shift;
	
	my @list = ();
	
	if( defined($remote_dir) && length($remote_dir) > 0 )
	{
		$LOGGER->debug("listing files in remote dir $remote_dir");
		$this->send("ls $remote_dir");
	}
	else
	{
		$LOGGER->debug("listing files in default remote dir (cwd)");
		$this->send("ls");
	}
	
	my $response = $this->recv();
	
	while( $response =~ /^[drwx\-sSt]+\s+\d+\s+\S+\s+\S+\s+\d+\s+[A-Za-z]+\s+\d+\s+[0-9\:]+\s+(.*?)$/msg )
	{
		my $item = $1;
		if( defined($remote_dir) && length($remote_dir) > 0 )
		{
			$item = $remote_dir."/".$item;
		}
		push @list, $item;
	}
	
	$LOGGER->debug("listed ".scalar(@list)." files");
	
	return \@list;	 
}


sub _empty_buffers
{
	my $this = shift;
	$this->{out} = '';
	$this->{err} = '';
}

sub DESTROY
{
	# Preserve any pre-existing error message stored in the $@ sign, since
	# disconnect might over-ride what's already stored in there
	my $err = $@;

	my $this = shift;

	# Disconnect from the device
	$this->disconnect();

	# Restore any pre-existing error message
	$@ = $err;
}

1;

__END__

=head1 NAME

ZipTie::SFTP - SFTP client implementation of the ZipTie::TransferProtocol interface/abstract base class.

=head1 SYNOPSIS

	use ZipTie::TransferProtocol;
	use ZipTie::TransferProtocolFactory;

	my $xfer_client = ZipTie::TransferProtocolFactory::create("SFTP");
	$xfer_client->connect("10.10.10.10", 22, somename, somepassword);
	$xfer_client->put("test.txt", "test.txt");
	$xfer_client->disconnect();

=head1 DESCRIPTION

ZipTie::SFTP represents a client implementing the SFTP protocol that adheres to the interface specified by the
ZipTie::TransferProtocol class.  The core backing of this implementation is the C<psftp> command line tool.

In order to conform to the interface provided by ZipTie::TransferProtocol, the following functionality
had to be implemented:  connecting to a remote server (this is supported by the C<connect> method), disconnecting
from a remote server (this is supported by the C<disconnect> method), retrieve/get a file from the server and
store it locally (this is supported by the C<get> method), and send/put a local file to the server (this is
supported by the C<put> method).

=head1 IMPLEMENTED METHODS

The following methods are all methods that have been implemented according to the abstract/interface methods that have been
specified in the C<ZipTie::TransferProtocol> module.

=over 12

=item C<connect($ip_address, $port, $username, $password)>

Connects to a server implementing the transfer protocol at a specified IP address and port.
In order to authenticate with the server, a username and password also needs to be specified.

$ip_address -	The hostname/IP address of the server.

$port -			The port to connect to on the server.

$username -		The username credential to authenticate with the server.

$password -		The password credential to authenticate with the server.
	
=item C<disconnect()>

Disconnects from the server implementing the transfer protocol.

=item C<get($remote_file, $local_file)>

Retrieves a file from the remote host/IP address that this ZipTie::TransferProtocol is connected to.

$remote_file -	A relative or absolute filepath on the host/server defining the file to be retrieved.
				If the filepath is relative, it is then relative to the default directory of the host/server.

$local_file -	Optional.  A filepath where the remote file should be transfer to on the local machine.
				This file path can either absolute or relative.  If the filepath is not specified, then the file
				name of the remote file being retrieved will be used and placed within the current working directory
				on the local machine.  The current working directory will also be used as the relative directory if
				the local file name is not absolute.

Returns I<true> if the file transfer was successful; I<false> otherwise.

=item C<put($local_file, $remote_file)>

Sends a file to the remote host/IP address that this ZipTie::TransferProtocol is connected to.

$local_file -	A filepath defining of the local file to put on the host/server.  This file path can either absolute or
				relative.  If the filepath is relative, then it is relative to the current working directory.

$remote_file -	Optional.  A filepath defining where the local file should be transfer to on the host/server.
				This file path can either absolute or relative.  If the filepath is not specified, then the file
				name of the local file being sent will be used and placed within the default directory of the host/server.

Returns I<true> if the file transfer was successful; I<false> otherwise.

=item C<list($remote_dir)>

Lists files in the remote dir on the remote host/IP address that this ZipTie::TransferProtocol is connected to.

$remote_dir -	Optional.  A filepath defining where on the host/server the listing should be made.
				This file path can either absolute or relative.  If the filepath is not specified, then the default
				directory will be used and returned files will have no additional path. May return an
				empty list if listing isn't allowed.

=back

=head1 INHERITED METHODS

The following methods are all methods that have been inherited from the C<ZipTie::TransferProtocol> module.

=over 12

=item C<get_ip_address()>

Retrieves the hostname/IP address of the server that the implementation of the ZipTie::TransferProtocol
class is connected to.

=item C<get_protocol_name()>

Retrieves the name of the transfer protocol that this implementation of the ZipTie::TransferProtocol
class represents.  For example: "SCP".

=item C<get_port()>

Retrieves the port that this implementation of the ZipTie::TransferProtocol
class has connected to on the server.

=item C<get_timeout()>

Retrieves the length (in seconds) that this implementation of the ZipTie::TransferProtocol class
should wait for the server during a transfer before a time-out error is declared.

=item C<set_timeout($timeout)>

Sets the length (in seconds) that this implementation of the ZipTie::TransferProtocol class
should wait for the server during a transfer before a time-out error is declared.

=back

=head1 LICENSE

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Ziptie Client Framework.

The Initial Developer of the Original Code is Q1 Labs.
Portions created by Q1 Labs are Copyright (C) 2011,
Q1 Labs, Inc. All Rights Reserved.

=head1 AUTHOR

Contributor(s): rory.bray (rory.bray@q1labs.com)
Date: Feb 3, 2011

=cut
