package ZipTie::Snort;

=pod

=head1 NAME

ZipTie::Snort - Snort parser.

=head1 SUBROUTINES

=cut

use strict;
use warnings;

use ZipTie::Adapters::Utils qw(get_crep);

my $LOGGER = ZipTie::Logger::get_logger();

=pod

B<new($config_name)>

Create a new ZipTie::Snort.

$config_name is the name of the SNORT config.

$get_config_callback is a reference to a subroutine that takes a single argument, the path and
filename to a SNORT configuration.  The callback is expected to return the content of the
configuration, or undef if the adapter is unable to obtain the configuration.

$change_action_callback is optional.  It's currently used by the GX adapter.  If defined and the
Snort rule contains an Sid the callback is called with a single argument, the SID of the rule.  If
the callback returns any value other than undef then the primaryAction of the rule is set to the
value returned by the callback.

$config_type is the type of config being parsed.  This will be appeneded to the object group names.
$config_type is optional, defaults to 'SNORT'.

=cut
sub new
{
	my ($class, $config_name, $get_config_callback, $config_type, $change_action_callback) = @_;

	if (!defined($config_type))
	{
		$config_type = 'SNORT';
	}

	my $self = {
			_uncategorized_vars => {},
			_network_groups => {},
			_service_groups => {},
			_applications => [],
			_preprocessor_rule_groups => {},
			_rule_groups => {},
			_preprocessor_evaluation_order => 'whitelist blacklist',
			_evaluation_order => 'activate dynamic pass drop sdrop reject alert log custom',
			_config_name => $config_type . '_' . $config_name,
			_applies_to => undef,
			_user_rule_types => {},
			_preprocessors => {},
			_parsed_configs => {},
			_get_config_callback => $get_config_callback,
			_change_action_callback => $change_action_callback,
			_rule_names => {},
			_config_type => $config_type,
			_final_rules => undef,
			_bound_snorts => [],
			_generated_names_to_msg => {}
		};

	bless($self, $class);

	return $self;
}

=pod

B<set_applies_to($applies_to)>

Set the appliesTo elements for a SNORT config.

=cut
sub set_applies_to
{
	my ($self, $applies_to) = @_;

	$self->{_applies_to} = $applies_to;
}

=pod

B<get_applies_to()>

Return the appliesTo elements for a SNORT config.

=cut
sub get_applies_to
{
	my ($self) = @_;

	return $self->{_applies_to};
}

=pod

B<has_rules()>

Returns true if the SNORT object has parsed valid rules, otherwise returns false.

=cut
sub has_rules
{
	my ($self) = @_;

	if (scalar(keys(%{$self->{_rule_groups}})))
	{
		return 1;
	}

	return 0;
}

=pod

B<get_rules()>

Returns the parsed SNORT rules.

=cut
sub get_rules
{
	my ($self) = @_;

	# If rules have already been parsed, return that list.
	if (defined($self->{_final_rules}))
	{
		return $self->{_final_rules};
	}

	my $process_order = 1;
	my @final_rules;

	my @groups = keys(%{$self->{_rule_groups}});
	my @ordered_groups = split(/\s+/, $self->{_evaluation_order});
	foreach my $group (@groups)
	{
		my $already_ordered;
		foreach my $ordered_group (@ordered_groups)
		{
			if ($ordered_group eq $group)
			{
				$already_ordered = 1;
				last;
			}
		}
		if (!$already_ordered)
		{
			push(@ordered_groups, $group);
		}
	}

	foreach my $rule_group (@ordered_groups)
	{
		RULE_LOOP: foreach my $rule (@{$self->{_rule_groups}->{$rule_group}})
		{
			# We allow parse_config, and parse_rules to be called in any order. Because some things
			# in parse_rules rely on configuration from parse_config not everything can be evaluated
			# in parse_rules. By the time get_rules is called all configuration should have been 
			# parsed, and we can do some late evaluation to ensure everything is in order.

			my $snort_action = $rule->{primaryAction};
			my ($primary_action, $log) = _parse_snort_action($self, $snort_action);
			if (!defined($primary_action))
			{
				$LOGGER->debug('Unsupported ' . $self->{_config_type} . " rule action: '$snort_action', skipping rule.");
				next;
			}

			foreach my $component ('sourceIpAddr', 'destinationIpAddr', 'sourceService', 'destinationService')
			{
				my @final_terms;
				if (defined($rule->{$component}) && scalar(@{$rule->{$component}}))
				{
					foreach my $term (@{$rule->{$component}})
					{
						foreach my $key (keys(%{$term}))
						{
							if ($key eq 'objectGroupReference')
							{
								my $object_group_reference = $term->{$key};

								my $object_group_reference_terms = _parse_object_group_reference($self, $component, $object_group_reference);
								if (defined($object_group_reference_terms))
								{
									push(@final_terms, @{$object_group_reference_terms});
								}
								else
								{
									$LOGGER->debug("Failed to parse objectGroupReference '$object_group_reference', skipping rule.");
									next RULE_LOOP;
								}
							}
							else
							{
								push(@final_terms, {$key => $term->{$key}});
							}
						}
					}

					$rule->{$component} = \@final_terms;
				}
			}

			if (defined($rule->{callback_action}))
			{
				$rule->{primaryAction} = $rule->{callback_action};
				delete($rule->{callback_action});
			}
			else
			{
				$rule->{primaryAction} = $primary_action;
			}

			$rule->{'log'} = $log;
			$rule->{processOrder} = $process_order++;

			push(@final_rules, $rule);
		}
	}

	return \@final_rules;
}

=pod

B<get_preprocessor_rules()>

Returns the parsed SNORT preprocessor rules.

=cut
sub get_preprocessor_rules
{
	my ($self) = @_;

	if ($self->{_preprocessors})
	{
		my @preprocessors = keys(%{$self->{_preprocessors}});
		foreach my $preprocessor (@preprocessors)
		{
			my $sub = "_parse_$preprocessor";
			my $sub_ref = \&$sub;

			if (defined(&$sub_ref))
			{
				&$sub_ref($self);
			}
			else
			{
				$LOGGER->debug("Function $sub does not exist.");
			}
		}
	}

	my @preprocessor_rules;
	my @groups = keys(%{$self->{_preprocessor_rule_groups}});
	my @ordered_groups = split(/\s+/, $self->{_preprocessor_evaluation_order});
	foreach my $group (@groups)
	{
		my $already_ordered;
		foreach my $ordered_group (@ordered_groups)
		{
			if ($ordered_group eq $group)
			{
				$already_ordered = 1;
				last;
			}
		}
		if (!$already_ordered)
		{
			push(@ordered_groups, $group);
		}
	}

	foreach my $preprocessor_rule_group (@ordered_groups)
	{
		if ($self->{_preprocessor_rule_groups}->{$preprocessor_rule_group})
		{
			push(@preprocessor_rules, @{$self->{_preprocessor_rule_groups}->{$preprocessor_rule_group}});
		}
	}

	return \@preprocessor_rules;
}

=pod

B<get_config_name()>

Returns the SNORT config name.  The config type will be appended to the front of the name.

=cut
sub get_config_name
{
	my ($self) = @_;

	return $self->{_config_name};
}

=pod

B<parse_config($config)>

Parses the SNORT configuration.

=cut
sub parse_config
{
	my ($self, $config_name) = @_;

	# If we already parsed this config, simply return
	if (defined($self->{_parsed_configs}->{$config_name}))
	{
		return;
	}
	$self->{_parsed_configs}->{$config_name} = 1;

	my $config = $self->{_get_config_callback}->($config_name);

	if (!defined($config))
	{
		$LOGGER->debug("Unable to get included config '$config_name', skipping config.");
		return;
	}

	$config =~ s/\s*\\\s*?\n\s*(?=\S)/ /g;

	_parse_vars($self, $config);

	$config = _replace_vars($self, $config);

	_parse_ruletype($self, $config);

	my $include_files = _parse_include_files($self, $config);

	_parse_preprocessors($self, $config);

	_parse_config_options($self, $config);

	foreach my $include_file (@{$include_files})
	{
		if ($include_file =~ /\/([^\/]+)\.rules$/)
		{
			my $rule_name = $1;
			my $rule_file = $self->{_get_config_callback}->($include_file);
			parse_rules($self, $rule_name, $rule_file);
		}
		else
		{
			parse_config($self, $include_file);
		}
	}
}

=pod

B<get_network_groups()>

Returns a list of SNORT network groups.

=cut
sub get_network_groups
{
	my ($self) = @_;

	# Since get_rules performs some last minute parsing of 'var' object/network groups me must
	# ensure it has been called before get_network_groups.
	if (!defined($self->{_final_rules}))
	{
		$self->{_final_rules} = $self->get_rules();
	}

	my @network_groups;

	foreach my $network_group_name (keys(%{$self->{_network_groups}}))
	{
		my $network_group = {
				id => $network_group_name
			};

		foreach my $term (@{$self->{_network_groups}->{$network_group_name}})
		{
			if ($term->{type} eq 'objectGroupReference')
			{
				my $terms = _parse_object_group_reference($self, 'IpAddr', $term->{value});
				if (defined($terms))
				{
					foreach my $term (@{$terms})
					{
						foreach my $type (keys(%{$term}))
						{
							push(@{$network_group->{$type}}, $term->{$type});
						}
					}
				}
				else
				{
					$LOGGER->debug('Failed to parse objectGroupReference \'' . $term->{value} . "' on networkGroup '$network_group_name'.");
				}
			}
			else
			{
				push(@{$network_group->{$term->{type}}}, $term->{value});
			}
		}

		# Verify there is at least one valid term.
		if (defined($network_group->{objectGroupReference}) || defined($network_group->{subnet}) ||
			defined($network_group->{host}) || defined($network_group->{range}))
		{
			push(@network_groups, $network_group);
		}
		else
		{
			$LOGGER->debug('No valid terms for ' . $self->{_config_type} . " network group: '$network_group_name'");
		}
	}

	return \@network_groups;
}

=pod

B<get_service_groups()>

Returns a list of SNORT service groups.

=cut
sub get_service_groups
{
	my ($self) = @_;

	# Since get_rules performs some last minute parsing of 'var' object/network groups me must
	# ensure it has been called before get_service_groups.
	if (!defined($self->{_final_rules}))
	{
		$self->{_final_rules} = $self->get_rules();
	}

	my @service_groups;

	foreach my $service_group_name (keys(%{$self->{_service_groups}}))
	{
		my $service_group = {
				id => $service_group_name
			};

		foreach my $term (@{$self->{_service_groups}->{$service_group_name}})
		{
			if ($term->{type} eq 'objectGroupReference')
			{
				my $terms = _parse_object_group_reference($self, 'Service', $term->{value});
				if (defined($terms))
				{
					foreach my $term (@{$terms})
					{
						foreach my $type (keys(%{$term}))
						{
							push(@{$service_group->{$type}}, $term->{$type});
						}
					}
				}
				else
				{
					$LOGGER->debug('Failed to parse objectGroupReference \'' . $term->{value} . "' on serviceGroup '$service_group_name'.");
				}
			}
			else
			{
				push(@{$service_group->{$term->{type}}}, $term->{value});
			}
		}

		# Verify there is at least one valid term.
		if (defined($service_group->{objectGroupReference}) || defined($service_group->{portExpression}) ||
			defined($service_group->{portRange}))
		{
			push(@service_groups, $service_group);
		}
		else
		{
			$LOGGER->debug('No valid terms for ' . $self->{_config_type} . " service group: '$service_group_name'");
		}
	}

	return \@service_groups;
}

=pod

B<get_applications()>

Returns a list of SNORT applications.

=cut
sub get_applications
{
	my ($self) = @_;

	return $self->{_applications};
}

=pod

B<parse_rules($rules_name, $rule_config)>

Parses SNORT rules. $rules_name is the name of the SNORT rule or SNORT rule file, $rule_config is
the SNORT rule or snort file contents.

Returns the number of rules found.  This is not the number of rules successfully parsed.

=cut
sub parse_rules
{
	my ($self, $rules_name, $rule_config) = @_;

	# Rules can be broken up over multiple lines using \ at the end of any line.  Convert any lines
	# broken up over multiple lines into a single line.
	# http://manual.snort.org/node28.html
	$rule_config =~ s/\s*\\\s*?\n\s*(?=\S)/ /g;

	my $count = 0;

	while ($rule_config =~ /^(\s*
			(\S+)\s+                   # Rule type
			(\S+)\s+                   # Protocol
			((?:\$\([^\)]+\))|\S+)\s+  # Source ip term
			((?:\$\([^\)]+\))|\S+)\s+  # Source port term
			((?:<-)|(?:->)|(?:<>))\s+  # Rule operator
			((?:\$\([^\)]+\))|\S+)\s+  # Destination ip term
			((?:\$\([^\)]+\))|\S+)\s+  # Destination port term
			\((.*)\)                   # Remainder of the rule, classifierName and classifierValue pairs
		)/mgx)
	{
		my $rule_string = $1;
		my $snort_action = _replace_vars($self, $2);
		my $protocol = _replace_vars($self, $3);
		my $source_ip = $4;
		my $source_service = $5;
		my $rule_operator = $6;
		my $destination_ip = $7;
		my $destination_service = $8;
		my $rule_parameters = $9;

		my $name = $rules_name . ++$count;

		my $sid;

		($name, $sid) = _create_application($self, $name, $rule_parameters);

		my $filter_entry = {};

		my $source_ip_terms = _parse_network_terms($self, $source_ip);
		if (!defined($source_ip_terms))
		{
			$LOGGER->debug('Problem encountered with source IP term in ' . $self->{_config_type} . " rule: '$rule_string', skipping rule.");
			next;
		}
		if (scalar(@{$source_ip_terms}))
		{
			$filter_entry->{sourceIpAddr} = [];
			foreach my $source_ip_term (@{$source_ip_terms})
			{
				my $type = $source_ip_term->{type};
				if ($type eq 'subnet')
				{
					$type = 'network';
				}
				push(@{$filter_entry->{sourceIpAddr}}, {$type => $source_ip_term->{value}});
			}
		}

		my $source_service_terms = _parse_service_terms($self, $source_service);
		if (!defined($source_service_terms))
		{
			$LOGGER->debug('Problem encountered with source service term in ' . $self->{_config_type} . " rule: '$rule_string', skipping rule.");
			next;
		}
		if (scalar(@{$source_service_terms}) && $source_service_terms->[0] ne 'any')
		{
			$filter_entry->{sourceService} = [];
			foreach my $source_service_term (@{$source_service_terms})
			{
				push(@{$filter_entry->{sourceService}}, {$source_service_term->{type} => $source_service_term->{value}});
			}
		}

		my $destination_ip_terms = _parse_network_terms($self, $destination_ip);
		if (!defined($destination_ip_terms))
		{
			$LOGGER->debug('Problem encountered with destination IP term in ' . $self->{_config_type} . " rule: '$rule_string', skipping rule.");
			next;
		}
		if (scalar(@{$destination_ip_terms}))
		{
			$filter_entry->{destinationIpAddr} = [];
			foreach my $destination_ip_term (@{$destination_ip_terms})
			{
				my $type = $destination_ip_term->{type};
				if ($type eq 'subnet')
				{
					$type = 'network';
				}
				push(@{$filter_entry->{destinationIpAddr}}, {$type => $destination_ip_term->{value}});
			}
		}

		my $destination_service_terms = _parse_service_terms($self, $destination_service);
		if (!defined($destination_service_terms))
		{
			$LOGGER->debug('Problem encountered with destination service term in ' . $self->{_config_type} . " rule: '$rule_string', skipping rule.");
			next;
		}
		if (scalar(@{$destination_service_terms}) && $destination_service_terms->[0] ne 'any')
		{
			$filter_entry->{destinationService} = [];
			foreach my $destination_service_term (@{$destination_service_terms})
			{
				push(@{$filter_entry->{destinationService}}, {$destination_service_term->{type} => $destination_service_term->{value}});
			}
		}

		$filter_entry->{protocol} = $protocol;

		# For now set primaryAction to the snort ruletype.  primaryAction and log will later be be
		# corrected by get_rules by incoperating custom ruletypes.
		$filter_entry->{primaryAction} = $snort_action;

		if (defined($self->{_change_action_callback}) && defined($sid))
		{
			my $callback_action = $self->{_change_action_callback}->($sid);
			if (defined($callback_action))
			{
				# The change action callback is used by the GX adapter.  If it returns a value other
				# than undef then the action of the rule has been changed in the tuing config.  Set the
				# callback action to the action returned by the callback.  This will be used by the get_rules
				# method when setting the actual primaryAction of the rule.
				$filter_entry->{callback_action} = $callback_action;
			}
		}

		$filter_entry->{application} = {applicationID => $name};

		$filter_entry->{name} = $name;

		# The Open Signatures manual claims support of a reverse direction operator.
		if ($rule_operator eq '<-')
		{
			# Reverse the direction of the rule.
			my $source_ip_addr = $filter_entry->{sourceIpAddr};
			my $source_service = $filter_entry->{sourceService};
			my $destination_ip_addr = $filter_entry->{destinationIpAddr};
			my $destination_service = $filter_entry->{destinationService};

			delete($filter_entry->{sourceIpAddr});
			delete($filter_entry->{sourceService});
			delete($filter_entry->{destinationIpAddr});
			delete($filter_entry->{destinationService});

			if (defined($destination_ip_addr))
			{
				$filter_entry->{sourceIpAddr} = $destination_ip_addr;
			}
			if (defined($destination_service))
			{
				$filter_entry->{sourceService} = $destination_service;
			}
			if (defined($source_ip_addr))
			{
				$filter_entry->{destinationIpAddr} = $source_ip_addr;
			}
			if (defined($source_service))
			{
				$filter_entry->{destinationService} = $source_service;
			}
		}

		push(@{$self->{_rule_groups}->{$snort_action}}, $filter_entry);

		# <> is the bi-directional operator, create a second rule in the opposite direction.
		# http://manual.snort.org/node29.html#SECTION00425000000000000000
		if ($rule_operator eq '<>')
		{
			my $reverse_filter_entry = {
					primaryAction => $filter_entry->{primaryAction},
					protocol => $protocol,
					application => $filter_entry->{application},
					name => 'Reverse ' . $filter_entry->{name}
				};

			if (defined($filter_entry->{callback_action}))
			{
				$reverse_filter_entry->{callback_action} = $filter_entry->{callback_action};
			}

			if (defined($filter_entry->{sourceIpAddr}))
			{
				$reverse_filter_entry->{destinationIpAddr} = $filter_entry->{sourceIpAddr};
			}
			if (defined($filter_entry->{sourceService}))
			{
				$reverse_filter_entry->{destinationService} = $filter_entry->{sourceService};
			}
			if (defined($filter_entry->{destinationIpAddr}))
			{
				$reverse_filter_entry->{sourceIpAddr} = $filter_entry->{destinationIpAddr};
			}
			if (defined($filter_entry->{destinationService}))
			{
				$reverse_filter_entry->{sourceService} = $filter_entry->{destinationService};
			}

			push(@{$self->{_rule_groups}->{$snort_action}}, $reverse_filter_entry);
		}
	}

	return $count;
}

=pod

B<alert_before_pass()>

It's possible to provide a command line option that forces alert rules to be processed before pass rules:
http://manual.snort.org/node6.html#SECTION00244000000000000000

If this option has been set on the command line, call this method.

=cut
sub alert_before_pass
{
	my ($self) = @_;

	# _evaluation_order is a space seperated list of values.

	# First remove alert from the list.
	$self->{_evaluation_order} =~ s/ alert / /;

	# Now add alert back into the list, before pass.
	$self->{_evaluation_order} =~ s/ pass / alert pass /;
}

=pod

B<get_bound_snorts()>

Returns any bound snort configurations.

Described here: http://manual.snort.org/node25.html

=cut
sub get_bound_snorts
{
	my ($self) = @_;

        return $self->{_bound_snorts};
}

=pod

B<application_dedup($snort_objects)>

This is a static subroutine.

$snort_objects is a reference to an array of ZipTie::Snort objects. Solves
the problem of having multiple applications of the same name.

=cut
sub application_dedup
{
	my ($snort_objects) = @_;

	my %applications;

	foreach my $snort (@{$snort_objects})
	{
		my $new_snort_applications_list = [];

		foreach my $application (@{$snort->{_applications}})
		{
			if (!defined($applications{$application->{'ztApp:appName'}}))
			{
				$applications{$application->{'ztApp:appName'}} = 1;
				push(@{$new_snort_applications_list}, $application);
			}
		}

		$snort->{_applications} = $new_snort_applications_list;
	}
}

=pod

B<get_generated_names_to_msg()>

Returns a mapping of generated names to the msg value that was already taken.

=cut
sub get_generated_names_to_msg
{
	my ($self) = @_;

	return $self->{_generated_names_to_msg};
}

# Recursive function, externally called with only $self and $snort_action.  Converts a snort
# ruletype into ziptie primaryAction and log values taking into account any custom ruletypes stored
# in $self->{_user_rule_types}.
sub _parse_snort_action
{
	my ($self, $snort_action, $seen) = @_;

	# This is used to avoid infinite recursion.
	if (!defined($seen))
	{
		$seen = {};
	}
	if (defined($seen->{$snort_action}))
	{
		return (undef, undef);
	}
	$seen->{$snort_action} = 1;

	my $primary_action;
	my $log;

	if (($snort_action eq 'alert') || ($snort_action eq 'log') || ($snort_action eq 'activate') || ($snort_action eq 'dynamic'))
	{
		$primary_action = 'next';
		$log = 'true';
	}
	elsif ($snort_action eq 'pass')
	{
		$primary_action = 'none';
		$log = 'false';
	}
	elsif ($snort_action eq 'drop')
	{
		$primary_action = 'drop';
		$log = 'true';
	}
	elsif ($snort_action eq 'sdrop')
	{
		$primary_action = 'drop';
		$log = 'false';
	}
	elsif ($snort_action eq 'reject')
	{
		$primary_action = 'deny';
		$log = 'true';
	}
	elsif (defined($self->{_user_rule_types}->{$snort_action}))
	{
		return _parse_snort_action($self, $self->{_user_rule_types}->{$snort_action}, $seen);
	}
	else
	{
		return (undef, undef);
	}

	return ($primary_action, $log);
}

sub _is_network_group
{
	my ($self, $group_name) = @_;

	if (defined($self->{_network_groups}->{$group_name}))
	{
		return 1;
	}

	if (defined($self->{_uncategorized_vars}->{$group_name}))
	{
		# If this group is currently being processed, simply return.
		if (defined($self->{_uncategorized_vars}->{$group_name}->{processing}))
		{
			return 1;
		}
		$self->{_uncategorized_vars}->{$group_name}->{processing} = 1;

		my $terms = _parse_network_terms($self, $self->{_uncategorized_vars}->{$group_name}->{value_string});
		if (!defined($terms))
		{
			$self->{_uncategorized_vars}->{$group_name}->{processing} = undef;
			$LOGGER->debug('Problem parsing ' . $self->{_config_type} . " network group: '" . $self->{_uncategorized_vars}->{$group_name}->{value_string} . "'.");
			return 0;
		}
		$self->{_network_groups}->{$group_name} = $terms;

		return 1;
	}

	return 0;
}

sub _is_service_group
{
	my ($self, $group_name) = @_;

	if (defined($self->{_service_groups}->{$group_name}))
	{
		return 1;
	}

	if (defined($self->{_uncategorized_vars}->{$group_name}))
	{
		# If this group is currently being processed, simply return.
		if (defined($self->{_uncategorized_vars}->{$group_name}->{processing}))
		{
			return 1;
		}
		$self->{_uncategorized_vars}->{$group_name}->{processing} = 1;

		my $terms = _parse_service_terms($self, $self->{_uncategorized_vars}->{$group_name}->{value_string});
		if (!defined($terms))
		{
			$self->{_uncategorized_vars}->{$group_name}->{processing} = undef;
			$LOGGER->debug('Problem parsing ' . $self->{_config_type} . " service group: '" . $self->{_uncategorized_vars}->{$group_name}->{value_string} . "'.");
			return 0;
		}
		if ($terms->[0] eq 'any')
		{
			$self->{_service_groups}->{$group_name} = [{type => 'portRange', value => {portStart => 0, portEnd => 65535}}];
		}
		else
		{
			$self->{_service_groups}->{$group_name} = $terms;
		}

		return 1;
	}

	return 0;
}

sub _parse_network_terms
{
	my ($self, $value_string) = @_;

	my $var_values = [];

	my $ipre = get_crep('cipm');

	# Test for a list of values
	if ($value_string =~ /^\[(.*)\]$/)
	{
		my $value_list = $1;

		while ($value_list =~ s/(\[.*\])//g)
		{
			my $list_value = $1;

			my $parsed_values = _parse_network_terms($self, $list_value);

			if (defined($parsed_values))
			{
				 push(@{$var_values}, @{$parsed_values});	
			}
			else
			{
				$LOGGER->debug('Unable to parse ' . $self->{_config_type} . " network term '$list_value' from list '$value_string'.");
			}
		}

		foreach my $single_value (split(/\s*,\s*/, $value_list))
		{
			my $parsed_single_value = _parse_network_terms($self, $single_value);
			if (defined($parsed_single_value))
			{
				push(@{$var_values}, @{$parsed_single_value});
			}
			else
			{
				$LOGGER->debug('Unable to parse ' . $self->{_config_type} . " network term '$single_value' from list '$value_string'.");
			}
		}
	}
	# Test for a CIDR
	elsif ($value_string =~ /^($ipre)\/(\d{1,2})$/)
	{
		my $ip = $1;
		my $mask = $2;
		push(@{$var_values}, {type => 'subnet', value => {address => $ip, mask => $mask}});
	}
	# Test for a single IP
	elsif ($value_string =~ /^($ipre)$/)
	{
		my $ip = $1;
		push(@{$var_values}, {type => 'host', value => $ip});
	}
	# Test for any
	elsif ($value_string =~ /^any$/i)
	{
		push(@{$var_values}, {type => 'subnet', value => {address => '0.0.0.0', mask => '0'}});
	}
	# Test for var reference
	elsif ($value_string =~ /^\$(.+)$/)
	{
		# Remove any ( ) and messages defined by :? from the vairable.
		my $object_group = $1;
		if ($object_group =~ /^\((\S+?)(?::\?.*)?\)$/)
		{
			$object_group = $1;
		}

		push(@{$var_values}, {type => 'objectGroupReference', value => $object_group});
	}
	# Test for a IP range
	elsif ($value_string =~ /^($ipre)-($ipre)$/)
	{
		# The Snort documentation doesn't mention anything about IP ranges being supported, but
		# they are mentioned in the Proventia Network IPS G/GX User Guide, in the Open Signature
		# Syntax section.
		my $start_address = $1;
		my $end_address = $2;
		push(@{$var_values}, {type => 'range', value => {startAddress => $start_address, endAddress => $end_address}});
	}
	elsif ($value_string =~ /!/)
	{
		$LOGGER->debug('Negated IP terms not supported in ' . $self->{_config_type} . " network terms: '$value_string'");
		return undef;
	}
	else
	{
		push(@{$var_values}, {type => 'other', value => $value_string});
	}

	if (!scalar(@{$var_values}))
	{
		$LOGGER->debug('No terms defined for ' . $self->{_config_type} . " network terms '$value_string'.");
		return undef;
	}

	return $var_values;
}

sub _parse_service_terms
{
	my ($self, $value_string) = @_;

	my $var_values = [];

	my $original_value_string = $value_string;

	# Check if the list of ports should be negated.
	my $negated = 0;
	if ($value_string =~ s/^!//)
	{
		$negated = 1;
	}

	# Test for a list of values
	if ($value_string =~ /^\[(.*)\]$/)
	{
		my $value_list = $1;

		while ($value_list =~ s/(\[.*\])//g)
		{
			my $list_value = $1;

			my $parsed_values = _parse_service_terms($self, $list_value);

			if (defined($parsed_values))
			{
				 push(@{$var_values}, @{$parsed_values});
			}
			else
			{
				$LOGGER->debug('Unable to parse ' . $self->{_config_type} . " service term '$list_value' from list '$value_string'.");
			}
		}

		foreach my $single_value (split(/\s*,\s*/, $value_list))
		{
			my $parsed_single_value = _parse_service_terms($self, $single_value);
			if (defined($parsed_single_value))
			{
				# If one of the terms is any, that's all we care about.
				if ($parsed_single_value->[0] eq 'any')
				{
					return $parsed_single_value;
				}

				push(@{$var_values}, @{$parsed_single_value});
			}
			else
			{
				$LOGGER->debug('Unable to parse ' . $self->{_config_type} . " service term '$single_value' from list '$value_string'.");
			}
		}
	}
	elsif ($value_string =~ /^\d+$/)
	{
		push(@{$var_values}, {type => 'portExpression', value => {operator => 'eq', port => $value_string}});
	}
	elsif ($value_string =~ /^any$/i)
	{
		return ['any'];
	}
	elsif ($value_string =~ /^(\d+)?:(\d+)?$/)
	{
		my $port_start = $1;
		my $port_end = $2;

		if (!defined($port_start))
		{
			$port_start = 0;
		}

		if (!defined($port_end))
		{
			$port_end = 65535;
		}

		push(@{$var_values}, {type => 'portRange', value => {portStart => $port_start, portEnd => $port_end}});
	}
	elsif ($value_string =~ /^\$(.+)$/)
	{
		# Remove any ( ) and messages defined by :? from the vairable.
		my $object_group = $1;
		if ($object_group =~ /^\((\S+?)(?::\?.*)?\)$/)
		{
			$object_group = $1;
		}

		push(@{$var_values}, {type => 'objectGroupReference', value => $object_group});
	}

	# Negate the list of ports
	if ($negated)
	{
		if ((scalar(@{$var_values}) == 1) && ($var_values->[0]->{type} eq 'portExpression'))
		{
			$var_values->[0]->{value}->{operator} = 'ne';
		}
		else
		{
			$LOGGER->debug('Negated port ranges in ' . $self->{_config_type} . " service term not supported: '$original_value_string'");
			return undef;
		}
	}

	if (!scalar(@{$var_values}))
	{
		$LOGGER->debug('No terms defined for ' . $self->{_config_type} . " service terms '$value_string'.");
		return undef;
	}

	return $var_values;
}

sub _parse_vars
{
	my ($self, $config) = @_;

	while ($config =~ /^\s*(?:(ip)|(port))?var\s+(\S+)\s+(.*?)\s*$/gm)
	{
		my $var_name = $self->{_config_name} . "_" . _replace_vars($self, $3);

		my $value_string;
		if (!defined($1) && !defined($2))
		{
			$value_string = _replace_vars($self, $4);
		}
		else
		{
			$value_string = $4;
		}

		# If the vairable alreast exists, relpace it.
		delete($self->{_network_groups}->{$var_name});
		delete($self->{_service_groups}->{$var_name});

		$self->{_uncategorized_vars}->{$var_name} = {value_string => $value_string};

		if (defined($1))
		{
			_is_network_group($self, $var_name);
		}
		elsif (defined($2))
		{
			_is_service_group($self, $var_name);
		}
	}
}

# Parses an object group reference
sub _parse_object_group_reference
{
	my ($self, $component, $object_group_reference) = @_;

	if ($component =~ /IpAddr/)
	{
		if (_is_network_group($self, $self->{_config_name} . "_$object_group_reference"))
		{
			return [{objectGroupReference => $self->{_config_name} . "_$object_group_reference"}];
		}
		elsif ($object_group_reference =~ /^(\S+?):-(.*)$/)
		{
			$object_group_reference = $1;
			my $default_term = $2;

			if (_is_network_group($self, $self->{_config_name} . "_$object_group_reference"))
			{
				return [{objectGroupReference => $self->{_config_name} . "_$object_group_reference"}];
			}
			else
			{
				my $terms = _parse_network_terms($self, $default_term);

				if (defined($terms))
				{
					my $parsed_terms = [];
					foreach my $term (@{$terms})
					{
						push(@{$parsed_terms}, {$term->{type} => $term->{value}});
					}
					return $parsed_terms;
				}
				else
				{
					$LOGGER->debug("Failed to parse default network group term: '$default_term'.");
					return undef;
				}
			}
		}
		else
		{
			$LOGGER->debug('Unable to find ' . $self->{_config_type} . " network group '$object_group_reference'");
			return undef;
		}
	}
	else
	{
		if (_is_service_group($self, $self->{_config_name} . "_$object_group_reference"))
		{
			return [{objectGroupReference => $self->{_config_name} . "_$object_group_reference"}];
		}
		elsif ($object_group_reference =~ /^(\S+?):-(.*)$/)
		{
			$object_group_reference = $1;
			my $default_term = $2;

			if (_is_service_group($self, $self->{_config_name} . "_$object_group_reference"))
			{
				return [{objectGroupReference => $self->{_config_name} . "_$object_group_reference"}];
			}
			else
			{
				my $terms = _parse_service_terms($self, $default_term);

				if (defined($terms))
				{
					my $parsed_terms = [];
					foreach my $term (@{$terms})
					{
						push(@{$parsed_terms}, {$term->{type} => $term->{value}});
					}
					return $parsed_terms;
				}
				else
				{
					$LOGGER->debug("Failed to parse default service group term: '$default_term'.");
					return undef;
				}
			}
		}
		else
		{
			$LOGGER->debug('Unable to find ' . $self->{_config_type} . " service group '$object_group_reference'");
			return undef;
		}
	}

	return undef;
}

# Parse custom ruletypes
# http://manual.snort.org/node29.html#SECTION00421000000000000000
sub _parse_ruletype
{
	my ($self, $config) = @_;

	while ($config =~ /^\s*ruletype\s+(\S+)[^{]*{[^}]*type\s+(\S+)/msg)
	{
		my $user_type_name = $1;
		my $type = $2;

		$self->{_user_rule_types}->{$user_type_name} = $type;
	}
}

sub _parse_include_files
{
	my ($self, $config) = @_;

	my $include_files = [];

	while ($config =~ /^\s*include\s+(\S+)\s*$/gm)
	{
		my $file = $1;

		$file =~ s/\/\//\//g;
		$file =~ s/\\\\/\\/g;

		push(@{$include_files}, $file);
	}

	return $include_files;
}

sub _parse_preprocessors
{
	my ($self, $config) = @_;

	while ($config =~ /^\s*preprocessor\s+(\S+)\s*\:\s*(.*?[^\\\s])\s*$/msg)
	{
		my $type = $1;
		my $options = $2;
		
		foreach my $option (split(/,/, $options))
		{
			if ($option =~ /^\s*(\S+)\s+(\S+)\s*$/)
			{
				my $option_name = $1;
				my $arg = $2;

				# blacklist/whitelist preprocessor
				# http://manual.snort.org/node176.html
				if ($type eq 'reputation')
				{
					$self->{_preprocessors}->{$type}->{$option_name} = $arg;
				}
			}
		}
	}
}

sub _parse_reputation
{
	my ($self) = @_;
	my $reputation = $self->{_preprocessors}->{reputation};
	my $ipre = get_crep('cipm');

	foreach my $type ('whitelist', 'blacklist')
	{
		if ($reputation->{$type})
		{
			my $ip_file = $self->{_get_config_callback}->($reputation->{$type});
			push(@{$reputation->{$type . '_contents'}}, $ip_file);
		}
	}

	if ($reputation->{shared_mem})
	{
		my $ip_files = $self->{_get_config_callback}->($reputation->{shared_mem} . '/*.wlf');
		push(@{$reputation->{'whitelist_contents'}}, $ip_files);

		$ip_files = $self->{_get_config_callback}->($reputation->{shared_mem} . '/*.blf');
		push(@{$reputation->{'blacklist_contents'}}, $ip_files);
	}

	my $white = 'unblack';
	if ($reputation->{white})
	{
		$white = $reputation->{white};
	}
	
	if ($white eq 'unblack')
	{
		$self->{_preprocessor_evaluation_order} = 'whitelist blacklist';
	}
	else
	{
		if ($reputation->{priority})
		{
			my $priority = $reputation->{priority};
			if ($priority eq 'blacklist')
			{
				$self->{_preprocessor_evaluation_order} = 'blacklist whitelist';
			}
			elsif ($priority eq 'whitelist')
			{
				$self->{_preprocessor_evaluation_order} = 'whitelist blacklist';
			}
		}
	}

	my $process_order = 0;
	foreach my $list ('whitelist', 'blacklist')
	{
		my $file_ref = $list . '_contents';
		if ($reputation->{$file_ref})
		{
			foreach my $file (@{$reputation->{$file_ref}})
			{
				while ($file =~ /^\s*($ipre)(\/(\d+))?.*$/mg)
				{
					my $ip = $1;
					my $mask = $3 // 32;
					my $ip_addr;

					if ($mask == 32)
					{
						$ip_addr = {
							host => $ip
						};
					}
					else
					{
						$ip_addr = {
							network => {
								address => $ip,
								mask => $mask
							}
						};
					}

					my $action = 'drop';
					if ($list eq 'whitelist')
					{
						if ($white eq 'unblack')
						{
							$action = 'next';
						}
						else
						{
							$action = 'permit';
						}
					}

					# TODO: Is this logged?

					my $source_filter = {
						primaryAction => $action,
						'log' => 'false',
						processOrder => ++$process_order
					};
					push(@{$source_filter->{sourceIpAddr}}, $ip_addr);

					my $dest_filter = {
						primaryAction => $action,
						'log' => 'false',
						processOrder => ++$process_order
					};
					push(@{$dest_filter->{destinationIpAddr}}, $ip_addr);

					push(@{$self->{_preprocessor_rule_groups}->{$list}}, ($source_filter, $dest_filter));
				}
			}
		}
	}
}

sub _parse_config_options
{
	my ($self, $config) = @_;

	if ($config =~ /^\s*config\s+order\:\s*((\S+\s*?)+)\s*$/m)
	{
		$self->{_evaluation_order} = $1;

		# Remove potential extra whitespace
		$self->{_evaluation_order} =~ s/\s{2,}/ /g;
	}

	if ($config =~ /^\s*config\s+policy_mode\:\s*(\S+)\s*$/m)
	{
		# TODO: Does the policy mode matter?
	}

	# Multiple snort configurations
	# http://manual.snort.org/node25.html
	# TODO: Support vlan id list / ip list
	if ($config =~ /^\s*config\s+binding\:\s*(\S+)\s*(.*?)\s*$/m)
	{
		my $snort_path = $1;
		my $arguments = $2;

		push(@{$self->{_bound_snorts}}, $snort_path);
	}
}

sub _create_application
{
	my ($self, $name, $rule_parameters) = @_;

	my $application = {
			'ztApp:appClassification' => []
		};

	my $sid = undef;

	# Avoid adding identical appClassifications.  Idential appClassifications cause exceptions in ziptie-server when adding the applications.
	my %dedup_check;

	while ($rule_parameters =~ /([^:;]+)(\s*:\s*(?:(?:(?:"[^"]+")|(?:[^;,]+))(?:\s*,\s*)?)+)?\s*;\s*/g)
	{
		my $classifier_name = $1;

		# Classifier name and classifier value is represented as a varchar(128) in the ziptie database.  Truncate to 128 characters if it's too long.
		if (length($classifier_name) > 128)
		{
			$LOGGER->debug("classifierName string longer than 128 characters, truncating: '$classifier_name'");
			$classifier_name = substr($classifier_name, 0, 128);
		}

		if (defined($2))
		{
			my $classifier_values_string = $2;

			while ($classifier_values_string =~ /(?:\s*:\s*)?((?:"[^"]+")|(?:[^;,]+))(?:\s*,\s*)?\s*/g)
			{
				my $classifier_value = $1;
				$classifier_value =~ s/^"(.*?)\s*"$/$1/;

				# Truncate to 128 characters, same as $classifier_name above.
				if (length($classifier_value) > 128)
				{
					$LOGGER->debug("classifierValue string longer than 128 characters, truncating: '$classifier_value'");
					$classifier_value = substr($classifier_value, 0, 128);
				}

				if ($classifier_name eq 'msg')
				{
					if (!defined($self->{_rule_names}->{$classifier_value}))
					{
						$self->{_rule_names}->{$classifier_value} = 1;
						$name = $classifier_value;
					}
					else
					{
						$self->{_generated_names_to_msg}->{$name} = $classifier_value;
					}
				}

				if ($classifier_name =~ /sid/i)
				{
					$sid = $classifier_value;
				}

				my $dedup_hash = "$classifier_name-$classifier_value";
				if (!defined($dedup_check{$dedup_hash}))
				{
					$dedup_check{$dedup_hash} = 1;
					push(@{$application->{'ztApp:appClassification'}}, {'ztApp:classifierName' => $classifier_name, 'ztApp:classifierValue' => $classifier_value});
				}
			}
		}
		else
		{
			my $dedup_hash = "$classifier_name-";
			if (!defined($dedup_check{$dedup_hash}))
			{
				$dedup_check{$dedup_hash} = 1;
				push(@{$application->{'ztApp:appClassification'}}, {'ztApp:classifierName' => $classifier_name, 'ztApp:classifierValue' => ''});
			}
		}
	}

	$application->{'ztApp:appName'} = $name;

	push(@{$self->{_applications}}, $application);

	return ($name, $sid);
}

sub _replace_vars
{
	my ($self, $string) = @_;
	my $original_string = $string;

	while ($string =~ /\$(\w+)/g)
	{
		my $short_name = $1;
		my $name = $self->{_config_name} . "_$short_name";

		if ($self->{_uncategorized_vars}->{$name})
		{
			my $value = $self->{_uncategorized_vars}->{$name}->{value_string};
			$string =~ s/\$$short_name/$value/g;
		}
		else{
			$LOGGER->debug("$short_name not found in _uncategorized_vars");
		}
	}

	return($string);
}

1;

__END__

=pod

=head1 AUTHOR(S)

David Payne
david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2014, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
