#!/usr/bin/perl

use strict;
use warnings;

use ZipTie::Logger;
use ZipTie::Client;
use XML::Twig;

my ($adapterId, $ipAddress) = @ARGV;

if (!defined($adapterId))
{
	exit(1);
}

my $LOGGER = ZipTie::Logger::get_logger();
if ($ENV{ADAPTER_LOG_TO_FILE})
{
	$LOGGER->enable_logging_to_file($adapterId, 'applications', 'all');
}

if ($adapterId !~ /^ZipTie::Adapters::[^\s\]]+::[^\s\]]+$/)
{
	$LOGGER->debug("$adapterId is not a valid adapter ID to call applications.");
	exit(1);
}

if (!defined($ipAddress))
{
	$LOGGER->debug('No IP address defined.');
	exit(1);
}

my $module = $adapterId . '::Applications';
eval("use $module;");
if ($@)
{
	$LOGGER->debug("$module doesn't exist.");
	exit(1);
}

# Build the connection path document.
my $ziptieClient = ZipTie::Client->new();
my $connectionPathXml = get_connection_path_xml($ziptieClient, $ipAddress, $adapterId);

$LOGGER->debug("Calling $module" . '::invoke()');
eval($module . '::invoke($connectionPathXml);');


############################################################
# get_credentials
############################################################
sub get_connection_path_xml
{
	my ($ziptieClient, $ip, $adapterId) = @_;

	my $credentialsService = $ziptieClient->credentials();
	my $devicesService = $ziptieClient->devices();

	# Get the management server IP
	my $managementServerIp = undef;
	my $device = $devicesService->getDevice(ipAddress => $ip, managedNetwork => undef);
	if (defined($device->{managementServerIp}))
	{
		$managementServerIp = $device->{managementServerIp};
	}
	my $deviceId = $device->{deviceId};

	# Get the correct credentials
	my @credentials = $credentialsService->getCredentialSetsByIpAddress(ipAddress => $ip);
	if (!scalar(@credentials))
	{
		$LOGGER->debug("No credentials found for ip $ip.");
		exit(1);
	}

	if (ref($credentials[0]->{credentials}) ne 'ARRAY')
	{
		$credentials[0]->{credentials} = [$credentials[0]->{credentials}];
	}
	my %connectionPathCredentials;
	foreach my $credential (@{$credentials[0]->{credentials}})
	{
		$connectionPathCredentials{$credential->{name}} = $credential->{value};
	}

	# Get the correct protocols
	my $protocols = $credentialsService->getAllEnabledProtocols(ipAddress => $ip, deviceId => $deviceId);
	if (!defined($protocols) || !defined($protocols->{protocols}))
	{
		$LOGGER->debug("No protocols found for ip $ip.");
		exit(1);
	}
	if (ref($protocols->{protocols}) ne 'ARRAY')
	{
		$protocols->{protocols} = [$protocols->{protocols}];
	}
	my @connectionPathProtocols;
	foreach my $protocol (@{$protocols->{protocols}})
	{
		if ($protocol->{enabled} eq 'true' && $protocol->{validatedOnDevice} eq 'true')
		{
			my $protocolParameters = {
					name => $protocol->{name},
					port => $protocol->{port},
					validatedOnDevice => $protocol->{validatedOnDevice},
					priority => $protocol->{priority}
				};

			if (defined($protocol->{properties}))
			{
				my %properties;
				if (ref($protocol->{properties}) ne 'ARRAY')
				{
					$protocol->{properties} = [$protocol->{properties}];
				}
				foreach my $property (@{$protocol->{properties}})
				{
					$properties{$property->{key}} = $property->{value};
				}
				$protocolParameters->{properties} = {property => \%properties};
			}

			push(@connectionPathProtocols, $protocolParameters);
		}
	}

	@connectionPathProtocols = sort { $a->{priority} <=> $b->{priority} } (@connectionPathProtocols);

	# We had to generate our own connection path XML so the protocols would be printed in the correct order.
	my $managementServerIpText = defined($managementServerIp) ? "managementServerIp=\"$managementServerIp\" " : '';
	my $connectionPathXml = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n";
	$connectionPathXml .= "<connectionPath $managementServerIpText" . "host=\"$ip\">\n";
	$connectionPathXml .= "<credentials>\n";
	foreach my $name (keys(%connectionPathCredentials))
	{
		$connectionPathXml .= "<credential name=\"$name\" value=\"" . $connectionPathCredentials{$name} . "\"/>\n";
	}
	$connectionPathXml .= "</credentials>\n";
	$connectionPathXml .= "<protocols>\n";
	foreach my $protocol (@connectionPathProtocols)
	{
		$connectionPathXml .= '<protocol name="' . $protocol->{name} . '" port="' . $protocol->{port} .
			'" validatedOnDevice="' . $protocol->{validatedOnDevice} . "\">\n";
		$connectionPathXml .= "<properties/>\n";
		$connectionPathXml .= "</protocol>\n";
	}
	$connectionPathXml .= "</protocols>\n";
        $connectionPathXml .= "<fileServers/>\n";
	$connectionPathXml .= "</connectionPath>\n";

	return $connectionPathXml;
}

__END__

=head1 NAME

applications.pl

=head1 DESCRIPTION

ZipTie plugin used to collect vendor supplied applications by calling Applications.pm specific to each adapter.

=head1 AUTHOR(S)

David Payne
david.payne@ca.ibm.com

=head1 COPYRIGHT

Copyright (c) 2014, IBM Corporation. All Rights Reserved.
This module is free software. It may be used, redistributed
and/or modified under the same terms as Perl itself.

=cut
