package Verification;

use strict;

use File::Find;
use Digest::MD5::File qw(file_md5_hex md5_hex);
use Data::Dumper;
use XML::Simple;

sub new
{
	my $className = shift;
	
	my $self = {
		zed_file  => undef,
		entries   => undef,
		final_sig => undef,
		file_exts => undef,
		found_files => undef,
		sig_seed => undef,
	};
	
	my @file_exts = ('.pl','.pm',);
	$self->{file_exts} = \@file_exts;
	
	bless( $self, $className);
}

my $found_files = ();
my $file_exts;
my $find_path;
sub file_filter
{
	foreach my $file_ext (@{$file_exts})
	{
		if( $File::Find::name =~ /.*$file_ext$/ )
		{
			my $local_name = $File::Find::name;
			$local_name =~ s/$find_path//;
			push @$found_files, $local_name;
		}
	}
}

sub find_files
{
	my $self = shift;
	my $path = shift;
	
	$find_path = $path;
	
	if( not defined $path or not -d $path )
	{
		warn "invalid path in find_files!\n";
		return 0;
	}
	
	$found_files = (); # wipe the old one
	$file_exts = $self->{file_exts};
	
	find \&file_filter, $path;
	
	$self->{found_files} = $found_files;
	
	return 1;
}

sub generate_file_sigs
{
	my $self = shift;
	
	#print "generate_file_sigs\n";
	
	foreach my $file (sort @{$self->{found_files}})
	{
		open CODE, $find_path.$file;
		my @lines = <CODE>;
		close CODE;
		my $content = join "", @lines;
		if( $self->{sig_seed} )
		{
			$content .= $self->{sig_seed};
		}
		my $entry = {
			name=>$file,
			value=>md5_hex($content),
		};
		push @{$self->{entries}}, $entry;
	}
	#printf Dumper($self->{entries})."\n";
}

# Note bene: overlap with Ziptie::Model::ZiptieSignatureAssembly, its ability to be configured from
# subsets of the SED, and ability to be added as a subelement payload to the very element trees 
# over which it was computed.  ZiptieSignatureAssembly is a 64-bit hex entity.
sub export_file_sigs
{
		my $self = shift;
		
		my $xml_text = XMLout($self->{entries}, RootName=>'sigs');
		return $xml_text;
}

sub import_file_sigs
{
	my $self = shift;
	my $xml_text = shift;
	
	#print "import_file_sigs\n";
	
	$self->{entries} = XMLin($xml_text, KeyAttr=>[]);
	#printf Dumper($self->{entries})."\n";
}

sub generate_zed_with_sig
{
	my $self = shift;
	my $file = shift;
	
	if( not defined $file or not -f $file )
	{
		warn "invalid file in generate_file_sig\n";
		return 0;
	}
	
	open ZED, "<".$file or die "unable to open file $file, $!\n";
	my @zed_lines = <ZED>;
	close ZED;
	my $zed_content = join("", @zed_lines);
	$zed_content .= $self->export_file_sigs();
	$zed_content .= "<verification>".md5_hex($zed_content)."</verification>\n";
	return $zed_content;
}

sub verify_zed_with_sig
{
	my $self = shift;
	my $sig_file = shift;
	
	open SIGNED, "<$sig_file" or die "unable to open file $sig_file in verify, $!\n";
	my @sig_lines = <SIGNED>;
	close SIGNED;
	my $sig_content = join "", @sig_lines;
	$sig_content =~ s/<verification>(.*)<\/verification>\n//msg;
	if( md5_hex($sig_content) eq $1 )
	{
		return 1;
	}
	return 0;
}

sub generate_complete_signed_zed
{
	my $self = shift;
	my $path = shift;
	my $zed_file = shift;
	
	if( not defined $path or not defined $zed_file )
	{
		warn "invalid args passed to generate_complete_signed_zed, need path and file. content will be empty\n";
		return "";
	}
	
	$self->find_files($path);
	open SEED_CONTENT, "<$zed_file";
	my @seed_lines = <SEED_CONTENT>;
	close SEED_CONTENT;
	my $seed_content = join "", @seed_lines;
	$self->{sig_seed} = md5_hex($seed_content);
	$self->generate_file_sigs();
	return $self->generate_zed_with_sig($zed_file);
}

sub super_verification
{
	my $self = shift;
	my $path = shift;
	my $sig_file = shift;
	my $supressed = shift;
	
	
	my $result = 1; # assume it all works out
	
	if( not defined $path or not defined $sig_file )
	{
		warn "invalid args passed to super_verification, need path and file. content will be empty\n";
		return "";
	}
	
	# first, get the listing
	$self->find_files($path);
	
	# next pull out the details of the given zed
	open SIGNED, "<$sig_file" or die "unable to open file $sig_file in verify, $!\n";
	my @sig_lines = <SIGNED>;
	close SIGNED;
	my $sig_content = join "", @sig_lines;
	$sig_content =~ /(.*)(<sigs>.*<\/sigs>\n)(.*)/msg;
	my $zed_content = $1;
	
	# now, generate a signature set
	$self->{sig_seed} = md5_hex($zed_content);
	$self->generate_file_sigs();
	my $local_entries = $self->{entries};
	$self->import_file_sigs($2);
	
	# walk the local and look for matches in the remote
	foreach my $local_entry (@{$local_entries})
	{
		my $local_file = $local_entry->{name};
		my $local_sig = $self->find_file_sig($local_file);
		if( not defined $local_sig )
		{
			print "MISSING  $local_file missing from $sig_file\n" if not $supressed;
			$result = 0;
		}
		else
		{
			if( $local_entry->{value} eq $local_sig )
			{
				printf "VERIFIED $local_file signature matches\n" if not $supressed;
			}
			else
			{
				printf "FAILED   $local_file mismatch\n" if not $supressed;
				$result = 0;
			}
		}
	}
	return $result; # if any failed, then return is false
}

sub find_file_sig()
{
	my $self = shift;
	my $filename = shift;
	foreach my $entry (@{$self->{entries}})
	{
		if( $entry->{name} eq $filename )
		{
			return $entry->{value};
		}
	}
	return undef;
}

1;
